Package: qtl2
Version: 0.44
Date: 2026-07-06
Title: Quantitative Trait Locus Mapping in Experimental Crosses
Description: Provides a set of tools to perform quantitative
    trait locus (QTL) analysis in experimental crosses. It is a
    reimplementation of the 'R/qtl' package to better handle
    high-dimensional data and complex cross designs.
    Broman et al. (2019) <doi:10.1534/genetics.118.301595>.
Author: Karl W Broman [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-4914-6671>),
  R Core Team [ctb]
Maintainer: Karl W Broman <broman@wisc.edu>
Authors@R: c(person("Karl W", "Broman", role=c("aut", "cre"),
             email="broman@wisc.edu", comment=c(ORCID = "0000-0002-4914-6671")),
             person("R Core Team", role="ctb"))
Copyright: Code for Brent's method for univariate function optimization
        was taken from R 3.2.2 (Copyright 1995, 1996 Robert Gentleman
        and Ross Ihaka, Copyright 2003-2004 The R Foundation, Copyright
        1998-2014 The R Core Team).
Depends: R (>= 3.1.0)
Imports: Rcpp (>= 1.0.7), yaml (>= 2.1.13), jsonlite (>= 0.9.17),
        data.table (>= 1.10.4-3), parallel, stats, utils, graphics,
        grDevices, RSQLite
Suggests: testthat, devtools, roxygen2, vdiffr, qtl
License: GPL-3
URL: https://kbroman.org/qtl2/, https://github.com/rqtl/qtl2
BugReports: https://github.com/rqtl/qtl2/issues
LazyData: true
Encoding: UTF-8
ByteCompile: true
LinkingTo: Rcpp, RcppEigen
Config/roxygen2/version: 8.0.0
NeedsCompilation: yes
Packaged: 2026-07-06 12:47:21 UTC; kbroman
Repository: CRAN
Date/Publication: 2026-07-06 14:10:07 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-07-06 15:36:27 UTC; unix
Archs: qtl2.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[",calc_genoprob)
S3method("[",cross2)
S3method("[",phasedgeno)
S3method("[",sim_geno)
S3method("[",viterbi)
S3method(c,scan1perm)
S3method(cbind,calc_genoprob)
S3method(cbind,phasedgeno)
S3method(cbind,scan1)
S3method(cbind,scan1perm)
S3method(cbind,sim_geno)
S3method(cbind,viterbi)
S3method(clean,calc_genoprob)
S3method(clean,scan1)
S3method(dim,calc_genoprob)
S3method(dimnames,calc_genoprob)
S3method(max,compare_geno)
S3method(max,scan1)
S3method(plot,calc_genoprob)
S3method(plot,compare_geno)
S3method(plot,scan1)
S3method(plot,scan1coef)
S3method(print,cross2)
S3method(print,summary.compare_geno)
S3method(print,summary.cross2)
S3method(print,summary.scan1perm)
S3method(rbind,calc_genoprob)
S3method(rbind,phasedgeno)
S3method(rbind,scan1)
S3method(rbind,scan1perm)
S3method(rbind,sim_geno)
S3method(rbind,viterbi)
S3method(replace_ids,calc_genoprob)
S3method(replace_ids,cross2)
S3method(replace_ids,data.frame)
S3method(replace_ids,matrix)
S3method(replace_ids,sim_geno)
S3method(replace_ids,viterbi)
S3method(subset,calc_genoprob)
S3method(subset,cross2)
S3method(subset,phasedgeno)
S3method(subset,scan1)
S3method(subset,sim_geno)
S3method(subset,viterbi)
S3method(summary,compare_geno)
S3method(summary,cross2)
S3method(summary,scan1perm)
export(add_threshold)
export(align_scan1_map)
export(batch_cols)
export(batch_vec)
export(bayes_int)
export(calc_entropy)
export(calc_errorlod)
export(calc_geno_freq)
export(calc_genoprob)
export(calc_grid)
export(calc_het)
export(calc_hotspots)
export(calc_kinship)
export(calc_raw_founder_maf)
export(calc_raw_geno_freq)
export(calc_raw_het)
export(calc_raw_maf)
export(calc_sdp)
export(cbind_expand)
export(check_cross2)
export(chisq_colpairs)
export(chr_lengths)
export(chr_names)
export(clean)
export(clean_genoprob)
export(clean_scan1)
export(compare_founder_geno)
export(compare_geno)
export(compare_genoprob)
export(compare_maps)
export(convert2cross2)
export(count_xo)
export(covar_names)
export(create_gene_query_func)
export(create_snpinfo)
export(create_variant_query_func)
export(decomp_kinship)
export(drop_markers)
export(drop_nullmarkers)
export(est_herit)
export(est_map)
export(find_dup_markers)
export(find_ibd_segments)
export(find_index_snp)
export(find_map_gaps)
export(find_marker)
export(find_markerpos)
export(find_peaks)
export(fit1)
export(founders)
export(fread_csv)
export(fread_csv_numer)
export(genoprob_to_alleleprob)
export(genoprob_to_snpprob)
export(get_common_ids)
export(get_x_covar)
export(guess_phase)
export(ind_ids)
export(ind_ids_covar)
export(ind_ids_geno)
export(ind_ids_gnp)
export(ind_ids_pheno)
export(index_snps)
export(insert_pseudomarkers)
export(interp_genoprob)
export(interp_map)
export(invert_sdp)
export(locate_xo)
export(lod_int)
export(map_to_grid)
export(marker_names)
export(mat2strata)
export(max_compare_geno)
export(max_scan1)
export(maxlod)
export(maxmarg)
export(n_chr)
export(n_covar)
export(n_founders)
export(n_ind)
export(n_ind_covar)
export(n_ind_geno)
export(n_ind_gnp)
export(n_ind_pheno)
export(n_mar)
export(n_missing)
export(n_pheno)
export(n_phenocovar)
export(n_typed)
export(pheno_names)
export(phenocovar_names)
export(plot_ci)
export(plot_cistrans)
export(plot_coef)
export(plot_coefCC)
export(plot_colorscale)
export(plot_compare_geno)
export(plot_genes)
export(plot_geno)
export(plot_genoprob)
export(plot_genoprobcomp)
export(plot_lodpeaks)
export(plot_onegeno)
export(plot_peaks)
export(plot_pxg)
export(plot_scan1)
export(plot_scan1_heatmap)
export(plot_sdp)
export(plot_snpasso)
export(predict_snpgeno)
export(probs_to_grid)
export(pull_genoprobint)
export(pull_genoprobpos)
export(pull_markers)
export(qtl2version)
export(read_cross2)
export(read_pheno)
export(recode_snps)
export(reduce_map_gaps)
export(reduce_markers)
export(replace_ids)
export(scale_kinship)
export(scan1)
export(scan1blup)
export(scan1coef)
export(scan1gen)
export(scan1max)
export(scan1perm)
export(scan1snps)
export(sdp2char)
export(sim_geno)
export(smooth_gmap)
export(subset_scan1)
export(summary_compare_geno)
export(summary_scan1perm)
export(top_snps)
export(tot_mar)
export(unsmooth_gmap)
export(viterbi)
export(write_control_file)
export(xpos_scan1)
export(zip_datafiles)
importFrom(RSQLite,SQLite)
importFrom(RSQLite,dbConnect)
importFrom(RSQLite,dbDisconnect)
importFrom(RSQLite,dbGetQuery)
importFrom(Rcpp,sourceCpp)
importFrom(data.table,fread)
importFrom(grDevices,gray)
importFrom(grDevices,heat.colors)
importFrom(grDevices,rainbow)
importFrom(grDevices,rgb)
importFrom(grDevices,terrain.colors)
importFrom(grDevices,topo.colors)
importFrom(graphics,abline)
importFrom(graphics,axis)
importFrom(graphics,box)
importFrom(graphics,hist)
importFrom(graphics,image)
importFrom(graphics,layout)
importFrom(graphics,lines)
importFrom(graphics,par)
importFrom(graphics,points)
importFrom(graphics,rect)
importFrom(graphics,rug)
importFrom(graphics,segments)
importFrom(graphics,strheight)
importFrom(graphics,strwidth)
importFrom(graphics,text)
importFrom(graphics,title)
importFrom(parallel,detectCores)
importFrom(stats,complete.cases)
importFrom(stats,lm)
importFrom(stats,quantile)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(utils,packageVersion)
useDynLib(qtl2, .registration=TRUE)

### [R/qtl2](https://kbroman.org/qtl2/) <a href="https://kbroman.org/qtl2/"><img src="https://kbroman.org/qtl2/assets/pics/qtl2_logo.png" align="right" height="138" alt="R/qtl2 logo"/></a>

[![R-CMD-check](https://github.com/rqtl/qtl2/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rqtl/qtl2/actions/workflows/R-CMD-check.yaml)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/qtl2)](https://cran.r-project.org/package=qtl2)
[![r-universe badge](https://rqtl.r-universe.dev/qtl2/badges/version)](https://rqtl.r-universe.dev/qtl2)
[![zenodo DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.3237772.svg)](https://doi.org/10.5281/zenodo.3237772)

[Karl Broman](https://kbroman.org)

[R/qtl2](https://kbroman.org/qtl2/) (aka qtl2) is a reimplementation of
the QTL analysis software [R/qtl](https://rqtl.org), to better handle
high-dimensional data and complex cross designs.

Also see the related packages,
[qtl2convert](https://github.com/rqtl/qtl2convert) (for converting
data among the R/qtl2, DOQTL, and R/qtl formats),
[qtl2fst](https://github.com/rqtl/qtl2fst) (for storing genotype
probabilities on disk), and [qtl2ggplot](https://github.com/byandell-sysgen/qtl2ggplot)
(for [ggplot2](https://ggplot2.tidyverse.org/)-based data visualizations).

---

### Installation

Install R/qtl2 from [CRAN](https://cran.r-project.org):

```r
install.packages("qtl2")
```

Alternatively, install it from [R
universe](https://rqtl.r-universe.dev):

```r
install.packages("qtl2", repos=c("https://rqtl.r-universe.dev",
                                 "https://cloud.r-project.org"))
```

Or use [remotes](https://remotes.r-lib.org) to install it from its GitHub source:

```r
install.packages("remotes")
remotes::install_github("rqtl/qtl2")
```

---

### Documentation

- [user guide](https://kbroman.org/qtl2/assets/vignettes/user_guide.html)
- [categorized list of functions in R/qtl2](https://kbroman.org/qtl2/pages/rqtl2_functions.html)
- [input file formats](https://kbroman.org/qtl2/assets/vignettes/input_files.html)
  (also see the
  [sample data files](https://kbroman.org/qtl2/pages/sampledata.html)
  and the [qtl2data repository](https://github.com/rqtl/qtl2data))
- [using qtl2fst for on-disk genotype probabilities](https://kbroman.org/qtl2/assets/vignettes/qtl2fst.html)
- [preparing diversity outbred mouse data for R/qtl2](https://kbroman.org/qtl2/pages/prep_do_data.html)
- [resources for mouse genome build GRCm39](https://kbroman.org/qtl2/pages/move_to_build39.html)
- [genotype diagnostics for diversity outbred mice](https://kbroman.org/qtl2/assets/vignettes/do_diagnostics.html)
- [identifying sample mix-ups in diversity outbred mice](https://kbroman.org/qtl2/assets/vignettes/do_mixups.html)
- [differences between R/qtl and R/qtl2](https://kbroman.org/qtl2/assets/vignettes/rqtl_diff.html)
- [developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html)
- [HMM benchmarks](https://kbroman.org/qtl2/assets/vignettes/hmm_benchmarks.html)
- [Tutorial on R/qtl2](https://smcclatchy.github.io/mapping/) by [Susan McClatchy](https://github.com/smcclatchy) and
  [Dan Gatti](https://github.com/dmgatti)

---

### Citation

To cite R/qtl2 in publications, use:

> Broman KW, Gatti DM, Simecek P, Furlotte NA, Prins P, Sen &#346;,
> Yandell BS, Churchill GA (2019)
> R/qtl2: software for mapping quantitative trait loci with
> high-dimensional data and multi-parent populations.
> Genetics 211:495-502
> [doi:10.1534/genetics.118.301595](https://doi.org/10.1534/genetics.118.301595)

---

### License

Licensed under [GPL-3](https://www.r-project.org/Licenses/GPL-3).

### [R/qtl2](https://kbroman.org/qtl2/)

[![R-CMD-check](https://github.com/rqtl/qtl2/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/rqtl/qtl2/actions/workflows/R-CMD-check.yaml)
[![CRAN_Status_Badge](https://www.r-pkg.org/badges/version/qtl2)](https://cran.r-project.org/package=qtl2)
[![zenodo DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.3237772.svg)](https://doi.org/10.5281/zenodo.3237772)

[Karl Broman](https://kbroman.org)

[R/qtl2](https://kbroman.org/qtl2/) (aka qtl2) is a reimplementation of
the QTL analysis software [R/qtl](https://rqtl.org), to better handle
high-dimensional data and complex cross designs.

Also see the related packages,
[qtl2convert](https://github.com/rqtl/qtl2convert) (for converting
data among the R/qtl2, DOQTL, and R/qtl formats),
[qtl2fst](https://github.com/rqtl/qtl2fst) (for storing genotype
probabilities on disk), and [qtl2ggplot](https://github.com/byandell-sysgen/qtl2ggplot)
(for [ggplot2](https://ggplot2.tidyverse.org/)-based data visualizations).

---

### Installation

Install R/qtl2 from [CRAN](https://cran.r-project.org):

    install.packages("qtl2")

---

### Documentation

- [user guide](https://kbroman.org/qtl2/assets/vignettes/user_guide.html)
- [categorized list of functions in R/qtl2](https://kbroman.org/qtl2/pages/rqtl2_functions.html)
- [input file formats](https://kbroman.org/qtl2/assets/vignettes/input_files.html)
  (also see the
  [sample data files](https://kbroman.org/qtl2/pages/sampledata.html)
  and the [qtl2data repository](https://github.com/rqtl/qtl2data))
- [using qtl2fst for on-disk genotype probabilities](https://kbroman.org/qtl2/assets/vignettes/qtl2fst.html)
- [preparing diversity outbred mouse data for R/qtl2](https://kbroman.org/qtl2/pages/prep_do_data.html)
- [resources for mouse genome build GRCm39](https://kbroman.org/qtl2/pages/move_to_build39.html)
- [genotype diagnostics for diversity outbred mice](https://kbroman.org/qtl2/assets/vignettes/do_diagnostics.html)
- [identifying sample mix-ups in diversity outbred mice](https://kbroman.org/qtl2/assets/vignettes/do_mixups.html)
- [differences between R/qtl and R/qtl2](https://kbroman.org/qtl2/assets/vignettes/rqtl_diff.html)
- [developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html)
- [HMM benchmarks](https://kbroman.org/qtl2/assets/vignettes/hmm_benchmarks.html)
- [Tutorial on R/qtl2](https://smcclatchy.github.io/mapping/) by [Susan McClatchy](https://github.com/smcclatchy) and
  [Dan Gatti](https://github.com/dmgatti)

---

### Citation

To cite R/qtl2 in publications, use:

> Broman KW, Gatti DM, Simecek P, Furlotte NA, Prins P, Sen &#346;,
> Yandell BS, Churchill GA (2019)
> R/qtl2: software for mapping quantitative trait loci with
> high-dimensional data and multi-parent populations.
> Genetics 211:495-502
> [doi:10.1534/genetics.118.301595](https://doi.org/10.1534/genetics.118.301595)

---

### License

Licensed under [GPL-3](https://www.r-project.org/Licenses/GPL-3).

## Scripts for creating databases of CC variants and mouse genes

This directory contains five R scripts, for creating
[SQLite](https://www.sqlite.org) databases for use with the
Collaborative Cross (CC) and Diversity Outbred (DO) mouse populations:

- `create_ccvariants.R` creates the database `cc_variants.sqlite` which
  combines data on SNPs, indels, and structural variants in the eight
  CC founders.

- `create_mousegenes.R` creates the database `mouse_genes.sqlite`
  which mouse gene locations from the
  [Mouse Genome Informatics (MGI)](http://www.informatics.jax.org/)
  database. Key fields in the resulting database include

  - `source` (e.g., `"MGI"`)
  - `type`
  - `start` (in basepairs)
  - `stop` (in basepairs)
  - `strand`
  - `Name`

- `create_ccvariants_small.R` creates
  `../extdata/cc_variants_small.sqlite`, a small version of
  `cc_variants.sqlite` for use in tests. This contains the variants within
  two small regions (one on chr 2 and one on chr 3).

- `create_mousegenes_small.R` creates
  `../extdata/mouse_genes_small.sqlite`, a small version of
  `mouse_genes.sqlite` for use in tests. This contains the genes with
  `source=="MGI"` that overlap two small regions (one on chr 2 and one
  on chr 3).

- `create_mousegenes_mgi.R` creates `mouse_genes_mgi.sqlite`, a version
  with just the records with `source=="MGI"`.

Using these scripts to constructing these databases requires the
following R packages:

- [data.table](https://github.com/Rdatatable/data.table/wiki)
- [RSQLite](https://rstats-db.github.io/RSQLite/)
- [VariantAnnotation](https://bioconductor.org/packages/release/bioc/html/VariantAnnotation.html)
- [R/qtl2](https://github.com/rqtl/qtl2)

The larger database files created by these scripts are available for
direct download from [figshare](https://figshare.com):

- [`cc_variants.sqlite` doi:10.6084/m9.figshare.5280229.v3](https://doi.org/10.6084/m9.figshare.5280229.v3)
- [`mouse_genes.sqlite` doi:10.6084/m9.figshare.5280238.v7](https://doi.org/10.6084/m9.figshare.5280238.v7)

A smaller version of the mouse genes database, with just the records
with `source=="MGI"`, is also available at
[figshare](https://figshare.com):

- [`mouse_genes_mgi.sqlite` doi:10.6084/m9.figshare.5286019.v8](https://doi.org/10.6084/m9.figshare.5286019.v8)

## Sample input data files (zipped) + small example variant and gene databases

### Sample input files

- [`grav2.zip`](grav2.zip), data from
  [Moore et al. (2013) Genetics 195:1077-1086](https://doi.org/10.1534/genetics.113.153346)
  (the second replicate of the RILs)

- [`iron.zip`](iron.zip), data from
  [Grant et al. (2006) Hepatology 44:174-185](https://doi.org/10.1002/hep.21233)

See the contents at <https://kbroman.org/qtl2/pages/sampledata.html>.

These can be read into R with `qtl2::read_cross2()`.

### Variant and gene databases

- [`cc_variants_small.sqlite`](cc_variants_small.sqlite) - SQLite
  database of variants in the Collaborative Cross founders in two
  small regions

- [`mouse_genes_small.sqlite`](mouse_genes_small.sqlite) - SQLite
  database of mosue genes in two small regions

The larger versions of these database files are available for
direct download from [figshare](https://figshare.com):

- [`cc_variants.sqlite` doi:10.6084/m9.figshare.5280229.v3](https://doi.org/10.6084/m9.figshare.5280229.v3)
- [`mouse_genes.sqlite` doi:10.6084/m9.figshare.5280238.v7](https://doi.org/10.6084/m9.figshare.5280238.v7)

A smaller version of the mouse genes database, with just the records
with `source=="MGI"`, is also available at
[figshare](https://figshare.com):

- [`mouse_genes_mgi.sqlite` doi:10.6084/m9.figshare.5286019.v8](https://doi.org/10.6084/m9.figshare.5286019.v8)

# create sqlite database with mouse genes, from MGI
#
# source (dated 2020-09-07):
#      http://www.informatics.jax.org/downloads/mgigff3/archive/monthly/
#          MGI.202009.gff3.gz
#
#      http://www.informatics.jax.org/downloads/mgigff/
#          MGI_GFF_Spec.docx (annotations, e.g. column names)

library(RSQLite)

### download files
site <- "http://www.informatics.jax.org/downloads/mgigff3/archive/monthly"
file <- "MGI.202009.gff3.gz"
date_source <- as.character(as.Date("20200907", format="%Y%m%d"))
genome_build <- "GRCm38/mm10"
url <- paste0(site, "/", file)
if(!file.exists(file))
    download.file(url, file)

# load gene table
file <- sub(".gz$", "", file)
tmpfile <- tempfile()
remove_tmpfile <- FALSE
if(!file.exists(file)) { # need to unzip
    system(paste0("gunzip -c ", file, ".gz > ", tmpfile))
    remove_tmpfile <- TRUE
    file <- tmpfile
}

# fields in the file
fields <- c("chr", "source", "type", "start", "stop", "score", "strand",
            "phase", "attributes")
# within attributes field:
attrib <- c("ID", "Name", "Parent", "Dbxref", "gene_id", "mgi_type", "description")

# read data
tab <- read.table(file, sep="\t", header=FALSE, comment.char="#",
                  na.strings=".", stringsAsFactors=FALSE,
                  quote="", fill=FALSE,
                  colClasses=rep(rep(c("character", "numeric"), 3), c(3,3,1,1,1,0)))
colnames(tab) <- fields

# remove temporary file
if(remove_tmpfile) {
    unlink(tmpfile)
}

# remove things not on chr 1-19, "X", "Y", "MT"
tab <- tab[tab$chr %in% c(1:19, "X", "Y", "MT"),]

# remove things where start > stop (all have stop==0 and start==999999999; presumably unknown position)
tab <- tab[tab$start <= tab$stop,]

# split 9th column at ';' then split at '=' and use first bit as key and second bit as value
tab9_spl <- strsplit(tab[,9], ";")
tab9_list <- lapply(tab9_spl, function(a) {
    spl <- strsplit(a, "=", fixed=TRUE)
    setNames(sapply(spl, "[", 2), sapply(spl, "[", 1)) })

# turn into a list with all of the attributes (with NAs as needed)
tab9_df <- lapply(attrib, function(lab) sapply(tab9_list, "[", lab))
names(tab9_df) <- attrib

# add to data as columns
tab <- cbind(tab[,1:8], as.data.frame(tab9_df, stringsAsFactors=FALSE))

# write to database
dbfile <- "mouse_genes.sqlite"
if(file.exists(dbfile)) unlink(dbfile)
db <- dbConnect(SQLite(), dbfile)
dbWriteTable(db, "genes", tab)
dbExecute(db, "CREATE INDEX chr_start_stop ON genes (chr, start, stop)")

# add description table
description <- data.frame(description="mouse gene information",
                          source="Mouse Genome Informatics (MGI), Jackson Lab",
                          url=url,
                          date_created=as.character(Sys.Date()),
                          date_source=date_source,
                          genome_build=genome_build,
                          stringsAsFactors=FALSE)
dbWriteTable(db, "description", description, append=TRUE)

dbDisconnect(db)

# create a SQLite database just the MGI genes

library(RSQLite)
db <- dbConnect(SQLite(), "mouse_genes.sqlite")
tab <- dbGetQuery(db, "SELECT * FROM genes WHERE source='MGI'")
description <- dbGetQuery(db, "SELECT * FROM description")
dbDisconnect(db)

# remove duplicates (they have multiple Dbxref but are otherwise identical)
#    ... pick a random one of each
tab_name <- table(tab$Name)
dup <- names(tab_name)[tab_name > 1]
if(length(dup) > 0) {
    omit <- NULL
    for(i in dup) {
        z <- which(tab$Name==i)
        dbxref <- unique(tab$Dbxref[z])
        keep <- sample(length(z), 1)
        tab$Dbxref[z[keep]] <- paste(dbxref, collapse=",")
        omit <- c(omit, z[-keep])
    }
    tab <- tab[-omit,]
}

# write to new database
dbfile <- "mouse_genes_mgi.sqlite"
if(file.exists(dbfile)) unlink(dbfile)
db <- dbConnect(SQLite(), dbfile)
dbWriteTable(db, "genes", tab)
dbExecute(db, "CREATE INDEX chr_start_stop ON genes (chr, start, stop)")

# add description table
description[1] <- paste(description[1], "(subset to source='MGI' records)")
dbWriteTable(db, "description", description, append=TRUE)

dbDisconnect(db)

# Create sqlite database with variants (SNPs/indels/SVs)
# in the 8 founder strains of the Collaborative Cross (CC)
######################################################################
# sources:
#   SNPs:
#     ftp://ftp-mouse.sanger.ac.uk/current_snps/mgp.v5.merged.snps_all.dbSNP142.vcf.gz
#     ftp://ftp-mouse.sanger.ac.uk/current_snps/mgp.v5.merged.snps_all.dbSNP142.vcf.gz.tbi
#   Indels:
#     ftp://ftp-mouse.sanger.ac.uk/current_indels/mgp.v5.merged.indels.dbSNP142.normed.vcf.gz
#     ftp://ftp-mouse.sanger.ac.uk/current_indels/mgp.v5.merged.indels.dbSNP142.normed.vcf.gz.tbi
#   SVs:
#     ftp://ftp-mouse.sanger.ac.uk/current_svs/28strains.REL-1410-SV.sdp.tab.gz
#     ftp://ftp-mouse.sanger.ac.uk/current_svs/28strains.REL-1410-SV.sdp.tab.gz.tbi
#
# Same files (except SVs), at JAX:
#   SNPs:
#     ftp://ftp.jax.org/SNPtools/variants/mgp.v5.merged.snps_all.dbSNP142.vcf.gz
#     ftp://ftp.jax.org/SNPtools/variants/mgp.v5.merged.snps_all.dbSNP142.vcf.gz.tbi
#   Indels:
#     ftp://ftp.jax.org/SNPtools/variants/mgp.v5.merged.indels.dbSNP142.normed.vcf.gz
#     ftp://ftp.jax.org/SNPtools/variants/mgp.v5.merged.indels.dbSNP142.normed.vcf.gz.tbi
######################################################################

##############################
### download files
##############################
site <- c("ftp://ftp.jax.org",
          "ftp://ftp.jax.org",
          "ftp://ftp-mouse.sanger.ac.uk")
subdir <- c("SNPtools/variants",
            "SNPtools/variants",
            "current_svs")
files <- c("mgp.v5.merged.snps_all.dbSNP142.vcf.gz",
           "mgp.v5.merged.indels.dbSNP142.normed.vcf.gz",
           "28strains.REL-1410-SV.sdp.tab.gz")
date_source <- c("2015-09-20",
                 "2015-09-20",
                 "2014-10-20")
genome_build <- rep("GRCm38/mm10", 3)

for(i in seq_along(files)) {
    file <- files[i]
    url <- paste0(site[i], "/", subdir[i], "/", file)
    tbi_file <- paste0(file, ".tbi")
    tbi_url <- paste0(site[i], "/", subdir[i], "/", tbi_file)

    if(!file.exists(file)) {
        cat(" -Downloading", file, "\n")
        download.file(url, file)
    }
    if(!file.exists(tbi_file)) {
        cat(" -Downloading", tbi_file, "\n")
        download.file(tbi_url, tbi_file)
    }
}

# function for re-formating snp/indel "consequence"
# Consequences: vectors of |-separated valued
#    2nd value is gene ID
#    3rd is transcript ID (ignored)
#    5th is consequence (potentially with multiple separated by &'s)
format_consequence <-
    function(csq_record)
{
        x <- strsplit(csq_record, "|", fixed=TRUE)
        genes <- sapply(x, "[", 2)
        csq <- sapply(x, "[", 5)
        csq <- unlist(lapply(seq_along(csq),
                             function(i) {
            if(genes[i] == "") {
                return(csq[i])
            } else {
                tmp <- unlist(strsplit(csq[i], "&", fixed=TRUE))
                return(paste(genes[i], tmp, sep=":"))
            }}))
        c(paste(unique(genes), collapse=","),
          paste(unique(csq), collapse=","))
}





##############################
### SNPs
##############################
chr <- c(1:19, "X", "Y", "MT")
cc_founders <- c("A/J", "C57BL/6J", "129S1/SvImJ", "NOD/ShiLtJ", "NZO/HlLtJ",
                 "CAST/EiJ", "PWK/PhJ", "WSB/EiJ")
strains <- sub("/", "_", cc_founders[-2])
n_strains <- length(strains)

library(VariantAnnotation)

library(RSQLite)
db_file <- "cc_variants.sqlite"
db <- dbConnect(SQLite(), dbname=db_file)

dbExecute(db, paste0("ATTACH '", db_file, "' AS NEW"))

cat(" -SNPs\n")
tabfile <- TabixFile(files[1], paste0(files[1], ".tbi"))

db_started <- FALSE
for(thechr in chr) {
    for(left in seq(0, 190, by=10)) {
        cat(thechr, left, "\n")

        # 10 Mbp range
        gr <- GRanges(seqnames=thechr, ranges=IRanges(start=left*1e6, end=(left+10)*1e6-1))

        # grab data
        param <- ScanVcfParam(geno = c("GT", "FI"), samples = strains,
                              which = gr)
        snps <- readVcf(tabfile, genome = "mm10", param = param)
        if(nrow(snps)==0) next

        # drop snps with any quality < 1
        fi <- geno(snps)$FI
        snps <- snps[rowSums(!is.na(fi) & fi==1) == n_strains]

        # drop snps that are all 0/0
        g <- geno(snps)$GT
        snps <- snps[rowSums(is.na(g)) == 0 & rowSums(g=="0/0") < n_strains]
        g <- geno(snps)$GT
        if(nrow(snps)==0) next

        # grab genotypes
        g <- geno(snps)$GT
        # add B6 genotypes (reference) and change column names
        g <- cbind(g[,1,drop=FALSE], C57BL_6J="0/0", g[,-1])
        colnames(g) <- cc_founders

        # alleles
        major <- as.character(ref(snps))
        minor <- CharacterList(alt(snps))
        alleles <- matrix("", nrow=nrow(snps), ncol=4)
        alleles[,1] <- major
        for(i in 2:4) {
            alleles[,i] <- sapply(minor, "[", i-1)
        }

        # numbers of each genotype
        rs <- sapply(c("0/0", "1/1", "2/2", "3/3"),
                     function(a) rowSums(g==a))

        # rows should all sum to 8
        stopifnot(all(rowSums(rs)==8))

        # find the most common allele and swap with major
        for(i in ncol(rs):2) {
            # this allele most common?
            wh <- (rs[,i] > rs[,1] & rowSums(rs <= rs[,i]) == ncol(rs))

            if(any(wh)) {
                # swap alleles
                tmp <- alleles[wh,i]
                alleles[wh,i] <- alleles[wh,1]
                alleles[wh,1] <- tmp

                # swap genotypes
                pat <- paste0(i-1, "/", i-1)
                gg <- g[wh,,drop=FALSE]
                gg[gg==pat] <- "x/x"
                gg[gg=="0/0"] <- pat
                gg[gg=="x/x"] <- "0/0"
                g[wh,] <- gg
            }
        }

        # version with A/C/G/T
        glet <- gnum <- matrix(nrow=nrow(g), ncol=ncol(g))
        dimnames(glet) <- dimnames(gnum) <- dimnames(g)
        for(i in 1:4) {
            pat <- paste0(i-1, "/", i-1)
            for(j in 1:ncol(g)) {
                wh <- (g[,j] == pat)
                glet[wh,j] <- alleles[wh,i]
                gnum[wh,j] <- i
            }
        }

        # NAs in alleles when not seen
        for(i in 1:4) {
            wh <- which(rowSums(gnum==i)==0)
            alleles[wh,i] <- NA
        }

        # alleles as a pattern A|C/G/T
        alleles_char <- paste(alleles[,1],
                              apply(alleles[,-1,drop=FALSE], 1,
                                    function(a) paste(a[!is.na(a)], collapse="/")),
                              sep="|")

        # gnum: turn it into consecutive numbers
        #   (if "2" is missing make 3 = 2)
        for(i in 3:2) {
            wh <- is.na(alleles[,i])
            if(any(wh)) {
                tmp <- gnum[wh,]
                tmp[tmp >= i] <- tmp[tmp >= i] - 1
                gnum[wh,] <- tmp
            }
        }

        # convert to 1/3
        gbin <- gnum
        gbin[gbin > 1] <- 3

        # first row = gene; 2nd row = consequence
        csq <- sapply(info(snps)$CSQ, format_consequence)

        # create full table of info
        snps <- data.frame(snp_id=rownames(g),
                           chr=as.vector(seqnames(snps)),
                           pos=start(snps),
                           alleles=alleles_char,
                           sdp=qtl2::calc_sdp(gbin),
                           ensembl_gene=csq[1,],
                           consequence=csq[2,],
                           gnum,
                           type="snp",
                           stringsAsFactors=FALSE)
        # make sure column names are what we want
        colnames(snps)[8:15] <- c(strains[1], "C57BL_6J", strains[-1])

        dbWriteTable(db, "variants", snps, row.names=FALSE, overwrite=!db_started,
                     append=db_started, field.types=NULL)
        db_started <- TRUE

    }
}

##############################
### indels
##############################
cat(" -InDels\n")
chr <- c(1:19, "X", "Y") # drop MT because not present in indel file
tabfile <- TabixFile(files[2], paste0(files[2], ".tbi"))

for(thechr in chr) {
    for(left in seq(0, 190, by=10)) {
        cat(thechr, left, "\n")

        # 10 Mbp range
        gr <- GRanges(seqnames=thechr, ranges=IRanges(start=left*1e6, end=(left+10)*1e6-1))

        # grab data
        param <- ScanVcfParam(geno = c("GT", "FI"), samples = strains,
                              which = gr)
        indels <- readVcf(tabfile, genome = "mm10", param = param)
        if(nrow(indels)==0) next

        # drop indels with any quality < 1
        fi <- geno(indels)$FI
        indels <- indels[rowSums(!is.na(fi) & fi==1) == n_strains]
        if(nrow(indels)==0) next

        # drop indels that are all 0/0
        g <- geno(indels)$GT
        indels <- indels[rowSums(is.na(g)) == 0 & rowSums(g=="0/0") < n_strains]
        g <- geno(indels)$GT
        if(nrow(indels)==0) next

        # add B6 genotypes (reference) and change column names
        g <- cbind(g[,1,drop=FALSE], C57BL_6J="0/0", g[,-1,drop=FALSE])
        colnames(g) <- cc_founders

        # alleles
        major <- as.character(ref(indels))
        minor <- CharacterList(alt(indels))
        max_minor <- max(sapply(minor, length))
        alleles <- matrix("", nrow=nrow(g), ncol=max_minor+1)
        alleles[,1] <- major
        for(i in 2:(max_minor+1))
            alleles[,i] <- sapply(minor, "[", i-1)

        # numbers of each genotype
        pat <- paste0(0:max_minor, "/", 0:max_minor)
        rs <- sapply(pat, function(a) rowSums(g==a))
        if(!is.matrix(rs)) {
            rs <- matrix(rs, nrow=1)
            dimnames(rs) <- list(rownames(g), pat)
        }

        # rows should all sum to 8
        stopifnot(all(rowSums(rs)==8))

        # find the most common allele and swap with major
        for(i in ncol(rs):2) {
            # this allele most common?
            wh <- (rs[,i] > rs[,1] & rowSums(rs <= rs[,i]) == ncol(rs))

            if(any(wh)) {
                # swap alleles
                tmp <- alleles[wh,i]
                alleles[wh,i] <- alleles[wh,1]
                alleles[wh,1] <- tmp

                # swap genotypes
                pat <- colnames(rs)[i]
                gg <- g[wh,,drop=FALSE]
                gg[gg==pat] <- "x/x"
                gg[gg=="0/0"] <- pat
                gg[gg=="x/x"] <- "0/0"
                g[wh,] <- gg
            }
        }

        # version with alleles
        glet <- gnum <- matrix(nrow=nrow(g), ncol=ncol(g))
        dimnames(glet) <- dimnames(gnum) <- dimnames(g)
        for(i in 1:ncol(rs)) {
            pat <- paste0(i-1, "/", i-1)
            for(j in 1:ncol(g)) {
                wh <- (g[,j] == pat)
                glet[wh,j] <- alleles[wh,i]
                gnum[wh,j] <- i
            }
        }

        # NAs in alleles when not seen
        for(i in 1:ncol(alleles)) {
            wh <- which(rowSums(gnum==i)==0)
            alleles[wh,i] <- NA
        }

        # alleles as a pattern A|C/G/T
        alleles_char <- paste(alleles[,1],
                              apply(alleles[,-1,drop=FALSE], 1,
                                    function(a) paste(a[!is.na(a)], collapse="/")),
                              sep="|")

        # gnum: turn it into consecutive numbers
        #   (if "2" is missing make 3 = 2)
        if(ncol(alleles) >= 3) {
            for(i in 3:2) {
                wh <- is.na(alleles[,i])
                if(any(wh)) {
                    tmp <- gnum[wh,]
                    tmp[tmp >= i] <- tmp[tmp >= i] - 1
                    gnum[wh,] <- tmp
                }
            }
        }

        # convert to 1/3
        gbin <- gnum
        gbin[gbin > 1] <- 3

        # first row = gene; 2nd row = consequence
        csq <- sapply(info(indels)$CSQ, format_consequence)

        # create full table of info
        indels <- data.frame(snp_id=rownames(g),
                             chr=thechr,
                             pos=start(indels),
                             alleles=alleles_char,
                             sdp=qtl2::calc_sdp(gbin),
                             ensembl_gene=csq[1,],
                             consequence=csq[2,],
                             gnum,
                             type="indel",
                             stringsAsFactors=FALSE)
        # make sure column names are what we want
        colnames(indels)[8:15] <- c(strains[1], "C57BL_6J", strains[-1])

        dbWriteTable(db, "variants", indels, row.names=FALSE, overwrite=FALSE,
                     append=TRUE, field.types=NULL)
    }
}


##############################
# structural variants (SVs)
##############################
cat(" -Stuctural variants\n")
tmpfile <- tempfile()
system(paste0("gunzip -c ", files[3], " > ", tmpfile))
svs <- data.table::fread(tmpfile, data.table=FALSE)
unlink(tmpfile)

# pull out genotypes
g <- svs[,colnames(svs) %in% strains]
g <- g[,strains]

# add B6 ref genotype
g <- cbind(g[,1,drop=FALSE], C57BL_6J="0", g[,-1,drop=FALSE], stringsAsFactors=FALSE)

# drop monomorphic ones
n_allele <- apply(g, 1, function(a) length(unique(a)))
svs <- svs[n_allele > 1,]
g <- g[n_allele > 1,]
g[g=="0"] <- "-"

# alleles
alleles <- apply(g, 1, function(a) {
    tab <- table(a)
    result <- names(sort(tab, decreasing=TRUE))
    if(length(result) < 8)
        result <- c(result, rep(NA, 8-length(result)))
    result })
alleles <- t(alleles)

# numeric version
gnum <- matrix(nrow=nrow(g), ncol=ncol(g))
dimnames(gnum) <- dimnames(g)
for(i in 1:ncol(alleles)) {
    for(j in 1:ncol(g)) {
        wh <- (!is.na(alleles[,i]) & g[,j] == alleles[,i])
        gnum[wh,j] <- i
    }
}

# alleles as a pattern A|C/G/T
alleles_char <- paste(alleles[,1],
                      apply(alleles[,-1,drop=FALSE], 1,
                            function(a) paste(a[!is.na(a)], collapse="/")),
                      sep="|")

# convert to 1/3
gbin <- gnum
gbin[gbin > 1] <- 3

# create full table of info
svs <- data.frame(snp_id=paste0("SV_", svs[,"#CHROM"], "_", svs[,"START"], "_", svs[,"END"]),
                  chr=svs[,"#CHROM"],
                  pos=round((svs[,"START"] + svs[,"END"])/2), # average of start and end
                  alleles=alleles_char,
                  sdp=qtl2::calc_sdp(gbin),
                  ensembl_gene=NA,
                  consequence=NA,
                  gnum,
                  type="SV",
                  stringsAsFactors=FALSE)
# make sure column names are what we want
colnames(svs)[8:15] <- c(strains[1], "C57BL_6J", strains[-1])

dbWriteTable(db, "variants", svs, row.names=FALSE, overwrite=FALSE,
             append=TRUE, field.types=NULL)


##############################
### Add description table
##############################

description <- data.frame(description=c("SNPs in Collaborative Cross founders",
                                        "Indels in Collaborative Cross founders",
                                        "SVs in Collaborative Cross founders"),
                          source=c("Mouse Genome Informatics (MGI), Jackson Lab",
                                   "Mouse Genome Informatics (MGI), Jackson Lab",
                                   "Sanger"),
                          url=paste0(site, "/", subdir, "/", files),
                          date_created=rep(as.character(Sys.Date()), 3),
                          date_source=date_source,
                          genome_build=genome_build,
                          stringsAsFactors=FALSE)
dbWriteTable(db, "description", description, append=TRUE)


##############################
### add index
##############################
dbExecute(db, "CREATE INDEX chr_pos ON variants(chr, pos)")
dbDisconnect(db)

# create a tiny example SQLite database with just the MGI genes that overlap two small regions:
#   chr 2, 2 Mbp interval centered at 97.5 Mbp
#   chr 3, 2 Mbp interval centered at 15.0 Mbp

library(RSQLite)
db <- dbConnect(SQLite(), "mouse_genes.sqlite")
tab <- dbGetQuery(db, paste("SELECT * FROM genes WHERE",
                            "((chr=='2' AND",
                            "((start >= 96500000 AND start <= 98500000) OR",
                            "(stop >=   96500000 AND stop <=  98500000) OR",
                            "(start <=  98500000 AND stop >=  96500000))) OR",
                            "(chr=='3' AND",
                            "((start >= 14000000 AND start <= 16000000) OR",
                            "(stop >=   14000000 AND stop <=  16000000) OR",
                            "(start <=  16000000 AND stop >=  14000000))))",
                            "AND source=='MGI'"))
description <- dbGetQuery(db, "SELECT * FROM description")
dbDisconnect(db)

# write to new database
dbfile <- "../extdata/mouse_genes_small.sqlite"
if(file.exists(dbfile)) unlink(dbfile)
db <- dbConnect(SQLite(), dbfile)
dbWriteTable(db, "genes", tab)
dbExecute(db, "CREATE INDEX chr_start_stop ON genes (chr, start, stop)")

# add description table
description[1] <- paste(description[1], "(subset to 2 regions)")
dbWriteTable(db, "description", description, append=TRUE)

dbDisconnect(db)

# create a tiny example SQLite database with just the variants that overlap two small regions:
#   chr 2, 1 Mbp interval centered at 97.5 Mbp
#   chr 3, 1 Mbp interval centered at 15.0 Mbp

library(RSQLite)
db <- dbConnect(SQLite(), "cc_variants.sqlite")

# description
description <- dbGetQuery(db, "SELECT * FROM description")
description$description <- paste(description$description, "(subset to 2 regions)")

tab <- dbGetQuery(db, paste("SELECT * FROM variants WHERE",
                            "(chr=='2' AND pos >= 97000000 AND pos <= 98000000) OR",
                            "(chr=='3' AND pos >= 14500000 AND pos <= 15500000)"))
dbDisconnect(db)

# write to new database
dbfile <- "../extdata/cc_variants_small.sqlite"
if(file.exists(dbfile)) unlink(dbfile)
db <- dbConnect(SQLite(), dbfile)
dbWriteTable(db, "variants", tab)
dbGetQuery(db, "CREATE INDEX chr_pos ON variants (chr, pos)")

# add description table
dbWriteTable(db, "description", description, append=TRUE)

dbDisconnect(db)

#' @name CCcolors
#' @aliases CCcolors
#'
#' @title Collaborative Cross colors
#'
#' @description A vector of 8 colors for use with the mouse
#' Collaborative Cross and Diversity Outbreds.
#'
#' @details
#' `CCorigcolors` are the original eight colors for the Collaborative Cross founder strains.
#' `CCaltcolors` are a slightly modified version, but still not color-blind friendly.
#' `CCcolors` are derived from the the Okabe-Ito color blind friendly palette in
#'  Wong (2011) Nature Methods \doi{10.1038/nmeth.1618}.
#'
#' @source <https://web.archive.org/web/20250215070655/https://csbio.unc.edu/CCstatus/index.py?run=AvailableLines.information>
#'
#' @keywords datasets
#'
#' @examples
#' data(CCcolors)
#' data(CCaltcolors)
#' data(CCorigcolors)
#  c(AJ = "#F0E442", B6 = "#555555", `129` = "#E69F00", NOD = "#0072B2",
#    NZO = "#56B4E9", CAST = "#009E73", PWK = "#D55E00", WSB = "#CC79A7")
NULL

#' @rdname CCcolors
#' @name CCorigcolors
#' @aliases CCorigcolors
# c(AJ = "#FFFF00", B6 = "#888888", `129` = "#FF8888", NOD = "#1111FF",
#   NZO = "#00CCFF", CAST = "#00AA00", PWK = "#FF0000", WSB = "#9900EE")
NULL

#' @rdname CCcolors
#' @name CCaltcolors
#' @aliases CCaltcolors
# c(AJ = "#FFDC00", B6 = "#888888", `129` = "#F08080", NOD = "#0064C9",
#   NZO = "#7FDBFF", CAST = "#2ECC40", PWK = "#FF4136", WSB = "#B10DC9")
NULL

# Calculate BLUPs of QTL effects in scan along one chromosome, with residual polygenic effect
scan1blup_pg <-
    function(genoprobs, pheno, kinship, addcovar=NULL, nullcovar=NULL,
             contrasts=NULL, se=FALSE, reml=TRUE,
             tol=1e-12, cores=1, quiet=TRUE)
{
    stopifnot(tol > 0)

    # check that the objects have rownames
    check4names(pheno, addcovar, NULL, NULL, nullcovar)

    # force things to be matrices
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(nullcovar)) {
        if(!is.matrix(nullcovar)) nullcovar <- as.matrix(nullcovar)
        if(!is.numeric(nullcovar)) stop("nullcovar is not numeric")
    }
    if(!is.null(contrasts)) {
        if(!is.matrix(contrasts)) contrasts <- as.matrix(contrasts)
        if(!is.numeric(contrasts)) stop("contrasts is not numeric")
    }

    # make sure pheno is a vector
    if(is.matrix(pheno) || is.data.frame(pheno)) {
        if(ncol(pheno) > 1)
            warning("Considering only the first phenotype.")
        rn <- rownames(pheno)
        pheno <- pheno[,1]
        names(pheno) <- rn
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }

    # genoprobs has more than one chromosome?
    if(length(genoprobs) > 1)
        warning("Using only the first chromosome, ", names(genoprobs)[1])
    genoprobs <- genoprobs[[1]]

    # make sure contrasts is square n_genotypes x n_genotypes
    if(!is.null(contrasts)) {
        ng <- ncol(genoprobs)
        if(ncol(contrasts) != ng || nrow(contrasts) != ng)
            stop("contrasts should be a square matrix, ", ng, " x ", ng)
    }

    # make sure kinship is for a single chromosome and get IDs
    did_decomp <- is_kinship_decomposed(kinship)
    kinship <- check_kinship_onechr(kinship)
    kinshipIDs <- check_kinship(kinship, 1)

    # multiply kinship matrix by 2; rest is using 2*kinship
    # see Almasy & Blangero (1998) https://doi.org/10.1086/301844
    kinship <- double_kinship(kinship)

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, pheno, kinshipIDs, addcovar, nullcovar, complete.cases=TRUE)
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    if(did_decomp) { # if did decomposition already, make sure it was with exactly
        if(length(kinshipIDs) != length(ind2keep) ||
           any(sort(kinshipIDs) != sort(ind2keep)))
            stop("Decomposed kinship matrix was with different individuals")
        else
            ind2keep <- kinshipIDs # force them in same order
    }

    # omit individuals not in common
    genoprobs <- genoprobs[ind2keep,,,drop=FALSE]
    pheno <- pheno[ind2keep]
    if(!is.null(addcovar)) addcovar <- addcovar[ind2keep,,drop=FALSE]
    if(!is.null(nullcovar)) nullcovar <- nullcovar[ind2keep,,drop=FALSE]

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)
    # add the intercept
    addcovar <- cbind(intercept=rep(1, length(ind2keep)),
                      addcovar)
    rownames(addcovar) <- ind2keep

    # eigen decomposition of kinship matrix
    if(!did_decomp)
        kinship <- decomp_kinship(kinship[ind2keep, ind2keep])

    eigenval <- kinship$values
    eigenvec <- kinship$vectors

    # rotate genoprobs, pheno, and addcovar
    gp_dn <- dimnames(genoprobs)
    genoprobs <- matrix_x_3darray(eigenvec, genoprobs)
    addcovar <- eigenvec %*% addcovar
    pheno <- eigenvec %*% pheno

    # estimate hsq (this doesn't take intercept)
    nullresult <- Rcpp_fitLMM(eigenval, pheno, cbind(addcovar, nullcovar),
                              reml=reml, check_boundary=TRUE, tol=tol)
    hsq <- nullresult$hsq

    # eigen-vectors and weights
    weights <- 1/sqrt(hsq*kinship$values + (1-hsq))

    #  multiply by weights
    genoprobs <- weighted_3darray(genoprobs, weights)
    pheno <- pheno * weights
    addcovar <- addcovar * weights
    dimnames(genoprobs) <- gp_dn

    # multiply genoprobs by contrasts
    if(!is.null(contrasts))
        genoprobs <- genoprobs_by_contrasts(genoprobs, contrasts)

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # determine batches
    n_pos <- dim(genoprobs)[[3]]
    batches <- batch_vec(seq_len(n_pos), ceiling(n_pos/n_cores(cores)))

    by_group_func <- function(i)
        scanblup(genoprobs[,,batches[[i]],drop=FALSE], pheno, addcovar, se, reml, tol)

    # scan to get BLUPs and coefficient estimates
    if(n_cores(cores)==1) {
        result <- scanblup(genoprobs, pheno, addcovar, se, reml, tol)
        if(se) SE <- t(result$SE)
        else SE <- NULL
        coef <- t(result$coef)
    } else {
        result <- cluster_lapply(cores, seq_along(batches), by_group_func)

        # check for problems (if clusters run out of memory, they'll return NULL)
        result_is_null <- vapply(result, is.null, TRUE)
        if(any(result_is_null))
            stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

        SE <- coef <- matrix(nrow=n_pos,ncol=nrow(result[[1]]$coef))
        for(i in seq_along(result)) {
            coef[batches[[i]],] <- t(result[[i]]$coef)
            if(se) SE[batches[[i]],] <- t(result[[i]]$SE)
            else SE <- NULL
        }
    }

    # add names
    coefnames <- scan1coef_names(genoprobs, addcovar, NULL)
    dimnames(coef) <- list(dimnames(genoprobs)[[3]], coefnames)
    if(se) dimnames(SE) <- dimnames(coef)

    result <- coef

    # add some attributes with details on analysis
    attr(result, "sample_size") <- length(ind2keep)
    attr(result, "SE") <- SE # include only if not NULL

    class(result) <- c("scan1coef", "scan1", "matrix")
    result
}

# probs_to_grid
#' Subset genotype probability array to pseudomarkers on a grid
#'
#' Subset genotype probability array (from [calc_genoprob()]
#' to a grid of pseudomarkers along each chromosome.
#'
#' @param probs Genotype probabilities as output from
#' [calc_genoprob()] with `stepwidth="fixed"`.
#'
#' @param grid List of logical vectors that indicate which positions
#' are on the grid and should be retained.
#'
#' @return An object of class `"calc_genoprob"`, like the input, subset to just include
#' pseudomarkers along a grid. See [calc_genoprob()].
#'
#' @details This only works if [calc_genoprob()] was run
#' with `stepwidth="fixed"`, so that the genotype
#' probabilities were calculated at a grid of
#' markers/pseudomarkers. When this is the case, we omit all but
#' the probabilities on this grid. Use [calc_grid()] to
#' find the grid positions.
#'
#' @export
#' @keywords utilities
#' @seealso [calc_grid()], [map_to_grid()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map_w_pmar <- insert_pseudomarkers(grav2$gmap, step=1)
#' probs <- calc_genoprob(grav2, map_w_pmar, error_prob=0.002)
#' sapply(probs, dim)
#' grid <- calc_grid(grav2$gmap, step=1)
#' probs_sub <- probs_to_grid(probs, grid)
#' sapply(probs_sub, dim)

probs_to_grid <-
    function(probs, grid)
{
    if(is.null(probs)) stop("probs is NULL")
    if(is.null(grid)) stop("grid is NULL")

    if(is.cross2(probs))
        stop('Input probs is a "cross2" object but should be genotype probabilities, as from calc_genoprob')

    attrs <- attributes(probs)
    chrID <- names(probs)
    result <- vector("list", length(chrID))
    names(result) <- chrID

    npos <- dim(probs)[3,]

    for(i in seq(along=chrID)) {
        # grab grid vector
        if(is.null(grid[[i]]) || all(grid[[i]])) {
            result[[i]] <- probs[[i]]
            next
        }

        # subset probs
        if(length(grid[[i]]) != npos[i])
            stop("length(grid) [", length(grid[[i]]), "] != dim(probs)[3] [",
                 dim(probs)[3], "] for chr ", chrID[i])
        result[[i]] <- probs[[i]][,,grid[[i]],drop=FALSE]
    }

    # Set up attributes. The result object is of class calc_genoprob.
    ignore <- match(c("names","class"), names(attrs))
    for(a in names(attrs)[-ignore])
      attr(result, a) <- attrs[[a]]

    class(result) <- c("calc_genoprob", "list")

    result
}

context("Reduce probabilities to pseudomarker grid")

test_that("probs_to_grid works", {

    # try it out
    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2, map, error_prob=0.002)
    orig_dim <- sapply(probs, dim)
    grid <- calc_grid(grav2$gmap, step=1)
    probs_sub <- probs_to_grid(probs, grid)
    new_dim <- sapply(probs_sub, dim)

    # reduced dimension match what we'd expect?
    npmar <- 1 + floor(sapply(map, function(a) diff(range(a))))
    expected_dim <- orig_dim; expected_dim[3,] <- npmar
    expect_equal(new_dim, expected_dim)

    # test results
    expected <- probs
    for(i in seq(along=probs)) {
        expected[[i]] <- probs[[i]][,,grid[[i]],drop=FALSE]
    }

    expect_equal(probs_sub, expected)

})

# find_dup_markers
#
#' Find markers with identical genotype data
#'
#' Identify sets of markers with identical genotype data.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#'
#' @param chr Optional vector specifying which chromosomes to consider.
#' This may be a logical, numeric, or character string vector.
#'
#' @param exact_only If TRUE, look only for markers that have matching
#' genotypes and the same pattern of missing data; if FALSE, also look for
#' cases where the observed genotypes at one marker match those at
#' another, and where the first marker has missing genotype whenever the
#' genotype for the second marker is missing.
#'
#' @param adjacent_only If TRUE, look only for sets of markers that are
#' adjacent to each other.
#'
#' @return A list of marker names; each component is a set of markers whose
#' genotypes match one other marker, and the name of the component is the
#'  name of the marker that they match.
#'
#' @details
#'  If `exact.only=TRUE`, we look only for groups of markers whose
#'  pattern of missing data and observed genotypes match exactly.  One
#'  marker (chosen at random) is selected as the name of the group (in the
#'  output of the function).
#'
#'  If `exact.only=FALSE`, we look also for markers whose observed genotypes
#'  are contained in the observed genotypes of another marker.  We use a
#'  pair of nested loops, working from the markers with the most observed
#'  genotypes to the markers with the fewest observed genotypes.
#'
#' @export
#' @keywords utilities
#'
#' @seealso [drop_markers()], [drop_nullmarkers()], [reduce_markers()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' dup <- find_dup_markers(grav2)
#' grav2_nodup <- drop_markers(grav2, unlist(dup))

find_dup_markers <-
    function(cross, chr, exact_only=TRUE, adjacent_only=FALSE)
{
    if(!missing(chr)) cross <- subset(cross, chr=chr)

    if(!is.cross2(cross))
        stop('Input cross should be a "cross2" object.')

    # genotypes as one matrix; include founder_geno

    g <- do.call("cbind", cross$geno)
    if(!is.null(cross$founder_geno)) {
        g <- rbind(g, do.call("cbind", cross$founder_geno))
    }

    markers <- colnames(g)
    markerloc <- lapply(n_mar(cross), function(a) 1:a)
    if(length(markerloc) > 1) {
        for(j in 2:length(markerloc))
            markerloc[[j]] <- markerloc[[j]] + max(markerloc[[j-1]]) + 10
    }
    markerloc <- unlist(markerloc)

    if(exact_only) {
        g[is.na(g)] <- 0

        # genotype data patterns
        pat <- apply(g, 2, paste, collapse="")

        # table of unique values
        tab <- table(pat)

        # no duplicates; return
        if(all(tab == 1)) return(NULL)

        wh <- which(tab > 1)
        theloc <- themar <- vector("list", length(wh))
        for(i in seq(along=wh)) {
            themar[[i]] <- names(pat)[pat==names(tab)[wh[i]]]
            theloc[[i]] <- markerloc[pat==names(tab)[wh[i]]]
        }

        if(adjacent_only) {
            extraloc <- list()
            extramar <- list()
            for(i in seq(along=theloc)) {
                d <- diff(theloc[[i]]) # look for adjacency
                if(any(d>1)) { # split into adjacent groups
                    temp <- which(d>1)
                    first <- c(1, temp+1)
                    last <- c(temp, length(theloc[[i]]))
                    for(j in 2:length(first)) {
                        extraloc[[length(extraloc)+1]] <- theloc[[i]][first[j]:last[j]]
                        extramar[[length(extramar)+1]] <- themar[[i]][first[j]:last[j]]
                    }
                    themar[[i]] <- themar[[i]][first[1]:last[1]]
                    theloc[[i]] <- theloc[[i]][first[1]:last[1]]
                }
            }
            themar <- c(themar, extramar)
            theloc <- c(theloc, extraloc)

            nm <- sapply(themar, length)
            if(all(nm==1)) return(NULL) # nothing left
            themar <- themar[nm>1]
            theloc <- theloc[nm>1]
        }

        # order by first locus
        o <- order(sapply(theloc, min))
        themar <- themar[o]

        randompics <- sapply(themar, function(a) sample(length(a), 1))
        for(i in seq(along=themar)) {
            names(themar)[i] <- themar[[i]][randompics[i]]
            themar[[i]] <- themar[[i]][-randompics[i]]
        }

    }
    else {
        themar <- NULL

        ntyp <- n_typed(cross, "marker")
        o <- order(ntyp, decreasing=TRUE)

        g[is.na(g)] <- 0
        result <- .find_dup_markers_notexact(g, o, markerloc, adjacent_only)

        if(all(result==0)) return(NULL)
        u <- unique(result[result != 0])
        themar <- vector("list", length(u))
        names(themar) <- colnames(g)[u]
        for(i in seq(along=themar))
            themar[[i]] <- colnames(g)[result==u[i]]
    }

    themar
}

context("find duplicate markers")

test_that("find_dup_markers matches qtl::findDupMarkers", {

    library(qtl)
    data(hyper)

    hyper2 <- convert2cross2(hyper)

    # exact only, not adjacent only
    set.seed(20231110)
    dup2 <- find_dup_markers(hyper2)

    set.seed(20231110)
    dup <- qtl::findDupMarkers(hyper)

    expect_equal(dup2, dup)

    # exact only, adjacent only
    set.seed(20231110)
    dup2 <- find_dup_markers(hyper2, adjacent_only=TRUE)

    set.seed(20231110)
    dup <- qtl::findDupMarkers(hyper, adjacent.only=TRUE)

    expect_equal(dup2, dup)

    # not exact only, not adjacent only
    set.seed(20231110)
    dup2 <- find_dup_markers(hyper2, exact_only=FALSE)

    set.seed(20231110)
    dup <- qtl::findDupMarkers(hyper, exact.only=FALSE)

    expect_equal(dup2, dup)

    # not exact only, adjacent only
    set.seed(20231110)
    dup2 <- find_dup_markers(hyper2, exact_only=FALSE, adjacent_only=TRUE)

    set.seed(20231110)
    dup <- qtl::findDupMarkers(hyper, exact.only=FALSE, adjacent.only=TRUE)

    expect_equal(dup2, dup)


})

#' Plot SNP associations
#'
#' Plot SNP associations, with possible expansion from distinct snps to all snps.
#'
#' @param scan1output Output of [scan1()] using
#' SNP probabilities derived by
#' [genoprob_to_snpprob()].
#'
#' @param snpinfo Data frame with SNP information with the following
#'     columns (the last three are generally derived from with
#'     [index_snps()]):
#' * `chr` - Character string or factor with chromosome
#' * `pos` - Position (in same units as in the `"map"`
#'     attribute in `genoprobs`.
#' * `sdp` - Strain distribution pattern: an integer, between
#'     1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
#'     binary encoding indicates the founder genotypes
#' * `snp` - Character string with SNP identifier (if
#'     missing, the rownames are used).
#' * `index` - Indices that indicate equivalent
#'     groups of SNPs.
#' * `intervals` - Indexes that indicate which marker
#'     intervals the SNPs reside.
#' * `on_map` - Indicate whether SNP coincides with a marker
#'     in the `genoprobs`
#'
#' @param genes Optional data frame containing gene information for
#' the region, with columns `start` and `stop` in Mbp, `strand`
#' (as `"-"`, `"+"`, or `NA`), and `Name`. If included, a
#' two-panel plot is produced, with SNP associations above and
#' gene locations below.
#'
#' @param lodcolumn LOD score column to plot (a numeric index, or a
#' character string for a column name). Only one value allowed.
#'
#' @param show_all_snps If TRUE, expand to show all SNPs.
#'
#' @param chr Vector of character strings with chromosome IDs to plot.
#'
#' @param add If TRUE, add to current plot (must have same map and
#' chromosomes).
#'
#' @param drop_hilit SNPs with LOD score within this amount of the maximum SNP association will be highlighted.
#'
#' @param col_hilit Color of highlighted points
#'
#' @param col Color of other points
#'
#' @param gap Gap between chromosomes. The default is 1% of the total genome length.
#'
#' @param minlod Minimum LOD to display. (Mostly for GWAS, in which
#'     case using `minlod=1` will greatly increase the plotting speed,
#'     since the vast majority of points would be omittted.
#'
#' @param sdp_panel Include a panel with the strain distribution patterns for the highlighted SNPs
#'
#' @param strain_labels Labels for the strains, if `sdp_panel=TRUE`.
#'
#' @param ... Additional graphics parameters.
#'
#' @return None.
#'
#' @section Hidden graphics parameters:
#' A number of graphics parameters can be passed via `...`. For
#' example, `bgcolor` to control the background color,`altbgcolor`
#' to control the background color on alternate chromosomes,
#' `altcol` to control the point color on alternate chromosomes,
#' `cex` for character expansion for the points (default 0.5),
#' `pch` for the plotting character for the points (default 16),
#' and `ylim` for y-axis limits.
#' If you are including genes and/or SDP panels, you can use
#' `panel_prop` to control the relative heights of the panels,
#' from top to bottom.
#'
#' @examples
#' \dontrun{
#' # load example DO data from web
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#'
#' # subset to chr 2
#' DOex <- DOex[,"2"]
#'
#' # calculate genotype probabilities and convert to allele probabilities
#' pr <- calc_genoprob(DOex, error_prob=0.002)
#' apr <- genoprob_to_alleleprob(pr)
#'
#' # query function for grabbing info about variants in region
#' snp_dbfile <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
#' query_variants <- create_variant_query_func(snp_dbfile)
#'
#' # SNP association scan
#' out_snps <- scan1snps(apr, DOex$pmap, DOex$pheno, query_func=query_variants,
#'                       chr=2, start=97, end=98, keep_all_snps=TRUE)
#'
#' # plot results
#' plot_snpasso(out_snps$lod, out_snps$snpinfo)
#'
#' # can also just type plot()
#' plot(out_snps$lod, out_snps$snpinfo)
#'
#' # plot just subset of distinct SNPs
#' plot(out_snps$lod, out_snps$snpinfo, show_all_snps=FALSE)
#'
#' # highlight the top snps (with LOD within 1.5 of max)
#' plot(out_snps$lod, out_snps$snpinfo, drop_hilit=1.5)
#'
#' # query function for finding genes in region
#' gene_dbfile <- system.file("extdata", "mouse_genes_small.sqlite", package="qtl2")
#' query_genes <- create_gene_query_func(gene_dbfile)
#' genes <- query_genes(2, 97, 98)
#'
#' # plot SNP association results with gene locations
#' plot(out_snps$lod, out_snps$snpinfo, drop_hilit=1.5, genes=genes)
#'
#' # plot SNP asso results with genes plus SDPs of highlighted SNPs
#' plot(out_snps$lod, out_snps$snpinfo, drop_hilit=2, genes=genes, sdp_panel=TRUE)
#' }
#'
#' @seealso [plot_scan1()], [plot_coef()], [plot_coefCC()]
#'
#' @importFrom stats setNames
#' @export
#'
plot_snpasso <-
    function(scan1output, snpinfo, genes=NULL, lodcolumn=1, show_all_snps=TRUE, chr=NULL,
             add=FALSE, drop_hilit=NA, col_hilit="violetred", col="darkslateblue",
             gap=NULL, minlod=0, sdp_panel=FALSE, strain_labels=names(qtl2::CCcolors), ...)
{
    if(is.null(scan1output)) stop("scan1output is NULL")
    if(is.null(snpinfo)) stop("snpinfo is NULL")
    if(!is_nonneg_number(minlod)) stop("minlod should be a single non-negative number")

    # pull out lod scores
    if(length(lodcolumn)==0) stop("lodcolumn has length 0")
    if(length(lodcolumn) > 1) { # If length > 1, take first value
        warning("lodcolumn should have length 1; only first element used.")
        lodcolumn <- lodcolumn[1]
    }
    if(is.character(lodcolumn)) { # turn column name into integer
        tmp <- match(lodcolumn, colnames(scan1output))
        if(is.na(tmp)) stop('lodcolumn "', lodcolumn, '" not found')
        lodcolumn <- tmp
    }
    if(lodcolumn < 1 || lodcolumn > ncol(scan1output))
        stop("lodcolumn [", lodcolumn, "] out of range (should be in 1, ..., ", ncol(scan1output), ")")
    scan1output <- scan1output[,lodcolumn,drop=FALSE]

    if(!is.null(chr)) { # subset by chr
        if(!any(chr %in% snpinfo$chr)) stop("None of the chr found in snpinfo")
        if(!all(chr %in% snpinfo$chr)) {
            warning("Some chr not found: ", paste(chr[!(chr %in% snpinfo$chr)], collapse=", "))
        }
        snpinfo <- snpinfo[snpinfo$chr %in% chr,,drop=FALSE]
        scan1output <- scan1output[rownames(scan1output) %in% snpinfo$snp,,drop=FALSE]
    }

    if(!is.null(genes)) {
        if(length(unique(snpinfo$chr)) > 1) {
            warning("genes ignored if there are multiple chromosomes")
        } else if(sdp_panel) {
            return( plot_snpasso_sdp_genes(scan1output, snpinfo, show_all_snps=show_all_snps,
                                           drop_hilit=drop_hilit, col_hilit=col_hilit,
                                           col=col, gap=gap, minlod=minlod, genes=genes,
                                           strain_labels=strain_labels, ...) )
        } else {
            return( plot_snpasso_and_genes(scan1output, snpinfo, show_all_snps=show_all_snps,
                                           drop_hilit=drop_hilit, col_hilit=col_hilit,
                                           col=col, gap=gap, minlod=minlod, genes=genes, ...) )
        }
    } else if(sdp_panel) {
        if(length(unique(snpinfo$chr)) > 1) {
            warning("sdp_panel ignored if there are multiple chromosomes")
        } else {
            return( plot_snpasso_and_sdp(scan1output, snpinfo, show_all_snps=show_all_snps,
                                         drop_hilit=drop_hilit, col_hilit=col_hilit,
                                         col=col, gap=gap, minlod=minlod,
                                         strain_labels=strain_labels, ...) )
        }
    }

    if(nrow(scan1output) == nrow(snpinfo) && all(rownames(scan1output) == snpinfo$snp)) {
        show_all_snps <- FALSE

        snpinfo <- snpinfo[scan1output[,1]>=minlod, , drop=FALSE]
        scan1output <- scan1output[scan1output[,1]>=minlod, 1, drop=FALSE]

        # skip the index business
        # snpinfo -> map
        map <- tapply(seq_len(nrow(snpinfo)), factor(snpinfo$chr, unique(snpinfo$chr)),
                      function(i) setNames(snpinfo$pos[i], snpinfo$snp[i]))
    }
    else {
        snpinfo_spl <- split(snpinfo, factor(snpinfo$chr, unique(snpinfo$chr)))

        uindex <- unlist(lapply(snpinfo_spl, function(a) unique(a$index)))
        if(length(uindex) != nrow(scan1output)) {
            stop("Something is wrong with snpinfo$index.\n",
                 "      no. unique indexes [",
                 length(uindex), "] != nrow(scan1output) [",
                 nrow(scan1output), "].")
        }

        for(i in seq_along(snpinfo_spl)) {
            uindex <- unique(snpinfo_spl[[i]]$index)
            if(any(snpinfo_spl[[i]]$index[uindex] != uindex)) {
                stop("Something is wrong with snpinfo$index.\n",
                     "      on each chr, index[u] should == u for each index u")
            }
        }

        map <- snpinfo_to_map(snpinfo)

        if(show_all_snps) {
            tmp <- expand_snp_results(scan1output, map, snpinfo)
            scan1output <- tmp$lod
            map <- tmp$map
        }
    }

    # maximum LOD
    maxlod <- max(unclass(scan1output)[,1], na.rm=TRUE)

    # set values < minlod to NA so they're not plotted
    scan1output[scan1output < minlod] <- NA

    # internal function to give defaults to hidden graphics parameters
    plot_snpasso_internal <-
        function(pch=16, cex=0.5, ylim=NULL, bgcolor="gray90",
                 altbgcolor="gray85",
                 drop_hilit=NA, col_hilit="violetred",
                 drop.hilit=NULL, col.hilit=NULL, ...)
    {
        if(!is.null(drop.hilit)) {
            warning("drop.hilit is deprecated; use drop_hilit")
            drop_hilit <- drop.hilit
        }

        if(!is.null(col.hilit)) {
            warning("col.hilit is deprecated; use col_hilit")
            col_hilit <- col.hilit
        }

        if(is.null(ylim))
            ylim <- c(0, maxlod*1.02)

        if(!is.na(drop_hilit) && !is.null(drop_hilit)) {
            col <- c(col, col_hilit)[(scan1output >= maxlod-drop_hilit)+1]
        }

        plot_scan1(scan1output, map, lodcolumn=1, bgcolor=bgcolor, altbgcolor=altbgcolor, ylim=ylim,
                   gap=gap, add=add, col = col, type="p", cex=cex, pch=pch, ...)
    }

    plot_snpasso_internal(drop_hilit=drop_hilit, col_hilit=col_hilit, ...)
}


# expand snp association results according to snpinfo
expand_snp_results <-
    function(snp_results, map, snpinfo)
{
    snpinfo <- split(snpinfo, factor(snpinfo$chr, unique(snpinfo$chr)))

    if(length(map) != length(snpinfo))
        stop("length(map) [", length(map), "] != length(snpinfo) [",
             length(snpinfo), "]")

    if(nrow(snp_results) != length(unlist(map)))
        stop("nrow(snp_results) [", nrow(snp_results), "] != length(unlist(map)) [",
             length(unlist(map)), "]")

    cnames <- rep(names(map), vapply(map, length, 0))
    lodindex <- split(seq_len(nrow(snp_results)), factor(cnames, unique(cnames)))

    result <- NULL
    for(i in seq(along=map)) {
        revindex <- rev_snp_index(snpinfo[[i]])

        map[[i]] <- snpinfo[[i]]$pos
        names(map[[i]]) <- snpinfo[[i]]$snp
        this_result <- unclass(snp_results)[lodindex[[i]],,drop=FALSE][revindex,,drop=FALSE]
        rownames(this_result) <- snpinfo[[i]]$snp

        result <- rbind(result, this_result)
    }

    list(lod=result,
         map=map)
}

# write_control_file
#' Write a control file for QTL data
#'
#' Write the control file (in [YAML](https://yaml.org) or [JSON](https://json.org)) needed
#' by [read_cross2()] for a set of QTL data.
#'
#' @param output_file File name (with path) of the
#' [YAML](https://yaml.org) or [JSON](https://json.org) file to be created, as a character
#' string. If extension is `.json`, JSON format is used; otherwise, YAML is used.
#'
#' @param crosstype Character string with the cross type.
#' @param geno_file File name for genotype data.
#' @param founder_geno_file File name for the founder genotype data.
#' @param gmap_file File name for genetic map.
#' @param pmap_file File name for the physical map.
#' @param pheno_file File name for the phenotype data.
#' @param covar_file File name for the covariate data.
#' @param phenocovar_file File name for the phenotype covariate data
#' (i.e., metadata about the phenotypes).
#' @param sex_file File name for the individuals' sex. (Specify just
#' one of `sex_file` or `sex_covar`.)
#' @param sex_covar Column name in the covariate data that corresponds
#' to sex. (Specify just one of `sex_file` or `sex_covar`.)
#' @param sex_codes Named vector of character strings specifying the
#' encoding of sex. The names attribute should be the codes used in
#' the data files; the values within the vector should be
#' `"female"` and `"male"`.
#' @param crossinfo_file File name for the `cross_info` data. (Specify just
#' one of `crossinfo_file` or `crossinfo_covar`.)
#' @param crossinfo_covar Column name in the covariate data that
#' corresponds to the `cross_info` data. (Specify just one of
#' `crossinfo_file` or `crossinfo_covar`.)
#' @param crossinfo_codes In the case that there is a single cross
#' info column (whether in a file or as a covariate), you can
#' provide a named vector of character strings specifying the
#' encoding of `cross_info`. The names attribute should be the
#' codes used; the values within the vector should be the codes to
#' which they will be converted (for example, `0` and `1` for an
#' intercross).
#' @param geno_codes Named vector specifying the encoding of
#' genotypes. The names attribute has the codes used within the
#' genotype and founder genotype data files; the values within the
#' vector should be the integers to which the genotypes will be
#' converted.
#' @param alleles Vector of single-character codes for the founder
#' alleles.
#' @param xchr Character string with the ID for the X chromosome.
#' @param sep Character string that separates columns in the data files.
#' @param na.strings Vector of character strings with codes to be
#' treated as missing values.
#' @param comment.char Character string that is used as initial
#' character in a set of leading comment lines in the data files.
#' @param geno_transposed If TRUE, genotype file is transposed (with markers as rows).
#' @param founder_geno_transposed If TRUE, founder genotype file is transposed (with markers as rows).
#' @param pheno_transposed If TRUE, phenotype file is transposed (with phenotypes as rows).
#' @param covar_transposed If TRUE, covariate file is transposed (with covariates as rows).
#' @param phenocovar_transposed If TRUE, phenotype covariate file is transposed (with phenotype covariates as rows).
#' @param description Optional character string describing the data.
#' @param comments Vector of character strings to be inserted as
#' comments at the top of the file (in the case of YAML), with each
#' string as a line. For JSON, the comments are instead included
#' within the control object.
#' @param overwrite If TRUE, overwrite file if it exists. If FALSE
#' (the default) and the file exists, stop with an error.
#'
#' @return (Invisibly) The data structure that was written.
#'
#' @details This function takes a set of parameters and creates the
#' control file (in [YAML](https://yaml.org) or [JSON](https://json.org) format) needed
#' for the new input data file format for
#' [R/qtl2](https://kbroman.org/qtl2/).  See the
#' [sample data files](https://kbroman.org/qtl2/pages/sampledata.html) and the
#' [vignette describing the input file format](https://kbroman.org/qtl2/assets/vignettes/input_files.html).
#'
#' @export
#' @keywords utilities
#' @seealso [read_cross2()], sample data files at
#' <https://kbroman.org/qtl2/pages/sampledata.html>
#' @examples
#' # Control file for the sample dataset, grav2
#' grav2_control_file <- file.path(tempdir(), "grav2.yaml")
#' write_control_file(grav2_control_file,
#'                    crosstype="riself",
#'                    geno_file="grav2_geno.csv",
#'                    gmap_file="grav2_gmap.csv",
#'                    pheno_file="grav2_pheno.csv",
#'                    phenocovar_file="grav2_phenocovar.csv",
#'                    geno_codes=c(L=1L, C=2L),
#'                    alleles=c("L", "C"),
#'                    na.strings=c("-", "NA"))
#'
#' # Control file for the sample dataset, iron
#' iron_control_file <- file.path(tempdir(), "iron.yaml")
#' write_control_file(iron_control_file,
#'                    crosstype="f2",
#'                    geno_file="iron_geno.csv",
#'                    gmap_file="iron_gmap.csv",
#'                    pheno_file="iron_pheno.csv",
#'                    covar_file="iron_covar.csv",
#'                    phenocovar_file="iron_phenocovar.csv",
#'                    geno_codes=c(SS=1L, SB=2L, BB=3L),
#'                    sex_covar="sex",
#'                    sex_codes=c(f="female", m="male"),
#'                    crossinfo_covar="cross_direction",
#'                    crossinfo_codes=c("(SxB)x(SxB)"=0L, "(BxS)x(BxS)"=1L),
#'                    xchr="X",
#'                    alleles=c("S", "B"),
#'                    na.strings=c("-", "NA"))
#'
#' # Remove these files, to clean up temporary directory
#' unlink(c(grav2_control_file, iron_control_file))
write_control_file <-
function(output_file, crosstype=NULL, geno_file=NULL, founder_geno_file=NULL, gmap_file=NULL,
         pmap_file=NULL, pheno_file=NULL, covar_file=NULL, phenocovar_file=NULL,
         sex_file=NULL, sex_covar=NULL, sex_codes=NULL,
         crossinfo_file=NULL, crossinfo_covar=NULL, crossinfo_codes=NULL,
         geno_codes=NULL, alleles=NULL, xchr=NULL,
         sep=",", na.strings=c("-", "NA"), comment.char="#",
         geno_transposed=FALSE, founder_geno_transposed=FALSE,
         pheno_transposed=FALSE, covar_transposed=FALSE,
         phenocovar_transposed=FALSE,
         description=NULL, comments=NULL, overwrite=FALSE)
{
    output_file <- path.expand(output_file)
    if(!overwrite && file.exists(output_file))
        stop("The output file (", output_file, ") already exists. Remove it first (or use overwrite=TRUE).")

    result <- list(description="", # stub to be replaced or removed
                   comments="", # stub to be replaced or removed
                   crosstype=crosstype, sep=sep, na.strings=na.strings,
                   comment.char=comment.char)
    if(!is.null(description) && description!="") {
        paste(description, collapse="\n")
        result$description <- description
    } else {
        result$description <- NULL
    }

    if(!is.null(geno_file))
        result$geno <- geno_file
    if(!is.null(founder_geno_file))
        result$founder_geno <- founder_geno_file
    if(!is.null(gmap_file))
        result$gmap <- gmap_file
    if(!is.null(pmap_file))
        result$pmap <- pmap_file
    if(!is.null(pheno_file))
        result$pheno <- pheno_file
    if(!is.null(covar_file))
        result$covar <- covar_file
    if(!is.null(phenocovar_file))
        result$phenocovar <- phenocovar_file
    if(!is.null(alleles))
        result$alleles <- alleles
    if(!is.null(xchr))
        result$x_chr <- xchr
    if(!is.null(geno_codes)) {
        storage.mode(geno_codes) <- "integer"
        result$genotypes <- as.list(geno_codes)
    }

    # transposed file?
    if(!is.null(geno_transposed) && geno_transposed)
        result$geno_transposed <- geno_transposed
    if(!is.null(founder_geno_transposed) && founder_geno_transposed)
        result$founder_geno_transposed <- founder_geno_transposed
    if(!is.null(pheno_transposed) && pheno_transposed)
        result$pheno_transposed <- pheno_transposed
    if(!is.null(covar_transposed) && covar_transposed)
        result$covar_transposed <- covar_transposed
    if(!is.null(phenocovar_transposed) && phenocovar_transposed)
        result$phenocovar_transposed <- phenocovar_transposed

    # sex
    if(!is.null(sex_file)) {
        if(!is.null(sex_covar))
            stop("Specify just one of sex_file and sex_covar")
        if(is.null(sex_codes))
            stop("if sex_file is specified, sex_codes must be as well")

        result$sex <- c(list(file=sex_file),
                        as.list(sex_codes))
    }
    else if(!is.null(sex_covar)) {
        if(is.null(sex_codes))
            stop("if sex_covar is specified, sex_codes must be as well")
        result$sex <- c(list(covar=sex_covar),
                        as.list(sex_codes))
    }

    # cross_info
    if(!is.null(crossinfo_file)) {
        if(!is.null(crossinfo_covar)) {
            stop("Specify just one of crossinfo_file and crossinfo_covar")
        }
        result$cross_info <- list(file=crossinfo_file)
    }
    else if(!is.null(crossinfo_covar)) {
        result$cross_info <- list(covar=crossinfo_covar)
    }
    if(!is.null(crossinfo_codes)) {
        if(is.null(crossinfo_covar) && is.null(crossinfo_file)) {
            stop("if crossinfo_codes is provided, crossinfo_covar or crossinfo_file must also be provided")
        }
        storage.mode(crossinfo_codes) <- "integer"
        result$cross_info <- c(result$cross_info, as.list(crossinfo_codes))
    }

    # JSON or YAML?
    if(grepl("\\.json$", output_file)) { # assume JSON
        if(is.null(comments))
            result$comments <- NULL
        else
            result$comments <- comments

        cat(jsonlite::toJSON(result, auto_unbox=TRUE, pretty=TRUE),
            file=output_file)
        cat("\n", file=output_file, append=TRUE) # add extra newline to avoid warning when reading
    }
    else {
        result$comments <- NULL # delete the placeholder

        # comments as a single string
        if(is.null(comments))
            comments <- ""
        else comments <- paste0("# ", comments, "\n", collapse="")

        # write data
        cat(comments, file=output_file)
        cat(yaml::as.yaml(result), file=output_file, append=TRUE)
    }
    invisible(result)
}

context("I/O of yaml/json control files")

test_that("write_control_file gives correct output for riself", {

    dir <- tempdir()
    file <- file.path(dir, "grav2.yaml")
    on.exit(unlink(file))

    # write file
    write_control_file(file,
                    crosstype="riself",
                    geno_file="grav2_geno.csv",
                    gmap_file="grav2_gmap.csv",
                    pheno_file="grav2_pheno.csv",
                    phenocovar_file="grav2_phenocovar.csv",
                    geno_codes=c(L=1L, C=2L),
                    alleles=c("L", "C"),
                    na.strings=c("-", "NA"))

    # read it back in
    control <-  read_control_file(file)

    expected <- list(crosstype = "riself", geno = "grav2_geno.csv", pheno = "grav2_pheno.csv",
                     phenocovar = "grav2_phenocovar.csv", gmap = "grav2_gmap.csv",
                     alleles = c("L", "C"), genotypes = list(L = 1L, C = 2L),
                     sep = ",", na.strings = c("-", "NA"), comment.char = "#")

    expect_equal(sort(names(control)), sort(names(expected)))
    expected <- expected[names(control)]
    expect_equal(control, expected)

})


test_that("write_control_file in JSON for riself", {

    dir <- tempdir()
    file <- file.path(dir, "grav2.json")
    on.exit(unlink(file))

    # write file
    write_control_file(file,
                    crosstype="riself",
                    geno_file="grav2_geno.csv",
                    gmap_file="grav2_gmap.csv",
                    pheno_file="grav2_pheno.csv",
                    phenocovar_file="grav2_phenocovar.csv",
                    geno_codes=c(L=1L, C=2L),
                    alleles=c("L", "C"),
                    na.strings=c("-", "NA"))

    # read it back in
    control <-  read_control_file(file)

    expected <- list(crosstype = "riself", geno = "grav2_geno.csv", pheno = "grav2_pheno.csv",
                     phenocovar = "grav2_phenocovar.csv", gmap = "grav2_gmap.csv",
                     alleles = c("L", "C"), genotypes = list(L = 1L, C = 2L),
                     sep = ",", na.strings = c("-", "NA"), comment.char = "#")

    expect_equal(sort(names(control)), sort(names(expected)))
    expected <- expected[names(control)]
    expect_equal(control, expected)

})


test_that("write_control_file gives correct output for intercross", {

    dir <- tempdir()
    file <- file.path(dir, "iron.yaml")
    on.exit(unlink(file))

    # write file
    write_control_file(file,
                       crosstype="f2",
                       geno_file="iron_geno.csv",
                       gmap_file="iron_gmap.csv",
                       pheno_file="iron_pheno.csv",
                       covar_file="iron_covar.csv",
                       phenocovar_file="iron_phenocovar.csv",
                       geno_codes=c(SS=1L, SB=2L, BB=3L),
                       sex_covar="sex",
                       sex_codes=c(f="female", m="male"),
                       crossinfo_covar="cross_direction",
                       crossinfo_codes=c("(SxB)x(SxB)"=0L, "(BxS)x(BxS)"=1L),
                       xchr="X",
                       alleles=c("S", "B"),
                       na.strings=c("-", "NA"))

    # read it back in
    control <-  read_control_file(file)

    expected <- list(crosstype = "f2", geno = "iron_geno.csv", pheno = "iron_pheno.csv",
                     phenocovar = "iron_phenocovar.csv", covar = "iron_covar.csv",
                     gmap = "iron_gmap.csv", alleles = c("S", "B"),
                     genotypes = list(SS = 1L, SB = 2L, BB = 3L),
                     sex = list(covar = "sex", f = "female", m = "male"),
                     cross_info = list(covar = "cross_direction", `(SxB)x(SxB)` = 0L, `(BxS)x(BxS)` = 1L),
                     x_chr = "X", sep = ",", na.strings = c("-", "NA"), comment.char = "#")

    expect_equal(sort(names(control)), sort(names(expected)))
    expected <- expected[names(control)]
    expect_equal(control, expected)

})


test_that("write_control_file in JSON intercross", {

    dir <- tempdir()
    file <- file.path(dir, "iron.json")
    on.exit(unlink(file))

    # write file
    write_control_file(file,
                       crosstype="f2",
                       geno_file="iron_geno.csv",
                       gmap_file="iron_gmap.csv",
                       pheno_file="iron_pheno.csv",
                       covar_file="iron_covar.csv",
                       phenocovar_file="iron_phenocovar.csv",
                       geno_codes=c(SS=1L, SB=2L, BB=3L),
                       sex_covar="sex",
                       sex_codes=c(f="female", m="male"),
                       crossinfo_covar="cross_direction",
                       crossinfo_codes=c("(SxB)x(SxB)"=0L, "(BxS)x(BxS)"=1L),
                       xchr="X",
                       alleles=c("S", "B"),
                       na.strings=c("-", "NA"))

    # read it back in
    control <-  read_control_file(file)

    expected <- list(crosstype = "f2", geno = "iron_geno.csv", pheno = "iron_pheno.csv",
                     phenocovar = "iron_phenocovar.csv", covar = "iron_covar.csv",
                     gmap = "iron_gmap.csv", alleles = c("S", "B"),
                     genotypes = list(SS = 1L, SB = 2L, BB = 3L),
                     sex = list(covar = "sex", f = "female", m = "male"),
                     cross_info = list(covar = "cross_direction", `(SxB)x(SxB)` = 0L, `(BxS)x(BxS)` = 1L),
                     x_chr = "X", sep = ",", na.strings = c("-", "NA"), comment.char = "#")

    expect_equal(sort(names(control)), sort(names(expected)))
    expected <- expected[names(control)]
    expect_equal(control, expected)

})

# subsetting scan1 output

#' Subset scan1 output
#'
#' Subset the output of [scan1()] by chromosome or column
#'
#' @param x An object of class `"scan1"` as returned by
#' [scan1()].
#' @param map A list of vectors of marker positions, as produced by
#' [insert_pseudomarkers()].
#' @param chr Vector of chromosomes.
#' @param lodcolumn Vector of integers or character strings indicating the LOD
#' score columns, either as a numeric indexes or column names.
#' @param ... Ignored
#'
#' @export
#' @return Object of class `"scan1"`, like the input, but subset by chromosome and/or column. See [scan1()].
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' # pull out chromosome 8
#' out_c8 <- subset(out, map, chr="8")
#'
#' # just the second column on chromosome 2
#' out_c2_spleen <- subset(out, map, "2", "spleen")
#'
#' # all positions, but just the "liver" column
#' out_spleen <- subset(out, map, lodcolumn="spleen")
subset_scan1 <-
    function(x, map=NULL, chr=NULL, lodcolumn=NULL, ...)
{
    # align scan1 output and map
    if(!is.null(map)) {
        tmp <- align_scan1_map(x, map)
        x <- tmp$scan1
        map <- tmp$map
    }

    x_attr <- attributes(x)
    x_attrnam <- names(x_attr)
    x_class <- class(x)

    # subset to markers within map
    if(!is.null(map)) {
        map_mnames <- map2markernames(map)
        x_mnames <- rownames(x)

        if(!all(x_mnames %in% map_mnames)) {
            x <- x[x_mnames %in% map_mnames,,drop=FALSE]
            x_mnames <- rownames(x)
        }
        if(!all(map_mnames %in% x_mnames)) {
            for(i in seq_along(map)) {
                mn <- names(map[[i]])
                map[[i]] <- map[[i]][mn %in% x_mnames]
            }
        }
    }

    # subset by chromosome
    if(!is.null(chr)) {
        if(is.null(map))
            stop("To subset by chromosome, you need to provide map object.")

        chr <- subset_chr(chr, names(map))

        # rows to keep
        row <- map2chr(map) %in% chr

        x <- x[row,,drop=FALSE]

        # attributes to subset by row
        for(obj in c("SE", "hsq")) {
            if(obj %in% x_attrnam) {
                if(all(chr %in% rownames(x_attr[[obj]])))
                    x_attr[[obj]] <- x_attr[[obj]][chr,,drop=FALSE]
            }
        }

    }

    # subset by column
    if(!is.null(lodcolumn)) {
        if(is.character(lodcolumn)) {
            cols <- colnames(x)

            tmp <- match(lodcolumn, cols)
            if(any(is.na(tmp)))
                stop("Some lodcolumn not found: ", paste(lodcolumn[is.na(tmp)], collapse=", "))
            lodcolumn <- tmp
        }
        if(is.logical(lodcolumn) && length(lodcolumn) != length(cols))
            stop("lodcolumn is logical but not the correct length (", length(cols), ")")

        x <- x[,lodcolumn,drop=FALSE]

        # attributes to subset list
        sublist <- c("sample_size")
        for(obj in sublist) {
            if(obj %in% x_attrnam)
                x_attr[[obj]] <- x_attr[[obj]][lodcolumn]
        }

        # attributes to subset by column
        for(obj in c("SE", "hsq")) {
            if(obj %in% x_attrnam) {
                x_attr[[obj]] <- x_attr[[obj]][,lodcolumn,drop=FALSE]
            }
        }
    }

    # restore attributes
    for(obj in c("SE", "hsq", "sample_size"))
        attr(x, obj) <- x_attr[[obj]]
    class(x) <- x_class

    x
}

#' @export
#' @rdname subset_scan1
subset.scan1 <- subset_scan1


# grab marker names as a vector
map2markernames <-
    function(map)
{
    nam <- unlist(lapply(map, names))
    names(nam) <- NULL
    nam
}

# grab chromosome IDs as a vector
map2chr <-
    function(map)
{
    chr <- rep(names(map), vapply(map, length, 0))
    names(chr) <- map2markernames(map)
    chr
}

# grab positions as a vector
map2pos <-
    function(map)
{
    pos <- unlist(map)
    names(pos) <- map2markernames(map)
    pos
}

context("find_peaks lod_int, and bayes_int")

# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

# calculate genotype probabilities
map <- insert_pseudomarkers(iron$gmap, step=1)
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

test_that("find_peaks works", {

    expected_2_1 <- structure(list(lodindex = c(1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L, 2L),
                                   lodcolumn = c("liver", "liver", "liver", "liver", "liver", "liver", "liver",
                                   "liver", "liver", "liver", "liver", "liver", "spleen", "spleen", "spleen"),
                                   chr = structure(c(1L, 2L, 4L, 7L, 7L, 8L, 11L, 13L, 15L, 16L, 17L, 19L, 8L, 9L, 10L),
                                                   levels = c("1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11",
                                                              "12", "13", "14", "15", "16", "17", "18", "19", "X"), class = "factor"),
                                   pos = c(90.3, 56.8, 10.9, 25.1, 49.1, 41, 26, 32.5, 49.2, 28.6, 4.3, 38.3, 13.6, 56.6, 37),
                                   lod = c(2.53157055941134, 4.85599006729575, 2.74279943572619, 3.77915164598982, 4.42062344751016,
                                   3.49656665028876, 2.61495248604883, 3.40360749696547, 4.03706766736211, 6.35264382891418,
                                   2.71848298281312, 3.0383533142527, 5.35976176735248, 12.5986057120873, 2.23257885452713)),
                              row.names = c(NA, 15L), class = "data.frame")

    expect_equal(find_peaks(out, map, 2, 1), expected_2_1)

    # test sort
    expect_equal(find_peaks(out, map, 2, 1, sort_by="pos"), expected_2_1[c(1,2,3,4,5,13,6,14,15,7,8,9,10,11,12),])
    expect_equal(find_peaks(out, map, 2, 1, sort_by="lod"), expected_2_1[c(14,10,13,2,5,9,4,6,8,12,3,11,7,1,15),])

    expected_4_Inf <- expected_2_1[c(2, 5, 9, 10, 13, 14),]
    rownames(expected_4_Inf) <- NULL
    expect_equal(find_peaks(out, map, 4, Inf), expected_4_Inf)

    expected_3_1 <- expected_2_1[c(2,4,5,6,8,9,10,12,13,14),]
    rownames(expected_3_1) <- NULL
    expect_equal(find_peaks(out, map, 3, 1), expected_3_1)

    # separate X threshold
    expect_equal(find_peaks(out, map, 2, 1, thresholdX=1.5),
                 rbind(expected_2_1,
                       data.frame(lodindex=2, lodcolumn="spleen",
                                  chr="X", pos=29.5, lod=1.65321447653698,
                                  stringsAsFactors=FALSE)))

    # column-specific thresholds
    result <- find_peaks(out, map, c(4,2), 1, thresholdX=c(0.3, 1.5), peakdropX=c(0.1, 1))
    expected <- rbind(expected_2_1[c(2,5,9,10),],
                      data.frame(lodindex=rep(1,2), lodcolumn=rep("liver",2),
                                 chr=rep("X",2), pos=c(29.5, 57.9),
                                 lod=c(0.832206643447098, 0.35793505359049),
                                 stringsAsFactors=FALSE),
                      expected_2_1[13:15,],
                      data.frame(lodindex=2, lodcolumn="spleen",
                                 chr="X", pos=29.5, lod=1.65321447653698,
                                 stringsAsFactors=FALSE))
    rownames(expected) <- NULL
    expect_equal(result, expected)

    # lod support intervals
    expect_equal(find_peaks(out, map, 4, 2, 1.5),
                 cbind(expected_4_Inf,
                       ci_lo=c(48.1,  1.1, 16.4,  6.6,  0.0, 43.7),
                       ci_hi=c(73.2, 53.6, 49.2, 40.4, 32.7, 61.2)))

    expect_equal(find_peaks(out, map, 4, 2, 2),
                 cbind(expected_4_Inf,
                       ci_lo=c(48.1,  1.1, 16.4,  6.6,  0.0, 43.7),
                       ci_hi=c(73.2, 53.6, 49.2, 40.4, 32.7, 61.2)))

    expected_3_1_lodint <- cbind(expected_3_1,
                                 ci_lo=c(48.1,  1.1, 37.2, 17.3, 17.5, 16.4,  6.6,  3.3,  0.0, 53.6),
                                 ci_hi=c(73.2, 28.4, 53.6, 69.9, 40.4, 49.2, 40.4, 38.3, 17.3, 61.2))

    expect_equal(find_peaks(out, map, 3, 1, 0.9), expected_3_1_lodint)

    skip_on_cran()

    # sorting
    expect_equal(find_peaks(out, map, 3, 1, 0.9, sort_by="pos"), expected_3_1_lodint[c(1,2,3,9,4,10,5,6,7,8),])
    expect_equal(find_peaks(out, map, 3, 1, 0.9, sort_by="lod"), expected_3_1_lodint[c(10,7,9,1,3,6,2,4,5,8),])

    # expand2markers=FALSE
    expect_equal(find_peaks(out, map, 3, 1, 0.9, expand2markers=FALSE),
                 cbind(expected_3_1,
                       ci_lo=c(53.3, 11.1, 38.1, 24.0, 20.5, 38.4, 21.6, 28.3,  8.0, 53.6),
                       ci_hi=c(66.3, 28.1, 53.6, 53.0, 40.4, 49.2, 32.6, 38.3, 17.0, 59.6)))

    # Bayes intervals
    expect_equal(find_peaks(out, map, 4, 2, prob=0.95),
                 cbind(expected_4_Inf,
                       ci_lo=c(48.1, 13.1, 16.4,  6.6,  0.0, 53.6),
                       ci_hi=c(73.2, 53.6, 49.2, 40.4, 32.7, 61.2)))

    expect_equal(find_peaks(out, map, 4, 2, prob=0.99),
                 cbind(expected_4_Inf,
                       ci_lo=c(48.1,  1.1, 16.4,  6.6,  0.0, 43.7),
                       ci_hi=c(73.2, 53.6, 49.2, 40.4, 32.7, 61.2)))

    expected_3_1_bayesint <- cbind(expected_3_1,
                                   ci_lo=c(48.1, 13.1, 37.2, 32.7, 17.5, 16.4,  6.6,  3.3,  0.0, 53.6),
                                   ci_hi=c(73.2, 28.4, 53.6, 69.9, 40.4, 49.2, 40.4, 38.3, 17.3, 61.2))

    expect_equal(find_peaks(out, map, 3, 1, prob=0.8), expected_3_1_bayesint)

    # sorting
    expect_equal(find_peaks(out, map, 3, 1, prob=0.8, sort_by="pos"), expected_3_1_bayesint[c(1,2,3,9,4,10,5,6,7,8),])
    expect_equal(find_peaks(out, map, 3, 1, prob=0.8, sort_by="lod"), expected_3_1_bayesint[c(10,7,9,1,3,6,2,4,5,8),])

    # expand2markers=FALSE
    expect_equal(find_peaks(out, map, 3, 1, 0.8, expand2markers=FALSE),
                 cbind(expected_3_1,
                       ci_lo=c(53.3, 12.1, 39.1, 26.0, 21.5, 39.4, 25.1, 28.3,  9.0, 53.6),
                       ci_hi=c(65.3, 28.1, 53.6, 52.0, 40.4, 49.2, 32.6, 38.3, 17.0, 59.6)))


    # test that it works if output or map are subsetted
    expect_equal( find_peaks(out, map["2"]), find_peaks(subset(out, map, chr="2"), map) )

    # test that it works if output has rows shuffled
    out_shuffled <- out[sample(1:nrow(out)),,drop=FALSE]
    class(out_shuffled) <- c("scan1", "matrix")
    expect_equal( find_peaks(out_shuffled, map), find_peaks(out, map))

    # test that find_peaks works if there are no peaks above threshold
    blank_output <- structure(list(lodindex = numeric(0),
                                   lodcolumn = character(0),
                                   chr = structure(integer(0), levels = c("1", "2", "3", "4", "5", "6", "7", "8", "9", "10",
                                                                          "11", "12", "13", "14", "15", "16", "17", "18", "19", "X"),
                                                   class = "factor"),
                                   pos = numeric(0),
                                   lod = numeric(0)),
                              row.names = integer(0), class = "data.frame")

    expect_equal( find_peaks(out, map, threshold=9999), blank_output)

    # test that find_peaks works if there are no peaks above threshold
    # like above, but also requesting LOD or Bayes intervals
    blank_output <- structure(list(lodindex = numeric(0),
                                   lodcolumn = character(0),
                                   chr = structure(integer(0), levels = c("1", "2", "3", "4", "5", "6", "7", "8", "9", "10",
                                                                          "11", "12", "13", "14", "15", "16", "17", "18", "19", "X"),
                                                   class = "factor"),
                                   pos = numeric(0),
                                   lod = numeric(0),
                                   ci_lo = numeric(0),
                                   ci_hi = numeric(0)),
                              row.names = integer(0), class = "data.frame")

    expect_equal( find_peaks(out, map, threshold=9999, drop=2), blank_output)
    expect_equal( find_peaks(out, map, threshold=9999, prob=0.95), blank_output)

})

test_that("lod_int works", {

    expected <- cbind(ci_lo=1.1, pos=49.1, ci_hi=53.6)
    rownames(expected) <- 1
    expect_equal(lod_int(out, map, 7, 1), expected)

    expected[1,c(1,3)] <- c(12.1, 53.6)
    expect_equal(lod_int(out, map, 7, 1, expand2markers=FALSE), expected)

    expected[1,c(1,3)] <- c(37.2, 53.6)
    expect_equal(lod_int(out, map, 7, 1, drop=0.5), expected)

    expected2 <- rbind("1"=c(1.1,25.1,28.4), "2"=c(37.2,49.1,53.6))
    colnames(expected2) <- colnames(expected)
    expect_equal(lod_int(out, map, 7, 1, 0, 1.5, 1), expected2)

    expected3 <- cbind(ci_lo=43.7, pos=56.6, ci_hi=61.2)
    rownames(expected3) <- 1
    expect_equal(lod_int(out, map, 9, 2), expected3)

    expected4 <- cbind(ci_lo=0.0, pos=13.6, ci_hi=32.7)
    rownames(expected4) <- 1
    expect_equal(lod_int(out, map, 8, 2), expected4)

    expect_equal( lod_int(out, map, chr="2"), lod_int(subset(out, map, chr="2"), map, chr="2"))
    expect_equal( lod_int(out, map, chr="2"), lod_int(out, map["2"], chr="2"))

})


test_that("bayes_int works", {

    expected <- cbind(ci_lo=13.1, pos=49.1, ci_hi=53.6)
    rownames(expected) <- 1
    expect_equal(bayes_int(out, map, 7, 1), expected)

    expected[1,c(1,3)] <- c(15.1, 53.6)
    expect_equal(bayes_int(out, map, 7, 1, expand2markers=FALSE), expected)

    expected[1,c(1,3)] <- c(37.2, 53.6)
    expect_equal(bayes_int(out, map, 7, 1, prob=0.8), expected)

    expected2 <- rbind("1"=c(1.1,25.1,28.4), "2"=c(37.2,49.1,53.6))
    colnames(expected2) <- colnames(expected)
    expect_equal(bayes_int(out, map, 7, 1, 0, 1.5, 0.95), expected2)

    expected3 <- cbind(ci_lo=53.6, pos=56.6, ci_hi=61.2)
    rownames(expected3) <- 1
    expect_equal(bayes_int(out, map, 9, 2), expected3)

    expected4 <- cbind(ci_lo=0.0, pos=13.6, ci_hi=32.7)
    rownames(expected4) <- 1
    expect_equal(bayes_int(out, map, 8, 2), expected4)

    expect_equal( bayes_int(out, map, chr="2"), bayes_int(subset(out, map, chr="2"), map, chr="2"))
    expect_equal( bayes_int(out, map, chr="2"), bayes_int(out, map["2"], chr="2"))

})


test_that("lod_int and bayes_int give same results as R/qtl", {

    skip_on_cran()

    out_rqtl <- data.frame(chr=map2chr(map),
                           pos=map2pos(map),
                           unclass(out))
    rownames(out_rqtl) <- map2markernames(map)
    class(out_rqtl) <- c("scanone", "data.frame")

    # 1st lod col: chr 2, 7, 15, 16
    # 2nd lod col: chr 8, 9
    for(lod in 1:2) {
        for(chr in list(c(2,7,15,16), c(8,9))[[lod]]) {

            for(drop in c(1, 1.5, 2)) {
                # expand to markers
                li_rqtl <- qtl::lodint(out_rqtl, lod=lod, chr=chr, expandtomarkers=TRUE, drop=drop)
                li_qtl2 <- lod_int(out, map, lod=lod, chr=chr, drop=drop)
                expect_equivalent(li_rqtl[,2], as.numeric(li_qtl2))

                # don't expand to markers
                li_rqtl <- qtl::lodint(out_rqtl, lod=lod, chr=chr, drop=drop)
                li_qtl2 <- lod_int(out, map, lod=lod, chr=chr, expand2markers=FALSE, drop=drop)
                expect_equivalent(li_rqtl[,2], as.numeric(li_qtl2))
            }

            for(prob in c(0.90, 0.95, 0.99)) {
                # expand to markers
                bi_rqtl <- qtl::bayesint(out_rqtl, lod=lod, chr=chr, expandtomarkers=TRUE, prob=prob)
                bi_qtl2 <- bayes_int(out, map, lod=lod, chr=chr, prob=prob)
               expect_equivalent(bi_rqtl[,2], as.numeric(bi_qtl2))

                # don't expand to markers
                bi_rqtl <- qtl::bayesint(out_rqtl, lod=lod, chr=chr, prob=prob)
                bi_qtl2 <- bayes_int(out, map, lod=lod, chr=chr, expand2markers=FALSE, prob=prob)
                expect_equivalent(bi_rqtl[,2], as.numeric(bi_qtl2))
            }
        }
    }


})

test_that("find_peaks works with snpinfo table", {

    skip_if(isnt_karl(), "this test only run locally")

    # load example data and calculate genotype probabilities
    file <- paste0("https://raw.githubusercontent.com/rqtl/",
                   "qtl2data/main/DOex/DOex.zip")
    DOex <- read_cross2(file)
    probs <- calc_genoprob(DOex[1:20,"2"], error_prob=0.002)

    snpdb_file <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
    queryf <- create_variant_query_func(snpdb_file)

    out <- scan1snps(probs, DOex$pmap, DOex$pheno, query_func=queryf, chr=2, start=97.2, end=97.3)

    # test max_scan1()
    expect_equal(max(out$lod, out$snpinfo),
                 data.frame(chr="2",
                            pos=out$snpinfo[6,"pos"],
                            OF_immobile_pct=out$lod[6],
                            row.names=out$snpinfo[6,"snp"],
                            stringsAsFactors=FALSE))

    # find_peaks (no peaks above threshold)
    expect_equal(find_peaks(out$lod, out$snpinfo),
                 data.frame(lodindex=numeric(0),
                            lodcolumn=character(0),
                            chr=factor(character(0), "2"),
                            pos=numeric(0),
                            lod=numeric(0),
                            stringsAsFactors=FALSE))
    expect_equal(find_peaks(out$lod, out$snpinfo, drop=1.5),
                 data.frame(lodindex=numeric(0),
                            lodcolumn=character(0),
                            chr=factor(character(0), "2"),
                            pos=numeric(0),
                            lod=numeric(0),
                            ci_lo=numeric(0),
                            ci_hi=numeric(0),
                            stringsAsFactors=FALSE))
    expect_equal(find_peaks(out$lod, out$snpinfo, prob=0.95),
                 data.frame(lodindex=numeric(0),
                            lodcolumn=character(0),
                            chr=factor(character(0), "2"),
                            pos=numeric(0),
                            lod=numeric(0),
                            ci_lo=numeric(0),
                            ci_hi=numeric(0),
                            stringsAsFactors=FALSE))

    # find peaks (one peak above threshold)
    expect_equal(find_peaks(out$lod, out$snpinfo, threshold=0.5),
                 data.frame(lodindex=1,
                            lodcolumn="OF_immobile_pct",
                            chr=factor("2", "2"),
                            pos=out$snpinfo[6,"pos"],
                            lod=out$lod[6],
                            row.names=1L,
                            stringsAsFactors=FALSE))
    expect_equal(find_peaks(out$lod, out$snpinfo, threshold=0.5, drop=1.5),
                 data.frame(lodindex=1,
                            lodcolumn="OF_immobile_pct",
                            chr=factor("2", "2"),
                            pos=out$snpinfo[6,"pos"],
                            lod=out$lod[6],
                            ci_lo=min(out$snpinfo$pos),
                            ci_hi=max(out$snpinfo$pos),
                            row.names=1L,
                            stringsAsFactors=FALSE))
    expect_equal(find_peaks(out$lod, out$snpinfo, threshold=0.5, prob=0.95),
                 data.frame(lodindex=1,
                            lodcolumn="OF_immobile_pct",
                            chr=factor("2", "2"),
                            pos=out$snpinfo[6,"pos"],
                            lod=out$lod[6],
                            ci_lo=min(out$snpinfo$pos),
                            ci_hi=max(out$snpinfo$pos),
                            row.names=1L,
                            stringsAsFactors=FALSE))

})

context("subset_scan1")

test_that("subset_scan1 works for intercross with two phenotypes", {

    # read data
    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    # calculate genotype probabilities
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    # grab phenotypes and covariates; ensure that covariates have names attribute
    pheno <- iron$pheno
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    # perform genome scan
    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

    # subset one column
    expected <- unclass(out)[,2,drop=FALSE]
    attr(expected, "sample_size") <- attr(out, "sample_size")[2]
    class(expected) <- c("scan1", "matrix")
    expect_equal(subset(out, map, lodcolumn=2), expected)

    # subset chr 2, 8, and 9
    keep <- (map2chr(map) %in% c("2", "8", "9"))
    expected <- unclass(out)[keep,,drop=FALSE]
    attr(expected, "sample_size") <- attr(out, "sample_size")
    class(expected) <- c("scan1", "matrix")
    expect_equal(subset(out, map, chr=c("2", "8", "9")), expected)

    # subset chr 2, 8, and 9 and lodcolumn 2
    expected <- unclass(expected)[,2,drop=FALSE]
    attr(expected, "sample_size") <- attr(out, "sample_size")[2]
    class(expected) <- c("scan1", "matrix")
    expect_equal(subset(out, map, c("2", "8", "9"), 2), expected)

})

# Round a number, preserving extra 0's. (taken from R/broman, https://github.com/kbroman/broman)
myround <-
    function(x, digits=1)
{
    if(digits < 1)
        stop("This is intended for the case digits >= 1.")

    if(length(digits) > 1) {
        digits <- digits[1]
        warning("Using only digits[1]")
    }

    tmp <- sprintf(paste("%.", digits, "f", sep=""), x)

    # deal with "-0.00" case
    zero <- paste0("0.", paste(rep("0", digits), collapse=""))
    tmp[tmp == paste0("-", zero)] <- zero

    tmp
}

# get_x_covar
#' Get X chromosome covariates
#'
#' Get the matrix of covariates to be used for the null hypothesis when
#' performing QTL analysis with the X chromosome.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#'
#' @return A matrix of size individuals x no. covariates.
#'
#' @details For most crosses, the result is either `NULL`
#' (indicating no additional covariates are needed) or a
#' matrix with a single column containing sex indicators (1 for males
#' and 0 for females).
#'
#' For an intercross, we also consider cross direction. There are
#' four cases: (1) All male or all female but just one direction: no
#' covariate; (2) All female but both directions: covariate indicating
#' cross direction; (3) Both sexes, one direction: covariate
#' indicating sex; (4) Both sexes, both directions: a covariate
#' indicating sex and a covariate that is 1 for females from the
#' reverse direction and 0 otherwise.
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' xcovar <- get_x_covar(iron)

get_x_covar <-
function(cross)
{
    # check input
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')

    x <- .get_x_covar(cross$crosstype, cross$is_female, cross$cross_info)

    if(ncol(x)==0) return(NULL)

    # add rownames
    nam <- names(cross$is_female)
    if(!is.null(nam)) rownames(x) <- nam

    x
}

context("scan1 permutations with scan_func")

test_that("scan1 permutations work with scan_func", {

    skip_if(isnt_karl(), "this test only run locally")

    seed <- 3025685
    RNGkind("L'Ecuyer-CMRG")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    Xcovar <- get_x_covar(iron)
    bin_pheno <- setNames(as.numeric(iron$pheno[,1] > median(iron$pheno[,1])),
                          rownames(iron$pheno))

    scanf <- function(genoprobs, pheno, kinship, addcovar, Xcovar, intcovar, weights)
        scan1(genoprobs, pheno=pheno, kinship=kinship, addcovar=addcovar,
              Xcovar=Xcovar, intcovar=intcovar, weights=weights, model="binary")

    set.seed(20260615)
    operm1 <- scan1perm(probs, bin_pheno, Xcovar=Xcovar, model="binary", n_perm=10)
    set.seed(20260615)
    operm2 <- scan1perm(probs, bin_pheno, Xcovar=Xcovar, scan_func=scanf, n_perm=10)

    expect_equal(operm1, operm2)

    ### glm with link=probit

    ll_glm <-
    function(pr, pheno, addcovar=NULL, ...)
    {
        formula <- ifelse(is.null(pr), "pheno ~ 1", "pheno ~ pr")
        if(!is.null(addcovar)) formula <- paste(formula, "+ addcovar")

        glm_out <- glm(as.formula(formula), family=binomial(link=probit))
        -glm_out$deviance/(2*log(10)) # log10 likelihood
    }

    set.seed(20260615)
    operm3 <- scan1perm(probs, bin_pheno, Xcovar=Xcovar, scan_func=scan1gen,
                        func=ll_glm, n_perm=10, cores=10)

    expected <- structure(c(2.17385652559186, 2.31427078248578, 3.43941883075001,
                            1.91898059153991, 2.01118425059423, 2.32202927408811, 1.72215414380619,
                            1.31001918257803, 1.37739899949347, 2.68954157272938),
                          dim = c(10L, 1L), dimnames = list(NULL, "pheno1"),
                          class = c("scan1perm", "matrix"))
    expect_equal(operm3, expected)

    set.seed(20260615)
    operm4 <- scan1perm(probs[,1:19], bin_pheno, scan_func=scan1gen,
                        func=ll_glm, n_perm=10, cores=10)

    expected <- structure(c(2.68609128205085, 1.77002204352716, 1.93368949267834,
                            2.31094501376531, 2.79567905567447, 1.79725590191482, 2.79244070919604,
                            1.3401617230496, 3.955798656798, 2.23867940226528),
                          dim = c(10L, 1L), dimnames = list(NULL, "pheno1"),
                          class = c("scan1perm", "matrix"))
    expect_equal(operm4, expected)


    set.seed(20260615)
    operm5 <- scan1perm(probs, bin_pheno, Xcovar=Xcovar, scan_func=scan1gen,
                        func=ll_glm, n_perm=10, cores=10, perm_Xsp=TRUE,
                        chr_lengths=chr_lengths(iron$gmap))

    expected_summary <- structure(list(A = structure(c(2.0702243751226, 2.20882194904352, 2.56447872617982, 2.60840127327936),
                                                     dim = c(4L, 1L), dimnames = list(c("0.5", "0.25", "0.1", "0.05"), "pheno1")),
                                       X = structure(c(2.79707347743163, 3.33791604605526, 3.61256992679585, 4.07772316813587),
                                                     dim = c(4L, 1L), dimnames = list(c("0.5", "0.25", "0.1", "0.05"), "pheno1"))),
                                  class = c("summary.scan1perm", "list"),
                                  n_perm = structure(c(10, 283), dim = 2:1, dimnames = list(c("A", "X"), "pheno1")))
    expect_equal(summary(operm5, c(0.5, 0.25, 0.1, 0.05)), expected_summary)


    # permutations with scan1snps
    file <- paste0("https://raw.githubusercontent.com/rqtl/",
                   "qtl2data/main/DOex/DOex.zip")
    DOex <- read_cross2(file)
    DOex <- DOex[,c("2", "3")] # subset to chr 2 and 3
    probs <- calc_genoprob(DOex, error_prob=0.002)

    snpdb_file <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
    queryf <- create_variant_query_func(snpdb_file)

    set.seed(20260616)
    operm6 <- scan1perm(genoprobs=probs, map=DOex$pmap, pheno=DOex$pheno,
                       scan_func=scan1snps, query_func=queryf, n_perm=10)
    expected <- structure(c(1.87603485437823, 2.78903657330818, 1.28168354734608,
                            1.96766231770538, 2.00872833761448, 2.64571234938997, 2.84375635883363,
                            1.86947672509228, 1.49270250736036, 1.7635254846055),
                          dim = c(10L, 1L), dimnames = list(NULL, "OF_immobile_pct"),
                          class = c("scan1perm", "matrix"))
    expect_equal(operm6, expected)

})

context("BLUP estimates of QTL effects")

calc_blup <-
    function(probs, pheno, kinship=NULL, addcovar=NULL, tol=1e-12,
             quiet=TRUE)
{
    addcovar <- cbind(rep(1, length(pheno)), addcovar)
    if(!is.null(kinship)) {
        Ke <- decomp_kinship(kinship)
        Keval <- Ke$values
        Kevec <- Ke$vectors

        pheno <- Kevec %*% pheno
        addcovar <- Kevec %*% addcovar
        probs <- Kevec %*% probs

        lmmfit <- Rcpp_fitLMM(Keval, pheno, addcovar, tol=tol)
        hsq <- lmmfit$hsq
        if(!quiet) message("hsq_pg: ", hsq)
        wts <- 1/sqrt(hsq * Keval + (1-hsq))

        pheno <- wts * pheno
        addcovar <- wts * addcovar
        probs <- wts * probs
    }

    k <- probs %*% t(probs)
    ke <- decomp_kinship(k)
    keval <- ke$values
    kevec <- ke$vectors

    pheno <- kevec %*% pheno
    addcovar <- kevec %*% addcovar

    lmmfit <- Rcpp_fitLMM(keval, pheno, addcovar, tol=tol)
    beta <- lmmfit$beta
    hsq <- lmmfit$hsq
    if(!quiet) message("hsq_qtl: ", hsq)
    wts <- hsq * keval + (1-hsq)
    u <- as.numeric( t(kevec %*% probs) %*% diag(hsq/wts) %*%  (pheno - addcovar %*% beta) )
    c(u, beta)
}


iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
iron <- iron[c(1:30, 190:210),]
phe <- iron$pheno[,1,drop=FALSE]

test_that("scan1blup works with no kinship matrix", {

    pr <- calc_genoprob(iron[,"16"])

    blup <- scan1blup(pr, phe)
    blup_se <- scan1blup(pr, phe, se=TRUE)
    expect_equivalent(blup, blup_se)

    # brute force BLUPs
    for(i in 1:dim(pr[[1]])[[3]]) {
        blup_alt <- calc_blup(pr[[1]][,,i], phe)
        names(blup_alt) <- colnames(blup)
        expect_equal(unclass(blup)[i,], blup_alt, tol=1e-6)
    }

    # repeat with an additive covariate
    sex <- as.numeric(iron$covar$sex=="m")
    names(sex) <- rownames(iron$covar)

    blup <- scan1blup(pr, phe, addcovar=sex)
    blup_se <- scan1blup(pr, phe, addcovar=sex, se=TRUE)
    expect_equivalent(blup, blup_se)

    # brute force BLUPs
    for(i in 1:dim(pr[[1]])[[3]]) {
        blup_alt <- calc_blup(pr[[1]][,,i], phe, addcovar=sex)
        names(blup_alt) <- colnames(blup)
        expect_equal(unclass(blup)[i,], blup_alt, tolerance=1e-5)
    }

})

test_that("scan1blup works with kinship matrix", {

    skip_on_cran()

    pr <- calc_genoprob(iron)
    K <- calc_kinship(pr[,c(1:15,17:19,"X")])
    pr <- pr[,"16"]

    # sex as an additive covariate
    sex <- as.numeric(iron$covar$sex=="m")
    names(sex) <- rownames(iron$covar)

    blup <- scan1blup(pr, phe, K, sex)
    blup_se <- scan1blup(pr, phe, K, sex, se=TRUE)
    expect_equivalent(blup, blup_se)

    # brute force BLUPs
    for(i in 1:dim(pr[[1]])[[3]]) {
        blup_alt <- calc_blup(pr[[1]][,,i], phe, K, sex)
        names(blup_alt) <- colnames(blup)
        expect_equal(unclass(blup)[i,], blup_alt, tolerance=1e-5)
    }

})

test_that("scan1blup works with kinship matrix on another chromosome", {

    skip_if(isnt_karl(), "this test only run locally")

    pr <- calc_genoprob(iron)
    K <- calc_kinship(pr[,c(1:10,12:19,"X")])
    pr <- pr[,"11"]

    # sex as an additive covariate
    sex <- as.numeric(iron$covar$sex=="m")
    names(sex) <- rownames(iron$covar)

    blup <- scan1blup(pr, phe, K, sex)
    blup_se <- scan1blup(pr, phe, K, sex, se=TRUE)
    expect_equivalent(blup, blup_se)

    # brute force BLUPs
    for(i in 1:dim(pr[[1]])[[3]]) {
        blup_alt <- calc_blup(pr[[1]][,,i], phe, K, sex)
        names(blup_alt) <- colnames(blup)
        expect_equal(unclass(blup)[i,], blup_alt, tolerance=1e-5)
    }

})

test_that("scan1blup deals with mismatching individuals", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs, "loco")[["3"]]
    probs <- probs[,"3"]
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    phe <- iron$pheno[,2,drop=FALSE]

    ind <- c(1:50, 101:150)
    expected <- scan1blup(probs[ind,], phe[ind,,drop=FALSE], kinship[ind,ind], addcovar=X[ind])
    expect_equal(scan1blup(probs[ind,], phe, kinship, addcovar=X), expected)
    expect_equal(scan1blup(probs, phe[ind,,drop=FALSE], kinship, addcovar=X), expected)
    expect_equal(scan1blup(probs, phe, kinship[ind,ind], addcovar=X), expected)
    expect_equal(scan1blup(probs, phe, kinship, addcovar=X[ind]), expected)

    # repeat with no kinship
    expected <- scan1blup(probs[ind,], phe[ind,,drop=FALSE], addcovar=X[ind])
    expect_equal(scan1blup(probs[ind,], phe, addcovar=X), expected)
    expect_equal(scan1blup(probs, phe[ind,,drop=FALSE], addcovar=X), expected)
    expect_equal(scan1blup(probs, phe, addcovar=X[ind]), expected)

})

context("effect scan by scan1coef")

convert_probs2qtl2 <-
    function(cross)
{
    result <- lapply(cross$geno, function(a) {
        pr <- aperm(a$prob, c(1,3,2))
        rownames(pr) <- paste(1:nrow(pr))
        pr })

    class(result) <- c("calc_genoprob", "list")
    result
}

# calc estimates via lm(), just one chromosome
eff_via_lm <-
    function(probs, pheno, addcovar=NULL, intcovar=NULL, weights=NULL,
             se=TRUE)
{
    npos <- dim(probs)[2]
    nind <- length(pheno)
    ngen <- dim(probs)[3]

    nadd <- ifelse(is.null(addcovar), 0, ncol(addcovar))
    nint <- ifelse(is.null(intcovar), 0, ncol(intcovar))
    ncoef <- ngen + nadd + (ngen-1)*nint

    result <- matrix(nrow=npos, ncol=ncoef)
    SEs <- matrix(nrow=npos, ncol=ncoef)

    if(is.null(weights)) wts <- rep(1, nind)
    else wts <- weights
    if(!is.null(intcovar)) intcovar <- as.matrix(intcovar)

    for(i in 1:npos) {
        X <- probs[,i,]
        if(!is.null(addcovar)) X <- cbind(X, addcovar)
        if(!is.null(intcovar)) {
            for(j in 1:ncol(intcovar))
                X <- cbind(X, probs[,i,-1]*intcovar[,j])
        }

        lm.out <- lm(pheno ~ -1 + X, weights=weights)
        result[i,] <- lm.out$coef
        if(se) # need to deal with NAs
            SEs[i,!is.na(lm.out$coef)] <- summary(lm.out)$coef[,2]
        else SEs <- NULL
    }

    attr(result, "SE") <- SEs

    class(result) <- c("scan1coef", "scan1", "matrix")
    result
}

test_that("scan1coef for backcross", {

    set.seed(9308594)

    library(qtl)
    data(hyper)
    hyper <- hyper[4,] # chr 4 only
    hyper <- calc.genoprob(hyper, step=5)
    prob <- hyper$geno[[1]]$prob
    phe <- hyper$pheno[,1]
    weights <- runif(nind(hyper), 0, 5)
    covar <- cbind(sex=sample(0:1, nind(hyper), replace=TRUE))
    rownames(prob) <- names(phe) <- names(weights) <- rownames(covar) <- paste(1:nind(hyper))

    # probs for qtl2 code
    prob2 <- convert_probs2qtl2(hyper)

    # no covariates
    expected <- eff_via_lm(prob, phe)
    coef <- scan1coef(prob2, phe, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA")))
    coef <- scan1coef(prob2, phe, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(colnames(prob), c("BB", "BA")))

    # no covariates, weighted
    expected <- eff_via_lm(prob, phe, weights=weights)
    coef <- scan1coef(prob2, phe, weights=weights, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA")))
    coef <- scan1coef(prob2, phe, weights=weights, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(colnames(prob), c("BB", "BA")))

    # one add've covariate
    expected <- eff_via_lm(prob, phe, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(colnames(prob), c("BB", "BA", "sex")))

    # one add've covariate, weighted
    expected <- eff_via_lm(prob, phe, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, weights=weights, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, weights=weights, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(colnames(prob), c("BB", "BA", "sex")))

    # one int've covariate
    expected <- eff_via_lm(prob, phe, covar, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex", "BA:sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex", "BA:sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(colnames(prob), c("BB", "BA", "sex", "BA:sex")))

    # one int've covariate, weighted
    expected <- eff_via_lm(prob, phe, covar, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, zerosum=FALSE)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex", "BA:sex")))
    expect_equivalent(coef, expected)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, se=TRUE, zerosum=FALSE)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex", "BA:sex")))
    expect_equivalent(coef, expected)
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(colnames(prob), c("BB", "BA", "sex", "BA:sex")))

    # two int've covariates
    covar <- cbind(covar, another=rnorm(nind(hyper)))
    expected <- eff_via_lm(prob, phe, covar, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, zerosum=FALSE)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex", "another", "BA:sex", "BA:another")))
    expect_equivalent(coef, expected)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex", "another", "BA:sex", "BA:another")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")),
                 list(colnames(prob), c("BB", "BA", "sex", "another", "BA:sex", "BA:another")))

    # two int've covariates, weighted
    expected <- eff_via_lm(prob, phe, covar, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex", "another", "BA:sex", "BA:another")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(colnames(prob), c("BB", "BA", "sex", "another", "BA:sex", "BA:another")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")),
                 list(colnames(prob), c("BB", "BA", "sex", "another", "BA:sex", "BA:another")))

})

test_that("scan1coef for backcross, with contrasts", {

    skip_on_cran()

    set.seed(9308594)

    library(qtl)
    data(hyper)
    hyper <- hyper[4,] # chr 4 only
    hyper <- calc.genoprob(hyper, step=5)
    prob <- hyper$geno[[1]]$prob
    phe <- hyper$pheno[,1]
    weights <- runif(nind(hyper), 0, 5)
    covar <- cbind(sex=sample(0:1, nind(hyper), replace=TRUE))
    rownames(prob) <- names(phe) <- names(weights) <- rownames(covar) <- paste(1:nind(hyper))

    # probs for qtl2 code
    prob2 <- convert_probs2qtl2(hyper)

    # use contrasts
    contrasts <- cbind(mu=c(1,1), a=c(-0.5, 0.5))

    # change to contrasts for use with lm()
    p <- prob
    dim(p) <- c(prod(dim(prob)[1:2]), dim(prob)[3])
    p <- p %*% contrasts
    dim(p) <- dim(prob)
    prob <- p
    dimnames(prob) <- list(rownames(prob2), dimnames(prob2[[1]])[[3]], colnames(contrasts))

    posnames <- dimnames(prob2[[1]])[[3]]

    # no covariates
    expected <- eff_via_lm(prob, phe)
    coef <- scan1coef(prob2, phe, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a")))
    coef <- scan1coef(prob2, phe, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a")))

    # no covariates, weighted
    expected <- eff_via_lm(prob, phe, weights=weights)
    coef <- scan1coef(prob2, phe, weights=weights, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a")))
    coef <- scan1coef(prob2, phe, weights=weights, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a")))

    # one add've covariate
    expected <- eff_via_lm(prob, phe, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "sex")))

    # one add've covariate, weighted
    expected <- eff_via_lm(prob, phe, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, weights=weights, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, weights=weights, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "sex")))

    # one int've covariate
    expected <- eff_via_lm(prob, phe, covar, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex", "a:sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex", "a:sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "sex", "a:sex")))

    # one int've covariate, weighted
    expected <- eff_via_lm(prob, phe, covar, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex", "a:sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex", "a:sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "sex", "a:sex")))

    # two int've covariates
    covar <- cbind(covar, another=rnorm(nind(hyper)))
    expected <- eff_via_lm(prob, phe, covar, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex", "another", "a:sex", "a:another")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex", "another", "a:sex", "a:another")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")),
                 list(posnames, c("mu", "a", "sex", "another", "a:sex", "a:another")))

    # two int've covariates, weighted
    expected <- eff_via_lm(prob, phe, covar, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex", "another", "a:sex", "a:another")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "sex", "another", "a:sex", "a:another")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")),
                 list(posnames, c("mu", "a", "sex", "another", "a:sex", "a:another")))

})



test_that("scan1coef for intercross", {

    skip_on_cran()

    set.seed(9308594)

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    prob2 <- calc_genoprob(iron[,7], map, error_prob=0.002) # just look at chr 7
    phe <- iron$pheno[,1] # liver phenotype
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    covar <- cbind(sex=covar)
    weights <- runif(n_ind(iron), 0, 5)
    names(weights) <- names(phe)

    # different organization of probs for qtl2 and lm() code
    prob <- aperm(prob2[[1]], c(1,3,2)) # rearrange as expected for lm()

    # no covariates
    expected <- eff_via_lm(prob, phe)
    coef <- scan1coef(prob2, phe, zerosum=FALSE)
    expect_equivalent(coef, expected)
    coef <- scan1coef(prob2, phe, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))

    posnames <- dimnames(prob2[[1]])[[3]]

    # no covariates, weighted
    expected <- eff_via_lm(prob, phe, weights=weights)
    coef <- scan1coef(prob2, phe, weights=weights, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("SS", "SB", "BB")))
    coef <- scan1coef(prob2, phe, weights=weights, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("SS", "SB", "BB")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("SS", "SB", "BB")))

    # one add've covariate
    expected <- eff_via_lm(prob, phe, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("SS", "SB", "BB", "sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("SS", "SB", "BB", "sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("SS", "SB", "BB", "sex")))

    # one add've covariate, weighted
    expected <- eff_via_lm(prob, phe, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, weights=weights, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("SS", "SB", "BB", "sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, weights=weights, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("SS", "SB", "BB", "sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("SS", "SB", "BB", "sex")))

    # one int've covariate
    expected <- eff_via_lm(prob, phe, covar, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("SS", "SB", "BB", "sex", "SB:sex", "BB:sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("SS", "SB", "BB", "sex", "SB:sex", "BB:sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("SS", "SB", "BB", "sex", "SB:sex", "BB:sex")))

    # one int've covariate, weighted
    expected <- eff_via_lm(prob, phe, covar, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("SS", "SB", "BB", "sex", "SB:sex", "BB:sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("SS", "SB", "BB", "sex", "SB:sex", "BB:sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("SS", "SB", "BB", "sex", "SB:sex", "BB:sex")))

    # two int've covariates
    covar <- cbind(covar, another=rnorm(n_ind(iron)))
    expected <- eff_via_lm(prob, phe, covar, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, zerosum=FALSE)
    expect_equal(dimnames(coef),
                 list(posnames, c("SS", "SB", "BB", "sex", "another", "SB:sex", "BB:sex",
                                        "SB:another", "BB:another")))
    expect_equivalent(coef, expected)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef),
                 list(posnames, c("SS", "SB", "BB", "sex", "another", "SB:sex", "BB:sex",
                                        "SB:another", "BB:another")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")),
                 list(posnames, c("SS", "SB", "BB", "sex", "another", "SB:sex", "BB:sex",
                                        "SB:another", "BB:another")))

    # two int've covariates, weighted
    expected <- eff_via_lm(prob, phe, covar, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef),
                 list(posnames, c("SS", "SB", "BB", "sex", "another", "SB:sex", "BB:sex",
                                        "SB:another", "BB:another")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, se=TRUE, zerosum=FALSE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef),
                 list(posnames, c("SS", "SB", "BB", "sex", "another", "SB:sex", "BB:sex",
                                        "SB:another", "BB:another")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")),
                 list(posnames, c("SS", "SB", "BB", "sex", "another", "SB:sex", "BB:sex",
                                        "SB:another", "BB:another")))

})

test_that("scan1coef for intercross, with contrasts", {

    skip_on_cran()

    set.seed(9308594)

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    prob2 <- calc_genoprob(iron[,7], map, error_prob=0.002) # just look at chr 7
    phe <- iron$pheno[,1] # liver phenotype
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    covar <- cbind(sex=covar)
    weights <- runif(n_ind(iron), 0, 5)
    names(weights) <- names(phe)

    # different organization of probs for qtl2 and lm() code
    prob <- aperm(prob2[[1]], c(1,3,2)) # rearrange as expected for lm()

    # use contrasts
    contrasts <- cbind(mu=c(1,1,1), a=c(-1, 0, 1), d=c(0, 1,0))

    # change to contrasts for use with lm()
    dn <- dimnames(prob)
    p <- prob
    dim(p) <- c(prod(dim(prob)[1:2]), dim(prob)[3])
    p <- p %*% contrasts
    dim(p) <- dim(prob)
    prob <- p
    dimnames(prob) <- list(dn[[1]], dn[[2]], colnames(contrasts))

    posnames <- dimnames(prob2[[1]])[[3]]

    # no covariates
    expected <- eff_via_lm(prob, phe)
    coef <- scan1coef(prob2, phe, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d")))
    coef <- scan1coef(prob2, phe, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "d")))

    # no covariates, weighted
    expected <- eff_via_lm(prob, phe, weights=weights)
    coef <- scan1coef(prob2, phe, weights=weights, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d")))
    coef <- scan1coef(prob2, phe, weights=weights, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "d")))

    # one add've covariate
    expected <- eff_via_lm(prob, phe, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "d", "sex")))

    # one add've covariate, weighted
    expected <- eff_via_lm(prob, phe, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, weights=weights, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, weights=weights, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "d", "sex")))

    # one int've covariate
    expected <- eff_via_lm(prob, phe, covar, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex", "a:sex", "d:sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex", "a:sex", "d:sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "d", "sex", "a:sex", "d:sex")))

    # one int've covariate, weighted
    expected <- eff_via_lm(prob, phe, covar, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex", "a:sex", "d:sex")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex", "a:sex", "d:sex")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "d", "sex", "a:sex", "d:sex")))

    # two int've covariates
    covar <- cbind(covar, another=rnorm(n_ind(iron)))
    expected <- eff_via_lm(prob, phe, covar, covar)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, contrasts=contrasts)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex", "another",
                                                              "a:sex", "d:sex", "a:another", "d:another")))
    expect_equivalent(coef, expected)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex", "another",
                                                              "a:sex", "d:sex", "a:another", "d:another")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "d", "sex", "another",
                                                              "a:sex", "d:sex", "a:another", "d:another")))

    # two int've covariates, weighted
    expected <- eff_via_lm(prob, phe, covar, covar, weights=weights)
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, contrasts=contrasts)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames,
                                      c("mu", "a", "d", "sex", "another", "a:sex", "d:sex",
                                        "a:another", "d:another")))
    coef <- scan1coef(prob2, phe, addcovar=covar, intcovar=covar, weights=weights, contrasts=contrasts, se=TRUE)
    expect_equivalent(coef, expected)
    expect_equal(dimnames(coef), list(posnames, c("mu", "a", "d", "sex", "another",
                                                              "a:sex", "d:sex", "a:another", "d:another")))
    expect_equivalent(attr(coef, "SE"), attr(expected, "SE"))
    expect_equal(dimnames(attr(coef, "SE")), list(posnames, c("mu", "a", "d", "sex", "another",
                                                              "a:sex", "d:sex", "a:another", "d:another")))

})



test_that("scan1coef deals with mismatching individuals", {

    skip_on_cran()

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    probs <- probs[,"3"]
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    phe <- iron$pheno[,2,drop=FALSE]

    ind <- c(1:50, 101:150)
    expected <- scan1coef(probs[ind,], phe[ind,,drop=FALSE], addcovar=X[ind])
    expect_equal(scan1coef(probs[ind,], phe, addcovar=X), expected)
    expect_equal(scan1coef(probs, phe[ind,,drop=FALSE], addcovar=X), expected)
    expect_equal(scan1coef(probs, phe, addcovar=X[ind]), expected)

})

context("plot_peaks")

test_that("plot_peaks works", {

    skip_if(isnt_karl(), "plot tests only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    pheno <- iron$pheno
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

    # find peaks above lod=3.5 (and calculate 1.5-LOD support intervals)
    peaks <- find_peaks(out, map, threshold=3.5, drop=1.5)

    test_plot_peaks <- function() plot_peaks(peaks, map)

    expect_doppelganger("plot_peaks", test_plot_peaks)

    peaks <- find_peaks(out, map, threshold=3.5)
    test_plot_peaks_noci <- function() plot_peaks(peaks, map)
    expect_doppelganger("plot_peaks_noci", test_plot_peaks_noci)

})

context("scan1 permutations with binary phenotype")

iron <- read_cross2(system.file("extdata","iron.zip", package="qtl2"))
iron <- iron[,c(18,19,"X")]
map <- insert_pseudomarkers(iron$gmap, step=1)
pr <- calc_genoprob(iron, map, err=0.002)
kinship <- calc_kinship(pr)
kinship_loco <- calc_kinship(pr, "loco")

pheno1 <- apply(iron$pheno, 2, function(a) as.numeric(a > quantile(a, 0.7)))
rownames(pheno1) <- rownames(iron$pheno)
pheno2 <- pheno1; pheno2[5,1] <- NA

Xcovar <- get_x_covar(iron)
sex <- as.numeric(iron$covar$sex=="m")
names(sex) <- rownames(iron$covar)
perm_strata <- mat2strata(Xcovar)

test_that("scan1 permutations work (regression test)", {

    seed <- 3025685
    RNGkind("L'Ecuyer-CMRG")

    # no covariates or missing data
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, model="binary", n_perm=3)
    expected <- cbind(liver= c(1.9400869101173725539, 1.3826531402385029423, 2.32771714636574245105),
                      spleen=c(1.7406845630378597889, 1.0816176320165595826, 0.86215751014691477394))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)

    skip_on_cran()

    # no covariates or missing data
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, model="binary", n_perm=3, perm_strata=perm_strata)
    expected <- cbind(liver= c(7.7800367429250911755, 8.2464978987470800575, 8.4332725697338304371),
                      spleen=c(2.6682023473596672147, 3.1865109293844255944, 3.9334182919580200632))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)

    # sex and X-chr covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, model="binary", addcovar=sex, Xcovar=Xcovar, n_perm=3)
    expected <- cbind(liver= c(1.0627090966702610331, 1.3019431062016479927, 1.4887177771883841615),
                      spleen=c(1.2907987947091044134, 1.0942134968225047942, 1.5908715555718089263))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, model="binary", addcovar=sex, Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected)

    # sex and X-chr covariates, plus sex interactive
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, model="binary", addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3)
    expected <- cbind(liver= c(1.6988140471307104917, 3.3492519382653398452, 1.4887177771883841615),
                      spleen=c(1.3563064772388884194, 2.3361493956548571305, 1.5908715555718089263))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, model="binary", addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected)

    # sex and additive covariates + X-sp perms
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, model="binary", addcovar=sex, Xcovar=Xcovar, n_perm=3,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map))
    expected <- list(A=cbind(liver= c(1.0627090966702610331,  0.54091905189068256732, 0.51646113492044776194),
                             spleen=c(1.2907987947091044134,  1.09421349682250479418, 0.97567385090644620504)),
                     X=cbind(liver= c(0.92947674285197479094, 0.82005313512853206248, 0.41809257354587714417, 0.51654137988325032893,
                                      0.37354377722229514802, 1.8944354102049061339,  0.64614008414334023200),
                             spleen=c(1.86312726596119659916, 0.42851226191818625466, 1.82692521628932524891, 1.54255046971590559224,
                                      1.17331067477128669907, 1.9748233131450518840,  0.14820359491564261134)))
    L <- c(A=61.200000000000002842,X=28.399999999999998579)
    attr(L, "is_x_chr") <- c(A=FALSE, X=TRUE)
    attr(expected, "chr_lengths") <- L
    class(expected) <- c("scan1perm", "list")
    expect_equal(operm, expected)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, model="binary", addcovar=sex, Xcovar=Xcovar, n_perm=3,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map), perm_strata=perm_strata)
    expect_equal(operm, expected)

    # one missing phenotype, no covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, model="binary", n_perm=3)
    expected <- cbind(liver= c(1.9396621996084775219, 1.3860820684821959503, 2.39451070079611838537),
                      spleen=c(1.7406845630378597889, 1.0816176320165595826, 0.86215751014691477394))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)

    # one missing phenotype, sex and X-chr covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, model="binary", addcovar=sex, Xcovar=Xcovar, n_perm=3)
    expected <- cbind(liver= c(1.0607794866188555716, 1.2669252823191214929, 1.4888030146012454225),
                      spleen=c(1.2907987947091044134, 1.0942134968225047942, 1.5908715555718089263))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, model="binary", addcovar=sex, Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected)

    # one missing phenotype, sex and X-chr covariates, plus sex interactive
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, model="binary", addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3)
    expected <- cbind(liver= c(1.6885642297968530556, 3.4770987617121420499, 1.4888030146012454225),
                      spleen=c(1.3563064772388884194, 2.3361493956548571305, 1.5908715555718089263))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, model="binary", addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected)

    # one missing phenotype, sex and X-chr covariates, X-sp perms
    set.seed(seed)
    perm_strata <- rep(1, nrow(pheno2)) # avoid the stratified permutations
    names(perm_strata) <- rownames(pheno2)
    operm <- scan1perm(pr, pheno2, model="binary", addcovar=sex, Xcovar=Xcovar, n_perm=3,
                       perm_Xsp=TRUE, chr_lengths=chr_lengths(map),
                       perm_strata=perm_strata)
    expected <- list(A=cbind(liver= c(1.7660810516288876215, 0.40146200349977334554, 0.41346547186287807563),
                             spleen=c(1.5238369101838173947, 1.21467164163189522696, 0.79071777746629834382)),
                     X=cbind(liver= c(1.1856908571482307480, 1.6393912628303013435,  1.2664516688820839363, 0.72714128449122483744,
                                      2.7436059094899150068, 1.3434021611743531821,  0.49535669351789124448),
                             spleen=c(1.4926905975660105241, 1.2193200324768156406,  3.1386909220195633452, 0.56076286804034225497,
                                      1.4488076063209831545, 0.3586987681249098614,  1.20528954216223382900)))
    L <- c(A=61.200000000000002842,X=28.399999999999998579)
    attr(L, "is_x_chr") <- c(A=FALSE, X=TRUE)
    attr(expected, "chr_lengths") <- L
    class(expected) <- c("scan1perm", "list")
    expect_equal(operm, expected)

})

context("get X covariates")

test_that("get_x_covar for riself", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))

    expect_equal(get_x_covar(grav2), NULL)

})

test_that("get_x_covar for intercross", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    n <- n_ind(iron)

    expected <- matrix(0.0, ncol=2, nrow=n)
    expected[!iron$is_female, 1] <- 1.0
    expected[iron$is_female & iron$cross_info==1, 2] <- 1.0
    dimnames(expected) <- list(names(iron$is_female), c("sex", "direction"))

    expect_equal(get_x_covar(iron), expected)

    # all male
    tmp <- iron
    tmp$is_female <- rep(FALSE, length(tmp$is_female))
    expect_equal(get_x_covar(tmp), NULL)

    # all female
    tmp <- iron
    tmp$is_female <- rep(TRUE, length(tmp$is_female))
    expected <- matrix(0.0, ncol=1, nrow=n)
    expected[iron$cross_info==1, 1] <- 1.0
    dimnames(expected) <- list(names(tmp$is_female), "direction")
    expect_equal(get_x_covar(tmp), expected)

    # all female, forward direction
    tmp <- iron
    tmp$is_female <- rep(TRUE, length(tmp$is_female))
    tmp$cross_info[tmp$cross_info==1] <- 0
    expect_equal(get_x_covar(tmp), NULL)

    # all female, reverse direction
    tmp <- iron
    tmp$is_female <- rep(TRUE, length(tmp$is_female))
    tmp$cross_info[tmp$cross_info==0] <- 1
    expect_equal(get_x_covar(tmp), NULL)

    # both sexes, one direction
    tmp <- iron
    tmp$cross_info[tmp$cross_info==1] <- 0
    expected <- matrix(0.0, ncol=1, nrow=n)
    expected[!iron$is_female] <- 1
    dimnames(expected) <- list(names(tmp$is_female), "sex")
    expect_equal(get_x_covar(tmp), expected)

})

context("scan1 permutations")

iron <- read_cross2(system.file("extdata","iron.zip", package="qtl2"))
iron <- iron[,c(18,19,"X")]
map <- insert_pseudomarkers(iron$gmap, step=1)
pr <- calc_genoprob(iron, map, err=0.002)
kinship <- calc_kinship(pr)
kinship_loco <- calc_kinship(pr, "loco")

pheno1 <- iron$pheno
pheno2 <- iron$pheno; pheno2[5,1] <- NA

Xcovar <- get_x_covar(iron)
sex <- as.numeric(iron$covar$sex=="m")
names(sex) <- rownames(iron$covar)
perm_strata <- mat2strata(Xcovar)

test_that("scan1 permutations work (regression test)", {

    seed <- 3025685
    RNGkind("L'Ecuyer-CMRG")

    # no covariates or missing data
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, n_perm=3)
    expected <- cbind(liver= c(1.3756640664429049536, 2.7572772758036521168, 0.52452198541287131661),
                      spleen=c(1.5153608426190210423, 1.4532027591584668613, 2.86857866542299611012))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)

    # no covariates or missing data
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, n_perm=3, perm_strata=perm_strata)
    expected <- cbind(liver= c(13.541670362333418254, 14.7024286141676103767, 14.1453256570137018144),
                      spleen=c( 5.930376703877245248,  7.4039049500925262493,  5.5851374704542742222))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)

    # sex and X-chr covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, addcovar=sex, Xcovar=Xcovar, n_perm=3)
    expected <- cbind(liver= c(1.56882977009222202014, 1.0308336167670439920, 1.6771296752179409850),
                      spleen=c(0.88478150549593514995, 1.1297760401050815915, 2.8456095237703333822))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, addcovar=sex, Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected)

    # sex and X-chr covariates, plus sex interactive
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3)
    expected <- cbind(liver= c(2.0685972274435684426, 1.5683121279062959275, 1.6771296752180671064),
                      spleen=c(1.0328759662065980507, 1.5091023963920129347, 2.8456095237703333822))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected)

    # sex and additive covariates + X-sp perms
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, addcovar=sex, Xcovar=Xcovar, n_perm=3,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map))
    expected <- list(A=cbind(liver= c(1.56882977009222202014, 0.17654731701161452406, 0.67967768984640919427),
                             spleen=c(0.88478150549593514995, 1.12977604010508159149, 0.95835278347981578406)),
                     X=cbind(liver= c(0.65635525818365536566, 0.59711143944156575003, 0.49516910044845729999, 0.70305870212073351411,
                                      0.44456114810071056809, 1.17745924272541380160, 0.69147359544272646303),
                             spleen=c(0.71360878105542902006, 0.54087014463646632123, 1.56708158211767667467, 1.21950357033675516050,
                                      1.49396169005535739416, 2.8719171298632524270, 1.08820611768204855707)))
    L <- c(A=61.200000000000002842,X=28.399999999999998579)
    attr(L, "is_x_chr") <- c(A=FALSE, X=TRUE)
    attr(expected, "chr_lengths") <- L
    class(expected) <- c("scan1perm", "list")
    expect_equal(operm, expected)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, addcovar=sex, Xcovar=Xcovar, n_perm=3,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map), perm_strata=perm_strata)
    expect_equal(operm, expected)

    # one missing phenotype, no covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, n_perm=3)
    expected <- cbind(liver= c(1.4246052689684201020, 1.8035976886303082267, 1.295827488548965345),
                      spleen=c(2.1450871714236026122, 1.9160146558007777884, 1.887126247743548646))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)

    # one missing phenotype, sex and X-chr covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, addcovar=sex, Xcovar=Xcovar, n_perm=3)
    expected <- cbind(liver= c(1.56455340995498248802, 1.0121930261312361843, 1.6731999914352853054),
                      spleen=c(0.88478150549593514995, 1.1297760401050815915, 2.8456095237703333822))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, addcovar=sex, Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected)

    # one missing phenotype, sex and X-chr covariates, plus sex interactive
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3)
    expected <- cbind(liver= c(2.0627670956684416304, 1.6261084401604977145, 1.6731999914352853054),
                      spleen=c(1.0328759662065980507, 1.5091023963920129347, 2.8456095237703333822))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected)
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected)

    # one missing phenotype, sex and X-chr covariates, X-sp perms
    set.seed(seed)
    perm_strata <- rep(1, nrow(pheno2)) # avoid the stratified permutations
    names(perm_strata) <- rownames(pheno2)
    operm <- scan1perm(pr, pheno2, addcovar=sex, Xcovar=Xcovar, n_perm=3,
                       perm_Xsp=TRUE, chr_lengths=chr_lengths(map),
                       perm_strata=perm_strata)
    expected <- list(A=cbind(liver= c(1.18430970191987361420, 0.85274796222984994287,  0.76981903233409942899),
                             spleen=c(1.96074450827722834840, 2.47749749792303752827,  1.20235911795879424346)),
                     X=cbind(liver= c(1.29867858019373416670, 0.25258741397366168968,  2.22594902965746355150, 0.69668535735455083824,
                                      1.84525108476481491948, 1.97947969062733486467, -0.07964898968422540193),
                             spleen=c(1.39377789098622173470, 1.24497337934968221873,  0.94154476905226225369, 0.50568976179249958136,
                                      0.70117350389718069437, 0.46582839190887348479,  0.98324516512323434370)))
    L <- c(A=61.200000000000002842,X=28.399999999999998579)
    attr(L, "is_x_chr") <- c(A=FALSE, X=TRUE)
    attr(expected, "chr_lengths") <- L
    class(expected) <- c("scan1perm", "list")
    expect_equal(operm, expected)

})


test_that("scan1 permutations work with single kinship matrix (regression test)", {

    skip_if(isnt_karl(), "this test only run locally")

    seed <- 3025685
    RNGkind("L'Ecuyer-CMRG")

    # no covariates or missing data
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship, n_perm=3)
    expected <- cbind(liver= c(1.5850327051161108294, 1.6225465841443282855, 1.1047230815073763033),
                      spleen=c(2.1929283392327243440, 1.9397727636298436327, 1.8698391778175107447))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=2e-7)

    # no covariates or missing data
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship, n_perm=3, perm_strata=perm_strata)
    expected <- cbind(liver= c(14.6442198530459659622, 14.6817005229127524046, 15.6345815802552241536),
                      spleen=c( 5.5167589745319025596,  5.6328648432420660441,  8.0494249754664739527))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=2e-7)

    # sex and X-chr covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship, addcovar=sex, Xcovar=Xcovar, n_perm=3)
    expected <- cbind(liver= c(1.29766020287924632726, 1.0197730259964858934, 2.0112145151147022837),
                      spleen=c(0.90427196894971517693, 1.0214860518121018362, 2.8832996501350738328))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=2e-7)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship, addcovar=sex, Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected, tolerance=2e-7)

    # sex and X-chr covariates, plus sex interactive
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3)
    expected <- cbind(liver= c(1.8645265744331560587, 1.6155547860418919548, 2.0112145151147022837),
                      spleen=c(1.0356350517366044173, 1.4490075143563088123, 2.8832996501350738328))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=2e-7)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected, tolerance=2e-7)

    # sex and additive covariates + X-sp perms
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship, addcovar=sex, Xcovar=Xcovar, n_perm=3,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map))
    expected <- list(A=cbind(liver= c(1.29766020287924632726, 0.14126238719321179693, 0.77580567004437706036),
                             spleen=c(0.90427196894971517693, 1.02148605181210183623, 0.93168604007222088903)),
                     X=cbind(liver= c(0.50034270818987047758, 0.76231900452964485027, 0.6011447003868983785, 0.65860596461113896094,
                                      0.43605696449176051255, 1.26565994163476069900, 0.65084718581631106904),
                             spleen=c(0.73143027950382133451, 0.53598997526824987414, 1.5569757625108506804, 1.21787150923781783973,
                                      1.49387241817333116245, 2.83703932517150425600, 1.07984190845146876825)))
    L <- c(A=61.200000000000002842,X=28.399999999999998579)
    attr(L, "is_x_chr") <- c(A=FALSE, X=TRUE)
    attr(expected, "chr_lengths") <- L
    class(expected) <- c("scan1perm", "list")
    expect_equal(operm, expected, tolerance=3e-7)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship, addcovar=sex, Xcovar=Xcovar, n_perm=3,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map), perm_strata=perm_strata)
    expect_equal(operm, expected, tolerance=3e-7)

    # one missing phenotype, no covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, kinship, n_perm=3)
    expected <- cbind(liver= c(1.5863339748532325757, 1.6385708572073809375, 1.1068373425454520742),
                      spleen=c(2.1929283392327243440, 1.9397727636298436327, 1.8698391778175107447))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=2e-7)

    # one missing phenotype, sex and X-chr covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, kinship, addcovar=sex, Xcovar=Xcovar, n_perm=3)
    expected <- cbind(liver= c(1.29518706542233608126, 0.99913640281598758985, 2.0054677281986590387),
                      spleen=c(0.90427196894971517693, 1.02148605181210183623, 2.8832996501350738328))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=2e-7)
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, kinship, addcovar=sex, Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected, tolerance=2e-7)

    # one missing phenotype, sex and X-chr covariates, plus sex interactive
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, kinship, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3)
    expected <- cbind(liver= c(1.8604662282096269266, 1.6877912266523966700, 2.0054677281986590387),
                      spleen=c(1.0356350517366044173, 1.4490075143563088123, 2.8832996501350738328))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=1e-7)
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, kinship, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected, tolerance=1e-7)

    # one missing phenotype, sex and X-chr covariates, X-sp perms
    set.seed(seed)
    perm_strata <- rep(1, nrow(pheno2)) # avoid the stratified permutations
    names(perm_strata) <- rownames(pheno2)
    operm <- scan1perm(pr, pheno2, kinship, addcovar=sex, Xcovar=Xcovar, n_perm=3,
                       perm_Xsp=TRUE, chr_lengths=chr_lengths(map),
                       perm_strata=perm_strata)
    expected <- list(A=cbind(liver= c(1.37030487476505635770, 0.58183562848782044430, 0.79833486499453443219),
                             spleen=c(2.01478095022342573730, 2.52716001067115758620, 1.26929860408445516207)),
                     X=cbind(liver= c(1.63673834214325419900, 0.67953625118790006443, 2.073666671975550102000, 1.06998409582742026736,
                                      1.89434763183230137074, 2.14236446791721579785, 0.086724900384358594163),
                             spleen=c(1.56170484328174996590, 1.56288576386423483378, 1.116409713761464583800, 0.78539826876917206988,
                                      0.82004164028519510588, 0.74049399535694393482, 1.310035745047248845196)))
    L <- c(A=61.200000000000002842,X=28.399999999999998579)
    attr(L, "is_x_chr") <- c(A=FALSE, X=TRUE)
    attr(expected, "chr_lengths") <- L
    class(expected) <- c("scan1perm", "list")
    expect_equal(operm, expected, tolerance=4e-7)

})


test_that("scan1 permutations work with LOCO kinship matrix (regression test)", {

    skip_if(isnt_karl(), "this test only run locally")

    seed <- 3025685
    RNGkind("L'Ecuyer-CMRG")

    # no covariates or missing data
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship_loco, n_perm=3)
    expected <- cbind(liver= c(1.4233905282897871825, 1.6883979222096348050, 1.1105319041474899233),
                      spleen=c(2.1849905121197314983, 1.9112972630703948251, 1.8871262477436099303))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=1e-7)

    # no covariates or missing data
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship_loco, n_perm=3, perm_strata=perm_strata)
    expected <- cbind(liver= c(15.0498110236199185152, 15.1559669590417573914, 16.1619790013112449856),
                      spleen=c( 5.6528544325570964091,  5.7416707117225200818,  8.1270354686771071329))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=2e-7)

    # sex and X-chr covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship_loco, addcovar=sex, Xcovar=Xcovar, n_perm=3)
    expected <- cbind(liver= c(1.56882977009233215426, 1.0513515452269666106, 2.0908041328949908966),
                      spleen=c(0.91750143650706161846, 1.0096069617524108253, 2.8456095237817189414))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=2e-7)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship_loco, addcovar=sex, Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected, tolerance=2e-7)

    # sex and X-chr covariates, plus sex interactive
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship_loco, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3)
    expected <- cbind(liver= c(2.0685972274436985607, 1.5532309407813880142, 2.0908041328949908966),
                      spleen=c(1.0453315747253293377, 1.4393283407079127123, 2.8456095237817189414))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=2e-7)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship_loco, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected, tolerance=2e-7)

    # sex and additive covariates + X-sp perms
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship_loco, addcovar=sex, Xcovar=Xcovar, n_perm=3,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map))
    expected <- list(A=cbind(liver= c(1.56882977009233215426, 0.15408564377888087082, 0.67967768984653986752),
                             spleen=c(0.91750143650706161846, 1.00960696175241082528, 0.91230958640896986367)),
                     X=cbind(liver= c(0.47525078870451631374, 0.81648974630841897326, 0.55021896067082298742, 0.68285545241312051168,
                                      0.46054636746004662395, 1.28223568383226194100, 0.66115351693360990826),
                             spleen=c(0.71360878106659053621, 0.54087014464766181021, 1.56708158212895898309, 1.21950357034804057754,
                                      1.49396169006666879042, 2.87191712987437774980, 1.08820611769328978724)))
    L <- c(A=61.200000000000002842,X=28.399999999999998579)
    attr(L, "is_x_chr") <- c(A=FALSE, X=TRUE)
    attr(expected, "chr_lengths") <- L
    class(expected) <- c("scan1perm", "list")
    expect_equal(operm, expected, tolerance=3e-7)
    set.seed(seed)
    operm <- scan1perm(pr, pheno1, kinship_loco, addcovar=sex, Xcovar=Xcovar, n_perm=3,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map), perm_strata=perm_strata)
    expect_equal(operm, expected, tolerance=3e-7)

    # one missing phenotype, no covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, kinship_loco, n_perm=3)
    expected <- cbind(liver= c(1.4246052689684871595, 1.7056415036570125032, 1.1168979618114607266),
                      spleen=c(2.1849905121197314983, 1.9112972630703948251, 1.8871262477436099303))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=1e-7)

    # one missing phenotype, sex and X-chr covariates
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, kinship_loco, addcovar=sex, Xcovar=Xcovar, n_perm=3)
    expected <- cbind(liver= c(1.56455340995486880118, 1.0297698518383779920, 2.0850243087458522062),
                      spleen=c(0.91750143650706161846, 1.0096069617524108253, 2.8456095237817189414))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=2e-7)
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, kinship_loco, addcovar=sex, Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected, tolerance=2e-7)

    # one missing phenotype, sex and X-chr covariates, plus sex interactive
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, kinship_loco, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3)
    expected <- cbind(liver= c(2.0627670956682830905, 1.6203949273894016070, 2.0850243087458522062),
                      spleen=c(1.0453315747253293377, 1.4393283407079127123, 2.8456095237817189414))
    class(expected) <- c("scan1perm", "matrix")
    expect_equal(operm, expected, tolerance=1e-7)
    set.seed(seed)
    operm <- scan1perm(pr, pheno2, kinship_loco, addcovar=sex, Xcovar=Xcovar, intcovar=sex, n_perm=3, perm_strata=perm_strata)
    expect_equal(operm, expected, tolerance=1e-7)

    # one missing phenotype, sex and X-chr covariates, X-sp perms
    set.seed(seed)
    perm_strata <- rep(1, nrow(pheno2)) # avoid the stratified permutations
    names(perm_strata) <- rownames(pheno2)
    operm <- scan1perm(pr, pheno2, kinship_loco, addcovar=sex, Xcovar=Xcovar, n_perm=3,
                       perm_Xsp=TRUE, chr_lengths=chr_lengths(map),
                       perm_strata=perm_strata)
    expected <- list(A=cbind(liver= c(1.18430970191976991930, 0.85274796222960802528,  0.76981903233390269747),
                             spleen=c(1.97990260877063839470, 2.49362417238307543244,  1.25357633693210002157)),
                     X=cbind(liver= c(1.38737185599694012870, 0.40916780292085591642,  1.64260630582905675645, 0.73029802514640995703,
                                      1.48936109005232153457, 1.76554289216658433230, -0.19556097227382970849),
                             spleen=c(1.39377789099751669970, 1.24497337936097429711,  0.94154476906363604449, 0.50568976180393732101,
                                      0.70117350390854737974, 0.46582839192008501650,  0.98324516513467219436)))
    L <- c(A=61.200000000000002842,X=28.399999999999998579)
    attr(L, "is_x_chr") <- c(A=FALSE, X=TRUE)
    attr(expected, "chr_lengths") <- L
    class(expected) <- c("scan1perm", "list")
    expect_equal(operm, expected, tolerance=2e-7)

})

context("reduced rank covariates")

test_that("scan1 etc work with reduced-rank covariates", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(1,12,"X")]

    Xcovar <- get_x_covar(iron)
    set.seed(3706216)
    n_batch <- 4
    batch <- sample(1:n_batch, n_ind(iron), replace=TRUE)
    X <- matrix(0, nrow=n_ind(iron), ncol=n_batch-1)
    dimnames(X) <- list(rownames(Xcovar), paste0("batch", 2:n_batch))
    for(i in 2:n_batch) X[batch==i,i-1] <- 1

    pr <- calc_genoprob(iron, error_prob=0.01)

    phe <- iron$pheno
    phe[batch==n_batch,1] <- NA
    phe[batch==1,2] <- NA

    # scan1 with no kinship
    out <- scan1(pr, phe, Xcovar=Xcovar, addcovar=X)
    expected <- cbind(scan1(pr, phe[!is.na(phe[,1]),1,drop=FALSE], Xcovar=Xcovar, addcovar=X[,-3]),
                      scan1(pr, phe[!is.na(phe[,2]),2,drop=FALSE], Xcovar=Xcovar, addcovar=X[,-3]))
    expect_equal(out, expected)

    skip_on_cran()

    # scan1 with kinship
    k <- calc_kinship(pr)
    out_k <- scan1(pr, phe, k, Xcovar=Xcovar, addcovar=X)
    expected_k <- cbind(scan1(pr, phe[!is.na(phe[,1]),1,drop=FALSE], k, Xcovar=Xcovar, addcovar=X[,-3]),
                        scan1(pr, phe[!is.na(phe[,2]),2,drop=FALSE], k, Xcovar=Xcovar, addcovar=X[,-3]))
    expect_equal(out_k, expected_k, tol=5e-6)

    # scan1coef with no kinship
    for(phecol in 1:2) {
        for(chr in names(pr)) {
            co <- scan1coef(pr[,chr], phe[,phecol,drop=FALSE], addcovar=X)
            Xcol2keep <- colnames(X)[colnames(X) %in% colnames(co)]
            expected <- scan1coef(pr[,chr], phe[!is.na(phe[,phecol]),phecol,drop=FALSE],
                                  addcovar=X[,Xcol2keep])
            expect_equal(co, expected)
        }
    }

    # scan1coef with kinship
    for(phecol in 1:2) {
        for(chr in names(pr)) {
            co <- scan1coef(pr[,chr], phe[,phecol,drop=FALSE], k, addcovar=X)
            Xcol2keep <- colnames(X)[colnames(X) %in% colnames(co)]
            expected <- scan1coef(pr[,chr], phe[!is.na(phe[,phecol]),phecol,drop=FALSE],
                                  k, addcovar=X[,Xcol2keep])
            expect_equal(co, expected)
        }
    }


    # scan1blup with no kinship
    for(phecol in 1:2) {
        for(chr in names(pr)) {
            co <- scan1blup(pr[,chr], phe[,phecol,drop=FALSE], addcovar=X)
            Xcol2keep <- colnames(X)[colnames(X) %in% colnames(co)]
            expected <- scan1blup(pr[,chr], phe[!is.na(phe[,phecol]),phecol,drop=FALSE],
                                  addcovar=X[,Xcol2keep])
            expect_equal(co, expected)
        }
    }

    # scan1blup with kinship
    for(phecol in 1:2) {
        for(chr in names(pr)) {
            co <- scan1blup(pr[,chr], phe[,phecol,drop=FALSE], k, addcovar=X)
            Xcol2keep <- colnames(X)[colnames(X) %in% colnames(co)]
            expected <- scan1blup(pr[,chr], phe[!is.na(phe[,phecol]),phecol,drop=FALSE],
                                  k, addcovar=X[,Xcol2keep])
            expect_equal(co, expected)
        }
    }

})

context("summary scan1perm")

test_that("summary_scan1perm works", {

    seed <- 9896433
    RNGkind("Mersenne-Twister")

    # read data
    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(18,19,"X")]

    # insert pseudomarkers into map
    map <- insert_pseudomarkers(iron$gmap, step=1)

    # calculate genotype probabilities
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    # grab phenotypes and covariates; ensure that covariates have names attribute
    pheno <- iron$pheno
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    # not X-chr-specific
    set.seed(seed)
    operm <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar, n_perm=3)

    result <- summary(operm)
    expected <- rbind("0.05" = c(liver=1.5480532515070430932,spleen=1.1945482736887691466))
    attr(expected, "n_perm") <- rbind(c(liver=3, spleen=3))
    class(expected) <- c("summary.scan1perm", "matrix")
    expect_equal(result, expected)

    # two thresholds
    result <- summary(operm, c(0.05, 0.2))
    expected <- rbind(expected, "0.2"=c(liver=1.5039885621071871213, spleen=1.0497165924333802245))
    attr(expected, "n_perm") <- rbind(c(liver=3, spleen=3))
    class(expected) <- c("summary.scan1perm", "matrix")
    expect_equal(result, expected)

    # X-chr-specific
    set.seed(seed)
    operm <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar,
                       n_perm=3, perm_Xsp=TRUE,
                       chr_lengths=chr_lengths(iron$gmap))
    result <- summary(operm)
    expected <- list(A=rbind("0.05" = c(liver=1.5244058138572063044, spleen=0.63833353205793441631)),
                     X=rbind("0.05" = c(liver=1.7400163974820828106, spleen= 2.8361536028418536937)))
    attr(expected, "n_perm") <- rbind(A=c(liver=3, spleen=3), X=c(liver=7, spleen=7))
    class(expected) <- c("summary.scan1perm", "list")
    expect_equal(result, expected)

    # two thresholds
    result <- summary(operm, c(0.05, 0.2))
    expected$A <- rbind(expected$A, "0.2"=c(liver=1.4053300973638913618, spleen=0.61906591053182846718))
    expected$X <- rbind(expected$X, "0.2"=c(liver=1.6152559270167379246, spleen=2.2975884236147399164))
    expect_equal(result, expected)

})

context("align_scan1_map")

test_that("align_scan1_map works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    # insert pseudomarkers into map
    map <- insert_pseudomarkers(iron$gmap, step=1)

    # calculate genotype probabilities
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    # grab phenotypes and covariates; ensure that covariates have names attribute
    pheno <- iron$pheno[,1]
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)

    Xcovar <- get_x_covar(iron)

    # genome scan
    out <- scan1(probs, iron$pheno, Xcovar=Xcovar, addcovar=covar)

    # calculate coefficients for chromosome 7
    coef <- scan1coef(probs[,7], pheno, addcovar=covar)

    # not subseted
    result <- align_scan1_map(out, map)
    expect_equal(result$scan1, out)
    expect_equal(result$map, map)

    # chr 7 for map
    result <- align_scan1_map(out, map[7])
    expect_equal( result$scan1, subset(out, map, chr="7") )
    expect_equal( result$map, map[7] )

    # chr 7 for scan
    result <- align_scan1_map(subset(out,map,chr="7"), map)
    expect_equal( result$scan1, subset(out, map, chr="7") )
    expected_map <- map[7]
    attr(expected_map, "is_x_chr") <- attr(map, "is_x_chr")[7]
    expect_equal( result$map, expected_map )

    # coefficients
    result <- align_scan1_map(coef, map)
    expect_equal( result$scan1, coef )
    expected_map <- map[7]
    attr(expected_map, "is_x_chr") <- attr(map, "is_x_chr")[7]
    expect_equal( result$map, expected_map )

    # shuffled output
    out_attr <- attributes(out)
    out_shuffled <- out[sample(1:nrow(out)),,drop=FALSE]
    for(a in c("sample_size", "class"))
        attr(out_shuffled, a) <- out_attr[[a]]
    result <- align_scan1_map(out_shuffled, map)
    expect_equal( result$scan1, out )
    expect_equal( result$map, map )

})

context("rbind/cbind scan1perm")

# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
iron <- iron[,c(18,19,"X")]

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# strata for permutations
perm_strata <- mat2strata(Xcovar)

test_that("rbind.scan1perm works", {

    # not X-chr-specific
    operm1 <- scan1perm(probs, pheno, Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    operm2 <- scan1perm(probs, pheno, Xcovar=Xcovar, n_perm=4, perm_strata=perm_strata)
    operm <- rbind(operm1, operm2)
    expected <- rbind(unclass(operm1), unclass(operm2))
    class(expected) <- class(operm1)
    expect_equal(operm, expected)

    # test with three
    operm <- rbind(operm1, operm2, operm1)
    expected <- rbind(unclass(operm1), unclass(operm2), unclass(operm1))
    class(expected) <- class(operm1)
    expect_equal(operm, expected)

    # test with one
    expect_equal(rbind(operm1), operm1)

    # X-chr-specific
    operm1 <- scan1perm(probs, pheno, Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map))
    operm2 <- scan1perm(probs, pheno, Xcovar=Xcovar, n_perm=4, perm_strata=perm_strata,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map))
    operm <- rbind(operm1, operm2)
    expected <- list(A=rbind(operm1$A, operm2$A),
                     X=rbind(operm1$X, operm2$X))
    attr(expected, "chr_lengths") <- attr(operm1, "chr_lengths")
    class(expected) <- class(operm1)
    expect_equal(operm, expected)

    # test with three
    operm <- rbind(operm1, operm2, operm1)
    expected <- list(A=rbind(operm1$A, operm2$A, operm1$A),
                     X=rbind(operm1$X, operm2$X, operm1$X))
    attr(expected, "chr_lengths") <- attr(operm1, "chr_lengths")
    class(expected) <- class(operm1)
    expect_equal(operm, expected)

    # test with one
    expect_equal(rbind(operm1), operm1)

})


test_that("cbind.scan1perm works", {

    # not X-chr-specific, same number of permutations
    operm1 <- scan1perm(probs, pheno[,1,drop=FALSE], Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    operm2 <- scan1perm(probs, pheno[,2,drop=FALSE], Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata)
    operm <- cbind(operm1, operm2)
    expected <- cbind(unclass(operm1), unclass(operm2))
    class(expected) <- class(operm1)
    expect_equal(operm, expected)

    # test with three
    operm <- cbind(operm1, operm2, operm1)
    expected <- cbind(unclass(operm1), unclass(operm2), unclass(operm1))
    class(expected) <- class(operm1)
    expect_equal(operm, expected)

    # test with one
    expect_equal(cbind(operm1), operm1)



    # different numbers of permutations
    operm2b <- operm2[1:2,,drop=FALSE]
    operm <- cbind(operm1, operm2b)
    expected <- cbind(operm1, operm2)
    expected[3,2] <- NA
    expect_equal(operm, expected)



    # X-chr-specific
    operm1 <- scan1perm(probs, pheno[,1,drop=FALSE], Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map))
    operm2 <- scan1perm(probs, pheno[,2,drop=FALSE], Xcovar=Xcovar, n_perm=3, perm_strata=perm_strata,
                        perm_Xsp=TRUE, chr_lengths=chr_lengths(map))
    operm <- cbind(operm1, operm2)
    expected <- list(A=cbind(operm1$A, operm2$A),
                     X=cbind(operm1$X, operm2$X))
    attr(expected, "chr_lengths") <- attr(operm1, "chr_lengths")
    class(expected) <- class(operm1)
    expect_equal(operm, expected)

    # test with three
    operm <- cbind(operm1, operm2, operm1)
    expected <- list(A=cbind(operm1$A, operm2$A, operm1$A),
                     X=cbind(operm1$X, operm2$X, operm1$X))
    attr(expected, "chr_lengths") <- attr(operm1, "chr_lengths")
    class(expected) <- class(operm1)
    expect_equal(operm, expected)

    # test with one
    expect_equal(cbind(operm1), operm1)



    # different numbers of permutations
    operm2b <- list(A=operm2$A[1:2,,drop=FALSE],
                    X=operm2$X[1:5,,drop=FALSE])
    attr(operm2b, "chr_lengths") <- attr(operm2, "chr_lengths")
    operm <- cbind(operm1, operm2b)
    expected <- cbind(operm1, operm2)
    expected$A[3,2] <- NA
    expected$X[6:7,2] <- NA
    expect_equal(operm, expected)

})

context("plot_scan1")

test_that("plot_scan1 works", {

    skip_if(isnt_karl(), "plot tests only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    pheno <- iron$pheno
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
    ylim <- c(0, maxlod(out)*1.02) # need to strip class to get overall max LOD
    chr <- c(2,7,8,9,15,16)

    test_plot_scan1 <- function() {
        plot(out, map, chr=chr, ylim=ylim)
        plot(out, map, lodcolumn=2, chr=chr, col="violetred", add=TRUE)
        legend("topleft", lwd=2, col=c("darkslateblue", "violetred"), colnames(out),
               bg="gray90") }

    expect_doppelganger("plot_scan1", test_plot_scan1)


    # single chromosome
    test_plot_scan1_onechr <- function() plot(out, map, chr=2)
    expect_doppelganger("plot_scan1_onechr", test_plot_scan1_onechr)

})

context("genome scan by scan1")

isx_from_map <-
    function(map)
{
    isx <- rep(FALSE, length(map))
    names(isx) <- names(map)
    isx["X"] <- TRUE

    isx
}


# calc lod scores via lm(), just one chromosome
lod_via_lm <-
    function(probs, pheno, addcovar=NULL, intcovar=NULL, weights=NULL, map)
{
    if(!is.matrix(pheno)) pheno <- as.matrix(pheno)
    if(is.null(colnames(pheno)))
        colnames(pheno) <- paste0("pheno", 1:ncol(pheno))

    d3 <- dim(probs)[3]
    n <- nrow(pheno)
    p <- ncol(pheno)

    result <- matrix(nrow=d3, ncol=p)

    addcovar <- cbind(rep(1, n), addcovar)
    if(is.null(weights)) wts <- rep(1, n)
    else wts <- weights

    sample_size <- rep(NA, ncol(pheno))
    names(sample_size) <- colnames(pheno)

    for(i in 1:p) {
        not_na <- !is.na(pheno[,i])
        if(!is.null(addcovar)) not_na <- not_na & complete.cases(addcovar)
        if(!is.null(intcovar)) not_na <- not_na & complete.cases(intcovar)

        rss0 <- sum(lm(pheno[,i] ~ -1 + addcovar, weights=wts)$resid^2*wts[not_na])
        for(j in 1:d3) {
            if(is.null(intcovar))
                rss1 <- sum(lm(pheno[,i] ~ -1 + addcovar + probs[,-1,j],
                               weights=wts)$resid^2*wts[not_na])
            else
                rss1 <- sum(lm(pheno[,i] ~ -1 + addcovar + probs[,-1,j]*intcovar,
                               weights=wts)$resid^2*wts[not_na])
            result[j,i] <- sum(not_na)/2 * (log10(rss0) - log10(rss1))
        }
        sample_size[i] <- sum(not_na)
    }

    dimnames(result) <- list(dimnames(probs)[[3]], colnames(pheno))

    attr(result, "sample_size") <- sample_size

    class(result) <- c("scan1", "matrix")
    result
}

# revise scanone results to be as expected from scan1
scanone2scan1 <-
    function(x, n=NULL, posnames=NULL, phenames=NULL, addcovar=NULL, intcovar=NULL, Xcovar=NULL, weights=FALSE)
{
    map <- split(x[,2], factor(x[,1], levels=unique(x[,1])))
    names(map) <- unique(x[,1])
    if(!is.null(posnames)) {
        mapn <- split(posnames, factor(x[,1], levels=unique(x[,1])))
        for(i in seq(along=map)) names(map[[i]]) <- mapn[[i]]
    }

    x <- as.matrix(x[,-(1:2), drop=FALSE])
    if(is.null(phenames)) phenames <- paste0("pheno", 1:ncol(x))
    dimnames(x) <- list(posnames, phenames)

    if(is.null(names(n))) names(n) <- phenames

    result <- x
    attr(result, "sample_size") <- n

    class(result) <- c("scan1", "matrix")
    result
}

convert_probs2qtl2 <-
    function(cross)
{
    result <- lapply(cross$geno, function(a) {
        pr <- aperm(a$prob, c(1,3,2))
        rownames(pr) <- paste(1:nrow(pr))
        pr })

    attr(result, "is_x_chr") <- isx_from_map(convert2cross2(cross)$gmap)
    attr(result, "crosstype") <- class(cross)[1]

    class(result) <- c("calc_genoprob", "list")
    result
}

# now finally to some tests
test_that("scan1 for backcross with one phenotype", {

    library(qtl)
    data(hyper)

    # scan by R/qtl
    hyper <- calc.genoprob(hyper, step=5)
    out <- scanone(hyper, method="hk")

    # inputs for R/qtl2
    pr <- convert_probs2qtl2(hyper)
    y <- hyper$pheno[,1]
    names(y) <- paste(1:nind(hyper))
    for(i in seq(along=pr)) rownames(pr[[i]]) <- names(y)
    n <- length(y)

    posnames <- unlist(lapply(pr, function(a) dimnames(a)[[3]]))
    names(posnames) <- NULL

    # scan
    out2 <- scan1(pr, y)

    # as expected?
    expect_equal(out2, scanone2scan1(out, 250, posnames, weights=FALSE))

    map <- lapply(hyper$geno, function(a) attr(a$prob, "map"))

    # cf lm() for chr 18
    out.lm <- lod_via_lm(pr[[18]], y, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

    ##############################
    # weighted scan
    set.seed(20151202)
    w <- runif(n, 1, 3)
    out <- scanone(hyper, method="hk", weights=w)

    names(w) <- names(y)
    out2 <- scan1(pr, y, weights=w)
    expect_equal(out2, scanone2scan1(out, 250, posnames, weights=TRUE))

    # cf lm() for chr 18
    out.lm <- lod_via_lm(pr[[18]], y, weights=w, map=map)
    expect_equal(subset(out2, map,"18"), out.lm)


    ##############################
    # additive covariate
    x <- sample(0:1, n, replace=TRUE)
    names(x) <- names(y)
    out <- scanone(hyper, method="hk", addcovar=x)
    out2 <- scan1(pr, y, addcovar=x)
    expect_equal(out2, scanone2scan1(out, 250, posnames, addcovar=x))

    # cf lm() for chr 18
    out.lm <- lod_via_lm(pr[[18]], y, x, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

    ##############################
    # additive covariate + weights
    out <- scanone(hyper, method="hk", addcovar=x, weights=w)
    out2 <- scan1(pr, y, addcovar=x, weights=w)
    expect_equal(out2, scanone2scan1(out, 250, posnames, colnames(y),
                                     addcovar=x, weights=TRUE))

    # cf lm() for chr 18
    out.lm <- lod_via_lm(pr[[18]], y, x, weights=w, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

    ##############################
    # interactive covariate
    out <- scanone(hyper, method="hk", addcovar=x, intcovar=x)
    out2 <- scan1(pr, y, addcovar=x, intcovar=x)
    expect_equal(out2, scanone2scan1(out, 250, posnames, colnames(y),
                                     addcovar=x, intcovar=x))

    # auto add intcovar?
    out2r <- scan1(pr, y, intcovar=x)
    expect_equal(out2r, scanone2scan1(out, 250, posnames, colnames(y),
                                      addcovar=x, intcovar=x))


    # cf lm() for chr 18
    out.lm <- lod_via_lm(pr[[18]], y, x, x, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

    ##############################
    # interactive covariate + weights
    out <- scanone(hyper, method="hk", addcovar=x, intcovar=x, weights=w)
    out2 <- scan1(pr, y, addcovar=x, intcovar=x, weights=w)
    expect_equal(out2, scanone2scan1(out, 250, posnames, colnames(y),
                                     addcovar=x, intcovar=x, weights=TRUE))

    # auto add intcovar?
    out2r <- scan1(pr, y, intcovar=x, weights=w)
    expect_equal(out2r, scanone2scan1(out, 250, posnames, colnames(y),
                                      addcovar=x, intcovar=x, weights=TRUE))

    # cf lm() for chr 18
    out.lm <- lod_via_lm(pr[[18]], y, x, x, weights=w, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

})


test_that("scan1 for backcross with multiple phenotypes with NAs", {

    skip_if(isnt_karl(), "this test only run locally")

    set.seed(20151202)
    library(qtl)
    data(hyper)

    # phenotypes
    n_phe <- 15
    n_ind <- nind(hyper)
    y <- matrix(rnorm(n_ind*n_phe), ncol=n_phe)
    # 5 batches
    spl <- split(sample(n_phe), rep(1:5, 3))
    nmis <- c(0, 5, 10, 15, 20)
    for(i in seq(along=spl)[-1])
        y[sample(n_ind, nmis[i]), spl[[i]]] <- NA

    # scan by R/qtl
    hyper$pheno <- cbind(y, hyper$pheno[,2,drop=FALSE])
    hyper <- calc.genoprob(hyper, step=2.5)
    out <- scanone(hyper, method="hk", pheno.col=1:n_phe)

    map <- lapply(hyper$geno, function(a) attr(a$prob, "map"))

    # inputs for R/qtl2
    pr <- convert_probs2qtl2(hyper)
    rownames(y) <- paste(1:n_ind)
    posnames <- unlist(lapply(pr, function(a) dimnames(a)[[3]]))
    names(posnames) <- NULL

    # scan
    out2 <- scan1(pr, y)

    # as expected?
    expect_equal(out2, scanone2scan1(out, colSums(!is.na(y)), posnames,
                                     colnames(y)))

    # cf lm() for chr 1
    out.lm <- lod_via_lm(pr[[18]], y, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

    ##############################
    # weighted scan
    w <- runif(n_ind, 1, 3)
    out <- scanone(hyper, method="hk", weights=w, pheno.col=1:n_phe)

    names(w) <- rownames(y)
    out2 <- scan1(pr, y, weights=w)
    expect_equal(out2, scanone2scan1(out, colSums(!is.na(y)), posnames,
                                     colnames(y), weights=TRUE))

    # cf lm() for chr 1
    out.lm <- lod_via_lm(pr[[18]], y, weights=w, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

    ##############################
    # additive covariate
    x <- sample(0:1, n_ind, replace=TRUE)
    names(x) <- rownames(y)
    out <- scanone(hyper, method="hk", addcovar=x, pheno.col=1:n_phe)
    out2 <- scan1(pr, y, addcovar=x)
    expect_equal(out2, scanone2scan1(out, colSums(!is.na(y)),
                                     posnames, colnames(y), addcovar=x))

    # cf lm() for chr 1
    out.lm <- lod_via_lm(pr[[18]], y, x, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

    ##############################
    # additive covariate + weights
    out <- scanone(hyper, method="hk", addcovar=x, weights=w, pheno.col=1:n_phe)
    out2 <- scan1(pr, y, addcovar=x, weights=w)
    expect_equal(out2, scanone2scan1(out, colSums(!is.na(y)),
                                     posnames, colnames(y), addcovar=x,
                                     weights=TRUE))

    # cf lm() for chr 1
    out.lm <- lod_via_lm(pr[[18]], y, x, weights=w, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

    ##############################
    # interactive covariate
    out <- scanone(hyper, method="hk", addcovar=x, intcovar=x, pheno.col=1:n_phe)
    out2 <- scan1(pr, y, addcovar=x, intcovar=x)
    expect_equal(out2, scanone2scan1(out, colSums(!is.na(y)),
                                     posnames, colnames(y), addcovar=x,
                                     intcovar=x))

    # auto add intcovar?
    out2r <- scan1(pr, y, intcovar=x)
    expect_equal(out2r, scanone2scan1(out, colSums(!is.na(y)),
                                      posnames, colnames(y), addcovar=x,
                                      intcovar=x))

    # cf lm() for chr 1
    out.lm <- lod_via_lm(pr[[18]], y, x, intcovar=x, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

    ##############################
    # interactive covariate + weights
    out <- scanone(hyper, method="hk", addcovar=x, intcovar=x, weights=w, pheno.col=1:n_phe)
    out2 <- scan1(pr, y, addcovar=x, intcovar=x, weights=w)
    expect_equal(out2, scanone2scan1(out, colSums(!is.na(y)),
                                     posnames, colnames(y), addcovar=x,
                                     intcovar=x, weights=TRUE))

    # auto add intcovar?
    out2r <- scan1(pr, y, intcovar=x, weights=w)
    expect_equal(out2r, scanone2scan1(out, colSums(!is.na(y)),
                                      posnames, colnames(y), addcovar=x,
                                      intcovar=x, weights=TRUE))

    # cf lm() for chr 1
    out.lm <- lod_via_lm(pr[[18]], y, x, intcovar=x, weights=w, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)

})


test_that("scan1 works with NAs in the covariates", {

    skip_if(isnt_karl(), "this test only run locally")

    set.seed(20151202)
    library(qtl)
    data(hyper)

    # phenotypes
    n_phe <- 15
    n_ind <- nind(hyper)
    y <- matrix(rnorm(n_ind*n_phe), ncol=n_phe)
    # 5 batches
    spl <- split(sample(n_phe), rep(1:5, 3))
    nmis <- c(0, 5, 10, 15, 20)
    for(i in seq(along=spl)[-1])
        y[sample(n_ind, nmis[i]), spl[[i]]] <- NA
    hyper$pheno <- cbind(y, hyper$pheno[,2,drop=FALSE])

    # genoprobs by R/qtl
    hyper <- calc.genoprob(hyper, step=2.5)

    # inputs for R/qtl2
    pr <- convert_probs2qtl2(hyper)
    rownames(y) <- paste(1:n_ind)
    posnames <- unlist(lapply(pr, function(a) dimnames(a)[[3]]))
    names(posnames) <- NULL

    ##############################
    # additive covariate
    x <- sample(0:1, n_ind, replace=TRUE)
    names(x) <- rownames(y)
    x[5] <- NA

    suppressWarnings(out <- scanone(hyper, method="hk", addcovar=x, pheno.col=1:n_phe))
    out2 <- scan1(pr, y, addcovar=x)
    expect_equal(out2, scanone2scan1(out, colSums(!(is.na(y) | is.na(x))),
                                     posnames, colnames(y), addcovar=x))

    map <- lapply(hyper$geno, function(a) attr(a$prob, "map"))

    # cf lm() for chr 1
    out.lm <- lod_via_lm(pr[[18]], y, x, map=map)
    expect_equal(subset(out2, map, "18"), out.lm)
})


test_that("scan1 aligns the individuals", {

    skip_if(isnt_karl(), "this test only run locally")

    set.seed(20151202)
    library(qtl)
    data(hyper)

    # phenotypes
    n_phe <- 15
    n_ind <- nind(hyper)
    y <- matrix(rnorm(n_ind*n_phe), ncol=n_phe)
    # 5 batches
    spl <- split(sample(n_phe), rep(1:5, 3))
    nmis <- c(0, 5, 10, 15, 20)
    for(i in seq(along=spl)[-1])
        y[sample(n_ind, nmis[i]), spl[[i]]] <- NA
    hyper$pheno <- cbind(y, hyper$pheno[,2,drop=FALSE])

    # genoprobs from R/qtl
    hyper <- calc.genoprob(hyper, step=2.5)

    # inputs for R/qtl2
    pr <- convert_probs2qtl2(hyper)
    rownames(y) <- paste(1:n_ind)
    posnames <- unlist(lapply(pr, function(a) dimnames(a)[[3]]))
    names(posnames) <- NULL

    # scan
    out <- scan1(pr, y)
    out_perm <- scan1(pr, y[sample(n_ind),])
    expect_equal(out_perm, out)

    class(pr) <- c("calc_genoprob", "list") # allows simpler reordering of individuals
    out_perm <- scan1(pr[sample(n_ind),], y)
    expect_equal(out_perm, out)

    ##############################
    # weighted scan
    w <- runif(n_ind, 1, 3)
    names(w) <- rownames(y)

    out <- scan1(pr, y, weights=w)
    out_perm <- scan1(pr[sample(n_ind),], y[sample(n_ind),], weights=w[sample(n_ind)])
    expect_equal(out_perm, out)

    ##############################
    # additive covariate
    x <- sample(0:1, n_ind, replace=TRUE)
    names(x) <- rownames(y)

    out <- scan1(pr, y, addcovar=x)
    out_perm <- scan1(pr[sample(n_ind),], y[sample(n_ind),], addcovar=x[sample(n_ind)])
    expect_equal(out_perm, out)

    ##############################
    # additive covariate + weights
    out <- scan1(pr, y, addcovar=x, weights=w)
    out_perm <- scan1(pr[sample(n_ind),], y[sample(n_ind),], addcovar=x[sample(n_ind)],
                      weights=w[sample(n_ind)])
    expect_equal(out_perm, out)

    ##############################
    # interactive covariate
    out <- scan1(pr, y, addcovar=x, intcovar=x)
    out_perm <- scan1(pr[sample(n_ind),], y[sample(n_ind),], addcovar=x[sample(n_ind)],
                      intcovar=x[sample(n_ind)])
    expect_equal(out_perm, out)

    # auto add intcovar?
    out_perm <- scan1(pr[sample(n_ind),], y[sample(n_ind),], intcovar=x[sample(n_ind)])
    expect_equal(out_perm, out)

    ##############################
    # interactive covariate + weights
    out <- scan1(pr, y, addcovar=x, intcovar=x, weights=w)
    out_perm <- scan1(pr[sample(n_ind),], y[sample(n_ind),], addcovar=x[sample(n_ind)],
                      intcovar=x[sample(n_ind)], weights=w[sample(n_ind)])
    expect_equal(out_perm, out)

    # auto add intcovar?
    out_perm <- scan1(pr[sample(n_ind),], y[sample(n_ind),],
                      intcovar=x[sample(n_ind)], weights=w[sample(n_ind)])
    expect_equal(out_perm, out)

})


test_that("multi-core scan1 works", {
    skip_if(isnt_karl(), "this test only run locally")

    set.seed(20151202)
    library(qtl)
    data(hyper)

    # phenotypes
    n_phe <- 15
    n_ind <- nind(hyper)
    y <- matrix(rnorm(n_ind*n_phe), ncol=n_phe)
    # 5 batches
    spl <- split(sample(n_phe), rep(1:5, 3))
    nmis <- c(0, 5, 10, 15, 20)
    for(i in seq(along=spl)[-1])
        y[sample(n_ind, nmis[i]), spl[[i]]] <- NA
    rownames(y) <- paste(1:n_ind)

    # inputs for R/qtl2
    hyper2 <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper2, step=2.5)
    pr <- calc_genoprob(hyper2, map)
    posnames <- unlist(lapply(pr, function(a) dimnames(a)[[3]]))
    names(posnames) <- NULL

    # scan
    out <- scan1(pr, y)
    out_multicore <- scan1(pr, y, cores=2)
    expect_equal(out_multicore, out)

    ##############################
    # weighted scan
    w <- runif(n_ind, 1, 3)
    names(w) <- rownames(y)

    out <- scan1(pr, y, weights=w)
    out_multicore <- scan1(pr, y, weights=w, cores=2)
    expect_equal(out_multicore, out)

    ##############################
    # additive covariate
    x <- sample(0:1, n_ind, replace=TRUE)
    names(x) <- rownames(y)

    out <- scan1(pr, y, addcovar=x)
    out_multicore <- scan1(pr, y, addcovar=x, cores=2)
    expect_equal(out_multicore, out)

    ##############################
    # additive covariate + weights
    out <- scan1(pr, y, addcovar=x, weights=w)
    out_multicore <- scan1(pr, y, addcovar=x, weights=w, cores=2)
    expect_equal(out_multicore, out)

    ##############################
    # interactive covariate
    out <- scan1(pr, y, addcovar=x, intcovar=x)
    out_multicore <- scan1(pr, y, addcovar=x, intcovar=x, cores=2)
    expect_equal(out_multicore, out)

    # auto add intcovar?
    out_multicore <- scan1(pr, y, intcovar=x, cores=2)
    expect_equal(out_multicore, out)

    ##############################
    # interactive covariate + weights
    out <- scan1(pr, y, addcovar=x, intcovar=x, weights=w)
    out_multicore <- scan1(pr, y, addcovar=x, intcovar=x, weights=w, cores=2)
    expect_equal(out_multicore, out)

    # auto add intcovar?
    out_multicore <- scan1(pr, y, intcovar=x, weights=w, cores=2)
    expect_equal(out_multicore, out)

})


test_that("scan1 LOD results don't depend on scale of x and y", {

    skip_if(isnt_karl(), "this test only run locally")

    set.seed(20151202)
    library(qtl)
    data(hyper)

    # phenotypes
    n_phe <- 15
    n_ind <- nind(hyper)
    y <- matrix(rnorm(n_ind*n_phe), ncol=n_phe)
    # 5 batches
    spl <- split(sample(n_phe), rep(1:5, 3))
    nmis <- c(0, 5, 10, 15, 20)
    for(i in seq(along=spl)[-1])
        y[sample(n_ind, nmis[i]), spl[[i]]] <- NA
    rownames(y) <- paste(1:n_ind)

    # inputs for R/qtl2
    hyper2 <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper2$gmap, step=2.5)
    pr <- calc_genoprob(hyper2, map)
    posnames <- unlist(lapply(pr, function(a) dimnames(a)[[3]]))
    names(posnames) <- NULL

    ybig <- y*100

    # scan
    out <- scan1(pr, y)
    outbig <- scan1(pr, ybig)
    expect_equal(outbig, out)

    ##############################
    # weighted scan
    w <- runif(n_ind, 1, 3)
    names(w) <- rownames(y)
    wbig <- w*5

    out <- scan1(pr, y, weights=w)
    outbig <- scan1(pr, ybig, weights=wbig)
    expect_equal(outbig, out)

    ##############################
    # additive covariate
    x <- sample(0:1, n_ind, replace=TRUE)
    names(x) <- rownames(y)
    xbig <- x*50

    out <- scan1(pr, y, addcovar=x)
    outbig <- scan1(pr, ybig, addcovar=xbig)
    expect_equal(outbig, out)

    ##############################
    # additive covariate + weights
    out <- scan1(pr, y, addcovar=x, weights=w)
    outbig <- scan1(pr, ybig, addcovar=xbig, weights=wbig)
    expect_equal(outbig, out)

    ##############################
    # interactive covariate
    out <- scan1(pr, y, addcovar=x, intcovar=x)
    outbig <- scan1(pr, ybig, addcovar=xbig, intcovar=xbig)
    expect_equal(outbig, out)

    ##############################
    # interactive covariate + weights
    out <- scan1(pr, y, addcovar=x, intcovar=x, weights=w)
    outbig <- scan1(pr, ybig, addcovar=xbig, intcovar=xbig, weights=wbig)
    expect_equal(outbig, out)

})

test_that("scan1 deals with mismatching individuals", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs, "loco")
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    ind <- c(1:50, 101:150)
    expected <- scan1(probs[ind,], iron$pheno[ind,,drop=FALSE], addcovar=X[ind], intcovar=X[ind],
                      Xcovar=Xc[ind,])
    expect_equal(scan1(probs[ind,], iron$pheno, addcovar=X, intcovar=X, Xcovar=Xc), expected)
    expect_equal(scan1(probs, iron$pheno[ind,], addcovar=X, intcovar=X, Xcovar=Xc), expected)
    expect_equal(scan1(probs, iron$pheno, addcovar=X[ind], intcovar=X, Xcovar=Xc), expected)
    expect_equal(scan1(probs, iron$pheno, addcovar=X, intcovar=X[ind], Xcovar=Xc), expected)
    expect_equal(scan1(probs, iron$pheno, addcovar=X, intcovar=X, Xcovar=Xc[ind,]), expected)

})

test_that("scan1 can handle decomposed kinship matrix", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    pr <- calc_genoprob(iron)
    k <- calc_kinship(pr)
    kd <- decomp_kinship(k)
    expect_equal( scan1(pr, iron$pheno, kd),
                  scan1(pr, iron$pheno, k) )

})


test_that("weights derived from table() are okay", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    pr <- calc_genoprob(iron)
    k <- calc_kinship(pr)

    w1 <- setNames(sample(1:10, nrow(iron$pheno), replace=TRUE), rownames(iron$pheno))
    out1 <- scan1(pr, iron$pheno, weights=w1)
    out1k <- scan1(pr, iron$pheno, k, weights=w1)

    w2 <- table( rep(names(w1), w1) )
    out2 <- scan1(pr, iron$pheno, weights=w2)
    out2k <- scan1(pr, iron$pheno, k, weights=w2)

    expect_equal(out1, out2)
    expect_equal(out1k, out2k)

})

test_that("scan1 can handle names in the phenotype rownames", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    pr <- calc_genoprob(iron, error_prob=0.002)
    pheno <- iron$pheno
    out <- scan1(pr, pheno)

    names(rownames(pheno)) <- paste0("mouse", rownames(pheno))
    expect_equal(scan1(pr, pheno), out)

    names(rownames(pheno)) <- rep(NA, nrow(pheno))
    expect_equal(scan1(pr, pheno), out)


    # the same with kinship matrix
    pheno <- iron$pheno
    k <- calc_kinship(pr)
    out <- scan1(pr, pheno, k)  # no error

    names(rownames(pheno)) <- paste0("mouse", rownames(pheno))
    expect_equal(scan1(pr, pheno, k), out)

    names(rownames(pheno)) <- rep(NA, nrow(pheno))
    expect_equal(scan1(pr, pheno, k), out)

})

context("scan1gen")

test_that("scan1gen works", {

    # function to fit glm
    ll_glm <-
        function(pr, pheno, addcovar=NULL, ...)
        {
            if(ncol(pheno)>1)
                return(sapply(1:ncol(pheno), function(i) ll_glm(pr, pheno[,i,drop=FALSE], addcovar=addcovar, ...)))

            formula <- ifelse(is.null(pr), "pheno ~ 1", "pheno ~ pr")
            if(!is.null(addcovar)) formula <- paste(formula, "+ addcovar")

            glm_out <- glm(as.formula(formula), family=binomial(link=logit))
            -glm_out$deviance/(2*log(10)) # log10 likelihood
        }

    # read data
    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c("1", "16", "X")] # subset to chr 1, 16, and X

    # insert pseudomarkers into map
    map <- insert_pseudomarkers(iron$gmap, step=1)

    # calculate genotype probabilities
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    # covariates for X chr under null
    Xcovar <- get_x_covar(iron)

    # create binary trait
    bin_pheno <- setNames(as.numeric(iron$pheno[,1] > median(iron$pheno[,1])),
                          rownames(iron$pheno))

    expect_equal(scan1gen(probs, bin_pheno, Xcovar=Xcovar, func=ll_glm),
                 scan1(probs, bin_pheno, Xcovar=Xcovar, model="binary"))


    # try a second binary trait
    bin_pheno <- cbind(bin_pheno,
                       as.numeric(iron$pheno[,1] > quantile(iron$pheno[,1], 0.75)))

    # vectorize the function
    expect_equal(scan1gen(probs, bin_pheno, Xcovar=Xcovar, func=ll_glm),
                 scan1(probs, bin_pheno, Xcovar=Xcovar, model="binary"))

    # don't vectorize the function
    expect_equal(scan1gen(probs, bin_pheno, Xcovar=Xcovar, func=ll_glm, vectorize_func=FALSE),
                 scan1(probs, bin_pheno, Xcovar=Xcovar, model="binary"))

})

context("scan1max")

test_that("scan1max works", {

    iron <- read_cross2( system.file("extdata", "iron.zip", package="qtl2") )
    iron <- iron[,c("1", "19", "X")]
    Xc <- get_x_covar(iron)
    X <- Xc[,1]
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    pr <- calc_genoprob(iron, map, error_prob=0.002)
    k <- calc_kinship(pr)
    k_loco <- calc_kinship(pr, "loco")
    chr <- factor(rep(names(map), sapply(map, length)), names(map))

    out <- scan1(pr, iron$pheno)
    expected <- apply(out, 2, max, na.rm=TRUE)
    expected_bychr <- apply(out, 2, tapply, chr, max, na.rm=TRUE)
    attr(expected, "sample_size") <- attr(expected_bychr, "sample_size") <- attr(out, "sample_size")
    expect_equal(scan1max(pr, iron$pheno), expected)
    expect_equal(scan1max(pr, iron$pheno, by_chr=TRUE), expected_bychr)

    out <- scan1(pr, iron$pheno, Xcovar=Xc)
    expected <- apply(out, 2, max, na.rm=TRUE)
    expected_bychr <- apply(out, 2, tapply, chr, max, na.rm=TRUE)
    attr(expected, "sample_size") <- attr(expected_bychr, "sample_size") <- attr(out, "sample_size")
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc), expected)
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, by_chr=TRUE), expected_bychr)

    out <- scan1(pr, iron$pheno, Xcovar=Xc, addcovar=X)
    expected <- apply(out, 2, max, na.rm=TRUE)
    expected_bychr <- apply(out, 2, tapply, chr, max, na.rm=TRUE)
    attr(expected, "sample_size") <- attr(expected_bychr, "sample_size") <- attr(out, "sample_size")
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, addcovar=X), expected)
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, addcovar=X, by_chr=TRUE), expected_bychr)

    out <- scan1(pr, iron$pheno, Xcovar=Xc, kinship=k)
    expected <- apply(out, 2, max, na.rm=TRUE)
    expected_bychr <- apply(out, 2, tapply, chr, max, na.rm=TRUE)
    for(at in c("sample_size", "hsq")) {
        attr(expected, at) <- attr(expected_bychr, at) <- attr(out, at)
    }
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, kinship=k), expected)
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, kinship=k, by_chr=TRUE), expected_bychr)

    out <- scan1(pr, iron$pheno, Xcovar=Xc, addcovar=X, kinship=k_loco)
    expected <- apply(out, 2, max, na.rm=TRUE)
    expected_bychr <- apply(out, 2, tapply, chr, max, na.rm=TRUE)
    for(at in c("sample_size", "hsq")) {
        attr(expected, at) <- attr(expected_bychr, at) <- attr(out, at)
    }
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, addcovar=X, kinship=k_loco), expected)
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, addcovar=X, kinship=k_loco, by_chr=TRUE), expected_bychr)

})


test_that("scan1max works with multicore", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2( system.file("extdata", "iron.zip", package="qtl2") )
    iron <- iron[,c("1", "19", "X")]
    Xc <- get_x_covar(iron)
    X <- Xc[,1]
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    pr <- calc_genoprob(iron, map, error_prob=0.002)
    k <- calc_kinship(pr)
    k_loco <- calc_kinship(pr, "loco")
    chr <- factor(rep(names(map), sapply(map, length)), names(map))

    out <- scan1(pr, iron$pheno, cores=2)
    expected <- apply(out, 2, max, na.rm=TRUE)
    expected_bychr <- apply(out, 2, tapply, chr, max, na.rm=TRUE)
    attr(expected, "sample_size") <- attr(expected_bychr, "sample_size") <- attr(out, "sample_size")
    expect_equal(scan1max(pr, iron$pheno, cores=2), expected)
    expect_equal(scan1max(pr, iron$pheno, by_chr=TRUE, cores=2), expected_bychr)

    out <- scan1(pr, iron$pheno, Xcovar=Xc, cores=2)
    expected <- apply(out, 2, max, na.rm=TRUE)
    expected_bychr <- apply(out, 2, tapply, chr, max, na.rm=TRUE)
    attr(expected, "sample_size") <- attr(expected_bychr, "sample_size") <- attr(out, "sample_size")
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, cores=2), expected)
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, by_chr=TRUE, cores=2), expected_bychr)

    out <- scan1(pr, iron$pheno, Xcovar=Xc, addcovar=X, cores=2)
    expected <- apply(out, 2, max, na.rm=TRUE)
    expected_bychr <- apply(out, 2, tapply, chr, max, na.rm=TRUE)
    attr(expected, "sample_size") <- attr(expected_bychr, "sample_size") <- attr(out, "sample_size")
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, addcovar=X, cores=2), expected)
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, addcovar=X, by_chr=TRUE, cores=2), expected_bychr)

    out <- scan1(pr, iron$pheno, Xcovar=Xc, kinship=k, cores=2)
    expected <- apply(out, 2, max, na.rm=TRUE)
    expected_bychr <- apply(out, 2, tapply, chr, max, na.rm=TRUE)
    for(at in c("sample_size", "hsq")) {
        attr(expected, at) <- attr(expected_bychr, at) <- attr(out, at)
    }
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, kinship=k, cores=2), expected)
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, kinship=k, by_chr=TRUE, cores=2), expected_bychr)

    out <- scan1(pr, iron$pheno, Xcovar=Xc, addcovar=X, kinship=k_loco, cores=2)
    expected <- apply(out, 2, max, na.rm=TRUE)
    expected_bychr <- apply(out, 2, tapply, chr, max, na.rm=TRUE)
    for(at in c("sample_size", "hsq")) {
        attr(expected, at) <- attr(expected_bychr, at) <- attr(out, at)
    }
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, addcovar=X, kinship=k_loco, cores=2), expected)
    expect_equal(scan1max(pr, iron$pheno, Xcovar=Xc, addcovar=X, kinship=k_loco, by_chr=TRUE, cores=2),
                 expected_bychr)

})

#' Genome scan with a single-QTL model
#'
#' Genome scan with a single-QTL model by Haley-Knott regression or a
#' linear mixed model, with possible allowance for covariates.
#'
#' @param genoprobs Genotype probabilities as calculated by
#' [calc_genoprob()].
#' @param pheno A numeric matrix of phenotypes, individuals x phenotypes.
#' @param kinship Optional kinship matrix, or a list of kinship matrices (one
#' per chromosome), in order to use the LOCO (leave one chromosome
#' out) method.
#' @param addcovar An optional numeric matrix of additive covariates.
#' @param Xcovar An optional numeric matrix with additional additive covariates used for
#' null hypothesis when scanning the X chromosome.
#' @param intcovar An numeric optional matrix of interactive covariates.
#' @param weights An optional numeric vector of positive weights for the
#' individuals. As with the other inputs, it must have `names`
#' for individual identifiers.
#' @param reml If `kinship` provided: if `reml=TRUE`, use
#' REML; otherwise maximum likelihood.
#' @param model Indicates whether to use a normal model (least
#'     squares) or binary model (logistic regression) for the phenotype.
#'     If `model="binary"`, the phenotypes must have values in \eqn{[0, 1]}.
#' @param hsq Considered only if `kinship` is provided, in which case
#'     this is taken as the assumed value for the residual
#'     heritability. It should be a vector with length corresponding
#'     to the number of columns in `pheno`, or (if `kinship`
#'     corresponds to a list of LOCO kinship matrices) a matrix with dimension
#'     `length(kinship) x ncol(pheno)`.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#' @param ... Additional control parameters; see Details.
#'
#' @return An object of class `"scan1"`: a matrix of LOD scores, positions x phenotypes.
#' Also contains one or more of the following attributes:
#' * `sample_size` - Vector of sample sizes used for each
#'    phenotype
#' * `hsq` - Included if `kinship` provided: A matrix of
#'    estimated heritabilities under the null hypothesis of no
#'    QTL. Columns are the phenotypes. If the `"loco"` method was
#'    used with [calc_kinship()] to calculate a list
#'    of kinship matrices, one per chromosome, the rows of `hsq`
#'    will be the heritabilities for the different chromosomes (well,
#'    leaving out each one). If `Xcovar` was not NULL, there will at
#'    least be an autosome and X chromosome row.
#'
#' @details
#' We first fit the model \eqn{y = X \beta + \epsilon}{y = Xb + e}
#' where \eqn{X} is a matrix of covariates (or just an intercept) and
#' \eqn{\epsilon}{e} is multivariate normal with mean 0 and covariance
#' matrix \eqn{\sigma^2 [h^2 (2 K) + (1-h^2) I]}{sigmasq*[hsq*2*K+(1-hsq)*I]} where
#' \eqn{K} is the kinship matrix and \eqn{I} is the identity matrix.
#'
#' We then take \eqn{h^2}{hsq} as fixed and then scan the genome, at
#' each genomic position fitting the model \eqn{y = P \alpha + X \beta
#' + \epsilon}{y = Xb + e} where \eqn{P} is a matrix of genotype
#' probabilities for the current position and again \eqn{X} is a
#' matrix of covariates \eqn{\epsilon}{e} is multivariate normal with
#' mean 0 and covariance matrix \eqn{\sigma^2 [h^2 (2 K) +
#' (1-h^2) I]}{sigmasq*[hsq*2*K+(1-hsq)*I]}, taking \eqn{h^2}{hsq} to be known.
#'
#' Note that if `weights` are provided, the covariance matrix becomes
#' \eqn{\sigma^2 [h^2 (2 K) + (1-h^2) D]}{sigmasq*[hsq*2*K+(1-hsq)*D]} where
#' \eqn{D} is a diagonal matrix with the reciprocal of the weights.
#'
#' For each of the inputs, the row names are used as
#' individual identifiers, to align individuals. The `genoprobs`
#' object should have a component `"is_x_chr"` that indicates
#' which of the chromosomes is the X chromosome, if any.
#'
#' The `...` argument can contain several additional control
#' parameters; suspended for simplicity (or confusion, depending on
#' your point of view). `tol` is used as a tolerance value for linear
#' regression by QR decomposition (in determining whether columns are
#' linearly dependent on others and should be omitted); default
#' `1e-12`. `intcovar_method` indicates whether to use a high-memory
#' (but potentially faster) method or a low-memory (and possibly
#' slower) method, with values `"highmem"` or `"lowmem"`; default
#' `"lowmem"`. `max_batch` indicates the maximum number of phenotypes
#' to run together; default is unlimited. `maxit` is the maximum
#' number of iterations for converence of the iterative algorithm
#' used when `model=binary`. `bintol` is used as a tolerance for
#' converence for the iterative algorithm used when `model=binary`.
#' `eta_max` is the maximum value for the "linear predictor" in the
#' case `model="binary"` (a bit of a technicality to avoid fitted
#' values exactly at 0 or 1).
#'
#' If `kinship` is absent, Haley-Knott regression is performed.
#' If `kinship` is provided, a linear mixed model is used, with a
#' polygenic effect estimated under the null hypothesis of no (major)
#' QTL, and then taken as fixed as known in the genome scan.
#'
#' If `kinship` is a single matrix, then the `hsq`
#' in the results is a vector of heritabilities (one value for each phenotype). If
#' `kinship` is a list (one matrix per chromosome), then
#' `hsq` is a matrix, chromosomes x phenotypes.
#'
#' @references Haley CS, Knott SA (1992) A simple
#' regression method for mapping quantitative trait loci in line
#' crosses using flanking markers.  Heredity 69:315--324.
#'
#' Kang HM, Zaitlen NA, Wade CM, Kirby A, Heckerman D, Daly MJ, Eskin
#' E (2008) Efficient control of population structure in model
#' organism association mapping. Genetics 178:1709--1723.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c("19", "X")] # subset to chr 19 and X}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' # leave-one-chromosome-out kinship matrices
#' kinship <- calc_kinship(probs, "loco")
#'
#' # genome scan with a linear mixed model
#' out_lmm <- scan1(probs, pheno, kinship, covar, Xcovar)
#'
#' @seealso [scan1perm()], [scan1coef()], [cbind.scan1()], [rbind.scan1()], [scan1max()]
#'
#' @export
scan1 <-
    function(genoprobs, pheno, kinship=NULL, addcovar=NULL, Xcovar=NULL,
             intcovar=NULL, weights=NULL, reml=TRUE,
             model=c("normal", "binary"), hsq=NULL, cores=1, ...)
{
    if(is.null(genoprobs)) stop("genoprobs is NULL")
    if(is.null(pheno)) stop("pheno is NULL")

    # grab dot args
    dotargs <- list(...)
    if("n_perm" %in% names(dotargs))
        stop("You included n_perm as an argument; you probably want to run scan1perm not scan1.")

    model <- match.arg(model)

    if(!is.null(kinship)) { # fit linear mixed model
        if(model=="binary") warning("Can't fit binary model with kinship matrix; using normal model")
        return(scan1_pg(genoprobs, pheno, kinship, addcovar, Xcovar, intcovar,
                        weights, reml, hsq, cores, ...))
    }

    # deal with the dot args
    tol <- grab_dots(dotargs, "tol", 1e-12)
    if(!is_pos_number(tol)) stop("tol should be a single positive number")
    intcovar_method <- grab_dots(dotargs, "intcovar_method", "lowmem",
                                 c("highmem", "lowmem"))
    quiet <- grab_dots(dotargs, "quiet", TRUE)
    max_batch <- grab_dots(dotargs, "max_batch", NULL)
    if(!is.null(max_batch) && !is_pos_number(max_batch)) stop("max_batch should be a single positive integer")
    if(model=="binary") {
        bintol <- grab_dots(dotargs, "bintol", sqrt(tol)) # for model="binary"
        if(!is_pos_number(bintol)) stop("bintol should be a single positive number")
        eta_max <- grab_dots(dotargs, "eta_max", log(1-tol)-log(tol)) # for model="binary"
        if(!is_pos_number(eta_max)) stop("eta_max should be a single positive number")
        maxit <- grab_dots(dotargs, "maxit", 100) # for model="binary"
        if(!is_nonneg_number(maxit)) stop("maxit should be a single non-negative integer")
        check_extra_dots(dotargs, c("tol", "intcovar_method", "quiet", "max_batch", "maxit", "bintol", "eta_max"))
    }
    else {
        check_extra_dots(dotargs, c("tol", "intcovar_method", "quiet", "max_batch"))
    }

    # check that the objects have rownames
    check4names(pheno, addcovar, Xcovar, intcovar)

    # force things to be matrices
    if(!is.matrix(pheno)) {
        pheno <- as.matrix(pheno)
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }
    if(is.null(colnames(pheno))) # force column names
        colnames(pheno) <- paste0("pheno", seq_len(ncol(pheno)))
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(Xcovar)) {
        if(!is.matrix(Xcovar)) Xcovar <- as.matrix(Xcovar)
        if(!is.numeric(Xcovar)) stop("Xcovar is not numeric")
    }
    if(!is.null(intcovar)) {
        if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)
        if(!is.numeric(intcovar)) stop("intcovar is not numeric")
    }

    # for binary model
    if(model=="binary") {
        if(!is.null(kinship))
            stop("Can't yet account for kinship with model = \"binary\"")
        pheno <- check_binary_pheno(pheno)
    }
    else {
        # square-root of weights (only if model="normal")
        weights <- sqrt_weights(weights) # also check >0 (and if all 1's, turn to NULL)
    }

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, addcovar, Xcovar, intcovar,
                               weights, complete.cases=TRUE)
    ind2keep <- get_common_ids(ind2keep, pheno[rowSums(is.finite(pheno)) > 0,,drop=FALSE])
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # make sure columns in intcovar are also in addcovar
    addcovar <- force_intcovar(addcovar, intcovar, tol)

    # drop things from Xcovar that are already in addcovar
    Xcovar <- drop_xcovar(addcovar, Xcovar, tol)

    # batch phenotypes by missing values
    phe_batches <- batch_cols(pheno[ind2keep,,drop=FALSE], max_batch)

    # drop cols in genotype probs that are all 0 (just looking at the X chromosome)
    genoprob_Xcol2drop <- genoprobs_col2drop(genoprobs)
    is_x_chr <- attr(genoprobs, "is_x_chr")
    if(is.null(is_x_chr)) is_x_chr <- rep(FALSE, length(genoprobs))

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # batches for analysis, to allow parallel analysis
    run_batches <- data.frame(chr=rep(seq_len(length(genoprobs)), length(phe_batches)),
                              phe_batch=rep(seq_along(phe_batches), each=length(genoprobs)))
    run_indexes <- seq_len(length(genoprobs)*length(phe_batches))

    # the function that does the work
    by_group_func <- function(i) {
        # deal with batch information, including individuals to drop due to missing phenotypes
        chr <- run_batches$chr[i]
        chrnam <- names(genoprobs)[chr]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        phecol <- phebatch$cols
        omit <- phebatch$omit
        these2keep <- ind2keep # individuals 2 keep for this batch
        if(length(omit) > 0) these2keep <- ind2keep[-omit]
        if(length(these2keep)<=2) return(NULL) # not enough individuals

        # subset the genotype probabilities: drop cols with all 0s, plus the first column
        Xcol2drop <- genoprob_Xcol2drop[[chrnam]]
        if(length(Xcol2drop) > 0) {
            pr <- genoprobs[[chr]][these2keep,-Xcol2drop,,drop=FALSE]
            pr <- pr[,-1,,drop=FALSE]
        }
        else
            pr <- genoprobs[[chr]][these2keep,-1,,drop=FALSE]

        # subset the rest
        ac <- addcovar; if(!is.null(ac)) { ac <- ac[these2keep,,drop=FALSE]; ac <- drop_depcols(ac, TRUE, tol) }
        Xc <- Xcovar;   if(!is.null(Xc)) Xc <- Xc[these2keep,,drop=FALSE]
        ic <- intcovar; if(!is.null(ic)) { ic <- ic[these2keep,,drop=FALSE]; ic <- drop_depcols(ic, TRUE, tol) }
        ph <- pheno[these2keep,phecol,drop=FALSE]
        wts <- weights[these2keep]

        # if X chr, paste X covariates onto additive covariates
        # (only for the null)
        if(is_x_chr[chr]) ac0 <- drop_depcols(cbind(ac, Xc), add_intercept=FALSE, tol)
        else ac0 <- ac

        if(model=="normal") {
            # FIX_ME: calculating null RSS multiple times :(
            nullrss <- nullrss_clean(ph, ac0, wts, add_intercept=TRUE, tol)

            # scan1 function taking clean data (with no missing values)
            rss <- scan1_clean(pr, ph, ac, ic, wts, add_intercept=TRUE, tol, intcovar_method)

            # calculate LOD score
            lod <- nrow(ph)/2 * (log10(nullrss) - log10(rss))
        }
        else { # binary traits
            # FIX_ME: calculating null LOD multiple times :(
            nulllod <- null_binary_clean(ph, ac0, wts, add_intercept=TRUE, maxit, bintol, tol, eta_max)

            # scan1 function taking clean data (with no missing values)
            lod <- scan1_binary_clean(pr, ph, ac, ic, wts, add_intercept=TRUE,
                                      maxit, bintol, tol, intcovar_method, eta_max)

            # calculate LOD score
            lod <- lod - nulllod
        }

        list(lod=lod, n=nrow(ph)) # return LOD & number of individuals used
    }

    # number of markers/pseudomarkers by chromosome, and their indexes to result matrix
    npos_by_chr <- dim(genoprobs)[3,]
    totpos <- sum(npos_by_chr)
    pos_index <- split(seq_len(totpos), rep(seq_len(length(genoprobs)), npos_by_chr))

    # object to contain the LOD scores; also attr to contain sample size
    result <- matrix(nrow=totpos, ncol=ncol(pheno))
    n <- rep(NA, ncol(pheno)); names(n) <- colnames(pheno)
    if(totpos==0) { # edge case of no genoprobs
        colnames(result) <- colnames(pheno)
        attr(result, "sample_size") <- n
        class(result) <- c("scan1", "matrix")
        return(result)
    }

    if(n_cores(cores)==1) { # no parallel processing
        for(i in run_indexes) {
            chr <- run_batches$chr[i]
            chrnam <- names(genoprobs)[chr]
            phebatch <- phe_batches[[run_batches$phe_batch[i]]]
            phecol <- phebatch$cols

            this_result <- by_group_func(i)
            if(!is.null(this_result)) {
                result[pos_index[[chr]], phecol] <- t(this_result$lod)
                if(chr==1) n[phecol] <- this_result$n
            }
        }
    }
    else {
        # calculations in parallel
        list_result <- cluster_lapply(cores, run_indexes, by_group_func)

        # check for problems (if clusters run out of memory, they'll return NULL)
        result_is_null <- vapply(list_result, is.null, TRUE)
        if(any(result_is_null))
            stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

        # reorganize results
        for(i in run_indexes) {
            chr <- run_batches$chr[i]
            chrnam <- names(genoprobs)[chr]
            phebatch <- phe_batches[[run_batches$phe_batch[i]]]
            phecol <- phebatch$cols

            if(!is.null(list_result[[i]])) {
                result[pos_index[[chr]], phecol] <- t(list_result[[i]]$lod)
                if(chr==1) n[phecol] <- list_result[[i]]$n
            }
        }
    }

    pos_names <- unlist(dimnames(genoprobs)[[3]])
    names(pos_names) <- NULL # this is just annoying
    dimnames(result) <- list(pos_names, colnames(pheno))

    # add some attributes with details on analysis
    attr(result, "sample_size") <- n

    class(result) <- c("scan1", "matrix")
    result
}


# scan1 function taking nicely aligned data with no missing values
#
# Here genoprobs is a plain 3d array
scan1_clean <-
    function(genoprobs, pheno, addcovar, intcovar,
             weights, add_intercept=TRUE, tol, intcovar_method)
{
    n <- nrow(pheno)
    if(add_intercept)
        addcovar <- cbind(rep(1,n), addcovar) # add intercept

    if(is.null(intcovar)) { # no interactive covariates

        if(is.null(weights)) { # no weights
            return( scan_hk_onechr(genoprobs, pheno, addcovar, tol) )
        } else { # weights included
            # note: pheno gets multiplied by weights in c++ (or really sqrt of original weights)
            return( scan_hk_onechr_weighted(genoprobs, pheno, addcovar, weights, tol) )
        }

    } else { # interactive covariates
        # high- and low-memory versions of functions
        if(intcovar_method=="highmem")
            scanf <- c(scan_hk_onechr_intcovar_highmem,
                       scan_hk_onechr_intcovar_weighted_highmem)
        else
            scanf <- c(scan_hk_onechr_intcovar_lowmem,
                       scan_hk_onechr_intcovar_weighted_lowmem)

        if(is.null(weights)) { # no weights
            return( scanf[[1]](genoprobs, pheno, addcovar, intcovar, tol) )
        } else { # weights included
            # note: pheno gets multiplied by weights in c++ (or really sqrt of original weights)
            return( scanf[[2]](genoprobs, pheno, addcovar, intcovar, weights, tol) )
        }

    }
}

# calculate null RSS, with nicely aligned data with no missing values
nullrss_clean <-
    function(pheno, addcovar, weights, add_intercept=TRUE, tol)
{
    n <- nrow(pheno)
    if(add_intercept)
        addcovar <- cbind(rep(1,n), addcovar) # add intercept

    if(is.null(weights) || length(weights)==0) { # no weights
        result <- calc_rss_linreg(addcovar, pheno, tol)
    } else { # weights included
        result <- calc_rss_linreg(addcovar*weights, pheno*weights, tol)
    }

    as.numeric(result)
}

# function to add column names as attribute
colnames4attr <-
    function(mat)
{
    if(is.null(mat)) return(NULL)

    if(!is.matrix(mat)) mat <- as.matrix(mat)
    if(ncol(mat)==0) return(NULL)
    cn <- colnames(mat)

    if(is.null(cn)) cn <- rep("", ncol(mat))

    if(any(cn=="")) cn[cn==""] <- paste0("unnamed", seq_len(sum(cn=="")))

    cn
}

# check that objects have rownames, if they are matrices
#   (or names, if not matrices)
check4names <-
    function(pheno=NULL, addcovar=NULL, Xcovar=NULL, intcovar=NULL, nullcovar=NULL)
{
    args <- list(pheno=pheno,
                 addcovar=addcovar,
                 Xcovar=Xcovar,
                 intcovar=intcovar,
                 nullcovar=nullcovar)

    for(i in seq_along(args)) {
        a <- args[[i]]
        if(!is.null(a)) {
            if(is.matrix(a)) {
                if(is.null(rownames(a)))
                    stop(names(args)[i], " has no rownames")
            }
            else {
                if(is.null(names(a)))
                    stop(names(args)[i], " has no names")
            }
        } # end if(!is.null)
    } # end loop

    TRUE
}

context("estimate heritability with LMM by est_herit")


test_that("est_herit works with intercross", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron[1:40, c("5", "19")]
    map <- insert_pseudomarkers(iron$gmap, step=5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs)

    # scan just chr 19; compare est_herit to those
    out_reml <- scan1(probs[,"19"], iron$pheno, kinship, reml=TRUE)
    out_ml <- scan1(probs[,"19"], iron$pheno, kinship, reml=FALSE)

    expect_equal(est_herit(iron$pheno, kinship, reml=TRUE)[1:2], # [1:2] to strip off attributes
                 attr(out_reml, "hsq")[1,]) # [1,] to convert to vector

    expect_equal(est_herit(iron$pheno, kinship, reml=FALSE)[1:2], # [1:2] to strip off attributes
                 attr(out_ml, "hsq")[1,]) # [1,] to convert to vector

    # try when some individuals missing from one or the other dataset
    subind <- 1:20
    expected <- est_herit(iron$pheno[subind,,drop=FALSE], kinship[subind,subind])
    expect_equal(est_herit(iron$pheno, kinship[subind, subind]), expected)
    expect_equal(est_herit(iron$pheno[subind,,drop=FALSE], kinship), expected)

})

test_that("est_herit with intercross with an additive covariate", {

    skip_on_cran()

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    out_reml <- scan1(probs[,"19"], iron$pheno, kinship, addcovar=X, reml=TRUE)
    out_ml <- scan1(probs[,"19"], iron$pheno, kinship, addcovar=X, reml=FALSE)

    expect_equal(est_herit(iron$pheno, kinship, addcovar=X, reml=TRUE)[1:2], # [1:2] to strip off attributes
                 attr(out_reml, "hsq")[1,]) # [1,] to convert to vector

    expect_equal(est_herit(iron$pheno, kinship, addcovar=X, reml=FALSE)[1:2], # [1:2] to strip off attributes
                 attr(out_ml, "hsq")[1,]) # [1,] to convert to vector

    # try when some individuals missing from one or the other dataset
    subind <- c(1:50,101:150)
    expected <- est_herit(iron$pheno[subind,,drop=FALSE], kinship[subind,subind], addcovar=X[subind])
    expect_equal(est_herit(iron$pheno, kinship[subind, subind], addcovar=X), expected)
    expect_equal(est_herit(iron$pheno[subind,,drop=FALSE], kinship, addcovar=X), expected)
    expect_equal(est_herit(iron$pheno, kinship, addcovar=X[subind]), expected)
    expect_equal(est_herit(iron$pheno[subind,,drop=FALSE], kinship[subind, subind], addcovar=X), expected)
    expect_equal(est_herit(iron$pheno[subind,,drop=FALSE], kinship, addcovar=X[subind]), expected)
    expect_equal(est_herit(iron$pheno, kinship[subind,subind], addcovar=X[subind]), expected)

    # give error when covariates are not numeric
    X_not_numer <-iron$covar[,"sex",drop=FALSE]
    expect_error(est_herit(iron$pheno, kinship, X_not_numer))

})

test_that("est_herit handles dependent covariate columns", {

    skip_on_cran()

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs)
    pheno <- iron$pheno
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)

    covar2 <- cbind(a=covar,b=0)
    pheno[1:2,1] <- NA
    covar2[1:2,2] <- 1
    hsq <- est_herit(pheno, kinship, covar)
    hsq2 <- est_herit(pheno, kinship, covar2)
    hsq3 <- est_herit(pheno, kinship, covar2[,-2, drop=FALSE])

    expect_equal(hsq[1], hsq2[1])
    expect_equal(hsq2[2], c(spleen=0.545220193658054), tolerance=1e-6)
    expect_equal(hsq, hsq3)

    pheno[1:2,1:2] <- NA
    hsq <- est_herit(pheno, kinship, covar)
    hsq2 <- est_herit(pheno, kinship, covar2)
    hsq3 <- est_herit(pheno, kinship, covar2[,-2, drop=FALSE])

    expect_equal(hsq, hsq2)
    expect_equal(hsq, hsq3)

})

context("cbind and rbind for scan1 objects")

grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
grav2 <- grav2[1:20,1:3]
map <- insert_pseudomarkers(grav2$gmap, step=1)
probs <- calc_genoprob(grav2, map, error_prob=0.002)
k <- calc_kinship(probs)
kloco <- calc_kinship(probs, "loco")


test_that("cbind.scan1() works for scan1() results", {

    out1 <- scan1(probs, grav2$pheno[,1,drop=FALSE])
    out2 <- scan1(probs, grav2$pheno[,2,drop=FALSE])
    out12 <- scan1(probs, grav2$pheno[,1:2])

    expect_equal(cbind(out1, out2), out12)

})

test_that("cbind.scan1() works for scan1/LMM results", {

    out1 <- scan1(probs, grav2$pheno[,1,drop=FALSE], k)
    out2 <- scan1(probs, grav2$pheno[,2,drop=FALSE], k)
    out12 <- scan1(probs, grav2$pheno[,1:2], k)
    expect_equal(cbind(out1, out2), out12, tolerance=1e-6)

    out1 <- scan1(probs, grav2$pheno[,1,drop=FALSE], kloco)
    out2 <- scan1(probs, grav2$pheno[,2,drop=FALSE], kloco)
    out12 <- scan1(probs, grav2$pheno[,1:2], kloco)
    expect_equal(cbind(out1, out2), out12, tolerance=1e-6)

})

test_that("rbind.scan1() works for scan1() results", {

    skip_on_cran()

    phe <- grav2$pheno[,1,drop=FALSE]
    out1 <- scan1(probs[,1], phe)
    out2 <- scan1(probs[,2:3], phe)
    out3 <- scan1(probs[,3], phe)
    out12 <- scan1(probs[,1:3], phe)
    out123 <- scan1(probs[,c(1:3,3)], phe)
    out2123 <- scan1(probs[,c(2:3,1,2:3,3)], phe)

    expect_equal(rbind(out1, out2), out12)
    expect_equal(rbind(out1, out2, out3), out123)
    expect_equal(rbind(out2, out1, out2, out3), out2123)

})


test_that("rbind.scan1() works for scan1() results with multiple columns", {

    skip_on_cran()

    phe <- grav2$pheno[,15:18,drop=FALSE]
    out1 <- scan1(probs[,1], phe)
    out2 <- scan1(probs[,2:3], phe)
    out3 <- scan1(probs[,3], phe)
    out12 <- scan1(probs[,1:3], phe)
    out123 <- scan1(probs[,c(1:3,3)], phe)
    out2123 <- scan1(probs[,c(2:3,1,2:3,3)], phe)

    expect_equal(rbind(out1, out2), out12)
    expect_equal(rbind(out1, out2, out3), out123)
    expect_equal(rbind(out2, out1, out2, out3), out2123)

})


test_that("rbind.scan1() works for scan1() results", {

    skip_on_cran()

    phe <- grav2$pheno[,1,drop=FALSE]
    out1 <- scan1(probs[,1], phe, k)
    out2 <- scan1(probs[,2:3], phe, k)
    out3 <- scan1(probs[,3], phe, k)
    out12 <- scan1(probs[,1:3], phe, k)
    attr(out12, "hsq") <- rbind(attr(out1, "hsq"), attr(out2, "hsq")) # small adjustment
    out123 <- scan1(probs[,c(1:3,3)], phe, k)
    attr(out123, "hsq") <- rbind(attr(out1, "hsq"), attr(out2, "hsq"), attr(out3, "hsq")) # small adjustment
    out2123 <- scan1(probs[,c(2:3,1,2:3,3)], phe, k)
    attr(out2123, "hsq") <- rbind(attr(out2, "hsq"), attr(out1, "hsq"), attr(out2, "hsq"), attr(out3, "hsq")) # small adjustment

    expect_equal(rbind(out1, out2), out12)
    expect_equal(rbind(out1, out2, out3), out123)
    expect_equal(rbind(out2, out1, out2, out3), out2123)

    out1 <- scan1(probs[,1], phe, kloco[1])
    out2 <- scan1(probs[,2:3], phe, kloco[2:3])
    out3 <- scan1(probs[,3], phe, kloco[3])
    out12 <- scan1(probs[,1:3], phe, kloco[1:3])
    out123 <- scan1(probs[,c(1:3,3)], phe, kloco[c(1:3,3)])
    out2123 <- scan1(probs[,c(2:3,1,2:3,3)], phe, kloco[c(2:3,1,2:3,3)])

    expect_equal(rbind(out1, out2), out12)
    expect_equal(rbind(out1, out2, out3), out123)
    expect_equal(rbind(out2, out1, out2, out3), out2123)

})


test_that("rbind.scan1() works for scan1() results with multiple columns", {

    skip_on_cran()

    phe <- grav2$pheno[,15:18,drop=FALSE]
    out1 <- scan1(probs[,1], phe, k)
    out2 <- scan1(probs[,2:3], phe, k)
    out3 <- scan1(probs[,3], phe, k)
    out12 <- scan1(probs[,1:3], phe, k)
    attr(out12, "hsq") <- rbind(attr(out1, "hsq"), attr(out2, "hsq")) # small adjustment
    out123 <- scan1(probs[,c(1:3,3)], phe, k)
    attr(out123, "hsq") <- rbind(attr(out1, "hsq"), attr(out2, "hsq"), attr(out3, "hsq")) # small adjustment
    out2123 <- scan1(probs[,c(2:3,1,2:3,3)], phe, k)
    attr(out2123, "hsq") <- rbind(attr(out2, "hsq"), attr(out1, "hsq"), attr(out2, "hsq"), attr(out3, "hsq")) # small adjustment

    expect_equal(rbind(out1, out2), out12)
    expect_equal(rbind(out1, out2, out3), out123)
    expect_equal(rbind(out2, out1, out2, out3), out2123)

    out1 <- scan1(probs[,1], phe, kloco[1])
    out2 <- scan1(probs[,2:3], phe, kloco[2:3])
    out3 <- scan1(probs[,3], phe, kloco[3])
    out12 <- scan1(probs[,1:3], phe, kloco[1:3])
    out123 <- scan1(probs[,c(1:3,3)], phe, kloco[c(1:3,3)])
    out2123 <- scan1(probs[,c(2:3,1,2:3,3)], phe, kloco[c(2:3,1,2:3,3)])

    expect_equal(rbind(out1, out2), out12)
    expect_equal(rbind(out1, out2, out3), out123)
    expect_equal(rbind(out2, out1, out2, out3), out2123)

})

context("clean scan1 output")

test_that("clean_scan1 works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c("2", "8", "9")]
    pr <- calc_genoprob(iron)
    out <- scan1(pr, iron$pheno)

    # add some messiness
    out[5,1] <- -out[5,1]
    out[7,2] <- -out[7,2]
    out[14,] <- NA

    # expected values:
    expected <- out[-14,]
    expected[5,1] <- NA
    expected[7,2] <- NA
    class(expected) <- class(out)
    attr(expected, "sample_size") <- attr(out, "sample_size")

    expect_equal(clean_scan1(out), expected)
    expect_equal(clean(out), expected)

    expect_equal(clean_scan1(expected), expected)
    expect_equal(clean(expected), expected)

})

context("eigen decomposition of kinship matrix")

test_that("eigen decomposition works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:30, c(1,3,5)]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002) # <- use just 3 chromosomes, for speed

    K <- calc_kinship(probs)
    Ke <- decomp_kinship(K)
    expected <- eigen(K)

    expect_equal(sort(Ke$values), sort(expected$values))
    expect_equivalent(K, t(Ke$vectors) %*% diag(Ke$values) %*% Ke$vectors)

    # run it through again and get same answer
    Ke2 <- decomp_kinship(Ke)
    expect_equal(Ke2, Ke)

    # list of matrices
    K <- calc_kinship(probs, "loco")
    Ke <- decomp_kinship(K)

    for(i in seq(along=K)) {
        expected <- eigen(K[[i]])

        expect_equal(sort(Ke[[i]]$values), sort(expected$values))
        expect_equivalent(K[[i]], t(Ke[[i]]$vectors) %*% diag(Ke[[i]]$values) %*% Ke[[i]]$vectors)
    }

    # run it through again and get same answer
    Ke2 <- decomp_kinship(Ke)
    expect_equal(Ke2, Ke)

    # make sure it gives an error input has 0 rows and 0 cols
    expect_error(decomp_kinship(K[[1]][numeric(0),numeric(0),drop=FALSE]))

    # make sure it gives an error input isn't square
    expect_error(decomp_kinship(K[[1]][1:5,1:4]))

    # make sure it gives an error input isn't a matrix
    expect_error(decomp_kinship(K[[1]][5,5]))

})

test_that("multi-core eigen decomposition works", {
    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    K <- calc_kinship(probs, "loco", cores=2)
    Ke <- decomp_kinship(K)
    Ke_multicore <- decomp_kinship(K, cores=2)
    expect_equal(Ke_multicore, Ke)

    K <- calc_kinship(probs, "chr", cores=2)
    Ke <- decomp_kinship(K)
    Ke_multicore <- decomp_kinship(K, cores=2)
    expect_equal(Ke_multicore, Ke)

})


test_that("eigen decomposition gives error with non-square matrix", {

    k <- matrix(runif(100), ncol=5)
    expect_error(decomp_eigen(k))

    k <- list(k, matrix(runif(100), ncol=20))
    expect_error(decomp_eigen(k))

})

test_that("eigen decomposition works with scan1", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:50, c(1,3,5)]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron[,c(1,3,5)], map, error_prob=0.002) # <- use just 3 chromosomes, for speed

    K <- calc_kinship(probs)
    Ke <- decomp_kinship(K)

    out <- scan1(probs, iron$pheno, K)
    expect_equal(scan1(probs, iron$pheno, Ke), out)

    K_loco <- calc_kinship(probs, "loco")
    Ke_loco <- decomp_kinship(K_loco)

    out_loco <- scan1(probs, iron$pheno, K_loco)
    expect_equal(scan1(probs, iron$pheno, Ke_loco), out_loco)

})

# functions to deal with covariates

# force interactive covariates into the additive covariate matrix
#
# addcovar and intcovar are two matrices
# intcovar columns should all be within the addcovar columns
# tol is tolerance for determining matching columns
#' @importFrom stats complete.cases
force_intcovar <-
    function(addcovar=NULL, intcovar=NULL, tol=1e-12)
{
    if(is.null(intcovar)) # no intcovar, so return addcovar w/o change
        return(addcovar)

    if(is.null(addcovar)) # no addcovar, so just return intcovar
        return(intcovar)

    if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
    if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)

    # IDs in both; omitting any individuals with missing values
    ids <- get_common_ids(addcovar[complete.cases(addcovar),,drop=FALSE],
                          intcovar[complete.cases(intcovar),,drop=FALSE])

    # look for matching columns, having reduced to common individuals
    full <- cbind(addcovar[ids,,drop=FALSE], intcovar[ids,,drop=FALSE])
    has_match <- find_matching_cols(full, tol)
    if(any(has_match > 0))
        full <- full[, has_match<0, drop=FALSE]

    if(ncol(full)==0) return(NULL)

    full
}

# drop linearly dependent columns
# if intercept=TRUE, add intercept before checking and then remove afterwards
#' @importFrom stats complete.cases
drop_depcols <-
    function(covar=NULL, add_intercept=FALSE, tol=1e-12)
{
    if(is.null(covar)) return(covar)

    if(!is.matrix(covar)) covar <- as.matrix(covar)
    if(add_intercept) covar <- cbind(rep(1, nrow(covar)), covar)

    if(ncol(covar) <= 1) return(covar)

    X <- covar[complete.cases(covar),,drop=FALSE]

    # deal with NAs by omitting those rows before
    indep_cols <- sort(find_lin_indep_cols(X, tol))

    if(add_intercept) {

        target_ncol <- length(indep_cols)

        n_it <- 0
        while(!(1 %in% indep_cols)) {
            # don't want to omit the first column (the intercept)
            # need to work harder...
            #  - drop one column at a time other the intercept
            #  - when you find a column that doesn't reduce the target number of columns, drop it
            #  - check again if the intercept is being retained; if not, repeat

            for(i in seq_len(ncol(X))[-1]) { # loop over all but the first column (the intercept)
                indep_cols <- find_lin_indep_cols(X[,-i,drop=FALSE], tol)
                if(length(indep_cols) == target_ncol) { # this one definitely dependent; omit it
                    X <- X[,-i,drop=FALSE]
                    break
                }
            }

            # new set of linearly independent columns
            indep_cols <- sort(find_lin_indep_cols(X, tol))

            n_it <- n_it+1 # count number of iterations and bail if it's large
            if(n_it > ncol(covar)+5) { # something is seriously messed up
                stop("Not above to find set of linearly independent columns")
            }
        }

        # now drop the intercept
        indep_cols <- indep_cols[-1]
    }
    if(length(indep_cols)==0) return(NULL)

    covar[, indep_cols, drop=FALSE]
}

# drop columns from X covariates that are already in addcovar
#' @importFrom stats complete.cases
drop_xcovar <-
    function(covar=NULL, Xcovar=NULL, tol=1e-12)
{
    if(is.null(Xcovar) || is.null(covar)) return(Xcovar)

    if(!is.matrix(covar)) covar <- as.matrix(covar)
    if(!is.matrix(Xcovar)) Xcovar <- as.matrix(Xcovar)

    # IDs in both; omitting any individuals with missing values
    ids <- get_common_ids(covar[complete.cases(covar),,drop=FALSE],
                          Xcovar[complete.cases(Xcovar),,drop=FALSE])

    # look for matching columns, having reduced to common individuals
    matches <- find_matching_cols(cbind(covar[ids,], Xcovar[ids,]), tol)[-seq_len(ncol(covar))]

    if(all(matches > 0)) return(NULL)

    # drop the columns with matches
    Xcovar[,matches<0,drop=FALSE]
}

context("force intcovar into addcovar")

test_that("find_matching_cols works", {

    set.seed(20151130)
    x <- cbind(1, sample(0:1, 200, repl=TRUE))

    expect_equal(find_matching_cols(x), c(-1, -1))
    expect_equal(find_matching_cols(cbind(x, 1)), c(-1, -1, 1))
    expect_equal(find_matching_cols(cbind(x, 2, x[,2])), c(-1, -1, -1, 2))
    expect_equal(find_matching_cols(cbind(x, 2, x[,2]+0.01)), c(-1, -1, -1, -1))
    expect_equal(find_matching_cols(cbind(x, 2, 1-x[,2])), c(-1, -1, -1, -1))

    y <- x
    y[5,1] <- NA
    expect_equal(find_matching_cols(cbind(x,y)), c(-1, -1, -1, 2))
    expect_equal(find_matching_cols(cbind(x,y,y[,1])), c(-1, -1, -1, 2, 3))

})


test_that("force_intcovar works", {

    set.seed(20151130)
    n.ind <- 100
    X <- cbind(1,
               sample(0:1, n.ind, replace=TRUE),
               sample(0:1, n.ind, replace=TRUE))
    rownames(X) <- paste(1:n.ind)

    expect_equal(force_intcovar(NULL, NULL), NULL)

    expect_equal(force_intcovar(X, NULL), X)
    expect_equal(force_intcovar(NULL, X), X)

    expect_equal(force_intcovar(X, X), X)

    for(i in 1:ncol(X))
        expect_equal(force_intcovar(X, X[,i]), X)

    expect_equal(force_intcovar(X[,1], X), X)
    expect_equal(force_intcovar(X[,2], X), X[,c(2,1,3)])
    expect_equal(force_intcovar(X[,3], X), X[,c(3,1,2)])

    expect_equal(force_intcovar(X[,-1], X[,1]), X[,c(2,3,1)])
    expect_equal(force_intcovar(X[,-2], X[,2]), X[,c(1,3,2)])
    expect_equal(force_intcovar(X[,-3], X[,3]), X)

})


test_that("force_intcovar works with missing data", {

    set.seed(20151130)
    n.ind <- 100
    X <- cbind(1,
               sample(0:1, n.ind, replace=TRUE),
               sample(0:1, n.ind, replace=TRUE))
    rownames(X) <- paste(1:n.ind)

    # add a few missing values
    X[1,1] <- X[2:3,2] <- X[3:5,3] <- NA

    expect_equal(force_intcovar(X, NULL), X)
    expect_equal(force_intcovar(NULL, X), X)

    expect_equal(force_intcovar(X, X), X[-(1:5),])

    for(i in 1:ncol(X))
        expect_equal(force_intcovar(X, X[,i]), X[-(1:5),])

    expect_equal(force_intcovar(X[,1], X), X[-(1:5),])
    expect_equal(force_intcovar(X[,2], X), X[-(1:5),c(2,1,3)])
    expect_equal(force_intcovar(X[,3], X), X[-(1:5),c(3,1,2)])

    expect_equal(force_intcovar(X[,-1], X[,1]), X[-(1:5),c(2,3,1)])
    expect_equal(force_intcovar(X[,-2], X[,2]), X[-(1:5),c(1,3,2)])
    expect_equal(force_intcovar(X[,-3], X[,3]), X[-(1:5),])

    expect_equal(force_intcovar(X[,1], X[,2]), X[-(1:3),-3])
    expect_equal(force_intcovar(X[,1], X[,3]), X[-c(1,3:5),-2])
    expect_equal(force_intcovar(X[,2], X[,1]), X[-(1:3),c(2,1)])
    expect_equal(force_intcovar(X[,2], X[,3]), X[-(2:5), -1])
    expect_equal(force_intcovar(X[,3], X[,1]), X[-c(1,3:5),c(3,1)])
    expect_equal(force_intcovar(X[,3], X[,2]), X[-(2:5),c(3,2)])

    # try some permutations
    expect_equal(force_intcovar(X, X[sample(1:nrow(X)),]), X[-(1:5),])
    expect_equal(force_intcovar(X[,-1], X[sample(1:nrow(X)),1]), X[-(1:5),c(2,3,1)])

})

context("drop linearly dependent columns")

test_that("find_lin_indep_cols works", {

    set.seed(20151130)
    x <- cbind(1, sample(0:1, 200, repl=TRUE))

    expect_equal(sort(find_lin_indep_cols(x)), c(1,2))
    expect_equal(sort(find_lin_indep_cols(cbind(x, 1))), c(1,2))
    expect_equal(sort(find_lin_indep_cols(cbind(x, x[,1] + 0.5*x[,2]))), c(2,3))

    X <- matrix(rnorm(1000*10), ncol=10)
    expect_equal(sort(find_lin_indep_cols(X)), 1:10)
    expect_equal(length(find_lin_indep_cols(cbind(rowSums(X), X))), 10)

})

test_that("drop_depcols works", {

    set.seed(20151130)
    n.ind <- 100
    x <- cbind(1,
               sample(0:1, n.ind, replace=TRUE),
               sample(0:1, n.ind, replace=TRUE))

    expect_equal(drop_depcols(NULL), NULL)
    expect_equal(drop_depcols(x), x)
    for(i in 1:ncol(x))
        expect_equal(drop_depcols(x[,i]), x[,i,drop=FALSE])

    X <- cbind(rowSums(x), x)
    expect_equal(ncol(drop_depcols(X)), 3)

})

test_that("drop_xcovar works", {

    set.seed(20151202)
    n <- 100
    Xcovar <- sample(0:1, n, replace=TRUE)
    int <- rep(1, n)
    names(Xcovar) <- names(int) <- paste(1:n)

    expect_equal( drop_xcovar(NULL, NULL), NULL)
    expect_equal( drop_xcovar(NULL, Xcovar), Xcovar)
    expect_equal( drop_xcovar(int, NULL), NULL)
    expect_equal( drop_xcovar(int, Xcovar), as.matrix(Xcovar))
    expect_equal( drop_xcovar(Xcovar, Xcovar), NULL)
    expect_equal( drop_xcovar(cbind(int, Xcovar), Xcovar), NULL)
    expect_equal( drop_xcovar(cbind(Xcovar, int), Xcovar), NULL)

    pgm <- sample(0:1, n, replace=TRUE)
    names(pgm) <- paste(1:n)
    Xcovar <- cbind(Xcovar, pgm)

    expect_equal( drop_xcovar(NULL, Xcovar), Xcovar)
    expect_equal( drop_xcovar(int, Xcovar), Xcovar)
    expect_equal( drop_xcovar(Xcovar, Xcovar), NULL)
    expect_equal( drop_xcovar(cbind(int, Xcovar), Xcovar), NULL)
    expect_equal( drop_xcovar(cbind(int, Xcovar[,1]), Xcovar), Xcovar[,2,drop=FALSE])
    expect_equal( drop_xcovar(cbind(int, Xcovar[,2]), Xcovar), Xcovar[,1,drop=FALSE])
    expect_equal( drop_xcovar(cbind(Xcovar, int), Xcovar), NULL)
    expect_equal( drop_xcovar(cbind(Xcovar[,1], int), Xcovar), Xcovar[,2,drop=FALSE])
    expect_equal( drop_xcovar(cbind(Xcovar[,2], int), Xcovar), Xcovar[,1,drop=FALSE])

})

test_that("drop_xcovar works with NAs", {

    set.seed(20151202)
    n <- 100
    Xcovar <- sample(0:1, n, replace=TRUE)
    int <- rep(1, n)
    names(Xcovar) <- names(int) <- paste(1:n)
    Xcovar[3:5] <- NA

    expect_equal( drop_xcovar(NULL, Xcovar), Xcovar)
    expect_equal( drop_xcovar(int, Xcovar), as.matrix(Xcovar))
    expect_equal( drop_xcovar(Xcovar, Xcovar), NULL)
    expect_equal( drop_xcovar(cbind(int, Xcovar), Xcovar), NULL)
    expect_equal( drop_xcovar(cbind(Xcovar, int), Xcovar), NULL)

    # try some permutations
    o <- sample(1:n)
    expect_equal( drop_xcovar(int, Xcovar[o]), as.matrix(Xcovar[o]))
    expect_equal( drop_xcovar(Xcovar, Xcovar[o]), NULL)
    expect_equal( drop_xcovar(cbind(int, Xcovar), Xcovar[o]), NULL)
    expect_equal( drop_xcovar(cbind(Xcovar, int), Xcovar[o]), NULL)

    # multiple X covar columns
    pgm <- sample(0:1, n, replace=TRUE)
    names(pgm) <- paste(1:n)
    Xcovar <- cbind(Xcovar, pgm)

    expect_equal( drop_xcovar(NULL, Xcovar), Xcovar)
    expect_equal( drop_xcovar(int, Xcovar), Xcovar)
    expect_equal( drop_xcovar(Xcovar, Xcovar), NULL)
    expect_equal( drop_xcovar(cbind(int, Xcovar), Xcovar), NULL)
    expect_equal( drop_xcovar(cbind(int, Xcovar[,1]), Xcovar), Xcovar[,2,drop=FALSE])
    expect_equal( drop_xcovar(cbind(int, Xcovar[,2]), Xcovar), Xcovar[,1,drop=FALSE])
    expect_equal( drop_xcovar(cbind(Xcovar, int), Xcovar), NULL)
    expect_equal( drop_xcovar(cbind(Xcovar[,1], int), Xcovar), Xcovar[,2,drop=FALSE])
    expect_equal( drop_xcovar(cbind(Xcovar[,2], int), Xcovar), Xcovar[,1,drop=FALSE])

    # try some permutations
    o <- sample(1:n)
    expect_equal( drop_xcovar(int, Xcovar[o,]), Xcovar[o,])
    expect_equal( drop_xcovar(Xcovar, Xcovar[o,]), NULL)
    expect_equal( drop_xcovar(cbind(int, Xcovar), Xcovar[o,]), NULL)
    expect_equal( drop_xcovar(cbind(Xcovar, int), Xcovar[o,]), NULL)
    expect_equal( drop_xcovar(cbind(int, Xcovar[,1]), Xcovar[o,]), Xcovar[o,2,drop=FALSE])
    expect_equal( drop_xcovar(cbind(int, Xcovar[,2]), Xcovar[o,]), Xcovar[o,1,drop=FALSE])
    expect_equal( drop_xcovar(cbind(Xcovar, int), Xcovar[o,]), NULL)
    expect_equal( drop_xcovar(cbind(Xcovar[,1], int), Xcovar[o,]), Xcovar[o,2,drop=FALSE])
    expect_equal( drop_xcovar(cbind(Xcovar[,2], int), Xcovar[o,]), Xcovar[o,1,drop=FALSE])

})

test_that("Fixed issue 22 re drop_depcols ", {

    # https://github.com/rqtl/qtl2/issues/22
    # ...case where indicators for all categories included

    # create example
    n_per <- 20
    n_group <- 4
    n <- n_per*n_group
    group <- rep(1:n_group, each=n_per)
    covar <- cbind(sex=rep(0:1,n_per*n_group),
                   diet=sample(100:200, n, replace=TRUE))
    for(i in 1:n_group)
        covar <- cbind(covar, (group==i)*1)
    colnames(covar)[-(1:2)] <- paste0("group", 1:n_group)

    # drop dependent columns
    covar_indep <- drop_depcols(covar, TRUE)

    # check that one of the groups got dropped
    expect_true("sex" %in% colnames(covar_indep))
    expect_true("diet" %in% colnames(covar_indep))
    expect_true(ncol(covar_indep) == ncol(covar)-1)

})

# lod_int
#' Calculate LOD support intervals
#'
#' Calculate LOD support intervals for a single LOD curve on a single
#' chromosome, with the ability to identify intervals for multiple LOD
#' peaks.
#'
#' @param scan1_output An object of class `"scan1"` as returned by
#' [scan1()].
#' @param map A list of vectors of marker positions, as produced by
#' [insert_pseudomarkers()].
#' @param chr Chromosome ID to consider (must be a single value).
#' @param lodcolumn LOD score column to consider (must be a single value).
#' @param threshold Minimum LOD score for a peak.
#' @param peakdrop Amount that the LOD score must drop between peaks,
#' if multiple peaks are to be defined on a chromosome.
#' @param drop Amount to drop in the support interval.  Must be
#' \eqn{\le} `peakdrop`
#' @param expand2markers If TRUE, QTL intervals are expanded so
#' that their endpoints are at genetic markers.
#'
#' @return A matrix with three columns:
#' * `ci_lo` - lower bound of interval
#' * `pos` - peak position
#' * `ci_hi` - upper bound of interval
#'
#' Each row corresponds to a different peak.
#'
#' @details We identify a set of peaks defined as local maxima that
#' exceed the specified `threshold`, with the requirement that
#' the LOD score must have dropped by at least `peakdrop` below
#' the lowest of any two adjacent peaks.
#'
#' At a given peak, if there are ties, with multiple positions jointly
#' achieving the maximum LOD score, we take the average of these
#' positions as the location of the peak.
#'
#' The default is to use `threshold=0`, `peakdrop=Inf`, and
#' `drop=1.5`. We then return results a single peak, no matter the
#' maximum LOD score, and give a 1.5-LOD support interval.
#'
#' @export
#'
#' @seealso [bayes_int()], [find_peaks()], [scan1()]
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,7]}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' # 1.5-LOD support interval for QTL on chr 7, first phenotype
#' lod_int(out, map, chr=7, lodcolum=1)
lod_int <-
    function(scan1_output, map, chr=NULL, lodcolumn=1, threshold=0,
             peakdrop=Inf, drop=1.5, expand2markers=TRUE)
{
    # align scan1_output and map
    tmp <- align_scan1_map(scan1_output, map)
    scan1_output <- tmp$scan1
    map <- tmp$map

    if(nrow(scan1_output) != length(unlist(map)))
        stop("nrow(scan1_output) [", nrow(scan1_output), "] != number of positions in map [",
             length(unlist(map)), "]")

    if(is.null(chr)) { # just use the first chr
        chr <- names(map)[1]
    }

    if(length(chr) > 1) {
        warning("chr should have length 1; using the first value")
        chr <- as.character(chr[1])
    }
    if(length(lodcolumn) == 0) stop("lodcolumn has length 0")
    if(length(lodcolumn) > 1) {
        warning("lodcolumn should have length 1; using the first value")
        lodcolumn <- lodcolumn[1]
    }
    if(length(threshold) == 0) stop("threshold has length 0")
    if(length(threshold) > 1) {
        warning("threshold should have length 1; using the first value")
        threshold <- threshold[1]
    }
    if(length(peakdrop) == 0) stop("peakdrop has length 0")
    if(length(peakdrop) > 1) {
        warning("peakdrop should have length 1; using the first value")
        peakdrop <- peakdrop[1]
    }
    if(length(drop) == 0) stop("drop has length 0")
    if(length(drop) > 1) {
        warning("drop should have length 1; using the first value")
        drop <- drop[1]
    }

    if(drop > peakdrop)
        stop("Must have drop <= peakdrop")

    if(is.character(lodcolumn)) {
        if(!(lodcolumn %in% colnames(scan1_output)))
            stop('lodcolumn "', lodcolumn, '" not found')
        lodcolumn <- match(lodcolumn, colnames(scan1_output))
    }
    if(lodcolumn < 1 || lodcolumn > ncol(scan1_output))
        stop("lodcolumn should be between 1 and ", ncol(scan1_output))

    # for chr to be character string
    chr <- as.character(chr)

    if(!(chr %in% names(map)))
        stop("Chromosome ", chr, " not found")

    scan1_output <- subset_scan1(scan1_output, map, chr, lodcolumn)

    lod <- unclass(scan1_output)
    map <- map[[chr]]

    peaks <- .find_peaks_and_lodint(lod, threshold, peakdrop, drop)
    n_peaks <- length(peaks)
    if(n_peaks==0) {
        result <- matrix(nrow=0, ncol=3)
        colnames(result) <- c("ci_lo", "pos", "ci_hi")
        return(result)
    }

    # calculate peak positions
    # (this deals with ties in the LOD scores:
    #  take average of multiple positions sharing the maximum LOD score)
    # and remember that .find_peaks returns indexes starting at 0
    ci_lo <- ci_hi <- rep(0, n_peaks)
    for(i in seq_len(n_peaks)) {
        if(expand2markers)
            ci <- expand_interval_to_markers(peaks[[i]][1:2]+1, map)
        else
            ci <- map[peaks[[i]][1:2]+1]
        ci_lo[i] <- ci[1]
        ci_hi[i] <- ci[2]
    }
    peak_pos <- vapply(peaks, function(a) mean(map[a[-(1:2)]+1]), 0.0)

    result <- cbind(ci_lo=ci_lo,
                    pos=peak_pos,
                    ci_hi=ci_hi)
    rownames(result) <- seq_len(nrow(result))

    result
}

#' cis-trans plot for eQTL results
#'
#' Scatterplot of gene location vs QTL location to display eQTL/pQTL results
#'
#' @param peaks Data frame of QTL results, as output by [find_peaks()]
#'     Should contain columns `lodcolumn`, `chr`, and `pos`.
#'
#' @param map Marker map, as a list of chromosomes, each being a
#'     vector of positions; used to get chromosome names and start/end
#'
#' @param pheno_pos Data frame with chromosome, position, and name of
#'     the phenotypes in `peaks`. Should contain columns `pheno`,
#'     `chr`, and `pos`.
#'
#' @param gap Gap between chromosomes in the plot
#'
#' @param cis_window Window size that defines cis-QTL (on same
#'     chromosome, and position of gene within this distance of QTL
#'     position)
#'
#' @param pch Point type for "trans" points
#'
#' @param col Point color for "trans" points
#'
#' @param bg Background color for "trans" points
#'
#' @param cex Character expansion size for "trans" points
#'
#' @param pch_cis Point type for "cis" points
#'
#' @param col_cis Point color for "cis" points
#'
#' @param bg_cis Background color for "cis" points
#'
#' @param cex_cis Character expansion size for "cis" points
#'
#' @param pty Plot type; the default "s" forces a square plot; use "m"
#'     to use the maximal plotting region.
#'
#' @param bgcolor Background color for chromosome rectangles
#'
#' @param altbgcolor Alternate background color for chromosome rectangles
#'
#' @param ... Additional graphics arguments passed to [graphics::points()]
#'
#' @return Invisibly returns a data frame with `peaks` merged with
#'     `pheno_pos`, with an additional column `cis` that indicates
#'     which QTL appear to be cis (vs trans).
#'
#' @seealso [plot_scan1()], [find_peaks()], [plot_lodpeaks()]
#'
#' @importFrom graphics rect par axis title points
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # download example pQTL results (from Keele et al. 2026, https://doi.org/10.1016/j.xgen.2025.101069)
#' # contains qtl, map, and pheno_pos
#' url <- "https://kbroman.org/qtl2/assets/sampledata/pqtl_data.RData"
#' tempfile <- file.path(tempdir(), basename(url))
#' download.file(url, tempfile)
#' load(tempfile)
#' unlink(tempfile)
#'
#' plot_cistrans(qtl, map, pheno_pos, cis_window=5)
#' }

plot_cistrans <-
    function(peaks, map, pheno_pos, gap=0, cis_window=1,
             pch=21, col="slateblue", bg="slateblue", cex=0.8,
             pch_cis=21, col_cis="violetred", bg_cis="violetred", cex_cis=0.8,
             pty="s", bgcolor="gray90", altbgcolor="gray80", ...)
{
    # check that peaks conforms to expectations
    if(!is.data.frame(peaks) || !all(c("lodcolumn", "chr", "pos") %in% colnames(peaks)))
        stop("peaks should be a data frame with columns lodcolumn, chr, and pos")
    # check that map conforms to expectations
    stopifnot(is.list(map))
    # check that pheno_pos conforms to expectations
    if(!is.data.frame(pheno_pos) || !all(c("pheno", "chr", "pos") %in% colnames(pheno_pos)))
        stop("peaks should be a data frame with columns pheno, chr, and pos")
    # check that the phenotypes in peaks are also in pheno_pos
    if(!all(peaks$lodcolumn %in% pheno_pos$pheno))
        stop("Not all lodcolumns in peaks are found in the pheno column of pheno_pos")

    # merge peaks and pheno_pos
    peaks <- base::merge(peaks, pheno_pos, by.x="lodcolumn", by.y="pheno")
    # drop loci that don't map to the provided map
    peaks <- peaks[peaks$chr.x %in% names(map) & peaks$chr.y %in% names(map), ]
    # should give xpos_scan1 a "strict" argument, and make it NA if outside of chr's range
    peaks$xpos <- xpos_scan1(map, gap=gap, thechr=peaks$chr.x, thepos=peaks$pos.x)
    peaks$ypos <- xpos_scan1(map, gap=gap, thechr=peaks$chr.y, thepos=peaks$pos.y)

    plot_cistrans_internal <-
        function(peaks, map, pheno_pos, gap=0, pty="s", xlab="QTL position",
                 ylab="Gene position", xlim=NULL, ylim=NULL,
                 mgp=NULL, mgp.x=NULL, mgp.y=NULL,
                 bgcolor="gray90", altbgcolor="gray80",
                 col=col, bg = bg, pch=pch, las=1, cex=cex,
                 xaxt="s", yaxt="s", type="p", ...)
    {
        # get x- and y-axis range
        if(is.null(xlim) || is.null(ylim)) {
            start <- xpos_scan1(map, chr=names(map), gap=gap,
                                names(map)[1], min(map[[1]]))
            end <- xpos_scan1(map, chr=names(map), gap=gap,
                              names(map)[length(map)], max(map[[length(map)]]))

            if(is.null(xlim)) xlim <- c(start, end)
            if(is.null(ylim)) ylim <- c(start, end)
        }

        if(is.null(mgp.x)) {
            if(!is.null(mgp)) mgp.x <- mgp
            else mgp.x <- c(1.7, 0.3, 0)
        }
        if(is.null(mgp.y)) {
            if(!is.null(mgp)) mgp.y <- mgp
            else mgp.y <- c(2.0, 0.4, 0)
        }


        par(pty=pty)
        plot(0,0, type="n", xlim=xlim, ylim=ylim, xaxs="i", yaxs="i",
             xlab="", ylab="", xaxt="n", yaxt="n")
        chr <- names(map)
        chr_midpt <- sapply(seq_along(map), function(i) xpos_scan1(map, names(map), gap=gap,
                                                                   names(map)[i], mean(range(map[[i]], na.rm=TRUE))))

        for(i in seq_along(chr_midpt)) {
            if(xaxt != "n") graphics::axis(side=1, at=chr_midpt[i], chr[i], mgp=mgp.x, tick=FALSE, las=las)
            if(yaxt != "n") graphics::axis(side=2, at=chr_midpt[i], chr[i], mgp=mgp.y, tick=FALSE, las=las)
        }
        graphics::title(xlab=xlab, mgp=mgp.x)
        graphics::title(ylab=ylab, mgp=mgp.y)

        # add background rectangles
        chrstart <- xpos_scan1(map, gap=gap, thechr=names(map), thepos=sapply(map, min))
        chrend <- xpos_scan1(map, gap=gap, thechr=names(map), thepos=sapply(map, max))

        rect(min(chrstart), min(chrstart), max(chrend), max(chrend), col=bgcolor, border=NA)
        for(chri in seq_along(map)) {
            for(chrj in seq_along(map)) {
                if((chri+chrj) %% 2 == 1) {
                    rect(chrstart[chri], chrstart[chrj], chrend[chri], chrend[chrj],
                         col=altbgcolor, border=NA)
                }
            }
        }

        # plot points
        cis <- (!is.na(peaks$chr.x) & !is.na(peaks$chr.y) & !is.na(peaks$pos.x) & !is.na(peaks$pos.y) &
                peaks$chr.x == peaks$chr.y & abs(peaks$pos.x - peaks$pos.y) < cis_window)

        if(type != "n") {
            if(any(!cis)) points(peaks$xpos[!cis], peaks$ypos[!cis], pch=pch, bg=bg, col=col, cex=cex, ...)
            if(any(cis)) points(peaks$xpos[cis], peaks$ypos[cis], pch=pch_cis, bg=bg_cis, col=col_cis, cex=cex_cis, ...)
        }

        peaks$cis <- cis
        invisible(peaks)
    }

    plot_cistrans_internal(peaks, map, pheno_pos, gap=0, pty=pty,
                           bgcolor=bgcolor, altbgcolor=altbgcolor,
                           pch=pch, cex=cex, bg=bg, col=col,
                           ...)
}

#' Clean an object
#'
#' Clean an object by removing messy values
#'
#' @param object Object to be cleaned
#' @param ... Other arguments
#' @return Input object with messy values cleaned up
#'
#' @export
#' @seealso [clean.scan1()], [clean.calc_genoprob()]
clean <-
    function(object, ...)
{
    UseMethod("clean")
}

#  grayplot (internal function taken from github.com/kbroman/broman)
#
# Scatterplot with a gray background
#
# Like the plot function, but using a gray background just
#   for the plot region.
#
# @param x Coordinates of points in the plot
#
# @param y Coordinates of points in the plot (optional)
#
# @param ... Optional graphics arguments
#
# @param type Plot type (points, lines, etc.)
#
# @param hlines Locations of horizontal grid lines; use `hlines=NA` to prevent horizontal grid lines
#
# @param hlines.col Colors of horizontal grid lines
#
# @param hlines.lty Line type of horizontal grid lines
#
# @param hlines.lwd Line width of horizontal grid lines
#
# @param vlines Locations of vertical grid lines; use `vlines=NA` to prevent vertical grid lines
#
# @param vlines.col Colors of vertical grid lines
#
# @param vlines.lty Line type of vertical grid lines
#
# @param vlines.lwd Line width of vertical grid lines
#
# @param xat Locations for x-axis labels; `xat=NA` indicates no labels
#
# @param yat Locations for y-axis labels; `yat=NA` indicates no labels
#
# @param bgcolor Background color
#
# @param pch point type
# @param bg Background color in points
# @param col Color of outer circle in points
#
# @param v_over_h If `TRUE`, place vertical grid lines on top of
# the horizontal ones.
#
# @details
# Calls `plot()` with `type="n"`, then [graphics::rect()] to
# get the background, and then [graphics::points()]. Additional
# arguments you can include: `mgp.x` and `mgp.y` (like `mgp`, for
# controlling parameters of axis labels, but separate for x- and
# y-axis).
#
# @importFrom graphics title rect axis abline points
#
# @return
# None.
#
# @examples
# \dontshow{set.seed(97536917)}
# x <- rnorm(100)
# y <- x+rnorm(100, 0, 0.7)
# grayplot(x, y, col="slateblue", pch=16)
# at <- seq(-3, 3)
# grayplot(x, y, col="violetred", pch=16, hlines=at, vlines=at)
# grayplot(x, col="Orchid", pch=16, bgcolor="gray80",
#          hlines=seq(-4, 4, by=0.5), hlines.lwd=c(3,1),
#          vlines=seq(0, 100, by=5), vlines.lwd=c(3,1,1,1))
#
# @seealso
# [dotplot()], [timeplot()], [graphics::par()], [graphics::rect()], [graphics::points()]
#
# @keywords
# graphics
qtl2_grayplot <-
    function(x, y=NULL, ..., type="p", hlines=NULL, hlines.col="white", hlines.lty=1, hlines.lwd=1,
             vlines=NULL, vlines.col="white", vlines.lty=1, vlines.lwd=1,
             xat=NULL, yat=NULL, bgcolor="gray90",
             pch=21, bg="lightblue", col="black",
             v_over_h=FALSE)
{
    if(missing(x) || is.null(x)) stop("x unspecified")

    # this is to deal with varying inputs (did "..." include xaxt or not?)
    hidegrayplot <-
        function(x, y, ..., type="p", hlines=NULL, hlines.col, hlines.lty, hlines.lwd,
                 vlines=NULL, vlines.col, vlines.lty, vlines.lwd,
                 xat=pretty(x), yat=pretty(y), bgcolor="gray90", xaxt="n", yaxt="n",
                 col.lab=par("col.lab"),
                 xlim=NULL, ylim=NULL,
                 xlab, ylab, xname, yname,
                 las=1, mgp=c(2.1, 0.5, 0), mgp.x=NULL, mgp.y=NULL,
                 pch=21, bg="lightblue", col="black",
                 v_over_h=FALSE)
        {
            if(is.null(mgp.x)) mgp.x <- mgp
            if(is.null(mgp.y)) mgp.y <- mgp

            if(is.null(y)) {
                if(missing(xlab)) xlab <- "Index"
                if(missing(ylab)) ylab <- xname
                y <- x
                x <- seq(along=x)
            }
            else {
                if(missing(xlab)) xlab <- xname
                if(missing(ylab)) ylab <- yname
            }

            if(is.null(ylim))
                ylim <- range(y, na.rm=TRUE)
            if(is.null(hlines)) {
                if(!is.null(yat))
                    hlines <- yat
                else
                    hlines <- pretty(ylim)
            }
            else if(length(hlines)==1 && is.na(hlines))
                hlines <- NULL

            if(is.null(xlim))
                xlim <- range(x, na.rm=TRUE)
            if(is.null(vlines)) {
                if(!is.null(xat))
                    vlines <- xat
                else
                    vlines <- pretty(xlim)
            }
            else if(length(vlines)==1 && is.na(vlines))
                vlines <- NULL

            # blank plot
            plot(x, y, ..., type="n", xaxt="n", yaxt="n", xlab="", ylab="",
                 xlim=xlim, ylim=ylim)

            # axis titles
            graphics::title(xlab=xlab, mgp=mgp.x, col.lab=col.lab)
            graphics::title(ylab=ylab, mgp=mgp.y, col.lab=col.lab)

            # add gray rectangle
            u <- par("usr")
            graphics::rect(u[1], u[3], u[2], u[4], col=bgcolor, border="black")

            # x axis: if adding white lines, skip the tick marks and move the numbers closer
            if(!(!is.null(xat) && length(xat)==1 && is.na(xat))) { # if a single NA, skip x-axis
                if(!is.null(xat)) {
                    if(!is.null(vlines))
                        graphics::axis(side=1, at=xat, mgp=mgp.x, tick=FALSE, las=las)
                    else
                        graphics::axis(side=1, at=xat, las=las)
                }
                else {
                    if(!is.null(vlines))
                        graphics::axis(side=1, mgp=mgp.x, tick=FALSE, las=las)
                    else
                        graphics::axis(side=1, las=las)
                }
            }

            # y axis: like the x-axis
            if(!(!is.null(yat) && length(yat)==1 && is.na(yat))) { # if a single NA, skip y-axis
                if(!is.null(yat)) {
                    if(!is.null(hlines))
                        graphics::axis(side=2, at=yat, mgp=mgp.y, tick=FALSE, las=las)
                    else
                        graphics::axis(side=2, at=yat, las=las)
                }
                else {
                    if(!is.null(hlines))
                        graphics::axis(side=2, mgp=mgp.y, tick=FALSE, las=las)
                    else
                        graphics::axis(side=2, las=las)
                }
            }

            if(!is.null(vlines) && !v_over_h)
                graphics::abline(v=vlines, col=vlines.col, lty=vlines.lty, lwd=vlines.lwd)
            if(!is.null(hlines))
                graphics::abline(h=hlines, col=hlines.col, lty=hlines.lty, lwd=hlines.lwd)
            if(!is.null(vlines) && v_over_h)
                graphics::abline(v=vlines, col=vlines.col, lty=vlines.lty, lwd=vlines.lwd)

            graphics::points(x, y, ..., pch=pch, bg=bg, col=col, type=type)

            # add black border again
            graphics::abline(v=u[1:2], h=u[3:4])
        }

    hidegrayplot(x=x, y=y, ..., type=type, hlines=hlines, hlines.col=hlines.col,
                 hlines.lty=hlines.lty, hlines.lwd=hlines.lwd,
                 vlines=vlines, vlines.col=vlines.col,
                 vlines.lty=vlines.lty, vlines.lwd=vlines.lwd,
                 xat=xat, yat=yat, bgcolor=bgcolor,
                 pch=pch, bg=bg, col=col,
                 xname=substitute(x), yname=substitute(y),
                 v_over_h=v_over_h)
    invisible()
}

# two-panel plot with both coefficients and LOD scores
# (for a single chromosome)
#
# calls plot_coef and plot_scan1
# internal function that is called by plot_coef
plot_coef_and_lod <-
    function(x, map, columns=NULL, col=NULL, scan1_output,
             gap=NULL, top_panel_prop=0.65, legend=NULL,
             ylim=NULL, bgcolor="gray90", altbgcolor="gray85",
             ylab="QTL effects",
             ylab_lod="LOD score", ylim_lod=NULL, col_lod="slateblue",
             xaxt=NULL,
             vlines=NULL, vlines_col="white", vlines_lwd=1, vlines_lty=1,
             legend_loc=NULL, legend_ncol=NULL,
             lodcolumn=1, main="", sub="", ...)
{
    if(is.null(map)) stop("map is NULL")

    # align scan1 output and map
    tmp <- align_scan1_map(x, map)
    x <- tmp$scan1
    map <- tmp$map

    if(nrow(x) != length(unlist(map)))
        stop("nrow(x) [", nrow(x), "] != number of positions in map [",
             length(unlist(map)), "]")

    # also, match markers and use map in coefficients object
    mar_in_coef <- rownames(x)
    mar_in_scan1 <- rownames(scan1_output)
    scan1_output <- unclass(scan1_output)[mar_in_scan1 %in% mar_in_coef, , drop=FALSE]
    mis_mar <- !(mar_in_coef %in% mar_in_scan1)
    if(any(mis_mar)) {
        n_new <- sum(mis_mar)
        new_lod <- matrix(NA, nrow=sum(mis_mar), ncol=scan1_output$lod)
        rownames(new_lod) <- mar_in_coef[mis_mar]
        scan1_output <- rbind(scan1_output, new_lod)[mar_in_coef,]
    }

    # 2 x 1 panels; adjust margins
    old_mfrow <- par("mfrow")
    old_mar <- par("mar")
    on.exit(par(mfrow=old_mfrow, mar=old_mar))
    layout(rbind(1,2), heights=c(top_panel_prop, 1-top_panel_prop))
    top_mar <- bottom_mar <- old_mar
    top_mar[1] <- 0.1
    bottom_mar[3] <- 0.1

    par(mar=top_mar)
    plot_coef(x, map, columns=columns, col=col, scan1_output=NULL,
              add=FALSE, gap=gap, ylim=ylim, bgcolor=bgcolor,
              altbgcolor=altbgcolor, ylab=ylab,
              xaxt="n", vlines=vlines, vlines_col=vlines_col,
              vlines_lwd=vlines_lwd, vlines_lty=vlines_lty,
              legend=legend, legend_loc=legend_loc, legend_ncol=legend_ncol,
              main=main, ...)

    par(mar=bottom_mar)
    plot_scan1(scan1_output, map, lodcolumn=lodcolumn, col=col_lod, ylab=ylab_lod,
               add=FALSE, gap=gap, bgcolor=bgcolor, altbgcolor=altbgcolor,
               vlines=vlines, vlines_col=vlines_col, ylim=ylim_lod,
               vlines_lwd=vlines_lwd, vlines_lty=vlines_lty, sub=sub, ...)
}

#' Replace individual IDs
#'
#' Replace the individual IDs in an object with new ones
#'
#' @param x Object whose IDs will be replaced
#'
#' @param ids Vector of character strings with the new individual IDs, with the names being the original IDs.
#'
#' @return The input `x` object, but with individual IDs replaced.
#'
#' @importFrom stats setNames
#' @export
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' ids <- as.numeric(ind_ids(iron))
#'
#' # replace the numeric IDs with IDs like "mouse003"
#' new_ids <- setNames( sprintf("mouse%03d", as.numeric(ids)), ids)
#'
#' iron <- replace_ids(iron, new_ids)
replace_ids <- function(x, ids) UseMethod("replace_ids")


# check that the new ids vector is okay
check_new_ids <-
    function(new_ids, old_ids)
{
    if(length(new_ids) != length(unique(new_ids)))
        stop("The new IDs are not unique")

    new_ids_names <- names(new_ids)
    if(is.null(new_ids_names)) stop("The vector of new ids needs names, indicating the corresponding old IDs")
    if(length(new_ids_names) != length(unique(new_ids_names)))
        stop("The names of the new IDs are not unique")

    if(length(old_ids) != length(unique(old_ids)))
        stop("The old IDs are not unique")

    if(!all(new_ids_names %in% old_ids)) {
        warning("The new IDs include some extras")
        new_ids <- new_ids[new_ids_names %in% old_ids]
        new_ids_names <- names(new_ids)
    }

    if(!all(old_ids %in% new_ids_names)) {
        warning("Some old IDs are missing from the new IDs object")
    }

    new_ids
}


#' @describeIn replace_ids Replace IDs in a `"cross2"` object
#' @export
replace_ids.cross2 <-
    function(x, ids)
{
    ids <- check_new_ids(ids, ind_ids(x))
    ids_names <- names(ids)
    names(ids) <- NULL

    # rownames for each chr in geno
    for(i in seq_along(x$geno)) {
        m <- match(rownames(x$geno[[i]]), ids_names)
        if(any(is.na(m))) x$geno[[i]] <- x$geno[[i]][!is.na(m),,drop=FALSE]
        rownames(x$geno[[i]]) <- ids[m[!is.na(m)]]
    }

    # rownames in pheno, covar, cross_info
    for(obj in c("pheno", "covar", "cross_info")) {
        if(!(obj %in% names(x))) next

        m <- match(rownames(x[[obj]]), ids_names)
        if(any(is.na(m))) x[[obj]] <- x[[obj]][!is.na(m),,drop=FALSE]
        rownames(x[[obj]]) <- ids[m[!is.na(m)]]
    }

    # names in is_female
    for(obj in c("is_female")) {
        if(!obj %in% names(x)) next

        m <- match(names(x[[obj]]), ids_names)
        if(any(is.na(m))) x[[obj]] <- x[[obj]][!is.na(m)]
        names(x[[obj]]) <- ids[m[!is.na(m)]]
    }

    x
}




#' @describeIn replace_ids Replace IDs in output from [calc_genoprob()]
#' @export
replace_ids.calc_genoprob <-
    function(x, ids)
{
    ids <- check_new_ids(ids, rownames(x[[1]]))
    ids_names <- names(ids)
    names(ids) <- NULL

    # replace the row names for each chromosome; possibly subsetting things
    for(i in seq_along(x)) {
        m <- match(rownames(x[[i]]), ids_names)
        x[[i]] <- x[[i]][!is.na(m),,,drop=FALSE]
        rownames(x[[i]]) <- ids[m[!is.na(m)]]
    }

    x
}

#' @describeIn replace_ids Replace IDs in output from [viterbi()]
#' @export
replace_ids.viterbi <-
    function(x, ids)
{
    ids <- check_new_ids(ids, rownames(x[[1]]))
    ids_names <- names(ids)
    names(ids) <- NULL

    # replace the row names for each chromosome; possibly subsetting things
    for(i in seq_along(x)) {
        m <- match(rownames(x[[i]]), ids_names)
        x[[i]] <- x[[i]][!is.na(m),,drop=FALSE]
        rownames(x[[i]]) <- ids[m[!is.na(m)]]
    }

    x

}

#' @describeIn replace_ids Replace IDs in output from [sim_geno()]
#' @export
replace_ids.sim_geno <- function(x, ids) replace_ids.calc_genoprob(x, ids)

#' @describeIn replace_ids Replace IDs in a matrix
#' @export
replace_ids.matrix <-
    function(x, ids)
{
    ids <- check_new_ids(ids, rownames(x))
    ids_names <- names(ids)
    names(ids) <- NULL

    m <- match(rownames(x), ids_names)
    x <- x[!is.na(m),,drop=FALSE]
    rownames(x) <- ids[m[!is.na(m)]]

    x
}

#' @describeIn replace_ids Replace IDs in a data frame
#' @export
replace_ids.data.frame <- replace_ids.matrix

context("replace individual IDs")

test_that("check_new_ids() works", {

    old_ids <- as.character(1:200)
    new_ids <- setNames(sprintf("mouse%03d", as.numeric(old_ids)), old_ids)

    # same ids in same order
    expect_equal(check_new_ids(new_ids, old_ids), new_ids)

    # same ids, but shuffled
    shuffled <- sample(new_ids)
    expect_equal(check_new_ids(shuffled, old_ids), shuffled)

    # error if duplicate IDs
    dup_new <- new_ids
    dup_new[5] <- dup_new[20]
    expect_error(check_new_ids(dup_new, old_ids))

    # error if duplicate names in the IDs
    dup_old <- new_ids
    names(dup_old)[5] <- names(dup_old)[20]
    expect_error(check_new_ids(dup_old, old_ids))

    # warning if extra IDs
    new_extra <- c(new_ids, "201"="mouse201", "202"="mouse202")
    expect_warning( expect_equal( check_new_ids(new_extra, old_ids), new_ids) )
    o <- sample(length(new_extra))
    expect_warning( expect_equal( check_new_ids(new_extra[o], old_ids),
                                  new_extra[o[o<=length(new_ids)]] ) )

    # warning if not all IDs are there
    new_missing <- sample(new_ids, length(new_ids)-5)
    expect_warning( expect_equal( check_new_ids(new_missing, old_ids),
                                  new_missing[names(new_missing) %in% old_ids] ))


})


test_that("replace_ids() works for a cross2 object", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    ids <- ind_ids(iron)
    new_ids <- setNames(paste0("mouse", ids), ids)
    change_back <- setNames(ids, paste0("mouse", ids))
    extra_ids <- sample(c(ids, 1001:1020))
    extra_ids <- setNames(paste0("mouse", extra_ids), extra_ids)

    # same ids, old and new
    expect_equal( replace_ids(iron, setNames(ids, ids)), iron)

    # simple replacement, everything in order
    expect_equal( replace_ids(replace_ids(iron, new_ids), change_back), iron)

    # simple replacement, but shuffled
    expect_equal( replace_ids(replace_ids(iron, sample(new_ids)), sample(change_back)), iron)

    # simple replacement, with some extras plus shuffled
    expect_warning(
        expect_equal( replace_ids(replace_ids(iron, extra_ids), sample(change_back)), iron)
    )

    # missing some individuals
    sub_ids <- sample(ids, length(ids)-10)
    sub_ids_ordered <- sub_ids[order(as.numeric(sub_ids))]
    expect_warning(
        expect_equal( replace_ids(iron, setNames(sub_ids, sub_ids)),
                      iron[sub_ids_ordered,])
    )

})


test_that("replace_ids() works for calc_genoprob output", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    ids <- ind_ids(iron)
    new_ids <- setNames(paste0("mouse", ids), ids)
    change_back <- setNames(ids, paste0("mouse", ids))
    extra_ids <- sample(c(ids, 1001:1020))
    extra_ids <- setNames(paste0("mouse", extra_ids), extra_ids)

    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    pr <- calc_genoprob(iron, map)

    # same ids, old and new
    expect_equal( replace_ids(pr, setNames(ids, ids)), pr)

    # simple replacement, everything in order
    expect_equal( replace_ids(replace_ids(pr, new_ids), change_back), pr)

    # simple replacement, but shuffled
    expect_equal( replace_ids(replace_ids(pr, sample(new_ids)), sample(change_back)), pr)

    # simple replacement, with some extras plus shuffled
    expect_warning(
        expect_equal( replace_ids(replace_ids(pr, extra_ids), sample(change_back)), pr)
    )

    # missing some individuals
    sub_ids <- sample(ids, length(ids)-10)
    sub_ids_ordered <- sub_ids[order(as.numeric(sub_ids))]
    expect_warning(
        expect_equal( replace_ids(pr, setNames(sub_ids, sub_ids)),
                      pr[sub_ids_ordered,])
    )

})


test_that("replace_ids() works for viterbi output", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    ids <- ind_ids(iron)
    new_ids <- setNames(paste0("mouse", ids), ids)
    change_back <- setNames(ids, paste0("mouse", ids))
    extra_ids <- sample(c(ids, 1001:1020))
    extra_ids <- setNames(paste0("mouse", extra_ids), extra_ids)

    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    v <- viterbi(iron, map)

    # same ids, old and new
    expect_equal( replace_ids(v, setNames(ids, ids)), v)

    # simple replacement, everything in order
    expect_equal( replace_ids(replace_ids(v, new_ids), change_back), v)

    # simple replacement, but shuffled
    expect_equal( replace_ids(replace_ids(v, sample(new_ids)), sample(change_back)), v)

    # simple replacement, with some extras plus shuffled
    expect_warning(
        expect_equal( replace_ids(replace_ids(v, extra_ids), sample(change_back)), v)
    )

    # missing some individuals
    sub_ids <- sample(ids, length(ids)-10)
    sub_ids_ordered <- sub_ids[order(as.numeric(sub_ids))]
    expect_warning(
        expect_equal( replace_ids(v, setNames(sub_ids, sub_ids)),
                      v[sub_ids_ordered,])
    )

})


test_that("replace_ids() works for sim_geno output", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    ids <- ind_ids(iron)
    new_ids <- setNames(paste0("mouse", ids), ids)
    change_back <- setNames(ids, paste0("mouse", ids))
    extra_ids <- sample(c(ids, 1001:1020))
    extra_ids <- setNames(paste0("mouse", extra_ids), extra_ids)

    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    d <- sim_geno(iron, map, n_draws=8)

    # same ids, old and new (changed back)
    expect_equal( replace_ids(d, setNames(ids, ids)), d)

    # simple replacement, everything in order
    expect_equal( replace_ids(replace_ids(d, new_ids), change_back), d)

    # simple replacement, but shuffled
    expect_equal( replace_ids(replace_ids(d, sample(new_ids)), sample(change_back)), d)

    # simple replacement, with some extras plus shuffled
    expect_warning(
        expect_equal( replace_ids(replace_ids(d, extra_ids), sample(change_back)), d)
    )

    # missing some individuals
    sub_ids <- sample(ids, length(ids)-10)
    sub_ids_ordered <- sub_ids[order(as.numeric(sub_ids))]
    expect_warning(
        expect_equal( replace_ids(d, setNames(sub_ids, sub_ids)),
                      d[sub_ids_ordered,])
    )

})





test_that("replace_ids() works for a matrix", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    ids <- ind_ids(iron)
    new_ids <- setNames(paste0("mouse", ids), ids)
    change_back <- setNames(ids, paste0("mouse", ids))
    extra_ids <- sample(c(ids, 1001:1020))
    extra_ids <- setNames(paste0("mouse", extra_ids), extra_ids)

    # create a matrix
    set.seed(20210712)
    n_col <- 12
    d <- matrix(rnorm(n_ind(iron)*n_col), ncol=n_col)
    dimnames(d) <- list(ids, paste("V", seq_len(n_col)))

    # same ids, old and new (changed back)
    expect_equal( replace_ids(d, setNames(ids, ids)), d)

    # simple replacement, everything in order
    expect_equal( replace_ids(replace_ids(d, new_ids), change_back), d)

    # simple replacement, but shuffled
    expect_equal( replace_ids(replace_ids(d, sample(new_ids)), sample(change_back)), d)

    # simple replacement, with some extras plus shuffled
    expect_warning(
        expect_equal( replace_ids(replace_ids(d, extra_ids), sample(change_back)), d)
    )

    # missing some individuals
    sub_ids <- sample(ids, length(ids)-10)
    sub_ids_ordered <- sub_ids[order(as.numeric(sub_ids))]
    expect_warning(
        expect_equal( replace_ids(d, setNames(sub_ids, sub_ids)),
                      d[sub_ids_ordered,,drop=FALSE])
    )


    ##############################
    # turn it into a data frame and do it all again
    ##############################
    d <- as.data.frame(d)

    # same ids, old and new (changed back)
    expect_equal( replace_ids(d, setNames(ids, ids)), d)

    # simple replacement, everything in order
    expect_equal( replace_ids(replace_ids(d, new_ids), change_back), d)

    # simple replacement, but shuffled
    expect_equal( replace_ids(replace_ids(d, sample(new_ids)), sample(change_back)), d)

    # simple replacement, with some extras plus shuffled
    expect_warning(
        expect_equal( replace_ids(replace_ids(d, extra_ids), sample(change_back)), d)
    )

    # missing some individuals
    sub_ids <- sample(ids, length(ids)-10)
    sub_ids_ordered <- sub_ids[order(as.numeric(sub_ids))]
    expect_warning(
        expect_equal( replace_ids(d, setNames(sub_ids, sub_ids)),
                      d[sub_ids_ordered,,drop=FALSE])
    )



})

# reduce snpinfo to the indexed snps
reduce_to_index_snps <- function(snpinfo)
{
    if(length(unique(snpinfo$chr)) > 1) { # more than one chromosome
        # rows for each chromosome
        rows_spl <- split(1:nrow(snpinfo), factor(snpinfo$chr, unique(snpinfo$chr)))

        # apply this function to one chromosome at a time
        result <- vector("list", length(rows_spl))
        for(chr in seq_along(rows_spl)) {
            result[[chr]] <- reduce_to_index_snps(snpinfo[rows_spl[[chr]],,drop=FALSE])
        }

        # re-combine by rows
        return( do.call("rbind", result) )
    }

    # now working with just one chromosome
    if(any(snpinfo$index < 1 | snpinfo$index > nrow(snpinfo))) {
        stop("index seems messed up for chromosome ", snpinfo$chr[1])
    }

    result <- snpinfo[unique(snpinfo$index), , drop=FALSE]
    result$index <- seq_len(nrow(result))

    result
}

context("reduce to index snps")

test_that("reduce_to_index_snps works", {

    # founder genotypes for a set of SNPs
    snpgeno <- rbind(m1=c(3,1,1,3,1,1,1,1),
                     m2=c(1,3,1,3,1,3,1,3),
                     m3=c(1,1,1,1,3,3,3,3),
                     m4=c(1,3,1,3,1,3,1,3),
                     m5=c(1,1,1,3,1,3,1,1),
                     m6=c(1,1,3,1,1,1,3,1))
    sdp <- calc_sdp(snpgeno)
    snpinfo <- data.frame(chr=c("19", "19", "X", "X", "19", "X"),
                          pos=c(40.36, 40.53, 110.91, 111.21, 56.480843, 56.480843),
                          sdp=sdp,
                          snp=c("m1", "m2", "m3", "m4", "m5", "m6"), stringsAsFactors=FALSE)

    snpinfo <- cbind(snpinfo[c(1,2,5,6,3,4),],
                     index=c(1:3, 1:3),
                     interval=c(91,91,132,112,237,237),
                     on_map=c(FALSE, FALSE, TRUE, FALSE, FALSE, FALSE))

    expect_equal(reduce_to_index_snps(snpinfo), snpinfo)
    expect_equal(reduce_to_index_snps(snpinfo[snpinfo$chr==19,]), snpinfo[snpinfo$chr==19,])
    expect_equal(reduce_to_index_snps(snpinfo[snpinfo$chr=="X",]), snpinfo[snpinfo$chr=="X",])

    # repeat with some redundant snps
    snpinfo_extra <- data.frame(chr=c("19", "19", "X", "X"),
                                pos=c(40.45, 40.37, 56.3, 110.83),
                                sdp=c(170, 9, 68, 170),
                                snp=c("m7", "m8", "m9", "m10"),
                                stringsAsFactors=FALSE)
    rownames(snpinfo_extra) <- snpinfo_extra$snp
    snpinfo_extra <- rbind(snpinfo[,1:4], snpinfo_extra)

    snpinfo_extra <- cbind(snpinfo_extra[c(1,8,7,2,3,  9,4,10,5,6),],
                      index=c(1,1,3,3,5,  1,1,3,4,3),
                      interval=c(91,91,91,91,132,  112,112,237,237,237),
                      on_map=c(FALSE,FALSE,FALSE,FALSE,TRUE,   FALSE,FALSE,FALSE,FALSE,FALSE))

    expected <- snpinfo_extra[c(1,3,5,6,8,9),]
    expected$index <- c(1,2,3,1,2,3)

    expect_equal(reduce_to_index_snps(snpinfo_extra), expected)
    expect_equal(reduce_to_index_snps(snpinfo_extra[snpinfo_extra$chr==19,]), expected[expected$chr==19,])
    expect_equal(reduce_to_index_snps(snpinfo_extra[snpinfo_extra$chr=="X",]), expected[expected$chr=="X",])

})

# Calculate QTL effects in scan along one chromosome, adjusting for polygenes with an LMM
#
scan1coef_pg <-
    function(genoprobs, pheno, kinship,
             addcovar=NULL, nullcovar=NULL, intcovar=NULL,
             weights=NULL, contrasts=NULL, zerosum=TRUE, se=FALSE,
             hsq=NULL, reml=TRUE, ...)
{
    # deal with the dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    if(!is_pos_number(tol)) stop("tol should be a single positive number")
    check_extra_dots(dotargs, "tol")

    # check that the objects have rownames
    check4names(pheno, addcovar, NULL, intcovar, nullcovar)

    # force things to be matrices
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(nullcovar)) {
        if(!is.matrix(nullcovar)) nullcovar <- as.matrix(nullcovar)
        if(!is.numeric(nullcovar)) stop("nullcovar is not numeric")
    }
    if(!is.null(intcovar)) {
        if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)
        if(!is.numeric(intcovar)) stop("intcovar is not numeric")
    }
    if(!is.null(contrasts)) {
        if(!is.matrix(contrasts)) contrasts <- as.matrix(contrasts)
        if(!is.numeric(contrasts)) stop("contrasts is not numeric")
    }

    # make sure pheno is a vector
    if(is.matrix(pheno) || is.data.frame(pheno)) {
        if(ncol(pheno) > 1)
            warning("Considering only the first phenotype.")
        rn <- rownames(pheno)
        pheno <- pheno[,1]
        names(pheno) <- rn
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }

    # genoprobs has more than one chromosome?
    if(length(genoprobs) > 1)
        warning("Using only the first chromosome, ", names(genoprobs)[1])
    genoprobs <- genoprobs[[1]]

    # make sure contrasts is square n_genotypes x n_genotypes
    if(!is.null(contrasts)) {
        ng <- ncol(genoprobs)
        if(ncol(contrasts) != ng || nrow(contrasts) != ng)
            stop("contrasts should be a square matrix, ", ng, " x ", ng)
    }

    # make sure kinship is for a single chromosome and get IDs
    did_decomp <- is_kinship_decomposed(kinship)
    kinship <- check_kinship_onechr(kinship)
    kinshipIDs <- check_kinship(kinship, 1)

    # multiply kinship matrix by 2; rest is using 2*kinship
    # see Almasy & Blangero (1998) https://doi.org/10.1086/301844
    kinship <- double_kinship(kinship)

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, pheno, kinshipIDs, weights,
                               addcovar, nullcovar, intcovar, complete.cases=TRUE)

    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    if(did_decomp) { # if did decomposition already, make sure it was with exactly
        if(length(kinshipIDs) != length(ind2keep) ||
           any(sort(kinshipIDs) != sort(ind2keep)))
            stop("Decomposed kinship matrix was with different individuals")
        else
            ind2keep <- kinshipIDs # force them in same order
    }

    # omit individuals not in common
    genoprobs <- genoprobs[ind2keep,,,drop=FALSE]
    pheno <- pheno[ind2keep]
    if(!is.null(addcovar)) addcovar <- addcovar[ind2keep,,drop=FALSE]
    if(!is.null(nullcovar)) nullcovar <- nullcovar[ind2keep,,drop=FALSE]
    if(!is.null(intcovar)) intcovar <- intcovar[ind2keep,,drop=FALSE]
    if(!is.null(weights)) weights <- weights[ind2keep]

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # make sure columns in intcovar are also in addcovar
    addcovar <- force_intcovar(addcovar, intcovar, tol)

    # square-root of weights (only if model="normal")
    weights <- sqrt_weights(weights) # also check >0 (and if all 1's, turn to NULL)

    # multiply stuff by weights
    kinship <- weight_kinship(kinship, weights)
    addcovar <- weight_matrix(addcovar, weights)
    intcovar <- weight_matrix(intcovar, weights)
    nullcovar <- weight_matrix(nullcovar, weights)
    pheno <- weight_matrix(pheno, weights)

    # eigen decomposition of kinship matrix
    if(!did_decomp)
        kinship <- decomp_kinship(kinship[ind2keep, ind2keep])

    # estimate hsq if necessary
    if(is.null(hsq)) {
        nullresult <- calc_hsq_clean(Ke=kinship, pheno=as.matrix(pheno),
                                     addcovar=cbind(addcovar, nullcovar), Xcovar=NULL,
                                     weights=weights, is_x_chr=FALSE, reml=reml,
                                     cores=1, check_boundary=TRUE, tol=tol)
        hsq <- as.numeric(nullresult$hsq)
    }

    # eigen-vectors and weights
    eigenvec <- kinship$vectors
    wts <- 1/sqrt(hsq*kinship$values + (1-hsq))

    # multiply genoprobs by contrasts
    if(!is.null(contrasts))
        genoprobs <- genoprobs_by_contrasts(genoprobs, contrasts)

    # multiply genoprobs by weights
    genoprobs <- weight_array(genoprobs[ind2keep,,,drop=FALSE], weights)

    if(se) { # also calculate SEs

        if(is.null(intcovar)) { # just addcovar
            if(is.null(addcovar)) addcovar <- matrix(nrow=length(ind2keep), ncol=0)
            result <- scancoefSE_pg_addcovar(genoprobs, pheno, addcovar, eigenvec, wts, tol)
        }
        else {                  # intcovar
            result <- scancoefSE_pg_intcovar(genoprobs, pheno, addcovar, intcovar,
                                              eigenvec, wts, tol)
        }

        SE <- t(result$SE) # transpose to positions x coefficients
        result <- result$coef
    } else { # don't calculate SEs

        if(is.null(intcovar)) { # just addcovar
            if(is.null(addcovar)) addcovar <- matrix(nrow=length(ind2keep), ncol=0)
            result <- scancoef_pg_addcovar(genoprobs, pheno, addcovar, eigenvec, wts, tol)
        }
        else {                  # intcovar
            result <- scancoef_pg_intcovar(genoprobs, pheno, addcovar, intcovar,
                                            eigenvec, wts, tol)
        }
        SE <- NULL
    }

    result <- t(result) # transpose to positions x coefficients

    # add names
    dimnames(result) <- list(dimnames(genoprobs)[[3]],
                             scan1coef_names(genoprobs, addcovar, intcovar))
    if(se) dimnames(SE) <- dimnames(result)

    # center the QTL effects at zero and add an intercept
    if(zerosum && is.null(contrasts)) {
        ng <- dim(genoprobs)[2]
        whcol <- seq_len(ng)
        mu <- rowMeans(result[,whcol,drop=FALSE], na.rm=TRUE)
        result <- cbind(result, intercept=mu)
        result[,whcol] <- result[,whcol] - mu

        if(se) {
            SE <- cbind(SE, intercept=sqrt(rowMeans(SE[,whcol,drop=FALSE]^2, na.rm=TRUE)))
        }
    }

    # add some attributes with details on analysis
    attr(result, "sample_size") <- length(ind2keep)
    attr(result, "SE") <- SE # include only if not NULL

    class(result) <- c("scan1coef", "scan1", "matrix")
    result
}

#' Get x-axis position for genomic location
#'
#' For a plot of [scan1()] results, get the x-axis
#' location that corresponds to a particular genomic location
#' (chromosome ID and position).
#'
#' @param map A list of vectors of marker positions, as produced by
#' [insert_pseudomarkers()].
#' @param chr Selected chromosomes that were plotted (if used in the
#' call to [plot_scan1()].
#' @param gap The gap between chromosomes used in the call to
#' [plot_scan1()].
#' @param thechr Vector of chromosome IDs
#' @param thepos Vector of chromosomal positions
#'
#' @return A vector of x-axis locations.
#'
#' @details `thechr` and `thepos` should be the same length,
#' or should have length 1 (in which case they are expanded to the
#' length of the other vector).
#'
#' @export
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' # plot the results for selected chromosomes
#' ylim <- c(0, maxlod(out)*1.02) # need to strip class to get overall max LOD
#' chr <- c(2,7,8,9,15,16)
#' plot(out, map, chr=chr, ylim=ylim)
#' plot(out, map, lodcolumn=2, chr=chr, col="violetred", add=TRUE)
#' legend("topleft", lwd=2, col=c("darkslateblue", "violetred"), colnames(out),
#'        bg="gray90")
#'
#' # Use xpos_scan1 to add points at the peaks
#' # first find the peaks with LOD > 3
#' peaks <- find_peaks(out, map)
#'
#' # keep just the peaks for chromosomes that were plotted
#' peaks <- peaks[peaks$chr %in% chr,]
#'
#' # find x-axis positions
#' xpos <- xpos_scan1(map, chr=chr, thechr=peaks$chr, thepos=peaks$pos)
#'
#' # point colors
#' ptcolor <- c("darkslateblue", "violetred")[match(peaks$lodcolumn, c("liver", "spleen"))]
#'
#' # plot points
#' points(xpos, peaks$lod, pch=21, bg=ptcolor)
xpos_scan1 <-
function(map, chr=NULL, gap=NULL, thechr, thepos)
{
    if(is.null(map)) stop("map is NULL")
    if(is.null(gap)) gap <- sum(chr_lengths(map))/100

    # subset chromosomes
    if(!is.null(chr)) map <- map[chr]

    # start position for each chromosome
    start <- map_to_boundaries(map, gap)[1,] - vapply(map, "[", 1, 1)
    # if more than one chromosome, add gap/2
    if(length(map) > 1) start <- start + gap/2
    # add chromosome names
    names(start) <- names(map)

    # chromosome to character strings
    thechr <- as.character(thechr)

    if(length(thechr)==1)
        thechr <- rep(thechr, length(thepos))
    if(length(thepos)==1)
        thepos <- rep(thepos, length(thechr))
    if(length(thechr) != length(thepos))
        stop("thechr (length ", length(thechr), ") and thepos (length ",
             length(thepos), ") must be the same length, or length 1")

    uchr <- unique(thechr)
    if(!all(uchr %in% names(start)))
        stop("Unknown chromosome IDs: ",
             paste(uchr[!(uchr %in% names(start))], collapse=", "))

    result <- start[thechr] + thepos
    names(result) <- NULL

    result
}

context("x-axis positions for scan1 plot")

test_that("xpos_scan1 works", {

    # single chromosome
    expect_equal(xpos_scan1(list("1"=1:5), thechr="1", thepos=1:5), 1:5)

    m <- sort(runif(100, 0, 100))
    expect_equal(xpos_scan1(list("1"=m), thechr="1", thepos=m), m)

    # two chromosomes
    map <- list("1"=c(10, 20, 50),
                "2"=c(5,  25, 85))
    expect_equal(xpos_scan1(map, thechr=c("1", "2"), thepos=c(20, 25), gap=25),
                 c(22.5, 97.5))
    expect_equal(xpos_scan1(map, thechr="1", thepos=c(15, 40), gap=25),
                 c(17.5, 42.5))
    expect_equal(xpos_scan1(map, thechr=c("1", "2"), thepos=10, gap=25),
                 c(12.5, 82.5))


    # two chromosomes, automatic gap
    expect_equal(xpos_scan1(map, thechr=c("1", "2"), thepos=c(20, 25)),
                 xpos_scan1(map, thechr=c("1", "2"), thepos=c(20, 25), gap=1.2))
    expect_equal(xpos_scan1(map, thechr="1", thepos=c(15, 40)),
                 xpos_scan1(map, thechr="1", thepos=c(15, 40), gap=1.2))
    expect_equal(xpos_scan1(map, thechr=c("1", "2"), thepos=10),
                 xpos_scan1(map, thechr=c("1", "2"), thepos=10, gap=1.2))
    expect_equal(xpos_scan1(map, thechr=c("1", "2"), thepos=c(20, 25)),
                 c(10.6, 61.8))
    expect_equal(xpos_scan1(map, thechr="1", thepos=c(15, 40)),
                 c(5.6, 30.6))
    expect_equal(xpos_scan1(map, thechr=c("1", "2"), thepos=10),
                 c(0.6, 46.8))

    # invalid inputs
    expect_error(xpos_scan1(map, thechr=c("1", "2", "2"), thepos=c(5, 10)))
    expect_error(xpos_scan1(map, thechr=c("1", "2"), thepos=c(5, 10, 15)))
    expect_error(xpos_scan1(map, thechr=c("1", "2", "3"), thepos=c(5, 10, 15)))

})

#' Smooth genetic map
#'
#' Smooth a genetic map by mixing it with a bit of constant
#' recombination (using a separate recombination rate for each
#' chromosome), to eliminate intervals that have exactly 0
#' recombination.
#'
#' @param gmap Genetic map, as a list of numeric vectors; each vector gives marker
#' positions for a single chromosome.
#' @param pmap Physical map, as a list of numeric vectors; each vector gives marker
#' positions for a single chromosome, with the same chromosomes and markers as `gmap`.
#' @param alpha Proportion of mixture to take from constant recombination.
#'
#' @return A genetic map like the input `gmap`, but smoothed by mixing
#' it with a proportion `alpha` of constant recombination on each
#' chromosome.
#'
#' @details An interval of genetic length \eqn{d_g}{dg} and physical
#'     length \eqn{d_p}{dp} is changed to have length
#'     \eqn{(1-\alpha)d_g + \alpha d_p r}{(1-alpha)dg + (alpha*dp*r)}
#'     where \eqn{r = L_g / L_p}{r = Lg/Lp} is the chromosome-specific
#'     recombination rate.
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' gmap_adj <- smooth_gmap(iron$gmap, iron$pmap)
#'
#' @export
#' @keywords utilities
#' @seealso [unsmooth_gmap()]

smooth_gmap <-
    function(gmap, pmap, alpha=0.02)
    {
        stopifnot(is.list(gmap), is.list(pmap))
        stopifnot(length(gmap)==length(pmap), all(names(gmap) == names(pmap)))
        stopifnot(length(alpha)==1, alpha >= 0, alpha <= 1)
        for(chr in seq(along=gmap)) {
            g <- gmap[[chr]]
            p <- pmap[[chr]]
            stopifnot(length(g) == length(p), all(names(g) == names(p)) )
            gd <- diff(g)
            pd <- diff(p)
            rr <- diff(range(g)) / diff(range(p))
            gd <- gd*(1-alpha) + alpha*pd*rr
            gmap[[chr]] <- setNames(cumsum(c(g[1], gd)), names(g))
        }
        gmap
    }

context("smooth_gmap")

test_that("smooth_gmap works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    # if alpha==0, nothing happens
    expect_equal(smooth_gmap(iron$gmap, iron$pmap, 0), iron$gmap)

    gmap_adj <- smooth_gmap(iron$gmap, iron$pmap, 0.02)

    # structure shouldn't change
    expect_equal(names(gmap_adj), names(iron$gmap))
    expect_equal(sapply(gmap_adj, length), sapply(iron$gmap, length))
    expect_equal(names(unlist(gmap_adj)), names(unlist(iron$gmap)))

    # chromosomes with just two markers are unchanged
    nmar <- sapply(gmap_adj, length)
    expect_equal( gmap_adj[nmar==2], iron$gmap[nmar==2] )

    # test unsmooth gmap
    gmap_back <- unsmooth_gmap(gmap_adj, iron$pmap, 0.02)
    expect_equal(iron$gmap, gmap_back)

    # test with a different alpha value
    expect_equal(unsmooth_gmap(smooth_gmap(iron$gmap, iron$pmap, 0.05), iron$pmap, 0.05), iron$gmap)

})

#' Interpolate between maps
#'
#' Use interpolate to convert from one map to another
#'
#' @param map The map to be interpolated; a list of vectors.
#' @param oldmap Map with positions in the original scale, as in `map`.
#' @param newmap Map with positions in the new scale.
#'
#' @return Object of same form as input `map` but in the units as in `newmap`.
#'
#' @examples
#' # load example data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#'
#' # positions to interpolate from cM to Mbp
#' tointerp <- list("7" = c(pos7.1= 5, pos7.2=15, pos7.3=25),
#'                  "9" = c(pos9.1=20, pos9.2=40))
#'
#' interp_map(tointerp, iron$gmap, iron$pmap)
#'
#' @export
interp_map <-
    function(map, oldmap, newmap)
{
    if(is.null(map)) stop("map is NULL")
    if(is.null(oldmap)) stop("oldmap is NULL")
    if(is.null(newmap)) stop("newmap is NULL")

    chr <- names(map)
    result <- vector("list", length(map))
    names(result) <- chr

    chr_found <- (chr %in% names(oldmap)) & (chr %in% names(newmap))
    if(!all(chr_found)) {
        stop("Chromosomes not found in both old and new maps: ",
             paste(chr[!chr_found], collapse=", "))
    }

    dropped_old <- dropped_new <- NULL
    for(thechr in chr) {
        om <- oldmap[[thechr]]
        nm <- newmap[[thechr]]
        nom <- names(om)
        nnm <- names(nm)

        tm <- map[[thechr]]
        ntm <- names(tm)

        # markers in oldmap that aren't in newmap?
        if((is.null(nom) || is.null(nnm)) && length(om) != length(nm))
            stop("If length(oldmap) != length(newmap), they both need marker names")

        if(!is.null(nom) && !is.null(nnm)) {
            if(length(unique(nom)) != length(nom) ||
               length(unique(nnm)) != length(nnm))
                stop("marker names should be distinct")

            drop_om <- !(nom %in% nnm)
            if(any(drop_om)) {
                dropped_old <- c(dropped_old, nom[drop_om])
                om <- om[!drop_om]
                nom <- names(om)
            }

            # markers in newmap that aren't in oldmap?
            drop_nm <- !(nnm %in% nom)
            if(any(drop_nm)) {
                dropped_new <- c(dropped_new, nnm[drop_nm])
                nm <- nm[!drop_nm]
                nnm <- names(nm)
            }
        }

        # need at least two markers left, and need them to be in the same order
        if(length(nm) < 2 || length(om) < 2)
            stop("Less than 2 matching markers on chr ", thechr)
        if(length(om) != length(nm) || (!is.null(nom) && !is.null(nnm) && !all(nom == nnm)))
            stop("Old and new maps not aligned on chr ", thechr)

        result[[thechr]] <- interpolate_map(tm, om, nm)
        names(result[[thechr]]) <- ntm

        # make sure markers that are in both map and newmap are kept in the newmap position
        if(!is.null(ntm) && !is.null(nom) && !is.null(nnm) && any(ntm %in% nom)
           && length(unique(ntm)) == length(ntm)) { # skip if marker names aren't distinct
            overlap <- ntm[ntm %in% nom] # markers in both old and to-be-transformed map
            overlap <- overlap[abs(tm[overlap] - om[overlap]) < 1e-6] # ...that are in the same position

            # insert the position in newmap
            result[[thechr]][overlap] <- nm[overlap]
        }
    }

    if(length(dropped_old) > 1)
        warning("Dropped ", length(dropped_old), " markers not present in new map")
    if(length(dropped_new) > 1)
        warning("Dropped ", length(dropped_new), " markers not present in old map")

    result
}

#' Permutation test for genome scan with a single-QTL model
#'
#' Permutation test for a enome scan with a single-QTL model by
#' Haley-Knott regression or a linear mixed model, with possible
#' allowance for covariates.
#'
#' @param genoprobs Genotype probabilities as calculated by
#' [calc_genoprob()].
#' @param pheno A numeric matrix of phenotypes, individuals x phenotypes.
#' @param kinship Optional kinship matrix, or a list of kinship matrices (one
#' per chromosome), in order to use the LOCO (leave one chromosome
#' out) method.
#' @param addcovar An optional numeric matrix of additive covariates.
#' @param Xcovar An optional numeric matrix with additional additive covariates used for
#' null hypothesis when scanning the X chromosome.
#' @param intcovar An optional numeric matrix of interactive covariates.
#' @param weights An optional numeric vector of positive weights for the
#' individuals. As with the other inputs, it must have `names`
#' for individual identifiers.
#' @param reml If `kinship` provided: if `reml=TRUE`, use
#' REML; otherwise maximum likelihood.
#' @param model Indicates whether to use a normal model (least
#'     squares) or binary model (logistic regression) for the phenotype.
#'     If `model="binary"`, the phenotypes must have values in \eqn{[0, 1]}.
#' @param n_perm Number of permutation replicates.
#' @param perm_Xsp If TRUE, do separate permutations for the autosomes
#' and the X chromosome.
#' @param perm_strata Vector of strata, for a stratified permutation
#' test. Should be named in the same way as the rows of
#' `pheno`. The unique values define the strata.
#' @param chr_lengths Lengths of the chromosomes; needed only if
#' `perm_Xsp=TRUE`. See [chr_lengths()].
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#' @param scan_func If provided, this function is used for the genome scans.
#' It must take arguments `genoprobs`, `pheno`, `kinship`, `addcovar`,
#' `Xcovar`, `intcovar`, `weights`, and possibly further arguments.
#' @param ... Additional control parameters; see Details.
#'
#' @return If `perm_Xsp=FALSE`, the result is matrix of
#' genome-wide maximum LOD scores, permutation replicates x
#' phenotypes. If `perm_Xsp=TRUE`, the result is a list of
#' two matrices, one for the autosomes and one for the X
#' chromosome. The object is given class `"scan1perm"`.
#'
#' @details
#' If `kinship` is not provided, so that analysis proceeds by
#' Haley-Knott regression, we permute the rows of the phenotype data;
#' the same permutations are also applied to the rows of the
#' covariates (`addcovar`, `Xcovar`, and `intcovar`)
#' are permuted.
#'
#' If `kinship` is provided, we instead permute the rows of the
#' genotype data and fit an LMM with the same residual heritability
#' (estimated under the null hypothesis of no QTL).
#'
#' If `Xcovar` is provided and `perm_strata=NULL`, we do a
#' stratified permutation test with the strata defined by the rows of
#' `Xcovar`. If a simple permutation test is desired, provide
#' `perm_strata` that is a vector containing a single repeated
#' value.
#'
#' The `...` argument can contain several additional control
#' parameters; suspended for simplicity (or confusion, depending on
#' your point of view). `tol` is used as a tolerance value for linear
#' regression by QR decomposition (in determining whether columns are
#' linearly dependent on others and should be omitted); default
#' `1e-12`. `maxit` is the maximum number of iterations for
#' converence of the iterative algorithm used when `model=binary`.
#' `bintol` is used as a tolerance for converence for the iterative
#' algorithm used when `model=binary`. `eta_max` is the maximum value
#' for the "linear predictor" in the case `model="binary"` (a bit of a
#' technicality to avoid fitted values exactly at 0 or 1).
#'
#' @references Churchill GA, Doerge RW (1994) Empirical threshold
#' values for quantitative trait mapping. Genetics 138:963--971.
#'
#' Manichaikul A, Palmer AA, Sen S, Broman KW (2007) Significance
#' thresholds for quantitative trait locus mapping under selective
#' genotyping. Genetics 177:1963--1966.
#'
#' Haley CS, Knott SA (1992) A simple regression method for mapping
#' quantitative trait loci in line crosses using flanking markers.
#' Heredity 69:315--324.
#'
#' Kang HM, Zaitlen NA, Wade CM, Kirby A, Heckerman D, Daly MJ, Eskin
#' E (2008) Efficient control of population structure in model
#' organism association mapping. Genetics 178:1709--1723.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c(10,18,"X")]}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # strata for permutations
#' perm_strata <- mat2strata(Xcovar)
#'
#' # permutations with genome scan (just 3 replicates, for illustration)
#' operm <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar,
#'                    n_perm=3, perm_strata=perm_strata)
#' summary(operm)
#'
#' # leave-one-chromosome-out kinship matrices
#' kinship <- calc_kinship(probs, "loco")
#'
#' # permutations of genome scan with a linear mixed model
#' operm_lmm <- scan1perm(probs, pheno, kinship, covar, Xcovar, n_perm=3,
#'                        perm_Xsp=TRUE, perm_strata=perm_strata,
#'                        chr_lengths=chr_lengths(map))
#' summary(operm_lmm)
#'
#'
#' # permutations of scan1gen, fitting glm with probit link
#' # fitting function
#' ll_glm <-
#'    function(pr, pheno, addcovar=NULL, ...)
#' {
#'    formula <- ifelse(is.null(pr), "pheno ~ 1", "pheno ~ pr")
#'    if(!is.null(addcovar)) formula <- paste(formula, "+ addcovar")
#'
#'     glm_out <- glm(as.formula(formula), family=binomial(link=probit))
#'     -glm_out$deviance/(2*log(10)) # log10 likelihood
#' }
#'
#' # create binary trait
#' bin_pheno <- setNames(as.numeric(iron$pheno[,1] > median(iron$pheno[,1])),
#'                       rownames(iron$pheno))
#'
#' # permutations with glm
#' operm_glm <- scan1perm(probs, bin_pheno, kinship, covar, Xcovar, n_perm=3,
#'                        perm_Xsp=TRUE, perm_strata=perm_strata,
#'                        chr_lengths=chr_lengths(map),
#'                        scan_func=scan1gen, func=ll_glm)
#' summary(operm_glm)
#'
#' \dontrun{
#' # example of scan1perm with scan1snps
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' DOex <- DOex[,c("2", "3")] # subset to chr 2 and 3
#' probs <- calc_genoprob(DOex, error_prob=0.002)
#'
#' snpdb_file <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
#' queryf <- create_variant_query_func(snpdb_file)
#'
#' operm <- scan1perm(genoprobs=probs, map=DOex$pmap, pheno=DOex$pheno,
#'                    scan_func=scan1snps, query_func=queryf, n_perm=3)
#' }
#'
#' @seealso [scan1()], [chr_lengths()], [mat2strata()], [scan1gen examples](https://kbroman.org/qtl2/assets/vignettes/scan1gen.html)
#'
#' @export
scan1perm <-
    function(genoprobs, pheno, kinship=NULL, addcovar=NULL, Xcovar=NULL,
             intcovar=NULL, weights=NULL, reml=TRUE, model=c("normal", "binary"),
             n_perm=1, perm_Xsp=FALSE, perm_strata=NULL, chr_lengths=NULL,
             cores=1, scan_func=NULL, ...)
{
    if(is.null(genoprobs)) stop("genoprobs is NULL")
    if(is.null(pheno)) stop("pheno is NULL")

    # grab tol from dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    if(!is_pos_number(tol)) stop("tol should be a single positive number")

    # normal or binary model?
    model <- match.arg(model)

    if(!is_pos_number(n_perm)) stop("n_perm should be a single positive integer")

    # check that the objects have rownames
    check4names(pheno, addcovar, Xcovar, intcovar)

    # if Xcovar provided, the default is to do a stratified permutation test
    if(is.null(perm_strata) && !is.null(Xcovar))
        perm_strata <- mat2strata(Xcovar)

    if(perm_Xsp) { # autosome/X chr-specific permutations
        if(is.null(chr_lengths))
            stop("Need to provide chr_lengths when perm_Xsp=TRUE")

        is_x_chr <- attr(genoprobs, "is_x_chr")

        # collapse to A/X
        chr_lengths <- collapse_chr_lengths_to_AX(chr_lengths, is_x_chr)
        if(any(chr_lengths < 1e-12))
            stop("Autosome or X chromosome of length 0; skip perm_Xsp")

        n_permX <- ceiling(n_perm * chr_lengths[["A"]] / chr_lengths[["X"]])

        A <- scan1perm(genoprobs=genoprobs[,!is_x_chr], pheno=pheno,
                       kinship=subset_kinship(kinship, chr=!is_x_chr),
                       addcovar=addcovar, Xcovar=NULL, intcovar=intcovar, weights=weights,
                       reml=reml, model=model, n_perm=n_perm, perm_Xsp=FALSE,
                       perm_strata=perm_strata, chr_lengths=NULL, cores=cores,
                       scan_func=scan_func, ...)
        X <- scan1perm(genoprobs=genoprobs[,is_x_chr], pheno=pheno,
                       kinship=subset_kinship(kinship, chr=is_x_chr),
                       addcovar=addcovar, Xcovar=Xcovar, intcovar=intcovar, weights=weights,
                       reml=reml, model=model, n_perm=n_permX, perm_Xsp=FALSE,
                       perm_strata=perm_strata, chr_lengths=NULL, cores=cores,
                       scan_func=scan_func, ...)
        result <- list(A=A, X=X)
        attr(result, "chr_lengths") <- chr_lengths

        class(result$A) <- class(result$X) <- "matrix"
        class(result) <- c("scan1perm", "list")
        return(result)
    }

    # force things to be matrices
    if(!is.matrix(pheno)) {
        pheno <- as.matrix(pheno)
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }
    if(is.null(colnames(pheno))) # force column names
        colnames(pheno) <- paste0("pheno", seq_len(ncol(pheno)))
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(Xcovar)) {
        if(!is.matrix(Xcovar)) Xcovar <- as.matrix(Xcovar)
        if(!is.numeric(Xcovar)) stop("Xcovar is not numeric")
    }
    if(!is.null(intcovar)) {
        if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)
        if(!is.numeric(intcovar)) stop("intcovar is not numeric")
    }

    # for binary model
    if(model=="binary") {
        if(!is.null(kinship))
            stop("Can't yet account for kinship with model = \"binary\"")
        pheno <- check_binary_pheno(pheno)
    }
    else {
        # square-root of weights (only if model="normal")
        weights <- sqrt_weights(weights) # also check >0 (and if all 1's, turn to NULL)
    }

    # check that kinship matrices are square with same IDs
    kinshipIDs <- check_kinship(kinship, length(genoprobs))

    # multiply kinship matrix by 2; rest is using 2*kinship
    # see Almasy & Blangero (1998) https://doi.org/10.1086/301844
    kinship <- double_kinship(kinship)

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, addcovar, Xcovar, intcovar,
                               kinshipIDs, weights, perm_strata, complete.cases=TRUE)
    ind2keep <- get_common_ids(ind2keep, pheno[rowSums(is.finite(pheno)) > 0,,drop=FALSE])
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # make sure columns in intcovar are also in addcovar
    addcovar <- force_intcovar(addcovar, intcovar, tol)

    # drop things from Xcovar that are already in addcovar
    Xcovar <- drop_xcovar(addcovar, Xcovar, tol)

    if(!is.null(scan_func)) {
        if(model != "normal") warning("model argument ignored if scan_func provided")
        # no covariates, no weights, no missing phenotypes
        if(is.null(addcovar) && is.null(Xcovar) && is.null(intcovar)
           && is.null(weights) && is.null(kinship) && sum(!is.finite(pheno[ind2keep,]))==0) {
            return(scan1perm_gen_simple(genoprobs=genoprobs, pheno=pheno,
                                        n_perm=n_perm, perm_strata=perm_strata,
                                        cores=cores,
                                        scan_func=scan_func, ind2keep=ind2keep, ...))
        } else {
            return(scan1perm_gen(genoprobs=genoprobs, pheno=pheno,
                                 kinship=kinship, addcovar=addcovar, Xcovar=Xcovar,
                                 intcovar=intcovar, weights=weights,
                                 n_perm=n_perm, perm_strata=perm_strata,
                                 cores=cores,
                                 scan_func=scan_func, ind2keep=ind2keep, ...))
        }
    }

    if(!is.null(kinship)) { # fit linear mixed model
        return(scan1perm_pg(genoprobs=genoprobs, pheno=pheno, kinship=kinship,
                            addcovar=addcovar, Xcovar=Xcovar, intcovar=intcovar,
                            weights=weights,
                            reml=reml, n_perm=n_perm, perm_strata=perm_strata,
                            cores=cores, ind2keep=ind2keep, ...))
    }

    if(model=="normal" && is.null(addcovar) && is.null(Xcovar) &&
       is.null(intcovar) && is.null(weights)
       && sum(!is.finite(pheno[ind2keep,]))==0) # no covariates, no weights, no missing phenotypes
        result <- scan1perm_nocovar(genoprobs=genoprobs,
                                    pheno=pheno,
                                    n_perm=n_perm,
                                    perm_strata=perm_strata,
                                    cores=cores,
                                    ind2keep=ind2keep,
                                    ...)
    else
        result <- scan1perm_covar(genoprobs=genoprobs,
                                  pheno=pheno,
                                  addcovar=addcovar,
                                  Xcovar=Xcovar,
                                  intcovar=intcovar,
                                  weights=weights,
                                  model=model,
                                  n_perm=n_perm,
                                  perm_strata=perm_strata,
                                  cores=cores,
                                  ind2keep=ind2keep,
                                  ...)

    result
}


# simplest version: no covariates, no weights, no missing phenotypes
scan1perm_nocovar <-
    function(genoprobs, pheno, n_perm=1, perm_strata=NULL,
             cores=1, ind2keep, ...)
{
    # deal with the dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    stopifnot(tol > 0)
    quiet <- grab_dots(dotargs, "quiet", TRUE)
    check_extra_dots(dotargs, c("tol", "intcovar_method", "quiet", "max_batch"))

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # max batch size
    max_batch <- grab_dots(dotargs, "max_batch",
                           min(1000, ceiling(n_perm*length(genoprobs)*ncol(pheno)/n_cores(cores))))

    # generate permutations
    perms <- gen_strat_perm(n_perm, ind2keep, perm_strata)

    # batch permutations
    phe_batches <- batch_vec( rep(seq_len(ncol(pheno)), n_perm), max_batch)
    perm_batches <- batch_vec( rep(seq_len(n_perm), each=ncol(pheno)), max_batch)

    # drop cols in genotype probs that are all 0 (just looking at the X chromosome)
    genoprob_Xcol2drop <- genoprobs_col2drop(genoprobs)
    is_x_chr <- attr(genoprobs, "is_x_chr")
    if(is.null(is_x_chr)) is_x_chr <- rep(FALSE, length(genoprobs))

    # batches for analysis, to allow parallel analysis
    run_batches <- data.frame(chr=rep(seq_len(length(genoprobs)), length(phe_batches)),
                              phe_batch=rep(seq_along(phe_batches), each=length(genoprobs)))

    run_indexes <- seq_len(length(genoprobs)*length(phe_batches))

    # subset the phenotypes
    pheno <- pheno[ind2keep,,drop=FALSE]
    nind <- nrow(pheno)

    nullrss <- apply(pheno, 2, function(ph) nullrss_clean(as.matrix(ph), NULL, NULL, add_intercept=TRUE, tol))

    # the function that does the work
    by_group_func <- function(i) {
        # deal with batch information, including individuals to drop due to missing phenotypes
        chr <- run_batches$chr[i]
        chrnam <- names(genoprobs)[chr]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        permbatch <- perm_batches[[run_batches$phe_batch[i]]]

        # subset the genotype probabilities: drop cols with all 0s, plus the first column
        Xcol2drop <- genoprob_Xcol2drop[[chrnam]]
        if(length(Xcol2drop) > 0) {
            pr <- genoprobs[[chr]][ind2keep,-Xcol2drop,,drop=FALSE]
            pr <- pr[,-1,,drop=FALSE]
        }
        else
            pr <- genoprobs[[chr]][ind2keep,-1,,drop=FALSE]

        ph <- pheno[,phebatch,drop=FALSE]
        for(col in seq_len(ncol(ph))) # permute columns
            ph[,col] <- ph[perms[,permbatch[col]] , col]

        # scan1 function taking clean data (with no missing values)
        rss <- scan1_clean(pr, ph, NULL, NULL, NULL, add_intercept=TRUE, tol, "lowmem")

        # calculate LOD score
        lod <- nind/2 * (log10(nullrss[phebatch]) - log10(rss))

        # return column maxima
        apply(lod, 1, max, na.rm=TRUE)
    }

    # calculations in parallel
    list_result <- cluster_lapply(cores, run_indexes, by_group_func)

    # check for problems (if clusters run out of memory, they'll return NULL)
    result_is_null <- vapply(list_result, is.null, TRUE)
    if(any(result_is_null))
        stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

    # reorganize results
    result <- array(dim=c(length(genoprobs), n_perm, ncol(pheno)))

    for(i in run_indexes) {
        chr <- run_batches$chr[i]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        permbatch <- perm_batches[[run_batches$phe_batch[i]]]

        for(j in seq_along(phebatch))
            result[chr,permbatch[j], phebatch[j]] <- list_result[[i]][j]

    }

    result <- apply(result, c(2,3), max)
    colnames(result) <- colnames(pheno)

    class(result) <- c("scan1perm", "matrix")
    result
}

# permutations with covariates and/or different batches of phenotypes
# also covers the case of a binary phenotype
scan1perm_covar <-
    function(genoprobs, pheno, addcovar=NULL, Xcovar=NULL, intcovar=NULL,
             weights=NULL, model=c("normal", "binary"),
             n_perm=1, perm_strata=NULL, cores=1, ind2keep, ...)
{
    model <- match.arg(model)

    # deal with the dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    if(!is_pos_number(tol)) stop("tol should be a single positive number")
    intcovar_method <- grab_dots(dotargs, "intcovar_method", "lowmem",
                                 c("highmem", "lowmem"))
    quiet <- grab_dots(dotargs, "quiet", TRUE)
    if(model=="binary") {
        bintol <- grab_dots(dotargs, "bintol", sqrt(tol))
        if(!is_pos_number(bintol)) stop("bintol should be a single positive number")
        eta_max <- grab_dots(dotargs, "eta_max", log(1-tol)-log(tol)) # for model="binary"
        if(!is_pos_number(eta_max)) stop("eta_max should be a single positive number")
        maxit <- grab_dots(dotargs, "maxit", 100)
        if(!is_nonneg_number(maxit)) stop("maxit should be a single non-negative integer")
        check_extra_dots(dotargs, c("tol", "intcovar_method", "quiet", "max_batch",
                                    "maxit", "bintol", "eta_max"))
    }
    else {
        check_extra_dots(dotargs, c("tol", "intcovar_method", "quiet", "max_batch"))
    }

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # max batch size
    max_batch <- grab_dots(dotargs, "max_batch",
                           min(1000, ceiling(n_perm*length(genoprobs)*ncol(pheno)/n_cores(cores))))

    # generate permutations
    perms <- gen_strat_perm(n_perm, ind2keep, perm_strata)

    # batch permutations
    phe_batches <- batch_cols(pheno[ind2keep,,drop=FALSE], max_batch)

    # drop cols in genotype probs that are all 0 (just looking at the X chromosome)
    genoprob_Xcol2drop <- genoprobs_col2drop(genoprobs)
    is_x_chr <- attr(genoprobs, "is_x_chr")
    if(is.null(is_x_chr)) is_x_chr <- rep(FALSE, length(genoprobs))

    # batches for analysis, to allow parallel analysis
    run_batches <- data.frame(chr=rep(seq_len(length(genoprobs)), length(phe_batches)*n_perm),
                              phe_batch=rep(seq_along(phe_batches), each=length(genoprobs)*n_perm),
                              perm_batch=rep(rep(seq_len(n_perm), each=length(genoprobs), length(phe_batches))))

    run_indexes <- seq_len(length(genoprobs)*length(phe_batches)*n_perm)

    # the function that does the work
    by_group_func <- function(i) {
        # deal with batch information, including individuals to drop due to missing phenotypes
        chr <- run_batches$chr[i]
        chrnam <- names(genoprobs)[chr]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        permbatch <- run_batches$perm_batch[i]
        phecol <- phebatch$cols
        omit <- phebatch$omit
        these2keep <- ind2keep # individuals 2 keep for this batch
        if(length(omit) > 0) these2keep <- ind2keep[-omit]
        if(length(these2keep)<=2) return(NULL) # not enough individuals

        # apply permutation to the probs
        #     (I initially thought we'd need to apply these just to the rownames, but
        #     actually we've already aligned probs and pheno and so forth
        #     via ind2keep/these2keep, so now we need to just permute the rows)
        #
        # (also need some contortions here to keep the sizes the same)
        pr <- genoprobs[[chr]][ind2keep,,,drop=FALSE]
        pr <- pr[perms[,permbatch],,,drop=FALSE]
        rownames(pr) <- ind2keep
        pr <- pr[these2keep,,,drop=FALSE]

        # subset the genotype probabilities: drop cols with all 0s, plus the first column
        Xcol2drop <- genoprob_Xcol2drop[[chrnam]]
        if(length(Xcol2drop) > 0) {
            pr <- pr[,-Xcol2drop,,drop=FALSE]
        }
        pr <- pr[,-1,,drop=FALSE]

        # subset the rest
        ac <- addcovar; if(!is.null(ac)) ac <- ac[these2keep,,drop=FALSE]
        Xc <- Xcovar;   if(!is.null(Xc)) Xc <- Xc[these2keep,,drop=FALSE]
        ic <- intcovar; if(!is.null(ic)) ic <- ic[these2keep,,drop=FALSE]
        wts <- weights; if(!is.null(wts)) wts <- wts[these2keep]
        ph <- pheno[these2keep, phecol, drop=FALSE]

        # if X chr, paste X covariates onto additive covariates
        # (only for the null)
        if(is_x_chr[chr]) ac0 <- drop_depcols(cbind(ac, Xc), add_intercept=FALSE, tol)
        else ac0 <- ac

        if(model=="normal") {
            # FIX_ME: calculating null RSS multiple times :(
            nullrss <- nullrss_clean(ph, ac0, wts, add_intercept=TRUE, tol)

            # scan1 function taking clean data (with no missing values)
            rss <- scan1_clean(pr, ph, ac, ic, wts, add_intercept=TRUE, tol, intcovar_method)

            # calculate LOD score
            lod <- nrow(ph)/2 * (log10(nullrss) - log10(rss))
        }
        else { # binary model
            # FIX_ME: calculating null LOD multiple times :(
            nulllod <- null_binary_clean(ph, ac0, wts, add_intercept=TRUE, maxit, bintol, tol, eta_max)

            # scan1 function taking clean data (with no missing values)
            lod <- scan1_binary_clean(pr, ph, ac, ic, wts, add_intercept=TRUE,
                                      maxit, bintol, tol, intcovar_method, eta_max)

            # calculate LOD score
            lod <- lod - nulllod
        }

        # return column maxima
        apply(lod, 1, max, na.rm=TRUE)
    }

    # calculations in parallel
    list_result <- cluster_lapply(cores, run_indexes, by_group_func)

    # check for problems (if clusters run out of memory, they'll return NULL)
    result_is_null <- vapply(list_result, is.null, TRUE)
    if(any(result_is_null))
        stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

    # reorganize results
    result <- array(dim=c(length(genoprobs), n_perm, ncol(pheno)))

    for(i in run_indexes) {
        chr <- run_batches$chr[i]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        permbatch <- run_batches$perm_batch[i]

        result[chr, permbatch, phebatch$cols] <- list_result[[i]]
    }

    result <- apply(result, c(2,3), max)
    colnames(result) <- colnames(pheno)

    class(result) <- c("scan1perm", "matrix")
    result
}



# generate (potentially) stratified permutations
#   (actual work is done in c++, see src/random.cpp)
gen_strat_perm <-
    function(n_perm, ind2keep, perm_strata=NULL)
{
    if(!is.null(perm_strata)) {
        perm_strata <- perm_strata[ind2keep]
        u <- unique(perm_strata)

        if(length(u) > 1) {

            strat_numeric <- match(perm_strata, u)-1

            return(permute_nvector_stratified(n_perm,
                                              seq_along(ind2keep),
                                              strat_numeric,
                                              length(u)))
        }
    }

    permute_nvector(n_perm, seq_along(ind2keep))
}

# calculate summary statistics from raw SNP data

#' Calculate estimated heterozygosity from raw SNP genotypes
#'
#' Calculate estimated heterozygosity for each individual from raw SNP genotypes
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param by Indicates whether to summarize by founder strain (`"individual"`) or by marker.
#'
#' @return A vector of heterozygosities, one for each individual or marker.
#'
#' @export
#' @keywords utilities
#' @seealso [recode_snps()], [calc_raw_maf()], [calc_raw_founder_maf()], [calc_raw_geno_freq()], [calc_het()]
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' DOex_het <- calc_raw_het(DOex)
#' }
calc_raw_het <-
    function(cross, by=c("individual", "marker"))
{
    by <- match.arg(by)

    g <- do.call("cbind", cross$geno)
    g[g < 1 | g > 3] <- NA

    if(by=="individual") return( rowMeans(g==2, na.rm=TRUE) )

    colMeans(g==2, na.rm=TRUE)
}



#' Calculate minor allele frequency from raw SNP genotypes
#'
#' Calculate minor allele frequency from raw SNP genotypes, by individual or by marker
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param by Indicates whether to summarize by founder strain (`"individual"`) or by marker.
#'
#' @return A vector of minor allele frequencies, one for each individual or marker.
#'
#' @export
#' @keywords utilities
#' @seealso [recode_snps()], [calc_raw_founder_maf()], [calc_raw_het()], [calc_raw_geno_freq()]
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' DOex_maf <- calc_raw_maf(DOex)
#' }
calc_raw_maf <-
    function(cross, by=c("individual", "marker"))
{
    by <- match.arg(by)

    g <- do.call("cbind", cross$geno)
    g[g < 1 | g > 3] <- NA

    if(by=="individual") {
        return( rowMeans(g==3, na.rm=TRUE) + rowMeans(g==2, na.rm=TRUE)/2 )
    }

    # else by marker
    colMeans(g==3, na.rm=TRUE) + colMeans(g==2, na.rm=TRUE)/2

}


#' Calculate founder minor allele frequencies from raw SNP genotypes
#'
#' Calculate minor allele frequency from raw SNP genotypes in founders, by founder strain or by marker
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param by Indicates whether to summarize by founder strain (`"individual"`) or by marker.
#'
#' @return A vector of minor allele frequencies, one for each founder strain or marker.
#'
#' @export
#' @keywords utilities
#' @seealso [recode_snps()], [calc_raw_maf()]
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' DOex_maf <- calc_raw_founder_maf(DOex)
#' }
calc_raw_founder_maf <-
    function(cross, by=c("individual", "marker"))
{
    if(!("founder_geno" %in% names(cross))) {
        stop("cross does not contain founder genotypes")
    }

    by <- match.arg(by)

    fg <- do.call("cbind", cross$founder_geno)
    fg[fg!=1 & fg!=3] <- NA

    if(by=="individual") {
        return( rowMeans(fg==3, na.rm=TRUE) )
    }

    # else by marker
    colMeans(fg==3, na.rm=TRUE)
}


#' Calculate genotype frequencies from raw SNP genotypes
#'
#' Calculate genotype frequencies from raw SNP genotypes, by individual or by marker
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param by Indicates whether to summarize by individual or by marker.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A matrix of genotypes frequencies with 3 columns (AA, AB,
#' and BB) and with rows being either individuals or markers.
#'
#' @export
#' @keywords utilities
#' @seealso [calc_raw_maf()], [calc_raw_het()], [recode_snps()], [calc_geno_freq()]
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' gfreq <- calc_raw_geno_freq(DOex)
#' }
calc_raw_geno_freq <-
    function(cross, by=c("individual", "marker"), cores=1)
{
    by <- match.arg(by)

    g <- do.call("cbind", cross$geno)
    g[g<1 | g>3] <- NA

    cores <- setup_cluster(cores)

    if(n_cores(cores)==1) {
        if(by=="individual") {
            result <- t(apply(g, 1, function(a) table(factor(a, levels=1:3))))
        } else {
            result <- t(apply(g, 2, function(a) table(factor(a, levels=1:3))))
        }
    } else {

        if(by=="individual") {

            rn <- rownames(g)
            result <- cluster_lapply(cores, seq_len(nrow(g)), function(row_index) table(factor(g[row_index,], levels=1:3)))
        } else {

            rn <- colnames(g)
            result <- cluster_lapply(cores, seq_len(ncol(g)), function(col_index) table(factor(g[,col_index], levels=1:3)))
        }

        result <- matrix(unlist(result), byrow=TRUE, ncol=3)
        rownames(result) <- rn
    }

    colnames(result) <- c("AA", "AB", "BB")

    result/rowSums(result)
}

context("calc raw summaries")

test_that("calculation of raw summaries work", {

    skip_if(isnt_karl(), "this test only run locally")

    # load example data
    file <- paste0("https://raw.githubusercontent.com/rqtl/",
                   "qtl2data/main/DOex/DOex.zip")
    DOex <- read_cross2(file)
    indIDs <- ind_ids(DOex)
    mnames <- marker_names(DOex)
    founder_ids <- rownames(DOex$founder_geno[[1]])

    g <- do.call("cbind", DOex$geno)
    fg <- do.call("cbind", DOex$founder_geno)
    g[g==0] <- NA
    fg[fg==0] <- NA

    # calc_raw_het()
    het <- calc_raw_het(DOex)
    expect_equal(names(het), indIDs)
    expect_true(all(het >= 0 & het <= 1))
    expect_equal(het, rowMeans(g==2, na.rm=TRUE))

    # calc_raw_maf() by marker
    het <- calc_raw_het(DOex, "marker")
    expect_equal(names(het), mnames)
    expect_true(all(het >= 0 & het <= 1))
    expect_equal(het, colMeans(g==2, na.rm=TRUE))

    # calc_raw_maf()
    maf_ind <- calc_raw_maf(DOex)
    maf_mar <- calc_raw_maf(DOex, "marker")
    expect_equal(names(maf_ind), indIDs)
    expect_equal(names(maf_mar), mnames)
    expect_equal(maf_ind, rowMeans(g==3, na.rm=TRUE) + rowMeans(g==2, na.rm=TRUE)/2)
    expect_equal(maf_mar, colMeans(g==3, na.rm=TRUE) + colMeans(g==2, na.rm=TRUE)/2)

    # calc_raw_founder_maf()
    maf_ind <- calc_raw_founder_maf(DOex)
    maf_mar <- calc_raw_founder_maf(DOex, "marker")
    expect_equal(names(maf_ind), founder_ids)
    expect_equal(names(maf_mar), mnames)
    expect_equal(maf_ind, rowMeans(fg==3, na.rm=TRUE))
    expect_equal(maf_mar, colMeans(fg==3, na.rm=TRUE))

    # calc_raw_geno_freq()
    gf_ind <- calc_raw_geno_freq(DOex)
    gf_mar <- calc_raw_geno_freq(DOex, "marker")
    expect_equal(dimnames(gf_ind), list(indIDs, c("AA", "AB", "BB")))
    expect_equal(dimnames(gf_mar), list(mnames, c("AA", "AB", "BB")))
    expect_equal(gf_ind, cbind(AA=rowMeans(g==1, na.rm=TRUE), AB=rowMeans(g==2, na.rm=TRUE), BB=rowMeans(g==3, na.rm=TRUE)))
    expect_equal(gf_mar, cbind(AA=colMeans(g==1, na.rm=TRUE), AB=colMeans(g==2, na.rm=TRUE), BB=colMeans(g==3, na.rm=TRUE)))

    # multi-core
    gf_ind_2 <- calc_raw_geno_freq(DOex, cores=2)
    gf_mar_2 <- calc_raw_geno_freq(DOex, "marker", cores=2)
    expect_equal(gf_ind_2, gf_ind)
    expect_equal(gf_mar_2, gf_mar)

})

# swap_colname
#
# replace column name in dataframe
# column labeled orig_name becomes new_name
# if there had already been a column labeled new_name,
#     we append "_old"
#
# (used in create_variant_query_func and create_gene_query_func)
swap_colname <-
    function(df, orig_name, new_name)
{
    if(orig_name == new_name) return(df)

    cn <- colnames(df)
    if(!any(cn == orig_name)) {
        stop('field "', orig_name, '" not found')
    }

    if(any(cn==new_name)) {
        cn[cn==new_name] <- paste0(new_name, "_old")
    }
    cn[cn==orig_name] <- new_name

    colnames(df) <- cn
    df
}

#' Batch columns by pattern of missing values
#'
#' Identify batches of columns of a matrix that have the same pattern
#' of missing values.
#'
#' @param mat A numeric matrix
#' @param max_batch Maximum batch size
#'
#' @return A list containing the batches, each with two components:
#' `cols` containing numeric indices of the columns in the
#' corresponding batch, and `omit` containing a vector of row indices
#' that have missing values in this batch.
#'
#' @export
#' @seealso [batch_vec()]
#' @examples
#' x <- rbind(c( 1,  2,  3, 13, 16),
#'            c( 4,  5,  6, 14, 17),
#'            c( 7, NA,  8, NA, 18),
#'            c(NA, NA, NA, NA, 19),
#'            c(10, 11, 12, 15, 20))
#' batch_cols(x)
batch_cols <-
    function(mat, max_batch=NULL)
{
    if(is.null(mat)) stop("mat is NULL")
    if(!is.matrix(mat)) mat <- as.matrix(mat)

    if(is.null(max_batch) || max_batch<=0)
        max_batch <- ncol(mat)

    mat <- !is.finite(mat)
    n <- nrow(mat)
    all_true <- rep(TRUE, n)
    result <- NULL

    # first pull out the columns that have 0 NAs
    n_na <- colSums(mat)
    no_na <- (n_na==0)
    if(any(no_na))
        result <- list(list(cols=which(no_na),
                                omit=numeric(0)))

    # now deal with the columns that have exactly 1 NA
    one_na <- (n_na==1)
    if(any(one_na)) {
        wh <- apply(mat[,one_na,drop=FALSE], 2, which)
        spl <- split(which(one_na), wh)
        part2 <- lapply(seq_along(spl), function(i)
                        list(cols=as.numeric(spl[[i]]),
                             omit=as.numeric(names(spl)[i])))

        if(is.null(result)) result <- part2
        else result <- c(result, part2)
    }

    # now the columns with >1 NA
    other_cols <- !(no_na | one_na)
    if(any(other_cols)) {
        other_cols <- seq_len(ncol(mat))[other_cols]

        # pattern of missing data (as character string with 0's and 1's
        pat <- apply(mat[,other_cols,drop=FALSE], 2, function(a) paste(which(a), collapse=":"))

        # the unique patterns
        u <- unique(pat)

        # the result
        part3 <- lapply(u, function(a)
                        list(cols=other_cols[pat==a],
                             omit=as.numeric(strsplit(a, ":")[[1]])))

    } else part3 <- NULL

    if(is.null(result)) result <- part3
    else result <- c(result, part3)

    # reduce to max_batch
    sizes <- vapply(result, function(a) length(a$cols), 1)
    big <- sizes > max_batch
    if(!any(big)) return(result)

    toreduce <- which(big)
    for(i in toreduce) {
        omit <- result[[i]]$omit
        col_list <- batch_vec(result[[i]]$cols, max_batch)
        this <- lapply(col_list, function(a) list(cols=a, omit=omit))
        result <- c(result, this)
    }

   result[-toreduce]
}

context("batch columns by NAs")

test_that("batch_cols works", {

    x <- rbind(c( 1,  2,  3),
               c( 4,  5,  6),
               c( 7, NA,  8),
               c(NA, NA,  9),
               c(10, 11, 12))

    expect_equal(batch_cols(x),
                 list(list(cols=3, omit=numeric(0)),
                      list(cols=1, omit=4),
                      list(cols=2, omit=c(3,4)))
                 )

    expect_equal(batch_cols( rbind(c(1,2,3)) ),
                 list(list(cols=c(1,2,3), omit=numeric(0)))
                 )

    expect_equal(batch_cols( rbind(1, 2, NA, 3) ),
                 list(list(cols=1, omit=3))
                 )

    y <- rbind(c( 1,  2,  3, 13, 16),
               c( 4,  5,  6, 14, 17),
               c( 7, NA,  8, NA, 18),
               c(NA, NA, NA, NA, 19),
               c(10, 11, 12, 15, 20))
    expect_equal(batch_cols(y),
                 list(list(cols=5, omit=numeric(0)),
                      list(cols=c(1,3), omit=4),
                      list(cols=c(2,4), omit=c(3,4)))

                 )
})

test_that("batch_vec works", {

    # one big batch
    expect_equal(batch_vec(1:100, 100), list(1:100))

    # one element
    expect_equal(batch_vec(1, 1), list(1))
    expect_equal(batch_vec(1, 8), list(1))

    # fewer elements than batch size
    expect_equal(batch_vec(1:5, 8), list(1:5))
    expect_equal(batch_vec(1:30, 32), list(1:30))

    # equal elements in each core
    expect <- as.list(as.data.frame(matrix(1:10, nrow=5, byrow=FALSE)))
    names(expect) <- NULL
    expect_equal(batch_vec(1:10, 5), expect)

    expect <- as.list(as.data.frame(matrix(1:10, nrow=2, byrow=FALSE)))
    names(expect) <- NULL
    expect_equal(batch_vec(1:10, 2), expect)

    expect <- as.list(as.data.frame(matrix(1:144, nrow=18, byrow=FALSE)))
    names(expect) <- NULL
    expect_equal(batch_vec(1:144, 18), expect)

    expect <- as.list(as.data.frame(matrix(1:144, nrow=8, byrow=FALSE)))
    names(expect) <- NULL
    expect_equal(batch_vec(1:144, 8), expect)

    # non-equal
    expect_equal(batch_vec(101:110, 3),
                      list(101:103, 104:106, 107:108, 109:110))
    expect_equal(batch_vec(1:35, 6),
                      list(1:6, 7:12, 13:18, 19:24, 25:30, 31:35))
    expect_equal(batch_vec((1:16)+3, 9),
                      list((1:8)+3, (9:16)+3))

    set.seed(20151202)
    vec <- sample((1:17)+4)
    expect_equal(batch_vec(vec, 9),
                      list(vec[1:9], vec[10:17]))

})


test_that("batch_cols with max_batch works", {

    set.seed(20151202)

    n <- 100
    p <- 1000
    x <- matrix(rnorm(n*p), ncol=p)
    expect_equal(batch_cols(x),
                 list(list(cols=1:p, omit=numeric(0))))

    expect <- vector("list", 4)
    for(i in 1:4)
        expect[[i]] <- list(cols=1:250+(i-1)*250, omit=numeric(0))
    expect_equal(batch_cols(x, max_batch=250), expect)

    x[16:20, 8:12] <- NA
    expect <- vector("list", 5)
    expect[[1]] <- list(cols=8:12, omit=16:20)
    rest <- (1:1000)[-(8:12)]
    expect[[2]] <- list(cols=rest[1:249], omit=numeric(0))
    expect[[3]] <- list(cols=rest[249+1:249], omit=numeric(0))
    expect[[4]] <- list(cols=rest[2*249+1:249], omit=numeric(0))
    expect[[5]] <- list(cols=rest[3*249+1:248], omit=numeric(0))
    expect_equal(batch_cols(x, max_batch=250), expect)

})

#' General genome scan with a single-QTL model
#'
#' General genome scan with a single-QTL model, with possible allowance for covariates
#' and a polygeneic effect, but taking a general function that calculates a LOD score
#' at a single position.
#'
#' @param genoprobs Genotype probabilities as calculated by
#' [calc_genoprob()].
#' @param pheno A numeric matrix of phenotypes, individuals x phenotypes.
#' @param kinship Optional kinship matrix, or a list of kinship matrices (one
#' per chromosome), in order to use the LOCO (leave one chromosome
#' out) method.
#' @param addcovar An optional numeric matrix of additive covariates.
#' @param Xcovar An optional numeric matrix with additional additive covariates used for
#' null hypothesis when scanning the X chromosome.
#' @param intcovar An numeric optional matrix of interactive covariates.
#' @param weights An optional numeric vector of positive weights for the
#' individuals. As with the other inputs, it must have `names`
#' for individual identifiers.
#' @param func Function to calculate log10 likelihood. It is called as
#' `func(pr, pheno, kinship, addcovar, intcovar, weights, ...)`
#' where `pr` is for the genotype probabilities at a fixed
#' position. It will also be used to calculate the null log10
#' likelihood, with `pr` being NULL in that case.
#' @param vectorize_func If TRUE (the default), assume that `func`
#' takes only a single phenotype. If FALSE, assume `func` is able
#' to take a matrix of phenotypes and return a vector of log10
#' likelihood values.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' @param ... Additional arguments passed to `func()`.
#'
#' @return An object of class `"scan1"`: a matrix of LOD scores, positions x phenotypes.
#'
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c("19", "X")] # subset to chr 19 and X}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # covariates for X chr under null
#' Xcovar <- get_x_covar(iron)
#'
#' # create binary trait
#' bin_pheno <- setNames(as.numeric(iron$pheno[,1] > median(iron$pheno[,1])),
#'                       rownames(iron$pheno))
#'
#' ll_glm <-
#'    function(pr, pheno, addcovar=NULL, ...)
#' {
#'    formula <- ifelse(is.null(pr), "pheno ~ 1", "pheno ~ pr")
#'    if(!is.null(addcovar)) formula <- paste(formula, "+ addcovar")
#'
#'     glm_out <- glm(as.formula(formula), family=binomial(link=probit))
#'     -glm_out$deviance/(2*log(10)) # log10 likelihood
#' }
#'
#' # perform genome scan using glm() with probit link function
#' out <- scan1gen(probs, bin_pheno, Xcovar=Xcovar, func=ll_glm)
#'
#' @seealso [scan1()], [scan1perm()], [scan1max()], [scan1gen examples](https://kbroman.org/qtl2/assets/vignettes/scan1gen.html)
#'
#' @export
scan1gen <-
    function(genoprobs, pheno, kinship=NULL, addcovar=NULL, Xcovar=NULL,
             intcovar=NULL, weights=NULL, func, vectorize_func=TRUE,
             cores=1, ...)
{
    if(is.null(genoprobs)) stop("genoprobs is NULL")
    if(is.null(pheno)) stop("pheno is NULL")

    # deal with the dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    quiet <- grab_dots(dotargs, "quiet", TRUE)
    max_batch <- grab_dots(dotargs, "max_batch", NULL)
    if(!is.null(max_batch) && !is_pos_number(max_batch)) stop("max_batch should be a single positive integer")
    else {
        check_extra_dots(dotargs, c("tol", "quiet", "max_batch"))
    }

    # check that the objects have rownames
    check4names(pheno, addcovar, Xcovar, intcovar)

    # force things to be matrices
    if(!is.matrix(pheno)) {
        pheno <- as.matrix(pheno)
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }
    if(is.null(colnames(pheno))) # force column names
        colnames(pheno) <- paste0("pheno", seq_len(ncol(pheno)))
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(Xcovar)) {
        if(!is.matrix(Xcovar)) Xcovar <- as.matrix(Xcovar)
        if(!is.numeric(Xcovar)) stop("Xcovar is not numeric")
    }
    if(!is.null(intcovar)) {
        if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)
        if(!is.numeric(intcovar)) stop("intcovar is not numeric")
    }

    # check that kinship matrices are square with same IDs
    kinshipIDs <- check_kinship(kinship, length(genoprobs))

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, addcovar, Xcovar, intcovar,
                               weights, kinshipIDs, complete.cases=TRUE)
    ind2keep <- get_common_ids(ind2keep, pheno[rowSums(is.finite(pheno)) > 0,,drop=FALSE])
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # make sure columns in intcovar are also in addcovar
    addcovar <- force_intcovar(addcovar, intcovar, tol)

    # drop things from Xcovar that are already in addcovar
    Xcovar <- drop_xcovar(addcovar, Xcovar, tol)

    phe_batches <- batch_cols(pheno[ind2keep,,drop=FALSE], max_batch)

    # drop cols in genotype probs that are all 0 (just looking at the X chromosome)
    genoprob_Xcol2drop <- genoprobs_col2drop(genoprobs)
    is_x_chr <- attr(genoprobs, "is_x_chr")
    if(is.null(is_x_chr)) is_x_chr <- rep(FALSE, length(genoprobs))

    # vectorize function if necessary
    if(ncol(pheno)>1 && vectorize_func) {
        vfunc <- function(pr, pheno, ...) {
            if(ncol(pheno) > 1)
                return(sapply(1:ncol(pheno), function(i) func(pr, pheno[,i,drop=FALSE], ...)))
            else return(func(pr, pheno, ...))
        }
    } else {
        vfunc <- func
    }

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # batches for analysis, to allow parallel analysis
    run_batches <- data.frame(chr=rep(seq_len(length(genoprobs)), length(phe_batches)),
                              phe_batch=rep(seq_along(phe_batches), each=length(genoprobs)))
    run_indexes <- seq_len(length(genoprobs)*length(phe_batches))

    # the function that does the work
    by_group_func <- function(i) {
        # deal with batch information, including individuals to drop due to missing phenotypes
        chr <- run_batches$chr[i]
        chrnam <- names(genoprobs)[chr]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        phecol <- phebatch$cols
        omit <- phebatch$omit
        these2keep <- ind2keep # individuals 2 keep for this batch
        if(length(omit) > 0) these2keep <- ind2keep[-omit]
        if(length(these2keep)<=2) return(NULL) # not enough individuals

        # subset the genotype probabilities: drop cols with all 0s, plus the first column
        Xcol2drop <- genoprob_Xcol2drop[[chrnam]]
        if(length(Xcol2drop) > 0) {
            pr <- genoprobs[[chr]][these2keep,-Xcol2drop,,drop=FALSE]
            pr <- pr[,-1,,drop=FALSE]
        } else pr <- genoprobs[[chr]][these2keep,-1,,drop=FALSE]

        # subset the rest
        K <- subset_kinship(kinship, ind=these2keep)
        ac <- addcovar; if(!is.null(ac)) { ac <- ac[these2keep,,drop=FALSE]; ac <- drop_depcols(ac, TRUE, tol) }
        Xc <- Xcovar;   if(!is.null(Xc)) Xc <- Xc[these2keep,,drop=FALSE]
        ic <- intcovar; if(!is.null(ic)) { ic <- ic[these2keep,,drop=FALSE]; ic <- drop_depcols(ic, TRUE, tol) }
        ph <- pheno[these2keep,phecol,drop=FALSE]
        wts <- weights[these2keep]

        # if X chr, paste X covariates onto additive covariates
        # (only for the null)
        if(is_x_chr[chr]) ac0 <- drop_depcols(cbind(ac, Xc), add_intercept=FALSE, tol)
        else ac0 <- ac

        # if kinship is list, subset to this chr
        if(is_kinship_list(K)) K <- K[[chr]]

        # FIX_ME: fitting null model multiple times :(
        ll0 <- vfunc(NULL, ph, addcovar=ac0, weights=wts, kinship=K)

        ll <- matrix(ncol=ncol(ph), nrow=dim(pr)[3])
        for(pos in 1:nrow(ll)) {
            # make sure genoprobs are passed as a matrix with column names
            p <- as.matrix(pr[,,pos])
            if(is.null(colnames(p))) colnames(p) <- paste0("pr", 1:ncol(p))

            ll[pos,] <- vfunc(p, ph, addcovar=ac, intcovar=ic, kinship=K, weights=wts)
        }

        # calculate LOD score
        lod <- t(ll) - ll0 # deal with multiple phenotypes

        list(lod=lod, n=nrow(ph)) # return LOD & number of individuals used
    }

    # number of markers/pseudomarkers by chromosome, and their indexes to result matrix
    npos_by_chr <- dim(genoprobs)[3,]
    totpos <- sum(npos_by_chr)
    pos_index <- split(seq_len(totpos), rep(seq_len(length(genoprobs)), npos_by_chr))

    # object to contain the LOD scores; also attr to contain sample size
    result <- matrix(nrow=totpos, ncol=ncol(pheno))
    n <- rep(NA, ncol(pheno)); names(n) <- colnames(pheno)
    if(totpos==0) { # edge case of no genoprobs
        colnames(result) <- colnames(pheno)
        attr(result, "sample_size") <- n
        class(result) <- c("scan1", "matrix")
        return(result)
    }

    if(n_cores(cores)==1) { # no parallel processing
        for(i in run_indexes) {
            chr <- run_batches$chr[i]
            chrnam <- names(genoprobs)[chr]
            phebatch <- phe_batches[[run_batches$phe_batch[i]]]
            phecol <- phebatch$cols

            this_result <- by_group_func(i)
            if(!is.null(this_result)) {
                result[pos_index[[chr]], phecol] <- t(this_result$lod)
                if(chr==1) n[phecol] <- this_result$n
            }
        }
    }
    else {
        # calculations in parallel
        list_result <- cluster_lapply(cores, run_indexes, by_group_func)

        # check for problems (if clusters run out of memory, they'll return NULL)
        result_is_null <- vapply(list_result, is.null, TRUE)
        if(any(result_is_null))
            stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

        # reorganize results
        for(i in run_indexes) {
            chr <- run_batches$chr[i]
            chrnam <- names(genoprobs)[chr]
            phebatch <- phe_batches[[run_batches$phe_batch[i]]]
            phecol <- phebatch$cols

            if(!is.null(list_result[[i]])) {
                result[pos_index[[chr]], phecol] <- t(list_result[[i]]$lod)
                if(chr==1) n[phecol] <- list_result[[i]]$n
            }
        }
    }

    pos_names <- unlist(dimnames(genoprobs)[[3]])
    names(pos_names) <- NULL # this is just annoying
    dimnames(result) <- list(pos_names, colnames(pheno))

    # add some attributes with details on analysis
    attr(result, "sample_size") <- n

    class(result) <- c("scan1", "matrix")
    result
}

# find_peaks_and_bayesint
#
# This is like find_peaks but the results also include Bayes credible intervals.
# The same input as find_peaks, which calls this function when prob is given.
find_peaks_and_bayesint <-
    function(scan1_output, map, threshold=3, peakdrop=Inf, prob=0.95,
             thresholdX=NULL, peakdropX=NULL, probX=NULL,
             expand2markers=TRUE, sort_by=c("column", "pos", "lod"), cores=1)
{
    sort_by <- match.arg(sort_by)

    lodnames <- colnames(scan1_output)
    n_lod <- length(lodnames)

    # make the thresholds have length n_lod
    if(length(threshold)==1) threshold <- rep(threshold, n_lod)
    if(length(threshold) != n_lod)
        stop("threshold should have length 1 or ", n_lod)
    if(is.null(thresholdX)) thresholdX <- threshold
    if(length(thresholdX)==1) thresholdX <- rep(thresholdX, n_lod)
    if(length(thresholdX) != n_lod)
        stop("thresholdX should have length 1 or ", n_lod)

    # make the peakdrops have length n_lod
    if(length(peakdrop)==1) peakdrop <- rep(peakdrop, n_lod)
    if(length(peakdrop) != n_lod)
        stop("peakdrop should have length 1 or ", n_lod)
    if(is.null(peakdropX)) peakdropX <- peakdrop
    if(length(peakdropX)==1) peakdropX <- rep(peakdropX, n_lod)
    if(length(peakdropX) != n_lod)
        stop("peakdropX should have length 1 or ", n_lod)

    # make the probs have length n_lod
    if(length(prob)==1) prob <- rep(prob, n_lod)
    if(length(prob) != n_lod)
        stop("prob should have length 1 or ", n_lod)
    if(is.null(probX)) probX <- prob
    if(length(probX)==1) probX <- rep(probX, n_lod)
    if(length(probX) != n_lod)
        stop("probX should have length 1 or ", n_lod)

    # split lod scores by column and by chromosome
    lod <- as.list(as.data.frame(unclass(scan1_output)))
    chr <- rep(seq_along(map), vapply(map, length, 1))
    lod <- lapply(lod, split, chr)

    # batch info for parallel-processing
    batch <- cbind(rep(seq_along(lod), vapply(lod, length, 1)),
                   unlist(lapply(lod, function(a) seq_along(a))))
    dimnames(batch) <- NULL
    batch_index <- seq_len(nrow(batch))

    # set up parallel analysis
    cores <- setup_cluster(cores)

    # X chr info
    is_x_chr <- attr(map, "is_x_chr")
    if(is.null(is_x_chr))
        is_x_chr <- rep(FALSE, length(map))

    # function to be applied to each batch
    by_batch_func <- function(i) {
        lodcol <- batch[i,1]
        chr <- batch[i,2]

        this_lod <- lod[[lodcol]][[chr]]

        peaks <- .find_peaks_and_bayesint(this_lod, map[[chr]],
                                          ifelse(is_x_chr[chr], thresholdX[lodcol], threshold[lodcol]),
                                          ifelse(is_x_chr[chr], peakdropX[lodcol], peakdrop[lodcol]),
                                          ifelse(is_x_chr[chr], probX[lodcol], prob[lodcol]))

        n_peaks <- length(peaks)
        if(n_peaks==0) return(NULL)

        # calculate peak positions
        # (this deals with ties in the LOD scores:
        #  take average of multiple positions sharing the maximum LOD score)
        # and remember that .find_peaks returns indexes starting at 0
        ci_lo <- ci_hi <- rep(0, n_peaks)
        for(i in seq_len(n_peaks)) {
            if(expand2markers)
                ci <- expand_interval_to_markers(peaks[[i]][1:2]+1, map[[chr]])
            else
                ci <- map[[chr]][peaks[[i]][1:2]+1]
            ci_lo[i] <- ci[1]
            ci_hi[i] <- ci[2]
        }
        peak_pos <- vapply(peaks, function(a) mean(map[[chr]][a[-(1:2)]+1]), 0.0)
        peak_lod <- vapply(peaks, function(a) this_lod[a[3]+1], 0.0)

        data.frame(lodindex=rep(lodcol,n_peaks),
                   lodcolumn=rep(lodnames[lodcol],n_peaks),
                   chr=rep(names(map)[chr],n_peaks),
                   pos=peak_pos,
                   lod=peak_lod,
                   ci_lo=ci_lo,
                   ci_hi=ci_hi,
                   row.names=seq_along(peaks),
                   stringsAsFactors=FALSE)
    }

    if(n_cores(cores)==1) {
        peaks <- lapply(batch_index, by_batch_func)
    } else {
        peaks <- cluster_lapply(cores, batch_index, by_batch_func)
    }

    # empty data frame to start
    result <- data.frame(lodindex=numeric(0),
                         lodcolumn=character(0),
                         chr=character(0),
                         pos=numeric(0),
                         lod=numeric(0),
                         ci_lo=numeric(0),
                         ci_hi=numeric(0),
                         row.names=character(0),
                         stringsAsFactors=FALSE)

    for(p in peaks) result <- rbind(result, p)

    rownames(result) <- NULL
    result$chr <- factor(result$chr, names(map))

    if(nrow(result) > 1) {
        if(sort_by == "pos") {

            result <- result[order(result$chr, result$pos, result$lodindex), ]

        } else if(sort_by == "lod") {

            result <- result[order(result$lod, decreasing=TRUE),]

        }
    }

    result
}

# Genome scan with a single-QTL and linear mixed model
#
# called by scan1max() when kinship() is provided.
scan1max_pg <-
    function(genoprobs, pheno, kinship, addcovar=NULL, Xcovar=NULL,
             intcovar=NULL, weights=NULL, reml=TRUE, hsq=NULL,
             by_chr=FALSE, cores=1, ...)
{
    # deal with the dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    if(!is_pos_number(tol)) stop("tol should be a single positive number")
    intcovar_method <- grab_dots(dotargs, "intcovar_method", "lowmem",
                                 c("highmem", "lowmem"))
    quiet <- grab_dots(dotargs, "quiet", TRUE)
    max_batch <- grab_dots(dotargs, "max_batch", NULL)
    if(!is.null(max_batch) && !is_pos_number(max_batch)) stop("max_batch should be a single positive integer")
    check_boundary <- grab_dots(dotargs, "check_boundary", TRUE)
    check_extra_dots(dotargs, c("tol", "intcovar_method", "check_boundary", "quiet", "max_batch"))

    # check that the objects have rownames
    check4names(pheno, addcovar, Xcovar, intcovar)

    # force things to be matrices
    if(!is.matrix(pheno)) {
        pheno <- as.matrix(pheno)
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }
    if(is.null(colnames(pheno))) # force column names
        colnames(pheno) <- paste0("pheno", seq_len(ncol(pheno)))
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(Xcovar)) {
        if(!is.matrix(Xcovar)) Xcovar <- as.matrix(Xcovar)
        if(!is.numeric(Xcovar)) stop("Xcovar is not numeric")
    }
    if(!is.null(intcovar)) {
        if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)
        if(!is.numeric(intcovar)) stop("intcovar is not numeric")
    }

    # check that kinship matrices are square with same IDs
    kinshipIDs <- check_kinship(kinship, length(genoprobs))

    # multiply kinship matrix by 2; rest is using 2*kinship
    # see Almasy & Blangero (1998) https://doi.org/10.1086/301844
    kinship <- double_kinship(kinship)

    # take square-root of weights
    weights <- sqrt_weights(weights)

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, addcovar, Xcovar, intcovar,
                               kinshipIDs, weights, complete.cases=TRUE)
    ind2keep <- get_common_ids(ind2keep, pheno[rowSums(is.finite(pheno)) > 0,,drop=FALSE])
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # make sure columns in intcovar are also in addcovar
    addcovar <- force_intcovar(addcovar, intcovar, tol)

    # drop things from Xcovar that are already in addcovar
    Xcovar <- drop_xcovar(addcovar, Xcovar, tol)

    # batch phenotypes by missing values
    phe_batches <- batch_cols(pheno[ind2keep,,drop=FALSE], max_batch)

    # drop cols in genotype probs that are all 0 (just looking at the X chromosome)
    genoprob_Xcol2drop <- genoprobs_col2drop(genoprobs)
    is_x_chr <- attr(genoprobs, "is_x_chr")
    if(is.null(is_x_chr)) is_x_chr <- rep(FALSE, length(genoprobs))

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    n_chr <- length(genoprobs)

    # to contain the results
    result <- matrix(nrow=n_chr, ncol=ncol(pheno))
    dimnames(result) <- list(names(genoprobs), colnames(pheno))

    # number of chr to consider under null
    if(is_kinship_list(kinship)) n_null_chr <- length(kinship)
    else if(!is.null(Xcovar)) n_null_chr <- any(is_x_chr) + any(!is_x_chr)
    else n_null_chr <- 1

    if(is.null(hsq)) {
        hsq <- matrix(nrow=n_null_chr, ncol=ncol(pheno))
        dimnames(hsq) <- hsq_dimnames(kinship, Xcovar, is_x_chr, pheno)
        estimate_hsq <- TRUE
    } else {
        if(length(hsq) != n_null_chr * ncol(pheno)) {
            stop("hsq should be NULL or a matrix of size ", n_null_chr, " x ", ncol(pheno))
        }
        if(!is.matrix(hsq)) hsq <- as.matrix(hsq, ncol=ncol(pheno))
        if(nrow(hsq) != n_null_chr) {
            stop("hsq should be NULL or a matrix of size ", n_null_chr, " x ", ncol(pheno))
        }
        if(any(hsq < 0 | hsq > 1)) {
            stop("hsq values should be between 0 and 1.")
        }
        estimate_hsq <- FALSE
    }
    n <- rep(NA, ncol(pheno)); names(n) <- colnames(pheno)

    chr_index <- rep(seq_len(n_chr), sapply(genoprobs, function(a) dim(a)[3]))

    # loop over batches of phenotypes with the same pattern of NAs
    for(batch in seq_along(phe_batches)) {

        # info about batch
        omit <- phe_batches[[batch]]$omit # ind to omit
        phecol <- phe_batches[[batch]]$cols # phenotype columns in batch

        # individuals to keep in this batch
        these2keep <- ind2keep
        if(length(omit)>0) these2keep <- ind2keep[-omit]
        n[phecol] <- length(these2keep)
        if(length(these2keep) <= 2) next # not enough individuals; skip this batch

        # subset the rest
        K <- subset_kinship(kinship, ind=these2keep)
        ac <- addcovar; if(!is.null(ac)) { ac <- ac[these2keep,,drop=FALSE]; ac <- drop_depcols(ac, TRUE, tol) }
        Xc <- Xcovar;   if(!is.null(Xc)) Xc <- Xc[these2keep,,drop=FALSE]
        ic <- intcovar; if(!is.null(ic)) { ic <- ic[these2keep,,drop=FALSE]; ic <- drop_depcols(ic, TRUE, tol) }
        wts <- weights; if(!is.null(wts)) wts <- wts[these2keep]
        ph <- pheno[these2keep,phecol,drop=FALSE]

        # multiply stuff by the weights
        K <- weight_kinship(K, wts)
        ac <- weight_matrix(ac, wts)
        Xc <- weight_matrix(Xc, wts)
        ph <- weight_matrix(ph, wts)

        # eigen decomposition of kinship matrix
        Ke <- decomp_kinship(K, cores=cores)

        # fit LMM for each phenotype, one at a time
        if(estimate_hsq) {
            nullresult <- calc_hsq_clean(Ke=Ke, pheno=ph, addcovar=ac, Xcovar=Xc,
                                         is_x_chr=is_x_chr, weights=wts, reml=reml,
                                         cores=cores, check_boundary=check_boundary, tol=tol)

            hsq[, phecol] <- nullresult$hsq
        }
        else {
            # for the log likelihood, calculate the reml=FALSE version
            loglik <- calc_nullLL_clean(hsq=hsq[,phecol,drop=FALSE], Ke=Ke, pheno=ph,
                                        addcovar=ac, Xcovar=Xc,
                                        is_x_chr=is_x_chr, weights=wts, reml=FALSE,
                                        cores=cores)
            nullresult <- list(hsq=hsq[,phecol,drop=FALSE],
                               loglik=loglik)
        }

        # weighted least squares genome scan, using cluster_lapply across chromosomes
        lod <- scan1_pg_clean(genoprobs, these2keep, Ke, ph, ac, ic, is_x_chr,
                              wts, genoprob_Xcol2drop,
                              nullresult$hsq, nullresult$loglik, reml, cores,
                              intcovar_method, tol)

        result[,phecol] <- apply(lod, 2, tapply, chr_index, max, na.rm=TRUE)
    }

    if(!by_chr) result <- apply(result, 2, max, na.rm=TRUE)

    # add attributes
    attr(result, "hsq") <- hsq
    attr(result, "sample_size") <- n

    result
}

#' Calculate QTL hotspots
#'
#' For a set of QTL locations, calculate a running count in a sliding
#' window across the genome.
#'
#' @param peaks Data frame of QTL results, as output by [find_peaks()]
#'     Should contain columns `chr` and `pos`.
#'
#' @param map Marker map, as a list of chromosomes, each being a vector of
#'     positions. Hotspot counts will be calculated at these
#'     positions.
#'
#' @param window Window size for counting QTL.
#'
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @param quiet If TRUE, don't print any messages.
#'
#' @return An object of class `"scan1"`: a matrix with a single
#'     column, of counts, with rownames being the marker names in
#'     `map`. The column name is `"num_qtl"`.
#'
#'
#' @seealso [find_peaks()], [plot_lodpeaks()], [plot_cistrans()]
#'
#' @export
#'
#' @examples
#' \dontrun{
#' # download example pQTL results (from Keele et al. 2026, https://doi.org/10.1016/j.xgen.2025.101069)
#' # contains qtl, map_endpoints, and pheno_pos
#' url <- "https://kbroman.org/qtl2/assets/sampledata/pqtl_data.RData"
#' tempfile <- file.path(tempdir(), basename(url))
#' download.file(url, tempfile)
#' load(tempfile)
#' unlink(tempfile)
#'
#' hotspots <- calc_hotspots(qtl, map, window=2)
#' plot(hotspots, map, ylab="No. QTL")
#' find_peaks(hotspots, map, threshold=20)
#' }

calc_hotspots <-
    function(peaks, map, window=1, cores=1, quiet=TRUE)
{
    if(!is.data.frame(peaks) || !all(c("chr", "pos") %in% names(peaks))) {
        stop("peaks should be a data frame with columns chr and pos")
    }
    stopifnot(is.list(map))

    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    bychr_func <-
        function(chr)
    {
        result <- .running_count(peaks$pos[peaks$chr==chr], map[[chr]], window)

        names(result) <- names(map[[chr]])
        result
    }

    result <- as.matrix(unlist(cluster_lapply(cores, names(map), bychr_func)))

    class(result) <- c("scan1", "matrix")
    colnames(result) <- "num_qtl"
    result
}

# calc_grid

#' Calculate indicators of which marker/pseudomarker positions are along a fixed grid
#'
#' Construct vectors of logical indicators that indicate which
#' positions correspond to locations along a grid
#'
#' @param map A list of numeric vectors; each vector gives marker
#' positions for a single chromosome.
#' @param step Distance between pseudomarkers and markers; if
#' `step=0` no pseudomarkers are inserted.
#' @param off_end Distance beyond terminal markers in which to insert
#' pseudomarkers.
#' @param tol Tolerance for determining whether a pseudomarker would
#' duplicate a marker position.
#'
#' @return A list of logical (TRUE/FALSE) vectors that indicate, for a
#'     marker/pseudomarker map created by
#'     [insert_pseudomarkers()] with `step`>0 and
#'     `stepwidth="fixed"`, which positions correspond to he
#'     locations along the fixed grid.
#'
#' @details The function [insert_pseudomarkers()], with
#'     `stepwidth="fixed"`, will insert a grid of pseudomarkers,
#'     to a marker map. The present function gives a series of
#'     TRUE/FALSE vectors that indicate which positions fall on the
#'     grid. This is for use with [probs_to_grid()], for
#'     reducing genotype probabilities, calculated with
#'     [calc_genoprob()], to just the positions on the grid.
#'     The main value of this is to speed up genome scan computations
#'     in the case of very dense markers, by focusing on just a grid
#'     of positions rather than on all marker locations.
#'
#' @export
#' @keywords utilities
#' @seealso [insert_pseudomarkers()], [probs_to_grid()],
#'     [map_to_grid()]
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' gmap_w_pmar <- insert_pseudomarkers(iron$gmap, step=1)
#' grid <- calc_grid(iron$gmap, step=1)
calc_grid <-
function(map, step=0, off_end=0, tol=0.01)
{
    # is input not a map but a cross2 object?
    cl <- class(map)
    if(length(cl)==1 && cl=="cross2") {
        map <- map$gmap
        if(is.null(map))
            stop("Input is a cross2 object, but no genetic map found.")
    }
    if(!is_nonneg_number(step)) stop("step should be a single non-negative number")
    if(!is_nonneg_number(off_end)) stop("off_end should be a single non-negative number")
    if(!is_nonneg_number(tol)) stop("tol should be a single non-negative number")

    grid <- map
    chr <- names(grid) <- names(map)
    for(i in seq(along=map)) {
        grid[[i]] <- calc_grid_onechr(map[[i]], step, off_end, tol,
                                      paste0("c", chr[i], ".loc"))
    }

    grid
}



# grid for pseudomarkers
#
# map = vector of marker positions; names = marker names
# step = distance between pseudomarkers
# off_end = amount to go beyond the terminal markers
# tol = tolerance for determining whether a marker hits a pseudomarker
# pmar_stem = leading part of name of pseudomarkers
calc_grid_onechr <-
function(map, step=0, off_end=0, tol=0.01, pmar_stem="loc")
{
    if(!is_nonneg_number(step)) stop("step should be a single non-negative number")
    if(!is_nonneg_number(off_end)) stop("off_end should be a single non-negative number")
    if(!is_nonneg_number(tol)) stop("tol should be a single non-negative number")

    if(step==0) {
        grid <- rep(TRUE, length(map))
        names(grid) <- names(map)
        return(grid)
    }

    if(any(is.na(map))) stop("map values can't be missing")

    if(tol < 0) stop("tol should be >= 0")

    # locations of pseudomarkers
    pmar_map <- seq(min(map)-off_end, max(map)+off_end, by=step)

    # for each pseudomarker, distance to nearest marker
    d <- abs(outer(map, pmar_map, "-"))
    mind <- apply(d, 2, min)

    # omit pseudomarkers that hit a marker
    to_omit <- (mind < tol)
    pmar_map <- pmar_map[!to_omit]
    if(length(pmar_map) == 0) {
        grid <- rep(TRUE, length(map))
        names(grid) <- names(map)
        return(grid)
    }

    # markers that are on pseudomarker grid
    grid <- rep(FALSE, length(map))
    names(grid) <- names(map)
    if(any(to_omit))
        grid[apply(d[,to_omit,drop=FALSE], 2, which.min)] <- TRUE

    if(is.null(names(pmar_map)))
        names(pmar_map) <- create_pseudomarker_names(pmar_map, step, pmar_stem)

    # logical vector indicating grid
    pmar_grid <- rep(TRUE, length(pmar_map))
    names(pmar_grid) <- names(pmar_map)
    grid <- c(grid, pmar_grid)

    # add pseudomarkers to markers
    map <- c(map, pmar_map)

    # sort the map
    o <- order(map)

    grid[o]
}

context("Reduce map to pseudomarker grid")

test_that("map_to_grid works", {

    # try it out
    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    grid <- calc_grid(grav2$gmap, step=1)

    reduced_map <- map
    for(i in seq(along=map)) {
        reduced_map[[i]] <- map[[i]][grid[[i]]]
    }

    expect_equal(reduced_map, map_to_grid(map, grid))
})

context("calculation of grid positions")
suppressMessages(library(qtl))

test_that("grid-based version works in simple case", {

    # equally-spaced map
    map <- seq(0, 50, by=2.5)
    names(map) <- paste0("map", seq(along=map))

    # step = marker distance
    pmap <- insert_pseudomarkers(list("1"=map), step=2.5, off_end=0, stepwidth="fixed")
    grid <- calc_grid(list("1"=map), step=2.5, off_end=0)
    expect_equal(length(pmap[[1]]), length(grid[[1]]))
    expect_equal(names(pmap[[1]]), names(grid[[1]]))

    expected_grid <- rep(TRUE, length(map))
    names(expected_grid) <- names(map)
    expect_equal(grid, list("1"=expected_grid))

    # step = 1
    pmap <- insert_pseudomarkers(list("1"=map), step=1, off_end=0, stepwidth="fixed")
    grid <- calc_grid(list("1"=map), step=1, off_end=0)

    # expected grid
    expected_grid <- !is.na(match(pmap[[1]], seq(0, 50, by=1)))
    names(expected_grid) <- names(pmap[[1]])
    expect_equal(grid, list("1"=expected_grid))

})

test_that("minimal version works in simple case", {

    # equally-spaced map
    map <- seq(0, 50, by=2.5)
    names(map) <- paste0("map", seq(along=map))

    # step = marker distance
    grid <- calc_grid(list("1"=map), step=2.5, off_end=0)
    expected_grid <- rep(TRUE, length(map))
    names(expected_grid) <- names(map)
    expect_equal(grid, list("1"=expected_grid))

})



test_that("insert_pseudomarkers gives distinct pseudomarker names with iron data", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    pmap <- insert_pseudomarkers(iron$gmap, step=1)
    grid <- calc_grid(iron$gmap, step=1)

    expect_equal(length(pmap), length(grid))
    expect_equal(names(pmap), names(grid))

    for(i in seq(along=pmap)) {
        expect_equal(length(pmap[[i]]), length(grid[[i]]))
        expect_equal(names(pmap[[i]]), names(grid[[i]]))
    }

})

# scan1 permutations by LMM (with a kinship matrix)
scan1perm_pg <-
    function(genoprobs, pheno, kinship, addcovar=NULL, Xcovar=NULL, intcovar=NULL,
             weights=NULL, reml=TRUE, n_perm=1, perm_strata=NULL, cores=1, ind2keep, ...)
{
    # deal with the dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    stopifnot(tol > 0)
    intcovar_method <- grab_dots(dotargs, "intcovar_method", "lowmem",
                                 c("highmem", "lowmem"))
    quiet <- grab_dots(dotargs, "quiet", TRUE)
    check_boundary <- grab_dots(dotargs, "check_boundary", TRUE)
    check_extra_dots(dotargs, c("tol", "intcovar_method", "quiet", "max_batch",
                                "check_boundary"))

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # max batch size
    max_batch <- grab_dots(dotargs, "max_batch",
                           min(1000, ceiling(n_perm*length(genoprobs)*ncol(pheno)/n_cores(cores))))

    # generate permutations
    perms <- gen_strat_perm(n_perm, ind2keep, perm_strata)

    # batch permutations
    phe_batches <- batch_cols(pheno[ind2keep,,drop=FALSE], max_batch)

    # drop cols in genotype probs that are all 0 (just looking at the X chromosome)
    genoprob_Xcol2drop <- genoprobs_col2drop(genoprobs)
    is_x_chr <- attr(genoprobs, "is_x_chr")
    if(is.null(is_x_chr)) is_x_chr <- rep(FALSE, length(genoprobs))

    ## decompose kinship matrices (multiplied by weights)
    # look for the unique groups of individuals omitted
    #   (we subset and then decompose the kinship matrix separately for each)
    index_batches <- index_batches_by_omits(phe_batches)
    # unique indexes
    uindex_batches <- unique(index_batches)
    decomp_func <- function(i) {
        these2keep <- ind2keep # individuals 2 keep for this batch
        omit <- phe_batches[[uindex_batches[i]]]$omit
        if(length(omit) > 0) these2keep <- ind2keep[-omit]

        k <- subset_kinship(kinship, ind=these2keep)
        wts <- weights[these2keep]
        k <- weight_kinship(k, wts)

        decomp_kinship(k, cores=1)
    }
    kinship_list <- cluster_lapply(cores, seq_along(uindex_batches), decomp_func)

    ## null results
    null_by_batch_func <- function(i) {
        omit <- phe_batches[[i]]$omit
        phecol <- phe_batches[[i]]$cols
        these2keep <- ind2keep # individuals 2 keep for this batch
        if(length(omit) > 0) these2keep <- ind2keep[-omit]
        if(length(these2keep)<=2) return(NULL) # not enough individuals

        ac <- addcovar; if(!is.null(ac)) ac <- ac[these2keep,,drop=FALSE]
        Xc <- Xcovar;   if(!is.null(Xc)) Xc <- Xc[these2keep,,drop=FALSE]
        wts <- weights; if(!is.null(wts)) wts <- wts[these2keep]
        ph <- pheno[these2keep, phecol, drop=FALSE]

        k <- kinship_list[[ index_batches[i] ]]

        # multiply stuff by weights (kinship is already done)
        ac <- weight_matrix(ac, wts)
        Xc <- weight_matrix(Xc, wts)
        ph <- weight_matrix(ph, wts)

        result <- calc_hsq_clean(Ke=k, pheno=ph, addcovar=ac, Xcovar=Xc,
                                 is_x_chr=is_x_chr, reml=reml, weights=wts,
                                 cores=1, check_boundary=check_boundary, tol=tol)

        # expand hsq and loglik to one row per chromosome
        n_chr <- length(is_x_chr)
        if(nrow(result$hsq) != n_chr) {
            if(nrow(result$hsq)==1) {
                for(i in seq_along(result))
                    result[[i]] <- matrix(rep(result[[i]], n_chr), byrow=TRUE, nrow=n_chr)
            }
            else if(nrow(result$hsq)==2) {
                for(i in seq_along(result)) {
                    tmp <- result[[i]]
                    result[[i]] <- matrix(nrow=n_chr, ncol=ncol(tmp))
                    for(j in seq_along(is_x_chr)) {
                        if(is_x_chr[j])
                            result[[i]][j,] <- tmp[2,]
                        else
                            result[[i]][j,] <- tmp[1,]
                    }
                }
            }
        }

        result
    }
    nullresult <- cluster_lapply(cores, seq_along(phe_batches), null_by_batch_func)

    # batches for analysis, to allow parallel analysis
    run_batches <- data.frame(chr=rep(seq_len(length(genoprobs)), length(phe_batches)*n_perm),
                              phe_batch=rep(seq_along(phe_batches), each=length(genoprobs)*n_perm),
                              perm_batch=rep(rep(seq_len(n_perm), each=length(genoprobs), length(phe_batches))))

    run_indexes <- seq_len(length(genoprobs)*length(phe_batches)*n_perm)

    # the function that does the work
    by_group_func <- function(i) {
        # deal with batch information, including individuals to drop due to missing phenotypes
        chr <- run_batches$chr[i]
        chrnam <- names(genoprobs)[chr]
        phebatchnum <- run_batches$phe_batch[i]
        phebatch <- phe_batches[[phebatchnum]]
        permbatch <- run_batches$perm_batch[i]
        phecol <- phebatch$cols
        omit <- phebatch$omit
        these2keep <- ind2keep # individuals 2 keep for this batch
        if(length(omit) > 0) these2keep <- ind2keep[-omit]
        if(length(these2keep)<=2) return(NULL) # not enough individuals

        # apply permutation to the probs
        #     (I initially thought we'd need to apply these just to the rownames, but
        #     actually we've already aligned probs and pheno and so forth
        #     via ind2keep/these2keep, so now we need to just permute the rows)
        #
        # (also need some contortions here to keep the sizes the same)
        pr <- genoprobs[[chr]][ind2keep,,,drop=FALSE]
        pr <- pr[perms[,permbatch],,,drop=FALSE]
        rownames(pr) <- ind2keep
        pr <- pr[these2keep,,,drop=FALSE]

        # subset the genotype probabilities: drop cols with all 0s, plus the first column
        Xcol2drop <- genoprob_Xcol2drop[[chrnam]]
        if(length(Xcol2drop) > 0) {
            pr <- pr[,-Xcol2drop,,drop=FALSE]
        }
        pr <- pr[,-1,,drop=FALSE]

        # subset the rest
        ac <- addcovar; if(!is.null(ac)) ac <- ac[these2keep,,drop=FALSE]
        ic <- intcovar; if(!is.null(ic)) ic <- ic[these2keep,,drop=FALSE]
        wts <- weights; if(!is.null(wts)) wts <- wts[these2keep]
        ph <- pheno[these2keep, phecol, drop=FALSE]

        # grab decomposed kinship matrix for this chromosome
        Ke <- subset_kinship(kinship_list[[index_batches[phebatchnum]]], chr=chrnam)

        # multiply stuff by weights
        ac <- weight_matrix(ac, wts)
        ic <- weight_matrix(ic, wts)
        ph <- weight_matrix(ph, wts)

        # weight the probabilities
        pr <- weight_array(pr, wts)

        # hsq, null_loglik for this batch and chr
        hsq <- nullresult[[phebatchnum]]$hsq[chr,]
        null_loglik <- nullresult[[phebatchnum]]$loglik[chr,]

        # scan this chromosome and calculate maximum LOD for each phenotype
        scan1perm_pg_onechr(pr, Ke, ph, ac, ic, wts, hsq, null_loglik, reml,
                            intcovar_method, tol)

    }

    # calculations in parallel
    list_result <- cluster_lapply(cores, run_indexes, by_group_func)

    # check for problems (if clusters run out of memory, they'll return NULL)
    result_is_null <- vapply(list_result, is.null, TRUE)
    if(any(result_is_null))
        stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

    # reorganize results
    result <- array(dim=c(length(genoprobs), n_perm, ncol(pheno)))

    for(i in run_indexes) {
        chr <- run_batches$chr[i]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        permbatch <- run_batches$perm_batch[i]

        result[chr, permbatch, phebatch$cols] <- list_result[[i]]
    }

    result <- apply(result, c(2,3), max)
    colnames(result) <- colnames(pheno)

    class(result) <- c("scan1perm", "matrix")
    result
}


# identify groups of equivalent batches, by individuals omitted
index_batches_by_omits <-
    function(phe_batches)
{
    omits <- vapply(phe_batches, function(a) paste(a$omit, collapse="|"), "")

    match(omits, omits)
}


# perform an LMM scan for a single chromosome, and returns the maximum LOD for each phenotype
# genoprobs is an array for a single chromosome
# Ke is the decomposed kinship matrix
scan1perm_pg_onechr <-
    function(genoprobs, Ke, pheno, addcovar, intcovar, weights,
             hsq, null_loglik, reml, intcovar_method, tol)
{
    Kevec <- Ke$vectors
    Keval <- Ke$values

    maxlod <- rep(NA, ncol(pheno))

    intercept <- weights; if(is_null_weights(weights)) intercept <- rep(1,nrow(pheno))
    ac <- cbind(intercept, addcovar)
    ic <- intcovar

    for(phecol in seq_len(ncol(pheno))) {
        y <- pheno[,phecol,drop=FALSE]

        w <- 1/(hsq[phecol]*Keval + (1-hsq[phecol]))
        w <- sqrt(w)

        if(is.null(ic))
            loglik <- scan_pg_onechr(genoprobs, y, ac, Kevec, w, tol)
        else if(intcovar_method=="highmem")
            loglik <- scan_pg_onechr_intcovar_highmem(genoprobs, y, ac, ic, Kevec, w, tol)
        else
            loglik <- scan_pg_onechr_intcovar_lowmem(genoprobs, y, ac, ic, Kevec, w, tol)

        maxlod[phecol] <- (max(loglik) - null_loglik[phecol])/log(10)
    }
    names(maxlod) <- colnames(pheno)

    maxlod
}

# genome scan at SNPs (or scan of defined region)
#' Single-QTL genome scan at imputed SNPs
#'
#' Perform a single-QTL scan across the genome or a defined region at
#' SNPs genotyped in the founders, by Haley-Knott regression or a
#' liear mixed model, with possible allowance for covariates.
#'
#' @param genoprobs Genotype probabilities as calculated by
#' [calc_genoprob()].
#' @param map Physical map for the positions in the `genoprobs`
#' object: A list of numeric vectors; each vector gives marker
#' positions for a single chromosome.
#' @param pheno A numeric matrix of phenotypes, individuals x phenotypes.
#' @param kinship Optional kinship matrix, or a list of kinship matrices (one
#' per chromosome), in order to use the LOCO (leave one chromosome
#' out) method.
#' @param addcovar An optional numeric matrix of additive covariates.
#' @param Xcovar An optional numeric matrix with additional additive covariates used for
#' null hypothesis when scanning the X chromosome.
#' @param intcovar An optional numeric matrix of interactive covariates.
#' @param weights An optional numeric vector of positive weights for the
#' individuals. As with the other inputs, it must have `names`
#' for individual identifiers.
#' @param reml If `kinship` provided: if `reml=TRUE`, use
#' REML; otherwise maximum likelihood.
#' @param model Indicates whether to use a normal model (least
#' squares) or binary model (logistic regression) for the phenotype.
#' If `model="binary"`, the phenotypes must have values in \eqn{[0, 1]}.
#' @param query_func Function for querying SNP information; see
#' [create_variant_query_func()]). Takes arguments
#' `chr`, `start`, `end`, (with `start` and `end` in the units in
#' `map`, generally Mbp), and returns a data frame containing
#' the columns `snp`, `chr`, `pos`, and `sdp`. (See `snpinfo` below.)
#' @param chr Chromosome or chromosomes to scan
#' @param start Position defining the start of an interval to scan.
#' Should be a single number, and if provided, `chr` should also
#' have length 1.
#' @param end Position defining the end of an interval to scan.
#' Should be a single number, and if provided, `chr` should also
#' have length 1.
#' @param snpinfo Optional data frame of SNPs to scan; if provided,
#' `query_func`, `chr`, `start`, and `end` are ignored. Should
#' contain the following columns:
#' * `chr` - Character string or factor with chromosome
#' * `pos` - Position (in same units as in the `"map"`).
#' * `sdp` - Strain distribution pattern: an integer, between
#'     1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
#'     binary encoding indicates the founder genotypes
#' * `snp` - Character string with SNP identifier (if
#'     missing, the rownames are used).
#' @param batch_length Interval length (in units of `map`, generally
#' Mbp) to scan at one time.
#' @param keep_all_snps SNPs are grouped into equivalence classes based
#'     on position and founder genotypes; if `keep_all_snps=FALSE`,
#'     the return value will contain information only on the indexed
#'     SNPs (one per equivalence class).
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#' @param ... Additional control parameters passed to [scan1()]
#'
#' @return A list with two components: `lod` (matrix of LOD scores)
#' and `snpinfo` (a data frame of SNPs that were scanned,
#' including columns `index` which indicates groups of equivalent
#' SNPs)
#'
#' @details
#' The analysis proceeds as follows:
#' * Call `query_func()` to grab all SNPs over a region.
#' * Use [index_snps()] to group SNPs into equivalence classes.
#' * Use [genoprob_to_snpprob()] to convert `genoprobs` to SNP probabilities.
#' * Use [scan1()] to do a single-QTL scan at the SNPs.
#'
#' @seealso [scan1()], [genoprob_to_snpprob()], [index_snps()], [create_variant_query_func()], [plot_snpasso()]
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' probs <- calc_genoprob(DOex, error_prob=0.002)
#'
#' snpdb_file <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
#' queryf <- create_variant_query_func(snpdb_file)
#'
#' out <- scan1snps(probs, DOex$pmap, DOex$pheno, query_func=queryf, chr=2, start=97, end=98)
#' }
#'
#' @export
#'
scan1snps <-
    function(genoprobs, map, pheno, kinship=NULL, addcovar=NULL, Xcovar=NULL,
             intcovar=NULL, weights=NULL, reml=TRUE, model=c("normal", "binary"),
             query_func=NULL, chr=NULL, start=NULL, end=NULL, snpinfo=NULL,
             batch_length=20, keep_all_snps=FALSE, cores=1, ...)
{
    model <- match.arg(model)

    if(is.null(genoprobs)) stop("genoprobs is NULL")
    if(is.null(map)) stop("map is NULL")
    if(is.null(pheno)) stop("pheno is NULL")

    # reduce to common chromosomes
    pchr <- names(genoprobs)
    mchr <- names(map)
    cchr <- pchr[pchr %in% mchr]
    if(length(cchr)==0) {
        warning("No common chromosomes among genoprobs and map")
        return(NULL)
    }
    genoprobs <- genoprobs[,cchr]
    map <- map[cchr]

    # check genoprobs and map have same markers in same order
    #  - subset to common markers if different
    #  - stop with error if different order
    subset_markers <- FALSE
    for(ichr in names(genoprobs)) {
        markers_genoprobs <- dimnames(genoprobs[[ichr]])[[3]]
        markers_map <- names(map[[ichr]])

        if(length(markers_genoprobs) != length(markers_map) ||
           any(markers_genoprobs != markers_map)) {
            subset_markers <- TRUE # later give warning
            markers <- markers_genoprobs[markers_genoprobs %in% markers_map]
            if(length(markers)==0) stop("No markers in common between genoprobs and map on chr ", ichr)
            genoprobs[[ichr]] <- genoprobs[[ichr]][,,markers_genoprobs %in% markers,drop=FALSE]
            map[[ichr]] <- map[[ichr]][markers_map %in% markers]

            markers_genoprobs <- dimnames(genoprobs[[ichr]])[[3]]
            markers_map <- names(map[[ichr]])
            if(any(markers_genoprobs != markers_map))
                stop("Markers in different order between genoprobs and map on chr ", ichr)
        }
    }
    if(subset_markers) warning("Subset to common markers between genoprobs and map")

    if(!is.null(snpinfo) && !is.null(query_func)) {
        warning("If snpinfo is provided, chr, start, end, and query_func are all ignored")
    }

    if(!is.null(chr)) chr <- unique(as.character(chr)) # make sure chr is character strings

    if(!is.null(chr) || !is.null(start) || !is.null(end)) { # defined region
        if(!is.null(snpinfo)) warning("If snpinfo provided, chr, start, end, and query_func are all ignored")
        if(is.null(chr)) stop("If start and/or end provided, need to give chr as well")
        if(is.null(query_func)) stop("If chr, start, and/or end provided, need to give query_func too")

        if(length(chr) == 1) {
            if(is.null(start)) start <- -1e15
            if(is.null(end)) end <- 1e15

            if(length(start) > 1 || length(end) > 1) {
                warning("start and end should have length 1; using the first values")
                start <- start[1]
                end <- end[1]
                if(start >= end) stop("Need start < end")
            }

            snpinfo <- query_func(chr, start, end)
        }
    }

    if(length(chr) > 1) {
        if(!is.null(start) || !is.null(end)) {
            warning("If length(chr) > 1, start end end are ignored.")
        }

        if(!all(chr %in% names(map))) {
            stop("Not all chromosomes found: ", paste(chr[!(chr %in% names(map))], collapse=", "))
        }
        map <- map[chr]
        genoprobs <- genoprobs[,chr]
    }

    if(!is.null(snpinfo)) {
        return( scan1snps_snpinfo(genoprobs=genoprobs, map=map, pheno=pheno, kinship=kinship,
                                  addcovar=addcovar, Xcovar=Xcovar, intcovar=intcovar,
                                  weights=weights, reml=reml, model=model,
                                  snpinfo=snpinfo, keep_all_snps=keep_all_snps, ...) )
    }

    if(is.null(query_func)) {
        stop("Need to provide either snpinfo or query_func()")
    }

    # split into batches
    if(!is_pos_number(batch_length)) stop("batch_length should be a single positive integer")
    chr <- names(map)
    chr_start <- vapply(map, min, 0)
    chr_end <- vapply(map, max, 0)
    n_batch <- ceiling((chr_end - chr_start)/batch_length)
    endpts <- lapply(seq_along(chr), function(i) seq(chr_start[i], chr_end[i], length.out=n_batch[i]+1))
    batches <- data.frame(chr=rep(chr, n_batch),
                          start=unlist(lapply(endpts, function(a) a[-length(a)])),
                          end=unlist(lapply(endpts, function(a) a[-1])),
                          stringsAsFactors=FALSE)

    # function that does the work
    by_batch_func <- function(i)
    {
        snpinfo <- query_func(batches$chr[i], batches$start[i], batches$end[i])

        if(nrow(snpinfo) == 0) return( list(lod=NULL, snpinfo=snpinfo) )

        # trim off end if necessary
        if(i < nrow(batches) && batches$chr[i] == batches$chr[i+1] &&
           batches$end[i] == batches$start[i+1])
            snpinfo <- snpinfo[snpinfo$pos < batches$end[i], , drop=FALSE]

        scan1snps_snpinfo(genoprobs=genoprobs, map=map, pheno=pheno,
                          kinship=subset_kinship(kinship, chr=batches$chr[i]),
                          addcovar=addcovar, Xcovar=Xcovar, intcovar=intcovar,
                          weights=weights, reml=reml, model=model, snpinfo=snpinfo,
                          keep_all_snps=keep_all_snps, ...)
    }

    # set up parallel analysis
    cores <- setup_cluster(cores)

    # the real work
    result <- cluster_lapply(cores, seq_len(nrow(batches)), by_batch_func)

    # combine the multiple results
    omit <- vapply(result, function(a) is.null(a$lod) || is.null(a$snpinfo) || nrow(a$snpinfo)==0, TRUE)
    result <- result[!omit]

    # one useful batch?
    if(length(result)==0) return(NULL)
    if(length(result)==1) return(result[[1]])

    ### lod scores are easy
    lod <- do.call("rbind", lapply(result, "[[", "lod"))

    ### snpinfo: need to revise the index column
    offset <- 0
    for(i in seq_along(result)[-1]) {
        offset <- offset + nrow(result[[i-1]]$snpinfo)
        if(result[[i]]$snpinfo$chr[1] == result[[i-1]]$snpinfo$chr[1]) { # same chromosome so advanced indexes
            result[[i]]$snpinfo$index <- result[[i]]$snpinfo$index + offset
        }
        else offset <- 0
    }
    snpinfo <- do.call("rbind", lapply(result, "[[", "snpinfo"))
    rownames(snpinfo) <- seq_len(nrow(snpinfo))

    list(lod=lod, snpinfo=snpinfo)
}


# scan1snps where snpinfo provided
#     snpinfo must contain snp_id, chr, pos, SDP
scan1snps_snpinfo <-
    function(genoprobs, map, pheno, kinship=NULL, addcovar=NULL, Xcovar=NULL, intcovar=NULL,
             weights=NULL, reml=TRUE, model=c("normal", "binary"), snpinfo,
             keep_all_snps=FALSE, ...)
{
    if(is.null(snpinfo) || nrow(snpinfo)==0) return(NULL) # no snps to study
    model <- match.arg(model)

    # reduce to common chromosomes
    pchr <- names(genoprobs)
    mchr <- names(map)
    schr <- unique(snpinfo$chr)
    cchr <- pchr[pchr %in% mchr & pchr %in% schr]
    if(length(cchr)==0) {
        warning("No common chromosomes among genoprobs, map, and snpinfo")
        return(NULL)
    }
    genoprobs <- genoprobs[,cchr]
    map <- map[cchr]
    snpinfo <- snpinfo[snpinfo$chr %in% cchr,,drop=FALSE]

    # snpinfo -> add index
    snpinfo <- index_snps(map, snpinfo)

    # genoprob -> snpprob
    snp_pr <- genoprob_to_snpprob(genoprobs, snpinfo)

    # scan1
    lod <- scan1(snp_pr, pheno=pheno, kinship=subset_kinship(kinship, chr=cchr),
                 addcovar=addcovar, Xcovar=Xcovar, intcovar=intcovar,
                 weights=weights, reml=reml, model=model, ...)

    if(!keep_all_snps) {
        snpinfo <- reduce_to_index_snps(snpinfo)
    }

    # return list with lod scores + indexed snpinfo
    list(lod=lod, snpinfo=snpinfo)
}

#' Compare two marker maps
#'
#' Compare two marker maps, identifying markers that are only in one
#' of the two maps, or that are in different orders on the two maps.
#'
#' @param map1 A list of numeric vectors; each vector gives marker
#' positions for a single chromosome.
#' @param map2 A second map, in the same format as `map1`.
#'
#' @return
#' A data frame containing
#' * `marker` - marker name
#' * `chr_map1` - chromosome ID on `map1`
#' * `pos_map1` - position on `map1`
#' * `chr_map2` - chromosome ID on `map2`
#' * `pos_map2` - position on `map2`
#'
#' @examples
#' # load some data
#' iron <- read_cross2( system.file("extdata", "iron.zip", package="qtl2") )
#' gmap <- iron$gmap
#' pmap <- iron$pmap
#'
#' # omit a marker from each map
#' gmap[[7]] <- gmap[[7]][-3]
#' pmap[[8]] <- pmap[[8]][-7]
#' # swap order of a couple of markers on the physical map
#' names(pmap[[9]])[3:4] <- names(pmap[[9]])[4:3]
#' # move a marker to a different chromosome
#' pmap[[10]] <- c(pmap[[10]], pmap[[1]][2])[c(1,3,2)]
#' pmap[[1]] <- pmap[[1]][-2]
#'
#' # compare these messed-up maps
#' compare_maps(gmap, pmap)
#'
#' @importFrom stats setNames
#' @export
compare_maps <-
    function(map1, map2)
{
    if(!is.list(map1) || !is.list(map2)) stop("map1 and map2 should be lists")

    # find markers that are not on both maps
    mar1 <- unlist(lapply(map1, names))
    mar2 <- unlist(lapply(map2, names))
    only1 <- mar1[!(mar1 %in% mar2)]
    only2 <- mar2[!(mar2 %in% mar1)]
    both <- mar1[mar1 %in% mar2]

    # vectors with chr ID and position
    chr1 <- setNames(rep(names(map1), vapply(map1, length, 0)), mar1)
    chr2 <- setNames(rep(names(map2), vapply(map2, length, 0)), mar2)
    pos1 <- setNames(unlist(map1), mar1)
    pos2 <- setNames(unlist(map2), mar2)

    result <- data.frame(marker=character(0),
                         chr_map1=character(0),
                         pos_map1=numeric(0),
                         chr_map2=character(0),
                         pos_map2=numeric(0),
                         stringsAsFactors=FALSE)

    if(length(only1) > 0) {
        result <- rbind(result,
                        data.frame(marker=only1,
                                   chr_map1=chr1[only1],
                                   pos_map1=pos1[only1],
                                   chr_map2=rep(NA, length(only1)),
                                   pos_map2=rep(NA, length(only1)),
                                   stringsAsFactors=FALSE))
    }
    if(length(only2) > 0) {
        result <- rbind(result,
                        data.frame(marker=only2,
                                   chr_map1=rep(NA, length(only2)),
                                   pos_map1=rep(NA, length(only2)),
                                   chr_map2=chr2[only2],
                                   pos_map2=pos2[only2],
                                   stringsAsFactors=FALSE))
    }

    # reduce to common markers
    for(i in seq_along(map1))
        map1[[i]] <- map1[[i]][names(map1[[i]]) %in% both]
    for(i in seq_along(map2))
        map2[[i]] <- map2[[i]][names(map2[[i]]) %in% both]

    # reduce chr and position vectors
    chr1 <- chr1[both]
    chr2 <- chr2[both]
    pos1 <- pos1[both]
    pos2 <- pos2[both]

    # markers on different chromosomes
    diff_chr <- names(chr1)[chr1 != chr2[names(chr1)]]
    if(length(diff_chr) > 0) {
        result <- rbind(result,
                        data.frame(marker=diff_chr,
                                   chr_map1=chr1[diff_chr],
                                   pos_map1=pos1[diff_chr],
                                   chr_map2=chr2[diff_chr],
                                   pos_map2=pos2[diff_chr],
                                   stringsAsFactors=FALSE))
    }

    # reduce to common markers
    same_chr <- both[!(both %in% diff_chr)]
    for(i in seq_along(map1))
        map1[[i]] <- map1[[i]][names(map1[[i]]) %in% same_chr]
    for(i in seq_along(map2))
        map2[[i]] <- map2[[i]][names(map2[[i]]) %in% same_chr]

    # reduce chr and position vectors
    chr1 <- chr1[same_chr]
    chr2 <- chr2[same_chr]
    pos1 <- pos1[same_chr]
    pos2 <- pos2[same_chr]

    # order markers by position on first chromosome
    map2 <- map2[names(map1)]
    for(i in seq_along(map1)) {
        map1[[i]] <- map1[[i]][order(map1[[i]])]
        map2[[i]] <- map2[[i]][names(map1[[i]])]
    }

    # order problems
    for(i in seq_along(map2)) {
        wh <- which(diff(map2[[i]]) < 0)
        wh <- sort(unique(c(wh, wh+1)))
        problems <- names(map1[[i]])[wh]
        if(length(problems) > 0) {
            result <- rbind(result,
                            data.frame(marker=problems,
                                       chr_map1=chr1[problems],
                                       pos_map1=pos1[problems],
                                       chr_map2=chr2[problems],
                                       pos_map2=pos2[problems],
                                       stringsAsFactors=FALSE))
        }
    }

    rownames(result) <- seq_len(nrow(result))

    result
}

context("compare maps")

test_that("compare_maps works", {

    iron <- read_cross2( system.file("extdata", "iron.zip", package="qtl2") )
    gmap <- iron$gmap
    pmap <- iron$pmap

    expected <- data.frame(marker=character(0),
                           chr_map1=character(0),
                           pos_map1=numeric(0),
                           chr_map2=character(0),
                           pos_map2=numeric(0),
                           stringsAsFactors=FALSE)

    expect_equal(compare_maps(gmap, pmap), expected)

    # omit a marker from each map
    gmap[[7]] <- gmap[[7]][-3]
    pmap[[8]] <- pmap[[8]][-7]
    # swap order of a couple of markers on the physical map
    names(pmap[[9]])[3:4] <- names(pmap[[9]])[4:3]
    # move a marker to a different chromosome
    pmap[[10]] <- c(pmap[[10]], pmap[[1]][2])[c(1,3,2)]
    pmap[[1]] <- pmap[[1]][-2]

    expected <- structure(list(marker = c("D8Mit120", "D7Nds5", "D1Mit80", "D9Mit10", "D9Mit182"),
                               chr_map1 = c("8", NA, "1", "9", "9"),
                               pos_map1 = c(69.9, NA, 51.4, 43.7, 53.6),
                               chr_map2 = c(NA, "7", "10", "9", "9"),
                               pos_map2 = c(NA, 44.201096, 86.377953, 101.52887, 89.977078)),
                          row.names = c(NA, 5L), class = "data.frame")

    expect_equal(compare_maps(gmap, pmap), expected)

})

#' Convert genotype probabilities to SNP probabilities
#'
#' For multi-parent populations, convert use founder genotypes at a
#' set of SNPs to convert founder-based genotype probabilities to SNP
#' genotype probabilities.
#'
#' @param genoprobs Genotype probabilities as
#' calculated by [calc_genoprob()].
#'
#' @param snpinfo Data frame with SNP information with the following
#'     columns (the last three are generally derived with
#'     [index_snps()]):
#' * `chr` - Character string or factor with chromosome
#' * `pos` - Position (in same units as in the `"map"`
#'     attribute in `genoprobs`.
#' * `sdp` - Strain distribution pattern: an integer, between
#'     1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
#'     binary encoding indicates the founder genotypes
#' * `snp` - Character string with SNP identifier (if
#'     missing, the rownames are used).
#' * `index` - Indices that indicate equivalent
#'     groups of SNPs, calculated by [index_snps()].
#' * `intervals` - Indexes that indicate which marker
#'     intervals the SNPs reside.
#' * `on_map` - Indicate whether SNP coincides with a marker
#'     in the `genoprobs`
#'
#' Alternatively, `snpinfo` can be a object of class `"cross2"`,
#' as output by [read_cross2()], containing the data for a multi-parent
#' population with founder genotypes, in which case the SNP
#' information for all markers with complete founder genotype data is
#' calculated and then used. But, in this case, the genotype
#' probabilities must be at the markers in the cross.
#'
#' @return An object of class `"calc_genoprob"`, like the input `genoprobs`,
#' but with imputed genotype probabilities at the selected SNPs indicated in
#' `snpinfo$index`. See [calc_genoprob()].
#'
#' If the input `genoprobs` is for allele probabilities, the
#' `probs` output has just two probability columns (for the two SNP
#' alleles). If the input has a full set of \eqn{n(n+1)/2}
#' probabilities for \eqn{n} strains, the `probs` output has 3 probabilities
#' (for the three SNP genotypes). If the input has full genotype
#' probabilities for the X chromosome (\eqn{n(n+1)/2} genotypes for
#' the females followed by \eqn{n} hemizygous genotypes for the
#' males), the output has 5 probabilities: the 3 female SNP genotypes
#' followed by the two male hemizygous SNP genotypes.
#'
#' @details We first split the SNPs by chromosome and use
#' `snpinfo$index` to subset to non-equivalent SNPs.
#' `snpinfo$interval` indicates the intervals in the genotype
#' probabilities that contain each. For SNPs contained within an
#' interval, we use the average of the probabilities for the two
#' endpoints. We then collapse the probabilities according to the
#' strain distribution pattern.
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DO_Recla/recla.zip")
#' recla <- read_cross2(file)
#' recla <- recla[c(1:2,53:54), c("19","X")] # subset to 4 mice and 2 chromosomes
#' probs <- calc_genoprob(recla, error_prob=0.002)
#'
#' # founder genotypes for a set of SNPs
#' snpgeno <- rbind(m1=c(3,1,1,3,1,1,1,1),
#'                  m2=c(1,3,1,3,1,3,1,3),
#'                  m3=c(1,1,1,1,3,3,3,3),
#'                  m4=c(1,3,1,3,1,3,1,3))
#' sdp <- calc_sdp(snpgeno)
#' snpinfo <- data.frame(chr=c("19", "19", "X", "X"),
#'                       pos=c(40.36, 40.53, 110.91, 111.21),
#'                       sdp=sdp,
#'                       snp=c("m1", "m2", "m3", "m4"), stringsAsFactors=FALSE)
#'
#' # identify groups of equivalent SNPs
#' snpinfo <- index_snps(recla$pmap, snpinfo)
#'
#' # collapse to SNP genotype probabilities
#' snpprobs <- genoprob_to_snpprob(probs, snpinfo)
#'
#' # could also first convert to allele probs
#' aprobs <- genoprob_to_alleleprob(probs)
#' snpaprobs <- genoprob_to_snpprob(aprobs, snpinfo)
#' }
#'
#' @seealso [index_snps()], [calc_genoprob()], [scan1snps()]
#' @export
genoprob_to_snpprob <-
    function(genoprobs, snpinfo)
{
    if(is.null(genoprobs)) stop("genoprobs is NULL")
    if(is.null(snpinfo)) stop("snpinfo is NULL")
    if(inherits(snpinfo, "cross2")) { # genoprobs -> snpprobs for all markers in a cross
        return(snpprob_from_cross(genoprobs, snpinfo))
    }

    if(nrow(snpinfo)==0) {
        result <- genoprobs[,names(genoprobs)[1]]
        if(length(attr(result, "alleles")) == ncol(result[[1]])) {
            result[[1]] <- result[[1]][,1:2,numeric(0),drop=FALSE]
            colnames(result[[1]]) <- c("A", "B")
        } else if(attr(result, "is_x_chr")[1]) {
            result[[1]] <- result[[1]][,1:5,numeric(0),drop=FALSE]
            colnames(result[[1]]) <- c("AA", "AB", "BB", "AY", "BY")
        } else {
            result[[1]] <- result[[1]][,1:3,numeric(0),drop=FALSE]
            colnames(result[[1]]) <- c("AA", "AB", "BB")
        }
        return(result)
    }


    uchr <- unique(snpinfo$chr)
    chrID <- names(genoprobs)
    if(!all(uchr %in% chrID)) {
        mischr <- uchr[!(uchr %in% chrID)]
        stop("Not all chr found in genoprobs: ", paste(mischr, collapse=","))
    }
    # reorder
    uchr <- factor(factor(uchr, levels=chrID))

    # check for index, interval, and on_map
    if(!all(c("sdp", "index", "interval", "on_map") %in% colnames(snpinfo)))
        stop('snpinfo should contain columns "sdp", "index", "interval", and "on_map".')

    # if more than one chromosome:
    if(length(uchr) > 1) {
        uchr <- as.character(uchr) # convert back to character

        ### split snpinfo by chr, ordered as in genoprobs
        chr <- factor(snpinfo$chr, levels=uchr)
        snpinfo_spl <- split(snpinfo, chr)

        ### loop over chromosomes, using recursion
        results <- vector("list", length(uchr))
        names(results) <- uchr
        for(i in seq_along(uchr))
            results[[i]] <- genoprob_to_snpprob(genoprobs, snpinfo_spl[[i]])[[1]]

        ### add attributes
        attr(results, "crosstype") <- "snps"
        attr(results, "alleles") <- c("A", "B")
        attr(results, "is_x_chr") <- attr(genoprobs, "is_x_chr")[uchr]
        class(results) <- c("calc_genoprob", "list")

        return(results)
    }

    # one chromosome

    # make chromosome a character string again
    uchr <- as.character(uchr)

    ### number of alleles
    alleles <- attr(genoprobs, "alleles")
    n_alleles <- length(alleles)

    ### subset the snpinfo
    snpinfo <- snpinfo[sort(unique(snpinfo$index)),,drop=FALSE]

    ### grab stuff
    sdp <- snpinfo$sdp
    interval <- snpinfo$interval
    on_map <- snpinfo$on_map

    if(any(interval < 0 | interval > dim(genoprobs[[uchr]])[3]-1))
        stop("interval values outside of the range [0, ",
             dim(genoprobs[[uchr]])[3]-1, "].")

    if(any(sdp < 1 | sdp > 2^n_alleles-1))
        stop("sdp values outside of the range [1, ",
             2^n_alleles-1, "].")

    # subset to the single chromosome
    results <- vector("list", 1)
    names(results) <- uchr

    ### genoprobs -> SNP genotype probs
    if(ncol(genoprobs[[uchr]]) == n_alleles) { # allele probs
        results[[1]] <- .alleleprob_to_snpprob(genoprobs[[uchr]], sdp, interval, on_map)
        coln <- c("A", "B")
    }
    else if(ncol(genoprobs[[uchr]]) == n_alleles*(n_alleles+1)/2) { # autosomal genotype probs
        results[[1]] <- .genoprob_to_snpprob(genoprobs[[uchr]], sdp, interval, on_map)
        coln <- c("AA", "AB", "BB")
    }
    else if(ncol(genoprobs[[uchr]]) == n_alleles + n_alleles*(n_alleles+1)/2) { # X chr genotype probs
        results[[1]] <- .Xgenoprob_to_snpprob(genoprobs[[uchr]], sdp, interval, on_map)
        coln <- c("AA", "AB", "BB", "AY", "BY")
    }
    else {
        stop("genoprobs has ", ncol(genoprobs[[uchr]]),
             " columns but there are ", n_alleles, " alleles")
    }

    # add dimnames
    snpnames <- snpinfo$snp
    if(is.null(snpnames)) snpnames <- rownames(snpinfo)
    dimnames(results[[1]]) <- list(rownames(genoprobs[[uchr]]),
                                        coln, snpnames)

    attr(results, "is_x_chr") <- attr(genoprobs, "is_x_chr")[uchr]
    attr(results, "crosstype") <- "snps"
    attr(results, "alleles") <- c("A", "B")
    class(results) <- c("calc_genoprob", "list")

    results
}

# function for subsetting by chromosome
#
# chr = logical, numeric, or character strings
#       (numeric converted to character strings)
# all_chr = vector of character strings with all chromosome names
#
# result = vector of character strings (the selected ones)
subset_chr <-
    function(chr, all_chr)
{
    if(is.logical(chr)) {
        if(length(chr) != length(all_chr))
            stop("chr is logical but length [", length(chr), "] != n_chr is x [",
                 length(all_chr), "]")
        chr <- all_chr[chr]
    } else {
        chr <- as.character(chr)

        # look for negatives; turn to positives
        if(any(grepl("^\\-", chr))) {
            if(!all(grepl("^\\-", chr)))
                stop("Can't mix negative and positive chr subscripts")
            chr <- sub("^\\-", "", chr)
            if(!all(chr %in% all_chr)) {
                if(!any(chr %in% all_chr))
                    stop("None of the chr found in the cross object")
                warning("Some chr not found: ", paste(chr[!(chr %in% all_chr)], collapse=", "))
                chr <- chr[chr %in% all_chr]
            }
            chr <- all_chr[!(all_chr %in% chr)]
        }

        if(!all(chr %in% all_chr)) {
            if(!any(chr %in% all_chr)) stop("None of the chromosomes in cross")
            warning("Some chr not in cross: ", paste(chr[!(chr %in% all_chr)], collapse=", "))
            chr <- chr[chr %in% all_chr]
        }
    }

    chr
}

# calc_genoprob2
# Calculate conditional genotype probabilities
# this version pre-calculates init, step, and emit (attempting to be faster for DO)
#
# Same input and output as calc_genoprob()
calc_genoprob2 <-
function(cross, map=NULL, error_prob=1e-4,
         map_function=c("haldane", "kosambi", "c-f", "morgan"),
         quiet=TRUE, cores=1)
{
    # check inputs
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')
    if(error_prob < 0)
        stop("error_prob must be > 0")
    map_function <- match.arg(map_function)

    # set up cluster; set quiet=TRUE if multi-core
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # pseudomarker map
    if(is.null(map)) {
        if(is.null(cross$gmap)) stop("If cross does not contain a genetic map, map must be provided.")
        map <- insert_pseudomarkers(cross$gmap)
    }
    # possibly subset the map
    if(length(map) != length(cross$geno) || !all(names(map) == names(cross$geno))) {
        chr <- names(cross$geno)
        if(!all(chr %in% names(map)))
            stop("map doesn't contain all of the necessary chromosomes")
        map <- map[chr]
    }
    # calculate marker index object
    index <- create_marker_index(lapply(cross$geno, colnames), map)

    probs <- vector("list", length(map))
    rf <- map2rf(map, map_function)

    # deal with missing information
    ind <- rownames(cross$geno[[1]])
    chrnames <- names(cross$geno)
    is_x_chr <- handle_null_isxchr(cross$is_x_chr, chrnames)
    cross$is_female <- handle_null_isfemale(cross$is_female, ind)
    cross$cross_info <- handle_null_isfemale(cross$cross_info, ind)

    founder_geno <- cross$founder_geno
    if(is.null(founder_geno))
        founder_geno <- create_empty_founder_geno(cross$geno)

    by_group_func <- function(i) {
        pr <- .calc_genoprob2(cross$crosstype, t(cross$geno[[chr]][group[[i]],,drop=FALSE]),
                              founder_geno[[chr]], cross$is_x_chr[chr], cross$is_female[group[[i]][1]],
                              cross$cross_info[group[[i]][1],], rf[[chr]], index[[chr]],
                              error_prob)
        aperm(pr, c(2,1,3))
    }

    # split individuals into groups with common sex and cross_info
    sex_crossinfo <- paste(cross$is_female, apply(cross$cross_info, 1, paste, collapse=":"), sep=":")
    group <- split(seq(along=sex_crossinfo), sex_crossinfo)
    names(group) <- NULL
    nc <- n_cores(cores)
    while(nc > length(group) && max(sapply(group, length)) > 1) { # successively split biggest group in half until there are as many groups as cores
        mx <- which.max(sapply(group, length))
        g <- group[[mx]]
        group <- c(group, list(g[seq(1, length(g), by=2)]))
        group[[mx]] <- g[seq(2, length(g), by=2)]
    }
    groupindex <- seq(along=group)

    probs <- vector("list", length(cross$geno))
    for(chr in seq(along=cross$geno)) {
        if(!quiet) message("Chr ", names(cross$geno)[chr])

        # calculations in parallel [if cores==1, it just does lapply()]
        temp <- cluster_lapply(cores, groupindex, by_group_func)

        # paste them back together
        d <- vapply(temp, dim, rep(0,3))
        nr <- sum(d[1,])
        probs[[chr]] <- array(dim=c(nr, d[2,1], d[3,1]))
        for(i in groupindex)
            probs[[chr]][group[[i]],,] <- temp[[i]]

        # genotype names
        alleles <- cross$alleles
        if(is.null(alleles)) # no alleles saved; use caps
            alleles <- LETTERS[1:ncol(probs[[chr]])]
        gnames <- geno_names(cross$crosstype,
                             alleles,
                             cross$is_x_chr[chr])
        if(length(gnames) != ncol(probs[[chr]])) {
            warning("genotype names has length (", length(gnames),
                    ") != genoprob dim (", ncol(probs[[chr]]), ")")
            gnames <- NULL
        }

        dimnames(probs[[chr]]) <- list(rownames(cross$geno[[chr]]),
                                       gnames,
                                       names(map[[chr]]))

    }

    names(probs) <- names(cross$geno)

    attr(probs, "crosstype") <- cross$crosstype
    attr(probs, "is_x_chr") <- cross$is_x_chr
    attr(probs, "alleles") <- cross$alleles
    attr(probs, "alleleprobs") <- FALSE

    class(probs) <- c("calc_genoprob", "list")

    probs
}

#' Recode SNPs by major allele
#'
#' For multi-parent populations with founder genotypes, recode the raw
#' SNP genotypes so that `1` means homozygous for the major allele in the
#' founders.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#'
#' @return The input cross object with the raw SNP genotypes recoded so that
#' `1` is homozygous for the major alleles in the founders.
#'
#' @export
#' @keywords utilities
#' @seealso [calc_raw_founder_maf()], [calc_raw_maf()]
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' DOex <- recode_snps(DOex)
#' }
recode_snps <-
    function(cross)
{
    if(!("founder_geno" %in% names(cross))) {
        stop("cross does not contain founder genotypes")
    }

    for(chr in seq_along(cross$founder_geno)) {
        fg <- cross$founder_geno[[chr]]
        fg[fg!=1 & fg!=3] <- NA
        p1 <- colMeans(fg==1, na.rm=TRUE)
        recode <- (!is.na(p1) & p1 < 0.5)

        g <- cross$geno[[chr]]
        g[g==0] <- NA

        g[,recode] <- 4 - g[,recode]
        fg[,recode] <- 4 - fg[,recode]

        g[is.na(g)] <- 0
        fg[is.na(fg)] <- 0

        cross$founder_geno[[chr]] <- fg
        cross$geno[[chr]] <- g
    }

    cross
}

context("recode_snps")

test_that("recode_snps works", {

    skip_if(isnt_karl(), "this test only run locally")

    # load example data
    file <- paste0("https://raw.githubusercontent.com/rqtl/",
                   "qtl2data/main/DOex/DOex.zip")
    DOex <- read_cross2(file)

    pr <- calc_genoprob(DOex, error_prob=0.002, cores=2)
    nmis_ind <- n_missing(DOex)
    nmis_mar <- n_missing(DOex, "mar")

    DOex <- recode_snps(DOex)

    # no change in number missing
    expect_equal(nmis_ind, n_missing(DOex))
    expect_equal(nmis_mar, n_missing(DOex, "mar"))

    # no change in genotype probabilities
    expect_equal(pr, calc_genoprob(DOex, error_prob=0.002, cores=2))

    fg <- do.call("cbind", DOex$founder_geno)
    fg[fg!=1 & fg!=3] <- NA
    expect_true(all(colMeans(fg==1, na.rm=TRUE) >= 0.5))

})

#' Clean scan1 output
#'
#' Clean scan1 output by replacing negative values with NA and remove
#' rows where all values are NA.
#'
#' @param object Output of [scan1()].
#' @param ... Ignored at present
#'
#' @return The input object with negative values replaced with NAs and then rows with all NAs removed.
#'
#' @export
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,"2"]}
#' pr <- calc_genoprob(iron)
#' out <- scan1(pr, iron$pheno)
#'
#' out <- clean(out)

clean_scan1 <-
    function(object, ...)
{
    cl <- class(object)
    at <- attributes(object)
    if(!("scan1" %in% cl)) {
        stop("object is not of class scan1")
    }

    object[object < 0] <- NA

    omit <- apply(object, 1, function(a) all(is.na(a)))
    if(any(omit)) {
        object <- object[!omit, , drop=FALSE]
        class(object) <- cl

        # retain attributes
        for(a in names(at)) {
            if(!(a %in% c("dim", "dimnames", "class"))) {
                attr(object, a) <- at[[a]]
            }
        }
    }

    object
}


#' @rdname clean_scan1
#' @export
clean.scan1 <- clean_scan1

# subset kinship matrix by individuals or chromosomes
subset_kinship <-
    function(kinship, ind=NULL, chr=NULL)
{
    if(is.null(kinship)) return(NULL)

    # already decomposed
    if(is_kinship_decomposed(kinship)) {
        if(!is.null(ind)) {
            # maybe it's not really needing to be subset by individual
            if(is_kinship_list(kinship)) {
                k_ind <- rownames(kinship[[1]]$vectors)
            } else {
                k_ind <- rownames(kinship$vectors)
            }
            new_ind <- subset_ind(ind, k_ind)
            if(length(new_ind) != length(k_ind) ||
               any(new_ind != k_ind)) {
                stop("Can't subset decomposed kinship matrices by individual")
            }
            # otherwise, not really being subset so can ignore ind argument

            if(!is_kinship_list(kinship)) { # can ignore chr argument
                return(kinship)
            }
        }

        if(!is.null(chr)) {
            if(is_kinship_list(kinship)) {
                chr <- subset_chr(chr, names(kinship))
                kinship <- kinship[chr]
            }
            else return(kinship) # if just one chromosmoe, just return it
        }
        if(length(kinship)==1) return(kinship[[1]])
        return(kinship)
    }

    # not yet decomposed
    if(!is.null(chr) && is.list(kinship)) {
        chr <- subset_chr(chr, names(kinship))
        kinship <- kinship[chr]

        if(is.list(kinship) && length(kinship) == 1)
            kinship <- kinship[[1]]
    }

    if(!is.null(ind)) {
        if(is.list(kinship)) {
            ind <- subset_ind(ind, rownames(kinship[[1]]))
            kinship <- lapply(kinship, function(a) a[ind, ind])
        }
        else {
            ind <- subset_ind(ind, rownames(kinship))
            kinship <- kinship[ind,ind]
        }
    }

    kinship
}

context("subset kinship matrices")

test_that("subset.kinship works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[c("2","3","4","5","6","7"), 2:6]
    pr <- calc_genoprob(iron)

    k <- calc_kinship(pr)
    k_loco <- calc_kinship(pr, "loco")

    # subset by individuals
    expect_equal(subset_kinship(k, ind=c(2,4)), k[c(2,4), c(2,4)])
    expect_equal(subset_kinship(k, ind=c("2","4")), k[c("2","4"), c("2","4")])

    # subset by individuals, loco case
    expect_equal(subset_kinship(k_loco, ind=c(2,4)), lapply(k_loco, function(a) a[c(2,4), c(2,4)]))
    expect_equal(subset_kinship(k_loco, ind=c("2","4")), lapply(k_loco, function(a) a[c("2","4"), c("2","4")]))

    # chr argument ignored if a plain matrix
    expect_equal(subset_kinship(k, ind=c(2,4), chr=5), k[c(2,4), c(2,4)])
    expect_equal(subset_kinship(k, ind=c("2","4"), chr=5), k[c("2","4"), c("2","4")])

    # chr argument, loco case, one chr in result
    expect_equal(subset_kinship(k_loco, ind=c(2,4), chr=5), k_loco[["5"]][c(2,4), c(2,4)])
    expect_equal(subset_kinship(k_loco, ind=c("2","4"), chr=5), k_loco[["5"]][c("2","4"), c("2","4")])

    # chr argument, loco case, mult chr in result
    expect_equal(subset_kinship(k_loco, ind=c(2,4), chr=c(2,5)), lapply(k_loco[c("2","5")], function(a) a[c(2,4), c(2,4)]))
    expect_equal(subset_kinship(k_loco, ind=c("2","4"), chr=c(2,5)), lapply(k_loco[c("2", "5")], function(a) a[c("2","4"), c("2","4")]))

    # chr argument, loco case, negative subscripts
    expect_equal(subset_kinship(k_loco, ind=c(2,4), chr=-c(2,4)), lapply(k_loco[c("3","5","6")], function(a) a[c(2,4), c(2,4)]))
    expect_equal(subset_kinship(k_loco, ind=c("2","4"), chr=c("-2","-4")), lapply(k_loco[c("3","5","6")], function(a) a[c("2","4"), c("2","4")]))

})

#' Fit single-QTL model at a single position
#'
#' Fit a single-QTL model at a single putative QTL position and get detailed results
#' about estimated coefficients and individuals contributions to the LOD score.
#'
#' @param genoprobs A matrix of genotype probabilities, individuals x genotypes.
#'     If NULL, we create a single intercept column, matching the individual IDs in `pheno`.
#' @param pheno A numeric vector of phenotype values (just one phenotype, not a matrix of them)
#' @param kinship Optional kinship matrix.
#' @param addcovar An optional numeric matrix of additive covariates.
#' @param nullcovar An optional numeric matrix of additional additive
#' covariates that are used under the null hypothesis (of no QTL)
#' but not under the alternative (with a QTL). This is needed for
#' the X chromosome, where we might need sex as a additive
#' covariate under the null hypothesis, but we wouldn't want to
#' include it under the alternative as it would be collinear with
#' the QTL effects.
#' @param intcovar An optional numeric matrix of interactive covariates.
#' @param weights An optional numeric vector of positive weights for the
#' individuals. As with the other inputs, it must have `names`
#' for individual identifiers.
#' @param contrasts An optional numeric matrix of genotype contrasts, size
#' genotypes x genotypes. For an intercross, you might use
#' `cbind(mu=c(1,1,1), a=c(-1, 0, 1), d=c(0, 1, 0))` to get
#' mean, additive effect, and dominance effect. The default is the
#' identity matrix.
#' @param model Indicates whether to use a normal model (least
#'     squares) or binary model (logistic regression) for the phenotype.
#'     If `model="binary"`, the phenotypes must have values in
#'     \eqn{[0, 1]}.
#' @param zerosum If TRUE, force the genotype or allele coefficients
#'     sum to 0 by subtracting their mean and add another column with
#'     the mean. Ignored if `contrasts` is provided.
#' @param se If TRUE, calculate the standard errors.
#' @param hsq (Optional) residual heritability; used only if
#' `kinship` provided.
#' @param reml If `kinship` provided: if `reml=TRUE`, use
#' REML; otherwise maximum likelihood.
#' @param blup If TRUE, fit a model with QTL effects being random, as in [scan1blup()].
#' @param var If TRUE, calculate the variance-covariance matrix of the coefficients
#'     (in which case `se` is ignored, and standard errors will also be included).
#'     But note that the variance-covariance matrix is only returned
#'     if the covariate matrix is full rank.
#' @param ... Additional control parameters; see Details;
#'
#' @return A list containing
#' * `coef` - Vector of estimated coefficients.
#' * `SE` - Vector of estimated standard errors (included if `se=TRUE`).
#' * `lod` - The overall lod score.
#' * `ind_lod` - Vector of individual contributions to the LOD score (not provided if `kinship` is used).
#' * `fitted`  - Fitted values.
#' * `resid`  - Residuals.
#' If `blup==TRUE`, only `coef` and `SE` are included at present.
#'
#' @details For each of the inputs, the row names are used as
#' individual identifiers, to align individuals.
#'
#' If `kinship` is absent, Haley-Knott regression is performed.
#' If `kinship` is provided, a linear mixed model is used, with a
#' polygenic effect estimated under the null hypothesis of no (major)
#' QTL, and then taken as fixed as known in the genome scan.
#'
#' If `contrasts` is provided, the genotype probability matrix,
#' \eqn{P}, is post-multiplied by the contrasts matrix, \eqn{A}, prior
#' to fitting the model. So we use \eqn{P \cdot A}{P A} as the \eqn{X}
#' matrix in the model. One might view the rows of
#' \ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}
#' as the set of contrasts, as the estimated effects are the estimated
#' genotype effects pre-multiplied by
#' \ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}.
#'
#' The `...` argument can contain several additional control
#' parameters; suspended for simplicity (or confusion, depending on
#' your point of view). `tol` is used as a tolerance value for linear
#' regression by QR decomposition (in determining whether columns are
#' linearly dependent on others and should be omitted); default
#' `1e-12`. `maxit` is the maximum number of iterations for
#' converence of the iterative algorithm used when `model=binary`.
#' `bintol` is used as a tolerance for converence for the iterative
#' algorithm used when `model=binary`. `eta_max` is the maximum value
#' for the "linear predictor" in the case `model="binary"` (a bit of a
#' technicality to avoid fitted values exactly at 0 or 1).
#'
#' @references Haley CS, Knott SA (1992) A simple
#' regression method for mapping quantitative trait loci in line
#' crosses using flanking markers.  Heredity 69:315--324.
#'
#' Kang HM, Zaitlen NA, Wade CM, Kirby A, Heckerman D, Daly MJ, Eskin
#' E (2008) Efficient control of population structure in model
#' organism association mapping. Genetics 178:1709--1723.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,"7"] # subset to chr 7}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=5)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno[,1]
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#'
#' # scan chromosome 7 to find peak
#' out <- scan1(probs[,"7"], pheno, addcovar=covar)
#'
#' # find peak position
#' max_pos <- max(out, map)
#'
#' # genoprobs at max position
#' pr_max <- pull_genoprobpos(probs, map, max_pos$chr, max_pos$pos)
#'
#' # fit QTL model just at that position
#' out_fit1 <- fit1(pr_max, pheno, addcovar=covar)
#'
#' # plot estimated effects
#' plot_ci(out_fit1$coef[1:3], out_fit1$SE[1:3],
#'         xlab="Genotype", ylab="QTL effects")
#'
#' @seealso [pull_genoprobpos()], [find_marker()], [plot_ci()]
#'
#' @importFrom stats setNames
#' @export
fit1 <-
    function(genoprobs, pheno, kinship=NULL, addcovar=NULL, nullcovar=NULL,
             intcovar=NULL, weights=NULL,
             contrasts=NULL, model=c("normal", "binary"),
             zerosum=TRUE, se=TRUE, hsq=NULL, reml=TRUE, blup=FALSE,
             var=FALSE, ...)
{
    if(is.null(pheno)) stop("pheno is NULL")

    if(missing(genoprobs) || is.null(genoprobs)) { # create matrix of 1's
        if(!is.matrix(pheno)) pheno <- cbind(pheno)
        genoprobs <- matrix(1, ncol=1, nrow=nrow(pheno))
        dimnames(genoprobs) <- list(rownames(pheno), "intercept")
        zerosum <- FALSE
    }

    if(zerosum && se) var <- TRUE # force calculation of full variance matrix

    model <- match.arg(model)

    if(blup) {
        if(!is.null(intcovar)) warning("If blup==TRUE, intcovar ignored")
        if(model != "normal") warning('If blup==TRUE, model taken to be "normal"')
        if(!is.null(weights)) warning("If blup==TRUE, weights ignored")
        if(!is.null(hsq)) warning("If blup==TRUE, hsq ignored")
        if(var) { warning("For now, if blup==TRUE, var ignored"); var <- FALSE }

        rn <- rownames(genoprobs)
        cn <- colnames(genoprobs)
        genoprobs <- list("1" = array(genoprobs, dim=c(dim(genoprobs), 1)))
        dimnames(genoprobs[[1]]) <- list(rn, cn, "marker")

        blup_result <- scan1blup(genoprobs, pheno, kinship=kinship, addcovar=addcovar,
                                 nullcovar=nullcovar, contrasts=contrasts, se=se, reml=reml,
                                 tol=grab_dots(list(...), "tol", 1e-12))
        result <- list(coef=blup_result[1,])
        if(se) result$SE <- attr(blup_result, "SE")[1,]
        return(result)
    }

    if(!is.null(kinship)) { # use LMM; see fit1_pg.R
        return(fit1_pg(genoprobs, pheno, kinship, addcovar, nullcovar,
                       intcovar, weights, contrasts, zerosum, se, hsq, reml, var, ...))
    }

    # deal with the dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    if(!is_pos_number(tol)) stop("tol should be a single positive number")
    if(model=="binary") {
        bintol <- grab_dots(dotargs, "bintol", sqrt(tol)) # for model="binary"
        if(!is_pos_number(bintol)) stop("bintol should be a single positive number")
        eta_max <- grab_dots(dotargs, "eta_max", log(1-tol)-log(tol)) # for model="binary"
        if(!is_pos_number(eta_max)) stop("eta_max should be a single positive number")
        maxit <- grab_dots(dotargs, "maxit", 100) # for model="binary"
        if(!is_nonneg_number(maxit)) stop("maxit should be a single non-negative integer")

        check_extra_dots(dotargs, c("tol", "bintol", "eta_max", "maxit"))
    }
    else { # not binary trait
        check_extra_dots(dotargs, "tol")
    }


    # check that the objects have rownames
    check4names(pheno, addcovar, NULL, intcovar, nullcovar)

    # force things to be matrices
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(nullcovar)) {
        if(!is.matrix(nullcovar)) nullcovar <- as.matrix(nullcovar)
        if(!is.numeric(nullcovar)) stop("nullcovar is not numeric")
    }
    if(!is.null(intcovar)) {
        if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)
        if(!is.numeric(intcovar)) stop("intcovar is not numeric")
    }
    if(!is.null(contrasts)) {
        if(!is.matrix(contrasts)) contrasts <- as.matrix(contrasts)
        if(!is.numeric(contrasts)) stop("contrasts is not numeric")
    }

    if(model=="binary") {
        if(!is.null(kinship))
            stop("Can't yet account for kinship with model = \"binary\"")
        if(any(!is.na(pheno) & (pheno < 0 | pheno > 1)))
            stop('with model="binary", pheno should be in [0,1]')
    }
    else {
        # square-root of weights
        weights <- sqrt_weights(weights) # also check >0 (and if all 1's, turn to NULL)
    }

    # make sure pheno is a vector
    if(is.matrix(pheno) || is.data.frame(pheno)) {
        if(ncol(pheno) > 1)
            warning("Considering only the first phenotype.")
        rn <- rownames(pheno)
        pheno <- pheno[,1]
        names(pheno) <- rn
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }

    # genoprobs is a matrix?
    if(!is.matrix(genoprobs))
        stop("genoprobs should be a matrix, individuals x genotypes")

    # make sure contrasts is square n_genotypes x n_genotypes
    if(!is.null(contrasts)) {
        ng <- ncol(genoprobs)
        if(ncol(contrasts) != ng || nrow(contrasts) != ng)
            stop("contrasts should be a square matrix, ", ng, " x ", ng)
    }

    # find individuals in common across all arguments
    ind2keep <- get_common_ids(genoprobs, pheno, addcovar, nullcovar, intcovar,
                               weights, complete.cases=TRUE)
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    # omit individuals not in common
    genoprobs <- genoprobs[ind2keep,,drop=FALSE]
    pheno <- pheno[ind2keep]
    if(!is.null(addcovar)) addcovar <- addcovar[ind2keep,,drop=FALSE]
    if(!is.null(nullcovar)) nullcovar <- nullcovar[ind2keep,,drop=FALSE]
    if(!is.null(intcovar)) intcovar <- intcovar[ind2keep,,drop=FALSE]
    if(!is.null(weights)) weights <- weights[ind2keep]

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # make sure columns in intcovar are also in addcovar
    addcovar <- force_intcovar(addcovar, intcovar, tol)

    # multiply genoprobs by contrasts
    if(!is.null(contrasts))
        genoprobs <- genoprobs %*% contrasts

    X0 <- drop_depcols(cbind(rep(1, length(pheno)), addcovar, nullcovar), FALSE, tol)
    if(model=="normal") {
        # if weights, adjust phenotypes
        if(!is.null(weights)) pheno <- weights * pheno

        # weights have 0 dimension if missing
        if(is.null(weights)) weights <- numeric(0)

        # null fit
        fit0 <- fit1_hk_addcovar(X0, # plug addcovar where genoprobs would be
                                 pheno,
                                 matrix(nrow=length(pheno), ncol=0),     # empty slot for addcovar
                                 weights, se=FALSE, var=FALSE, tol)

        if(is.null(intcovar)) { # just addcovar
            if(is.null(addcovar)) addcovar <- matrix(nrow=length(ind2keep), ncol=0)
            fitA <- fit1_hk_addcovar(genoprobs, pheno, addcovar, weights, se=se, var=var, tol)
        }
        else {                  # intcovar
            fitA <- fit1_hk_intcovar(genoprobs, pheno, addcovar, intcovar,
                                     weights, se=se, var=var, tol)
        }

        # lod score
        n <- length(pheno)
        lod <- (n/2)*log10(fit0$rss/fitA$rss)

        # residual SDs using 1/n
        sigsq0 <- fit0$sigma^2/n*fit0$df
        sigsqA <- fitA$sigma^2/n*fitA$df

        # individual contributions to the lod score
        ind_lod <- 0.5*(fit0$resid^2/sigsq0 - fitA$resid^2/sigsqA + log(sigsq0) - log(sigsqA))/log(10)
        names(ind_lod) <- names(pheno)

        fitted <- setNames(fitA$fitted, names(pheno))
        resid <- pheno - fitted
    }
    else { # binary phenotype
        # null fit
        if(is.null(weights)) { # no weights
            fit0 <- fit_binreg(X0, pheno, FALSE, FALSE, maxit, bintol, tol, eta_max)
            weights <- numeric(0)
        }
        else {
            fit0 <- fit_binreg_weighted(X0, pheno, weights, FALSE, FALSE, maxit, bintol, tol, eta_max)
        }

        if(is.null(intcovar)) { # just addcovar
            if(is.null(addcovar)) addcovar <- matrix(nrow=length(ind2keep), ncol=0)
            fitA <- fit1_binary_addcovar(genoprobs, pheno, addcovar, weights, se=se, var=var,
                                         maxit, bintol, tol, eta_max)
        }
        else {                  # intcovar
            fitA <- fit1_binary_intcovar(genoprobs, pheno, addcovar, intcovar,
                                         weights, se=se, var=var, maxit, bintol, tol, eta_max)
        }

        lod <- fitA$log10lik - fit0$log10lik

        # individual contributions to LOD
        if(length(weights)==0) weights <- rep(1, length(pheno))
        pA <- fitA$fitted_probs
        ind_lodA <- weights * (pheno*log10(pA) + (1-pheno)*log10(1-pA))
        p0 <- fit0$fitted_probs
        ind_lod0 <- weights * (pheno*log10(p0) + (1-pheno)*log10(1-p0))
        ind_lod <- ind_lodA - ind_lod0

        fitted <- stats::setNames(pA, names(pheno))
        resid <- pheno - fitted
    }

    # names of coefficients
    coef_names <- scan1coef_names(genoprobs, addcovar, intcovar)

    # center the QTL effects at zero and add an intercept
    if(zerosum && is.null(contrasts)) {
        ng <- dim(genoprobs)[2]
        whval <- seq_len(ng)

        A <- matrix(0, length(fitA$coef)+1, length(fitA$coef))
        diag(A) <- 1
        for(i in whval) { A[i,whval] <- -1/ng; A[i,i] <- 1-1/ng }
        A[nrow(A),whval] <- 1/ng

        fitA$coef <- as.numeric(A %*% fitA$coef)
        coef_names <- names(fitA$coef) <- c(coef_names, "intercept")

        if(var && !is.null(fitA$var)) {
            fitA$var <- A %*% fitA$var %*% t(A)
            fitA$SE <- sqrt(diag(fitA$var))
        }
        else if(se) {
            fitA$SE <- c(fitA$SE, sqrt(mean(fitA$SE[whval]^2, na.rm=TRUE)))
        }
    }

    if(var && !is.null(fitA$var)) { # results include var-cov matrix, if full rank
        dimnames(fitA$var) <- list(coef_names, coef_names)
        return(list(lod=lod, ind_lod=ind_lod,
                    coef=stats::setNames(fitA$coef, coef_names),
                    SE=stats::setNames(fitA$SE, coef_names),
                    var=fitA$var,
                    fitted=fitted, resid=resid))
    }
    else if(se) # results include standard errors
        return(list(lod=lod, ind_lod=ind_lod,
                    coef=stats::setNames(fitA$coef, coef_names),
                    SE=stats::setNames(fitA$SE, coef_names),
                    fitted=fitted, resid=resid))
    else
        return(list(lod=lod, ind_lod=ind_lod,
                    coef=stats::setNames(fitA$coef, coef_names),
                    fitted=fitted, resid=resid))
}

# Genome scan with a single-QTL and linear mixed model
#
# called by scan1() when kinship() is provided.
scan1_pg <-
    function(genoprobs, pheno, kinship, addcovar=NULL, Xcovar=NULL,
             intcovar=NULL, weights=NULL, reml=TRUE, hsq=NULL,
             cores=1, ...)
{
    # deal with the dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    if(!is_pos_number(tol)) stop("tol should be a single positive number")
    intcovar_method <- grab_dots(dotargs, "intcovar_method", "lowmem",
                                 c("highmem", "lowmem"))
    quiet <- grab_dots(dotargs, "quiet", TRUE)
    max_batch <- grab_dots(dotargs, "max_batch", NULL)
    if(!is.null(max_batch) && !is_pos_number(max_batch)) stop("max_batch should be a single positive integer")
    check_boundary <- grab_dots(dotargs, "check_boundary", TRUE)
    check_extra_dots(dotargs, c("tol", "intcovar_method", "check_boundary", "quiet", "max_batch"))

    # check that the objects have rownames
    check4names(pheno, addcovar, Xcovar, intcovar)

    # force things to be matrices
    if(!is.matrix(pheno)) {
        pheno <- as.matrix(pheno)
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }
    if(is.null(colnames(pheno))) # force column names
        colnames(pheno) <- paste0("pheno", seq_len(ncol(pheno)))
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(Xcovar)) {
        if(!is.matrix(Xcovar)) Xcovar <- as.matrix(Xcovar)
        if(!is.numeric(Xcovar)) stop("Xcovar is not numeric")
    }
    if(!is.null(intcovar)) {
        if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)
        if(!is.numeric(intcovar)) stop("intcovar is not numeric")
    }

    # check that kinship matrices are square with same IDs
    kinshipIDs <- check_kinship(kinship, length(genoprobs))

    # multiply kinship matrix by 2; rest is using 2*kinship
    # see Almasy & Blangero (1998) https://doi.org/10.1086/301844
    kinship <- double_kinship(kinship)

    # take square-root of weights
    weights <- sqrt_weights(weights)

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, addcovar, Xcovar, intcovar,
                               kinshipIDs, weights, complete.cases=TRUE)
    ind2keep <- get_common_ids(ind2keep, pheno[rowSums(is.finite(pheno)) > 0,,drop=FALSE])
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # make sure columns in intcovar are also in addcovar
    addcovar <- force_intcovar(addcovar, intcovar, tol)

    # drop things from Xcovar that are already in addcovar
    Xcovar <- drop_xcovar(addcovar, Xcovar, tol)

    # batch phenotypes by missing values
    phe_batches <- batch_cols(pheno[ind2keep,,drop=FALSE], max_batch)

    # drop cols in genotype probs that are all 0 (just looking at the X chromosome)
    genoprob_Xcol2drop <- genoprobs_col2drop(genoprobs)
    is_x_chr <- attr(genoprobs, "is_x_chr")
    if(is.null(is_x_chr)) is_x_chr <- rep(FALSE, length(genoprobs))

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # number of markers/pseudomarkers by chromosome, and their indexes to result matrix
    npos_by_chr <- dim(genoprobs)[3,]
    totpos <- sum(npos_by_chr)
    pos_index <- split(seq_len(totpos), rep(seq_len(length(genoprobs)), npos_by_chr))
    pos_names <- unlist(dimnames(genoprobs)[[3]])
    names(pos_names) <- NULL # this is just annoying

    # to contain the results
    result <- matrix(nrow=totpos, ncol=ncol(pheno))
    dimnames(result) <- list(pos_names, colnames(pheno))

    # number of chr to consider under null
    if(is_kinship_list(kinship)) n_null_chr <- length(kinship)
    else if(!is.null(Xcovar)) n_null_chr <- any(is_x_chr) + any(!is_x_chr)
    else n_null_chr <- 1

    if(is.null(hsq)) {
        hsq <- matrix(nrow=n_null_chr, ncol=ncol(pheno))
        dimnames(hsq) <- hsq_dimnames(kinship, Xcovar, is_x_chr, pheno)
        estimate_hsq <- TRUE
    } else {
        if(length(hsq) != n_null_chr * ncol(pheno)) {
            stop("hsq should be NULL or a matrix of size ", n_null_chr, " x ", ncol(pheno))
        }
        if(!is.matrix(hsq)) hsq <- as.matrix(hsq, ncol=ncol(pheno))
        if(nrow(hsq) != n_null_chr) {
            stop("hsq should be NULL or a matrix of size ", n_null_chr, " x ", ncol(pheno))
        }
        if(any(hsq < 0 | hsq > 1)) {
            stop("hsq values should be between 0 and 1.")
        }
        estimate_hsq <- FALSE
    }
    n <- rep(NA, ncol(pheno)); names(n) <- colnames(pheno)


    # loop over batches of phenotypes with the same pattern of NAs
    for(batch in seq_along(phe_batches)) {

        # info about batch
        omit <- phe_batches[[batch]]$omit # ind to omit
        phecol <- phe_batches[[batch]]$cols # phenotype columns in batch

        # individuals to keep in this batch
        these2keep <- ind2keep
        if(length(omit)>0) these2keep <- ind2keep[-omit]
        n[phecol] <- length(these2keep)
        if(length(these2keep) <= 2) next # not enough individuals; skip this batch

        # subset the rest
        K <- subset_kinship(kinship, ind=these2keep)
        ac <- addcovar; if(!is.null(ac)) { ac <- ac[these2keep,,drop=FALSE]; ac <- drop_depcols(ac, TRUE, tol) }
        Xc <- Xcovar;   if(!is.null(Xc)) Xc <- Xc[these2keep,,drop=FALSE]
        ic <- intcovar; if(!is.null(ic)) { ic <- ic[these2keep,,drop=FALSE]; ic <- drop_depcols(ic, TRUE, tol) }
        wts <- weights; if(!is.null(wts)) wts <- wts[these2keep]
        ph <- pheno[these2keep,phecol,drop=FALSE]

        # multiply stuff by the weights
        K <- weight_kinship(K, wts)
        ac <- weight_matrix(ac, wts)
        Xc <- weight_matrix(Xc, wts)
        ph <- weight_matrix(ph, wts)

        # eigen decomposition of kinship matrix
        Ke <- decomp_kinship(K, cores=cores)

        # fit LMM for each phenotype, one at a time
        if(estimate_hsq) {
            nullresult <- calc_hsq_clean(Ke=Ke, pheno=ph, addcovar=ac, Xcovar=Xc,
                                         is_x_chr=is_x_chr, weights=wts, reml=reml,
                                         cores=cores, check_boundary=check_boundary, tol=tol)

            hsq[, phecol] <- nullresult$hsq
        }
        else {
            # for the log likelihood, calculate the reml=FALSE version
            loglik <- calc_nullLL_clean(hsq=hsq[,phecol,drop=FALSE], Ke=Ke, pheno=ph,
                                        addcovar=ac, Xcovar=Xc,
                                        is_x_chr=is_x_chr, weights=wts, reml=FALSE,
                                        cores=cores)
            nullresult <- list(hsq=hsq[,phecol,drop=FALSE],
                               loglik=loglik)
        }

        # weighted least squares genome scan, using cluster_lapply across chromosomes
        lod <- scan1_pg_clean(genoprobs, these2keep, Ke, ph, ac, ic, is_x_chr,
                              wts, genoprob_Xcol2drop,
                              nullresult$hsq, nullresult$loglik, reml, cores,
                              intcovar_method, tol)

        result[,phecol] <- lod
    }

    # add attributes
    attr(result, "hsq") <- hsq
    attr(result, "sample_size") <- n
    class(result) <- c("scan1", "matrix")

    result
}


# fit LMM for each of a matrix of phenotypes
# Ke is eigendecomposition of 2*kinship
calc_hsq_clean <-
    function(Ke, pheno, addcovar=NULL, Xcovar=NULL, is_x_chr=FALSE, weights=NULL,
             reml=TRUE, cores=1, check_boundary=FALSE, tol=1e-12)
{
    n <- nrow(pheno)
    nphe <- ncol(pheno)

    # if just one kinship matrix, force it to be a list
    if(!is_kinship_list(Ke)) {
        # X chromosome with special covariates
        if(!is.null(Xcovar) && any(is_x_chr) && any(!is_x_chr)) {
            Ke <- list(Ke, Ke)
            is_x_chr <- c(FALSE, TRUE)
        }
        else { Ke <- list(Ke) }
    }

    # function that does the work
    by_chr_func <-
        function(chr)
        {
            # premultiply phenotypes and covariates by transposed eigenvectors
            y <- Ke[[chr]]$vectors %*% pheno
            intercept <- weights; if(is_null_weights(weights)) intercept <- rep(1,n)
            ac <- cbind(intercept, addcovar)
            if(!is.null(Xcovar) && is_x_chr[chr]) # add Xcovar if necessary
                ac <- drop_depcols(cbind(ac, Xcovar), FALSE, tol)
            logdetXpX = Rcpp_calc_logdetXpX(ac)
            ac <- Ke[[chr]]$vectors %*% ac

            Rcpp_fitLMM_mat(Ke[[chr]]$values, y, ac, reml, check_boundary,
                            logdetXpX, tol)
        }

    # now do the work
    result <- cluster_lapply(cores, seq_along(Ke), by_chr_func)

    # check for problems (if clusters run out of memory, they'll return NULL)
    result_is_null <- vapply(result, is.null, TRUE)
    if(any(result_is_null))
        stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

    # re-arrange results
    hsq <- matrix(unlist(lapply(result, function(a) a$hsq)), byrow=TRUE,
                  ncol=nphe)
    loglik <- matrix(unlist(lapply(result, function(a) a$loglik)), byrow=TRUE,
                  ncol=nphe)

    sigmasq <- matrix(unlist(lapply(result, function(a) a$sigmasq)), byrow=TRUE,
                      ncol=nphe)

    list(hsq=hsq, loglik=loglik, sigmasq=sigmasq)
}

# calculate log likelihood for given hsq and
# for each column of a matrix of phenotypes
# Ke is eigendecomposition of 2*kinship
#
# hsq is a matrix with ncol = ncol(pheno)
calc_nullLL_clean <-
    function(hsq, Ke, pheno, addcovar=NULL, Xcovar=NULL, is_x_chr=FALSE,
             weights=NULL, reml=FALSE, cores=1, tol=1e-12)
{
    n <- nrow(pheno)
    nphe <- ncol(pheno)
    if(ncol(hsq) != nphe) stop("ncol(hsq) != ncol(pheno)")

    # if just one kinship matrix, force it to be a list
    if(!is_kinship_list(Ke)) {
        # X chromosome with special covariates
        if(!is.null(Xcovar) && any(is_x_chr) && any(!is_x_chr)) {
            Ke <- list(Ke, Ke)
            is_x_chr <- c(FALSE, TRUE)
        }
        else { Ke <- list(Ke) }
    }
    if(nrow(hsq) != length(Ke)) stop("nrow(hsq) != no. chr")

    # function that does the work
    by_chr_func <-
        function(chr)
        {
            # premultiply phenotypes and covariates by transposed eigenvectors
            y <- Ke[[chr]]$vectors %*% pheno
            intercept <- weights; if(is_null_weights(weights)) intercept <- rep(1,n)
            ac <- cbind(intercept, addcovar)
            if(!is.null(Xcovar) && is_x_chr[chr]) # add Xcovar if necessary
                ac <- drop_depcols(cbind(ac, Xcovar), FALSE, tol)
            logdetXpX = Rcpp_calc_logdetXpX(ac)
            ac <- Ke[[chr]]$vectors %*% ac

            Rcpp_calcLL_mat(hsq[chr,], Ke[[chr]]$values, y, ac, reml, logdetXpX)
        }

    # now do the work
    result <- cluster_lapply(cores, seq_along(Ke), by_chr_func)

    # check for problems (if clusters run out of memory, they'll return NULL)
    result_is_null <- vapply(result, is.null, TRUE)
    if(any(result_is_null))
        stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

    # re-arrange results
    matrix(unlist(result), byrow=TRUE, ncol=nphe)
}

# perform the LMM scan
# genoprobs is still a big complicated calc_genoprob object
scan1_pg_clean <-
    function(genoprobs, ind2keep, Ke, pheno, addcovar, intcovar, is_x_chr,
             weights, genoprob_Xcol2drop,
             hsq, null_loglik, reml, cores, intcovar_method, tol)
{
    n <- nrow(pheno)
    nphe <- ncol(pheno)

    if(!is.list(Ke[[1]])) {
        loco <- FALSE
        if(nrow(hsq)==1) no_x <- TRUE
        else no_x <- FALSE
    } else loco <- TRUE

    # This creates a batch for every chr, pheno and set of positions.
    # If cores == 1, use all positions on each chr (original code).
    # if cores != 1, then
    #   for each chr, divide positions into contiguous intervals
    #   number of intervals across chr and phenos =~ cores
    npos_by_chr <- dim(genoprobs)[3,]
    # Genoprob names not retained for qtl2fst
    names(npos_by_chr) <- names(genoprobs)
    if(n_cores(cores) == 1) {
      # Add beginning and end pos of each chr to original batches list.
      batches <- list(chr=rep(seq_len(length(genoprobs)), ncol(pheno)),
                      pos1 = rep(1, length(genoprobs) * ncol(pheno)),
                      pos2 = rep(npos_by_chr, ncol(pheno)),
                      phecol=rep(seq_len(ncol(pheno)), each=length(genoprobs)))
    } else {
      # Divide each chr into intervals pos1:pos2 contiguous from 1 to length of chr.
      tmpfn <- function(x) {
        npos <- ceiling(x * ncol(pheno) * length(npos_by_chr) / n_cores(cores))
        pos1 <- seq(1, x, by = npos)
        pos2 <- c(pos1[-1] - 1, x)
        list(pos1 = pos1, pos2 = pos2)
      }
      npos <- lapply(npos_by_chr, tmpfn)
      
      # Collapse pos1 and pos2 into vectors.
      pos1 <- pos2 <- NULL
      for(i in names(npos)) {
        pos1 <- c(pos1, npos[[i]]$pos1)
        pos2 <- c(pos2, npos[[i]]$pos2)
      }
      # npos = number of intervals by chr
      npos <- sapply(npos, function(x) length(x$pos1))
      
      # Batches list
      batches <- list(chr=rep(rep(seq_len(length(genoprobs)),npos), ncol(pheno)),
                      pos1=rep(pos1, ncol(pheno)),
                      pos2=rep(pos2, ncol(pheno)),
                      phecol=rep(seq_len(ncol(pheno)), each=sum(npos)))
    }
    
    # function that does the work
    by_batch_func <-
        function(batch)
        {
            chr <- batches$chr[batch]
            pos1 <- batches$pos1[batch]
            pos2 <- batches$pos2[batch]
            phecol <- batches$phecol[batch]

            if(loco) {
                Kevec <- Ke[[chr]]$vectors
                Keval <- Ke[[chr]]$values
            }
            else {
                Kevec <- Ke$vectors
                Keval <- Ke$values
            }

            # prep phenotype and covariates
            y <- pheno[,phecol,drop=FALSE]
            intercept <- weights; if(is_null_weights(weights)) intercept <- rep(1,n)
            ac <- cbind(intercept, addcovar)
            ic <- intcovar

            # subset the genotype probabilities: drop cols with all 0s, plus the first column
            Xcol2drop <- genoprob_Xcol2drop[[chr]]
            if(length(Xcol2drop) > 0) {
                pr <- genoprobs[[chr]][ind2keep,-Xcol2drop,pos1:pos2,drop=FALSE]
                pr <- pr[,-1,,drop=FALSE]
            }
            else {
                pr <- genoprobs[[chr]][ind2keep,-1,pos1:pos2,drop=FALSE]
            }
            # weight the probabilities
            pr <- weight_array(pr, weights)

            # calculate weights for this chromosome
            if(loco) {
                lmm_wts <- 1/(hsq[chr,phecol]*Keval + (1-hsq[chr,phecol]))
                nullLL <- null_loglik[chr,phecol]
            }
            else {
                if(no_x || !is_x_chr[chr]) {
                    lmm_wts <- 1/(hsq[1,phecol]*Keval + (1-hsq[1,phecol]))
                    nullLL <- null_loglik[1,phecol]
                }
                else {
                    lmm_wts <- 1/(hsq[2,phecol]*Keval + (1-hsq[2,phecol]))
                    nullLL <- null_loglik[2,phecol]
                }
            }
            lmm_wts <- sqrt(lmm_wts)

            if(is.null(ic))
                loglik <- scan_pg_onechr(pr, y, ac, Kevec, lmm_wts, tol)
            else if(intcovar_method=="highmem")
                loglik <- scan_pg_onechr_intcovar_highmem(pr, y, ac, ic, Kevec, lmm_wts, tol)
            else
                loglik <- scan_pg_onechr_intcovar_lowmem(pr, y, ac, ic, Kevec, lmm_wts, tol)
            lod <- (loglik - nullLL)/log(10)
        }

    # now do the work
    lod_list <- cluster_lapply(cores, seq_along(batches$chr), by_batch_func)

    # check for problems (if clusters run out of memory, they'll return NULL)
    result_is_null <- vapply(lod_list, is.null, TRUE)
    if(any(result_is_null))
        stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

    matrix(nrow=sum(npos_by_chr), ncol=ncol(pheno), unlist(lod_list))
}


# dimnames for hsq
# (a bit awkward due to loco or not, and xchr special or not)
hsq_dimnames <-
    function(kinship, Xcovar, is_x_chr, pheno)
{
    if(is_kinship_list(kinship)) {
        rn <- names(kinship)
    }
    else if(!is.null(Xcovar) && any(is_x_chr) + any(!is_x_chr) == 2) {
        rn <- c(names(is_x_chr)[!is_x_chr][1],
                names(is_x_chr)[is_x_chr][1])
    }
    else {
        rn <- names(is_x_chr)[!is_x_chr][1]
    }

    list(rn, colnames(pheno))
}

# maxmarg
#' Find genotypes with maximum marginal probabilities
#'
#' For each individual at each position, find the genotype with the maximum marginal probability.
#'
#' @param probs Genotype probabilities, as calculated from
#' [calc_genoprob()].
#' @param map Map of pseudomarkers in `probs`. Used only if `chr` and `pos` are provided.
#' @param minprob Minimum probability for making a call. If maximum
#' probability is less then this value, give `NA`.
#' @param chr If provided (along with `pos`), consider only the single specified position.
#' @param pos If provided (along with `chr`), consider only the single specified position.
#' @param return_char If TRUE, return genotype names as character strings.
#' @param quiet IF `FALSE`, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#' @param tol Tolerance value; genotypes with probability that are within
#'     `tol` of each other are treated as equivalent.
#'
#' @return If `chr` and `pos` are provided, a vector of
#' genotypes is returned. In this case, `map` is needed.
#'
#' Otherwise, the result is a object like that returned by [viterbi()],
#' A list of two-dimensional arrays of imputed genotypes,
#' individuals x positions. Also includes these attributes:
#' * `crosstype` - The cross type of the input `cross`.
#' * `is_x_chr` - Logical vector indicating whether chromosomes
#'     are to be treated as the X chromosome or not, from input `cross`.
#' * `alleles` - Vector of allele codes, from input
#'     `cross`.
#'
#' @details
#' If multiple genotypes share the maximum probability, one is chosen at random.
#'
#' @seealso [sim_geno()], [viterbi()]
#'
#' @export
#'
#' @examples
#' # load data and calculate genotype probabilities
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' pr <- calc_genoprob(iron, error_prob=0.002)
#'
#' # full set of imputed genotypes
#' ginf <- maxmarg(pr)
#'
#' # imputed genotypes at a fixed position
#' g <- maxmarg(pr, iron$gmap, chr=8, pos=45.5)
#'
#' # return genotype names rather than integers
#' g <- maxmarg(pr, iron$gmap, chr=8, pos=45.5, return_char=TRUE)
maxmarg <-
    function(probs, map=NULL, minprob=0.95, chr=NULL, pos=NULL,
             return_char=FALSE, quiet=TRUE, cores=1, tol=1e-13)
{
    if(is.null(probs)) stop("probs is NULL")

    if(!is_nonneg_number(minprob)) stop("minprob should be a single non-negative number")
    if(!is_nonneg_number(tol)) stop("tol should be a single non-negative number")

    if(!is.null(chr) || !is.null(pos)) {
        if(is.null(chr) || is.null(pos) || is.null(map))
            stop("Provide all of chr, pos, and map, or none of them")
        if(length(chr) > 1 || length(pos) > 1) {
            warning("chr and pos should have length 1; using the first values")
            chr <- chr[1]
            pos <- pos[1]
        }

        # possibly subset the map
        if(length(map) != length(probs) || !all(names(map) == names(probs))) {
            chr <- names(probs)
            if(!all(chr %in% names(map)))
                stop("map doesn't contain all of the necessary chromosomes")
            map <- map[chr]
        }

        chr <- as.character(chr)
        marker <- find_marker(map, chr, pos)

        if(class(probs)[1]=="feather_calc_genoprob")
            probs <- list("1"=probs[,,marker][[1]])
        else
            probs <- list("1"=probs[[chr]][,,marker,drop=FALSE])
        names(probs) <- chr
        class(probs) <- c("calc_genoprob", "list")
    }
    else {
        if(!is.null(map)) # give warning that map was provided but not needed (or used)
            warning("map provided but not used (only needed in chr and pos are provided)")
    }


    if(return_char && is.null(colnames(probs[[1]]))) {
        warning("No genotype names included in probs; returning integers")
        return_char <- FALSE
    }

    dn <- dimnames(probs)
    probs_attr <- attributes(probs)

    # function that does the work
    by_chr_func <- function(chr, return_char) {
        if(!quiet) message(" - Chr ", names(probs)[chr])
        result <- .maxmarg(aperm(probs[[chr]], c(2,3,1)), minprob=minprob, tol=tol) # convert to pos x gen x ind
        if(return_char && !is.null(dn[[2]][[chr]]))
            result[,1:ncol(result)] <- dn[[2]][[chr]][result]
        dimnames(result) <- list(dn[[1]], dn[[3]][[chr]])

        result
    }

    # set up cluster; set quiet=TRUE if multi-core
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # run and combine results
    result <- cluster_lapply(cores, seq_len(length(probs)), by_chr_func, return_char=return_char)
    names(result) <- names(probs)

    # if chr and pos given, return a vector
    if(!is.null(pos)) return(result[[1]][,1])

    # otherwise, return an object like viterbi()
    attr(result, "crosstype") <- probs_attr$crosstype
    attr(result, "alleles") <- probs_attr$alleles
    attr(result, "is_x_chr") <- probs_attr$is_x_chr
    class(result) <- c("viterbi", "list")
    result
}

context("manipulation of phased genotypes")

test_that("subset.phasedgeno works", {

    skip_if(isnt_karl(), "this test only run locally")

    file <- paste0("https://raw.githubusercontent.com/rqtl/",
                   "qtl2data/main/DOex/DOex.zip")
    DOex <- read_cross2(file)
    pr <- calc_genoprob(DOex, DOex$gmap, cores=15)
    m <- maxmarg(pr)
    ph <- guess_phase(DOex, m)

    ph_sub <- ph[1:10,]
    expected <- lapply(ph, function(a) a[1:10,,])
    class(expected) <- c("phasedgeno", "list")
    expect_equal(ph_sub, expected)

    ph_sub <- subset.phasedgeno(ph, chr=2)
    expected <- unclass(ph)["2"]
    class(expected) <- c("phasedgeno", "list")
    expect_equal(ph_sub, expected)

    ph_sub <- subset.phasedgeno(ph, ind=11:20, chr=2)
    expected <- lapply(ph, function(a) a[11:20,,])["2"]
    class(expected) <- c("phasedgeno", "list")
    expect_equal(ph_sub, expected)

    # test rbind
    phA <- ph[1:10,]
    phB <- ph[11:20,]
    expect_equal(rbind(phA, phB), ph[1:20,])

    # test cbind
    phA <- ph[,2]
    phB <- ph[,3]
    expect_equal(cbind(phA, phB), ph[,2:3])

})

context("plot_pxg")

test_that("plot_pxg works", {

    skip_if(isnt_karl(), "plot tests only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    set.seed(62610474)
    geno <- maxmarg(probs, map, chr=16, pos=28.6, return_char=TRUE)

    test_plot_pxg <- function() plot_pxg(geno, log10(iron$pheno[,1]), ylab="log liver")
    expect_doppelganger("plot_pxg", test_plot_pxg)

    test_plot_pxg_se <- function() plot_pxg(geno, log10(iron$pheno[,1]), ylab="log liver", SEmult=2)
    expect_doppelganger("plot_pxg_se", test_plot_pxg_se)

})

context("maxmarg")

test_that("maxmarg works for F2 data", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,8] # only chr 8

    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    pr <- calc_genoprob(iron, map, error_prob=0.01)

    z <- maxmarg(pr, minprob=0)
    g <- z[[1]]
    expected <- apply(pr[[1]], c(1,3), function(a) seq(along=a)[a==max(a)])
    expect_equal(g, expected)

    # single locus
    zz <- maxmarg(pr, map, minprob=0, chr="8", pos=45.5)
    expect_equal(zz, g[,"D8Mit40"])

    # repeat with minprob=0.95
    z <- maxmarg(pr, minprob=0.95)
    g <- z[[1]]
    expected <- apply(pr[[1]], c(1,3), function(a) { b <- max(a); ifelse(b > 0.95, which.max(a), NA) })
    expect_equal(g, expected)

    # single locus
    zz <- maxmarg(pr, map, minprob=0.95, chr="8", pos=45.5)
    expect_equal(zz, g[,"D8Mit40"])

    # repeat with return_value="genotypes"
    zchar <- maxmarg(pr, minprob=0.95, return_char=TRUE)
    gchar <- g;gchar[,1:ncol(g)] <- c("SS", "SB", "BB")[g]
    expect_equal(zchar[[1]], gchar)

    # single locus
    zzchar <- maxmarg(pr, map, minprob=0.95, chr="8", pos=45.5, return_char=TRUE)
    gg <- c("SS", "SB", "BB")[zz]
    names(gg) <- names(zz)
    expect_equal(zzchar, gg)
})

context("count_xo and locate_xo")

test_that("count_xo and locate_xo work for intercross", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:5, c(7,8,"X")]
    pr <- calc_genoprob(iron, error_prob=0.002, map_function="c-f")
    v <- maxmarg(pr)

    n_xo <- count_xo(v)
    expected <- structure(c(2L, 0L, 3L, 2L, 0L, 1L, 0L, 2L, 0L, 1L, 0L, 0L, 0L,
                            0L, 0L), dim = c(5L, 3L),
                          dimnames = list(c("1", "2", "3", "4", "5"), c("7", "8", "X")))
    expect_equal(n_xo, expected)

    pos_xo <- locate_xo(v, iron$gmap)
    pos_xoB <- locate_xo(v, iron$gmap[c("7", "8", "X")])
    expect_equal(pos_xo, pos_xoB)

    midpts <- sapply(iron$gmap[c(7,8,"X")], function(a) setNames((a[-1] + a[-length(a)])/2, NULL))
    expected <- list("7"=list("1"=midpts[[1]][c(4,6)],
                              "2"=numeric(0),
                              "3"=midpts[[1]][c(3,3,6)],
                              "4"=midpts[[1]][c(3,6)],
                              "5"=numeric(0)),
                     "8"=list("1"=midpts[[2]][3],
                              "2"=numeric(0),
                              "3"=midpts[[2]][c(4,4)],
                              "4"=numeric(0),
                              "5"=midpts[[2]][6]),
                     "X"=list("1"=numeric(0),
                              "2"=numeric(0),
                              "3"=numeric(0),
                              "4"=numeric(0),
                              "5"=numeric(0)))
    expect_equal(pos_xo, expected)

    n_xo_derived <- sapply(pos_xo, sapply, length)
    expect_equal(n_xo, n_xo_derived)


    # sim_geno
    set.seed(85309395)
    dr <- sim_geno(iron, n_draws=4, error_prob=0.002, map_function="c-f")
    n_xo <- count_xo(dr)
    expect_equal(dim(n_xo), c(5,3,4))

    expected <- structure(c(2L, 2L, 3L, 2L, 4L, 3L, 0L, 2L, 0L, 3L, 0L, 0L, 0L,
                            0L, 0L, 2L, 3L, 3L, 2L, 1L, 3L, 0L, 2L, 0L, 3L, 0L, 0L, 0L, 0L,
                            0L, 2L, 0L, 3L, 2L, 0L, 3L, 0L, 2L, 0L, 3L, 0L, 1L, 0L, 0L, 1L,
                            2L, 2L, 3L, 2L, 1L, 2L, 0L, 2L, 0L, 3L, 0L, 0L, 0L, 0L, 1L),
                          dim = c(5L, 3L, 4L), dimnames = list(c("1", "2", "3", "4", "5"),
                                                               c("7", "8", "X"), NULL))
    expect_equal(n_xo, expected)

})

#' Create a function to query variants
#'
#' Create a function that will connect to a SQLite database of founder
#' variant information and return a data frame with variants for a
#' selected region.
#'
#' @param dbfile Name of database file
#' @param db Optional database connection (provide one of `file` and `db`).
#' @param table_name Name of table in the database
#' @param chr_field Name of chromosome field
#' @param pos_field Name of position field
#' @param id_field Name of SNP/variant ID field
#' @param sdp_field Name of strain distribution pattern (SDP) field
#' @param filter Additional SQL filter (as a character string)
#'
#' @return Function with three arguments, `chr`, `start`,
#'     and `end`, which returns a data frame with the variants in
#'     that region, with `start` and `end` being in Mbp. The
#'     output should contain at least the columns `chr` and
#'     `pos`, the latter being position in Mbp.
#'
#' @details Note that this function assumes that the database has a
#'     `pos` field that is in basepairs, but the selection uses
#'     `start` and `end` positions in Mbp, and the output
#'     data frame should have `pos` in Mbp.
#'
#' Also note that a SQLite database of variants in the founder strains
#' of the mouse Collaborative Cross is available at figshare:
#' \doi{10.6084/m9.figshare.5280229.v3}
#'
#' @export
#' @importFrom RSQLite SQLite dbConnect dbDisconnect dbGetQuery
#'
#' @examples
#' # create query function by connecting to file
#' dbfile <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
#' query_variants <- create_variant_query_func(dbfile)
#' # query_variants will connect and disconnect each time
#' variants <- query_variants("2", 97.0, 98.0)
#'
#' # create query function to just grab SNPs
#' query_snps <- create_variant_query_func(dbfile, filter="type=='snp'")
#' # query_variants will connect and disconnect each time
#' snps <- query_snps("2", 97.0, 98.0)
#'
#' # connect and disconnect separately
#' library(RSQLite)
#' db <- dbConnect(SQLite(), dbfile)
#' query_variants <- create_variant_query_func(db=db)
#' variants <- query_variants("2", 97.0, 98.0)
#' dbDisconnect(db)

create_variant_query_func <-
    function(dbfile=NULL, db=NULL, table_name="variants",
             chr_field="chr", pos_field="pos",
             id_field="snp_id", sdp_field="sdp",
             filter=NULL)
{
    if(!is.null(db)) {
        if(!is.null(dbfile))
            warning("Provide just one of dbfile or db; using db")

        query_func <- function(chr, start, end) {

            # convert start and end to basepairs
            start <- round(start*1e6)
            end <- round(end*1e6)

            # bits of the query
            chrselect <- paste0(chr_field, " == '", chr, "'")
            startselect <- paste0(pos_field, " >= ", start)
            endselect <- paste0(pos_field, " <= ", end)

            # the full query
            query <- paste0("SELECT * FROM ", table_name, " WHERE ",
                            chrselect, " AND ",
                            startselect, " AND ",
                            endselect)

            if(!is.null(filter) && filter != "")
                query <- paste0(query, " AND (", filter, ")")

            # do the query
            result <- RSQLite::dbGetQuery(db, query)

            # include pos column in Mbp
            result$pos <- result[,pos_field]/1e6

            # name and sdp fields
            result <- swap_colname(result, id_field, "snp")
            result <- swap_colname(result, sdp_field, "sdp")

            result
        }

    }
    else {
        if(is.null(dbfile) || dbfile=="")
            stop("Provide either dbfile or db")

        query_func <- function(chr, start, end)
        {
            if(!file.exists(dbfile))
                stop("File ", dbfile, " doesn't exist")

            db <- RSQLite::dbConnect(RSQLite::SQLite(), dbfile)
            on.exit(RSQLite::dbDisconnect(db)) # disconnect on exit

            # convert start and end to basepairs
            start <- round(start*1e6)
            end <- round(end*1e6)

            # bits of the query
            chrselect <- paste0(chr_field, " == '", chr, "'")
            startselect <- paste0(pos_field, " >= ", start)
            endselect <- paste0(pos_field, " <= ", end)

            # the full query
            query <- paste0("SELECT * FROM ", table_name, " WHERE ",
                            chrselect, " AND ",
                            startselect, " AND ",
                            endselect)

            if(!is.null(filter) && filter != "")
                query <- paste0(query, " AND (", filter, ")")

            # do the query
            result <- RSQLite::dbGetQuery(db, query)

            # include pos column in Mbp
            result$pos <- result[,pos_field]/1e6

            # name and sdp fields
            result <- swap_colname(result, id_field, "snp")
            result <- swap_colname(result, sdp_field, "sdp")

            result
        }
    }

    query_func
}

context("create_variant_query_func")

test_that("create_variant_query_func works", {

    # use file name
    dbfile <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
    qf <- create_variant_query_func(dbfile)

    expected <- structure(list(snp = c("rs213863525", "rs235573572", "rs253240367"),
                               chr = c("2", "2", "2"), pos = c(97.300098, 97.30014, 97.300197),
                               alleles = c("A|T", "T|C", "G|T"), sdp = c(32L, 64L, 64L),
                               ensembl_gene = rep("ENSMUSG00000050587", 3),
                               consequence = rep("ENSMUSG00000050587:intron_variant", 3),
                               A_J = c(1L, 1L, 1L),
                               C57BL_6J = c(1L, 1L, 1L),
                               "129S1_SvImJ" = c(1L, 1L, 1L),
                               NOD_ShiLtJ = c(1L, 1L, 1L),
                               NZO_HlLtJ = c(1L, 1L, 1L),
                               CAST_EiJ = c(2L, 1L, 1L),
                               PWK_PhJ = c(1L, 2L, 2L),
                               WSB_EiJ = c(1L, 1L, 1L),
                               type = c("snp", "snp", "snp")),
                          row.names = c(NA, -3L), class = "data.frame")

    expect_equal(qf(2, 97.3, 97.3002), expected)

    # use db connection
    library(RSQLite)
    db <- dbConnect(SQLite(), dbfile)
    qf2 <- create_variant_query_func(db=db)
    expect_equal(qf2(2, 97.3, 97.3002), expected)
    dbDisconnect(db)

    # include filter
    qf3 <- create_variant_query_func(dbfile, filter="snp_id = 'rs235573572'")
    expected_sub <- expected[2,,drop=FALSE]
    rownames(expected_sub) <- 1L
    expect_equal(qf3(2, 0, 200), expected_sub)

})

# subset sim_geno objects

#' Subsetting imputed genotypes
#'
#' Pull out a specified set of individuals and/or chromosomes from
#' the results of [sim_geno()].
#'
#' @param x Imputed genotypes as output from [sim_geno()].
#' @param ind A vector of individuals: numeric indices, logical
#' values, or character string IDs
#' @param chr A vector of chromosomes: logical values, or character
#' string IDs. Numbers are interpreted as character string IDs.
#' @param ... Ignored.
#'
#' @return An object of class `"sim_geno"`, like the input
#' with the selected individuals and/or chromsomes; see [sim_geno()].
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' \dontshow{grav2 <- grav2[1:8,c(1,2)]}
#' dr <- sim_geno(grav2, n_draws=4)
#' # keep just individuals 1:5, chromosome 2
#' drsub <- dr[1:5,2]
#' # keep just chromosome 2
#' drsub2 <- dr[,2]
subset.sim_geno <-
    function(x, ind=NULL, chr=NULL, ...)
    subset.calc_genoprob(x, ind, chr, ...)

#' @export
#' @rdname subset.sim_geno
`[.sim_geno` <-
    function(x, ind=NULL, chr=NULL)
    subset(x, ind, chr)

#' Plot QTL peak locations
#'
#' Plot QTL peak locations (possibly with intervals) for multiple traits.
#'
#' @param peaks Data frame such as that produced by
#'     [find_peaks()]) containing columns
#'     `chr`, `pos`, `lodindex`, and `lodcolumn`.
#'     May also contain columns `ci_lo` and `ci_hi`, in
#'     which case intervals will be plotted.
#' @param map Marker map, used to get chromosome lengths (and start
#'     and end positions).
#' @param chr Selected chromosomes to plot; a vector of character
#'     strings.
#' @param tick_height Height of tick marks at the peaks (a number between 0 and 1).
#' @param gap Gap between chromosomes. The default is 1% of the total genome length.
#' @param lod_labels If TRUE, plot LOD scores near the intervals. Uses
#'     three hidden graphics parameters, `label_gap` (distance between
#'     CI and LOD text label), `label_left` (vector that indicates
#'     whether the labels should go on the left side; TRUE=on left,
#'     FALSE=on right, NA=put into larger gap on that chromosome), and
#'     `label_cex` that controls the size of these labels
#' @param ... Additional graphics parameters
#'
#' @seealso [find_peaks()], [plot_lodpeaks()]
#'
#' @section Hidden graphics parameters:
#' A number of graphics parameters can be passed via `...`. For
#' example, `bgcolor` to control the background color and
#' `altbgcolor` to control the background color on alternate chromosomes.
#' These are not included as formal parameters in order to avoid
#' cluttering the function definition.
#'
#' @export
#' @importFrom graphics segments abline par axis title box rect
#'
#' @return None.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c(1,2,7,8,9,13,15,16,19)]}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' # find peaks above lod=3.5 (and calculate 1.5-LOD support intervals)
#' peaks <- find_peaks(out, map, threshold=3.5, drop=1.5)
#'
#' plot_peaks(peaks, map)
#'
#' # show LOD scores
#' plot_peaks(peaks, map, lod_labels=TRUE)
#'
#' # show LOD scores, controlling whether they go on the left or right
#' plot_peaks(peaks, map, lod_labels=TRUE,
#'            label_left=c(TRUE, FALSE, TRUE, FALSE, TRUE, TRUE))
plot_peaks <-
    function(peaks, map, chr=NULL, tick_height = 0.3,
             gap=NULL, lod_labels=FALSE, ...)
{
    if(is.null(peaks)) stop("peaks is NULL")
    if(is.null(map)) stop("map is NULL")
    if(is.null(gap)) gap <- sum(chr_lengths(map))/100

    if(!is_nonneg_number(gap)) stop("gap should be a non-negative number")
    if(!is_nonneg_number(tick_height)) stop("tick_height should be a non-negative number")

    if(!is.list(map)) map <- list(" "=map) # if a vector, treat it as a list with no names

    if(!is.null(chr)) {
        chri <- match(chr, names(map))
        if(any(is.na(chri)))
            stop("Chromosomes ", paste(chr[is.na(chri)], collapse=", "), " not found")
        map <- map[chri]
        peaks <- peaks[peaks$chr %in% names(map),,drop=FALSE]
    }

    if(nrow(peaks) == 0)
        stop("There are no peaks on the selected chromosomes")

    plot_peaks_internal <-
        function(peaks, map, tick_height,
                 gap, bgcolor="gray90", altbgcolor="gray85",
                 lwd=2, col="slateblue", xlab=NULL, ylab="",
                 xlim=NULL, ylim=NULL, xaxs="i", yaxs="i",
                 main="", mgp.x=c(2.6, 0.5, 0), mgp.y=c(2.6, 0.5, 0),
                 mgp=NULL, las=1, lend=1, ljoin=1,
                 hlines=NULL, hlines_col="white", hlines_lwd=1, hlines_lty=1,
                 vlines=NULL, vlines_col="white", vlines_lwd=1, vlines_lty=1,
                 sub="", lod_labels=FALSE, label_left=NULL, label_gap=NULL,
                 label_cex=NULL, label_col=NULL, ...)
        {
            dots <- list(...)
            onechr <- (length(map)==1) # single chromosome

            xpos <- map_to_xpos(map, gap)
            chrbound <- map_to_boundaries(map, gap)

            unique_lodindex <- sort(unique(peaks$lodindex))

            if(is.null(ylim))
                ylim <- c(length(unique_lodindex)+0.5, 0.5)

            if(is.null(xlim)) {
                xlim <- range(xpos, na.rm=TRUE)
                if(!onechr) xlim <- xlim + c(-gap/2, gap/2)
            }

            if(is.null(xlab)) {
                if(onechr) {
                    if(names(map) == " ") xlab <- "Position"
                    else xlab <- paste("Chr", names(map), "position")
                }
                else xlab <- "Chromosome"
            }

            # margin parameters
            if(!is.null(mgp)) mgp.x <- mgp.y <- mgp

            # make basic plot
            plot(0,0, xlab="", ylab="", xlim=xlim, ylim=ylim,
                 xaxs=xaxs, yaxs=yaxs, xaxt="n", yaxt="n", type="n",
                 main=main, sub=sub)

            # add background rectangles
            u <- par("usr")
            if(!is.null(bgcolor))
                rect(u[1], u[3], u[2], u[4], col=bgcolor, border=NA)
            if(!is.null(altbgcolor) && !onechr) {
                for(i in seq(2, ncol(chrbound), by=2))
                    rect(chrbound[1,i], u[3], chrbound[2,i], u[4], col=altbgcolor, border=NA)
            }

            # include axis labels?
            if(is.null(dots$xaxt)) dots$xaxt <- par("xaxt")
            if(is.null(dots$yaxt)) dots$yaxt <- par("yaxt")

            # add x axis unless par(xaxt="n")
            if(dots$xaxt != "n") {
                if(onechr) {
                    axis(side=1, at=pretty(xlim), mgp=mgp.x, las=las, tick=FALSE)
                }
                else {
                    loc <- colMeans(chrbound)
                    odd <- seq(1, length(map), by=2)
                    even <- seq(2, length(map), by=2)
                    axis(side=1, at=loc[odd], names(map)[odd],
                         mgp=mgp.x, las=las, tick=FALSE)
                    axis(side=1, at=loc[even], names(map)[even],
                         mgp=mgp.x, las=las, tick=FALSE)
                }
            }

            # add y axis unless par(yaxt="n")
            if(dots$yaxt != "n") {
                axis(side=2, at=seq(along=unique_lodindex),
                     peaks$lodcolumn[match(unique_lodindex, peaks$lodindex)], # <- the y-axis labels
                     mgp=mgp.y, las=las, tick=FALSE)
            }
            dots$xaxt <- dots$yaxt <- NULL # delete those

            # x and y axis labels
            title(xlab=xlab, mgp=mgp.x)
            title(ylab=ylab, mgp=mgp.y)

            # grid lines
            if(onechr && !(length(vlines)==1 && is.na(vlines))) { # if vlines==NA (or mult chr), skip lines
                if(is.null(vlines)) vlines <- pretty(xlim)
                abline(v=vlines, col=vlines_col, lwd=vlines_lwd, lty=vlines_lty)
            }
            if(!(length(hlines)==1 && is.na(hlines))) { # if hlines==NA, skip lines
                if(is.null(hlines)) hlines <- seq_along(unique_lodindex)
                abline(h=hlines, col=hlines_col, lwd=hlines_lwd, lty=hlines_lty)
            }

            y <- match(peaks$lodindex, unique_lodindex)
            x <- xpos_scan1(map, names(map), gap, peaks$chr, peaks$pos)

            segments(x, y-tick_height/2, x, y+tick_height/2,
                     col=col, lwd=lwd, lend=lend, ljoin=ljoin)

            if("ci_lo" %in% names(peaks) && "ci_hi" %in% names(peaks)) {
                x_lo <- xpos_scan1(map, names(map), gap, peaks$chr, peaks$ci_lo)
                x_hi <- xpos_scan1(map, names(map), gap, peaks$chr, peaks$ci_hi)

                segments(x_lo, y, x_hi, y, col=col, lwd=lwd,
                         lend=lend, ljoin=ljoin)
            }

            # add lod labels
            if(lod_labels) {
                add_lod_labels(peaks, map, gap=gap,
                               label_left=label_left, label_gap=label_gap,
                               label_cex=label_cex, label_col=label_col)
            }

        } # end of internal function

    plot_peaks_internal(peaks=peaks, map=map, tick_height=tick_height,
                        gap=gap, lod_labels=lod_labels, ...)

}


# Add LOD scores as text labels
add_lod_labels <-
    function(peaks, map, gap,
             label_left=NULL, label_gap=NULL,
             label_cex=NULL, label_col=NULL)
    {
        # end points of CIs
        if("ci_lo" %in% names(peaks) && "ci_hi" %in% names(peaks)) {
            ci_lo <- peaks$ci_lo
            ci_hi <- peaks$ci_hi
        } else {
            ci_lo <- ci_hi <- peaks$pos
        }

        # x and y positions
        x_lo <- xpos_scan1(map, names(map), gap, peaks$chr, ci_lo)
        x_hi <- xpos_scan1(map, names(map), gap, peaks$chr, ci_hi)

        unique_lodindex <- sort(unique(peaks$lodindex))
        y <- match(peaks$lodindex, unique_lodindex)

        # determine whether they go on the left or right
        chr_start <- vapply(map, min, na.rm=TRUE, 0.0)[peaks$chr]
        chr_end <- vapply(map, max, na.rm=TRUE, 0.0)[peaks$chr]
        left_gap <- ci_lo - chr_start
        right_gap <- chr_end - ci_hi

        if(is.null(label_left)) {
            label_left <- rep(NA, nrow(peaks))
        } else if(length(label_left) != nrow(peaks)) {
            if(length(label_left) < nrow(peaks)) {
                label_left <- rep(label_left, nrow(peaks))
            }
            label_left <- label_left[seq_len(nrow(peaks))]
        }
        if(any(is.na(label_left))) {
            wh <- which(is.na(label_left))
            label_left[wh] <- TRUE
            label_left[wh][right_gap[wh] > left_gap[wh]] <- FALSE
        }

        # expand label_gap to vector
        if(is.null(label_gap)) {
            label_gap <- gap/2
        }
        if(length(label_gap) != nrow(peaks)) {
            if(length(label_gap) < nrow(peaks)) {
                label_gap <- rep(label_gap, nrow(peaks))
            }
            label_gap <- label_gap[seq_len(nrow(peaks))]
        }
        # expand label_cex to vector
        if(is.null(label_cex)) {
            label_cex <- par("cex")
        }
        if(length(label_cex) != nrow(peaks)) {
            if(length(label_cex) < nrow(peaks)) {
                label_cex <- rep(label_cex, nrow(peaks))
            }
            label_cex <- label_cex[seq_len(nrow(peaks))]
        }
        # expand label_col to vector
        if(is.null(label_col)) {
            label_col <- par("fg")
        }
        if(length(label_col) != nrow(peaks)) {
            if(length(label_col) < nrow(peaks)) {
                label_col <- rep(label_col, nrow(peaks))
            }
            label_col <- label_col[seq_len(nrow(peaks))]
        }

        # position for text
        label_gap[label_left] <- -label_gap[label_left]
        x <- ci_hi
        label_adj <- rep(0, nrow(peaks))
        if(any(label_left)) {
            x[label_left] <- ci_lo[label_left]
            label_adj[label_left] <- 1
        }
        x <- xpos_scan1(map, names(map), gap, peaks$chr, x) + label_gap

        # add the text
        for(i in seq_len(nrow(peaks))) {
            text(x[i], y[i], myround(peaks$lod[i], 1),
                 label_adj[i], cex=label_cex[i], col=label_col[i])
        }
    }

#' Create index of equivalent SNPs
#'
#' For a set of SNPs and a map of marker/pseudomarkers, partition the
#' SNPs into groups that are contained within common intervals and
#' have the same strain distribution pattern, and then create an index
#' to a set of distinct SNPs, one per partition.
#'
#' @param map Physical map of markers and pseudomarkers; generally
#'     created from [insert_pseudomarkers()] and
#'     used for a set of genotype probabilities (calculated with
#'     [calc_genoprob()]) that are to be used to
#'     interpolate SNP genotype probabilities (with
#'     [genoprob_to_snpprob()]).
#' @param snpinfo Data frame with SNP information with the following columns:
#' * `chr` - Character string or factor with chromosome
#' * `pos` - Position (in same units as in the `"map"`).
#' * `sdp` - Strain distribution pattern: an integer, between
#'     1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
#'     binary encoding indicates the founder genotypes
#' * `snp` - Character string with SNP identifier (if
#'     missing, the rownames are used).
#' @param tol Tolerance for determining whether a SNP is exactly at a
#' position at which genotype probabilities were already calculated.
#'
#' @return A data frame containing the input `snpinfo` with three
#' added columns: `"index"` (which indicates the groups of
#' equivalent SNPs), `"interval"` (which indicates the map
#' interval containing the SNP, with values starting at 0), and
#' `on_map` (which indicates that the SNP is within
#' `tol` of a position on the map). The rows get reordered,
#' so that they are ordered by chromosome and position, and the
#' values in the `"index"` column are _by chromosome_.
#'
#' @details We split the SNPs by chromosome and identify the intervals
#' in the `map` that contain each. For SNPs within `tol`
#' of a position at which the genotype probabilities were
#' calculated, we take the SNP to be at that position. For each
#' marker position or interval, we then partition the SNPs into
#' groups that have distinct strain distribution patterns, and
#' choose a single index SNP for each partition.
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DO_Recla/recla.zip")
#' recla <- read_cross2(file)
#'
#' # founder genotypes for a set of SNPs
#' snpgeno <- rbind(m1=c(3,1,1,3,1,1,1,1),
#'                  m2=c(1,3,1,3,1,3,1,3),
#'                  m3=c(1,1,1,1,3,3,3,3),
#'                  m4=c(1,3,1,3,1,3,1,3))
#' sdp <- calc_sdp(snpgeno)
#' snpinfo <- data.frame(chr=c("19", "19", "X", "X"),
#'                       pos=c(40.36, 40.53, 110.91, 111.21),
#'                       sdp=sdp,
#'                       snp=c("m1", "m2", "m3", "m4"), stringsAsFactors=FALSE)
#'
#' # update snp info by adding the SNP index column
#' snpinfo <- index_snps(recla$pmap, snpinfo)
#' }
#'
#' @seealso [genoprob_to_snpprob()], [scan1snps()], [find_index_snp()]
#' @export
index_snps <-
    function(map, snpinfo, tol=1e-8)
{
    if(is.null(map)) stop("map is NULL")
    if(is.null(snpinfo)) stop("snpinfo is NULL")
    if(!is_nonneg_number(tol)) stop("tol should be a single non-negative number")

    if(nrow(snpinfo)==0) {
        # no rows; return data frame with no rows
        return( cbind(snpinfo, index=numeric(0), interval=numeric(0), on_map=logical(0)) )
    }

    uchr <- unique(snpinfo$chr)
    chrID <- names(map)
    if(!all(uchr %in% chrID)) {
        mischr <- uchr[!(uchr %in% chrID)]
        stop("Not all chr found in genoprobs: ", paste(mischr, collapse=","))
    }
    # reorder
    uchr <- factor(factor(uchr, levels=chrID))

    # reorder snp info by chromosome and position
    snpinfo <- snpinfo[order(factor(snpinfo$chr, levels=uchr), snpinfo$pos),,drop=FALSE]

    # if more than one chromosome:
    if(length(uchr) > 1) {
        uchr <- as.character(uchr) # convert back to character

        ### split snpinfo by chr, ordered as in genoprobs
        chr <- factor(snpinfo$chr, levels=uchr)
        snpinfo_spl <- split(snpinfo, chr)

        ### loop over chromosomes, using recursion
        for(i in seq_along(uchr))
            snpinfo_spl[[i]] <- index_snps(map, snpinfo_spl[[i]])

        # combine the results
        snpinfo <- snpinfo_spl[[1]]
        for(i in 2:length(uchr))
            snpinfo <- rbind(snpinfo, snpinfo_spl[[i]])

        return(snpinfo)
    }

    # one chromosome

    # make chromosome a character string again
    uchr <- as.character(uchr)

    ### find snps in map
    this_map <- map[[uchr]]
    if(any(is.na(this_map))) stop("Missing values in map on chr ", uchr)
    snploc <- find_intervals(snpinfo$pos, this_map, tol)
    interval <- snploc[,1]
    on_map <- (snploc[,2]==1)

    # drop snps outside of range
    snps2drop <- (interval < 0 | (interval >= length(this_map)-1 & !on_map))
    if(any(snps2drop)) {
        snpinfo <- snpinfo[!snps2drop,,drop=FALSE]
        interval <- interval[!snps2drop]
        on_map <- on_map[!snps2drop]
    }
    if(nrow(snpinfo) == 0)
        stop("No SNPs within range")

    ### find unique (interval, on_map, sdp) patterns
    pat <- paste(interval, on_map, snpinfo$sdp, sep=":")
    upat <- unique(pat)
    ### order unique patterns by position
    pos <- snpinfo$pos[match(upat, pat)]
    upat <- upat[order(pos)]
    snpinfo$index <- match(pat, upat)
    snpinfo$interval <- interval
    snpinfo$on_map <- on_map
    ### first row with each unique pattern
    urow <- seq_len(nrow(snpinfo))[match(upat, pat)]
    ### update index as linking to row number
    snpinfo$index <- urow[snpinfo$index]

    snpinfo
}

context("index snps")

test_that("index_snps works", {

    skip_if(isnt_karl(), "this test only run locally")

    # load example data and calculate genotype probabilities
    file <- paste0("https://raw.githubusercontent.com/rqtl/",
                   "qtl2data/main/DO_Recla/recla.zip")
    recla <- read_cross2(file)

    # founder genotypes for a set of SNPs
    snpgeno <- rbind(m1=c(3,1,1,3,1,1,1,1),
                     m2=c(1,3,1,3,1,3,1,3),
                     m3=c(1,1,1,1,3,3,3,3),
                     m4=c(1,3,1,3,1,3,1,3),
                     m5=c(1,1,1,3,1,3,1,1),
                     m6=c(1,1,3,1,1,1,3,1))
    sdp <- calc_sdp(snpgeno)
    snpinfo <- data.frame(chr=c("19", "19", "X", "X", "19", "X"),
                          pos=c(40.36, 40.53, 110.91, 111.21, 56.480843, 56.480843),
                          sdp=sdp,
                          snp=c("m1", "m2", "m3", "m4", "m5", "m6"), stringsAsFactors=FALSE)

    snpinfo_windex <- index_snps(recla$pmap, snpinfo)
    snpinfo19_windex <- index_snps(recla$pmap, snpinfo[snpinfo$chr=="19",])
    snpinfoX_windex <- index_snps(recla$pmap, snpinfo[snpinfo$chr=="X",])

    expected <- cbind(snpinfo[c(1,2,5,6,3,4),],
                      index=c(1:3, 1:3),
                      interval=c(91,91,132,112,237,237),
                      on_map=c(FALSE, FALSE, TRUE, FALSE, FALSE, FALSE))

    expect_equal(snpinfo_windex, rbind(snpinfo19_windex, snpinfoX_windex))
    expect_equal(snpinfo_windex, expected)
    expect_equal(snpinfo19_windex, expected[1:3,])
    expect_equal(snpinfoX_windex, expected[4:6,])

    # repeat with some redundant snps
    snpinfo_extra <- data.frame(chr=c("19", "19", "X", "X"),
                                pos=c(40.45, 40.37, 56.3, 110.83),
                                sdp=c(170, 9, 68, 170),
                                snp=c("m7", "m8", "m9", "m10"),
                                stringsAsFactors=FALSE)
    rownames(snpinfo_extra) <- snpinfo_extra$snp
    snpinfo <- rbind(snpinfo, snpinfo_extra)

    expected <- cbind(snpinfo[c(1,8,7,2,5,  9,6,10,3,4),],
                      index=c(1,1,3,3,5,  1,1,3,4,3),
                      interval=c(91,91,91,91,132,  112,112,237,237,237),
                      on_map=c(FALSE,FALSE,FALSE,FALSE,TRUE,   FALSE,FALSE,FALSE,FALSE,FALSE))

    snpinfo_windex <- index_snps(recla$pmap, snpinfo)
    snpinfo19_windex <- index_snps(recla$pmap, snpinfo[snpinfo$chr=="19",])
    snpinfoX_windex <- index_snps(recla$pmap, snpinfo[snpinfo$chr=="X",])

    expect_equal(snpinfo_windex, rbind(snpinfo19_windex, snpinfoX_windex))
    expect_equal(snpinfo_windex, expected)
    expect_equal(snpinfo19_windex, expected[1:5,])
    expect_equal(snpinfoX_windex, expected[6:10,])

})

#' Heat map color scale
#'
#' Plot heat map color scale
#'
#' @param x Values that define the color limits (matrix, data frame, or vector); ignored if `zlim` is provided.
#'
#' @param zlim Z-axis limits (a pair of numbers)
#'
#' @param color_scheme Color scheme for the heatmap (ignored if `col` is provided).
#'
#' @param col Optional vector of colors for the heatmap.
#'
#' @param n_colors Number of z-axis colors (ignored if `col` is provided).
#'
#' @param gridlines Color of gridlines (NULL for default choices; NA to skip them).
#'
#' @param swap_axes If TRUE, make the scale horizontal
#'
#' @param ... Additional graphics paramaters.
#'
#' @return None.
#'
#' @seealso [plot_scan1_heatmap()]
#'
#' @importFrom graphics image par axis
#' @importFrom grDevices rainbow heat.colors terrain.colors topo.colors gray
#'
#' @export
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' probs <- calc_genoprob(grav2, map, error_prob=0.002)
#' out <- scan1(probs, grav2$pheno)
#'
#' layout(cbind(1,2), widths=c(4, 1))
#' par(mar=c(4.1, 4.1, 1.1, 1.1))
#' plot_scan1_heatmap(out, map, xlab="Chromosome", ylab="Phenotype", mgp.y=c(2.6, 0.3, 0))
#' par(mar=c(15.1, 3.1, 15.1, 1.1))
#' plot_colorscale(out)

plot_colorscale <-
    function(x=NULL, zlim=NULL,
             color_scheme=c("viridis", "gray", "revgray", "heat", "terrain", "topo", "rainbow"),
             col=NULL, n_colors=256, gridlines=NULL, swap_axes=FALSE, ...)
{
    if(is.null(zlim)) {
        if(!is.null(x)) {
            x <- as.numeric(unlist(x))
            zlim <- c(0, max(x, na.rm=TRUE))
        } else {
            stop("Provide either x or zlim")
        }
    }

    if(is.null(col)) {
        color_scheme <- match.arg(color_scheme)

        i <- seq(0, 1, length=n_colors)

        if(color_scheme=="viridis") {
            col <- viridis_qtl2(n_colors)
            if(is.null(gridlines)) gridlines <- "white"
        } else if(color_scheme=="gray") {
            col <- grDevices::gray(i)
            if(is.null(gridlines)) gridlines <- "violetred"
        } else if(color_scheme=="revgray") {
            col <- rev(grDevices::gray(i))
            if(is.null(gridlines)) gridlines <- "violetred"
        } else if(color_scheme=="heat") {
            col <- grDevices::heat.colors(n_colors)
            if(is.null(gridlines)) gridlines <- "white"
        } else if(color_scheme=="terrain") {
            col <- grDevices::terrain.colors(n_colors)
            if(is.null(gridlines)) gridlines <- "black"
        } else if(color_scheme=="topo") {
            col <- grDevices::topo.colors(n_colors)
            if(is.null(gridlines)) gridlines <- "white"
        } else if(color_scheme=="rainbow") {
            col <- rev(grDevices::rainbow(n_colors, start=0, end=2/3))
            if(is.null(gridlines)) gridlines <- "white"
        } else stop('color_scheme "', color_scheme, '" not allowed')
    } else {
        if(is.null(gridlines)) gridlines <- "white"
    }

    plot_colorscale_internal <-
        function(zlim, col, swap_axes=FALSE,
                 gridlines=NULL, gridlines_lwd=1,
                 mgp=NULL, mgp.x=NULL, mgp.y=NULL,
                 las=1, xaxt="s", yaxt="s",
                 xlab="", ylab="", ...)
    {
        xpos <- 1
        ypos <- seq(zlim[1], zlim[2], length=length(col))
        z <- as.matrix(ypos)

        if(swap_axes) {
            graphics::image(ypos, xpos, z, xaxs="i", yaxs="i", zlim=zlim, xlab="", ylab="", xaxt="n", yaxt="n", col=col)
        } else {
            graphics::image(xpos, ypos, t(z), xaxs="i", yaxs="i", zlim=zlim, xlab="", ylab="", xaxt="n", yaxt="n", col=col)
        }

        if(is.null(mgp.x)) {
            if(!is.null(mgp)) mgp.x <- mgp
            else mgp.x <- c(1.7, 0.3, 0)
        }
        if(is.null(mgp.y)) {
            if(!is.null(mgp)) mgp.y <- mgp
            else mgp.y <- c(2.0, 0.4, 0)
        }

        # lod column axis labels
        if(!swap_axes && yaxt != "n") {
            graphics::axis(side=2, at=pretty(ypos), mgp=mgp.y, tick=FALSE, las=las)
            abline(h=pretty(ypos), col=gridlines, lwd=gridlines_lwd)
        }
        if(swap_axes && xaxt != "n") {
            graphics::axis(side=1, at=pretty(ypos), mgp=mgp.x, tick=FALSE, las=las)
            abline(v=pretty(ypos), col=gridlines, lwd=gridlines_lwd)
        }

        # axis titles
        if(!is.null(xlab) && xlab != "") graphics::title(xlab=xlab, mgp=mgp.x)
        if(!is.null(ylab) && ylab != "") graphics::title(ylab=ylab, mgp=mgp.y)
    }

    plot_colorscale_internal(zlim=zlim, col=col, swap_axes=swap_axes, gridlines=gridlines, ...)

}

#' Combine data from scan1perm objects
#'
#' Row-bind multiple scan1perm objects with the same set of columns
#'
#' @param ... A set of permutation results from
#' [scan1perm()] (objects of class `"scan1perm"`).
#' They must have the same set of columns. If any include
#' autosome/X chromosome-specific permutations, they must all be
#' such.
#'
#' @return The combined row-binded input, as an object of class `"scan1perm"`; see [scan1perm()].
#'
#' @details The aim of this function is to concatenate the results
#' from multiple runs of a permutation test with
#' [scan1perm()], to assist in the case that such
#' permutations are done on multiple processors in parallel.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c("19","X")] # subset to chr 19 and X}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # permutations with genome scan (just 3 replicates, for illustration)
#' operm1 <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar, n_perm=3)
#' operm2 <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar, n_perm=3)
#'
#' operm <- rbind(operm1, operm2)
#'
#' @seealso [cbind.scan1perm()], [scan1perm()], [scan1()]
#'
#' @export
rbind.scan1perm <-
    function(...)
{
    dots <- list(...)

    if(length(dots)==1) return(dots[[1]])

    if(is.matrix(dots[[1]])) {
        if(!all(vapply(dots, is.matrix, TRUE)))
            stop("Inputs cannot be a mixture of X-chr-specific permutations and not")
        n_col <- vapply(dots, ncol, 1)
        if(length(unique(n_col)) != 1)
            stop("Inputs must all have the same number of columns.")
        for(i in seq_along(dots)[-1]) {
            if(!all(colnames(dots[[1]]) == colnames(dots[[i]])))
                warning("Inputs have different column names; using those from the first input.")
        }

        result <- do.call("rbind", lapply(dots, unclass))
        colnames(result) <- colnames(dots[[1]])
        class(result) <- class(dots[[1]])
        return(result)
    }
    else { # perm_Xsp=TRUE
        # check they're all lists of A and X
        if(!all(vapply(dots, is.list, TRUE)))
            stop("Inputs cannot be a mixture of X-chr-specific permutations and not")
        if(!all(vapply(dots, length, 1)==2) || !all(vapply(dots, function(a) all(sort(names(a)) == c("A", "X")), TRUE)))
            stop("Inputs should all be a list with two components")

        # check that the columns are the same
        n_colA <- vapply(dots, function(a) ncol(a$A), 1)
        n_colX <- vapply(dots, function(a) ncol(a$X), 1)
        if(length(unique(c(n_colA, n_colX))) != 1)
            stop("Inputs must all have the same number of columns.")
        for(i in seq_along(dots)[-1]) {
            if(!all(colnames(dots[[1]]$A) == colnames(dots[[i]]$A)) ||
               !all(colnames(dots[[1]]$X) == colnames(dots[[i]]$X)))
                warning("Inputs have different column names; using those from the first input.")
        }

        # check the chr_lengths attributes
        chr_lengths <- lapply(dots, function(a) attr(a, "chr_lengths"))
        if(any(vapply(chr_lengths, is.null, TRUE)))
            stop("Missing chr_lengths attribute from some inputs")
        if(any(vapply(chr_lengths, length, 1) != 2) ||
           any(vapply(chr_lengths, function(a) any(sort(names(a)) != c("A", "X")), TRUE)))
            stop('chr_lengths attributes should be length 2 vectors with names "A" and "X"')
        if(max(vapply(chr_lengths, function(a) max(abs(chr_lengths[[1]] - a)), 1.0)) > 1e-6) {
            stop("chr_lengths attributes differ.")
        }

        # check the is_x_chr attributes of the chr_lengths attributes
        is_x_chr <- lapply(chr_lengths, function(a) attr(a, "is_x_chr"))
        if(!any(vapply(is_x_chr, is.null, TRUE))) { # if some missing; just skip this
            if(any(vapply(is_x_chr, length, 1) != 2) ||
               any(vapply(is_x_chr, function(a) any(sort(names(a)) != c("A", "X")), TRUE)))
                stop('is_x_chr attributes should be length 2 vectors with names "A" and "X"')
            for(i in seq_along(is_x_chr)[-1]) {
                if(is_x_chr[[1]]["A"] != is_x_chr[[i]]["A"] ||
                   is_x_chr[[1]]["X"] != is_x_chr[[i]]["X"])
                    stop("is_x_chr attributes differ")
            }
            is_x_chr <- is_x_chr[[1]]
        }
        else is_x_chr <- NULL # if some missing, just skip it

        # rbind
        A <- do.call("rbind", lapply(dots, function(a) a$A))
        X <- do.call("rbind", lapply(dots, function(a) a$X))

        # add attributes
        colnames(A) <- colnames(dots[[1]]$A)
        colnames(X) <- colnames(dots[[1]]$X)
        result <- list(A=A, X=X)
        attr(chr_lengths[[1]], "is_x_chr") <- is_x_chr
        attr(result, "chr_lengths") <- chr_lengths[[1]]
        class(result) <- class(dots[[1]])

        return(result)
    }

}


#' @export
#' @rdname rbind.scan1perm
c.scan1perm <- rbind.scan1perm


#' Combine columns from multiple scan1 permutation results
#'
#' Column-bind multiple scan1perm objects with the same numbers of rows.
#'
#' @param ... A set of permutation results from
#' [scan1perm()] (objects of class `"scan1perm"`. If
#' different numbers of permutation replicates were used, those
#' columns with fewer replicates are padded with missing values
#' `NA`. However, if any include autosome/X
#' chromosome-specific permutations, they must all be such.
#'
#' @return The combined column-binded input, as an object of class `"scan1perm"`; see [scan1perm()].
#'
#' @details The aim of this function is to concatenate the results
#' from multiple runs of a permutation test with
#' [scan1perm()], generally with different phenotypes
#' and/or methods, to be used in parallel with
#' [rbind.scan1perm()].
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c("19","X")] # subset to chr 19 and X}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # permutations with genome scan (just 3 replicates, for illustration)
#' operm1 <- scan1perm(probs, pheno[,1,drop=FALSE], addcovar=covar, Xcovar=Xcovar, n_perm=3)
#' operm2 <- scan1perm(probs, pheno[,2,drop=FALSE], addcovar=covar, Xcovar=Xcovar, n_perm=3)
#'
#' operm <- cbind(operm1, operm2)
#'
#' @seealso [rbind.scan1perm()], [scan1perm()], [scan1()]
#'
#' @export
cbind.scan1perm <-
    function(...)
{
    dots <- list(...)

    if(length(dots)==1) return(dots[[1]])

    if(is.matrix(dots[[1]])) {
        if(!all(vapply(dots, is.matrix, TRUE)))
            stop("Inputs cannot be a mixture of X-chr-specific permutations and not")

        result <- do.call("cbind_expand_noalign", lapply(dots, unclass))
        class(result) <- class(dots[[1]])
        return(result)
    }
    else { # perm_Xsp=TRUE
        # check they're all lists of A and X
        if(!all(vapply(dots, is.list, TRUE)))
            stop("Inputs cannot be a mixture of X-chr-specific permutations and not")
        if(!all(vapply(dots, length, 1)==2) || !all(vapply(dots, function(a) all(sort(names(a)) == c("A", "X")), TRUE)))
            stop("Inputs should all be a list with two components")

        # check the chr_lengths attributes
        chr_lengths <- lapply(dots, function(a) attr(a, "chr_lengths"))
        if(any(vapply(chr_lengths, is.null, TRUE)))
            stop("Missing chr_lengths attribute from some inputs")
        if(any(vapply(chr_lengths, length, 1) != 2) ||
           any(vapply(chr_lengths, function(a) any(sort(names(a)) != c("A", "X")), TRUE)))
            stop('chr_lengths attributes should be length 2 vectors with names "A" and "X"')
        if(max(vapply(chr_lengths, function(a) max(abs(chr_lengths[[1]] - a)), 1.0)) > 1e-6) {
            stop("chr_lengths attributes differ.")
        }

        # check the is_x_chr attributes of the chr_lengths attributes
        is_x_chr <- lapply(chr_lengths, function(a) attr(a, "is_x_chr"))
        if(!any(vapply(is_x_chr, is.null, TRUE))) { # if some missing; just skip this
            if(any(vapply(is_x_chr, length, 1) != 2) ||
               any(vapply(is_x_chr, function(a) any(sort(names(a)) != c("A", "X")), TRUE)))
                stop('is_x_chr attributes should be length 2 vectors with names "A" and "X"')
            for(i in seq_along(is_x_chr)[-1]) {
                if(is_x_chr[[1]]["A"] != is_x_chr[[i]]["A"] ||
                   is_x_chr[[1]]["X"] != is_x_chr[[i]]["X"])
                    stop("is_x_chr attributes differ")
            }
            is_x_chr <- is_x_chr[[1]]
        }
        else is_x_chr <- NULL # if some missing, just skip it

        # rbind
        A <- do.call("cbind_expand_noalign", lapply(dots, function(a) a$A))
        X <- do.call("cbind_expand_noalign", lapply(dots, function(a) a$X))

        # add attributes
        result <- list(A=A, X=X)
        attr(chr_lengths[[1]], "is_x_chr") <- is_x_chr
        attr(result, "chr_lengths") <- chr_lengths[[1]]
        class(result) <- class(dots[[1]])

        return(result)
    }

}

#' Join genotype imputations for different chromosomes
#'
#' Join multiple genotype imputation objects, as produced by
#' [sim_geno()], for the same individuals but different
#' chromosomes.
#'
#' @param ... Genotype imputation objects as produced by
#' [sim_geno()]. Must have the same set of individuals.
#'
#' @return An object of class `"sim_geno"`, like the input; see [sim_geno()].
#'
#' @seealso [rbind.sim_geno()], [sim_geno()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' drawsA <- sim_geno(grav2[1:5,1:2], map, error_prob=0.002, n_draws=4)
#' drawsB <- sim_geno(grav2[1:5,3:4], map, error_prob=0.002, n_draws=4)
#' draws <- cbind(drawsA, drawsB)
#'
#' @export
cbind.sim_geno <-
    function(...)
    cbind.calc_genoprob(...)

#' Join genotype probabilities for different individuals
#'
#' Join multiple genotype probability objects, as produced by
#' [calc_genoprob()], for the same set of markers and genotypes but for
#' different individuals.
#'
#' @param ... Genotype probability objects as produced by
#' [calc_genoprob()]. Must have the same set of markers and
#' genotypes.
#'
#' @return An object of class `"calc_genoprob"`, like the input; see [calc_genoprob()].
#'
#' @seealso [cbind.calc_genoprob()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' probsA <- calc_genoprob(grav2[1:5,], map, error_prob=0.002)
#' probsB <- calc_genoprob(grav2[6:12,], map, error_prob=0.002)
#' probs <- rbind(probsA, probsB)
#'
#' @export
rbind.calc_genoprob <-
    function(...)
{
    args <- list(...)

    # to rbind: the data
    # to pass through (must match): crosstype, is_x_chr, alleles, alleleprobs

    result <- args[[1]]
    if(length(args) == 1) return(result)

    # check that things match
    other_stuff <- c("crosstype", "is_x_chr", "alleles", "alleleprobs")
    for(i in 2:length(args)) {
        for(obj in other_stuff) {
            if(!is_same(attr(args[[1]], obj), attr(args[[i]], obj)))
                stop("Input objects 1 and ", i, " differ in their ", obj)
        }
    }

    # create space for result
    nind <- vapply(args, function(a) nrow(a[[1]]), 1)
    totind <- sum(nind)
    index <- split(1:totind, rep(seq(along=nind), nind))

    result <- vector("list", length(args[[1]]))
    names(result) <- names(args[[1]])
    for(chr in names(result)) {
        dimn <- dimnames(args[[1]][[chr]])
        dimv <- dim(args[[1]][[chr]])
        result[[chr]] <- array(dim=c(totind, dimv[2], dimv[3]))
        dimnames(result[[chr]]) <- list(paste(1:totind), dimn[[2]], dimn[[3]])
    }

    # paste stuff together
    for(i in 1:length(args)) {
        if(!is_same(names(args[[1]]), names(args[[i]])))
            stop("Input objects 1 and ", i, " have different chromosome names")
        for(chr in names(args[[1]])) {
            dimn1 <- dimnames(args[[1]][[chr]])
            dimni <- dimnames(args[[i]][[chr]])
            if(!is_same(dimn1[-1], dimni[-1]))
                stop("Input objects 1 and ", i, " differ in shape on chromosome ", chr)

            result[[chr]][index[[i]],,] <- args[[i]][[chr]]
            rownames(result[[chr]])[index[[i]]] <- rownames(args[[i]][[chr]])
        }
    }

    # paste in the attributes
    for(obj in other_stuff)
        attr(result, obj) <- attr(args[[1]], obj)
    class(result) <- class(args[[1]])

    result
}

#' Read phenotype data
#'
#' Read phenotype data from a CSV file (and, optionally, phenotype
#' covariate data from a separate CSV file). The CSV files may be
#' contained in zip files, separately or together.
#'
#' @param file Character string with path to the phenotype data file
#' (or a zip file containing both the phenotype and phenotype
#' covariate files).
#' @param phenocovarfile Character string with path to the phenotype
#' covariate file. This can be a separate CSV or zip file; if a zip
#' file, it must contain exactly one CSV file. Alternatively, if the
#' `file` argument indicates a zip file that contains two files
#' (phenotypes and phenotype covariates), then this
#' `phenocovarfile` argument must indicate the base name for the
#' phenotype covariate file.
#' @param sep the field separator character
#' @param na.strings a character vector of strings which are to be
#' interpreted as `NA` values.
#' @param comment.char A character vector of length one containing a
#' single character to denote comments within the CSV files.
#' @param transpose If TRUE, the phenotype data will be transposed. The
#' phenotype covariate information is **never** transposed.
#' @param quiet If `FALSE`, print progress messages.
#'
#' @return Either a matrix of phenotype data, or a list containing
#' `pheno` (phenotype matrix) and `phenocovar` (phenotype
#' covariate matrix).
#'
#' @export
#' @keywords IO
#' @seealso [read_cross2()],
#' sample data files at <https://kbroman.org/qtl2/pages/sampledata.html>
#' and <https://github.com/rqtl/qtl2data>
#'
#' @examples
#' \dontrun{
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/Gough/gough_pheno.csv")
#' phe <- read_pheno(file)
#'
#' phecovfile <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                      "qtl2data/main/Gough/gough_phenocovar.csv")
#' phe_list <- read_pheno(file, phecovfile)
#' }
read_pheno <-
    function(file, phenocovarfile=NULL, sep=",", na.strings=c("-", "NA"),
             comment.char="#", transpose=FALSE, quiet=TRUE)
{
    # handle zip file, which may contain phenocovarfile
    if(grepl("\\.zip$", file)) { # zip file
        dir <- tempdir()
        if(is_web_file(file)) {
            tmpfile <- tempfile()
            if(!quiet) message(" - downloading ", file, "\n       to ", tmpfile)
            utils::download.file(file, tmpfile, quiet=TRUE)
            file <- tmpfile
            on.exit(unlink(tmpfile))
        }

        if(!quiet) message(" - unzipping ", file, "\n       to ", dir)
        file <- path.expand(file)
        stop_if_no_file(file)
        unzipped_files <- utils::unzip(file, exdir=dir)
        basenames <- basename(unzipped_files)

        on.exit({ # clean up when done
            if(!quiet) message(" - cleaning up")
            unlink(unzipped_files)
        }, add=TRUE)

        other_files <- unzipped_files
        if(!is.null(phenocovarfile) && any(basenames == phenocovarfile)) {
            # look for phenocovarfile in the unzipped files
            if(sum(basenames==phenocovarfile) > 1)
                stop("Multiple copies of ", phenocovarfile, " in zip file")
            other_files <- unzipped_files[basenames != phenocovarfile]
            phenocovarfile <- unzipped_files[basenames==phenocovarfile]
        }

        # halt if more than one file left
        if(length(other_files) > 1) {
            stop("Unclear which file to read: ",
                 paste(basename(other_files), collapse=" "))
        }

        # redefine file with new name
        file <- other_files
    }

    # directory containing the data
    file <- path.expand(file)

    # read file
    stop_if_no_file(file)
    if(!quiet) message(" - Reading ", basename(file))
    pheno <- fread_csv_numer(file, sep=sep, na.strings=na.strings,
                             comment.char=comment.char, transpose=transpose)
    pheno <- as.matrix(pheno)

    # read phenocovarfile
    if(!is.null(phenocovarfile)) {
        if(!quiet) message(" - Reading ", basename(phenocovarfile))
        phenocovar <- read_phenocovar(phenocovarfile, sep=sep, na.strings=na.strings,
                                      comment.char=comment.char, quiet=quiet)
        return(list(pheno=pheno, phenocovar=phenocovar))
    }

    pheno
}

# read phenocovar file (to handle case where it's also zipped)
read_phenocovar <-
    function(file, phenocovarfile=NULL, sep=",", na.strings=c("-", "NA"),
             comment.char="#", quiet=TRUE)
{
    # handle zip file, which may contain phenocovarfile
    if(grepl("\\.zip$", file)) { # zip file
        dir <- tempdir()
        if(is_web_file(file)) {
            tmpfile <- tempfile()
            if(!quiet) message(" - downloading ", file, "\n       to ", tmpfile)
            utils::download.file(file, tmpfile, quiet=TRUE)
            file <- tmpfile
            on.exit(unlink(tmpfile))
        }

        if(!quiet) message(" - unzipping ", file, "\n       to ", dir)
        file <- path.expand(file)
        stop_if_no_file(file)
        unzipped_files <- utils::unzip(file, exdir=dir)

        on.exit({ # clean up when done
            if(!quiet) message(" - cleaning up")
            unlink(unzipped_files)
        }, add=TRUE)

        if(length(unzipped_files) > 1) {
            stop("Unclear which file to read: ",
                 paste(basename(unzipped_files), collapse=" "))
        }

        # redefine file with new name
        file <- unzipped_files
    }

    # directory containing the data
    file <- path.expand(file)

    # read file
    stop_if_no_file(file)
    if(!quiet) message(" - Reading ", basename(file))
    result <- fread_csv(file, sep=sep, na.strings=na.strings,
                        comment.char=comment.char, transpose=FALSE)

    convert_numeric_columns(result)
}


convert_numeric_columns <-
    function(df)
{
    # for each column, look if it's numeric and if so convert
    if(ncol(df) > 0) {
        for(i in 1:ncol(df)) {
            x <- df[,i]
            x <- x[!is.na(x)]
            if(all(grepl("^(-)*[0-9]+\\.*[0-9]*$", x))) { # looks numeric
                df[,i] <- as.numeric(df[,i])
            }
        }
    }
    df
}

context("input/output")

test_that("can read grav2 data", {

    zip_file <- system.file("extdata", "grav2.zip", package="qtl2")
    grav2 <- read_cross2(zip_file)

    # check that it contains the same stuff
    expect_equal(sort(names(grav2)),
                 c("alleles", "cross_info", "crosstype", "geno", "gmap", "is_female",
                   "is_x_chr", "pheno", "phenocovar"))

    # check summary
    expected <- structure(list(crosstype = "riself", nind = 162L, nind_geno = 162L,
                               nind_pheno = 162L, nind_gnp = 162L, nchr = 5L,
                               nmar = c(`1` = 26, `2` = 42, `3` = 64, `4` = 35, `5` = 67),
                               npheno = 241L, ncovar = 0, nphenocovar = 1L, totmar = 234),
                          class = c("summary.cross2", "list"))
    expect_equal(summary(grav2), expected)

    # calculate QTL genotype probabilities
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    pr <- calc_genoprob(grav2, map)

})

test_that("can read iron data", {

    zip_file <- system.file("extdata", "iron.zip", package="qtl2")
    iron <- read_cross2(zip_file)

    # check that it contains the same stuff
    expect_equal(sort(names(iron)),
                 c("alleles", "covar", "cross_info", "crosstype", "geno", "gmap",
                   "is_female", "is_x_chr", "pheno", "phenocovar", "pmap"))

    # check summary
    expected <- structure(list(crosstype = "f2", nind = 284L, nind_geno = 284L,
                               nind_pheno = 284L, nind_gnp = 284L, nchr = 20L,
                               nmar = c(`1` = 3, `2` = 5, `3` = 2, `4` = 2, `5` = 2, `6` = 2, `7` = 7, `8` = 8,
                                        `9` = 5, `10` = 2, `11` = 7, `12` = 2, `13` = 2, `14` = 2,
                                        `15` = 2, `16` = 5, `17` = 2, `18` = 2, `19` = 2, X = 2),
                               npheno = 2L, ncovar = 2L, nphenocovar = 1L, totmar = 66),
                          class = c("summary.cross2", "list"))
    expect_equal(summary(iron), expected)


    # calculate QTL genotype probabilities
    map <- insert_pseudomarkers(iron$gmap, step=1)
    pr <- calc_genoprob(iron, map)

})

test_that("read_pheno works", {

    # iron data
    ironfile <- system.file("extdata", "iron.zip", package="qtl2")
    dir <- tempdir()

    # read full data
    iron <- read_cross2(ironfile)

    # unzip data
    unzipped_files <- utils::unzip(ironfile, exdir=dir)

    # names of data files to be used
    phefile <- unzipped_files[grep("_pheno\\.csv$", unzipped_files)]
    phecovfile <- unzipped_files[grep("_phenocovar\\.csv$", unzipped_files)]

    # zip files to be created
    phezipfile <- paste0(phefile, ".zip")
    phecovzipfile <- paste0(phecovfile, ".zip")
    bothzipfile <- paste0(phefile, "_both.zip")

    # clean up
    on.exit(unlink(c(unzipped_files, phezipfile, phecovzipfile, bothzipfile)))

    # create zip files (-j: don't store file name, -q: be quiet)
    zip(phezipfile, phefile, flags="-j -q")
    zip(phecovzipfile, phecovfile, flags="-j -q")
    zip(bothzipfile, c(phefile, phecovfile), flags="-j -q")

    # read pheno as plain file
    expect_equal(read_pheno(phefile), iron$pheno)

    # read pheno as zip file
    expect_equal(read_pheno(phezipfile), iron$pheno)
    unzip(phezipfile, exdir=dirname(phezipfile)) # restore file

    # read pheno + phenocovar each in plain files
    phelist <- read_pheno(phefile, phecovfile)
    expect_equal(phelist$pheno, iron$pheno)
    expect_equal(phelist$phenocovar, iron$phenocovar)

    # read pheno + phenocovar in one zip file
    expect_error(read_pheno(bothzipfile))
    phelist <- read_pheno(bothzipfile, basename(phecovfile))
    expect_equal(phelist$pheno, iron$pheno)
    expect_equal(phelist$phenocovar, iron$phenocovar)
    unzip(bothzipfile, exdir=dirname(bothzipfile)) # restore files

    # read pheno from one zip file and phenocovar from plain file
    phelist <- read_pheno(phezipfile, phecovfile)
    expect_equal(phelist$pheno, iron$pheno)
    expect_equal(phelist$phenocovar, iron$phenocovar)
    unzip(phezipfile, exdir=dirname(phezipfile)) # restore file

    # read pheno from one zip file and phenocovar from another
    phelist <- read_pheno(phezipfile, phecovzipfile)
    expect_equal(phelist$pheno, iron$pheno)
    expect_equal(phelist$phenocovar, iron$phenocovar)
    unzip(phezipfile, exdir=dirname(phezipfile)) # restore file
    unzip(phecovzipfile, exdir=dirname(phecovzipfile)) # restore file

    # read pheno from zip file and phenocovar from plain file
    phelist <- read_pheno(phefile, phecovzipfile)
    expect_equal(phelist$pheno, iron$pheno)
    expect_equal(phelist$phenocovar, iron$phenocovar)
    unzip(phecovzipfile, exdir=dirname(phecovzipfile)) # restore file

})

test_that("Create zip file works", {

    # unzip iron data to a temporary directory
    ironfile <- system.file("extdata", "iron.zip", package="qtl2")
    dir <- tempdir()
    unzipped_files <- utils::unzip(ironfile, exdir=dir)
    on.exit(unlink(unzipped_files)) # clean up

    # zip the files
    zip_datafiles(file.path(dir, "iron.yaml"))
    on.exit(unlink(file.path(dir, "iron.zip")), add=TRUE)

    # file created?
    zipfile <- file.path(dir, "iron.zip")
    expect_true( file.exists(zipfile) )

    # unzip to tmp dir
    tmp_dir <- file.path(dir, "tmp")
    new_unzipped_files <- utils::unzip(zipfile, exdir=tmp_dir)
    on.exit(unlink(tmp_dir, recursive=TRUE), add=TRUE) # clean up

    # sample files as originally?
    ofiles <- sort( basename( unzipped_files ) )
    nfiles <- sort( basename( new_unzipped_files ) )
    expect_equal( ofiles, nfiles )

    # read them?
    x <- read_cross2(file.path(dir, "iron.yaml"))
    y <- read_cross2(file.path(dir, "iron.zip"))
    expect_equal(y, x)

})

#' Plot genotype probabilities for one individual on one chromosome.
#'
#' Plot the genotype probabilities for one individual on one chromosome, as a heat map.
#'
#' @param probs Genotype probabilities (as produced by [calc_genoprob()])
#' or allele dosages (as produced by [genoprob_to_alleleprob()]).
#' @param map Marker map (a list of vectors of marker positions).
#' @param ind Individual to plot, either a numeric index or an ID.
#' @param chr Selected chromosome to plot; a single character string.
#' @param geno Optional vector of genotypes or alleles to be shown
#' (vector of integers or character strings)
#' @param color_scheme Color scheme for the heatmap (ignored if `col` is provided).
#' @param col Optional vector of colors for the heatmap.
#' @param threshold Threshold for genotype probabilities; only genotypes that achieve
#' this value somewhere on the chromosome will be shown.
#' @param swap_axes If TRUE, swap the axes, so that the genotypes are
#' on the x-axis and the chromosome position is on the y-axis.
#' @param ... Additional graphics parameters passed to [graphics::image()].
#'
#' @return None.
#'
#' @section Hidden graphics parameters:
#' A number of graphics parameters can be passed via `...`. For
#' example, `hlines`, `hlines_col`, `hlines_lwd`, and `hlines_lty` to
#' control the horizontal grid lines. (Use `hlines=NA` to avoid
#' plotting horizontal grid lines.) Similarly `vlines`, `vlines_col`,
#' `vlines_lwd`, and `vlines_lty` for vertical grid lines. You can
#' also use many standard graphics parameters like `xlab` and `xlim`.
#' These are not included as formal parameters in order to avoid
#' cluttering the function definition.
#'
#' @seealso [plot_genoprobcomp()], [plot_onegeno()], [plot_geno()]
#'
#' @examples
#' # load data and calculate genotype probabilities
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' iron <- iron[,"2"] # subset to chr 2
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#' pr <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # plot the probabilities for the individual labeled "262"
#' #  (white = 0, black = 1)
#' plot_genoprob(pr, map, ind="262")
#'
#' # change the x-axis label
#' plot_genoprob(pr, map, ind="262", xlab="Position (cM)")
#'
#' # swap the axes so that the chromosome runs vertically
#' plot_genoprob(pr, map, ind="262", swap_axes=TRUE, ylab="Position (cM)")
#'
#' # This is more interesting for a Diversity Outbred mouse example
#' \dontrun{
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' # subset to chr 2 and X and individuals labeled "232" and "256"
#' DOex <- DOex[c("232", "256"), c("2", "X")]
#' pr <- calc_genoprob(DOex, error_prob=0.002)
#' # plot individual "256" on chr 2 (default is to pick first chr in the probs)
#' plot_genoprob(pr, DOex$pmap, ind="256")
#'
#' # omit states that never have probability >= 0.5
#' plot_genoprob(pr, DOex$pmap, ind="256", threshold=0.05)
#'
#' # X chr male 232: just show the AY-HY genotype probabilities
#' plot_genoprob(pr, DOex$pmap, ind="232", chr="X", geno=paste0(LETTERS[1:8], "Y"))
#' # could also indicate genotypes by number
#' plot_genoprob(pr, DOex$pmap, ind="232", chr="X", geno=37:44)
#' # and can use negative indexes
#' plot_genoprob(pr, DOex$pmap, ind="232", chr="X", geno=-(1:36))
#'
#' # X chr female 256: just show the first 36 genotype probabilities
#' plot_genoprob(pr, DOex$pmap, ind="256", chr="X", geno=1:36)
#'
#' # again, can give threshold to omit genotypes whose probabilities never reach that threshold
#' plot_genoprob(pr, DOex$pmap, ind="256", chr="X", geno=1:36, threshold=0.5)
#'
#' # can also look at the allele dosages
#' apr <- genoprob_to_alleleprob(pr)
#' plot_genoprob(apr, DOex$pmap, ind="232")
#' }
#'
#' @export
#' @importFrom graphics image par axis title box
#' @importFrom grDevices gray
plot_genoprob <-
    function(probs, map, ind=1, chr=NULL, geno=NULL,
             color_scheme=c("gray", "viridis"), col=NULL,
             threshold=0, swap_axes=FALSE, ...)
{
    # check inputs
    if(is.null(map)) stop("map is NULL")
    if(is.null(probs)) stop("probs is NULL")
    if(!is_nonneg_number(threshold)) stop("threshold should be a non-negative number")

    if(is.null(chr)) chr <- names(probs)[1]
    if(length(chr) > 1) {
        warning("chr should have length 1; using the first value")
        chr <- chr[1]
    }
    if(!(chr %in% names(probs))) stop("chr ", chr, " not found in probs")
    if(!(chr %in% names(map))) stop("chr ", chr, " not found in map")
    if(length(ind) > 1) {
        warning("ind should have length 1; using the first value")
        ind <- ind[1]
    }

    # pull out selected chromosomes
    probs <- probs[[chr]]
    map <- map[[chr]]
    if(dim(probs)[[3]] != length(map)) stop("Different numbers of positions in probs and map")

    # pull out individual's probs; make it positions x probs
    if(is.character(ind) && !(ind %in% rownames(probs)))
        stop("ind ", ind, " not found in probs")
    if(is.numeric(ind) && (ind < 1 || ind > nrow(probs)))
        stop("ind ", ind, " should be in the range [1, ", nrow(probs), "]")
    probs <- t(probs[ind,,])

    # pull out selected genotypes
    if(!is.null(geno)) {
        if(is.numeric(geno)) {
            if(all(geno < 0)) {
                if(any(geno > -1 | geno < -ncol(probs)))
                    stop("negative geno should be in the range [", -ncol(probs), " , -1]")
                geno <- seq_len(ncol(probs))[geno]
            }
            if(any(geno < 1 | geno > ncol(probs)))
                stop("numeric geno should be in the range [1, ", ncol(probs))
            geno <- colnames(probs)[geno]
        }
        if(!all(geno %in% colnames(probs)))
            stop("Not all geno in probs")
        probs <- probs[,geno,drop=FALSE]
    }

    # drop genotypes that do exceed threshold
    if(threshold > 0) {
        geno_keep <- colSums(probs >= threshold) > 0
        if(sum(geno_keep) == 0) stop("No genotype probabilities exceed the threshold")
        probs <- probs[,geno_keep,drop=FALSE]
    }

    # set up colors
    if(is.null(col)) {
        color_scheme <- match.arg(color_scheme)
        if(color_scheme=="gray") col <- gray((256:0)/256)
        else if(color_scheme=="viridis") col <- viridis_qtl2(256)
    }

    plot_genoprob_internal(probs, map, col=col, swap_axes=swap_axes, ...)

}



# the function that does the work
plot_genoprob_internal <-
    function(probs, map, col=NULL, swap_axes=FALSE,
             zlim=c(0,1), xlab=NULL, ylab=NULL, las=NULL,
             hlines=NULL, hlines_col="#B3B3B370", hlines_lwd=1, hlines_lty=1,
             vlines=NULL, vlines_col="#B3B3B370", vlines_lwd=1, vlines_lty=1,
             mgp.x=c(2.6,0.5,0), mgp.y=c(2.6,0.5,0), mgp=NULL,
             ...)
{
    dots <- list(...)
    if(!is.null(mgp)) mgp.x <- mgp.y <- mgp
    if(is.null(dots$xaxt)) dots$xaxt <- par("xaxt")
    if(is.null(dots$yaxt)) dots$yaxt <- par("yaxt")

    # separate positions if necessary
    tol <- 1e-6
    if(any(diff(map) < tol))
        map <- map + seq(0, tol, length.out=length(map))

    if(swap_axes) {
        probs <- t(probs)

        x <- seq_len(nrow(probs))
        y <- map

        if(is.null(xlab)) xlab <- ""
        if(is.null(ylab)) ylab <- "Position"

        ytick <- pretty(map)
        yticklab <- NULL

        xtick <- x
        xticklab <- rownames(probs)

        if(is.null(hlines)) hlines <- pretty(map, n=10)
        if(is.null(vlines)) vlines <- seq_len(nrow(probs))

        if(is.null(las)) las <- 2
    } else {
        # reverse order of genotypes, so that the first appears at the top
        probs <- probs[,ncol(probs):1,drop=FALSE]

        x <- map
        y <- seq_len(ncol(probs))

        if(is.null(xlab)) xlab <- "Position"
        if(is.null(ylab)) ylab <- ""

        xtick <- pretty(map)
        xticklab <- NULL

        ytick <- y
        yticklab <- colnames(probs)

        if(is.null(hlines)) hlines <- seq_len(nrow(probs))
        if(is.null(vlines)) vlines <- pretty(map, n=10)

        if(is.null(las)) las <- 1
    }

    image(x, y, probs,
          ylab="", yaxt="n", xlab="", xaxt="n",
          las=1, zlim=zlim, col=col, ...)

    if(dots$xaxt != "n")
        axis(side=1, at=xtick, labels=xticklab, las=las, mgp=mgp.x, tick=FALSE)
    if(dots$yaxt != "n")
        axis(side=2, at=ytick, labels=yticklab, las=las, mgp=mgp.y, tick=FALSE)


    # add grid lines
    if(!(length(hlines)==1 && is.na(hlines)))
        abline(h=hlines, lty=hlines_lty, lwd=hlines_lwd, col=hlines_col)
    if(!(length(vlines)==1 && is.na(vlines)))
        abline(v=vlines, lty=vlines_lty, lwd=vlines_lwd, col=vlines_col)

    title(xlab=xlab, mgp=mgp.x)
    title(ylab=ylab, mgp=mgp.y)

    # add outer box
    box()
}


#' @export
#' @rdname plot_genoprob
#' @param x Genotype probabilities (as produced by
#' [calc_genoprob()]) or allele dosages (as produced by
#' [genoprob_to_alleleprob()]). (For the S3 type plot
#' function, this has to be called `x`.)
plot.calc_genoprob <- function(x, ...) plot_genoprob(x, ...)

context("plot_genoprob")

test_that("plot_genoprob works", {

    skip_if(isnt_karl(), "plot tests only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[c("116", "232"),2]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    test_plot_genoprob <- function() {
        par(mfrow=c(2,1))
        plot_genoprob(probs, map, ind="116", main="116")
        plot_genoprob(probs, map, ind="232", main="232")
    }

    expect_doppelganger("plot_genoprob", test_plot_genoprob)

})

#' Scale kinship matrix
#'
#' Scale kinship matrix to be like a correlation matrix.
#'
#' @param kinship A kinship matrix, or a list of such in the case of
#'     the "leave one chromosome out" method, as calculated by
#'     [calc_kinship()].
#'
#' @return A matrix or list of matrices, as with the input, but with
#'     the matrices scaled to be like correlation matrices.
#'
#' @details We take
#' \ifelse{html}{\out{<em>c<sub>ij</sub> = k<sub>ij</sub> / &radic;(k<sub>ii</sub> k<sub>jj</sub>)</em>}}{\eqn{c_{ij} = k_{ij} / \sqrt{k_{ii} k_{jj}}}{c_ij = k_ij / sqrt(k_ii k_jj)}}
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' probs <- calc_genoprob(grav2, map, error_prob=0.002)
#' K <- calc_kinship(probs)
#' Ka <- scale_kinship(K)

scale_kinship <-
    function(kinship)
{
    if(is_kinship_decomposed(kinship))
        stop("scale_kinship cannot be applied, because kinship has been decomposed")

    k_attr <- attributes(kinship)

    if(is.list(kinship))
        result <- lapply(kinship, scale_kinship)
    else {
        d <- diag(kinship)
        if(any(d <= 0)) stop("Some diagonal elements are <= 0")
        d <- sqrt(d)
        result <- t(t(kinship/d) / d)
    }

    for(a in names(k_attr)) attr(result, a) <- k_attr[[a]]

    result
}

context("Scale kinship matrix")

test_that("scale_kinship works for RIL", {

    set.seed(49265251)

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2, map, error_prob=0.002)
    sim <- calc_kinship(probs)

    scaled <- scale_kinship(sim)

    # row and colnames okay
    expect_equal(dimnames(scaled), dimnames(sim))

    # check some values
    expect_equal(diag(scaled), setNames(rep(1, nrow(sim)), colnames(sim)))

    d <- sqrt(diag(sim))
    n <- 10
    pairs <- matrix(sample(ncol(sim), 2*n, replace=TRUE), ncol=2)
    for(i in 1:n)
        expect_equal(scaled[pairs[i,1], pairs[i,2]],
                     setNames(sim[pairs[i,1], pairs[i,2]] / (d[pairs[i,1]] * d[pairs[i,2]]), NULL) )

    # test with "loco" method
    sim_loco <- calc_kinship(probs, "loco")
    scaled_loco <- scale_kinship(sim_loco)

    expected <- sim_loco
    for(i in seq_along(sim_loco))
        expected[[i]] <- scale_kinship(sim_loco[[i]])
    expect_equal(scaled_loco, expected)

})

#' @keywords internal
#' @importFrom Rcpp sourceCpp
#' @useDynLib qtl2, .registration=TRUE
#'
#' @section Vignettes:
#' * [user guide](https://kbroman.org/qtl2/assets/vignettes/user_guide.html)
#' * [categorized list of functions in R/qtl2](https://kbroman.org/qtl2/pages/rqtl2_functions.html)
#' * [input file formats](https://kbroman.org/qtl2/assets/vignettes/input_files.html)
#' * [using qtl2fst for on-disk genotype probabilities](https://kbroman.org/qtl2/assets/vignettes/qtl2fst.html)
#' * [preparing DO mouse data for R/qtl2](https://kbroman.org/qtl2/pages/prep_do_data.html)
#' * [genotype diagnostics for diversity outbred mice](https://kbroman.org/qtl2/assets/vignettes/do_diagnostics.html)
#' * [identifying sample mix-ups in diversity outbred mice](https://kbroman.org/qtl2/assets/vignettes/do_mixups.html)
#' * [differences between R/qtl and R/qtl2](https://kbroman.org/qtl2/assets/vignettes/rqtl_diff.html)
#' * [developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html)
#' * [Tutorial on R/qtl2](https://smcclatchy.github.io/mapping/) by
#'   [Susan McClatchy](https://github.com/smcclatchy) and [Dan Gatti](https://github.com/dmgatti)
#' * [scan1gen examples](https://kbroman.org/qtl2/assets/vignettes/scan1gen.html)
#'
#' @section Related packages:
#' * [qtl2convert](https://github.com/rqtl/qtl2convert), for converting
#'   data among the R/qtl2, DOQTL, and R/qtl formats
#' * [qtl2fst](https://github.com/rqtl/qtl2fst), for storing genotype
#'   probabilities on disk
#' * [qtl2ggplot](https://github.com/byandell-sysgen/qtl2ggplot),
#'   for [ggplot2](https://ggplot2.tidyverse.org/)-based data visualizations
#'
#' @section Google groups:
#' * For discussion/questions about R/qtl2, join the [rqtl2-disc google group](https://groups.google.com/forum/#!forum/rqtl2-disc)
#' * Or join [rqtl-announce](https://groups.google.com/forum/#!forum/rqtl-announce) for announcements about R/qtl and R/qtl2
#'
#'
"_PACKAGE"

#' Add thresholds to genome scan plot
#'
#' Draw line segments at significance thresholds for a genome scan plot
#'
#' @param map Marker map used in the genome scan plot
#' @param thresholdA Autosomal threshold. Numeric or a list. (If a
#' list, the `"A"` component is taken to be `thresholdA` and the
#' `"X"` component is taken to be `thresholdX`.)
#' @param thresholdX X chromosome threshold (if missing, assumed to be the same as `thresholdA`)
#' @param chr Chromosomes that were included in the plot
#' @param gap Gap between chromosomes in the plot. Default is 1% of the total genome length.
#' @param ... Additional arguments passed to [graphics::segments()]
#'
#' @return None.
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c(3,15,"X")]}
#' map <- insert_pseudomarkers(iron$gmap, step=5)
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#' Xcovar <- get_x_covar(iron)
#' out <- scan1(probs, iron$pheno[,1], Xcovar=Xcovar)
#' # run just 3 permutations, as a fast illustration
#' operm <- scan1perm(probs, iron$pheno[,1], addcovar=Xcovar,
#'                    n_perm=3, perm_Xsp=TRUE, chr_lengths=chr_lengths(map))
#'
#' plot(out, map)
#' add_threshold(map, summary(operm), col="violetred", lty=2)
#' @importFrom graphics segments
#' @importFrom stats setNames
#' @export
add_threshold <-
    function(map, thresholdA, thresholdX=NULL, chr=NULL, gap=NULL, ...)
{
    if(is.null(gap)) gap <- sum(chr_lengths(map))/100

    if(is.list(thresholdA)) {
        thresholdX <- thresholdA$X
        thresholdA <- thresholdA$A
    }
    if(is.null(thresholdX)) thresholdX <- thresholdA

    # which chromosomes are X chr?
    is_x_chr <- attr(map, "is_x_chr")
    if(is.null(is_x_chr)) {
        is_x_chr <- setNames(names(map) %in% c("X", "x"), names(map))
    }

    # subset map and is_x_chr
    if(!is.null(chr)) {
        map <- map[chr]
        is_x_chr <- is_x_chr[chr]
    }

    if(length(map) == 1) { # one chromosome
        abline(h=ifelse(is_x_chr, thresholdX, thresholdA), ...)
    }
    else {
        start <- vapply(map, min, na.rm=TRUE, 0) - gap/2
        end <- vapply(map, max, na.rm=TRUE, 0) + gap/2

        for(chr in names(map)) {
            h <- ifelse(is_x_chr[chr], thresholdX, thresholdA)
            segments(xpos_scan1(map, gap=gap, thechr=chr, thepos=start[chr]), h,
                     xpos_scan1(map, gap=gap, thechr=chr, thepos=end[chr]), h, ...)
        }
    }
}

#' Join phased geno results for different individuals
#'
#' Join phased genotype objects, as produced by [guess_phase()],
#' for the same set of markers but for different individuals.
#'
#' @param ... Imputed genotype objects as produced by
#' [guess_phase()]. Must have the same set of markers.
#'
#' @return An object of class `"phasedgeno"`, like the input; see [guess_phase()].
#'
#' @seealso [cbind.phasedgeno()], [subset.phasedgeno()], [guess_phase()]
#'
#' @examples
#' \dontrun{
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' pr <- calc_genoprob(DOex, error_prob=0.002)
#' m <- maxmarg(pr)
#' phA <- guess_phase(DOex[1:10,], m[1:10,])
#' phB <- guess_phase(DOex[11:20,], m[11:20,])
#' ph <- rbind(phA, phB)
#' }
#'
#' @export
rbind.phasedgeno <-
    rbind.calc_genoprob

# two-panel plot with both snp asso and genes
# (for a single chromosome)
#
# calls plot_snpasso and plot_genes
# internal function that is called by plot_snpasso
plot_snpasso_and_genes <-
    function(scan1output, snpinfo, show_all_snps=TRUE,
             drop_hilit=NA, col_hilit="violetred", col="darkslateblue",
             gap=NULL, minlod=0,
             genes, minrow=4, padding=0.2,
             colors=c("black", "red3", "green4", "blue3", "orange"),
             scale_pos=1, start_field="start", stop_field="stop",
             strand_field="strand", name_field="Name",
             top_panel_prop=0.65, panel_prop=NULL, xlim=NULL, xaxt="s",
             xlab=NULL, main="", sub="", ...)
{
    if(is.null(panel_prop)) panel_prop <- c(top_panel_prop, 1-top_panel_prop)

    # 2 x 1 panels; adjust margins
    old_mfrow <- par("mfrow")
    old_mar <- par("mar")
    on.exit(par(mfrow=old_mfrow, mar=old_mar))
    layout(rbind(1,2), heights=panel_prop)
    top_mar <- bottom_mar <- old_mar
    top_mar[1] <- 0.1
    bottom_mar[3] <- 0.1


    if(is.null(xlim)) xlim <- range(snpinfo$pos)

    if(is.null(xlab)) {
        if(length(unique(snpinfo$chr))==1)
            xlab <- paste("Chr", snpinfo$chr[1], "position (Mbp)")
        else
            xlab <- "Position (Mbp)"
    }

    par(mar=top_mar)
    plot_snpasso(scan1output, snpinfo, show_all_snps=show_all_snps,
                 drop_hilit=drop_hilit, col_hilit=col_hilit, col=col,
                 gap=gap, minlod=minlod, xlim=xlim, xaxt="n", xlab="",
                 main=main, ...)

    par(mar=bottom_mar)
    plot_genes(genes, minrow=minrow, padding=padding, colors=colors,
               scale_pos=scale_pos, start_field=start_field, stop_field=stop_field,
               strand_field=strand_field, name_field=name_field, xlim=xlim,
               xaxt=xaxt, xlab=xlab, sub=sub, ...)

}

#' Read a csv file that has numeric columns
#'
#' Read a csv file via [data.table::fread()] using a
#' particular set of options, including the ability to transpose the
#' result. This version assumes that the contents other than the first
#' column and the header row are strictly numeric.
#'
#' @param filename Name of input file
#' @param sep Field separator
#' @param na.strings Missing value codes
#' @param comment.char Comment character; rest of line after this character is ignored
#' @param transpose If TRUE, transpose the result
#' @param rownames_included If TRUE, the first column is taken to be row names.
#'
#' @return Data frame
#'
#' @details Initial two lines can contain comments with number of rows
#' and columns. Number of columns includes an ID column; number of
#' rows does not include the header row.
#'
#' The first column is taken to be a set of row names
#'
#' @seealso [fread_csv()]
#'
#' @importFrom data.table fread
#' @export
#'
#' @examples
#' \dontrun{mydata <- fread_csv_numer("myfile.csv", transpose=TRUE)}
fread_csv_numer <-
    function(filename, sep=",", na.strings=c("NA", "-"), comment.char="#", transpose=FALSE,
             rownames_included=TRUE)
{
    # read header and extract expected number of rows and columns
    # (number of columns includes ID column; number of rows does *not* include header row)
    header <- read_header(filename, comment.char=comment.char)
    expected_dim <- extract_dim_from_header(header)

    # determine expected columns
    expected_ncol <- expected_dim[2]
    if(is.na(expected_ncol)) { # no header about no. columns
        x <- data.table::fread(filename, na.strings=na.strings, sep=sep, header=TRUE,
                               verbose=FALSE, showProgress=FALSE, data.table=FALSE,
                               nrow=1, skip=length(header))
        expected_ncol <- ncol(x)
    }
    if(expected_ncol == 1) colClasses <- "character"
    else colClasses <- c("character", rep("numeric", expected_ncol-1))

    # read the data
    x <- data.table::fread(filename, na.strings=na.strings, sep=sep, header=TRUE,
                           verbose=FALSE, showProgress=FALSE, data.table=FALSE,
                           colClasses=colClasses, skip=length(header))

    # check that number of rows and columns match expected from header
    for(i in 1:2) {
        labels <- c("rows", "columns")
        if(!is.na(expected_dim[i])) { # nrows given
            if(dim(x)[i] != expected_dim[i])
                stop('In file "', filename, '", no. ', labels[i],
                     ' (', dim(x)[i], ') != expected (', expected_dim[i], ')')
        }
    }

    # move first column to row names
    if(rownames_included)
        x <- firstcol2rownames(x, filename)

    # transpose if requested
    if(transpose)
        x <- as.data.frame(t(x), stringAsFactors=FALSE)

    x
}

# find_markerpos
#' Find positions of markers
#'
#' Find positions of markers within a cross object
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' Can also be a map (as a list of vectors of marker positions).
#' @param markers A vector of marker names.
#' @param na.rm If TRUE, don't include not-found markers in the
#' results (but issue a warning if some markers weren't found). If
#' FALSE, include those markers with `NA` for chr and position.
#'
#' @return A data frame with chromosome and genetic and physical
#' positions (in columns `"gmap"` and `"pmap"`), with markers as
#' row names. If the input `cross` is not a cross2 object but
#' rather a map, the output contains `chr` and `pos`.
#'
#' @seealso [find_marker()]
#' @export
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#'
#' # find markers
#' find_markerpos(iron, c("D8Mit294", "D11Mit101"))
find_markerpos <-
    function(cross, markers, na.rm=TRUE)
{
    output_generic <- FALSE
    if(!is.cross2(cross)) { # assume input is a map
        output_generic <- TRUE

        # make the map look like a cross2 object
        cross <- list(geno=cross, pmap=cross)
        cross$geno <- lapply(cross$geno, function(a) t(as.matrix(a)))
    }

    if(!any(c("gmap", "pmap") %in% names(cross)))
        stop('cross contains neither "gmap" nor "pmap"')

    chr <- rep(NA, length(markers))
    gmap <- rep(NA, length(markers))
    pmap <- rep(NA, length(markers))

    for(i in seq(along=cross$geno)) {
        mn <- colnames(cross$geno[[i]])
        on_this_chr <- (markers %in% mn)
        if(any(on_this_chr)) {
            chr[on_this_chr] <- names(cross$geno)[i]

            if("gmap" %in% names(cross))
                gmap[on_this_chr] <- cross$gmap[[i]][markers[on_this_chr]]

            if("pmap" %in% names(cross))
                pmap[on_this_chr] <- cross$pmap[[i]][markers[on_this_chr]]
        }
    }

    if(na.rm) {
        if(any(is.na(chr)))
            warning("Some markers not found: ", paste(markers[is.na(chr)], collapse=", "))
        gmap <- gmap[!is.na(chr)]
        pmap <- pmap[!is.na(chr)]
        markers <- markers[!is.na(chr)]
        chr <- chr[!is.na(chr)]
    }

    result <- data.frame(chr=chr, gmap=gmap, pmap=pmap,
                         stringsAsFactors=FALSE)
    rownames(result) <- markers
    if(!("pmap" %in% names(cross)))
        result <- result[,-3,drop=FALSE]
    if(!("gmap" %in% names(cross)))
        result <- result[,-2,drop=FALSE]

    if(output_generic) colnames(result)[2] <- "pos"

    result
}

context("find_marker and find_markerpos")

test_that("find_marker works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    # find markers by their genetic map positions
    expect_equal(find_marker(iron$gmap, c(8, 11), c(37.7, 56.9)), c("D8Mit294", "D11Mit101"))

    # find markers by their physical map positions (two markers on chr 7)
    expect_equal(find_marker(iron$pmap, 7, c(44.2, 108.9)), c("D7Nds5", "D7Mit17"))

    # find markers in an interval
    expect_equal(find_marker(iron$pmap, 16, interval=c(35, 80)), c("D16Mit4", "D16Mit30", "D16Mit19"))

    # error if you give both pos and interval
    expect_error(find_marker(iron$gmap, 17, 20, c(5,6)))

    # error if chr and pos are different lengths > 1
    expect_error(find_marker(iron$gmap, c(2,17), c(20, 40, 50)))

    # error if length(interval) != 2
    expect_error(find_marker(iron$pmap, 18, interval=20))
    expect_error(find_marker(iron$pmap, 18, interval=c(20, 40, 50)))

    # error if interval provided and length(chr) != 1
    expect_error(find_marker(iron$gmap, c(18, 19), interval=c(20, 30)))

    # works if input map is a data frame
    map <- data.frame(snp_id=unlist(lapply(iron$pmap, names)),
                      chr=rep(names(iron$pmap), sapply(iron$pmap, length)),
                      pos=unlist(iron$pmap))
    expect_equal(find_marker(map, c(8,11), c(82, 112)), c("D8Mit294", "D11Mit101"))
    expect_equal(find_marker(map, 7, c(44.2, 108.9)), c("D7Nds5", "D7Mit17"))
    expect_equal(find_marker(map, 16, interval=c(35, 80)), c("D16Mit4", "D16Mit30", "D16Mit19"))
    expect_error(find_marker(map, 17, 20, c(5,6)))

})

test_that("find_markerpos works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    expected <- data.frame(chr=c("8", "11"),
                           gmap=c(39.1, 56.9),
                           pmap=c(82.491410, 112.653628),
                           stringsAsFactors=FALSE)
    rownames(expected) <- c("D8Mit294", "D11Mit101")
    expect_equal(find_markerpos(iron, c("D8Mit294", "D11Mit101")), expected)

    # not-found marker with na.rm=TRUE
    expect_equal(suppressWarnings(find_markerpos(iron, c("D8Mit294", "blah", "D11Mit101"))), expected)
    expect_warning(find_markerpos(iron, c("D8Mit294", "blah", "D11Mit101")))

    # not-found marker with na.rm=FALSE
    expected <- rbind(expected[1,], "blah"=data.frame(chr=NA, gmap=NA, pmap=NA), expected[2,])
    expect_equal(find_markerpos(iron, c("D8Mit294", "blah", "D11Mit101"), na.rm=FALSE), expected)

    # one marker
    expect_equal(find_markerpos(iron, "D8Mit294"), expected[1,])

    # no gmap
    iron_nogmap <- iron
    iron_nogmap$gmap <- NULL
    expect_equal(find_markerpos(iron_nogmap, c("D8Mit294", "D11Mit101")), expected[c(1,3),-2])
    expect_equal(find_markerpos(iron_nogmap, "D11Mit101"), expected[3,-2])

    # no pmap
    iron_nopmap <- iron
    iron_nopmap$pmap <- NULL
    expect_equal(find_markerpos(iron_nopmap, c("D8Mit294", "D11Mit101")), expected[c(1,3),-3])
    expect_equal(find_markerpos(iron_nopmap, "D11Mit101"), expected[3,-3])

    # no map at all
    iron_nomap <- iron_nopmap
    iron_nomap$gmap <- NULL
    expect_error(find_markerpos(iron_nomap, "D11Mit101"))

    # input as a map
    expected2 <- expected[-2,-2,drop=FALSE]
    colnames(expected2)[2] <- "pos"
    expect_equal(find_markerpos(iron$pmap, c("D8Mit294", "D11Mit101")), expected2)
    expected3 <- expected[-2,-3,drop=FALSE]
    colnames(expected3)[2] <- "pos"
    expect_equal(find_markerpos(iron$gmap, c("D8Mit294", "D11Mit101")), expected3)

})

#' Find IBD segments for a set of strains
#'
#' Find IBD segments (regions with a lot of shared SNP genotypes) for
#' a set of strains
#'
#' @param geno List of matrices of founder genotypes. The matrices
#'     correspond to the genotypes on chromosomes and are arrayed as
#'     founders x markers.
#' @param map List of vectors of marker positions
#' @param min_lod Threshold for minimum LOD score for a segment
#' @param error_prob Genotyping error/mutation probability
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A data frame whose rows are IBD segments and whose columns
#'     are:
#' * Strain 1
#' * Strain 2
#' * Chromosome
#' * Left marker
#' * Right marker
#' * Left position
#' * Right position
#' * Left marker index
#' * Right marker index
#' * Interval length
#' * Number of markers
#' * Number of mismatches
#' * LOD score
#'
#' @details For each strain pair on each chromosome, we consider all
#' marker intervals and calculate a LOD score comparing the two
#' hypotheses: that the strains are IBD in the interval, vs. that
#' they are not. We assume that the two strains are homozygous at
#' all markers, and use the model from Broman and Weber (1999),
#' which assumes linkage equilibrium between markers and uses a
#' simple model for genotype frequencies in the presence of genotyping
#' errors or mutations.
#'
#' Note that inference of IBD segments is heavily dependent on how
#' SNPs were chosen to be genotyped. (For example, were the SNPs ascertained
#' based on their polymorphism between a particular strain pair?)
#'
#' @references
#' Broman KW, Weber JL (1999) Long homozygous chromosomal segments in
#' reference families from the Centre d’Étude du Polymorphisme Humain.
#' Am J Hum Genet 65:1493--1500.
#'
#' @examples
#' \dontrun{
#' # load DO data from Recla et al. (2014) Mamm Genome 25:211-222.
#' recla <- read_cross2("https://raw.githubusercontent.com/rqtl/qtl2data/main/DO_Recla/recla.zip")
#'
#' # grab founder genotypes and physical map
#' fg <- recla$founder_geno
#' pmap <- recla$pmap
#'
#' # find shared segments
#' (segs <- find_ibd_segments(fg, pmap, min_lod=10, error_prob=0.0001))
#' }
#'
#' @export
find_ibd_segments <-
    function(geno, map, min_lod=15, error_prob=0.001, cores=1)
{
    if(is.null(geno)) stop("geno is NULL")
    if(is.null(map)) stop("map is NULL")

    if(!is_number(min_lod)) stop("min_lod should be a single number")

    n_str <- vapply(geno, nrow, 1)
    if(length(unique(n_str)) != 1)
        stop("The geno matrices have different numbers of rows; they should be strains x markers")
    n_str <- n_str[1]
    if(n_str < 2)
        stop("Need at least two strains")

    nmar <- vapply(geno, ncol, 1)
    nmar_map <- vapply(map, length, 1)
    if(length(nmar) != length(nmar_map))
        stop("length(geno) != length(map)")
    if(!all(nmar == nmar_map))
        stop("geno and map have different numbers of markers")

    if(!is_pos_number(error_prob) || error_prob >= 1)
        stop("error_prob should be a single number in (0,1)")

    # get strain pairs
    str1 <- matrix(rep(1:n_str, n_str), ncol=n_str)
    str2 <- matrix(rep(1:n_str, n_str), ncol=n_str, byrow=TRUE)
    str_list <- data.frame(strain1=str1[upper.tri(str1)],
                           strain2=str2[upper.tri(str2)])
    str_list <- str_list[order(str_list[,1], str_list[,2]),]
    n_pair <- nrow(str_list)

    n_chr <- length(geno)
    run_list <- data.frame(str1 = rep(str_list[,1], n_chr),
                           str2 = rep(str_list[,2], n_chr),
                           chr = rep(1:n_chr, each=n_pair))

    freq <- lapply(geno, function(a) colMeans(a==1, na.rm=TRUE))

    str_names <- rownames(geno[[1]])
    if(is.null(str_names)) {
        if(n_str <= 26) str_names <- LETTERS[1:n_str]
        else str_names <- as.character(1:n_str)
    }

    by_run_func <- function(i) {
        chr <- run_list$chr[i]
        chrnam <- names(map)[chr]
        str1 <- run_list$str1[i]
        str2 <- run_list$str2[i]

        g1 <- geno[[chr]][str1,]
        g2 <- geno[[chr]][str2,]
        m <- map[[chr]]
        p <- freq[[chr]]

        keep <- which(!is.na(g1) & !is.na(g2) & p > 0 & p < 1)

        result <- .find_ibd_segments(g1[keep], g2[keep], p[keep], error_prob)
        result <- as.data.frame(result)

        # drop rows (low LOD or overlapping interval with higher LOD)
        result <- result[result[,3] >= min_lod & result[,6] > 0.5,,drop=FALSE]

        if(nrow(result)==0) return(NULL)

        data.frame(strain1=str_names[str1],
                   strain2=str_names[str2],
                   chr=chrnam,
                   left_marker=names(m)[result[,1]],
                   right_marker=names(m)[result[,2]],
                   left_pos=m[result[,1]],
                   right_pos=m[result[,2]],
                   left_index=as.integer(keep[result[,1]]),
                   right_index=as.integer(keep[result[,2]]),
                   int_length=m[result[,2]] - m[result[,1]],
                   n_mar = as.integer(result[,2] - result[,1] + 1),
                   n_mismatch=as.integer(result[,5]),
                   lod=result[,3],
                   stringsAsFactors=FALSE)

    }

    # set up cluster; set quiet=TRUE if multi-core
    quiet <- TRUE
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    result_list <- cluster_lapply(cores, 1:nrow(run_list), by_run_func)
    if(all(vapply(result_list, is.null, TRUE))) return(NULL) # no segments at all!
    result_list <- result_list[!vapply(result_list, is.null, TRUE)] # just save runs that returned some rows

    # combine the results into a single data frame
    result <- do.call("rbind", result_list)

    # reorder by chromosome, left pos, right pos, strain 1, strain 2
    result <- result[order(factor(result$chr, names(map)),
                           result$left_pos,
                           result$right_pos,
                           factor(result$strain1, str_names),
                           factor(result$strain2, str_names)), ]

    rownames(result) <- as.character(1:nrow(result))
    result
}

# create_marker_index
#
# find marker names in map and create index of locations
# (indexes start at 0)

create_marker_index <-
    function(marker_names, map)
{
    if(length(marker_names) != length(map))
        stop("Different numbers of chromosomes")
    if(!all(names(marker_names) == names(map)))
        stop("Different chromosome names")

    index <- vector("list", length(map))
    for(i in seq(along=marker_names)) {
        if(!all(marker_names[[i]] %in% names(map[[i]])))
            stop("Some markers not found in map")
        index[[i]] <- match(names(map[[i]]), marker_names[[i]]) - 1
        index[[i]][is.na(index[[i]])] <- -1
    }
    names(index) <- names(map)

    index
}

context("create marker index")
suppressMessages(library(qtl))

test_that("grid-based version works in simple case", {

    # equally-spaced map
    map <- seq(0, 50, by=2.5)
    names(map) <- paste0("map", seq(along=map))

    # step = marker distance
    pmap <- insert_pseudomarkers(list("1"=map), step=2.5, off_end=0, stepwidth="fixed")
    index <- create_marker_index(list("1"=names(map)), pmap)
    expect_equal(index, list("1"=seq(along=map)-1))

    # step = 1
    pmap <- insert_pseudomarkers(list("1"=map), step=1, off_end=0, stepwidth="fixed")

    # expected index
    index <- rep(-1, length(pmap[[1]]))
    names(index) <- names(pmap[[1]])
    index[names(map)] <- seq(along=map)-1
    names(index) <- NULL

    expect_equal(list("1"=index), create_marker_index(list("1"=names(map)), pmap))
})

test_that("minimal version works in simple case", {

    # equally-spaced map
    map <- seq(0, 50, by=2.5)
    names(map) <- paste0("map", seq(along=map))

    # step = marker distance
    pmap <- insert_pseudomarkers(list("1"=map), step=2.5, off_end=0, stepwidth="max")
    expect_equal(list("1"=seq(along=map)-1), create_marker_index(list("1"=names(map)), pmap))

    # step = 1
    pmap <- insert_pseudomarkers(list("1"=map), step=1, off_end=0, stepwidth="max")

    # expected index
    index <- rep(-1, length(pmap[[1]]))
    names(index) <- names(pmap[[1]])
    index[names(map)] <- seq(along=map)-1
    names(index) <- NULL

    expect_equal(list("1"=index), create_marker_index(list("1"=names(map)), pmap))
})

test_that("minimal version works in more realistic case", {

    data(hyper)
    map <- qtl::pull.map(hyper, chr=1)

    pmap <- insert_pseudomarkers(map, step=1.55, off_end=4, stepwidth="max")

    expected_index <- c(0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0,
                        0, 0, 3, 0, 4, 0, 5, 0, 0, 6, 0, 7, 8, 0, 0, 0, 9, 0, 0, 0, 10,
                        0, 0, 0, 0, 0, 0, 11, 0, 0, 12, 0, 13, 0, 0, 14, 15, 0, 0, 0,
                        0, 16, 17, 18, 19, 0, 0, 20, 0, 0, 0, 0, 21, 0, 0, 0, 0, 0, 0,
                        0, 0, 0, 0, 0, 0, 0, 0, 22, 0, 0)-1
    expect_equal(list("1"=expected_index), create_marker_index(list("1"=names(map[[1]])), pmap))

})

test_that("insert_pseudomarkers works with a custom pseudomarker map", {

    data(hyper)
    map <- qtl::pull.map(hyper)

    set.seed(99735998)
    pseudomarker_map <- vector("list", length(map))
    names(pseudomarker_map) <- names(map)
    for(i in seq(along=map)) {
        n.pmar <- 10
        pseudomarker_map[[i]] <- sort(runif(n.pmar, 0, max(map[[i]])))
        names(pseudomarker_map[[i]]) <- paste0("c", names(map)[i], ".loc", 1:n.pmar)
    }

    combined_map <- insert_pseudomarkers(map, pseudomarker_map = pseudomarker_map)

    for(i in seq(along=map)) {
        expected <- match(names(combined_map[[i]]), names(map[[i]]))
        expected[is.na(expected)] <- 0
        expected <- expected - 1
        expect_equal(expected, create_marker_index(list("1"=names(map[[i]])), list("1"=combined_map[[i]]))[[1]])
    }
})

# scan1perm with general function but no covariates, kinship, weights
scan1perm_gen_simple <-
    function(genoprobs, pheno, n_perm=1, perm_strata=NULL,
             cores=1, scan_func, ind2keep, ..., max_batch=NULL)
{
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    quiet <- grab_dots(dotargs, "quiet", TRUE)

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # max batch size
    if(is.null(max_batch))
        max_batch <- min(1000, ceiling(n_perm*length(genoprobs)*ncol(pheno)/n_cores(cores)))

    # generate permutations
    perms <- gen_strat_perm(n_perm, ind2keep, perm_strata)

    # batch permutations
    phe_batches <- batch_vec( rep(seq_len(ncol(pheno)), n_perm), max_batch)
    perm_batches <- batch_vec( rep(seq_len(n_perm), each=ncol(pheno)), max_batch)

    # drop cols in genotype probs that are all 0 (just looking at the X chromosome)
    genoprob_Xcol2drop <- genoprobs_col2drop(genoprobs)
    is_x_chr <- attr(genoprobs, "is_x_chr")
    if(is.null(is_x_chr)) is_x_chr <- rep(FALSE, length(genoprobs))

    # batches for analysis, to allow parallel analysis
    run_batches <- data.frame(chr=rep(seq_len(length(genoprobs)), length(phe_batches)),
                              phe_batch=rep(seq_along(phe_batches), each=length(genoprobs)))

    run_indexes <- seq_len(length(genoprobs)*length(phe_batches))

    # subset the phenotypes
    pheno <- pheno[ind2keep,,drop=FALSE]
    nind <- nrow(pheno)

    # the function that does the work
    by_group_func <- function(i) {
        # deal with batch information, including individuals to drop due to missing phenotypes
        chr <- run_batches$chr[i]
        chrnam <- names(genoprobs)[chr]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        permbatch <- perm_batches[[run_batches$phe_batch[i]]]

        pr <- genoprobs[[chr]][ind2keep,,,drop=FALSE]
        pr <- setNames(list(pr), chrnam)
        attr(pr, "is_x_chr") <- attr(genoprobs, "is_x_chr")[chr]
        attr(pr, "crosstype") <- attr(genoprobs, "crosstype")
        attr(pr, "alleles") <- attr(genoprobs, "alleles")
        attr(pr, "alleleprobs") <- attr(genoprobs, "alleleprobs")
        class(pr) <- class(genoprobs)

        ph <- pheno[,phebatch,drop=FALSE]
        for(col in seq_len(ncol(ph))) # permute columns
            ph[,col] <- ph[perms[,permbatch[col]] , col]

        lod <- scan_func(genoprobs=pr, pheno=ph,
                         addcovar=NULL, Xcovar=NULL,
                         intcovar=NULL, kinship=NULL,
                         cores=1, ...)
        lod <- fix_output_if_snps(lod)

        # return column maxima
        apply(lod, 2, max, na.rm=TRUE)
    }

    # calculations in parallel
    list_result <- cluster_lapply(cores, run_indexes, by_group_func)

    # check for problems (if clusters run out of memory, they'll return NULL)
    result_is_null <- vapply(list_result, is.null, TRUE)
    if(any(result_is_null))
        stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

    # reorganize results
    result <- array(dim=c(length(genoprobs), n_perm, ncol(pheno)))

    for(i in run_indexes) {
        chr <- run_batches$chr[i]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        permbatch <- perm_batches[[run_batches$phe_batch[i]]]

        for(j in seq_along(phebatch))
            result[chr,permbatch[j], phebatch[j]] <- list_result[[i]][j]

    }

    result <- apply(result, c(2,3), max)
    colnames(result) <- colnames(pheno)

    class(result) <- c("scan1perm", "matrix")
    result

}

scan1perm_gen <-
    function(genoprobs, pheno, kinship=NULL, addcovar=NULL, Xcovar=NULL,
             intcovar=NULL, weights=NULL,
             n_perm=1, perm_strata=NULL,
             cores=1, scan_func, ind2keep, ..., max_batch=NULL)
{
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    quiet <- grab_dots(dotargs, "quiet", TRUE)

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # max batch size
    if(is.null(max_batch))
        max_batch <- min(1000, ceiling(n_perm*length(genoprobs)*ncol(pheno)/n_cores(cores)))

    # generate permutations
    perms <- gen_strat_perm(n_perm, ind2keep, perm_strata)

    # batch permutations
    phe_batches <- batch_cols(pheno[ind2keep,,drop=FALSE], max_batch)

    # batches for analysis, to allow parallel analysis
    run_batches <- data.frame(chr=rep(seq_len(length(genoprobs)), length(phe_batches)*n_perm),
                              phe_batch=rep(seq_along(phe_batches), each=length(genoprobs)*n_perm),
                              perm_batch=rep(rep(seq_len(n_perm), each=length(genoprobs), length(phe_batches))))

    run_indexes <- seq_len(length(genoprobs)*length(phe_batches)*n_perm)

    # the function that does the work
    by_group_func <- function(i) {
        # deal with batch information, including individuals to drop due to missing phenotypes
        chr <- run_batches$chr[i]
        chrnam <- names(genoprobs)[chr]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        permbatch <- run_batches$perm_batch[i]
        phecol <- phebatch$cols
        omit <- phebatch$omit
        these2keep <- ind2keep # individuals 2 keep for this batch
        if(length(omit) > 0) these2keep <- ind2keep[-omit]
        if(length(these2keep)<=2) return(NULL) # not enough individuals

        # apply permutation to the probs
        #     (I initially thought we'd need to apply these just to the rownames, but
        #     actually we've already aligned probs and pheno and so forth
        #     via ind2keep/these2keep, so now we need to just permute the rows)
        #
        # (also need some contortions here to keep the sizes the same)
        pr <- genoprobs[[chr]][ind2keep,,,drop=FALSE]
        pr <- pr[perms[,permbatch],,,drop=FALSE]
        rownames(pr) <- ind2keep

        # make the probabilities back into a genoprobs object
        pr <- setNames(list(pr), chrnam)
        attr(pr, "is_x_chr") <- attr(genoprobs, "is_x_chr")[chr]
        attr(pr, "crosstype") <- attr(genoprobs, "crosstype")
        attr(pr, "alleles") <- attr(genoprobs, "alleles")
        attr(pr, "alleleprobs") <- attr(genoprobs, "alleleprobs")
        class(pr) <- class(genoprobs)

        # subset kinship and pheno
        k <- kinship; if(is_kinship_list(k)) k <- k[[chr]]
        ph <- pheno[, phecol, drop=FALSE]

        # scan1 function taking clean data (with no missing values)
        lod <- scan_func(genoprobs=pr, pheno=ph, kinship=k, addcovar=addcovar,
                         intcovar=intcovar, Xcovar=Xcovar, weights=weights, ...)
        lod <- fix_output_if_snps(lod)

        # return column maxima
        apply(lod, 2, max, na.rm=TRUE)
    }

    # calculations in parallel
    list_result <- cluster_lapply(cores, run_indexes, by_group_func)

    # check for problems (if clusters run out of memory, they'll return NULL)
    result_is_null <- vapply(list_result, is.null, TRUE)
    if(any(result_is_null))
        stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

    # reorganize results
    result <- array(dim=c(length(genoprobs), n_perm, ncol(pheno)))

    for(i in run_indexes) {
        chr <- run_batches$chr[i]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        permbatch <- run_batches$perm_batch[i]

        result[chr, permbatch, phebatch$cols] <- list_result[[i]]
    }

    result <- apply(result, c(2,3), max)
    colnames(result) <- colnames(pheno)

    class(result) <- c("scan1perm", "matrix")
    result
}



# generate (potentially) stratified permutations
#   (actual work is done in c++, see src/random.cpp)
gen_strat_perm <-
    function(n_perm, ind2keep, perm_strata=NULL)
{
    if(!is.null(perm_strata)) {
        perm_strata <- perm_strata[ind2keep]
        u <- unique(perm_strata)

        if(length(u) > 1) {

            strat_numeric <- match(perm_strata, u)-1

            return(permute_nvector_stratified(n_perm,
                                              seq_along(ind2keep),
                                              strat_numeric,
                                              length(u)))
        }
    }

    permute_nvector(n_perm, seq_along(ind2keep))
}

# detect if scan1snps out (list with "lod" and "snpinfo")
fix_output_if_snps <-
    function(scan_output)
{
    if(is.list(scan_output) && length(scan_output)==2
       && all(names(scan_output) == c("lod", "snpinfo")))
        return(scan_output$lod)

    scan_output
}

# find_peaks_and_lodint
#
# This is like find_peaks but the results also include lod support intervals.
# The same input as find_peaks, which calls this function when drop is given.
find_peaks_and_lodint <-
    function(scan1_output, map, threshold=3, peakdrop=Inf, drop=1.8,
             thresholdX=NULL, peakdropX=NULL, dropX=NULL,
             expand2markers=TRUE, sort_by=c("column", "pos", "lod"), cores=1)
{
    sort_by <- match.arg(sort_by)

    lodnames <- colnames(scan1_output)
    n_lod <- length(lodnames)

    # make the thresholds have length n_lod
    if(length(threshold)==1) threshold <- rep(threshold, n_lod)
    if(length(threshold) != n_lod)
        stop("threshold should have length 1 or ", n_lod)
    if(is.null(thresholdX)) thresholdX <- threshold
    if(length(thresholdX)==1) thresholdX <- rep(thresholdX, n_lod)
    if(length(thresholdX) != n_lod)
        stop("thresholdX should have length 1 or ", n_lod)

    # make the peakdrops have length n_lod
    if(length(peakdrop)==1) peakdrop <- rep(peakdrop, n_lod)
    if(length(peakdrop) != n_lod)
        stop("peakdrop should have length 1 or ", n_lod)
    if(is.null(peakdropX)) peakdropX <- peakdrop
    if(length(peakdropX)==1) peakdropX <- rep(peakdropX, n_lod)
    if(length(peakdropX) != n_lod)
        stop("peakdropX should have length 1 or ", n_lod)

    # make the drops have length n_lod
    if(length(drop)==1) drop <- rep(drop, n_lod)
    if(length(drop) != n_lod)
        stop("drop should have length 1 or ", n_lod)
    if(is.null(dropX)) dropX <- drop
    if(length(dropX)==1) dropX <- rep(dropX, n_lod)
    if(length(dropX) != n_lod)
        stop("dropX should have length 1 or ", n_lod)

    if(any(drop > peakdrop | dropX > peakdropX))
        stop("Must have drop <= peakdrop")

    # split lod scores by column and by chromosome
    lod <- as.list(as.data.frame(unclass(scan1_output)))
    chr <- rep(seq_along(map), vapply(map, length, 1))
    lod <- lapply(lod, split, chr)

    # batch info for parallel-processing
    batch <- cbind(rep(seq_along(lod), vapply(lod, length, 1)),
                   unlist(lapply(lod, function(a) seq_along(a))))
    dimnames(batch) <- NULL
    batch_index <- seq_len(nrow(batch))

    # set up parallel analysis
    cores <- setup_cluster(cores)

    # X chr info
    is_x_chr <- attr(map, "is_x_chr")
    if(is.null(is_x_chr))
        is_x_chr <- rep(FALSE, length(map))

    # function to be applied to each batch
    by_batch_func <- function(i) {
        lodcol <- batch[i,1]
        chr <- batch[i,2]

        this_lod <- lod[[lodcol]][[chr]]

        peaks <- .find_peaks_and_lodint(this_lod,
                                        ifelse(is_x_chr[chr], thresholdX[lodcol], threshold[lodcol]),
                                        ifelse(is_x_chr[chr], peakdropX[lodcol], peakdrop[lodcol]),
                                        ifelse(is_x_chr[chr], dropX[lodcol], drop[lodcol]))

        n_peaks <- length(peaks)
        if(n_peaks==0) return(NULL)

        # calculate peak positions
        # (this deals with ties in the LOD scores:
        #  take average of multiple positions sharing the maximum LOD score)
        # and remember that .find_peaks returns indexes starting at 0
        ci_lo <- ci_hi <- rep(0, n_peaks)
        for(i in seq_len(n_peaks)) {
            if(expand2markers)
                ci <- expand_interval_to_markers(peaks[[i]][1:2]+1, map[[chr]])
            else
                ci <- map[[chr]][peaks[[i]][1:2]+1]
            ci_lo[i] <- ci[1]
            ci_hi[i] <- ci[2]
        }
        peak_pos <- vapply(peaks, function(a) mean(map[[chr]][a[-(1:2)]+1]), 0.0)
        peak_lod <- vapply(peaks, function(a) this_lod[a[3]+1], 0.0)

        data.frame(lodindex=rep(lodcol,n_peaks),
                   lodcolumn=rep(lodnames[lodcol],n_peaks),
                   chr=rep(names(map)[chr],n_peaks),
                   pos=peak_pos,
                   lod=peak_lod,
                   ci_lo=ci_lo,
                   ci_hi=ci_hi,
                   row.names=seq_along(peaks),
                   stringsAsFactors=FALSE)
    }

    if(n_cores(cores)==1) {
        peaks <- lapply(batch_index, by_batch_func)
    } else {
        peaks <- cluster_lapply(cores, batch_index, by_batch_func)
    }

    # empty data frame to start
    result <- data.frame(lodindex=numeric(0),
                         lodcolumn=character(0),
                         chr=character(0),
                         pos=numeric(0),
                         lod=numeric(0),
                         ci_lo=numeric(0),
                         ci_hi=numeric(0),
                         row.names=character(0),
                         stringsAsFactors=FALSE)

    for(p in peaks) result <- rbind(result, p)

    rownames(result) <- NULL
    result$chr <- factor(result$chr, names(map))

    if(nrow(result) > 1) {
        if(sort_by == "pos") {

            result <- result[order(result$chr, result$pos, result$lodindex), ]

        } else if(sort_by == "lod") {

            result <- result[order(result$lod, decreasing=TRUE),]

        }
    }

    result
}


# expand LOD interval endpoints to be at markers
# (rather than pseudomarkers)
#
# input is a pair of indexes
# output is a pair of positions
expand_interval_to_markers <-
    function(ci_index, map)
{
    mn <- names(map)
    pmar_pattern <- "^c.+\\.loc-*[0-9]+(\\.[0-9]+)*$"
    pmar <- grepl(pmar_pattern, mn)

    mar_index <- seq_along(map)[!pmar]

    lo_index <- ci_index[1]
    hi_index <- ci_index[2]

    if(pmar[lo_index]) {
        mar_index <= lo_index
        lo_index <- ifelse(any(mar_index <= lo_index),
                           max(mar_index[mar_index <= lo_index]),
                           1)
    }

    if(pmar[hi_index]) {
        mar_index >= hi_index
        hi_index <- ifelse(any(mar_index >= hi_index),
                           min(mar_index[mar_index >= hi_index]),
                           length(map))
    }

    map[c(lo_index, hi_index)]
}

# find columns in a genotype probability array that should be dropped
#
# Xonly: If TRUE and probs is a list, just look at the X chromosome
genoprobs_col2drop <-
    function(probs, Xonly=TRUE, tol=1e-8)
{
    if(is.list(probs)) { # proper calc_genoprob object, hopefully
        is_x_chr <- attr(probs, "is_x_chr")
        if(Xonly) {
            result <- lapply(seq_len(length(probs)), function(a) numeric(0))
            names(result) <- names(probs)

            if(is.null(is_x_chr) || !any(is_x_chr)) return(result) # no X chr

            w <- which(is_x_chr)
            for(i in w)
                result[[i]] <- genoprobs_col2drop(probs[[i]])

            return(result)
        }
        else {
            result <- lapply(seq_len(length(probs)), function(a) numeric(0))
            names(result) <- names(probs)

            for(i in seq_len(length(probs)))
              result[[i]] <- genoprobs_col2drop(probs[[i]])

            return(result)
        }
    }

    # put genotypes first and take rowMeans
    wh <- which(rowMeans(aperm(probs, c(2,1,3))) < tol)

    names(wh) <- NULL # eliminate names
    wh
}

context("genoprobs columns to drop")

test_that("genoprobs_cols2drop works", {

    # load example data from R/qtl
    library(qtl)
    data(fake.f2)
    fake.f2 <- fake.f2[c(18,19,"X"),] # subset chromosomes
    fake.f2 <- convert2cross2(fake.f2)

    # calculate genoprobs
    pr <- calc_genoprob(fake.f2)

    # just X chromosome
    expect_equal(genoprobs_col2drop(pr$X), numeric(0))

    # all chromosomes
    expected <- vector("list", length(pr))
    names(expected) <- names(pr)
    for(i in seq(along=expected)) expected[[i]] <- numeric(0)
    expect_equal(genoprobs_col2drop(pr), expected)
    expect_equal(genoprobs_col2drop(pr, FALSE), expected)

    #####
    # all males
    fake.f2$is_female[fake.f2$is_female] <- FALSE
    pr <- calc_genoprob(fake.f2)

    # just X chromosome
    expect_equal(genoprobs_col2drop(pr$X), 1:4)

    # all chromosomes
    expected$X <- 1:4
    expect_equal(genoprobs_col2drop(pr), expected)
    expect_equal(genoprobs_col2drop(pr, FALSE), expected)

    # test Xonly=FALSE
    expected2 <- expected;expected2$X <- numeric(0)
    prA <- pr
    attr(prA, "is_x_chr") <- rep(FALSE, length(prA))
    expect_equal(genoprobs_col2drop(prA), expected2)
    expect_equal(genoprobs_col2drop(prA, FALSE), expected)

    #####
    # all females
    fake.f2$is_female[!fake.f2$is_female] <- TRUE
    pr <- calc_genoprob(fake.f2)

    # just X chromosome
    expect_equal(genoprobs_col2drop(pr$X), 5:6)

    # all chromosomes
    expected$X <- 5:6
    expect_equal(genoprobs_col2drop(pr), expected)
    expect_equal(genoprobs_col2drop(pr, FALSE), expected)

    # test Xonly=FALSE
    expected2 <- expected;expected2$X <- numeric(0)
    prA <- pr
    attr(prA, "is_x_chr") <- rep(FALSE, length(prA))
    expect_equal(genoprobs_col2drop(prA), expected2)
    expect_equal(genoprobs_col2drop(prA, FALSE), expected)

})

# used to skip tests when not on *my* computer in interactive mode
# needed for testing mult-core code, which isn't allowed within R CMD check
isnt_karl <-
    function()
{
    !(interactive() &&
      identical(Sys.getenv("KARL_LOCAL"), "true"))
}

# is a number? (from assertthat)
is_number <- function(x) is.numeric(x) && length(x)==1
is_nonneg_number <- function(x) is_number(x) && x >= 0
is_pos_number <- function(x) is_number(x) && x > 0

context("Conversion of genotype probabilities to allele probabilities")

test_that("genoprob_to_alleleprob works for RIL", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2, map, error_prob=0.002)
    allele_probs <- genoprob_to_alleleprob(probs)

    # expected result, hardly changed
    expected <- probs
    attr(expected, "alleleprobs") <- TRUE
    for(i in seq(along=probs))
        colnames(expected[[i]]) <- c("L", "C")

    expect_equal(allele_probs, expected)

})

test_that("genoprob_to_alleleprob works for F2", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    allele_probs <- genoprob_to_alleleprob(probs)

    f2_geno2alle <-
        function(prob, x_chr=FALSE)
        {
            if(x_chr) {
                cn <- colnames(prob)
                prob[,1,] <- prob[,1,]+prob[,2,]/2+prob[,3,]/2 + prob[,5,]
                prob[,2,] <- prob[,4,]+prob[,2,]/2+prob[,3,]/2 + prob[,6,]
                prob <- prob[,1:2,]
                colnames(prob) <- substr(cn[c(1,3)], 1, 1)
                return(prob)

            } else {
                cn <- colnames(prob)
                prob[,1,] <- prob[,1,]+prob[,2,]/2
                prob[,2,] <- prob[,3,]+prob[,2,]/2
                prob <- prob[,1:2,]
                colnames(prob) <- substr(cn[c(1,3)], 1, 1)
                return(prob)
            }
        }

    expected <- probs
    is_x_chr <- attr(probs, "is_x_chr")
    for(i in seq(along=probs)) # loop over chromosomes
        expected[[i]] <- f2_geno2alle(probs[[i]], is_x_chr[i])
    attr(expected, "alleleprobs") <- TRUE

    expect_equal(allele_probs, expected)

})

test_that("genoprob_to_alleleprob works when multi-core", {
    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    allele_probs <- genoprob_to_alleleprob(probs)
    allele_probs_mc <- genoprob_to_alleleprob(probs, cores=2)
    expect_equal(allele_probs_mc, allele_probs)

    # following shouldn't really matter, since no calculations are done
    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2, map, error_prob=0.002)
    allele_probs <- genoprob_to_alleleprob(probs)
    allele_probs_mc <- genoprob_to_alleleprob(probs, cores=2)
    expect_equal(allele_probs_mc, allele_probs)

})

context("plot_coef")

test_that("plot_coef works", {

    skip_if(isnt_karl(), "plot tests only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,2]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    pheno <- iron$pheno[,1,drop=FALSE]
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)

    coef <- scan1coef(probs, pheno, addcovar=covar)

    test_plot_coef <- function() plot(coef, map, columns=1:3)

    expect_doppelganger("plot_coef", test_plot_coef)

})

context("calc_genoprob2")
suppressMessages(library(qtl))

test_that("backcross autosome", {

    data(hyper)
    chr <- c(1, 3, 4, 17, 19)
    hyper <- hyper[chr,]

    hyper <- convert2cross2(hyper)
    pr <- calc_genoprob(hyper, error_prob=0.002, lowmem=TRUE)
    pr2 <- calc_genoprob(hyper, error_prob=0.002, lowmem=FALSE)

    expect_equal(pr, pr2)

})

test_that("intercross autosome", {

    data(listeria)
    chr <- c(1, 4, 14, 18)
    listeria <- listeria[chr,]

    listeria <- convert2cross2(listeria)
    map <- insert_pseudomarkers(listeria$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(listeria, map, error_prob=0.01, lowmem=TRUE)
    pr2 <- calc_genoprob(listeria, map, error_prob=0.01, lowmem=FALSE)

    expect_equal(pr, pr2)

})


test_that("risib autosome", {

    data(hyper)
    chr <- c(1, 3, 4, 17, 19)
    hyper <- hyper[chr,]
    class(hyper)[1] <- "risib"

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(hyper, map, error_prob=0.002, lowmem=TRUE)
    pr2 <- calc_genoprob(hyper, map, error_prob=0.002, lowmem=FALSE)

    expect_equal(pr, pr2)

})

test_that("riself autosome", {

    data(hyper)
    chr <- c(1, 3, 4, 17, 19)
    hyper <- hyper[chr,]
    class(hyper)[1] <- "riself"

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(hyper, map, error_prob=0.002, lowmem=TRUE)
    pr2 <- calc_genoprob(hyper, map, error_prob=0.002, lowmem=FALSE)

    expect_equal(pr, pr2)

})

test_that("f2 X chr", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]

    fake.f2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(fake.f2, map, error_prob=0.01, lowmem=TRUE)
    pr2 <- calc_genoprob(fake.f2, map, error_prob=0.01, lowmem=FALSE)

    expect_equal(pr, pr2)

})

test_that("bc X chr", {

    data(hyper)
    hyper <- hyper["X",]
    # make it half female, half male
    hyper$pheno$sex <- rep(c("female", "male"), nind(hyper)/2)

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(hyper, map, error_prob=0.02, lowmem=TRUE)
    pr2 <- calc_genoprob(hyper, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(pr, pr2)

})

test_that("f2 X chr all males", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 1

    fake.f2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(fake.f2, map, error_prob=0.02, lowmem=TRUE)
    pr2 <- calc_genoprob(fake.f2, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(pr, pr2)

})


test_that("f2 X chr all females", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 0

    fake.f2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(fake.f2, map, error_prob=0.02, lowmem=TRUE)
    pr2 <- calc_genoprob(fake.f2, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(pr, pr2)

})

test_that("f2 X chr all females forw", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 0
    fake.f2$pheno$pgm <- 0

    fake.f2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(fake.f2, map, error_prob=0.0001, lowmem=TRUE)
    pr2 <- calc_genoprob(fake.f2, map, error_prob=0.0001, lowmem=FALSE)

    expect_equal(pr, pr2)

})

test_that("f2 X chr all females rev", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 0
    fake.f2$pheno$pgm <- 1

    fake.f2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(fake.f2, map, error_prob=0.0001, lowmem=TRUE)
    pr2 <- calc_genoprob(fake.f2, map, error_prob=0.0001, lowmem=FALSE)

    expect_equal(pr, pr2)

})

test_that("bc X chr all males", {

    data(hyper)
    hyper <- hyper["X",]

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(hyper, map, error_prob=0.02, lowmem=TRUE)
    pr2 <- calc_genoprob(hyper, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(pr, pr2)

})


test_that("bc X chr all females", {

    data(hyper)
    hyper <- hyper["X",]
    hyper$pheno$sex <- "female"

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(hyper, map, error_prob=0.02, lowmem=TRUE)
    pr2 <- calc_genoprob(hyper, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(pr, pr2)

})

test_that("doubled haploids", {

    data(hyper)
    hyper <- hyper[1,]
    hyper$pheno <- hyper$pheno[,1,drop=FALSE]
    class(hyper)[1] <- "dh"

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(hyper, map, error_prob=0.02, lowmem=TRUE)
    pr2 <- calc_genoprob(hyper, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(pr, pr2)

})


test_that("haploids calc_genoprob", {

    data(hyper)
    hyper <- hyper[2,]
    hyper$pheno <- hyper$pheno[,1,drop=FALSE]
    class(hyper)[1] <- "haploid"

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(hyper, map, error_prob=0.02, lowmem=TRUE)
    pr2 <- calc_genoprob(hyper, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(pr, pr2)

})

test_that("backcross autosome with markers at same location", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    grav2 <- grav2[1:4,1]

    # put some markers at same location
    grav2$gmap[[1]][4] <- grav2$gmap[[1]][3]

    pr <- calc_genoprob(grav2, err=0, lowmem=TRUE)
    pr2 <- calc_genoprob(grav2, err=0, lowmem=FALSE)

    expect_true( all(!is.na(pr[[1]])) )
    expect_equal(pr, pr2)

})

test_that("calc_genoprob works when multi-core", {
    skip_if(isnt_karl(), "this test only run locally")

    data(hyper)
    hyper2 <- convert2cross2(hyper)
    pr <- calc_genoprob(hyper2, error_prob=0.002, lowmem=TRUE)
    pr_mc <- calc_genoprob(hyper2, error_prob=0.002, cores=2, lowmem=TRUE)
    expect_equal(pr_mc, pr)

    pr2 <- calc_genoprob(hyper2, error_prob=0.002, lowmem=FALSE)
    expect_equal(pr2, pr)
    pr2_mc <- calc_genoprob(hyper2, error_prob=0.002, cores=2, lowmem=FALSE)
    expect_equal(pr2_mc, pr)

    data(listeria)
    listeria2 <- convert2cross2(listeria)
    map <- insert_pseudomarkers(listeria2$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(listeria2, map, error_prob=0.01, lowmem=TRUE)
    pr_mc <- calc_genoprob(listeria2, map, error_prob=0.01, cores=2, lowmem=TRUE)
    expect_equal(pr_mc, pr)

    pr2 <- calc_genoprob(listeria2, map, error_prob=0.01, lowmem=FALSE)
    expect_equal(pr2, pr)
    pr2_mc <- calc_genoprob(listeria2, map, error_prob=0.01, cores=2, lowmem=FALSE)
    expect_equal(pr2_mc, pr)

})

context("compare_geno")

test_that("compare_geno works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:5, c(18:19,"X")]

    cg <- compare_geno(iron)
    expected <- structure(c(6, 0, 6, 6, 0, NA, 0, 0, 0, 0, 0.333333333333333,
                            NA, 6, 6, 0, 0.166666666666667, NA, 0.666666666666667, 6, 0,
                            NA, NA, NA, NA, 0), dim = c(5L, 5L),
                          dimnames = list(c("1", "2", "3", "4", "5"), c("1", "2", "3", "4", "5")),
                          proportion=TRUE,
                          class = c("compare_geno", "matrix"))
    expect_equal(cg, expected)

    cg_X <- compare_geno(iron[,"X"])
    expectedX <- structure(c(2, 0, 2, 2, 0, NA, 0, 0, 0, 0, 0, NA, 2, 2, 0, 0,
                             NA, 1, 2, 0, NA, NA, NA, NA, 0), dim = c(5L, 5L),
                           dimnames = list(c("1", "2", "3", "4", "5"), c("1", "2", "3", "4", "5")),
                           proportion=TRUE,
                           class = c("compare_geno", "matrix"))
    expect_equal(cg_X, expectedX)

    cg_noX <- compare_geno(iron, omit_x=TRUE)
    expected_noX <- structure(c(4, 0, 4, 4, 0, NA, 0, 0, 0, 0, 0.5, NA, 4, 4, 0,
                                0.25, NA, 0.5, 4, 0, NA, NA, NA, NA, 0), dim = c(5L, 5L),
                              dimnames = list(c("1", "2", "3", "4", "5"), c("1", "2", "3", "4", "5")),
                              proportion=TRUE,
                              class = c("compare_geno", "matrix"))
    expect_equal(cg_noX, expected_noX)

    # if proportion=FALSE, total is sum of X and autosomes
    expect_equal(compare_geno(iron, proportion=FALSE),
                 compare_geno(iron[,"X"], proportion=FALSE) +
                 compare_geno(iron, omit_x=TRUE, proportion=FALSE))

    # test summary()
    expected <- structure(list(ind1 = character(0), ind2 = character(0), prop_match = numeric(0),
                               n_mismatch = numeric(0), n_typed = numeric(0), n_match = numeric(0),
                               index1=numeric(0), index2=numeric(0)),
                          row.names = integer(0), class = c("summary.compare_geno", "data.frame"),
                          threshold = 0.9)
    expect_equal(summary(cg), expected)

    expected <- structure(list(ind1 = "3", ind2 = "4", prop_match = 0.666666666666667,
                               n_mismatch = 2, n_typed = 6, n_match = 4, index1=3, index2=4),
                          row.names = 1L, class = c("summary.compare_geno", "data.frame"),
                          threshold = 0.5)
    expect_equal(summary(cg, 0.5), expected)

    # test max()
    attr(expected, "threshold") <- NULL
    expect_equal(max(cg), expected)

    # summary and max should give the same results whether you use proportion=TRUE or FALSE
    expect_equal(summary(compare_geno(iron, proportion=TRUE)),
                 summary(compare_geno(iron, proportion=FALSE)))
    expect_equal(summary(compare_geno(iron, proportion=TRUE), threshold=0.7),
                 summary(compare_geno(iron, proportion=FALSE), threshold=0.7))
    expect_equal(max(compare_geno(iron, proportion=TRUE)),
                 max(compare_geno(iron, proportion=FALSE)))
    expect_equal(summary(compare_geno(iron, omit_x=TRUE, proportion=TRUE)),
                 summary(compare_geno(iron, omit_x=TRUE, proportion=FALSE)))
    expect_equal(summary(compare_geno(iron, omit_x=TRUE, proportion=TRUE), threshold=0.7),
                 summary(compare_geno(iron, omit_x=TRUE, proportion=FALSE), threshold=0.7))
    expect_equal(max(compare_geno(iron, omit_x=TRUE, proportion=TRUE)),
                 max(compare_geno(iron, omit_x=TRUE, proportion=FALSE)))

})

test_that("compare_geno works when multi-core", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:5, c(18:19,"X")]

    cg <- compare_geno(iron)
    cg_mc <- compare_geno(iron, cores=2)
    expect_equal(cg, cg_mc)

})

context("est_map")
suppressMessages(library(qtl))

test_that("est_map for backcross autosome matches R/qtl", {

    data(hyper)
    chr <- c(3, 4, 17, 19)
    hyper <- hyper[chr,]
    newmap <- est.map(hyper, err=0.002, tol=1e-8)
    newmap <- lapply(newmap, unclass)

    hyper2 <- convert2cross2(hyper)
    newmap2 <- est_map(hyper2, err=0.002, tol=1e-8)

    expect_equivalent(newmap, newmap2)
    expect_equal(lapply(newmap, attr, "loglik"),
                 lapply(newmap2, attr, "loglik"))

})

test_that("est_map for intercross autosome matches R/qtl", {

    skip_on_cran()

    data(listeria)
    chr <- c(4, 14, 18)
    listeria <- listeria[chr,]
    newmap <- est.map(listeria, err=0.01, tol=1e-8)
    newmap <- lapply(newmap, unclass)

    listeria2 <- convert2cross2(listeria)
    newmap2 <- est_map(listeria2, err=0.01, tol=1e-8)

    expect_equivalent(newmap, newmap2)
    expect_equal(lapply(newmap, attr, "loglik"),
                 lapply(newmap2, attr, "loglik"))


})

test_that("f2 X chr est_map matches R/qtl", {

    skip_on_cran()

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    newmap <- est.map(fake.f2, err=0.01, tol=1e-8)
    newmap <- lapply(newmap, unclass)

    fake.f2.2 <- convert2cross2(fake.f2)
    newmap2 <- est_map(fake.f2.2, err=0.01, tol=1e-8)

    expect_equivalent(newmap, newmap2)
    expect_equal(lapply(newmap, attr, "loglik"),
                 lapply(newmap2, attr, "loglik"))

})

test_that("bc X chr calc_genoprob matches R/qtl", {

    skip_on_cran()

    set.seed(19115167)
    xmap <- sim.map(100, n.mar=11, anchor.tel=TRUE, include.x=TRUE, eq.spacing=TRUE)
    n.ind <- 100
    cross <- sim.cross(xmap, n.ind=n.ind, type="bc",
                       error.prob=0.01, missing.prob=0.05)
    cross$pheno$sex <- rep(c(0,1), n.ind/2)
    sexpgm <- getsex(cross)

    newmap <- est.map(cross, err=0.01, tol=1e-8)
    newmap <- lapply(newmap, unclass)

    cross2 <- convert2cross2(cross)
    newmap2 <- est_map(cross2, err=0.01, tol=1e-8)

    expect_equivalent(newmap, newmap2)
    expect_equal(lapply(newmap, attr, "loglik"),
                 lapply(newmap2, attr, "loglik"))

})

test_that("est_map for RIself matches R/qtl", {

    skip_on_cran()

    data(hyper)
    chr <- c(3, 4, 17, 19)
    hyper <- hyper[chr,]
    class(hyper)[1] <- "riself"
    newmap <- est.map(hyper, err=0.002, tol=1e-8)
    newmap <- lapply(newmap, unclass)

    hyper2 <- convert2cross2(hyper)
    newmap2 <- est_map(hyper2, err=0.002, tol=1e-8)

    expect_equivalent(newmap, newmap2)
    expect_equal(lapply(newmap, attr, "loglik"),
                 lapply(newmap2, attr, "loglik"))

})


test_that("est_map for RIsib matches R/qtl", {

    skip_on_cran()

    data(hyper)
    chr <- c(3, 4, 17, 19)
    hyper <- hyper[chr,]
    class(hyper)[1] <- "risib"
    newmap <- est.map(hyper, err=0.002, tol=1e-8)
    newmap <- lapply(newmap, unclass)

    hyper2 <- convert2cross2(hyper)
    newmap2 <- est_map(hyper2, err=0.002, tol=1e-8)

    expect_equivalent(newmap, newmap2)
    expect_equal(lapply(newmap, attr, "loglik"),
                 lapply(newmap2, attr, "loglik"))

})

test_that("est_map for doubled haploids matches R/qtl", {

    skip_on_cran()

    data(hyper)
    hyper <- hyper[3,]
    hyper$pheno <- hyper$pheno[,1,drop=FALSE]
    class(hyper)[1] <- "dh"

    newmap <- est.map(hyper, err=0.002, tol=1e-8)
    newmap <- lapply(newmap, unclass)

    hyper2 <- convert2cross2(hyper)
    newmap2 <- est_map(hyper2, err=0.002, tol=1e-8)

    expect_equivalent(newmap, newmap2)
    expect_equal(lapply(newmap, attr, "loglik"),
                 lapply(newmap2, attr, "loglik"))

})

test_that("est_map for haploids matches R/qtl", {

    skip_on_cran()

    data(hyper)
    hyper <- hyper[4,]
    hyper$pheno <- hyper$pheno[,1,drop=FALSE]
    class(hyper)[1] <- "haploid"

    newmap <- est.map(hyper, err=0.002, tol=1e-8)
    newmap <- lapply(newmap, unclass)

    hyper2 <- convert2cross2(hyper)
    newmap2 <- est_map(hyper2, err=0.002, tol=1e-8)

    expect_equivalent(newmap, newmap2)
    expect_equal(lapply(newmap, attr, "loglik"),
                 lapply(newmap2, attr, "loglik"))

})

test_that("est_map works in case of 2 markers", {

    skip_on_cran()

    data(hyper)
    hyper <- pull.markers(hyper, markers=markernames(hyper, chr=6)[2:3])
    hyper <- shiftmap(hyper)

    newmap <- est.map(hyper, err=0.002, tol=1e-8)
    newmap <- lapply(newmap, unclass)

    hyper2 <- convert2cross2(hyper)
    newmap2 <- est_map(hyper2, err=0.002, tol=1e-8)

    expect_equivalent(newmap, newmap2)
    expect_equal(lapply(newmap, attr, "loglik"),
                 lapply(newmap2, attr, "loglik"))
})

test_that("est_map works in case of 2 markers in intercross", {

    skip_on_cran()

    data(fake.f2)
    fake.f2 <- fake.f2[18,]
    newmap <- est.map(fake.f2, err=0.01, tol=1e-8)
    newmap <- lapply(newmap, unclass)

    fake.f2.2 <- convert2cross2(fake.f2)
    newmap2 <- est_map(fake.f2.2, err=0.01, tol=1e-8)

    expect_equivalent(newmap, newmap2)
    expect_equal(lapply(newmap, attr, "loglik"),
                 lapply(newmap2, attr, "loglik"))
})

test_that("est_map works when multi-core", {
    skip_if(isnt_karl(), "this test only run locally")

    data(hyper)
    hyper2 <- convert2cross2(hyper)
    newmap2 <- est_map(hyper2, err=0.002, tol=1e-8)
    newmap2_mc <- est_map(hyper2, err=0.002, tol=1e-8, cores=2)
    expect_equal(newmap2_mc, newmap2)

    data(listeria)
    listeria2 <- convert2cross2(listeria)
    newmap2 <- est_map(listeria2, err=0.01, tol=1e-8)
    newmap2_mc <- est_map(listeria2, err=0.01, tol=1e-8, cores=2)
    expect_equal(newmap2_mc, newmap2)

})

context("Calculate heterozygosities")

test_that("calc_het works for an intercross", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    pr <- calc_genoprob(iron, err=0.002)

    # omit_x works?
    expect_equal(calc_het(pr[,1:19]), calc_het(pr, omit_x=TRUE))
    expect_equal(calc_het(pr[,1:19], "marker"), calc_het(pr, "marker", omit_x=TRUE))

    # by individual
    ind <- c(88,105,231)
    expected <- c(`88` = 0.437768300719857, `105` = 0.530653642640462, `231` = 0.51155729816745)
    expect_equal(calc_het(pr[ind,]), expected)
    expect_equal(calc_het(pr[ind,], omit_x=TRUE), expected)
    expected_wX <- c(`88` = 0.424502594637437, `105` = 0.514573229227115, `231` = 0.526341691848004)
    expect_equal(calc_het(pr[ind,], omit_x=FALSE), expected_wX)

    # by marker
    expected <- c(D19Mit68 = 0.48461632550778, D19Mit37 = 0.477608930381241,
                  DXMit16 = 0.251799101291949, DXMit186 = 0.258740713762659)
    expect_equal(calc_het(pr[,c(19,"X")], "marker"), expected[1:2])
    expect_equal(calc_het(pr[,c(19,"X")], "marker", omit_x=TRUE), expected[1:2])
    expect_equal(calc_het(pr[,c(19,"X")], "marker", omit_x=FALSE), expected)

})

test_that("calc_het works with multi-core", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    pr <- calc_genoprob(iron, err=0.002)

    # test multicore
    expect_equal(calc_het(pr), calc_het(pr, cores=4))
    expect_equal(calc_het(pr, "marker"), calc_het(pr, "marker", cores=4))
    expect_equal(calc_het(pr, omit_x=FALSE), calc_het(pr, omit_x=FALSE, cores=4))
    expect_equal(calc_het(pr, "marker", omit_x=FALSE), calc_het(pr, "marker", omit_x=FALSE, cores=4))

})

context("sim_geno")

# these are really just regression tests

test_that("sim_geno riself", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))

    RNGkind("Mersenne-Twister")
    set.seed(20150918)

    map <- insert_pseudomarkers(grav2$gmap, step=1)
    dr <- sim_geno(grav2, map, n_draws=2, err=0.002)

    expected <- structure(c(2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
                            1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L, 1L,
                            1L, 1L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L, 2L,
                            2L), dim = c(135L, 2L),
                          dimnames = list(c("PVV4", "c1.loc1",
                          "c1.loc2", "c1.loc3", "c1.loc4", "c1.loc5", "c1.loc6", "AXR-1",
                          "c1.loc7", "c1.loc8", "c1.loc9", "HH.335C-Col/PhyA", "c1.loc10",
                          "c1.loc11", "c1.loc12", "EC.480C", "c1.loc13", "c1.loc14", "c1.loc15",
                          "c1.loc16", "c1.loc17", "c1.loc18", "EC.66C", "c1.loc19", "GD.86L",
                          "c1.loc20", "c1.loc21", "c1.loc22", "c1.loc23", "c1.loc24", "c1.loc25",
                          "CH.160L-Col", "CC.98L-Col/101C", "c1.loc26", "c1.loc27", "c1.loc28",
                          "c1.loc29", "c1.loc30", "c1.loc31", "c1.loc32", "c1.loc33", "c1.loc34",
                          "c1.loc35", "AD.121C", "c1.loc36", "AD.106L-Col", "c1.loc37",
                          "c1.loc38", "c1.loc39", "c1.loc40", "c1.loc41", "c1.loc42", "c1.loc43",
                          "c1.loc44", "GB.112L", "c1.loc45", "c1.loc46", "c1.loc47", "c1.loc48",
                          "c1.loc49", "c1.loc50", "GD.97L", "c1.loc51", "c1.loc52", "c1.loc53",
                          "EG.113L/115C", "c1.loc54", "c1.loc55", "c1.loc56", "c1.loc57",
                          "CD.89C", "c1.loc58", "c1.loc59", "c1.loc60", "c1.loc61", "c1.loc62",
                          "BF.206L-Col", "c1.loc63", "c1.loc64", "c1.loc65", "c1.loc66",
                          "c1.loc67", "CH.200C", "c1.loc68", "c1.loc69", "c1.loc70", "c1.loc71",
                          "c1.loc72", "EC.88C", "c1.loc73", "c1.loc74", "c1.loc75", "c1.loc76",
                          "GD.160C", "c1.loc77", "c1.loc78", "c1.loc79", "c1.loc80", "c1.loc81",
                          "HH.375L", "CH.215L", "c1.loc82", "c1.loc83", "c1.loc84", "c1.loc85",
                          "c1.loc86", "c1.loc87", "c1.loc88", "BF.116C", "c1.loc89", "c1.loc90",
                          "GH.157L-Col", "c1.loc91", "c1.loc92", "c1.loc93", "c1.loc94",
                          "c1.loc95", "c1.loc96", "c1.loc97", "CC.318C", "c1.loc98", "c1.loc99",
                          "CD.173L/175C-Col", "c1.loc100", "c1.loc101", "c1.loc102", "c1.loc103",
                          "c1.loc104", "GH.127L-Col/ADH", "c1.loc105", "c1.loc106", "c1.loc107",
                          "c1.loc108", "c1.loc109", "HH.360L-Col"), NULL))

    expect_equal(dr[[1]][138,,], expected)

})

test_that("sim_geno f2", {

    skip_on_cran()

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    RNGkind("Mersenne-Twister")
    set.seed(20150918)
    map <- insert_pseudomarkers(iron$gmap, step=1)
    dr <- sim_geno(iron, map, n_draws=2, err=0.002)

    expected <- structure(c(5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
                            4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
                            4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
                            4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 6L, 4L, 6L, 4L, 5L,
                            4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
                            4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
                            4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L,
                            4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L, 5L, 4L),
                          dim = c(2L, 30L, 2L),
                          dimnames = list(c("145", "146"), c("DXMit16", "cX.loc30.5",
                          "cX.loc31.5", "cX.loc32.5", "cX.loc33.5", "cX.loc34.5", "cX.loc35.5", "cX.loc36.5",
                          "cX.loc37.5", "cX.loc38.5", "cX.loc39.5", "cX.loc40.5", "cX.loc41.5", "cX.loc42.5",
                          "cX.loc43.5", "cX.loc44.5", "cX.loc45.5", "cX.loc46.5", "cX.loc47.5", "cX.loc48.5",
                          "cX.loc49.5", "cX.loc50.5", "cX.loc51.5", "cX.loc52.5", "cX.loc53.5", "cX.loc54.5",
                          "cX.loc55.5", "cX.loc56.5", "cX.loc57.5", "DXMit186"), NULL))

    expect_equal(dr[["X"]][145:146,,], expected)

})

test_that("sim_geno works when multi-core", {
    skip_if(isnt_karl(), "this test only run locally")

    # can't really tell if I'm getting the same answers as w/o multi-core
    # but I can tell if I get the same thing twise, when run from same seed
    # (need to use RNGkind)

    library(qtl)
    data(hyper)
    hyper2 <- convert2cross2(hyper)

    RNGkind("L'Ecuyer-CMRG")
    set.seed(20151209)
    dr <- sim_geno(hyper2, n_draws=2, error_prob=0.002, cores=2)
    set.seed(20151209)
    dr_mc <- sim_geno(hyper2, n_draws=2, error_prob=0.002, cores=2)
    expect_equal(dr_mc, dr)

    data(listeria)
    listeria2 <- convert2cross2(listeria)
    set.seed(20151209)
    dr <- sim_geno(listeria2, n_draws=2, error_prob=0.002, cores=2)
    set.seed(20151209)
    dr_mc <- sim_geno(listeria2, n_draws=2, error_prob=0.002, cores=2)
    expect_equal(dr_mc, dr)

    # re-set RNGkind
    RNGkind("Mersenne-Twister")
})


test_that("sim_geno riself gives same result for lowmem=TRUE and =FALSE", {

    skip_on_cran()

    RNGkind("Mersenne-Twister")
    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    set.seed(20150918)
    dr <- sim_geno(grav2, map, n_draws=2, err=0.002, lowmem=FALSE)
    set.seed(20150918)
    dr2 <- sim_geno(grav2, map, n_draws=2, err=0.002, lowmem=TRUE)

    expect_equal(dr, dr2)

})

test_that("sim_geno f2 gives same result for lowmem=TRUE and =FALSE", {

    skip_on_cran()

    RNGkind("Mersenne-Twister")
    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    # order individuals to be sure that the get run in the same order
    p <- paste(iron$is_female, apply(iron$cross_info, 1, paste, collapse=":"), sep=":")
    iron <- iron[order(p),]

    map <- insert_pseudomarkers(iron$gmap, step=1)

    set.seed(20150918)
    dr <- sim_geno(iron, map, n_draws=2, err=0.002, lowmem=FALSE)
    set.seed(20150918)
    dr2 <- sim_geno(iron, map, n_draws=2, err=0.002, lowmem=TRUE)

    expect_equal(dr, dr2)

})

context("zip data files")

test_that("zip_datafiles() works", {

    skip_if(isnt_karl(), "this test only run locally")

    # put files in temp directory
    tmpdir <- file.path(tempdir(), "test_zip_datafiles")

    # grav2 zip file, unzip into temp directory
    orig_zip_file <- system.file("extdata", "grav2.zip", package="qtl2")
    unzip(orig_zip_file, exdir=tmpdir)

    # read original version
    grav2a <- read_cross2(orig_zip_file)

    # create new version
    new_zip_file <- zip_datafiles(file.path(tmpdir, "grav2.yaml"), overwrite=TRUE)

    # load newly created file
    grav2b <- read_cross2(new_zip_file)
    expect_equal(grav2a, grav2b)

    # zip into another directory
    other_dir <- file.path(tempdir(), "test_zip_file_again")
    if(!dir.exists(other_dir)) {
        dir.create(other_dir)
    }
    zip_file <- file.path(other_dir, "other_zip.zip")
    zip_datafiles(file.path(tmpdir, "grav2.yaml"), zip_file, overwrite=TRUE)
    grav2c <- read_cross2(zip_file)
    expect_equal(grav2a, grav2c)

    # clean up
    unlink(other_dir, recursive=TRUE)
    unlink(tmpdir, recursive=TRUE)

})

context("calc_genoprob")
suppressMessages(library(qtl))

grab_prob_rqtl <- function(cross) lapply(cross$geno, function(a) aperm(a$prob, c(1,3,2)))


test_that("backcross autosome calc_genoprob matches R/qtl", {

    data(hyper)
    chr <- c(1, 3, 4, 17, 19)
    hyper <- hyper[chr,]

    hyper <- calc.genoprob(hyper, err=0.002)
    pr <- grab_prob_rqtl(hyper)

    hyper2 <- convert2cross2(hyper)
    pr2 <- calc_genoprob(hyper2, error_prob=0.002)

    expect_equivalent(pr, pr2)

})

test_that("intercross autosome calc_genoprob matches R/qtl", {

    data(listeria)
    chr <- c(1, 4, 14, 18)
    listeria <- listeria[chr,]

    listeria <- calc.genoprob(listeria, step=1, stepwidth="max", err=0.01)
    pr <- grab_prob_rqtl(listeria)

    listeria2 <- convert2cross2(listeria)
    map <- insert_pseudomarkers(listeria2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(listeria2, map, error_prob=0.01)

    expect_equivalent(pr, pr2)

})


test_that("risib autosome calc_genoprob matches R/qtl", {

    data(hyper)
    chr <- c(1, 3, 4, 17, 19)
    hyper <- hyper[chr,]
    class(hyper)[1] <- "risib"

    hyper <- calc.genoprob(hyper, step=1, stepwidth="max", err=0.002)
    pr <- grab_prob_rqtl(hyper)

    hyper2 <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(hyper2, map, error_prob=0.002)

    expect_equivalent(pr, pr2)

})

test_that("riself autosome calc_genoprob matches R/qtl", {

    data(hyper)
    chr <- c(1, 3, 4, 17, 19)
    hyper <- hyper[chr,]
    class(hyper)[1] <- "riself"

    hyper <- calc.genoprob(hyper, step=1, stepwidth="max", err=0.002)
    pr <- grab_prob_rqtl(hyper)

    hyper2 <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(hyper2, map, error_prob=0.002)

    expect_equivalent(pr, pr2)

})

test_that("f2 X chr calc_genoprob matches R/qtl", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]

    fake.f2 <- calc.genoprob(fake.f2, step=1, stepwidth="max", error.prob=0.01)
    pr <- fake.f2$geno[[1]]$prob
    pr <- reviseXdata("f2", "full", getsex(fake.f2), prob=pr, cross.attr=attributes(fake.f2))
    pr <- aperm(pr, c(1,3,2))

    fake.f2.2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2.2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(fake.f2.2, map, error_prob=0.01)[[1]]

    expect_equivalent(pr, pr2)

})

test_that("bc X chr calc_genoprob matches R/qtl", {

    data(hyper)
    hyper <- hyper["X",]
    # make it half female, half male
    hyper$pheno$sex <- rep(c("female", "male"), nind(hyper)/2)

    hyper <- calc.genoprob(hyper, step=1, stepwidth="max", error.prob=0.02)
    pr <- hyper$geno[["X"]]$prob
    pr <- reviseXdata("bc", "full", getsex(hyper), prob=pr, cross.attr=attributes(hyper))
    pr <- aperm(pr, c(1,3,2))

    hyper2 <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(hyper2, map, error_prob=0.02)[[1]]

    expect_equivalent(pr, pr2)

})

test_that("f2 X chr all males calc_genoprob matches R/qtl", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 1

    fake.f2 <- calc.genoprob(fake.f2, step=1, stepwidth="max", error.prob=0.02)
    pr <- fake.f2$geno[["X"]]$prob
    pr <- aperm(pr, c(1,3,2))

    fake.f2.2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2.2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(fake.f2.2, map, error_prob=0.02)[[1]][,5:6,]

    expect_equivalent(pr, pr2)

})


test_that("f2 X chr all females calc_genoprob matches R/qtl", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 0

    fake.f2 <- calc.genoprob(fake.f2, step=1, stepwidth="max", error.prob=0.02)
    pr <- fake.f2$geno[["X"]]$prob
    pr <- aperm(pr, c(1,3,2))

    fake.f2.2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2.2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(fake.f2.2, map, error_prob=0.02)[[1]]
    pr2 <- pr2[,1:2,] + pr2[,4:3,] # recode as in R/qtl

    expect_equivalent(pr, pr2)

})

test_that("f2 X chr all females forw calc_genoprob matches R/qtl", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 0
    fake.f2$pheno$pgm <- 0

    fake.f2 <- calc.genoprob(fake.f2, step=1, stepwidth="max", error.prob=0.0001)
    pr <- fake.f2$geno[["X"]]$prob
    pr <- aperm(pr, c(1,3,2))

    fake.f2.2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2.2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(fake.f2.2, map, error_prob=0.0001)[[1]][,1:2,]

    expect_equivalent(pr, pr2)

})

test_that("f2 X chr all females rev calc_genoprob matches R/qtl", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 0
    fake.f2$pheno$pgm <- 1
    fake.f2 <- calc.genoprob(fake.f2, step=1, stepwidth="max", error.prob=0.0001)
    pr <- fake.f2$geno[["X"]]$prob
    pr <- aperm(pr, c(1,3,2))

    fake.f2.2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2.2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(fake.f2.2, map, error_prob=0.0001)[[1]][,4:3,]

    expect_equivalent(pr, pr2)

})

test_that("bc X chr all males calc_genoprob matches R/qtl", {

    data(hyper)
    hyper <- hyper["X",]
    sexpgm <- getsex(hyper)

    hyper <- calc.genoprob(hyper, step=1, stepwidth="max", error.prob=0.02)
    pr <- hyper$geno[["X"]]$prob
    pr <- reviseXdata("bc", "full", getsex(hyper), prob=pr, cross.attr=attributes(hyper))
    pr <- aperm(pr, c(1,3,2))

    hyper2 <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(hyper2, map, error_prob=0.02)[[1]][,3:4,]

    expect_equivalent(pr, pr2)

})


test_that("bc X chr all females calc_genoprob matches R/qtl", {

    data(hyper)
    hyper <- hyper["X",]
    hyper$pheno$sex <- "female"

    hyper <- calc.genoprob(hyper, step=1, stepwidth="max", error.prob=0.02)
    pr <- hyper$geno[["X"]]$prob
    pr <- reviseXdata("bc", "full", getsex(hyper), prob=pr, cross.attr=attributes(hyper))
    pr <- aperm(pr, c(1,3,2))

    hyper2 <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(hyper2, map, error_prob=0.02)[[1]]
    pr2 <- pr2[,1:2,] + pr2[,3:4,]

    expect_equivalent(pr, pr2)

})

test_that("doubled haploids calc_genoprob matches R/qtl", {

    data(hyper)
    hyper <- hyper[1,]
    hyper$pheno <- hyper$pheno[,1,drop=FALSE]
    class(hyper)[1] <- "dh"

    hyper <- calc.genoprob(hyper, step=1, stepwidth="max", error.prob=0.02)
    pr <- hyper$geno[[1]]$prob
    pr <- aperm(pr, c(1,3,2))

    hyper2 <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(hyper2, map, error_prob=0.02)[[1]]

    expect_equivalent(pr, pr2)

})


test_that("haploids calc_genoprob matches R/qtl", {

    data(hyper)
    hyper <- hyper[2,]
    hyper$pheno <- hyper$pheno[,1,drop=FALSE]
    class(hyper)[1] <- "haploid"

    hyper <- calc.genoprob(hyper, step=1, stepwidth="max", error.prob=0.02)
    pr <- hyper$geno[[1]]$prob
    pr <- aperm(pr, c(1,3,2))

    hyper2 <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper2$gmap, step=1, stepwidth="max")
    pr2 <- calc_genoprob(hyper2, map, error_prob=0.02)[[1]]

    expect_equivalent(pr, pr2)

})

test_that("backcross autosome calc_genoprob with markers at same location", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    grav2 <- grav2[1:4,1]

    # put some markers at same location
    grav2$gmap[[1]][4] <- grav2$gmap[[1]][3]

    pr <- calc_genoprob(grav2, err=0)

    expect_true( all(!is.na(pr[[1]])) )

})

test_that("calc_genoprob works when multi-core", {
    skip_if(isnt_karl(), "this test only run locally")

    data(hyper)
    hyper2 <- convert2cross2(hyper)
    pr <- calc_genoprob(hyper2, error_prob=0.002)
    pr_mc <- calc_genoprob(hyper2, error_prob=0.002, cores=2)
    expect_equal(pr_mc, pr)


    data(listeria)
    listeria2 <- convert2cross2(listeria)
    map <- insert_pseudomarkers(listeria2$gmap, step=1, stepwidth="max")
    pr <- calc_genoprob(listeria2, map, error_prob=0.01)
    pr_mc <- calc_genoprob(listeria2, map, error_prob=0.01, cores=2)

    expect_equal(pr_mc, pr)

})

context("create snpinfo from cross2 object")

test_that("create_snpinfo works", {

    skip_if(isnt_karl(), "this test only run locally")

    file <- paste0("https://raw.githubusercontent.com/rqtl/",
                   "qtl2data/main/DOex/DOex.zip")

    DOex <- read_cross2(file)

    snpinfo <- create_snpinfo(DOex)

    fg <- do.call("cbind", DOex$founder_geno)

    expect_equal(sum(colSums(fg==0)==0), nrow(snpinfo))

    n <- sapply(DOex$founder_geno, function(a) sum(colSums(a==0)==0))
    expect_equivalent(unclass(table(snpinfo$chr)), n)

    expect_equivalent(snpinfo$sdp, calc_sdp(t(fg[,colSums(fg==0)==0])))

})


test_that("create_snpinfo works for arab data", {

    skip_if(isnt_karl(), "this test only run locally")

    file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/ArabMAGIC/arabmagic_tair9.zip")
    arab <- read_cross2(file)

    snpinfo <- create_snpinfo(arab)

    # expected markers, dropping non-informative or with missing founder genotype
    tmp <- drop_nullmarkers(arab)
    fg <- do.call("cbind", tmp$founder_geno)
    fg <- fg[,colSums(is.na(fg) | fg==0) == 0]

    expect_equal(nrow(snpinfo), ncol(fg))
    expect_equal(rownames(snpinfo), colnames(fg))
})

#' @export
# dimensions of calc_genoprob object
dim.calc_genoprob <-
    function(x)
{
  vapply(x, dim, rep(1,3))
}

#' @export
# dimnames of calc_genoprob object
dimnames.calc_genoprob <-
    function(x)
{
  dnames <- lapply(x, dimnames)

  list(ind = dnames[[1]][[1]],
       gen = lapply(dnames, '[[', 2),
       mar = lapply(dnames, '[[', 3))
}

#' Calculate heterozygosities
#'
#' Calculate heterozygosites, by individual or by marker
#'
#' @param probs List of arrays of genotype probabilities, as
#'     calculated by [calc_genoprob()].
#' @param by Whether to summarize by individual or marker
#'
#' @param omit_x If TRUE, omit the X chromosome.
#'
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @details `calc_het()` looks at the genotype names (the 2nd
#' dimension of the dimnames of the input `probs`), which must be
#' two-letter names, and assumes that when the two letters are
#' different it's a heterozygous genotype while if they're the
#' same it's a homozygous genotype
#'
#' @seealso [calc_raw_het()], [calc_geno_freq()]
#'
#' @return
#' The result is a vector of estimated heterozygosities
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' p <- calc_genoprob(iron, err=0.002)
#'
#' # heterozygosities by individual
#' het_ind <- calc_het(p)
#'
#' # heterozygosities by marker
#' het_mar <- calc_het(p, "marker")
#'
#' @export
calc_het <-
    function(probs, by=c("individual", "marker"), omit_x=TRUE, cores=1)
{
    by <- match.arg(by)
    if(is.null(probs)) stop("probs is NULL")
    if(is.cross2(probs))
        stop('Input probs is a "cross2" object but should be genotype probabilities, as from calc_genoprob')

    aprob_attr <- attr(probs, "alleleprobs")
    if(!is.null(aprob_attr) && aprob_attr) {
        stop("Input probs should not be allele dosages")
    }

    # set up cluster
    cores <- setup_cluster(cores, quiet=TRUE)

    is_x_chr <- attr(probs, "is_x_chr")
    if(any(is_x_chr) && !all(is_x_chr) && omit_x) {
        probs <- probs[,!is_x_chr]
        is_x_chr <- is_x_chr[!is_x_chr]
        omit_x <- FALSE
    }

    n_chr <- length(probs)

    # determine which columns are het
    het_col <- vector("list", n_chr)
    geno <- dimnames(probs)[[2]]

    if(any(nchar(unlist(geno)) != 2)) {
        stop("calc_het requires genotypes to have two-letter names")
    }

    for(chr in seq_len(n_chr)) {
        a1 <- substr(geno[[chr]], 1, 1)
        a2 <- substr(geno[[chr]], 2, 2)
        if(is_x_chr[chr]) het_col[[chr]] <- (a1 != a2 & a2 != "Y")
        else het_col[[chr]] <- (a1 != a2)
    }

    if(by=="individual") {
        # total markers
        total_mar <- sum(dim(probs)[3,])

        # summarize each chromosome
        result <- cluster_lapply(cores, seq_len(n_chr), function(chr) apply(probs[[chr]][,het_col[[chr]],,drop=FALSE], 1, sum))

        if(length(result)>1) {
            for(i in seq_along(result)[-1])
                result[[1]] <- result[[1]] + result[[i]]
        }
        return(result[[1]]/total_mar)
    }

    # else: by marker
    unlist(cluster_lapply(cores, seq_len(n_chr), function(chr) apply(probs[[chr]][,het_col[[chr]],,drop=FALSE], 3, sum)/
                                                          nrow(probs[[chr]])))
}

# snpprob_from_cross: genoprob_to_snpprob where snpinfo is a cross2 object
# ...construct snpinfo for all markers with no missing founder genotypes
# ...apply genoprob_to_snpprob with that

snpprob_from_cross <-
    function(genoprobs, cross)
{
    if(!("founder_geno" %in% names(cross)))
        stop("cross doesn't contain founder_geno")

    ### calculate snp info from cross
    # drop markers that are missing any founder genotypes or that are monomorphic
    fg <- do.call("cbind", cross$founder_geno)
    mar2drop <- colnames(fg)[colSums(fg != 1 & fg != 3) > 0 | colSums(fg==1)==0 | colSums(fg==3)==0]
    cross <- drop_markers(cross, mar2drop)

    # subset to common chr
    chr_pr <- names(genoprobs)
    chr_cross <- chr_names(cross)
    chr <- chr_pr[chr_pr %in% chr_cross]
    cross <- cross[,chr]
    genoprobs <- genoprobs[,chr]

    # get physical map, if available; otherwise genetic map
    map <- cross$pmap
    if(is.null(map)) map <- cross$gmap

    if(cross$crosstype %in% c("hsf1", "dof1")) { # if HSF1 or DOF1, drop last founder
        cross$founder_geno <- cross$founder_geno[,-ncol(cross$founder_geno)]
    }

    sdp <- calc_sdp(t(do.call("cbind", cross$founder_geno)))

    snpinfo <- data.frame(chr=rep(chr_names(cross), n_mar(cross)),
                          pos=unlist(map),
                          sdp=sdp,
                          snp=names(sdp))

    mar2keep <- NULL
    for(i in seq_along(genoprobs)) {
        mar <- colnames(cross$geno[[i]])
        mar <- mar[mar %in% dimnames(genoprobs)[[3]][[i]]]
        if(length(mar) == 0) {
            stop("No overlap between genoprobs and markers on chr ", names(genoprobs)[i])
        }
        map[[i]] <- map[[i]][mar]

        if(inherits(genoprobs, "fst_genoprob")) {
            mar2keep <- c(mar2keep, mar)
        } else {
            genoprobs[[i]] <- genoprobs[[i]][,,mar,drop=FALSE]
        }
    }
    if(inherits(genoprobs, "fst_genoprob")) {
        genoprobs <- subset(genoprobs, mar=mar2keep)
    }

    snpinfo <- index_snps(map, snpinfo)

    genoprob_to_snpprob(genoprobs, snpinfo)
}

# function for subsetting by individual
#
# ind = logical, numeric, or character strings
# all_ind = vector of character strings with all individual names
#
# result = vector of character strings (the selected ones)
subset_ind <-
    function(ind, all_ind, allow_logical=TRUE)
{
    if(is.logical(ind)) {
        if(!allow_logical) stop("ind can't be logical if different individuals in geno, pheno, covar")

        if(length(ind) != length(all_ind))
            stop("ind is logical but length [", length(ind), "] != no. ind in cross [",
                 length(all_ind), "]")
        ind <- all_ind[ind]
    }
    if(is.numeric(ind)) { # treat as numeric indexes
        if(!allow_logical) stop("ind can't be numeric if different individuals in geno, pheno, covar")
        if(any(ind < 0)) { # deal with negatives
            if(!all(ind < 0)) stop("Can't mix negative and positive ind subscripts")
                ind <- (seq_along(all_ind))[ind]
        }

        if(any(ind < 1 | ind > length(all_ind))) {
            ind <- ind[ind >= 1 & ind <= length(all_ind)]
            if(length(ind)==0) stop("All ind out of range")
            warning("some ind out of range [1,", length(all_ind), "]")
        }

        ind <- all_ind[ind]
    }

    # now treat ind as character strings
    ind <- as.character(ind)

    # look for negatives; turn to positives
    if(!any(grepl("^\\-", all_ind)) && any(grepl("^\\-", ind))) { # if "-" used in actual IDs, don't allow negative subscripts
        if(!all(grepl("^\\-", ind)))
            stop("Can't mix negative and positive ind subscripts")
        ind <- sub("^\\-", "", ind)
        if(!all(ind %in% all_ind)) {
            if(!any(ind %in% all_ind)) stop("None of the individuals in cross")
            warning("Some individuals not in cross: ", paste(ind[!(ind %in% all_ind)], collapse=", "))
        }
        ind <- all_ind[!(all_ind %in% ind)]
    }

    if(!all(ind %in% all_ind)) {
        if(!any(ind %in% all_ind)) stop("None of the individuals in cross")
        warning("Some ind not in cross: ", paste(ind[!(ind %in% all_ind)], collapse=", "))
        ind <- ind[ind %in% all_ind]
    }

    ind
}

#' Pull genotype probabilities for an interval
#'
#' Pull out the genotype probabilities for a given genomic interval
#'
#' @param genoprobs Genotype probabilities as calculated by
#' [calc_genoprob()].
#' @param map The marker map for the genotype probabilities
#' @param chr Chromosome ID (single character sting)
#' @param interval Interval (pair of numbers)
#'
#' @return A list containing a single 3d array of genotype probabilities, like the input `genoprobs`
#' but for the designated interval.
#'
#' @seealso [find_marker()], [pull_genoprobpos()]
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,"8"]}
#' gmap <- insert_pseudomarkers(iron$gmap, step=1)
#' pr <- calc_genoprob(iron, gmap, error_prob=0.002)
#'
#' pr_sub <- pull_genoprobint(pr, gmap, "8", c(25, 35))
#'
#' @export
pull_genoprobint <-
    function(genoprobs, map, chr, interval)
{
    if(is.null(genoprobs)) stop("genoprobs is NULL")

    # get vector of marker names that will be kept
    markers <- find_marker(map, chr, interval=interval)
    if(length(markers)==0) stop("No markers/pseudomarkers in the interval")

    # reduce to the one chromosome
    genoprobs <- genoprobs[,chr]

    # reduce to the markers in the interval
    genoprobs[[1]] <- genoprobs[[1]][,,markers,drop=FALSE]

    genoprobs
}

context("pull genotype probabilities for an interval")

test_that("pull_genoprobint works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(2,8,"X")]
    gmap <- insert_pseudomarkers(iron$gmap, step=1)
    pr <- calc_genoprob(iron, gmap, error_prob=0.002)

    pr_sub <- pull_genoprobint(pr, gmap, "8", c(25, 30))

    expected <- pr[,"8"]
    gm8 <- gmap[["8"]]
    markers <- names(gm8)[gm8 >= 25 & gm8 <= 30]
    expected[[1]] <- expected[[1]][,,markers,drop=FALSE]

    expect_equal(pr_sub, expected)

})

# utility for grabbing "..." args
#
# dotargs = list(...) from function call
# argname = character string of the argument to grab
# default = default value for argument
# values  = optional vector of character strings with possible values
grab_dots <-
    function(dotargs, argname, default, values=NULL)
{
    if(argname %in% names(dotargs)) {
        arg <- dotargs[[argname]]
        if(!is.null(values) && !(arg %in% values)) {
            warning(argname, ' "', arg, '" not valid; using "',
                    default, '".')
            arg <- default
        }
    }
    else arg <- default
    arg
}

# give warning if there are extra named arguments
#
# dotargs = list(...) from function call
# expected = vector of character strings of the names of possible args anticipated
check_extra_dots <-
    function(dotargs, expected)
{
    args <- names(dotargs)
    if(!all(args %in% expected)) {
        extra <- args[!(args %in% expected)]
        if(length(extra) == 1)
            warning("Extra argument ignored: ", extra)
        else
            warning("Extra arguments ignored: ",
                    paste(extra, collapse=" "))
    }
}

# check that two objects are the same
is_same <-
    function(a, b)
{
    if(is.null(a) && is.null(b)) return(TRUE)
    if(is.null(a) || is.null(b)) return(FALSE)

    if(is.list(a) && is.list(b)) {
        if(length(a) != length(b)) return(FALSE)
        for(i in seq_along(a))
            if(!is_same(a[[i]], b[[i]])) return(FALSE)
        return(TRUE)
    }
    if(is.list(a) || is.list(b)) return(FALSE)

    attra <- attributes(a)
    attrb <- attributes(b)

    if(length(a) != length(b)) return(FALSE)
    if(!all((is.na(a) & is.na(b)) |
            !is.na(a) & !is.na(b) & a==b)) return(FALSE)

    if(is.null(names(attra))) {
        if(!is.null(names(attrb))) return(FALSE)
    }
    else {
        if(length(attra) != length(attrb)) return(FALSE)
        attra <- attrb[order(names(attra))]
        attrb <- attrb[order(names(attrb))]
        if(!all(names(attra) == names(attrb))) return(FALSE)
        nam <- names(attra)
        nam <- nam[nam != "names"]
        for(i in names(attra)) {
            if(!is_same(attra[[i]], attrb[[i]])) return(FALSE)
        }
    }

    TRUE
}

context("is_same for comparing objects")

test_that("is_same works for calc_genoprob results", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2[6:12,], map, error_prob=0.002)

    expect_true(is_same(probs, probs))

    for(i in seq(along=probs))
        expect_true(is_same(probs[[i]], probs[[i]]))

    expect_false(is_same(probs, probs[1:2,]))
})

#' Clean genotype probabilities
#'
#' Clean up genotype probabilities by setting small values to 0 and
#' for a genotype column where the maximum value is rather small, set
#' all values in that column to 0.
#'
#' @param object Genotype probabilities as calculated by
#'     [calc_genoprob()].
#' @param value_threshold Probabilities below this value will be set
#'     to 0.
#' @param column_threshold For genotype columns where the maximum
#'     value is below this threshold, all values will be set to 0.
#'     This must be less than \eqn{1/k} where \eqn{k} is the number of
#'     genotypes.
#' @param ind Optional vector of individuals (logical, numeric, or
#'     character). If provided, only the genotype probabilities for
#'     these individuals will be cleaned, though the full set will be
#'     returned.
#' @param cores Number of CPU cores to use, for parallel calculations.
#'     (If `0`, use [parallel::detectCores()].)
#'     Alternatively, this can be links to a set of cluster sockets, as
#'     produced by [parallel::makeCluster()].
#' @param ... Ignored at this point.
#'
#' @return A cleaned version of the input genotype probabilities
#'     object, `object`.
#'
#' @details
#' In cases where a particular genotype is largely absent,
#' `scan1coef()` and `fit1()` can give unstable estimates of the
#' genotype effects. Cleaning up the genotype probabilities by setting
#' small values to 0 helps to ensure that such effects get set to
#' `NA`.
#'
#' At each position and for each genotype column, we find the maximum
#' probability across individuals. If that maximum is <
#' `column_threshold`, all values in that genotype column at that
#' position are set to 0.
#'
#' In addition, any genotype probabilities that are < `value_threshold`
#' (generally < `column_threshold`) are set to 0.
#'
#' The probabilities are then re-scaled so that the probabilities for
#' each individual at each position sum to 1.
#'
#' If `ind` is provided, the function is applied only to the
#' designated subset of individuals. This may be useful when only a
#' subset of individuals have been phenotyped, as you may want to zero
#' out genotype columns where that subset of individuals has only
#' negligible probability values.
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c("19", "X")] # subset to chr 19 and X}
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, error_prob=0.002)
#'
#' # clean the genotype probabilities
#' # (doesn't really do anything in this case, because there are no small but non-zero values)
#' probs_clean <- clean(probs)
#'
#' # clean only the females' genotype probabilities
#' probs_cleanf <- clean(probs, ind=names(iron$is_female)[iron$is_female])

#' @export
clean_genoprob <-
    function(object, value_threshold=1e-6, column_threshold=0.01, ind=NULL, cores=1, ...)
{
    if(!is.null(ind)) {
        # clean the selected subset of individuals
        probs_sub <- clean_genoprob(subset(object, ind=ind), value_threshold=value_threshold,
                                    column_threshold=column_threshold, cores=cores)

        # paste the result into the main object
        for(i in seq_along(object))
            object[[i]][rownames(probs_sub[[i]]),,] <- probs_sub[[i]]

        # exit the function
        return(object)
    }

    attrib <- attributes(object)

    cores <- setup_cluster(cores)

    result <- cluster_lapply(cores, seq_along(object),
                             function(i) {
        this_result <- .clean_genoprob(object[[i]], value_threshold, column_threshold)
        dimnames(this_result) <- dimnames(object[[i]])
        this_result })

    for(a in names(attrib)) {
        attr(result, a) <- attrib[[a]]
    }

    result
}

#' @rdname clean_genoprob
#' @export
clean.calc_genoprob <- clean_genoprob

context("Clean genotype probabilities")

test_that("clean_genoprob works", {

    # read data
    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c("19", "X")]

    map <- insert_pseudomarkers(iron$gmap, step=1)
    pr <- calc_genoprob(iron, map, error_prob=0.002, map_function="c-f")

    # there are no *really* small values here
    pr_clean <- clean_genoprob(pr)
    expect_equal(pr_clean, pr)

    # use a bigger tolerance for the individual values
    pr_clean <- clean_genoprob(pr, 0.001, 0.01)

    for(i in seq_along(pr_clean)) {
        # ensure that things still sum to 1
        expect_true( max(abs(apply(pr_clean[[i]], c(1,3), sum) - 1)) < 1e-12 )
    }

    # create a column that is largely missing
    pr[[1]][,3,20] <- runif(nrow(pr[[1]]), 0, 1e-4)
    # re-scale so values sum to 1
    pr[[1]][,,20] <- pr[[1]][,,20] / rowSums(pr[[1]][,,20])

    pr_clean <- clean_genoprob(pr, column_threshold=0.01)
    expect_true( all(pr_clean[[1]][,3,20] == 0) )
    expect_equal( pr_clean[[1]][,,-20], pr[[1]][,,-20] )
    expect_equal( pr_clean[[2]], pr[[2]] )

    expect_true( max(abs(apply(pr_clean[[1]], c(1,3), sum) - 1)) < 1e-12 )

    # try this with a subset of individuals
    # create another column that is largely missing, for a subset of individuals
    set.seed(20180223)
    ind <- sample(ind_ids(iron), 50)
    pr[[1]][ind,1,15] <- runif(length(ind), 0, 1e-4)
    # re-scale so values sum to 1
    pr[[1]][,,15] <- pr[[1]][,,15] / rowSums(pr[[1]][,,15])


    pr_clean <- clean_genoprob(pr, column_threshold=0.01, ind=ind)
    other_ind <- ind_ids(iron)[!(ind_ids(iron) %in% ind)]
    expect_equal(pr_clean[other_ind,], pr[other_ind,])
    expect_true( max(abs(apply(pr_clean[[1]], c(1,3), sum) - 1)) < 1e-12 )
    expect_true( all(pr_clean[[1]][ind,1,15] == 0) )

    expect_equal( pr_clean[[1]][,,-c(15,20)], pr[[1]][,,-c(15,20)] )
    expect_equal( pr_clean[[2]], pr[[2]] )
})

# match ids
#
# takes two vectors of character strings
# returns a list with
#   first,second = indexes in both vectors, to make them aligned
#   firstonly = indexes in first vector that are not in the second
#   secondonly = indexes in second vector that are not in the first
match_ids <-
    function(x,y)
{
    mxy <- match(x, y)
    myx <- match(y, x)

    result <- list(first=which(!is.na(mxy)),
                   second=mxy[!is.na(mxy)],
                   firstonly=which(is.na(mxy)),
                   secondonly=which(is.na(myx)))

    if(length(result$first) > 0)
        names(result$first) <- x[result$first]
    if(length(result$firstonly) > 0)
        names(result$firstonly) <- x[result$firstonly]
    if(length(result$second) > 0)
        names(result$second) <- y[result$second]
    if(length(result$secondonly) > 0)
        names(result$secondonly) <- y[result$secondonly]

    result
}

# find common ids across a set of vectors of character string
# return value: ids that are present in *all* of the inputs
find_common_ids <-
    function(...)
{
    inp <- list(...)
    n_inp <- length(inp)

    if(n_inp==0)
        stop("No input")

    if(n_inp==1) {
        if(length(inp[[1]])==0) return(character(0))
        return(inp[[1]])
    }

    reduced <- inp[[1]]
    for(i in 2:n_inp) {
        m <- match_ids(reduced, inp[[i]])
        reduced <- reduced[m$first]
    }

    if(length(reduced)==0) return(character(0))

    reduced
}

# align genotypes, is_female and cross_info
# (because we'll do this alot)
align_geno_sex_cross <-
    function(geno, is_female, cross_info, hold_off_warning=FALSE)
{
    if(is.list(geno)) {
        give_warning <- FALSE
        dropped_ind <- NULL
        for(i in seq(along=geno)) {
            orig_nind <- nrow(geno[[i]])
            orig_ind <- rownames(geno[[i]])
            tmp <- align_geno_sex_cross(geno[[i]],
                                        is_female,
                                        cross_info,
                                        hold_off_warning=TRUE)
            if(orig_nind > nrow(tmp$geno)) {
                new_ind <- rownames(tmp$geno)
                dropped_ind <- unique(c(dropped_ind,
                                        orig_ind[!(orig_ind %in% new_ind)]))
                give_warning <- TRUE
            }
            is_female <- tmp$is_female
            cross_info <- tmp$cross_info
            geno[[i]] <- tmp$geno
        }
        if(give_warning)
            warning("Omitting genotypes for ", length(dropped_ind),
                    ifelse(length(dropped_ind)==1, " individual", " individuals"),
                    " with no sex/cross info.")

        return(list(geno=geno, is_female=is_female, cross_info=cross_info))
    }

    ind <- rownames(geno)
    is_female <- handle_null_isfemale(is_female, ind)
    cross_info <- handle_null_crossinfo(cross_info, ind)

    keep <- find_common_ids(rownames(geno), names(is_female), rownames(cross_info))

    if(length(keep) < length(ind) && !hold_off_warning)
        warning("Omitting genotypes for ", length(ind) - length(keep),
                ifelse(length(ind)-length(keep)==1, " individual", " individuals"),
                " with no sex/cross info.")

    list(geno=geno[keep,,drop=FALSE],
         is_female=is_female[keep],
         cross_info=cross_info[keep,,drop=FALSE])
}

context("find_common_ids and match_ids")

test_that("match_ids and find_common_ids work", {

    set.seed(20151006)
    first <- LETTERS[1:10]
    second <- sample(LETTERS[c(1:4,7:12)])

    m12 <- match(first, second)
    expected <- list(first=c(1:4, 7:10),
                     second=m12[!is.na(m12)],
                     firstonly=5:6,
                     secondonly=which(second %in% LETTERS[11:12]))
    names(expected$first) <- first[expected$first]
    names(expected$second) <- second[expected$second]
    names(expected$firstonly) <- first[expected$firstonly]
    names(expected$secondonly) <- second[expected$secondonly]

    expect_equal(match_ids(first, second), expected)
    expect_equal(first[expected$first], second[expected$second])

    expect_equal(find_common_ids(first, second), first[expected$first])

})

test_that("match_ids and find_common_ids work with no overlap", {

    set.seed(20151006)
    first <- LETTERS[1:10]
    second <- LETTERS[11:20]

    expected <- list(first=numeric(0),
                     second=numeric(0),
                     firstonly=1:10,
                     secondonly=1:10)
    names(expected$firstonly) <- first
    names(expected$secondonly) <- second

    expect_equal(match_ids(first, second), expected)
    expect_equal(first[expected$first], second[expected$second])

    expect_equal(find_common_ids(first, second), first[expected$first])

})

test_that("match_ids and find_common_ids work with null vectors", {

    set.seed(20151006)
    first <- LETTERS[1:10]
    second <- character(0)

    expected <- list(first=numeric(0),
                     second=numeric(0),
                     firstonly=1:10,
                     secondonly=numeric(0))
    names(expected$firstonly) <- first

    expect_equal(match_ids(first, second), expected)
    expect_equal(first[expected$first], second[expected$second])
    expect_equal(find_common_ids(first, second), first[expected$first])

    # switch the role of the two
    expected$secondonly <- expected$firstonly
    expected$firstonly <- numeric(0)
    expect_equal(match_ids(second, first), expected)
    expect_equal(second[expected$first], first[expected$second])
    expect_equal(find_common_ids(second, first), second[expected$first])

    # same, when second is NULL
    second <- NULL

    expected <- list(first=numeric(0),
                     second=numeric(0),
                     firstonly=1:10,
                     secondonly=numeric(0))
    names(expected$firstonly) <- first

    expect_equal(match_ids(first, second), expected)
    expect_equal(find_common_ids(first, second), first[expected$first])

    # switch the role of the two (when second is NULL)
    expected$secondonly <- expected$firstonly
    expected$firstonly <- numeric(0)
    expect_equal(match_ids(second, first), expected)

    # both NULL
    expected <- list(first=numeric(0),
                     second=numeric(0),
                     firstonly=numeric(0),
                     secondonly=numeric(0))
    expect_equal(match_ids(NULL, NULL), expected)
    expect_equal(find_common_ids(NULL, NULL), character(0))

    expect_equal(match_ids(character(0), character(0)), expected)
    expect_equal(find_common_ids(character(0), character(0)), character(0))

})


test_that("find_common_ids with <2 inputs", {

    expect_error(find_common_ids())

    expect_equal(find_common_ids(LETTERS[1:5]), LETTERS[1:5])
    expect_equal(find_common_ids(NULL), character(0))
    expect_equal(find_common_ids(character(0)), character(0))

})


test_that("find_common_ids with >2 inputs", {

    set.seed(20151008)
    first <- LETTERS[1:10]
    second <- sample(LETTERS[3:12])
    third <- sample(LETTERS[5:14])

    expect_equal(find_common_ids(first, second, third), LETTERS[5:10])
    expect_equal(sort(find_common_ids(second, first, third)), LETTERS[5:10])
    expect_equal(sort(find_common_ids(third, second, first)), LETTERS[5:10])

    expect_equal(find_common_ids(third, second, NULL, first), character(0))

    fourth <- sample(c("H", "E", "L", "P"))
    expect_equal(sort(find_common_ids(third, second, fourth, first)), c("E", "H"))

})

#' Get common set of IDs from objects
#'
#' For a set objects with IDs as row names (or, for a vector, just
#' names), find the IDs that are present in all of the objects.
#'
#' @param ... A set of objects: vectors, lists, matrices, data frames,
#' and/or arrays. If one is a character vector with no names
#' attribute, it's taken to be a set of IDs, itself.
#' @param complete.cases If TRUE, look at matrices and non-character
#' vectors and keep only individuals with no missing values.
#'
#' @return A vector of character strings for the individuals that are
#' in common.
#'
#' @details This is used (mostly internally) to align phenotypes,
#' genotype probabilities, and covariates in preparation for a genome
#' scan. The `complete.cases` argument is used to omit
#' individuals with any missing covariate values.
#'
#' @examples
#' x <- matrix(0, nrow=10, ncol=5); rownames(x) <- LETTERS[1:10]
#' y <- matrix(0, nrow=5, ncol=5);  rownames(y) <- LETTERS[(1:5)+7]
#' z <- LETTERS[5:15]
#' get_common_ids(x, y, z)
#'
#' x[8,1] <- NA
#' get_common_ids(x, y, z)
#' get_common_ids(x, y, z, complete.cases=TRUE)
#'
#' @export
get_common_ids <-
    function(..., complete.cases=FALSE)
{
    args <- list(...)
    if(length(args)==0) {
        return(character(0))
    }

    # find the IDs in common across all
    id <- NULL
    for(i in seq_along(args)) {
        if(is.null(args[[i]])) next

        if(is.matrix(args[[i]]) || is.data.frame(args[[i]]) || is.array(args[[i]])) {
            if(length(dim(args[[i]])) > 3)
                stop("Can't handle arrays with >3 dimensions")
            these <- rownames(args[[i]])
            if(complete.cases && (is.matrix(args[[i]]) || is.data.frame(args[[i]])))
                these <- these[rowSums(!is.finite(args[[i]]))==0]
        }
        else if(is.list(args[[i]]) && !is.null(rownames(args[[i]][[1]]))) {
            these <- rownames(args[[i]][[1]])
        }
        else if(is.vector(args[[i]])) {
            if(is.character(args[[i]])) {
                if(is.null(names(args[[i]]))) {
                    these <- args[[i]]
                } else {
                    these <- names(args[[i]])
                    if(complete.cases) {
                        these <- these[!is.na(args[[i]])]
                    }
                }
            }
            else {
                these <- names(args[[i]])
                if(complete.cases) {
                    these <- these[is.finite(args[[i]])]
                }
            }
        }
        else if(is.character(args[[i]])) { # character but not vector
            if(is.null(names(args[[i]]))) {
                these <- args[[i]]
            } else {
                these <- names(args[[i]])
                if(complete.cases) {
                    these <- these[!is.na(args[[i]])]
                }
            }
        }
        else if(!is.null(names(args[[i]]))) { # not a vector but has names
            these <- names(args[[i]])
        }
        else {
            stop("Not sure what to do with object of class ", class(args[[i]]))
        }

        if(length(unique(these)) != length(these))
            stop("Duplicate names in argument ", i)

        if(is.null(id)) id <- these
        else id <- id[id %in% these]
    }

    names(id) <- NULL # just in case; should be bare vector of character strings
    id
}

context("get common ids from set of objects")

test_that("get_common_ids works", {

    x <- matrix(1:20, nrow=5, ncol=4)
    rownames(x) <- LETTERS[1:5]

    y <- 1:5
    names(y) <- c("A", "C", "D", "E", "F")

    z <- array(1:(5*2*3), dim=c(5, 2, 3))
    dimnames(z) <- list(c("G", "C", "D", "G", "A"),
                        c("col1", "col2"),
                        c("depth1", "depth2", "depth3"))

    # error due to duplicate name
    expect_error(get_common_ids(x, y, z))
    expect_error(get_common_ids(z, y, x))
    expect_error(get_common_ids(x, z, y))

    rownames(z)[4] <- "H" # fix dup
    expect_equal(get_common_ids(x, y, z), c("A", "C", "D"))
    expect_equal(get_common_ids(x, z, y), c("A", "C", "D"))
    expect_equal(get_common_ids(z, y, x), c("C", "D", "A"))

    rownames(z)[4] <- "H" # fix dup
    chvec <- c("A", "I", "D", "E")
    expect_equal(get_common_ids(x, y, z, chvec), c("A", "D"))
    expect_equal(get_common_ids(x, z, chvec, y), c("A", "D"))
    expect_equal(get_common_ids(z, chvec, y, x), c("D", "A"))

    # try data.frame
    x <- as.data.frame(x)
    rownames(z)[4] <- "G"
    expect_error(get_common_ids(x, y, z)) # duplicate names
    expect_error(get_common_ids(z, y, x)) # duplicate names
    expect_error(get_common_ids(x, z, y)) # duplicate names

    rownames(z)[4] <- "H"
    expect_equal(get_common_ids(x, y, z), c("A", "C", "D"))
    expect_equal(get_common_ids(x, z, y), c("A", "C", "D"))
    expect_equal(get_common_ids(z, y, x), c("C", "D", "A"))

})

test_that("get_common_ids works when complete.cases=TRUE", {

    n <- 10
    x <- matrix(rnorm(10*n), ncol=10)
    rownames(x) <- LETTERS[1:10]

    y <- matrix(rnorm(5*(n-1)), ncol=5)
    rownames(y) <- LETTERS[(1:(n-1))+2]

    z <- LETTERS[c(3:8, 15:18)]

    expected <- LETTERS[3:8]
    expect_equal(get_common_ids(x, y, z, complete.cases=TRUE), expected)
    expect_equal(get_common_ids(y, x, z, complete.cases=TRUE), expected)
    expect_equal(get_common_ids(z, y, x, complete.cases=TRUE), expected)

    # add some missing values
    x[1,c(5,7)] <- NA    # A
    x[3,c(2,8,9)] <- NA  # C
    x[5,4] <- NA         # E
    y[3,5] <- NA         # C
    y[6,] <- NA          # H

    expected <- c("D", "F", "G")
    expect_equal(get_common_ids(x, y, z, complete.cases=TRUE), expected)
    expect_equal(get_common_ids(y, x, z, complete.cases=TRUE), expected)
    expect_equal(get_common_ids(z, y, x, complete.cases=TRUE), expected)

})

#' Join genotype imputations for different individuals
#'
#' Join multiple genotype imputation objects, as produced by
#' [sim_geno()], for the same set of markers but for different
#' individuals.
#'
#' @param ... Genotype imputation objects as produced by
#' [sim_geno()]. Must have the same set of markers and
#' genotypes.
#'
#' @return An object of class `"sim_geno"`, like the input; see [sim_geno()].
#'
#' @seealso [cbind.sim_geno()], [sim_geno()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' drawsA <- sim_geno(grav2[1:5,], map, error_prob=0.002, n_draws=4)
#' drawsB <- sim_geno(grav2[6:12,], map, error_prob=0.002, n_draws=4)
#' draws <- rbind(drawsA, drawsB)
#'
#' @export
rbind.sim_geno <-
    function(...)
    rbind.calc_genoprob(...)

# viterbi
#' Calculate most probable sequence of genotypes
#'
#' Uses a hidden Markov model to calculate arg max Pr(g | O) where g
#' is the underlying sequence of true genotypes and O is the observed
#' multipoint marker data, with possible allowance for genotyping
#' errors.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param map Genetic map of markers. May include pseudomarker
#' locations (that is, locations that are not within the marker
#' genotype data). If NULL, the genetic map in `cross` is used.
#' @param error_prob Assumed genotyping error probability
#' @param map_function Character string indicating the map function to
#' use to convert genetic distances to recombination fractions.
#' @param lowmem If `FALSE`, split individuals into groups with
#' common sex and crossinfo and then precalculate the transition
#' matrices for a chromosome; potentially a lot faster but using more
#' memory.
#' @param quiet If `FALSE`, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return An object of class `"viterbi"`: a list of two-dimensional
#' arrays of imputed genotypes, individuals x positions.
#' Also contains three attributes:
#' * `crosstype` - The cross type of the input `cross`.
#' * `is_x_chr` - Logical vector indicating whether chromosomes
#'     are to be treated as the X chromosome or not, from input `cross`.
#' * `alleles` - Vector of allele codes, from input
#'     `cross`.
#'
#' @details We use a hidden Markov model to find, for each individual
#' on each chromosome, the most probable sequence of underlying
#' genotypes given the observed marker data.
#'
#' Note that we break ties at random, and our method for doing this
#' may introduce some bias.
#'
#' Consider the results with caution; the most probable sequence can
#' have very low probability, and can have features that are quite
#' unusual (for example, the number of recombination events can be too
#' small). In most cases, the results of a single imputation with
#' [sim_geno()] will be more realistic.
#'
#' @seealso [sim_geno()], [maxmarg()], [cbind.viterbi()], [rbind.viterbi()]
#'
#' @export
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map_w_pmar <- insert_pseudomarkers(grav2$gmap, step=1)
#' g <- viterbi(grav2, map_w_pmar, error_prob=0.002)

viterbi <-
    function(cross, map=NULL, error_prob=1e-4,
             map_function=c("haldane", "kosambi", "c-f", "morgan"),
             lowmem=FALSE, quiet=TRUE, cores=1)
{
    # check inputs
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')
    if(error_prob < 0)
        stop("error_prob must be > 0")
    map_function <- match.arg(map_function)

    if(!is_nonneg_number(error_prob)) stop("error_prob should be a single non-negative number")

    if(!lowmem)
        return(viterbi2(cross=cross, map=map, error_prob=error_prob,
                        map_function=map_function, quiet=quiet,
                        cores=cores))

    # set up cluster; make quiet=FALSE if cores>1
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores) > 1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # no more messages
    }


    # pseudomarker map
    if(is.null(map)) {
        if(is.null(cross$gmap)) stop("If cross does not contain a genetic map, map must be provided.")
        map <- insert_pseudomarkers(cross$gmap)
    }
    # possibly subset the map
    if(length(map) != length(cross$geno) || !all(names(map) == names(cross$geno))) {
        chr <- names(cross$geno)
        if(!all(chr %in% names(map)))
            stop("map doesn't contain all of the necessary chromosomes")
        map <- map[chr]
    }
    # calculate marker index object
    index <- create_marker_index(lapply(cross$geno, colnames), map)

    rf <- map2rf(map, map_function)

    # deal with missing information
    ind <- rownames(cross$geno[[1]])
    chrnames <- names(cross$geno)
    is_x_chr <- handle_null_isxchr(cross$is_x_chr, chrnames)
    cross$is_female <- handle_null_isfemale(cross$is_female, ind)
    cross$cross_info <- handle_null_isfemale(cross$cross_info, ind)

    founder_geno <- cross$founder_geno
    if(is.null(founder_geno))
        founder_geno <- create_empty_founder_geno(cross$geno)

    by_group_func <- function(i) {
        .viterbi(cross$crosstype, t(cross$geno[[chr]][group[[i]],,drop=FALSE]),
                 founder_geno[[chr]], cross$is_x_chr[chr], cross$is_female[group[[i]]],
                 t(cross$cross_info[group[[i]],,drop=FALSE]), rf[[chr]], index[[chr]],
                 error_prob)
    }

    group <- parallel::splitIndices(nrow(cross$geno[[1]]), n_cores(cores))
    groupindex <- seq(along=group)

    result <- vector("list", length(cross$geno))
    names(result) <- names(cross$geno)
    for(chr in seq(along=cross$geno)) {
        if(!quiet) message("Chr ", names(cross$geno)[chr])

        if(n_cores(cores)==1) { # no parallel processing
            # calculations in one group
            result[[chr]] <- by_group_func(1)
        }
        else {
            # calculations in parallel
            temp <- cluster_lapply(cores, groupindex, by_group_func)

            # paste them back together
            d <- vapply(temp, dim, rep(0,2))
            nr <- sum(d[1,])
            result[[chr]] <- matrix(nrow=nr, ncol=d[2,1])
            for(i in groupindex)
                result[[chr]][group[[i]],] <- temp[[i]]
        }

        dimnames(result[[chr]]) <- list(rownames(cross$geno[[chr]]),
                                        names(map[[chr]]))
    }

    names(result) <- names(cross$geno)
    attr(result, "crosstype") <- cross$crosstype
    attr(result, "is_x_chr") <- cross$is_x_chr
    attr(result, "alleles") <- cross$alleles

    class(result) <- c("viterbi", "list")
    result
}

context("rbind and cbind viterbi")

test_that("rbind.viterbi works for grav", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    g <- viterbi(grav2, map, error_prob=0.002)
    gA <- g[1:5,]
    gB <- g[6:12,]
    gC <- g[13:20,]
    gAB <- g[1:12,]
    gABC <- g[1:20,]
    gBACA <- g[c(6:12, 1:5, 13:20, 1:5),]

    expect_equal(rbind(gA, gB), gAB)
    expect_equal(rbind(gA, gB, gC), gABC)
    expect_equal(rbind(gB, gA, gC, gA), gBACA)

})


test_that("rbind.viterbi works for iron", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    g <- viterbi(iron, map, error_prob=0.002)
    gA <- g[2:20,]
    gB <- g[41:60,]
    gC <- g[102:201,]
    gAB <- g[c(2:20,41:60),]
    gABC <- g[c(2:20,41:60,102:201),]
    gBACA <- g[c(41:60,2:20,102:201,2:20),]

    expect_equal(rbind(gA, gB), gAB)
    expect_equal(rbind(gA, gB, gC), gABC)
    expect_equal(rbind(gB, gA, gC, gA), gBACA)

})

test_that("cbind.viterbi works for grav", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    g <- viterbi(grav2[1:10,], map, error_prob=0.002)
    gA <- g[,1:2]
    gB <- g[,5]
    gC <- g[,3:4]
    gAB <- g[,c(1:2,5)]
    gABC <- g[,c(1,2,5,3,4)]
    gBACA <- g[,c(5,1:2,3:4,1:2)]

    expect_equal(cbind(gA, gB), gAB)
    expect_equal(cbind(gA, gB, gC), gABC)
    expect_equal(cbind(gB, gA, gC, gA), gBACA)

})


test_that("cbind.viterbi works for iron", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    g <- viterbi(iron[6:21,], map, error_prob=0.002)
    gA <- g[,2:3]
    gB <- g[,c(4,5,8)]
    gC <- g[,c(19,"X")]
    gAB <- g[,c(2:3,4,5,8)]
    gABC <- g[,c(2:3,4,5,8,19,"X")]
    gBACA <- g[,c(4,5,8,2,3,19,"X",2,3)]

    expect_equal(cbind(gA, gB), gAB)
    expect_equal(cbind(gA, gB, gC), gABC)
    expect_equal(cbind(gB, gA, gC, gA), gBACA)

})

context("viterbi")
suppressMessages(library(qtl))

test_that("backcross autosome", {

    data(hyper)
    chr <- c(1, 4)
    hyper <- hyper[chr,]

    hyper <- convert2cross2(hyper)
    set.seed(20150524)
    g <- viterbi(hyper, error_prob=0.002, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(hyper, error_prob=0.002, lowmem=FALSE)

    expect_equal(g, g2)

})

test_that("intercross autosome", {

    data(listeria)
    chr <- c(1, 4, 14, 18)
    listeria <- listeria[chr,]

    listeria <- convert2cross2(listeria)
    map <- insert_pseudomarkers(listeria$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(listeria, map, error_prob=0.01, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(listeria, map, error_prob=0.01, lowmem=FALSE)

    expect_equal(g, g2)

})


test_that("risib autosome", {

    data(hyper)
    chr <- c(1, 4)
    hyper <- hyper[chr,]
    class(hyper)[1] <- "risib"

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(hyper, map, error_prob=0.002, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(hyper, map, error_prob=0.002, lowmem=FALSE)

    expect_equal(g, g2)

})

test_that("riself autosome", {

    data(hyper)
    chr <- c(1, 4)
    hyper <- hyper[chr,]
    class(hyper)[1] <- "riself"

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(hyper, map, error_prob=0.002, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(hyper, map, error_prob=0.002, lowmem=FALSE)

    expect_equal(g, g2)

})

test_that("f2 X chr", {

    skip_on_cran()

    data(fake.f2)
    fake.f2 <- fake.f2["X",]

    fake.f2 <- convert2cross2(fake.f2)

    # order individuals by sex/cross
    sexcr <- paste(fake.f2$is_female, apply(fake.f2$cross_info, 1, paste, collapse=":"), sep=":")
    o <- order(sexcr)
    fake.f2 <- fake.f2[o,]

    map <- insert_pseudomarkers(fake.f2$gmap, step=1, stepwidth="max")

    set.seed(20150524)
    g <- viterbi(fake.f2, map, error_prob=0.001, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(fake.f2, map, error_prob=0.001, lowmem=FALSE)

    expect_equal(g, g2)

})

test_that("bc X chr", {

    data(hyper)
    hyper <- hyper["X",]
    # make it half female, half male
    hyper$pheno$sex <- rep(c("female", "male"), nind(hyper)/2)

    hyper <- convert2cross2(hyper)

    # order individuals by sex/cross
    sexcr <- paste(hyper$is_female, apply(hyper$cross_info, 1, paste, collapse=":"), sep=":")
    o <- order(sexcr)
    hyper <- hyper[o,]

    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")

    set.seed(20150524)
    g <- viterbi(hyper, map, error_prob=0.02, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(hyper, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(g, g2)

})

test_that("f2 X chr all males", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 1

    fake.f2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(fake.f2, map, error_prob=0.02, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(fake.f2, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(g, g2)

})


test_that("f2 X chr all females", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 0

    fake.f2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(fake.f2, map, error_prob=0.02, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(fake.f2, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(g, g2)

})

test_that("f2 X chr all females forw", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 0
    fake.f2$pheno$pgm <- 0

    fake.f2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(fake.f2, map, error_prob=0.0001, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(fake.f2, map, error_prob=0.0001, lowmem=FALSE)

    expect_equal(g, g2)

})

test_that("f2 X chr all females rev", {

    data(fake.f2)
    fake.f2 <- fake.f2["X",]
    fake.f2$pheno$sex <- 0
    fake.f2$pheno$pgm <- 1

    fake.f2 <- convert2cross2(fake.f2)
    map <- insert_pseudomarkers(fake.f2$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(fake.f2, map, error_prob=0.0001, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(fake.f2, map, error_prob=0.0001, lowmem=FALSE)

    expect_equal(g, g2)

})

test_that("bc X chr all males", {

    data(hyper)
    hyper <- hyper["X",]

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(hyper, map, error_prob=0.02, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(hyper, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(g, g2)

})


test_that("bc X chr all females", {

    data(hyper)
    hyper <- hyper["X",]
    hyper$pheno$sex <- "female"

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(hyper, map, error_prob=0.02, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(hyper, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(g, g2)

})

test_that("doubled haploids", {

    data(hyper)
    hyper <- hyper[1,]
    hyper$pheno <- hyper$pheno[,1,drop=FALSE]
    class(hyper)[1] <- "dh"

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(hyper, map, error_prob=0.02, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(hyper, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(g, g2)

})


test_that("haploids viterbi", {

    data(hyper)
    hyper <- hyper[2,]
    hyper$pheno <- hyper$pheno[,1,drop=FALSE]
    class(hyper)[1] <- "haploid"

    hyper <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper$gmap, step=1, stepwidth="max")
    set.seed(20150524)
    g <- viterbi(hyper, map, error_prob=0.02, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(hyper, map, error_prob=0.02, lowmem=FALSE)

    expect_equal(g, g2)

})

test_that("backcross autosome with markers at same location", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    grav2 <- grav2[1:4,1]

    # put some markers at same location
    grav2$gmap[[1]][4] <- grav2$gmap[[1]][3]

    set.seed(20150524)
    g <- viterbi(grav2, error_prob=0, lowmem=TRUE)
    set.seed(20150524)
    g2 <- viterbi(grav2, error_prob=0, lowmem=FALSE)

    expect_true( all(!is.na(g[[1]])) )
    expect_equal(g, g2)

})

test_that("R/qtl2 gives same viterbi results as R/qtl when step=0, for backcross", {

    library(qtl)
    data(hyper)

    # step=0, error_prob=0, chromosomes where everyone has *some* data
    for(chr in c(1, 4, 7, 17)) {
        hypersub <- hyper[chr,]
        hyper2sub <- convert2cross2(hypersub)
        hypersub <- argmax.geno(hypersub, step=0, error.prob=0)
        agm <- hypersub$geno[[1]]$argmax
        agm2 <- viterbi(hyper2sub, error_prob=0)[[1]]
        expect_equivalent(agm, agm2)

        # lowmem version
        agm2b <- viterbi(hyper2sub, error_prob=0, lowmem=TRUE)[[1]]
        expect_equivalent(agm, agm2b)
    }

    # step=0, error_prob=0.002, chromosomes where everyone has *some* data
    for(chr in c(1, 4, 7, 17)) {
        hypersub <- hyper[chr,]
        hyper2sub <- convert2cross2(hypersub)
        hypersub <- argmax.geno(hypersub, step=0, error.prob=0.002)
        agm <- hypersub$geno[[1]]$argmax
        agm2 <- viterbi(hyper2sub, error_prob=0.002)[[1]]
        expect_equivalent(agm, agm2)

        # lowmem version
        agm2b <- viterbi(hyper2sub, error_prob=0.002, lowmem=TRUE)[[1]]
        expect_equivalent(agm, agm2b)
    }

})


test_that("R/qtl2 gives same viterbi results as R/qtl when step=0, for intercross", {

    library(qtl)
    data(listeria)

    # step=0, error_prob=0, chromosomes where everyone has *some* data
    for(chr in c(1, 4, 7, 17)) {
        listeriasub <- listeria[chr,]
        listeria2sub <- convert2cross2(listeriasub)
        listeriasub <- argmax.geno(listeriasub, step=0, error.prob=0)
        agm <- listeriasub$geno[[1]]$argmax
        agm2 <- viterbi(listeria2sub, error_prob=0)[[1]]
        expect_equivalent(agm, agm2)

        # lowmem version
        agm2b <- viterbi(listeria2sub, error_prob=0, lowmem=TRUE)[[1]]
        expect_equivalent(agm, agm2b)
    }

    # step=0, error_prob=0.002, chromosomes where everyone has *some* data
    for(chr in c(1, 4, 7, 17)) {
        listeriasub <- listeria[chr,]
        listeria2sub <- convert2cross2(listeriasub)
        listeriasub <- argmax.geno(listeriasub, step=0, error.prob=0.002)
        agm <- listeriasub$geno[[1]]$argmax
        agm2 <- viterbi(listeria2sub, error_prob=0.002)[[1]]
        expect_equivalent(agm, agm2)

        # lowmem version
        agm2b <- viterbi(listeria2sub, error_prob=0.002, lowmem=TRUE)[[1]]
        expect_equivalent(agm, agm2b)
    }

})

# check that kinship matrices are square
# and have same row and column IDs
#
# returns vector of IDs
check_kinship <-
    function(kinship, n_chr)
{
    if(is_kinship_decomposed(kinship)) {
        if(is_kinship_list(kinship)) {
            if(length(kinship) != n_chr)
                stop("length(kinship) != no. chromosomes (", n_chr, ")")
            return(rownames(kinship[[1]]$vectors))
        }
        else {
            return(rownames(kinship$vectors))
        }
    }

    if(!is_kinship_list(kinship)) { # one kinship matrix
        stopifnot(nrow(kinship) == ncol(kinship))
        stopifnot( all(rownames(kinship) == colnames(kinship)) )
        return(rownames(kinship))
    } else {
        if(length(kinship) != n_chr)
            stop("length(kinship) != no. chromosomes (", n_chr, ")")

        kinship_square <- vapply(kinship, function(mat) nrow(mat) == ncol(mat), TRUE)
        stopifnot( all(kinship_square) )

        kinship_sameIDs <- vapply(kinship, function(mat) (nrow(mat) == nrow(kinship[[1]])) &&
                                  all((rownames(mat) == rownames(kinship[[1]])) &
                                      (colnames(mat) == colnames(kinship[[1]])) &
                                      (rownames(mat) == colnames(kinship[[1]]))), TRUE)
        if(!all(kinship_sameIDs))
            stop("All kinship matrices should be the same size ",
                 "and have the same row and column names")

        return(rownames(kinship[[1]]))
    }
}

# multiply kinship matrix by 2
# see Almasy & Blangero (1998) https://doi.org/10.1086/301844
#
# This can also handle the case of "loco", and of having eigen decomposition pre-computed
double_kinship <-
    function(kinship)
{
    if(is.null(kinship)) return(NULL)

    if(is_kinship_decomposed(kinship)) { # already did decomposition
        if(is_kinship_list(kinship)) { # list of decomposed kinship matrices
            kinship <- lapply(kinship, function(a) { a$values <- 2*a$values; a })
        }
        else { # one decomposed kinship matrix
            kinship$values <- 2*kinship$values
        }
    }
    else {
        if(is.list(kinship))
            kinship <- lapply(kinship, function(a) a*2)
        else kinship <- 2*kinship
    }

    kinship
}


# check if alread decomposed
is_kinship_decomposed <-
    function(kinship)
{
    decomp <- attr(kinship, "eigen_decomp")

    (!is.null(decomp) && decomp) || # should have attribute
        (length(kinship)==2 && all(names(kinship) == c("values", "vectors"))) || # single-chr case missing attribute
        (is.list(kinship) && all(vapply(kinship, length, 1)==2) && all(vapply(kinship, function(a) all(names(a)==c("values", "vectors")), TRUE))) # multi-chr case
}

# is kinship a list with (potentially) multiple chromosomes
is_kinship_list <-
    function(kinship)
{
    if(is_kinship_decomposed(kinship)) {
        if(length(kinship)==2 && all(names(kinship) == c("values", "vectors"))) { # just one chromosome
            return(FALSE)
        }
        else return(TRUE)
    }
    else {
        return(is.list(kinship))
    }
}

# check that kinship concerns one chromosome
# and remove outer list if necessary
check_kinship_onechr <-
    function(kinship)
{
    if(is_kinship_list(kinship)) {
        if(length(kinship) > 1)
            stop("kinship should concern just one chromosome")
        decomp <- attr(kinship, "eigen_decomp") ## preserve attribute
        kinship <- kinship[[1]]
        attr(kinship, "eigen_decomp") <- decomp
    }

    kinship
}

# multiply kinship by weights (from and back)
# assuming weights are really square-root weights
#' @importFrom stats setNames
weight_kinship <-
    function(kinship, weights=NULL, tol=1e-8)
{
    # if null weights are all 1's, just skip it
    if(is.null(weights) || max(abs(weights-1)) < tol) return(kinship)

    if(is_kinship_list(kinship)) {
        for(i in seq_along(kinship)) {
            kinship[[i]] <- weight_kinship(kinship[[i]], weights)
        }
        return(kinship)
    }

    # if kinship was decomposed, expand it and then decompose it later
    do_decomp <- FALSE
    if(is_kinship_decomposed(kinship)) {
        do_decomp <- TRUE
        # expand out the decomposition
        kinship <- t(kinship$vectors) %*% diag(kinship$values) %*% kinship$vectors
    }

    # line them up
    ind2keep <- get_common_ids(setNames(rownames(kinship), NULL), setNames(names(weights), NULL))
    weights <- weights[ind2keep]
    kinship <- kinship[ind2keep, ind2keep, drop=FALSE]

    # multiply kinship matrix by weights
    kinship <- kinship * outer(weights, weights, "*")

    # decompose kinship again
    if(do_decomp) kinship <- decomp_kinship(kinship)

    kinship
}

context("kinship utilities")

test_that("is_kinship and is_kinship_decomposed work", {

    library(qtl2)
    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:100,c(11,12,16,"X")]
    pr <- calc_genoprob(iron)

    k <- calc_kinship(pr)
    kloco <- calc_kinship(pr, "loco")

    kd <- decomp_kinship(k)
    kdloco <- decomp_kinship(kloco)

    expect_false(is_kinship_list(k))
    expect_false(is_kinship_list(kd))
    expect_true(is_kinship_list(kloco))
    expect_true(is_kinship_list(kdloco))

    expect_false(is_kinship_decomposed(k))
    expect_false(is_kinship_decomposed(kloco))
    expect_true(is_kinship_decomposed(kd))
    expect_true(is_kinship_decomposed(kdloco))

    expect_error(check_kinship_onechr(kloco))
    expect_error(check_kinship_onechr(kdloco))

    expect_equal(check_kinship_onechr(kloco[2]), kloco[[2]])
    expect_equivalent(check_kinship_onechr(kdloco[2]), kdloco[[2]]) # former will have the eigen_decomp attribute
    expect_equal(check_kinship_onechr(k), k)
    expect_equal(check_kinship_onechr(kd), kd)

})

#' Join genotype probabilities for different chromosomes
#'
#' Join multiple genotype probability objects, as produced by
#' [calc_genoprob()], for the same set of individuals but different
#' chromosomes.
#'
#' @param ... Genotype probability objects as produced by
#' [calc_genoprob()]. Must have the same set of individuals.
#'
#' @return An object of class `"calc_genoprob"`, like the input; see [calc_genoprob()].
#'
#' @seealso [rbind.calc_genoprob()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' probsA <- calc_genoprob(grav2[1:5,1:2], map, error_prob=0.002)
#' probsB <- calc_genoprob(grav2[1:5,3:4], map, error_prob=0.002)
#' probs <- cbind(probsA, probsB)
#'
#' @export
cbind.calc_genoprob <-
    function(...)
{
    args <- list(...)

    # to cbind: probs, is_x_chr
    # to pass through (must match): crosstype, alleles, alleleprobs

    result <- args[[1]]
    if(length(args) == 1) return(result)

    # paste stuff together
    for(i in 2:length(args)) {
        if(nrow(args[[1]][[1]]) != nrow(args[[i]][[1]]) ||
           !all(rownames(args[[1]][[1]]) == rownames(args[[i]][[1]])))
            stop("Input objects 1 and ", i, " have different individuals")

        result <- c(result, args[[i]])
    }

    other_stuff <- c("is_x_chr")
    for(obj in other_stuff)
        attr(result, obj) <- attr(args[[1]], obj)
    for(i in 2:length(args)) {
        for(obj in other_stuff) {
            if(is.null(attr(args[[1]], obj)) && is.null(attr(args[[i]], obj))) next # not present
            if(is.null(attr(args[[1]], obj)) || is.null(attr(args[[i]], obj)))
                stop(obj, " not present in all inputs")
            attr(result, obj) <- c(attr(result, obj), attr(args[[i]], obj))
        }
    }

    # check that things match
    other_stuff <- c("crosstype", "alleles", "alleleprobs")
    for(i in 2:length(args)) {
        for(obj in other_stuff) {
            if(!is_same(attr(args[[1]], obj), attr(args[[i]], obj)))
                stop("Input objects 1 and ", i, " differ in their ", obj)
        }
    }
    for(obj in other_stuff)
        attr(result, obj) <- attr(args[[1]], obj)

    class(result) <- class(args[[1]])

    result
}

# n_missing
#' Count missing genotypes
#'
#' Number (or proportion) of missing (or non-missing) genotypes by individual or marker
#'
#' @param cross An object of class `"cross2"`, as output by [read_cross2()]. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param by Whether to summarize by individual or marker
#' @param summary Whether to take count or proportion
#'
#' @return Vector of counts (or proportions) of missing (or non-missing) genotypes.
#'
#' @describeIn n_missing Count missing genotypes
#' @export
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' nmis_ind <- n_missing(iron)
#' pmis_mar <- n_typed(iron, "mar", "proportion")
#' plot(nmis_ind, xlab="Individual", ylab="No. missing genotypes")
#' plot(pmis_mar, xlab="Markers", ylab="Prop. genotyped")
n_missing <-
    function(cross, by=c("individual", "marker"), summary=c("count", "proportion"))
{
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')
    by <- match.arg(by)
    summary <- match.arg(summary)

    result <- NULL
    denom <- 0
    for(i in seq(along=cross$geno)) {
        if(by=="individual") {
            this_result <- rowSums(cross$geno[[i]]==0)
            denom <- denom + ncol(cross$geno[[i]])
        }
        else
            this_result <- colSums(cross$geno[[i]]==0)


        if(is.null(result)) result <- this_result
        else {
            if(by=="individual")
                result <- result + this_result
            else
                result <- c(result, this_result)
        }
    }

    if(summary == "proportion") {
        if(by=="marker")
            denom <- nrow(cross$geno[[1]])
        result <- result/denom
    }

    result
}

#' @describeIn n_missing Count genotypes
#' @export
n_typed <-
    function(cross, by=c("individual", "marker"), summary=c("count", "proportion"))
{
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')
    by <- match.arg(by)
    summary <- match.arg(summary)

    opp <- n_missing(cross, by, summary)
    if(summary=="proportion") return(1-opp)
    else if(by=="individual") return(tot_mar(cross) - opp)
    else return(n_ind(cross) - opp)
}

#' Calculate strain distribution pattern from SNP genotypes
#'
#' Calculate the strain distribution patterns (SDPs) from the strain
#' genotypes at a set of SNPs.
#'
#' @param geno Matrix of SNP genotypes, markers x strains, coded as 1
#' (AA) and 3 (BB). Markers with values other than 1 or 3 are omitted,
#' and monomorphic markers, are omitted.
#'
#' @return A vector of strain distribution patterns: integers between
#' 1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
#' binary representation indicates the strain genotypes.
#'
#' @seealso [invert_sdp()], [sdp2char()]
#'
#' @export
#' @examples
#' x <- rbind(m1=c(3, 1, 1, 1, 1, 1, 1, 1),
#'            m2=c(1, 3, 3, 1, 1, 1, 1, 1),
#'            m3=c(1, 1, 1, 1, 3, 3, 3, 3))
#' calc_sdp(x)
calc_sdp <-
    function(geno)
{
    # tolerate data frames, but convert to matrix
    if(!is.matrix(geno)) {
        if(is.data.frame(geno)) {
            geno <- as.matrix(geno)
        } else {
            geno <- rbind(geno)
            dimnames(geno) <- NULL
        }
    }

    n_str <- ncol(geno)

    n_missing <- rowSums(is.na(geno) | (geno != 1 & geno != 3))
    if(any(n_missing > 0)) {
        geno <- geno[n_missing==0,,drop=FALSE]
        warning("Omitting ", sum(n_missing>0), " markers with missing or invalid genotypes")
        if(nrow(geno) == 0)
            stop("No markers with valid data.")
    }

    n_AA <- rowSums(geno==1)
    monomorphic <- (n_AA==0 | n_AA==n_str)
    if(any(monomorphic)) {
        geno <- geno[!monomorphic,,drop=FALSE]
        warning("Omitting ", sum(monomorphic), " monomorphic markers")
    }

    if(nrow(geno) == 0)
        stop("No polymorphic markers.")

    result <- .calc_sdp((geno- 1)/2) # convert matrix to 0/1
    names(result) <- rownames(geno)
    result
}


#' Calculate SNP genotype matrix from strain distribution patterns
#'
#' Calculate the matrix of SNP genotypes from a vector of strain distribution patterns (SDPs).
#'
#' @param sdp Vector of strain distribution patterns (integers between
#' 1 and \eqn{2^n-2} where \eqn{n} is the number of strains.
#' @param n_strains Number of strains
#'
#' @return Matrix of SNP genotypes, markers x strains, coded as 1
#' (AA) and 3 (BB). Markers with values other than 1 or 3 are omitted,
#' and monomorphic markers, are omitted.
#'
#' @seealso [sdp2char()], [calc_sdp()]
#'
#' @export
#' @examples
#' sdp <- c(m1=1, m2=12, m3=240)
#' invert_sdp(sdp, 8)
invert_sdp <-
    function(sdp, n_strains)
{
    if(!is.numeric(sdp)) stop("sdp should be a vector of integers")
    if(!is_pos_number(n_strains)) stop("n_strains should be a single positive integer")

    geno <- .invert_sdp(sdp, n_strains)
    rownames(geno) <- names(sdp)
    geno <- geno*2+1 # convert 0/1 to 1/3

    geno
}

#' Convert strain distribution patterns to character strings
#'
#' Convert a vector of numeric codes for strain distribution patterns to character strings.
#'
#' @param sdp Vector of strain distribution patterns (integers between
#' 1 and \eqn{2^n-2} where \eqn{n} is the number of strains.
#' @param n_strains Number of founder strains (if missing but
#' `strains` is provided, we use the length of `strains`)
#' @param strains Vector of single-letter codes for the strains
#'
#' @return Vector of character strings with the two groups of alleles
#'     separated by a vertical bar (|).
#'
#' @seealso [invert_sdp()], [calc_sdp()]
#'
#' @export
#' @examples
#' sdp <- c(m1=1, m2=12, m3=240)
#' sdp2char(sdp, 8)
#' sdp2char(sdp, strains=c("A", "B", "1", "D", "Z", "C", "P", "W"))
sdp2char <-
    function(sdp, n_strains=NULL, strains=NULL)
{
    if(is.null(n_strains) && !is.null(strains)) {
        n_strains <- length(strains)
    }
    if(!is.null(n_strains) && is.null(strains)) {
        strains <- LETTERS[seq_len(n_strains)]
    }
    if(is.null(n_strains) && is.null(strains)) {
        stop("Provide at least one of n_strains or strains")
    }

    # convert to matrix
    pat <- invert_sdp(sdp, n_strains)

    # convert to character strings
    result <- apply(pat, 1, function(p)
        paste(paste(strains[p==3], collapse=""),
              paste(strains[p==1], collapse=""), sep="|") )

    names(result) <- names(sdp)

    result
}

# convert2cross2
#' Convert R/qtl cross object to new format
#'
#' Convert a cross object from the R/qtl format to the R/qtl2 format
#'
#' @param cross An object of class `"cross"`; see
#' [qtl::read.cross()] for details.
#'
#' @return Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#'
#' @export
#' @keywords utilities
#' @seealso [read_cross2()]
#' @examples
#' library(qtl)
#' data(hyper)
#' hyper2 <- convert2cross2(hyper)
convert2cross2 <-
function(cross)
{
    if(is.null(cross)) stop("cross is NULL")
    crosstype <- rqtl1_crosstype(cross)
    check_crosstype(crosstype)
    result <- list(crosstype=crosstype)
    n.ind <- rqtl_nind(cross)

    # genetic map, and grab chrtype
    result$gmap <- rqtl_pull_map(cross)
    class(result$gmap) <- "list"
    result$is_x_chr <- vapply(result$gmap, function(a) rqtl1_chrtype(a)=="X", TRUE)
    for(i in seq(along=result$gmap))
        class(result$gmap[[i]]) <- "numeric"

    # ids
    ids <- rqtl_getid(cross)
    if(is.null(ids)) ids <- as.character(1:n.ind)

    # split out genotype data
    result$geno <- lapply(cross$geno, "[[", "data")
    for(i in seq(along=result$geno))
        rownames(result$geno[[i]])  <- ids

    # sex/pgm
    sexpgm <- rqtl_getsex(cross)
    if(is.null(sexpgm$sex))
        result$is_female <- rep(FALSE, n.ind)
    else result$is_female <- (sexpgm$sex == 0)
    names(result$is_female) <- ids

    if(is.null(sexpgm$pgm))
        result$cross_info <- matrix(0L, ncol=1, nrow=n.ind) # if missing, assume they're all 0's
    else result$cross_info <- matrix(as.integer(sexpgm$pgm))
    rownames(result$cross_info) <- ids

    # convert X chr genotypes
    if(any(result$is_x_chr) && crosstype %in% c("bc", "f2", "bcsft")) { # bcsft not really supported yet
        for(i in which(result$is_x_chr)) {
            result$geno[[i]] <- rqtl_reviseXdata(crosstype, "simple",
                                                sexpgm, geno=result$geno[[i]],
                                                cross.attr=attributes(cross), force=TRUE)
        }
    }

    # in genotypes, replace NAs with 0s
    for(i in seq(along=result$geno)) {
        result$geno[[i]][is.na(result$geno[[i]])] <- 0L
        storage.mode(result$geno[[i]]) <- "integer"
    }

    # phenotypes: pull out numeric columns and put the rest in covariates
    phe <- cross$pheno

    # put sex and pgm in covar data even if numeric
    phe.names <- colnames(phe)
    is_sex <- is_pgm <- rep(FALSE, ncol(phe))
    is_sex[grep("^[Ss][Ee][Xx]$", phe.names)] <- TRUE
    is_pgm[grep("^[Pp][Gg][Mm]$", phe.names)] <- TRUE
    numercol <- vapply(phe, is.numeric, TRUE) & (!is_sex) & (!is_pgm)
    covar <- phe[,!numercol,drop=FALSE]
    if(ncol(covar) > 0) {
        result$covar <- covar
        rownames(result$covar) <- ids
    }
    phe <- phe[, numercol, drop=FALSE]
    if(ncol(phe) > 0) {
        result$pheno <- as.matrix(phe)
        rownames(result$pheno) <- ids
        storage.mode(result$pheno) <- "double"
    }

    # alleles
    alleles <- attr(cross, "alleles")
    if(!is.null(alleles)) result$alleles <- alleles
    else result$alleles <- LETTERS[seq_len(nalleles(result$crosstype))]

    class(result) <- "cross2"

    check_cross2(result) # double-check

    result
}


rqtl_nind <-
    function(object)
{
    if(!inherits(object, "cross"))
        stop("Input should have class \"cross\".")

    n.ind1 <- nrow(object$pheno)
    n.ind2 <- sapply(object$geno,function(x) nrow(x$data))
    if(any(n.ind2 != n.ind1))
        stop("Different numbers of individuals in genotypes and phenotypes.")
    n.ind1
}

# R/qtl pull.map
rqtl_pull_map <-
    function(cross)
{
    if(!inherits(cross, "cross"))
        stop("Input should have class \"cross\".")

    result <- lapply(cross$geno,function(a) {
        b <- a$map
        class(b) <- as.character(class(a))
        b })
    class(result) <- "map"
    result
}

# R/qtl getid
rqtl_getid <-
    function(cross)
{
    phe <- cross$pheno
    nam <- names(phe)
    if("id" %in% nam) {
        id <- phe$id
        phenam <- "id"
    }
    else if("ID" %in% nam) {
        id <- phe$ID
        phenam <- "ID"
    }
    else if("Id" %in% nam) {
        id <- phe$Id
        phenam <- "Id"
    }
    else if("iD" %in% nam) {
        id <- phe$iD
        phenam <- "iD"
    }
    else {
        id <- NULL
        phenam <- NULL
    }

    if(is.factor(id))
        id <- as.character(id)

    attr(id, "phenam") <- phenam

    id
}

# R/qtl getsex
rqtl_getsex <-
    function(cross)
{
    type <- rqtl1_crosstype(cross)
    if(type != "bc" && type != "f2" && type != "4way") return(list(sex=NULL, pgm=NULL))

    phe.names <- names(cross$pheno)

    sex.column <- grep("^[Ss][Ee][Xx]$", phe.names)
    pgm.column <- grep("^[Pp][Gg][Mm]$", phe.names)

    if(length(sex.column)==0) { # no sex included
        sex <- NULL
    }
    else {
        if(length(sex.column)>1)
            warning("'sex' included multiple times.  Using the first one.")
        temp <- cross$pheno[,sex.column[1]]
        if(is.numeric(temp)) {
            if(any(!is.na(temp) & temp != 0 & temp != 1)) {
                warning("Sex column should be coded as 0=female 1=male; sex ignored.")
                sex <- NULL
            }
            else sex <- temp
        }
        else {
            if(!is.factor(temp)) temp <- as.factor(temp)

            if(length(levels(temp)) == 1) {
                if(levels(temp) == "F" || levels(temp)=="f" ||
                   toupper(levels(temp)) == "FEMALE") sex <- rep(0,rqtl_nind(cross))
                else if(levels(temp) == "M" || levels(temp)=="m" ||
                        toupper(levels(temp)) == "MALE") sex <- rep(1,rqtl_nind(cross))
                else
                    warning("Sex column should be coded as 0=female 1=male; sex ignored.")
            }
            else if(length(levels(temp)) > 2) {
                warning("Sex column should be coded as a two-level factor; sex ignored.")
                sex <- NULL
            }
            else { # is a factor with two levels
                lev <- levels(temp)
                if(length(grep("^[Ff]",lev))>0 &&
                   length(males <- grep("^[Mm]",lev))>0) {
                    temp <- as.character(temp)
                    sex <- rep(0,length(temp))
                    sex[is.na(temp)] <- NA
                    sex[!is.na(temp) & temp==lev[males]] <- 1
                }
                else
                    warning("Don't understand levels in sex column; sex ignored.")
            }
        }
    }

    if(length(pgm.column)==0 || type=="4way") { # no pgm included
        pgm <- NULL
    }
    else {
        if(length(pgm.column)>1)
            warning("'pgm' included multiple times.  Using the first one.")
        temp <- cross$pheno[,pgm.column[1]]
        if(!is.numeric(temp))
            temp <- as.numeric(temp)-1
        if(any(!is.na(temp) & temp != 0 & temp != 1)) {
            warning("pgm column should be coded as 0/1; pgm ignored.")
            pgm <- NULL
        }
        else pgm <- temp
    }

    if(!is.null(sex) && any(is.na(sex))) {
        if(all(sex[!is.na(sex)]==1)) {
            warning(sum(is.na(sex)), " individuals with missing sex; assuming they're male like the others")
            sex[is.na(sex)] <- 1
        }
        else if(all(sex[!is.na(sex)]==0)) {
            warning(sum(is.na(sex)), " individuals with missing sex; assuming they're female like the others")
            sex[is.na(sex)] <- 0
        }
        else {
            warning(sum(is.na(sex)), " individuals with missing sex; assuming they're female")
            sex[is.na(sex)] <- 0
        }
    }

    if(!is.null(pgm) && any(is.na(pgm))) {
        if(all(pgm[!is.na(pgm)]==1)) {
            warning(sum(is.na(pgm)), " individuals with missing pgm; assuming pgm==1 like the others")
            pgm[is.na(pgm)] <- 1
        }
        else if(all(pgm[!is.na(pgm)]==0)) {
            warning(sum(is.na(pgm)), " individuals with missing pgm; assuming pgm==0 like the others")
            pgm[is.na(pgm)] <- 0
        }
        else {
            warning(sum(is.na(pgm)), " individuals with missing pgm; assuming pgm==0")
            pgm[is.na(pgm)] <- 0
        }
    }

    list(sex=sex,pgm=pgm)
}

# R/qtl getgenonames
#
# get names of genotypes
# used in discan, effectplot, plotPXG, scanone, scantwo, vbscan, reviseXdata
# cross.attr gives the cross attributes
rqtl_getgenonames <-
    function(type=c("f2","bc","riself","risib","4way","dh","haploid","special","bcsft"),
             chrtype=c("A","X"), expandX=c("simple","standard","full"),
             sexpgm, cross.attr=NULL)
{
    type <- match.arg(type)
    chrtype <- match.arg(chrtype)
    expandX <- match.arg(expandX)

    ## Treat bcsft as bc if no intercross generations; otherwise as f2.
    if(type == "bcsft") {
        if(cross.attr$scheme[2] == 0)
            type <- "bc"
        else
            type <- "f2"
    }

    if(chrtype=="X") {
        sex <- sexpgm$sex
        pgm <- sexpgm$pgm
    }

    if(type=="special") return(cross.attr$genotypes)

    if(is.null(cross.attr) || !("alleles" %in% names(cross.attr))) {
        if(type == "4way") alleles <- LETTERS[1:4]
        else alleles <- LETTERS[1:2]
    }
    else
        alleles <- cross.attr$alleles

    tempgn <- c(paste(rep(alleles[1],2),collapse=""),
                paste(alleles,collapse=""),
                paste(rep(alleles[2],2),collapse=""),
                paste(alleles[1],"Y",sep=""),
                paste(alleles[2],"Y",sep=""))

    # get rid of missing sex and pgm values, if there are any
    if(chrtype=="X") {
        if(length(sex)>0) sex <- sex[!is.na(sex)]
        if(length(pgm)>0) pgm <- pgm[!is.na(pgm)]
    }

    if(type=="riself" || type=="risib" || type=="dh")
        gen.names <- tempgn[c(1,3)]

    else if(type=="haploid")
        gen.names <- alleles

    else if(type == "4way") {
        if(chrtype=="A")
            gen.names <- c(paste(alleles[1],alleles[3],sep=""),
                           paste(alleles[2],alleles[3],sep=""),
                           paste(alleles[1],alleles[4],sep=""),
                           paste(alleles[2],alleles[4],sep=""))
        else
            gen.names <- c(paste(alleles[1],alleles[3],sep=""),
                           paste(alleles[2],alleles[3],sep=""),
                           paste(alleles[1],"Y",sep=""),
                           paste(alleles[2],"Y",sep=""))
    }

    else if(type == "bc") {

        if(chrtype=="A") # autosome
            gen.names <- tempgn[1:2] # AA AB

        else { # X chromosome

            #                 simple     standard       full
            #   -both sexes   A-/AB/BY   AA/AB/AY/BY    same as std
            #   -all females  AA/AB      same           same
            #   -all males    AY/BY      same           same

            if(length(sex)==0 || all(sex==0)) # all females
                gen.names <- tempgn[1:2] # AA AB
            else if(all(sex==1)) # all males
                gen.names <- tempgn[4:5] # AY BY
            else { # some of each
                if(expandX == "simple")
                    gen.names <- c(paste(alleles[1], "-", sep=""),
                                   tempgn[c(2,5)]) # A-, AB, BY
                else gen.names <- tempgn[c(1,2,4,5)]  # AA,AB,AY,BY
            }
        }
    }

    else { # intercross
        if(chrtype == "A")  # autosomal
            gen.names <- tempgn[1:3]
        else { # X chromsome

            # both crosses     simple     standard         full
            #   -both sexes   A-/AB/B-    AA/AB/BB/AY/BY   AA/AB1/AB2/BB/AY/BY
            #   -all females  AA/AB/BB    same as simple   AA/AB1/AB2/BB
            #   -all males    AY/BY       same             same
            # forw cross
            #   -both sexes   A-/AB/BY    AA/AB/AY/BY      same as std
            #   -all females  AA/AB       same             same
            #   -all males    AY/BY       same             same
            # backw cross
            #   -both sexes   B-/AB/AY    BB/AB/AY/BY      same as std
            #   -all females  BB/AB       same             same
            #   -all males    AY/BY       same             same

            if(length(sex)==0 || all(sex==0)) { # all females
                if(length(pgm)==0 || all(pgm==0)) # all forw dir
                    gen.names <- tempgn[1:2] # AA AB
                else if(all(pgm==1))  # all backw dir
                    gen.names <- tempgn[3:2] # BB AB
                else { # some of each direction
                    if(expandX=="full")
                        gen.names <- c(tempgn[1],
                                       paste(tempgn[2],c("f","r"), sep=""),
                                       tempgn[3])
                    else gen.names <- tempgn[1:3]
                }
            }
            else if(all(sex==1))  # all males
                gen.names <- tempgn[4:5]
            else { # some of each sex
                if(length(pgm)==0 || all(pgm==0)) { # all forw
                    if(expandX=="simple")
                        gen.names <- c(paste(alleles[1],"-", sep=""),
                                       tempgn[c(2,5)])
                    else gen.names <- tempgn[c(1,2,4,5)]
                }
                else if (all(pgm==1)) { # all backw
                    if(expandX=="simple")
                        gen.names <- c(paste(alleles[2], "-",sep=""),
                                       tempgn[c(2,4)])
                    else gen.names <- tempgn[c(3,2,4,5)]
                }
                else { # some of each dir
                    if(expandX=="simple")
                        gen.names <- c(paste(alleles[1],"-",sep=""),
                                       tempgn[2],
                                       paste(alleles[2],"-",sep=""))
                    else if(expandX=="standard")
                        gen.names <- tempgn
                    else
                        gen.names <- c(tempgn[1],
                                       paste(tempgn[2],c("f","r"),sep=""),
                                       tempgn[3:5])
                }
            }
        }
    }

    gen.names
}

# R/qtl reviseXdata reduced to geno case
rqtl_reviseXdata <-
    function(type=c("f2","bc","bcsft"), expandX=c("simple","standard","full"),
             sexpgm, geno, cross.attr, force=FALSE)
{
    type <- match.arg(type)
    expandX <- match.arg(expandX)

    ## Treat bcsft as bc if no intercross generations; otherwise as f2.
    if(type == "bcsft") {
        if(cross.attr$scheme[2] == 0)
            type <- "bc"
        else
            type <- "f2"
    }

    sex <- sexpgm$sex
    pgm <- sexpgm$pgm

    # get genonames
    genonames <- rqtl_getgenonames(type, "X", expandX, sexpgm, cross.attr)

    if(type == "bc") { # backcross

        if(length(sex)==0 || ((all(sex==0) || all(sex==1)) && !force)) { # all one sex
            # no changes necessary
            return(geno)
        }

        else { # both sexes

            gmale <- geno[sex==1,]
            if(expandX=="simple")
                gmale[!is.na(gmale) & gmale==2] <- 3
            else {
                gmale[!is.na(gmale) & gmale==1] <- 3
                gmale[!is.na(gmale) & gmale==2] <- 4
            }
            geno[sex==1,] <- gmale
            return(geno)
        } # end of "both sexes" / backcross

    } # end of backcross

    else { # intercross

        if(length(sex)==0 || all(sex==0)) { # all females

            if(length(pgm)==0 || ((all(pgm==0) || all(pgm==1)) && !force)) { # one dir, females
                return(geno)
            }

            else { # both dir, females
                gback <- geno[pgm==1,]
                if(expandX!="full") {
                    gback[!is.na(gback) & gback==1] <- 3
                    geno[pgm==1,] <- gback
                }
                else {
                    gback[!is.na(gback) & gback==1] <- 4
                    gback[!is.na(gback) & gback==2] <- 3
                    geno[pgm==1,] <- gback
                }
                return(geno)
            }
        }
        else if(all(sex==1) && !force)  { # all males
            return(geno)
        }

        else { # both sexes

            if(length(pgm)==0 || all(pgm==0)) { # both sexes, forw dir
                gmale <- geno[sex==1,]
                if(expandX=="simple")
                    gmale[!is.na(gmale) & gmale==2] <- 3
                else {
                    gmale[!is.na(gmale) & gmale==1] <- 3
                    gmale[!is.na(gmale) & gmale==2] <- 4
                }
                geno[sex==1,] <- gmale
                return(geno)

            } # both sexes, forw dir

            if(all(pgm==1) && !force) { # both sexes, backw dir
                gmale <- geno[sex==1,]
                if(expandX!="full") {
                    gmale[!is.na(gmale) & gmale==1] <- 3
                    gmale[!is.na(gmale) & gmale==2] <- 1
                }
                else {
                    gmale[!is.na(gmale) & gmale==1] <- 3
                    gmale[!is.na(gmale) & gmale==2] <- 4
                }
                geno[sex==1,] <- gmale
                return(geno)
            } # both sexes, backw dir

            else { # both dir, both sexes
                gmale <- geno[sex==1,]
                gfemaler <- geno[sex==0 & pgm==1,]
                if(expandX=="simple") {
                    gmale[!is.na(gmale) & gmale==2] <- 3
                    gfemaler[!is.na(gfemaler) & gfemaler==1] <- 3
                }
                else if(expandX=="standard") {
                    gmale[!is.na(gmale) & gmale==1] <- 4
                    gmale[!is.na(gmale) & gmale==2] <- 5
                    gfemaler[!is.na(gfemaler) & gfemaler==1] <- 3
                }
                else {
                    gmale[!is.na(gmale) & gmale==1] <- 5
                    gmale[!is.na(gmale) & gmale==2] <- 6
                    gfemaler[!is.na(gfemaler) & gfemaler==1] <- 4
                    gfemaler[!is.na(gfemaler) & gfemaler==2] <- 3
                }
                geno[sex==1,] <- gmale
                geno[sex==0 & pgm==1,] <- gfemaler
                return(geno)
            }
        }

    } # end of intercross

}

# R/qtl1 cross type
rqtl1_crosstype <-
    function(cross)
    {
        type <- class(cross)
        type <- type[type != "cross" & type != "list"]
        if(length(type) > 1) {
            warning("cross has multiple classes")
        }
        type[1]
    }

rqtl1_chrtype <-
    function(object)
    {
        if(inherits(object, "X")) return("X")
        "A"
    }

context("est_map gives same results w/ lowmem=T or F")

test_that("est_map2 works with intercross", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:20,c(3,19,"X")]

    map_lomem <- est_map(iron, lowmem=TRUE)
    map_himem <- est_map(iron, lowmem=FALSE)
    expect_equal(map_himem, map_lomem)

})

test_that("est_map2 works with RIL by selfing", {

    skip_on_cran()

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    grav2 <- grav2[,4:5]

    map_lomem <- est_map(grav2, lowmem=TRUE)
    map_himem <- est_map(grav2, lowmem=FALSE)
    expect_equal(map_himem, map_lomem)

})


test_that("est_map2 works with a backcross with both sexes", {

    skip_on_cran()

    library(qtl)
    data(hyper)
    set.seed(35288832)
    hyper$pheno$sex <- factor(sample(c("female", "male"), nind(hyper), replace=TRUE))

    hyper <- convert2cross2(hyper)
    hyper <- hyper[,c(4,19,"X")]


    map_lomem <- est_map(hyper, lowmem=TRUE)
    map_himem <- est_map(hyper, lowmem=FALSE)
    expect_equal(map_himem, map_lomem)

})

context("convert2cross2")
suppressMessages(library(qtl))

test_that("convert2cross2 works appropriately for hyper data", {

    data(hyper)
    hyper2 <- convert2cross2(hyper)

    ids <- as.character(1:nind(hyper))

    # class and crosstype
    expect_equal(class(hyper2), "cross2")
    expect_equal(hyper2$crosstype, "bc")

    # X chr
    is_x <- rep(FALSE, nchr(hyper))
    names(is_x) <- c(1:19, "X")
    is_x["X"] <- TRUE
    expect_equal(hyper2$is_x_chr, is_x)

    # genotypes
    for(i in seq(along=hyper$geno)) {
        go <- hyper$geno[[i]]$data
        if(is_x[i])
            go <- reviseXdata("bc", "simple", getsex(hyper),
                              geno=go, cross.attr=attributes(hyper),
                              force=TRUE)
        gn <- hyper2$geno[[i]]
        expect_true(all(is.na(go) | go > 0))
        go[is.na(go)] <- 0
        rownames(go) <- ids
        expect_equal(gn, go)
    }

    # gmap
    gmap <- lapply(pull.map(hyper),
                   function(a) { class(a) <- "numeric"; a })
    expect_equal(hyper2$gmap, gmap)

    # sex
    is_female <- rep(FALSE, nind(hyper))
    names(is_female) <- ids
    expect_equal(hyper2$is_female, is_female)

    # cross_info
    cross_info <- matrix(0L, ncol=1, nrow=nind(hyper))
    rownames(cross_info) <- ids
    expect_equal(hyper2$cross_info, cross_info)

    # pheno
    phe <- hyper$pheno[,1,drop=FALSE]
    rownames(phe) <- ids
    phe <- as.matrix(phe)
    expect_equal(hyper2$pheno, phe)

    # covar
    covar <- hyper$pheno[,2,drop=FALSE]
    rownames(covar) <- ids
    expect_equal(hyper2$covar, covar)

})

test_that("convert2cross2 works appropriately for fake.f2 data", {

    data(fake.f2)
    fake.f2.2 <- convert2cross2(fake.f2)

    ids <- as.character(1:nind(fake.f2))

    # class and crosstype
    expect_equal(class(fake.f2.2), "cross2")
    expect_equal(fake.f2.2$crosstype, "f2")

    # X chr
    is_x <- rep(FALSE, nchr(fake.f2))
    names(is_x) <- c(1:19, "X")
    is_x["X"] <- TRUE
    expect_equal(fake.f2.2$is_x_chr, is_x)

    # genotypes
    for(i in seq(along=fake.f2.2$geno)) {
        go <- fake.f2$geno[[i]]$data
        if(is_x[i])
            go <- reviseXdata("f2", "simple", getsex(fake.f2),
                              geno=go, cross.attr=attributes(fake.f2))
        gn <- fake.f2.2$geno[[i]]
        expect_true(all(is.na(go) | go > 0))
        go[is.na(go)] <- 0
        rownames(go) <- ids
        expect_equal(gn, go)
    }

    # gmap
    gmap <- lapply(pull.map(fake.f2), function(a) { class(a) <- "numeric"; a })
    expect_equal(fake.f2.2$gmap, gmap)

    # sex
    sexpgm <- getsex(fake.f2)
    is_female <- (sexpgm$sex == 0)
    names(is_female) <- ids
    expect_equal(fake.f2.2$is_female, is_female)

    # cross_info
    cross_info <- as.matrix(sexpgm$pgm)
    rownames(cross_info) <- ids
    expect_equal(fake.f2.2$cross_info, cross_info)

    # pheno
    phe <- fake.f2$pheno[,1,drop=FALSE]
    rownames(phe) <- ids
    phe <- as.matrix(phe)
    expect_equal(fake.f2.2$pheno, phe)

    # covar
    covar <- fake.f2$pheno[,2:3,drop=FALSE]
    rownames(covar) <- ids
    expect_equal(fake.f2.2$covar, covar)

})

#' Join viterbi results for different chromosomes
#'
#' Join multiple viterbi objects, as produced by [viterbi()], for the
#' same set of individuals but different chromosomes.
#'
#' @param ... Imputed genotype objects as produced by
#' [viterbi()]. Must have the same set of individuals.
#'
#' @return An object of class `"viterbi"`, like the input; see [viterbi()].
#'
#' @seealso [rbind.viterbi()], [viterbi()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' gA <- viterbi(grav2[1:5,1:2], map, error_prob=0.002)
#' gB <- viterbi(grav2[1:5,3:4], map, error_prob=0.002)
#' g <- cbind(gA, gB)
#'
#' @export
cbind.viterbi <-
    cbind.calc_genoprob

#' Installed version of R/qtl2
#'
#' Get installed version of R/qtl2
#'
#' @return A character string with the installed version of the R/qtl2 package.
#'
#' @examples
#' qtl2version()
#'
#' @importFrom utils packageVersion
#' @export
qtl2version <-
    function()
{
    version <- unlist(packageVersion("qtl2"))


    if(length(version) == 3) {
        # make it like #.#-#
        return( paste(c(version, ".", "-")[c(1,4,2,5,3)], collapse="") )
    }

    paste(version, collapse=".")
}

#' Calculate chromosome lengths
#'
#' Calculate chromosome lengths for a map object
#'
#' @param map A list of vectors, each specifying locations of the markers.
#' Can also be a cross2 object, in which case the genetic map (if present) is used.
#'
#' @param collapse_to_AX If TRUE, collapse to the total lengths of the
#' autosomes and X chromosome.
#'
#' @return A vector of chromosome lengths. If
#' `collapse_to_AX=TRUE`, the result is a vector of length 2
#' (autosomal and X chromosome lengths).
#'
#' @details We take `diff(range(v))` for each vector, `v`.
#'
#' @export
#' @seealso [scan1perm()]
chr_lengths <-
    function(map, collapse_to_AX=FALSE)
{
    if(is.null(map)) stop("map is NULL")

    if(is.cross2(map)) {
        if("gmap" %in% names(map)) {
            map <- map$gmap
        } else if("pmap" %in% names(map)) {
            warning("input is a cross2 object that has no genetic map; using physical map")
            map <- map$pmap
        } else {
            stop("input is a cross2 object that has neither a genetic not physical map")
        }
    }

    result <- vapply(map, function(a) diff(range(a, na.rm=TRUE)), 1)
    attr(result, "is_x_chr") <- attr(map, "is_x_chr")

    if(collapse_to_AX) return(collapse_chr_lengths_to_AX(result))

    result
}

# turn lengths into sums of autosomal and X-chr lengths
collapse_chr_lengths_to_AX <-
function(lengths, is_x_chr=NULL)
{
    if(length(lengths)==2 && (is.null(names(lengths)) || all(names(lengths)==c("A","X")))) {
        # here, assume that the lengths have already been collapsed
        result <- lengths
        names(result) <- c("A", "X")
        attr(result, "is_x_chr") <- c(A=FALSE,X=TRUE)
        return(result)
    }

    if(is.null(is_x_chr)) # not provided as argument; use attribute
        is_x_chr <- attr(lengths, "is_x_chr")

    if(is.null(is_x_chr)) { # still not provided; assume all autosomes
        warning("No is_x_chr attribute found; assuming all autosomes.")
        is_x_chr <- rep(FALSE, length(lengths))
        names(is_x_chr) <- names(lengths)
    }

    if(length(is_x_chr) != length(lengths)) {
        if(!all(names(lengths) %in% names(is_x_chr)))
           stop("Mismatch between lengths and its is_x_chr attribute")
        is_x_chr <- is_x_chr[names(lengths)]
    }

    result <- c(A=sum(lengths[!is_x_chr]),
                X=sum(lengths[is_x_chr]))
    attr(result, "is_x_chr") <- c(A=FALSE, X=TRUE)

    result
}

context("chromosome lengths")


test_that("chr_lengths works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    # regular case
    expected <- c("1"=83.1, "2"=34.9, "3"=29.5, "4"=42.7, "5"=44.8,
                  "6"=25.2, "7"=52.5, "8"=75.3, "9"=54.6, "10"=33.9,
                  "11"=56.9, "12"=38.2, "13"=22.9, "14"=32.8, "15"=32.8,
                  "16"=44.8, "17"=36.0, "18"=26.2, "19"=35.0, "X"=28.4)
    is_x_chr <- (names(expected)=="X")
    names(is_x_chr) <- names(expected)
    attr(expected, "is_x_chr") <- is_x_chr
    expect_equal(chr_lengths(iron$gmap), expected)

    # check collapsing -> A, X lengths
    expected_collapsed <- c(A=802.1, X=28.4)
    attr(expected_collapsed, "is_x_chr") <- c(A=FALSE, X=TRUE)
    expect_equal(chr_lengths(iron$gmap, TRUE), expected_collapsed)
    expect_equal(collapse_chr_lengths_to_AX(expected), expected_collapsed)

    # problems with is_x_chr
    is_x_chr <- attr(expected, "is_x_chr")
    attr(expected, "is_x_chr") <- NULL
    expect_warning(collapse_chr_lengths_to_AX(expected))
    expect_equal(collapse_chr_lengths_to_AX(expected, is_x_chr), expected_collapsed)

    # already collapsed
    pre_collapsed <- c(A=1300, X=100)
    expected <- pre_collapsed
    attr(expected, "is_x_chr") <- c(A=FALSE, X=TRUE)
    expect_equal(collapse_chr_lengths_to_AX(pre_collapsed), expected)
    expect_equal(collapse_chr_lengths_to_AX(pre_collapsed, c(FALSE,FALSE)), expected)
    expect_equal(collapse_chr_lengths_to_AX(pre_collapsed, c(TRUE,TRUE)), expected)
    expect_equal(collapse_chr_lengths_to_AX(pre_collapsed, c(TRUE,FALSE)), expected)

    # already collapsed, no names
    names(pre_collapsed) <- NULL
    expect_equal(collapse_chr_lengths_to_AX(pre_collapsed), expected)
    expect_equal(collapse_chr_lengths_to_AX(pre_collapsed, c(FALSE,FALSE)), expected)
    expect_equal(collapse_chr_lengths_to_AX(pre_collapsed, c(TRUE,TRUE)), expected)
    expect_equal(collapse_chr_lengths_to_AX(pre_collapsed, c(TRUE,FALSE)), expected)

    # length two but not collapsed, really
    only_A <- pre_collapsed
    names(only_A) <- c("I", "II")
    attr(only_A, "is_x_chr") <- c(I=FALSE, II=FALSE)
    expected <- c("A"=1400, "X"=0)
    attr(expected, "is_x_chr") <- c(A=FALSE, X=TRUE)
    expect_equal(collapse_chr_lengths_to_AX(only_A), expected)
    expect_equal(collapse_chr_lengths_to_AX(only_A, c(FALSE,FALSE)), expected)
    expected[1] <- 0; expected[2] <- 1400
    expect_equal(collapse_chr_lengths_to_AX(only_A, c(TRUE,TRUE)), expected)
    expected[1] <- 100; expected[2] <- 1300
    expect_equal(collapse_chr_lengths_to_AX(only_A, c(TRUE,FALSE)), expected)

})

test_that("chr_lengths works on cross2 objects", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    expect_equal(chr_lengths(iron$gmap), chr_lengths(iron))

    # if no genetic map, use physical map and give warning
    iron$gmap <- NULL
    expect_warning(result <- chr_lengths(iron))
    expect_equal(chr_lengths(iron$pmap), result)

    # if no map, give error
    iron$pmap <- NULL
    expect_error(chr_lengths(iron))

})

#' Compare two sets of genotype probabilities
#'
#' Compare two sets of genotype probabilities for one individual on a single chromosome.
#'
#' @param probs1 Genotype probabilities (as produced by [calc_genoprob()])
#' or allele dosages (as produced by [genoprob_to_alleleprob()]).
#' @param probs2 A second set of genotype probabilities, just like `probs1`.
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param ind Individual to plot, either a numeric index or an ID.
#' @param chr Selected chromosome; a single character string.
#' @param minprob Minimum probability for inferring genotypes (passed to [maxmarg()]).
#' @param minmarkers Minimum number of markers in results.
#' @param minwidth Minimum width in results.
#' @param annotate If TRUE, add some annotations to the `geno1` and
#' `geno2` columns to indicate, where they differ, which one
#' matches what appears to be the best genotype. (`*` = matches
#' the best genotype; `-` = lower match).
#'
#' @details
#' The function does the following:
#' - Reduce the probabilities to a set of common locations that also appear in `cross`.
#' - Use [maxmarg()] to infer the genotype at every position using each set of probabilities.
#' - Identify intervals where the two inferred genotypes are constant.
#' - Within each segment, compare the observed SNP genotypes to the founders' genotypes.
#'
#' @return
#' A data frame with each row corresponding to an interval over which
#' `probs1` and `probs2` each have a fixed inferred genotype. Columns
#' include the two inferred genotypes, the start and end points and
#' width of the interval, and when founder genotypes are in `cross`,
#' the proportions of SNPs where the individual matches each possible
#' genotypes.
#'
#' @seealso [plot_genoprobcomp()]
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' iron <- iron[1,"2"]   # subset to first individual on chr 2
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # in presence of a genotyping error, how much does error_prob matter?
#' iron$geno[[1]][1,3] <- 3
#' pr_e <- calc_genoprob(iron, map, error_prob=0.002)
#' pr_ne <- calc_genoprob(iron, map, error_prob=1e-15)
#'
#' compare_genoprob(pr_e, pr_ne, iron, minmarkers=1, minprob=0.5)
#'
#' @export
compare_genoprob <-
    function(probs1, probs2, cross, ind=1, chr=NULL,
             minprob=0.95, minmarkers=10, minwidth=0,
             annotate=FALSE)
{
    if(is.null(probs1)) stop("probs1 is NULL")
    if(is.null(probs2)) stop("probs2 is NULL")

    # check inputs
    if(is.null(chr)) chr <- names(probs1)[1]
    if(length(chr) > 1) {
        warning("chr should have length 1; using the first value")
        chr <- chr[1]
    }
    if(!(chr %in% names(probs1) && chr %in% names(probs2)))
        stop("chr ", chr, " not found in both probs")
    if(!(chr %in% chr_names(cross))) stop("chr ", chr, " not found in cross")
    if(length(ind) > 1) {
        warning("ind should have length 1; using the first value")
        ind <- ind[1]
    }

    # pull out individual's probs; make it positions x probs
    if(is.character(ind) && !(ind %in% rownames(probs1[[chr]]) && ind %in% rownames(probs2[[chr]])))
        stop("ind ", ind, " not found in both sets of probs")
    if(is.numeric(ind)) {
        if(nrow(probs1[[chr]]) != nrow(probs2[[chr]]))
            stop("If ind is numeric, must have nrow(probs1) == nrow(probs2)")
        if(ind < 1 || ind > nrow(probs1[[chr]]))
            stop("ind ", ind, " should be in the range [1, ", nrow(probs1[[chr]]), "]")
        ind <- rownames(probs1[[1]])[ind]
    }

    # make sure they have the same genotypes
    if(ncol(probs1[[chr]]) != ncol(probs2[[chr]]) || !all(colnames(probs1[[chr]]) == colnames(probs2[[chr]])))
        stop("Need ncol(probs1) == ncol(probs2) and same colnames")
    geno_names <- colnames(probs1[[chr]])

    # infer genotypes
    ginf1 <- maxmarg(probs1[ind,chr], minprob=minprob)[[1]]
    ginf2 <- maxmarg(probs2[ind,chr], minprob=minprob)[[1]]

    if("pmap" %in% names(cross)) {
        map <- cross$pmap
        map_units <- "Mbp"
    } else {
        map <- cross$gmap
        map_units <- "cM"
    }
    gmap <- cross$gmap
    pmap <- cross$pmap

    # pull out selected chromosomes
    probs1 <- probs1[[chr]]
    probs2 <- probs2[[chr]]
    map <- map[[chr]]

    if(!is.null(gmap)) gmap <- gmap[[chr]]
    if(!is.null(pmap)) pmap <- pmap[[chr]]
    geno <- cross$geno[[chr]]

    # find common markers
    mar1 <- dimnames(probs1)[[3]]
    mar2 <- dimnames(probs2)[[3]]
    marm <- names(map)
    mar <- mar1[mar1 %in% mar2 & mar1 %in% marm]
    if(length(mar) < 2) stop("<2 markers in common")
    map <- map[mar]
    if(!is.null(gmap)) gmap <- gmap[mar]
    if(!is.null(pmap)) pmap <- pmap[mar]
    geno <- geno[,mar,drop=FALSE]
    ginf1 <- ginf1[,mar]
    ginf2 <- ginf2[,mar]

    if(!is.null(cross$founder_geno)) {
        founder_geno <- cross$founder_geno[[chr]]
        founder_geno <- founder_geno[,mar,drop=FALSE]
    }

    # now pull out individual from the genotypes
    geno <- geno[ind,]

    # find segments with constant (and non-missing calls)
    ginf <- paste(geno_names[ginf1], geno_names[ginf2], sep=":")
    ginf_num <- (ginf1-1) * ncol(probs2) + ginf2
    ginf_num[is.na(ginf_num)] <- -1
    jump <- which(diff(ginf_num) != 0)

    value_num <- c(ginf_num[1], ginf_num[jump+1])
    value <- c(ginf[1], ginf[jump+1])
    left_i <- c(1, jump+1)
    right_i <- c(jump, length(map))
    n_markers=right_i - left_i + 1
    left <- c(map[1], map[jump+1])
    right <- c(map[jump], map[length(map)])
    width <- right - left
    if(!is.null(gmap)) gwidth <- c(gmap[jump], gmap[length(gmap)]) - c(gmap[1], gmap[jump+1])
    if(!is.null(pmap)) pwidth <- c(pmap[jump], pmap[length(pmap)]) - c(pmap[1], pmap[jump+1])

    result <- data.frame(geno1=geno_names[c(ginf1[1], ginf1[jump+1])],
                         geno2=geno_names[c(ginf2[1], ginf2[jump+1])],
                         left_index = left_i,
                         right_index = right_i,
                         n_markers = n_markers,
                         left = left,
                         right = right,
                         width_Mbp = width,
                         width_cM = width,
                         stringsAsFactors=FALSE)

    # add units to left/right
    colnames(result)[6:7] <- paste(colnames(result)[6:7], map_units, sep="_")

    col2drop <- NULL
    if(!is.null(gmap)) result$width_cM <- gwidth
    else col2drop <- 9
    if(!is.null(pmap)) result$width_Mbp <- pwidth
    else col2drop <- c(col2drop, 8)
    if(!is.null(col2drop)) result <- result[,-col2drop,drop=FALSE]

    result <- result[value_num != -1 & n_markers >= minmarkers & width >= minwidth, , drop=FALSE]
    if(nrow(result) > 0) rownames(result) <- 1:nrow(result)

    ### if no founder_geno in cross, return what we've got
    if(is.null(cross$founder_geno)) return(result)

    # construct F1-type individuals
    f1 <- infer_f1_geno(founder_geno)

    # adjust genotype labels
    if("is_x_chr" %in% names(cross) && cross$is_x_chr[chr] &&
       "is_female" %in% names(cross) && !cross$is_female[ind])
        rownames(founder_geno) <- paste0(rownames(founder_geno), "Y")
    else
        rownames(founder_geno) <- paste0(rownames(founder_geno), rownames(founder_geno))

    # add F1s to founders
    founder_geno <- rbind(founder_geno, f1)

    # calculate proportion SNPs that match each possible genotype
    p_match <- matrix(nrow=nrow(result), ncol=nrow(founder_geno))
    colnames(p_match) <- rownames(founder_geno)
    for(k in 1:nrow(result)) {
        gsub <- geno[result$left_index[k]:result$right_index[k]]
        p_match[k,] <- apply(founder_geno[,result$left_index[k]:result$right_index[k],drop=FALSE], 1,
                             function(a) sum(a != 0 & gsub != 0 & gsub == a) / sum(a != 0 & gsub != 0))
    }

    # paste detailed results onto the end
    result <- cbind(result,
                    match1=rep(0, nrow(result)),
                    match2=rep(0, nrow(result)),
                    match_best=rep(0, nrow(result)),
                    what_best=rep("", nrow(result)),
                    match_next=rep(0, nrow(result)),
                    what_next=rep("", nrow(result)),
                    stringsAsFactors=FALSE)

    for(i in 1:nrow(result)) {
        result$match1[i] <- p_match[i,result$geno1[i]]
        result$match2[i] <- p_match[i,result$geno2[i]]
        z <- sort(p_match[i,], decreasing=TRUE)
        result$match_best[i] <- z[1]
        result$match_next[i] <- z[2]
        result$what_best[i] <- names(z)[1]
        result$what_next[i] <- names(z)[2]
    }

    if(annotate) result <- compare_genoprob_add_annotation(result)

    # drop the index columns
    result <- result[,-grep("index", colnames(result))]

    # add matrix of proportion of matches for each possible genotype, as an attribute
    rownames(p_match) <- rownames(result)
    attr(result, "prop_match") <- p_match

    result
}

# from founders, infer all F1 SNP genotypes
infer_f1_geno <-
    function(founders)
{
    n <- nrow(founders)
    n_f1 <- choose(n, 2)
    result <- matrix(0, nrow=n_f1, ncol=ncol(founders))
    dimnames(result) <- list(1:n_f1, colnames(founders))

    f <- rownames(founders)
    k <- 1
    for(i in 1:(n-1)) {
        for(j in (i+1):n) {
            rownames(result)[k] <- paste0(f[i], f[j])

            result[k,founders[i,]==1 & founders[j,]==1] <- 1
            result[k,founders[i,]==3 & founders[j,]==3] <- 3
            result[k,founders[i,]==1 & founders[j,]==3] <- 2
            result[k,founders[i,]==3 & founders[j,]==1] <- 2

            k <- k + 1
        }
    }

    result
}

compare_genoprob_add_annotation <-
    function(x)
{
    if("match_best" %in% names(x)) { # add some annotations
        for(i in 1:nrow(x)) {
            if(x[i,"geno1"]==x[i,"geno2"]) {
                x[i,"geno1"] <- paste0(x[i,"geno1"], " ")
                x[i,"geno2"] <- paste0(x[i,"geno2"], " ")
            } else {
                if(x$match1[i] >= x$match_best[i])
                    x[i,"geno1"] <- paste0(x[i,"geno1"], "*")
                else
                    x[i,"geno1"] <- paste0(x[i,"geno1"], "-")
                if(x$match2[i] >= x$match_best[i])
                    x[i,"geno2"] <- paste0(x[i,"geno2"], "*")
                else
                    x[i,"geno2"] <- paste0(x[i,"geno2"], "-")
            }
        }
    }

    x
}

# two-panel plot with both snp asso and SDP
# (for a single chromosome)
#
# calls plot_snpasso and plot_sdp
# internal function that is called by plot_snpasso
plot_snpasso_and_sdp <-
    function(scan1output, snpinfo, show_all_snps=TRUE,
             drop_hilit=NA, col_hilit="violetred", col="darkslateblue",
             gap=NULL, minlod=0,
             minrow=4, padding=0.2,
             colors=c("black", "red3", "green4", "blue3", "orange"),
             scale_pos=1, start_field="start", stop_field="stop",
             strand_field="strand", name_field="Name",
             top_panel_prop=0.3, panel_prop=NULL, xlim=NULL, xaxt="s",
             xlab=NULL, main="", sub="",
             strain_labels=names(qtl2::CCcolors), ...)
{
    if(is.null(panel_prop)) panel_prop <- c(top_panel_prop, 1-top_panel_prop)

    # 2 x 1 panels; adjust margins
    old_mfrow <- par("mfrow")
    old_mar <- par("mar")
    on.exit(par(mfrow=old_mfrow, mar=old_mar))
    layout(rbind(1,2), heights=panel_prop)
    top_mar <- bottom_mar <- old_mar
    top_mar[1] <- 0.1
    bottom_mar[3] <- 0.1

    if(is.null(xlim)) xlim <- range(snpinfo$pos)

    if(is.null(xlab)) {
        if(length(unique(snpinfo$chr))==1)
            xlab <- paste("Chr", snpinfo$chr[1], "position (Mbp)")
        else
            xlab <- "Position (Mbp)"
    }

    # determine snps to show in SDP plot
    # maybe expand snp info
    map <- snpinfo_to_map(snpinfo)
    if(show_all_snps) {
        tmp <- expand_snp_results(scan1output, map, snpinfo)
        scan1output <- tmp$lod
        map <- tmp$map
    }
    if(is.na(drop_hilit)) drop_hilit <- Inf
    snps2show <- rownames(scan1output)[max(scan1output[,1]) - scan1output[,1] <= drop_hilit]
    snpinfo_sub <- snpinfo[snpinfo$snp %in% snps2show,,drop=FALSE]

    par(mar=top_mar)
    plot_sdp(snpinfo_sub$pos, snpinfo_sub$sdp, strain_labels=strain_labels,
             xlim=xlim, xaxt="n", xlab="", main=main, ...)

    par(mar=bottom_mar)
    plot_snpasso(scan1output, snpinfo, show_all_snps=show_all_snps,
                 drop_hilit=drop_hilit, col_hilit=col_hilit, col=col,
                 gap=gap, minlod=minlod, xlim=xlim, xaxt="n", xlab="",
                 main=main, ...)

}

# viterbi
# Calculate most probable sequence of genotypes
# this version pre-calculates init, step, and emit (attempting to be faster for DO)
#
# Same input and output as viterbi()
viterbi2 <-
    function(cross, map=NULL, error_prob=1e-4,
             map_function=c("haldane", "kosambi", "c-f", "morgan"),
             quiet=TRUE, cores=1)
{
    # check inputs
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')
    if(error_prob < 0)
        stop("error_prob must be > 0")
    map_function <- match.arg(map_function)

    # set up cluster; make quiet=FALSE if cores>1
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores) > 1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # no more messages
    }

    # pseudomarker map
    if(is.null(map)) {
        if(is.null(cross$gmap)) stop("If cross does not contain a genetic map, map must be provided.")
        map <- insert_pseudomarkers(cross$gmap)
    }
    # possibly subset the map
    if(length(map) != length(cross$geno) || !all(names(map) == names(cross$geno))) {
        chr <- names(cross$geno)
        if(!all(chr %in% names(map)))
            stop("map doesn't contain all of the necessary chromosomes")
        map <- map[chr]
    }
    # calculate marker index object
    index <- create_marker_index(lapply(cross$geno, colnames), map)

    rf <- map2rf(map, map_function)

    # deal with missing information
    ind <- rownames(cross$geno[[1]])
    chrnames <- names(cross$geno)
    is_x_chr <- handle_null_isxchr(cross$is_x_chr, chrnames)
    cross$is_female <- handle_null_isfemale(cross$is_female, ind)
    cross$cross_info <- handle_null_isfemale(cross$cross_info, ind)

    founder_geno <- cross$founder_geno
    if(is.null(founder_geno))
        founder_geno <- create_empty_founder_geno(cross$geno)

    by_group_func <- function(i) {
        .viterbi2(cross$crosstype, t(cross$geno[[chr]][group[[i]],,drop=FALSE]),
                  founder_geno[[chr]], cross$is_x_chr[chr], cross$is_female[group[[i]][1]],
                  cross$cross_info[group[[i]][1],], rf[[chr]], index[[chr]],
                  error_prob)
    }

    # split individuals into groups with common sex and cross_info
    sex_crossinfo <- paste(cross$is_female, apply(cross$cross_info, 1, paste, collapse=":"), sep=":")
    group <- split(seq(along=sex_crossinfo), sex_crossinfo)
    names(group) <- NULL
    nc <- n_cores(cores)
    while(nc > length(group) && max(sapply(group, length)) > 1) { # successively split biggest group in half until there are as many groups as cores
        mx <- which.max(sapply(group, length))
        g <- group[[mx]]
        group <- c(group, list(g[seq(1, length(g), by=2)]))
        group[[mx]] <- g[seq(2, length(g), by=2)]
    }
    groupindex <- seq(along=group)

    result <- vector("list", length(cross$geno))
    names(result) <- names(cross$geno)
    for(chr in seq(along=cross$geno)) {
        if(!quiet) message("Chr ", names(cross$geno)[chr])

        # calculations in parallel [if cores==1, it just does lapply()]
        temp <- cluster_lapply(cores, groupindex, by_group_func)

        # paste them back together
        d <- vapply(temp, dim, rep(0,2))
        nr <- sum(d[1,])
        result[[chr]] <- matrix(nrow=nr, ncol=d[2,1])
        for(i in groupindex)
            result[[chr]][group[[i]],] <- temp[[i]]

        dimnames(result[[chr]]) <- list(rownames(cross$geno[[chr]]),
                                        names(map[[chr]]))
    }

    names(result) <- names(cross$geno)
    attr(result, "crosstype") <- cross$crosstype
    attr(result, "is_x_chr") <- cross$is_x_chr
    attr(result, "alleles") <- cross$alleles

    class(result) <- c("viterbi", "list")
    result
}

# interp_genoprob
#' Interpolate genotype probabilities
#'
#' Linear interpolation of genotype probabilities, mostly to get two sets onto the same map for comparison purposes.
#'
#' @param probs Genotype probabilities, as calculated from
#' [calc_genoprob()].
#' @param map List of vectors of map positions.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return An object of class `"calc_genoprob"`, like the input,
#' but with additional positions present in `map`. See [calc_genoprob()].
#'
#' @details We reduce `probs` to the positions present in `map` and then
#' interpolate the genotype probabilities at additional positions
#' in `map` by linear interpolation using the two adjacent
#' positions. Off the ends, we just copy over the first or last
#' value unchanged.
#'
#' In general, it's better to use [insert_pseudomarkers()] and
#' [calc_genoprob()] to get genotype probabilities at additional
#' positions along a chromosome. This function is a **very** crude
#' alternative that was implemented in order to compare genotype
#' probabilities derived by different methods, where we first need to
#' get them onto a common set of positions.
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[1:20,c("1", "X")]}
#' probs <- calc_genoprob(iron, iron$gmap, error_prob=0.002)
#'
#' # you generally wouldn't want to do this, but this is an illustration
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#' probs_map <- interp_genoprob(probs, map)
#'
#' @seealso [calc_genoprob()]
#'
#' @export
interp_genoprob <-
    function(probs, map, cores=1)
{
    if(is.null(probs)) stop("probs is NULL")
    if(is.null(map)) stop("map is NULL")

    gchr <- names(probs)
    mchr <- names(map)
    if(!all(gchr %in% mchr) || !all(mchr %in% gchr)) {
        cchr <- gchr[gchr %in% mchr]
        probs <- probs[,cchr]
        map <- map[cchr]
    }

    # set up cluster
    cores <- setup_cluster(cores)
    result <- cluster_lapply(cores, seq_along(map), interp_genoprob_onechr, probs=probs, map=map)

    names(result) <- names(probs)
    for(x in c("crosstype", "is_x_chr", "alleles", "alleleprobs", "class"))
        attr(result, x) <- attr(probs, x)

    result
}


interp_genoprob_onechr <-
    function(chr, probs, map)
{
    probs <- probs[[chr]]
    map <- map[[chr]]

    d <- dim(probs)
    markers <- dimnames(probs)[[3]]
    pmar <- names(map)

    keep <- (markers %in% pmar)
    if(!any(keep)) stop("No overlapping markers on chr ", chr)
    if(!all(keep)) {
        probs <- probs[,,keep]
        markers <- markers[keep]
    }

    # check that the marker order didn't change
    m <- match(markers, pmar)
    if(any(diff(m) < 0)) stop("probs positions out of order on chr ", chr)

    pos_index <- match(pmar, markers) - 1 # indexes start at 0
    if(!any(is.na(pos_index))) return(probs) # no new positions
    pos_index[is.na(pos_index)] <- -1

    result <- .interp_genoprob_onechr(probs, map, pos_index)
    dimnames(result) <- c(dimnames(probs)[1:2], list(pmar))

    result
}

context("interpolation of genotype probabilities")

test_that("interp_genoprob works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:20,c("1", "2", "X")]
    probs <- calc_genoprob(iron, iron$gmap, error_prob=0.002)

    # no added positions
    expect_equal(interp_genoprob(probs, iron$gmap), probs)

    # mess up the order
    map <- iron$gmap
    map[[2]] <- map[[2]][c(1,3,2,4,5)]
    map[[2]][2:3] <- c(48.1, 56.8)
    expect_error(interp_genoprob(probs, map))

    # add a couple of pseudomarkers
    map <- insert_pseudomarkers(iron$gmap, step=20, stepwidth="max")
    expected <- probs
    expected[[1]] <- array(dim=c(nrow(probs[[1]]), ncol(probs[[1]]), length(map[[1]])))
    dimnames(expected[[1]]) <- c(dimnames(probs[[1]])[1:2], list(names(map[[1]])))
    expected[[1]][,,c(1,3,6)] <- probs[[1]]
    expected[[1]][,,2] <- (probs[[1]][,,1] + probs[[1]][,,2])/2
    p <- (map[[1]][4]-map[[1]][3])/(map[[1]][6]-map[[1]][3])
    expected[[1]][,,4] <- probs[[1]][,,2]*(1-p) + probs[[1]][,,3]*p
    p <- (map[[1]][5]-map[[1]][3])/(map[[1]][6]-map[[1]][3])
    expected[[1]][,,5] <- probs[[1]][,,2]*(1-p) + probs[[1]][,,3]*p
    expected[[3]] <- array(dim=c(nrow(probs[[3]]), ncol(probs[[3]]), length(map[[3]])))
    dimnames(expected[[3]]) <- c(dimnames(probs[[3]])[1:2], list(names(map[[3]])))
    expected[[3]][,,c(1,3)] <- probs[[3]]
    expected[[3]][,,2] <- (probs[[3]][,,1] + probs[[3]][,,2])/2

    expect_equal(interp_genoprob(probs, map), expected)

    # add some positions off the ends
    map[[1]] <- c(off_c1=10, map[[1]])
    map[[2]] <- c(off_c2a=10, off_c2b=20, map[[2]], off_c2c=80)
    map[[3]] <- c(off_cXa=10, map[[3]], off_cXb=60)

    expected2 <- expected
    for(i in 1:3) {
        expected2[[i]] <- array(dim=c(nrow(probs[[i]]), ncol(probs[[i]]), length(map[[i]])))
        dimnames(expected2[[i]]) <- c(dimnames(probs[[i]])[1:2], list(names(map[[i]])))
    }
    expected2[[1]][,,1:7] <- expected[[1]][,,c(1,1:6)]
    expected2[[2]][,,1:8] <- expected[[2]][,,c(1,1,1:5,5)]
    expected2[[3]][,,1:5] <- expected[[3]][,,c(1,1:3,3)]

    expect_equal(interp_genoprob(probs, map), expected2)

})

#' Count numbers of crossovers
#'
#' Estimate the numbers of crossovers in each individual on each chromosome.
#'
#' @param geno List of matrices of genotypes (output of [maxmarg()] or [viterbi()])
#' or a list of 3d-arrays of genotypes (output of [sim_geno()]).
#' @param quiet If FALSE, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A matrix of crossover counts, individuals x chromosomes, or
#' (if the input was the output of [sim_geno()]) a
#' 3d-array of crossover counts, individuals x chromosomes x
#' imputations.
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[1:20,1:2] # subset to first 20 individuals and to chr 1 and 2}
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#' pr <- calc_genoprob(iron, map, error_prob=0.002, map_function="c-f")
#' g <- maxmarg(pr)
#' n_xo <- count_xo(g)
#'
#' # imputations
#' imp <- sim_geno(iron, map, error_prob=0.002, map_function="c-f", n_draws=32)
#' n_xo_imp <- count_xo(imp)
#' # sums across chromosomes
#' tot_xo_imp <- apply(n_xo_imp, c(1,3), sum)
#' # mean and SD across imputations
#' summary_xo <- cbind(mean=rowMeans(tot_xo_imp),
#'                     sd=apply(tot_xo_imp, 1, sd))
#'
#' @seealso [locate_xo()]
#'
#' @importFrom stats setNames
#' @export

count_xo <-
    function(geno, quiet=TRUE, cores=1)
{
    if(is.cross2(geno))
        stop('Input geno is a "cross2" object but should be genotypes as from viterbi, maxmarg, or sim_geno')

    # set up cluster; set quiet=TRUE if multi-core
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    crosstype <- attr(geno, "crosstype")
    if(is.null(crosstype))
        stop("Input geno needs to include a crosstype attribute.")
    is_x_chr <- attr(geno, "is_x_chr")
    if(is.null(is_x_chr))
        is_x_chr <- rep(FALSE, length(geno))

    # the case of 3d-arrays from sim_geno
    if(length(dim(geno[[1]])) == 3) {
        nind <- vapply(geno, nrow, 1)
        if(length(unique(nind)) > 1)
            stop("Input chromosomes have varying numbers of individuals")
        nchr <- length(geno)
        nind <- nind[1]
        ndraws <- vapply(geno, function(a) dim(a)[3], 1)
        if(length(unique(ndraws)) > 1)
            stop("Input chromosomes have varying numbers of imputations")
        ndraws <- ndraws[1]

        result <- array(dim=stats::setNames(c(nind, ndraws, nchr), NULL))
        dimnames(result) <- list(rownames(geno[[1]]), NULL, names(geno))

        by_chr_func <- function(chr) {
            result <- .count_xo_3d(aperm(geno[[chr]], c(2,1,3)), crosstype, is_x_chr[chr])
            names(result) <- rownames(geno[[chr]])
            result
        }

        result_list <- cluster_lapply(cores, seq(along=geno), by_chr_func)

        for(i in seq(nchr))
            result[,,i] <- result_list[[i]]

        return(aperm(result, c(1,3,2)))
    }

    # the case of matrices from viterbi or maxmarg
    by_chr_func <- function(chr) {
        result <- .count_xo(t(geno[[chr]]), crosstype, is_x_chr[chr])
        names(result) <- rownames(geno[[chr]])
        result
    }

    result_list <- cluster_lapply(cores, seq(along=geno), by_chr_func)

    result <- do.call("cbind", result_list)
    colnames(result) <- names(geno)
    result
}

#' Plot gene locations for a genomic interval
#'
#' Plot gene locations for a genomic interval, as rectangles with gene
#' symbol (and arrow indicating strand/direction) below.
#'
#' @param genes Data frame containing `start` and `stop` in
#' Mbp, `strand` (as `"-"`, `"+"`, or `NA`), and
#' `Name`.
#' @param minrow Minimum number of rows of genes in the plot
#' @param padding Proportion to pad with white space around the genes
#' @param colors Vectors of colors, used sequentially and then re-used.
#' @param scale_pos Factor by which to scale position (for example, to convert basepairs to Mbp)
#' @param start_field Character string with name of column containing the genes' start positions.
#' @param stop_field Character string with name of column containing the genes' stop positions.
#' @param strand_field Character string with name of column containing the genes' strands.
#'   (The values of the corresponding field can be character strings `"+"` or `"-"`, or numeric +1 or -1.)
#' @param name_field Character string with name of column containing the genes' names.
#' @param ... Optional arguments passed to `plot()`.
#'
#' @return None.
#'
#' @keywords hgraphics
#' @export
#' @importFrom graphics strheight strwidth text par rect abline box
#'
#' @section Hidden graphics parameters:
#' Graphics parameters can be passed via `...`. For
#' example, `xlim` to control the x-axis limits.
#' These are not included as formal
#'
#' @examples
#' genes <- data.frame(chr = c("6", "6", "6", "6", "6", "6", "6", "6"),
#'                     start = c(139988753, 140680185, 141708118, 142234227, 142587862,
#'                               143232344, 144398099, 144993835),
#'                     stop  = c(140041457, 140826797, 141773810, 142322981, 142702315,
#'                               143260627, 144399821, 145076184),
#'                     strand = c("-", "+", "-", "-", "-", NA, "+", "-"),
#'                     Name = c("Plcz1", "Gm30215", "Gm5724", "Slco1a5", "Abcc9",
#'                              "4930407I02Rik", "Gm31777", "Bcat1"),
#'                     stringsAsFactors=FALSE)
#'
#' # use scale_pos=1e-6 because data in bp but we want the plot in Mbp
#' plot_genes(genes, xlim=c(140, 146), scale_pos=1e-6)

# create an empty plot with test x- and y-axis limits
plot_genes <-
    function(genes, minrow=4, padding=0.2,
             colors=c("black", "red3", "green4", "blue3", "orange"),
             scale_pos=1, start_field="start", stop_field="stop",
             strand_field="strand", name_field="Name", ...)
{
    if(is.null(genes)) stop("genes is NULL")

    # make sure the columns are there
    fields <- c(start_field, stop_field, strand_field, name_field)
    fields_found <- fields %in% colnames(genes)
    if(!all(fields_found)) {
        stop("Columns not found: ", paste(fields[!fields_found], collapse=", "))
    }

    if(!is_pos_number(minrow)) stop("minrow should be a positive integer")
    if(!is_nonneg_number(padding)) stop("padding should be a non-negative number")
    if(!is_pos_number(scale_pos)) stop("scale_pos should be a positive number")

    # grab just the start and stop
    start <- genes[,start_field]
    end <- genes[,stop_field]

    # drop genes with missing start or stop
    missing_pos <- is.na(start) | is.na(end)
    if(any(missing_pos)) {
        warning("Dropping ", sum(missing_pos), " rows with missing positions")
        genes <- genes[!missing_pos, , drop=FALSE]
    }

    # make sure genes are ordered by their start values
    if(any(diff(start) < 0))
        genes <- genes[order(start, end),]

    # grab data
    start <- genes[,start_field]*scale_pos # convert to Mbp
    end <- genes[,stop_field]*scale_pos    # convert to Mbp
    strand <- as.character(genes[,strand_field])
    name <- as.character(genes[,name_field])

    # deal with +/- 1 for strand
    strand[!is.na(strand) & (strand=="1" | strand=="+1")] <- "+"
    strand[!is.na(strand) & strand=="-1"] <- "-"

    plot_genes_internal <-
        function(xlab=NULL, xaxs="i",
                 bgcolor="gray92", xat=NULL,
                 mgp=c(1.6,0.2,0), xlim=NULL,
                 vlines=NULL, vlines_col="white",
                 vlines_lwd=1, vlines_lty=1,
                 xaxt="s", ylab="", ...)
        {
            if(is.null(xlab)) {
                if(length(unique(genes$chr)) == 1)
                    xlab <- paste("Chr", genes$chr[1], "position (Mbp)")
                else xlab <- "Position (Mbp)"
            }

            if(is.null(xlim)) {
                xlim <- range(c(start, end), na.rm=TRUE)
            }

            plot(0, 0, type="n",
                 xlim=xlim,   xlab=xlab, xaxs=xaxs, xaxt="n",
                 ylim=c(1,0), ylab=ylab, yaxs="i", yaxt="n", mgp=mgp, ...)

            # gray background
            u <- par("usr")
            rect(u[1], u[3], u[2], u[4], col=bgcolor, border="black")

            # axis
            if(is.null(xat)) xat <- pretty(xlim)
            if((length(xat) > 1 || !is.na(xat)) && xaxt != "n")
                axis(side=1, at=xat, mgp=mgp, tick=FALSE)

            # vertical lines
            if(is.null(vlines)) vlines <- xat
            if(length(vlines) > 1 || !is.na(vlines))
                abline(v=vlines, col=vlines_col, lwd=vlines_lwd, lty=vlines_lty)

            box()
        }

    plot_genes_internal(...)

    # drop genes that are not in plotting region
    u <- par("usr")
    omit <- (end < u[1] | start > u[2])
    if(any(omit)) {
        keep <- !omit
        start <- start[keep]
        end <- end[keep]
        strand <- strand[keep]
        name <- name[keep]
    }
    if(length(start) == 0) # no genes to plot
        return(invisible(NULL))

    # missing names: use ?
    name[is.na(name)] <- "?"

    # arrow annotation re direction, to place after gene name
    dir_symbol <- rep(' ', length(name))
    right <- !is.na(strand) & strand == "+"
    if(any(right))
        dir_symbol[right] <- expression(phantom('') %->% phantom(''))
    left <- !is.na(strand) & strand == "-"
    if(any(left))
        dir_symbol[left] <- expression(phantom('') %<-% phantom(''))

    # initial determination of text size
    maxy <- minrow
    height <- 1/maxy
    text_cex <- 1

    # adjust text size and determine vertical location of genes
    for(it in 1:2) { # go through all of this twice

        while(max(abs(strheight(name, cex=text_cex))) > height*(1-padding)) {
            text_cex <- text_cex * 0.99
        }

        # horizontal padding
        space <- strwidth(' ', cex=text_cex)

        # figure out how to arrange genes vertically
        #     + number of rows of genes
        # (function defined in src/arrange_genes.cpp)
        y <- arrange_genes(start, end + space + strwidth(name, cex=text_cex) + strwidth(dir_symbol, cex=text_cex))

        maxy <- max(c(y, minrow))
        height <- 1/maxy
    }

    ypos <- seq(height/2, by=height, length=maxy)
    y <- ypos[y]
    rect_height <- height*(1-padding)
    rect_top <- y - rect_height/2
    rect_bottom <- y + rect_height/2

    colors <- rep(colors, length(y))
    for(i in seq(along=start)) {
        rect(start[i], rect_top[i],
             end[i],   rect_bottom[i],
             col=colors[i], border=colors[i],
             lend=1, ljoin=1)
        text(end[i] + space, y[i],
             name[i], adj=c(0, 0.5), col=colors[i],
             cex=text_cex)
        if(!is.na(strand[i]) && (strand[i] == "+" || strand[i] == '-'))
            text(end[i] + space + strwidth(name[i], cex=text_cex), y[i],
                 dir_symbol[i], adj=c(0, 0.5), col=colors[i],
                 cex=text_cex)
    }
}

# calc_genoprob
#' Calculate conditional genotype probabilities
#'
#' Uses a hidden Markov model to calculate the probabilities of the
#' true underlying genotypes given the observed multipoint marker
#' data, with possible allowance for genotyping errors.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param map Genetic map of markers. May include pseudomarker
#' locations (that is, locations that are not within the marker
#' genotype data). If NULL, the genetic map in `cross` is used.
#' @param error_prob Assumed genotyping error probability
#' @param map_function Character string indicating the map function to
#' use to convert genetic distances to recombination fractions.
#' @param lowmem If `FALSE`, split individuals into groups with
#' common sex and crossinfo and then precalculate the transition
#' matrices for a chromosome; potentially a lot faster but using more
#' memory.
#' @param quiet If `FALSE`, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return An object of class `"calc_genoprob"`: a list of three-dimensional arrays of probabilities,
#'     individuals x genotypes x positions. (Note that the arrangement is
#'     different from R/qtl.) Also contains four attributes:
#' * `crosstype` - The cross type of the input `cross`.
#' * `is_x_chr` - Logical vector indicating whether chromosomes
#'     are to be treated as the X chromosome or not, from input `cross`.
#' * `alleles` - Vector of allele codes, from input
#'     `cross`.
#' * `alleleprobs` - Logical value (`FALSE`) that
#'     indicates whether the probabilities are compressed to allele
#'     probabilities, as from [genoprob_to_alleleprob()].
#'
#' @details
#'   Let \eqn{O_k}{O[k]} denote the observed marker genotype at position
#'  \eqn{k}, and \eqn{g_k}{g[k]} denote the corresponding true underlying
#'  genotype.
#'
#'  We use the forward-backward equations to calculate
#'  \eqn{\alpha_{kv} = \log Pr(O_1, \ldots, O_k, g_k = v)}{%
#'    a[k][v] = log Pr(O[1], \ldots, O[k], g[k] = v)}
#'  and
#'  \eqn{\beta_{kv} = \log Pr(O_{k+1}, \ldots, O_n | g_k = v)}{%
#'    b[k][v] = log Pr(O[k+1], \ldots, O[n] | g[k] = v)}
#'
#'  We then obtain
#'  \eqn{Pr(g_k | O_1, \ldots, O_n) = \exp(\alpha_{kv} + \beta_{kv}) / s}{%
#'    Pr(g[k] | O[1], \ldots, O[n] = exp(a[k][v] + b[k][v]) / s}
#'  where
#'  \eqn{s = \sum_v \exp(\alpha_{kv} + \beta_{kv})}{%
#'    s = sum_v exp(a[k][v] + b[k][v])}
#'
#' @export
#' @keywords utilities
#' @seealso [insert_pseudomarkers()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' gmap_w_pmar <- insert_pseudomarkers(grav2$gmap, step=1)
#' probs <- calc_genoprob(grav2, gmap_w_pmar, error_prob=0.002)

calc_genoprob <-
function(cross, map=NULL, error_prob=1e-4,
         map_function=c("haldane", "kosambi", "c-f", "morgan"),
         lowmem=FALSE, quiet=TRUE, cores=1)
{
    # check inputs
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')
    if(!is_nonneg_number(error_prob)) stop("error_prob should be a single non-negative number")
    map_function <- match.arg(map_function)

    if(!lowmem) { # use other version
        return(calc_genoprob2(cross=cross, map=map,
                              error_prob=error_prob, map_function=map_function,
                              quiet=quiet, cores=cores))
    }

    # set up cluster; set quiet=TRUE if multi-core
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # pseudomarker map
    if(is.null(map)) {
        if(is.null(cross$gmap)) stop("If cross does not contain a genetic map, map must be provided.")
        map <- insert_pseudomarkers(cross$gmap)
    }
    # possibly subset the map
    if(length(map) != length(cross$geno) || !all(names(map) == names(cross$geno))) {
        chr <- names(cross$geno)
        if(!all(chr %in% names(map)))
            stop("map doesn't contain all of the necessary chromosomes")
        map <- map[chr]
    }
    # calculate marker index object
    index <- create_marker_index(lapply(cross$geno, colnames), map)

    probs <- vector("list", length(map))
    rf <- map2rf(map, map_function)

    # deal with missing information
    ind <- rownames(cross$geno[[1]])
    chrnames <- names(cross$geno)
    is_x_chr <- handle_null_isxchr(cross$is_x_chr, chrnames)
    cross$is_female <- handle_null_isfemale(cross$is_female, ind)
    cross$cross_info <- handle_null_isfemale(cross$cross_info, ind)

    founder_geno <- cross$founder_geno
    if(is.null(founder_geno))
        founder_geno <- create_empty_founder_geno(cross$geno)

    by_group_func <- function(i) {
        pr <- .calc_genoprob(cross$crosstype, t(cross$geno[[chr]][group[[i]],,drop=FALSE]),
                             founder_geno[[chr]], cross$is_x_chr[chr], cross$is_female[group[[i]]],
                             t(cross$cross_info[group[[i]],,drop=FALSE]), rf[[chr]], index[[chr]],
                             error_prob)
        aperm(pr, c(2,1,3))
    }

    group <- parallel::splitIndices(nrow(cross$geno[[1]]), n_cores(cores))
    groupindex <- seq(along=group)

    probs <- vector("list", length(cross$geno))
    for(chr in seq(along=cross$geno)) {
        if(!quiet) message("Chr ", names(cross$geno)[chr])

        if(n_cores(cores) == 1) { # no parallel processing
            # calculations in one group
            probs[[chr]] <- by_group_func(1)
        }
        else {
            # calculations in parallel
            temp <- cluster_lapply(cores, groupindex, by_group_func)

            # paste them back together
            d <- vapply(temp, dim, rep(0,3))
            nr <- sum(d[1,])
            probs[[chr]] <- array(dim=c(nr, d[2,1], d[3,1]))
            for(i in groupindex)
                probs[[chr]][group[[i]],,] <- temp[[i]]
        }

        # genotype names
        alleles <- cross$alleles
        if(is.null(alleles)) # no alleles saved; use caps
            alleles <- LETTERS[1:ncol(probs[[chr]])]
        gnames <- geno_names(cross$crosstype,
                             alleles,
                             cross$is_x_chr[chr])
        if(length(gnames) != ncol(probs[[chr]])) {
            warning("genotype names has length (", length(gnames),
                    ") != genoprob dim (", ncol(probs[[chr]]), ")")
            gnames <- NULL
        }

        dimnames(probs[[chr]]) <- list(rownames(cross$geno[[chr]]),
                                       gnames,
                                       names(map[[chr]]))

    }

    names(probs) <- names(cross$geno)

    attr(probs, "crosstype") <- cross$crosstype
    attr(probs, "is_x_chr") <- cross$is_x_chr
    attr(probs, "alleles") <- cross$alleles
    attr(probs, "alleleprobs") <- FALSE

    class(probs) <- c("calc_genoprob", "list")

    probs
}

context("rbind and cbind calc_genoprob")

test_that("rbind.calc_genoprob works for grav", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2, map, error_prob=0.002)
    probsA <- probs[1:5,]
    probsB <- probs[6:12,]
    probsC <- probs[13:20,]
    probsAB <- probs[1:12,]
    probsABC <- probs[1:20,]
    probsBACA <- probs[c(6:12, 1:5, 13:20, 1:5),]

    expect_equal(rbind(probsA, probsB), probsAB)
    expect_equal(rbind(probsA, probsB, probsC), probsABC)
    expect_equal(rbind(probsB, probsA, probsC, probsA), probsBACA)

})


test_that("rbind.calc_genoprob works for iron", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    probsA <- probs[2:20,]
    probsB <- probs[41:60,]
    probsC <- probs[102:201,]
    probsAB <- probs[c(2:20,41:60),]
    probsABC <- probs[c(2:20,41:60,102:201),]
    probsBACA <- probs[c(41:60,2:20,102:201,2:20),]

    expect_equal(rbind(probsA, probsB), probsAB)
    expect_equal(rbind(probsA, probsB, probsC), probsABC)
    expect_equal(rbind(probsB, probsA, probsC, probsA), probsBACA)

})

test_that("cbind.calc_genoprob works for grav", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2[1:10,], map, error_prob=0.002)
    probsA <- probs[,1:2]
    probsB <- probs[,5]
    probsC <- probs[,3:4]
    probsAB <- probs[,c(1:2,5)]
    probsABC <- probs[,c(1,2,5,3,4)]
    probsBACA <- probs[,c(5,1:2,3:4,1:2)]

    expect_equal(cbind(probsA, probsB), probsAB)
    expect_equal(cbind(probsA, probsB, probsC), probsABC)
    expect_equal(cbind(probsB, probsA, probsC, probsA), probsBACA)

})


test_that("cbind.calc_genoprob works for iron", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron[6:21,], map, error_prob=0.002)
    probsA <- probs[,2:3]
    probsB <- probs[,c(4,5,8)]
    probsC <- probs[,c(19,"X")]
    probsAB <- probs[,c(2:3,4,5,8)]
    probsABC <- probs[,c(2:3,4,5,8,19,"X")]
    probsBACA <- probs[,c(4,5,8,2,3,19,"X",2,3)]

    expect_equal(cbind(probsA, probsB), probsAB)
    expect_equal(cbind(probsA, probsB, probsC), probsABC)
    expect_equal(cbind(probsB, probsA, probsC, probsA), probsBACA)

})

context("pull genotype probabilities for a specified position")

test_that("pull_genoprobpos works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(2,8,"X")]
    gmap <- insert_pseudomarkers(iron$gmap, step=1)
    pr <- calc_genoprob(iron, gmap, error_prob=0.002)

    pmar <- find_marker(gmap, 8, 40)
    pr_8_40 <- pull_genoprobpos(pr, pmar)
    expect_equal(pr_8_40, pr[["8"]][,,"c8.loc40"])
    expect_equal(pr_8_40, pull_genoprobpos(pr, marker=pmar))

    expect_equal(pull_genoprobpos(pr["275",], pmar),
                 pr_8_40["275",,drop=FALSE])

    expect_warning(pull_genoprobpos(pr, c("c8.loc40", find_marker(gmap, 8, 70))))

    expect_error(pull_genoprobpos(pr, "karl"))


    # same result if you provide gmap, chr, pos
    expect_equal(pr_8_40, pull_genoprobpos(pr, gmap, 8, 40))

})

# functions to check the integrity of QTL cross data

# check if a crosstype is supported
#   if give_error=TRUE, problem results in error
#   if give_error=FALSE, returns TRUE if supported; FALSE if not
check_crosstype <-
function(crosstype, give_error=TRUE)
{
    z <- try(.crosstype_supported(crosstype), silent=TRUE)

    if(inherits(z, "try-error") || !z) {
        if(give_error)
            stop("Cross type ", crosstype, " not yet supported.")
        return(invisible(FALSE))
    }

    invisible(TRUE) # if no error, return TRUE
}


# count the number of invalid genotypes
# returns a matrix individuals x chromosomes
count_invalid_genotypes <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input cross must have class "cross2"')

    result <- matrix(nrow=nrow(cross2$geno[[1]]),
                     ncol=length(cross2$geno))
    dimnames(result) <- list(rownames(cross2$geno[[1]]),
                             names(cross2$geno))

    ind <- rownames(cross2$geno[[1]])

    # handle case of missing cross_info or is_female
    cross_info <- handle_null_crossinfo(cross2$cross_info, ind)
    is_female <- handle_null_isfemale(cross2$is_female, ind)
    is_x_chr <- handle_null_isxchr(cross2$is_x_chr, names(cross2$geno))

    cross_info <- t(cross_info)

    for(i in seq(along=cross2$geno))
        result[,i] <- .count_invalid_genotypes(cross2$crosstype,
                                               t(cross2$geno[[i]]),
                                               is_x_chr[i],
                                               is_female,
                                               cross_info)

    result
}

# check_cross2
#' Check a cross2 object
#'
#' Check the integrity of the data within a cross2 object.
#'
#' @param cross2 An object of class `"cross2"`, as output by [read_cross2()]. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#'
#' @return If everything is correct, returns `TRUE`; otherwise `FALSE`,
#' with attributes that give the problems.
#'
#' @details Checks whether a cross2 object meets the
#' specifications. Problems are issued as warnings.
#'
#' @export
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' check_cross2(grav2)

check_cross2 <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input cross must have class "cross2"')

    result <- TRUE

    # required pieces
    #     crosstype
    #     geno
    #     gmap
    #     is_x_chr

    # sometimes required
    #     is_female
    #     cross_info
    #     founder_geno (required for some crosstypes)

    # optional pieces
    #     pheno
    #     covar
    #     phenocovar
    #     pmap
    #     alleles

    crosstype <- cross2$crosstype
    if(is.null(crosstype)) {
        result <- FALSE
        warning("crosstype is missing")
    }
    else if(!check_crosstype(cross2$crosstype)) {
        result <- FALSE
        warning("Crosstype ", cross2$crosstype, " is not supported")
    }

    # check for required pieces
    geno <- cross2$geno
    if(is.null(geno)) {
        result <- FALSE
        warning("geno is missing")
    }

    gmap <- cross2$gmap
    if(is.null(gmap)) {
        result <- FALSE
        warning("gmap is missing")
    }

    # if either of those pieces is missing, we'll just return
    if(!result) return(result)

    # generally required pieces, but not worth stopping
    is_x_chr <- cross2$is_x_chr
    if(is.null(is_x_chr)) {
        result <- FALSE
        warning("is_x_chr is missing")
        is_x_chr <- rep(FALSE, length(geno))
        names(is_x_chr) <- names(geno)
        cross2$is_x_chr <- is_x_chr
    }
    else {
        chr <- names(cross2$geno)
        # double-check X chr in case that one of the chromosomes is called "X" or "x"
        if(any(chr=="X" | chr=="x")) {
          if(!all(is_x_chr[chr=="X" | chr=="x"]) ||
             any(is_x_chr[chr!="X" & chr!="x"]))
              warning("Suspicious indication of X chromosome:\n",
                      "    autosomes: ", paste(chr[!is_x_chr], collapse=" "), "\n",
                      "    X chr: ", paste(chr[is_x_chr], collapse=" "))
      }
    }

    if(!check_handle_x_chr(crosstype, any(is_x_chr))) {
        result <- FALSE
        warning("X chr not handled for cross type ", crosstype)
    }

    is_female <- cross2$is_female
    to_pass <- is_female
    if(is.null(to_pass)) to_pass <- logical(0)
    if(!check_is_female_vector(crosstype, to_pass, any(is_x_chr))) {
        result <- FALSE
        if(is.null(is_female)) warning("is_female is missing")
        else warning("is_female is misspecified")
    }

    cross_info <- cross2$cross_info
    to_pass <- cross_info
    if(is.null(to_pass)) to_pass <- matrix(0L,0,0)
    if(!check_crossinfo(crosstype, to_pass, any(is_x_chr))) {
        result <- FALSE
        if(is.null(cross_info)) warning("cross_info is missing")
        else warning("cross_info is misspecified")
    }

    # pheno
    pheno <- cross2$pheno
    if(!is.null(pheno)) { # pheno optional
        if(is.null(rownames(pheno))) {
            result <- FALSE
            warning("rownames(pheno) is NULL")
        }
        if(is.null(colnames(pheno))) {
            result <- FALSE
            warning("colnames(pheno) is NULL")
        }
        if(!is.matrix(pheno)) {
            result <- FALSE
            warning("pheno is not a matrix")
        }
        if(storage.mode(pheno) != "double") {
            result <- FALSE
            warning("pheno is not stored as doubles but rather ", storage.mode(pheno))
        }

        # covar
        covar <- cross2$covar
        if(!is.null(covar)) { # covar is optional
            if(nrow(pheno) != nrow(covar)) {
                result <- FALSE
                warning("nrow(pheno) (", nrow(pheno), ") != nrow(covar) (", nrow(covar), ")")
            }
            else if(any(rownames(pheno) != rownames(covar))) {
                result <- FALSE
                warning("rownames(pheno) != rownames(covar)")
            }
        }

        # phenocovar
        phenocovar <- cross2$phenocovar
        if(!is.null(phenocovar)) { # phenocovar is optional
            if(ncol(pheno) != nrow(phenocovar)) {
                result <- FALSE
                warning("ncol(pheno) (", ncol(pheno), ") != nrow(phenocovar) (", nrow(phenocovar), ")")
            }
            else if(any(colnames(pheno) != rownames(phenocovar))) {
                result <- FALSE
                warning("colnames(pheno) != rownames(phenocovar)")
            }
        }
    }

    # check dimnames of geno
    if(is.null(names(geno))) {
        result <- FALSE
        warning("names(geno) is missing")
    }
    for(i in seq(along=geno)) {
        if(is.null(rownames(geno[[i]]))) {
            result <- FALSE
            warning("missing geno rownames for chr ", names(geno)[i])
        }
        if(is.null(colnames(geno[[i]]))) {
            result <- FALSE
            warning("missing geno colnames for chr ", names(geno)[i])
        }
        if(nrow(geno[[i]]) != nrow(geno[[1]])) {
            result <- FALSE
            warning("nrow in geno for chr ", names(geno)[i],
                    " (", nrow(geno[[i]]), ") doesn't match that for chr ",
                    names(geno)[1], " (", nrow(geno[[1]]), ")")
        }
    }

    # compare geno to gmap
    if(length(geno) != length(gmap)) {
        result <- FALSE
        warning("length(geno) (", length(geno), ") != length(gmap) (", length(gmap), ")")
    }
    else {
        if(is.null(names(gmap))) {
            result <- FALSE
            warning("names(geno) is missing")
        }
        else if(any(names(geno) != names(gmap))) {
            result <- FALSE
            warning("names(geno) != names(gmap)")
        }
        for(i in seq(along=geno)) {
            if(ncol(geno[[i]]) != length(gmap[[i]])) {
                result <- FALSE
                warning("Mismatch between geno and gmap in no. markers on chr ", names(geno)[i])
            }
            else if(any(colnames(geno[[i]]) != names(gmap[[i]]))) {
                result <- FALSE
                warning("Mismatch in marker names between geno and gmap on chr ", names(geno)[i])
            }
        }
    }

    # Markers in gmap in non-decreasing order
    d <- vapply(gmap, function(x) ifelse(length(x)==1, 0, min(diff(x))), 0)
    if(any(d < 0)) {
        result <- FALSE
        warning("Genetic map out of order on chr ", paste(names(gmap)[d < 0], collapse=", "))
    }

    # compare geno to is_female
    is_female <- cross2$is_female
    if(!is.null(is_female)) {
        if(nrow(geno[[1]]) != length(is_female)) {
            result <- FALSE
            warning("length(is_female) (", length(is_female), ") != nrow(geno[[1]]) (", nrow(geno[[1]]), ")")
        }
        else if(any(names(is_female) != rownames(geno[[1]]))) {
            result <- FALSE
            warning("names(is_female) != rownames(geno[[1]])")
        }
        if(!is.logical(is_female)) {
            result <- FALSE
            warning("is_female is not logical")
        }
        if(any(is.na(is_female))) {
            result <- FALSE
            warning("is_female has ", sum(is.na(is_female)), " missing values")
        }
    }

    # compare geno to cross_info
    if(!is.null(cross_info)) {
        if(nrow(geno[[1]]) != nrow(cross_info)) {
            result <- FALSE
            warning("nrow(cross_info) (", nrow(cross_info), ") != nrow(geno[[1]]) (", nrow(geno[[1]]), ")")
        }
        else if(any(rownames(cross_info) != rownames(geno[[1]]))) {
            result <- FALSE
            warning("rownames(cross_info) != rownames(geno[[1]])")
        }
        if(!is.matrix(cross_info)) {
            result <- FALSE
            warning("cross_info is not a matrix")
        }
        if(storage.mode(cross_info) != "integer") {
            result <- FALSE
            warning("cross_info is not stored as integers but rather ", storage.mode(cross_info))
        }
    }

    # compare individuals in geno and pheno
    if(!is.null(pheno)) {
        id <- find_common_ids(rownames(pheno), rownames(geno[[1]]))
        if(length(id) == 0)
            warning("No individuals in common between geno and pheno")
    }

    # check is_x_chr
    if(length(is_x_chr) != length(geno)) {
        result <- FALSE
        warning("length(is_x_chr) (", length(is_x_chr), ") != length(geno) (", length(geno), ")")
    }
    else if(any(names(is_x_chr) != names(geno))) {
        result <- FALSE
        warning("names(is_x_chr) != names(geno)")
    }

    # founder_geno
    founder_geno <- cross2$founder_geno
    if(need_founder_geno(crosstype)) {
        if(is.null(founder_geno)) {
            result <- FALSE
            warning("founder_geno not provided but needed.")
        }
        else {

            if(length(geno) != length(founder_geno)) {
                result <- FALSE
                warning("length(geno) (", length(geno), ") != length(founder_geno) (", length(founder_geno), ")")
            }
            else {
                if(is.null(names(founder_geno))) {
                    result <- FALSE
                    warning("names(founder_geno) is NULL")
                }
                else if(any(names(geno) != names(founder_geno))) {
                    result <- FALSE
                    warning("names(geno) != names(founder_geno)")
                }
                for(i in seq(along=geno)) {
                    if(!check_founder_geno_size(crosstype, founder_geno[[i]], ncol(geno[[i]]))) {
                        result <- FALSE
                        warning("founder_geno has incorrect size on chr ", names(geno)[i])
                    }
                }
                for(i in seq(along=geno)) {
                    if(ncol(geno[[i]]) != ncol(founder_geno[[i]])) {
                        result <- FALSE
                        warning("Mismatch in number of markers between geno and founder_geno on chr ", names(geno)[i])
                    }
                    else if(any(colnames(geno[[i]]) != colnames(founder_geno[[i]]))) {
                        result <- FALSE
                        warning("Mismatch in marker names between geno and founder_geno on chr ", names(geno)[i])
                    }

                    # check values
                    if(!test_founder_geno_values(crosstype, founder_geno[[i]])) {
                        result <- FALSE
                    }
                }

                # checkout founder strain labels
                founders <- founders(cross2)
                for(i in seq(along=geno)) {
                    if(length(founders) != nrow(founder_geno[[i]]) ||
                       any(founders != rownames(founder_geno[[i]]))) {
                        warning("Mismatch in founder strain labels on chr ", names(geno)[i])
                        result <- FALSE
                    }
                }

            }
        }
    } else { # founder genotypes not needed
        if(!is.null(founder_geno)) {
            result <- FALSE
            warning('founder_geno provided but not needed for cross type "', crosstype, '"')
        }
    }

    # pmap
    pmap <- cross2$pmap
    if(!is.null(pmap)) { # pmap is optional
        if(length(gmap) != length(pmap)) {
            result <- FALSE
            warning("length(gmap) (", length(gmap), ") != length(pmap) (", length(pmap), ")")
        }
        else {
            if(is.null(names(pmap))) {
                result <- FALSE
                warning("names(pmap) is NULL")
            }
            else if(any(names(gmap) != names(pmap))) {
                result <- FALSE
                warning("names(gmap) != names(pmap)")
            }
            for(i in seq(along=gmap)) {
                if(length(gmap[[i]]) != length(pmap[[i]])) {
                    result <- FALSE
                    warning("Mismatch between gmap and pmap in no. markers on chr ", names(gmap)[i])
                }
                else if(any(colnames(gmap[[i]]) != names(pmap[[i]]))) {
                    result <- FALSE
                    warning("Mismatch in marker names between gmap and pmap on chr ", names(gmap)[i])
                }
            }
        }

        # markers in order?
        d <- vapply(pmap, function(x) ifelse(length(x)==1, 0, min(diff(x))), 0)
        if(any(d < 0)) {
            result <- FALSE
            warning("Physical map out of order on chr ", paste(names(pmap)[d < 0], collapse=", "))
        }
    }

    # check alleles
    alleles <- cross2$alleles
    if(!is.null(alleles)) {
        if(!is.character(alleles)) {
            result <- FALSE
            warning("alleles should be a vector of character strings but is a ",
                    class(alleles))
        }

        n_alleles <- length(alleles)
        expected_n <- nalleles(crosstype)
        if(n_alleles != expected_n) {
            result <- FALSE
            warning("length(alleles) (", n_alleles, ") != ", expected_n)
        }
        else if(length(unique(alleles)) != n_alleles) {
            result <- FALSE
            warning("alleles are not distinct")
        }
    }

    if(result) { # check genotypes only if everything else is okay
        n_invalid <- count_invalid_genotypes(cross2)
        if(sum(n_invalid)>0) {
            result <- FALSE
            warning(sum(n_invalid), " invalid genotypes in cross")
        }
    }

    result
}

handle_null_crossinfo <-
    function(crossinfo, ids)
{
    n_ind <- length(ids)
    if(is.null(crossinfo)) {
        crossinfo <- matrix(0L, nrow=n_ind, ncol=0)
        rownames(crossinfo) <- ids
    }
    crossinfo
}

handle_null_isfemale <-
    function(isfemale, ids)
{
    n_ind <- length(ids)
    if(is.null(isfemale)) {
        isfemale <- rep(FALSE, n_ind)
        names(isfemale) <- ids
    }
    isfemale
}

handle_null_isxchr <-
    function(is_x_chr, chrnames)
{
    if(is.null(is_x_chr)) {
        is_x_chr <- rep(FALSE, length(chrnames))
        names(is_x_chr) <- chrnames
    }
    is_x_chr
}

check_is_female_vector <-
    function(crosstype, is_female, any_x_chr)
{
    if(is.null(crosstype) || !is.character(crosstype))
        stop("crosstype should be a character string.")
    if(!is.logical(is_female))
        stop("is_female should be a logical vector")
    if(!is.logical(any_x_chr))
        stop("any_x_chr should be a TRUE/FALSE value")

    .check_is_female_vector(crosstype, is_female, any_x_chr)

}

check_crossinfo <-
    function(crosstype, cross_info, any_x_chr)
{
    if(is.null(crosstype) || !is.character(crosstype))
        stop("crosstype should be a character string.")
    if(!is.numeric(cross_info) || !is.matrix(cross_info))
        stop("cross_info should be a numeric matrix")
    if(!is.logical(any_x_chr))
        stop("any_x_chr should be a TRUE/FALSE value")

    .check_crossinfo(crosstype, cross_info, any_x_chr)

}

context("check cross")

test_that("check_crosstype works appropriately", {

    good_types <- c("bc", "f2", "riself", "risib", "dh", "haploid")
    for(type in good_types)
        expect_true(check_crosstype(type))

    expect_error(check_crosstype("f2pk"))
    expect_false(check_crosstype("f2pk", FALSE))

    expect_error(check_crosstype("bada1gakadsf"))
    expect_false(check_crosstype("bada1gakadsf", FALSE))

})

test_that("count_invalid_genotypes works appropriately", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    expect_true(check_cross2(grav2))
    count <- count_invalid_genotypes(grav2)
    expect_true(all(count==0))

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    expect_true(check_cross2(iron))
    count <- count_invalid_genotypes(iron)
    expect_true(all(count==0))

    # add some errors
    set.seed(79648025)
    g <- grav2$geno[[1]]
    err <- sample(0:5, prod(dim(g)), replace=TRUE, prob=c(0.95, rep(0.01, 5)))
    grav2$geno[[1]] <- g+err
    count <- count_invalid_genotypes(grav2)
    expect_equivalent(count[,1] , rowSums(g+err > 2))
    expect_true(all(count[,-1] == 0))

    # add some errors to the other data
    g <- iron$geno[[1]]
    err <- sample(0:5, prod(dim(g)), replace=TRUE, prob=c(0.95, rep(0.01, 5)))
    iron$geno[[1]] <- g+err
    count <- count_invalid_genotypes(iron)
    expect_equivalent(count[,1] , rowSums(g+err > 5))
    expect_true(all(count[,-1] == 0))

})

test_that("check_cross2 gives proper warnings", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    # missing geno
    iron_bad <- iron; iron_bad$geno <- NULL
    expect_warning(check_cross2(iron_bad))

    # missing gmap
    iron_bad <- iron; iron_bad$gmap <- NULL
    expect_warning(check_cross2(iron_bad))

    # missing cross_info
    iron_bad <- iron; iron_bad$cross_info <- NULL
    expect_warning(check_cross2(iron_bad))

    # missing is_female
    iron_bad <- iron; iron_bad$is_female <- NULL
    expect_warning(check_cross2(iron_bad))

    # missing is_x_chr
    iron_bad <- iron; iron_bad$is_x_chr <- NULL
    expect_warning(check_cross2(iron_bad))

    # missing alleles okay
    iron_bad <- iron; iron_bad$alleles <- NULL
    expect_true(check_cross2(iron_bad))

    # missing pmap okay
    iron_bad <- iron; iron_bad$pmap <- NULL
    expect_true(check_cross2(iron_bad))

    # gmap has no names()
    iron_bad <- iron; names(iron_bad$gmap) <- NULL
    expect_warning(check_cross2(iron_bad))

    # pmap has no names()
    iron_bad <- iron; names(iron_bad$pmap) <- NULL
    expect_warning(check_cross2(iron_bad))

    # gmap names != geno names
    iron_bad <- iron; names(iron_bad$gmap) <- 1:20
    expect_warning(check_cross2(iron_bad))

    # pmap names != geno names
    iron_bad <- iron; names(iron_bad$pmap) <- 1:20
    expect_warning(check_cross2(iron_bad))

    # founder genotypes included but not needed
    iron_bad <- iron; iron_bad$founder_geno <- iron_bad$geno
    expect_warning(check_cross2(iron_bad))

})


test_that("check_cross2 works for MPP data", {

    skip_if(isnt_karl(), "this test only run locally")

    do <- read_cross2("https://raw.githubusercontent.com/rqtl/qtl2data/main/DO_Recla/recla.zip")
    expect_true(check_cross2(do))

    # no founder genotypes
    do_bad <- do
    do_bad$founder_geno <- NULL
    expect_warning(check_cross2(do_bad))

    # no names in founder genotypes
    do_bad <- do
    names(do_bad$founder_geno) <- NULL
    expect_warning(check_cross2(do_bad))

    # mismatch in names of founder genotypes
    do_bad <- do
    names(do_bad$founder_geno) <- 1:20
    expect_warning(check_cross2(do_bad))

    # missing a founder in founder_geno
    do_bad <- do
    do_bad$founder_geno[[10]] <- do$founder_geno[[10]][-4,]
    expect_warning(check_cross2(do_bad))

    # missing a marker in founder_geno
    do_bad <- do
    do_bad$founder_geno[[12]] <- do$founder_geno[[12]][,-50]
    expect_warning(check_cross2(do_bad))

    # markers in founder_geno out of order
    do_bad <- do
    colnames(do_bad$founder_geno[[13]])[50:52] <- colnames(do$founder_geno[[13]])[52:50]
    expect_warning(check_cross2(do_bad))

    ### HS need founder geno
    hs <- do
    hs$crosstype <- "hs"
    expect_true(check_cross2(hs))

    hs_bad <- hs
    hs_bad$founder_geno <- NULL
    expect_warning(check_cross2(hs_bad))

    ### AIL3
    ail3 <- do
    ail3$crosstype <- "ail3"
    ail3$alleles <- ail3$alleles[1:3]
    for(i in seq_along(ail3$founder_geno))
        ail3$founder_geno[[i]] <- ail3$founder_geno[[i]][1:3,,drop=FALSE]
    expect_true(check_cross2(ail3))

    ail3_bad <- ail3
    ail3_bad$founder_geno <- NULL
    expect_warning(check_cross2(ail3_bad))

    ### riself4
    riself4 <- do
    riself4$crosstype <- "riself4"
    riself4$alleles <- riself4$alleles[1:4]
    for(x in c("geno", "founder_geno", "is_x_chr", "gmap", "pmap"))
        riself4[[x]] <- riself4[[x]][1:19]
    riself4$cross_info <- cbind(riself4$cross_info, 1L, 2L, 3L, 4L)[,-1,drop=FALSE]
    for(i in seq_along(riself4$founder_geno)) {
        riself4$founder_geno[[i]] <- riself4$founder_geno[[i]][1:4,,drop=FALSE]
        riself4$geno[[i]][riself4$geno[[i]] == 2] <- 0
    }
    expect_true(check_cross2(riself4))

    riself4_bad <- riself4
    riself4_bad$founder_geno <- NULL
    expect_warning(check_cross2(riself4_bad))

    ### riself8
    riself8 <- do
    riself8$crosstype <- "riself8"
    for(x in c("geno", "founder_geno", "is_x_chr", "gmap", "pmap"))
        riself8[[x]] <- riself8[[x]][1:19]
    riself8$cross_info <- cbind(riself8$cross_info, 1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L)[,-1,drop=FALSE]
    for(i in seq_along(riself8$founder_geno)) {
        riself8$geno[[i]][riself8$geno[[i]] == 2] <- 0
    }
    expect_true(check_cross2(riself8))

    riself8_bad <- riself8
    riself8_bad$founder_geno <- NULL
    expect_warning(check_cross2(riself8_bad))

    ### riself16
    riself16 <- do
    riself16$crosstype <- "riself16"
    riself16$alleles <- LETTERS[1:16]
    for(x in c("geno", "founder_geno", "is_x_chr", "gmap", "pmap"))
        riself16[[x]] <- riself16[[x]][1:19]
    riself16$cross_info <- cbind(riself16$cross_info, 1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L,
                                 9L, 10L, 11L, 12L, 13L, 14L, 15L, 16L)[,-1,drop=FALSE]
    for(i in seq_along(riself16$founder_geno)) {
        riself16$founder_geno[[i]] <- rbind(riself16$founder_geno[[i]], riself16$founder_geno[[i]])
        riself16$geno[[i]][riself16$geno[[i]] == 2] <- 0
    }
    expect_true(check_cross2(riself16))

    riself16_bad <- riself16
    riself16_bad$founder_geno <- NULL
    expect_warning(check_cross2(riself16_bad))

    ### risib4
    risib4 <- do
    risib4$crosstype <- "risib4"
    risib4$alleles <- risib4$alleles[1:4]
    risib4$cross_info <- cbind(risib4$cross_info, 1L, 2L, 3L, 4L)[,-1,drop=FALSE]
    for(i in seq_along(risib4$founder_geno)) {
        risib4$founder_geno[[i]] <- risib4$founder_geno[[i]][1:4,,drop=FALSE]
        risib4$geno[[i]][risib4$geno[[i]] == 2] <- 0
    }
    expect_true(check_cross2(risib4))

    risib4_bad <- risib4
    risib4_bad$founder_geno <- NULL
    expect_warning(check_cross2(risib4_bad))

    ### risib8
    risib8 <- do
    risib8$crosstype <- "risib8"
    risib8$cross_info <- cbind(risib8$cross_info, 1L, 2L, 3L, 4L, 5L, 6L, 7L, 8L)[,-1,drop=FALSE]
    for(i in seq_along(risib8$founder_geno)) {
        risib8$geno[[i]][risib8$geno[[i]] == 2] <- 0
    }
    expect_true(check_cross2(risib8))

    risib8_bad <- risib8
    risib8_bad$founder_geno <- NULL
    expect_warning(check_cross2(risib8_bad))

    ### dh6
    dh6 <- do
    dh6$crosstype <- "dh6"
    dh6$alleles <- dh6$alleles[1:6]
    for(x in c("geno", "founder_geno", "is_x_chr", "gmap", "pmap"))
        dh6[[x]] <- dh6[[x]][1:19]
    for(i in seq_along(dh6$founder_geno)) {
        dh6$founder_geno[[i]] <- dh6$founder_geno[[i]][1:6,]
        dh6$geno[[i]][dh6$geno[[i]] == 2] <- 0
    }
    expect_true(check_cross2(dh6))

    dh6_bad <- dh6
    dh6_bad$founder_geno <- NULL
    expect_warning(check_cross2(dh6_bad))

    ### dof1
    dof1 <- do
    dof1$crosstype <- "dof1"
    for(i in seq_along(dof1$founder_geno)) {
        dof1$founder_geno[[i]] <- dof1$founder_geno[[i]][c(1:8, 5),,drop=FALSE]
    }
    expect_true(check_cross2(dof1))

    dof1_bad <- dof1
    dof1_bad$founder_geno <- NULL
    expect_warning(check_cross2(dof1_bad))

})

context("check cross_info, sex, and handling of X")

test_that("checks of cross_info, sex, and X are correct", {

    null_crossinfo <- matrix(0L, 0, 0)
    null_isfemale <- logical(0)

    # dh, haploid, riself don't want X chr
    for(crosstype in c("dh", "haploid", "riself")) {
        expect_true(check_handle_x_chr(crosstype, FALSE))
        expect_false(check_handle_x_chr(crosstype, TRUE))
    }

    # others handle X chr ok
    for(crosstype in c("bc", "f2", "f2pk", "risib", "ail")) {
        expect_true(check_handle_x_chr(crosstype, FALSE))
        expect_true(check_handle_x_chr(crosstype, TRUE))
    }

    # dh, haploid, riself, risib ignore is_female
    for(crosstype in c("dh", "haploid", "riself", "risib")) {
        expect_true(check_is_female_vector(crosstype, null_isfemale, FALSE))
        expect_true(check_is_female_vector(crosstype, null_isfemale, TRUE))
        expect_true(check_is_female_vector(crosstype, c(TRUE, FALSE), FALSE))
        expect_true(check_is_female_vector(crosstype, c(TRUE, FALSE), TRUE))
        expect_true(check_is_female_vector(crosstype, c(TRUE, NA, FALSE), FALSE))
        expect_true(check_is_female_vector(crosstype, c(TRUE, NA, FALSE), TRUE))
    }

    # backcross, intercross, AIL, and DO need is_female
    for(crosstype in c("bc", "f2", "f2pk", "ail", "do")) {
        expect_true(check_is_female_vector(crosstype, null_isfemale, FALSE))
        expect_false(check_is_female_vector(crosstype, null_isfemale, TRUE))
        expect_true(check_is_female_vector(crosstype, c(TRUE, FALSE), TRUE))
        expect_true(check_is_female_vector(crosstype, c(TRUE, FALSE), FALSE))
        expect_false(check_is_female_vector(crosstype, c(TRUE, NA, FALSE), TRUE))
        expect_true(check_is_female_vector(crosstype, c(TRUE, NA, FALSE), FALSE))
    }

    # if is_female is numeric, should give an error
    expect_error( check_is_female_vector("do", c(1,0,1,1,0,0,0), TRUE) )
    expect_error( check_is_female_vector("do", c(1,0,1,1,0,0,0), 5) )

    # dh, haploid, riself, bc ignore cross_info
    for(crosstype in c("dh", "haploid", "riself", "bc")) {
        expect_true(check_crossinfo(crosstype, null_crossinfo, FALSE))
        expect_true(check_crossinfo(crosstype, null_crossinfo, TRUE))
        expect_true(check_crossinfo(crosstype, cbind(c(0,1)), TRUE))
    }

    # intercross, phase-known intercross, and risib need cross_info for X
    for(crosstype in c("f2", "f2pk", "risib")) {
        expect_true(check_crossinfo(crosstype, null_crossinfo, FALSE))
        expect_false(check_crossinfo(crosstype, null_crossinfo, TRUE))
        expect_true(check_crossinfo(crosstype, cbind(c(0,1)), TRUE))
        expect_true(check_crossinfo(crosstype, cbind(c(0,1)), FALSE))
        expect_false(check_crossinfo(crosstype, cbind(c(0,NA,1)), TRUE))
        expect_true(check_crossinfo(crosstype, cbind(c(0,NA,1)), FALSE))
        expect_false(check_crossinfo(crosstype, cbind(c(0,2,1)), TRUE))
        expect_true(check_crossinfo(crosstype, cbind(c(0,2,1)), FALSE))
        expect_false(check_crossinfo(crosstype, cbind(0,0,1), TRUE))
        expect_true(check_crossinfo(crosstype, cbind(0,1,1), FALSE))
    }

    # AIL needs no. generations + (for X chromosome) direction (0=AxB, 1=BxA, 2=balanced)
    expect_false(check_crossinfo("ail", null_crossinfo, FALSE))
    expect_false(check_crossinfo("ail", null_crossinfo, TRUE))
    expect_true(check_crossinfo("ail", cbind(c(2,3,25,50)), FALSE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50)), TRUE))
    expect_true(check_crossinfo("ail", cbind(c(2,3,25,50),c(0,0,1,2)), FALSE))
    expect_true(check_crossinfo("ail", cbind(c(2,3,25,50),c(0,0,1,2)), TRUE))

    # AIL: no. generations >= 2 and not missing
    expect_false(check_crossinfo("ail", cbind(c(1,3,25,50),c(0,0,1,2)), FALSE))
    expect_false(check_crossinfo("ail", cbind(c(2,NA,25,50),c(0,0,1,2)), FALSE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,1,50),c(0,0,1,2)), FALSE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,NA),c(0,0,1,2)), FALSE))
    expect_false(check_crossinfo("ail", cbind(c(1,3,25,50),c(0,0,1,2)), TRUE))
    expect_false(check_crossinfo("ail", cbind(c(2,NA,25,50),c(0,0,1,2)), TRUE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,1,50),c(0,0,1,2)), TRUE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,NA),c(0,0,1,2)), TRUE))

    # AIL: dir = 0,1,2 and not missing
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50),c(0,-1,1,2)), FALSE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50),c(0,0,1,-1)), FALSE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50),c(NA,0,1,2)), FALSE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50),c(0,0,NA,2)), FALSE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50),c(0,0,1,NA)), FALSE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50),c(0,-1,1,2)), TRUE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50),c(0,0,1,-1)), TRUE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50),c(NA,0,1,2)), TRUE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50),c(0,0,NA,2)), TRUE))
    expect_false(check_crossinfo("ail", cbind(c(2,3,25,50),c(0,0,1,NA)), TRUE))

    # DO needs no. generations
    expect_false(check_crossinfo("do", null_crossinfo, FALSE))
    expect_false(check_crossinfo("do", null_crossinfo, TRUE))
    expect_true(check_crossinfo("do", cbind(c(2,3,1,50)), FALSE))
    expect_true(check_crossinfo("do", cbind(c(2,3,1,50)), TRUE))

    # DO: no. generations >= 1 and not missing
    expect_false(check_crossinfo("do", cbind(c(0,3,25,50)), FALSE))
    expect_false(check_crossinfo("do", cbind(c(2,NA,25,50)), FALSE))
    expect_false(check_crossinfo("do", cbind(c(2,3,0,50)), FALSE))
    expect_false(check_crossinfo("do", cbind(c(2,3,25,NA)), FALSE))
    expect_false(check_crossinfo("do", cbind(c(0,3,25,50)), TRUE))
    expect_false(check_crossinfo("do", cbind(c(2,NA,25,50)), TRUE))
    expect_false(check_crossinfo("do", cbind(c(2,3,0,50)), TRUE))
    expect_false(check_crossinfo("do", cbind(c(2,3,25,NA)), TRUE))

    # if cross_info is not a numeric matrix, should give an error
    expect_error( check_crossinfo("do", c(0, 1, 2, 1), TRUE) )
    expect_error( check_crossinfo("do", matrix(letters[1:4], 2, 2), TRUE) )
    expect_error( check_crossinfo("do", matrix(letters[1:4], 2, 2), 5) )


})

# map functions
mf <-
function(d, map.function=c("haldane", "kosambi", "c-f", "morgan"))
{
    map.function <- match.arg(map.function)
    switch(map.function,
           haldane=mf.h(d),
           kosambi=mf.k(d),
           "c-f"=  mf.cf(d),
           morgan= mf.m(d))
}

imf <-
function(r, map.function=c("haldane", "kosambi", "c-f", "morgan"))
{
    map.function <- match.arg(map.function)
    switch(map.function,
           haldane=imf.h(r),
           kosambi=imf.k(r),
           "c-f"=  imf.cf(r),
           morgan= imf.m(r))
}


mf.h <-
function(d)
    0.5*(1-exp(-d/50))
imf.h <-
function(r)
{
    r[r >= 0.5] <- 0.5-1e-14
    -50*log(1-2*r)
}

mf.k <-
function(d)
    0.5*tanh(d/50)

imf.k <-
function(r)
{
    r[r >= 0.5] <- 0.5-1e-14
    50*atanh(2*r)
}

mf.m <-
function(d)
    pmin(d/100, 0.5)

imf.m <-
function(r)
    pmin(r*100, 50)

# carter-falconer
imf.cf <- function(r) {
    r[r >= 0.5] <- 0.5-1e-14
    12.5*(log(1+2*r)-log(1-2*r))+25*atan(2*r)
}

mf.cf <-
function(d)
{
    d[d >= 300] <- 300

    icf <- function(r,d)
        imf.cf(r)-d

    sapply(d,function(a) {
        if(a==0) return(0)
        stats::uniroot(icf, c(0,0.5-1e-14),d=a,tol=1e-12)$root })
}

context("map functions")

test_that("map functions are giving the invertible", {

    mapfunc <- c("haldane", "kosambi", "c-f", "morgan")

    d <- seq(0, 200, by=10)
    for(m in mapfunc) {
        if(m == "morgan") d <- d[d <= 50]

        expect_equal(imf(mf(d, m), m), d)
    }

    r <- seq(0, 0.5, by=0.5)
    for(m in mapfunc)
        expect_equal(imf(mf(r, m), m), r)

})

#' Plot LOD scores vs QTL peak locations
#'
#' Create a scatterplot of LOD scores vs QTL peak locations (possibly with intervals) for multiple traits.
#'
#' @param peaks Data frame such as that produced by
#'     [find_peaks()]) containing columns
#'     `chr`, `pos`, `lodindex`, and `lodcolumn`.
#'     May also contain columns `ci_lo` and `ci_hi`, in
#'     which case intervals will be plotted.
#' @param map Marker map, used to get chromosome lengths (and start
#'     and end positions).
#' @param chr Selected chromosomes to plot; a vector of character
#'     strings.
#' @param gap Gap between chromosomes. The default is 1% of the total genome length.
#' @param intervals If TRUE and `peaks` contains QTL intervals, plot the intervals.
#' @param ... Additional graphics parameters
#'
#' @seealso [find_peaks()], [plot_peaks()], [plot_cistrans()], [plot_scan1()], [plot_scan1_heatmap()]
#'
#' @section Hidden graphics parameters:
#' A number of graphics parameters can be passed via `...`. For
#' example, `bgcolor` to control the background color and
#' `altbgcolor` to control the background color on alternate chromosomes.
#' These are not included as formal parameters in order to avoid
#' cluttering the function definition.
#'
#' @export
#' @importFrom graphics segments abline par axis title box rect
#'
#' @return None.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c(1,2,7,8,9,13,15,16,19)]}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' # find peaks above lod=3.5 (and calculate 1.5-LOD support intervals)
#' peaks <- find_peaks(out, map, threshold=3.5, drop=1.5)
#'
#' plot_lodpeaks(peaks, map)

plot_lodpeaks <-
    function(peaks, map, chr=NULL, gap=NULL, intervals=FALSE, ...)
{
    if(is.null(peaks)) stop("peaks is NULL")
    if(is.null(map)) stop("map is NULL")

    if(!is.list(map)) map <- list(" "=map) # if a vector, treat it as a list with no names

    if(!is.null(chr)) {
        chri <- match(chr, names(map))
        if(any(is.na(chri)))
            stop("Chromosomes ", paste(chr[is.na(chri)], collapse=", "), " not found")
        map <- map[chri]
        peaks <- peaks[peaks$chr %in% names(map),,drop=FALSE]
    }

    if(nrow(peaks) == 0)
        stop("There are no peaks on the selected chromosomes")

    plot_lodpeaks_internal <-
        function(peaks, map, gap, intervals=FALSE,
                 bgcolor="gray90", altbgcolor="gray85",
                 lwd=2, pch=21, col="black", bg="slateblue",
                 xlab=NULL, ylab="LOD score",
                 xlim=NULL, ylim=NULL, xaxs="i", yaxs="i",
                 main="", mgp.x=c(2.6, 0.5, 0), mgp.y=c(2.6, 0.5, 0),
                 mgp=NULL, las=1,
                 hlines=NULL, hlines_col="white", hlines_lwd=1, hlines_lty=1,
                 vlines=NULL, vlines_col="white", vlines_lwd=1, vlines_lty=1,
                 lend=1, ljoin=1, sub="", ...)
        {
            dots <- list(...)
            onechr <- (length(map)==1) # single chromosome

            if(is.null(ylim)) {
                ylim <- c(0, max(peaks$lod, na.rm=TRUE)*1.05)
            }

            # plot empty frame
            emptylod <- as.matrix(rep(NA, sum(sapply(map, length))))
            rownames(emptylod) <- unlist(lapply(map, names))
            plot_scan1(emptylod, map, gap=gap, type="n",
                       xlab=xlab, ylab=ylab, xlim=xlim, ylim=ylim, xaxs=xaxs, yaxs=yaxs,
                       bgcolor=bgcolor, altbgcolor=altbgcolor,
                       main=main, sub=sub, las=las,
                       hlines=hlines, hlines_col=hlines_col, hlines_lwd=hlines_lwd, hlines_lty=hlines_lty,
                       vlines=vlines, vlines_col=vlines_col, vlines_lwd=vlines_lwd, vlines_lty=vlines_lty)

            # add QTL intervals
            if(intervals && all(c("ci_lo", "ci_hi") %in% names(peaks))) {
                xpos_lo <- xpos_scan1(map, gap=gap, thechr=peaks$chr, thepos=peaks$ci_lo)
                xpos_hi <- xpos_scan1(map, gap=gap, thechr=peaks$chr, thepos=peaks$ci_hi)

                segments(xpos_lo, peaks$lod, xpos_hi, peaks$lod, lend=lend, ljoin=ljoin)
            }

            # add points at LOD scores
            xpos <- xpos_scan1(map, gap=gap, thechr=peaks$chr, thepos=peaks$pos)
            points(xpos, peaks$lod, pch=pch, col=col, bg=bg, ...)
        }

    plot_lodpeaks_internal(peaks=peaks, map=map, gap=gap, intervals=intervals, ...)

}

#' Reduce markers to a subset of more-evenly-spaced ones
#'
#' Find the largest subset of markers such that no two adjacent
#' markers are separated by less than some distance.
#'
#' @param map A list with each component being a vector with the
#' marker positions for a chromosome.
#' @param min_distance Minimum distance between markers.
#' @param weights A (optional) list of weights on the markers; same
#' size as `map`.
#' @param max_batch Maximum number of markers to consider in a batch
#' @param batch_distance_mult If working with batches of markers,
#' reduce `min_distance` by this multiple.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A list like the input `map`, but with the selected
#' subset of markers.
#'
#' @details Uses a dynamic programming algorithm to find, for each
#' chromosome, the subset of markers for with max(`weights`) is
#' maximal, subject to the constraint that no two adjacent markers may
#' be separated by more than `min_distance`.
#'
#' The computation time for the algorithm grows with like the square
#' of the number of markers, like 1 sec for 10k markers
#' but 30 sec for 50k markers. If the number of markers on a chromosome
#' is greater than `max_batch`, the markers are split into batches and
#' the algorithm applied to each batch with min_distance smaller by a
#' factor `min_distance_mult`, and then merged together for one last pass.
#'
#' @seealso [find_dup_markers()], [drop_markers()]
#'
#' @references Broman KW, Weber JL (1999) Method for constructing
#' confidently ordered linkage maps. Genet Epidemiol 16:337--343
#'
#' @examples
#' # read data
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#'
#' # grab genetic map
#' gmap <- grav2$gmap
#'
#' # subset to markers that are >= 1 cM apart
#' gmap_sub <- reduce_markers(gmap, 1)
#'
#' # drop all of the other markers from the cross
#' markers2keep <- unlist(lapply(gmap_sub, names))
#' grav2_sub <- pull_markers(grav2, markers2keep)
#' @export
reduce_markers <-
    function(map, min_distance=1, weights=NULL,
             max_batch=10000, batch_distance_mult=1, cores=1)
{
    if(is.null(map)) stop("map is NULL")
    if(is.cross2(map))
        stop('Input map is a "cross2" object but should be a genetic map')

    if(!is.list(map) && !is.list(weights)) { # single chromosome
        if(!is.null(weights)) weights <- list("1"=weights)
        return(reduce_markers(list("1"=map), min_distance, weights)[[1]])
    }

    if(is.null(weights))
        weights <- lapply(map, function(a) rep(1, length(a)))

    stopifnot(length(map) == length(weights))
    nmar <- vapply(map, length, 1)
    nwts <- vapply(weights, length, 1)
    if(!all(nmar == nwts))
        stop("Different numbers of markers and weights on chr",
             paste(names(map)[nmar != nwts], collapse=" "))

    if(!is_pos_number(min_distance)) stop("min_distance should be a single positive number")

    for(i in seq(along=map)) {
        if(length(map[[i]]) < 2) next
        map[[i]] <- reduce_markers_onechr(map[[i]], min_distance, weights[[i]],
                                          max_batch, batch_distance_mult, cores)
    }

    map
}

reduce_markers_onechr <-
    function(marker_pos, min_distance=1, weights=NULL,
             max_batch=10000, batch_distance_mult=1, cores=1)
{
    num_mar <- length(marker_pos)
    if(is.null(weights)) {
        weights <- rep(1, num_mar)
    }
    stopifnot(num_mar == length(weights))

    cores <- setup_cluster(cores)

    # repeat the following until number hasn't decreased and
    while(num_mar > max_batch) {
        # split markers into groups of length
        batches <- batch_vec(seq_len(num_mar), max_batch, n_cores=1)

#        message(num_mar, " in ", length(batches), " batches")

        batch_func <- function(batch) {
            result <- .reduce_markers(marker_pos[batch], min_distance/batch_distance_mult, weights[batch])
            batch[result]
        }

        sub_map <- unlist( cluster_lapply(cores, batches, batch_func) )

        marker_pos <- marker_pos[sub_map]
        weights <- weights[sub_map]

        if(length(marker_pos) == num_mar) break # there's been no change in number of markers
        num_mar <- length(marker_pos)
    }

    marker_pos[.reduce_markers(marker_pos, min_distance, weights)]
}

# utilities for weights

# are the weights NULL or all 0's?
is_null_weights <-
    function(weights, tol=1e-12)
{
    if(is.null(weights) || max(abs(weights-1)) < tol)
        return(TRUE)

    FALSE
}


# check vector of weights & take square-roots
#
# - all positive?
#
# - if no missing values and all close to 1,
#   just use NULL rather than the weights

sqrt_weights <-
    function(weights, tol=1e-12)
{
    if(is.null(weights)) return(weights)

    if(any(weights <= 0))
        stop("weights must all be positive")

    if(all(!is.na(weights) & abs(weights - 1)<tol))
        return(NULL)

    weights <- stats::setNames( as.numeric(weights), names(weights) )

    return(sqrt(weights))
}


# multiply a vector by a set of weights
#' @importFrom stats setNames
weight_vector <-
    function(vec, weights, tol=1e-12)
{
    if(is_null_weights(weights, tol) || is.null(vec)) return(vec)

    # align and multiply
    id <- get_common_ids(setNames(names(vec), NULL), setNames(names(weights), NULL))
    vec[id] * weights[id]
}


# multiply a matrix by a set of weights
#' @importFrom stats setNames
weight_matrix <-
    function(mat, weights, tol=1e-12)
{
    if(is_null_weights(weights, tol) || is.null(mat)) return(mat)

    # force the input mat to be a matrix
    if(!is.matrix(mat)) mat <- as.matrix(mat)

    # align and multiply
    id <- get_common_ids(setNames(rownames(mat), NULL), setNames(names(weights), NULL))
    mat[id,,drop=FALSE] * weights[id]
}

# multiply an array by a set of weights
#' @importFrom stats setNames
weight_array <-
    function(arr, weights, tol=1e-12)
{
    if(is_null_weights(weights, tol) || is.null(arr)) return(arr)

    # force the input mat to be a matrix
    if(!is.array(arr) || length(dim(arr)) != 3)
        stop("arr should be a 3-dimensional array")

    # align and multiply
    id <- get_common_ids(setNames(rownames(arr), NULL), setNames(names(weights), NULL))
    arr[id,,,drop=FALSE] * weights[id]
}

#' Calculate QTL effects in scan along one chromosome
#'
#' Calculate QTL effects in scan along one chromosome with a
#' single-QTL model using Haley-Knott regression or a linear mixed
#' model (the latter to account for a residual polygenic effect), with
#' possible allowance for covariates.
#'
#' @param genoprobs Genotype probabilities as calculated by
#' [calc_genoprob()].
#' @param pheno A numeric vector of phenotype values (just one phenotype, not a matrix of them)
#' @param kinship Optional kinship matrix, or a list of kinship matrices (one
#' per chromosome), in order to use the LOCO (leave one chromosome
#' out) method.
#' @param addcovar An optional numeric matrix of additive covariates.
#' @param nullcovar An optional numeric matrix of additional additive
#' covariates that are used under the null hypothesis (of no QTL) but
#' not under the alternative (with a QTL). This is needed for the X
#' chromosome, where we might need sex as a additive covariate under
#' the null hypothesis, but we wouldn't want to include it under the
#' alternative as it would be collinear with the QTL effects. Only
#' used if `kinship` is provided but `hsq` is not, to get
#' estimate of residual heritability.
#' @param intcovar An optional numeric matrix of interactive covariates.
#' @param weights An optional numeric vector of positive weights for the
#' individuals. As with the other inputs, it must have `names`
#' for individual identifiers.
#' @param contrasts An optional numeric matrix of genotype contrasts, size
#' genotypes x genotypes. For an intercross, you might use
#' `cbind(mu=c(1,1,1), a=c(-1, 0, 1), d=c(0, 1, 0))` to get
#' mean, additive effect, and dominance effect. The default is the
#' identity matrix.
#' @param model Indicates whether to use a normal model (least
#'     squares) or binary model (logistic regression) for the phenotype.
#'     If `model="binary"`, the phenotypes must have values in \eqn{[0, 1]}.
#' @param zerosum If TRUE, force the genotype or allele coefficients
#'     sum to 0 by subtracting their mean and add another column with
#'     the mean. Ignored if `contrasts` is provided.
#'     **Note**: the standard errors aren't quite right when `zerosum=TRUE`;
#'     use [fit1()] to get proper standard errors.
#' @param se If TRUE, also calculate the standard errors.
#' @param hsq (Optional) residual heritability; used only if
#' `kinship` provided.
#' @param reml If `kinship` provided: if `reml=TRUE`, use
#' REML; otherwise maximum likelihood.
#' @param ... Additional control parameters; see Details;
#'
#' @return An object of class `"scan1coef"`: a matrix of estimated regression coefficients, of dimension
#'     positions x number of effects. The number of effects is
#'     `n_genotypes + n_addcovar + (n_genotypes-1)*n_intcovar`.
#' May also contain the following attributes:
#' * `SE` - Present if `se=TRUE`: a matrix of estimated
#'   standard errors, of same dimension as `coef`.
#' * `sample_size` - Vector of sample sizes used for each
#'   phenotype
#'
#' @details For each of the inputs, the row names are used as
#' individual identifiers, to align individuals.
#'
#' If `kinship` is absent, Haley-Knott regression is performed.
#' If `kinship` is provided, a linear mixed model is used, with a
#' polygenic effect estimated under the null hypothesis of no (major)
#' QTL, and then taken as fixed as known in the genome scan.
#'
#' If `contrasts` is provided, the genotype probability matrix,
#' \eqn{P}, is post-multiplied by the contrasts matrix, \eqn{A}, prior
#' to fitting the model. So we use \eqn{P \cdot A}{P A} as the \eqn{X}
#' matrix in the model. One might view the rows of
#' \ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}
#' as the set of contrasts, as the estimated effects are the estimated
#' genotype effects pre-multiplied by
#' \ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}.
#'
#' The `...` argument can contain several additional control
#' parameters; suspended for simplicity (or confusion, depending on
#' your point of view). `tol` is used as a tolerance value for linear
#' regression by QR decomposition (in determining whether columns are
#' linearly dependent on others and should be omitted); default
#' `1e-12`. `maxit` is the maximum number of iterations for
#' converence of the iterative algorithm used when `model=binary`.
#' `bintol` is used as a tolerance for converence for the iterative
#' algorithm used when `model=binary`. `eta_max` is the maximum value
#' for the "linear predictor" in the case `model="binary"` (a bit of a
#' technicality to avoid fitted values exactly at 0 or 1).
#'
#' @references Haley CS, Knott SA (1992) A simple
#' regression method for mapping quantitative trait loci in line
#' crosses using flanking markers.  Heredity 69:315--324.
#'
#' Kang HM, Zaitlen NA, Wade CM, Kirby A, Heckerman D, Daly MJ, Eskin
#' E (2008) Efficient control of population structure in model
#' organism association mapping. Genetics 178:1709--1723.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c(7,19)] # reduce to chr 7 and 19}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#' \dontshow{probs[["7"]] <- probs[["7"]][,,1:5] # reduce to very small number}
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno[,1]
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#'
#' # calculate coefficients for chromosome 7
#' coef <- scan1coef(probs[,"7"], pheno, addcovar=covar)
#'
#' # leave-one-chromosome-out kinship matrix for chr 7
#' kinship7 <- calc_kinship(probs, "loco")[["7"]]
#'
#' # calculate coefficients for chromosome 7, adjusting for residual polygenic effect
#' coef_pg <- scan1coef(probs[,"7"], pheno, kinship7, addcovar=covar)
#'
#' @export
scan1coef <-
    function(genoprobs, pheno, kinship=NULL, addcovar=NULL, nullcovar=NULL,
             intcovar=NULL, weights=NULL,
             contrasts=NULL, model=c("normal", "binary"), zerosum=TRUE,
             se=FALSE, hsq=NULL, reml=TRUE, ...)
{
    if(is.null(genoprobs)) stop("genoprobs is NULL")
    if(is.null(pheno)) stop("pheno is NULL")

    if(!is.null(kinship)) { # use LMM; see scan1_pg.R
        return(scan1coef_pg(genoprobs, pheno, kinship, addcovar, nullcovar,
                            intcovar, weights, contrasts, zerosum, se, hsq, reml, ...))
    }

    model <- match.arg(model)

    # deal with the dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    if(!is_pos_number(tol)) stop("tol should be a single positive number")
    if(model=="binary") {
        bintol <- grab_dots(dotargs, "bintol", sqrt(tol)) # for model="binary"
        if(!is_pos_number(bintol)) stop("bintol should be a single positive number")
        eta_max <- grab_dots(dotargs, "eta_max", log(1-tol)-log(tol)) # for model="binary"
        if(!is_pos_number(eta_max)) stop("eta_max should be a single positive number")
        maxit <- grab_dots(dotargs, "maxit", 100) # for model="binary"
        if(!is_nonneg_number(maxit)) stop("maxit should be a single non-negative integer")

        check_extra_dots(dotargs, c("tol", "bintol", "eta_max", "maxit"))
    }
    else { # not binary trait
        check_extra_dots(dotargs, "tol")
    }

    # check that the objects have rownames
    check4names(pheno, addcovar, NULL, intcovar, nullcovar)

    # force things to be matrices
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(nullcovar)) {
        if(!is.matrix(nullcovar)) nullcovar <- as.matrix(nullcovar)
        if(!is.numeric(nullcovar)) stop("nullcovar is not numeric")
    }
    if(!is.null(intcovar)) {
        if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)
        if(!is.numeric(intcovar)) stop("intcovar is not numeric")
    }
    if(!is.null(contrasts)) {
        if(!is.matrix(contrasts)) contrasts <- as.matrix(contrasts)
        if(!is.numeric(contrasts)) stop("contrasts is not numeric")
    }

    # make sure pheno is a vector
    if(is.matrix(pheno) || is.data.frame(pheno)) {
        if(ncol(pheno) > 1)
            warning("Considering only the first phenotype.")
        rn <- rownames(pheno)
        pheno <- pheno[,1]
        names(pheno) <- rn
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }

    # genoprobs has more than one chromosome?
    if(length(genoprobs) > 1)
        warning("Using only the first chromosome, ", names(genoprobs)[1])
    chrid <- names(genoprobs)[1]
    genoprobs <- genoprobs[[1]]

    # make sure contrasts is square n_genotypes x n_genotypes
    if(!is.null(contrasts)) {
        ng <- ncol(genoprobs)
        if(ncol(contrasts) != ng || nrow(contrasts) != ng)
            stop("contrasts should be a square matrix, ", ng, " x ", ng)
    }

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, pheno, addcovar, intcovar,
                               weights, complete.cases=TRUE)
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    # omit individuals not in common
    genoprobs <- genoprobs[ind2keep,,,drop=FALSE]
    pheno <- pheno[ind2keep]
    if(!is.null(addcovar)) addcovar <- addcovar[ind2keep,,drop=FALSE]
    if(!is.null(intcovar)) intcovar <- intcovar[ind2keep,,drop=FALSE]
    if(!is.null(weights)) weights <- weights[ind2keep]

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # make sure columns in intcovar are also in addcovar
    addcovar <- force_intcovar(addcovar, intcovar, tol)

    # normal or binary model?
    if(model=="binary") {
        if(!is.null(kinship))
            stop("Can't yet account for kinship with model = \"binary\"")
        if(any(!is.na(pheno) & (pheno < 0 | pheno > 1)))
            stop('with model="binary", pheno should be in [0,1]')
    }
    else {
        # square-root of weights (only if model="normal")
        weights <- sqrt_weights(weights) # also check >0 (and if all 1's, turn to NULL)

        # if weights, adjust phenotypes
        if(!is.null(weights)) pheno <- weights * pheno
    }

    # weights have 0 dimension if missing
    if(is.null(weights)) weights <- numeric(0)

    # multiply genoprobs by contrasts
    if(!is.null(contrasts))
        genoprobs <- genoprobs_by_contrasts(genoprobs, contrasts)

    if(se) { # also calculate SEs

        if(is.null(intcovar)) { # just addcovar
            if(is.null(addcovar)) addcovar <- matrix(nrow=length(ind2keep), ncol=0)
            if(model=="normal")
                result <- scancoefSE_hk_addcovar(genoprobs, pheno, addcovar, weights, tol)
            else # binary trait
                result <- scancoefSE_binary_addcovar(genoprobs, pheno, addcovar, weights, maxit, bintol, tol, eta_max)
        }
        else {                  # intcovar
            if(model=="normal")
                result <- scancoefSE_hk_intcovar(genoprobs, pheno, addcovar, intcovar,
                                                 weights, tol)
            else
                result <- scancoefSE_binary_intcovar(genoprobs, pheno, addcovar, intcovar,
                                                 weights, maxit, bintol, tol, eta_max)
        }

        SE <- t(result$SE) # transpose to positions x coefficients
        result <- result$coef
    } else { # don't calculate SEs

        if(is.null(intcovar)) { # just addcovar
            if(is.null(addcovar)) addcovar <- matrix(nrow=length(ind2keep), ncol=0)
            if(model=="normal")
                result <- scancoef_hk_addcovar(genoprobs, pheno, addcovar, weights, tol)
            else
                result <- scancoef_binary_addcovar(genoprobs, pheno, addcovar, weights, maxit, bintol, tol, eta_max)
        }
        else {                  # intcovar
            if(model=="normal")
                result <- scancoef_hk_intcovar(genoprobs, pheno, addcovar, intcovar,
                                               weights, tol)
            else
                result <- scancoef_binary_intcovar(genoprobs, pheno, addcovar, intcovar,
                                                   weights, maxit, bintol, tol, eta_max)
        }
        SE <- NULL
    }

    result <- t(result) # transpose to positions x coefficients

    # add names
    dimnames(result) <- list(dimnames(genoprobs)[[3]],
                             scan1coef_names(genoprobs, addcovar, intcovar))
    if(se) dimnames(SE) <- dimnames(result)

    # center the QTL effects at zero and add an intercept
    if(zerosum && is.null(contrasts)) {
        ng <- dim(genoprobs)[2]
        whcol <- seq_len(ng)
        mu <- rowMeans(result[,whcol,drop=FALSE], na.rm=TRUE)
        result <- cbind(result, intercept=mu)
        result[,whcol] <- result[,whcol] - mu

        if(se) {
            SE <- cbind(SE, intercept=sqrt(rowMeans(SE[,whcol,drop=FALSE]^2, na.rm=TRUE)))
        }
    }

    attr(result, "sample_size") <- length(ind2keep)
    attr(result, "SE") <- SE # include only if not NULL

    class(result) <- c("scan1coef", "scan1", "matrix")
    result
}


# genoprob x contrasts
genoprobs_by_contrasts <-
    function(genoprobs, contrasts)
{
    dg <- dim(genoprobs)
    dc <- dim(contrasts)
    if(dc[1] != dc[2] || dc[1] != dg[2])
        stop("contrasts should be a square matrix, ", dg[2], " x ", dg[2])

    # rearrange to put genotypes in last position
    dn <- dimnames(genoprobs)
    genoprobs <- aperm(genoprobs, c(1,3,2))
    dim(genoprobs) <- c(dg[1]*dg[3], dg[2])

    # multiply by contrasts
    genoprobs <- genoprobs %*% contrasts
    dim(genoprobs) <- dg[c(1,3,2)]

    genoprobs <- aperm(genoprobs, c(1,3,2))
    if(!is.null(colnames(contrasts))) dn[[2]] <- colnames(contrasts)
    dimnames(genoprobs) <- dn

    genoprobs
}

# coefficient names
scan1coef_names <-
    function(genoprobs, addcovar, intcovar)
{
    qtl_names <-colnames(genoprobs)
    if(is.null(qtl_names))
        qtl_names <- paste0("qtleff", seq_len(ncol(genoprobs)))

    if(is.null(addcovar) || ncol(addcovar)==0) { # no additive covariates
        return(qtl_names)
    }
    else { # some additive covariates
        add_names <- colnames(addcovar)
        if(is.null(add_names) || all(add_names==""))
            add_names <- paste0("ac", seq_len(ncol(addcovar)))
        else if(all(add_names[-1] == "")) # all but first is empty
            add_names[-1] <- paste0("ac", seq_len(ncol(addcovar)-1))

        if(is.null(intcovar)) { # no interactive covariates
            return(c(qtl_names, add_names))
        }

        int_names <- colnames(intcovar)
        if(is.null(int_names) || all(int_names==""))
            int_names <- paste0("ic", seq_len(ncol(intcovar)))
        else if(all(int_names[-1] == "")) # all but first is empty
            int_names[-1] <- paste0("ic", seq_len(ncol(intcovar)-1))

        result <- c(qtl_names, add_names)
        for(ic in int_names)
            result <- c(result, paste(qtl_names[-1], ic, sep=":"))

        return(result)
    }
}

#' Join genome scan results for different phenotypes.
#'
#' Join multiple [scan1()] results for different phenotypes;
#' must have the same map.
#'
#' @param ... Genome scan objects of class `"scan1"`, as produced by [scan1()].
#' Must have the same map.
#'
#' @return An object of class `"scan1", like the inputs, but with the lod score columns
#' from the inputs combined as multiple columns in a single object.
#'
#' @details If components `addcovar()`, `Xcovar`,
#' `intcovar`, `weights` do not match between objects, we
#' omit this information.
#'
#' If `hsq` present but has differing numbers of rows, we omit this information.
#'
#' @seealso [rbind.scan1()], [scan1()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' probs <- calc_genoprob(grav2, map, error_prob=0.002)
#' phe1 <- grav2$pheno[,1,drop=FALSE]
#' phe2 <- grav2$pheno[,2,drop=FALSE]
#'
#' out1 <- scan1(probs, phe1) # phenotype 1
#' out2 <- scan1(probs, phe2) # phenotype 2
#' out <- cbind(out1, out2)
#'
#' @export
cbind.scan1 <-
    function(...)
{
    args <- list(...)
    if(length(args) == 1) return(result)

    # to cbind: main data, attributes SE, hsq
    # to c(): attribute sample_size

    # grab attributes
    args_attr <- lapply(args, attributes)

    result <- unclass(args[[1]])

    # cbind the data
    for(i in 2:length(args)) {
        if(!is_same(rownames(result), rownames(args[[i]])))
            stop("Input objects 1 and ", i, " have different rownames.")
        result <- cbind(result, unclass(args[[i]]))
    }

    # combine sample_size
    for(obj in c("sample_size")) {
        for(i in 2:length(args)) {
            drop_arg <- FALSE
            if(is.null(args_attr[[1]][[obj]]) && is.null(args_attr[[i]][[obj]])) next
            if(is.null(args_attr[[1]][[obj]]) || is.null(args_attr[[i]][[obj]])) {
                drop_arg <- TRUE
            } else {
                args_attr[[1]][[obj]] <- c(args_attr[[1]][[obj]], args_attr[[i]][[obj]])
            }
        }
        if(drop_arg) {
            warning('attribute "', obj, '" not present in all objects')
            args_attr[[1]][[obj]] <- NULL
        }

    }

    # attributes to combine by cbind()
    drop_arg <- c(hsq=FALSE, SE=FALSE)
    diff_rn_arg <- c(hsq=FALSE, SE=FALSE)
    for(obj in c("hsq", "SE")) {
        drop_arg <- FALSE
        drop_rn_arg <- FALSE
        for(i in 2:length(args)) {
            if(is.null(args_attr[[1]][[obj]]) && is.null(args_attr[[i]][[obj]])) next
            if(is.null(args_attr[[1]][[obj]]) || is.null(args_attr[[i]][[obj]])) {
                drop_arg <- TRUE
            } else if(!is_same(rownames(args_attr[[1]][[obj]]), rownames(args_attr[[i]][[obj]]))) {
                diff_rn_arg <- TRUE
            } else {
                args_attr[[1]][[obj]] <- cbind(args_attr[[1]][[obj]], args_attr[[i]][[obj]])
            }
        }
        if(drop_arg) {
            warning('attribute "', obj, '" not present in all objects')
            args_attr[[1]][[obj]] <- NULL
        } else if(drop_rn_arg) {
            warning('attribute "', obj, '" does not have the same row names in all objects')
            args_attr[[1]][[obj]] <- NULL
        }
    }

    for(obj in c("sample_size", "hsq", "SE"))
        attr(result, obj) <- args_attr[[1]][[obj]]
    class(result) <- class(args[[1]])

    result
}

# fit a single-QTL model at a single position, adjusting for polygenes with an LMM
#
fit1_pg <-
    function(genoprobs, pheno, kinship,
             addcovar=NULL, nullcovar=NULL, intcovar=NULL,
             weights=NULL, contrasts=NULL, zerosum=TRUE, se=FALSE,
             hsq=NULL, reml=TRUE, var=FALSE, ...)
{
    if(zerosum && se) var <- TRUE # force calculation of full variance matrix

    # deal with the dot args
    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    if(!is_pos_number(tol)) stop("tol should be a single positive number")
    check_extra_dots(dotargs, "tol")

    # check that the objects have rownames
    check4names(pheno, addcovar, NULL, intcovar, nullcovar)

    # force things to be matrices
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(nullcovar)) {
        if(!is.matrix(nullcovar)) nullcovar <- as.matrix(nullcovar)
        if(!is.numeric(nullcovar)) stop("nullcovar is not numeric")
    }
    if(!is.null(intcovar)) {
        if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)
        if(!is.numeric(intcovar)) stop("intcovar is not numeric")
    }
    if(!is.null(contrasts)) {
        if(!is.matrix(contrasts)) contrasts <- as.matrix(contrasts)
        if(!is.numeric(contrasts)) stop("contrasts is not numeric")
    }

    # make sure pheno is a vector
    if(is.matrix(pheno) || is.data.frame(pheno)) {
        if(ncol(pheno) > 1)
            warning("Considering only the first phenotype.")
        rn <- rownames(pheno)
        pheno <- pheno[,1]
        names(pheno) <- rn
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }

    # genoprobs is a matrix?
    if(!is.matrix(genoprobs))
        stop("genoprobs should be a matrix, individuals x genotypes")

    # make sure contrasts is square n_genotypes x n_genotypes
    if(!is.null(contrasts)) {
        ng <- ncol(genoprobs)
        if(ncol(contrasts) != ng || nrow(contrasts) != ng)
            stop("contrasts should be a square matrix, ", ng, " x ", ng)
    }

    # make sure kinship is for a single chromosome and get IDs
    did_decomp <- is_kinship_decomposed(kinship)
    kinship <- check_kinship_onechr(kinship)
    kinshipIDs <- check_kinship(kinship, 1)

    # multiply kinship matrix by 2; rest is using 2*kinship
    # see Almasy & Blangero (1998) https://doi.org/10.1086/301844
    kinship <- double_kinship(kinship)

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, pheno, kinshipIDs, weights,
                               addcovar, nullcovar, intcovar, complete.cases=TRUE)

    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    if(did_decomp) { # if did decomposition already, make sure it was with exactly
        if(length(kinshipIDs) != length(ind2keep) ||
           any(sort(kinshipIDs) != sort(ind2keep)))
            stop("Decomposed kinship matrix was with different individuals")
        else
            ind2keep <- kinshipIDs # force them in same order
    }

    # omit individuals not in common
    genoprobs <- genoprobs[ind2keep,,drop=FALSE]
    pheno <- pheno[ind2keep]
    if(!is.null(addcovar)) addcovar <- addcovar[ind2keep,,drop=FALSE]
    if(!is.null(nullcovar)) nullcovar <- nullcovar[ind2keep,,drop=FALSE]
    if(!is.null(intcovar)) intcovar <- intcovar[ind2keep,,drop=FALSE]
    if(!is.null(weights)) weights <- weights[ind2keep]

    # square-root of weights; multiply things by weights
    weights <- sqrt_weights(weights)
    kinship <- weight_kinship(kinship, weights)
    pheno <- weight_matrix(pheno, weights)
    addcovar <- weight_matrix(addcovar, weights)
    intcovar <- weight_matrix(intcovar, weights)
    nullcovar <- weight_matrix(nullcovar, weights)
    genoprobs <- weight_matrix(genoprobs, weights)
    intercept <- weights; if(is_null_weights(weights)) intercept <- rep(1,length(pheno))

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # make sure columns in intcovar are also in addcovar
    addcovar <- force_intcovar(addcovar, intcovar, tol)

    # eigen decomposition of kinship matrix
    if(!did_decomp)
        kinship <- decomp_kinship(kinship[ind2keep, ind2keep])

    # estimate hsq if necessary
    if(is.null(hsq)) {
        nullresult <- calc_hsq_clean(Ke=kinship, pheno=as.matrix(pheno),
                                     addcovar=cbind(addcovar, nullcovar),
                                     Xcovar=NULL, is_x_chr=FALSE, weights=weights, reml=reml,
                                     cores=1, check_boundary=TRUE, tol=tol)
        hsq <- as.numeric(nullresult$hsq)
    }

    # eigen-vectors and weights
    eigenvec <- kinship$vectors
    wts <- 1/sqrt(hsq*kinship$values + (1-hsq))

    # fit null model
    fit0 <- fit1_pg_addcovar(cbind(intercept, addcovar, nullcovar),
                             pheno,
                             matrix(ncol=0, nrow=length(pheno)),
                             eigenvec, wts, FALSE, FALSE, tol)

    # multiply genoprobs by contrasts
    if(!is.null(contrasts))
        genoprobs <- genoprobs %*% contrasts

    if(is.null(intcovar)) { # just addcovar
        if(is.null(addcovar)) addcovar <- matrix(nrow=length(ind2keep), ncol=0)
        fitA <- fit1_pg_addcovar(genoprobs, pheno, addcovar, eigenvec, wts, se, var, tol)
    }
    else {                  # intcovar
        fitA <- fit1_pg_intcovar(genoprobs, pheno, addcovar, intcovar,
                                 eigenvec, wts, se, var, tol)
    }

    # lod score
    n <- length(pheno)
    lod <- (n/2)*log10(fit0$rss/fitA$rss)

    # names of coefficients
    coef_names <- scan1coef_names(genoprobs, addcovar, intcovar)

    # fitted values
    fitted <- setNames(fitA$fitted, names(pheno))
    resid <- pheno - fitted

    # center the QTL effects at zero and add an intercept
    if(zerosum && is.null(contrasts)) {
        ng <- dim(genoprobs)[2]
        whval <- seq_len(ng)

        A <- matrix(0, length(fitA$coef)+1, length(fitA$coef))
        diag(A) <- 1
        for(i in whval) { A[i,whval] <- -1/ng; A[i,i] <- 1-1/ng }
        A[nrow(A),whval] <- 1/ng

        fitA$coef <- as.numeric(A %*% fitA$coef)
        coef_names <- names(fitA$coef) <- c(coef_names, "intercept")

        if(var && !is.null(fitA$var)) {
            fitA$var <- A %*% fitA$var %*% t(A)
            fitA$SE <- sqrt(diag(fitA$var))
        }
        else if(se) {
            fitA$SE <- c(fitA$SE, sqrt(mean(fitA$SE[whval]^2, na.rm=TRUE)))
        }
    }

    if(var && !is.null(fitA$var)) { # include variance-covariance matrix, if full rank
        dimnames(fitA$var) <- list(coef_names, coef_names)
        return(list(lod=lod,
                    coef=stats::setNames(fitA$coef, coef_names),
                    SE=stats::setNames(fitA$SE, coef_names),
                    var=fitA$var,
                    fitted=fitted, resid=resid))
    }
    else if(se) # results include standard errors
        return(list(lod=lod,
                    coef=stats::setNames(fitA$coef, coef_names),
                    SE=stats::setNames(fitA$SE, coef_names),
                    fitted=fitted, resid=resid))
    else
        return(list(lod=lod,
                    coef=stats::setNames(fitA$coef, coef_names),
                    fitted=fitted, resid=resid))
}

#' plot strain distribution patterns for SNPs
#'
#' plot the strain distribution patterns of SNPs
#' using tracks of tick-marks for each founder strain
#'
#' @param pos vector of SNP positions
#' @param sdp vector of strain distribution patterns (as integers)
#' @param strain_labels names of the strains
#' @param ... additional graphic arguments
#'
#' @return None.
#'
#' @keywords hgraphics
#' @export
#' @importFrom graphics segments axis title rect box abline
#'
#' @seealso [calc_sdp()], [invert_sdp()]
#'
#' @details
#' Additional arguments, such as `xlab`, `ylab`, `xlim`, and `main`,
#' are passed via `...`; also `bgcolor` to control the color of the
#' background, and `col` and `lwd` to control the color and thickness
#' of the tick marks.
#'
#' @examples
#' n_tick <- 50
#' plot_sdp(runif(n_tick, 0, 100), sample(0:255, n_tick, replace=TRUE))

plot_sdp <-
    function(pos, sdp, strain_labels=names(qtl2::CCcolors), ...)
{
    n_str <- length(strain_labels)

    stopifnot(length(pos) == length(sdp))
    stopifnot(all(sdp < 2^n_str & sdp >= 0))

    alleles <- invert_sdp(sdp, n_str)

    y <- seq_len(n_str)

    plot_sdp_internal <-
        function(xlim=range(pos), ylim=c(max(y)+0.5, min(y)-0.5),
                 xlab="Position (Mbp)", ylab="",
                 xaxs="i", yaxs="i",
                 main="", mgp.x=c(2.6, 0.5, 0), mgp.y=c(2.6, 0.5, 0),
                 mgp=NULL, las=1,
                 hlines=NULL, hlines_col="black", hlines_lwd=1, hlines_lty=1,
                 vlines=NULL, vlines_col="white", vlines_lwd=1, vlines_lty=1,
                 bgcolor="gray90", lwd=2, col="darkslateblue", sub="", ...)
        {
            dots <- list(...)

            # margin parameters
            if(!is.null(mgp)) mgp.x <- mgp.y <- mgp

            # make basic plot
            plot(0, 0, xlab="", ylab="", xlim=xlim, ylim=ylim,
                 xaxs=xaxs, yaxs=yaxs, xaxt="n", yaxt="n", type="n",
                 main=main, sub=sub)

            # add background rectangle
            u <- par("usr")
            if(!is.null(bgcolor))
                rect(u[1], u[3], u[2], u[4], col=bgcolor, border=NA)

            # include axis labels?
            if(is.null(dots$xaxt)) dots$xaxt <- par("xaxt")
            if(is.null(dots$yaxt)) dots$yaxt <- par("yaxt")

            if(dots$xaxt != "n") {
                axis(side=1, at=pretty(xlim), mgp=mgp.x, las=las, tick=FALSE)
            }

            if(dots$yaxt != "n") {
                axis(side=2, at=y, strain_labels, mgp=mgp.y, las=las, tick=FALSE)
            }

            # add axis titles
            title(xlab=xlab, mgp=mgp.x)
            title(ylab=ylab, mgp=mgp.y)

            # grid lines
            if(!(length(hlines)==1 && is.na(hlines))) { # if hlines==NA, skip lines
                if(is.null(hlines)) hlines <- seq(0.5, n_str+0.5)
                abline(h=hlines, col=hlines_col, lwd=hlines_lwd, lty=hlines_lty)
            }
            if(!(length(vlines)==1 && is.na(vlines))) { # if vlines==NA (or mult chr), skip lines
                if(is.null(vlines)) vlines <- pretty(xlim)
                abline(v=vlines, col=vlines_col, lwd=vlines_lwd, lty=vlines_lty)
            }


            for(i in seq_len(n_str)) {
                if(all(alleles[,i] == 1)) next
                wh <- which(alleles[,i]==3)
                yy <- rep(y[i], length(wh))
                segments(pos[wh], yy-0.5, pos[wh], yy+0.5, lwd=lwd, col=col)
            }

            box()
        }

    plot_sdp_internal(...)

}

#' Estimate heritability with a linear mixed model
#'
#' Estimate the heritability of a set of traits via a linear mixed
#' model, with possible allowance for covariates.
#'
#' @param pheno A numeric matrix of phenotypes, individuals x phenotypes.
#' @param kinship A kinship matrix.
#' @param addcovar An optional numeric matrix of additive covariates.
#' @param weights An optional numeric vector of positive weights for the
#' individuals. As with the other inputs, it must have `names`
#' for individual identifiers.
#' @param reml If true, use REML; otherwise, use maximimum likelihood.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#' @param ... Additional control parameters (see details).
#'
#' @return A vector of estimated heritabilities, corresponding to the
#' columns in `pheno`. The result has attributes `"sample_size"`,
#' `"log10lik"` and `"resid_sd"`.
#'
#' @details
#' We fit the model \eqn{y = X \beta + \epsilon}{y = Xb + e} where
#' \eqn{\epsilon}{e} is multivariate normal with mean 0 and covariance
#' matrix \eqn{\sigma^2 [h^2 (2 K) + (1-h^2) I]}{sigmasq*[hsq*2*K+(1-hsq)*I]} where
#' \eqn{K} is the kinship matrix and \eqn{I} is the identity matrix.
#'
#' If `weights` are provided, the covariance matrix becomes
#' \eqn{\sigma^2 [h^2 (2 K) + (1-h^2) D]}{sigmasq*[hsq*2*K+(1-hsq)*D]} where
#' \eqn{D} is a diagonal matrix with the reciprocal of the weights.
#'
#' For each of the inputs, the row names are used as
#' individual identifiers, to align individuals.
#'
#' If `reml=TRUE`, restricted maximum likelihood (reml) is used
#' to estimate the heritability, separately for each phenotype.
#'
#' Additional control parameters include `tol` for the tolerance
#' for convergence, `quiet` for controlling whether messages will
#' be display, `max_batch` for the maximum number of phenotypes
#' in a batch, and `check_boundary` for whether the 0 and 1
#' boundary values for the estimated heritability will be checked
#' explicitly.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c("19","X")] # subset to two chromosomes}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # kinship matrix
#' kinship <- calc_kinship(probs)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#'
#' # perform genome scan
#' hsq <- est_herit(pheno, kinship, covar)
#'
#' @export
est_herit <-
    function(pheno, kinship, addcovar=NULL, weights=NULL, reml=TRUE, cores=1, ...)
{
    if(is.null(pheno)) stop("pheno is NULL")

    dotargs <- list(...)
    tol <- grab_dots(dotargs, "tol", 1e-12)
    stopifnot(tol > 0)
    quiet <- grab_dots(dotargs, "quiet", TRUE)
    max_batch <- grab_dots(dotargs, "max_batch", NULL)
    check_boundary <- grab_dots(dotargs, "check_boundary", TRUE)
    check_extra_dots(dotargs, c("tol", "quiet", "max_batch",
                                "check_boundary"))

    # force things to be matrices
    if(!is.matrix(pheno))
        pheno <- as.matrix(pheno)
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }

    # check that kinship matrices are square with same IDs
    if(!is.matrix(kinship) || nrow(kinship) != ncol(kinship) || any(rownames(kinship) != colnames(kinship)))
        stop("kinship should be a square matrix with common row and column names")
    kinshipIDs <- rownames(kinship)

    # multiply kinship matrix by 2; rest is using 2*kinship
    # see Almasy & Blangero (1998) https://doi.org/10.1086/301844
    kinship <- double_kinship(kinship)

    # square-root of weights
    weights <- sqrt_weights(weights) # also check >0 (and if all 1's, turn to NULL)

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(kinshipIDs, addcovar, weights, complete.cases=TRUE)
    ind2keep <- get_common_ids(ind2keep, pheno[rowSums(is.finite(pheno)) > 0,,drop=FALSE])
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # batch phenotypes by missing values
    phe_batches <- batch_cols(pheno[ind2keep,,drop=FALSE], max_batch)

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # to contain the results
    hsq <- rep(NA, ncol(pheno))
    names(hsq) <- colnames(pheno)
    sigma <- n <- nullLL <- hsq

    # loop over batches of phenotypes with the same pattern of NAs
    for(batch in seq_along(phe_batches)) {

        # info about batch
        omit <- phe_batches[[batch]]$omit # ind to omit
        phecol <- phe_batches[[batch]]$cols # phenotype columns in batch

        # individuals to keep in this batch
        these2keep <- ind2keep
        if(length(omit)>0) these2keep <- ind2keep[-omit]
        n[phecol] <- length(these2keep)
        if(length(these2keep) <= 2) next # not enough individuals; skip this batch

        # subset the rest
        K <- kinship[these2keep, these2keep]
        ac <- addcovar; if(!is.null(ac)) { ac <- ac[these2keep,,drop=FALSE]; ac <- drop_depcols(ac, TRUE, tol) }
        ph <- pheno[these2keep,phecol,drop=FALSE]
        wts <- weights; if(!is.null(wts)) { wts <- wts[these2keep] }

        # multiply stuff by the weights
        K <- weight_kinship(K, wts)
        ac <- weight_matrix(ac, wts)
        ph <- weight_matrix(ph, wts)

        # eigen decomposition of kinship matrix
        Ke <- decomp_kinship(K, cores=cores)

        # fit LMM for each phenotype, one at a time
        nullresult <- calc_hsq_clean(Ke=Ke, pheno=ph, addcovar=ac, Xcovar=NULL,
                                     is_x_chr=FALSE, weights=wts, reml=reml, cores=cores,
                                     check_boundary=check_boundary, tol=tol)
        hsq[phecol] <- nullresult$hsq
        nullLL[phecol] <- nullresult$loglik
        sigma[phecol] <- sqrt(nullresult$sigmasq)
    }

    attr(hsq, "sample_size") <- n
    attr(hsq, "log10lik") <- nullLL/log(10)
    attr(hsq, "resid_sd") <- sigma

    hsq
}

#' Plot comparison of two sets of genotype probabilities
#'
#' Plot a comparison of two sets of genotype probabilities for one individual on one chromosome, as a heat map.
#'
#' @param probs1 Genotype probabilities (as produced by [calc_genoprob()])
#' or allele dosages (as produced by [genoprob_to_alleleprob()]).
#' @param probs2 A second set of genotype probabilities, just like `probs1`.
#' @param map Marker map (a list of vectors of marker positions).
#' @param ind Individual to plot, either a numeric index or an ID.
#' @param chr Selected chromosome to plot; a single character string.
#' @param geno Optional vector of genotypes or alleles to be shown
#' (vector of integers or character strings)
#' @param threshold Threshold for genotype probabilities; only genotypes that achieve
#' this value somewhere on the chromosome (in one or the other set of probabilities) will be shown.
#' @param n_colors Number of colors in each color scale.
#' @param swap_axes If TRUE, swap the axes, so that the genotypes are
#' on the x-axis and the chromosome position is on the y-axis.
#' @param ... Additional graphics parameters passed to [graphics::image()].
#'
#' @return None.
#'
#' @details
#' We plot the first set of probabilities in the range white to blue
#' and the second set in the range white to red and attempt to combine
#' them, for colors that are white, some amount of blue or red, or
#' where both are large something like blackish purple.
#'
#' @seealso [plot_genoprob()]
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' iron <- iron[228,"1"]   # subset to one individual on chr 1
#' map <- insert_pseudomarkers(iron$gmap, step=5)
#'
#' # introduce genotype error and look at difference in genotype probabilities
#' pr_ne <- calc_genoprob(iron, map, error_prob=0.002)
#' iron$geno[[1]][1,2] <- 3
#' pr_e <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # image of probabilities + comparison
#' \dontshow{old_mfrow <- par("mfrow")}
#' par(mfrow=c(3,1))
#' plot_genoprob(pr_ne, map, main="No error")
#' plot_genoprob(pr_e, map, main="With an error")
#' plot_genoprobcomp(pr_ne, pr_e, map, main="Comparison")
#' \dontshow{par(mfrow=old_mfrow)}
#'
#' @export
#' @importFrom grDevices rgb
plot_genoprobcomp <-
    function(probs1, probs2, map, ind=1, chr=NULL, geno=NULL,
             threshold=0, n_colors=256, swap_axes=FALSE, ...)
{
    # check inputs
    if(is.null(map)) stop("map is NULL")
    if(is.null(probs1)) stop("probs1 is NULL")
    if(is.null(probs2)) stop("probs2 is NULL")
    if(!is_nonneg_number(threshold)) stop("threshold should be a non-negative number")
    if(!is_pos_number(n_colors)) stop("n_colors should be a positive integer")

    if(is.null(chr)) chr <- names(probs1)[1]
    if(length(chr) > 1) {
        warning("chr should have length 1; using the first value")
        chr <- chr[1]
    }
    if(!(chr %in% names(probs1) && chr %in% names(probs2)))
        stop("chr ", chr, " not found in both probs")
    if(!(chr %in% names(map))) stop("chr ", chr, " not found in map")
    if(length(ind) > 1) {
        warning("ind should have length 1; using the first value")
        ind <- ind[1]
    }

    # pull out selected chromosomes
    probs1 <- probs1[[chr]]
    probs2 <- probs2[[chr]]
    map <- map[[chr]]

    # find common markers
    mar1 <- dimnames(probs1)[[3]]
    mar2 <- dimnames(probs2)[[3]]
    marm <- names(map)
    mar <- mar1[mar1 %in% mar2 & mar1 %in% marm]
    if(length(mar) < 2) stop("<2 markers in common")
    probs1 <- probs1[,,mar,drop=FALSE]
    probs2 <- probs2[,,mar,drop=FALSE]
    map <- map[mar]

    # make sure they have the same genotypes
    if(ncol(probs1) != ncol(probs2) || !all(colnames(probs1) == colnames(probs2)))
        stop("Need ncol(probs1) == ncol(probs2) and same colnames")

    # pull out individual's probs; make it positions x probs
    if(is.character(ind) && !(ind %in% rownames(probs1) && ind %in% rownames(probs2)))
        stop("ind ", ind, " not found in both sets of probs")
    if(is.numeric(ind)) {
        if(nrow(probs1) != nrow(probs2))
            stop("If ind is numeric, must have nrow(probs1) == nrow(probs2)")
        if(ind < 1 || ind > nrow(probs1))
            stop("ind ", ind, " should be in the range [1, ", nrow(probs1), "]")
    }

    # now pull out individual and transpose
    probs1 <- t(probs1[ind,,])
    probs2 <- t(probs2[ind,,])

    # pull out selected genotypes
    if(!is.null(geno)) {
        if(is.numeric(geno)) {
            if(all(geno < 0)) {
                if(any(geno > -1 | geno < -ncol(probs1)))
                    stop("negative geno should be in the range [", -ncol(probs1), " , -1]")
                geno <- seq_len(ncol(probs1))[geno]
            }
            if(any(geno < 1 | geno > ncol(probs1)))
                stop("numeric geno should be in the range [1, ", ncol(probs1))
            geno <- colnames(probs1)[geno]
        }
        if(!all(geno %in% colnames(probs1)))
            stop("Not all geno in probs")
        probs1 <- probs1[,geno,drop=FALSE]
        probs2 <- probs2[,geno,drop=FALSE]
    }

    # drop genotypes that do exceed threshold
    if(threshold > 0) {
        geno_keep <- colSums(probs1 >= threshold) + colSums(probs2 >= threshold) > 0
        if(sum(geno_keep) == 0) stop("No genotype probabilities exceed the threshold")
        probs1 <- probs1[,geno_keep,drop=FALSE]
        probs2 <- probs2[,geno_keep,drop=FALSE]
    }

    # save dimnames
    dn <- dimnames(probs1)

    # split probabilities into integer levels
    #    need to screw around a bit with case of 1+epsilon
    probs1[probs1 > 1] <- 1
    probs1 <- matrix(as.numeric(cut(probs1, breaks=seq(0, 1, length=n_colors+1), include.lowest=TRUE, right=TRUE)),
                     ncol=ncol(probs1), nrow=nrow(probs1))

    probs2[probs2 > 1] <- 1
    probs2 <- matrix(as.numeric(cut(probs2, breaks=seq(0, 1, length=n_colors+1), include.lowest=TRUE, right=TRUE)),
                     ncol=ncol(probs2), nrow=nrow(probs2))

    # combine into one object
    probs <- (probs1-1)*n_colors + probs2
    dimnames(probs) <- dn

    # create color palette
    red <- blue <- matrix(1, ncol=3, nrow=n_colors^2)
    # using 0.2 - 1 for the range of colors so that the combination is more purple rather than black
    # (also if ending at 0, the blue is a bit ugly)
    blue[,1] <- blue[,2] <- rep(seq(1, 0.2, length=n_colors), each=n_colors)
    red[,2] <- red[,3] <-   rep(seq(1, 0.2, length=n_colors), n_colors)
    joint_colors <- apply(blue*red, 1, function(a) rgb(a[1], a[2], a[3], maxColorValue=1))

    plot_genoprob_internal(probs, map, col=joint_colors, swap_axes=swap_axes,
                           zlim=c(1,n_colors^2), ...)

}

#' Plot one individual's genome-wide genotypes
#'
#' Plot one individual's genome-wide genotypes
#'
#' @param geno Imputed phase-known genotypes, as a list of matrices
#'     (as produced by [maxmarg()]) or a list of
#'     three-dimensional arrays (as produced by [guess_phase()]).
#' @param map Marker map (a list of vectors of marker positions).
#' @param ind Individual to plot, either a numeric index or an ID.
#' @param chr Selected chromosomes to plot; a vector of character strings.
#' @param col Vector of colors for the different genotypes.
#' @param na_col Color for missing segments.
#' @param swap_axes If TRUE, swap the axes, so that the chromosomes run horizontally.
#' @param border Color of outer border around chromosome rectangles.
#' @param shift If TRUE, shift the chromosomes so they all start at 0.
#' @param chrwidth Total width of rectangles for each chromosome, as a
#'     fraction of the distance between them.
#' @param ... Additional graphics parameters
#'
#' @return None.
#'
#' @seealso [plot_geno()], [plot_genoprob()]
#'
#' @section Hidden graphics parameters:
#' A number of graphics parameters can be passed via `...`. For
#' example, `bgcolor` to control the background color.
#' These are not included as formal parameters in order to avoid
#' cluttering the function definition.
#'
#' @examples
#' # load data and calculate genotype probabilities
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' iron <- iron["146", ] # subset to individual 146
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#' pr <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # infer genotypes, as those with maximal marginal probability
#' m <- maxmarg(pr)
#'
#' # guess phase
#' ph <- guess_phase(iron, m)
#'
#' # plot phased genotypes
#' plot_onegeno(ph, map, shift=TRUE, col=c("slateblue", "Orchid"))
#'
#' # this is more interesting for Diversity Outbred mouse data
#' \dontrun{
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' # subset to individuals labeled "232" and "256"
#' DOex <- DOex[c("232", "256"), ]
#' pr <- calc_genoprob(DOex, error_prob=0.002)
#'
#' # infer genotypes, as those with maximal marginal probability
#' m <- maxmarg(pr, minprob=0.5)
#' # guess phase
#' ph <- guess_phase(DOex, m)
#'
#' # plot phased genotypes
#' plot_onegeno(ph, DOex$gmap, shift=TRUE)
#' plot_onegeno(ph, DOex$gmap, ind="256", shift=TRUE)
#' }
#'
#' @export
#' @importFrom graphics rect par axis title abline box
plot_onegeno <-
    function(geno, map, ind=1, chr=NULL,
             col=NULL, na_col="white",
             swap_axes=FALSE,
             border="black", shift=FALSE,
             chrwidth=0.5, ...)
{
    if(is.null(geno)) stop("geno is NULL")
    if(is.null(map)) stop("map is NULL")
    if(!is_pos_number(chrwidth) || chrwidth >= 1) stop("chrwidth should be a number in (0, 1)")

    # ignore class of geno object
    geno <- unclass(geno)

    # drop all but the target individual
    if(length(ind)>1) {
        ind <- ind[1]
        warning("Only using the first individual")
    }
    for(i in seq_along(geno)) {
        if(is.matrix(geno[[i]]))
            geno[[i]] <- geno[[i]][ind,,drop=FALSE]
        else {
            if(!is.array(geno[[i]]) || length(dim(geno[[i]])) != 3 ||
               dim(geno[[i]])[3] != 2)
                stop("geno should be an array individuals x positions x 2 haplotypes")
            geno[[i]] <- geno[[i]][ind,,,drop=FALSE]
        }
    }

    # find common chr
    chr_geno <- names(geno)
    chr_map <- names(map)
    common_chr <- chr_geno[chr_geno %in% chr_map]
    if(length(common_chr) == 0)
        stop("No chr in common between geno and map")
    geno <- geno[common_chr]
    map <- map[common_chr]

    # subset chr if necessary
    if(!is.null(chr)) {
        if(any(!(chr %in% common_chr))) {
            chr <- chr[chr %in% common_chr]
            if(length(chr) == 0)
                stop("No chromosomes in common between geno, map, and chr")
            warning("Dropping some chr not found in geno and/or map")
        }
        geno <- geno[chr]
        map <- map[chr]
    }

    # same numbers of markers?
    nmar_map <- sapply(map, length)
    nmar_geno <- sapply(geno, ncol)
    if(any(nmar_geno != nmar_map))
        stop("Mismatch between numbers of markers between geno and map on chr ",
             paste(names(geno)[nmar_geno != nmar_map], collapse=", "))
    for(i in seq_along(geno)) {
        if(!all(names(map[[i]]) == colnames(geno[[i]])))
            stop("Mismatch between marker names on chr ", names(geno)[i])
    }

    # shift map to start at 0
    if(shift) map <- lapply(map, function(a) a-min(a,na.rm=TRUE))

    plot_onegeno_internal <-
        function(geno, map, col=NULL, na_col="white",
                 swap_axes=FALSE,
                 border="black", bgcolor="gray90",
                 chrwidth=0.5,
                 xlab=NULL, ylab=NULL,
                 xlim=NULL, ylim=NULL, las=1, xaxs=NULL, yaxs=NULL,
                 mgp.x=c(2.6,0.5,0), mgp.y=c(2.6,0.5,0), mgp=NULL,
                 hlines=NULL, hlines_col="white", hlines_lwd=1, hlines_lty=1,
                 vlines=NULL, vlines_col="gray80", vlines_lwd=1, vlines_lty=1,
                 ...)
    {
        dots <- list(...)

        nchr <- length(map)

        # margin parameters
        if(!is.null(mgp)) mgp.x <- mgp.y <- mgp

        if(swap_axes) {
            if(is.null(xlab)) xlab <- "Position"
            if(is.null(ylab)) ylab <- "Chromosome"

            if(is.null(xlim)) xlim <- range(unlist(map), na.rm=TRUE)
            if(is.null(ylim)) ylim <- c(nchr+0.5, 0.5)

            if(is.null(hlines)) hlines <- seq_len(nchr)
            if(is.null(vlines)) vlines <- pretty(xlim)

            if(is.null(xaxs)) xaxs <- "r"
            if(is.null(yaxs)) yaxs <- "i"
        }
        else {
            if(is.null(xlab)) xlab <- "Chromosome"
            if(is.null(ylab)) ylab <- "Position"

            if(is.null(xlim)) xlim <- c(0.5, nchr+0.5)
            if(is.null(ylim)) ylim <- rev(range(unlist(map), na.rm=TRUE))

            if(is.null(hlines)) hlines <- pretty(ylim)
            if(is.null(vlines)) vlines <- seq_len(nchr)

            if(is.null(xaxs)) xaxs <- "i"
            if(is.null(yaxs)) yaxs <- "r"
        }

        plot(0, 0, type="n", xlab="", ylab="",
             xaxs=xaxs, yaxs=yaxs,
             xaxt="n", yaxt="n", xlim=xlim, ylim=ylim, ...)
        u <- par("usr")
        if(!is.null(bgcolor))
            rect(u[1], u[3], u[2], u[4], col=bgcolor, border=NA)

        # include axis labels?
        if(is.null(dots$xaxt)) dots$xaxt <- par("xaxt")
        if(is.null(dots$yaxt)) dots$yaxt <- par("yaxt")

        if(swap_axes) {
            # add y axis unless par(yaxt="n")
            if(dots$yaxt != "n") {
                odd <- seq(1, nchr, by=2)
                axis(side=2, at=odd, names(map)[odd],
                     mgp=mgp.y, las=las, tick=FALSE)
                if(nchr > 1) {
                    even <- seq(2, nchr, by=2)
                    axis(side=2, at=even, names(map)[even],
                         mgp=mgp.y, las=las, tick=FALSE)
                }
            }
            # add x axis unless par(xaxt="n")
            if(dots$xaxt != "n") {
                axis(side=1, at=pretty(xlim), mgp=mgp.x, las=las, tick=FALSE)
            }
        } else {
            # add x axis unless par(xaxt="n")
            if(dots$xaxt != "n") {
                odd <- seq(1, nchr, by=2)
                axis(side=1, at=odd, names(map)[odd],
                     mgp=mgp.x, las=las, tick=FALSE)
                if(nchr > 1) {
                    even <- seq(2, nchr, by=2)
                    axis(side=1, at=even, names(map)[even],
                         mgp=mgp.x, las=las, tick=FALSE)
                }
            }
            # add y axis unless par(yaxt="n")
            if(dots$yaxt != "n") {
                axis(side=2, at=pretty(ylim), mgp=mgp.y, las=las, tick=FALSE)
            }
        }

        # grid lines
        if(!(length(vlines)==1 && is.na(vlines))) {
            abline(v=vlines, col=vlines_col, lwd=vlines_lwd, lty=vlines_lty)
        }
        if(!(length(hlines)==1 && is.na(hlines))) {
            abline(h=hlines, col=hlines_col, lwd=hlines_lwd, lty=hlines_lty)
        }

        # x and y axis labels
        title(xlab=xlab, mgp=mgp.x)
        title(ylab=ylab, mgp=mgp.y)

        max_geno <- max(unlist(geno), na.rm=TRUE)
        if(is.null(col)) {
            if(max_geno <= 8) {
                col <- qtl2::CCcolors
            }
            else {
                warning("With ", max_geno, " genotypes, you need to provide the vector of colors; recycling some")
                col <- rep(qtl2::CCcolors, max_geno)
            }
        }
        else if(max_geno > length(col)) {
            warning("not enough colors; recycling them")
            col <- rep(col, max_geno)
        }

        for(i in seq_len(nchr)) {
            g <- geno[[i]]

            # if completely missing the second chr but not the first, treat as if we have just the one
            #   (this is a kludge to deal with males on X chr;
            #    really should use is_x_chr and is_female but we don't have it)
            this_chrwidth <- chrwidth
            if(!is.matrix(g) && !all(is.na(g[,,1])) && all(is.na(g[,,2]))) {
                g <- rbind(g[,,1]) # make it a row matrix
                this_chrwidth <- this_chrwidth/2
            }

            if(is.matrix(g)) { # phase-known
                if(swap_axes) {
                    rect(min(map[[i]], na.rm=TRUE), i-this_chrwidth/2,
                         max(map[[i]], na.rm=TRUE), i+this_chrwidth/2,
                         col=na_col, border=border, lend=1, ljoin=1)
                } else {
                    rect(i-this_chrwidth/2, min(map[[i]], na.rm=TRUE),
                         i+this_chrwidth/2, max(map[[i]], na.rm=TRUE),
                         col=na_col, border=border, lend=1, ljoin=1)
                }

                addgenorect(g[1,], map[[i]], i-this_chrwidth/2, i+this_chrwidth/2,
                            col=col, swap_axes=swap_axes)

                if(swap_axes) {
                    rect(min(map[[i]], na.rm=TRUE), i-this_chrwidth/2,
                         max(map[[i]], na.rm=TRUE), i+this_chrwidth/2,
                         col=NULL, border=border, lend=1, ljoin=1)
                } else {
                    rect(i-this_chrwidth/2, min(map[[i]], na.rm=TRUE),
                         i+this_chrwidth/2, max(map[[i]], na.rm=TRUE),
                         col=NULL, border=border, lend=1, ljoin=1)
                }
            }
            else {
                if(swap_axes) {
                    rect(min(map[[i]], na.rm=TRUE), i-this_chrwidth/2,
                         max(map[[i]], na.rm=TRUE), i,
                         col=na_col, border=border, lend=1, ljoin=1)
                    rect(min(map[[i]], na.rm=TRUE), i+this_chrwidth/2,
                         max(map[[i]], na.rm=TRUE), i,
                         col=na_col, border=border, lend=1, ljoin=1)
                } else {
                    rect(i-this_chrwidth/2, min(map[[i]], na.rm=TRUE),
                         i, max(map[[i]], na.rm=TRUE),
                         col=na_col, border=border, lend=1, ljoin=1)
                    rect(i+this_chrwidth/2, min(map[[i]], na.rm=TRUE),
                         i, max(map[[i]], na.rm=TRUE),
                         col=na_col, border=border, lend=1, ljoin=1)
                }

                addgenorect(g[1,,1], map[[i]], i-this_chrwidth/2, i,
                            col=col, swap_axes=swap_axes)
                addgenorect(g[1,,2], map[[i]], i+this_chrwidth/2, i,
                            col=col, swap_axes=swap_axes)

                if(swap_axes) {
                    rect(min(map[[i]], na.rm=TRUE), i-this_chrwidth/2,
                         max(map[[i]], na.rm=TRUE), i,
                         col=NULL, border=border, lend=1, ljoin=1)
                    rect(min(map[[i]], na.rm=TRUE), i+this_chrwidth/2,
                         max(map[[i]], na.rm=TRUE), i,
                         col=NULL, border=border, lend=1, ljoin=1)
                } else {
                    rect(i-this_chrwidth/2, min(map[[i]], na.rm=TRUE),
                         i, max(map[[i]], na.rm=TRUE),
                         col=NULL, border=border, lend=1, ljoin=1)
                    rect(i+this_chrwidth/2, min(map[[i]], na.rm=TRUE),
                         i, max(map[[i]], na.rm=TRUE),
                         col=NULL, border=border, lend=1, ljoin=1)
                }
            }
        }

        box()
    }

    plot_onegeno_internal(geno, map, col=col, na_col=na_col,
                          swap_axes=swap_axes, border=border,
                          chrwidth=chrwidth, ...)


}

# add rectangles for the genotypes
addgenorect <-
    function(geno, map, x1, x2, col, swap_axes=FALSE)
{
    intervals <- geno2intervals(geno, map)
    if(is.null(intervals) || nrow(intervals) < 1) return(NULL)

    for(i in seq_len(nrow(intervals))) {
        if(swap_axes) {
            rect(intervals[i,1], x1,
                 intervals[i,2], x2,
                 col=col[intervals[i,3]],
                 border=NA, lend=1, ljoin=1)
        } else{
            rect(x1, intervals[i,1],
                 x2, intervals[i,2],
                 col=col[intervals[i,3]],
                 border=NA, lend=1, ljoin=1)
        }
    }
}


# convert vector of integer genotypes to intervals with common genotypes
# (start, end, genotype)
geno2intervals <-
    function(geno, map)
{
    if(all(is.na(geno))) return(NULL)

    stopifnot(length(geno) == length(map))

    # drop missing values
    map <- map[!is.na(geno)]
    geno <- geno[!is.na(geno)]

    d <- diff(geno)
    xo_int <- which(d != 0)

    data.frame(lo=map[c(1,xo_int+1)],
               hi=map[c(xo_int, length(map))],
               geno=geno[c(xo_int, length(map))])

}

# align scan1 output with a map
#' @describeIn qtl2-internal Align `scan1()` output with a map
#' @keywords internal
#' @export
align_scan1_map <-
    function(scan1_output, map)
{
    if(is.null(scan1_output)) stop("scan1_output is NULL")
    if(is.null(map)) stop("map is NULL")

    if(!is.list(map)) stop("map should be a list")

    scan1_names <- rownames(scan1_output)
    map_names <- map2markernames(map)
    map_chr <- map2chr(map)

    # perfectly fine
    if(length(scan1_names) == length(map_names) &&
       all(scan1_names == map_names))
        return(list(scan1_output=scan1_output, map=map))

    if(!any(scan1_names %in% map_names))
        stop("scan1 output and map have no markers in common")

    # subset scan1_output to markers in the map
    #    and use order as in map
    scan1_attr <- attributes(scan1_output)
    names_ordered <- map_names[map_names %in% scan1_names]
    scan1_output <- scan1_output[names_ordered,,drop=FALSE]
    for(a in c("SE", "sample_size", "hsq", "class"))
        attr(scan1_output, a) <- scan1_attr[[a]]
    scan1_names <- rownames(scan1_output)

    # subset map to markers in scan1_output
    if(any(!(map_names %in% scan1_names))) {
        map_attr <- attributes(map)
        keep <- map_names %in% scan1_names
        uchr <- unique(map_chr[keep])
        map <- map[uchr]
        if("is_x_chr" %in% names(map_attr))
            map_attr$is_x_chr <- map_attr$is_x_chr[uchr]
        for(i in seq_along(map))
            map[[i]] <- map[[i]][names(map[[i]]) %in% scan1_names]
        for(a in c("is_x_chr", "class"))
            attr(map, a) <- map_attr[[a]]
    }

    list(scan1_output=scan1_output, map=map)
}

#' Pull genotype probabilities for a particular position
#'
#' Pull out the genotype probabilities for a particular position (by
#' name)
#'
#' @param genoprobs Genotype probabilities as calculated by
#' [calc_genoprob()].
#' @param map A map object: a list (corresponding to chromosomes) of
#'     vectors of marker positions. Can also be a snpinfo object (data
#'     frame with columns `chr` and `pos`; marker names taken from
#'     column `snp` or if that doesn't exist from the row names)
#' @param chr A chromosome ID
#' @param pos A numeric position
#' @param marker A single character string with the name of the
#' position to pull out.
#'
#' @return A matrix of genotype probabilities for the specified position.
#'
#' @details Provide either a marker/pseudomarker name (with the argument `marker`)
#' or all of `map`, `chr`, and `pos`.
#'
#' @seealso [find_marker()], [fit1()], [pull_genoprobint()]
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,"8"]}
#' gmap <- insert_pseudomarkers(iron$gmap, step=1)
#' pr <- calc_genoprob(iron, gmap, error_prob=0.002)
#'
#' pmar <- find_marker(gmap, 8, 40)
#' pr_8_40 <- pull_genoprobpos(pr, pmar)
#'
#' pr_8_40_alt <- pull_genoprobpos(pr, gmap, 8, 40)
#'
#' @export
pull_genoprobpos <-
    function(genoprobs, map=NULL, chr=NULL, pos=NULL, marker=NULL)
{
    if(is.null(genoprobs)) stop("genoprobs is NULL")

    if(is.character(map) && is.null(chr) && is.null(pos) && is.null(marker)) {
        # treat this as map -> marker
        marker <- map
        map <- NULL
    }

    if(!is.null(marker) && (!is.null(map) || !is.null(chr) || !is.null(pos))) {
        warning("Provide either {map,chr,pos} or marker, not both; using marker")
    }
    if(is.null(marker)) {
        marker <- find_marker(map, chr, pos)
    }

    if(length(marker) == 0) stop("marker has length 0")
    if(length(marker) > 1) {
        marker <- marker[1]
        warning("marker should have length 1; using the first value")
    }

    # pull out marker names
    markers <- dimnames(genoprobs)[[3]]
    # vector of chromosomes
    chr <- rep(names(genoprobs), lapply(markers, length))
    # vector of indexes
    index <- unlist(lapply(markers, function(a) seq_along(a)))
    # markers into single vector
    markers <- unlist(markers)

    # identify the particular position
    wh <- (marker == markers)

    # check that it's there exactly onces
    n_found <- sum(wh)
    if(n_found==0) stop('marker "', marker, '" not found')
    if(n_found > 1) stop('marker "', marker, '" appears ', n_found, ' times')

    # pull out that set of probabilities
    result <- genoprobs[[chr[wh]]][,,index[wh], drop=FALSE]

    # make sure it's a plain matrix
    d <- dim(result)
    dn <- dimnames(result)
    result <- matrix(result[,,1], nrow=d[1], ncol=d[2])
    dimnames(result) <- dn[1:2]

    result
}

# drop_markers
#' Drop markers from a cross2 object
#'
#' Drop a vector of markers from a cross2 object.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param markers A vector of marker names.
#'
#' @return The input `cross` with the specified markers removed.
#'
#' @export
#' @seealso [pull_markers()], [drop_nullmarkers()], [reduce_markers()], [find_dup_markers()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' markers2drop <- c("BH.342C/347L-Col", "GH.94L", "EG.357C/359L-Col", "CD.245L", "ANL2")
#' grav2_rev <- drop_markers(grav2, markers2drop)
drop_markers <-
    function(cross, markers)
{
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')

    # to keep track of which markers were found
    found <- rep(FALSE, length(markers))
    names(found) <- markers

    chr2drop <- NULL
    for(i in seq(along=cross$geno)) {
        mn <- colnames(cross$geno[[i]])
        to_drop <- mn %in% markers

        if(any(to_drop)) {
            cross$geno[[i]] <- cross$geno[[i]][,!to_drop,drop=FALSE]
            found[mn[to_drop]] <- TRUE

            if("founder_geno" %in% names(cross))
                cross$founder_geno[[i]] <- cross$founder_geno[[i]][,!to_drop,drop=FALSE]
            if("gmap" %in% names(cross))
                cross$gmap[[i]] <- cross$gmap[[i]][!to_drop]
            if("pmap" %in% names(cross))
                cross$pmap[[i]] <- cross$pmap[[i]][!to_drop]

            # nothing left?
            if(all(to_drop)) chr2drop <- c(chr2drop, i)
        }
    }

    if(length(chr2drop) > 0) {
        cross$geno <- cross$geno[-chr2drop]

        if("founder_geno" %in% names(cross))
            cross$founder_geno <- cross$founder_geno[-chr2drop]
        if("gmap" %in% names(cross))
            cross$gmap <- cross$gmap[-chr2drop]
        if("pmap" %in% names(cross))
            cross$pmap <- cross$pmap[-chr2drop]
        if("is_x_chr" %in% names(cross))
            cross$is_x_chr <- cross$is_x_chr[-chr2drop]
    }
    if(length(cross$geno) == 0)
        stop("No markers left")

    if(any(!found))
        warning("Some markers not found: ", paste(markers[!found], collapse=", "))

    cross
}


# drop_nullmarkers
#' Drop markers with no genotype data
#'
#' Drop markers with no genotype data (or no informative genotypes)
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param quiet If FALSE, print information about how many markers were dropped.
#'
#' @return The input `cross` with the uninformative markers removed.
#'
#' @details We omit any markers that have completely missing data, or
#' if founder genotypes are present (e.g., for Diversity Outbreds),
#' the founder genotypes are missing or are all the same.
#'
#' @export
#' @seealso [drop_markers()], [pull_markers()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' # make a couple of markers missing
#' grav2$geno[[2]][,c(3,25)] <- 0
#' grav2_rev <- drop_nullmarkers(grav2)
drop_nullmarkers <-
    function(cross, quiet=FALSE)
{
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')

    ngeno <- n_typed(cross, "marker", "count")
    if(any(ngeno==0)) {
        if(!quiet) message("Dropping ", sum(ngeno==0), " markers with no data")
        cross <- drop_markers(cross, names(ngeno)[ngeno==0])
    }

    # look at founder_geno
    if("founder_geno" %in% names(cross)) {
        to_drop <- NULL
        for(i in seq(along=cross$founder_geno)) {
            fg <- cross$founder_geno[[i]]

            # any completely missing?
            ntyped <- colSums(fg!=0)
            if(any(ntyped==0))
                to_drop <- c(to_drop, names(ntyped)[ntyped==0])

            # any with none missing but all the same
            noninf <- apply(fg[,ntyped==nrow(fg),drop=FALSE], 2,
                            function(a) length(unique(a))==1 & all(a != 0))
            if(any(noninf))
                to_drop <- c(to_drop, names(noninf)[noninf])
        }

        if(length(to_drop) > 0) {
            if(!quiet) message("Dropping ", length(to_drop), " noninformative markers")
            cross <- drop_markers(cross, to_drop)
        }
    }

    cross
}


# pull_markers
#' Drop all but a specified set of markers
#'
#' Drop all markers from a cross2 object expect those in a specified vector.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param markers A vector of marker names.
#'
#' @return The input `cross` with only the specified markers.
#'
#' @export
#' @seealso [drop_markers()], [drop_nullmarkers()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' markers2drop <- c("BH.342C/347L-Col", "GH.94L", "EG.357C/359L-Col", "CD.245L", "ANL2")
#' grav2_rev <- pull_markers(grav2, markers2drop)
pull_markers <-
    function(cross, markers)
{
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')

    mn <- marker_names(cross)
    to_drop <- !(mn %in% markers)
    not_found <- !(markers %in% mn)

    if(all(to_drop))
        stop("No markers found")

    if(any(not_found))
        warning("Some markers not found: ", paste(markers[not_found], collapse=", "))

    drop_markers(cross, mn[to_drop])
}

context("read_cross2 handling of file inconsistencies")

test_that("read_cross2 deals with missing marker info", {

    ironfile <- system.file("extdata", "iron.zip", package="qtl2")
    iron <- read_cross2(ironfile)

    # unzip iron data to a temporary directory
    dir <- tempdir()
    unzipped_files <- utils::unzip(ironfile, exdir=dir)
    on.exit(unlink(unzipped_files)) # clean up

    # drop two markers from the geno file
    geno_file <- file.path(dir, "iron_geno.csv")
    g <- fread_csv(geno_file, rownames_included=FALSE)
    dropped_index <- c(20,28)
    dropped_markers <- colnames(g)[dropped_index]
    write.table(g[,-dropped_index], file=geno_file, sep=",",
                row.names=FALSE, col.names=TRUE, quote=FALSE)

    yaml_file <- file.path(dir, "iron.yaml")
    expect_warning( iron_sub <- read_cross2(yaml_file) )

    expect_equal(iron_sub, drop_markers(iron, dropped_markers))

    # drop two markers from the gmap file
    # and give another two missing positions
    gmap_file <- file.path(dir, "iron_gmap.csv")
    gmap <- fread_csv(gmap_file, rownames_included=FALSE)
    dropped_index2 <- c(14, 51)
    dropped_markers2 <- gmap[dropped_index2,1]
    na_index2 <- c(5, 56)
    na_markers2 <- gmap[na_index2,1]
    gmap[na_index2,3] <- NA
    write.table(gmap[-dropped_index2,], file=gmap_file, sep=",",
                row.names=FALSE, col.names=TRUE, quote=FALSE)

    expect_warning( iron_sub2 <- read_cross2(yaml_file) ) # this gives an error at the moment

    expect_equal(iron_sub2, drop_markers(iron_sub, c(dropped_markers2, na_markers2)))

    dropped_index <- c(dropped_index-1, dropped_index2, na_index2)
    dropped_markers <- c(dropped_markers, dropped_markers2, na_markers2)

    # drop two markers from the pmap file
    # and give another one a missing position
    pmap_file <- file.path(dir, "iron_pmap.csv")
    pmap <- fread_csv(pmap_file, rownames_included=FALSE)
    dropped_index3 <- c(4, 29)
    dropped_markers3 <- pmap[dropped_index3,1]
    na_index3 <- 33
    na_markers3 <- pmap[na_index3,1]
    pmap[na_index3,3] <- NA
    write.table(pmap[-dropped_index3,], file=pmap_file, sep=",",
                row.names=FALSE, col.names=TRUE, quote=FALSE)

    expect_warning( iron_sub3 <- read_cross2(yaml_file) ) # this gives an error at the moment
    expect_equal(iron_sub3, drop_markers(iron, c(dropped_markers, dropped_markers3, na_markers3)))

    # markers out of order in genotypes
    g <- fread_csv(geno_file)
    g <- g[,sample(ncol(g)),drop=FALSE]
    write.table(cbind(marker=rownames(g), g), file=geno_file, sep=",",
                row.names=FALSE, col.names=TRUE, quote=FALSE)

    expect_warning( iron_randgeno <- read_cross2(yaml_file) )
    expect_equal(iron_sub3, iron_randgeno)

    # markers out of order in physical map
    pmap <- fread_csv(pmap_file, rownames_included=FALSE)
    pmap_rand <- pmap
    pmap_rand <- pmap[sample(nrow(pmap)), , drop=FALSE]
    write.table(pmap_rand, file=pmap_file, sep=",",
                row.names=FALSE, col.names=TRUE, quote=FALSE)

    expect_warning( iron_randpmap <- read_cross2(yaml_file) )
    expect_equal(iron_sub3, iron_randpmap)

    # markers out of order within a chromosome on genetic map
    gmap <- fread_csv(gmap_file, rownames_included=FALSE)
    gmap_reordered <- gmap
    gmap_reordered[6:7,1] <- gmap[7:6,1] # swap two marker names
    write.table(gmap_reordered, file=gmap_file, sep=",",
                row.names=FALSE, col.names=TRUE, quote=FALSE)

    expect_warning( iron_marorder <- read_cross2(yaml_file) )
    expected <- iron_sub3
    expected$pmap[[2]] <- expected$pmap[[2]][c(2,1,3)]
    names(expected$gmap[[2]]) <- names(expected$gmap[[2]])[c(2,1,3)]
    expected$geno[[2]] <- expected$geno[[2]][,c(2,1,3)]
    expect_equal(iron_marorder, expected)

    # markers out of order within a chromosome on physical map
    ### first, put gmap and pmap back in order
    write.table(gmap, file=gmap_file, sep=",",
                row.names=FALSE, col.names=TRUE, quote=FALSE)
    write.table(pmap, file=pmap_file, sep=",",
                row.names=FALSE, col.names=TRUE, quote=FALSE)
    pmap_reordered <- pmap
    neworder <- c(40,37,43,41,38,42,39)
    pmap_reordered[37:43,1] <- pmap[neworder,1] # reorder marker names on chr 11
    write.table(pmap_reordered, file=pmap_file, sep=",",
                row.names=FALSE, col.names=TRUE, quote=FALSE)

    expect_warning( iron_marorder <- read_cross2(yaml_file) )
    expected <- iron_sub3
    tmp <- expected$pmap[[11]]
    names(tmp) <- names(tmp)[order(order(neworder))]
    expected$pmap[[11]] <- tmp[names(expected$pmap[[11]])]
    expect_equal(iron_marorder, expected)

    # a marker on different chromosomes in gmap and pmap
    pmap_mischr <- pmap
    pmap_mischr[50,2] <- "14" # marker on chr 15 -> 14
    pmap_mischr[50,3] <- 90
    write.table(pmap_mischr, file=pmap_file, sep=",",
                row.names=FALSE, col.names=TRUE, quote=FALSE)
    expect_warning( iron_mischr <- read_cross2(yaml_file) )
    expect_equal( iron_mischr, drop_markers(iron_sub3, pmap_mischr$marker[50]) )

})

context("drop_markers, drop_nullmarkers, pull_markers")

test_that("drop_markers works with riself", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))

    # extra markers in list
    markers2drop <- c("BH.342C/347L-Col", "GH.94L", "EG.357C/359L-Col", "blah", "CD.245L", "karl", "caleb", "ANL2")
    expect_warning(drop_markers(grav2, markers2drop), "Some markers not found: blah, karl, caleb")

    # drop some markers
    markers2drop <- markers2drop[-c(4,6,7)]
    result <- drop_markers(grav2, markers2drop)
    expected <- grav2
    for(i in c(2,4)) {
        mn <- marker_names(grav2[,i])
        expected$geno[[i]] <- expected$geno[[i]][,!(mn %in% markers2drop)]
        expected$gmap[[i]] <- expected$gmap[[i]][!(mn %in% markers2drop)]
    }
    expect_equal(result, expected)

    # drop a whole chromosome
    mn <- sample(marker_names(grav2[,3]))
    expect_equal(drop_markers(grav2, mn), grav2[,-3])

})

test_that("drop_markers works with f2", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    # drop some markers
    markers2drop <- c("D1Mit80", "D8Mit195", "D3Mit18", "D8Mit31", "D8Mit124",
                      "D5Mit30", "D9Mit10", "D4Mit352", "D1Mit17", "D16Mit70")
    result <- drop_markers(iron, markers2drop)
    expected <- iron
    for(i in c(1,3,4,5,8,9,16)) {
        mn <- marker_names(iron[,i])
        expected$geno[[i]] <- expected$geno[[i]][,!(mn %in% markers2drop),drop=FALSE]
        expected$gmap[[i]] <- expected$gmap[[i]][!(mn %in% markers2drop)]
        expected$pmap[[i]] <- expected$pmap[[i]][!(mn %in% markers2drop)]
    }
    expect_equal(result, expected)

    # drop a few chromosomes
    chr2drop <- c(5,6,9,13)
    mn <- sample(marker_names(iron[,chr2drop]))
    expect_equal(drop_markers(iron, mn), iron[,-chr2drop])

})

test_that("drop_nullmarkers works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    iron$geno[[1]][,1:3] <- 0
    iron$geno[[17]][,1] <- 0

    mn <- c(marker_names(iron[,1]),
            marker_names(iron[,17])[1])

    expect_equal(drop_nullmarkers(iron), drop_markers(iron, mn))

})

test_that("pull_markers works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    mn <- marker_names(iron)
    to_keep <- sample(mn, 50)
    to_drop <- mn[!(mn %in% to_keep)]

    expect_equal(pull_markers(iron, to_keep), drop_markers(iron, to_drop))

})

#' Calculate genotype frequencies
#'
#' Calculate genotype frequencies, by individual or by marker
#'
#' @param probs List of arrays of genotype probabilities, as
#'     calculated by [calc_genoprob()].
#' @param by Whether to summarize by individual or marker
#' @param omit_x If TRUE, results are just for the autosomes. If
#'     FALSE, results are a list of length two, containing the results
#'     for the autosomes and those for the X chromosome.
#'
#' @return
#' If `omit_x=TRUE`, the result is a matrix of genotype
#' frequencies; columns are genotypes and rows are either individuals
#' or markers.
#'
#' If necessary (that is, if `omit_x=FALSE`, the data include the
#' X chromosome, and the set of genotypes on the X chromosome are
#' different than on the autosomes), the result is a list with two
#' components (for the autosomes and for the X chromosome), each being
#' a matrix of genotype frequencies.
#'
#' @seealso [calc_raw_geno_freq()], [calc_het()]
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' p <- calc_genoprob(iron, err=0.002)
#'
#' # genotype frequencies by marker
#' tab_g <- calc_geno_freq(p, "marker")
#'
#' # allele frequencies by marker
#' ap <- genoprob_to_alleleprob(p)
#' tab_a <- calc_geno_freq(ap, "marker")
#'
#' @export
calc_geno_freq <-
    function(probs, by=c("individual", "marker"), omit_x=TRUE)
{
    by <- match.arg(by)

    if(is.null(probs)) stop("probs is NULL")

    if(is.cross2(probs))
        stop('Input probs is a "cross2" object but should be genotype probabilities, as from calc_genoprob')

    is_x_chr <- attr(probs, "is_x_chr")
    if(any(is_x_chr) && !all(is_x_chr) && omit_x) {
        probs <- probs[,!is_x_chr]
        is_x_chr <- is_x_chr[!is_x_chr]
        omit_x <- FALSE
    }

    if(any(is_x_chr) && any(!is_x_chr)) { # some autosome, some X chr
        ng <- dim(probs)[2,]
        g <- dimnames(probs)[[2]]
        if(length(unique(ng)) > 1 ||  # not all the same number of genotypes
           !all(vapply(g[-1], function(a) all(a==g[[1]]),TRUE))) { # not all the same genotypes
            return( list(A=calc_geno_freq(probs[,!is_x_chr], by, FALSE),
                         X=calc_geno_freq(probs[,is_x_chr], by, FALSE)) )
        }
    }

    # for rest, can assume that they're all one group

    n_chr <- length(probs)

    if(by=="individual") {
        # total markers
        total_mar <- sum(dim(probs)[3,])

        # summarize each chromosome
        result <- lapply(seq_len(n_chr), function(chr) apply(probs[[chr]], 1:2, sum))

        if(length(result)>1) {
            for(i in seq_along(result)[-1])
                result[[1]] <- result[[1]] + result[[i]]
        }
        return(result[[1]]/total_mar)
    }

    # else: by marker
    t(do.call("cbind", lapply(seq_len(n_chr), function(chr) apply(probs[[chr]], 2:3, mean))))
}

#' Calculate BLUPs of QTL effects in scan along one chromosome
#'
#' Calculate BLUPs of QTL effects in scan along one chromosome, with a
#' single-QTL model treating the QTL effects as random, with possible
#' allowance for covariates and for a residual polygenic effect.
#'
#' @param genoprobs Genotype probabilities as calculated by
#' [calc_genoprob()].
#' @param pheno A numeric vector of phenotype values (just one phenotype, not a matrix of them)
#' @param kinship Optional kinship matrix, or a list of kinship matrices (one
#' per chromosome), in order to use the LOCO (leave one chromosome
#' out) method.
#' @param addcovar An optional numeric matrix of additive covariates.
#' @param nullcovar An optional numeric matrix of additional additive
#' covariates that are used under the null hypothesis (of no QTL) but
#' not under the alternative (with a QTL). This is needed for the X
#' chromosome, where we might need sex as a additive covariate under
#' the null hypothesis, but we wouldn't want to include it under the
#' alternative as it would be collinear with the QTL effects. Only
#' used if `kinship` is provided but `hsq` is not, to get
#' estimate of residual heritability.
#' @param contrasts An optional numeric matrix of genotype contrasts, size
#' genotypes x genotypes. For an intercross, you might use
#' `cbind(mu=c(1,0,0), a=c(-1, 0, 1), d=c(0, 1, 0))` to get
#' mean, additive effect, and dominance effect. The default is the
#' identity matrix.
#' @param se If TRUE, also calculate the standard errors.
#' @param reml If `reml=TRUE`, use
#' REML to estimate variance components; otherwise maximum likelihood.
#' @param tol Tolerance value for convergence of linear mixed model fit.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#' @param quiet If FALSE, print message about number of cores used when multi-core.
#'
#' @return An object of class `"scan1coef"`: a matrix of estimated regression coefficients, of dimension
#'     positions x number of effects. The number of effects is
#'     `n_genotypes + n_addcovar`.
#' May also contain the following attributes:
#' * `SE` - Present if `se=TRUE`: a matrix of estimated
#'   standard errors, of same dimension as `coef`.
#' * `sample_size` - Vector of sample sizes used for each
#'   phenotype
#'
#' @details For each of the inputs, the row names are used as
#' individual identifiers, to align individuals.
#'
#' If `kinship` is provided, the linear mixed model accounts for
#' a residual polygenic effect, with a the polygenic variance
#' estimated under the null hypothesis of no (major) QTL, and then
#' taken as fixed as known in the scan to estimate QTL effects.
#'
#' If `contrasts` is provided, the genotype probability matrix,
#' \eqn{P}, is post-multiplied by the contrasts matrix, \eqn{A}, prior
#' to fitting the model. So we use \eqn{P \cdot A}{P A} as the \eqn{X}
#' matrix in the model. One might view the rows of
#' \ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}
#' as the set of contrasts, as the estimated effects are the estimated
#' genotype effects pre-multiplied by
#' \ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}.
#'
#' @references Haley CS, Knott SA (1992) A simple
#' regression method for mapping quantitative trait loci in line
#' crosses using flanking markers.  Heredity 69:315--324.
#'
#' Kang HM, Zaitlen NA, Wade CM, Kirby A, Heckerman D, Daly MJ, Eskin
#' E (2008) Efficient control of population structure in model
#' organism association mapping. Genetics 178:1709--1723.
#'
#' Robinson GK (1991) That BLUP is a good thing: The estimation of
#' random effects. Statist Sci 6:15--32.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c(7,19)] # reduce to chr 7 and 19}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#' \dontshow{probs[["7"]] <- probs[["7"]][,,1:5] # reduce to very small number}
#'
#' # convert to allele probabilities
#' aprobs <- genoprob_to_alleleprob(probs)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno[,1]
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#'
#' # calculate BLUPs of coefficients for chromosome 7
#' blup <- scan1blup(aprobs[,"7"], pheno, addcovar=covar)
#'
#' # leave-one-chromosome-out kinship matrix for chr 7
#' kinship7 <- calc_kinship(probs, "loco")[["7"]]
#'
#' # calculate BLUPs of coefficients for chromosome 7, adjusting for residual polygenic effect
#' blup_pg <- scan1blup(aprobs[,"7"], pheno, kinship7, addcovar=covar)
#'
#' @export
scan1blup <-
    function(genoprobs, pheno, kinship=NULL, addcovar=NULL, nullcovar=NULL,
             contrasts=NULL, se=FALSE, reml=TRUE,
             tol=1e-12, cores=1, quiet=TRUE)
{
    if(is.null(genoprobs)) stop("genoprobs is NULL")
    if(is.null(pheno)) stop("pheno is NULL")

    if(!is.null(kinship)) { # use LMM; see scan1_pg.R
        return(scan1blup_pg(genoprobs, pheno, kinship, addcovar, nullcovar,
                            contrasts, se, reml, tol, cores, quiet))
    }

    if(!is_pos_number(tol)) stop("tol should be a single positive number")

    # check that the objects have rownames
    check4names(pheno, addcovar, NULL, NULL, nullcovar)

    # force things to be matrices
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(nullcovar)) {
        if(!is.matrix(nullcovar)) nullcovar <- as.matrix(nullcovar)
        if(!is.numeric(nullcovar)) stop("nullcovar is not numeric")
    }
    if(!is.null(contrasts)) {
        if(!is.matrix(contrasts)) contrasts <- as.matrix(contrasts)
        if(!is.numeric(contrasts)) stop("contrasts is not numeric")
    }

    # make sure pheno is a vector
    if(is.matrix(pheno) || is.data.frame(pheno)) {
        if(ncol(pheno) > 1)
            warning("Considering only the first phenotype.")
        rn <- rownames(pheno)
        pheno <- pheno[,1]
        names(pheno) <- rn
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }

    # genoprobs has more than one chromosome?
    if(length(genoprobs) > 1)
        warning("Using only the first chromosome, ", names(genoprobs)[1])
    genoprobs <- genoprobs[[1]]

    # make sure contrasts is square n_genotypes x n_genotypes
    if(!is.null(contrasts)) {
        ng <- ncol(genoprobs)
        if(ncol(contrasts) != ng || nrow(contrasts) != ng)
            stop("contrasts should be a square matrix, ", ng, " x ", ng)
    }

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, pheno, addcovar, complete.cases=TRUE)
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    # omit individuals not in common
    genoprobs <- genoprobs[ind2keep,,,drop=FALSE]
    pheno <- pheno[ind2keep]
    if(!is.null(addcovar)) addcovar <- addcovar[ind2keep,,drop=FALSE]

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)
    # add the intercept
    addcovar <- cbind(intercept=rep(1, length(ind2keep)),
                      addcovar)
    rownames(addcovar) <- ind2keep

    # multiply genoprobs by contrasts
    if(!is.null(contrasts))
        genoprobs <- genoprobs_by_contrasts(genoprobs, contrasts)

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # determine batches
    n_pos <- dim(genoprobs)[[3]]
    batches <- batch_vec(seq_len(n_pos), ceiling(n_pos/n_cores(cores)))

    by_group_func <- function(i)
        scanblup(genoprobs[,,batches[[i]],drop=FALSE], pheno, addcovar, se, reml, tol)

    # scan to get BLUPs and coefficient estimates
    if(n_cores(cores)==1) {
        result <- scanblup(genoprobs, pheno, addcovar, se, reml, tol)
        coef <- t(result$coef)
        if(se) SE <- t(result$SE)
        else SE <- NULL
    } else {
        result <- cluster_lapply(cores, seq_along(batches), by_group_func)

        # check for problems (if clusters run out of memory, they'll return NULL)
        result_is_null <- vapply(result, is.null, TRUE)
        if(any(result_is_null))
            stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

        SE <- coef <- matrix(nrow=n_pos,ncol=nrow(result[[1]]$coef))
        for(i in seq_along(result)) {
            coef[batches[[i]],] <- t(result[[i]]$coef)
            if(se) SE[batches[[i]],] <- t(result[[i]]$SE)
            else SE <- NULL
        }
    }

    # add names
    coefnames <- scan1coef_names(genoprobs, addcovar, NULL)
    dimnames(coef) <- list(dimnames(genoprobs)[[3]], coefnames)
    if(se) dimnames(SE) <- dimnames(coef)

    result <- coef

    # add some attributes with details on analysis
    attr(result, "sample_size") <- length(ind2keep)
    attr(result, "SE") <- SE # include only if not NULL

    class(result) <- c("scan1coef", "scan1", "matrix")
    result
}

# assign some allele codes based on how many you need
assign_allele_codes <-
    function(num_alleles, genotypes=NULL)
{
    if(!is.null(genotypes)) { # genotypes provided; find unique letters
        alleles <- sort(unique(unlist(strsplit(genotypes, ""))))
        if(num_alleles <= length(alleles))
            return(alleles[1:num_alleles])
    }

    # otherwise, use capital letters
    if(num_alleles <= length(LETTERS))
        return(LETTERS[1:num_alleles])

    # ack! need more than 26. Use 2-letter codes.
    alleles <- paste0(LETTERS, rep(LETTERS, each=length(LETTERS)))
    if(num_alleles <= length(alleles))
        return(alleles[1:num_alleles])

    # ack! need more than 26^2. Just use numbers
    as.character(1:num_alleles)
}

context("assign allele codes")

test_that("assign_allele_codes works", {

    expect_equal(assign_allele_codes(2, c("AA", "AB", "BB")), c("A", "B"))
    expect_equal(assign_allele_codes(2, c("AA", "AC", "CC")), c("A", "C"))

    expect_equal(assign_allele_codes(2), c("A", "B"))

    expect_equal(assign_allele_codes(3, c("AA", "AB", "BA", "BB", "AC", "CA", "CC", "BC", "CB")), c("A", "B", "C"))

})

# summary functions

#' Basic summaries of a cross2 object
#'
#' Basic summaries of a cross2 object.
#'
#' @param cross2 An object of class `"cross2"`, as output by [read_cross2()]. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#'
#' @return Variously a number, vector of numbers, or vector of character strings.
#'
#' @name basic_summaries
#' @keywords utilities
#' @seealso [summary.cross2()]
NULL

#' @describeIn basic_summaries Number of individuals (either genotyped or phenotyped)
#' @export
n_ind <-
function(cross2)
{
    length( ind_ids(cross2) )
}

#' @describeIn basic_summaries Number of genotyped individuals
#' @export
n_ind_geno <-
function(cross2)
{
    length( ind_ids_geno(cross2) )
}

#' @describeIn basic_summaries Number of phenotyped individuals
#' @export
n_ind_pheno <-
function(cross2)
{
    length( ind_ids_pheno(cross2) )
}

#' @describeIn basic_summaries Number of individuals with covariate data
#' @export
n_ind_covar <-
function(cross2)
{
    length( ind_ids_covar(cross2) )
}

#' @describeIn basic_summaries Number of individuals with both genotype and phenotype data
#' @export
n_ind_gnp <-
function(cross2)
{
    length( ind_ids_gnp(cross2) )
}

#' @describeIn basic_summaries IDs of individuals (either genotyped or phenotyped)
#' @export
ind_ids <-
function(cross2)
{
    # going to sneak in use for calc_genoprob (or fst_genoprob)
    # this is probably really bad form
    if(inherits(cross2, "calc_genoprob")) {
        return(dimnames(cross2)[[1]])
    }

    if(!is.cross2(cross2))
        stop('Input cross must have class "cross2"')
    unique( c(ind_ids_geno(cross2), ind_ids_pheno(cross2), ind_ids_covar(cross2)) )
}

#' @describeIn basic_summaries IDs of genotyped individuals
#' @export
ind_ids_geno <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$geno)) return(NULL)

    rownames(cross2$geno[[1]])
}

#' @describeIn basic_summaries IDs of phenotyped individuals
#' @export
ind_ids_pheno <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$pheno)) return(NULL)

    rownames(cross2$pheno)
}

#' @describeIn basic_summaries IDs of individuals with covariate data
#' @export
ind_ids_covar <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$covar)) return(NULL)

    rownames(cross2$covar)
}

#' @describeIn basic_summaries IDs of individuals with both genotype and phenotype data
#' @export
ind_ids_gnp <-
function(cross2)
{
    pheID <- ind_ids_pheno(cross2)
    genID <- ind_ids_geno(cross2)

    # if either is empty, return NULL
    if(is.null(pheID) || is.null(genID)) return(NULL)

    id <- find_common_ids(genID, pheID)

    # if no individuals in common, return NULL
    if(length(id) == 0) return(NULL)

    id
}

#' @describeIn basic_summaries Number of chromosomes
#' @export
n_chr <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$geno)) {
        warning("No genotypes present")
        return(0)
    }

    length(cross2$geno)
}

#' @describeIn basic_summaries Number of founder strains
#' @export
n_founders <-
    function(cross2)
{
    if("founder_geno" %in% names(cross2)) {
        return(nrow(cross2$founder_geno[[1]]))
    }

    return(2) # if no founder_geno, must be just two founders, right?
}

#' @describeIn basic_summaries Names of founder strains
#' @export
founders <-
    function(cross2)
{
    if("founder_geno" %in% names(cross2)) {
        return(rownames(cross2$founder_geno[[1]]))
    }

    return(cross2$alleles)
}

#' @describeIn basic_summaries Chromosome names
#' @export
chr_names <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$geno)) {
        warning("No genotypes present")
        return(NULL)
    }

    names(cross2$geno)
}

#' @describeIn basic_summaries Total number of markers
#' @export
tot_mar <-
function(cross2)
{
    sum(n_mar(cross2))
}

#' @describeIn basic_summaries Number of markers on each chromosome
#' @export
n_mar <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$geno)) {
        warning("No genotypes present")
        return(0)
    }

    vapply(cross2$geno, ncol, 1)
}

#' @describeIn basic_summaries Marker names
#' @export
marker_names <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$geno)) {
        warning("No genotypes present")
        return(NULL)
    }

    nam <- unlist(lapply(cross2$geno, colnames))
    names(nam) <- NULL
    nam
}


#' @describeIn basic_summaries Number of phenotypes
#' @export
n_pheno <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$pheno)) {
        return(NULL)
    }

    ncol(cross2$pheno)
}

#' @describeIn basic_summaries Phenotype names
#' @export
pheno_names <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$pheno)) {
        return(NULL)
    }

    colnames(cross2$pheno)
}

#' @describeIn basic_summaries Number of covariates
#' @export
n_covar <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$covar)) {
        return(0)
    }

    ncol(cross2$covar)
}

#' @describeIn basic_summaries Covariate names
#' @export
covar_names <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$covar)) {
        return(NULL)
    }

    colnames(cross2$covar)
}
#' @describeIn basic_summaries Number of phenotype covariates
#' @export
n_phenocovar <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$phenocovar)) {
        return(0)
    }

    ncol(cross2$phenocovar)
}

#' @describeIn basic_summaries Phenotype covariate names
#' @export
phenocovar_names <-
function(cross2)
{
    if(!is.cross2(cross2))
        stop('Input must have class "cross2"')

    if(is.null(cross2$phenocovar)) {
        return(NULL)
    }

    colnames(cross2$phenocovar)
}


# print cross
#' Print a cross2 object
#'
#' Print a summary of a cross2 object
#'
#' @param x An object of class `"cross2"`, as output by [read_cross2()]. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param ... Ignored.
#'
#' @return None.
#'
#' @export
#' @keywords utilities
print.cross2 <-
function(x, ...)
{
    print(summary(x))

    invisible(NULL)
}

# summary_cross
#' Summary of cross2 object
#'
#' Summarize a cross2 object
#'
#' @param object An object of class `"cross2"`, as output by [read_cross2()]. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param ... Ignored.
#'
#' @return None.
#'
#' @export
#' @keywords utilities
#' @seealso [basic_summaries]
summary.cross2 <-
function(object, ...)
{
    # run checks
    check <- check_cross2(object)

    result <- list(crosstype=object$crosstype,
                   nind=n_ind(object),
                   nind_geno=n_ind_geno(object),
                   nind_pheno=n_ind_pheno(object),
                   nind_gnp=n_ind_gnp(object),
                   nchr=n_chr(object),
                   nmar=n_mar(object),
                   npheno=n_pheno(object),
                   ncovar=n_covar(object),
                   nphenocovar=n_phenocovar(object),
                   totmar=tot_mar(object))

    class(result) <- c("summary.cross2", "list")
    result
}

#' @export
print.summary.cross2 <-
function(x, ...)
{
    cat('Object of class cross2 (crosstype "', x$crosstype, '")\n\n', sep='')

    toprint <- c("Total individuals"=           x$nind,
                 "No. genotyped individuals"=   x$nind_geno,
                 "No. phenotyped individuals"=  x$nind_pheno,
                 "No. with both geno & pheno"=  x$nind_gnp,
                 "\nNo. phenotypes"=            x$npheno,
                 "No. covariates"=              x$ncovar,
                 "No. phenotype covariates"=    x$nphenocovar,
                 "\nNo. chromosomes"=           x$nchr,
                 "Total markers"=               x$totmar)

    print_aligned(toprint)

    cat("\nNo. markers by chr:\n")
    print(x$nmar)

    invisible(x)
}

print_aligned <-
function(x)
{
    newline_before <- grepl("\\n", names(x))
    names(x) <- sub("^\\n", "", names(x))

    ndig <- ceiling(log10(max(x, na.rm=TRUE)))
    n.char <- max(nchar(names(x)))

    format <- paste0("%-", n.char+1, "s  %", ndig, "d\n")

    for(i in seq(along=x)) {
        if(newline_before[i]) cat("\n")
        cat(sprintf(format, names(x)[i], x[i]))
    }
}

# is an object a cross2 object?
# look for "cross2" within class()
is.cross2 <-
    function(x)
    inherits(x, "cross2")

context("scan1 with binary phenotype")

test_that("scan1 with binary phenotype gives same result as R/qtl", {

    library(qtl)
    data(hyper)
    hyper <- hyper[c(2,3,8),] # subset to 3 chromosomes
    hyper <- calc.genoprob(hyper, error.prob=0.002, step=1)
    hyper$pheno <- cbind(hyper$pheno, bp_binary=as.numeric(hyper$pheno[,1] > median(hyper$pheno[,1])))
    out1 <- scanone(hyper, pheno.col=3, model="binary", method="hk")

    # out1 map to list
    map <- split(out1[,2], out1[,1])
    mapn <- split(rownames(out1), out1[,1])
    for(i in seq_along(map)) names(map[[i]]) <- mapn[[i]]

    # convert for R/qtl2
    hyper2 <- convert2cross2(hyper)
    pr <- calc_genoprob(hyper2, map, error_prob=0.002)
    out2 <- scan1(pr, hyper2$pheno[,"bp_binary",drop=FALSE], model="binary")

    expect_equal(out1[,3] , as.numeric(out2))

    # add a couple of random covariates
    set.seed(45170702)
    X <- matrix(rnorm(nind(hyper)*2, 20, 2), nrow=nind(hyper), ncol=2)
    rownames(X) <- ind_ids(hyper2)

    out1_ac <- scanone(hyper, pheno.col=3, model="binary", method="hk", addcovar=X)
    out2_ac <- scan1(pr, hyper2$pheno[,"bp_binary",drop=FALSE], model="binary", addcovar=X)
    expect_equal(out1_ac[,3] , as.numeric(out2_ac))

    # make one of them an interactive covariate
    out1_ic <- scanone(hyper, pheno.col=3, model="binary", method="hk",
                       addcovar=X, intcovar=X[,1])
    out2_ic <- scan1(pr, hyper2$pheno[,"bp_binary",drop=FALSE], model="binary",
                     addcovar=X, intcovar=X[,1,drop=FALSE])
    expect_equal(out1_ic[,3] , as.numeric(out2_ic))

    # make both interactive covariates
    out1_ic2 <- scanone(hyper, pheno.col=3, model="binary", method="hk",
                        addcovar=X, intcovar=X)
    out2_ic2 <- scan1(pr, hyper2$pheno[,"bp_binary",drop=FALSE], model="binary",
                      addcovar=X, intcovar=X)
    expect_equal(out1_ic2[,3] , as.numeric(out2_ic2))

})


test_that("scan1 with binary phenotype in intercross gives same results as R/qtl", {

    skip_on_cran()

    library(qtl)
    data(listeria)
    listeria <- listeria[c(4,11,16),] # subset to 3 chromosomes
    listeria <- calc.genoprob(listeria, error.prob=0.002, step=1)
    listeria$pheno <- cbind(listeria$pheno, binary=as.numeric(listeria$pheno[,1] == 264))
    expect_warning(out1 <- scanone(listeria, pheno.col=3, model="binary", method="hk"))

    # out1 map to list
    map <- split(out1[,2], out1[,1])
    mapn <- split(rownames(out1), out1[,1])
    for(i in seq_along(map)) names(map[[i]]) <- mapn[[i]]

    # convert for R/qtl2
    listeria2 <- convert2cross2(listeria)
    pr <- calc_genoprob(listeria2, map, error_prob=0.002)
    out2 <- scan1(pr, listeria2$pheno[,"binary",drop=FALSE], model="binary")

    expect_equal(out1[,3] , as.numeric(out2))

    # add a couple of random covariates
    set.seed(45170702)
    X <- matrix(rnorm(nind(listeria)*2, 20, 2), nrow=nind(listeria), ncol=2)
    rownames(X) <- ind_ids(listeria2)

    expect_warning(out1_ac <- scanone(listeria, pheno.col=3, model="binary", method="hk", addcovar=X))
    out2_ac <- scan1(pr, listeria2$pheno[,"binary",drop=FALSE], model="binary", addcovar=X)
    expect_equal(out1_ac[,3] , as.numeric(out2_ac))

    # make one of them an interactive covariate
    expect_warning(out1_ic <- scanone(listeria, pheno.col=3, model="binary", method="hk",
                                      addcovar=X, intcovar=X[,1]))
    out2_ic <- scan1(pr, listeria2$pheno[,"binary",drop=FALSE], model="binary",
                     addcovar=X, intcovar=X[,1,drop=FALSE])
    expect_equal(out1_ic[,3] , as.numeric(out2_ic))

    # make both interactive covariates
    expect_warning(out1_ic2 <- scanone(listeria, pheno.col=3, model="binary", method="hk",
                                       addcovar=X, intcovar=X))
    out2_ic2 <- scan1(pr, listeria2$pheno[,"binary",drop=FALSE], model="binary",
                      addcovar=X, intcovar=X)
    expect_equal(out1_ic2[,3] , as.numeric(out2_ic2))

})


test_that("scan1 with multiple binary phenotype gives same results as R/qtl", {

    skip_on_cran()

    library(qtl)
    data(listeria)
    listeria <- listeria[c(1,3,12),] # subset to 3 chromosomes
    listeria <- calc.genoprob(listeria, error.prob=0.002, step=1)
    listeria$pheno <- cbind(listeria$pheno, binary1=as.numeric(listeria$pheno[,1] == 264),
                            binary2 = as.numeric(listeria$pheno[,1] > 116.5),
                            binary3 = as.numeric(listeria$pheno[,1] > 91.1))

    expect_warning(out1 <- scanone(listeria, pheno.col=3:5, model="binary", method="hk"))

    # out1 map to list
    map <- split(out1[,2], out1[,1])
    mapn <- split(rownames(out1), out1[,1])
    for(i in seq_along(map)) names(map[[i]]) <- mapn[[i]]

    # convert for R/qtl2
    listeria2 <- convert2cross2(listeria)
    pr <- calc_genoprob(listeria2, map, error_prob=0.002)
    out2 <- scan1(pr, listeria2$pheno[,2:4], model="binary")

    for(i in 1:3) expect_equal(out1[,i+2], as.numeric(out2[,i]))

    # add a couple of random covariates
    set.seed(45170702)
    X <- matrix(rnorm(nind(listeria)*2, 20, 2), nrow=nind(listeria), ncol=2)
    rownames(X) <- ind_ids(listeria2)

    expect_warning(out1_ac <- scanone(listeria, pheno.col=3:5, model="binary", method="hk", addcovar=X))
    out2_ac <- scan1(pr, listeria2$pheno[,2:4], model="binary", addcovar=X)
    for(i in 1:3) expect_equal(out1_ac[,i+2], as.numeric(out2_ac[,i]))

    # make one of them an interactive covariate
    expect_warning(out1_ic <- scanone(listeria, pheno.col=3:5, model="binary", method="hk",
                                      addcovar=X, intcovar=X[,1]))
    out2_ic <- scan1(pr, listeria2$pheno[,2:4], model="binary",
                     addcovar=X, intcovar=X[,1,drop=FALSE])
    for(i in 1:3) expect_equal(out1_ic[,i+2], as.numeric(out2_ic[,i]))

    # make both interactive covariates
    expect_warning(out1_ic2 <- scanone(listeria, pheno.col=3:5, model="binary", method="hk",
                                       addcovar=X, intcovar=X))
    out2_ic2 <- scan1(pr, listeria2$pheno[,2:4], model="binary",
                      addcovar=X, intcovar=X)
    for(i in 1:3) expect_equal(out1_ic2[,i+2], as.numeric(out2_ic2[,i]))

})


test_that("scan1 with weights gives the same answer as glm()", {

    skip_on_cran()

    library(qtl)
    data(listeria)
    set.seed(20180717)
    listeria$pheno$sex <- sample(0:1, nind(listeria), replace=TRUE)
    listeria <- listeria[c(1,3,"X"), ] # subset to 3 chromosomes
    listeria <- convert2cross2(listeria)
    Xcovar <- get_x_covar(listeria)

    phe <- cbind(binary1=as.numeric(listeria$pheno[,1] == 264),
                 binary2 = as.numeric(listeria$pheno[,1] > 116.5),
                 binary3 = as.numeric(listeria$pheno[,1] > 91.1))
    rownames(phe) <- rownames(listeria$pheno)

    map <- insert_pseudomarkers(listeria$gmap, step=2.5)
    pr <- calc_genoprob(listeria, map)

    out <- scan1(pr, phe, model="binary", Xcovar=Xcovar)

    nmar <- sapply(map, length) # no. markers/pseudomarkers per chromosome
    csmar <- cumsum(c(0, nmar))
    pos <- c(37, 7, 9)

    # null loglik (X chr needs sex as covariate)
    dev_glm0 <- apply(phe, 2, function(y)
        glm(y ~ 1, family=binomial(link=logit))$deviance)
    dev_glm0X <- apply(phe, 2, function(y)
        glm(y ~ Xcovar, family=binomial(link=logit))$deviance)
    dev_glm0 <- list(dev_glm0, dev_glm0, dev_glm0X)

    for(i in 1:3) {
        dev_glm <- apply(phe, 2, function(y)
            glm(y ~ -1 + pr[[i]][,,pos[i]], family=binomial(link=logit))$deviance)

        lod_glm <- (dev_glm0[[i]] - dev_glm)/(2*log(10))

        expect_equal(out[csmar[i] + pos[i],], lod_glm)
    }

    # repeat with weights
    weights <- setNames(sample(1:4, n_ind(listeria), replace=TRUE), rownames(phe))

    out <- scan1(pr, phe, model="binary", Xcovar=Xcovar, weights=weights)

    # null loglik (X chr needs sex as covariate)
    dev_glm0 <- apply(phe, 2, function(y)
        glm(y ~ 1, family=binomial(link=logit), weights=weights)$deviance)
    dev_glm0X <- apply(phe, 2, function(y)
        glm(y ~ Xcovar, family=binomial(link=logit), weights=weights)$deviance)
    dev_glm0 <- list(dev_glm0, dev_glm0, dev_glm0X)

    for(i in 1:3) {
        dev_glm <- apply(phe, 2, function(y)
            glm(y ~ -1 + pr[[i]][,,pos[i]], family=binomial(link=logit), weights=weights)$deviance)

        lod_glm <- (dev_glm0[[i]] - dev_glm)/(2*log(10))

        expect_equal(out[csmar[i] + pos[i],], lod_glm)
    }

})

context("Calculation of kinship matrix")

test_that("calc_kinship works for RIL", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    grav2 <- grav2[1:10,c(1,3)]
    map <- insert_pseudomarkers(grav2$gmap, step=5)
    probs <- calc_genoprob(grav2, map, error_prob=0.002)
    grid <- calc_grid(grav2$gmap, step=5)
    probs_sub <- probs_to_grid(probs, grid)
    sim <- calc_kinship(probs_sub)

    # row and colnames okay
    expect_equal(rownames(sim), rownames(grav2$geno[[1]]))
    expect_equal(colnames(sim), rownames(grav2$geno[[1]]))

    # check a few pairs
    set.seed(88213118)
    pairs <- list(c(1,2))
    n <- n_ind(grav2)
    for(i in 1:5)
        pairs <- c(pairs, list(sample(n, 2), sample(n, 2)))
    expected <- rep(0, length(pairs))
    tot_pos <- 0
    for(i in seq(along=probs)) {
        for(k in seq(along=pairs))
            expected[k] <- expected[k] + sum(probs_sub[[i]][pairs[[k]][1],,] * probs_sub[[i]][pairs[[k]][2],,])
        tot_pos <- tot_pos + dim(probs_sub[[i]])[3]
    }
    expected <- expected/tot_pos
    for(k in seq(along=pairs))
        expect_equal(sim[pairs[[k]][1],pairs[[k]][2]], expected[k])

})

test_that("calc_kinship works for F2", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:10,c(1,3)]
    map <- insert_pseudomarkers(iron$gmap, step=5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    grid <- calc_grid(iron$gmap, step=5)
    probs_sub <- probs_to_grid(probs, grid)
    sim <- calc_kinship(probs_sub, omit_x=TRUE)

    # row and colnames okay
    expect_equal(rownames(sim), rownames(iron$geno[[1]]))
    expect_equal(colnames(sim), rownames(iron$geno[[1]]))

    f2_geno2alle <-
        function(prob, x_chr=FALSE)
        {
            if(x_chr) {
                prob[,1,] <- prob[,1,]+prob[,2,]/2+prob[,3,]/2+prob[,5,]
                prob[,2,] <- prob[,4,]+prob[,2,]/2+prob[,3,]/2+prob[,6,]
                return(prob[,1:2,])

            } else {
                prob[,1,] <- prob[,1,]+prob[,2,]/2
                prob[,2,] <- prob[,3,]+prob[,2,]/2
                return(prob[,1:2,])
            }
        }

    # check a few values
    set.seed(54028069)
    n <- nrow(probs[[1]])
    pairs <- list(c(1,2))
    for(i in 1:5)
        pairs <- c(pairs, list(sample(n, 2), sample(n, 2)))
    expected <- rep(0, length(pairs))
    tot_pos <- 0
    is_x_chr <- attr(probs_sub, "is_x_chr")
    for(i in which(!is_x_chr)) { # just use autosomes
        for(k in seq(along=pairs)) {
            prob_1 <- probs_sub[[i]][pairs[[k]][1],,,drop=FALSE]
            prob_2 <- probs_sub[[i]][pairs[[k]][2],,,drop=FALSE]
            # autosomes: convert to allele probs
            prob_1 <- f2_geno2alle(prob_1)
            prob_2 <- f2_geno2alle(prob_2)

            expected[k] <- expected[k] + sum(prob_1 * prob_2)
        }
        tot_pos <- tot_pos + dim(probs_sub[[i]])[3]
    }
    expected <- expected/tot_pos
    for(k in seq(along=pairs))
        expect_equal(sim[pairs[[k]][1],pairs[[k]][2]], expected[k])

    # version using genotype probabilities
    sim <- calc_kinship(probs_sub, use_allele_probs=FALSE, omit_x=TRUE)
    sim2 <- calc_kinship(probs_sub, use_allele_probs=FALSE, omit_x=TRUE)
    expect_equal(sim, sim2)

    # check a few values
    expected <- rep(0, length(pairs))
    tot_pos <- 0
    for(i in which(!is_x_chr)) { # just use autosomes
        for(k in seq(along=pairs)) {
            prob_1 <- probs_sub[[i]][pairs[[k]][1],,,drop=FALSE]
            prob_2 <- probs_sub[[i]][pairs[[k]][2],,,drop=FALSE]

            expected[k] <- expected[k] + sum(prob_1 * prob_2)
        }
        tot_pos <- tot_pos + dim(probs_sub[[i]])[3]
    }
    expected <- expected/tot_pos
    for(k in seq(along=pairs))
        expect_equal(sim[pairs[[k]][1],pairs[[k]][2]], expected[k])


    # also try with X chr (now the default)
    sim <- calc_kinship(probs_sub)

    # row and colnames okay
    expect_equal(rownames(sim), rownames(iron$geno[[1]]))
    expect_equal(colnames(sim), rownames(iron$geno[[1]]))

    # check a few values
    set.seed(54028069)
    n <- nrow(probs[[1]])
    pairs <- list(c(1,2))
    for(i in 1:5)
        pairs <- c(pairs, list(sample(n, 2), sample(n, 2)))
    expected <- rep(0, length(pairs))
    tot_pos <- 0
    for(i in seq(along=probs_sub)) {
        for(k in seq(along=pairs)) {
            prob_1 <- probs_sub[[i]][pairs[[k]][1],,,drop=FALSE]
            prob_2 <- probs_sub[[i]][pairs[[k]][2],,,drop=FALSE]
            prob_1 <- f2_geno2alle(prob_1, is_x_chr[i])
            prob_2 <- f2_geno2alle(prob_2, is_x_chr[i])

            expected[k] <- expected[k] + sum(prob_1 * prob_2)
        }
        tot_pos <- tot_pos + dim(probs_sub[[i]])[3]
    }
    expected <- expected/tot_pos
    for(k in seq(along=pairs)) {
        expect_equal(sim[pairs[[k]][1],pairs[[k]][2]], expected[k])
    }

    # version using genotype probabilities
    sim <- calc_kinship(probs_sub, omit_x=FALSE, use_allele_probs=FALSE)
    sim2 <- calc_kinship(probs_sub, omit_x=FALSE, use_allele_probs=FALSE)
    expect_equal(sim, sim2)

    # check a few values
    expected <- rep(0, length(pairs))
    tot_pos <- 0
    for(i in seq(along=probs_sub)) {
        for(k in seq(along=pairs)) {
            prob_1 <- probs_sub[[i]][pairs[[k]][1],,,drop=FALSE]
            prob_2 <- probs_sub[[i]][pairs[[k]][2],,,drop=FALSE]

            expected[k] <- expected[k] + sum(prob_1 * prob_2)
        }
        tot_pos <- tot_pos + dim(probs_sub[[i]])[3]
    }
    expected <- expected/tot_pos
    for(k in seq(along=pairs))
        expect_equal(sim[pairs[[k]][1],pairs[[k]][2]], expected[k])

})

test_that("calc_kinship chr & loco work for F2", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:10, c(1,3)]
    map <- insert_pseudomarkers(iron$gmap, step=5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    grid <- calc_grid(iron$gmap, step=5)
    probs_sub <- probs_to_grid(probs, grid)
    sim <- calc_kinship(probs_sub, omit_x=TRUE)

    sim_chr <- calc_kinship(probs_sub, "chr", omit_x=TRUE)
    sim_loco <- calc_kinship(probs_sub, "loco", omit_x=TRUE)

    # combine results from sim_chr and compare to sim
    sim_combchr <- sim_chr[[1]]*attr(sim_chr[[1]], "n_pos")
    for(i in seq_len(n_chr(iron))[-1])
        sim_combchr <- sim_combchr + sim_chr[[i]]*attr(sim_chr[[i]], "n_pos")
    totpos <- sum(sapply(sim_chr[seq_len(n_chr(iron))], attr, "n_pos"))
    sim_combchr <- sim_combchr/totpos
    attr(sim_combchr, "n_pos") <- totpos
    expect_equal(sim_combchr, sim)

    # calculate results with one chromosome at a time
    sim_alt <- vector("list", length(probs))
    names(sim_alt) <- names(probs)
    for(i in names(probs))
        sim_alt[[i]] <- calc_kinship(probs_sub[,i], omit_x=FALSE)
    expect_equal(sim_alt, sim_chr)

    # compare sim - sim_chr with sim_loco
    sim_loco_alt <- vector("list", length(probs))
    names(sim_loco_alt) <- names(probs)
    is_x_chr <- attr(probs, "is_x_chr")
    totpos <- attr(sim, "n_pos")
    for(i in seq(along=probs)) {
        if(is_x_chr[i])
            sim_loco_alt[[i]] <- sim
        else {
            npos <- attr(sim_chr[[i]], "n_pos")
            sim_loco_alt[[i]] <- (sim*totpos - sim_chr[[i]]*npos)/(totpos-npos)
            attr(sim_loco_alt[[i]], "n_pos") <- totpos-npos
        }
    }
    expect_equal(sim_loco_alt, sim_loco)

})

test_that("calc_kinship chr & loco work for F2, when including X", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:10,c(1,3,"X")]
    map <- insert_pseudomarkers(iron$gmap, step=5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    grid <- calc_grid(iron$gmap, step=5)
    probs_sub <- probs_to_grid(probs, grid)
    sim <- calc_kinship(probs_sub, omit_x=FALSE) # *not* omitting X chr

    sim_chr <- calc_kinship(probs_sub, "chr")
    sim_loco <- calc_kinship(probs_sub, "loco", omit_x=FALSE) # *not* omitting X chr

    # combine results from sim_chr and compare to sim
    sim_combchr <- sim_chr[[1]]*attr(sim_chr[[1]], "n_pos")
    for(i in seq_len(n_chr(iron))[-1])
        sim_combchr <- sim_combchr + sim_chr[[i]]*attr(sim_chr[[i]], "n_pos")
    totpos <- sum(sapply(sim_chr, attr, "n_pos"))
    sim_combchr <- sim_combchr/totpos
    attr(sim_combchr, "n_pos") <- totpos
    expect_equal(sim_combchr, sim)

    # compare sim - sim_chr with sim_loco
    sim_loco_alt <- vector("list", length(probs))
    names(sim_loco_alt) <- names(probs)
    totpos <- attr(sim, "n_pos")
    for(i in seq(along=probs)) {
        npos <- attr(sim_chr[[i]], "n_pos")
        sim_loco_alt[[i]] <- (sim*totpos - sim_chr[[i]]*npos)/(totpos-npos)
        attr(sim_loco_alt[[i]], "n_pos") <- totpos-npos
    }
    expect_equal(sim_loco_alt, sim_loco)

})


test_that("calc_kinship chr & loco work when multi-core", {
    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    grid <- calc_grid(iron$gmap, step=1)
    probs_sub <- probs_to_grid(probs, grid)

    sim_chr <- calc_kinship(probs_sub, "chr")
    sim_chr_mc <- calc_kinship(probs_sub, "chr", cores=2)
    expect_equal(sim_chr_mc, sim_chr)

    sim_loco <- calc_kinship(probs_sub, "loco")
    sim_loco_mc <- calc_kinship(probs_sub, "loco", cores=2)
    expect_equal(sim_loco_mc, sim_loco)

    sim_loco <- calc_kinship(probs_sub, "loco", omit_x=FALSE)
    sim_loco_mc <- calc_kinship(probs_sub, "loco", omit_x=FALSE, cores=2)
    expect_equal(sim_loco_mc, sim_loco)


    ## RIL
    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2, map, error_prob=0.002)
    grid <- calc_grid(grav2$gmap, step=1)
    probs_sub <- probs_to_grid(probs, grid)

    sim_chr <- calc_kinship(probs_sub, "chr")
    sim_chr_mc <- calc_kinship(probs_sub, "chr", cores=2)
    expect_equal(sim_chr_mc, sim_chr)

    sim_loco <- calc_kinship(probs_sub, "loco")
    sim_loco_mc <- calc_kinship(probs_sub, "loco", cores=2)
    expect_equal(sim_loco_mc, sim_loco)

})

context("fit1 for binary traits")

test_that("fit1 for binary traits works in intercross", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(18:19,"X")]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    pheno <- setNames(as.numeric(iron$pheno[,1] > median(iron$pheno[,1])),
                      ind_ids_pheno(iron))
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    # calculate LOD scores
    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar, model="binary")

    # estimate coefficients; no covariates for X chromosome
    coef <- lapply(seq_len(length(probs)), function(i) {
        if(i==3) cov <- NULL
        else cov <- covar
        scan1coef(subset(probs, chr=names(probs)[i]), pheno, addcovar=cov, model="binary", zerosum=FALSE) })

    # fit1, no missing data
    npos <- sapply(probs, function(a) dim(a)[3])
    pmar <- c(3, 4, 12)
    out_fit1 <- lapply(seq(along=pmar),
                       function(i) {
        if(i==3) { nullcov <- Xcovar; cov <- NULL } # need Xcovar under null on X chr but no other covariates
        else { nullcov <- NULL; cov <- covar }      # sex as covariate; no additional covariates under null
        fit1(probs[[i]][,,pmar[i]], pheno, addcovar=cov, nullcovar=nullcov, model="binary", zerosum=FALSE) })

    pos <- cumsum(c(0, npos[-3])) + pmar
    # check LOD vs scan1, plus ind'l contributions to LOD
    for(i in 1:3) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1])
        expect_equal(sum(out_fit1[[i]]$ind_lod), out_fit1[[i]]$lod)
    }

    # check coefficients
    for(i in 1:3)
        expect_equal(out_fit1[[i]]$coef, coef[[i]][pmar[i],])

    # repeat the whole thing with a couple of missing phenotypes
    pheno[c(187, 244)] <- NA

    # calculate LOD scores
    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar, model="binary")

    # estimate coefficients; no covariates for X chromosome
    coef <- lapply(seq_len(length(probs)), function(i) {
        if(i==3) cov <- NULL
        else cov <- covar
        scan1coef(subset(probs, chr=names(probs)[i]), pheno, addcovar=cov, se=TRUE, model="binary", zerosum=FALSE) })

    # fit1, missing data
    out_fit1 <- lapply(seq(along=pmar),
                       function(i) {
        if(i==3) { nullcov <- Xcovar; cov <- NULL } # need Xcovar under null on X chr but no other covariates
        else { nullcov <- NULL; cov <- covar }      # sex as covariate; no additional covariates under null
        fit1(probs[[i]][,,pmar[i]], pheno, addcovar=cov, nullcovar=nullcov, se=TRUE, model="binary", zerosum=FALSE) })

    # check LOD vs scan1, plus ind'l contributions to LOD
    for(i in 1:3) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1])
        expect_equal(sum(out_fit1[[i]]$ind_lod), out_fit1[[i]]$lod)
    }

    # check coefficients
    for(i in 1:3)
        expect_equal(out_fit1[[i]]$coef, coef[[i]][pmar[i],])

    # check SEs
    for(i in 1:3)
        expect_equal(out_fit1[[i]]$SE, attr(coef[[i]], "SE")[pmar[i],])

    # direct calculations, chr 18
    glm0 <- glm(pheno ~ covar, family=binomial(link=logit))
    X <- cbind(probs[[1]][,,pmar[1]], covar)
    colnames(X) <- c("SS", "SB", "BB", "ac1")
    glm1 <- glm(pheno ~ -1 + X, family=binomial(link=logit))
    glm_lod <- (glm1$deviance - glm0$deviance)/(-2*log(10))
    p1 <- glm1$fitted
    p0 <- glm0$fitted
    y <- pheno[!is.na(pheno)]
    glm_ind_lod <- (y * log10(p1) + (1-y)*log10(1-p1)) -
        (y * log10(p0) + (1-y)*log10(1-p0))

    expect_equal(out_fit1[[1]]$lod, glm_lod)
    expect_equal(out_fit1[[1]]$ind_lod, glm_ind_lod)

    expect_equal(out_fit1[[1]]$coef, stats::setNames(glm1$coef, c("SS", "SB", "BB", "ac1")))
    expect_equal(out_fit1[[1]]$SE, stats::setNames(summary(glm1)$coef[,2], c("SS", "SB", "BB", "ac1")), tol=1e-6)

    # direct calculations, chr X
    glm0 <- glm(pheno ~ Xcovar, family=binomial(link=logit))
    X <- probs[[3]][,,pmar[3]]
    colnames(X) <- c("SS", "SB", "BS", "BB", "SY", "BY")
    glm1 <- glm(pheno ~ -1 + X, family=binomial(link=logit))
    glm_lod <- (glm1$deviance - glm0$deviance)/(-2*log(10))
    p1 <- glm1$fitted
    p0 <- glm0$fitted
    y <- pheno[!is.na(pheno)]
    glm_ind_lod <- (y * log10(p1) + (1-y)*log10(1-p1)) -
        (y * log10(p0) + (1-y)*log10(1-p0))

    expect_equal(out_fit1[[3]]$lod, glm_lod)
    expect_equal(out_fit1[[3]]$ind_lod, glm_ind_lod)
    expect_equal(out_fit1[[3]]$coef, stats::setNames(glm1$coef, c("SS", "SB", "BS", "BB", "SY", "BY")))
    expect_equal(out_fit1[[3]]$SE, stats::setNames(summary(glm1)$coef[,2], c("SS", "SB", "BS", "BB", "SY", "BY")), tol=1e-6)

})


test_that("fit1 by H-K works in riself", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    grav2 <- grav2[,4:5]
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2, map, error_prob=0.002)

    pheno <- setNames(as.numeric(grav2$pheno[,219] > 101.25), ind_ids_pheno(grav2))

    # calculate LOD scores
    out <- scan1(probs, pheno, model="binary")

    # estimate coefficients
    coef <- lapply(seq_len(length(probs)), function(i) scan1coef(subset(probs, chr=names(probs)[i]), pheno, model="binary", zerosum=FALSE))

    # fit1, no missing data
    npos <- sapply(probs, function(a) dim(a)[3])
    pmar <- c(1, 172)
    out_fit1 <- lapply(seq(along=pmar), function(i) fit1(probs[[i]][,,pmar[i]], pheno, model="binary", zerosum=FALSE))

    pos <- c(0,npos[1]) + pmar
    # check LOD vs scan1, plus ind'l contributions to LOD
    for(i in 1:2) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1])
        expect_equal(sum(out_fit1[[i]]$ind_lod), out_fit1[[i]]$lod)
    }

    # check coefficients
    for(i in 1:2)
        expect_equal(out_fit1[[i]]$coef, coef[[i]][pmar[i],])

    # repeat the whole thing with a couple of missing phenotypes
    pheno[c(24, 106)] <- NA

    # calculate LOD scores
    out <- scan1(probs, pheno, model="binary")

    # estimate coefficients
    coef <- lapply(seq_len(length(probs)), function(i) scan1coef(subset(probs, chr=names(probs)[i]), pheno, se=TRUE, model="binary", zerosum=FALSE))

    # fit1, missing data
    out_fit1 <- lapply(seq(along=pmar), function(i) fit1(probs[[i]][,,pmar[i]], pheno, se=TRUE, model="binary", zerosum=FALSE))

    # check LOD vs scan1, plus ind'l contributions to LOD
    for(i in 1:2) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1])
        expect_equal(sum(out_fit1[[i]]$ind_lod), out_fit1[[i]]$lod)
    }

    # check coefficients
    for(i in 1:2)
        expect_equal(out_fit1[[i]]$coef, coef[[i]][pmar[i],])


    # check SEs
    for(i in 1:2)
        expect_equal(out_fit1[[i]]$SE, attr(coef[[i]], "SE")[pmar[i],])

    # direct calculations, chr 18
    glm0 <- glm(pheno ~ 1, family=binomial(link=logit))
    X <- probs[[1]][,,pmar[1]]
    colnames(X) <- c("LL", "CC")
    glm1 <- glm(pheno ~ -1 + X, family=binomial(link=logit))
    glm_lod <- (glm1$deviance - glm0$deviance)/(-2*log(10))
    p1 <- glm1$fitted
    p0 <- glm0$fitted
    y <- pheno[!is.na(pheno)]
    glm_ind_lod <- (y * log10(p1) + (1-y)*log10(1-p1)) -
        (y * log10(p0) + (1-y)*log10(1-p0))

    expect_equal(out_fit1[[1]]$lod, glm_lod)
    expect_equal(out_fit1[[1]]$ind_lod, glm_ind_lod)
    expect_equal(out_fit1[[1]]$coef, stats::setNames(glm1$coef, c("LL", "CC")))
    expect_equal(out_fit1[[1]]$SE, stats::setNames(summary(glm1)$coef[,2], c("LL", "CC")), tol=1e-6)

})

test_that("fit1 works for binary traits with weights", {

    set.seed(17262911)

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(2,"X")]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, err=0.002)

    phe <- iron$pheno[,1]
    phe <- setNames(as.numeric(phe > quantile(phe, 0.7)),
                    ind_ids(iron))
    phe[c(108,142,268)] <- NA
    weights <- setNames(sample(1:10, n_ind(iron), replace=TRUE), names(phe))

    npos <- dim(probs)[3,]
    pos <- sapply(npos, sample, 1)
    pr <- list(probs[[1]][,,pos[1]],
               probs[[2]][,,pos[2]])

    out_fit1_1 <- fit1(pr[[1]], phe, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    out_fit1_2 <- fit1(pr[[2]], phe, model="binary", se=TRUE, weights=weights, zerosum=FALSE)

    # coefficients and SEs
    co2 <- scan1coef(probs[,"2"], phe, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    expect_equal(out_fit1_1$coef, co2[pos[1],])
    expect_equal(out_fit1_1$SE, attr(co2, "SE")[pos[1],])
    coX <- scan1coef(probs[,"X"], phe, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    expect_equal(out_fit1_2$coef, coX[pos[2],])
    expect_equal(out_fit1_2$SE, attr(coX, "SE")[pos[2],])

    # lod
    out <- scan1(probs, phe, model="binary", weights=weights)
    expect_equal(out_fit1_1$lod, out[pos[1]])
    expect_equal(out_fit1_2$lod, out[npos[1] + pos[2]])

    # compare to glm
    out_glm_1 <- glm(phe ~ -1 + pr[[1]], family=binomial(link=logit), weights=weights,
                     control=list(epsilon=1e-11))
    out_glm_2 <- glm(phe ~ -1 + pr[[2]], family=binomial(link=logit), weights=weights,
                     control=list(epsilon=1e-11))
    out_glm_0 <- glm(phe ~ 1, family=binomial(link=logit), weights=weights)
    sum_glm_1 <- summary(out_glm_1)
    sum_glm_2 <- summary(out_glm_2)

    # function to calc lod
    lod_glm <- function(out_glm_alt, out_glm_null)
        (out_glm_null$deviance - out_glm_alt$deviance)/(2*log(10))

    # lod scores
    expect_equivalent(out_fit1_1$lod, lod_glm(out_glm_1, out_glm_0))
    expect_equivalent(out_fit1_2$lod, lod_glm(out_glm_2, out_glm_0))

    # coefficients and SEs
    expect_equivalent(out_fit1_1$coef, out_glm_1$coef)
    expect_equivalent(out_fit1_2$coef, out_glm_2$coef)
    expect_equivalent(out_fit1_1$SE, sum_glm_1$coef[,2])
    expect_equivalent(out_fit1_2$SE, sum_glm_2$coef[,2])

    # add a covariate
    X <- setNames(rnorm(n_ind(iron)), names(phe))

    out_fit1_1 <- fit1(pr[[1]], phe, model="binary", se=TRUE, weights=weights, addcovar=X, zerosum=FALSE)
    out_fit1_2 <- fit1(pr[[2]], phe, model="binary", se=TRUE, weights=weights, addcovar=X, zerosum=FALSE)

    # coefficients and SEs
    co2 <- scan1coef(probs[,"2"], phe, model="binary", se=TRUE, weights=weights, addcovar=X, zerosum=FALSE)
    expect_equal(out_fit1_1$coef, co2[pos[1],])
    expect_equal(out_fit1_1$SE, attr(co2, "SE")[pos[1],])
    coX <- scan1coef(probs[,"X"], phe, model="binary", se=TRUE, weights=weights, addcovar=X, zerosum=FALSE)
    expect_equal(out_fit1_2$coef, coX[pos[2],])
    expect_equal(out_fit1_2$SE, attr(coX, "SE")[pos[2],])

    # lod
    out <- scan1(probs, phe, model="binary", weights=weights, addcovar=X)
    expect_equal(out_fit1_1$lod, out[pos[1]])
    expect_equal(out_fit1_2$lod, out[npos[1] + pos[2]])

    # compare to glm
    out_glm_1 <- glm(phe ~ -1 + pr[[1]] + X, family=binomial(link=logit), weights=weights,
                     control=list(epsilon=1e-11))
    out_glm_2 <- glm(phe ~ -1 + pr[[2]] + X, family=binomial(link=logit), weights=weights,
                     control=list(epsilon=1e-11))
    out_glm_0 <- glm(phe ~ X, family=binomial(link=logit), weights=weights)
    sum_glm_1 <- summary(out_glm_1)
    sum_glm_2 <- summary(out_glm_2)

    # lod scores
    expect_equivalent(out_fit1_1$lod, lod_glm(out_glm_1, out_glm_0))
    expect_equivalent(out_fit1_2$lod, lod_glm(out_glm_2, out_glm_0))

    # coefficients and SEs
    expect_equivalent(out_fit1_1$coef, out_glm_1$coef)
    expect_equivalent(out_fit1_2$coef, out_glm_2$coef)
    expect_equivalent(out_fit1_1$SE, sum_glm_1$coef[,2])
    expect_equivalent(out_fit1_2$SE, sum_glm_2$coef[,2])

    # interactive covariate, autosome only
    out_fit1_1 <- fit1(pr[[1]], phe, model="binary", se=TRUE, weights=weights, addcovar=X, intcovar=X, zerosum=FALSE)

    # coefficients and SEs
    co2 <- scan1coef(probs[,"2"], phe, model="binary", se=TRUE, weights=weights, addcovar=X, intcovar=X, zerosum=FALSE)
    expect_equal(out_fit1_1$coef, co2[pos[1],])
    expect_equal(out_fit1_1$SE, attr(co2, "SE")[pos[1],])

    # lod
    out <- scan1(probs, phe, model="binary", weights=weights, addcovar=X, intcovar=X)
    expect_equal(out_fit1_1$lod, out[pos[1]])

    # compare to glm
    out_glm_1 <- glm(phe ~ -1 + pr[[1]] + X + I(pr[[1]][,2]*X) + I(pr[[1]][,3]*X),
                     family=binomial(link=logit), weights=weights,
                     control=list(epsilon=1e-11))
    out_glm_0 <- glm(phe ~ X, family=binomial(link=logit), weights=weights)
    sum_glm_1 <- summary(out_glm_1)

    # lod scores
    expect_equivalent(out_fit1_1$lod, lod_glm(out_glm_1, out_glm_0))

    # coefficients and SEs
    expect_equivalent(out_fit1_1$coef, out_glm_1$coef)
    expect_equivalent(out_fit1_1$SE, sum_glm_1$coef[,2])

})

test_that("fit one for binary traits handles NA case", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c("2", "X")]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, err=0.002)

    phe <- iron$pheno[,1]
    phe <- setNames(as.numeric(phe > quantile(phe, 0.95)),
                    ind_ids(iron))

    suppressWarnings(out <- scan1(probs, phe, model="binary"))

    mar <- c("D2Mit304", "DXMit186")
    p1 <- pull_genoprobpos(probs, mar[1])
    p2 <- pull_genoprobpos(probs, mar[2])

    suppressWarnings(ll1a <- glm(phe ~ -1 + p1, family=binomial(link=logit))$deviance)
    suppressWarnings(ll1b <- glm(phe ~ -1 + p2, family=binomial(link=logit))$deviance)
    ll0 <- glm(phe ~ 1, family=binomial(link=logit))$deviance

    expect_equal(out[mar[1],1], (ll0-ll1a)/2/log(10), tol=1e-5)
    expect_equal(out[mar[2],1], (ll0-ll1b)/2/log(10), tol=1e-6)

    # repeat with weights
    set.seed(20180720)
    w <- setNames( runif(length(phe), 1, 5), names(phe))
    suppressWarnings(outw <- scan1(probs, phe, model="binary", weights=w))

    suppressWarnings(ll1a_w <- glm(phe ~ -1 + p1, family=binomial(link=logit), weights=w)$deviance)
    suppressWarnings(ll1b_w <- glm(phe ~ -1 + p2, family=binomial(link=logit), weights=w)$deviance)
    suppressWarnings(ll0_w <- glm(phe ~ 1, family=binomial(link=logit), weights=w)$deviance)

    expect_equal(outw[mar[1],1], (ll0_w-ll1a_w)/2/log(10), tol=1e-4)
    expect_equal(outw[mar[2],1], (ll0_w-ll1b_w)/2/log(10), tol=1e-6)

})

test_that("fit one for binary traits handles NA case with DO data", {

    skip_if(isnt_karl(), "this test only run locally")

    file <- paste0("https://raw.githubusercontent.com/rqtl/",
                   "qtl2data/main/DOex/DOex.zip")
    DOex <- read_cross2(file)
    probs <- calc_genoprob(DOex, error_prob=0.002, cores=2)
    apr <- genoprob_to_alleleprob(probs)

    phe <- setNames((DOex$pheno[,1] > quantile(DOex$pheno[,1], 0.95, na.rm=TRUE))*1, rownames(DOex$pheno))

    suppressWarnings(out <- scan1(apr, phe, model="binary"))

    mar <- c("backupUNC021331957", "UNC020459944")
    p1 <- pull_genoprobpos(apr, mar[1])
    p2 <- pull_genoprobpos(apr, mar[2])

    suppressWarnings(ll1a <- glm(phe ~ -1 + p1, family=binomial(link=logit))$deviance)
    suppressWarnings(ll1b <- glm(phe ~ -1 + p2, family=binomial(link=logit))$deviance)
    ll0 <- glm(phe ~ 1, family=binomial(link=logit))$deviance

    expect_equal(out[mar[1],1], (ll0-ll1a)/2/log(10))
    expect_equal(out[mar[2],1], (ll0-ll1b)/2/log(10))


    # repeat with weights
    set.seed(20180720)
    w <- setNames( runif(length(phe), 1, 5), names(phe))
    suppressWarnings(outw <- scan1(apr, phe, model="binary", weights=w))

    suppressWarnings(ll1a_w <- glm(phe ~ -1 + p1, family=binomial(link=logit), weights=w)$deviance)
    suppressWarnings(ll1b_w <- glm(phe ~ -1 + p2, family=binomial(link=logit), weights=w)$deviance)
    suppressWarnings(ll0_w <- glm(phe ~ 1, family=binomial(link=logit), weights=w)$deviance)

    expect_equal(outw[mar[1],1], (ll0_w-ll1a_w)/2/log(10))
    expect_equal(outw[mar[2],1], (ll0_w-ll1b_w)/2/log(10), tol=1e-6)

})


test_that("fit1 for binary trait gives same answer as glm(), including variance matrix", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,18]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    pheno <- setNames(as.numeric(iron$pheno[,1] > quantile(iron$pheno[,1], 0.6)), rownames(iron$pheno))
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)

    # max positions
    out <- scan1(probs, pheno, addcovar=covar)
    mx <- max(out, map)

    # pull prob
    pr <- pull_genoprobpos(probs, map, chr=mx$chr, pos=mx$pos)

    # estimate coefficients; no covariates for X chromosome
    out_glm <- glm(pheno ~ -1 + pr + covar, family=binomial(link=logit))
    sum_glm <- summary(out_glm)
    coef_glm <- out_glm$coef
    se_glm <- sum_glm$coef[,2]
    var_glm <- sum_glm$cov.scaled

    expect_equal(sqrt(diag(var_glm)), se_glm)

    out_fit1_se <- fit1(pr, pheno, addcovar=covar, model="binary", zerosum=FALSE, se=TRUE, var=FALSE)
    out_fit1_var <- fit1(pr, pheno, addcovar=covar, model="binary", zerosum=FALSE, var=TRUE)

    names(coef_glm) <- names(se_glm) <- rownames(var_glm) <- colnames(var_glm) <- names(out_fit1_se$coef)

    expect_equal(out_fit1_se$coef,  coef_glm)
    expect_equal(out_fit1_se$SE,    se_glm, tol=1e-5)
    expect_equal(out_fit1_var$coef, coef_glm)
    expect_equal(out_fit1_var$SE,   se_glm, tol=1e-5)
    expect_equal(out_fit1_var$var,  var_glm, tol=1e-5)

    # zerosum version
    A <- rbind(c(2/3, -1/3, -1/3, 0),
               c(-1/3, 2/3, -1/3, 0),
               c(-1/3,-1/3,  2/3, 0),
               c(  0,   0,    0,  1),
               c(1/3,  1/3,  1/3, 0))
    coef_glm <- setNames(as.numeric(A %*% coef_glm), c(names(coef_glm), "intercept"))
    var_glm <- A %*% var_glm %*% t(A)
    se_glm <- sqrt(diag(var_glm))
    names(se_glm) <- rownames(var_glm) <- colnames(var_glm) <- names(coef_glm)

    out_fit1_se <- fit1(pr, pheno, addcovar=covar, model="binary", zerosum=TRUE, se=TRUE, var=FALSE)
    out_fit1_var <- fit1(pr, pheno, addcovar=covar, model="binary", zerosum=TRUE, var=TRUE)

    expect_equal(out_fit1_se$coef,  coef_glm)
    expect_equal(out_fit1_se$SE,    se_glm, tol=1e-5)
    expect_equal(out_fit1_var$coef, coef_glm)
    expect_equal(out_fit1_var$SE,   se_glm, tol=1e-5)
    expect_equal(out_fit1_var$var,  var_glm, tol=1e-5)

})

context("snp association scan")

test_that("scan1snps works", {

    skip_if(isnt_karl(), "this test only run locally")

    RNGkind("Mersenne-Twister") # make sure we're using the standard RNG
    set.seed(20180727)

    # load example data and calculate genotype probabilities
    file <- paste0("https://raw.githubusercontent.com/rqtl/",
                   "qtl2data/main/DOex/DOex.zip")
    DOex <- read_cross2(file)
    probs <- calc_genoprob(DOex[1:20,"2"], error_prob=0.002)

    snpdb_file <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
    queryf <- create_variant_query_func(snpdb_file)

    expected <- list(lod = structure(c(0.0792733888900488, 0.692764338278575, 0.427585764384135,  0.794318319047203,
                                       0.252436101716556,  1.10965528916839,  0.0052242742952302, 0.044886580360961,
                                       0.77913769683386,   0.723589209750504, 0.130713797731086,  0.150391046233995,
                                       0.254799450048782,  0.00724275232762306, 0.19073862233745, 0.27168521752353,
                                       0.0404789019635032, 0.171541072518635, 0.369411627351868),
                                     dim = c(19L, 1L),
                                     dimnames = list(c("rs259104594", "rs49962811", "rs250912493", "rs49310503",
                                                       "rs27363853", "rs27363851",  "rs33556222",
                                                       "rs387021772;rs258896118;rs220544684",
                                                       "rs227317919", "rs33673239", "rs33438111", "rs27363804",
                                                       "rs29764604", "rs241221617", "rs27413286", "rs27395539",
                                                       "rs27395529", "rs27395515", "2:97277447_ATTT/ATT"),
                                                      "OF_immobile_pct"),
                                     sample_size = c(OF_immobile_pct = 20L), class = c("scan1", "matrix")),
                     snpinfo = structure(list(snp = c("rs259104594", "rs49962811", "rs250912493", "rs49310503",
                                                         "rs27363853", "rs27363851", "rs33556222",
                                                         "rs387021772;rs258896118;rs220544684", "rs227317919",
                                                         "rs33673239", "rs33438111", "rs27363804", "rs29764604",
                                                         "rs241221617", "rs27413286", "rs27395539", "rs27395529",
                                                         "rs27395515", "2:97277447_ATTT/ATT"),
                                              chr = rep("2", 19),
                                              pos = c(97.20006, 97.200082, 97.200598, 97.200765, 97.2008,
                                                      97.200955, 97.201002, 97.205127, 97.20862, 97.209716,
                                                      97.210417, 97.215528, 97.219902, 97.221925, 97.232949,
                                                      97.248752, 97.24921, 97.249953, 97.277447),
                                              alleles = c("T|A", "T|C", "G|A", "A|C", "T|A", "G|A", "T|G",
                                                          "CT|CTTTT/CTT/C", "T|C", "C|T", "C|T/G", "C|A/T",
                                                          "T|C", "A|T", "A|G", "T|C", "T|C", "G|A",
                                                          "ATT|ATTT/AT/A"),
                                              sdp = c(32L, 8L, 64L, 96L, 12L, 4L, 108L, 23L, 128L, 17L,
                                                      40L, 72L, 2L, 44L, 100L, 36L, 68L, 76L, 170L),
                                              ensembl_gene = c(rep("ENSMUSG00000050587", 14),
                                                               "ENSMUSG00000088633,ENSMUSG00000050587",
                                                               rep("ENSMUSG00000050587", 4)),
                                              consequence = c(rep("ENSMUSG00000050587:intron_variant", 14),
                                     "ENSMUSG00000088633:upstream_gene_variant,ENSMUSG00000050587:intron_variant",
                                                              rep("ENSMUSG00000050587:intron_variant", 4)),
                                              A_J = c(1, 1, 1, 1, 1, 1, 1, 4, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1),
                                              C57BL_6J = c(1, 1, 1, 1, 1, 1, 1, 3, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 2),
                                              `129S1_SvImJ` = c(1, 1, 1, 1, 2, 2, 2, 2, 1, 1, 1, 1, 1, 2,2,2,2,2,1),
                                              NOD_ShiLtJ = c(1, 2, 1, 1, 2,1,2,1, 1, 1, 2, 2, 1, 2, 1, 1, 1, 2, 3),
                                              NZO_HlLtJ = c(1, 1, 1, 1, 1, 1, 1, 4, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1),
                                              CAST_EiJ = c(2, 1, 1, 2, 1, 1, 2, 1, 1, 1, 3, 1, 1, 2, 2, 2, 1, 1, 4),
                                              PWK_PhJ = c(1, 1, 2, 2, 1, 1, 2, 1, 1, 1, 1, 3, 1, 1, 2, 1, 2, 2, 1),
                                              WSB_EiJ = c(1, 1, 1, 1, 1, 1, 1, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 3),
                                              type = c("snp", "snp", "snp", "snp", "snp", "snp", "snp", "indel",
                                                       "snp", "snp", "snp", "snp", "snp", "snp", "snp", "snp",
                                                       "snp", "snp", "indel"),
                                              index = 1:19,
                                              interval = rep(65L, 19),
                                              on_map = rep(FALSE, 19)),
                                              row.names = c(1L, 2L, 14L, 22L, 26L, 34L, 36L, 152L, 228L, 256L,
                                                            285L, 448L, 543L, 569L, 780L, 950L, 963L, 982L, 1324L),
                                              class = "data.frame"))

    # using query function for a defined region
    out <- scan1snps(probs, DOex$pmap, DOex$pheno, query_func=queryf, chr=2, start=97.2, end=97.3)
    expect_equal(out, expected)

    # if probs and map don't conform, should get a warning
    ## (one fewer marker in map)
    junk_map <- DOex$pmap
    junk_map[[1]] <- junk_map[[1]][-1]
    expect_warning( scan1snps(probs, junk_map, DOex$pheno, query_func=queryf, chr=2,
                              start=97.2, end=97.3) )
    ## (one fewere marker in probs)
    junk_probs <- probs
    junk_probs[["2"]] <- junk_probs[["2"]][,,-1]
    expect_warning( scan1snps(junk_probs, DOex$pmap, DOex$pheno, query_func=queryf, chr=2,
                              start=97.2, end=97.3) )
    ## markers out of order gives an error though
    junk_map <- DOex$pmap
    names(junk_map[["2"]])[1:5] <- names(junk_map[["2"]])[5:1]
    expect_error( scan1snps(probs, junk_map, DOex$pheno, query_func=queryf, chr=2,
                            start=97.2, end=97.3) )


    # using a pre-defined table of snps
    snpinfo <- queryf(2, 97.2, 97.3)
    out2 <- scan1snps(probs, DOex$pmap, DOex$pheno, snpinfo=snpinfo)
    expect_equal(out, out2)

    # using a pre-defined table of snps
    out3 <- scan1snps(probs, DOex$pmap, DOex$pheno, query_func=queryf)
    expect_equal(dim(out3$lod), c(49,1))
    expect_equal(dim(out3$snpinfo), c(49,19))

    # same, keeping all snps
    out3 <- scan1snps(probs, DOex$pmap, DOex$pheno, query_func=queryf, keep_all_snps=TRUE)
    expect_equal(dim(out3$lod), c(49,1))
    expect_equal(dim(out3$snpinfo), c(13836,19))

    # with weights, vs lm()
    w <- setNames(runif(n_ind(DOex), 0, 10), ind_ids(DOex))
    outw <- scan1snps(probs, DOex$pmap, DOex$pheno, query_func=queryf, chr=2, start=97.2, end=97.3,
                      weights=w)

    snpinfo <- queryf(2, 97.2, 97.3)
    snpinfo <- index_snps(DOex$pmap, snpinfo)
    snp_probs <- genoprob_to_snpprob(probs, snpinfo)

    p <- pull_genoprobpos(snp_probs, "rs227317919")

    y <- DOex$pheno[rownames(p),,drop=FALSE]
    w <- w[rownames(p)]
    out_lm <- lm(y ~ -1 + p, weights=w)
    out_lm0 <- lm(y ~ 1, weights=w)

    expect_equal(outw$lod["rs227317919", 1], nrow(p)/2*log10(sum(w*out_lm0$resid^2)/sum(w*out_lm$resid^2)))

    # binary trait
    binphe <- setNames((DOex$pheno > quantile(DOex$pheno, 0.8, na.rm=TRUE))*1, rownames(DOex$pheno))
    biny <- binphe[rownames(p)]

    outbin <- scan1snps(probs, DOex$pmap, binphe, query_func=queryf, chr=2, start=97.2, end=97.3,
                        model="binary")

    out_glm <- glm(biny ~ -1 + p, family=binomial(link=logit))
    out_glm0 <- glm(biny ~ 1, family=binomial(link=logit))
    expect_equal(outbin$lod["rs227317919", 1], (out_glm0$deviance - out_glm$deviance)/2/log(10))

    # binary trait with weights
    w <- ceiling(w) # avoid warning from glm() about weights not being integers
    outbinw <- scan1snps(probs, DOex$pmap, binphe, query_func=queryf,
                         chr=2, start=97.2, end=97.3, model="binary",
                         weights=w, eta_max=25, maxit=1000, tol=1e-4)

    out_glm <- glm(biny ~ -1 + p, family=binomial(link=logit), weights=w)
    out_glm0 <- glm(biny ~ 1, family=binomial(link=logit), weights=w)

    expect_equal(outbinw$lod["rs227317919", 1],
                 (out_glm0$deviance - out_glm$deviance)/2/log(10),
                 tolerance=5e-5)

})

context("fit1")

test_that("fit1 by H-K works in intercross", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(18:19,"X")]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    pheno <- iron$pheno[,1]
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    # calculate LOD scores
    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

    # estimate coefficients; no covariates for X chromosome
    coef <- lapply(seq_len(length(probs)), function(i) {
        if(i==3) cov <- NULL
        else cov <- covar
        scan1coef(subset(probs, chr=names(probs)[i]), pheno, addcovar=cov, zerosum=FALSE) })

    # fit1, no missing data
    npos <- sapply(probs, function(a) dim(a)[3])
    pmar <- c(3, 4, 12)
    out_fit1 <- lapply(seq(along=pmar),
                       function(i) {
        if(i==3) { nullcov <- Xcovar; cov <- NULL } # need Xcovar under null on X chr but no other covariates
        else { nullcov <- NULL; cov <- covar }      # sex as covariate; no additional covariates under null
        fit1(probs[[i]][,,pmar[i]], pheno, addcovar=cov, nullcovar=nullcov, zerosum=FALSE) })

    pos <- cumsum(c(0, npos[-3])) + pmar
    # check LOD vs scan1, plus ind'l contributions to LOD
    for(i in 1:3) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1])
        expect_equal(sum(out_fit1[[i]]$ind_lod), out_fit1[[i]]$lod)
    }

    # check coefficients
    for(i in 1:3)
        expect_equal(out_fit1[[i]]$coef, coef[[i]][pmar[i],])

    # repeat the whole thing with a couple of missing phenotypes
    pheno[c(187, 244)] <- NA

    # calculate LOD scores
    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

    # estimate coefficients; no covariates for X chromosome
    coef <- lapply(seq_len(length(probs)), function(i) {
        if(i==3) cov <- NULL
        else cov <- covar
        scan1coef(subset(probs, chr=names(probs)[i]), pheno, addcovar=cov, se=TRUE, zerosum=FALSE) })

    # fit1, missing data
    out_fit1 <- lapply(seq(along=pmar),
                       function(i) {
        if(i==3) { nullcov <- Xcovar; cov <- NULL } # need Xcovar under null on X chr but no other covariates
        else { nullcov <- NULL; cov <- covar }      # sex as covariate; no additional covariates under null
        fit1(probs[[i]][,,pmar[i]], pheno, addcovar=cov, nullcovar=nullcov, se=TRUE, zerosum=FALSE) })

    # check LOD vs scan1, plus ind'l contributions to LOD
    for(i in 1:3) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1])
        expect_equal(sum(out_fit1[[i]]$ind_lod), out_fit1[[i]]$lod)
    }

    # check coefficients
    for(i in 1:3)
        expect_equal(out_fit1[[i]]$coef, coef[[i]][pmar[i],])

    # check SEs
    for(i in 1:3)
        expect_equal(out_fit1[[i]]$SE, attr(coef[[i]], "SE")[pmar[i],])

    # direct calculations, chr 18
    lm0 <- lm(pheno ~ covar)
    X <- cbind(probs[[1]][,,pmar[1]], covar)
    colnames(X) <- c("SS", "SB", "BB", "ac1")
    lm1 <- lm(pheno ~ -1 + X)
    n <- sum(!is.na(pheno))
    sigsq0 <- sum(lm0$resid^2)/n
    sigsq1 <- sum(lm1$resid^2)/n
    expect_equal(out_fit1[[1]]$lod, n/2*log10(sum(lm0$resid^2)/sum(lm1$resid^2)))
    expect_equal(out_fit1[[1]]$ind_lod, (dnorm(lm1$resid,0,sqrt(sigsq1), TRUE) - dnorm(lm0$resid,0,sqrt(sigsq0),TRUE))/log(10))
    expect_equal(out_fit1[[1]]$coef, stats::setNames(lm1$coef, c("SS", "SB", "BB", "ac1")))
    expect_equal(out_fit1[[1]]$SE, stats::setNames(summary(lm1)$coef[,2], c("SS", "SB", "BB", "ac1")))

    # direct calculations, chr X
    lm0 <- lm(pheno ~ Xcovar)
    X <- probs[[3]][,,pmar[3]]
    colnames(X) <- c("SS", "SB", "BS", "BB", "SY", "BY")
    lm1 <- lm(pheno ~ -1 + X)
    n <- sum(!is.na(pheno))
    sigsq0 <- sum(lm0$resid^2)/n
    sigsq1 <- sum(lm1$resid^2)/n
    expect_equal(out_fit1[[3]]$lod, n/2*log10(sum(lm0$resid^2)/sum(lm1$resid^2)))
    expect_equal(out_fit1[[3]]$ind_lod, (dnorm(lm1$resid,0,sqrt(sigsq1), TRUE) - dnorm(lm0$resid,0,sqrt(sigsq0),TRUE))/log(10))
    expect_equal(out_fit1[[3]]$coef, stats::setNames(lm1$coef, c("SS", "SB", "BS", "BB", "SY", "BY")))
    expect_equal(out_fit1[[3]]$SE, stats::setNames(summary(lm1)$coef[,2], c("SS", "SB", "BS", "BB", "SY", "BY")))

})


test_that("fit1 by H-K works in intercross, with weights", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(18:19,"X")]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    pheno <- iron$pheno[,1]
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    w <- setNames(runif(n_ind(iron), 1, 10), ind_ids(iron))

    # calculate LOD scores
    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar, weights=w)

    # estimate coefficients; no covariates for X chromosome
    coef <- lapply(seq_len(length(probs)), function(i) {
        if(i==3) cov <- NULL
        else cov <- covar
        scan1coef(subset(probs, chr=names(probs)[i]), pheno, addcovar=cov, weights=w, zerosum=FALSE) })

    # fit1, no missing data
    npos <- sapply(probs, function(a) dim(a)[3])
    pmar <- c(3, 4, 12)
    out_fit1 <- lapply(seq(along=pmar),
                       function(i) {
        if(i==3) { nullcov <- Xcovar; cov <- NULL } # need Xcovar under null on X chr but no other covariates
        else { nullcov <- NULL; cov <- covar }      # sex as covariate; no additional covariates under null
        fit1(probs[[i]][,,pmar[i]], pheno, addcovar=cov, nullcovar=nullcov, weights=w, zerosum=FALSE) })

    pos <- cumsum(c(0, npos[-3])) + pmar
    # check LOD vs scan1, plus ind'l contributions to LOD
    for(i in 1:3) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1])
        expect_equal(sum(out_fit1[[i]]$ind_lod), out_fit1[[i]]$lod)
    }

    # check coefficients
    for(i in 1:3)
        expect_equal(out_fit1[[i]]$coef, coef[[i]][pmar[i],])

    # repeat the whole thing with a couple of missing phenotypes
    pheno[c(187, 244)] <- NA

    # calculate LOD scores
    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar, weights=w)

    # estimate coefficients; no covariates for X chromosome
    coef <- lapply(seq_len(length(probs)), function(i) {
        if(i==3) cov <- NULL
        else cov <- covar
        scan1coef(subset(probs, chr=names(probs)[i]), pheno, addcovar=cov, se=TRUE, weights=w, zerosum=FALSE) })

    # fit1, missing data
    out_fit1 <- lapply(seq(along=pmar),
                       function(i) {
        if(i==3) { nullcov <- Xcovar; cov <- NULL } # need Xcovar under null on X chr but no other covariates
        else { nullcov <- NULL; cov <- covar }      # sex as covariate; no additional covariates under null
        fit1(probs[[i]][,,pmar[i]], pheno, addcovar=cov, nullcovar=nullcov, se=TRUE, weights=w, zerosum=FALSE) })

    # check LOD vs scan1, plus ind'l contributions to LOD
    for(i in 1:3) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1])
        expect_equal(sum(out_fit1[[i]]$ind_lod), out_fit1[[i]]$lod)
    }

    # check coefficients
    for(i in 1:3)
        expect_equal(out_fit1[[i]]$coef, coef[[i]][pmar[i],])

    # check SEs
    for(i in 1:3)
        expect_equal(out_fit1[[i]]$SE, attr(coef[[i]], "SE")[pmar[i],])

})


test_that("fit1 by H-K works in riself", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    grav2 <- grav2[,4:5]
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2, map, error_prob=0.002)

    pheno <- grav2$pheno[,219]

    # calculate LOD scores
    out <- scan1(probs, pheno)

    # estimate coefficients
    coef <- lapply(seq_len(length(probs)), function(i) scan1coef(subset(probs, chr=names(probs)[i]), pheno, zerosum=FALSE))

    # fit1, no missing data
    npos <- sapply(probs, function(a) dim(a)[3])
    pmar <- c(1, 172)
    out_fit1 <- lapply(seq(along=pmar), function(i) fit1(probs[[i]][,,pmar[i]], pheno, zerosum=FALSE))

    pos <- c(0,npos[1]) + pmar
    # check LOD vs scan1, plus ind'l contributions to LOD
    for(i in 1:2) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1])
        expect_equal(sum(out_fit1[[i]]$ind_lod), out_fit1[[i]]$lod)
    }

    # check coefficients
    for(i in 1:2)
        expect_equal(out_fit1[[i]]$coef, coef[[i]][pmar[i],])

    # repeat the whole thing with a couple of missing phenotypes
    pheno[c(24, 106)] <- NA

    # calculate LOD scores
    out <- scan1(probs, pheno)

    # estimate coefficients
    coef <- lapply(seq_len(length(probs)), function(i) scan1coef(subset(probs, chr=names(probs)[i]), pheno, se=TRUE, zerosum=FALSE))

    # fit1, missing data
    out_fit1 <- lapply(seq(along=pmar), function(i) fit1(probs[[i]][,,pmar[i]], pheno, se=TRUE, zerosum=FALSE))

    # check LOD vs scan1, plus ind'l contributions to LOD
    for(i in 1:2) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1])
        expect_equal(sum(out_fit1[[i]]$ind_lod), out_fit1[[i]]$lod)
    }

    # check coefficients
    for(i in 1:2)
        expect_equal(out_fit1[[i]]$coef, coef[[i]][pmar[i],])


    # check SEs
    for(i in 1:2)
        expect_equal(out_fit1[[i]]$SE, attr(coef[[i]], "SE")[pmar[i],])

    # direct calculations, chr 18
    lm0 <- lm(pheno ~ 1)
    X <- probs[[1]][,,pmar[1]]
    colnames(X) <- c("LL", "CC")
    lm1 <- lm(pheno ~ -1 + X)
    n <- sum(!is.na(pheno))
    sigsq0 <- sum(lm0$resid^2)/n
    sigsq1 <- sum(lm1$resid^2)/n
    expect_equal(out_fit1[[1]]$lod, n/2*log10(sum(lm0$resid^2)/sum(lm1$resid^2)))
    expect_equal(out_fit1[[1]]$ind_lod, (dnorm(lm1$resid,0,sqrt(sigsq1), TRUE) - dnorm(lm0$resid,0,sqrt(sigsq0),TRUE))/log(10))
    expect_equal(out_fit1[[1]]$coef, stats::setNames(lm1$coef, c("LL", "CC")))
    expect_equal(out_fit1[[1]]$SE, stats::setNames(summary(lm1)$coef[,2], c("LL", "CC")))

})


test_that("fit1 by LMM works in intercross", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs)
    kinship_loco <- calc_kinship(probs, "loco")
    probs <- probs[,c(7,19,"X")]
    kinship_loco <- kinship_loco[c(7,19,"X")]

    pheno <- iron$pheno[,1]
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    # calculate LOD scores
    out <- scan1(probs, pheno, kinship, addcovar=covar, Xcovar=Xcovar)
    out_loco <- scan1(probs, pheno, kinship_loco, addcovar=covar, Xcovar=Xcovar)

    # estimate coefficients (with SEs)
    coef <- lapply(seq_len(length(probs)), function(i) {
        if(i==3) { cov <- NULL; nullcov <- Xcovar }
        else { cov <- covar; nullcov <- NULL }
        scan1coef(subset(probs, chr=names(probs)[i]), pheno, kinship,
                  addcovar=cov, nullcovar=nullcov, se=TRUE, zerosum=FALSE) })

    coef_loco <- lapply(seq_len(length(probs)), function(i) {
        if(i==3) { cov <- NULL; nullcov <- Xcovar }
        else { cov <- covar; nullcov <- NULL }
        scan1coef(subset(probs, chr=names(probs)[i]), pheno, kinship_loco[i],
                  addcovar=cov, nullcovar=nullcov, se=TRUE, zerosum=FALSE) })

    names(coef) <- names(coef_loco) <- names(kinship_loco)

    # positions to test fit1()
    chr <- rep(c(7,19,"X"), each=4)
    n_pos <- sapply(probs, function(a) dim(a)[3])
    index <- c(1,6,7,58, 1,15,31,36,  1,20,27,30)
    pos <- index + rep(cumsum(c(0,n_pos)), c(4,4,4,0))

    # fit1, no missing data
    out_fit1 <- lapply(seq(along=chr),
                       function(i) {
        if(chr[i]=="X") { nullcov <- Xcovar; cov <- NULL } # need Xcovar under null on X chr but no other covariates
        else { nullcov <- NULL; cov <- covar }      # sex as covariate; no additional covariates under null
        fit1(probs[[chr[i]]][,,index[i]], pheno, kinship, addcovar=cov, nullcovar=nullcov, se=TRUE, zerosum=FALSE) })

    out_fit1_loco <- lapply(seq(along=chr),
                       function(i) {
        if(chr[i]=="X") { nullcov <- Xcovar; cov <- NULL } # need Xcovar under null on X chr but no other covariates
        else { nullcov <- NULL; cov <- covar }      # sex as covariate; no additional covariates under null
        fit1(probs[[chr[i]]][,,index[i]], pheno, kinship_loco[[chr[i]]], addcovar=cov, nullcovar=nullcov, se=TRUE, zerosum=FALSE) })

    # check LOD vs scan1
    for(i in seq(along=out_fit1)) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1], tol=1e-6)
    }
    # same with _loco version
    for(i in seq(along=out_fit1)) {
        expect_equal(out_fit1_loco[[i]]$lod, out_loco[pos[i],1], tol=1e-6)
    }

    # check coefficients
    for(i in seq(along=out_fit1))
        expect_equal(out_fit1[[i]]$coef, coef[[chr[i]]][index[i],])
    # same, _loco version
    for(i in seq(along=out_fit1))
        expect_equal(out_fit1_loco[[i]]$coef, coef_loco[[chr[i]]][index[i],])

    # check SEs
    for(i in seq(along=out_fit1))
        expect_equal(out_fit1[[i]]$SE, attr(coef[[chr[i]]], "SE")[index[i],])
    # same, _loco version
    for(i in seq(along=out_fit1))
        expect_equal(out_fit1_loco[[i]]$SE, attr(coef_loco[[chr[i]]], "SE")[index[i],])

})


test_that("fit1 by LMM works in intercross, with weights", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs)
    kinship_loco <- calc_kinship(probs, "loco")
    probs <- probs[,c(7,19,"X")]
    kinship_loco <- kinship_loco[c(7,19,"X")]

    set.seed(2661343)
    w <- setNames(runif(n_ind(iron), 1, 8), ind_ids(iron))

    pheno <- iron$pheno[,1]
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    # calculate LOD scores
    out <- scan1(probs, pheno, kinship, addcovar=covar, Xcovar=Xcovar, weights=w)
    out_loco <- scan1(probs, pheno, kinship_loco, addcovar=covar, Xcovar=Xcovar, weights=w)

    # estimate coefficients (with SEs)
    coef <- lapply(seq_len(length(probs)), function(i) {
        if(i==3) { cov <- NULL; nullcov <- Xcovar }
        else { cov <- covar; nullcov <- NULL }
        scan1coef(subset(probs, chr=names(probs)[i]), pheno, kinship,
                  addcovar=cov, nullcovar=nullcov, se=TRUE, weights=w, zerosum=FALSE) })

    coef_loco <- lapply(seq_len(length(probs)), function(i) {
        if(i==3) { cov <- NULL; nullcov <- Xcovar }
        else { cov <- covar; nullcov <- NULL }
        scan1coef(subset(probs, chr=names(probs)[i]), pheno, kinship_loco[i],
                  addcovar=cov, nullcovar=nullcov, se=TRUE, weights=w, zerosum=FALSE) })

    names(coef) <- names(coef_loco) <- names(kinship_loco)

    # positions to test fit1()
    chr <- rep(c(7,19,"X"), each=4)
    n_pos <- sapply(probs, function(a) dim(a)[3])
    index <- c(1,6,7,58, 1,15,31,36,  1,20,27,30)
    pos <- index + rep(cumsum(c(0,n_pos)), c(4,4,4,0))

    # fit1, no missing data
    out_fit1 <- lapply(seq(along=chr),
                       function(i) {
        if(chr[i]=="X") { nullcov <- Xcovar; cov <- NULL } # need Xcovar under null on X chr but no other covariates
        else { nullcov <- NULL; cov <- covar }      # sex as covariate; no additional covariates under null
        fit1(probs[[chr[i]]][,,index[i]], pheno, kinship, addcovar=cov, nullcovar=nullcov, se=TRUE, weights=w, zerosum=FALSE) })

    out_fit1_loco <- lapply(seq(along=chr),
                       function(i) {
        if(chr[i]=="X") { nullcov <- Xcovar; cov <- NULL } # need Xcovar under null on X chr but no other covariates
        else { nullcov <- NULL; cov <- covar }      # sex as covariate; no additional covariates under null
        fit1(probs[[chr[i]]][,,index[i]], pheno, kinship_loco[[chr[i]]], addcovar=cov, nullcovar=nullcov,
             se=TRUE, weights=w, zerosum=FALSE) })

    # check LOD vs scan1
    for(i in seq(along=out_fit1)) {
        expect_equal(out_fit1[[i]]$lod, out[pos[i],1], tol=1e-6)
    }
    # same with _loco version
    for(i in seq(along=out_fit1)) {
        expect_equal(out_fit1_loco[[i]]$lod, out_loco[pos[i],1], tol=1e-6)
    }

    # check coefficients
    for(i in seq(along=out_fit1)) {
        expect_equal(out_fit1[[i]]$coef, coef[[chr[i]]][index[i],])
    }
    # same, _loco version
    for(i in seq(along=out_fit1)) {
        expect_equal(out_fit1_loco[[i]]$coef, coef_loco[[chr[i]]][index[i],])
    }

    # check SEs
    for(i in seq(along=out_fit1)) {
        expect_equal(out_fit1[[i]]$SE, attr(coef[[chr[i]]], "SE")[index[i],])
    }
    # same, _loco version
    for(i in seq(along=out_fit1)) {
        expect_equal(out_fit1_loco[[i]]$SE, attr(coef_loco[[chr[i]]], "SE")[index[i],])
    }

})




test_that("fit1 handles contrasts properly in a backcross", {

    library(qtl)
    data(fake.bc)
    fake.bc <- subset(fake.bc, chr=c(2,5))

    fake_bc <- convert2cross2(fake.bc)

    fake.bc <- calc.genoprob(fake.bc, error.prob=0.002, map.function="c-f")
    pr <- calc_genoprob(fake_bc, error_prob=0.002, map_function="c-f")

    # fit single-QTL model at two positions
    qtl <- makeqtl(fake.bc, c(2,5), pos=c(44.8, 9.8), what="prob")
    out.fq1 <- fitqtl(fake.bc, qtl=qtl, formula=y~Q1, get.ests=TRUE, method="hk")
    co1 <- summary(out.fq1)$ests
    out.fq2 <- fitqtl(fake.bc, qtl=qtl, formula=y~Q2, get.ests=TRUE, method="hk")
    co2 <- summary(out.fq2)$ests

    contrasts <- cbind(mu=c(1,1), a=c(-0.5,0.5))
    out_fit1a <- fit1(pull_genoprobpos(pr, "D2M336"), fake_bc$pheno[,1],
                      contrasts=contrasts)
    expect_equivalent(co1[,"est"], out_fit1a$coef)
    expect_equivalent(co1[,"SE"], out_fit1a$SE)

    out_fit1b <- fit1(pull_genoprobpos(pr, "D5M394"), fake_bc$pheno[,1],
                      contrasts=contrasts)
    expect_equivalent(co2[,"est"], out_fit1b$coef)
    expect_equivalent(co2[,"SE"], out_fit1b$SE)

    # compare to scan1coef()
    ests_c2 <- scan1coef(pr[,"2"], fake_bc$pheno[,1], contrasts=contrasts, se=TRUE)
    expect_equal(ests_c2["D2M336",], out_fit1a$coef)
    expect_equal(attr(ests_c2, "SE")["D2M336",], out_fit1a$SE)

    ests_c5 <- scan1coef(pr[,"5"], fake_bc$pheno[,1], contrasts=contrasts, se=TRUE)
    expect_equal(ests_c5["D5M394",], out_fit1b$coef)
    expect_equal(attr(ests_c5, "SE")["D5M394",], out_fit1b$SE)

})

test_that("fit1 handles contrasts properly in an intercross", {

    library(qtl)
    data(fake.f2)
    fake.f2 <- subset(fake.f2, chr=c(1,13))

    fake_f2 <- convert2cross2(fake.f2)

    fake.f2 <- calc.genoprob(fake.f2, error.prob=0.002, map.function="c-f")
    pr <- calc_genoprob(fake_f2, error_prob=0.002, map_function="c-f")

    # fit single-QTL model at two positions
    qtl <- makeqtl(fake.f2, c(1,13), pos=c(37.1, 24.0), what="prob")
    out.fq1 <- fitqtl(fake.f2, qtl=qtl, formula=y~Q1, get.ests=TRUE, method="hk")
    co1 <- summary(out.fq1)$ests
    out.fq2 <- fitqtl(fake.f2, qtl=qtl, formula=y~Q2, get.ests=TRUE, method="hk")
    co2 <- summary(out.fq2)$ests

    # R/qtl1 and R/qtl2 give different answers for the intercept,
    #    so we'll just look at the additive and dominance effects
    contrasts <- cbind(mu=c(1,1,1), a=c(-1,0,1), d=c(0,1,0))
    out_fit1a <- fit1(pull_genoprobpos(pr, "D1M437"), fake_f2$pheno[,1],
                      contrasts=contrasts)
    expect_equivalent(co1[-1,"est"], out_fit1a$coef[-1])
    expect_equivalent(co1[-1,"SE"], out_fit1a$SE[-1])

    out_fit1b <- fit1(pull_genoprobpos(pr, "D13M254"), fake_f2$pheno[,1],
                      contrasts=contrasts)
    expect_equivalent(co2[-1,"est"], out_fit1b$coef[-1])
    expect_equivalent(co2[-1,"SE"], out_fit1b$SE[-1])

    # compare to scan1coef()
    ests_c1 <- scan1coef(pr[,"1"], fake_f2$pheno[,1], contrasts=contrasts, se=TRUE)
    expect_equal(ests_c1["D1M437",], out_fit1a$coef)
    expect_equal(attr(ests_c1, "SE")["D1M437",], out_fit1a$SE)

    ests_c13 <- scan1coef(pr[,"13"], fake_f2$pheno[,1], contrasts=contrasts, se=TRUE)
    expect_equal(ests_c13["D13M254",], out_fit1b$coef)
    expect_equal(attr(ests_c13, "SE")["D13M254",], out_fit1b$SE)

})

test_that("fit1 works with blup=TRUE", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    Xcovar <- get_x_covar(iron)
    kinship <- calc_kinship(probs, "loco")

    # pull out a position
    marker <- find_marker(map, "2", 56.8)
    pr_pos <- pull_genoprobpos(probs, marker=marker)

    # reduce the probs to a few positions
    probs[[2]] <- probs[[2]][,, which(dimnames(probs[[2]])[[3]]==marker) + -2:1]

    out_scan1blup <- scan1blup(probs[,2], iron$pheno[,1], kinship[[2]])
    out_scan1blup_nok <- scan1blup(probs[,2], iron$pheno[,1])

    expect_equal( fit1(pr_pos, iron$pheno[,1], kinship[[2]], blup=TRUE, se=FALSE)$coef,
                 out_scan1blup[marker,] )

    expect_equal( fit1(pr_pos, iron$pheno[,1], blup=TRUE, se=FALSE)$coef,
                 out_scan1blup_nok[marker,] )

})


test_that("fit1 fitted values don't depend on order of individuals", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    pheno <- iron$pheno[,1]

    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)

    out <- scan1(probs[,"7"], pheno, addcovar=covar)
    max_pos <- max(out, map)
    pr_max <- pull_genoprobpos(probs, map, max_pos$chr, max_pos$pos)

    out_fit1 <- fit1(pr_max, pheno, addcovar=covar)

    ind <- rownames(pr_max)
    pr_max_samp <- pr_max[sample(ind),]
    out_fit1b <- fit1(pr_max_samp, pheno, addcovar=covar)

    testthat::expect_equal(out_fit1$fitted[ind] , out_fit1b$fitted[ind])

    k <- calc_kinship(probs)

    out_fit1 <- fit1(pr_max, pheno, addcovar=covar, kinship=k)
    out_fit1b <- fit1(pr_max_samp[ind,], pheno[ind], addcovar=covar[ind], kinship=k[ind,ind])

    testthat::expect_equal(out_fit1$fitted[ind] , out_fit1b$fitted[ind])
})

test_that("fit1 works without genoprobs", {

    set.seed(20201215)
    n <- 100
    nam <- paste0("mouse", sample(10*n, n))

    phe <- setNames(rnorm(n), nam)
    cov <- setNames(sample(0:1, n, replace=TRUE), nam)

    lm_out <- lm(phe ~ cov)
    lm_sum <- summary(lm_out)

    coef_names <- c("intercept", "ac1")
    expected <- list(lod=0,
                     ind_lod=setNames(rep(0, n), nam),
                     coef=setNames(lm_out$coef, coef_names),
                     SE=setNames(lm_sum$coef[,"Std. Error"], coef_names),
                     fitted=lm_out$fitted,
                     resid=lm_out$resid)

    expect_equal(fit1(pheno=phe, addcovar=cov), expected)
    expect_equal(fit1(pheno=phe, addcovar=cov, zerosum=TRUE), expected)

    # test var-covariance matrix
    expected$var <- lm_sum$cov.unscaled * lm_sum$sigma^2
    dimnames(expected$var) <- list(coef_names, coef_names)
    expected <- expected[c(1,2,3,4,7,5,6)]
    expect_equal(fit1(pheno=phe, addcovar=cov, var=TRUE), expected)
    expect_equal(fit1(pheno=phe, addcovar=cov, var=TRUE, zerosum=TRUE), expected)

    k <- matrix(0.5,ncol=n, nrow=n)
    diag(k) <- 1
    dimnames(k) <- list(nam, nam)

    # just test that this works
    should_work <- fit1(pheno=phe, addcovar=cov, kinship=k)

})

test_that("fit1 by H-K gives same answer as lm(), including variance matrix", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,18]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    pheno <- iron$pheno[,1]
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)

    # max positions
    out <- scan1(probs, pheno, addcovar=covar)
    mx <- max(out, map)

    # pull prob
    pr <- pull_genoprobpos(probs, map, chr=mx$chr, pos=mx$pos)

    # estimate coefficients; no covariates for X chromosome
    out_lm <- lm(pheno ~ -1 + pr + covar)
    sum_lm <- summary(out_lm)
    coef_lm <- out_lm$coef
    se_lm <- sum_lm$coef[,2]
    var_lm <- sum_lm$cov.unscaled * sum_lm$sigma^2

    expect_equal(sqrt(diag(var_lm)), se_lm)

    out_fit1_se <- fit1(pr, pheno, addcovar=covar, zerosum=FALSE, se=TRUE, var=FALSE)
    out_fit1_var <- fit1(pr, pheno, addcovar=covar, zerosum=FALSE, var=TRUE)

    names(coef_lm) <- names(se_lm) <- rownames(var_lm) <- colnames(var_lm) <- names(out_fit1_se$coef)

    expect_equal(out_fit1_se$coef,  coef_lm)
    expect_equal(out_fit1_se$SE,    se_lm)
    expect_equal(out_fit1_var$coef, coef_lm)
    expect_equal(out_fit1_var$SE,   se_lm)
    expect_equal(out_fit1_var$var,  var_lm)

    # zerosum version
    A <- rbind(c(2/3, -1/3, -1/3, 0),
               c(-1/3, 2/3, -1/3, 0),
               c(-1/3,-1/3,  2/3, 0),
               c(  0,   0,    0,  1),
               c(1/3,  1/3,  1/3, 0))
    coef_lm <- setNames(as.numeric(A %*% coef_lm), c(names(coef_lm), "intercept"))
    var_lm <- A %*% var_lm %*% t(A)
    se_lm <- sqrt(diag(var_lm))
    names(se_lm) <- rownames(var_lm) <- colnames(var_lm) <- names(coef_lm)

    out_fit1_se <- fit1(pr, pheno, addcovar=covar, zerosum=TRUE, se=TRUE, var=FALSE)
    out_fit1_var <- fit1(pr, pheno, addcovar=covar, zerosum=TRUE, var=TRUE)

    expect_equal(out_fit1_se$coef,  coef_lm)
    expect_equal(out_fit1_se$SE,    se_lm)
    expect_equal(out_fit1_var$coef, coef_lm)
    expect_equal(out_fit1_var$SE,   se_lm)
    expect_equal(out_fit1_var$var,  var_lm)

})


test_that("fit1 gives correction variance matrix when incl kinsip", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    map <- insert_pseudomarkers(grav2$gmap, step=1)
    probs <- calc_genoprob(grav2, map, error_prob=0.002)

    pheno <- grav2$pheno[,"T342"]

    k <- calc_kinship(probs, "loco")

    # max positions
    out <- scan1(probs, pheno, kinship=k)
    mx <- max(out, map)

    # pull prob
    pr <- pull_genoprobpos(probs, map, chr=mx$chr, pos=mx$pos)

    n <- n_ind(grav2)
    hsq <- attr(out, "hsq")[mx$chr,1]
    kd <- double_kinship(decomp_kinship(k))
    wts <- 1/sqrt(hsq*kd[[mx$chr]]$values + (1-hsq))

    rotphe <- kd[[mx$chr]]$vectors %*% pheno * wts
    rotX <- kd[[mx$chr]]$vectors %*% pr
    for(i in 1:ncol(rotX)) rotX[,i] <- rotX[,i] * wts

    # estimate coefficients; no covariates for X chromosome
    out_lm <- lm(rotphe ~ -1 + rotX)
    sum_lm <- summary(out_lm)
    coef_lm <- out_lm$coef
    se_lm <- sum_lm$coef[,2]
    var_lm <- sum_lm$cov.unscaled * sum_lm$sigma^2

    expect_equal(sqrt(diag(var_lm)), se_lm)

    out_fit1_se <- fit1(pr, pheno, kinship=k[4], zerosum=FALSE, se=TRUE, var=FALSE)
    out_fit1_var <- fit1(pr, pheno, kinship=k[4], zerosum=FALSE, var=TRUE)

    names(coef_lm) <- names(se_lm) <- rownames(var_lm) <- colnames(var_lm) <- names(out_fit1_se$coef)

    expect_equal(out_fit1_se$SE, out_fit1_var$SE)
    expect_equal(out_fit1_se$SE, sqrt(diag(out_fit1_var$var)))

    expect_equal(out_fit1_se$coef,  coef_lm)
    expect_equal(out_fit1_se$SE,    se_lm)
    expect_equal(out_fit1_var$coef, coef_lm)
    expect_equal(out_fit1_var$SE,   se_lm)
    expect_equal(out_fit1_var$var,  var_lm)

    # zerosum version
    A <- rbind(c(1/2, -1/2),
               c(-1/2, 1/2),
               c(1/2,  1/2))
    coef_lm <- setNames(as.numeric(A %*% coef_lm), c(names(coef_lm), "intercept"))
    var_lm <- A %*% var_lm %*% t(A)
    se_lm <- sqrt(diag(var_lm))
    names(se_lm) <- rownames(var_lm) <- colnames(var_lm) <- names(coef_lm)

    out_fit1_se <- fit1(pr, pheno, kinship=k[4], zerosum=TRUE, se=TRUE, var=FALSE)
    out_fit1_var <- fit1(pr, pheno, kinship=k[4], zerosum=TRUE, var=TRUE)

    expect_equal(out_fit1_se$coef,  coef_lm)
    expect_equal(out_fit1_se$SE,    se_lm)
    expect_equal(out_fit1_var$coef, coef_lm)
    expect_equal(out_fit1_var$SE,   se_lm)
    expect_equal(out_fit1_var$var,  var_lm)

})

context("LMM genome scan by scan1 with kinship matrix")


test_that("scan1 with kinship with intercross, vs ported lmmlite code", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[1:10,c(1,3)]
    map <- insert_pseudomarkers(iron$gmap, step=5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs)

    out_reml <- scan1(probs, iron$pheno, kinship, reml=TRUE)
    out_ml <- scan1(probs, iron$pheno, kinship, reml=FALSE)

    # "by hand" calculation
    y <- iron$pheno
    X <- cbind(rep(1, nrow(iron$pheno)))
    Ke <- decomp_kinship(kinship) # eigen decomp
    yp <- Ke$vectors %*% y
    Xp <- Ke$vectors %*% X
    # double the eigenvalues (== kinship matrix * 2)
    Ke$values <- Ke$values*2

    byhand1_reml <- Rcpp_fitLMM(Ke$values, yp[,1], Xp, reml=TRUE, tol=1e-12)
    byhand2_reml <- Rcpp_fitLMM(Ke$values, yp[,2], Xp, reml=TRUE, tol=1e-12)
    byhand1_ml <- Rcpp_fitLMM(Ke$values, yp[,1], Xp, reml=FALSE, tol=1e-12)
    byhand2_ml <- Rcpp_fitLMM(Ke$values, yp[,2], Xp, reml=FALSE, tol=1e-12)

    # hsq the same?
    expect_equal(as.numeric(attr(out_reml, "hsq")[1,]),
                 c(byhand1_reml$hsq, byhand2_reml$hsq), tol=1e-7)
    expect_equal(as.numeric(attr(out_ml, "hsq")[1,]),
                 c(byhand1_ml$hsq, byhand2_ml$hsq), tol=1e-7)

    # compare chromosome 1 LOD scores
    d <- dim(probs[[1]])[3]
    loglik_reml1 <- loglik_reml2 <-
        loglik_ml1 <- loglik_ml2 <- rep(NA, d)
    for(i in 1:d) {
        Xp <- Ke$vectors %*% cbind(X, probs[[1]][,-1,i])
        # calculate likelihoods using plain ML (not the residual log likelihood)
        loglik_reml1[i] <- Rcpp_calcLL(byhand1_reml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_reml2[i] <- Rcpp_calcLL(byhand2_reml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
        loglik_ml1[i] <- Rcpp_calcLL(byhand1_ml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_ml2[i] <- Rcpp_calcLL(byhand2_ml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
    }
    lod_reml1 <- (loglik_reml1 - byhand1_reml$loglik)/log(10)
    lod_reml2 <- (loglik_reml2 - byhand2_reml$loglik)/log(10)
    lod_ml1 <- (loglik_ml1 - byhand1_ml$loglik)/log(10)
    lod_ml2 <- (loglik_ml2 - byhand2_ml$loglik)/log(10)

    out_reml <- unclass(out_reml)
    out_ml <- unclass(out_ml)
    dimnames(out_reml) <- dimnames(out_ml) <- NULL
    expect_equal(out_reml[1:d,1], lod_reml1, tol=1e-7)
    expect_equal(out_reml[1:d,2], lod_reml2, tol=1e-7)
    expect_equal(out_ml[1:d,1], lod_ml1, tol=1e-7)
    expect_equal(out_ml[1:d,2], lod_ml2, tol=1e-7)

})

test_that("scan1 with intercross with X covariates for null", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs)
    Xc <- get_x_covar(iron)

    out_reml <- scan1(probs, iron$pheno, kinship, Xcovar=Xc, reml=TRUE)
    out_ml <- scan1(probs, iron$pheno, kinship, Xcovar=Xc, reml=FALSE)

    # "by hand" calculation
    y <- iron$pheno
    X <- cbind(rep(1, nrow(iron$pheno)))
    Ke <- decomp_kinship(kinship) # eigen decomp
    yp <- Ke$vectors %*% y
    Xp <- Ke$vectors %*% X
    Xcp <- Ke$vectors %*% Xc
    # double the eigenvalues (== kinship matrix * 2)
    Ke$values <- Ke$values*2

    byhand1_reml <- Rcpp_fitLMM(Ke$values, yp[,1], cbind(Xp, Xcp), reml=TRUE, tol=1e-12)
    byhand2_reml <- Rcpp_fitLMM(Ke$values, yp[,2], cbind(Xp, Xcp), reml=TRUE, tol=1e-12)
    byhand1_ml <- Rcpp_fitLMM(Ke$values, yp[,1], cbind(Xp, Xcp), reml=FALSE, tol=1e-12)
    byhand2_ml <- Rcpp_fitLMM(Ke$values, yp[,2], cbind(Xp, Xcp), reml=FALSE, tol=1e-12)

    # hsq the same?
    expect_equal(as.numeric(attr(out_reml, "hsq")[2,]),
                 c(byhand1_reml$hsq, byhand2_reml$hsq), tolerance=1e-6)
    expect_equal(as.numeric(attr(out_ml, "hsq")[2,]),
                 c(byhand1_ml$hsq, byhand2_ml$hsq))

    # compare chromosome X LOD scores
    d <- dim(probs[["X"]])[3]
    loglik_reml1 <- loglik_reml2 <-
        loglik_ml1 <- loglik_ml2 <- rep(NA, d)
    for(i in 1:d) {
        Xp <- Ke$vectors %*% cbind(1, probs[["X"]][,-1,i])
        # calculate likelihoods using plain ML (not the residual log likelihood)
        loglik_reml1[i] <- Rcpp_calcLL(byhand1_reml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_reml2[i] <- Rcpp_calcLL(byhand2_reml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
        loglik_ml1[i] <- Rcpp_calcLL(byhand1_ml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_ml2[i] <- Rcpp_calcLL(byhand2_ml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
    }
    lod_reml1 <- (loglik_reml1 - byhand1_reml$loglik)/log(10)
    lod_reml2 <- (loglik_reml2 - byhand2_reml$loglik)/log(10)
    lod_ml1 <- (loglik_ml1 - byhand1_ml$loglik)/log(10)
    lod_ml2 <- (loglik_ml2 - byhand2_ml$loglik)/log(10)

    index <- nrow(out_reml) - rev(1:d) + 1
    out_reml <- unclass(out_reml)
    out_ml <- unclass(out_ml)
    dimnames(out_reml) <- dimnames(out_ml) <- NULL
    expect_equal(out_reml[index,1], lod_reml1, tolerance=1e-6)
    expect_equal(out_reml[index,2], lod_reml2, tolerance=1e-6)
    expect_equal(out_ml[index,1], lod_ml1)
    expect_equal(out_ml[index,2], lod_ml2)

})


test_that("scan1 with kinship with intercross with an additive covariate", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs)
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    out_reml <- scan1(probs, iron$pheno, kinship, addcovar=X, Xcovar=Xc, reml=TRUE, tol=1e-12)
    out_ml <- scan1(probs, iron$pheno, kinship, addcovar=X, Xcovar=Xc, reml=FALSE, tol=1e-12)

    # "by hand" calculation
    y <- iron$pheno
    Ke <- decomp_kinship(kinship) # eigen decomp
    yp <- Ke$vectors %*% y
    Xp <- Ke$vectors %*% cbind(1, X)
    Xcp <- Ke$vectors %*% Xc
    # double the eigenvalues (== kinship matrix * 2)
    Ke$values <- Ke$values*2

    # autosome null
    byhand1A_reml <- Rcpp_fitLMM(Ke$values, yp[,1], Xp, reml=TRUE, tol=1e-12)
    byhand2A_reml <- Rcpp_fitLMM(Ke$values, yp[,2], Xp, reml=TRUE, tol=1e-12)
    byhand1A_ml <- Rcpp_fitLMM(Ke$values, yp[,1], Xp, reml=FALSE, tol=1e-12)
    byhand2A_ml <- Rcpp_fitLMM(Ke$values, yp[,2], Xp, reml=FALSE, tol=1e-12)

    expect_equal(as.numeric(attr(out_reml, "hsq")[1,]),
                 c(byhand1A_reml$hsq, byhand2A_reml$hsq))
    expect_equal(as.numeric(attr(out_ml, "hsq")[1,]),
                 c(byhand1A_ml$hsq, byhand2A_ml$hsq))

    # X chr null
    byhand1X_reml <- Rcpp_fitLMM(Ke$values, yp[,1], cbind(Xp, Xcp[,-1]), reml=TRUE, tol=1e-12)
    byhand2X_reml <- Rcpp_fitLMM(Ke$values, yp[,2], cbind(Xp, Xcp[,-1]), reml=TRUE, tol=1e-12)
    byhand1X_ml <- Rcpp_fitLMM(Ke$values, yp[,1], cbind(Xp, Xcp[,-1]), reml=FALSE, tol=1e-12)
    byhand2X_ml <- Rcpp_fitLMM(Ke$values, yp[,2], cbind(Xp, Xcp[,-1]), reml=FALSE, tol=1e-12)

    # hsq the same?
    expect_equal(as.numeric(attr(out_reml, "hsq")[2,]),
                 c(byhand1X_reml$hsq, byhand2X_reml$hsq), tolerance=1e-6)
    expect_equal(as.numeric(attr(out_ml, "hsq")[2,]),
                 c(byhand1X_ml$hsq, byhand2X_ml$hsq))


    # compare chromosome 2 LOD scores
    d <- dim(probs[["2"]])[3]
    loglik_reml1 <- loglik_reml2 <-
        loglik_ml1 <- loglik_ml2 <- rep(NA, d)
    for(i in 1:d) {
        Xp <- Ke$vectors %*% cbind(1, X, probs[["2"]][,-1,i])
        # calculate likelihoods using plain ML (not the residual log likelihood)
        loglik_reml1[i] <- Rcpp_calcLL(byhand1A_reml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_reml2[i] <- Rcpp_calcLL(byhand2A_reml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
        loglik_ml1[i] <- Rcpp_calcLL(byhand1A_ml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_ml2[i] <- Rcpp_calcLL(byhand2A_ml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
    }
    lod_reml1 <- (loglik_reml1 - byhand1A_reml$loglik)/log(10)
    lod_reml2 <- (loglik_reml2 - byhand2A_reml$loglik)/log(10)
    lod_ml1 <- (loglik_ml1 - byhand1A_ml$loglik)/log(10)
    lod_ml2 <- (loglik_ml2 - byhand2A_ml$loglik)/log(10)

    index <- dim(probs[["1"]])[3] + 1:dim(probs[["2"]])[3]
    out_reml <- unclass(out_reml)
    out_ml <- unclass(out_ml)
    dimnames(out_reml) <- dimnames(out_ml) <- NULL
    expect_equal(out_reml[index,1], lod_reml1)
    expect_equal(out_reml[index,2], lod_reml2)
    expect_equal(out_ml[index,1], lod_ml1)
    expect_equal(out_ml[index,2], lod_ml2)

    # compare chromosome X LOD scores
    d <- dim(probs[["X"]])[3]
    loglik_reml1 <- loglik_reml2 <-
        loglik_ml1 <- loglik_ml2 <- rep(NA, d)
    for(i in 1:d) {
        Xp <- Ke$vectors %*% cbind(1, probs[["X"]][,-1,i])
        # calculate likelihoods using plain ML (not the residual log likelihood)
        loglik_reml1[i] <- Rcpp_calcLL(byhand1X_reml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_reml2[i] <- Rcpp_calcLL(byhand2X_reml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
        loglik_ml1[i] <- Rcpp_calcLL(byhand1X_ml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_ml2[i] <- Rcpp_calcLL(byhand2X_ml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
    }
    lod_reml1 <- (loglik_reml1 - byhand1X_reml$loglik)/log(10)
    lod_reml2 <- (loglik_reml2 - byhand2X_reml$loglik)/log(10)
    lod_ml1 <- (loglik_ml1 - byhand1X_ml$loglik)/log(10)
    lod_ml2 <- (loglik_ml2 - byhand2X_ml$loglik)/log(10)

    index <- nrow(out_reml) - rev(1:d) + 1
    out_reml <- unclass(out_reml)
    out_ml <- unclass(out_ml)
    dimnames(out_reml) <- dimnames(out_ml) <- NULL
    ## FIX_ME
    ## REML not yet working on X chromosome, when (X, probs) is not full rank
#    expect_equal(out_reml[index,1], lod_reml1)
#    expect_equal(out_reml[index,2], lod_reml2)
    expect_equal(out_ml[index,1], lod_ml1)
    expect_equal(out_ml[index,2], lod_ml2)

})


test_that("scan1 with kinship with intercross with an interactive covariate", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs)
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    out_reml <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                      Xcovar=Xc, reml=TRUE, tol=1e-12, intcovar_method="lowmem")
    out_ml <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                    Xcovar=Xc, reml=FALSE, tol=1e-12, intcovar_method="lowmem")
    out_reml_himem <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                            Xcovar=Xc, reml=TRUE, tol=1e-12, intcovar_method="highmem")
    out_ml_himem <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                          Xcovar=Xc, reml=FALSE, tol=1e-12, intcovar_method="highmem")

    # same result using "highmem" and "lowmem" methods
    expect_equal(out_reml_himem, out_reml)
    expect_equal(out_ml_himem, out_ml)

    # "by hand" calculation
    y <- iron$pheno
    Ke <- decomp_kinship(kinship) # eigen decomp
    yp <- Ke$vectors %*% y
    Xp <- Ke$vectors %*% cbind(1, X)
    Xcp <- Ke$vectors %*% Xc
    # double the eigenvalues (== kinship matrix * 2)
    Ke$values <- Ke$values*2

    # autosome null (same as w/o interactive covariate)
    byhand1A_reml <- Rcpp_fitLMM(Ke$values, yp[,1], Xp, reml=TRUE, tol=1e-12)
    byhand2A_reml <- Rcpp_fitLMM(Ke$values, yp[,2], Xp, reml=TRUE, tol=1e-12)
    byhand1A_ml <- Rcpp_fitLMM(Ke$values, yp[,1], Xp, reml=FALSE, tol=1e-12)
    byhand2A_ml <- Rcpp_fitLMM(Ke$values, yp[,2], Xp, reml=FALSE, tol=1e-12)

    expect_equal(as.numeric(attr(out_reml, "hsq")[1,]),
                 c(byhand1A_reml$hsq, byhand2A_reml$hsq))
    expect_equal(as.numeric(attr(out_ml, "hsq")[1,]),
                 c(byhand1A_ml$hsq, byhand2A_ml$hsq))

    # X chr null (same as w/o interactive covariate)
    byhand1X_reml <- Rcpp_fitLMM(Ke$values, yp[,1], cbind(Xp, Xcp[,-1]), reml=TRUE, tol=1e-12)
    byhand2X_reml <- Rcpp_fitLMM(Ke$values, yp[,2], cbind(Xp, Xcp[,-1]), reml=TRUE, tol=1e-12)
    byhand1X_ml <- Rcpp_fitLMM(Ke$values, yp[,1], cbind(Xp, Xcp[,-1]), reml=FALSE, tol=1e-12)
    byhand2X_ml <- Rcpp_fitLMM(Ke$values, yp[,2], cbind(Xp, Xcp[,-1]), reml=FALSE, tol=1e-12)

    # hsq the same?
    expect_equal(as.numeric(attr(out_reml, "hsq")[2,]),
                 c(byhand1X_reml$hsq, byhand2X_reml$hsq), tolerance=1e-6)
    expect_equal(as.numeric(attr(out_ml, "hsq")[2,]),
                 c(byhand1X_ml$hsq, byhand2X_ml$hsq))


    # compare chromosome 4 LOD scores
    npos <- sapply(probs, function(a) dim(a)[[3]])
    d <- npos["4"]
    loglik_reml1 <- loglik_reml2 <-
        loglik_ml1 <- loglik_ml2 <- rep(NA, d)
    for(i in 1:d) {
        Xp <- Ke$vectors %*% cbind(1, X, probs[["4"]][,-1,i], probs[["4"]][,-1,i]*X)
        # calculate likelihoods using plain ML (not the residual log likelihood)
        loglik_reml1[i] <- Rcpp_calcLL(byhand1A_reml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_reml2[i] <- Rcpp_calcLL(byhand2A_reml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
        loglik_ml1[i] <- Rcpp_calcLL(byhand1A_ml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_ml2[i] <- Rcpp_calcLL(byhand2A_ml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
    }
    lod_reml1 <- (loglik_reml1 - byhand1A_reml$loglik)/log(10)
    lod_reml2 <- (loglik_reml2 - byhand2A_reml$loglik)/log(10)
    lod_ml1 <- (loglik_ml1 - byhand1A_ml$loglik)/log(10)
    lod_ml2 <- (loglik_ml2 - byhand2A_ml$loglik)/log(10)

    index <- sum(npos[1:3]) + 1:npos[4]
    out_reml <- unclass(out_reml)
    out_ml <- unclass(out_ml)
    dimnames(out_reml) <- dimnames(out_ml) <- NULL
    expect_equal(out_reml[index,1], lod_reml1)
    expect_equal(out_reml[index,2], lod_reml2)
    expect_equal(out_ml[index,1], lod_ml1)
    expect_equal(out_ml[index,2], lod_ml2)

    # compare chromosome X LOD scores
    d <- dim(probs[["X"]])[3]
    loglik_reml1 <- loglik_reml2 <-
        loglik_ml1 <- loglik_ml2 <- rep(NA, 3)
    for(i in 1:d) {
        Xp <- Ke$vectors %*% cbind(1, X, probs[["X"]][,-1,i], probs[["X"]][,-1,i]*X)
        # calculate likelihoods using plain ML (not the residual log likelihood)
        loglik_reml1[i] <- Rcpp_calcLL(byhand1X_reml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_reml2[i] <- Rcpp_calcLL(byhand2X_reml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
        loglik_ml1[i] <- Rcpp_calcLL(byhand1X_ml$hsq, Ke$values, yp[,1], Xp, reml=FALSE)
        loglik_ml2[i] <- Rcpp_calcLL(byhand2X_ml$hsq, Ke$values, yp[,2], Xp, reml=FALSE)
    }
    lod_reml1 <- (loglik_reml1 - byhand1X_reml$loglik)/log(10)
    lod_reml2 <- (loglik_reml2 - byhand2X_reml$loglik)/log(10)
    lod_ml1 <- (loglik_ml1 - byhand1X_ml$loglik)/log(10)
    lod_ml2 <- (loglik_ml2 - byhand2X_ml$loglik)/log(10)

    index <- nrow(out_reml) - rev(1:d) + 1
    ## FIX ME
    ## Not yet working on X chromosome, when (X, probs) is not full rank
#    expect_equal(out_reml[index,1], lod_reml1)
#    expect_equal(out_reml[index,2], lod_reml2)
#    expect_equal(out_ml[index,1], lod_ml1)
#    expect_equal(out_ml[index,2], lod_ml2)

})

test_that("scan1 with kinship works with LOCO, additive covariates", {

    skip_on_cran()

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs, "loco")
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    out_reml <- scan1(probs, iron$pheno, kinship, addcovar=X,
                      Xcovar=Xc, reml=TRUE, tol=1e-12)
    out_ml <- scan1(probs, iron$pheno, kinship, addcovar=X,
                    Xcovar=Xc, reml=FALSE, tol=1e-12)

    y <- iron$pheno
    Ke <- decomp_kinship(kinship) # eigen decomp
    # double the eigenvalues (== kinship matrix * 2)
    Ke <- lapply(Ke, function(a) { a$values <- 2*a$values; a})

    # compare chromosomes 1, 6, 9, 18
    chrs <- paste(c(1,6,9,18))
    npos <- sapply(probs, function(a) dim(a)[[3]])

    for(chr in chrs) {
        nchr <- which(names(npos) == chr)
        d <- npos[chr]

        yp <- Ke[[chr]]$vectors %*% y
        Xp <- Ke[[chr]]$vectors %*% cbind(1, X)

        # autosome null
        byhand1_reml <- Rcpp_fitLMM(Ke[[chr]]$values, yp[,1], Xp, reml=TRUE, tol=1e-12)
        byhand2_reml <- Rcpp_fitLMM(Ke[[chr]]$values, yp[,2], Xp, reml=TRUE, tol=1e-12)
        byhand1_ml <- Rcpp_fitLMM(Ke[[chr]]$values, yp[,1], Xp, reml=FALSE, tol=1e-12)
        byhand2_ml <- Rcpp_fitLMM(Ke[[chr]]$values, yp[,2], Xp, reml=FALSE, tol=1e-12)

        expect_equal(as.numeric(attr(out_reml, "hsq")[nchr,]),
                     c(byhand1_reml$hsq, byhand2_reml$hsq), tolerance=1e-5)
        expect_equal(as.numeric(attr(out_ml, "hsq")[nchr,]),
                     c(byhand1_ml$hsq, byhand2_ml$hsq), tolerance=1e-6)

        # chromosome scan
        loglik_reml1 <- loglik_reml2 <-
            loglik_ml1 <- loglik_ml2 <- rep(NA, d)
        for(i in 1:d) {
            Xp <- Ke[[chr]]$vectors %*% cbind(1, X, probs[[chr]][,-1,i])
            # calculate likelihoods using plain ML (not the residual log likelihood)
            loglik_reml1[i] <- Rcpp_calcLL(byhand1_reml$hsq, Ke[[chr]]$values, yp[,1], Xp, reml=FALSE)
            loglik_reml2[i] <- Rcpp_calcLL(byhand2_reml$hsq, Ke[[chr]]$values, yp[,2], Xp, reml=FALSE)
            loglik_ml1[i] <- Rcpp_calcLL(byhand1_ml$hsq, Ke[[chr]]$values, yp[,1], Xp, reml=FALSE)
            loglik_ml2[i] <- Rcpp_calcLL(byhand2_ml$hsq, Ke[[chr]]$values, yp[,2], Xp, reml=FALSE)
        }
        lod_reml1 <- (loglik_reml1 - byhand1_reml$loglik)/log(10)
        lod_reml2 <- (loglik_reml2 - byhand2_reml$loglik)/log(10)
        lod_ml1 <- (loglik_ml1 - byhand1_ml$loglik)/log(10)
        lod_ml2 <- (loglik_ml2 - byhand2_ml$loglik)/log(10)

        if(nchr > 1) index <- sum(npos[1:(nchr-1)]) + 1:d
        else index <- 1:d
        out_reml <- unclass(out_reml)
        out_ml <- unclass(out_ml)
        dimnames(out_reml) <- dimnames(out_ml) <- NULL
        expect_equal(out_reml[index,1], lod_reml1, tolerance=1e-6)
        expect_equal(out_reml[index,2], lod_reml2, tolerance=1e-5)
        expect_equal(out_ml[index,1], lod_ml1)
        expect_equal(out_ml[index,2], lod_ml2)
    }

})

test_that("scan1 with kinship works with LOCO, interactive covariates", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs, "loco")
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    out_reml <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                      Xcovar=Xc, reml=TRUE, tol=1e-12)
    out_ml <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                    Xcovar=Xc, reml=FALSE, tol=1e-12)


    y <- iron$pheno
    Ke <- decomp_kinship(kinship) # eigen decomp
    # double the eigenvalues (== kinship matrix * 2)
    Ke <- lapply(Ke, function(a) { a$values <- 2*a$values; a})

    # compare chromosomes 1, 6, 9, 18
    chrs <- paste(c(1,6,9,18))
    npos <- sapply(probs, function(a) dim(a)[[3]])

    for(chr in chrs) {
        nchr <- which(names(npos) == chr)
        d <- npos[chr]

        yp <- Ke[[chr]]$vectors %*% y
        Xp <- Ke[[chr]]$vectors %*% cbind(1, X)

        # autosome null (same as w/o interactive covariate)
        byhand1_reml <- Rcpp_fitLMM(Ke[[chr]]$values, yp[,1], Xp, reml=TRUE, tol=1e-12)
        byhand2_reml <- Rcpp_fitLMM(Ke[[chr]]$values, yp[,2], Xp, reml=TRUE, tol=1e-12)
        byhand1_ml <- Rcpp_fitLMM(Ke[[chr]]$values, yp[,1], Xp, reml=FALSE, tol=1e-12)
        byhand2_ml <- Rcpp_fitLMM(Ke[[chr]]$values, yp[,2], Xp, reml=FALSE, tol=1e-12)

        expect_equal(as.numeric(attr(out_reml, "hsq")[nchr,]),
                     c(byhand1_reml$hsq, byhand2_reml$hsq), tolerance=1e-5)
        expect_equal(as.numeric(attr(out_ml, "hsq")[nchr,]),
                     c(byhand1_ml$hsq, byhand2_ml$hsq), tolerance=1e-6)

        # chromosome scan
        loglik_reml1 <- loglik_reml2 <-
            loglik_ml1 <- loglik_ml2 <- rep(NA, d)
        for(i in 1:d) {
            Xp <- Ke[[chr]]$vectors %*% cbind(1, X, probs[[chr]][,-1,i], probs[[chr]][,-1,i]*X)
            # calculate likelihoods using plain ML (not the residual log likelihood)
            loglik_reml1[i] <- Rcpp_calcLL(byhand1_reml$hsq, Ke[[chr]]$values, yp[,1], Xp, reml=FALSE)
            loglik_reml2[i] <- Rcpp_calcLL(byhand2_reml$hsq, Ke[[chr]]$values, yp[,2], Xp, reml=FALSE)
            loglik_ml1[i] <- Rcpp_calcLL(byhand1_ml$hsq, Ke[[chr]]$values, yp[,1], Xp, reml=FALSE)
            loglik_ml2[i] <- Rcpp_calcLL(byhand2_ml$hsq, Ke[[chr]]$values, yp[,2], Xp, reml=FALSE)
        }
        lod_reml1 <- (loglik_reml1 - byhand1_reml$loglik)/log(10)
        lod_reml2 <- (loglik_reml2 - byhand2_reml$loglik)/log(10)
        lod_ml1 <- (loglik_ml1 - byhand1_ml$loglik)/log(10)
        lod_ml2 <- (loglik_ml2 - byhand2_ml$loglik)/log(10)

        if(nchr > 1) index <- sum(npos[1:(nchr-1)]) + 1:d
        else index <- 1:d
        out_reml <- unclass(out_reml)
        out_ml <- unclass(out_ml)
        dimnames(out_reml) <- dimnames(out_ml) <- NULL
        expect_equal(out_reml[index,1], lod_reml1)
        expect_equal(out_reml[index,2], lod_reml2, tolerance=1e-6)
        expect_equal(out_ml[index,1], lod_ml1)
        expect_equal(out_ml[index,2], lod_ml2)
    }


})


test_that("scan1 with kinship works with multicore", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs, "loco")
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    out_reml <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                      Xcovar=Xc, reml=TRUE, tol=1e-12)
    out_reml_4core <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                            Xcovar=Xc, reml=TRUE, tol=1e-12, cores=2)
    expect_equal(out_reml, out_reml_4core)


    out_ml <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                    Xcovar=Xc, reml=FALSE, tol=1e-12)
    out_ml_4core <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                          Xcovar=Xc, reml=FALSE, tol=1e-12, cores=2)
    expect_equal(out_ml, out_ml_4core)

})


test_that("scan1 with kinship LOD results invariant to change in scale to pheno and covar", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs, "loco")
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    out_reml <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                      Xcovar=Xc, reml=TRUE, tol=1e-12)
    out_reml_scale <- scan1(probs, iron$pheno/100, kinship, addcovar=X*2, intcovar=X*2,
                            Xcovar=Xc*2, reml=TRUE, tol=1e-12)
    expect_equal(out_reml, out_reml_scale, tol=1e-6)


    out_ml <- scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                    Xcovar=Xc, reml=FALSE, tol=1e-12)
    out_ml_scale <- scan1(probs, iron$pheno/100, kinship, addcovar=X*4, intcovar=X*4,
                          Xcovar=Xc*4, reml=FALSE, tol=1e-12)
    expect_equal(out_ml, out_ml_scale, tol=1e-6)

})

test_that("scan1 deals with mismatching individuals", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs, "loco")
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    ind <- c(1:50, 101:150)
    subK <- lapply(kinship, "[", ind, ind)
    expected <- scan1(probs[ind,], iron$pheno[ind,,drop=FALSE], subK, addcovar=X[ind], intcovar=X[ind],
                      Xcovar=Xc[ind,], reml=TRUE, tol=1e-12)
    expect_equal(scan1(probs[ind,], iron$pheno, kinship, addcovar=X, intcovar=X,
                      Xcovar=Xc, reml=TRUE, tol=1e-12), expected)
    expect_equal(scan1(probs, iron$pheno[ind,], kinship, addcovar=X, intcovar=X,
                      Xcovar=Xc, reml=TRUE, tol=1e-12), expected)
    expect_equal(scan1(probs, iron$pheno, subK, addcovar=X, intcovar=X,
                      Xcovar=Xc, reml=TRUE, tol=1e-12), expected)
    expect_equal(scan1(probs, iron$pheno, kinship, addcovar=X[ind], intcovar=X,
                      Xcovar=Xc, reml=TRUE, tol=1e-12), expected)
    expect_equal(scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X[ind],
                      Xcovar=Xc, reml=TRUE, tol=1e-12), expected)
    expect_equal(scan1(probs, iron$pheno, kinship, addcovar=X, intcovar=X,
                      Xcovar=Xc[ind,], reml=TRUE, tol=1e-12), expected)

})


test_that("scan1 with weights and kinship", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs, "loco")
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    chr <- c("7", "12", "X")
    probs <- probs[,chr]
    kinship <- kinship[chr]

    RNGkind("Mersenne-Twister") # make sure we're using the standard RNG
    set.seed(28915967)
    weights <- stats::setNames(sample(1:10, n_ind(iron), replace=TRUE), ind_ids(iron))

    out_reml <- scan1(probs, iron$pheno, kinship, addcovar=X,
                      Xcovar=Xc, weights=weights, reml=TRUE, tol=1e-12)

    ### results via regress library
    # library(regress)
    # hsq <- matrix(nrow=3, ncol=2)
    # dimnames(hsq) <- list(c("7", "12", "X"), c("liver", "spleen"))
    # for(i in 1:2) {
    #     for(j in 1:3) {
    #         k <- kinship[[j]]*2
    #         if(j==3) co <- Xc else co <- X
    #         out_regress <- regress(iron$pheno[,i] ~ co, ~ k + diag(1/weights), identity=FALSE, tol=1e-8)
    #         sig <- out_regress$sigma
    #         hsq[j,i] <- sig[1]/sum(sig)
    #    }
    # }
    hsq <- structure(c(0.281878438466624, 0.353558919335662, 0.354315217715963,
                       0.20945078570429, 0.217671918799751, 0.201682575651958),
                     dim = 3:2, dimnames = list(c("7", "12", "X"), c("liver", "spleen")))

    expect_equal(attr(out_reml, "hsq"), hsq, tol=1e-5)

    ### calculate lod scors from scan
    # hsq <- attr(out_reml, "hsq")
    # result <- vector("list", 6)
    # for(i in 1:2) {
    #     for(j in 1:3) {
    #         # cholesky decomp of kinhip matrix
    #         v <- hsq[j,i]*2*kinship[[j]] + (1-hsq[j,i])*diag(1/weights)
    #         d <- solve(t(chol( v )))
    #         # pre-multiply phenotype and covar by t(d)
    #         y <- d %*% iron$pheno[,i]
    #         if(j==3) co <- Xc else co <- X
    #         co <- d %*% cbind(1, co)
    #
    #         # log lik under null
    #         r <- lm(y ~ -1 + co)$resid
    #         llik0 <- sum(dnorm(r, 0, sqrt(mean(r^2)), log=TRUE))
    #
    #         llik1 <- apply(probs[[j]], 3, function(z) {
    #             thisX <- cbind(co, d %*% z[,-1,drop=FALSE])
    #             r <- lm(y ~ -1 + thisX)$resid
    #             sum(dnorm(r, 0, sqrt(mean(r^2)), log=TRUE)) })
    #
    #         result[[(i-1)*3+j]] <- (llik1 - llik0)/log(10)
    #
    #     }
    # }
    # result <- rbind(cbind(result[[1]], result[[4]]),
    #                 cbind(result[[2]], result[[5]]),
    #                 cbind(result[[3]], result[[6]]))
    # dimnames(result) <- dimnames(out_reml)
    # class(result) <- class(out_reml)

    result <- structure(c(0.765379546370331, 1.33697830149196, 2.05141217249269,
                          2.82748389184348, 3.57825928288821, 4.11482576884623,
                          4.18208695088672, 4.5010472223741, 4.77640847444171,
                          4.98121670508953, 5.07224467963965, 5.28435499128555,
                          5.10772158608095, 3.83479814287182, 3.89931966500793,
                          4.24832760135268, 4.18515409936768, 4.61791967696627,
                          4.70308738505572, 4.58758777783123, 4.88425659502959,
                          5.39302146958257, 5.86125538183219, 6.27973204425698,
                          6.62245498126028, 6.84349072896544, 6.91418168584823,
                          0.492648140597935, 0.483021280268028, 0.465046322957041,
                          0.435712712087884, 0.391979816567206, 0.332013881699469,
                          0.257330299826467, 0.174985513110401, 0.0975448556296765,
                          0.0390292238934769, 0.0087954970835623, 0.0081071126160765,
                          0.0315063370227757, 0.0706853835370534, 0.117761588763882,
                          0.166806834360401, 0.180329873353274, 0.146302909441694,
                          0.161137797908763, 0.192675853649471, 0.243358929100302,
                          0.313640413846475, 0.401241925760445, 0.501210880937258,
                          0.606957139785506, 0.711817493234455, 0.810381904299065,
                          0.899112752704855, 0.976303366208711, 1.00118015673376,
                          0.68879116519848, 1.13593060562675, 1.59214070065802,
                          1.93410070637396, 2.08715532129677, 2.08161469298463,
                          2.11418090796465, 2.24129850312586, 2.28750006399176,
                          2.24152015532401, 2.15435105671462, 2.16059659637594,
                          1.40659065214545, 0.389546568631669, 0.392567989473253,
                          0.432129729053773, 0.445384879284572, 0.468919719042007,
                          0.561846602500791, 0.61487493505232, 0.680306211651546,
                          0.828892839652596, 1.03747954481922, 1.3178288863555,
                          1.64752359905175, 1.97570834069646, 2.25872692788406,
                          1.76552836513385, 1.9624748694232, 2.179534029569,
                          2.40709075366897, 2.62706355895772, 2.81242765604838,
                          2.93166656075001, 2.95963946310072, 2.89030386259124,
                          2.74110806468527, 2.54394356128059, 2.33040574729971,
                          2.12258054254516, 1.93206224497564, 1.76308130810039,
                          1.61583945469733, 1.57830829020771, 1.91298146944818,
                          1.89937590028851, 1.86320092206996, 1.79734091344691,
                          1.69690361501555, 1.56240479342725, 1.4020247631517,
                          1.23069582035597, 1.06561266027339, 0.920732232061234,
                          0.803561188099648, 0.715279174632663, 0.690066084958058),
                        dim = c(57L, 2L),
                        dimnames = list(c("D7Mit74", "c7.loc4", "c7.loc6", "c7.loc9",
                                          "c7.loc11", "D7Mit25", "c7.loc14", "c7.loc16",
                                          "c7.loc19", "c7.loc21", "D7Nds5", "c7.loc24",
                                          "c7.loc26", "D7mit30", "c7.loc29", "c7.loc31",
                                          "D7Mit31", "c7.loc34", "c7.loc36", "D7Mit17",
                                          "c7.loc39", "c7.loc41", "c7.loc44", "c7.loc46",
                                          "c7.loc49", "c7.loc51", "D7Mit71", "D12Mit88",
                                          "c12.loc22", "c12.loc25", "c12.loc27", "c12.loc30",
                                          "c12.loc32", "c12.loc35", "c12.loc37", "c12.loc40",
                                          "c12.loc42", "c12.loc45", "c12.loc47", "c12.loc50",
                                          "c12.loc52", "c12.loc55", "c12.loc57", "D12Mit134",
                                          "DXMit16", "cX.loc32", "cX.loc34", "cX.loc37",
                                          "cX.loc40", "cX.loc42", "cX.loc44", "cX.loc47",
                                          "cX.loc50", "cX.loc52", "cX.loc54",
                                          "cX.loc57", "DXMit186"), c("liver", "spleen")),
                        class = c("scan1", "matrix"))

    for(at in c("hsq", "sample_size"))
        attr(result, at) <- attr(out_reml, at)
    expect_equal(out_reml, result, tol=1e-6)

})

test_that("scan1 works with hsq specified", {

    skip_if(isnt_karl(), "this test only run locally")

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs)
    kinship_loco <- calc_kinship(probs, "loco")

    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    # no covariates, single kinship
    out <- scan1(probs, iron$pheno, kinship)
    expect_equal(scan1(probs, iron$pheno, kinship, hsq=attr(out, "hsq")), out)

    # no covariates, loco kinship
    out <- scan1(probs, iron$pheno, kinship_loco)
    expect_equal(scan1(probs, iron$pheno, kinship_loco, hsq=attr(out, "hsq")), out)


    # X covariates, single kinship
    out <- scan1(probs, iron$pheno, kinship, Xcovar=Xc)
    expect_equal(scan1(probs, iron$pheno, kinship, Xcovar=Xc, hsq=attr(out, "hsq")), out)

    # X covariates, loco kinship
    out <- scan1(probs, iron$pheno, kinship_loco, Xcovar=Xc)
    expect_equal(scan1(probs, iron$pheno, kinship_loco, Xcovar=Xc, hsq=attr(out, "hsq")), out)


    # covariates, single kinship
    out <- scan1(probs, iron$pheno, kinship, addcovar=X)
    expect_equal(scan1(probs, iron$pheno, kinship, addcovar=X, hsq=attr(out, "hsq")), out)

    # covariates, loco kinship
    out <- scan1(probs, iron$pheno, kinship_loco, addcovar=X)
    expect_equal(scan1(probs, iron$pheno, kinship_loco, addcovar=X, hsq=attr(out, "hsq")), out)



    # covariates + X covar, single kinship
    out <- scan1(probs, iron$pheno, kinship, addcovar=X, Xcovar=Xc)
    expect_equal(scan1(probs, iron$pheno, kinship, addcovar=X, Xcovar=Xc, hsq=attr(out, "hsq")), out)

    # covariates + X covar, loco kinship
    out <- scan1(probs, iron$pheno, kinship_loco, addcovar=X, Xcovar=Xc)
    expect_equal(scan1(probs, iron$pheno, kinship_loco, addcovar=X, Xcovar=Xc, hsq=attr(out, "hsq")), out)

})

context("scan1coef for binary traits")

test_that("scan1coef for binary traits works with intercross", {

    set.seed(17262911)

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(2,"X")]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, err=0.002)

    phe <- iron$pheno[,1]
    phe <- setNames(as.numeric(phe > quantile(phe, 0.7)),
                    ind_ids(iron))

    # no covariates, autosome
    co <- scan1coef(probs[,"2"], phe, model="binary", zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, model="binary", se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) glm(phe ~ -1 + a, family=binomial(link=logit),
                                                    control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    # no covariates, X chromosome
    co <- scan1coef(probs[,"X"], phe, model="binary", zerosum=FALSE)
    coSE <- scan1coef(probs[,"X"], phe, model="binary", se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[2]], 3, function(a) glm(phe ~ -1 + a, family=binomial(link=logit),
                                                    control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BS", "BB", "SY", "BY")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    sex <- setNames(as.numeric(iron$is_female), names(iron$is_female))

    # add covariate, autosome
    co <- scan1coef(probs[,"2"], phe, addcovar=sex, model="binary", zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, addcovar=sex, model="binary", se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) glm(phe ~ -1 + a + sex, family=binomial(link=logit),
                                                    control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB", "ac1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    covar <- setNames(rnorm(n_ind(iron)), names(sex))

    # add covariate, X chromosome
    co <- scan1coef(probs[,"X"], phe, addcovar=covar, model="binary", zerosum=FALSE)
    coSE <- scan1coef(probs[,"X"], phe, addcovar=covar, model="binary", se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[2]], 3, function(a) glm(phe ~ -1 + a + covar, family=binomial(link=logit),
                                                    control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BS", "BB", "SY", "BY", "ac1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    # int covariate, autosome
    co <- scan1coef(probs[,"2"], phe, addcovar=sex, intcovar=sex, model="binary", zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, addcovar=sex, intcovar=sex, model="binary", se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) {
        X <- cbind(a, sex, a[,-1]*sex)
        glm(phe ~ -1 + X, family=binomial(link=logit),
            control=list(epsilon=1e-12)) })
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB", "ac1", "SB:ic1", "BB:ic1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

})


test_that("scan1coef for binary traits works some missing phenotypes", {

    set.seed(17262911)

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(2,"X")]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, err=0.002)

    phe <- iron$pheno[,1]
    phe <- setNames(as.numeric(phe > quantile(phe, 0.7)),
                    ind_ids(iron))
    phe[c(41,153)] <- NA

    # no covariates, autosome
    co <- scan1coef(probs[,"2"], phe, model="binary", zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, model="binary", se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) glm(phe ~ -1 + a, family=binomial(link=logit),
                                                    control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    # no covariates, X chromosome
    co <- scan1coef(probs[,"X"], phe, model="binary", zerosum=FALSE)
    coSE <- scan1coef(probs[,"X"], phe, model="binary", se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[2]], 3, function(a) glm(phe ~ -1 + a, family=binomial(link=logit),
                                                    control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BS", "BB", "SY", "BY")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    sex <- setNames(as.numeric(iron$is_female), names(iron$is_female))

    # add covariate, autosome
    co <- scan1coef(probs[,"2"], phe, addcovar=sex, model="binary", zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, addcovar=sex, model="binary", se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) glm(phe ~ -1 + a + sex, family=binomial(link=logit),
                                                    control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB", "ac1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    covar <- setNames(rnorm(n_ind(iron)), names(sex))

    # add covariate, X chromosome
    co <- scan1coef(probs[,"X"], phe, addcovar=covar, model="binary", zerosum=FALSE)
    coSE <- scan1coef(probs[,"X"], phe, addcovar=covar, model="binary", se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[2]], 3, function(a) glm(phe ~ -1 + a + covar, family=binomial(link=logit),
                                                    control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BS", "BB", "SY", "BY", "ac1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    # int covariate, autosome
    co <- scan1coef(probs[,"2"], phe, addcovar=sex, intcovar=sex, model="binary", zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, addcovar=sex, intcovar=sex, model="binary", se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) {
        X <- cbind(a, sex, a[,-1]*sex)
        glm(phe ~ -1 + X, family=binomial(link=logit),
            control=list(epsilon=1e-12)) })
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB", "ac1", "SB:ic1", "BB:ic1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

})

test_that("scan1coef for binary traits works with weights", {

    set.seed(17262911)

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(2,"X")]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, err=0.002)

    phe <- iron$pheno[,1]
    phe <- setNames(as.numeric(phe > quantile(phe, 0.7)),
                    ind_ids(iron))
    weights <- setNames(sample(1:10, n_ind(iron), replace=TRUE), names(phe))

    # no covariates, autosome
    co <- scan1coef(probs[,"2"], phe, model="binary", weights=weights, zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, model="binary", , weights=weights, se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) glm(phe ~ -1 + a, family=binomial(link=logit),
                                                    weights=weights, control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    # no covariates, X chromosome
    co <- scan1coef(probs[,"X"], phe, model="binary", weights=weights, zerosum=FALSE)
    coSE <- scan1coef(probs[,"X"], phe, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[2]], 3, function(a) glm(phe ~ -1 + a, family=binomial(link=logit),
                                                    weights=weights, control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BS", "BB", "SY", "BY")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    sex <- setNames(as.numeric(iron$is_female), names(iron$is_female))

    # add covariate, autosome
    co <- scan1coef(probs[,"2"], phe, addcovar=sex, model="binary", weights=weights, zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, addcovar=sex, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) glm(phe ~ -1 + a + sex, family=binomial(link=logit),
                                                    weights=weights, control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB", "ac1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    covar <- setNames(rnorm(n_ind(iron)), names(sex))

    # add covariate, X chromosome
    co <- scan1coef(probs[,"X"], phe, addcovar=covar, model="binary", weights=weights, zerosum=FALSE)
    coSE <- scan1coef(probs[,"X"], phe, addcovar=covar, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[2]], 3, function(a) glm(phe ~ -1 + a + covar, family=binomial(link=logit),
                                                    weights=weights, control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BS", "BB", "SY", "BY", "ac1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    # int covariate, autosome
    co <- scan1coef(probs[,"2"], phe, addcovar=sex, intcovar=sex, model="binary", weights=weights, zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, addcovar=sex, intcovar=sex, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) {
        X <- cbind(a, sex, a[,-1]*sex)
        glm(phe ~ -1 + X, family=binomial(link=logit),
            control=list(epsilon=1e-12), weights=weights) })
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB", "ac1", "SB:ic1", "BB:ic1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

})

test_that("scan1coef for binary traits works with weights and missing phenotypes", {

    set.seed(17262911)

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    iron <- iron[,c(2,"X")]
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, err=0.002)

    phe <- iron$pheno[,1]
    phe <- setNames(as.numeric(phe > quantile(phe, 0.7)),
                    ind_ids(iron))
    phe[c(108,142,268)] <- NA
    weights <- setNames(sample(1:10, n_ind(iron), replace=TRUE), names(phe))

    # no covariates, autosome
    co <- scan1coef(probs[,"2"], phe, model="binary", weights=weights, zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, model="binary", , weights=weights, se=TRUE, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) glm(phe ~ -1 + a, family=binomial(link=logit),
                                                    weights=weights, control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    # no covariates, X chromosome
    co <- scan1coef(probs[,"X"], phe, model="binary", weights=weights, zerosum=FALSE)
    coSE <- scan1coef(probs[,"X"], phe, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[2]], 3, function(a) glm(phe ~ -1 + a, family=binomial(link=logit),
                                                    weights=weights, control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BS", "BB", "SY", "BY")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    sex <- setNames(as.numeric(iron$is_female), names(iron$is_female))

    # add covariate, autosome
    co <- scan1coef(probs[,"2"], phe, addcovar=sex, model="binary", weights=weights, zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, addcovar=sex, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) glm(phe ~ -1 + a + sex, family=binomial(link=logit),
                                                    weights=weights, control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB", "ac1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    covar <- setNames(rnorm(n_ind(iron)), names(sex))

    # add covariate, X chromosome
    co <- scan1coef(probs[,"X"], phe, addcovar=covar, model="binary", weights=weights, zerosum=FALSE)
    coSE <- scan1coef(probs[,"X"], phe, addcovar=covar, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[2]], 3, function(a) glm(phe ~ -1 + a + covar, family=binomial(link=logit),
                                                    weights=weights, control=list(epsilon=1e-12)))
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BS", "BB", "SY", "BY", "ac1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

    # int covariate, autosome
    co <- scan1coef(probs[,"2"], phe, addcovar=sex, intcovar=sex, model="binary", weights=weights, zerosum=FALSE)
    coSE <- scan1coef(probs[,"2"], phe, addcovar=sex, intcovar=sex, model="binary", se=TRUE, weights=weights, zerosum=FALSE)
    expect_equivalent(co, coSE)

    out_glm <- apply(probs[[1]], 3, function(a) {
        X <- cbind(a, sex, a[,-1]*sex)
        glm(phe ~ -1 + X, family=binomial(link=logit),
            control=list(epsilon=1e-12), weights=weights) })
    glm_coef <- t(sapply(out_glm, function(a) a$coef))
    glm_se <- t(sapply(out_glm, function(a) summary(a)$coef[,2]))
    colnames(glm_se) <- colnames(glm_coef) <- c("SS", "SB", "BB", "ac1", "SB:ic1", "BB:ic1")
    expect_equivalent(unclass(co), glm_coef)
    expect_equal(attr(coSE, "SE"), glm_se, tol=1e-6)

})

context("effect scan by scan1coef with polygenic effect")

# calc estimates via lm(), just one chromosome
eff_via_lm <-
    function(probs, pheno, kinship, addcovar=NULL, intcovar=NULL,
             se=TRUE)
{
    kinship <- qtl2:::double_kinship(kinship) # need 2*kinship for LMM
    kinship <- decomp_kinship(kinship)
    eigenvec <- kinship$vectors
    hsq <- qtl2:::calc_hsq_clean(kinship, as.matrix(pheno), addcovar, NULL, FALSE, NULL,
                                 reml=TRUE, cores=1, check_boundary=TRUE, tol=1e-12)$hsq[1,1]

    wts <- 1/(hsq*kinship$values + (1-hsq))

    npos <- dim(probs)[3]
    nind <- length(pheno)
    ngen <- dim(probs)[2]

    nadd <- ifelse(is.null(addcovar), 0, ncol(addcovar))
    nint <- ifelse(is.null(intcovar), 0, ncol(intcovar))
    ncoef <- ngen + nadd + (ngen-1)*nint

    result <- matrix(nrow=npos, ncol=ncoef)
    SEs <- matrix(nrow=npos, ncol=ncoef)

    if(!is.null(intcovar)) intcovar <- as.matrix(intcovar)

    pheno <- eigenvec %*% pheno

    for(i in 1:npos) {
        X <- probs[,,i]
        if(!is.null(addcovar)) X <- cbind(X, addcovar)
        if(!is.null(intcovar)) {
            for(j in 1:ncol(intcovar))
                X <- cbind(X, probs[,-1,i]*intcovar[,j])
        }
        X <- eigenvec %*% X

        lm.out <- lm(pheno ~ -1 + X, weights=wts)
        result[i,] <- lm.out$coef
        if(se) # need to deal with NAs
            SEs[i,!is.na(lm.out$coef)] <- summary(lm.out)$coef[,2]
        else SEs <- NULL
    }

    attr(result, "SE") <- SEs

    class(result) <- c("scan1coef", "scan1", "matrix")
    result
}

# function to maybe swap NAs between 2nd and 3rd columns in 22nd row
swap_nas <-
    function(mat, row=22, col=2:3, also_se=TRUE)
{
    if(is.na(mat[row,col[1]]) && !is.na(mat[row, col[2]])) {
        mat[row,col] <- mat[row,rev(col)]
    }

    if(also_se) {
        attr(mat, "SE") <- swap_nas(attr(mat, "SE"), row, col, also_se=FALSE)
    }

    mat
}

test_that("scan1coef_pg for grav", {

    set.seed(9308594)

    grav <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    grav <- grav[1:30,c(3,4)]
    map <- insert_pseudomarkers(grav$gmap, step=1)
    pr <- calc_genoprob(grav, map)
    K <- calc_kinship(pr)
    phe <- grav$pheno[,"T330",drop=FALSE]

    est <- scan1coef(pr[,"3"], phe, K, se=FALSE, zerosum=FALSE)
    est_lm <- eff_via_lm(pr[["3"]], phe, K)
    expect_equivalent(est, est_lm)

    est <- scan1coef(pr[,"3"], phe, K, se=TRUE, zerosum=FALSE)
    expect_equivalent(est, est_lm)
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))

    # pre-decomp kinship
    Ke <- decomp_kinship(K)
    est <- scan1coef(pr[,"3"], phe, Ke, zerosum=FALSE)
    expect_equivalent(est, est_lm)

    est <- scan1coef(pr[,"3"], phe, Ke, se=TRUE, zerosum=FALSE)
    expect_equivalent(est, est_lm)
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))

    # kinship is a list of length 1
    Klist <- list("1"=K)
    est <- scan1coef(pr[,"3"], phe, Klist, zerosum=FALSE)
    expect_equivalent(est, est_lm)

    est <- scan1coef(pr[,"3"], phe, Klist, se=TRUE, zerosum=FALSE)
    expect_equivalent(est, est_lm)
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))

    # include covariate
    covar <- cbind(chr3=pr[["3"]][,2,"CC.266L"])
    est <- scan1coef(pr[,"3"], phe, K, covar, se=FALSE, zerosum=FALSE)
    est <- swap_nas(est, also_se=FALSE)
    est_lm <- eff_via_lm(pr[["3"]], phe, K, covar)
    est_lm <- swap_nas(est_lm)
    expect_equivalent(est, est_lm)

    est <- scan1coef(pr[,"3"], phe, K, covar, se=TRUE, zerosum=FALSE)
    est <- swap_nas(est)
    expect_equivalent(est, est_lm)
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))

    # pre-computed eigen decomp
    est <- scan1coef(pr[,"3"], phe, Ke, covar, zerosum=FALSE)
    est <- swap_nas(est, also_se=FALSE)
    expect_equivalent(est, est_lm)  ## <- fix problem here
    est <- scan1coef(pr[,"3"], phe, Ke, covar, se=TRUE, zerosum=FALSE)
    est <- swap_nas(est)
    expect_equivalent(est, est_lm)  ## <- fix problem here
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))  ## <- fix problem here

    # interactive covariate
    est <- scan1coef(pr[,"3"], phe, K, covar, intcovar=covar, se=FALSE, zerosum=FALSE)
    est_lm <- eff_via_lm(pr[["3"]], phe, K, covar, covar)
    expect_equivalent(est, est_lm)

    est <- scan1coef(pr[,"3"], phe, K, covar, intcovar=covar, se=TRUE, zerosum=FALSE)
    expect_equivalent(est, est_lm)
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))

    # pre-computed eigen decomp
    est <- scan1coef(pr[,"3"], phe, Ke, covar, intcovar=covar, zerosum=FALSE)
    expect_equivalent(est, est_lm)
    est <- scan1coef(pr[,"3"], phe, Ke, covar, intcovar=covar, se=TRUE, zerosum=FALSE)
    expect_equivalent(est, est_lm)
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))

    skip_on_cran()

    # two covariates
    covar <- cbind(covar, chr4=pr[["4"]][,2,"CD.329C-Col"])
    est <- scan1coef(pr[,"3"], phe, K, covar, se=FALSE, zerosum=FALSE)
    est <- swap_nas(est, also_se=FALSE)
    est_lm <- eff_via_lm(pr[["3"]], phe, K, covar)
    est_lm <- swap_nas(est_lm)
    expect_equivalent(est, est_lm)

    est <- scan1coef(pr[,"3"], phe, K, covar, se=TRUE, zerosum=FALSE)
    est <- swap_nas(est)
    expect_equivalent(est, est_lm)
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))

    # pre-computed eigen decomp
    est <- scan1coef(pr[,"3"], phe, Ke, covar, zerosum=FALSE)
    est <- swap_nas(est, also_se=FALSE)
    expect_equivalent(est, est_lm)
    est <- scan1coef(pr[,"3"], phe, Ke, covar, se=TRUE, zerosum=FALSE)
    est <- swap_nas(est)
    expect_equivalent(est, est_lm)
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))

    # two interactive covariates
    est <- scan1coef(pr[,"3"], phe, K, covar, intcovar=covar, se=FALSE, zerosum=FALSE)
    est_lm <- eff_via_lm(pr[["3"]], phe, K, covar, covar)
    expect_equivalent(est, est_lm)

    est <- scan1coef(pr[,"3"], phe, K, covar, intcovar=covar, se=TRUE, zerosum=FALSE)
    expect_equivalent(est, est_lm)
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))

    # pre-computed eigen decomp
    est <- scan1coef(pr[,"3"], phe, Ke, covar, intcovar=covar, zerosum=FALSE)
    expect_equivalent(est, est_lm)
    est <- scan1coef(pr[,"3"], phe, Ke, covar, intcovar=covar, se=TRUE, zerosum=FALSE)
    expect_equivalent(est, est_lm)
    expect_equivalent(attr(est, "SE"), attr(est_lm, "SE"))

})


test_that("scan1coef deals with mismatching individuals", {

    skip_on_cran()

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs, "loco")[["3"]]
    probs <- probs[,"3"]
    Xc <- get_x_covar(iron)
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)

    phe <- iron$pheno[,2,drop=FALSE]

    ind <- c(1:50, 101:150)
    expected <- scan1coef(probs[ind,], phe[ind,,drop=FALSE], kinship[ind,ind], addcovar=X[ind])
    expect_equal(scan1coef(probs[ind,], phe, kinship, addcovar=X), expected)
    expect_equal(scan1coef(probs, phe[ind,,drop=FALSE], kinship, addcovar=X), expected)
    expect_equal(scan1coef(probs, phe, kinship[ind,ind], addcovar=X), expected)
    expect_equal(scan1coef(probs, phe, kinship, addcovar=X[ind]), expected)

})

test_that("scan1coef with weights works", {

    skip_on_cran()

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=2.5)
    probs <- calc_genoprob(iron, map, error_prob=0.002)
    kinship <- calc_kinship(probs, "loco")[["3"]]
    probs <- probs[,"3"]
    X <- match(iron$covar$sex, c("f", "m"))-1
    names(X) <- rownames(iron$covar)
    phe <- iron$pheno[,2,drop=FALSE]

    weights <- setNames(runif(n_ind(iron), 1, 10), ind_ids(iron))

    # plain least squares match LMM when hsq==0?
    coef_hk <- scan1coef(probs, phe, addcovar=X, se=TRUE, zerosum=FALSE)
    coef_lmm0 <- scan1coef(probs, phe, kinship, X, se=TRUE, hsq=0, zerosum=FALSE)
    expect_equal(coef_hk, coef_lmm0)

    # plain least squares match LMM when hsq==0, with weights?
    coef_hk <- scan1coef(probs, phe, addcovar=X, weights=weights, se=TRUE, zerosum=FALSE)
    coef_lmm0 <- scan1coef(probs, phe, kinship, X, weights=weights, se=TRUE, hsq=0, zerosum=FALSE)
    expect_equal(coef_hk, coef_lmm0)

    # same with contrasts, no weights
    coef_hk <- scan1coef(probs, phe, addcovar=X, se=TRUE,
                    contrasts=cbind(mu=c(1,1,1), a=c(-1, 0, 1), d=c(0, 1, 0)))
    coef_lmm0 <- scan1coef(probs, phe, kinship, X, se=TRUE, hsq=0,
                    contrasts=cbind(mu=c(1,1,1), a=c(-1, 0, 1), d=c(0, 1, 0)))
    expect_equal(coef_hk, coef_lmm0)

    # same with contrasts, weights
    coef_hk <- scan1coef(probs, phe, addcovar=X, weights=weights, se=TRUE,
                    contrasts=cbind(mu=c(1,1,1), a=c(-1, 0, 1), d=c(0, 1, 0)))
    coef_lmm0 <- scan1coef(probs, phe, kinship, X, weights=weights, se=TRUE, hsq=0,
                    contrasts=cbind(mu=c(1,1,1), a=c(-1, 0, 1), d=c(0, 1, 0)))
    expect_equal(coef_hk, coef_lmm0)


    # check LMM with hsq not 0
    coef_lmm <- scan1coef(probs, phe, kinship, X, weights=weights, se=TRUE, zerosum=FALSE)

    ### compare to regress::regress()
    # k <- 2*kinship
    # v <- diag(1/weights)
    # out0_regress <- regress::regress(phe ~ X, ~ k + v, identity=FALSE)
    # p <- probs[[1]][,,6]
    # hsq <- out0_regress$sigma[1]/sum(out0_regress$sigma)
    # V <- hsq*k + (1-hsq)*v
    # out1_regress <- regress::regress(phe ~ -1 + p + X, ~ V, identity=FALSE)
    # co <- cbind(coef_lmm[6,], attr(coef_lmm, "SE")[6,])

})

context("subset cross2, calc_genoprob, sim_geno, viterbi")

test_that("subset.cross2 works (riself)", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))

    # possible problems in indexes
    expect_error(grav2[,c(18:19,"X")])
    expect_warning(grav2[,c(1,2,18)])
    expect_equal(grav2[,1:2], suppressWarnings(grav2[,c(1:2,18)]))
    expect_error(grav2[201:250,])
    expect_warning(grav2[101:250,])
    expect_equal(grav2[101:162,], suppressWarnings(grav2[101:250,]))

    chr <- c(3,5)
    grav2sub <- grav2[,chr]

    expected <- grav2
    expected$geno <- expected$geno[chr]
    expected$gmap <- expected$gmap[chr]
    expected$is_x_chr <- expected$is_x_chr[chr]

    expect_equal(grav2sub, expected)

    ind <- c("79", "100", "123", "160")
    grav2subA <- grav2sub[ind,]
    grav2subB <- grav2[ind,chr]

    for(i in seq(along=expected$geno))
        expected$geno[[i]] <- expected$geno[[i]][ind,,drop=FALSE]
    expected$pheno <- expected$pheno[ind,,drop=FALSE]
    expected$cross_info <- expected$cross_info[ind,,drop=FALSE]
    expected$is_female <- expected$is_female[ind]

    expect_equal(grav2subA, expected)
    expect_equal(grav2subB, expected)

    # subset cross with individuals that aren't present
    expect_warning(
        expect_equal(grav2sub[c(ind, "karl", "fisher"),], expected)
    )

})

test_that("subset.cross2 works (F2)", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    chr <- rep(FALSE, 20)
    chr[c(4, 20)] <- TRUE
    ironsub <- iron[,chr]

    expected <- iron
    expected$geno <- expected$geno[chr]
    expected$gmap <- expected$gmap[chr]
    expected$pmap <- expected$pmap[chr]
    expected$is_x_chr <- expected$is_x_chr[chr]

    expect_equal(ironsub, expected)

    ind <- c("79", "100", "123", "160", "284")
    ironsubA <- ironsub[ind,]
    ironsubB <- iron[ind,chr]

    for(i in seq(along=expected$geno))
        expected$geno[[i]] <- expected$geno[[i]][ind,,drop=FALSE]
    expected$pheno <- expected$pheno[ind,,drop=FALSE]
    expected$covar <- expected$covar[ind,,drop=FALSE]
    expected$cross_info <- expected$cross_info[ind,,drop=FALSE]
    expected$is_female <- expected$is_female[ind]

    expect_equal(ironsubA, expected)
    expect_equal(ironsubB, expected)

    # subset cross with individuals that aren't present
    expect_warning(
        expect_equal(ironsub[c(ind, "karl", "fisher"),], expected)
    )

    # test negative subscripts
    expect_equal(iron[,"-X"], iron[,1:19])
    expect_equal(iron[,"-4"], iron[,c(1:3,5:19,"X")])
    expect_equal(iron[c("-189", "-190"),], iron[-c(189,190),])

    # test ind logical
    ind_logic <- rep(FALSE, n_ind(iron))
    ind_logic[1:20] <- TRUE
    ind_num <- 1:20
    ind_char <- rownames(iron$geno[[1]])[1:20]
    expect_equal(iron[ind_logic,], iron[ind_char,])
    expect_equal(iron[ind_num,], iron[ind_char,])

    # test case of different numbers of genotyped and phenotyped individuals
    iron_orig <- iron
    iron$pheno <- iron$pheno[1:10,]
    expect_error(iron[1:20,])
    iron$covar <- iron$covar[1:20,]
    expect_error(iron[rep(TRUE, n_ind_geno(iron)),])
    ind <- ind_ids_geno(iron)[1:10]
    expect_equal(iron[ind,], iron_orig[ind,])

})

test_that("subset.calc_genoprob works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    ironsub <- iron[,c(4,"X")]

    map <- insert_pseudomarkers(iron$gmap, step=5)

    pr <- calc_genoprob(iron, map, err=0.01)
    prsub <- pr[,"X"]

    expected <- unclass(pr)["X"]
    class(expected) <- class(pr)
    attr(expected, "is_x_chr") <- attr(pr, "is_x_chr")["X"]
    for(obj in c("alleles", "alleleprobs", "crosstype"))
        attr(expected, obj) <- attr(pr, obj)

    expect_equal(prsub, expected)

    ind <- c("5", "50", "55", "280")
    prsub <- pr[ind, "X"]

    expected[["X"]] <- expected[["X"]][ind,,,drop=FALSE]

    expect_equal(prsub, expected)

    # subset with individuals that aren't present
    expect_warning(
        expect_equal(pr[c(ind, "karl", "fisher"), "X"], expected)
    )

})

test_that("subset.sim_geno works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    ironsub <- iron[,c(4,"X")]

    map <- insert_pseudomarkers(iron$gmap, step=5)

    dr <- sim_geno(iron, map, err=0.01)
    drsub <- dr[,"X"]

    expected <- unclass(dr)["X"]
    class(expected) <- class(dr)
    attr(expected, "is_x_chr") <- attr(dr, "is_x_chr")["X"]
    for(obj in c("alleles", "crosstype"))
        attr(expected, obj) <- attr(dr, obj)

    expect_equal(drsub, expected)

    ind <- c("5", "50", "55", "280")
    drsub <- dr[ind, "X"]

    expected[["X"]] <- expected[["X"]][ind,,,drop=FALSE]

    expect_equal(drsub, expected)

    # subset with individuals that aren't present
    expect_warning(
        expect_equal(dr[c(ind, "karl", "fisher"), "X"], expected)
    )

})

test_that("subset.calc_genoprob works with reduction to grid and/or allele prob", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    ironsub <- iron[,c(5,8,"X")]

    map <- insert_pseudomarkers(ironsub$gmap, step=2.5)

    pr <- calc_genoprob(ironsub, map, err=0.002)

    grid <- calc_grid(ironsub$gmap, step=2.5)
    pr_grid <- probs_to_grid(pr, grid)
    pr_a <- genoprob_to_alleleprob(pr)

    set.seed(20150918)
    ind <- sample(rownames(iron$geno[[1]]), 10, replace=FALSE)
    chr <- c("5", "X")
    pr_sub <- pr[ind,chr]

    pr_grid_sub <- pr_grid[ind,chr]
    pr_sub_grid <- probs_to_grid(pr_sub, grid[chr])
    expect_equal(pr_grid_sub, pr_sub_grid)

    pr_a_sub <- pr_a[ind,chr]
    pr_sub_a <- genoprob_to_alleleprob(pr_sub)
    expect_equal(pr_a_sub, pr_sub_a)

})

test_that("subset.viterbi works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    ironsub <- iron[,c(4,"X")]

    map <- insert_pseudomarkers(iron$gmap, step=5)
    g <- viterbi(iron, map, err=0.01)
    gsub <- g[,"X"]

    expected <- unclass(g)["X"]
    class(expected) <- class(g)
    attr(expected, "is_x_chr") <- attr(g, "is_x_chr")["X"]
    attr(expected, "crosstype") <- attr(g, "crosstype")
    attr(expected, "alleles") <- attr(g, "alleles")

    expect_equal(gsub, expected)

    ind <- c("5", "50", "55", "280")
    gsub <- g[ind, "X"]

    expected[["X"]] <- expected[["X"]][ind,,drop=FALSE]

    expect_equal(gsub, expected)

    # subset with individuals that aren't present
    expect_warning(
        expect_equal(g[c(ind, "karl", "fisher"), "X"], expected)
    )

})

# insert_pseudomarkers

#' Insert pseudomarkers into a marker map
#'
#' Insert pseudomarkers into a map of genetic markers
#'
#' @param map A list of numeric vectors; each vector gives marker
#' positions for a single chromosome.
#' @param step Distance between pseudomarkers and markers; if
#' `step=0` no pseudomarkers are inserted.
#' @param off_end Distance beyond terminal markers in which to insert
#' pseudomarkers.
#' @param stepwidth Indicates whether to use a fixed grid
#' (`stepwidth="fixed"`) or to use the maximal distance between
#' pseudomarkers to ensure that no two adjacent markers/pseudomarkers
#' are more than `step` apart.
#' @param pseudomarker_map A map of pseudomarker locations; if provided the
#' `step`, `off_end`, and `stepwidth` arguments are
#' ignored.
#' @param tol Tolerance for determining whether a pseudomarker would duplicate a marker position.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A list like the input `map` with pseudomarkers
#' inserted. Will also have an attribute `"is_x_chr"`, taken
#' from the input `map`.
#'
#' @details If `stepwidth="fixed"`, a grid of pseudomarkers is
#' added to the marker map.
#'
#' If `stepwidth="max"`, a minimal set of pseudomarkers are
#' added, so that the maximum distance between adjacent markers or
#' pseudomarkers is at least `step`. If two adjacent markers are
#' separated by less than `step`, no pseudomarkers will be added
#' to the interval. If they are more then `step` apart, a set of
#' equally-spaced pseudomarkers will be added.
#'
#' If `pseudomarker_map` is provided, then the `step`,
#' `off_end`, and `stepwidth` arguments are ignored, and the
#' input `pseudomarker_map` is taken to be the set of
#' pseudomarker positions.
#'
#' @export
#' @keywords utilities
#' @seealso [calc_genoprob()], [calc_grid()]
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' gmap_w_pmar <- insert_pseudomarkers(iron$gmap, step=1)
insert_pseudomarkers <-
function(map, step=0, off_end=0, stepwidth=c("fixed", "max"),
         pseudomarker_map=NULL, tol=0.01, cores=1)
{
    if(is.null(map))
        stop("Input map is NULL")

    # is input not a map but a cross2 object?
    cl <- class(map)
    if(length(cl)==1 && cl=="cross2") {
        map <- map$gmap
        if(is.null(map))
            stop("Input is a cross2 object, but no genetic map found.")
    }

    # check inputs
    stepwidth <- match.arg(stepwidth)
    if(is.null(pseudomarker_map)) {
        pseudomarker_map <- vector("list", length(map))
    }
    else { # pseudomarker map provided
        if(!is.list(pseudomarker_map))
            stop("pseudomarker_map should be a list")

        if(length(map) == length(pseudomarker_map) &&
           is.null(names(pseudomarker_map))) { # pseudomarker map has no naems, but two are same length
            warning("pseudomarker_map has no names; assuming the same as map")
            names(pseudomarker_map) <- names(map) # assume they are the same
        }

        if(is.null(names(pseudomarker_map)))
            stop("pseudomarker_map needs names")

        pmar_names <- unlist(lapply(pseudomarker_map, names))
        map_names <- unlist(lapply(map, names))
        if(any(pmar_names %in% map_names))
            stop("pseudomarker_map names can't appear among the map names")
        if(length(pmar_names) != length(unique(pmar_names)))
            stop("pseudomarker_map names must be unique")

        if(!all(names(pseudomarker_map) %in% names(map))) {
            extra_chr <- names(pseudomarker_map)[!(names(pseudomarker_map) %in% names(map))]
            stop("Some chr in pseudomarker_map are not in map: ", paste(extra_chr, collapse=", "))
        }


        if(!all(names(map) %in% names(pseudomarker_map))) { # some chr in map but not pseudomarker_map
            map[names(pseudomarker_map)] <-
                insert_pseudomarkers(map[names(pseudomarker_map)],
                                     pseudomarker_map=pseudomarker_map,
                                     tol=tol, cores=cores)
            return(map)
        }
    }
    if(!is_nonneg_number(step)) stop("step should be a single non-negative number")
    if(!is_nonneg_number(off_end)) stop("off_end should be a single non-negative number")
    if(!is_nonneg_number(tol)) stop("tol should be a single non-negative number")

    chr <- names(map)

    # set up cluster
    cores <- setup_cluster(cores)

    # do the work, possible with multiple CPU
    newmap <- cluster_lapply(cores, seq(along=map), function(i)
        insert_pseudomarkers_onechr(map[[i]], step=step, off_end=off_end, stepwidth=stepwidth,
                                    pseudomarker_map[[i]], tol=tol, paste0("c", chr[i], ".loc")))
    names(newmap) <- chr

    attr(newmap, "is_x_chr") <- attr(map, "is_x_chr")
    newmap
}



insert_pseudomarkers_onechr <-
function(map, step=0, off_end=0, stepwidth=c("fixed", "max"),
         pseudomarker_map=NULL, tol=0.01, pmar_stem="loc")
{
    if(any(is.na(map))) stop("map values can't be missing")

    if(tol < 0) stop("tol should be >= 0")

    if(is.null(pseudomarker_map)) {
        stepwidth <- match.arg(stepwidth)
        if(step < 0) stop("step should be >= 0")
        if(step > 0 && tol >= step) stop("tol should be << step")
        if(off_end < 0) stop("off_end should be >= 0")
    }
    else {
        if(any(is.na(pseudomarker_map)))
            stop("pseudomarker_map values can't be missing")
        stepwidth <- "custom"
    }

    switch(stepwidth,
           fixed=insert_pseudomarkers_grid(map, step, off_end, tol, pmar_stem),
           max=insert_pseudomarkers_minimal(map, step, off_end, tol, pmar_stem),
           custom=combine_markers_with_pseudomarkers(map, pseudomarker_map, tol, pmar_stem))
}

# pseudomarker map as grid, ignoring markers
#
# map = vector of marker positions; names = marker names
# step = distance between pseudomarkers
# off_end = amount to go beyond the terminal markers
# tol = tolerance for determining whether a marker hits a pseudomarker
# pmar_stem = leading part of name of pseudomarkers
insert_pseudomarkers_grid <-
function(map, step, off_end=0, tol=0.01, pmar_stem="loc")
{
    if(step==0) return(map)

    # locations of pseudomarkers
    pmar <- seq(min(map)-off_end, max(map)+off_end, by=step)

    combine_markers_with_pseudomarkers(map, pmar, tol, pmar_stem, step)
}

# combine the markers in "map" with the pseudomarker positions in
# "pmar". treat the positions in "pmar" as a grid, and return a
# logical vector of which values in output are on the grid.
combine_markers_with_pseudomarkers <-
function(map, pmar_map, tol=0.01, pmar_stem="loc", step=0.1)
{
    # for each pseudomarker, distance to nearest marker
    d <- abs(outer(map, pmar_map, "-"))
    mind <- apply(d, 2, min)

    # omit pseudomarkers that hit a marker
    to_omit <- (mind < tol)
    pmar_map <- pmar_map[!to_omit]
    if(length(pmar_map) == 0) return(map)

    if(is.null(names(pmar_map)))
        names(pmar_map) <- create_pseudomarker_names(pmar_map, step, pmar_stem)

    # add pseudomarkers to markers and sort
    sort(c(map, pmar_map))
}

# create names for pseudomarkers
create_pseudomarker_names <-
function(pmar_map, step, pmar_stem="loc")
{
    digits <- ceiling(-log10(step))
    digits <- ifelse(digits < 0, 0, digits)

    pmar_nam <- paste0(pmar_stem, round(pmar_map, digits))

   while(any_duplicates(pmar_nam)) {
       digits <- digits + 1
       pmar_nam <- paste0(pmar_stem, round(pmar_map, digits))
   }
   pmar_nam
}

any_duplicates <-
    function(x)
    length(unique(x)) < length(x)


# pseudomarker map, minimal number of pseudomarkers to add
#
# map = vector of marker positions; names = marker names
# step = min distance between markers or pseudomarkers
# off_end = amount to go beyond the terminal markers
# tol = tolerance for determining whether a marker hits a pseudomarker
# pmar_stem = leading part of name of pseudomarkers
insert_pseudomarkers_minimal <-
function(map, step, off_end=0, tol=0.01, pmar_stem="loc")
{
    if(step==0) return(map)

    # distance between markers
    d <- diff(map)
    if(any(d < 0))
        stop("marker positions should be sorted")

    pmar_map <- c(seq(min(map), min(map)-off_end, by=-step)[-1],
                  seq(max(map), max(map)+off_end, by=step)[-1])

    if(length(map) == 1) { # just one marker
        map <- c(map, pmar_map)
        o <- order(map)
        map <- map[o]
        return(map)
    }

    # intervals needing pseudomarkers (d = distances between adjacent markers
    int2fill <- which(d > step)

    pmar_map <- c(pmar_map, unlist(lapply(int2fill, function(a) {
        n <- ceiling((map[a+1]-map[a])/step)+1
        seq(map[a], map[a+1], length.out=n)[-c(1, n)] })))

    if(length(pmar_map) == 0) {
        return(map)
    }

    names(pmar_map) <- create_pseudomarker_names(pmar_map, step, pmar_stem)

    # add pseudomarkers to markers and sort
    sort(c(map, pmar_map))
}

context("rbind and cbind sim_geno")

test_that("rbind.sim_geno works for grav", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    grav2 <- grav2[1:30,1:2]
    map <- insert_pseudomarkers(grav2$gmap, step=5)
    draws <- sim_geno(grav2, map, error_prob=0.002, n_draws=8)
    drawsA <- draws[1:5,]
    drawsB <- draws[6:12,]
    drawsC <- draws[13:20,]
    drawsAB <- draws[1:12,]
    drawsABC <- draws[1:20,]
    drawsBACA <- draws[c(6:12, 1:5, 13:20, 1:5),]

    expect_equal(rbind(drawsA, drawsB), drawsAB)
    expect_equal(rbind(drawsA, drawsB, drawsC), drawsABC)
    expect_equal(rbind(drawsB, drawsA, drawsC, drawsA), drawsBACA)

})


test_that("rbind.sim_geno works for iron", {

    skip_on_cran()

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    draws <- sim_geno(iron, map, error_prob=0.002, n_draws=5)
    drawsA <- draws[2:20,]
    drawsB <- draws[41:60,]
    drawsC <- draws[102:201,]
    drawsAB <- draws[c(2:20,41:60),]
    drawsABC <- draws[c(2:20,41:60,102:201),]
    drawsBACA <- draws[c(41:60,2:20,102:201,2:20),]

    expect_equal(rbind(drawsA, drawsB), drawsAB)
    expect_equal(rbind(drawsA, drawsB, drawsC), drawsABC)
    expect_equal(rbind(drawsB, drawsA, drawsC, drawsA), drawsBACA)

})

test_that("cbind.simgeno for grav", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    grav2 <- grav2[1:30,]
    map <- insert_pseudomarkers(grav2$gmap, step=5)
    draws <- sim_geno(grav2[1:10,], map, error_prob=0.002, n_draws=8)
    drawsA <- draws[,1:2]
    drawsB <- draws[,5]
    drawsC <- draws[,3:4]
    drawsAB <- draws[,c(1:2,5)]
    drawsABC <- draws[,c(1,2,5,3,4)]
    drawsBACA <- draws[,c(5,1:2,3:4,1:2)]

    expect_equal(cbind(drawsA, drawsB), drawsAB)
    expect_equal(cbind(drawsA, drawsB, drawsC), drawsABC)
    expect_equal(cbind(drawsB, drawsA, drawsC, drawsA), drawsBACA)

})


test_that("cbind.sim_geno works for iron", {

    skip_on_cran()

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- insert_pseudomarkers(iron$gmap, step=1)
    draws <- sim_geno(iron[6:21,], map, error_prob=0.002, n_draws=6)
    drawsA <- draws[,2:3]
    drawsB <- draws[,c(4,5,8)]
    drawsC <- draws[,c(19,"X")]
    drawsAB <- draws[,c(2:3,4,5,8)]
    drawsABC <- draws[,c(2:3,4,5,8,19,"X")]
    drawsBACA <- draws[,c(4,5,8,2,3,19,"X",2,3)]

    expect_equal(cbind(drawsA, drawsB), drawsAB)
    expect_equal(cbind(drawsA, drawsB, drawsC), drawsABC)
    expect_equal(cbind(drawsB, drawsA, drawsC, drawsA), drawsBACA)

})

context("construction of pseudomarker map")
suppressMessages(library(qtl))

test_that("insert_pseudomarkers gives an error if the input is NULL", {
    expect_error(insert_pseudomarkers(NULL, step=1))

    library(qtl)
    data(hyper)
    expect_error(insert_pseudomarkers(hyper$gmap, step=1, error.prob=0.01))

})


test_that("grid-based version works in simple case", {

    # equally-spaced map
    map <- seq(0, 50, by=2.5)

    # step = marker distance
    pmap <- insert_pseudomarkers(list("1"=map), step=2.5, off_end=0, stepwidth="fixed")
    expect_equivalent(map, pmap[[1]])

    # step = 1
    pmap <- insert_pseudomarkers(list("1"=map), step=1, off_end=0, stepwidth="fixed")
    pmap_qtl <- qtl::create.map(map, step=1, off.end=0)
    expect_equivalent(pmap[[1]], pmap_qtl)

})

test_that("minimal version works in simple case", {

    # equally-spaced map
    map <- seq(0, 50, by=2.5)

    # step = marker distance
    pmap <- insert_pseudomarkers(list("1"=map), step=2.5, off_end=0, stepwidth="max")
    expect_equivalent(map, pmap[[1]])

    # step = 1
    pmap <- insert_pseudomarkers(list("1"=map), step=1, off_end=0, stepwidth="max")
    expect_equivalent(pmap[[1]], seq(0, 50, by=5/6))

})

test_that("minimal version works in more realistic case", {

    data(hyper)
    map <- qtl::pull.map(hyper, chr=1)

    pmap <- insert_pseudomarkers(map, step=1.55, off_end=4, stepwidth="max")

    expected <- c(0.2, 1.75, 3.3, 4.79090909091818, 6.28181818183636, 7.77272727275454,
                  9.26363636367273, 10.7545454545909, 12.2454545455091, 13.7363636364273,
                  15.2272727273455, 16.7181818182636, 18.2090909091818, 19.7000000001,
                  21.1555555556667, 22.6111111112333, 24.0666666668, 25.5222222223667,
                  26.9777777779333, 28.4333333335, 29.8888888890667, 31.3444444446333,
                  32.8000000002, 33.90000000025, 35.0000000003, 36.10000000035,
                  37.2000000004, 38.6333333337667, 40.0666666671333, 41.5000000005,
                  42.60000000055, 43.7000000006, 43.7000000007, 45.075000000725,
                  46.45000000075, 47.825000000775, 49.2000000008, 50.550000000825,
                  51.90000000085, 53.250000000875, 54.6000000009, 56.0142857152,
                  57.4285714295, 58.8428571438, 60.2571428581, 61.6714285724, 63.0857142867,
                  64.500000001, 65.6000000010333, 66.7000000010667, 67.8000000011,
                  68.85000000115, 69.9000000012, 71.3666666679, 72.8333333346,
                  74.3000000013, 75.4000000014, 76.72000000142, 78.04000000144,
                  79.36000000146, 80.68000000148, 82.0000000015, 82.0000000016,
                  82.0000000017, 82.0000000018, 83.4333333351667, 84.8666666685333,
                  86.3000000019, 87.84000000192, 89.38000000194, 90.92000000196,
                  92.46000000198, 94.000000002, 95.45333333534, 96.90666666868,
                  98.36000000202, 99.81333333536, 101.2666666687, 102.72000000204,
                  104.17333333538, 105.62666666872, 107.08000000206, 108.5333333354,
                  109.98666666874, 111.44000000208, 112.89333333542, 114.34666666876,
                  115.8000000021, 117.3500000021, 118.9000000021)
    pmap_names <- c("c1.loc0", "c1.loc2", "D1Mit296", "c1.loc5", "c1.loc6", "c1.loc8", "c1.loc9",
                    "c1.loc11", "c1.loc12", "c1.loc14", "c1.loc15", "c1.loc17", "c1.loc18", "D1Mit123",
                    "c1.loc21", "c1.loc23", "c1.loc24", "c1.loc26", "c1.loc27", "c1.loc28", "c1.loc30",
                    "c1.loc31", "D1Mit156", "c1.loc34", "D1Mit178", "c1.loc36", "D1Mit19",
                    "c1.loc39", "c1.loc40", "D1Mit7", "c1.loc43", "D1Mit46", "D1Mit132", "c1.loc45",
                    "c1.loc46", "c1.loc48", "D1Mit334", "c1.loc51", "c1.loc52", "c1.loc53", "D1Mit305",
                    "c1.loc56", "c1.loc57", "c1.loc59", "c1.loc60", "c1.loc62", "c1.loc63", "D1Mit26",
                    "c1.loc66", "c1.loc67", "D1Mit94", "c1.loc69", "D1Mit218", "c1.loc71", "c1.loc73",
                    "D1Mit100", "D1Mit102", "c1.loc77", "c1.loc78", "c1.loc79", "c1.loc81", "D1Mit14",
                    "D1Mit105", "D1Mit159", "D1Mit267", "c1.loc83", "c1.loc85", "D1Mit15",
                    "c1.loc88", "c1.loc89", "c1.loc91", "c1.loc92", "D1Mit456", "c1.loc95", "c1.loc97",
                    "c1.loc98", "c1.loc100", "c1.loc101", "c1.loc103", "c1.loc104", "c1.loc106", "c1.loc107",
                    "c1.loc109", "c1.loc110", "c1.loc111", "c1.loc113", "c1.loc114", "D1Mit155",
                    "c1.loc117", "c1.loc119")

    expect_equivalent(pmap[[1]], expected)
    expect_equal(names(pmap[[1]]), pmap_names)
})

test_that("insert_pseudomarkers works with a custom pseudomarker map", {

    data(hyper)
    map <- qtl::pull.map(hyper)

    set.seed(99735998)
    pseudomarker_map <- vector("list", length(map))
    names(pseudomarker_map) <- names(map)
    for(i in seq(along=map)) {
        n.pmar <- 10
        pseudomarker_map[[i]] <- sort(runif(n.pmar, 0, max(map[[i]])))
        names(pseudomarker_map[[i]]) <- paste0("c", names(map)[i], ".loc", 1:n.pmar)
    }

    combined_map <- insert_pseudomarkers(map, pseudomarker_map = pseudomarker_map)

    expect_equal(sapply(map, length) + sapply(pseudomarker_map, length),
                 sapply(combined_map, length))

    for(i in seq(along=map)) {
        combined_map_2 <- sort(c(map[[i]], pseudomarker_map[[i]]))
        expect_equivalent(combined_map_2, combined_map[[i]])
        expect_equal(names(combined_map_2), names(combined_map[[i]]))
    }

    expect_equal(names(map), names(combined_map))

})

test_that("insert_pseudomarkers gives distinct pseudomarker names with iron data", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    pmap <- insert_pseudomarkers(iron$gmap, step=1)

    expect_true(all(vapply(pmap$map, function(a) !any_duplicates(names(a)), TRUE)))

})

test_that("insert_pseudomarkers works with multi-core", {

    skip_if(isnt_karl(), "this test only run locally")

    data(hyper)
    map <- qtl::pull.map(hyper)

    set.seed(99735998)
    pseudomarker_map <- vector("list", length(map))
    names(pseudomarker_map) <- names(map)
    for(i in seq(along=map)) {
        n.pmar <- 10
        pseudomarker_map[[i]] <- sort(runif(n.pmar, 0, max(map[[i]])))
        names(pseudomarker_map[[i]]) <- paste0("c", names(map)[i], ".loc", 1:n.pmar)
    }

    combined_map <- insert_pseudomarkers(map, pseudomarker_map = pseudomarker_map, cores=2)
    expect_equal(sapply(map, length) + sapply(pseudomarker_map, length),
                 sapply(combined_map, length))

    for(i in seq(along=map)) {
        combined_map_2 <- sort(c(map[[i]], pseudomarker_map[[i]]))
        expect_equivalent(combined_map_2, combined_map[[i]])
        expect_equal(names(combined_map_2), names(combined_map[[i]]))
    }

    expect_equal(names(map), names(combined_map))

})


test_that("insert_pseudomarkers works with partial pseudomarker_map", {

    library(qtl)
    data(hyper)

    map <- pull.map(hyper)

    # single chromosome
    pmap <- list("2"=c(pmar1=10, pmar2=20))
    result <- insert_pseudomarkers(map, pseudomarker_map=pmap)
    expect_equal(result[-2], map[-2])
    expect_equal(result[[2]], sort(c(map[[2]], pmap[["2"]])))

    # a second chromosome
    pmap <- c(pmap, list("X"=c(pmar3=5)))
    result <- insert_pseudomarkers(map, pseudomarker_map=pmap)
    expect_equal(result[-c(2,20)], map[-c(2,20)])
    expect_equal(result[["2"]], sort(c(map[["2"]], pmap[["2"]])))
    expect_equal(result[["X"]], sort(c(map[["X"]], pmap[["X"]])))

    # warning if missing chr names but is the same length
    pmap <- lapply(map, function(a) setNames(a+1, paste0(names(a), "plus1")))
    names(pmap) <- NULL
    expect_warning(insert_pseudomarkers(map, pseudomarker_map=pmap))

    # error if names in common
    pmap <- lapply(map, function(a) a+1)
    expect_error(insert_pseudomarkers(map, pseudomarker_map=pmap))

})

# read_cross2
#' Read QTL data from files
#'
#' Read QTL data from a set of files
#'
#' @param file Character string with path to the
#' [YAML](https://yaml.org) or [JSON](https://json.org) file containing all of the control
#' information. This could instead be a zip file containing all of the
#' data files, in which case the contents are unzipped to a temporary
#' directory and then read.
#' @param quiet If `FALSE`, print progress messages.
#'
#' @return Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#'
#' @details
#' A control file in [YAML](https://yaml.org) or
#' [JSON](https://json.org) format contains information
#' about basic parameters as well as the names of the series of
#' data files to be read. See the
#' [sample data files](https://kbroman.org/qtl2/pages/sampledata.html) and the
#' [vignette describing the input file format](https://kbroman.org/qtl2/assets/vignettes/input_files.html).
#'
#' @export
#' @keywords IO
#' @seealso [read_pheno()], [write_control_file()],
#' sample data files at <https://kbroman.org/qtl2/pages/sampledata.html>
#' and <https://github.com/rqtl/qtl2data>
#'
#' @examples
#' \dontrun{
#' yaml_file <- "https://kbroman.org/qtl2/assets/sampledata/grav2/grav2.yaml"
#' grav2 <- read_cross2(yaml_file)
#' }
#' zip_file <- system.file("extdata", "grav2.zip", package="qtl2")
#' grav2 <- read_cross2(zip_file)
read_cross2 <-
function(file, quiet=TRUE)
{
    if(length(grep("\\.zip$", file)) > 0) { # zip file
        dir <- qtl2_temp_dir()

        if(is_web_file(file)) {
            tmpfile <- tempfile()
            if(!quiet) message(" - downloading ", file, "\n       to ", tmpfile)
            utils::download.file(file, tmpfile, quiet=TRUE)
            file <- tmpfile
            on.exit(unlink(tmpfile))
        }

        if(!quiet) message(" - unzipping ", file, "\n       to ", dir)
        file <- path.expand(file)
        stop_if_no_file(file)
        unzipped_files <- utils::unzip(file, exdir=dir)

        # remove temporary directory on exit
        on.exit({
            if(!quiet) message(" - cleaning up")
            unlink(dir, recursive=TRUE)
        }, add=TRUE)

        # ignore "__MACOSX/._" files
        unzipped_files <- grep("__MACOSX/._", unzipped_files, fixed=TRUE, invert=TRUE, value=TRUE)

        if(any(grepl("\\.yaml$", unzipped_files))) {
            file <- grep("\\.yaml$", unzipped_files, value=TRUE)
            if(length(file) > 1)
                stop("The zip file contains multiple yaml files")
            if(any(grepl("\\.json$", unzipped_files)))
                warning("The zip file contains both YAML and JSON files; using the YAML file.")
        }
        else if(any(grepl("\\.json$", unzipped_files))) {
            file <- grep("\\.json$", unzipped_files, value=TRUE)
            if(length(file) > 1)
                stop("The zip file contains multiple json files")
        }
        else {
            stop('No ".yaml" or ".json" control file found')
        }
    }

    # directory containing the data
    file <- path.expand(file)
    dir <- dirname(file)

    # load the control file
    stop_if_no_file(file)
    control <-  read_control_file(file)

    # for keeping track of use of control stuff
    used_control <- rep(FALSE, length(control))
    names(used_control) <- names(control)
    used_control["sep"] <- used_control["na.strings"] <-
        used_control["comment.char"] <- used_control["description"] <-
            used_control["comments"] <- TRUE

    # grab cross type
    if("crosstype" %in% names(control)) {
        output <- list(crosstype=control$crosstype)
        used_control["crosstype"] <- TRUE # indicate that we used it
    }
    else
        stop("crosstype not found")

    # grab genotype encodings or use defaults
    if("genotypes" %in% names(control)) {
        genotypes <- control$genotypes
        used_control["genotypes"] <- TRUE # indicate that we used it
    }
    else
        genotypes <- list(A=1, H=2, B=3, D=4, C=5)

    # grab the major bits
    sections <- c("geno", "gmap", "pmap", "pheno", "covar", "phenocovar", "founder_geno")
    for(section in sections) {
        if(section %in% names(control)) {
            used_control[section] <- TRUE # indicate that we used it
            if(!quiet) message(" - reading ", section)

            # transposed?
            tr <- paste0(section, "_transposed")
            used_control[tr] <- TRUE # indicate that we used it
            tr <- tr %in% names(control) && control[[tr]]

            filename <- control[[section]]

            if(length(filename)==1) { # single file
                filename <- file.path(dir, filename)
                stop_if_no_file(filename)

                # read file
                sheet <- fread_csv(filename, na.strings=control$na.strings, sep=control$sep,
                                   comment.char=control$comment.char, transpose=tr,
                                   rownames_included=TRUE)
            }
            else { # vector of files
                # add dir to paths
                filenames <- vapply(filename, function(a) file.path(dir, a), "")

                # check that the all exist
                lapply(filenames, stop_if_no_file)

                if(section=="gmap" || section=="pmap") {
                    # rbind the chromosomes together
                    sheet <- NULL
                    for(filename in filenames)
                        sheet <- rbind(sheet,
                                       fread_csv(filename, na.strings=control$na.strings,
                                                 sep=control$sep, comment.char=control$comment.char,
                                                 transpose=tr, rownames_included=TRUE))
                }
                else {
                    # read all of the files and cbind(), matching on row names
                    sheet <- fread_mult_csv(filenames, na.strings=control$na.strings,
                                            sep=control$sep, comment.char=control$comment.char,
                                            transpose=tr)
                }
            }
            if(length(unique(colnames(sheet))) != ncol(sheet))
                warning("Duplicate column names in ", section, " data")

            # change genotype codes and convert phenotypes to numeric matrix
            if(section=="geno" || section=="founder_geno") {
                if(!quiet) message(" - encoding ", section)
                sheet <- recode_geno(sheet, genotypes)
            }
            else if(section=="pheno") {
                sheet <- pheno2matrix(sheet)
            }
            else if(section=="phenocovar") {
                sheet <- convert_numeric_columns(sheet)
            }

            output[[section]] <- sheet
        }
    }

    # stop if no genotypes
    if(!("geno" %in% names(output)))
       stop("No genotype data found.")

    # omit markers not in common among geno, gmap, pmap, founder_geno
    output <- drop_incomplete_markers( output )

    # pull out a map; make it numeric
    if("gmap" %in% names(output))
        map <- output$gmap
    else if("pmap" %in% names(output))
        map <- output$pmap
    else stop("Need a genetic or physical marker map")
    map[,2] <- as.numeric(map[,2])
    map <- reorder_map_table(map)

    # split maps by chr
    maps <- c("gmap", "pmap")
    for(section in maps) {
        if(section %in% names(output)) {
            if(!quiet) message(" - splitting up ", section)
            output[[section]] <- split_map(output[[section]], unique(map[,1]))
        }
    }

    # split genotypes by chromosome
    geno <- c("geno", "founder_geno")
    for(section in geno) {
        if(section %in% names(output)) {
            if(!quiet) message(" - splitting up ", section)
            output[[section]] <- split_geno(output[[section]], map)
        }
    }

    # X chr?
    chr <- names(output$geno)
    output$is_x_chr <- rep(FALSE, length(chr))
    names(output$is_x_chr) <- chr
    if("x_chr" %in% names(control)) {
        x_chr <- control$x_chr # name of X chromosome
        used_control["x_chr"] <- TRUE # indicate that we used it

        # deal with case of multiple chromosomes, and with not all present in genotypes
        found <- x_chr[x_chr %in% names(output$is_x_chr)]
        output$is_x_chr[found] <- TRUE
        notfound <- x_chr[!(x_chr %in% names(output$is_x_chr))]
        if(length(notfound) == 1) {
            warning('Chromosome "', notfound, '" not in genotype data but in control file as x_chr')
        } else if(length(notfound) > 1) {
            warning('Chromosomes ', paste0('"', notfound, '"', collapse=", "),
                    ' not in genotype data but in control file as x_chr')
        }
    }

    # sex
    output$is_female <- convert_sex(control$sex, output$covar, control$sep,
                                    control$comment.char, dir, quiet=quiet)
    if(is.null(output$is_female)) { # missing; assume all TRUE
        output$is_female <- rep(TRUE, nrow(output$geno[[1]]))
        names(output$is_female) <- rownames(output$geno[[1]])

        # give warning if sex is needed
        if(suppressMessages(!check_is_female_vector(output$crosstype, logical(0), any(output$is_x_chr)))) {
            warning("No sex information; assuming all female")
        }
    } else {
        used_control["sex"] <- TRUE # indicate that we used it
    }

    # cross_info
    output$cross_info <- convert_cross_info(control$cross_info, output$covar, control$sep,
                                            control$comment.char, dir, quiet=quiet)
    if(is.null(output$cross_info)) { # missing; make a 0-column matrix
        output$cross_info <- matrix(0L, ncol=0, nrow=nrow(output$geno[[1]]))
        rownames(output$cross_info) <- rownames(output$geno[[1]])
    }
    used_control["cross_info"] <- TRUE # indicate that we used it

    # alleles?
    n_alleles <- nalleles(output$crosstype)
    if("alleles" %in% names(control)) {
        output$alleles <- control$alleles
        used_control["alleles"] <- TRUE # indicate that we used it
        if(n_alleles != length(output$alleles)) {
            warning("length(alleles) [", length(output$alleles),
                    "] != expected number [", n_alleles, "]")

            if(length(output$alleles) < n_alleles) {
                # pad with a bunch of letters
                output$alleles <- c(output$alleles,
                                    rep(LETTERS[!(LETTERS %in% output$alleles)], n_alleles))
            }
            # trim to n_alleles
            output$alleles <- output$alleles[seq_len(n_alleles)]
        }
    } else {
        output$alleles <- LETTERS[seq_len(n_alleles)]
    }

    class(output) <- "cross2"

    # force genotypes, is_female, and cross_info to be aligned
    gfc <- align_geno_sex_cross(output$geno,
                                output$is_female,
                                output$cross_info)
    output$geno <- gfc$geno
    output$is_female <- gfc$is_female
    output$cross_info <- gfc$cross_info

    # force same markers in geno, founder_geno, and cross_info
    for(i in seq(along=output$geno)) {
        gmar <- colnames(output$geno[[i]])
        mmar <- names(output$gmap[[i]])
        pmar <- names(output$pmap[[i]])
        if(is.null(pmar)) pmar <- mmar
        if(is.null(mmar)) mmar <- pmar
        fmar <- colnames(output$founder_geno[[i]])
        if(is.null(fmar)) fmar <- gmar

        mar <- find_common_ids(gmar, mmar, pmar, fmar)
        tmar <- length(unique(c(gmar, mmar, pmar, fmar)))
        if(length(mar) < tmar)
            warning("Omitting ", tmar - length(mar),
                    " markers on chr ", names(output$geno)[i],
                    " that ", ifelse(tmar-length(mar)==1, "is", "are"),
                    " missing position, genotypes, or founder genotypes.")

        output$geno[[i]] <- output$geno[[i]][,mar,drop=FALSE]
        output$gmap[[i]] <- output$gmap[[i]][mar]
        if(!is.null(output$pmap))
            output$pmap[[i]] <- output$pmap[[i]][mar]
        if(!is.null(output$founder_geno))
            output$founder_geno[[i]] <- output$founder_geno[[i]][,mar,drop=FALSE]
    }

    # if pheno and geno have same individuals, make same order
    if(!is.null(output$pheno)) {
        ind_g <- rownames(output$geno[[1]])
        ind_p <- rownames(output$pheno)
        if(length(ind_g) == length(ind_p)) {
            if(!all(ind_g == ind_p) && all(sort(ind_g) == sort(ind_p))) {
                output$pheno <- output$pheno[ind_g,,drop=FALSE]
            }
        }
    }
    # same with geno and covar
    if(!is.null(output$covar)) {
        ind_g <- rownames(output$geno[[1]])
        ind_p <- rownames(output$covar)
        if(length(ind_g) == length(ind_p)) {
            if(!all(ind_g == ind_p) && all(sort(ind_g) == sort(ind_p))) {
                output$covar <- output$covar[ind_g,,drop=FALSE]
            }
        }
    }

    check_cross2(output) # run all the checks

    if(any(!used_control))
        warning("Unused control information: ",
                paste0('"', names(used_control)[!used_control], '"', collapse=", "))

    # add is_x_chr as attributes to gmap and pmap
    if("is_x_chr" %in% names(output)) {
        for(obj in c("gmap", "pmap")) {
            if(obj %in% names(output))
                attr(output[[obj]], "is_x_chr") <- output$is_x_chr
        }
    }

    output
}

# convert genotype data, using genotype encodings
# genotypes is a list with names = code in data
#                     and values = new numeric code
recode_geno <-
function(geno, genotypes)
{
    newgeno <- geno <- as.matrix(geno)

    if(any(unlist(genotypes)==0))
        stop("Can't encode genotypes as 0, that's used for missing values.")

    # genotype codes
    gnames <- names(genotypes)
    codes <- genotypes

    # any mismatches?
    gch <- as.character(geno)
    mismatch <- !is.na(gch) & is.na(match(gch,gnames))
    if(any(mismatch)) {
        warning(sum(mismatch), " genotypes treated as missing: ",
                paste0('"', unique(gch[mismatch]), '"', collapse=", "))
        newgeno[mismatch] <- NA
    }

    # re-code
    for(g in gnames) {
        if(g != genotypes[[g]]) # if g==genotypes[[g]], don't need to switch
            newgeno[!is.na(geno) & geno==g] <- genotypes[[g]]
    }

    # turn missing values to 0s
    newgeno[is.na(newgeno)] <- "0"

    storage.mode(newgeno) <- "integer"

    newgeno
}

# convert a phenotype data.frame to a matrix of doubles
pheno2matrix <-
function(pheno)
{
    nc <- ncol(pheno)
    dn <- dimnames(pheno)
    pheno <- matrix(as.numeric(unlist(pheno)), ncol=nc)
    dimnames(pheno) <- dn

    storage.mode(pheno) <- "double"

    pheno
}

# make the first column of a data frame the row names
firstcol2rownames <-
function(mat, name="")
{
    # assume the first column is the ID
    idcolumn <- mat[,1]

    # look for duplicates
    check4duplicates(idcolumn, name)

    # drop that column and make it the row names
    mat <- mat[,-1,drop=FALSE]
    rownames(mat) <- as.character(idcolumn)

    mat
}

# check a list of IDs for duplicates
check4duplicates <-
function(ids, name="")
{
    if(name != "") name <- paste(" in", name)
    dup <- duplicated(ids)
    if(any(dup)) {
        stop("Not all ids", name, " are unique: ",
             paste0('"', unique(ids[dup]), '"', collapse=", "))
    }
    TRUE
}

# split genotype data into list of chromosomes
split_geno <-
function(geno, map)
{
    gmark <- colnames(geno)
    mmark <- rownames(map)

    m <- match(gmark, mmark)
    if(any(is.na(m)))
        stop("Some markers in genotype data not found in map: ",
             paste0('"', gmark[is.na(m)], '"', collapse=", "))
    map <- map[mmark %in% gmark,,drop=FALSE] # toss from map any markers not in geno

    # sort map
    map <- map[order(factor(map[,1], levels=unique(map[,1])), map[,2], seq_len(nrow(map))),,drop=FALSE]

    geno <- geno[,rownames(map),drop=FALSE]

    chr <- map[,1]
    uchr <- unique(chr)
    newgeno <- vector("list", length(uchr))
    names(newgeno) <- uchr
    for(i in uchr) {
        newgeno[[i]] <- geno[,which(chr==i), drop=FALSE]
    }

    newgeno
}

# split a data frame with chr, pos (and marker names as rows)
#   into a list of chromosomes
split_map <-
function(map, chr_names=NULL)
{
    map <- reorder_map_table(map, chr_names=chr_names)

    pos <- as.numeric(map[,2])
    chr <- map[,1]
    uchr <- unique(chr)

    names(pos) <- rownames(map)

    lapply(split(pos, factor(chr, uchr)), sort)
}

# grab sex information
convert_sex <-
function(sex_control, covar, sep, comment.char, dir, quiet=TRUE)
{
    if(is.null(sex_control)) return(NULL)

    if("covar" %in% names(sex_control)) { # sex within the covariates
        if(!(sex_control$covar %in% colnames(covar)))
            stop('sex column "', sex_control$covar, '" not found in covar')
        sex <- covar[,sex_control$covar, drop=FALSE]
    }
    else if("file" %in% names(sex_control)) { # look for file
        if(!quiet) message(" - reading sex")
        file <- file.path(dir, sex_control$file)
        stop_if_no_file(file)
        sex <- fread_csv(file, sep=sep, na.strings=NULL, comment.char=comment.char,
                         rownames_included=TRUE)
    }
    else return(NULL)

    # convert to vector
    id <- rownames(sex)
    sex <- sex[,1]
    names(sex) <- id

    # missing values?
    if(any(is.na(sex))) {
        stop(sum(is.na(sex)), " missing sexes (sex can't be missing).")
    }

    # grab the rest of sex_control as conversion codes
    codes <- sex_control[is.na(match(names(sex_control), c("file", "covar")))]
    codes <- convert_sexcodes(codes) # convert to 0 and 1
    sexcode <- names(codes)

    if(length(codes)==0) { # no codes, pass over as is
        storage.mode(sex) <- "integer"
        return(sex==0)
    }

    # any mismatches?
    sexch <- as.character(sex)
    mismatch <- !is.na(sexch) & is.na(match(sexch,sexcode))
    if(any(mismatch)) {
        stop(sum(mismatch), " sexes don't match the codes (sex can't be missing): ",
                paste0('"', unique(sexch[mismatch]), '"', collapse=", "))
    }

    # re-code
    newsex <- sex
    for(code in sexcode)
        newsex[sex==code] <- codes[[code]]

    storage.mode(newsex) <- "integer"

    (newsex==0)
}

convert_sexcodes <-
function(codes)
{
    result <- match(substr(tolower(codes), 1, 1), c("f", "m")) - 1

    names(result) <- names(codes)
    result
}


# grab cross_info
convert_cross_info <-
function(cross_info_control, covar, sep, comment.char, dir, quiet=TRUE)
{
    if(is.null(cross_info_control)) return(NULL)

    if(!is.list(cross_info_control)) { # provided file name directly?
        if(file.exists(file.path(dir, cross_info_control)))
            cross_info_control <- list(file=cross_info_control)
    }

    if("file" %in% names(cross_info_control)) { # look for file
        if(!quiet) message(" - reading cross_info")

        file <- file.path(dir, cross_info_control$file)
        stop_if_no_file(file)
        cross_info <- fread_csv(file, sep=sep, comment.char=comment.char, rownames_included=TRUE)

        if(any(is.na(cross_info)))
            stop(sum(is.na(cross_info)), " missing values in cross_info (cross_info can't be missing.")
    }
    else if("covar" %in% names(cross_info_control)) { # cross_info within the covariates
        if(!all(cross_info_control$covar %in% colnames(covar))) {
            not_found <- cross_info_control$covar[!(cross_info_control$covar %in% colnames(covar))]
            not_found <- paste0('"', not_found, '"', collapse=", ")
            stop('cross_info column ', not_found, '" not found in covar')
        }
        cross_info <- covar[,cross_info_control$covar, drop=FALSE]
    }
    else return(NULL)

    # make it a matrix
    cross_info <- as.matrix(cross_info)

    if(any(is.na(cross_info))) {
        stop(sum(is.na(cross_info)), " missing values in cross_info (cross_info can't be missing).")
    }

    # grab the rest of cross_info_control as conversion codes
    codes <- cross_info_control[is.na(match(names(cross_info_control), c("file", "covar")))]
    cicode <- names(codes)

    if(length(codes)==0) { # no codes, pass over as is
        storage.mode(cross_info) <- "integer"

        # again check for missing values
        if(any(is.na(cross_info))) {
            stop(sum(is.na(cross_info)), " missing values in cross_info (cross_info can't be missing, and must be integers).")
        }

        return(cross_info)
    }

    if(ncol(cross_info) > 1) {
        warning("cross_info has more than one column but uses conversion codes (which are intended for a single column)")
    }

    # any mismatches?
    cich <- as.character(cross_info)
    mismatch <- !is.na(cich) & is.na(match(cich,cicode))
    if(any(mismatch)) {
        stop(sum(mismatch), " cross_info vals don't match the codes (cross_info can't be missing): ",
                paste0('"', unique(cich[mismatch]), '"', collapse=", "))
    }
    # re-code
    newci <- cross_info
    for(code in cicode)
        newci[cross_info==code] <- codes[[code]]

    storage.mode(newci) <- "integer"

    newci
}

# is file on web (starts with http://, https://, or file://)
is_web_file <-
function(file)
{
    patterns <- c("^http://", "^https://", "^file://", "^ftp://")
    any(vapply(patterns, function(a,b) grepl(a, b), logical(1), file))
}

stop_if_no_file <-
function(filename)
{
    if(is_web_file(filename)) return(TRUE)
    if(!file.exists(filename))
        stop('file "', filename, '" does not exist.')
}


# read multiple CSV files and cbind results
fread_mult_csv <-
    function(filenames, sep=",", na.strings=c("NA", "-"), comment.char="#", transpose=FALSE)
{
    result <- NULL
    for(file in filenames) {
        this <- fread_csv(file, sep=sep, na.strings=na.strings,
                          comment.char=comment.char, transpose=transpose,
                          rownames_included=TRUE)

        if(is.null(result)) result <- this
        else result <- cbind_expand(result, this)
    }
    result
}

# read control file, as either YAML or JSON
read_control_file <-
function(filename)
{
    # ends in yaml?
    if(grepl("\\.yaml$", filename)) {
        control <- yaml::yaml.load_file(filename)
    }
    else if(grepl("\\.json$", filename)) {
        control <- jsonlite::fromJSON(readLines(filename))
    }
    else stop(paste('Control file', filename, 'should have extension ".yaml" or ".json"'))

    # default values for sep, na.strings, and comment.char
    if(is.null(control$sep)) control$sep <- ","
    if(is.null(control$na.strings)) control$na.strings <- "NA"
    if(is.null(control$comment.char)) control$comment.char <- "#"

    control
}

# read header lines
read_header <-
    function(filename, comment.char="#")
{
    con <- file(filename, "rt")
    on.exit(close(con))
    header <- NULL
    while(length(line <- readLines(con, 1))>0) {
        if(grepl(paste0("^", comment.char), line))
            header <- c(header, line)
        else break
    }
    header
}

# get expected dimensions from header
# (expecting header rows like "# nrow 5201" and "# ncol 59")
extract_dim_from_header <-
    function(header, comment.char="#")
{
    result <- c(NA, NA)
    nam <- c("nrow", "ncol")
    for(i in seq(along=nam)) {
        matchhead <- grepl(paste0("^", comment.char, "\\s*", nam[i], "\\s*\\d+"), header)
        if(sum(matchhead) > 1) # more than one matches
            stop('Multiple header lines with "', nam[i], '"')
        if(sum(matchhead) == 0) next
        spl <- strsplit(header[matchhead], "\\s+")[[1]]
        number <- spl[length(spl)]
        number <- gsub("(^[0-9]+).*$", "\\1", number)
        result[i] <- as.numeric(number)
    }

    result
}

# drop markers that do not appear in all of geno, gmap, pmap, founder_geno
drop_incomplete_markers <-
    function( output )
{
    geno_mar <- colnames( output$geno )

    if(!is.null(output$gmap)) {
        pos <- suppressWarnings( as.numeric(output$gmap[,2]) )
        if(any(is.na(pos))) {
            n_na <- sum(is.na(pos))
            warning(n_na, ifelse(n_na==1, " marker", " markers"), " with missing genetic map position.")
        }
        gmap_mar <- rownames(output$gmap)[!is.na(pos)]
        output$gmap <- output$gmap[!is.na(pos),]
    }
    else gmap_mar <- geno_mar

    if(!is.null(output$pmap)) {
        pos <- suppressWarnings( as.numeric(output$pmap[,2]) )
        if(any(is.na(pos))) {
            n_na <- sum(is.na(pos))
            warning(n_na, ifelse(n_na==1, " marker", " markers"), " with missing physical map position.")
        }
        pmap_mar <- rownames(output$pmap)[!is.na(pos)]
        output$pmap <- output$pmap[!is.na(pos),]
    }
    else pmap_mar <- geno_mar

    if(!is.null(output$founder_geno))
        fg_mar <- colnames(output$founder_geno)
    else fg_mar <- geno_mar

    tot_mar <- unique(c(geno_mar, gmap_mar, pmap_mar, fg_mar))
    keep_mar <- find_common_ids(geno_mar, gmap_mar, pmap_mar, fg_mar)

    if(length(keep_mar) < length(tot_mar)) {
        n_drop <- length( tot_mar[!(tot_mar %in% keep_mar)] )
        warning("Omitting ", n_drop, " markers that are not in both genotypes and maps")

        output$geno <- output$geno[,geno_mar %in% keep_mar, drop=FALSE]
        if(!is.null(output$gmap))
            output$gmap <- output$gmap[gmap_mar %in% keep_mar, , drop=FALSE]
        if(!is.null(output$pmap))
            output$pmap <- output$pmap[pmap_mar %in% keep_mar, , drop=FALSE]
        if(!is.null(output$founder_geno))
            output$founder_geno <- output$founder_geno[, fg_mar %in% keep_mar, drop=FALSE]
    }
    if(length(keep_mar) == 0) {
        stop("No markers in common between map and genotypes")
    }

    output
}

reorder_map_table <-
    function(map_tab, chr_col=1, pos_col=2, chr_names=NULL)
{
    chr <- map_tab[,chr_col]

    if(is.null(chr_names)) chr_names <- unique(chr)

    chr <- factor(chr, levels=chr_names)
    pos <- suppressWarnings( as.numeric(map_tab[,pos_col]) )

    map_tab[ order(chr, pos, seq_len(nrow(map_tab))), , drop=FALSE]
}

# create unique temporary directory name
qtl2_temp_dir <-
    function(n_letters=15, initial_bit="qtl2")
{
    file.path(tempdir(),
              paste0(initial_bit, "_", paste(sample(c(letters, 0:9), n_letters, replace=TRUE), collapse="")) )

}

context("reduce gaps in a map")

test_that("reduce_map_gaps works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    map <- iron$gmap

    expect_equal(reduce_map_gaps(map, 1000), map)

    expected <- map
    expected[[1]][3] <- expected[[1]][2] + 40
    expected[[4]][2] <- expected[[4]][1] + 40
    expected[[5]][2] <- expected[[5]][1] + 40

    expect_equal(reduce_map_gaps(map, 40), expected)

    map2 <- map
    map2[[2]][3:5] <- map2[[2]][3:5]+50
    expected[[2]][3:5] <- expected[[2]][3:5] - diff(expected[[2]][2:3]) + 40
    expect_equal(reduce_map_gaps(map2, 40), expected)

})

context("is_phase_known")

test_that("is_phase_known works for grav2 and iron", {

    grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
    expect_true(is_phase_known(grav2))

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    expect_false(is_phase_known(iron))

})


test_that(".is_phase_known works for all cross types", {

    for(crosstype in c("bc", "haploid", "dh", "riself", "risib",
                       "riself4", "riself8", "riself16", "risib4", "risib8",
                       "magic19", "dh6", "dof1")) {
        expect_true(.is_phase_known(crosstype))
    }

    for(crosstype in c("f2", "ail", "ail3", "do", "hs")) {
        expect_false(.is_phase_known(crosstype))
    }

})

context("find gaps in a map")

test_that("find_map_gaps works", {

    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
    expect_equal(find_map_gaps(iron$gmap, 40),
                 data.frame(chr=c("1", "4", "5"),
                            left_marker=c("D1Mit80", "D4Mit2", "D5Mit11"),
                            left_index=c(2,1,1),
                            right_marker=c("D1Mit17", "D4Mit352", "D5Mit30"),
                            right_index=c(3,2,2),
                            gap=c(59, 42.7, 44.8),
                            stringsAsFactors=FALSE))

    expect_equal(find_map_gaps(iron$gmap, 38),
                 data.frame(chr=c("1", "4", "5", "12"),
                            left_marker=c("D1Mit80", "D4Mit2", "D5Mit11", "D12Mit88"),
                            left_index=c(2,1,1,1),
                            right_marker=c("D1Mit17", "D4Mit352", "D5Mit30", "D12Mit134"),
                            right_index=c(3,2,2,2),
                            gap=c(59, 42.7, 44.8, 38.2),
                            stringsAsFactors=FALSE))

    expect_equal(find_map_gaps(iron$gmap, 1000),
                 data.frame(chr=character(0),
                            left_marker=character(0),
                            left_index=numeric(0),
                            right_marker=character(0),
                            right_index=numeric(0),
                            gap=numeric(0),
                            stringsAsFactors=FALSE))

})

# code related to clusters

# test if input is a prepared cluster (vs. just a number)
is_cluster <-
    function(cores)
{
    inherits(cores, "cluster") && inherits(cores, "SOCKcluster")
}

# number of cores being used
n_cores <-
    function(cores)
{
    if(is_cluster(cores)) return( length(cores) )
    cores
}

# set up a cluster
setup_cluster <-
    function(cores, quiet=TRUE)
{
    if(is_cluster(cores)) return(cores)

    if(is.null(cores) || is.na(cores)) cores <- 1
    if(cores==0) cores <- max(1, parallel::detectCores()-1, na.rm=TRUE) # if 0, detect cores
    if(is.na(cores)) cores <- 1

    if(cores > 1 && Sys.info()[1] == "Windows") { # windows doesn't support mclapply
        cores <- parallel::makeCluster(cores)
        # the following calls on.exit() in the function that called this one
        # see https://stackoverflow.com/a/20998531
        do.call("on.exit",
                list(quote(parallel::stopCluster(cores))),
                envir=parent.frame())
    }
    cores
}

# run code by cluster (generalizes lapply, parLapply, and mclapply)
# (to deal with different methods on different architectures)
# if cores==1, just use lapply
cluster_lapply <-
    function(cores, ...)
{
    if(is_cluster(cores)) { # cluster object; use mclapply
        return( parallel::parLapply(cores, ...) )
    } else {
        if(cores==1) return( lapply(...) )
        return( parallel::mclapply(..., mc.cores=cores) )
    }
}

#' Combine matrices by columns, expanding and aligning rows
#'
#' This is like [base::cbind()] but using row names to align the rows and expanding
#' with missing values if there are rows in some matrices but not others.
#'
#' @param ... A set of matrices or data frames
#'
#' @return The matrices combined by columns, using row names to align the rows, and expanding with missing values if there are rows in some matrices but not others.
#'
#' @export
#'
#' @examples
#' df1 <- data.frame(x=c(1,2,3,NA,4), y=c(5,8,9,10,11), row.names=c("A", "B", "C", "D", "E"))
#' df2 <- data.frame(w=c(7,8,0,9,10), z=c(6,NA,NA,9,10), row.names=c("A", "B", "F", "C", "D"))
#' cbind_expand(df1, df2)
cbind_expand <-
    function(...)
{
    input <- list(...)

    if(length(input)<2) return(input[[1]])

    # check IDs
    id <- lapply(input, rownames)
    if(any(vapply(id, is.null, TRUE)))
        stop("All input matrices must have rownames (containing individual IDs)")
    if(any(vapply(id, function(a) length(unique(a)) != length(a), TRUE)))
        stop("Some input matrices have duplicate rownames")
    uid <- unique(unlist(id))

    for(i in seq(along=input)) {
        not_in <- !(uid %in% id[[i]])
        if(!any(not_in)) next
        missing <- uid[not_in]

        # create new rows
        new_rows <- matrix(NA, ncol=ncol(input[[i]]), nrow=length(missing))
        dimnames(new_rows) <- list(missing, colnames(input[[i]]))

        input[[i]] <- rbind(input[[i]], new_rows)
    }

    for(i in seq(along=input)) {
        if(i==1)
            input[[1]] <- input[[1]][uid,,drop=FALSE]
        else
            input[[1]] <- cbind(input[[1]], input[[i]][uid,,drop=FALSE])
    }

    input[[1]]
}


# this is the version that was used in qtl2scan, with no attempt to align the rows
cbind_expand_noalign <-
    function(...)
{
    input <- list(...)

    if(length(input)<2) return(input[[1]])

    if(!all(vapply(input, is.matrix, TRUE)))
        stop("Not all inputs are matrices")

    nrow <- vapply(input, nrow, 1)
    max_nrow <- max(nrow)

    for(i in seq_along(input)) {
        rownames(input[[i]]) <- NULL # strip off the rownames
        if(nrow[i] < max_nrow) { # need to pad
            new_rows <- matrix(NA, ncol=ncol(input[[i]]), nrow=max_nrow - nrow[i])
            colnames(new_rows) <- colnames(input[[i]])
            input[[i]] <- rbind(input[[i]], new_rows)
        }
    }

    do.call("cbind", input)
}

context("cbind_expand")

test_that("cbind_expand works", {

    set.seed(20151109)
    x <- data.frame(x=sample(LETTERS[1:5]), y=sample(letters[6:10]), stringsAsFactors=FALSE)
    rownames(x) <- c("1", "3", "4", "5", "6")
    y <- data.frame(a=runif(4), b=rnorm(4))
    rownames(y) <- c("5", "2", "9", "4")

    expected <- data.frame(x=c(x$x, NA, NA), y=c(x$y, NA, NA),
                           a=rep(NA, 7), b=rep(NA, 7),
                           stringsAsFactors=FALSE)
    rownames(expected) <- c("1", "3", "4", "5", "6", "2", "9")
    expected[rownames(y),"a"] <- y$a
    expected[rownames(y),"b"] <- y$b


    expect_equal(cbind_expand(x, y), expected)

    # reverse inputs
    expect_equal(cbind_expand(y, x), expected[c(4,6,7,3,1,2,5), c(3,4,1,2)])

    # duplicate inputs
    expect_equal(cbind_expand(x, x), cbind(x, x))
    expect_equal(cbind_expand(y, y), cbind(y, y))

    # expect error due to duplicate row names
    expect_error(cbind_expand( rbind("1"=1:3, "1"=4:6), rbind("1"=4:9) ) )


    # add a third input
    z <- data.frame(c=sample(LETTERS[11:17]), d=sample(1L:5L, 7, replace=TRUE),
                    stringsAsFactors=FALSE)
    rownames(z) <- c("3", "1", "9", "4", "6", "5", "8")

    expected <- data.frame(x=c(x$x, NA, NA, NA), y=c(x$y, NA, NA, NA),
                           a=rep(NA, 8), b=rep(NA, 8),
                           c=rep("", 8), d=rep(NA, 8),
                           stringsAsFactors=FALSE)
    rownames(expected) <- c("1", "3", "4", "5", "6", "2", "9", "8")
    expected[rownames(y),"a"] <- y$a
    expected[rownames(y),"b"] <- y$b
    expected[rownames(z),"d"] <- z$d
    expected[expected$c=="", "c"] <- NA
    expected[rownames(z),"c"] <- z$c
    expect_equal(cbind_expand(x, y, z), expected)


    # a few more simple tests
    df1 <- data.frame(x=c(1,2,3,NA,4), y=c(5,8,9,10,11), row.names=c("A", "B", "C", "D", "E"))
    df2 <- data.frame(w=c(7,8,0,9,10), z=c(6,NA,NA,9,10), row.names=c("A", "B", "F", "C", "D"))

    df1n2 <- data.frame(x=c(1,2,3,NA,4,NA), y=c(5,8,9,10,11,NA),w=c(7,8,9,10,NA,0),z=c(6,NA,9,10,NA,NA),
                        row.names=c("A", "B", "C", "D", "E","F"))
    df2n1 <- df1n2[c(1,2,6,3:5), c(3:4,1:2)]

    expect_equal(cbind_expand(df1, df2), df1n2)
    expect_equal(cbind_expand(df2, df1), df2n1)

    expect_equal(cbind_expand(df1, df1), cbind(df1, df1))
    expect_equal(cbind_expand(df2, df2), cbind(df2, df2))

})

#' Split vector into batches
#'
#' Split a vector into batches, each no longer than `batch_size` and
#' creating at least `n_cores` batches, for use in parallel
#' calculations.
#'
#' @param vec A vector to be split into batches
#' @param batch_size Maximum size for each batch
#' @param n_cores Number of compute cores, to be used as a minimum number of batches.
#'
#' @return A list of vectors, each no longer than `batch_size`, and with at least `n_cores` componenets.
#'
#' @export
#' @importFrom parallel detectCores
#' @seealso [batch_cols()]
#' @examples
#' vec_split <- batch_vec(1:304, 50, 8)
#' vec_split2 <- batch_vec(1:304, 50)
batch_vec <-
    function(vec, batch_size=NULL, n_cores=1)
{
    n <- length(vec)

    if(n_cores==0) n_cores <- parallel::detectCores()

    if(is.null(batch_size)) n_batches <- n_cores
    else {
        n_batches <- ceiling(n/batch_size)
        if(n_batches < n_cores) n_batches <- n_cores
    }

    n_per_batch <- rep(floor(n/n_batches), n_batches)
    d <- n - sum(n_per_batch)
    if(d >= 1)
        n_per_batch[seq_len(d)] <- n_per_batch[seq_len(d)]+1

    result <- split(vec, rep(seq_len(n_batches), n_per_batch))
    names(result) <- NULL
    result
}

# calc_errorlod
#' Calculate genotyping error LOD scores
#'
#' Use the genotype probabilities calculated with
#' [calc_genoprob()] to calculate genotyping error LOD
#' scores, to help identify potential genotyping errors (and problem
#' markers and/or individuals).
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param probs Genotype probabilities as calculated from [calc_genoprob()].
#' @param quiet If `FALSE`, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A list of matrices of genotyping error LOD scores. Each
#'     matrix corresponds to a chromosome and is arranged as
#'     individuals x markers.
#'
#' @details
#'   Let \eqn{O_k}{O[k]} denote the observed marker genotype at position
#'  \eqn{k}, and \eqn{g_k}{g[k]} denote the corresponding true underlying
#'  genotype.
#'
#'  Following Lincoln and Lander (1992), we calculate
#'  LOD = \eqn{log_{10} [ Pr(O_k | g_k = O_k) / Pr(O_k | g_k \ne O_K) ]}{
#'  log10[ Pr(O[k] | g[k] = O[k]) / Pr(O[k] | g[k] != O[k]) ]}
#'
#' @references
#' Lincoln SE, Lander ES (1992) Systematic detection of errors in genetic linkage data. Genomics 14:604--610.
#'
#' @export
#' @keywords utilities
#' @seealso [calc_genoprob()]
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' probs <- calc_genoprob(iron, error_prob=0.002, map_function="c-f")
#' errorlod <- calc_errorlod(iron, probs)
#'
#' # combine into one matrix
#' errorlod <- do.call("cbind", errorlod)

calc_errorlod <-
function(cross, probs, quiet=TRUE, cores=1)
{
    if(is.null(probs)) stop("probs is NULL")

    # check inputs
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')
    alleleprobs <- attr(probs, "alleleprobs")
    if(!is.null(alleleprobs) && alleleprobs)
        stop("probs must contain full genotype probabilities, not haplotype/allele probabilities")

    if(length(cross$geno) != length(probs) ||
       !all(names(cross$geno) == names(probs))) {
        chr1 <- names(cross$geno)
        chr2 <- names(probs)
        chr <- find_common_ids(chr1, chr2)
        if(length(chr)==0)
            stop("cross and probs have no chromosomes in common")
        cross <- cross[,chr]
        probs <- subset(probs, chr=chr)
    }

    # set up cluster; set quiet=TRUE if multi-core
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # deal with missing information
    ind <- rownames(cross$geno[[1]])
    chrnames <- names(cross$geno)
    is_x_chr <- handle_null_isxchr(cross$is_x_chr, chrnames)
    cross$is_female <- handle_null_isfemale(cross$is_female, ind)
    cross$cross_info <- handle_null_isfemale(cross$cross_info, ind)

    founder_geno <- cross$founder_geno
    if(is.null(founder_geno))
        founder_geno <- create_empty_founder_geno(cross$geno)

    dn <- dimnames(probs)

    # align individuals
    if(length(ind) != length(dn[[1]]) ||
       !all(ind == dn[[1]])) {
        ind <- find_common_ids(ind, dn[[1]])
        if(length(ind)==0)
            stop("No individuals in common between cross and probs")
        cross <- cross[ind,]
    }

    by_group_func <- function(i) {
        mn <- colnames(cross$geno[[chr]])
        if(!all(mn %in% dn[[3]][[chr]]))
            stop("Some markers in cross are not in probs on chr ", chrnames[chr])
        pr <- aperm(probs[[chr]][ind,,mn,drop=FALSE], c(2,1,3)) # genotype, ind, marker
        errorlod <- t(.calc_errorlod(cross$crosstype, pr[,group[[i]],,drop=FALSE],
                                     t(cross$geno[[chr]][group[[i]],,drop=FALSE]),
                                     founder_geno[[chr]], cross$is_x_chr[chr], cross$is_female[group[[i]][1]],
                                     t(cross$cross_info[group[[i]][1],])))

    }

    # split individuals into groups with common sex and cross_info
    sex_crossinfo <- paste(cross$is_female, apply(cross$cross_info, 1, paste, collapse=":"), sep=":")
    group <- split(seq(along=sex_crossinfo), sex_crossinfo)
    names(group) <- NULL
    nc <- n_cores(cores)
    while(nc > length(group) && max(sapply(group, length)) > 1) { # successively split biggest group in half until there are as many groups as cores
        mx <- which.max(sapply(group, length))
        g <- group[[mx]]
        group <- c(group, list(g[seq(1, length(g), by=2)]))
        group[[mx]] <- g[seq(2, length(g), by=2)]
    }
    groupindex <- seq(along=group)

    errorlod <- vector("list", length(probs))
    names(errorlod) <- chrnames

    for(chr in seq_len(length(chrnames))) {
        if(!quiet) message("Chr ", chrnames[chr])

        temp <- cluster_lapply(cores, groupindex, by_group_func)

        # paste them back together
        d <- vapply(temp, dim, c(0,0))
        nr <- sum(d[1,])
        errorlod[[chr]] <- matrix(nrow=sum(vapply(temp, nrow, 1)),
                                  ncol=ncol(temp[[1]]))
        for(i in groupindex)
            errorlod[[chr]][group[[i]],] <- temp[[i]]

        dimnames(errorlod[[chr]]) <- dimnames(cross$geno[[chr]])
    }

    errorlod
}

#' Create snp information table for a cross
#'
#' Create a table of snp information from a cross, for use with [scan1snps()].
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#'
#' @return A data frame of SNP information with the following columns:
#' * `chr` - Character string or factor with chromosome
#' * `pos` - Position (in same units as in the `"map"`
#'     attribute in `genoprobs`.
#' * `snp` - Character string with SNP identifier (if
#'     missing, the rownames are used).
#' * `sdp` - Strain distribution pattern: an integer, between
#'     1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
#'     binary encoding indicates the founder genotypes
#' SNPs with missing founder genotypes are omitted.
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DO_Recla/recla.zip")
#' recla <- read_cross2(file)
#' snpinfo <- create_snpinfo(recla)
#'
#' # calculate genotype probabilities
#' pr <- calc_genoprob(recla, error_prob=0.002, map_function="c-f")
#'
#' # index the snp information
#' snpinfo <- index_snps(recla$pmap, snpinfo)
#'
#' # sex covariate
#' sex <- setNames((recla$covar$Sex=="female")*1, rownames(recla$covar))
#'
#' # perform a SNP scan
#' out <- scan1snps(pr, recla$pmap, recla$pheno[,"bw"], addcovar=sex, snpinfo=snpinfo)
#'
#' # plot the LOD scores
#' plot(out$lod, snpinfo, altcol="green3")
#' }
#'
#' @seealso [index_snps()], [scan1snps()], [genoprob_to_snpprob()]
#' @export
create_snpinfo <-
    function(cross)
{
    if(!is.cross2(cross)) {
        stop("Input should be a cross2 object")
    }

    if("pmap" %in% names(cross)) map <- cross$pmap
    else if("gmap" %in% names(cross)) map <- cross$gmap
    else stop("cross contains neither pmap nor gmap")

    # convert map to a data frame
    nmar <- vapply(map, length, 1) # no. markers per chromosome
    markers <- unlist(lapply(map, names))
    map <- data.frame(chr=rep(names(map), nmar),
                      pos=unlist(map),
                      snp=markers,
                      stringsAsFactors=FALSE)
    rownames(map) <- map$snp

    # founder genotypes -> SDP
    if(!("founder_geno" %in% names(cross))) {
        stop("cross does not contain founder_geno")
    }
    fg <- do.call("cbind", cross$founder_geno)

    # drop markers with missing founders or only one allele present in founders
    mar2drop <- (colSums(is.na(fg) | fg==0) > 0) |
        apply(fg, 2, function(g) length(unique(g[!is.na(g) & g!=0]))==1)

    # drop markers with missing data; calculate founder strain distribution patterns (SDPs)
    cbind(map[!mar2drop,,drop=FALSE],
          sdp=calc_sdp(t(fg[,!mar2drop])))
}

# zip_datafiles
#' Zip a set of data files
#'
#' Zip a set of data files (in format read by [read_cross2()]).
#'
#' @param control_file Character string with path to the control file
#' ([YAML](https://yaml.org) or [JSON](https://json.org))
#' containing all of the control information.
#' @param zip_file Name of zip file to use. If NULL, we use the
#' stem of `control_file` but with a `.zip` extension.
#' @param overwrite If `TRUE`, overwrite file if it exists. If `FALSE`
#' (the default) and the file exists, stop with an error.
#' @param quiet If `FALSE`, print progress messages.
#'
#' @return Character string with the file name of the zip file that
#' was created.
#'
#' @details The input `control_file` is the control file (in
#' [YAML](https://yaml.org) or [JSON](https://json.org) format)
#' to be read by [read_cross2()].  (See the
#' [sample data files](https://kbroman.org/qtl2/pages/sampledata.html) and the
#' [vignette describing the input file format](https://kbroman.org/qtl2/assets/vignettes/input_files.html).)
#'
#' The [utils::zip()] function is used to do the zipping.
#'
#' The files should all be contained within the directory where the
#' `control_file` sits, or in a subdirectory of that directory.
#' If file paths use `..`, these get stripped by zip, and so the
#' resulting zip file may not work with [read_cross2()].
#'
#' @export
#' @keywords IO
#' @seealso [read_cross2()], sample data files at <https://kbroman.org/qtl2/pages/sampledata.html>
#' @examples
#' \dontrun{
#' zipfile <- file.path(tempdir(), "grav2.zip")
#' zip_datafiles("grav2.yaml", zipfile)
#' }
zip_datafiles <-
function(control_file, zip_file=NULL, overwrite=FALSE, quiet=TRUE)
{
    control_file <- path.expand(control_file)
    if(!(file.exists(control_file)))
        stop("The control file (", control_file, ") doesn't exist.")

    dir <- dirname(control_file)

    move_zip_file <- TRUE
    if(is.null(zip_file)) {
        zip_file <- sub("\\.[a-z]+$", ".zip", basename(control_file))
        move_zip_file <- FALSE
    }

    # read control file
    control <-  read_control_file(control_file)

    # get all of the file names
    sections <- c("geno", "gmap", "pmap", "pheno", "covar", "phenocovar", "founder_geno")
    files <- basename(control_file)
    for(section in sections) {
        if(section %in% names(control))
            files <- c(files, control[[section]])
    }

    # sex and cross_info as files?
    sections <- c("sex", "cross_info")
    for(section in sections) {
        if(section %in% names(control)) {
            if("file" %in% names(control[[section]]))
                files <- c(files, control[[section]][["file"]])
        }
    }

    # flag for quiet
    zip_flags <- ifelse(quiet, "-q", "")

    # move to the directory with the files
    cwd <- getwd()
    on.exit(setwd(cwd)) # move back on exit
    setwd(dir)

    # check for ".." in file paths
    patterns <- c("^\\.\\.\\/", "\\/\\.\\.\\/")
    patterns <- gsub("/", .Platform$file.sep, patterns, fixed=TRUE)
    if(any(grepl(patterns[1], files) | grepl(patterns[2], files))) {
        warning('zip strips ".." from file paths, so ', zip_file,
                ' may not work with read_cross2().')
    }

    # check if file exists
    if(file.exists(zip_file)) {
        if(overwrite) { # first delete it
            unlink(zip_file)
        } else { # refuse to overwrite
            stop("The zip file (", zip_file, ") already exists. Remove it first (or use overwrite=TRUE).")
        }
    }

    # do the zipping
    utils::zip(basename(zip_file), files, flags=zip_flags)

    # if not being zipped into the data file directory, move it
    if(move_zip_file) {
        setwd(cwd)
        file.rename(file.path(dir, basename(zip_file)), zip_file)
    } else {
        zip_file <- file.path(dir, zip_file)
    }

    invisible(zip_file)
}

# subset phasedgeno objects

#' Subsetting phased genotype objects
#'
#' Pull out a specified set of individuals and/or chromosomes from
#' the results of [guess_phase()]
#'
#' @param x Imputed genotypes as output from [guess_phase()], a list of 3d arrays.
#' @param ind A vector of individuals: numeric indices, logical
#' values, or character string IDs
#' @param chr A vector of chromosomes: logical values, or character
#' string IDs. Numbers are interpreted as character string IDs.
#' @param ... Ignored.
#'
#' @return An object of class `"phasedgeno"`, like the input, with the
#' selected individuals and/or chromosomes; see [guess_phase()].
#'
#' @seealso [guess_phase()], [cbind.phasedgeno()], [rbind.phasedgeno()]
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' pr <- calc_genoprob(DOex, error_prob=0.002)
#' m <- maxmarg(pr)
#' ph <- guess_phase(DOex, m)
#' ph <- ph[1:10,] # first ten individuals
#' ph <- ph[,2]  # chromosome 2
#' }
subset.phasedgeno <-
    subset.calc_genoprob

#' @export
#' @rdname subset.phasedgeno
`[.phasedgeno` <-
    function(x, ind=NULL, chr=NULL)
    subset(x, ind, chr)

#' Calculate eigen decomposition of kinship matrix
#'
#' Calculate the eigen decomposition of a kinship matrix, or of a list of such matrices.
#'
#' @param kinship A square matrix, or a list of square matrices.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return The eigen values and the **transposed** eigen vectors,
#' as a list containing a vector `values` and a matrix
#' `vectors`.
#'
#' @details The result contains an attribute `"eigen_decomp"`.
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[1:30,18:19] # subset to 30 individuals and two chromosomes}
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#' K <- calc_kinship(probs)
#'
#' Ke <- decomp_kinship(K)
#'
#' @export
decomp_kinship <-
    function(kinship, cores=1)
{
    # already done?
    if(is_kinship_decomposed(kinship))
        return(kinship) # no need to do it again

    if(is.matrix(kinship)) {
        if(ncol(kinship) != nrow(kinship))
            stop("matrix must be square")
        if(ncol(kinship) == 0)
            stop("matrix has dimension (0,0)")
        return(Rcpp_eigen_decomp(kinship))
    }

    if(!is.list(kinship))
        stop("kinship should be either a square matrix or a list of square matrices")

    cores <- setup_cluster(cores)

    result <- cluster_lapply(cores, kinship, Rcpp_eigen_decomp)
    attr(result, "eigen_decomp") <- TRUE
    result
}

# genoprob_to_alleleprob
#' Convert genotype probabilities to allele probabilities
#'
#' Reduce genotype probabilities (as calculated by
#' [calc_genoprob()]) to allele probabilities.
#'
#' @param probs Genotype probabilities, as calculated from
#' [calc_genoprob()].
#' @param quiet IF `FALSE`, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return An object of class `"calc_genoprob"`, like the input `probs`,
#' but with probabilities collapsed to alleles rather than genotypes. See [calc_genoprob()].
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' gmap_w_pmar <- insert_pseudomarkers(iron, step=1)
#' probs <- calc_genoprob(iron, gmap_w_pmar, error_prob=0.002)
#' allele_probs <- genoprob_to_alleleprob(probs)

genoprob_to_alleleprob <-
    function(probs, quiet=TRUE, cores=1)
{
    if(is.null(probs)) stop("probs is NULL")

    # already converted?
    ap <- attr(probs, "alleleprobs")
    if(!is.null(ap) && ap) return(probs)

    is_x_chr <- attr(probs, "is_x_chr")

    # set up cluster; make quiet=FALSE if cores>1
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores) > 1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # no more messages
    }

    probs_attr <- attributes(probs)

    # alleles attribute?
    alleles <- probs_attr$alleles
    if(is.null(alleles))
        warning("probs has no alleles attribute; guessing allele codes.")

    by_chr_func <- function(chr) {
        if(!quiet) message(" - Chr ", names(probs)[chr])
        result <- aperm(.genoprob_to_alleleprob(attr(probs, "crosstype"),
                                                aperm(probs[[chr]], c(2, 1, 3)), # reorg -> geno x ind x pos
                                                is_x_chr[chr]),
                        c(2, 1, 3)) # reorg back to ind x geno x pos

        # allele names
        dn <- dimnames(probs)
        if(is.null(alleles) || length(alleles) < ncol(result)) {
            alleles <- assign_allele_codes(ncol(result), dn[[2]][[chr]])
        }
        dn[[2]][[chr]] <- alleles
        dimnames(result) <- list(dn[[1]], dn[[2]][[chr]], dn[[3]][[chr]])

        result
    }

    chrID <- names(probs)
    chrs <- seq(along=chrID)

    probs <- cluster_lapply(cores, chrs, by_chr_func) # if cores==1, this uses lapply()
    names(probs) <- chrID

    attr(probs, "crosstype") <- probs_attr$crosstype
    attr(probs, "is_x_chr") <- probs_attr$is_x_chr
    attr(probs, "alleles") <- probs_attr$alleles
    attr(probs, "alleleprobs") <- TRUE
    class(probs) <- c("calc_genoprob", "list")

    probs
}

# functions to summarize scan1 results

#' Find position with maximum LOD score
#'
#' Return data frame with the positions having maximum LOD score for a
#' particular LOD score column
#'
#' @param scan1_output An object of class `"scan1"` as returned by
#' [scan1()].
#' @param map A list of vectors of marker positions, as produced by
#' [insert_pseudomarkers()]. Can also be an indexed SNP info table,
#' as from [index_snps()] or [scan1snps()].
#' @param lodcolumn An integer or character string indicating the LOD
#' score column, either as a numeric index or column name.
#' If `NULL`, return maximum for all columns.
#' @param chr Optional vector of chromosomes to consider.
#' @param na.rm Ignored (take to be TRUE)
#' @param ... Ignored
#'
#' @importFrom stats setNames
#' @export
#'
#' @return If `map` is NULL, the genome-wide maximum LOD score for the selected column is returned.
#' If also `lodcolumn` is NULL, you get a vector with the maximum LOD for each column.
#'
#' If `map` is provided, the return value is a data.frame with three columns: chr, pos, and lod score.
#' But if `lodcolumn` is NULL, you get the maximum for each lod score column, in the format provided by
#' [find_peaks()], so a data.frame with five columns: lodindex, lodcolumn, chr, pos, and lod.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' # maximum of first column
#' max(out, map)
#'
#' # maximum of spleen column
#' max(out, map, lodcolumn="spleen")
#'
#' # maximum of first column on chr 2
#' max(out, map, chr="2")
max_scan1 <-
    function(scan1_output, map=NULL, lodcolumn=1, chr=NULL, na.rm=TRUE, ...)
{
    if(is.null(scan1_output)) stop("scan1_output is NULL")

    # force column names
    if(is.null(colnames(scan1_output))) {
        colnames(scan1_output) <- paste0("lod", seq_len(ncol(scan1_output)))
    }

    if(is.null(lodcolumn)) {
        return(maxall_scan1(scan1_output, map=map, chr=chr, na.rm=na.rm, ...))
    }

    if(length(lodcolumn) > 1) {
        lodcolumn <- lodcolumn[1]
        warning("lodcolumn should have length 1; using the first value")
    }
    if(is.character(lodcolumn)) {
        lodcolumn_num <- match(lodcolumn, colnames(scan1_output))
        if(is.na(lodcolumn_num)) stop('LOD column "', lodcolumn, '" not found.')
        lodcolumn <- lodcolumn_num
    }
    if(lodcolumn < 1 || lodcolumn > ncol(scan1_output)) {
        stop("column [", lodcolumn, "] out of range (should be in 1 - ", ncol(scan1_output), ")")
    }

    if(is.null(map)) {
        if(!is.null(chr)) warning("chr ignored if map is not provided")

        return( setNames( max(scan1_output[,lodcolumn], na.rm=TRUE), colnames(scan1_output)[lodcolumn]) )
    }

    if(is.data.frame(map) && "index" %in% names(map)) { # looks like snpinfo table
        map <- snpinfo_to_map(map)
    }

    # align scan1_output and map
    tmp <- align_scan1_map(scan1_output, map)
    scan1_output <- tmp$scan1
    map <- tmp$map

    if(nrow(scan1_output) != length(unlist(map))) {
        stop("nrow(scan1_output) [", nrow(scan1_output), "] != number of positions in map [",
             length(unlist(map)), "]")
    }

    # to handle output of either scan1() or scan1coef()
    # for coef(), look at the sign
    lod <- unclass(scan1_output)
    sign <- (lod >= 0)*2-1 # sign +1

    if(inherits(scan1_output, "scan1coef")) lod <- abs(lod)

    coln <- colnames(lod)
    mnames <- rownames(lod)

    if(length(lodcolumn) > 1) { # If length > 1, take first value
        warning("lodcolumn should have length 1; one first element used.")
        lodcolumn <- lodcolumn[1]
    }
    if(is.character(lodcolumn)) { # turn column name into integer
        tmp <- match(lodcolumn, coln)
        if(is.na(tmp))
            stop('column "', lodcolumn, '" not found')
        lodcolumn <- tmp
    }
    lod <- lod[,lodcolumn]
    sign <- sign[,lodcolumn]

    thechr <- map2chr(map)
    thepos <- map2pos(map)

    # subset chromosomes
    if(!is.null(chr)) {
        keep <- (thechr %in% chr)
        thechr <- thechr[keep]
        thepos <- thepos[keep]
        mnames <- mnames[keep]
        lod <- lod[keep]*sign[keep] # sign is a kludge to handle scan1coef output
    }

    themax <- which.max(lod)

    result <- data.frame(chr=thechr[themax],
                         pos=thepos[themax],
                         lod=lod[themax],
                         stringsAsFactors=FALSE)
    rownames(result) <- mnames[themax]
    names(result)[3] <- coln[lodcolumn]

    result
}

#' @export
#' @rdname max_scan1
max.scan1 <-
    function(scan1_output, map=NULL, lodcolumn=1, chr=NULL, na.rm=TRUE, ...)
    max_scan1(scan1_output, map, lodcolumn, chr, na.rm, ...)

#' Overall maximum LOD score
#'
#' Find overall maximum LOD score in genome scan results, across all positions and columns.
#'
#' @param scan1_output An object of class `"scan1"` as returned by
#' [scan1()].
#' @param map A list of vectors of marker positions, as produced by
#' [insert_pseudomarkers()].
#' @param chr Optional vector of chromosomes to consider.
#' @param lodcolumn An integer or character string indicating the LOD
#' score column, either as a numeric index or column name.
#' If `NULL`, return maximum for all columns.
#'
#' @export
#' @return A single number: the maximum LOD score across all columns and positions for
#' the selected chromosomes.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' # overall maximum
#' maxlod(out)
#'
#' # maximum on chromosome 2
#' maxlod(out, map, "2")
maxlod <-
    function(scan1_output, map=NULL, chr=NULL, lodcolumn=NULL)
{
    if(is.null(scan1_output)) stop("scan1_output is NULL")

    if(is.null(map) || is.null(chr)) {
        if(!is.null(map) || !is.null(chr))
            stop("Provide both map and chr, or neither.")
    }
    else {
        # subset by chromosome
        scan1_output <- subset(scan1_output, map=map, chr=chr)
    }

    if(!is.null(lodcolumn)) {
        scan1_output <- subset(scan1_output, lodcolumn=lodcolumn)
    }

    # to handle output of either scan1() or scan1coef()
    # for coef(), look at the sign
    if(inherits(scan1_output, "scan1coef")) {
        coef <- unclass(scan1_output)
        sign <- (coef >= 0)*2-1 # sign +1/-1
        coef <- abs(coef)
        maxcoef <- max(coef, na.rm=TRUE)

        # deal with ties
        wh <- which(!is.na(coef) & coef == maxcoef)
        if(length(wh)>1) wh <- sample(wh, 1)

        return(maxcoef * sign[wh])
    }
    else {
        lod <- unclass(scan1_output)
        return(max(lod, na.rm=TRUE))
    }
}

# return maximum for all lod score columns
maxall_scan1 <-
    function(scan1_output, map=NULL, chr=NULL, na.rm=TRUE, ...)
{
    if(is.null(map)) {
        if(!is.null(chr)) warning("chr ignored if map is not provided.")

        return(apply(scan1_output, 2, max, na.rm=na.rm))
    }

    res <- lapply(1:ncol(scan1_output), function(lodcol) max_scan1(scan1_output, map=map, lodcolumn=lodcol,
                                                                   chr=chr, na.rm=na.rm, ...))

    data.frame(lodindex=1:ncol(scan1_output),
               lodcolumn=colnames(scan1_output),
               chr=sapply(res, function(a) a[[1]][1]),
               pos=sapply(res, function(a) a[[2]][1]),
               lod=sapply(res, function(a) a[[3]][1]),
               stringsAsFactors=FALSE)

}

context("max_scan1")

test_that("max_scan1 works for intercross with two phenotypes", {

    # read data
    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    # calculate genotype probabilities
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    # grab phenotypes and covariates; ensure that covariates have names attribute
    pheno <- iron$pheno
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    # perform genome scan
    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

    # maximum of first column
    expected <- data.frame(chr="16",
                           pos=28.6,
                           liver=max(out[,1]),
                           stringsAsFactors=FALSE)
    rownames(expected) <- "c16.loc29"
    expect_equal(max(out,map), expected)

    # maximum of spleen column
    expected <- data.frame(chr="9",
                           pos=56.6,
                           spleen=max(out[,2]),
                           stringsAsFactors=FALSE)
    rownames(expected) <- "c9.loc57"
    expect_equal(max(out, map, lodcolumn="spleen"), expected)

    # maximum of first column on chr 2
    expected <- data.frame(chr="2",
                           pos=56.8,
                           liver=maxlod(subset(out, map, "2", 1)),
                           stringsAsFactors=FALSE)
    rownames(expected) <- "D2Mit17"
    expect_equal(max(out, map, chr="2"), expected)

    # test that it works if output or map are subsetted
    expect_equal( max(out, map, chr="2"), max(subset(out, map, chr="2"), map) )
    expect_equal( max(out, map, chr="2"), max(out, map["2"]) )

    # test that it works if output has rows shuffled
    out_shuffled <- out[sample(1:nrow(out)),,drop=FALSE]
    class(out_shuffled) <- c("scan1", "matrix")
    expect_equal( max(out_shuffled, map), max(out, map))

    expect_equal(max_scan1(out), c(liver= 6.35264382891418))
    expect_equal(max_scan1(out, lodcolumn=2), c(spleen=12.5986057120873))
    expect_equal(max_scan1(out, lodcolumn="spleen"), c(spleen=12.5986057120873))

    # warning if you give lodcolumn as a vector
    expect_warning( expect_equal(max(out, lodcolumn=1:2), c(liver= 6.35264382891418)) )
    expected <- data.frame(chr="16",
                           pos=28.6,
                           liver=max(out[,1]),
                           stringsAsFactors=FALSE)
    rownames(expected) <- "c16.loc29"
    expect_warning( expect_equal( max(out, map, lodcolumn=1:2), expected) )

    # warning if you give chr but not map
    expect_warning( expect_equal( max(out, lodcolumn=2, chr=16), c(spleen=12.5986057120873)) )

    # results for all LOD score columns if lodcolumn=NULL
    expect_equal( max(out, lodcolumn=NULL),
                 c(liver= 6.35264382891418, spleen=12.5986057120873) )


    expected <- data.frame(lodindex=1:ncol(out),
                           lodcolumn=colnames(out),
                           chr=c("16","9"),
                           pos=c(28.6, 56.6),
                           lod=apply(out, 2, max),
                           stringsAsFactors=FALSE)
    rownames(expected) <- NULL
    expect_equal( max(out, map, lodcolumn=NULL), expected )

    expected <- data.frame(lodindex=1:ncol(out),
                           lodcolumn=colnames(out),
                           chr=c("16","16"),
                           pos=c(28.6, 30.6),
                           lod=apply(subset(out, map, chr=16), 2, max),
                           stringsAsFactors=FALSE)
    rownames(expected) <- NULL
    expect_equal( max(out, map, lodcolumn=NULL, chr=16), expected )

    expected <- data.frame(lodindex=1:ncol(out),
                           lodcolumn=colnames(out),
                           chr=c("3","9"),
                           pos=c(25.1, 56.6),
                           lod=apply(subset(out, map, chr=c(3,9)), 2, max),
                           stringsAsFactors=FALSE)
    rownames(expected) <- NULL
    expect_equal( max(out, map, lodcolumn=NULL, chr=c(3,9)), expected )
    expect_equal( max(out, map, lodcolumn=NULL, chr=c(9,3)), expected )

    # max handles NA (here, just guessing that I'm unlikely to NA-out the max
    set.seed(1331427)
    out_na <- out
    out_na[5,2] <- NA
    out_na[sample(nrow(out_na), 10),1] <- NA
    expect_equal(max(out), max(out_na))
    expect_equal(max(out, map), max(out_na, map))
    expect_equal(max(out, lodcolumn=NULL), max(out_na, lodcolumn=NULL, na.rm=TRUE))
    expect_equal(max(out, map, lodcolumn=NULL), max(out_na, map, lodcolumn=NULL, na.rm=TRUE))

    # max gives error if lodcolumn out of range
    expect_error(max(out, lodcolumn="blah"), 'LOD column "blah" not found.')
    expect_error(max(out, lodcolumn=0),
                 "column [0] out of range (should be in 1 - 2)", fixed=TRUE)
    expect_error(max(out, lodcolumn=3),
                 "column [3] out of range (should be in 1 - 2)", fixed=TRUE)

    # no column names
    colnames(out) <- colnames(out_na) <- NULL
    expect_equal(max(out), max(out_na))
    expect_equal(max(out, map), max(out_na, map))
    expect_equal(max(out, lodcolumn=2), max(out_na, lodcolumn=2))
    expect_equal(max(out, map, lodcolumn=2), max(out_na, map, lodcolumn=2))
    expect_equal(max(out, lodcolumn=NULL), max(out_na, lodcolumn=NULL, na.rm=TRUE))
    expect_equal(max(out, map, lodcolumn=NULL), max(out_na, map, lodcolumn=NULL, na.rm=TRUE))

})

test_that("maxlod works for intercross with two phenotypes", {

    # read data
    iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

    # calculate genotype probabilities
    map <- insert_pseudomarkers(iron$gmap, step=1)
    probs <- calc_genoprob(iron, map, error_prob=0.002)

    # grab phenotypes and covariates; ensure that covariates have names attribute
    pheno <- iron$pheno
    covar <- match(iron$covar$sex, c("f", "m")) # make numeric
    names(covar) <- rownames(iron$covar)
    Xcovar <- get_x_covar(iron)

    # perform genome scan
    out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

    # overall max
    expect_equal(maxlod(out), max(unclass(out)))

    expect_equal(maxlod(out, lodcolumn="liver"), max(out[,"liver"], na.rm=TRUE))
    expect_equal(maxlod(out, lodcolumn="spleen"), max(out[,"spleen"], na.rm=TRUE))
    expect_equal(maxlod(out, lodcolumn=1), max(out[,"liver"], na.rm=TRUE))
    expect_equal(maxlod(out, lodcolumn=2), max(out[,"spleen"], na.rm=TRUE))

    expect_equal(maxlod(out, map, c("2", "9")),
                 max(unclass(subset(out, map, c("2", "9")))))

    expect_equal(maxlod(out, map, "2"), max(unclass(subset(out, map,"2")), na.rm=TRUE))
    expect_equal(maxlod(out, map, "9"), max(unclass(subset(out, map,"9")), na.rm=TRUE))

    expect_equal( maxlod(out, map, c("2","8", "11")),
                 max( maxlod(out, map, "2"), maxlod(out, map, "8"), maxlod(out, map, "11") ) )

    # really could be using integers
    expect_equal( maxlod(out, map, c(2,8, 11)),
                 max( maxlod(out, map, 2), maxlod(out, map, 8), maxlod(out, map, 11) ) )

    # test that it works if output or map are subsetted
    expect_equal( max(out, map, chr="2"), max(subset(out, map, chr="2"), map) )
    expect_equal( max(out, map, chr="2"), max(out, map["2"]) )

    # test that it works if output has rows shuffled
    out_shuffled <- out[sample(1:nrow(out)),,drop=FALSE]
    class(out_shuffled) <- c("scan1", "matrix")
    expect_equal( max(out_shuffled, map), max(out, map))

    # maxlod handles NA (here, just guessing that I'm unlikely to NA-out the max
    set.seed(1331427)
    out_na <- out
    out_na[5,2] <- NA
    out_na[sample(nrow(out_na), 10),1] <- NA
    expect_equal(maxlod(out), maxlod(out_na))
    expect_equal(maxlod(out, map, 11), maxlod(out_na, map, 11))

})

#' Locate crossovers
#'
#' Estimate the locations of crossovers in each individual on each chromosome.
#'
#' @param geno List of matrices of genotypes (output of [maxmarg()] or [viterbi()]).
#' @param map List of vectors with the map positions of the markers.
#' @param quiet If FALSE, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A list of lists of estimated crossover locations, with
#'     crossovers placed at the midpoint of the intervals that contain
#'     them.
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#' pr <- calc_genoprob(iron, map, error_prob=0.002, map_function="c-f")
#' g <- maxmarg(pr)
#' pos <- locate_xo(g, iron$gmap)
#'
#' @seealso [count_xo()]
#'
#' @export

locate_xo <-
    function(geno, map, quiet=TRUE, cores=1)
{
    if(is.null(geno)) stop("geno is NULL")
    if(is.null(map)) stop("map is NULL")

    crosstype <- attr(geno, "crosstype")
    if(is.null(crosstype))
        stop("Input geno needs to include a crosstype attribute.")
    is_x_chr <- attr(geno, "is_x_chr")
    if(is.null(is_x_chr))
        is_x_chr <- rep(FALSE, length(geno))
    names(is_x_chr) <- names(geno)

    if(length(geno) != length(map) ||
       any(names(geno) != names(map))) { # force matching chromosomes
        chr <- find_common_ids(names(geno), names(map))
        if(length(chr)==0)
            stop("geno and map have no chromosomes in common")
        geno <- geno[,chr]
        map <- map[chr]
    }
    for(i in seq_along(geno)) { # get matching markers
        if(ncol(geno[[i]]) != length(map[[i]]) ||
           !all(colnames(geno[[i]]) == names(map[[i]]))) {
            mar <- find_common_ids(colnames(geno[[i]]), names(map[[i]]))
            if(length(mar) == 0) {
                warning("No markers in common on chr ", names(map)[i])
                geno <- geno[-i]
                map <- map[-i]
            }
            else {
                geno[[i]] <- geno[[i]][,mar,drop=FALSE]
                map[[i]] <- map[[i]][mar]
            }
        }
    }
    if(length(geno) == 0)
        stop("geno and map have no chromosomes/markers in common")
    is_x_chr <- is_x_chr[names(geno)]

    # set up cluster; set quiet=TRUE if multi-core
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    by_chr_func <- function(chr) {
        result <- .locate_xo(t(geno[[chr]]), map[[chr]], crosstype, is_x_chr[chr])
        names(result) <- rownames(geno[[chr]])
        result
    }

    result <- cluster_lapply(cores, seq(along=geno), by_chr_func)
    names(result) <- names(geno)
    result
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

arrange_genes <- function(start, end) {
    .Call(`_qtl2_arrange_genes`, start, end)
}

.bayes_int_plain <- function(lod, pos, prob) {
    .Call(`_qtl2_R_bayes_int_plain`, lod, pos, prob)
}

calc_ll_binreg <- function(X, y, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_ll_binreg`, X, y, maxit, tol, qr_tol, eta_max)
}

calc_coef_binreg <- function(X, y, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_coef_binreg`, X, y, maxit, tol, qr_tol, eta_max)
}

calc_coefSE_binreg <- function(X, y, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_coefSE_binreg`, X, y, maxit, tol, qr_tol, eta_max)
}

fit_binreg <- function(X, y, se = TRUE, var = FALSE, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_fit_binreg`, X, y, se, var, maxit, tol, qr_tol, eta_max)
}

calc_ll_binreg_eigenchol <- function(X, y, maxit = 100L, tol = 1e-6, eta_max = 30.0) {
    .Call(`_qtl2_calc_ll_binreg_eigenchol`, X, y, maxit, tol, eta_max)
}

calc_ll_binreg_eigenqr <- function(X, y, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_ll_binreg_eigenqr`, X, y, maxit, tol, qr_tol, eta_max)
}

calc_coef_binreg_eigenqr <- function(X, y, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_coef_binreg_eigenqr`, X, y, maxit, tol, qr_tol, eta_max)
}

calc_coefSE_binreg_eigenqr <- function(X, y, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_coefSE_binreg_eigenqr`, X, y, maxit, tol, qr_tol, eta_max)
}

fit_binreg_eigenqr <- function(X, y, se = TRUE, var = FALSE, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_fit_binreg_eigenqr`, X, y, se, var, maxit, tol, qr_tol, eta_max)
}

calc_ll_binreg_weighted <- function(X, y, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_ll_binreg_weighted`, X, y, weights, maxit, tol, qr_tol, eta_max)
}

calc_coef_binreg_weighted <- function(X, y, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_coef_binreg_weighted`, X, y, weights, maxit, tol, qr_tol, eta_max)
}

calc_coefSE_binreg_weighted <- function(X, y, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_coefSE_binreg_weighted`, X, y, weights, maxit, tol, qr_tol, eta_max)
}

fit_binreg_weighted <- function(X, y, weights, se = TRUE, var = FALSE, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_fit_binreg_weighted`, X, y, weights, se, var, maxit, tol, qr_tol, eta_max)
}

calc_ll_binreg_weighted_eigenchol <- function(X, y, weights, maxit = 100L, tol = 1e-6, eta_max = 30.0) {
    .Call(`_qtl2_calc_ll_binreg_weighted_eigenchol`, X, y, weights, maxit, tol, eta_max)
}

calc_ll_binreg_weighted_eigenqr <- function(X, y, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_ll_binreg_weighted_eigenqr`, X, y, weights, maxit, tol, qr_tol, eta_max)
}

calc_coef_binreg_weighted_eigenqr <- function(X, y, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_coef_binreg_weighted_eigenqr`, X, y, weights, maxit, tol, qr_tol, eta_max)
}

calc_coefSE_binreg_weighted_eigenqr <- function(X, y, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_calc_coefSE_binreg_weighted_eigenqr`, X, y, weights, maxit, tol, qr_tol, eta_max)
}

fit_binreg_weighted_eigenqr <- function(X, y, weights, se = TRUE, var = FALSE, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_fit_binreg_weighted_eigenqr`, X, y, weights, se, var, maxit, tol, qr_tol, eta_max)
}

.calc_kinship <- function(prob_array) {
    .Call(`_qtl2_calc_kinship`, prob_array)
}

.crosstype_supported <- function(crosstype) {
    .Call(`_qtl2_crosstype_supported`, crosstype)
}

.count_invalid_genotypes <- function(crosstype, genotypes, is_X_chr, is_female, cross_info) {
    .Call(`_qtl2_count_invalid_genotypes`, crosstype, genotypes, is_X_chr, is_female, cross_info)
}

.check_crossinfo <- function(crosstype, cross_info, any_x_chr) {
    .Call(`_qtl2_check_crossinfo`, crosstype, cross_info, any_x_chr)
}

.check_is_female_vector <- function(crosstype, is_female, any_x_chr) {
    .Call(`_qtl2_check_is_female_vector`, crosstype, is_female, any_x_chr)
}

check_handle_x_chr <- function(crosstype, any_x_chr) {
    .Call(`_qtl2_check_handle_x_chr`, crosstype, any_x_chr)
}

.chisq_colpairs <- function(input) {
    .Call(`_qtl2_chisq_colpairs`, input)
}

.clean_genoprob <- function(prob_array, value_threshold = 1e-6, column_threshold = 0.01) {
    .Call(`_qtl2_clean_genoprob`, prob_array, value_threshold, column_threshold)
}

.compare_geno <- function(geno) {
    .Call(`_qtl2_compare_geno`, geno)
}

.count_xo <- function(geno, crosstype, is_X_chr) {
    .Call(`_qtl2_count_xo`, geno, crosstype, is_X_chr)
}

.count_xo_3d <- function(geno_array, crosstype, is_X_chr) {
    .Call(`_qtl2_count_xo_3d`, geno_array, crosstype, is_X_chr)
}

mpp_encode_alleles <- function(allele1, allele2, n_alleles, phase_known) {
    .Call(`_qtl2_mpp_encode_alleles`, allele1, allele2, n_alleles, phase_known)
}

mpp_decode_geno <- function(true_gen, n_alleles, phase_known) {
    .Call(`_qtl2_mpp_decode_geno`, true_gen, n_alleles, phase_known)
}

mpp_is_het <- function(true_gen, n_alleles, phase_known) {
    .Call(`_qtl2_mpp_is_het`, true_gen, n_alleles, phase_known)
}

mpp_geno_names <- function(alleles, is_x_chr) {
    .Call(`_qtl2_mpp_geno_names`, alleles, is_x_chr)
}

invert_founder_index <- function(cross_info) {
    .Call(`_qtl2_invert_founder_index`, cross_info)
}

.is_phase_known <- function(crosstype) {
    .Call(`_qtl2_is_phase_known`, crosstype)
}

.find_dup_markers_notexact <- function(Geno, order, markerloc, adjacent_only) {
    .Call(`_qtl2_find_dup_markers_notexact`, Geno, order, markerloc, adjacent_only)
}

.find_ibd_segments <- function(g1, g2, p, error_prob) {
    .Call(`_qtl2_find_ibd_segments`, g1, g2, p, error_prob)
}

.find_peaks <- function(lod, threshold, peakdrop) {
    .Call(`_qtl2_R_find_peaks`, lod, threshold, peakdrop)
}

.find_peaks_and_lodint <- function(lod, threshold, peakdrop, drop) {
    .Call(`_qtl2_R_find_peaks_and_lodint`, lod, threshold, peakdrop, drop)
}

.find_peaks_and_bayesint <- function(lod, pos, threshold, peakdrop, prob) {
    .Call(`_qtl2_R_find_peaks_and_bayesint`, lod, pos, threshold, peakdrop, prob)
}

fit1_binary_addcovar <- function(genoprobs, pheno, addcovar, weights, se = FALSE, var = FALSE, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_fit1_binary_addcovar`, genoprobs, pheno, addcovar, weights, se, var, maxit, tol, qr_tol, eta_max)
}

fit1_binary_intcovar <- function(genoprobs, pheno, addcovar, intcovar, weights, se = TRUE, var = FALSE, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_fit1_binary_intcovar`, genoprobs, pheno, addcovar, intcovar, weights, se, var, maxit, tol, qr_tol, eta_max)
}

fit1_hk_addcovar <- function(genoprobs, pheno, addcovar, weights, se = FALSE, var = FALSE, tol = 1e-12) {
    .Call(`_qtl2_fit1_hk_addcovar`, genoprobs, pheno, addcovar, weights, se, var, tol)
}

fit1_hk_intcovar <- function(genoprobs, pheno, addcovar, intcovar, weights, se = FALSE, var = FALSE, tol = 1e-12) {
    .Call(`_qtl2_fit1_hk_intcovar`, genoprobs, pheno, addcovar, intcovar, weights, se, var, tol)
}

fit1_pg_addcovar <- function(genoprobs, pheno, addcovar, eigenvec, weights, se = FALSE, var = FALSE, tol = 1e-12) {
    .Call(`_qtl2_fit1_pg_addcovar`, genoprobs, pheno, addcovar, eigenvec, weights, se, var, tol)
}

fit1_pg_intcovar <- function(genoprobs, pheno, addcovar, intcovar, eigenvec, weights, se = FALSE, var = FALSE, tol = 1e-12) {
    .Call(`_qtl2_fit1_pg_intcovar`, genoprobs, pheno, addcovar, intcovar, eigenvec, weights, se, var, tol)
}

geno_names <- function(crosstype, alleles, is_x_chr) {
    .Call(`_qtl2_geno_names`, crosstype, alleles, is_x_chr)
}

nalleles <- function(crosstype) {
    .Call(`_qtl2_nalleles`, crosstype)
}

.genoprob_to_alleleprob <- function(crosstype, prob_array, is_x_chr) {
    .Call(`_qtl2_genoprob_to_alleleprob`, crosstype, prob_array, is_x_chr)
}

.get_x_covar <- function(crosstype, is_female, cross_info) {
    .Call(`_qtl2_get_x_covar`, crosstype, is_female, cross_info)
}

.guess_phase_f2A <- function(geno, deterministic) {
    .Call(`_qtl2_guess_phase_f2A`, geno, deterministic)
}

.guess_phase_f2X <- function(geno, deterministic) {
    .Call(`_qtl2_guess_phase_f2X`, geno, deterministic)
}

.guess_phase_A <- function(geno, crosstype, deterministic) {
    .Call(`_qtl2_guess_phase_A`, geno, crosstype, deterministic)
}

.guess_phase_X <- function(geno, crosstype, is_female, deterministic) {
    .Call(`_qtl2_guess_phase_X`, geno, crosstype, is_female, deterministic)
}

.calc_errorlod <- function(crosstype, probs, genotypes, founder_geno, is_X_chr, is_female, cross_info) {
    .Call(`_qtl2_calc_errorlod`, crosstype, probs, genotypes, founder_geno, is_X_chr, is_female, cross_info)
}

.calc_genoprob <- function(crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob) {
    .Call(`_qtl2_calc_genoprob`, crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob)
}

.calc_genoprob2 <- function(crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob) {
    .Call(`_qtl2_calc_genoprob2`, crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob)
}

.est_map <- function(crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, error_prob, max_iterations, tol, verbose) {
    .Call(`_qtl2_est_map`, crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, error_prob, max_iterations, tol, verbose)
}

.est_map2 <- function(crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, cross_group, unique_cross_group, rec_frac, error_prob, max_iterations, tol, verbose) {
    .Call(`_qtl2_est_map2`, crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, cross_group, unique_cross_group, rec_frac, error_prob, max_iterations, tol, verbose)
}

.sim_geno <- function(crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob, n_draws) {
    .Call(`_qtl2_sim_geno`, crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob, n_draws)
}

.sim_geno2 <- function(crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob, n_draws) {
    .Call(`_qtl2_sim_geno2`, crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob, n_draws)
}

addlog <- function(a, b) {
    .Call(`_qtl2_addlog`, a, b)
}

subtractlog <- function(a, b) {
    .Call(`_qtl2_subtractlog`, a, b)
}

.viterbi <- function(crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob) {
    .Call(`_qtl2_viterbi`, crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob)
}

.viterbi2 <- function(crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob) {
    .Call(`_qtl2_viterbi2`, crosstype, genotypes, founder_geno, is_X_chr, is_female, cross_info, rec_frac, marker_index, error_prob)
}

.interp_genoprob_onechr <- function(genoprob, map, pos_index) {
    .Call(`_qtl2_interp_genoprob_onechr`, genoprob, map, pos_index)
}

interpolate_map <- function(oldpos, oldmap, newmap) {
    .Call(`_qtl2_interpolate_map`, oldpos, oldmap, newmap)
}

find_intervals <- function(pos, map, tol = 1e-8) {
    .Call(`_qtl2_find_intervals`, pos, map, tol)
}

calc_rss_linreg <- function(X, Y, tol = 1e-12) {
    .Call(`_qtl2_calc_rss_linreg`, X, Y, tol)
}

calc_coef_linreg <- function(X, y, tol = 1e-12) {
    .Call(`_qtl2_calc_coef_linreg`, X, y, tol)
}

calc_coefSE_linreg <- function(X, y, tol = 1e-12) {
    .Call(`_qtl2_calc_coefSE_linreg`, X, y, tol)
}

calc_resid_linreg <- function(X, Y, tol = 1e-12) {
    .Call(`_qtl2_calc_resid_linreg`, X, Y, tol)
}

calc_resid_linreg_3d <- function(X, P, tol = 1e-12) {
    .Call(`_qtl2_calc_resid_linreg_3d`, X, P, tol)
}

fit_linreg <- function(X, y, se = TRUE, var = FALSE, tol = 1e-12) {
    .Call(`_qtl2_fit_linreg`, X, y, se, var, tol)
}

fit_linreg_eigenchol <- function(X, y, se, var = FALSE) {
    .Call(`_qtl2_fit_linreg_eigenchol`, X, y, se, var)
}

calc_coef_linreg_eigenchol <- function(X, y) {
    .Call(`_qtl2_calc_coef_linreg_eigenchol`, X, y)
}

calc_coefSE_linreg_eigenchol <- function(X, y) {
    .Call(`_qtl2_calc_coefSE_linreg_eigenchol`, X, y)
}

calc_rss_eigenchol <- function(X, y) {
    .Call(`_qtl2_calc_rss_eigenchol`, X, y)
}

calc_fitted_linreg_eigenchol <- function(X, y) {
    .Call(`_qtl2_calc_fitted_linreg_eigenchol`, X, y)
}

fit_linreg_eigenqr <- function(X, y, se, var = FALSE, tol = 1e-12) {
    .Call(`_qtl2_fit_linreg_eigenqr`, X, y, se, var, tol)
}

calc_coef_linreg_eigenqr <- function(X, y, tol = 1e-12) {
    .Call(`_qtl2_calc_coef_linreg_eigenqr`, X, y, tol)
}

calc_coefSE_linreg_eigenqr <- function(X, y, tol = 1e-12) {
    .Call(`_qtl2_calc_coefSE_linreg_eigenqr`, X, y, tol)
}

calc_rss_eigenqr <- function(X, y, tol = 1e-12) {
    .Call(`_qtl2_calc_rss_eigenqr`, X, y, tol)
}

calc_fitted_linreg_eigenqr <- function(X, y, tol = 1e-12) {
    .Call(`_qtl2_calc_fitted_linreg_eigenqr`, X, y, tol)
}

calc_mvrss_eigenchol <- function(X, Y) {
    .Call(`_qtl2_calc_mvrss_eigenchol`, X, Y)
}

calc_mvrss_eigenqr <- function(X, Y, tol = 1e-12) {
    .Call(`_qtl2_calc_mvrss_eigenqr`, X, Y, tol)
}

calc_resid_eigenchol <- function(X, Y) {
    .Call(`_qtl2_calc_resid_eigenchol`, X, Y)
}

calc_resid_eigenqr <- function(X, Y, tol = 1e-12) {
    .Call(`_qtl2_calc_resid_eigenqr`, X, Y, tol)
}

Rcpp_eigen_decomp <- function(A) {
    .Call(`_qtl2_Rcpp_eigen_decomp`, A)
}

Rcpp_eigen_rotation <- function(K, y, X) {
    .Call(`_qtl2_Rcpp_eigen_rotation`, K, y, X)
}

Rcpp_calc_logdetXpX <- function(X) {
    .Call(`_qtl2_Rcpp_calc_logdetXpX`, X)
}

Rcpp_calcLL <- function(hsq, Kva, y, X, reml = TRUE, logdetXpX = NA_real_) {
    .Call(`_qtl2_Rcpp_calcLL`, hsq, Kva, y, X, reml, logdetXpX)
}

Rcpp_calcLL_mat <- function(hsq, Kva, Y, X, reml = TRUE, logdetXpX = NA_real_) {
    .Call(`_qtl2_Rcpp_calcLL_mat`, hsq, Kva, Y, X, reml, logdetXpX)
}

Rcpp_fitLMM <- function(Kva, y, X, reml = TRUE, check_boundary = TRUE, logdetXpX = NA_real_, tol = 1e-4) {
    .Call(`_qtl2_Rcpp_fitLMM`, Kva, y, X, reml, check_boundary, logdetXpX, tol)
}

Rcpp_fitLMM_mat <- function(Kva, Y, X, reml = TRUE, check_boundary = TRUE, logdetXpX = NA_real_, tol = 1e-4) {
    .Call(`_qtl2_Rcpp_fitLMM_mat`, Kva, Y, X, reml, check_boundary, logdetXpX, tol)
}

.locate_xo <- function(geno, map, crosstype, is_X_chr) {
    .Call(`_qtl2_locate_xo`, geno, map, crosstype, is_X_chr)
}

.lod_int_plain <- function(lod, drop) {
    .Call(`_qtl2_R_lod_int_plain`, lod, drop)
}

find_matching_cols <- function(mat, tol = 1e-12) {
    .Call(`_qtl2_find_matching_cols`, mat, tol)
}

find_lin_indep_cols <- function(mat, tol = 1e-12) {
    .Call(`_qtl2_find_lin_indep_cols`, mat, tol)
}

formX_intcovar <- function(probs, addcovar, intcovar, position, has_intercept = TRUE) {
    .Call(`_qtl2_formX_intcovar`, probs, addcovar, intcovar, position, has_intercept)
}

expand_genoprobs_intcovar <- function(probs, intcovar) {
    .Call(`_qtl2_expand_genoprobs_intcovar`, probs, intcovar)
}

weighted_matrix <- function(mat, weights) {
    .Call(`_qtl2_weighted_matrix`, mat, weights)
}

weighted_3darray <- function(array, weights) {
    .Call(`_qtl2_weighted_3darray`, array, weights)
}

matrix_x_matrix <- function(X, Y) {
    .Call(`_qtl2_matrix_x_matrix`, X, Y)
}

matrix_x_vector <- function(X, y) {
    .Call(`_qtl2_matrix_x_vector`, X, y)
}

matrix_x_3darray <- function(X, A) {
    .Call(`_qtl2_matrix_x_3darray`, X, A)
}

.maxmarg <- function(prob_array, minprob, tol) {
    .Call(`_qtl2_maxmarg`, prob_array, minprob, tol)
}

.predict_snpgeno <- function(allele1, allele2, founder_geno) {
    .Call(`_qtl2_predict_snpgeno`, allele1, allele2, founder_geno)
}

random_int <- function(n, low, high) {
    .Call(`_qtl2_random_int`, n, low, high)
}

get_permutation <- function(n) {
    .Call(`_qtl2_get_permutation`, n)
}

permute_nvector <- function(n_perm, x) {
    .Call(`_qtl2_permute_nvector`, n_perm, x)
}

permute_ivector <- function(n_perm, x) {
    .Call(`_qtl2_permute_ivector`, n_perm, x)
}

permute_nvector_stratified <- function(n_perm, x, strata, n_strata = -1L) {
    .Call(`_qtl2_permute_nvector_stratified`, n_perm, x, strata, n_strata)
}

permute_ivector_stratified <- function(n_perm, x, strata, n_strata = -1L) {
    .Call(`_qtl2_permute_ivector_stratified`, n_perm, x, strata, n_strata)
}

.reduce_markers <- function(pos, min_dist, weights) {
    .Call(`_qtl2_reduce_markers`, pos, min_dist, weights)
}

.running_count <- function(pos, result_pos, window) {
    .Call(`_qtl2_running_count`, pos, result_pos, window)
}

scan_binary_onechr <- function(genoprobs, pheno, addcovar, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_scan_binary_onechr`, genoprobs, pheno, addcovar, maxit, tol, qr_tol, eta_max)
}

scan_binary_onechr_weighted <- function(genoprobs, pheno, addcovar, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_scan_binary_onechr_weighted`, genoprobs, pheno, addcovar, weights, maxit, tol, qr_tol, eta_max)
}

scan_binary_onechr_intcovar_highmem <- function(genoprobs, pheno, addcovar, intcovar, maxit = 100L, tol = 1e-6, qr_tol = 1e-12) {
    .Call(`_qtl2_scan_binary_onechr_intcovar_highmem`, genoprobs, pheno, addcovar, intcovar, maxit, tol, qr_tol)
}

scan_binary_onechr_intcovar_weighted_highmem <- function(genoprobs, pheno, addcovar, intcovar, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12) {
    .Call(`_qtl2_scan_binary_onechr_intcovar_weighted_highmem`, genoprobs, pheno, addcovar, intcovar, weights, maxit, tol, qr_tol)
}

scan_binary_onechr_intcovar_lowmem <- function(genoprobs, pheno, addcovar, intcovar, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_scan_binary_onechr_intcovar_lowmem`, genoprobs, pheno, addcovar, intcovar, maxit, tol, qr_tol, eta_max)
}

scan_binary_onechr_intcovar_weighted_lowmem <- function(genoprobs, pheno, addcovar, intcovar, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_scan_binary_onechr_intcovar_weighted_lowmem`, genoprobs, pheno, addcovar, intcovar, weights, maxit, tol, qr_tol, eta_max)
}

scan_hk_onechr_nocovar <- function(genoprobs, pheno, tol = 1e-12) {
    .Call(`_qtl2_scan_hk_onechr_nocovar`, genoprobs, pheno, tol)
}

scan_hk_onechr <- function(genoprobs, pheno, addcovar, tol = 1e-12) {
    .Call(`_qtl2_scan_hk_onechr`, genoprobs, pheno, addcovar, tol)
}

scan_hk_onechr_weighted <- function(genoprobs, pheno, addcovar, weights, tol = 1e-12) {
    .Call(`_qtl2_scan_hk_onechr_weighted`, genoprobs, pheno, addcovar, weights, tol)
}

scan_hk_onechr_intcovar_highmem <- function(genoprobs, pheno, addcovar, intcovar, tol = 1e-12) {
    .Call(`_qtl2_scan_hk_onechr_intcovar_highmem`, genoprobs, pheno, addcovar, intcovar, tol)
}

scan_hk_onechr_intcovar_weighted_highmem <- function(genoprobs, pheno, addcovar, intcovar, weights, tol = 1e-12) {
    .Call(`_qtl2_scan_hk_onechr_intcovar_weighted_highmem`, genoprobs, pheno, addcovar, intcovar, weights, tol)
}

scan_hk_onechr_intcovar_lowmem <- function(genoprobs, pheno, addcovar, intcovar, tol = 1e-12) {
    .Call(`_qtl2_scan_hk_onechr_intcovar_lowmem`, genoprobs, pheno, addcovar, intcovar, tol)
}

scan_hk_onechr_intcovar_weighted_lowmem <- function(genoprobs, pheno, addcovar, intcovar, weights, tol = 1e-12) {
    .Call(`_qtl2_scan_hk_onechr_intcovar_weighted_lowmem`, genoprobs, pheno, addcovar, intcovar, weights, tol)
}

scan_pg_onechr <- function(genoprobs, pheno, addcovar, eigenvec, weights, tol = 1e-12) {
    .Call(`_qtl2_scan_pg_onechr`, genoprobs, pheno, addcovar, eigenvec, weights, tol)
}

scan_pg_onechr_intcovar_highmem <- function(genoprobs, pheno, addcovar, intcovar, eigenvec, weights, tol = 1e-12) {
    .Call(`_qtl2_scan_pg_onechr_intcovar_highmem`, genoprobs, pheno, addcovar, intcovar, eigenvec, weights, tol)
}

scan_pg_onechr_intcovar_lowmem <- function(genoprobs, pheno, addcovar, intcovar, eigenvec, weights, tol = 1e-12) {
    .Call(`_qtl2_scan_pg_onechr_intcovar_lowmem`, genoprobs, pheno, addcovar, intcovar, eigenvec, weights, tol)
}

scanblup <- function(genoprobs, pheno, addcovar, se, reml, tol = 1e-12) {
    .Call(`_qtl2_scanblup`, genoprobs, pheno, addcovar, se, reml, tol)
}

scancoef_binary_addcovar <- function(genoprobs, pheno, addcovar, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_scancoef_binary_addcovar`, genoprobs, pheno, addcovar, weights, maxit, tol, qr_tol, eta_max)
}

scancoef_binary_intcovar <- function(genoprobs, pheno, addcovar, intcovar, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_scancoef_binary_intcovar`, genoprobs, pheno, addcovar, intcovar, weights, maxit, tol, qr_tol, eta_max)
}

scancoefSE_binary_addcovar <- function(genoprobs, pheno, addcovar, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_scancoefSE_binary_addcovar`, genoprobs, pheno, addcovar, weights, maxit, tol, qr_tol, eta_max)
}

scancoefSE_binary_intcovar <- function(genoprobs, pheno, addcovar, intcovar, weights, maxit = 100L, tol = 1e-6, qr_tol = 1e-12, eta_max = 30.0) {
    .Call(`_qtl2_scancoefSE_binary_intcovar`, genoprobs, pheno, addcovar, intcovar, weights, maxit, tol, qr_tol, eta_max)
}

scancoef_hk_addcovar <- function(genoprobs, pheno, addcovar, weights, tol = 1e-12) {
    .Call(`_qtl2_scancoef_hk_addcovar`, genoprobs, pheno, addcovar, weights, tol)
}

scancoef_hk_intcovar <- function(genoprobs, pheno, addcovar, intcovar, weights, tol = 1e-12) {
    .Call(`_qtl2_scancoef_hk_intcovar`, genoprobs, pheno, addcovar, intcovar, weights, tol)
}

scancoefSE_hk_addcovar <- function(genoprobs, pheno, addcovar, weights, tol = 1e-12) {
    .Call(`_qtl2_scancoefSE_hk_addcovar`, genoprobs, pheno, addcovar, weights, tol)
}

scancoefSE_hk_intcovar <- function(genoprobs, pheno, addcovar, intcovar, weights, tol = 1e-12) {
    .Call(`_qtl2_scancoefSE_hk_intcovar`, genoprobs, pheno, addcovar, intcovar, weights, tol)
}

scancoef_pg_addcovar <- function(genoprobs, pheno, addcovar, eigenvec, weights, tol = 1e-12) {
    .Call(`_qtl2_scancoef_pg_addcovar`, genoprobs, pheno, addcovar, eigenvec, weights, tol)
}

scancoef_pg_intcovar <- function(genoprobs, pheno, addcovar, intcovar, eigenvec, weights, tol = 1e-12) {
    .Call(`_qtl2_scancoef_pg_intcovar`, genoprobs, pheno, addcovar, intcovar, eigenvec, weights, tol)
}

scancoefSE_pg_addcovar <- function(genoprobs, pheno, addcovar, eigenvec, weights, tol = 1e-12) {
    .Call(`_qtl2_scancoefSE_pg_addcovar`, genoprobs, pheno, addcovar, eigenvec, weights, tol)
}

scancoefSE_pg_intcovar <- function(genoprobs, pheno, addcovar, intcovar, eigenvec, weights, tol = 1e-12) {
    .Call(`_qtl2_scancoefSE_pg_intcovar`, genoprobs, pheno, addcovar, intcovar, eigenvec, weights, tol)
}

.calc_sdp <- function(geno) {
    .Call(`_qtl2_calc_sdp`, geno)
}

.invert_sdp <- function(sdp, n_str) {
    .Call(`_qtl2_invert_sdp`, sdp, n_str)
}

.alleleprob_to_snpprob <- function(alleleprob, sdp, interval, on_map) {
    .Call(`_qtl2_alleleprob_to_snpprob`, alleleprob, sdp, interval, on_map)
}

genocol_to_snpcol <- function(n_str, sdp) {
    .Call(`_qtl2_genocol_to_snpcol`, n_str, sdp)
}

.genoprob_to_snpprob <- function(genoprob, sdp, interval, on_map) {
    .Call(`_qtl2_genoprob_to_snpprob`, genoprob, sdp, interval, on_map)
}

Xgenocol_to_snpcol <- function(n_str, sdp) {
    .Call(`_qtl2_Xgenocol_to_snpcol`, n_str, sdp)
}

.Xgenoprob_to_snpprob <- function(genoprob, sdp, interval, on_map) {
    .Call(`_qtl2_Xgenoprob_to_snpprob`, genoprob, sdp, interval, on_map)
}

test_init <- function(crosstype, true_gen, is_x_chr, is_female, cross_info) {
    .Call(`_qtl2_test_init`, crosstype, true_gen, is_x_chr, is_female, cross_info)
}

test_emit <- function(crosstype, obs_gen, true_gen, error_prob, founder_geno, is_x_chr, is_female, cross_info) {
    .Call(`_qtl2_test_emit`, crosstype, obs_gen, true_gen, error_prob, founder_geno, is_x_chr, is_female, cross_info)
}

test_step <- function(crosstype, gen_left, gen_right, rec_frac, is_x_chr, is_female, cross_info) {
    .Call(`_qtl2_test_step`, crosstype, gen_left, gen_right, rec_frac, is_x_chr, is_female, cross_info)
}

test_check_geno <- function(crosstype, gen, is_observed_value, is_x_chr, is_female, cross_info) {
    .Call(`_qtl2_test_check_geno`, crosstype, gen, is_observed_value, is_x_chr, is_female, cross_info)
}

test_possible_gen <- function(crosstype, is_x_chr, is_female, cross_info) {
    .Call(`_qtl2_test_possible_gen`, crosstype, is_x_chr, is_female, cross_info)
}

test_ngen <- function(crosstype, is_x_chr) {
    .Call(`_qtl2_test_ngen`, crosstype, is_x_chr)
}

test_nrec <- function(crosstype, gen_left, gen_right, is_x_chr, is_female, cross_info) {
    .Call(`_qtl2_test_nrec`, crosstype, gen_left, gen_right, is_x_chr, is_female, cross_info)
}

test_founder_geno_values <- function(crosstype, founder_geno) {
    .Call(`_qtl2_test_founder_geno_values`, crosstype, founder_geno)
}

need_founder_geno <- function(crosstype) {
    .Call(`_qtl2_need_founder_geno`, crosstype)
}

check_founder_geno_size <- function(crosstype, founder_geno, n_markers) {
    .Call(`_qtl2_check_founder_geno_size`, crosstype, founder_geno, n_markers)
}

test_emitmatrix <- function(crosstype, error_prob, max_obsgeno, founder_geno, is_x_chr, is_female, cross_info) {
    .Call(`_qtl2_test_emitmatrix`, crosstype, error_prob, max_obsgeno, founder_geno, is_x_chr, is_female, cross_info)
}

test_stepmatrix <- function(crosstype, rec_frac, is_x_chr, is_female, cross_info) {
    .Call(`_qtl2_test_stepmatrix`, crosstype, rec_frac, is_x_chr, is_female, cross_info)
}

test_initvector <- function(crosstype, is_x_chr, is_female, cross_info) {
    .Call(`_qtl2_test_initvector`, crosstype, is_x_chr, is_female, cross_info)
}

context("basic HMM functions in haploids")

test_that("haploids nalleles works", {
    expect_equal(nalleles("haploid"), 2)
})

test_that("haploids check_geno works", {

    # autosome
    expect_true(test_check_geno("haploid", 0, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("haploid", 1, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("haploid", 2, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("haploid", 1, FALSE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("haploid", 2, FALSE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("haploid", 3, TRUE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("haploid", 0, FALSE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("haploid", 3, FALSE, FALSE, FALSE, integer(0)))
})

test_that("haploids n_gen works", {

    expect_equal(test_ngen("haploid", FALSE), 2)

})

test_that("haploids possible_gen works", {

    expect_equal(test_possible_gen("haploid", FALSE, FALSE, integer(0)), 1:2)
})

test_that("haploids nrec works", {

    # autosome
    expect_equal(test_nrec("haploid", 1, 1, FALSE, FALSE, integer(0)), 0)
    expect_equal(test_nrec("haploid", 1, 2, FALSE, FALSE, integer(0)), 1)
    expect_equal(test_nrec("haploid", 2, 1, FALSE, FALSE, integer(0)), 1)
    expect_equal(test_nrec("haploid", 2, 2, FALSE, FALSE, integer(0)), 0)

})

test_that("haploids init works", {

    expect_equal(test_init("haploid", 1, FALSE, FALSE, integer(0)), log(0.5))
    expect_equal(test_init("haploid", 2, FALSE, FALSE, integer(0)), log(0.5))

})

test_that("haploids emit works", {

    eps <- 0.01
    expect_equal(test_emit("haploid", 0, 1, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("haploid", 0, 2, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("haploid", 1, 1, eps, integer(0), FALSE, FALSE, integer(0)), log(1-eps))
    expect_equal(test_emit("haploid", 1, 2, eps, integer(0), FALSE, FALSE, integer(0)), log(eps))
    expect_equal(test_emit("haploid", 2, 1, eps, integer(0), FALSE, FALSE, integer(0)), log(eps))
    expect_equal(test_emit("haploid", 2, 2, eps, integer(0), FALSE, FALSE, integer(0)), log(1-eps))

    # errors
    expect_equal(test_emit("haploid", 0, 0, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("haploid", 0, 3, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("haploid", 3, 1, eps, integer(0), FALSE, FALSE, integer(0)), 0)

})

test_that("haploids step works", {

    rf <- 0.01
    expect_equal(test_step("haploid", 1, 1, rf, FALSE, FALSE, integer(0)), log(1-rf))
    expect_equal(test_step("haploid", 1, 2, rf, FALSE, FALSE, integer(0)), log(rf))
    expect_equal(test_step("haploid", 2, 1, rf, FALSE, FALSE, integer(0)), log(rf))
    expect_equal(test_step("haploid", 2, 2, rf, FALSE, FALSE, integer(0)), log(1-rf))

})

test_that("geno_names works", {
    expect_equal(geno_names("haploid", c("B", "R"), FALSE), c("B", "R"))
    expect_equal(geno_names("haploid", c("B", "R"), TRUE), c("B", "R"))
})

context("linear regression")
library(qtl)

test_that("lin regr works for simple example", {

    set.seed(27343534)

    n <- 1000
    x <- rnorm(n, 50, 10)
    X <- cbind(1, x)
    y <- 30 + 0.5*x + rnorm(n, 0, 2.5)
    Y <- as.matrix(y)

    lm.out <- lm(y ~ x)
    resid <- lm.out$resid
    names(resid) <- NULL
    rss <- sum(resid^2)

    # QR-based regression
    qrFit <- fit_linreg_eigenqr(X, y, TRUE)
    expect_equal(rss, qrFit$rss)
    expect_equivalent(lm.out$coef, qrFit$coef)
    expect_equivalent(lm.out$fitted, qrFit$fitted)
    expect_equivalent(summary(lm.out)$coef[,2], qrFit$SE)

    qrRSS <- calc_rss_eigenqr(X, y)
    expect_equal(rss, qrRSS)

    # eigen residuals
    eigenqr_resid <- calc_resid_eigenqr(X,Y)
    expect_equal(eigenqr_resid, as.matrix(resid))
    eigenchol_resid <- calc_resid_eigenqr(X,Y)
    expect_equal(eigenchol_resid, as.matrix(resid))

    # generic
    linreg_rss <- calc_rss_linreg(X,Y)
    expect_equal(linreg_rss, rss)
    linreg_resid <- calc_resid_linreg(X,Y)
    expect_equal(linreg_resid, as.matrix(resid))

    # just the coefficients
    coef <- calc_coef_linreg(X,Y)
    expect_equal(coef, as.numeric(lm.out$coef))

    # coefficients and SEs
    coefSE <- calc_coefSE_linreg(X,Y)
    expected <- summary(lm.out)$coef
    expected <- list(coef=as.numeric(expected[,1]),
                     SE=as.numeric(expected[,2]))
    expect_equal(coefSE, expected)

})

test_that("lin regr works for reduced-rank example", {

    dd <- data.frame(f1 = gl(4, 6, labels = LETTERS[1:4]),
                     f2 = gl(3, 2, labels = letters[1:3]))[-(7:8), ]
    mm <- model.matrix(~ f1*f2, dd)
    y <- mm %*% seq_len(ncol(mm)) + rnorm(nrow(mm), sd = 0.1)
    Y <- as.matrix(y)

    lm.out <- lm(y ~ -1 + mm)
    resid <- lm.out$resid
    names(resid) <- NULL
    rss <- sum(resid^2)
    lm_se <- lm.out$coef
    lm_se[!is.na(lm_se)] <- summary(lm.out)$coef[,2]

    # (cholesky-based regression won't work here)

    # QR-based regression
    qrFit <- fit_linreg_eigenqr(mm, y, TRUE)
    expect_equal(rss, qrFit$rss)
    expect_equivalent(lm.out$fitted, qrFit$fitted)
    lm_reduced <- lm(y ~ -1 + mm[,!is.na(qrFit$coef)])
    lm_reduced_se <- lm_reduced$coef
    lm_reduced_se <- summary(lm_reduced)$coef[,2]
    expect_equivalent(lm_reduced$coef, qrFit$coef[!is.na(qrFit$coef)])
    expect_equivalent(lm_reduced_se, qrFit$SE[!is.na(qrFit$coef)])

    qrRSS <- calc_rss_eigenqr(mm, y)
    expect_equal(rss, qrRSS)

    # eigen resid
    eigenqr_resid <- calc_resid_eigenqr(mm, Y)
    expect_equal(eigenqr_resid, as.matrix(resid))

    # generic
    linreg_rss <- calc_rss_linreg(mm,Y)
    expect_equal(linreg_rss, rss)
    linreg_resid <- calc_resid_linreg(mm,Y)
    expect_equal(linreg_resid, as.matrix(resid))

    # just the coefficients
    coef <- calc_coef_linreg(mm, y)
    expect_equivalent(coef[!is.na(coef)], lm_reduced$coef)

    # coefficients and SEs
    coefSE <- calc_coefSE_linreg(mm, y)
    expect_equivalent(coefSE$coef[!is.na(coefSE$coef)], lm_reduced$coef)
    expect_equivalent(coefSE$SE[!is.na(coefSE$coef)], lm_reduced_se)

})


test_that("lin regr works in a QTL situation", {

    data(hyper)
    hyper <- calc.genoprob(hyper, step=1, err=0.001)
    qtl <- makeqtl(hyper, c(1,4,6,15), c(68.3, 30, 66.7, 17.5),
                   what="prob")

    n <- nind(hyper)
    y <- hyper$pheno[,1]
    Y <- as.matrix(y)
    X <- matrix(ncol=6, nrow=n)
    X[,1] <- 1
    for(i in 2:5) X[,i] <- qtl$prob[[i-1]][,2]
    X[,6] <- X[,4]*X[,5]

    lodfull <- fitqtl(hyper, qtl=qtl, method="hk", dropone=FALSE,
                      formula=y~q1+q2+q3*q4)$lod
    lodadd <- fitqtl(hyper, qtl=qtl, method="hk", dropone=FALSE,
                     formula=y~q1+q2+q3+q4)$lod
    lod14 <- fitqtl(hyper, qtl=qtl, method="hk", dropone=FALSE,
                    formula=y~q1+q2)$lod
    lod4 <- fitqtl(hyper, qtl=qtl, method="hk", dropone=FALSE,
                    formula=y~q2)$lod

    # eigen qr
    rss0 <- calc_rss_eigenqr(X[,1,drop=FALSE], y)
    expect_equal(rss0, sum(lm(y~1)$resid^2))

    rssfull <- calc_rss_eigenqr(X, y)
    expect_equal(rssfull, sum(lm(y ~ X)$resid^2))
    expect_equal(lodfull, (n/2)*log10(rss0/rssfull))

    rssadd <- calc_rss_eigenqr(X[,1:5], y)
    expect_equal(rssadd, sum(lm(y ~ X[,2:5])$resid^2))
    expect_equal(lodadd, (n/2)*log10(rss0/rssadd))

    rss14 <- calc_rss_eigenqr(X[,1:3], y)
    expect_equal(rss14, sum(lm(y ~ X[,2:3])$resid^2))
    expect_equal(lod14, (n/2)*log10(rss0/rss14))

    rss4 <- calc_rss_eigenqr(X[,c(1,3)], y)
    expect_equal(rss4, sum(lm(y ~ X[,3,drop=FALSE])$resid^2))
    expect_equal(lod4, (n/2)*log10(rss0/rss4))

})

test_that("lin regr works for multiple columns", {

    set.seed(27343534)

    n <- 1000
    x <- rnorm(n, 50, 10)
    X <- cbind(1, x)
    y <- 30 + 0.5*x + rnorm(n, 0, 2.5)

    ncolY <- 10
    Y <- permute_nvector(10, y)

    resid <- lm.fit(X,Y)$resid
    lm.rss <- colSums(resid^2)

    # RSS
    expect_equal(lm.rss, calc_mvrss_eigenchol(X, Y))
    expect_equal(lm.rss, calc_mvrss_eigenqr(X, Y))

    # residuals
    expect_equal(resid, calc_resid_eigenqr(X, Y))
    expect_equal(resid, calc_resid_eigenchol(X, Y))

    # generic
    expect_equal(lm.rss, calc_rss_linreg(X, Y))
    expect_equal(as.matrix(resid), calc_resid_linreg(X, Y))

})

test_that("lin regr works for multiple columns, reduced-rank X", {

    dd <- data.frame(f1 = gl(4, 6, labels = LETTERS[1:4]),
                     f2 = gl(3, 2, labels = letters[1:3]))[-(7:8), ]
    mm <- model.matrix(~ f1*f2, dd)
    y <- mm %*% seq_len(ncol(mm)) + rnorm(nrow(mm), sd = 0.1)
    Y <- permute_nvector(10, y)

    resid <- lm.fit(mm, Y)$resid
    lm.rss <- colSums(resid^2)

    # RSS
    expect_equal(lm.rss, calc_mvrss_eigenqr(mm, Y))

    # resid
    expect_equal(resid, calc_resid_eigenqr(mm, Y))

    # generic
    expect_equal(lm.rss, calc_rss_linreg(mm, Y))
    expect_equal(as.matrix(resid), calc_resid_linreg(mm, Y))
})


test_that("calculation of residuals for 3d arrays works", {

    library(qtl)
    data(hyper)
    hyper <- hyper[1,]
    hyper2 <- convert2cross2(hyper)
    map <- insert_pseudomarkers(hyper2$gmap, step=1)
    pr <- calc_genoprob(hyper2, map, error_prob=0.002)
    pr <- aperm(pr[[1]], c(1,3,2)) # reorient to have genomic position last

    # residuals with intercept plus the phenotype
    X <- cbind(1, hyper$pheno[,1])
    expected <- array(0, dim=dim(pr))
    for(i in 1:dim(pr)[3])
        expected[,,i] <- lm(pr[,,i] ~ X)$resid

    resid <- calc_resid_linreg_3d(X, pr)
    expect_equal(expected, resid)

    # residuals with just the intercept
    expected <- array(0, dim=dim(pr))
    for(i in 1:dim(pr)[3])
        expected[,,i] <- lm(pr[,,i] ~ 1)$resid

    resid <- calc_resid_linreg_3d(X[,1,drop=FALSE], pr)
    expect_equal(expected, resid)

})

test_that("linreg full variance-covariance matrix is correct", {

    set.seed(20260608)
    n <- 100
    x <- cbind(1, rnorm(n))
    y <- x %*% c(1,1) + rnorm(n)

    out_lm <- summary(lm(y ~ -1 + x))

    # eigen cholesky method
    out_var <- fit_linreg_eigenchol(x, y, FALSE, TRUE)
    out_se <- fit_linreg_eigenchol(x, y, FALSE, TRUE)

    expect_equivalent(out_se$SE, out_var$SE)
    expect_equivalent(out_var$SE, out_lm$coefficients[,2])
    expect_equivalent(out_var$var, out_lm$cov.unscaled * out_lm$sigma^2)

    # eigen QR method
    out_qr_var <- fit_linreg_eigenqr(x, y, FALSE, TRUE)
    out_qr_se <- fit_linreg_eigenqr(x, y, FALSE, TRUE)

    expect_equivalent(out_se$SE, out_qr_se$SE)
    expect_equivalent(out_se$SE, out_qr_var$SE)
    expect_equivalent(out_var$var, out_qr_var$var)

    # generic version (which just calls fit_linreg_eigenqr)
    out_linreg_var <- fit_linreg(x, y, FALSE, TRUE)
    out_linreg_se <- fit_linreg(x, y, FALSE, TRUE)

    expect_equivalent(out_var$SE, out_linreg_var$SE)
    expect_equivalent(out_var$var, out_linreg_var$var)

})

context("interpolate gen <-> phys maps")

test_that("interpolate_map works", {

    set.seed(31728290)
    oldmap <- sort(runif(20, 0, 100))
    newmap <- sort(runif(20, 0, 200))

    oldpos <- c(-2, seq(0, 100, by=10), 104)

    newpos <- interpolate_map(oldpos, oldmap, newmap)

    # calculate what result should be in R
    expected <- numeric(length(oldpos))
    wh_left <- (oldpos < min(oldmap))
    if(any(wh_left))
        expected[wh_left] <- min(newmap) - (min(oldmap) - oldpos[wh_left])*diff(range(newmap))/diff(range(oldmap))
    wh_right <- (oldpos >= max(oldmap))
    if(any(wh_right))
        expected[wh_right] <- min(newmap) + (oldpos[wh_right] - min(oldmap))*diff(range(newmap))/diff(range(oldmap))
    other <- (!wh_left & !wh_right)
    if(any(other)) {
        index <- sapply(oldpos[other], function(a,b) sum(b <= a), oldmap)
        expected[other] <- newmap[index] + (oldpos[other] - oldmap[index]) *
            (newmap[index+1] - newmap[index]) / (oldmap[index+1] - oldmap[index])
    }

    expect_equal(newpos, expected)

})

test_that("find_intervals works", {
    set.seed(11703070)

    library(qtl)
    data(hyper)
    map <- pull.map(hyper)[[1]]

    pos <- round(runif(1000, 0, 120), 1)
    result <- find_intervals(pos, map)
    interval <- result[,1]
    on_map <- (result[,2]==1)

    # none of the pos are on the map
    expect_equal(pos %in% map, on_map)

    # intervals correct?
    expect_true(all(pos[interval==-1] < min(map)))
    expect_true(all(pos[interval==length(map)-1] >= max(map)))
    middle <- (interval >= 0 & interval < length(map)-1)
    expect_true(all(pos[middle] >= map[interval[middle]+1] &
                    pos[middle] < map[interval[middle]+2]))

    # include some positions that are right on the map
    pos <- sample(c(pos, sample(map, 10)))
    result <- find_intervals(pos, map)
    interval <- result[,1]
    on_map <- (result[,2]==1)

    # intervals correct?
    expect_true(all(pos[interval==-1] < min(map)))
    expect_true(all(pos[interval==length(map)-1] >= max(map)))
    middle <- (interval >= 0 & interval < length(map)-1)
    expect_true(all(pos[middle] >= map[interval[middle]+1] &
                    pos[middle] < map[interval[middle]+2]))

    # also check whether 2nd column is correct
    expect_equal(sum(on_map), 10)
    expect_equal(pos[on_map], map[interval[on_map]+1])

})


test_that("interp_map works in simplest case", {

    set.seed(31728290)
    oldmap <- list("1"=sort(runif(20, 0, 100)),
                   "2"=sort(runif(20, 0, 80)),
                   "3"=sort(runif(20, 0, 60)))

    # newmap is just double oldmap
    newmap <- oldmap
    for(i in seq(along=newmap)) newmap[[i]] <- oldmap[[i]]*2

    chr2 <- list("2"=c(0, 5, 20, 30))
    chr2_doubled <- list("2"=c(0, 10, 40, 60))
    expect_equal(interp_map(chr2, oldmap, newmap),
                 chr2_doubled)

    chr23 <- list("2"=c(0, 5, 20, 30), "3"=c(-5, 20, 40, 100))
    chr23_doubled <- list("2"=c(0, 10, 40, 60), "3"=c(-10, 40, 80, 200))
    expect_equal(interp_map(chr23, oldmap, newmap),
                 chr23_doubled)

})

test_that("interp_map preserves positions", {

    oldmap <- list("1"=c(0, 5,  5,  5, 10, 20))
    newmap <- list("1"=c(0, 5, 10, 15, 20, 30))
    names(oldmap[[1]]) <- names(newmap[[1]]) <- paste0("marker", seq_along(oldmap[[1]]))

    expect_equal(newmap, interp_map(oldmap, oldmap, newmap))

})

context("c++ matrix utilities")

test_that("formX_intcovar works", {

    set.seed(20151130)

    # create a prob matrix
    library(qtl)
    data(fake.f2)
    prob <- aperm(calc.genoprob(fake.f2["2",], step=5)$geno[["2"]]$prob[,,-1],
                  c(1,3,2))
    dimnames(prob) <- NULL

    n <- nind(fake.f2)
    addcovar <- cbind(rep(1,n), sample(0:1, n, replace=TRUE))
    intcovar <- addcovar[,2,drop=FALSE]

    X <- formX_intcovar(prob, addcovar, intcovar, 0)
    expected <- cbind(addcovar, prob[,,1], prob[,,1]*intcovar[,1]) # the [,1] makes intcovar an ordinary vector
    expect_equal(X, expected)

    X <- formX_intcovar(prob, addcovar, intcovar, 5)
    expected <- cbind(addcovar, prob[,,6], prob[,,6]*intcovar[,1]) # the [,1] makes intcovar an ordinary vector
    expect_equal(X, expected)

    # no interactive covariates
    expect_equal(formX_intcovar(prob, addcovar, matrix(ncol=0, nrow=n), 2), cbind(addcovar, prob[,,3]))

    # neither interactive nor additive covariates
    expect_equal(formX_intcovar(prob, matrix(ncol=0, nrow=n), matrix(ncol=0, nrow=n), 3), prob[,,4])

    # mismatch in rows
    expect_error(formX_intcovar(prob, addcovar[-n,], intcovar, 0))
    expect_error(formX_intcovar(prob, addcovar, intcovar[-1,], 2))

    # two interactive covariates
    addcovar <- cbind(addcovar, sample(0:1, n, replace=TRUE))

    intcovar <- addcovar[,-1]

    X <- formX_intcovar(prob, addcovar, intcovar, 4)
    expected <- cbind(addcovar, prob[,,5], prob[,,5]*intcovar[,1], prob[,,5]*intcovar[,2])
    expect_equal(X, expected)

})

test_that("expand_genoprobs_intcovar works", {

    set.seed(20151201)
    library(qtl)
    data(fake.f2)
    pr <- aperm(calc.genoprob(fake.f2["2",], step=5)$geno[["2"]]$prob[,,-1],
                c(1,3,2))
    dimnames(pr) <- NULL
    rownames(pr) <- paste(1:nind(fake.f2))

    intcovar <- cbind(sample(0:1, nrow(pr), replace=TRUE),
                      sample(0:1, nrow(pr), replace=TRUE))

    result <- expand_genoprobs_intcovar(pr, intcovar)

    expect_equal(dim(result), dim(pr)*c(1,ncol(intcovar)+1,1))
    expect_equal(result[,1:2,] , pr)
    expect_equal(result[,3:4,], pr*intcovar[,1])
    expect_equal(result[,5:6,], pr*intcovar[,2])

    # single interactive covariate
    result2 <- expand_genoprobs_intcovar(pr, intcovar[,1,drop=FALSE])

    expect_equal(dim(result2), dim(pr)*c(1,2,1))
    expect_equal(result2[,1:2,] , pr)
    expect_equal(result2[,3:4,], pr*intcovar[,1])

})


test_that("weighted_matrix works", {

    set.seed(20151201)
    n <- 100
    p <- 10
    X <- matrix(rnorm(n*p), ncol=p)
    w <- runif(n, 1, 4)

    result <- weighted_matrix(X, w)

    expect_equal(result, X*w)
    for(i in 1:p) expect_equal(result[,i], X[,i]*w)

})

test_that("weighted_3darray works", {

    set.seed(20151201)
    n <- 100
    p <- 3
    q <- 8
    X <- array(rnorm(n*p*q), dim=c(n, p, q))
    w <- runif(n, 1, 4)

    result <- weighted_3darray(X, w)

    expect_equal(result, X*w)
    for(i in 1:p)
        for(j in 1:q)
            expect_equal(result[,i,j], X[,i,j]*w)

})

test_that("matrix_x_matrix works", {
    X <- matrix(rnorm(200), ncol=10)
    Y <- matrix(rnorm(50), nrow=10)
    expect_equal(matrix_x_matrix(X, Y), X %*% Y)
})

test_that("matrix_x_vector works", {
    X <- matrix(rnorm(200), ncol=10)
    Y <- matrix(rnorm(30), nrow=10)
    expect_equal(matrix_x_vector(X, Y[,1]), (X %*% Y[,1])[,1])
    expect_equal(matrix_x_vector(X, Y[,2]), (X %*% Y[,2])[,1])
    expect_equal(matrix_x_vector(X, Y[,3]), (X %*% Y[,3])[,1])
})

test_that("matrix_x_3darray works", {
    p <- 25
    n <- 50
    g <- 8
    pos <- 100
    X <- matrix(rnorm(p*n), ncol=n)
    Y <- array(rnorm(n*g*pos), dim=c(n, g, pos))

    expect <- array(dim=c(nrow(X), dim(Y)[2:3]))
    for(i in 1:dim(expect)[3])
        expect[,,i] <- X %*% Y[,,i]

    expect_equal(matrix_x_3darray(X, Y), expect)

})

context("basic HMM functions in phase-known heterogeneous stock")

test_that("Phase-known HS nalleles works", {
    expect_equal(nalleles("hspk"), 8)
})

test_that("Phase-known HS check_geno works", {

    # observed genotypes
    for(i in 0:5) {
        # Autosome
        expect_true(test_check_geno("hspk", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("hspk", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_true(test_check_geno("hspk", i, TRUE, TRUE, FALSE, 20))
    }
    for(i in c(-1, 6)) {
        # Autosome
        expect_false(test_check_geno("hspk", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("hspk", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_false(test_check_geno("hspk", i, TRUE, TRUE, FALSE, 20))
    }

    # true genotypes, autosome and female X
    for(i in 1:64) {
        # Autosome
        expect_true(test_check_geno("hspk", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("hspk", i, FALSE, TRUE, TRUE, 20))
    }
    for(i in c(0, 65)) {
        # Autosome
        expect_false(test_check_geno("hspk", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("hspk", i, FALSE, TRUE, TRUE, 20))
    }

    # true genotypes, autosome and female X
    for(i in 64 + 1:8) {
        expect_true(test_check_geno("hspk", i, FALSE, TRUE, FALSE, 20))
    }
    for(i in c(0, 1, 2, 63, 64, 64+9)) {
        expect_false(test_check_geno("hspk", i, FALSE, TRUE, FALSE, 20))
    }

})

test_that("Phase-known HS n_gen works", {

    expect_equal(test_ngen("hspk", FALSE), 64)
    expect_equal(test_ngen("hspk", TRUE),  64+8)

})

test_that("Phase-known HS possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("hspk", FALSE, FALSE, 20), 1:64)

    # X female
    expect_equal(test_possible_gen("hspk", TRUE, TRUE, 20), 1:64)

    # X male
    expect_equal(test_possible_gen("hspk", TRUE, FALSE, 20), 64+(1:8))

})

test_that("Phase-known HS init works", {

    # autosome and female X
    for(i in 1:36) {
        expect_equal(test_init("hspk", i, FALSE, FALSE, 20), log(1/64))
        expect_equal(test_init("hspk", i, TRUE,  TRUE,  20), log(1/64))
    }

    # male X
    for(i in 64 + 1:8)
        expect_equal(test_init("hspk", i, TRUE,  FALSE, 20), log(1/8))

})


test_that("Phase-known HS emit works", {

    fgen <- c(1,3,0,1,3,0,1,3) # founder genotypes; 0=missing, 1=AA, 3=BB
    err <- 0.01

    # Autosome or female X
    # truth = homA: AA (1), AD (7), AG (22), DD (10), DG (25), GG (28)
    expected <- log(c(1-err, err/2, err/2, 1-err/2, err))
    for(trueg in c(1,7,22,10,25,28)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = het: AB (2), AE (11), AH (29), BD (8), BG (23), DE (14), DH (32), EG (26), GH (35)
    expected <- log(c(err/2, 1-err, err/2, 1-err/2, 1-err/2))
    for(trueg in c(2,11,29,8,23,14,32,26,35)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = homB: BB (3), BE (12), BH (30), EE (15), EH (33), HH (36)
    expected <- log(c(err/2, err/2, 1-err, err, 1-err/2))
    for(trueg in c(3,12,30,15,33,36)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = A-: AC (4), AF (16), CD (9), DF (19), CG (24), FG (27)
    expected <- log(c(1-err,1,err,1-err,err))
    for(trueg in c(4,16,9,19,24,27)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = B-: BC (5), BF (17), CE (13), EF (20), CH (31), FH (34)
    expected <- log(c(err,1,1-err,err,1-err))
    for(trueg in c(5,17,13,20,31,34)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }

    # male X: treat het as missing
    # truth = hemA: A (1+64), D (4+64), G (7+64)
    expected <- log(c(1-err, 1, err, 1-err, err))
    for(trueg in 64+c(1,4,7))
        for(obsg in 1:5)
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])
    # truth = hemB: BB (2+64), E (5+64), H (8+64)
    expected <- log(c(err, 1, 1-err, err, 1-err))
    for(trueg in 64+c(2,5,8))
        for(obsg in 1:5)
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])
    # truth = missing: C (3+64), F (6+64)
    expected <- rep(0,5)
    for(trueg in 64+c(3,6))
        for(obsg in 1:5)
            expect_equal(test_emit("hspk", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])


})

test_that("Phase-known HS step works", {

    # autosome
    ng <- 64
    trmat <- matrix(nrow=ng, ncol=ng)
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("hspk", gl, gr, rf, FALSE, FALSE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))
            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

    # female X
    ng <- 64
    trmat <- matrix(nrow=ng, ncol=ng)
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("hspk", gl, gr, rf, FALSE, FALSE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))
            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

    # male X
    ng <- 8
    trmat <- matrix(nrow=ng, ncol=ng)
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("do", 64+gl, 64+gr, rf, TRUE, FALSE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))
            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

})

context("basic HMM functions in 3-way AIL")

test_that("AIL3 nalleles works", {
    expect_equal(nalleles("ail3"), 3)
})

test_that("AIL3 check_geno works", {

    # observed genotypes
    for(i in 0:5) {
        # Autosome
        expect_true(test_check_geno("ail3", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("ail3", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_true(test_check_geno("ail3", i, TRUE, TRUE, FALSE, 20))
    }
    for(i in c(-1, 6)) {
        # Autosome
        expect_false(test_check_geno("ail3", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("ail3", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_false(test_check_geno("ail3", i, TRUE, TRUE, FALSE, 20))
    }

    # true genotypes, autosome and female X
    for(i in 1:6) {
        # Autosome
        expect_true(test_check_geno("ail3", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("ail3", i, FALSE, TRUE, TRUE, 20))
    }
    for(i in c(0, 7)) {
        # Autosome
        expect_false(test_check_geno("ail3", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("ail3", i, FALSE, TRUE, TRUE, 20))
    }

    # true genotypes, autosome and female X
    for(i in 6 + 1:3) {
        expect_true(test_check_geno("ail3", i, FALSE, TRUE, FALSE, 20))
    }
    for(i in c(0:6, 6+4)) {
        expect_false(test_check_geno("ail3", i, FALSE, TRUE, FALSE, 20))
    }

})

test_that("AIL3 n_gen works", {

    expect_equal(test_ngen("ail3", FALSE), 6)
    expect_equal(test_ngen("ail3", TRUE),  9)

})

test_that("AIL3 possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("ail3", FALSE, FALSE, 20), 1:6)

    # X female
    expect_equal(test_possible_gen("ail3", TRUE, TRUE, 20), 1:6)

    # X male
    expect_equal(test_possible_gen("ail3", TRUE, FALSE, 20), 6+(1:3))

})

test_that("AIL3 init works", {

    hom <- cumsum(1:3)
    het <- (1:6)[!((1:6) %in% cumsum(1:3))]
    male <- 6 + (1:3)

    for(i in hom) {
        # autosome and female X
        expect_equal(test_init("ail3", i, FALSE, FALSE, 20), log(1/9))
        expect_equal(test_init("ail3", i, TRUE,  TRUE,  20), log(1/9))
    }

    for(i in het) {
        # autosome and female X
        expect_equal(test_init("ail3", i, FALSE, FALSE, 20), log(2/9))
        expect_equal(test_init("ail3", i, TRUE,  TRUE,  20), log(2/9))
    }

    for(i in male)
        expect_equal(test_init("ail3", i, TRUE,  FALSE, 20), log(1/3))

})


test_that("AIL3 emit works", {

    fgen <- c(1,3,0) # founder genotypes; 0=missing, 1=AA, 3=BB
    err <- 0.01

    # Autosome or female X
    # truth = homA: AA (1)
    expected <- log(c(1-err, err/2, err/2, 1-err/2, err))
    for(trueg in 1) {
        for(obsg in 1:5) {
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = het: AB (2)
    expected <- log(c(err/2, 1-err, err/2, 1-err/2, 1-err/2))
    for(trueg in 2) {
        for(obsg in 1:5) {
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = homB: BB (3)
    expected <- log(c(err/2, err/2, 1-err, err, 1-err/2))
    for(trueg in 3) {
        for(obsg in 1:5) {
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = A-: AC (4)
    expected <- log(c(1-err,1,err,1-err,err))
    for(trueg in 4) {
        for(obsg in 1:5) {
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = B-: BC (5)
    expected <- log(c(err,1,1-err,err,1-err))
    for(trueg in 5) {
        for(obsg in 1:5) {
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }

    # male X: treat het as missing
    # truth = hemA: A (1+6)
    expected <- log(c(1-err, 1, err, 1-err, err))
    for(trueg in 7)
        for(obsg in 1:5)
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])
    # truth = hemB: BB (2+6)
    expected <- log(c(err, 1, 1-err, err, 1-err))
    for(trueg in 8)
        for(obsg in 1:5)
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])
    # truth = missing: C (3+6)
    expected <- rep(0,5)
    for(trueg in 9)
        for(obsg in 1:5)
            expect_equal(test_emit("ail3", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])

})

test_that("AIL3 step works for autosome", {

    pAA <- function(n, r, k=3)
    {
        stopifnot(n >= 2)
        (1 - (1 - k + k*r)*(1-r)^(n-2))/k^2
    }

    rf <- 0.02
    ngen <- 10

    pAA <- pAA(ngen, rf)
    R <- (1 - 3*pAA)

    # expected values
    expected <- matrix(ncol=6, nrow=6)
    # AA->AA
    expected[1,1] <- expected[3,3] <- expected[6,6] <- (1-R)^2
    # AB->AB
    expected[2,2] <- expected[4,4] <- expected[5,5] <- (1-R)^2 + (R/2)^2
    # AA->BB
    expected[1,3] <- expected[1,6] <- expected[3,6] <- (R/2)^2
    # AA->AB
    expected[1,2] <- expected[1,4] <- expected[3,5] <- expected[2,3] <-
        expected[4,6] <- expected[5,6] <- (1-R)*(R/2)
    # AA->BC
    expected[1,5] <- expected[3,4] <- expected[2,6] <- (R/2)^2
    # AB->AC
    expected[2,4] <- expected[2,5] <- expected[4,5] <- (1-R)*(R/2) + (R/2)^2
    expected[lower.tri(expected)] <- t(expected)[lower.tri(expected)]
    expected <- log(expected)

    result <- matrix(ncol=6, nrow=6)
    for(i in 1:6)
        for(j in 1:6)
            result[i,j] <- test_step("ail3", i, j, rf, FALSE, FALSE, ngen)

    expect_equal(result, expected)

})


test_that("AIL3 step works for X chromosome", {

    pAA <- function(n, r, k=3)
    {
        stopifnot(n >= 2)
        z <- sqrt((1-r)*(9-r))

        male <- (1-r)/k * ( (1-r+z)/(2*z) * ((1-r-z)/4)^(n-2) +
                            (-1+r+z)/(2*z) * ((1-r+z)/4)^(n-2)) +
            (2-r)/(2*k) * ( (1-r-z)/2 * (1-r+z)/(2*z) * ((1-r-z)/4)^(n-2) +
                            (1-r+z)/2 * (-1+r+z)/(2*z) * ((1-r+z)/4)^(n-2)) +
            ( (r^2 + r*(z-5))/(k^2*(3+r+z)) * (1-r+z)/(2*z) * ((1-r-z)/4)^(n-2) +
              (r^2 - r*(z+5))/(k^2*(3+r-z)) * (-1+r+z)/(2*z) * ((1-r+z)/4)^(n-2) + 1/k^2)

        female <- (1-r)/k * ( (-1/z)*((1-r-z)/4)^(n-2) +
                              (1/z)*((1-r+z)/4)^(n-2)) +
            (2-r)/(2*k) * ( (1-r-z)/2 * (-1/z)*((1-r-z)/4)^(n-2) +
                            (1-r+z)/2 *  (1/z)*((1-r+z)/4)^(n-2)) +
            ( (r^2 + r*(z-5))/(k^2*(3+r+z)) * (-1/z)*((1-r-z)/4)^(n-2) +
              (r^2 - r*(z+5))/(k^2*(3+r-z)) * (1/z)*((1-r+z)/4)^(n-2)  + 1/k^2)

        if(any(r==0))
            male[r==0] <- female[r==0] <- 1/k

        if(any(r==1)) {
            if(n==2) male[r==1] <- 0
            if(n>2) male[r==1] <- 1/k^2
            if(n==2) female[r==1] <- 1/(2*k)
            if(n==3) female[r==1] <- 1/(2*k^2)
            if(n>3)  female[r==1] <- 1/k^2
        }

        list(female=female, male=male)
    }

    rf <- 0.02
    ngen <- 10

    pAA <- pAA(ngen, rf)

    # female X
    R <- (1 - 3*pAA$female)

    # expected values
    expected <- matrix(ncol=6, nrow=6)
    # AA->AA
    expected[1,1] <- expected[3,3] <- expected[6,6] <- (1-R)^2
    # AB->AB
    expected[2,2] <- expected[4,4] <- expected[5,5] <- (1-R)^2 + (R/2)^2
    # AA->BB
    expected[1,3] <- expected[1,6] <- expected[3,6] <- (R/2)^2
    # AA->AB
    expected[1,2] <- expected[1,4] <- expected[3,5] <- expected[2,3] <-
        expected[4,6] <- expected[5,6] <- (1-R)*(R/2)
    # AA->BC
    expected[1,5] <- expected[3,4] <- expected[2,6] <- (R/2)^2
    # AB->AC
    expected[2,4] <- expected[2,5] <- expected[4,5] <- (1-R)*(R/2) + (R/2)^2
    expected[lower.tri(expected)] <- t(expected)[lower.tri(expected)]
    expected <- log(expected)

    result <- matrix(ncol=6, nrow=6)
    for(i in 1:6)
        for(j in 1:6)
            result[i,j] <- test_step("ail3", i, j, rf, TRUE, TRUE, ngen)

    expect_equal(result, expected)

    # male X
    R <- (1 - 3*pAA$male)

    # expected values
    expected <- matrix(R/2, ncol=3, nrow=3)
    diag(expected) <- 1-R
    expected <- log(expected)

    result <- matrix(ncol=3, nrow=3)
    for(i in 1:3)
        for(j in 1:3)
            result[i,j] <- test_step("ail3", 6+i, 6+j, rf, TRUE, FALSE, ngen)

    expect_equal(result, expected)

})



test_that("geno_names works", {
    auto <- c("AA", "AB", "BB", "AC", "BC", "CC")
    X <- c(auto, paste0(LETTERS[1:3], "Y"))

    expect_equal(geno_names("ail3", LETTERS[1:3], FALSE), auto)
    expect_equal(geno_names("ail3", LETTERS[1:3], TRUE), X)
})



test_that("nrec works", {

    # autosome genotypes = 1:6
    expected <- rbind(c(0,1,2,1,2,2),
                      c(1,0,1,1,1,2),
                      c(2,1,0,2,1,2),
                      c(1,1,2,0,1,1),
                      c(2,1,1,1,0,1),
                      c(2,2,2,1,1,0))
    for(i in 1:6)
        for(j in 1:6) {
            expect_equal(test_nrec("ail3", i, j, FALSE, FALSE, 0), expected[i,j])
            expect_equal(test_nrec("ail3", i, j, TRUE, TRUE, 0), expected[i,j])
        }

    # X chromosome male
    expected <- matrix(1, ncol=3, nrow=3) - diag(rep(1,3))
    for(i in 1:3)
        for(j in 1:3)
            expect_equal(test_nrec("ail3", i+6, j+6, TRUE, FALSE, 0), expected[i,j])

})

context("basic HMM functions in riself")

test_that("riself nalleles works", {
    expect_equal(nalleles("riself"), 2)
})

test_that("riself check_geno works", {

    expect_true(test_check_geno("riself", 0, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("riself", 1, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("riself", 2, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("riself", 1, FALSE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("riself", 2, FALSE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("riself", 3, TRUE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("riself", 0, FALSE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("riself", 3, FALSE, FALSE, FALSE, integer(0)))

})

test_that("riself ngen works", {
    expect_equal(test_ngen("riself", FALSE), 2)
})

test_that("riself possible_gen works", {

    expect_equal(test_possible_gen("riself", FALSE, FALSE, integer(0)), 1:2)

})

test_that("riself nrec works", {

    expect_equal(test_nrec("riself", 1, 1, FALSE, FALSE, integer(0)), 0)
    expect_equal(test_nrec("riself", 1, 2, FALSE, FALSE, integer(0)), 1)
    expect_equal(test_nrec("riself", 2, 1, FALSE, FALSE, integer(0)), 1)
    expect_equal(test_nrec("riself", 2, 2, FALSE, FALSE, integer(0)), 0)

})

test_that("riself init works", {

    expect_equal(test_init("riself", 1, FALSE, FALSE, integer(0)), log(0.5))
    expect_equal(test_init("riself", 2, FALSE, FALSE, integer(0)), log(0.5))

})

test_that("riself emit works", {

    eps <- 0.01
    expect_equal(test_emit("riself", 0, 1, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("riself", 0, 2, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("riself", 1, 1, eps, integer(0), FALSE, FALSE, integer(0)), log(1-eps))
    expect_equal(test_emit("riself", 1, 2, eps, integer(0), FALSE, FALSE, integer(0)), log(eps))
    expect_equal(test_emit("riself", 2, 1, eps, integer(0), FALSE, FALSE, integer(0)), log(eps))
    expect_equal(test_emit("riself", 2, 2, eps, integer(0), FALSE, FALSE, integer(0)), log(1-eps))

    # errors
    expect_equal(test_emit("riself", 0, 0, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("riself", 0, 3, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("riself", 3, 1, eps, integer(0), FALSE, FALSE, integer(0)), 0)

})

test_that("riself step works", {

    rf <- 0.01
    RF <- 2*rf/(1+2*rf)
    expect_equal(test_step("riself", 1, 1, rf, FALSE, FALSE, integer(0)), log(1-RF))
    expect_equal(test_step("riself", 1, 2, rf, FALSE, FALSE, integer(0)), log(RF))
    expect_equal(test_step("riself", 2, 1, rf, FALSE, FALSE, integer(0)), log(RF))
    expect_equal(test_step("riself", 2, 2, rf, FALSE, FALSE, integer(0)), log(1-RF))

})

test_that("geno_names works", {
    expect_equal(geno_names("riself", c("B", "R"), FALSE), c("BB", "RR"))
    expect_equal(geno_names("riself", c("B", "R"), TRUE), c("BB", "RR"))
})

context("genoprobs -> snpprobs")

test_that(".alleleprob_to_snpprob works in simple cases", {

    # conversion correct, with 1 map position
    pr <- array(c(0.15, 0.03, 0.20, 0.08, 0.18, 0.09, 0.15, 0.11),
                dim=c(1, 8, 1))
    sdp <- 1:255
    interval <- rep(0, length(sdp))
    on_map <- rep(TRUE, length(sdp))

    # sdp -> genotype matrix
    g <- matrix(rep(sdp,rep(8, length(sdp))), ncol=length(sdp))
    g <- (g %/% 2^(0:7)) %% 2

    expected <- array(dim=c(1, 2, length(sdp)))
    for(i in 1:length(sdp)) {
        expected[1,1,i] <- sum(pr[1,g[,i]==0,1])
        expected[1,2,i] <- sum(pr[1,g[,i]==1,1])
    }

    expect_equal(.alleleprob_to_snpprob(pr, sdp, interval, on_map),
                 expected)


    # add a second map position
    pr <- array(c(0.15, 0.03, 0.20, 0.08, 0.18, 0.09, 0.15, 0.11,
                  0.07, 0.05, 0.17, 0.02, 0.46, 0.03, 0.11, 0.08),
                dim=c(1, 8, 2))

    prsum <- (pr[,,1] + pr[,,2])/2
    interval[201:255] <- 1   # last 50 at 2nd marker
    on_map[151:200] <- FALSE # 50 before that in-between

    expected <- array(dim=c(1, 2, length(sdp)))
    for(i in 1:150) {
        expected[1,1,i] <- sum(pr[1,g[,i]==0,1])
        expected[1,2,i] <- sum(pr[1,g[,i]==1,1])
    }
    for(i in 151:200) {
        expected[1,1,i] <- sum((pr[1,g[,i]==0,1] + pr[1,g[,i]==0,2])/2)
        expected[1,2,i] <- sum((pr[1,g[,i]==1,1] + pr[1,g[,i]==1,2])/2)
    }
    for(i in 201:255) {
        expected[1,1,i] <- sum(pr[1,g[,i]==0,2])
        expected[1,2,i] <- sum(pr[1,g[,i]==1,2])
    }

    expect_equal(.alleleprob_to_snpprob(pr, sdp, interval, on_map),
                 expected)

})

test_that("genocol_to_snpcol works with 8 alleles", {

    # A has one allele and rest have another
    expected <- rep(0, 36)
    expected[1] <- 2
    expected[c(2,4,7,11,16,22,29)] <- 1
    expect_equal(genocol_to_snpcol(8,1), expected)
    # opposite
    expect_equal(genocol_to_snpcol(8,254), 2-expected)

    # B has one allele and rest have another
    expected <- rep(0, 36)
    expected[3] <- 2
    expected[c(2,5,8,12,17,23,30)] <- 1
    expect_equal(genocol_to_snpcol(8,2), expected)
    # opposite
    expect_equal(genocol_to_snpcol(8,253), 2-expected)

    # E has one allele and rest have another
    expected <- rep(0, 36)
    expected[15] <- 2
    expected[c(11,12,13,14,20,26,33)] <- 1
    expect_equal(genocol_to_snpcol(8,16), expected)
    # opposite
    expect_equal(genocol_to_snpcol(8,239), 2-expected)

    # H has one allele and rest have another
    expected <- rep(0, 36)
    expected[36] <- 2
    expected[29:35] <- 1
    expect_equal(genocol_to_snpcol(8,128), expected)
    # opposite
    expect_equal(genocol_to_snpcol(8,127), 2-expected)

    # ABCD have one allele and EFGH has
    expected <- rep(1, 36)
    expected[1:10] <- 2
    expected[c(15,20,21,26:28,33:36)] <- 0
    expect_equal(genocol_to_snpcol(8,15), expected)
    # opposite
    expect_equal(genocol_to_snpcol(8,255-15), 2-expected)

    # ACEG have one allele and BDFH has
    expected <- rep(1, 36)
    expected[c(1,4,6,11,13,15,22,24,26,28)] <- 2
    expected[c(3,8,10,17,19,21,30,32,34,36)] <- 0
    expect_equal(genocol_to_snpcol(8,85), expected)
    # opposite
    expect_equal(genocol_to_snpcol(8,255-85), 2-expected)

    # BE have one allele and ACDFGH have other
    expected <- rep(0, 36)
    expected[c(3,12,15)] <- 2
    expected[c(2,5,8,17,23,30,11,13,14,20,26,33)] <- 1
    expect_equal(genocol_to_snpcol(8,18), expected)
    # opposite
    expect_equal(genocol_to_snpcol(8,255-18), 2-expected)

    # test all opposites
    for(g in 1:254)
        expect_equal(genocol_to_snpcol(8, g),
                     2 - genocol_to_snpcol(8, 255-g))

})

test_that("genocol_to_snpcol works with 4 alleles", {

    # A,BCD (1)
    expect_equal(genocol_to_snpcol(4, 1),
                 c(2,1,0,1,0,0,1,0,0,0))
    expect_equal(genocol_to_snpcol(4, 15-1),
                 2-c(2,1,0,1,0,0,1,0,0,0))
    # B,ACD (2)
    expect_equal(genocol_to_snpcol(4, 2),
                 c(0,1,2,0,1,0,0,1,0,0))
    expect_equal(genocol_to_snpcol(4, 15-2),
                 2-c(0,1,2,0,1,0,0,1,0,0))
    # C,ABD (4)
    expect_equal(genocol_to_snpcol(4, 4),
                 c(0,0,0,1,1,2,0,0,1,0))
    expect_equal(genocol_to_snpcol(4, 15-4),
                 2-c(0,0,0,1,1,2,0,0,1,0))
    # D,ABC (8)
    expect_equal(genocol_to_snpcol(4, 8),
                 c(0,0,0,0,0,0,1,1,1,2))
    expect_equal(genocol_to_snpcol(4, 15-8),
                 2-c(0,0,0,0,0,0,1,1,1,2))
    # AB,CD (3)
    expect_equal(genocol_to_snpcol(4, 3),
                 c(2,2,2,1,1,0,1,1,0,0))
    expect_equal(genocol_to_snpcol(4, 15-3),
                 2-c(2,2,2,1,1,0,1,1,0,0))
    # AC,BD (5)
    expect_equal(genocol_to_snpcol(4, 5),
                 c(2,1,0,2,1,2,1,0,1,0))
    expect_equal(genocol_to_snpcol(4, 15-5),
                 2-c(2,1,0,2,1,2,1,0,1,0))
    # AD,BC (9)
    expect_equal(genocol_to_snpcol(4, 9),
                 c(2,1,0,1,0,0,2,1,1,2))
    expect_equal(genocol_to_snpcol(4, 15-9),
                 2-c(2,1,0,1,0,0,2,1,1,2))
})

test_that(".genoprob_to_snpprob works with 8 alleles", {

    # make a fake probability matrix
    set.seed(36319176)
    npos <- 8
    nind <- 5
    x <- matrix(runif(npos*nind*36), ncol=36)
    x <- x/rowSums(x)
    pr <- array(dim=c(nind, 36, npos))
    k <- 1
    for(i in 1:nind) {
        for(j in 1:npos) {
            pr[i,,j] <- x[k,]
            k <- k+1
        }
    }

    # errors
    expect_error(.genoprob_to_snpprob(pr[,-1,], 1, 0, TRUE)) # not enough columns
    expect_error(.genoprob_to_snpprob(pr[,-(1:2),], 1, 0, TRUE)) # not enough columns
    expect_error(.genoprob_to_snpprob(pr, c(1,0,200), c(0,1,2), c(TRUE,FALSE,FALSE))) # SDP out of range
    expect_error(.genoprob_to_snpprob(pr, c(1,2,256), c(0,1,2), c(TRUE,FALSE,FALSE))) # SDP out of range
    expect_error(.genoprob_to_snpprob(pr, c(1,255,200), c(0,1,-1), c(TRUE,FALSE,FALSE))) # snp out of map range
    expect_error(.genoprob_to_snpprob(pr, c(1,2,254), c(0,1,npos-1), c(TRUE,FALSE,FALSE)))    # snp out of map range

    nsnp <- 200
    sdp <- sample(255, nsnp, replace=TRUE)
    interval <- sample(0:(npos-1), nsnp, replace=TRUE)
    on_map <- sample(c(FALSE, TRUE), nsnp, replace=TRUE)
    on_map[interval==npos-1] <- TRUE

    # calculate snp genotype probabilities
    result <- .genoprob_to_snpprob(pr, sdp, interval, on_map)
    expect_equal(dim(result), c(nind,3,nsnp))

    # calculate in R
    expected <- array(dim=c(nind, 3, nsnp))
    for(snp in 1:nsnp) {
        snpcol <- genocol_to_snpcol(8, sdp[snp])
        for(snpg in 1:3) {
            if(on_map[snp])
                expected[,snpg,snp] <- rowSums(pr[,snpcol==snpg-1,interval[snp]+1,drop=FALSE])
            else
                expected[,snpg,snp] <- rowSums((pr[,snpcol==snpg-1,interval[snp]+1,drop=FALSE] +
                                                pr[,snpcol==snpg-1,interval[snp]+2,drop=FALSE])/2)
        }
    }
    expect_equal(result, expected)

})


test_that("Xgenocol_to_snpcol works with 8 alleles", {

    # flip expected when flipping SDP
    opp <- function(x) c(2-x[1:36],7-x[37:44])

    # A has one allele and rest have another
    expected <- rep(0, 44)
    expected[1] <- 2
    expected[c(2,4,7,11,16,22,29)] <- 1
    expected[37:44] <- c(4,3,3,3,3,3,3,3)
    expect_equal(Xgenocol_to_snpcol(8,1), expected)
    # opposite
    expect_equal(Xgenocol_to_snpcol(8,254), opp(expected))

    # B has one allele and rest have another
    expected <- rep(0, 44)
    expected[3] <- 2
    expected[c(2,5,8,12,17,23,30)] <- 1
    expected[37:44] <- c(3,4,3,3,3,3,3,3)
    expect_equal(Xgenocol_to_snpcol(8,2), expected)
    # opposite
    expect_equal(Xgenocol_to_snpcol(8,253), opp(expected))

    # E has one allele and rest have another
    expected <- rep(0, 44)
    expected[15] <- 2
    expected[c(11,12,13,14,20,26,33)] <- 1
    expected[37:44] <- c(3,3,3,3,4,3,3,3)
    expect_equal(Xgenocol_to_snpcol(8,16), expected)
    # opposite
    expect_equal(Xgenocol_to_snpcol(8,239), opp(expected))

    # H has one allele and rest have another
    expected <- rep(0, 44)
    expected[36] <- 2
    expected[29:35] <- 1
    expected[37:44] <- c(3,3,3,3,3,3,3,4)
    expect_equal(Xgenocol_to_snpcol(8,128), expected)
    # opposite
    expect_equal(Xgenocol_to_snpcol(8,127), opp(expected))

    # ABCD have one allele and EFGH has
    expected <- rep(1, 44)
    expected[1:10] <- 2
    expected[c(15,20,21,26:28,33:36)] <- 0
    expected[37:44] <- c(4,4,4,4,3,3,3,3)
    expect_equal(Xgenocol_to_snpcol(8,15), expected)
    # opposite
    expect_equal(Xgenocol_to_snpcol(8,255-15), opp(expected))

    # ACEG have one allele and BDFH has
    expected <- rep(1, 44)
    expected[c(1,4,6,11,13,15,22,24,26,28)] <- 2
    expected[c(3,8,10,17,19,21,30,32,34,36)] <- 0
    expected[37:44] <- c(4,3,4,3,4,3,4,3)
    expect_equal(Xgenocol_to_snpcol(8,85), expected)
    # opposite
    expect_equal(Xgenocol_to_snpcol(8,255-85), opp(expected))

    # BE have one allele and ACDFGH have other
    expected <- rep(0, 36)
    expected[c(3,12,15)] <- 2
    expected[c(2,5,8,17,23,30,11,13,14,20,26,33)] <- 1
    expected[37:44] <- c(3,4,3,3,4,3,3,3)
    expect_equal(Xgenocol_to_snpcol(8,18), expected)
    # opposite
    expect_equal(Xgenocol_to_snpcol(8,255-18), opp(expected))

    # test all opposites
    for(g in 1:254)
        expect_equal(Xgenocol_to_snpcol(8, g),
                     opp(Xgenocol_to_snpcol(8, 255-g)))

})

test_that("Xgenocol_to_snpcol works with 4 alleles", {

    # A,BCD (1)
    expect_equal(Xgenocol_to_snpcol(4, 1),
                 c(2,1,0,1,0,0,1,0,0,0,4,3,3,3))
    expect_equal(Xgenocol_to_snpcol(4, 15-1),
                 c(0,1,2,1,2,2,1,2,2,2,3,4,4,4))
    # B,ACD (2)
    expect_equal(Xgenocol_to_snpcol(4, 2),
                 c(0,1,2,0,1,0,0,1,0,0,3,4,3,3))
    expect_equal(Xgenocol_to_snpcol(4, 15-2),
                 c(2,1,0,2,1,2,2,1,2,2,4,3,4,4))
    # C,ABD (4)
    expect_equal(Xgenocol_to_snpcol(4, 4),
                 c(0,0,0,1,1,2,0,0,1,0,3,3,4,3))
    expect_equal(Xgenocol_to_snpcol(4, 15-4),
                 c(2,2,2,1,1,0,2,2,1,2,4,4,3,4))
    # D,ABC (8)
    expect_equal(Xgenocol_to_snpcol(4, 8),
                 c(0,0,0,0,0,0,1,1,1,2,3,3,3,4))
    expect_equal(Xgenocol_to_snpcol(4, 15-8),
                 c(2,2,2,2,2,2,1,1,1,0,4,4,4,3))
    # AB,CD (3)
    expect_equal(Xgenocol_to_snpcol(4, 3),
                 c(2,2,2,1,1,0,1,1,0,0,4,4,3,3))
    expect_equal(Xgenocol_to_snpcol(4, 15-3),
                 c(0,0,0,1,1,2,1,1,2,2,3,3,4,4))
    # AC,BD (5)
    expect_equal(Xgenocol_to_snpcol(4, 5),
                 c(2,1,0,2,1,2,1,0,1,0,4,3,4,3))
    expect_equal(Xgenocol_to_snpcol(4, 15-5),
                 c(0,1,2,0,1,0,1,2,1,2,3,4,3,4))
    # AD,BC (9)
    expect_equal(Xgenocol_to_snpcol(4, 9),
                 c(2,1,0,1,0,0,2,1,1,2,4,3,3,4))
    expect_equal(Xgenocol_to_snpcol(4, 15-9),
                 c(0,1,2,1,2,2,0,1,1,0,3,4,4,3))
})


test_that(".Xgenoprob_to_snpprob works with 8 alleles", {

    # make a fake probability matrix
    set.seed(20939472)
    npos <- 8
    nind <- 5
    x <- matrix(runif(npos*nind*36), ncol=36)
    x <- x/rowSums(x)
    x2 <- matrix(runif(npos*nind*8), ncol=8)
    x2 <- x2/rowSums(x2)
    x <- rbind(cbind(x, matrix(0, ncol=8, nrow=nrow(x))),
               cbind(matrix(0, ncol=36, nrow=nrow(x2)), x2))
    nind <- nind*2

    pr <- array(dim=c(nind, 44, npos))
    k <- 1
    for(i in 1:nind) {
        for(j in 1:npos) {
            pr[i,,j] <- x[k,]
            k <- k+1
        }
    }

    # errors
    expect_error(.Xgenoprob_to_snpprob(pr[,-1,], 1, 0, TRUE)) # not enough columns
    expect_error(.Xgenoprob_to_snpprob(pr[,-(1:2),], 1, 0, TRUE)) # not enough columns
    expect_error(.Xgenoprob_to_snpprob(pr, c(1,0,200), c(0,1,2), c(TRUE,FALSE,FALSE))) # SDP out of range
    expect_error(.Xgenoprob_to_snpprob(pr, c(1,2,256), c(0,1,2), c(TRUE,FALSE,FALSE))) # SDP out of range
    expect_error(.Xgenoprob_to_snpprob(pr, c(1,255,200), c(0,1,-1), c(TRUE,FALSE,FALSE))) # snp out of map range
    expect_error(.Xgenoprob_to_snpprob(pr, c(1,2,254), c(0,1,npos-1), c(TRUE,FALSE,FALSE)))    # snp out of map range

    nsnp <- 200
    sdp <- sample(255, nsnp, replace=TRUE)
    interval <- sample(0:(npos-1), nsnp, replace=TRUE)
    on_map <- sample(c(FALSE, TRUE), nsnp, replace=TRUE)
    on_map[interval==npos-1] <- TRUE

    # calculate snp genotype probabilities
    result <- .Xgenoprob_to_snpprob(pr, sdp, interval, on_map)
    expect_equal(dim(result), c(nind,5,nsnp))

    # calculate in R
    expected <- array(dim=c(nind, 5, nsnp))
    for(snp in 1:nsnp) {
        snpcol <- Xgenocol_to_snpcol(8, sdp[snp])
        for(snpg in 1:5) {
            if(on_map[snp])
                expected[,snpg,snp] <- rowSums(pr[,snpcol==snpg-1,interval[snp]+1,drop=FALSE])
            else
                expected[,snpg,snp] <- rowSums((pr[,snpcol==snpg-1,interval[snp]+1,drop=FALSE] +
                                                pr[,snpcol==snpg-1,interval[snp]+2,drop=FALSE])/2)
        }
    }
    expect_equal(result, expected)

})

test_that("the genoprob_to_snpprob R function works", {

    skip_if(isnt_karl(), "this test only run locally")

    # load example data and calculate genotype probabilities
    file <- paste0("https://raw.githubusercontent.com/rqtl/",
                   "qtl2data/main/DO_Recla/recla.zip")
    recla <- read_cross2(file)
    recla <- recla[c(1:2,53:54), c("19","X")] # subset to 4 mice and 2 chromosomes
    probs <- calc_genoprob(recla, err=0.002)

    # founder genotypes for a set of SNPs
    snpgeno <- rbind(m1=c(3,1,1,3,1,1,1,1),
                     m2=c(1,3,1,3,1,3,1,3),
                     m3=c(1,1,1,1,3,3,3,3),
                     m4=c(1,3,1,3,1,3,1,3),
                     m5=c(1,1,1,3,1,3,1,1),
                     m6=c(1,1,3,1,1,1,3,1))
    sdp <- calc_sdp(snpgeno)
    snpinfo <- data.frame(chr=c("19", "19", "X", "X", "19", "X"),
                          pos=c(40.36, 40.53, 110.91, 111.21, 56.480843, 56.480843),
                          sdp=sdp,
                          snp=c("m1", "m2", "m3", "m4", "m5", "m6"), stringsAsFactors=FALSE)

    # identify equivalent snps
    snpinfo19 <- index_snps(recla$pmap, snpinfo[snpinfo$chr=="19",])
    snpinfoX <- index_snps(recla$pmap, snpinfo[snpinfo$chr=="X",])
    snpinfo <- index_snps(recla$pmap, snpinfo)

    snpprob <- genoprob_to_snpprob(probs, snpinfo)
    snpprob19 <- genoprob_to_snpprob(probs, snpinfo19)
    snpprobX <- genoprob_to_snpprob(probs, snpinfoX)

    # test combination
    expect_equal(snpprob, cbind(snpprob19, snpprobX))

    # construct expected for chr 19
    int19 <- find_intervals(snpinfo$pos[snpinfo$chr=="19"], recla$pmap[["19"]])
    expected19 <- array(dim=c(nrow(probs[["19"]]), 3, nrow(int19)))
    for(i in 1:nrow(int19)) {
        gencol <- genocol_to_snpcol(8, snpinfo$sdp[snpinfo$chr=="19"][i])
        for(j in 1:3) {
            if(int19[i,2]==0)
                expected19[,j,i] <- rowSums((probs[["19"]][,gencol==j-1,int19[i,1]+1] +
                                           probs[["19"]][,gencol==j-1,int19[i,1]+2])/2)
            else
                expected19[,j,i] <- rowSums(probs[["19"]][,gencol==j-1,int19[i,1]+1])
        }
    }
    dimnames(expected19) <- dimnames(snpprob19[["19"]])
    expect_equivalent(snpprob19[["19"]], expected19)

    # construct expected for X chr
    intX <- find_intervals(snpinfo$pos[snpinfo$chr=="X"], recla$pmap$X)
    expectedX <- array(dim=c(nrow(probs[["X"]]), 5, nrow(intX)))
    for(i in 1:nrow(intX)) {
        sdp <- snpinfo$sdp[snpinfo$chr=="X"][i]
        gencol <- genocol_to_snpcol(8, sdp)
        yg <- invert_sdp(sdp, 8)
        gencol <- c(gencol, (yg-1)/2+3) # add hemizygous males
        for(j in 1:5) {
            if(intX[i,2]==0)
                expectedX[,j,i] <- rowSums((probs[["X"]][,gencol==j-1,intX[i,1]+1] +
                                             probs[["X"]][,gencol==j-1,intX[i,1]+2])/2)
            else
                expectedX[,j,i] <- rowSums(probs[["X"]][,gencol==j-1,intX[i,1]+1])
        }
    }
    dimnames(expectedX) <- dimnames(snpprob$X)
    expect_equivalent(snpprobX$X, expectedX)
    expect_equivalent(snpprob, list("19"=expected19, X=expectedX))

    # repeat with some redundant SNPs
    snpinfo2 <- rbind(snpinfo[,1:4], data.frame(chr=c("19", "19", "X", "X"),
                                                pos=c(40.41, 40.49, 110.99, 111.01),
                                                sdp=c(170,9,170,240),
                                                snp=c("m7","m8","m9","m10"),
                                                stringsAsFactors=FALSE))

    # identify equivalent snps
    snpinfo19 <- index_snps(recla$pmap, snpinfo2[snpinfo2$chr=="19",])
    snpinfoX <- index_snps(recla$pmap, snpinfo2[snpinfo2$chr=="X",])
    snpinfo2 <- index_snps(recla$pmap, snpinfo2)

    snpprob2 <- genoprob_to_snpprob(probs, snpinfo2)
    expect_equivalent(snpprob2$probs, snpprob$probs)

    # convert to allele probabilities
    aprobs <- genoprob_to_alleleprob(probs)

    # collapse to snp probabilities
    snpaprob <- genoprob_to_snpprob(aprobs, snpinfo2)
    snpaprob19 <- genoprob_to_snpprob(aprobs, snpinfo19)
    snpaprobX <- genoprob_to_snpprob(aprobs, snpinfoX)

    # test combination
    expect_equal(snpaprob, cbind(snpaprob19, snpaprobX))

    # construct expected for chr 19
    expected19 <- array(dim=c(nrow(probs[["19"]]), 2, nrow(int19)))
    for(i in 1:nrow(int19)) {
        gencol <- (invert_sdp(snpinfo$sdp[snpinfo$chr=="19"][i], 8)-1)/2
        for(j in 1:2) {
            if(int19[i,2]==0)
                expected19[,j,i] <- rowSums((aprobs[["19"]][,gencol==j-1,int19[i,1]+1] +
                                             aprobs[["19"]][,gencol==j-1,int19[i,1]+2])/2)
            else
                expected19[,j,i] <- rowSums(aprobs[["19"]][,gencol==j-1,int19[i,1]+1])
        }
    }
    expect_equivalent(snpaprob19, list("19"=expected19))

    # construct expected for X chr
    intX <- find_intervals(snpinfo$pos[snpinfo$chr=="X"], recla$pmap$X)
    expectedX <- array(dim=c(nrow(probs[["X"]]), 2, nrow(int19)))
    for(i in 1:nrow(intX)) {
        sdp <- snpinfo$sdp[snpinfo$chr=="X"][i]
        gencol <- (invert_sdp(sdp, 8)-1)/2
        for(j in 1:2) {
            if(intX[i,2]==0)
                expectedX[,j,i] <- rowSums((aprobs[["X"]][,gencol==j-1,intX[i,1]+1] +
                                            aprobs[["X"]][,gencol==j-1,intX[i,1]+2])/2)
            else
                expectedX[,j,i] <- rowSums(aprobs[["X"]][,gencol==j-1,intX[i,1]+1])
        }
    }
    expect_equivalent(snpaprobX, list(X=expectedX))
    expect_equivalent(snpaprob, list("19"=expected19, X=expectedX))

    # repeat with some redundant SNPs
    snpaprob2 <- genoprob_to_snpprob(aprobs, snpinfo2)
    expect_equivalent(snpprob2, snpprob)


    # expect error if snpinfo is missing columns
    snpinfo2a <- snpinfo2[,colnames(snpinfo2) != "sdp"]
    expect_error(genoprob_to_snpprob(aprobs, snpinfo2a))
    snpinfo2a <- snpinfo2[,colnames(snpinfo2) != "index"]
    expect_error(genoprob_to_snpprob(aprobs, snpinfo2a))
    snpinfo2a <- snpinfo2[,colnames(snpinfo2) != "interval"]
    expect_error(genoprob_to_snpprob(aprobs, snpinfo2a))
    snpinfo2a <- snpinfo2[,colnames(snpinfo2) != "on_map"]
    expect_error(genoprob_to_snpprob(aprobs, snpinfo2a))

})

context("basic HMM functions in 4- and 8-way RIL by sib-mating")

test_that("risib4-8 n_gen, n_alleles work", {

    expect_equal(nalleles("risib4"), 4)
    expect_equal(nalleles("risib8"), 8)

    expect_equal(test_ngen("risib4", FALSE), 4)
    expect_equal(test_ngen("risib8", FALSE), 8)

    # X chromosome
    expect_equal(test_ngen("risib4", TRUE), 4)
    expect_equal(test_ngen("risib8", TRUE), 8)

})

test_that("risib4-8 possible_gen work", {

    expect_equal(test_possible_gen("risib4", FALSE, FALSE, 1:4), 1:4)
    expect_equal(test_possible_gen("risib8", FALSE, FALSE, 1:8), 1:8)

    # X chr
    expect_equal(test_possible_gen("risib4", TRUE, FALSE, 1:4), 1:3)
    expect_equal(test_possible_gen("risib8", TRUE, FALSE, 1:8), c(1:3,5:6))

    # X chr, different cross order
    expect_equal(test_possible_gen("risib4", TRUE, FALSE, c(3,4,1,2)), c(3,4,1))
    expect_equal(test_possible_gen("risib8", TRUE, FALSE, c(4,5,1,3,2,7,8,6)), c(4,5,1,2,7))

    # one more time
    expect_equal(test_possible_gen("risib4", TRUE, FALSE, c(2,1,4,3)), c(2,1,4))
    expect_equal(test_possible_gen("risib8", TRUE, FALSE, c(1,5,6,8,4,2,7,3)), c(1,5,6,4,2))

})

# FIX_ME: test check_geno

test_that("risib4-8 init work", {

    expect_equal( sapply(1:4, function(i) test_init("risib4", i, FALSE, FALSE, 1:4)), log(rep(1/4, 4)) )
    expect_equal( sapply(1:8, function(i) test_init("risib8", i, FALSE, FALSE, 1:8)), log(rep(1/8, 8)) )

    # X chromosome
    expect_equal( sapply(1:3, function(i) test_init("risib4", i, TRUE, FALSE, 1:4)), log(rep(1/3, 3)) )
    expect_equal( sapply(c(1:3,5,6), function(i) test_init("risib8", i, TRUE, FALSE, 1:8)), log(c(1,1,2,1,1)/6) )

    # X chromosoem, different order
    expect_equal( sapply(c(3,4,1), function(i) test_init("risib4", i, TRUE, FALSE, c(3,4,1,2))), log(rep(1/3, 3)) )
    expect_equal( sapply(c(4,5,1,2,7), function(i) test_init("risib8", i, TRUE, FALSE, c(4,5,1,3,2,7,8,6))), log(c(1,1,2,1,1)/6) )

    # one more time
    expect_equal( sapply(c(1,2,4), function(i) test_init("risib4", i, TRUE, FALSE, c(2,1,4,3))), log(rep(1/3, 3)) )
    expect_equal( sapply(c(1,5,6,4,2), function(i) test_init("risib8", i, TRUE, FALSE, c(1,5,6,8,4,2,7,3))), log(c(1,1,2,1,1)/6) )

})

# FIX_ME: test emit

test_that("risib4 step works for autosome", {

    for(rf in c(0.01, 0.1, 0.45)) {
        expected <- matrix(rf*2/(1+6*rf), ncol=4, nrow=4)
        diag(expected) <- 1/(1+6*rf)

        result <- matrix(ncol=4, nrow=4)
        for(i in 1:4) {
            for(j in 1:4) {
                result[i,j] <- test_step("risib4", i, j, rf, FALSE, FALSE, 1:4)
            }
        }

        expect_equal(result, log(expected))
    }

    # test_stepmatrix: output has rows that sum to 1
    obs <- vapply(test_stepmatrix("risib4", c(0.01, 0.1, 0.45), FALSE, FALSE, 1:4),
                  function(a) rowSums(exp(a)), rep(1,4))
    expect_equal(obs, matrix(1, ncol=3, nrow=4))

    obs <- vapply(test_stepmatrix("risib4", c(0.01, 0.1, 0.45), FALSE, FALSE, c(3,1,2,4)),
                  function(a) rowSums(exp(a)), rep(1,4))
    expect_equal(obs, matrix(1, ncol=3, nrow=4))

    obs <- vapply(test_stepmatrix("risib4", c(0.01, 0.1, 0.45), TRUE, FALSE, c(3,1,2,4)),
                  function(a) rowSums(exp(a)), rep(1,3))
    expect_equal(obs, matrix(1, ncol=3, nrow=3))

})


test_that("risib8 step works for autosome", {

    for(rf in c(0.01, 0.1, 0.45)) {
        expected <- matrix(rf/(1+6*rf), ncol=8, nrow=8)
        diag(expected) <- (1-rf)/(1+6*rf)

        result <- matrix(ncol=8, nrow=8)
        for(i in 1:8) {
            for(j in 1:8) {
                result[i,j] <- test_step("risib8", i, j, rf, FALSE, FALSE, 1:8)
            }
        }

        expect_equal(result, log(expected))
    }

    # test_stepmatrix: output has rows that sum to 1
    obs <- vapply(test_stepmatrix("risib8", c(0.01, 0.1, 0.45), FALSE, FALSE, 1:8),
                  function(a) rowSums(exp(a)), rep(1,8))
    expect_equal(obs, matrix(1, ncol=3, nrow=8))

    obs <- vapply(test_stepmatrix("risib8", c(0.01, 0.1, 0.45), FALSE, FALSE, c(3,1,5:8,2,4)),
                  function(a) rowSums(exp(a)), rep(1,8))
    expect_equal(obs, matrix(1, ncol=3, nrow=8))

    obs <- vapply(test_stepmatrix("risib8", c(0.01, 0.1, 0.45), TRUE, FALSE, c(3,1,5:8,2,4)),
                  function(a) rowSums(exp(a)), rep(1,5))
    expect_equal(obs, matrix(1, ncol=3, nrow=5))

})



test_that("risib4-8 geno_names work", {

    expect_equal( geno_names("risib4", LETTERS[5:8], FALSE), paste0(LETTERS[5:8], LETTERS[5:8]) )
    expect_equal( geno_names("risib8", LETTERS[2:9], FALSE), paste0(LETTERS[2:9], LETTERS[2:9]) )

})

test_that("risib4-8 nrec work", {

    x <- matrix(ncol=8, nrow=8)
    x <- matrix(as.numeric(col(x) != row(x)), ncol=8)

    res4 <- matrix(ncol=4, nrow=4)
    res8 <- matrix(ncol=8, nrow=8)
    for(i in 1:8) {
        for(j in 1:8) {
            if(i<5 && j<5) res4[i,j] <- test_nrec("risib4", i, j, FALSE, FALSE, 1:4)
            if(i<9 && j<9) res8[i,j] <- test_nrec("risib8", i, j, FALSE, FALSE, 1:8)
        }
    }

    expect_equal( res4, x[1:4,1:4])
    expect_equal( res8, x)

})

context("chromosome scan by basic Haley-Knott functions")
library(qtl)

test_that("chromosome scan by Haley-Knott works", {

    # data for chr 6
    data(hyper)
    hyper <- hyper[6,]
    hyper2 <- convert2cross2(hyper)

    # scan by R/qtl
    hyper <- calc.genoprob(hyper, step=1)
    out <- scanone(hyper, method="hk")
    lod0 <- out[,3]

    # inputs for R/qtl2
    map <- insert_pseudomarkers(hyper2$gmap, step=1)
    pr <- calc_genoprob(hyper2, map)[[1]][,2,,drop=FALSE]
    y <- hyper2$pheno[,1]
    n <- length(y)

    # scan
    rss1 <- scan_hk_onechr(pr, as.matrix(y), as.matrix(rep(1, n)))
    lod1 <- n/2 * (log10(sum((y-mean(y))^2)) - log10(rss1))
    lod1 <- as.numeric(lod1)

    # as expected?
    expect_equal(lod0, lod1)

    ###
    # try it by centering first
    intercept <- cbind(rep(1, n))
    pr_r <- calc_resid_linreg_3d(intercept, pr)
    y_r <- calc_resid_linreg(intercept, as.matrix(y))

    # scan
    rss2 <- scan_hk_onechr_nocovar(pr_r, y_r)
    lod2 <- n/2 * (log10(sum(y_r^2)) - log10(rss2))
    lod2 <- as.numeric(lod2)

    # as expected?
    expect_equal(lod0, lod2)
    expect_equal(rss1, rss2)

    ###
    # direct calculation with lm()
    rss3 <- apply(pr, 3, function(a) sum(lm(y ~ a)$resid^2))
    names(rss3) <- NULL
    expect_equal(as.numeric(rss1), rss3)

    ##############################
    # weighted scan
    w <- runif(n, 1, 3)
    outw <- scanone(hyper, method="hk", weights=w)
    lodw0 <- outw[,3]

    rssw1 <- scan_hk_onechr_weighted(pr, as.matrix(y), as.matrix(rep(1, n)), sqrt(w))
    lodw1 <- n/2 * (log10(sum(lm(y ~ 1, weights=w)$resid^2*w)) - log10(rssw1))
    lodw1 <- as.numeric(lodw1)

    rssw2 <- scan_hk_onechr(pr*sqrt(w), as.matrix(y)*sqrt(w), as.matrix(rep(1, n))*sqrt(w))
    lodw2 <- n/2 * (log10(sum(lm(y ~ 1, weights=w)$resid^2*w)) - log10(rssw2))
    lodw2 <- as.numeric(lodw2)

    # as expected?
    expect_equal(lodw0, lodw1)
    expect_equal(lodw1, lodw2)

    ###
    # direct calculation with lm()
    rssw3 <- apply(pr, 3, function(a) sum(lm(y ~ a, weights=w)$resid^2*w))
    names(rssw3) <- NULL
    expect_equal(as.numeric(rssw1), rssw3)

})

test_that("chromosome scan by Haley-Knott with multiple phenotypes works", {

    skip_if(isnt_karl(), "this test only run locally")

    set.seed(20151201)
    # data for chr 6
    data(hyper)
    hyper <- hyper[6,]
    hyper2 <- convert2cross2(hyper)
    n_phe <- 200
    hyper$pheno <- cbind(permute_nvector(n_phe, hyper$pheno[,1]),
                         hyper$pheno[,2,drop=FALSE]) # sex column left at the end

    # scan by R/qtl
    hyper <- calc.genoprob(hyper, step=1)
    out <- scanone(hyper, method="hk", pheno.col=1:n_phe)
    lod0 <- t(out[,-(1:2)])
    dimnames(lod0) <- NULL

    # inputs for R/qtl2
    map <- insert_pseudomarkers(hyper2$gmap, step=1)
    pr <- calc_genoprob(hyper2, map)[[1]][,2,,drop=FALSE]
    y <- as.matrix(hyper$pheno[,1:n_phe])
    n <- nrow(y)

    # scan
    rss1 <- scan_hk_onechr(pr, y, as.matrix(rep(1, n)))
    lod1 <- n/2 * (log10(colSums((y-colMeans(y))^2)) - log10(rss1))

    # as expected?
    expect_equal(lod0, lod1)

    ###
    # try it by centering first
    intercept <- cbind(rep(1, n))
    pr_r <- calc_resid_linreg_3d(intercept, pr)
    y_r <- calc_resid_linreg(intercept, as.matrix(y))

    # scan
    rss2 <- scan_hk_onechr_nocovar(pr_r, y_r)
    lod2 <- n/2 * (log10(colSums(y_r^2)) - log10(rss2))

    # as expected?
    expect_equal(lod0, lod2)
    expect_equal(rss1, rss2)

    ###
    # direct calculation with lm()
    rss3 <- apply(pr, 3, function(a) colSums(lm(y ~ a)$resid^2))
    dimnames(rss3) <- NULL
    expect_equal(rss1, rss3)

    ##############################
    # weighted scan
    w <- runif(n, 1, 3)
    outw <- scanone(hyper, method="hk", weights=w, phe=1:n_phe)
    lodw0 <- t(outw[,-(1:2)])
    dimnames(lodw0) <- NULL

    rssw1 <- scan_hk_onechr_weighted(pr, y, as.matrix(rep(1, n)), sqrt(w))
    lodw1 <- n/2 * (log10(colSums(lm(y~1, weights=w)$resid^2*w)) - log10(rssw1))

    rssw2 <- scan_hk_onechr(pr*sqrt(w), y*sqrt(w), as.matrix(rep(1, n))*sqrt(w))
    lodw2 <- n/2 * (log10(colSums(lm(y~1, weights=w)$resid^2*w)) - log10(rssw2))

    # as expected?
    expect_equal(lodw0, lodw1)
    expect_equal(lodw1, lodw2)

    ###
    # direct calculation with lm()
    rssw3 <- apply(pr, 3, function(a) colSums(lm(y ~ a, weights=w)$resid^2*w))
    dimnames(rssw3) <- NULL
    expect_equal(rssw1, rssw3)

})


test_that("chromosome scan by Haley-Knott works with additive covariates", {

    skip_if(isnt_karl(), "this test only run locally")

    set.seed(20151201)

    # data for chr 6
    data(hyper)
    hyper <- hyper[6,]
    hyper2 <- convert2cross2(hyper)

    # scan by R/qtl
    hyper <- calc.genoprob(hyper, step=1)
    p <- hyper$pheno[,1]; p <- (p-min(p))/max(p)
    x <- rbinom(nind(hyper), 1, prob=p)
    out <- scanone(hyper, addcovar=x, method="hk")
    lod0 <- out[,3]

    # inputs for R/qtl2
    map <- insert_pseudomarkers(hyper2$gmap, step=1)
    pr <- calc_genoprob(hyper2, map)[[1]][,2,,drop=FALSE]
    y <- hyper2$pheno[,1]
    n <- length(y)

    # scan
    rss1 <- scan_hk_onechr(pr, as.matrix(y), cbind(1, x))
    lod1 <- n/2 * (log10(sum(lm(y~x)$resid^2)) - log10(rss1))
    lod1 <- as.numeric(lod1)

    # as expected?
    expect_equal(lod0, lod1)

    ###
    # direct calculation with lm()
    rss2 <- apply(pr, 3, function(a) sum(lm(y ~ x+a)$resid^2))
    names(rss2) <- NULL
    expect_equal(as.numeric(rss1), rss2)

    ##############################
    # weighted scan
    w <- runif(n, 1, 3)
    outw <- scanone(hyper, method="hk", weights=w, addcovar=x)
    lodw0 <- outw[,3]

    rssw1 <- scan_hk_onechr_weighted(pr, as.matrix(y), cbind(1,x), sqrt(w))
    lodw1 <- n/2 * (log10(sum(lm(y ~ x, weights=w)$resid^2*w)) - log10(rssw1))
    lodw1 <- as.numeric(lodw1)

    # as expected?
    expect_equal(lodw0, lodw1)

    ###
    # direct calculation with lm()
    rssw2 <- apply(pr, 3, function(a) sum(lm(y ~ x+a, weights=w)$resid^2*w))
    names(rssw2) <- NULL
    expect_equal(as.numeric(rssw1), rssw2)

})

test_that("chromosome scan by Haley-Knott with multiple phenotypes and an additive covariate works", {

    skip_if(isnt_karl(), "this test only run locally")

    set.seed(20151201)
    # data for chr 6
    data(hyper)
    hyper <- hyper[6,]
    hyper2 <- convert2cross2(hyper)
    n_phe <- 200
    hyper$pheno <- cbind(permute_nvector(n_phe, hyper$pheno[,1]),
                         hyper$pheno[,2,drop=FALSE]) # sex column left at the end

    # scan by R/qtl
    hyper <- calc.genoprob(hyper, step=1)
    p <- hyper$pheno[,1]; p <- (p-min(p))/max(p)
    x <- rbinom(nind(hyper), 1, prob=p)
    out <- scanone(hyper, method="hk", addcovar=x, pheno.col=1:n_phe)
    lod0 <- t(out[,-(1:2)])
    dimnames(lod0) <- NULL

    # inputs for R/qtl2
    map <- insert_pseudomarkers(hyper2$gmap, step=1)
    pr <- calc_genoprob(hyper2, map)[[1]][,2,,drop=FALSE]
    y <- as.matrix(hyper$pheno[,1:n_phe])
    n <- nrow(y)

    # scan
    rss1 <- scan_hk_onechr(pr, y, cbind(1,x))
    lod1 <- n/2 * (log10(colSums(lm(y~x)$resid^2)) - log10(rss1))

    # as expected?
    expect_equal(lod0, lod1)

    ###
    # direct calculation with lm()
    rss2 <- apply(pr, 3, function(a) colSums(lm(y ~ x+a)$resid^2))
    dimnames(rss2) <- NULL
    expect_equal(rss1, rss2)

    ##############################
    # weighted scan
    w <- runif(n, 1, 3)
    outw <- scanone(hyper, method="hk", weights=w, addcovar=x, phe=1:n_phe)
    lodw0 <- t(outw[,-(1:2)])
    dimnames(lodw0) <- NULL

    rssw1 <- scan_hk_onechr_weighted(pr, y, cbind(1,x), sqrt(w))
    lodw1 <- n/2 * (log10(colSums(lm(y~x, weights=w)$resid^2*w)) - log10(rssw1))

    # as expected?
    expect_equal(lodw0, lodw1)

    ###
    # direct calculation with lm()
    rssw2 <- apply(pr, 3, function(a) colSums(lm(y ~ x+a, weights=w)$resid^2*w))
    dimnames(rssw2) <- NULL
    expect_equal(rssw1, rssw2)

})


test_that("chromosome scan by Haley-Knott works with interactive covariates", {

    skip_if(isnt_karl(), "this test only run locally")

    set.seed(20151201)

    # data for chr 6
    data(hyper)
    hyper <- hyper[6,]
    hyper2 <- convert2cross2(hyper)

    # scan by R/qtl
    hyper <- calc.genoprob(hyper, step=1)
    p <- hyper$pheno[,1]; p <- (p-min(p))/max(p)
    x <- rbinom(nind(hyper), 1, prob=p)
    out <- scanone(hyper, addcovar=x, intcovar=x, method="hk")
    lod0 <- out[,3]

    # inputs for R/qtl2
    map <- insert_pseudomarkers(hyper2$gmap, step=1)
    pr <- calc_genoprob(hyper2, map)[[1]][,2,,drop=FALSE]
    y <- hyper2$pheno[,1]
    n <- length(y)

    # scan
    rss1 <- scan_hk_onechr_intcovar_highmem(pr, as.matrix(y), cbind(1, x), as.matrix(x))
    rss1_lm <- scan_hk_onechr_intcovar_lowmem(pr, as.matrix(y), cbind(1, x), as.matrix(x))
    expect_equal(rss1, rss1_lm)
    lod1 <- n/2 * (log10(sum(lm(y~x)$resid^2)) - log10(rss1))
    lod1 <- as.numeric(lod1)

    # as expected?
    expect_equal(lod0, lod1)

    ###
    # direct calculation with lm()
    rss2 <- apply(pr, 3, function(a) sum(lm(y ~ x*a)$resid^2))
    names(rss2) <- NULL
    expect_equal(as.numeric(rss1), rss2)

    ##############################
    # weighted scan
    w <- runif(n, 1, 3)
    outw <- scanone(hyper, method="hk", weights=w, addcovar=x, intcovar=x)
    lodw0 <- outw[,3]

    rssw1 <- scan_hk_onechr_intcovar_weighted_highmem(pr, as.matrix(y), cbind(1,x),
                                                      as.matrix(x), sqrt(w))
    rssw1_lm <- scan_hk_onechr_intcovar_weighted_lowmem(pr, as.matrix(y), cbind(1,x),
                                                        as.matrix(x), sqrt(w))
    expect_equal(rssw1, rssw1_lm)
    lodw1 <- n/2 * (log10(sum(lm(y ~ x, weights=w)$resid^2*w)) - log10(rssw1))
    lodw1 <- as.numeric(lodw1)

    # as expected?
    expect_equal(lodw0, lodw1)

    ###
    # direct calculation with lm()
    rssw2 <- apply(pr, 3, function(a) sum(lm(y ~ x*a, weights=w)$resid^2*w))
    names(rssw2) <- NULL
    expect_equal(as.numeric(rssw1), rssw2)

})

test_that("chromosome scan by Haley-Knott with multiple phenotypes and an interactive covariate works", {

    skip_if(isnt_karl(), "this test only run locally")

    set.seed(20151201)
    # data for chr 6
    data(hyper)
    hyper <- hyper[6,]
    hyper2 <- convert2cross2(hyper)
    n_phe <- 200
    hyper$pheno <- cbind(permute_nvector(n_phe, hyper$pheno[,1]),
                         hyper$pheno[,2,drop=FALSE]) # sex column left at the end

    # scan by R/qtl
    hyper <- calc.genoprob(hyper, step=1)
    p <- hyper$pheno[,1]; p <- (p-min(p))/max(p)
    x <- rbinom(nind(hyper), 1, prob=p)
    out <- scanone(hyper, method="hk", addcovar=x, intcovar=x, pheno.col=1:n_phe)
    lod0 <- t(out[,-(1:2)])
    dimnames(lod0) <- NULL

    # inputs for R/qtl2
    map <- insert_pseudomarkers(hyper2$gmap, step=1)
    pr <- calc_genoprob(hyper2, map)[[1]][,2,,drop=FALSE]
    y <- as.matrix(hyper$pheno[,1:n_phe])
    n <- nrow(y)

    # scan
    rss1 <- scan_hk_onechr_intcovar_highmem(pr, y, cbind(1,x), as.matrix(x))
    rss1_lm <- scan_hk_onechr_intcovar_lowmem(pr, y, cbind(1,x), as.matrix(x))
    expect_equal(rss1, rss1_lm)
    lod1 <- n/2 * (log10(colSums(lm(y~x)$resid^2)) - log10(rss1))

    # as expected?
    expect_equal(lod0, lod1)

    ###
    # direct calculation with lm()
    rss2 <- apply(pr, 3, function(a) colSums(lm(y ~ x*a)$resid^2))
    dimnames(rss2) <- NULL
    expect_equal(rss1, rss2)

    ##############################
    # weighted scan
    w <- runif(n, 1, 3)
    outw <- scanone(hyper, method="hk", weights=w, addcovar=x, intcovar=x, phe=1:n_phe)
    lodw0 <- t(outw[,-(1:2)])
    dimnames(lodw0) <- NULL

    rssw1 <- scan_hk_onechr_intcovar_weighted_highmem(pr, y, cbind(1,x), as.matrix(x), sqrt(w))
    rssw1_lm <- scan_hk_onechr_intcovar_weighted_lowmem(pr, y, cbind(1,x), as.matrix(x), sqrt(w))
    expect_equal(rssw1, rssw1_lm)
    lodw1 <- n/2 * (log10(colSums(lm(y~x, weights=w)$resid^2*w)) - log10(rssw1))

    # as expected?
    expect_equal(lodw0, lodw1)

    ###
    # direct calculation with lm()
    rssw2 <- apply(pr, 3, function(a) colSums(lm(y ~ x*a, weights=w)$resid^2*w))
    dimnames(rssw2) <- NULL
    expect_equal(rssw1, rssw2)

})

context("linear mixed models")

set.seed(20151124)
n <- 100
x <- matrix(rnorm(20*n), ncol=20)
k <- as.matrix(-dist(x))
k <- (k-min(k))/(max(k)-min(k))
dimnames(k) <- NULL

X <- cbind(1, sample(0:1, n, replace=TRUE))
d <- chol(0.25*k + diag(rep(0.75, n)))
y <- as.numeric(matrix(rnorm(n), nrow=1) %*% d) + X %*% c(1, 2)


test_that("eigen decomposition works", {

    e <- Rcpp_eigen_decomp(k)

    # same determinants?
    expect_equal(sum(log(e$values)), sum(log(eigen(k)$values)))

    # multiply back
    expect_equal(t(e$vectors) %*% diag(e$values) %*% e$vectors, k)

    # inverse
    expect_equal(t(e$vectors) %*% diag(1/e$values) %*% e$vectors, solve(k))

})

test_that("eigen + rotation works", {

    e <- Rcpp_eigen_rotation(k, y, X)

    # rotate back
    expect_equal(t(e$Kve_t) %*% e$y, y)
    expect_equal(t(e$Kve_t) %*% e$X, X)

    # same determinants?
    expect_equal(sum(log(e$Kva)), sum(log(eigen(k)$values)))

    # multiply back
    expect_equal(t(e$Kve_t) %*% diag(e$Kva) %*% e$Kve_t, k)

    # inverse
    expect_equal(t(e$Kve_t) %*% diag(1/e$Kva) %*% e$Kve_t, solve(k))

})

test_that("fitLMM works", {

    expected_reml <- list(loglik = -217.595158548486,
                          hsq = 0.0367120313458048,
                          sigmasq = 0.79659576062727,
                          beta = c(0.886161208390094, 1.74687862579568))

    expected_ml <- list(loglik = -217.180784548857,
                        hsq = 0,
                        sigmasq = 0.769853921083333,
                        beta = c(0.887111992802203, 1.74121566369042))

    e <- Rcpp_eigen_rotation(k, y, X)
    expect_equal(Rcpp_fitLMM(e$Kva, e$y, e$X), expected_reml)
    expect_equal(Rcpp_fitLMM(e$Kva, e$y, e$X, FALSE), expected_ml)

})

# test multi-parent population utilities
context("test MPP utilities")

g <- matrix(ncol=8, nrow=8)
cur <- 1
for(i in 1:8) {
    for(j in 1:i) {
        g[j,i] <- g[i,j] <- cur
        cur <- cur + 1
    }
}

test_that("mpp_encode_alleles works for 8 alleles, phase unknown", {
    homg <- c(1,3,6,10,15,21,28,36)
    for(i in seq(along=homg))
        expect_equal(mpp_encode_alleles(i, i, 8, FALSE), homg[i])

    for(i in 1:8) {
        for(j in 1:8) {
            expect_equal(mpp_encode_alleles(i, j, 8, FALSE), g[i,j])
        }
    }

    for(i in 1:8) {
        expect_equal(mpp_encode_alleles(NA, i, 8, FALSE), as.integer(NA))
        expect_equal(mpp_encode_alleles(i, NA, 8, FALSE), as.integer(NA))
        expect_equal(mpp_encode_alleles(0, i, 8, FALSE),  as.integer(NA))
        expect_equal(mpp_encode_alleles(9, i, 8, FALSE),  as.integer(NA))
        expect_equal(mpp_encode_alleles(i, 0, 8, FALSE),  as.integer(NA))
        expect_equal(mpp_encode_alleles(i, 9, 8, FALSE),  as.integer(NA))
    }

})

test_that("mpp_decode_geno works for 8 alleles, phase unknown", {
    homg <- c(1,3,6,10,15,21,28,36)
    for(i in seq(along=homg))
        expect_equal(mpp_decode_geno(homg[i], 8, FALSE), c(i,i))

    for(i in 1:8) {
        for(j in 1:8) {
            expect_equal(mpp_decode_geno(g[i, j], 8, FALSE), sort(c(i,j)))
        }
    }

    expect_equal(mpp_decode_geno(NA, 8, FALSE), rep(as.integer(NA), 2))
    expect_equal(mpp_decode_geno(0, 8, FALSE), rep(as.integer(NA), 2))
    expect_equal(mpp_decode_geno(37, 8, FALSE), rep(as.integer(NA), 2))

})

test_that("mpp_is_het works for 8 alleles, phase unknown", {
    homg <- c(1,3,6,10,15,21,28,36)
    for(i in seq(along=homg))
        expect_equal(mpp_is_het(homg[i], 8, FALSE), FALSE)

    for(i in 1:8) {
        for(j in 1:8) {
            expect_equal(mpp_is_het(g[i, j], 8, FALSE), ifelse(i==j, FALSE, TRUE))
        }
    }
})

test_that("mpp_geno_names works for 8 alleles", {
    a <- LETTERS[1:8]

    # genotype names
    gn <- paste0(a[row(g)], a[col(g)])[upper.tri(g, TRUE)]

    expect_equal(mpp_geno_names(a, FALSE), gn)
    expect_equal(mpp_geno_names(a, TRUE), c(gn, paste0(a, "Y")))
})


# phase-known case
cur <- 37
for(i in 2:8) {
    for(j in 1:(i-1)) {
        g[i,j] <- cur
        cur <- cur + 1
    }
}

test_that("mpp_encode_alleles works for 8 alleles, phase known", {
    homg <- c(1,3,6,10,15,21,28,36)
    for(i in seq(along=homg))
        expect_equal(mpp_encode_alleles(i, i, 8, TRUE), homg[i])

    for(i in 1:8) {
        for(j in 1:8) {
            expect_equal(mpp_encode_alleles(i, j, 8, TRUE), g[i,j])
        }
    }

    for(i in 1:8) {
        expect_equal(mpp_encode_alleles(NA, i, 8, FALSE), as.integer(NA))
        expect_equal(mpp_encode_alleles(i, NA, 8, FALSE), as.integer(NA))
        expect_equal(mpp_encode_alleles(0, i, 8, FALSE),  as.integer(NA))
        expect_equal(mpp_encode_alleles(9, i, 8, FALSE),  as.integer(NA))
        expect_equal(mpp_encode_alleles(i, 0, 8, FALSE),  as.integer(NA))
        expect_equal(mpp_encode_alleles(i, 9, 8, FALSE),  as.integer(NA))
    }

})

test_that("mpp_decode_geno works for 8 alleles, phase known", {
    homg <- c(1,3,6,10,15,21,28,36)
    for(i in seq(along=homg))
        expect_equal(mpp_decode_geno(homg[i], 8, TRUE), c(i,i))

    for(i in 1:8) {
        for(j in 1:8) {
            expect_equal(mpp_decode_geno(g[i, j], 8, TRUE), c(i,j))
        }
    }

    expect_equal(mpp_decode_geno(NA, 8, TRUE), rep(as.integer(NA), 2))
    expect_equal(mpp_decode_geno(0, 8, TRUE), rep(as.integer(NA), 2))
    expect_equal(mpp_decode_geno(65, 8, TRUE), rep(as.integer(NA), 2))

})

test_that("mpp_is_het works for 8 alleles, phase known", {
    homg <- c(1,3,6,10,15,21,28,36)
    for(i in seq(along=homg))
        expect_equal(mpp_is_het(homg[i], 8, TRUE), FALSE)

    for(i in 1:8) {
        for(j in 1:8) {
            expect_equal(mpp_is_het(g[i, j], 8, TRUE), ifelse(i==j, FALSE, TRUE))
        }
    }
})

test_that("invert_founder_index works", {

    expect_equal(invert_founder_index(c(2,3,1,4)), c(2,0,1,3))
    expect_equal(invert_founder_index(c(2,4,3,1)), c(3,0,2,1))
    expect_equal(invert_founder_index(c(7,8,3,5,4,1,6,2)), c(5,7,2,4,3,6,0,1))

})

context("basic HMM functions in general RIL")

test_that("genril n_gen, n_alleles work", {

    expect_equal(nalleles("genril38"), 38)
    expect_equal(nalleles("genril6"), 6)

    expect_equal(test_ngen("genril38", FALSE), 38)
    expect_equal(test_ngen("genril8", FALSE), 8)
    expect_equal(test_ngen("genril38", TRUE), 38)
    expect_equal(test_ngen("genril8", TRUE), 8)

    # throw error for "genril1"
    expect_error(nalleles("genril1"))

})

test_that("genril possible_gen work", {

    expect_equal(test_possible_gen("genril38", FALSE, FALSE, 1:38), 1:38)

})

# FIX_ME: test check_geno

test_that("genril init work", {

    set.seed(20181105)
    alpha <- sample(1:10, 38, replace=TRUE)
    init<- alpha/sum(alpha)
    expect_equal( sapply(1:38, function(i) test_init("genril38", i, FALSE, FALSE, c(8,alpha))), log(init))

    expect_equal( sapply(1:38, function(i) test_init("genril38", i, FALSE, FALSE, c(8,alpha))), log(init))

})

# FIX_ME: test emit

test_that("genril step works", {

    skip_on_cran()

    nf <- 38
    alpha_int <- sample(1:10, nf, replace=TRUE)
    alpha <- alpha_int/sum(alpha_int)

    for(rf in c(0.01, 0.1, 0.45)) {
        for(ngen in c(3, 5)) {
            expected <- matrix(rep(alpha*(1-(1-rf)^ngen),nf), ncol=nf, byrow=TRUE)
            diag(expected) <- alpha + (1-alpha)*(1-rf)^ngen

            result <- matrix(ncol=nf, nrow=nf)
            for(i in 1:nf) {
                for(j in 1:nf) {
                    result[i,j] <- test_step(paste0("genril",nf), i, j, rf, FALSE, FALSE, c(ngen, alpha_int))
                }
            }

            # rows sum to 1?
            expect_equal(rowSums(exp(result)), rep(1, nf))

            # match what I expected?
            expect_equal(result, log(expected))

        }
    }

})


test_that("genril geno_names work", {

    # if 38 founders, using upper case and lower case letters. Ugh.
    alleles <- c(LETTERS, letters[1:(38-26)])
    expect_equal( geno_names("genril38", alleles, FALSE), paste0(alleles, alleles) )

    # could also use two-letter allele codes, but ugly
    alleles <- c(paste0("A",LETTERS), paste0("B", LETTERS[1:(38-26)]))
    expect_equal( geno_names("genril38", alleles, FALSE), paste0(alleles, alleles) )

})

test_that("genril nrec work", {

    skip_on_cran()

    x <- matrix(ncol=38, nrow=38)
    x <- matrix(as.numeric(col(x) != row(x)), ncol=38)

    res38 <- matrix(ncol=38, nrow=38)
    for(i in 1:38) {
        for(j in 1:38) {
            res38[i,j] <- test_nrec("genril38", i, j, FALSE, FALSE, c(3, rep(1, 38)))
        }
    }

    expect_equal( res38, x )

})

context("basic HMM functions for magic19 (19-way RIL by selfing)")

test_that("magic19 n_gen, n_alleles work", {

    expect_equal(nalleles("magic19"), 19)

    expect_equal(test_ngen("magic19", FALSE), 19)

})

test_that("magic19 possible_gen work", {

    expect_equal(test_possible_gen("magic19", FALSE, FALSE, 1:19), 1:19)

})

# FIX_ME: test check_geno

test_that("magic19 init work", {

    expect_equal( sapply(1:19, function(i) test_init("magic19", i, FALSE, FALSE, integer(0))), log(rep(1/19, 19)) )

})

# FIX_ME: test emit

test_that("magic19 step works", {

    for(r in c(0.01, 0.1, 0.45)) {
        expected <- matrix(r*(90 - 54*r + 18*r^2)/(1+2*r)/19/18, ncol=19, nrow=19)
        diag(expected) <- (19 - 52*r + 54*r^2 - 18*r^3)/(1+2*r)/19

        result <- matrix(ncol=19, nrow=19)
        for(i in 1:19) {
            for(j in 1:19) {
                result[i,j] <- test_step("magic19", i, j, r, FALSE, FALSE, integer(0))
            }
        }

        expect_equal(result, log(expected))
    }

})



test_that("magic19 geno_names work", {

    expect_equal( geno_names("magic19", LETTERS[1:19], FALSE), paste0(LETTERS[1:19], LETTERS[1:19]) )

})

test_that("magic19 nrec work", {

    x <- matrix(ncol=19, nrow=19)
    x <- matrix(as.numeric(col(x) != row(x)), ncol=19)

    res19 <- matrix(ncol=19, nrow=19)
    for(i in 1:19) {
        for(j in 1:19) {
            res19[i,j] <- test_nrec("magic19", i, j, FALSE, FALSE, integer(0))
        }
    }

    expect_equal( res19, x)

})

context("basic HMM functions in Diversity Outcross F1 (crossed to a 9th inbred strain)")

test_that("DOF1 nalleles works", {
    expect_equal(nalleles("dof1"), 8)
})

test_that("DOF1 check_geno works", {

    # observed genotypes
    for(i in 0:5) {
        # Autosome
        expect_true(test_check_geno("dof1", i, TRUE, FALSE, FALSE, 8))
        # Female X
        expect_true(test_check_geno("dof1", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_true(test_check_geno("dof1", i, TRUE, TRUE, FALSE, 20))
    }
    for(i in c(-1, 6)) {
        # Autosome
        expect_false(test_check_geno("dof1", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("dof1", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_false(test_check_geno("dof1", i, TRUE, TRUE, FALSE, 20))
    }

    # true genotypes, autosome and female X
    for(i in 1:8) {
        # Autosome
        expect_true(test_check_geno("dof1", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("dof1", i, FALSE, TRUE, TRUE, 20))
    }
    for(i in c(0, 9)) {
        # Autosome
        expect_false(test_check_geno("dof1", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("dof1", i, FALSE, TRUE, TRUE, 20))
    }

    # true genotypes, male X
    for(i in c(0, 9:18)) {
        expect_false(test_check_geno("dof1", i, FALSE, TRUE, FALSE, 20))
    }
    for(i in c(1:8)) {
        expect_true(test_check_geno("dof1", i, FALSE, TRUE, FALSE, 20))
    }

})

test_that("DOF1 n_gen works", {

    expect_equal(test_ngen("dof1", FALSE), 8)
    expect_equal(test_ngen("dof1", TRUE),  8)

})

test_that("DOF1 possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("dof1", FALSE, FALSE, 20), 1:8)

    # X female
    expect_equal(test_possible_gen("dof1", TRUE, TRUE, 20), 1:8)

    # X male
    expect_equal(test_possible_gen("dof1", TRUE, FALSE, 20), 1:8)

})

test_that("DOF1 init works", {

    hom <- 1:8
    male <- 8 + (1:8)

    for(i in hom) {
        # autosome and female X
        expect_equal(test_init("dof1", i, FALSE, FALSE, 20), log(1/8))
        expect_equal(test_init("dof1", i, TRUE,  TRUE,  20), log(1/8))
    }

    for(i in male)
        expect_equal(test_init("dof1", i, TRUE,  FALSE, 20), log(1/8))

})


test_that("DOF1 emit works", {

    fgen <- c(1,3,0,1,3,0,1,3,1) # founder genotypes; 0=missing, 1=AA, 3=BB
    err <- 0.01

    expected <- log(c(1-err, err/2, err/2, 1-err/2, err))
    for(trueg in c(1,4,7)) { # A, D, G
        for(obsg in 1:5) {
            expect_equal(test_emit("dof1", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("dof1", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
            expect_equal(test_emit("dof1", obsg, trueg, err, fgen, TRUE, FALSE, 20), expected[obsg])
        }
    }
    expected <- log(c(err/2, 1-err, err/2, 1-err/2, 1-err/2))
    expected_male <- log(c(err/2, err/2, 1-err, err, 1-err/2))
    for(trueg in c(2,5,8)) { # B, E, H
        for(obsg in 1:5) {
            expect_equal(test_emit("dof1", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("dof1", obsg, trueg, err, fgen, TRUE,  TRUE, 20),  expected[obsg])
            expect_equal(test_emit("dof1", obsg, trueg, err, fgen, TRUE, FALSE, 20),  expected_male[obsg])
        }
    }
    expected <- log(c(1-err, 1, err, 1-err, err))
    expected_male <- log(c(1, 1, 1, 1, 1))
    for(trueg in c(3,6)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("dof1", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("dof1", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
            expect_equal(test_emit("dof1", obsg, trueg, err, fgen, TRUE, FALSE, 20), expected_male[obsg])
        }
    }

})

test_that("DOF1 step works", {

    ng <- 8
    trmat <- matrix(nrow=ng, ncol=ng)
    # autosome
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("dof1", gl, gr, rf, FALSE, FALSE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))

            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

    # female X
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("dof1", gl, gr, rf, TRUE, TRUE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))

            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

    # male X
    ng <- 8
    trmat <- matrix(nrow=ng, ncol=ng)
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("dof1", 8+gl, 8+gr, rf, TRUE, FALSE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))

            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

})

test_that("geno_names works", {
    auto <- LETTERS[1:8]

    expect_equal(geno_names("dof1", LETTERS[1:8], FALSE), auto)
    expect_equal(geno_names("dof1", LETTERS[1:8], TRUE), auto)
})

context("addlog and subtractlog")

test_that("addlog works", {

  a <- 50
  b <- 60
  d <- 2
  expect_equal(addlog(a,b), log(exp(a)+exp(b)))
  expect_equal(addlog(b,a), log(exp(a)+exp(b)))
  expect_equal(addlog(a,d), log(exp(a)+exp(d)))
  expect_equal(addlog(d,a), log(exp(a)+exp(d)))
  expect_equal(addlog(b,d), log(exp(b)+exp(d)))
  expect_equal(addlog(d,b), log(exp(b)+exp(d)))
  expect_equal(addlog(a,a+300), a+300);
  expect_equal(addlog(a,a-300), a);
  expect_equal(addlog(a+300,a), a+300);
  expect_equal(addlog(a-300,a), a);

})


test_that("subtractlog works", {

  a <- 60
  b <- 50
  d <- 2
  e <- 5
  expect_equal(subtractlog(a,b), log(exp(a)-exp(b)))
  expect_equal(subtractlog(a,d), log(exp(a)-exp(d)))
  expect_equal(subtractlog(b,d), log(exp(b)-exp(d)))
  expect_equal(subtractlog(e,d), log(exp(e)-exp(d)))
  expect_equal(subtractlog(e,d), log(exp(e)-exp(d)))
  expect_equal(subtractlog(a+300,a), 360);
  expect_equal(subtractlog(a,a-300), a);

  expect_equal(subtractlog(0.0, -a), log(1.0-exp(-a)));
  expect_equal(subtractlog(0.0, -b), log(1.0-exp(-b)));
  expect_equal(subtractlog(0.0, -d), log(1.0-exp(-d)));

})

context("logistic regression");

test_that("binreg_eigen functions work", {

    set.seed(94651642)

    n <- 200
    X <- cbind(1, sample(0:1, n, replace=TRUE), rnorm(n))
    nu <- X %*% c(0, 0.5, 0.3)
    p <- exp(nu)/(1+exp(nu))
    y <- rbinom(n, 1, p)

    out <- glm(y ~ -1 + X, family=binomial(link=logit))
    pi <- out$fitted
    expected <- sum(y*log10(pi) + (1-y)*log10(1-pi))

    # just the log likelihood
    expect_equal(calc_ll_binreg_eigenchol(X, y), expected)
    expect_equal(calc_ll_binreg_eigenqr(X, y), expected)
    expect_equal(calc_ll_binreg(X, y), expected)
    out_fit_binreg <- fit_binreg(X, y)
    expect_equal(out_fit_binreg$log10lik, expected);
    expect_equal(expected, out$deviance/(-2*log(10)))


    # coefficients
    coef <- setNames(out$coef, NULL)
    expect_equal(calc_coef_binreg_eigenqr(X, y), coef)
    expect_equal(calc_coef_binreg(X, y), coef)
    out_coefSE_binreg <- calc_coefSE_binreg(X, y)
    expect_equal(out_coefSE_binreg$coef, coef)
    expect_equal(out_fit_binreg$coef, coef)

    # SEs
    SE <- setNames(summary(out)$coef[,2], NULL)
    expect_equal(out_coefSE_binreg$SE, SE, tol=1e-6)
    expect_equal(out_fit_binreg$SE, SE, tol=1e-6)

    # reduced rank matrix
    XX <- cbind(X, X[,1]+X[,2])

    out <- glm(y ~ -1 + XX, family=binomial(link=logit))
    pi <- out$fitted
    expected2 <- sum(y*log10(pi) + (1-y)*log10(1-pi))

    # just the log likelihood
    expect_equal(calc_ll_binreg_eigenchol(XX, y), expected2)
    expect_equal(calc_ll_binreg_eigenqr(XX, y), expected2)
    expect_equal(calc_ll_binreg(XX, y), expected2)
    out_fit_binreg <- fit_binreg(XX, y)
    expect_equal(out_fit_binreg$log10lik, expected2);
    expect_equal(expected, out$deviance/(-2*log(10)))

})


test_that("weighted binreg_eigen functions work", {

    set.seed(94651642)

    n <- 200
    X <- cbind(1, sample(0:1, n, replace=TRUE), rnorm(n))
    nu <- X %*% c(0, 0.5, 0.3)
    p <- exp(nu)/(1+exp(nu))
    wt <- sample(1:10, n, replace=TRUE)
    y <- rbinom(n, wt, p)/wt

    out <- glm(y ~ -1 + X, family=binomial(link=logit), weights=wt)
    pi <- out$fitted
    expected <- sum((y*log10(pi) + (1-y)*log10(1-pi))*wt)

    # just the log likelihood
    expect_equal(calc_ll_binreg_weighted_eigenchol(X, y, wt), expected)
    expect_equal(calc_ll_binreg_weighted_eigenqr(X, y, wt), expected)
    expect_equal(calc_ll_binreg_weighted(X, y, wt), expected)
    out_fit_binreg_weighted <- fit_binreg_weighted(X, y, wt)
    expect_equal(out_fit_binreg_weighted$log10lik, expected);

    # not sure about constant factor in deviance, so calc LOD vs model with just intercept
    out0 <- glm(y ~ 1, family=binomial(link=logit), weights=wt)
    expect_equal(calc_ll_binreg_weighted_eigenqr(X, y, wt) -
                 calc_ll_binreg_weighted_eigenqr(cbind(rep(1, n)), y, wt),
                 (out$deviance - out0$deviance)/(-2*log(10)))

    # coefficients
    coef <- setNames(out$coef, NULL)
    expect_equal(calc_coef_binreg_weighted_eigenqr(X, y, wt), coef)
    expect_equal(calc_coef_binreg_weighted(X, y, wt), coef)
    out_coefSE_binreg_weighted <- calc_coefSE_binreg_weighted(X, y, wt)
    expect_equal(out_coefSE_binreg_weighted$coef, coef)
    expect_equal(out_fit_binreg_weighted$coef, coef)

    # SEs
    SE <- setNames(summary(out)$coef[,2], NULL)
    expect_equal(out_coefSE_binreg_weighted$SE, SE, tol=1e-6)
    expect_equal(out_fit_binreg_weighted$SE, SE, tol=1e-6)

    # reduced rank matrix
    XX <- cbind(X, X[,1]+X[,2])

    out <- glm(y ~ -1 + XX, family=binomial(link=logit), weights=wt)
    pi <- out$fitted
    expected2 <- sum((y*log10(pi) + (1-y)*log10(1-pi))*wt)

    # just the log likelihood
    expect_equal(calc_ll_binreg_weighted_eigenchol(XX, y, wt), expected2)
    expect_equal(calc_ll_binreg_weighted_eigenqr(XX, y, wt), expected2)
    expect_equal(calc_ll_binreg_weighted(XX, y, wt), expected2)
    out_fit_binreg_weighted <- fit_binreg_weighted(XX, y, wt)
    expect_equal(out_fit_binreg_weighted$log10lik, expected2);

    # not sure about constant factor in deviance, so calc LOD vs model with just intercept
    out0 <- glm(y ~ 1, family=binomial(link=logit), weights=wt)
    expect_equal(calc_ll_binreg_weighted_eigenqr(XX, y, wt) -
                 calc_ll_binreg_weighted_eigenqr(cbind(rep(1, n)), y, wt),
                 (out$deviance - out0$deviance)/(-2*log(10)))

})



test_that("binreg full variance-covariance matrix is correct", {

    set.seed(20260608)
    n <- 100
    x <- cbind(1, rnorm(n))
    y <- x %*% c(1,1) + rnorm(n)
    z <- as.numeric(y > 0.2)
    w <- sample(1:6, n, replace=TRUE) # weights, for later

    out_glm <- summary(glm(z ~ -1 + x, family=binomial(link=logit)))
    coef_glm <- out_glm$coef[,1]
    se_glm <- out_glm$coef[,2]
    var_glm <- out_glm$cov.scaled

    expect_equal(sqrt(diag(var_glm)), se_glm)

    # eigen QR
    out_var <- fit_binreg(x, z, FALSE, TRUE)
    out_se <- fit_binreg(x, z, TRUE, FALSE)

    expect_equivalent(coef_glm, out_se$coef)
    expect_equivalent(se_glm, out_se$SE, tol=1e-5)
    expect_equivalent(coef_glm, out_var$coef)
    expect_equivalent(se_glm, out_var$SE, tol=1e-5)
    expect_equivalent(var_glm, out_var$var, tol=1e-5)

    # weighted
    out_glm <- summary(glm(z ~ -1 + x, family=binomial(link=logit), weights=w))
    coef_glm <- out_glm$coef[,1]
    se_glm <- out_glm$coef[,2]
    var_glm <- out_glm$cov.scaled

    expect_equal(sqrt(diag(var_glm)), se_glm)

    # eigen QR
    out_var <- fit_binreg_weighted(x, z, w, TRUE, TRUE)
    out_se <- fit_binreg_weighted(x, z, w, TRUE, FALSE)

    expect_equivalent(coef_glm, out_se$coef)
    expect_equivalent(se_glm, out_se$SE, tol=1e-5)
    expect_equivalent(coef_glm, out_var$coef)
    expect_equivalent(se_glm, out_var$SE, tol=1e-5)
    expect_equivalent(var_glm, out_var$var, tol=1e-5)

})

context("basic HMM functions in 6-way doubled haploids")

test_that("dh6 n_gen, n_alleles work", {

    expect_equal(nalleles("dh6"), 6)

    expect_equal(test_ngen("dh6", FALSE), 6)

})

test_that("dh6 possible_gen work", {

    expect_equal(test_possible_gen("dh6", FALSE, FALSE, 1:6), 1:6)

})

# FIX_ME: test check_geno

test_that("dh6 init work", {

    expect_equal( sapply(1:6, function(i) test_init("dh6", i, FALSE, FALSE, 1:6)), log(rep(1/6, 6)) )

})

# FIX_ME: test emit

test_that("dh6 step works", {

    for(rf in c(0.01, 0.1, 0.45)) {
        for(ngen in c(3, 5)) {
            expected <- matrix((5 - (5-6*rf)*(1-rf)^(ngen-2))/30, ncol=6, nrow=6)
            diag(expected) <- (1 + (5-6*rf)*(1-rf)^(ngen-2))/6

            result <- matrix(ncol=6, nrow=6)
            for(i in 1:6) {
                for(j in 1:6) {
                    result[i,j] <- test_step("dh6", i, j, rf, FALSE, FALSE, ngen)
                }
            }

            expect_equal(result, log(expected))
        }
    }

})


test_that("dh6 geno_names work", {

    expect_equal( geno_names("dh6", LETTERS[5:10], FALSE), paste0(LETTERS[5:10], LETTERS[5:10]) )

})

test_that("dh6 nrec work", {

    x <- matrix(ncol=6, nrow=6)
    x <- matrix(as.numeric(col(x) != row(x)), ncol=6)

    res6 <- matrix(ncol=6, nrow=6)
    for(i in 1:6) {
        for(j in 1:6) {
            res6[i,j] <- test_nrec("dh6", i, j, FALSE, FALSE, 3)
        }
    }

    expect_equal( res6, x )

})

context("basic HMM functions in Diversity Outcross")

test_that("DO nalleles works", {
    expect_equal(nalleles("do"), 8)
})

test_that("DO check_geno works", {

    # observed genotypes
    for(i in 0:5) {
        # Autosome
        expect_true(test_check_geno("do", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("do", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_true(test_check_geno("do", i, TRUE, TRUE, FALSE, 20))
    }
    for(i in c(-1, 6)) {
        # Autosome
        expect_false(test_check_geno("do", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("do", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_false(test_check_geno("do", i, TRUE, TRUE, FALSE, 20))
    }

    # true genotypes, autosome and female X
    for(i in 1:36) {
        # Autosome
        expect_true(test_check_geno("do", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("do", i, FALSE, TRUE, TRUE, 20))
    }
    for(i in c(0, 37)) {
        # Autosome
        expect_false(test_check_geno("do", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("do", i, FALSE, TRUE, TRUE, 20))
    }

    # true genotypes, autosome and female X
    for(i in 36 + 1:8) {
        expect_true(test_check_geno("do", i, FALSE, TRUE, FALSE, 20))
    }
    for(i in c(0:36, 36+9)) {
        expect_false(test_check_geno("do", i, FALSE, TRUE, FALSE, 20))
    }

})

test_that("DO n_gen works", {

    expect_equal(test_ngen("do", FALSE), 36)
    expect_equal(test_ngen("do", TRUE),  36+8)

})

test_that("DO possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("do", FALSE, FALSE, 20), 1:36)

    # X female
    expect_equal(test_possible_gen("do", TRUE, TRUE, 20), 1:36)

    # X male
    expect_equal(test_possible_gen("do", TRUE, FALSE, 20), 36+(1:8))

})

test_that("DO init works", {

    hom <- cumsum(1:8)
    het <- (1:36)[!((1:36) %in% cumsum(1:8))]
    male <- 36 + (1:8)

    for(i in hom) {
        # autosome and female X
        expect_equal(test_init("do", i, FALSE, FALSE, 20), log(1/64))
        expect_equal(test_init("do", i, TRUE,  TRUE,  20), log(1/64))
    }

    for(i in het) {
        # autosome and female X
        expect_equal(test_init("do", i, FALSE, FALSE, 20), log(1/32))
        expect_equal(test_init("do", i, TRUE,  TRUE,  20), log(1/32))
    }

    for(i in male)
        expect_equal(test_init("do", i, TRUE,  FALSE, 20), log(1/8))

})


test_that("DO emit works", {

    fgen <- c(1,3,0,1,3,0,1,3) # founder genotypes; 0=missing, 1=AA, 3=BB
    err <- 0.01

    # Autosome or female X
    # truth = homA: AA (1), AD (7), AG (22), DD (10), DG (25), GG (28)
    expected <- log(c(1-err, err/2, err/2, 1-err/2, err))
    for(trueg in c(1,7,22,10,25,28)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("do", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("do", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = het: AB (2), AE (11), AH (29), BD (8), BG (23), DE (14), DH (32), EG (26), GH (35)
    expected <- log(c(err/2, 1-err, err/2, 1-err/2, 1-err/2))
    for(trueg in c(2,11,29,8,23,14,32,26,35)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("do", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("do", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = homB: BB (3), BE (12), BH (30), EE (15), EH (33), HH (36)
    expected <- log(c(err/2, err/2, 1-err, err, 1-err/2))
    for(trueg in c(3,12,30,15,33,36)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("do", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("do", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = A-: AC (4), AF (16), CD (9), DF (19), CG (24), FG (27)
    expected <- log(c(1-err,1,err,1-err,err))
    for(trueg in c(4,16,9,19,24,27)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("do", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("do", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = B-: BC (5), BF (17), CE (13), EF (20), CH (31), FH (34)
    expected <- log(c(err,1,1-err,err,1-err))
    for(trueg in c(5,17,13,20,31,34)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("do", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("do", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }

    # male X: treat het as missing
    # truth = hemA: A (1+36), D (4+36), G (7+36)
    expected <- log(c(1-err, 1, err, 1-err, err))
    for(trueg in 36+c(1,4,7))
        for(obsg in 1:5)
            expect_equal(test_emit("do", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])
    # truth = hemB: BB (2+36), E (5+36), H (8+36)
    expected <- log(c(err, 1, 1-err, err, 1-err))
    for(trueg in 36+c(2,5,8))
        for(obsg in 1:5)
            expect_equal(test_emit("do", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])
    # truth = missing: C (3+36), F (6+36)
    expected <- rep(0,5)
    for(trueg in 36+c(3,6))
        for(obsg in 1:5)
            expect_equal(test_emit("do", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])


})

test_that("DO step works", {

    skip_on_cran()

    ng <- 36
    trmat <- matrix(nrow=ng, ncol=ng)
    # autosome
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("do", gl, gr, rf, FALSE, FALSE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))

            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

    # female X
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("do", gl, gr, rf, TRUE, TRUE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))

            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

    # male X
    ng <- 8
    trmat <- matrix(nrow=ng, ncol=ng)
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("do", 36+gl, 36+gr, rf, TRUE, FALSE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))

            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

})

test_that("geno_names works", {
    auto <- c("AA", "AB", "BB", "AC", "BC", "CC", "AD", "BD", "CD", "DD",
              "AE", "BE", "CE", "DE", "EE", "AF", "BF", "CF", "DF", "EF", "FF",
              "AG", "BG", "CG", "DG", "EG", "FG", "GG", "AH", "BH", "CH", "DH",
              "EH", "FH", "GH", "HH")
    X <- c(auto, paste0(LETTERS[1:8], "Y"))

    expect_equal(geno_names("do", LETTERS[1:8], FALSE), auto)
    expect_equal(geno_names("do", LETTERS[1:8], TRUE), X)
})

test_that("nrec works", {

    skip_on_cran()

    # X chr male
    for(i in 36+(1:8)) {
        for(j in 36+(1:8)) {
            expect_equal(test_nrec("do", i, j, TRUE, FALSE, 0), as.numeric(i!=j))
            # hs should be the same, too
            expect_equal(test_nrec("hs", i, j, TRUE, FALSE, 0), as.numeric(i!=j))
        }
    }

    # autosome or X chr female
    g <- sapply(1:36, mpp_decode_geno, 8, FALSE)
    expected <- resultA <- resultX <- matrix(ncol=ncol(g), nrow=ncol(g))
    resultAhs <- resultXhs <- resultA
    for(i in 1:ncol(g)) {
        for(j in 1:ncol(g)) {
            if((g[1,i] == g[1,j] && g[2,i]==g[2,j]) ||
               (g[1,i] == g[2,j] && g[2,i]==g[1,j])) expected[i,j] <- 0
            else if(g[1,i] != g[1,j] && g[2,i]!=g[2,j] &&
                    g[1,i] != g[2,j] && g[2,i]!=g[1,j]) expected[i,j] <- 2
            else expected[i,j] <- 1

            resultA[i,j] <- test_nrec("do", i, j, FALSE, FALSE, 0)
            resultX[i,j] <- test_nrec("do", i, j, TRUE, TRUE, 0)
            resultAhs[i,j] <- test_nrec("hs", i, j, FALSE, FALSE, 0)
            resultXhs[i,j] <- test_nrec("hs", i, j, TRUE, TRUE, 0)
        }
    }
    expect_equal(resultA, expected)
    expect_equal(resultX, expected)
    expect_equal(resultAhs, expected)
    expect_equal(resultXhs, expected)
})

context("Arrange genes vertically so they don't overlap")

test_that("arrange_genes gives appropriate errors", {

#    expect_error(arrange_genes(NULL, NULL))
    expect_error(arrange_genes(numeric(0), numeric(0)))
    expect_error(arrange_genes(c(1,2), 4))
    expect_error(arrange_genes(c(1,2), c(4,5,6)))

})

test_that("arrange_genes works in simple cases", {

    # no overlaps
    expect_equal(arrange_genes(1,2), 1)
    expect_equal(arrange_genes(c(1,5), c(2,6)), c(1,1))
    expect_equal(arrange_genes(c(1,5,10), c(2,6,11)), c(1,1,1))

    # some overlap
    expect_equal(arrange_genes(c(1,5), c(6,10)), c(1,2))
    expect_equal(arrange_genes(c(1,5,11), c(6,10,12)), c(1,2,1))
    expect_equal(arrange_genes(c(1,5,9), c(6,10,12)), c(1,2,1))
    expect_equal(arrange_genes(c(1,5,5.5), c(6,10,12)), c(1,2,3))

})


test_that("the example is arranged properly", {
    # this was a bug in my initial implementation
    # was giving 1, 1, 1, 1, 2, 1, 1, 2
    expect_equal( arrange_genes(c(139.99, 140.68, 141.71, 142.23, 142.59, 143.23, 144.40, 144.99),
                                c(140.77, 141.89, 142.74, 143.25, 143.48, 144.53, 145.46, 145.82)),

                  c(1, 2, 1, 2, 3, 1, 2, 1) )
})

context("basic HMM functions in backcross")

test_that("backcross nalleles works", {
    expect_equal(nalleles("bc"), 2)
})

test_that("backcross check_geno works", {

    # autosome
    expect_true(test_check_geno("bc", 0, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("bc", 1, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("bc", 2, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("bc", 1, FALSE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("bc", 2, FALSE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("bc", 3, TRUE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("bc", 0, FALSE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("bc", 3, FALSE, FALSE, FALSE, integer(0)))

    # X chromosome female
    expect_true(test_check_geno("bc", 0, TRUE, TRUE, TRUE, integer(0)))
    expect_true(test_check_geno("bc", 1, TRUE, TRUE, TRUE, integer(0)))
    expect_true(test_check_geno("bc", 2, TRUE, TRUE, TRUE, integer(0)))
    expect_true(test_check_geno("bc", 1, FALSE, TRUE, TRUE, integer(0)))
    expect_true(test_check_geno("bc", 2, FALSE, TRUE, TRUE, integer(0)))
    expect_false(test_check_geno("bc", 3, TRUE, TRUE, TRUE, integer(0)))
    expect_false(test_check_geno("bc", 0, FALSE, TRUE, TRUE, integer(0)))
    expect_false(test_check_geno("bc", 3, FALSE, TRUE, TRUE, integer(0)))

    # X chromosome male
    expect_true(test_check_geno("bc", 0, TRUE, TRUE, FALSE, integer(0)))
    expect_true(test_check_geno("bc", 1, TRUE, TRUE, FALSE, integer(0)))
    expect_true(test_check_geno("bc", 3, TRUE, TRUE, FALSE, integer(0)))
    expect_true(test_check_geno("bc", 3, FALSE, TRUE, FALSE, integer(0)))
    expect_true(test_check_geno("bc", 4, FALSE, TRUE, FALSE, integer(0)))
    expect_false(test_check_geno("bc", 2, TRUE, TRUE, FALSE, integer(0)))
    expect_false(test_check_geno("bc", 4, TRUE, TRUE, FALSE, integer(0)))
    expect_false(test_check_geno("bc", 0, FALSE, TRUE, FALSE, integer(0)))
    expect_false(test_check_geno("bc", 1, FALSE, TRUE, FALSE, integer(0)))
    expect_false(test_check_geno("bc", 2, FALSE, TRUE, FALSE, integer(0)))
    expect_false(test_check_geno("bc", 5, FALSE, TRUE, FALSE, integer(0)))

})

test_that("backcross n_gen works", {

    expect_equal(test_ngen("bc", FALSE), 2)
    expect_equal(test_ngen("bc", TRUE), 4)

})

test_that("backcross possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("bc", FALSE, FALSE, integer(0)), 1:2)
    # X female
    expect_equal(test_possible_gen("bc", TRUE, TRUE, integer(0)), 1:2)
    # X male
    expect_equal(test_possible_gen("bc", TRUE, FALSE, integer(0)), 3:4)

})

test_that("backcross nrec works", {

    # autosome
    expect_equal(test_nrec("bc", 1, 1, FALSE, FALSE, integer(0)), 0)
    expect_equal(test_nrec("bc", 1, 2, FALSE, FALSE, integer(0)), 1)
    expect_equal(test_nrec("bc", 2, 1, FALSE, FALSE, integer(0)), 1)
    expect_equal(test_nrec("bc", 2, 2, FALSE, FALSE, integer(0)), 0)
    # X female
    expect_equal(test_nrec("bc", 1, 1, TRUE, TRUE, integer(0)), 0)
    expect_equal(test_nrec("bc", 1, 2, TRUE, TRUE, integer(0)), 1)
    expect_equal(test_nrec("bc", 2, 1, TRUE, TRUE, integer(0)), 1)
    expect_equal(test_nrec("bc", 2, 2, TRUE, TRUE, integer(0)), 0)
    # X male
    expect_equal(test_nrec("bc", 3, 3, TRUE, FALSE, integer(0)), 0)
    expect_equal(test_nrec("bc", 3, 4, TRUE, FALSE, integer(0)), 1)
    expect_equal(test_nrec("bc", 4, 3, TRUE, FALSE, integer(0)), 1)
    expect_equal(test_nrec("bc", 4, 4, TRUE, FALSE, integer(0)), 0)

})

test_that("backcross init works", {

    # autosome
    expect_equal(test_init("bc", 1, FALSE, FALSE, integer(0)), log(0.5))
    expect_equal(test_init("bc", 2, FALSE, FALSE, integer(0)), log(0.5))
    # X female
    expect_equal(test_init("bc", 1, TRUE, TRUE, integer(0)), log(0.5))
    expect_equal(test_init("bc", 2, TRUE, TRUE, integer(0)), log(0.5))
    # X male
    expect_equal(test_init("bc", 3, TRUE, FALSE, integer(0)), log(0.5))
    expect_equal(test_init("bc", 4, TRUE, FALSE, integer(0)), log(0.5))

})

test_that("backcross emit works", {

    # autosome
    eps <- 0.01
    expect_equal(test_emit("bc", 0, 1, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("bc", 0, 2, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("bc", 1, 1, eps, integer(0), FALSE, FALSE, integer(0)), log(1-eps))
    expect_equal(test_emit("bc", 1, 2, eps, integer(0), FALSE, FALSE, integer(0)), log(eps))
    expect_equal(test_emit("bc", 2, 1, eps, integer(0), FALSE, FALSE, integer(0)), log(eps))
    expect_equal(test_emit("bc", 2, 2, eps, integer(0), FALSE, FALSE, integer(0)), log(1-eps))
    # X female
    expect_equal(test_emit("bc", 0, 1, eps, integer(0), TRUE, TRUE, integer(0)), 0)
    expect_equal(test_emit("bc", 0, 2, eps, integer(0), TRUE, TRUE, integer(0)), 0)
    expect_equal(test_emit("bc", 1, 1, eps, integer(0), TRUE, TRUE, integer(0)), log(1-eps))
    expect_equal(test_emit("bc", 1, 2, eps, integer(0), TRUE, TRUE, integer(0)), log(eps))
    expect_equal(test_emit("bc", 2, 1, eps, integer(0), TRUE, TRUE, integer(0)), log(eps))
    expect_equal(test_emit("bc", 2, 2, eps, integer(0), TRUE, TRUE, integer(0)), log(1-eps))
    # X male
    expect_equal(test_emit("bc", 0, 3, eps, integer(0), TRUE, FALSE, integer(0)), 0)
    expect_equal(test_emit("bc", 0, 4, eps, integer(0), TRUE, FALSE, integer(0)), 0)
    expect_equal(test_emit("bc", 1, 3, eps, integer(0), TRUE, FALSE, integer(0)), log(1-eps))
    expect_equal(test_emit("bc", 1, 4, eps, integer(0), TRUE, FALSE, integer(0)), log(eps))
    expect_equal(test_emit("bc", 3, 3, eps, integer(0), TRUE, FALSE, integer(0)), log(eps))
    expect_equal(test_emit("bc", 3, 4, eps, integer(0), TRUE, FALSE, integer(0)), log(1-eps))

    # errors
    expect_equal(test_emit("bc", 0, 0, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("bc", 0, 3, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("bc", 3, 1, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    # X female
    expect_equal(test_emit("bc", 0, 0, eps, integer(0), TRUE, TRUE, integer(0)), 0)
    expect_equal(test_emit("bc", 0, 3, eps, integer(0), TRUE, TRUE, integer(0)), 0)
    expect_equal(test_emit("bc", 3, 1, eps, integer(0), TRUE, TRUE, integer(0)), 0)
    # X male
    expect_equal(test_emit("bc", 0, 0, eps, integer(0), TRUE, FALSE, integer(0)), 0)
    expect_equal(test_emit("bc", 0, 1, eps, integer(0), TRUE, FALSE, integer(0)), 0)
    expect_equal(test_emit("bc", 0, 2, eps, integer(0), TRUE, FALSE, integer(0)), 0)
    expect_equal(test_emit("bc", 2, 1, eps, integer(0), TRUE, FALSE, integer(0)), 0)

})

test_that("backcross step works", {

    # autosome
    rf <- 0.01
    expect_equal(test_step("bc", 1, 1, rf, FALSE, FALSE, integer(0)), log(1-rf))
    expect_equal(test_step("bc", 1, 2, rf, FALSE, FALSE, integer(0)), log(rf))
    expect_equal(test_step("bc", 2, 1, rf, FALSE, FALSE, integer(0)), log(rf))
    expect_equal(test_step("bc", 2, 2, rf, FALSE, FALSE, integer(0)), log(1-rf))
    # X female
    expect_equal(test_step("bc", 1, 1, rf, TRUE, TRUE, integer(0)), log(1-rf))
    expect_equal(test_step("bc", 1, 2, rf, TRUE, TRUE, integer(0)), log(rf))
    expect_equal(test_step("bc", 2, 1, rf, TRUE, TRUE, integer(0)), log(rf))
    expect_equal(test_step("bc", 2, 2, rf, TRUE, TRUE, integer(0)), log(1-rf))
    # X male
    expect_equal(test_step("bc", 3, 3, rf, TRUE, FALSE, integer(0)), log(1-rf))
    expect_equal(test_step("bc", 3, 4, rf, TRUE, FALSE, integer(0)), log(rf))
    expect_equal(test_step("bc", 4, 3, rf, TRUE, FALSE, integer(0)), log(rf))
    expect_equal(test_step("bc", 4, 4, rf, TRUE, FALSE, integer(0)), log(1-rf))

})

test_that("geno_names works", {
    expect_equal(geno_names("bc", c("B", "R"), FALSE), c("BB", "BR"))
    expect_equal(geno_names("bc", c("B", "R"), TRUE), c("BB", "BR", "BY", "RY"))
})

context("basic HMM functions in phase-known 3-way AIL")

test_that("AIL3PK nalleles works", {
    expect_equal(nalleles("ail3pk"), 3)
})

test_that("AIL3PK check_geno works", {

    # observed genotypes
    for(i in 0:5) {
        # Autosome
        expect_true(test_check_geno("ail3pk", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("ail3pk", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_true(test_check_geno("ail3pk", i, TRUE, TRUE, FALSE, 20))
    }
    for(i in c(-1, 6)) {
        # Autosome
        expect_false(test_check_geno("ail3pk", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("ail3pk", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_false(test_check_geno("ail3pk", i, TRUE, TRUE, FALSE, 20))
    }

    # true genotypes, autosome and female X
    for(i in 1:9) {
        # Autosome
        expect_true(test_check_geno("ail3pk", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("ail3pk", i, FALSE, TRUE, TRUE, 20))
    }
    for(i in c(0, 10)) {
        # Autosome
        expect_false(test_check_geno("ail3pk", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("ail3pk", i, FALSE, TRUE, TRUE, 20))
    }

    # true genotypes, autosome and female X
    for(i in 9 + 1:3) {
        expect_true(test_check_geno("ail3pk", i, FALSE, TRUE, FALSE, 20))
    }
    for(i in c(0:9, 9+4)) {
        expect_false(test_check_geno("ail3pk", i, FALSE, TRUE, FALSE, 20))
    }

})

test_that("AIL3PK n_gen works", {

    expect_equal(test_ngen("ail3pk", FALSE), 9)
    expect_equal(test_ngen("ail3pk", TRUE),  12)

})

test_that("AIL3PK possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("ail3pk", FALSE, FALSE, 20), 1:9)

    # X female
    expect_equal(test_possible_gen("ail3pk", TRUE, TRUE, 20), 1:9)

    # X male
    expect_equal(test_possible_gen("ail3pk", TRUE, FALSE, 20), 9+(1:3))

})

test_that("AIL3PK init works", {

    for(i in 1:9) {
        # autosome and female X
        expect_equal(test_init("ail3pk", i, FALSE, FALSE, 20), log(1/9))
        expect_equal(test_init("ail3pk", i, TRUE,  TRUE,  20), log(1/9))
    }

    for(i in 10:12)
        expect_equal(test_init("ail3pk", i, TRUE,  FALSE, 20), log(1/3))

})


test_that("AIL3PK emit works", {

    fgen <- c(1,3,0) # founder genotypes; 0=missing, 1=AA, 3=BB
    err <- 0.01

    # Autosome or female X
    # truth = homA: AA (1)
    expected <- log(c(1-err, err/2, err/2, 1-err/2, err))
    for(trueg in 1) {
        for(obsg in 1:5) {
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = het: AB (2) or BA (7)
    expected <- log(c(err/2, 1-err, err/2, 1-err/2, 1-err/2))
    for(trueg in c(2,7)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = homB: BB (3)
    expected <- log(c(err/2, err/2, 1-err, err, 1-err/2))
    for(trueg in 3) {
        for(obsg in 1:5) {
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = A-: AC (4) or CA (8)
    expected <- log(c(1-err,1,err,1-err,err))
    for(trueg in c(4,8)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }
    # truth = B-: BC (5) or CB (9)
    expected <- log(c(err,1,1-err,err,1-err))
    for(trueg in c(5,9)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
        }
    }

    # male X: treat het as missing
    # truth = hemA: A (1+9)
    expected <- log(c(1-err, 1, err, 1-err, err))
    for(trueg in 10)
        for(obsg in 1:5)
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])
    # truth = hemB: BB (2+9)
    expected <- log(c(err, 1, 1-err, err, 1-err))
    for(trueg in 11)
        for(obsg in 1:5)
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])
    # truth = missing: C (3+9)
    expected <- rep(0,5)
    for(trueg in 12)
        for(obsg in 1:5)
            expect_equal(test_emit("ail3pk", obsg, trueg, err, fgen,  TRUE, FALSE, 20), expected[obsg])

})

test_that("AIL3PK step works for autosome", {

    pAA <- function(n, r, k=3)
    {
        stopifnot(n >= 2)
        (1 - (1 - k + k*r)*(1-r)^(n-2))/k^2
    }

    rf <- 0.02
    ngen <- 10

    pAA <- pAA(ngen, rf)
    R <- (1 - 3*pAA)

    # expected values
    expected <- matrix(1, ncol=9, nrow=9)
    a1 <- c("A", "A", "B", "A", "B", "C", "B", "C", "C")
    a2 <- c("A", "B", "B", "C", "C", "C", "A", "A", "B")

    for(i in 1:9) {
        for(j in 1:9) {
            expected[i,j] <- expected[i,j] * ifelse(a1[i]==a1[j], 1-R, R/2)
            expected[i,j] <- expected[i,j] * ifelse(a2[i]==a2[j], 1-R, R/2)
        }
    }
    expected <- log(expected)

    result <- matrix(ncol=9, nrow=9)
    for(i in 1:9)
        for(j in 1:9)
            result[i,j] <- test_step("ail3pk", i, j, rf, FALSE, FALSE, ngen)

    expect_equal(result, expected)

})


test_that("AIL3PK step works for X chromosome", {

    pAA <- function(n, r, k=3)
    {
        stopifnot(n >= 2)
        z <- sqrt((1-r)*(9-r))

        male <- (1-r)/k * ( (1-r+z)/(2*z) * ((1-r-z)/4)^(n-2) +
                            (-1+r+z)/(2*z) * ((1-r+z)/4)^(n-2)) +
            (2-r)/(2*k) * ( (1-r-z)/2 * (1-r+z)/(2*z) * ((1-r-z)/4)^(n-2) +
                            (1-r+z)/2 * (-1+r+z)/(2*z) * ((1-r+z)/4)^(n-2)) +
            ( (r^2 + r*(z-5))/(k^2*(3+r+z)) * (1-r+z)/(2*z) * ((1-r-z)/4)^(n-2) +
              (r^2 - r*(z+5))/(k^2*(3+r-z)) * (-1+r+z)/(2*z) * ((1-r+z)/4)^(n-2) + 1/k^2)

        female <- (1-r)/k * ( (-1/z)*((1-r-z)/4)^(n-2) +
                              (1/z)*((1-r+z)/4)^(n-2)) +
            (2-r)/(2*k) * ( (1-r-z)/2 * (-1/z)*((1-r-z)/4)^(n-2) +
                            (1-r+z)/2 *  (1/z)*((1-r+z)/4)^(n-2)) +
            ( (r^2 + r*(z-5))/(k^2*(3+r+z)) * (-1/z)*((1-r-z)/4)^(n-2) +
              (r^2 - r*(z+5))/(k^2*(3+r-z)) * (1/z)*((1-r+z)/4)^(n-2)  + 1/k^2)

        if(any(r==0))
            male[r==0] <- female[r==0] <- 1/k

        if(any(r==1)) {
            if(n==2) male[r==1] <- 0
            if(n>2) male[r==1] <- 1/k^2
            if(n==2) female[r==1] <- 1/(2*k)
            if(n==3) female[r==1] <- 1/(2*k^2)
            if(n>3)  female[r==1] <- 1/k^2
        }

        list(female=female, male=male)
    }

    rf <- 0.02
    ngen <- 10

    pAA <- pAA(ngen, rf)

    # female X
    R <- (1 - 3*pAA$female)

    # expected values
    expected <- matrix(1, ncol=9, nrow=9)
    a1 <- c("A", "A", "B", "A", "B", "C", "B", "C", "C")
    a2 <- c("A", "B", "B", "C", "C", "C", "A", "A", "B")

    for(i in 1:9) {
        for(j in 1:9) {
            expected[i,j] <- expected[i,j] * ifelse(a1[i]==a1[j], 1-R, R/2)
            expected[i,j] <- expected[i,j] * ifelse(a2[i]==a2[j], 1-R, R/2)
        }
    }
    expected <- log(expected)

    result <- matrix(ncol=9, nrow=9)
    for(i in 1:9)
        for(j in 1:9)
            result[i,j] <- test_step("ail3pk", i, j, rf, TRUE, TRUE, ngen)

    expect_equal(result, expected)

    # male X
    R <- (1 - 3*pAA$male)

    # expected values
    expected <- matrix(R/2, ncol=3, nrow=3)
    diag(expected) <- 1-R
    expected <- log(expected)

    result <- matrix(ncol=3, nrow=3)
    for(i in 1:3)
        for(j in 1:3)
            result[i,j] <- test_step("ail3pk", 9+i, 9+j, rf, TRUE, FALSE, ngen)

    expect_equal(result, expected)

})

context("basic HMM functions in phase-known AIL")

test_that("phase-known AIL nalleles works", {
    expect_equal(nalleles("ailpk"), 2)
})

test_that("phase-known AIL check_geno works", {

    # autosome
    for(i in 0:5)
        expect_true(test_check_geno("ailpk", i, TRUE, FALSE, FALSE, c(20, 0)))
    for(i in 1:4)
        expect_true(test_check_geno("ailpk", i, FALSE, FALSE, FALSE, c(20, 0)))
    for(i in c(-1, 6))
        expect_false(test_check_geno("ailpk", i, TRUE, FALSE, FALSE, c(20, 0)))
    for(i in c(0, 5))
        expect_false(test_check_geno("ailpk", i, FALSE, FALSE, FALSE, c(20, 0)))

    for(dir in 0:2) {
        # X chromosome female
        for(i in 0:5)
            expect_true(test_check_geno("ailpk", i, TRUE, TRUE, TRUE, c(20, dir)))
        for(i in 1:4)
            expect_true(test_check_geno("ailpk", i, FALSE, TRUE, TRUE, c(20, dir)))
        for(i in c(-1, 6))
            expect_false(test_check_geno("ailpk", i, TRUE, TRUE, TRUE, c(20, dir)))
        for(i in c(0, 5:6))
            expect_false(test_check_geno("ailpk", i, FALSE, TRUE, TRUE, c(20, dir)))

        # X chromosome male
        for(i in 0:5)
            expect_true(test_check_geno("ailpk", i, TRUE, TRUE, FALSE, c(20, dir)))
        for(i in c(5,6))
            expect_true(test_check_geno("ailpk", i, FALSE, TRUE, FALSE, c(20, dir)))
        for(i in c(-1, 6))
            expect_false(test_check_geno("ailpk", i, TRUE, TRUE, FALSE, c(20, dir)))
        for(i in c(0:4, 7))
            expect_false(test_check_geno("ailpk", i, FALSE, TRUE, FALSE, c(20, dir)))
    }

})

test_that("phase-known AIL n_gen works", {

    expect_equal(test_ngen("ailpk", FALSE), 4)
    expect_equal(test_ngen("ailpk", TRUE),  6)

})

test_that("phase-known AIL possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("ailpk", FALSE, FALSE, c(20,0)), 1:4)

    for(dir in 0:2) {
        # X female
        expect_equal(test_possible_gen("ailpk", TRUE, TRUE, c(20,dir)), 1:4)
        # X male
        expect_equal(test_possible_gen("ailpk", TRUE, FALSE, c(20,dir)), 5:6)
    }

})

test_that("phase-known AIL init works", {

    for(n_gen in c(3,9,12,15)) {
        # autosome
        for(i in 1:4)
            expect_equal(test_init("ailpk", i, FALSE, FALSE, c(n_gen, 0)), log(0.25))

        # X chr balanced
        #     female
        for(i in 1:4)
            expect_equal(test_init("ailpk", i, TRUE, TRUE, c(n_gen, 2)), log(0.25))

        #     male
        for(i in 5:6)
            expect_equal(test_init("ailpk", i, TRUE, FALSE, c(n_gen, 2)), log(0.5))
    }

    for(n_gen in c(3,9,12,15)) {
        fprob <- (2/3) + (1/3)*(-1/2)^n_gen
        mprob <- (2/3) + (1/3)*(-1/2)^(n_gen-1)

        # X chr, AxB
        #     female
        expect_equal(test_init("ailpk", 1, TRUE, TRUE, c(n_gen, 0)), log(fprob^2))
        expect_equal(test_init("ailpk", 2, TRUE, TRUE, c(n_gen, 0)), log(fprob*(1-fprob)))
        expect_equal(test_init("ailpk", 3, TRUE, TRUE, c(n_gen, 0)), log(fprob*(1-fprob)))
        expect_equal(test_init("ailpk", 4, TRUE, TRUE, c(n_gen, 0)), log((1-fprob)^2))

        #     male
        expect_equal(test_init("ailpk", 5, TRUE, FALSE, c(n_gen, 0)), log(mprob))
        expect_equal(test_init("ailpk", 6, TRUE, FALSE, c(n_gen, 0)), log(1-mprob))

        # X chr, BxA
        #     female
        expect_equal(test_init("ailpk", 4, TRUE, TRUE, c(n_gen, 1)), log(fprob^2))
        expect_equal(test_init("ailpk", 3, TRUE, TRUE, c(n_gen, 1)), log(fprob*(1-fprob)))
        expect_equal(test_init("ailpk", 2, TRUE, TRUE, c(n_gen, 1)), log(fprob*(1-fprob)))
        expect_equal(test_init("ailpk", 1, TRUE, TRUE, c(n_gen, 1)), log((1-fprob)^2))

        #     male
        expect_equal(test_init("ailpk", 6, TRUE, FALSE, c(n_gen, 1)), log(mprob))
        expect_equal(test_init("ailpk", 5, TRUE, FALSE, c(n_gen, 1)), log(1-mprob))
    }

})

test_that("phase-known AIL emit works", {

    # autosome
    eps <- 0.01
    for(i in 1:4)
        expect_equal(test_emit("ailpk", 0, i, eps, integer(0), FALSE, FALSE, c(20,0)), 0)

    expect_equal(test_emit("ailpk", 1, 1, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ailpk", 2, 1, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ailpk", 3, 1, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ailpk", 4, 1, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps/2))
    expect_equal(test_emit("ailpk", 5, 1, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps))

    for(i in 2:3) {
        expect_equal(test_emit("ailpk", 1, i, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
        expect_equal(test_emit("ailpk", 2, i, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps))
        expect_equal(test_emit("ailpk", 3, i, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
        expect_equal(test_emit("ailpk", 4, i, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps/2))
        expect_equal(test_emit("ailpk", 5, i, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps/2))
    }

    expect_equal(test_emit("ailpk", 1, 4, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ailpk", 2, 4, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ailpk", 3, 4, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ailpk", 4, 4, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps))
    expect_equal(test_emit("ailpk", 5, 4, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps/2))


    # X female
    for(i in 1:4)
        expect_equal(test_emit("ailpk", 0, i, eps, integer(0), TRUE, TRUE, c(20,0)), 0)

    expect_equal(test_emit("ailpk", 1, 1, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ailpk", 2, 1, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ailpk", 3, 1, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ailpk", 4, 1, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps/2))
    expect_equal(test_emit("ailpk", 5, 1, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps))

    for(i in 2:3) {
        expect_equal(test_emit("ailpk", 1, i, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
        expect_equal(test_emit("ailpk", 2, i, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps))
        expect_equal(test_emit("ailpk", 3, i, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
        expect_equal(test_emit("ailpk", 4, i, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps/2))
        expect_equal(test_emit("ailpk", 5, i, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps/2))
    }

    expect_equal(test_emit("ailpk", 1, 4, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ailpk", 2, 4, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ailpk", 3, 4, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ailpk", 4, 4, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps))
    expect_equal(test_emit("ailpk", 5, 4, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps/2))

    # X male
    for(i in 5:6)
        expect_equal(test_emit("ailpk", 0, i, eps, integer(0), TRUE, FALSE, c(20,0)), 0)
    expect_equal(test_emit("ailpk", 1, 5, eps, integer(0), TRUE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ailpk", 2, 5, eps, integer(0), TRUE, FALSE, c(20,0)), 0)
    expect_equal(test_emit("ailpk", 3, 5, eps, integer(0), TRUE, FALSE, c(20,0)), log(eps))
    expect_equal(test_emit("ailpk", 4, 5, eps, integer(0), TRUE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ailpk", 5, 5, eps, integer(0), TRUE, FALSE, c(20,0)), log(eps))
    expect_equal(test_emit("ailpk", 1, 6, eps, integer(0), TRUE, FALSE, c(20,0)), log(eps))
    expect_equal(test_emit("ailpk", 2, 6, eps, integer(0), TRUE, FALSE, c(20,0)), 0)
    expect_equal(test_emit("ailpk", 3, 6, eps, integer(0), TRUE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ailpk", 4, 6, eps, integer(0), TRUE, FALSE, c(20,0)), log(eps))
    expect_equal(test_emit("ailpk", 5, 6, eps, integer(0), TRUE, FALSE, c(20,0)), log(1-eps))

})

test_that("phase-known AIL step works", {

    skip_on_cran()

    # autosome
    for(rf in c(0.01, 0.0001)) {
        for(ngen in c(3, 9, 12, 15)) {
            R <- 1 - 2*( 0.25*(1 + (1-2*rf)*(1-rf)^(ngen-2)))

            expect_equal(test_step("ailpk", 1, 1, rf, FALSE, FALSE, c(ngen, 0)), log((1-R)^2))
            expect_equal(test_step("ailpk", 1, 2, rf, FALSE, FALSE, c(ngen, 0)), log(R*(1-R)))
            expect_equal(test_step("ailpk", 1, 3, rf, FALSE, FALSE, c(ngen, 0)), log(R*(1-R)))
            expect_equal(test_step("ailpk", 1, 4, rf, FALSE, FALSE, c(ngen, 0)), log(R^2))
            expect_equal(test_step("ailpk", 2, 1, rf, FALSE, FALSE, c(ngen, 0)), log(R*(1-R)))
            expect_equal(test_step("ailpk", 2, 2, rf, FALSE, FALSE, c(ngen, 0)), log((1-R)^2))
            expect_equal(test_step("ailpk", 2, 3, rf, FALSE, FALSE, c(ngen, 0)), log(R^2))
            expect_equal(test_step("ailpk", 2, 4, rf, FALSE, FALSE, c(ngen, 0)), log(R*(1-R)))
            expect_equal(test_step("ailpk", 3, 1, rf, FALSE, FALSE, c(ngen, 0)), log(R*(1-R)))
            expect_equal(test_step("ailpk", 3, 2, rf, FALSE, FALSE, c(ngen, 0)), log(R^2))
            expect_equal(test_step("ailpk", 3, 3, rf, FALSE, FALSE, c(ngen, 0)), log((1-R)^2))
            expect_equal(test_step("ailpk", 3, 4, rf, FALSE, FALSE, c(ngen, 0)), log(R*(1-R)))
            expect_equal(test_step("ailpk", 4, 1, rf, FALSE, FALSE, c(ngen, 0)), log(R^2))
            expect_equal(test_step("ailpk", 4, 2, rf, FALSE, FALSE, c(ngen, 0)), log(R*(1-R)))
            expect_equal(test_step("ailpk", 4, 3, rf, FALSE, FALSE, c(ngen, 0)), log(R*(1-R)))
            expect_equal(test_step("ailpk", 4, 4, rf, FALSE, FALSE, c(ngen, 0)), log((1-R)^2))
        }
    }

    # X chromosome, balanced
    for(rf in c(0.01, 0.0001)) {
        for(ngen in c(3, 9, 12, 15)) {
            z <- sqrt((1-rf)*(9-rf))
            w <- (1 - rf + z)/4
            y <- (1 - rf - z)/4
            mR <- 1 - 0.25*(2 + (1-2*rf)*(w^(ngen-2) + y^(ngen-2)) +
                            (3 - 5*rf + 2*rf^2)/z*(w^(ngen-2) - y^(ngen-2)))
            fR <- 1 - 0.25*(2 + (1-2*rf)*(w^(ngen-2) + y^(ngen-2)) +
                            (3 - 6*rf + rf^2)/z*(w^(ngen-2) - y^(ngen-2)))

            expect_equal(test_step("ailpk", 1, 1, rf, TRUE, TRUE, c(ngen, 2)), log((1-fR)^2))
            expect_equal(test_step("ailpk", 1, 2, rf, TRUE, TRUE, c(ngen, 2)), log(fR*(1-fR)))
            expect_equal(test_step("ailpk", 1, 3, rf, TRUE, TRUE, c(ngen, 2)), log(fR*(1-fR)))
            expect_equal(test_step("ailpk", 1, 4, rf, TRUE, TRUE, c(ngen, 2)), log(fR^2))
            expect_equal(test_step("ailpk", 2, 1, rf, TRUE, TRUE, c(ngen, 2)), log(fR*(1-fR)))
            expect_equal(test_step("ailpk", 2, 2, rf, TRUE, TRUE, c(ngen, 2)), log((1-fR)^2))
            expect_equal(test_step("ailpk", 2, 3, rf, TRUE, TRUE, c(ngen, 2)), log(fR^2))
            expect_equal(test_step("ailpk", 2, 4, rf, TRUE, TRUE, c(ngen, 2)), log(fR*(1-fR)))
            expect_equal(test_step("ailpk", 3, 1, rf, TRUE, TRUE, c(ngen, 2)), log(fR*(1-fR)))
            expect_equal(test_step("ailpk", 3, 2, rf, TRUE, TRUE, c(ngen, 2)), log(fR^2))
            expect_equal(test_step("ailpk", 3, 3, rf, TRUE, TRUE, c(ngen, 2)), log((1-fR)^2))
            expect_equal(test_step("ailpk", 3, 4, rf, TRUE, TRUE, c(ngen, 2)), log(fR*(1-fR)))
            expect_equal(test_step("ailpk", 4, 1, rf, TRUE, TRUE, c(ngen, 2)), log(fR^2))
            expect_equal(test_step("ailpk", 4, 2, rf, TRUE, TRUE, c(ngen, 2)), log(fR*(1-fR)))
            expect_equal(test_step("ailpk", 4, 3, rf, TRUE, TRUE, c(ngen, 2)), log(fR*(1-fR)))
            expect_equal(test_step("ailpk", 4, 4, rf, TRUE, TRUE, c(ngen, 2)), log((1-fR)^2))

            expect_equal(test_step("ailpk", 5, 5, rf, TRUE, FALSE, c(ngen, 2)), log(1-mR))
            expect_equal(test_step("ailpk", 5, 6, rf, TRUE, FALSE, c(ngen, 2)), log(mR))
            expect_equal(test_step("ailpk", 6, 5, rf, TRUE, FALSE, c(ngen, 2)), log(mR))
            expect_equal(test_step("ailpk", 6, 6, rf, TRUE, FALSE, c(ngen, 2)), log(1-mR))

        }
    }


    # calculate Pr(A) for females, AxB
    calc_q <- function(n_gen)
        2/3 + (1/3)*(-1/2)^n_gen

    # calc Pr(AA haplotype) in males and females, recursively
    calc_p11 <- function(n_gen, rf) {
        if(n_gen == 1) return(c(1, 0.5))
        last <- calc_p11(n_gen-1, rf)
        z <- c((1-rf)*last[2] + rf*calc_q(n_gen-2)*calc_q(n_gen-3),
               0.5*last[1] + (1-rf)/2*last[2] + (rf/2)*calc_q(n_gen-2)*calc_q(n_gen-3))
        z
    }

    # X chromosome, AxB and BxA
    for(rf in c(0.01, 0.0001)) {
        for(ngen in c(3, 9, 12, 15)) {

            qf <- calc_q(ngen)
            qm <- calc_q(ngen-1) # Pr(A) in males is Pr(A) in females for prev generation
            p11 <- calc_p11(ngen, rf) # Pr(AA haplotype) in males and females
            m11 <- p11[1]
            f11 <- p11[2]

            # turn into conditional probabilities along one haplotype
            m1to1 <- m11/qm
            m1to2 <- 1 - m1to1
            m2to1 <- (qm - m11)/(1-qm)
            m2to2 <- 1 - m2to1

            f1to1 <- f11/qf
            f1to2 <- 1 - f1to1
            f2to1 <- (qf - f11)/(1-qf)
            f2to2 <- 1 - f2to1

            expect_equal(test_step("ailpk", 1, 1, rf, TRUE, TRUE, c(ngen, 0)), log(f1to1^2))
            expect_equal(test_step("ailpk", 1, 2, rf, TRUE, TRUE, c(ngen, 0)), log(f1to1*f1to2))
            expect_equal(test_step("ailpk", 1, 3, rf, TRUE, TRUE, c(ngen, 0)), log(f1to1*f1to2))
            expect_equal(test_step("ailpk", 1, 4, rf, TRUE, TRUE, c(ngen, 0)), log(f1to2^2))
            expect_equal(test_step("ailpk", 2, 1, rf, TRUE, TRUE, c(ngen, 0)), log(f1to1*f2to1))
            expect_equal(test_step("ailpk", 2, 2, rf, TRUE, TRUE, c(ngen, 0)), log(f1to1*f2to2))
            expect_equal(test_step("ailpk", 2, 3, rf, TRUE, TRUE, c(ngen, 0)), log(f2to1*f1to2))
            expect_equal(test_step("ailpk", 2, 4, rf, TRUE, TRUE, c(ngen, 0)), log(f1to2*f2to2))
            expect_equal(test_step("ailpk", 3, 1, rf, TRUE, TRUE, c(ngen, 0)), log(f1to1*f2to1))
            expect_equal(test_step("ailpk", 3, 3, rf, TRUE, TRUE, c(ngen, 0)), log(f1to1*f2to2))
            expect_equal(test_step("ailpk", 3, 2, rf, TRUE, TRUE, c(ngen, 0)), log(f2to1*f1to2))
            expect_equal(test_step("ailpk", 3, 4, rf, TRUE, TRUE, c(ngen, 0)), log(f1to2*f2to2))
            expect_equal(test_step("ailpk", 4, 1, rf, TRUE, TRUE, c(ngen, 0)), log(f2to1^2))
            expect_equal(test_step("ailpk", 4, 2, rf, TRUE, TRUE, c(ngen, 0)), log(f2to2*f2to1))
            expect_equal(test_step("ailpk", 4, 3, rf, TRUE, TRUE, c(ngen, 0)), log(f2to2*f2to1))
            expect_equal(test_step("ailpk", 4, 4, rf, TRUE, TRUE, c(ngen, 0)), log(f2to2^2))

            expect_equal(test_step("ailpk", 5, 5, rf, TRUE, FALSE, c(ngen, 0)), log(m1to1))
            expect_equal(test_step("ailpk", 5, 6, rf, TRUE, FALSE, c(ngen, 0)), log(m1to2))
            expect_equal(test_step("ailpk", 6, 5, rf, TRUE, FALSE, c(ngen, 0)), log(m2to1))
            expect_equal(test_step("ailpk", 6, 6, rf, TRUE, FALSE, c(ngen, 0)), log(m2to2))

            expect_equal(test_step("ailpk", 1, 1, rf, TRUE, TRUE, c(ngen, 1)), log(f2to2^2))
            expect_equal(test_step("ailpk", 1, 2, rf, TRUE, TRUE, c(ngen, 1)), log(f2to2*f2to1))
            expect_equal(test_step("ailpk", 1, 3, rf, TRUE, TRUE, c(ngen, 1)), log(f2to2*f2to1))
            expect_equal(test_step("ailpk", 1, 4, rf, TRUE, TRUE, c(ngen, 1)), log(f2to1^2))
            expect_equal(test_step("ailpk", 2, 1, rf, TRUE, TRUE, c(ngen, 1)), log(f2to2*f1to2))
            expect_equal(test_step("ailpk", 2, 2, rf, TRUE, TRUE, c(ngen, 1)), log(f1to1*f2to2))
            expect_equal(test_step("ailpk", 2, 3, rf, TRUE, TRUE, c(ngen, 1)), log(f1to2*f2to1))
            expect_equal(test_step("ailpk", 2, 4, rf, TRUE, TRUE, c(ngen, 1)), log(f2to1*f1to1))
            expect_equal(test_step("ailpk", 3, 1, rf, TRUE, TRUE, c(ngen, 1)), log(f2to2*f1to2))
            expect_equal(test_step("ailpk", 3, 3, rf, TRUE, TRUE, c(ngen, 1)), log(f1to1*f2to2))
            expect_equal(test_step("ailpk", 3, 2, rf, TRUE, TRUE, c(ngen, 1)), log(f1to2*f2to1))
            expect_equal(test_step("ailpk", 3, 4, rf, TRUE, TRUE, c(ngen, 1)), log(f2to1*f1to1))
            expect_equal(test_step("ailpk", 4, 1, rf, TRUE, TRUE, c(ngen, 1)), log(f1to2^2))
            expect_equal(test_step("ailpk", 4, 2, rf, TRUE, TRUE, c(ngen, 1)), log(f1to1*f1to2))
            expect_equal(test_step("ailpk", 4, 3, rf, TRUE, TRUE, c(ngen, 1)), log(f1to1*f1to2))
            expect_equal(test_step("ailpk", 4, 4, rf, TRUE, TRUE, c(ngen, 1)), log(f1to1^2))

            expect_equal(test_step("ailpk", 5, 5, rf, TRUE, FALSE, c(ngen, 1)), log(m2to2))
            expect_equal(test_step("ailpk", 5, 6, rf, TRUE, FALSE, c(ngen, 1)), log(m2to1))
            expect_equal(test_step("ailpk", 6, 5, rf, TRUE, FALSE, c(ngen, 1)), log(m1to2))
            expect_equal(test_step("ailpk", 6, 6, rf, TRUE, FALSE, c(ngen, 1)), log(m1to1))

        }
    }


})

test_that("nrec works for AIL-pk", {

    skip_on_cran()

    # autosome or female X chr
    expected <- rbind(c(0,1,1,2),
                      c(1,0,2,1),
                      c(1,2,0,1),
                      c(2,1,1,0))
    for(i in 1:4) {
        for(j in 1:4) {
            expect_equal(test_nrec("ailpk", i, j, FALSE, FALSE, c(20,0)), expected[i,j])
            expect_equal(test_nrec("ailpk", i, j, TRUE, TRUE, c(20,0)), expected[i,j])
        }
    }

    # male X chr
    for(i in 1:2) {
        for(j in 1:2) {
            expect_equal(test_nrec("ailpk", i, j, TRUE, FALSE, c(20,0)), expected[i,j])
        }
    }

})

context("basic HMM functions in phase-known intercross")

test_that("p-k intercross nalleles works", {
    expect_equal(nalleles("f2pk"), 2)
})

test_that("p-k intercross check_geno works", {

    # autosome
    for(i in 0:5)
        expect_true(test_check_geno("f2pk", i, TRUE, FALSE, FALSE, 0))
    for(i in 1:4)
        expect_true(test_check_geno("f2pk", i, FALSE, FALSE, FALSE, 0))
    for(i in c(-1, 6))
        expect_false(test_check_geno("f2pk", i, TRUE, FALSE, FALSE, 0))
    for(i in c(0, 5))
        expect_false(test_check_geno("f2pk", i, FALSE, FALSE, FALSE, 0))

    # X chromosome female, forward cross
    for(i in 0:5)
        expect_true(test_check_geno("f2pk", i, TRUE, TRUE, TRUE, 0))
    for(i in 1:2)
        expect_true(test_check_geno("f2pk", i, FALSE, TRUE, TRUE, 0))
    for(i in c(-1, 6))
        expect_false(test_check_geno("f2pk", i, TRUE, TRUE, TRUE, 0))
    for(i in c(0, 3, 4))
        expect_false(test_check_geno("f2pk", i, FALSE, TRUE, TRUE, 0))

    # X chromosome female, reverse cross
    for(i in 0:5)
        expect_true(test_check_geno("f2pk", i, TRUE, TRUE, TRUE, 1))
    for(i in 3:4)
        expect_true(test_check_geno("f2pk", i, FALSE, TRUE, TRUE, 1))
    for(i in c(-1, 6))
        expect_false(test_check_geno("f2pk", i, TRUE, TRUE, TRUE, 1))
    for(i in c(0, 1, 2, 5))
        expect_false(test_check_geno("f2pk", i, FALSE, TRUE, TRUE, 1))

    # X chromosome male
    for(i in 0:5)
        expect_true(test_check_geno("f2pk", i, TRUE, TRUE, FALSE, 0))
    for(i in 5:6)
        expect_true(test_check_geno("f2pk", i, FALSE, TRUE, FALSE, 0))
    for(i in c(-1, 6))
        expect_false(test_check_geno("f2pk", i, TRUE, TRUE, FALSE, 0))
    for(i in c(0:4, 7))
        expect_false(test_check_geno("f2pk", i, FALSE, TRUE, FALSE, 0))

    # X chromosome male
    for(i in 0:5)
        expect_true(test_check_geno("f2pk", i, TRUE, TRUE, FALSE, 1))
    for(i in 5:6)
        expect_true(test_check_geno("f2pk", i, FALSE, TRUE, FALSE, 1))
    for(i in c(-1, 6))
        expect_false(test_check_geno("f2pk", i, TRUE, TRUE, FALSE, 1))
    for(i in c(0:4, 7))
        expect_false(test_check_geno("f2pk", i, FALSE, TRUE, FALSE, 1))

})


test_that("p-k intercross ngen works", {

    expect_equal(test_ngen("f2pk", FALSE), 4)
    expect_equal(test_ngen("f2pk", TRUE),  6)
})

test_that("p-k intercross possible_gen works", {

    expect_equal(test_possible_gen("f2pk", FALSE, FALSE, 0), 1:4)
    expect_equal(test_possible_gen("f2pk", TRUE, TRUE, 0), 1:2)
    expect_equal(test_possible_gen("f2pk", TRUE, TRUE, 1), 3:4)
    expect_equal(test_possible_gen("f2pk", TRUE, FALSE, 0), 5:6)
    expect_equal(test_possible_gen("f2pk", TRUE, FALSE, 1), 5:6)
})


test_that("p-k intercross nrec works", {

    # autosome
    expect_equal(test_nrec("f2pk", 1, 1, FALSE, FALSE, 0), 0)
    expect_equal(test_nrec("f2pk", 1, 2, FALSE, FALSE, 0), 1)
    expect_equal(test_nrec("f2pk", 1, 3, FALSE, FALSE, 0), 1)
    expect_equal(test_nrec("f2pk", 1, 4, FALSE, FALSE, 0), 2)
    expect_equal(test_nrec("f2pk", 2, 1, FALSE, FALSE, 0), 1)
    expect_equal(test_nrec("f2pk", 2, 2, FALSE, FALSE, 0), 0)
    expect_equal(test_nrec("f2pk", 2, 3, FALSE, FALSE, 0), 2)
    expect_equal(test_nrec("f2pk", 2, 4, FALSE, FALSE, 0), 1)
    expect_equal(test_nrec("f2pk", 3, 1, FALSE, FALSE, 0), 1)
    expect_equal(test_nrec("f2pk", 3, 2, FALSE, FALSE, 0), 2)
    expect_equal(test_nrec("f2pk", 3, 3, FALSE, FALSE, 0), 0)
    expect_equal(test_nrec("f2pk", 3, 4, FALSE, FALSE, 0), 1)
    expect_equal(test_nrec("f2pk", 4, 1, FALSE, FALSE, 0), 2)
    expect_equal(test_nrec("f2pk", 4, 2, FALSE, FALSE, 0), 1)
    expect_equal(test_nrec("f2pk", 4, 3, FALSE, FALSE, 0), 1)
    expect_equal(test_nrec("f2pk", 4, 4, FALSE, FALSE, 0), 0)
    # X female forward
    expect_equal(test_nrec("f2pk", 1, 1, TRUE, TRUE, 0), 0)
    expect_equal(test_nrec("f2pk", 1, 2, TRUE, TRUE, 0), 1)
    expect_equal(test_nrec("f2pk", 2, 1, TRUE, TRUE, 0), 1)
    expect_equal(test_nrec("f2pk", 2, 2, TRUE, TRUE, 0), 0)
    # X female reverse
    expect_equal(test_nrec("f2pk", 3, 3, TRUE, TRUE, 1), 0)
    expect_equal(test_nrec("f2pk", 3, 4, TRUE, TRUE, 1), 1)
    expect_equal(test_nrec("f2pk", 4, 3, TRUE, TRUE, 1), 1)
    expect_equal(test_nrec("f2pk", 4, 4, TRUE, TRUE, 1), 0)
    # X male
    expect_equal(test_nrec("f2pk", 5, 5, TRUE, FALSE, 0), 0)
    expect_equal(test_nrec("f2pk", 5, 6, TRUE, FALSE, 0), 1)
    expect_equal(test_nrec("f2pk", 6, 5, TRUE, FALSE, 0), 1)
    expect_equal(test_nrec("f2pk", 6, 6, TRUE, FALSE, 0), 0)
    # X male reverse
    expect_equal(test_nrec("f2pk", 5, 5, TRUE, FALSE, 1), 0)
    expect_equal(test_nrec("f2pk", 5, 6, TRUE, FALSE, 1), 1)
    expect_equal(test_nrec("f2pk", 6, 5, TRUE, FALSE, 1), 1)
    expect_equal(test_nrec("f2pk", 6, 6, TRUE, FALSE, 1), 0)

})

test_that("p-k intercross init works", {

    # autosome
    for(i in 1:4)
        expect_equal(test_init("f2pk", i, FALSE, FALSE, 0), log(0.25))
    # X female forward
    for(i in 1:2)
        expect_equal(test_init("f2pk", i, TRUE, TRUE, 0), log(0.5))
    # X female reverse
    for(i in 3:4)
        expect_equal(test_init("f2pk", i, TRUE, TRUE, 1), log(0.5))
    # X male
    for(i in 5:6)
        expect_equal(test_init("f2pk", i, TRUE, FALSE, 0), log(0.5))
    # X male reverse
    for(i in 5:6)
        expect_equal(test_init("f2pk", i, TRUE, FALSE, 1), log(0.5))

})

test_that("p-k intercross emit works", {

    # autosome
    eps <- 0.02
    for(i in 1:4)
        expect_equal(test_emit("f2pk", 0, i, eps, integer(0), FALSE, FALSE, 0), 0)

    expect_equal(test_emit("f2pk", 1, 1, eps, integer(0), FALSE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2pk", 2, 1, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
    expect_equal(test_emit("f2pk", 3, 1, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
    expect_equal(test_emit("f2pk", 4, 1, eps, integer(0), FALSE, FALSE, 0), log(1-eps/2))
    expect_equal(test_emit("f2pk", 5, 1, eps, integer(0), FALSE, FALSE, 0), log(eps))

    for(i in 2:3) {
        expect_equal(test_emit("f2pk", 1, i, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
        expect_equal(test_emit("f2pk", 2, i, eps, integer(0), FALSE, FALSE, 0), log(1-eps))
        expect_equal(test_emit("f2pk", 3, i, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
        expect_equal(test_emit("f2pk", 4, i, eps, integer(0), FALSE, FALSE, 0), log(1-eps/2))
        expect_equal(test_emit("f2pk", 5, i, eps, integer(0), FALSE, FALSE, 0), log(1-eps/2))
    }

    expect_equal(test_emit("f2pk", 1, 4, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
    expect_equal(test_emit("f2pk", 2, 4, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
    expect_equal(test_emit("f2pk", 3, 4, eps, integer(0), FALSE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2pk", 4, 4, eps, integer(0), FALSE, FALSE, 0), log(eps))
    expect_equal(test_emit("f2pk", 5, 4, eps, integer(0), FALSE, FALSE, 0), log(1-eps/2))

    # X female forward
    for(i in 1:2)
        expect_equal(test_emit("f2pk", 0, i, eps, integer(0), TRUE, TRUE, 0), 0)

    expect_equal(test_emit("f2pk", 1, 1, eps, integer(0), TRUE, TRUE, 0), log(1-eps))
    expect_equal(test_emit("f2pk", 2, 1, eps, integer(0), TRUE, TRUE, 0), log(eps))
    expect_equal(test_emit("f2pk", 3, 1, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("f2pk", 4, 1, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("f2pk", 5, 1, eps, integer(0), TRUE, TRUE, 0), log(eps))

    expect_equal(test_emit("f2pk", 1, 2, eps, integer(0), TRUE, TRUE, 0), log(eps))
    expect_equal(test_emit("f2pk", 2, 2, eps, integer(0), TRUE, TRUE, 0), log(1-eps))
    expect_equal(test_emit("f2pk", 3, 2, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("f2pk", 4, 2, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("f2pk", 5, 2, eps, integer(0), TRUE, TRUE, 0), log(1-eps))

    # X female reverse
    for(i in 3:4)
        expect_equal(test_emit("f2pk", 0, i, eps, integer(0), TRUE, TRUE, 1), 0)
    expect_equal(test_emit("f2pk", 1, 4, eps, integer(0), TRUE, TRUE, 1), 0)
    expect_equal(test_emit("f2pk", 2, 4, eps, integer(0), TRUE, TRUE, 1), log(eps))
    expect_equal(test_emit("f2pk", 3, 4, eps, integer(0), TRUE, TRUE, 1), log(1-eps))
    expect_equal(test_emit("f2pk", 4, 4, eps, integer(0), TRUE, TRUE, 1), log(eps))
    expect_equal(test_emit("f2pk", 5, 4, eps, integer(0), TRUE, TRUE, 1), 0)
    expect_equal(test_emit("f2pk", 1, 3, eps, integer(0), TRUE, TRUE, 1), 0)
    expect_equal(test_emit("f2pk", 2, 3, eps, integer(0), TRUE, TRUE, 1), log(1-eps))
    expect_equal(test_emit("f2pk", 3, 3, eps, integer(0), TRUE, TRUE, 1), log(eps))
    expect_equal(test_emit("f2pk", 4, 3, eps, integer(0), TRUE, TRUE, 1), log(1-eps))
    expect_equal(test_emit("f2pk", 5, 3, eps, integer(0), TRUE, TRUE, 1), 0)

    # X male
    for(i in 5:6)
        expect_equal(test_emit("f2pk", 0, i, eps, integer(0), TRUE, FALSE, 0), 0)
    expect_equal(test_emit("f2pk", 1, 5, eps, integer(0), TRUE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2pk", 2, 5, eps, integer(0), TRUE, FALSE, 0), 0)
    expect_equal(test_emit("f2pk", 3, 5, eps, integer(0), TRUE, FALSE, 0), log(eps))
    expect_equal(test_emit("f2pk", 4, 5, eps, integer(0), TRUE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2pk", 5, 5, eps, integer(0), TRUE, FALSE, 0), log(eps))
    expect_equal(test_emit("f2pk", 1, 6, eps, integer(0), TRUE, FALSE, 0), log(eps))
    expect_equal(test_emit("f2pk", 2, 6, eps, integer(0), TRUE, FALSE, 0), 0)
    expect_equal(test_emit("f2pk", 3, 6, eps, integer(0), TRUE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2pk", 4, 6, eps, integer(0), TRUE, FALSE, 0), log(eps))
    expect_equal(test_emit("f2pk", 5, 6, eps, integer(0), TRUE, FALSE, 0), log(1-eps))

    # X male reverse
    for(i in 5:6)
        expect_equal(test_emit("f2pk", 0, i, eps, integer(0), TRUE, FALSE, 1), 0)
    expect_equal(test_emit("f2pk", 1, 5, eps, integer(0), TRUE, FALSE, 1), log(1-eps))
    expect_equal(test_emit("f2pk", 2, 5, eps, integer(0), TRUE, FALSE, 1), 0)
    expect_equal(test_emit("f2pk", 3, 5, eps, integer(0), TRUE, FALSE, 1), log(eps))
    expect_equal(test_emit("f2pk", 4, 5, eps, integer(0), TRUE, FALSE, 1), log(1-eps))
    expect_equal(test_emit("f2pk", 5, 5, eps, integer(0), TRUE, FALSE, 1), log(eps))
    expect_equal(test_emit("f2pk", 1, 6, eps, integer(0), TRUE, FALSE, 1), log(eps))
    expect_equal(test_emit("f2pk", 2, 6, eps, integer(0), TRUE, FALSE, 1), 0)
    expect_equal(test_emit("f2pk", 3, 6, eps, integer(0), TRUE, FALSE, 1), log(1-eps))
    expect_equal(test_emit("f2pk", 4, 6, eps, integer(0), TRUE, FALSE, 1), log(eps))
    expect_equal(test_emit("f2pk", 5, 6, eps, integer(0), TRUE, FALSE, 1), log(1-eps))

})

test_that("p-k intercross step works", {

    # autosome
    rf <- 0.15
    expect_equal(test_step("f2pk", 1, 1, rf, FALSE, FALSE, 0), log((1-rf)^2))
    expect_equal(test_step("f2pk", 1, 2, rf, FALSE, FALSE, 0), log(rf*(1-rf)))
    expect_equal(test_step("f2pk", 1, 3, rf, FALSE, FALSE, 0), log(rf*(1-rf)))
    expect_equal(test_step("f2pk", 1, 4, rf, FALSE, FALSE, 0), log(rf^2))
    expect_equal(test_step("f2pk", 2, 1, rf, FALSE, FALSE, 0), log(rf*(1-rf)))
    expect_equal(test_step("f2pk", 2, 2, rf, FALSE, FALSE, 0), log((1-rf)^2))
    expect_equal(test_step("f2pk", 2, 3, rf, FALSE, FALSE, 0), log(rf^2))
    expect_equal(test_step("f2pk", 2, 4, rf, FALSE, FALSE, 0), log(rf*(1-rf)))
    expect_equal(test_step("f2pk", 3, 1, rf, FALSE, FALSE, 0), log(rf*(1-rf)))
    expect_equal(test_step("f2pk", 3, 2, rf, FALSE, FALSE, 0), log(rf^2))
    expect_equal(test_step("f2pk", 3, 3, rf, FALSE, FALSE, 0), log((1-rf)^2))
    expect_equal(test_step("f2pk", 3, 4, rf, FALSE, FALSE, 0), log(rf*(1-rf)))
    expect_equal(test_step("f2pk", 4, 1, rf, FALSE, FALSE, 0), log(rf^2))
    expect_equal(test_step("f2pk", 4, 2, rf, FALSE, FALSE, 0), log(rf*(1-rf)))
    expect_equal(test_step("f2pk", 4, 3, rf, FALSE, FALSE, 0), log(rf*(1-rf)))
    expect_equal(test_step("f2pk", 4, 4, rf, FALSE, FALSE, 0), log((1-rf)^2))
    # X female
    expect_equal(test_step("f2pk", 1, 1, rf, TRUE, TRUE, 0), log(1-rf))
    expect_equal(test_step("f2pk", 1, 2, rf, TRUE, TRUE, 0), log(rf))
    expect_equal(test_step("f2pk", 2, 1, rf, TRUE, TRUE, 0), log(rf))
    expect_equal(test_step("f2pk", 2, 2, rf, TRUE, TRUE, 0), log(1-rf))
    # X female reverse
    expect_equal(test_step("f2pk", 3, 3, rf, TRUE, TRUE, 1), log(1-rf))
    expect_equal(test_step("f2pk", 3, 4, rf, TRUE, TRUE, 1), log(rf))
    expect_equal(test_step("f2pk", 4, 3, rf, TRUE, TRUE, 1), log(rf))
    expect_equal(test_step("f2pk", 4, 4, rf, TRUE, TRUE, 1), log(1-rf))
    # X male
    expect_equal(test_step("f2pk", 5, 5, rf, TRUE, FALSE, 0), log(1-rf))
    expect_equal(test_step("f2pk", 5, 6, rf, TRUE, FALSE, 0), log(rf))
    expect_equal(test_step("f2pk", 6, 5, rf, TRUE, FALSE, 0), log(rf))
    expect_equal(test_step("f2pk", 6, 6, rf, TRUE, FALSE, 0), log(1-rf))
    # X male reverse
    expect_equal(test_step("f2pk", 5, 5, rf, TRUE, FALSE, 1), log(1-rf))
    expect_equal(test_step("f2pk", 5, 6, rf, TRUE, FALSE, 1), log(rf))
    expect_equal(test_step("f2pk", 6, 5, rf, TRUE, FALSE, 1), log(rf))
    expect_equal(test_step("f2pk", 6, 6, rf, TRUE, FALSE, 1), log(1-rf))

})

context("basic HMM functions in 4-, 8-, and 16-way RIL by selfing")

test_that("riself4-8-16 n_gen, n_alleles work", {

    expect_equal(nalleles("riself4"), 4)
    expect_equal(nalleles("riself8"), 8)
    expect_equal(nalleles("riself16"), 16)

    expect_equal(test_ngen("riself4", FALSE), 4)
    expect_equal(test_ngen("riself8", FALSE), 8)
    expect_equal(test_ngen("riself16", FALSE), 16)

})

test_that("riself4-8-16 possible_gen work", {

    expect_equal(test_possible_gen("riself4", FALSE, FALSE, 1:4), 1:4)
    expect_equal(test_possible_gen("riself8", FALSE, FALSE, 1:8), 1:8)
    expect_equal(test_possible_gen("riself16", FALSE, FALSE, 1:16), 1:16)

})

# FIX_ME: test check_geno

test_that("riself4-8-16 init work", {

    expect_equal( sapply(1:4, function(i) test_init("riself4", i, FALSE, FALSE, 1:4)), log(rep(1/4, 4)) )
    expect_equal( sapply(1:8, function(i) test_init("riself8", i, FALSE, FALSE, 1:8)), log(rep(1/8, 8)) )
    expect_equal( sapply(1:16, function(i) test_init("riself16", i, FALSE, FALSE, 1:16)), log(rep(1/16, 16)) )

})

# FIX_ME: test emit

test_that("riself4 step works", {

    for(rf in c(0.01, 0.1, 0.45)) {
        expected <- matrix(rf/(1+2*rf), ncol=4, nrow=4)
        diag(expected) <- (1-rf)/(1+2*rf)

        result <- matrix(ncol=4, nrow=4)
        for(i in 1:4) {
            for(j in 1:4) {
                result[i,j] <- test_step("riself4", i, j, rf, FALSE, FALSE, 1:4)
            }
        }

        expect_equal(result, log(expected))
    }

    # test_stepmatrix: output has rows that sum to 1
    obs <- vapply(test_stepmatrix("riself4", c(0.01, 0.1, 0.45), FALSE, FALSE, 1:4),
                  function(a) rowSums(exp(a)), rep(1,4))
    expect_equal(obs, matrix(1, ncol=3, nrow=4))

    # test_stepmatrix: output has rows that sum to 1
    obs <- vapply(test_stepmatrix("riself4", c(0.01, 0.1, 0.45), FALSE, FALSE, c(3,1,2,4)),
                  function(a) rowSums(exp(a)), rep(1,4))
    expect_equal(obs, matrix(1, ncol=3, nrow=4))

})


test_that("riself8 step works", {

    for(rf in c(0.01, 0.1, 0.45)) {
        expected <- matrix(rf/2/(1+2*rf), ncol=8, nrow=8)
        diag(expected) <- (1-rf)^2/(1+2*rf)
        block <- list(c(1,2), c(3,4), c(5,6), c(7,8))
        for(i in seq_along(block)) {
            b <- block[[i]]
            expected[b[1],b[2]] <- expected[b[2],b[1]] <- rf*(1-rf)/(1+2*rf)
        }

        result <- matrix(ncol=8, nrow=8)
        for(i in 1:8) {
            for(j in 1:8) {
                result[i,j] <- test_step("riself8", i, j, rf, FALSE, FALSE, 1:8)
            }
        }

        expect_equal(result, log(expected))
    }

    # test with different order of founders
    forder <- c(3, 8, 5, 2, 1, 7, 6, 4)
    forder_rev <- c(5,4,1,8,3,7,6,2)

    for(rf in c(0.01, 0.1, 0.45)) {
        expected <- matrix(rf/2/(1+2*rf), ncol=8, nrow=8)
        diag(expected) <- (1-rf)^2/(1+2*rf)
        block <- list(c(1,2), c(3,4), c(5,6), c(7,8))
        for(i in seq_along(block)) {
            b <- block[[i]]
            expected[b[1],b[2]] <- expected[b[2],b[1]] <- rf*(1-rf)/(1+2*rf)
        }
        expected <- expected[forder_rev, forder_rev]

        result <- matrix(ncol=8, nrow=8)
        for(i in 1:8) {
            for(j in 1:8) {
                result[i,j] <- test_step("riself8", i, j, rf, FALSE, FALSE, forder)
            }
        }

        expect_equal(result, log(expected))
    }


    # test_stepmatrix: output has rows that sum to 1
    obs <- vapply(test_stepmatrix("riself8", c(0.01, 0.1, 0.45), FALSE, FALSE, 1:8),
                  function(a) rowSums(exp(a)), rep(1,8))
    expect_equal(obs, matrix(1, ncol=3, nrow=8))

    obs <- vapply(test_stepmatrix("riself8", c(0.01, 0.1, 0.45), FALSE, FALSE, forder),
                  function(a) rowSums(exp(a)), rep(1,8))
    expect_equal(obs, matrix(1, ncol=3, nrow=8))

})




test_that("riself16 step works", {

    for(rf in c(0.01, 0.1, 0.45)) {
        expected <- matrix(rf/4/(1+2*rf), ncol=16, nrow=16)
        expected[1:4,1:4] <- expected[5:8,5:8] <-
            expected[9:12,9:12] <- expected[13:16,13:16] <- rf*(1-rf)/2/(1+2*rf)
        diag(expected) <- (1-rf)^3/(1+2*rf)
        block <- list(c(1,2), c(3,4), c(5,6), c(7,8),c(9,10),c(11,12),c(13,14),c(15,16))
        for(i in seq_along(block)) {
            b <- block[[i]]
            expected[b[1],b[2]] <- expected[b[2],b[1]] <- rf*(1-rf)^2/(1+2*rf)
        }

        result <- matrix(ncol=16, nrow=16)
        for(i in 1:16) {
            for(j in 1:16) {
                result[i,j] <- test_step("riself16", i, j, rf, FALSE, FALSE, 1:16)
            }
        }

        expect_equal(result, log(expected))
    }

    # test with founders in a different order
    forder <- c(12, 5, 15, 13, 14, 1, 2, 3, 9, 10, 6, 11, 8, 4, 16, 7)
    forder_rev <- c(6,7,8,14,2,11,16,13,9,10,12,1,4,5,3,15)
    for(rf in c(0.01, 0.1, 0.45)) {
        expected <- matrix(rf/4/(1+2*rf), ncol=16, nrow=16)
        expected[1:4,1:4] <- expected[5:8,5:8] <-
            expected[9:12,9:12] <- expected[13:16,13:16] <- rf*(1-rf)/2/(1+2*rf)
        diag(expected) <- (1-rf)^3/(1+2*rf)
        block <- list(c(1,2), c(3,4), c(5,6), c(7,8),c(9,10),c(11,12),c(13,14),c(15,16))
        for(i in seq_along(block)) {
            b <- block[[i]]
            expected[b[1],b[2]] <- expected[b[2],b[1]] <- rf*(1-rf)^2/(1+2*rf)
        }
        expected <- expected[forder_rev, forder_rev]

        result <- matrix(ncol=16, nrow=16)
        for(i in 1:16) {
            for(j in 1:16) {
                result[i,j] <- test_step("riself16", i, j, rf, FALSE, FALSE, forder)
            }
        }

        expect_equal(result, log(expected))
    }


    # test_stepmatrix: output has rows that sum to 1
    obs <- vapply(test_stepmatrix("riself16", c(0.01, 0.1, 0.45), FALSE, FALSE, 1:16),
                  function(a) rowSums(exp(a)), rep(1,16))
    expect_equal(obs, matrix(1, ncol=3, nrow=16))

    # test_stepmatrix: output has rows that sum to 1
    obs <- vapply(test_stepmatrix("riself16", c(0.01, 0.1, 0.45), FALSE, FALSE, forder),
                  function(a) rowSums(exp(a)), rep(1,16))
    expect_equal(obs, matrix(1, ncol=3, nrow=16))

})




test_that("riself4-8-16 geno_names work", {

    expect_equal( geno_names("riself4", LETTERS[5:8], FALSE), paste0(LETTERS[5:8], LETTERS[5:8]) )
    expect_equal( geno_names("riself8", LETTERS[2:9], FALSE), paste0(LETTERS[2:9], LETTERS[2:9]) )
    expect_equal( geno_names("riself16", LETTERS[11:26], FALSE), paste0(LETTERS[11:26], LETTERS[11:26]) )

})

test_that("riself4-8-16 nrec work", {

    x <- matrix(ncol=16, nrow=16)
    x <- matrix(as.numeric(col(x) != row(x)), ncol=16)

    res4 <- matrix(ncol=4, nrow=4)
    res8 <- matrix(ncol=8, nrow=8)
    res16 <- matrix(ncol=16, nrow=16)
    for(i in 1:16) {
        for(j in 1:16) {
            if(i<5 && j<5) res4[i,j] <- test_nrec("riself4", i, j, FALSE, FALSE, 1:4)
            if(i<9 && j<9) res8[i,j] <- test_nrec("riself8", i, j, FALSE, FALSE, 1:8)
            res16[i,j] <- test_nrec("riself16", i, j, FALSE, FALSE, 1:16)
        }
    }

    expect_equal( res4, x[1:4,1:4])
    expect_equal( res8, x[1:8,1:8])
    expect_equal( res16, x)

})

context("basic HMM functions in risib")

test_that("risib nalleles works", {
    expect_equal(nalleles("risib"), 2)
})

test_that("risib check_geno works", {

    # autosome
    expect_true(test_check_geno("risib", 0, TRUE, FALSE, FALSE, 0))
    expect_true(test_check_geno("risib", 1, TRUE, FALSE, FALSE, 0))
    expect_true(test_check_geno("risib", 2, TRUE, FALSE, FALSE, 0))
    expect_true(test_check_geno("risib", 1, FALSE, FALSE, FALSE, 0))
    expect_true(test_check_geno("risib", 2, FALSE, FALSE, FALSE, 0))
    expect_false(test_check_geno("risib", 3, TRUE, FALSE, FALSE, 0))
    expect_false(test_check_geno("risib", 0, FALSE, FALSE, FALSE, 0))
    expect_false(test_check_geno("risib", 3, FALSE, FALSE, FALSE, 0))

    # X chromosome
    expect_true(test_check_geno("risib", 0, TRUE, TRUE, TRUE, 0))
    expect_true(test_check_geno("risib", 1, TRUE, TRUE, TRUE, 0))
    expect_true(test_check_geno("risib", 2, TRUE, TRUE, TRUE, 0))
    expect_true(test_check_geno("risib", 1, FALSE, TRUE, TRUE, 0))
    expect_true(test_check_geno("risib", 2, FALSE, TRUE, TRUE, 0))
    expect_false(test_check_geno("risib", 3, TRUE, TRUE, TRUE, 0))
    expect_false(test_check_geno("risib", 0, FALSE, TRUE, TRUE, 0))
    expect_false(test_check_geno("risib", 3, FALSE, TRUE, TRUE, 0))

    # X chr reverse
    expect_true(test_check_geno("risib", 0, TRUE, TRUE, TRUE, 1))
    expect_true(test_check_geno("risib", 1, TRUE, TRUE, TRUE, 1))
    expect_true(test_check_geno("risib", 2, TRUE, TRUE, TRUE, 1))
    expect_true(test_check_geno("risib", 1, FALSE, TRUE, TRUE, 1))
    expect_true(test_check_geno("risib", 2, FALSE, TRUE, TRUE, 1))
    expect_false(test_check_geno("risib", 3, TRUE, TRUE, TRUE, 1))
    expect_false(test_check_geno("risib", 0, FALSE, TRUE, TRUE, 1))
    expect_false(test_check_geno("risib", 3, FALSE, TRUE, TRUE, 1))

})

test_that("risib ngen works", {

    expect_equal(test_ngen("risib", FALSE), 2)
    expect_equal(test_ngen("risib", TRUE), 2)

})

test_that("risib possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("risib", FALSE, FALSE, 0), 1:2)
    # X chr
    expect_equal(test_possible_gen("risib", TRUE, TRUE, 0), 1:2)
    # X chr reverse
    expect_equal(test_possible_gen("risib", TRUE, TRUE, 1), 1:2)

})

test_that("risib nrec works", {

    # autosome
    expect_equal(test_nrec("risib", 1, 1, FALSE, FALSE, 0), 0)
    expect_equal(test_nrec("risib", 1, 2, FALSE, FALSE, 0), 1)
    expect_equal(test_nrec("risib", 2, 1, FALSE, FALSE, 0), 1)
    expect_equal(test_nrec("risib", 2, 2, FALSE, FALSE, 0), 0)
    # X chr
    expect_equal(test_nrec("risib", 1, 1, TRUE, TRUE, 0), 0)
    expect_equal(test_nrec("risib", 1, 2, TRUE, TRUE, 0), 1)
    expect_equal(test_nrec("risib", 2, 1, TRUE, TRUE, 0), 1)
    expect_equal(test_nrec("risib", 2, 2, TRUE, TRUE, 0), 0)
    # X chr reverse
    expect_equal(test_nrec("risib", 1, 1, TRUE, TRUE, 1), 0)
    expect_equal(test_nrec("risib", 1, 2, TRUE, TRUE, 1), 1)
    expect_equal(test_nrec("risib", 2, 1, TRUE, TRUE, 1), 1)
    expect_equal(test_nrec("risib", 2, 2, TRUE, TRUE, 1), 0)

})

test_that("risib init works", {

    # autosome
    expect_equal(test_init("risib", 1, FALSE, FALSE, 0), log(0.5))
    expect_equal(test_init("risib", 2, FALSE, FALSE, 0), log(0.5))
    # X chr
    expect_equal(test_init("risib", 1, TRUE, TRUE, 0), log(2/3))
    expect_equal(test_init("risib", 2, TRUE, TRUE, 0), log(1/3))
    # X chr reverse
    expect_equal(test_init("risib", 1, TRUE, TRUE, 1), log(1/3))
    expect_equal(test_init("risib", 2, TRUE, TRUE, 1), log(2/3))

})

test_that("risib emit works", {

    # autosome
    eps <- 0.01
    expect_equal(test_emit("risib", 0, 1, eps, integer(0), FALSE, FALSE, 0), 0)
    expect_equal(test_emit("risib", 0, 2, eps, integer(0), FALSE, FALSE, 0), 0)
    expect_equal(test_emit("risib", 1, 1, eps, integer(0), FALSE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("risib", 1, 2, eps, integer(0), FALSE, FALSE, 0), log(eps))
    expect_equal(test_emit("risib", 2, 1, eps, integer(0), FALSE, FALSE, 0), log(eps))
    expect_equal(test_emit("risib", 2, 2, eps, integer(0), FALSE, FALSE, 0), log(1-eps))
    # X chr
    expect_equal(test_emit("risib", 0, 1, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("risib", 0, 2, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("risib", 1, 1, eps, integer(0), TRUE, TRUE, 0), log(1-eps))
    expect_equal(test_emit("risib", 1, 2, eps, integer(0), TRUE, TRUE, 0), log(eps))
    expect_equal(test_emit("risib", 2, 1, eps, integer(0), TRUE, TRUE, 0), log(eps))
    expect_equal(test_emit("risib", 2, 2, eps, integer(0), TRUE, TRUE, 0), log(1-eps))
    # X chr reverse
    expect_equal(test_emit("risib", 0, 1, eps, integer(0), TRUE, FALSE, 1), 0)
    expect_equal(test_emit("risib", 0, 2, eps, integer(0), TRUE, FALSE, 1), 0)
    expect_equal(test_emit("risib", 1, 1, eps, integer(0), TRUE, FALSE, 1), log(1-eps))
    expect_equal(test_emit("risib", 1, 2, eps, integer(0), TRUE, FALSE, 1), log(eps))
    expect_equal(test_emit("risib", 2, 1, eps, integer(0), TRUE, FALSE, 1), log(eps))
    expect_equal(test_emit("risib", 2, 2, eps, integer(0), TRUE, FALSE, 1), log(1-eps))

    # errors
    expect_equal(test_emit("risib", 0, 0, eps, integer(0), FALSE, FALSE, 0), 0)
    expect_equal(test_emit("risib", 0, 3, eps, integer(0), FALSE, FALSE, 0), 0)
    expect_equal(test_emit("risib", 3, 1, eps, integer(0), FALSE, FALSE, 0), 0)
    # X chr
    expect_equal(test_emit("risib", 0, 0, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("risib", 0, 3, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("risib", 3, 1, eps, integer(0), TRUE, TRUE, 0), 0)
    # X chr reverse
    expect_equal(test_emit("risib", 0, 0, eps, integer(0), TRUE, FALSE, 1), 0)
    expect_equal(test_emit("risib", 0, 3, eps, integer(0), TRUE, FALSE, 1), 0)
    expect_equal(test_emit("risib", 3, 1, eps, integer(0), TRUE, FALSE, 1), 0)

})

test_that("risib step works", {

    # autosome
    rf <- 0.01
    RF <- 4*rf/(1+6*rf)
    expect_equal(test_step("risib", 1, 1, rf, FALSE, FALSE, 0), log(1-RF))
    expect_equal(test_step("risib", 1, 2, rf, FALSE, FALSE, 0), log(RF))
    expect_equal(test_step("risib", 2, 1, rf, FALSE, FALSE, 0), log(RF))
    expect_equal(test_step("risib", 2, 2, rf, FALSE, FALSE, 0), log(1-RF))
    # X chr (see table 3 in Broman, Genetics 169:1133-1146, 2005)
    t11 <- (1+2*rf)/(1+4*rf)
    t12 <- 2*rf/(1+4*rf)
    t21 <- 4*rf/(1+4*rf)
    t22 <- 1/(1+4*rf)
    expect_equal(test_step("risib", 1, 1, rf, TRUE, TRUE, 0), log(t11))
    expect_equal(test_step("risib", 1, 2, rf, TRUE, TRUE, 0), log(t12))
    expect_equal(test_step("risib", 2, 1, rf, TRUE, TRUE, 0), log(t21))
    expect_equal(test_step("risib", 2, 2, rf, TRUE, TRUE, 0), log(t22))
    # X chr reverse
    expect_equal(test_step("risib", 1, 1, rf, TRUE, FALSE, 1), log(t22))
    expect_equal(test_step("risib", 1, 2, rf, TRUE, FALSE, 1), log(t21))
    expect_equal(test_step("risib", 2, 1, rf, TRUE, FALSE, 1), log(t12))
    expect_equal(test_step("risib", 2, 2, rf, TRUE, FALSE, 1), log(t11))

})

test_that("geno_names works", {
    expect_equal(geno_names("risib", c("B", "R"), FALSE), c("BB", "RR"))
    expect_equal(geno_names("risib", c("B", "R"), TRUE), c("BB", "RR"))
})

context("basic HMM functions in general AIL")

### TODO none of this has been revised yet

test_that("genail n_gen, n_alleles work", {

    expect_equal(nalleles("genail38"), 38)
    expect_equal(nalleles("genail6"), 6)

    expect_equal(test_ngen("genail38", FALSE), 741)
    expect_equal(test_ngen("genail8", FALSE), 36)
    expect_equal(test_ngen("genail38", TRUE), 779)
    expect_equal(test_ngen("genail8", TRUE), 44)

    # throw error for "genail1"
    expect_error(nalleles("genail1"))

})

test_that("genail possible_gen work", {

    expect_equal(test_possible_gen("genail38", FALSE, FALSE, c(15,rep(1,38))), 1:741)
    expect_equal(test_possible_gen("genail38", TRUE, TRUE, c(15,rep(1,38))), 1:741)
    expect_equal(test_possible_gen("genail38", TRUE, FALSE, c(15,rep(1,38))), 742:779)

})

# FIX_ME: test check_geno

test_that("genail init work", {

    set.seed(20181105)
    alpha <- sample(1:10, 38, replace=TRUE)
    init<- alpha/sum(alpha)

    calcFX <- calcA <- expected <- matrix(nrow=38, ncol=38)
    calcMX <- rep(NA, 38)

    for(i in 1:38) {
        for(j in i:38) {
            g <- mpp_encode_alleles(i, j, 38, FALSE)

            calcA[i,j] <- test_init("genail38", g, FALSE, FALSE, c(8, alpha))
            expected[i,j] <- ifelse(i==j, 2*log(init[i]), log(2) + log(init[i]) + log(init[j]))

            calcFX[i,j] <- test_init("genail38", g, TRUE, TRUE, c(8, alpha))

        }

        calcMX[i] <- test_init("genail38", i+741, TRUE, FALSE, c(8, alpha))
    }

    expect_equal(calcMX, log(init))
    expect_equal(calcA, expected)
    expect_equal(calcFX, expected)

})

# FIX_ME: test emit

test_that("genail step works", {

    skip_on_cran()

    nf <- 7
    ng <- nf + choose(nf,2)
    alpha_int <- sample(1:10, nf, replace=TRUE)
    alpha <- alpha_int/sum(alpha_int)

    g <- t(sapply(1:ng, mpp_decode_geno, nf, FALSE)) # genotypes

    for(rf in c(0.01, 0.1, 0.45)) {
        for(ngen in c(3, 5)) {

            # male X chr
            R <- alpha*(1-(1-rf)^ngen)
            expected <- matrix(rep(alpha*(1-(1-rf)^(ngen*2/3)),nf), ncol=nf, byrow=TRUE)
            diag(expected) <- alpha + (1-alpha)*(1-rf)^(ngen*2/3)

            result <- matrix(ncol=nf, nrow=nf)
            for(i in 1:nf) {
                for(j in 1:nf) {
                    result[i,j] <- test_step(paste0("genail", nf), i+ng, j+ng, rf, TRUE, FALSE, c(ngen, alpha_int))
                }
            }
            # rows sum to 1?
            expect_equal(rowSums(exp(result)), rep(1, nf))

            # matches what I expected?
            expect_equal(result, log(expected))


            # autosome
            result <- matrix(ncol=ng, nrow=ng)
            for(i in 1:ng) {
                for(j in 1:ng) {
                    result[i,j] <- test_step(paste0("genail", nf), i, j, rf, FALSE, FALSE, c(ngen, alpha_int))
                }
            }
            # rows sum to 1?
            expect_equal(rowSums(exp(result)), rep(1, ng))


            # female X chr
            result <- matrix(ncol=ng, nrow=ng)
            for(i in 1:ng) {
                for(j in 1:ng) {
                    result[i,j] <- test_step(paste0("genail", nf), i, j, rf, TRUE, TRUE, c(ngen, alpha_int))
                }
            }
            # rows sum to 1?
            expect_equal(rowSums(exp(result)), rep(1, ng))

        }
    }

})


test_that("genail geno_names work", {

    # if 38 founders, using upper case and lower case letters. Ugh.
    alleles <- c(LETTERS, letters[1:(38-26)])
    expected <- sapply(1:741, function(a) paste(alleles[mpp_decode_geno(a, 38, FALSE)], collapse=""))
    expect_equal( geno_names("genail38", alleles, FALSE), expected)

    hemi <- paste(alleles, "Y", sep="")
    expect_equal( geno_names("genail38", alleles, TRUE), c(expected, hemi))

    # could also use two-letter allele codes, but ugly
    alleles <- c(paste0("A",LETTERS), paste0("B", LETTERS[1:(38-26)]))
    expected <- sapply(1:741, function(a) paste(alleles[mpp_decode_geno(a, 38, FALSE)], collapse=""))
    expect_equal( geno_names("genail38", alleles, FALSE), expected)

    hemi <- paste(alleles, "Y", sep="")
    expect_equal( geno_names("genail38", alleles, TRUE), c(expected, hemi))

})

test_that("genail nrec work", {

    skip_on_cran()

    nf <- 19
    ng <- nf + choose(nf, 2)
    crosstype <- paste0("genail", nf)

    # male X chromosome
    calculated <- matrix(nrow=nf, ncol=nf)
    for(i in 1:nf) {
        for(j in 1:nf) {
            calculated[i,j] <- test_nrec(crosstype, i+ng, j+ng, TRUE, FALSE, c(20, rep(1, nf)))
        }
    }
    expect_equal(calculated, 1-diag(nf))

    calculatedA <- calculatedfX <- expected <- matrix(nrow=ng, ncol=ng)
    g <- sapply(1:ng, function(g) mpp_decode_geno(g, nf, FALSE))
    # female X chromosome or autosome
    for(i in 1:ng) {
        for(j in 1:ng) {

            calculatedA[i,j] <- test_nrec(crosstype, i, j, FALSE, FALSE, c(20, rep(1, nf)))
            calculatedfX[i,j] <- test_nrec(crosstype, i, j, TRUE, TRUE, c(20, rep(1, nf)))

            expected[i,j] <- min(sum(g[,i]!=g[,j]), sum(g[,i]!=rev(g[,j])))
        }
    }
    expect_equal(calculatedA, expected)
    expect_equal(calculatedfX, expected)

})

context("basic HMM functions in intercross")

test_that("intercross nalleles works", {
    expect_equal(nalleles("f2"), 2)
})

test_that("intercross check_geno works", {

    # autosome
    for(i in 0:5)
        expect_true(test_check_geno("f2", i, TRUE, FALSE, FALSE, 0))
    for(i in 1:3)
        expect_true(test_check_geno("f2", i, FALSE, FALSE, FALSE, 0))
    for(i in c(-1, 6))
        expect_false(test_check_geno("f2", i, TRUE, FALSE, FALSE, 0))
    for(i in c(0, 4))
        expect_false(test_check_geno("f2", i, FALSE, FALSE, FALSE, 0))

    # X chromosome female, forward cross
    for(i in 0:5)
        expect_true(test_check_geno("f2", i, TRUE, TRUE, TRUE, 0))
    for(i in 1:2)
        expect_true(test_check_geno("f2", i, FALSE, TRUE, TRUE, 0))
    for(i in c(-1, 6))
        expect_false(test_check_geno("f2", i, TRUE, TRUE, TRUE, 0))
    for(i in c(0, 3, 4))
        expect_false(test_check_geno("f2", i, FALSE, TRUE, TRUE, 0))

    # X chromosome female, reverse cross
    for(i in 0:5)
        expect_true(test_check_geno("f2", i, TRUE, TRUE, TRUE, 1))
    for(i in 3:4)
        expect_true(test_check_geno("f2", i, FALSE, TRUE, TRUE, 1))
    for(i in c(-1, 6))
        expect_false(test_check_geno("f2", i, TRUE, TRUE, TRUE, 1))
    for(i in c(0, 1, 2, 5, 6))
        expect_false(test_check_geno("f2", i, FALSE, TRUE, TRUE, 1))

    # X chromosome male
    for(i in 0:5)
        expect_true(test_check_geno("f2", i, TRUE, TRUE, FALSE, 0))
    for(i in c(5,6))
        expect_true(test_check_geno("f2", i, FALSE, TRUE, FALSE, 0))
    for(i in c(-1, 6))
        expect_false(test_check_geno("f2", i, TRUE, TRUE, FALSE, 0))
    for(i in c(0:4, 7))
        expect_false(test_check_geno("f2", i, FALSE, TRUE, FALSE, 0))

    # X chromosome male reverse cross
    for(i in 0:5)
        expect_true(test_check_geno("f2", i, TRUE, TRUE, FALSE, 1))
    for(i in c(5,6))
        expect_true(test_check_geno("f2", i, FALSE, TRUE, FALSE, 1))
    for(i in c(-1, 6))
        expect_false(test_check_geno("f2", i, TRUE, TRUE, FALSE, 1))
    for(i in c(0:4, 7))
        expect_false(test_check_geno("f2", i, FALSE, TRUE, FALSE, 1))

})

test_that("intercross n_gen works", {

    expect_equal(test_ngen("f2", FALSE), 3)
    expect_equal(test_ngen("f2", TRUE),  6)

})

test_that("intercross possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("f2", FALSE, FALSE, 0), 1:3)
    # X female forward
    expect_equal(test_possible_gen("f2", TRUE, TRUE, 0), 1:2)
    # X female reverse
    expect_equal(test_possible_gen("f2", TRUE, TRUE, 1), 3:4)
    # X male
    expect_equal(test_possible_gen("f2", TRUE, FALSE, 0), 5:6)
    # X male reverse
    expect_equal(test_possible_gen("f2", TRUE, FALSE, 1), 5:6)

})

test_that("intercross init works", {

    # autosome
    expect_equal(test_init("f2", 1, FALSE, FALSE, 0), log(0.25))
    expect_equal(test_init("f2", 2, FALSE, FALSE, 0), log(0.5))
    expect_equal(test_init("f2", 3, FALSE, FALSE, 0), log(0.25))
    # X female forward
    for(i in 1:2)
        expect_equal(test_init("f2", i, TRUE, TRUE, 0), log(0.5))
    # X female reverse
    for(i in 3:4)
        expect_equal(test_init("f2", i, TRUE, TRUE, 1), log(0.5))
    # X male
    for(i in 5:6)
        expect_equal(test_init("f2", i, TRUE, FALSE, 0), log(0.5))
    # X male reverse
    for(i in 5:6)
        expect_equal(test_init("f2", i, TRUE, FALSE, 1), log(0.5))

})

test_that("intercross emit works", {

    # autosome
    eps <- 0.01
    for(i in 1:3)
        expect_equal(test_emit("f2", 0, i, eps, integer(0), FALSE, FALSE, 0), 0)

    expect_equal(test_emit("f2", 1, 1, eps, integer(0), FALSE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2", 2, 1, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
    expect_equal(test_emit("f2", 3, 1, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
    expect_equal(test_emit("f2", 4, 1, eps, integer(0), FALSE, FALSE, 0), log(1-eps/2))
    expect_equal(test_emit("f2", 5, 1, eps, integer(0), FALSE, FALSE, 0), log(eps))

    expect_equal(test_emit("f2", 1, 2, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
    expect_equal(test_emit("f2", 2, 2, eps, integer(0), FALSE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2", 3, 2, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
    expect_equal(test_emit("f2", 4, 2, eps, integer(0), FALSE, FALSE, 0), log(1-eps/2))
    expect_equal(test_emit("f2", 5, 2, eps, integer(0), FALSE, FALSE, 0), log(1-eps/2))

    expect_equal(test_emit("f2", 1, 3, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
    expect_equal(test_emit("f2", 2, 3, eps, integer(0), FALSE, FALSE, 0), log(eps/2))
    expect_equal(test_emit("f2", 3, 3, eps, integer(0), FALSE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2", 4, 3, eps, integer(0), FALSE, FALSE, 0), log(eps))
    expect_equal(test_emit("f2", 5, 3, eps, integer(0), FALSE, FALSE, 0), log(1-eps/2))

    # X female forward
    for(i in 1:2)
        expect_equal(test_emit("f2", 0, i, eps, integer(0), TRUE, TRUE, 0), 0)

    expect_equal(test_emit("f2", 1, 1, eps, integer(0), TRUE, TRUE, 0), log(1-eps))
    expect_equal(test_emit("f2", 2, 1, eps, integer(0), TRUE, TRUE, 0), log(eps))
    expect_equal(test_emit("f2", 3, 1, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("f2", 4, 1, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("f2", 5, 1, eps, integer(0), TRUE, TRUE, 0), log(eps))

    expect_equal(test_emit("f2", 1, 2, eps, integer(0), TRUE, TRUE, 0), log(eps))
    expect_equal(test_emit("f2", 2, 2, eps, integer(0), TRUE, TRUE, 0), log(1-eps))
    expect_equal(test_emit("f2", 3, 2, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("f2", 4, 2, eps, integer(0), TRUE, TRUE, 0), 0)
    expect_equal(test_emit("f2", 5, 2, eps, integer(0), TRUE, TRUE, 0), log(1-eps))

    # X female reverse
    for(i in 3:4)
        expect_equal(test_emit("f2", 0, i, eps, integer(0), TRUE, TRUE, 1), 0)
    expect_equal(test_emit("f2", 1, 4, eps, integer(0), TRUE, TRUE, 1), 0)
    expect_equal(test_emit("f2", 2, 4, eps, integer(0), TRUE, TRUE, 1), log(eps))
    expect_equal(test_emit("f2", 3, 4, eps, integer(0), TRUE, TRUE, 1), log(1-eps))
    expect_equal(test_emit("f2", 4, 4, eps, integer(0), TRUE, TRUE, 1), log(eps))
    expect_equal(test_emit("f2", 5, 4, eps, integer(0), TRUE, TRUE, 1), 0)
    expect_equal(test_emit("f2", 1, 3, eps, integer(0), TRUE, TRUE, 1), 0)
    expect_equal(test_emit("f2", 2, 3, eps, integer(0), TRUE, TRUE, 1), log(1-eps))
    expect_equal(test_emit("f2", 3, 3, eps, integer(0), TRUE, TRUE, 1), log(eps))
    expect_equal(test_emit("f2", 4, 3, eps, integer(0), TRUE, TRUE, 1), log(1-eps))
    expect_equal(test_emit("f2", 5, 3, eps, integer(0), TRUE, TRUE, 1), 0)

    # X male
    for(i in 5:6)
        expect_equal(test_emit("f2", 0, i, eps, integer(0), TRUE, FALSE, 0), 0)
    expect_equal(test_emit("f2", 1, 5, eps, integer(0), TRUE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2", 2, 5, eps, integer(0), TRUE, FALSE, 0), 0)
    expect_equal(test_emit("f2", 3, 5, eps, integer(0), TRUE, FALSE, 0), log(eps))
    expect_equal(test_emit("f2", 4, 5, eps, integer(0), TRUE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2", 5, 5, eps, integer(0), TRUE, FALSE, 0), log(eps))
    expect_equal(test_emit("f2", 1, 6, eps, integer(0), TRUE, FALSE, 0), log(eps))
    expect_equal(test_emit("f2", 2, 6, eps, integer(0), TRUE, FALSE, 0), 0)
    expect_equal(test_emit("f2", 3, 6, eps, integer(0), TRUE, FALSE, 0), log(1-eps))
    expect_equal(test_emit("f2", 4, 6, eps, integer(0), TRUE, FALSE, 0), log(eps))
    expect_equal(test_emit("f2", 5, 6, eps, integer(0), TRUE, FALSE, 0), log(1-eps))

    # X male reverse
    for(i in 5:6)
        expect_equal(test_emit("f2", 0, i, eps, integer(0), TRUE, FALSE, 1), 0)
    expect_equal(test_emit("f2", 1, 5, eps, integer(0), TRUE, FALSE, 1), log(1-eps))
    expect_equal(test_emit("f2", 2, 5, eps, integer(0), TRUE, FALSE, 1), 0)
    expect_equal(test_emit("f2", 3, 5, eps, integer(0), TRUE, FALSE, 1), log(eps))
    expect_equal(test_emit("f2", 4, 5, eps, integer(0), TRUE, FALSE, 1), log(1-eps))
    expect_equal(test_emit("f2", 5, 5, eps, integer(0), TRUE, FALSE, 1), log(eps))
    expect_equal(test_emit("f2", 1, 6, eps, integer(0), TRUE, FALSE, 1), log(eps))
    expect_equal(test_emit("f2", 2, 6, eps, integer(0), TRUE, FALSE, 1), 0)
    expect_equal(test_emit("f2", 3, 6, eps, integer(0), TRUE, FALSE, 1), log(1-eps))
    expect_equal(test_emit("f2", 4, 6, eps, integer(0), TRUE, FALSE, 1), log(eps))
    expect_equal(test_emit("f2", 5, 6, eps, integer(0), TRUE, FALSE, 1), log(1-eps))

})

test_that("intercross step works", {

    # autosome
    rf <- 0.01
    expect_equal(test_step("f2", 1, 1, rf, FALSE, FALSE, 0), log((1-rf)^2))
    expect_equal(test_step("f2", 1, 2, rf, FALSE, FALSE, 0), log(2*rf*(1-rf)))
    expect_equal(test_step("f2", 1, 3, rf, FALSE, FALSE, 0), log(rf^2))
    expect_equal(test_step("f2", 2, 1, rf, FALSE, FALSE, 0), log(rf*(1-rf)))
    expect_equal(test_step("f2", 2, 2, rf, FALSE, FALSE, 0), log((1-rf)^2+rf^2))
    expect_equal(test_step("f2", 2, 3, rf, FALSE, FALSE, 0), log(rf*(1-rf)))
    expect_equal(test_step("f2", 3, 1, rf, FALSE, FALSE, 0), log(rf^2))
    expect_equal(test_step("f2", 3, 2, rf, FALSE, FALSE, 0), log(2*rf*(1-rf)))
    expect_equal(test_step("f2", 3, 3, rf, FALSE, FALSE, 0), log((1-rf)^2))
    # X female
    expect_equal(test_step("f2", 1, 1, rf, TRUE, TRUE, 0), log(1-rf))
    expect_equal(test_step("f2", 1, 2, rf, TRUE, TRUE, 0), log(rf))
    expect_equal(test_step("f2", 2, 1, rf, TRUE, TRUE, 0), log(rf))
    expect_equal(test_step("f2", 2, 2, rf, TRUE, TRUE, 0), log(1-rf))
    # X female reverse
    expect_equal(test_step("f2", 3, 3, rf, TRUE, TRUE, 1), log(1-rf))
    expect_equal(test_step("f2", 3, 4, rf, TRUE, TRUE, 1), log(rf))
    expect_equal(test_step("f2", 4, 3, rf, TRUE, TRUE, 1), log(rf))
    expect_equal(test_step("f2", 4, 4, rf, TRUE, TRUE, 1), log(1-rf))
    # X male
    expect_equal(test_step("f2", 5, 5, rf, TRUE, FALSE, 0), log(1-rf))
    expect_equal(test_step("f2", 5, 6, rf, TRUE, FALSE, 0), log(rf))
    expect_equal(test_step("f2", 6, 5, rf, TRUE, FALSE, 0), log(rf))
    expect_equal(test_step("f2", 6, 6, rf, TRUE, FALSE, 0), log(1-rf))
    # X male reverse
    expect_equal(test_step("f2", 5, 5, rf, TRUE, FALSE, 1), log(1-rf))
    expect_equal(test_step("f2", 5, 6, rf, TRUE, FALSE, 1), log(rf))
    expect_equal(test_step("f2", 6, 5, rf, TRUE, FALSE, 1), log(rf))
    expect_equal(test_step("f2", 6, 6, rf, TRUE, FALSE, 1), log(1-rf))

})

test_that("geno_names works", {
    expect_equal(geno_names("f2", c("B", "R"), FALSE), c("BB", "BR", "RR"))
    expect_equal(geno_names("f2", c("B", "R"), TRUE), c("BB", "BR", "RB", "RR", "BY", "RY"))
})

test_that("nrec works", {

    # autosome genotypes = 1:3
    expected <- rbind(c(0,1,2), c(1,0,1), c(2,1,0))
    for(i in 1:3)
        for(j in 1:3)
            expect_equal(test_nrec("f2", i, j, FALSE, FALSE, 0), expected[i,j])

    # X chromosome female, forward
    expected <- rbind(c(0,1), c(1,0))
    for(i in 1:2)
        for(j in 1:2)
            expect_equal(test_nrec("f2", i, j, TRUE, TRUE, 0), expected[i,j])

    # X chromosome female, reverse
    expected <- rbind(c(0,1), c(1,0))
    for(i in 1:2)
        for(j in 1:2)
            expect_equal(test_nrec("f2", i+2, j+2, TRUE, TRUE, 1), expected[i,j])

    # X chromosome male
    expected <- rbind(c(0,1), c(1,0))
    for(i in 1:2)
        for(j in 1:2)
            expect_equal(test_nrec("f2", i+4, j+4, TRUE, FALSE, 0), expected[i,j])

    # X chromosome male
    expected <- rbind(c(0,1), c(1,0))
    for(i in 1:2)
        for(j in 1:2)
            expect_equal(test_nrec("f2", i+4, j+4, TRUE, FALSE, 1), expected[i,j])

})

context("basic HMM functions in AIL")

test_that("AIL nalleles works", {
    expect_equal(nalleles("ail"), 2)
})

test_that("AIL check_geno works", {

    # autosome
    for(i in 0:5)
        expect_true(test_check_geno("ail", i, TRUE, FALSE, FALSE, c(20, 0)))
    for(i in 1:3)
        expect_true(test_check_geno("ail", i, FALSE, FALSE, FALSE, c(20, 0)))
    for(i in c(-1, 6))
        expect_false(test_check_geno("ail", i, TRUE, FALSE, FALSE, c(20, 0)))
    for(i in c(0, 4))
        expect_false(test_check_geno("ail", i, FALSE, FALSE, FALSE, c(20, 0)))

    for(dir in 0:2) {
        # X chromosome female
        for(i in 0:5)
            expect_true(test_check_geno("ail", i, TRUE, TRUE, TRUE, c(20, dir)))
        for(i in 1:3)
            expect_true(test_check_geno("ail", i, FALSE, TRUE, TRUE, c(20, dir)))
        for(i in c(-1, 6))
            expect_false(test_check_geno("ail", i, TRUE, TRUE, TRUE, c(20, dir)))
        for(i in c(0, 4:6))
            expect_false(test_check_geno("ail", i, FALSE, TRUE, TRUE, c(20, dir)))

        # X chromosome male
        for(i in 0:5)
            expect_true(test_check_geno("ail", i, TRUE, TRUE, FALSE, c(20, dir)))
        for(i in c(4,5))
            expect_true(test_check_geno("ail", i, FALSE, TRUE, FALSE, c(20, dir)))
        for(i in c(-1, 6))
            expect_false(test_check_geno("ail", i, TRUE, TRUE, FALSE, c(20, dir)))
        for(i in c(0:3, 6))
            expect_false(test_check_geno("ail", i, FALSE, TRUE, FALSE, c(20, dir)))
    }

})

test_that("AIL n_gen works", {

    expect_equal(test_ngen("ail", FALSE), 3)
    expect_equal(test_ngen("ail", TRUE),  5)

})

test_that("AIL possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("ail", FALSE, FALSE, c(20,0)), 1:3)

    for(dir in 0:2) {
        # X female
        expect_equal(test_possible_gen("ail", TRUE, TRUE, c(20,dir)), 1:3)
        # X male
        expect_equal(test_possible_gen("ail", TRUE, FALSE, c(20,dir)), 4:5)
    }

})
test_that("AIL init works", {

    for(n_gen in c(3,9,12,15)) {
        # autosome
        expect_equal(test_init("ail", 1, FALSE, FALSE, c(n_gen, 0)), log(0.25))
        expect_equal(test_init("ail", 2, FALSE, FALSE, c(n_gen, 0)), log(0.5))
        expect_equal(test_init("ail", 3, FALSE, FALSE, c(n_gen, 0)), log(0.25))

        # X chr balanced
        #     female
        expect_equal(test_init("ail", 1, TRUE, TRUE, c(n_gen, 2)), log(0.25))
        expect_equal(test_init("ail", 2, TRUE, TRUE, c(n_gen, 2)), log(0.5))
        expect_equal(test_init("ail", 3, TRUE, TRUE, c(n_gen, 2)), log(0.25))

        #     male
        expect_equal(test_init("ail", 4, TRUE, FALSE, c(n_gen, 2)), log(0.5))
        expect_equal(test_init("ail", 5, TRUE, FALSE, c(n_gen, 2)), log(0.5))
    }

    for(n_gen in c(3,9,12,15)) {
        fprob <- (2/3) + (1/3)*(-1/2)^n_gen
        mprob <- (2/3) + (1/3)*(-1/2)^(n_gen-1)

        # X chr, AxB
        #     female
        expect_equal(test_init("ail", 1, TRUE, TRUE, c(n_gen, 0)), log(fprob^2))
        expect_equal(test_init("ail", 2, TRUE, TRUE, c(n_gen, 0)), log(2*fprob*(1-fprob)))
        expect_equal(test_init("ail", 3, TRUE, TRUE, c(n_gen, 0)), log((1-fprob)^2))

        #     male
        expect_equal(test_init("ail", 4, TRUE, FALSE, c(n_gen, 0)), log(mprob))
        expect_equal(test_init("ail", 5, TRUE, FALSE, c(n_gen, 0)), log(1-mprob))

        # X chr, BxA
        #     female
        expect_equal(test_init("ail", 3, TRUE, TRUE, c(n_gen, 1)), log(fprob^2))
        expect_equal(test_init("ail", 2, TRUE, TRUE, c(n_gen, 1)), log(2*fprob*(1-fprob)))
        expect_equal(test_init("ail", 1, TRUE, TRUE, c(n_gen, 1)), log((1-fprob)^2))

        #     male
        expect_equal(test_init("ail", 5, TRUE, FALSE, c(n_gen, 1)), log(mprob))
        expect_equal(test_init("ail", 4, TRUE, FALSE, c(n_gen, 1)), log(1-mprob))
    }

})


test_that("AIL emit works", {

    # autosome
    eps <- 0.01
    for(i in 1:3)
        expect_equal(test_emit("ail", 0, i, eps, integer(0), FALSE, FALSE, c(20,0)), 0)

    expect_equal(test_emit("ail", 1, 1, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ail", 2, 1, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 3, 1, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 4, 1, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps/2))
    expect_equal(test_emit("ail", 5, 1, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps))

    expect_equal(test_emit("ail", 1, 2, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 2, 2, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ail", 3, 2, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 4, 2, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps/2))
    expect_equal(test_emit("ail", 5, 2, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps/2))

    expect_equal(test_emit("ail", 1, 3, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 2, 3, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 3, 3, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ail", 4, 3, eps, integer(0), FALSE, FALSE, c(20,0)), log(eps))
    expect_equal(test_emit("ail", 5, 3, eps, integer(0), FALSE, FALSE, c(20,0)), log(1-eps/2))


    # X female
    for(i in 1:3)
        expect_equal(test_emit("ail", 0, i, eps, integer(0), TRUE, TRUE, c(20,0)), 0)

    expect_equal(test_emit("ail", 1, 1, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ail", 2, 1, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 3, 1, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 4, 1, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps/2))
    expect_equal(test_emit("ail", 5, 1, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps))

    expect_equal(test_emit("ail", 1, 2, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 2, 2, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ail", 3, 2, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 4, 2, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps/2))
    expect_equal(test_emit("ail", 5, 2, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps/2))

    expect_equal(test_emit("ail", 1, 3, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 2, 3, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps/2))
    expect_equal(test_emit("ail", 3, 3, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ail", 4, 3, eps, integer(0), TRUE, TRUE, c(20,0)), log(eps))
    expect_equal(test_emit("ail", 5, 3, eps, integer(0), TRUE, TRUE, c(20,0)), log(1-eps/2))

    # X male
    for(i in 4:5)
        expect_equal(test_emit("ail", 0, i, eps, integer(0), TRUE, FALSE, c(20,0)), 0)
    expect_equal(test_emit("ail", 1, 4, eps, integer(0), TRUE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ail", 2, 4, eps, integer(0), TRUE, FALSE, c(20,0)), 0)
    expect_equal(test_emit("ail", 3, 4, eps, integer(0), TRUE, FALSE, c(20,0)), log(eps))
    expect_equal(test_emit("ail", 4, 4, eps, integer(0), TRUE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ail", 5, 4, eps, integer(0), TRUE, FALSE, c(20,0)), log(eps))
    expect_equal(test_emit("ail", 1, 5, eps, integer(0), TRUE, FALSE, c(20,0)), log(eps))
    expect_equal(test_emit("ail", 2, 5, eps, integer(0), TRUE, FALSE, c(20,0)), 0)
    expect_equal(test_emit("ail", 3, 5, eps, integer(0), TRUE, FALSE, c(20,0)), log(1-eps))
    expect_equal(test_emit("ail", 4, 5, eps, integer(0), TRUE, FALSE, c(20,0)), log(eps))
    expect_equal(test_emit("ail", 5, 5, eps, integer(0), TRUE, FALSE, c(20,0)), log(1-eps))

})

test_that("AIL step works", {

    # autosome
    for(rf in c(0.01, 0.0001)) {
        for(ngen in c(3, 9, 12, 15)) {
            R <- 1 - 2*( 0.25*(1 + (1-2*rf)*(1-rf)^(ngen-2)))

            expect_equal(test_step("ail", 1, 1, rf, FALSE, FALSE, c(ngen, 0)), log((1-R)^2))
            expect_equal(test_step("ail", 1, 2, rf, FALSE, FALSE, c(ngen, 0)), log(2*R*(1-R)))
            expect_equal(test_step("ail", 1, 3, rf, FALSE, FALSE, c(ngen, 0)), log(R^2))
            expect_equal(test_step("ail", 2, 1, rf, FALSE, FALSE, c(ngen, 0)), log(R*(1-R)))
            expect_equal(test_step("ail", 2, 2, rf, FALSE, FALSE, c(ngen, 0)), log((1-R)^2+R^2))
            expect_equal(test_step("ail", 2, 3, rf, FALSE, FALSE, c(ngen, 0)), log(R*(1-R)))
            expect_equal(test_step("ail", 3, 1, rf, FALSE, FALSE, c(ngen, 0)), log(R^2))
            expect_equal(test_step("ail", 3, 2, rf, FALSE, FALSE, c(ngen, 0)), log(2*R*(1-R)))
            expect_equal(test_step("ail", 3, 3, rf, FALSE, FALSE, c(ngen, 0)), log((1-R)^2))
        }
    }

    # X chromosome, balanced
    for(rf in c(0.01, 0.0001)) {
        for(ngen in c(3, 9, 12, 15)) {
            z <- sqrt((1-rf)*(9-rf))
            w <- (1 - rf + z)/4
            y <- (1 - rf - z)/4
            mR <- 1 - 0.25*(2 + (1-2*rf)*(w^(ngen-2) + y^(ngen-2)) +
                            (3 - 5*rf + 2*rf^2)/z*(w^(ngen-2) - y^(ngen-2)))
            fR <- 1 - 0.25*(2 + (1-2*rf)*(w^(ngen-2) + y^(ngen-2)) +
                            (3 - 6*rf + rf^2)/z*(w^(ngen-2) - y^(ngen-2)))

            expect_equal(test_step("ail", 1, 1, rf, TRUE, TRUE, c(ngen, 2)), log((1-fR)^2))
            expect_equal(test_step("ail", 1, 2, rf, TRUE, TRUE, c(ngen, 2)), log(2*fR*(1-fR)))
            expect_equal(test_step("ail", 1, 3, rf, TRUE, TRUE, c(ngen, 2)), log(fR^2))
            expect_equal(test_step("ail", 2, 1, rf, TRUE, TRUE, c(ngen, 2)), log(fR*(1-fR)))
            expect_equal(test_step("ail", 2, 2, rf, TRUE, TRUE, c(ngen, 2)), log((1-fR)^2+fR^2))
            expect_equal(test_step("ail", 2, 3, rf, TRUE, TRUE, c(ngen, 2)), log(fR*(1-fR)))
            expect_equal(test_step("ail", 3, 1, rf, TRUE, TRUE, c(ngen, 2)), log(fR^2))
            expect_equal(test_step("ail", 3, 2, rf, TRUE, TRUE, c(ngen, 2)), log(2*fR*(1-fR)))
            expect_equal(test_step("ail", 3, 3, rf, TRUE, TRUE, c(ngen, 2)), log((1-fR)^2))

            expect_equal(test_step("ail", 4, 4, rf, TRUE, FALSE, c(ngen, 2)), log(1-mR))
            expect_equal(test_step("ail", 4, 5, rf, TRUE, FALSE, c(ngen, 2)), log(mR))
            expect_equal(test_step("ail", 5, 4, rf, TRUE, FALSE, c(ngen, 2)), log(mR))
            expect_equal(test_step("ail", 5, 5, rf, TRUE, FALSE, c(ngen, 2)), log(1-mR))

        }
    }


    # calculate Pr(A) for females, AxB
    calc_q <- function(n_gen)
        2/3 + (1/3)*(-1/2)^n_gen

    # calc Pr(AA haplotype) in males and females, recursively
    calc_p11 <- function(n_gen, rf) {
        if(n_gen == 1) return(c(1, 0.5))
        last <- calc_p11(n_gen-1, rf)
        z <- c((1-rf)*last[2] + rf*calc_q(n_gen-2)*calc_q(n_gen-3),
               0.5*last[1] + (1-rf)/2*last[2] + (rf/2)*calc_q(n_gen-2)*calc_q(n_gen-3))
        z
    }

    # X chromosome, AxB and BxA
    for(rf in c(0.01, 0.0001)) {
        for(ngen in c(3, 9, 12, 15)) {

            qf <- calc_q(ngen)
            qm <- calc_q(ngen-1) # Pr(A) in males is Pr(A) in females for prev generation
            p11 <- calc_p11(ngen, rf) # Pr(AA haplotype) in males and females
            m11 <- p11[1]
            f11 <- p11[2]

            # turn into conditional probabilities along one haplotype
            m1to1 <- m11/qm
            m1to2 <- 1 - m1to1
            m2to1 <- (qm - m11)/(1-qm)
            m2to2 <- 1 - m2to1

            f1to1 <- f11/qf
            f1to2 <- 1 - f1to1
            f2to1 <- (qf - f11)/(1-qf)
            f2to2 <- 1 - f2to1

            expect_equal(test_step("ail", 1, 1, rf, TRUE, TRUE, c(ngen, 0)), log(f1to1^2))
            expect_equal(test_step("ail", 1, 2, rf, TRUE, TRUE, c(ngen, 0)), log(2*f1to1*f1to2))
            expect_equal(test_step("ail", 1, 3, rf, TRUE, TRUE, c(ngen, 0)), log(f1to2^2))
            expect_equal(test_step("ail", 2, 1, rf, TRUE, TRUE, c(ngen, 0)), log(f1to1*f2to1))
            expect_equal(test_step("ail", 2, 2, rf, TRUE, TRUE, c(ngen, 0)), log(f2to1*f1to2 + f1to1*f2to2))
            expect_equal(test_step("ail", 2, 3, rf, TRUE, TRUE, c(ngen, 0)), log(f1to2*f2to2))
            expect_equal(test_step("ail", 3, 1, rf, TRUE, TRUE, c(ngen, 0)), log(f2to1^2))
            expect_equal(test_step("ail", 3, 2, rf, TRUE, TRUE, c(ngen, 0)), log(2*f2to2*f2to1))
            expect_equal(test_step("ail", 3, 3, rf, TRUE, TRUE, c(ngen, 0)), log(f2to2^2))

            expect_equal(test_step("ail", 4, 4, rf, TRUE, FALSE, c(ngen, 0)), log(m1to1))
            expect_equal(test_step("ail", 4, 5, rf, TRUE, FALSE, c(ngen, 0)), log(m1to2))
            expect_equal(test_step("ail", 5, 4, rf, TRUE, FALSE, c(ngen, 0)), log(m2to1))
            expect_equal(test_step("ail", 5, 5, rf, TRUE, FALSE, c(ngen, 0)), log(m2to2))

            expect_equal(test_step("ail", 1, 1, rf, TRUE, TRUE, c(ngen, 1)), log(f2to2^2))
            expect_equal(test_step("ail", 1, 2, rf, TRUE, TRUE, c(ngen, 1)), log(2*f2to2*f2to1))
            expect_equal(test_step("ail", 1, 3, rf, TRUE, TRUE, c(ngen, 1)), log(f2to1^2))
            expect_equal(test_step("ail", 2, 1, rf, TRUE, TRUE, c(ngen, 1)), log(f2to2*f1to2))
            expect_equal(test_step("ail", 2, 2, rf, TRUE, TRUE, c(ngen, 1)), log(f1to2*f2to1 + f1to1*f2to2))
            expect_equal(test_step("ail", 2, 3, rf, TRUE, TRUE, c(ngen, 1)), log(f2to1*f1to1))
            expect_equal(test_step("ail", 3, 1, rf, TRUE, TRUE, c(ngen, 1)), log(f1to2^2))
            expect_equal(test_step("ail", 3, 2, rf, TRUE, TRUE, c(ngen, 1)), log(2*f1to1*f1to2))
            expect_equal(test_step("ail", 3, 3, rf, TRUE, TRUE, c(ngen, 1)), log(f1to1^2))

            expect_equal(test_step("ail", 4, 4, rf, TRUE, FALSE, c(ngen, 1)), log(m2to2))
            expect_equal(test_step("ail", 4, 5, rf, TRUE, FALSE, c(ngen, 1)), log(m2to1))
            expect_equal(test_step("ail", 5, 4, rf, TRUE, FALSE, c(ngen, 1)), log(m1to2))
            expect_equal(test_step("ail", 5, 5, rf, TRUE, FALSE, c(ngen, 1)), log(m1to1))

        }
    }


})

test_that("geno_names works", {
    expect_equal(geno_names("ail", c("B", "R"), FALSE), c("BB", "BR", "RR"))
    expect_equal(geno_names("ail", c("B", "R"), TRUE), c("BB", "BR", "RR", "BY", "RY"))
})


test_that("nrec works", {

    # autosome genotypes = 1:3
    expected <- rbind(c(0,1,2), c(1,0,1), c(2,1,0))
    for(i in 1:3)
        for(j in 1:3) {
            expect_equal(test_nrec("ail", i, j, FALSE, FALSE, 0), expected[i,j])
            expect_equal(test_nrec("ail", i, j, TRUE, TRUE, 0), expected[i,j])
            expect_equal(test_nrec("ail", i, j, TRUE, TRUE, 0), expected[i,j])
        }

    # X chromosome male
    expected <- rbind(c(0,1), c(1,0))
    for(i in 1:2)
        for(j in 1:2)
            expect_equal(test_nrec("ail", i+3, j+3, TRUE, FALSE, 0), expected[i,j])

})

context("basic HMM functions in Heterogeneous Stock F1 (crossed to a 9th inbred strain)")

test_that("HSF1 nalleles works", {
    expect_equal(nalleles("hsf1"), 8)
})

test_that("HSF1 check_geno works", {

    # observed genotypes
    for(i in 0:5) {
        # Autosome
        expect_true(test_check_geno("hsf1", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("hsf1", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_true(test_check_geno("hsf1", i, TRUE, TRUE, FALSE, 20))
    }
    for(i in c(-1, 6)) {
        # Autosome
        expect_false(test_check_geno("hsf1", i, TRUE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("hsf1", i, TRUE, TRUE, TRUE, 20))
        # Male X
        expect_false(test_check_geno("hsf1", i, TRUE, TRUE, FALSE, 20))
    }

    # true genotypes, autosome and female X
    for(i in 1:8) {
        # Autosome
        expect_true(test_check_geno("hsf1", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_true(test_check_geno("hsf1", i, FALSE, TRUE, TRUE, 20))
    }
    for(i in c(0, 9)) {
        # Autosome
        expect_false(test_check_geno("hsf1", i, FALSE, FALSE, FALSE, 20))
        # Female X
        expect_false(test_check_geno("hsf1", i, FALSE, TRUE, TRUE, 20))
    }

    # true genotypes, male X
    for(i in c(0, 8 + 1:8)) {
        expect_false(test_check_geno("hsf1", i, FALSE, TRUE, FALSE, 20))
    }
    for(i in 1:8) {
        expect_true(test_check_geno("hsf1", i, FALSE, TRUE, FALSE, 20))
    }

})

test_that("HSF1 n_gen works", {

    expect_equal(test_ngen("hsf1", FALSE), 8)
    expect_equal(test_ngen("hsf1", TRUE),  8)

})

test_that("HSF1 possible_gen works", {

    # autosome
    expect_equal(test_possible_gen("hsf1", FALSE, FALSE, 20), 1:8)

    # X female
    expect_equal(test_possible_gen("hsf1", TRUE, TRUE, 20), 1:8)

    # X male
    expect_equal(test_possible_gen("hsf1", TRUE, FALSE, 20), 1:8)

})

test_that("HSF1 init works", {

    hom <- 1:8
    male <- 8 + (1:8)

    for(i in hom) {
        # autosome and female X
        expect_equal(test_init("hsf1", i, FALSE, FALSE, 20), log(1/8))
        expect_equal(test_init("hsf1", i, TRUE,  TRUE,  20), log(1/8))
    }

    for(i in male)
        expect_equal(test_init("hsf1", i, TRUE,  FALSE, 20), log(1/8))

})


test_that("HSF1 emit works", {

    fgen <- c(1,3,0,1,3,0,1,3,1) # founder genotypes; 0=missing, 1=AA, 3=BB
    err <- 0.01

    expected <- log(c(1-err, err/2, err/2, 1-err/2, err))
    for(trueg in c(1,4,7)) { # A, D, G
        for(obsg in 1:5) {
            expect_equal(test_emit("hsf1", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("hsf1", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
            expect_equal(test_emit("hsf1", obsg, trueg, err, fgen, TRUE, FALSE, 20), expected[obsg])
        }
    }
    expected <- log(c(err/2, 1-err, err/2, 1-err/2, 1-err/2))
    expected_male <- log(c(err/2, err/2, 1-err, err, 1-err/2))
    for(trueg in c(2,5,8)) { # B, E, H
        for(obsg in 1:5) {
            expect_equal(test_emit("hsf1", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("hsf1", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
            expect_equal(test_emit("hsf1", obsg, trueg, err, fgen, TRUE, FALSE, 20), expected_male[obsg])
        }
    }
    expected <- log(c(1-err, 1, err, 1-err, err))
    expected_male <- log(c(1, 1, 1, 1, 1))
    for(trueg in c(3,6)) {
        for(obsg in 1:5) {
            expect_equal(test_emit("hsf1", obsg, trueg, err, fgen, FALSE, FALSE, 20), expected[obsg])
            expect_equal(test_emit("hsf1", obsg, trueg, err, fgen,  TRUE,  TRUE, 20), expected[obsg])
            expect_equal(test_emit("hsf1", obsg, trueg, err, fgen, TRUE, FALSE, 20), expected_male[obsg])
        }
    }

})

test_that("HSF1 step works", {

    ng <- 8
    trmat <- matrix(nrow=ng, ncol=ng)
    # autosome
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("hsf1", gl, gr, rf, FALSE, FALSE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))

            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

    # female X
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("hsf1", gl, gr, rf, TRUE, TRUE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))

            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

    # male X
    ng <- 8
    trmat <- matrix(nrow=ng, ncol=ng)
    for(rf in c(0.01, 0.001, 0.0001)) {
        for(ngen in c(6, 12, 50)) {

            for(gl in 1:ng)
                for(gr in 1:ng)
                    trmat[gl,gr] <- test_step("hsf1", 8+gl, 8+gr, rf, TRUE, FALSE, ngen)

            # no missing values
            expect_true(all(!is.na(trmat)))
            # all in (-Inf, 0]
            expect_true(all(trmat > -Inf & trmat <= 0))
            # rows sum to 1
            expect_equal( rowSums(exp(trmat)), rep(1, ng))

            # maximum value on the diagonal
            expect_equal( apply(trmat, 1, which.max), 1:ng)
        }
    }

})

test_that("geno_names works", {
    auto <- LETTERS[1:8]

    expect_equal(geno_names("hsf1", LETTERS[1:8], FALSE), auto)
    expect_equal(geno_names("hsf1", LETTERS[1:8], TRUE), auto)
})

context("basic HMM functions in doubled haploids")

test_that("doubled haploids nalleles works", {
    expect_equal(nalleles("dh"), 2)
})

test_that("doubled haploids check_geno works", {

    # autosome
    expect_true(test_check_geno("dh", 0, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("dh", 1, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("dh", 2, TRUE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("dh", 1, FALSE, FALSE, FALSE, integer(0)))
    expect_true(test_check_geno("dh", 2, FALSE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("dh", 3, TRUE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("dh", 0, FALSE, FALSE, FALSE, integer(0)))
    expect_false(test_check_geno("dh", 3, FALSE, FALSE, FALSE, integer(0)))
})

test_that("doubled haploids n_gen works", {

    expect_equal(test_ngen("dh", FALSE), 2)

})

test_that("doubled haploids possible_gen works", {

    expect_equal(test_possible_gen("dh", FALSE, FALSE, integer(0)), 1:2)
})

test_that("doubled haploids nrec works", {

    # autosome
    expect_equal(test_nrec("dh", 1, 1, FALSE, FALSE, integer(0)), 0)
    expect_equal(test_nrec("dh", 1, 2, FALSE, FALSE, integer(0)), 1)
    expect_equal(test_nrec("dh", 2, 1, FALSE, FALSE, integer(0)), 1)
    expect_equal(test_nrec("dh", 2, 2, FALSE, FALSE, integer(0)), 0)

})

test_that("doubled haploids init works", {

    expect_equal(test_init("dh", 1, FALSE, FALSE, integer(0)), log(0.5))
    expect_equal(test_init("dh", 2, FALSE, FALSE, integer(0)), log(0.5))

})

test_that("doubled haploids emit works", {

    eps <- 0.01
    expect_equal(test_emit("dh", 0, 1, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("dh", 0, 2, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("dh", 1, 1, eps, integer(0), FALSE, FALSE, integer(0)), log(1-eps))
    expect_equal(test_emit("dh", 1, 2, eps, integer(0), FALSE, FALSE, integer(0)), log(eps))
    expect_equal(test_emit("dh", 2, 1, eps, integer(0), FALSE, FALSE, integer(0)), log(eps))
    expect_equal(test_emit("dh", 2, 2, eps, integer(0), FALSE, FALSE, integer(0)), log(1-eps))

    # errors
    expect_equal(test_emit("dh", 0, 0, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("dh", 0, 3, eps, integer(0), FALSE, FALSE, integer(0)), 0)
    expect_equal(test_emit("dh", 3, 1, eps, integer(0), FALSE, FALSE, integer(0)), 0)

})

test_that("doubled haploids step works", {

    rf <- 0.01
    expect_equal(test_step("dh", 1, 1, rf, FALSE, FALSE, integer(0)), log(1-rf))
    expect_equal(test_step("dh", 1, 2, rf, FALSE, FALSE, integer(0)), log(rf))
    expect_equal(test_step("dh", 2, 1, rf, FALSE, FALSE, integer(0)), log(rf))
    expect_equal(test_step("dh", 2, 2, rf, FALSE, FALSE, integer(0)), log(1-rf))

})

test_that("geno_names works", {
    expect_equal(geno_names("dh", c("B", "R"), FALSE), c("BB", "RR"))
    expect_equal(geno_names("dh", c("B", "R"), TRUE), c("BB", "RR"))
})

#' Join genome scan results for different chromosomes.
#'
#' Join multiple [scan1()] results for different
#' chromosomes; must have the same set of lod score column.
#'
#' @param ... Genome scan objects of class `"scan1"`, as produced by [scan1()].
#' Must have the same lod score columns.
#'
#' @return An object of class `"scan1", like the inputs, but with
#' the results for different sets of chromosomes combined.
#'
#' @details If components `addcovar`, `Xcovar`,
#' `intcovar`, `weights`, `sample_size` do not match
#' between objects, we omit this information.
#'
#' If `hsq` present, we simply `rbind()` the contents.
#'
#' @seealso [cbind.scan1()], [scan1()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' probs <- calc_genoprob(grav2, map, error_prob=0.002)
#' phe <- grav2$pheno[,1,drop=FALSE]
#'
#' out1 <- scan1(probs[,1], phe) # chr 1
#' out2 <- scan1(probs[,5], phe) # chr 5
#' out <- rbind(out1, out2)
#'
#' @export
rbind.scan1 <-
    function(...)
{
    args <- list(...)
    if(length(args) == 1) return(result)


    # to rbind: main data, attributes SE, hsq
    # drop if not matching: sample_size

    # grab attributes
    args_attr <- lapply(args, attributes)

    result <- unclass(args[[1]])

    # cbind the main data
    for(i in 2:length(args)) {
        if(!is_same(colnames(result), colnames(args[[i]])))
           stop("Input objects 1 and ", i, " have different columns.")
           result <- rbind(result, unclass(args[[i]]))
    }

    # rbind attributes
    to_rbind <- c("hsq", "SE")
    for(obj in to_rbind) {
        drop_arg <- FALSE
        for(i in 2:length(args)) {
            if(is.null(args_attr[[1]][[obj]]) && is.null(args_attr[[i]][[obj]])) next # not present
            if(is.null(args_attr[[1]][[obj]]) || is.null(args_attr[[i]][[obj]])) {
                drop_arg <- TRUE
            } else {
                args_attr[[1]][[obj]] <- rbind(args_attr[[1]][[obj]], args_attr[[i]][[obj]])
            }
        }
        if(drop_arg) {
            warning(obj, " not present in all inputs")
            arg_attr[[1]][[obj]] <- NULL
        }
    }

    # drop if not matching
    other_stuff <- c("sample_size")
    for(i in 2:length(args)) {
        for(obj in other_stuff) {
            if(!is_same(args_attr[[1]][[obj]], args_attr[[i]][[obj]]))
                args_attr[[1]][[obj]] <- NULL
        }
    }

    # copy attributes
    for(obj in c("sample_size", "hsq", "SE"))
        attr(result, obj) <- args_attr[[1]][[obj]]
    class(result) <- class(args[[1]])

    result
}

# subset calc_genoprob objects

#' Subsetting genotype probabilities
#'
#' Pull out a specified set of individuals and/or chromosomes from
#' the results of [calc_genoprob()].
#'
#' @param x Genotype probabilities as output from [calc_genoprob()].
#' @param ind A vector of individuals: numeric indices, logical
#' values, or character string IDs
#' @param chr A vector of chromosomes: logical values, or character
#' string IDs. Numbers are interpreted as character string IDs.
#' @param ... Ignored.
#'
#' @return An object of class `"calc_genoprob"`, like the input, with the selected
#' individuals and/or chromsomes; see [calc_genoprob()].
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' \dontshow{grav2 <- grav2[1:8,c(1,2)]}
#' pr <- calc_genoprob(grav2)
#' # keep just individuals 1:5, chromosome 2
#' prsub <- pr[1:5,2]
#' # keep just chromosome 2
#' prsub2 <- pr[,2]

subset.calc_genoprob <-
    function(x, ind=NULL, chr=NULL, ...)
{
    if(is.null(ind) && is.null(chr))
        stop("You must specify either ind or chr.")

    if(!is.null(chr)) {
        chr <- subset_chr(chr, names(x))

        if(length(chr) == 0)
            stop("Must retain at least one chromosome.")

        cl <- class(x)
        class(x) <- "list"

        attr_to_sub <- c("is_x_chr")
        attr_to_keep <- c("crosstype", "alleles", "alleleprobs")
        x_attr <- attributes(x)
        x_attrnam <- names(x_attr)
        x <- x[chr]
        for(a in attr_to_sub) {
            if(a %in% x_attrnam) {
                attr(x, a) <- x_attr[[a]][chr]
            }
        }
        for(a in attr_to_keep) {
            if(a %in% attr_to_keep)
                attr(x, a) <- x_attr[[a]]
        }
        class(x) <- cl
    }

    if(!is.null(ind)) {
        if(inherits(x, "calc_genoprob"))
            all_ind <- dimnames(x)[[1]]
        else all_ind <- rownames(x[[1]])

        ind <- subset_ind(ind, all_ind)

        if(length(ind) == 0)
            stop("Must retain at least one individual.")

        cl <- class(x)
        class(x) <- "list"

        for(i in names(x)) # loop over chromosomes
            x[[i]] <- x[[i]][ind,,,drop=FALSE]

        class(x) <- cl
    }

    x
}

#' @export
#' @rdname subset.calc_genoprob
`[.calc_genoprob` <-
    function(x, ind=NULL, chr=NULL)
    subset(x, ind, chr)

#' Plot multiple individuals' genome-wide genotypes
#'
#' Plot multiple individuals' genome-wide genotypes
#'
#' @param geno Imputed phase-known genotypes, as a list of matrices
#'     (as produced by [maxmarg()]) or a list of
#'     three-dimensional arrays (as produced by [guess_phase()]).
#' @param map Marker map (a list of vectors of marker positions).
#' @param ind Individuals to plot, either a numeric indexes or IDs.
#' @param chr Selected chromosomes to plot; a vector of character strings.
#' @param gap Gap between chromosomes
#' @param col Vector of colors for the different genotypes.
#' @param na_col Color for missing segments.
#' @param chrlines Color for lines separating chromosomes
#' @param swap_axes If TRUE, swap the axes, so that the chromosomes run horizontally.
#' @param ... Additional graphics parameters
#'
#' @return None.
#'
#' @seealso [plot_onegeno()], [plot_genoprob()]
#'
#' @examples
#' # load data and calculate genotype probabilities
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' iron <- iron[1:50, ] # subset to first 50 individuals
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#' pr <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # infer genotypes, as those with maximal marginal probability
#' m <- maxmarg(pr, minprob=0.5)
#'
#' # re-code the X chr, (5,6) -> (1,3)
#' m[["X"]] <- (m[["X"]] - 5)*2 + 1
#'
#' # plot phased genotypes
#' plot_geno(m, map, col=c("#FFDC00", "#00C800", "#0064C9"))
#'
#' # this is more interesting for Diversity Outbred mouse data
#' \dontrun{
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' # subset to first 25 individuals
#' DOex <- DOex[1:25, ]
#' pr <- calc_genoprob(DOex, error_prob=0.002)
#'
#' # infer genotypes, as those with maximal marginal probability
#' m <- maxmarg(pr, minprob=0.5)
#' # guess phase
#' ph <- guess_phase(DOex, m)
#'
#' # plot phased genotypes
#' plot_geno(ph, DOex$gmap)
#' }
#'
#' @export
#' @importFrom graphics image
plot_geno <-
    function(geno, map, ind=NULL, chr=NULL, gap=0,
             col=NULL, na_col="white", chrlines="white",
             swap_axes=FALSE, ...)

{
    if(is.null(geno)) stop("geno is NULL")
    if(is.null(map)) stop("map is NULL")

    # ignore class of geno object
    geno <- unclass(geno)

    # subset individuals
    if(!is.null(ind)) {
        for(i in seq_along(geno)) {
            if(is.matrix(geno[[i]]))
                geno[[i]] <- geno[[i]][ind,,drop=FALSE]
            else {
                if(!is.array(geno[[i]]) || length(dim(geno[[i]])) != 3 ||
                   dim(geno[[i]])[3] != 2)
                    stop("geno should be an array individuals x positions x 2 haplotypes")
                geno[[i]] <- geno[[i]][ind,,,drop=FALSE]
            }
        }
    }

    # find common chr
    chr_geno <- names(geno)
    chr_map <- names(map)
    common_chr <- chr_geno[chr_geno %in% chr_map]
    if(length(common_chr) == 0)
        stop("No chr in common between geno and map")
    geno <- geno[common_chr]
    map <- map[common_chr]

    # subset chr if necessary
    if(!is.null(chr)) {
        if(any(!(chr %in% common_chr))) {
            chr <- chr[chr %in% common_chr]
            if(length(chr) == 0)
                stop("No chromosomes in common between geno, map, and chr")
            warning("Dropping some chr not found in geno and/or map")
        }
        geno <- geno[chr]
        map <- map[chr]
    }

    # same numbers of markers?
    nmar_map <- sapply(map, length)
    nmar_geno <- sapply(geno, ncol)
    if(any(nmar_geno != nmar_map))
        stop("Mismatch between numbers of markers between geno and map on chr ",
             paste(names(geno)[nmar_geno != nmar_map], collapse=", "))
    for(i in seq_along(geno)) {
        if(!all(names(map[[i]]) == colnames(geno[[i]])))
            stop("Mismatch between marker names on chr ", names(geno)[i])
    }

    if(length(dim(geno[[1]]))>2) { # genotypes are 2d arrays; stack them
        geno <- lapply(geno, function(a) {
            d3 <- dim(a)[3]
            result <- matrix(nrow=nrow(a)*d3, ncol=ncol(a))
            rownames(result) <- 1:nrow(result)
            colnames(result) <- colnames(geno)
            for(i in 1:d3) {
                result[(1:nrow(a))*d3-(d3-i),] <- a[,,i]
                rownames(result)[(1:nrow(a))*d3-(d3-i)] <- rownames(a)
            }
            result
        })
    }

    plot_geno_internal <-
        function(geno, map, col=NULL, na_col="white",
                 chrlines="white", chrlines_lwd=2,
                 swap_axes=FALSE,
                 xlab=NULL, ylab=NULL,
                 xlim=NULL, ylim=NULL, las=1,
                 mgp.x=c(2.6,0.5,0), mgp.y=c(2.6,0.5,0), mgp=NULL,
                 xaxt="s", yaxt="s",
                 ...)
    {
        # margin parameters
        if(!is.null(mgp)) mgp.x <- mgp.y <- mgp

        # paste genos together
        geno <- do.call("cbind", geno)

        # z-axis range and breaks
        geno[is.na(geno)] <- 0
        n_colors <- max(geno)+1
        zlim <- c(0, max(geno))
        breaks <- seq(-0.5, n_colors-0.5, by=1)

        if(is.null(col)) {
            if(n_colors==3) col <- c("#FFDC00", "#0064C9")
            else if(n_colors==4) col <- c("#FFDC00", "#00C800", "#0064C9")
            else col <- qtl2::CCcolors
        }
        col <- c(na_col, col)
        if(length(breaks) > length(col)+1)
            stop("Need more colors: at least ", n_colors-1)
        col <- col[1:(length(breaks)-1)]

        # get x-axis range
        if(is.null(xlim)) {
            start <- xpos_scan1(map, chr=names(map), gap=gap,
                                names(map)[1], min(map[[1]]))
            end <- xpos_scan1(map, chr=names(map), gap=gap,
                              names(map)[length(map)], max(map[[length(map)]]))

            xlim <- c(start, end)
        }
        if(is.null(ylim)) {
            ylim <- c(0.5, nrow(geno)+0.5)
        }

        xpos <- unlist(lapply(seq_along(map), function(chr) xpos_scan1(map, chr=names(map), gap=gap, names(map)[chr], map[[chr]])))

        if(any(diff(xpos)==0)) { # deal with identical positions
            xpos <- xpos + seq(0, 1e-6, length=length(xpos))
        }

        ypos <- 1:nrow(geno)

        if(!swap_axes) image(xpos, ypos, t(geno), breaks=breaks, col=col, xlab="", ylab="", xaxs="i", yaxs="i", xaxt="n", yaxt="n")
        if(swap_axes)  image(ypos, xpos, geno,    breaks=breaks, col=col, xlab="", ylab="", xaxs="i", yaxs="i", xaxt="n", yaxt="n")

        # chromosome axis labels
        chr <- names(map)
        if(length(chr) > 1) {
            chr_midpt <- sapply(seq_along(map), function(i) xpos_scan1(map, names(map), gap=gap,
                                                                       names(map)[i], mean(range(map[[i]], na.rm=TRUE))))

            for(i in seq_along(chr_midpt)) {
                if(!swap_axes && xaxt != "n") graphics::axis(side=1, at=chr_midpt[i], chr[i], mgp=mgp.x, tick=FALSE, las=las)
                if(swap_axes && yaxt != "n") graphics::axis(side=2, at=chr_midpt[i], chr[i], mgp=mgp.y, tick=FALSE, las=las)
            }
        } else { # single chromosome; put position on axis
            if(!swap_axes && xaxt != "n") graphics::axis(side=1, at=pretty(xpos, n=7), mgp=mgp.x, tick=FALSE, las=las)
            if(swap_axes && yaxt != "n") graphics::axis(side=2, at=pretty(xpos, n=7), mgp=mgp.y, tick=FALSE, las=las)
        }

        # ind column axis labels
        if(!swap_axes && yaxt != "n") graphics::axis(side=2, at=1:nrow(geno), rownames(geno), mgp=mgp.y, tick=FALSE, las=las)
        if(swap_axes && xaxt != "n") graphics::axis(side=1, at=1:nrow(geno), rownames(geno), mgp=mgp.x, tick=FALSE, las=las)

        # axis titles
        if(!is.null(xlab) && xlab != "") graphics::title(xlab=xlab, mgp=mgp.x)
        if(!is.null(ylab) && ylab != "") graphics::title(ylab=ylab, mgp=mgp.y)

        if(!is.na(chrlines) && !is.null(chrlines) && length(map)>1) {
            minpos <- sapply(map, min)
            chrlines_pos <- sapply(seq_along(map)[-1], function(chri) xpos_scan1(map, names(map), gap=gap, names(map)[chri], minpos[chri])-gap/2)
            if(swap_axes) abline(h=chrlines_pos, col=chrlines, lwd=chrlines_lwd)
            else abline(v=chrlines_pos, col=chrlines, lwd=chrlines_lwd)
        }

    }

    plot_geno_internal(geno, map, gap=gap, col=col, na_col=na_col,
                       swap_axes=swap_axes, chrlines=chrlines, ...)


}

# subset viterbi objects

#' Subsetting Viterbi results
#'
#' Pull out a specified set of individuals and/or chromosomes from
#' the results of [viterbi()]
#'
#' @param x Imputed genotypes as output from [viterbi()].
#' @param ind A vector of individuals: numeric indices, logical
#' values, or character string IDs
#' @param chr A vector of chromosomes: logical values, or character
#' string IDs. Numbers are interpreted as character string IDs.
#' @param ... Ignored.
#'
#' @return An object of class `"viterbi"`, like the input, with the
#' selected individuals and/or chromosomes; see [viterbi()].
#'
#' @seealso [viterbi()], [cbind.viterbi()], [rbind.viterbi()]
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' \dontshow{grav2 <- grav2[1:8,c(1,2)]}
#' g <- viterbi(grav2)
#' # keep just individuals 1:5, chromosome 2
#' gsub <- g[1:5,2]
#' # keep just chromosome 2
#' gsub2 <- g[,2]
subset.viterbi <-
    function(x, ind=NULL, chr=NULL, ...)
{
    if(is.null(ind) && is.null(chr))
        stop("You must specify either ind or chr.")

    x_attr <- attributes(x)
    x_attrnam <- names(x_attr)
    x_class <- class(x)

    if(!is.null(chr)) {
        chr <- subset_chr(chr, names(x))
        if(length(chr) == 0)
            stop("Must retain at least one chromosome.")

        x <- unclass(x)[chr]
        to_sub <- "is_x_chr"
        for(obj in to_sub) {
            if(obj %in% x_attrnam)
                x_attr[[obj]] <- x_attr[[obj]][chr]
        }

    }

    if(!is.null(ind)) {
        ind <- subset_ind(ind, rownames(x[[1]]))
        if(length(ind) == 0)
            stop("Must retain at least one individual.")

        for(i in names(x)) # loop over chromosomes
            x[[i]] <- unclass(x)[[i]][ind,,drop=FALSE]
    }

    for(obj in c("alleles", "is_x_chr", "crosstype"))
        attr(x, obj) <- x_attr[[obj]]
    class(x) <- x_class

    x
}

#' @export
#' @rdname subset.viterbi
`[.viterbi` <-
    function(x, ind=NULL, chr=NULL)
    subset(x, ind, chr)

#' Plot phenotype vs genotype
#'
#' Plot phenotype vs genotype for a single putative QTL and a single phenotype.
#'
#' @param geno Vector of genotypes, for example as produced by
#' [maxmarg()] with specific `chr` and `pos`.
#' @param pheno Vector of phenotypes.
#' @param sort If TRUE, sort genotypes from largest to smallest.
#' @param SEmult If specified, interval estimates of the within-group
#' averages will be displayed, as `mean +/- SE * SEmult`.
#' @param pooledSD If TRUE and `SEmult` is specified, calculated
#' a pooled within-group SD. Otherwise, get separate estimates of
#' the within-group SD for each group.
#' @param swap_axes If TRUE, swap the axes, so that the genotypes are
#' on the y-axis and the phenotype is on the x-axis.
#' @param jitter Amount to jitter the points horizontally, if a vector
#' of length > 0, it is taken to be the actual jitter amounts
#' (with values between -0.5 and 0.5).
#' @param force_labels If TRUE, force all genotype labels to be shown.
#' @param alternate_labels If TRUE, place genotype labels in two rows
#' @param omit_points If TRUE, omit the points, just plotting the averages (and, potentially, the +/- SE intervals).
#' @param ... Additional graphics parameters, passed to `plot()`.
#'
#' @return (Invisibly) A matrix with rows being the genotype groups
#' and columns for the means and (if `SEmult` is specified) the SEs.
#'
#' @export
#' @importFrom graphics par segments title axis points
#' @importFrom stats lm runif sd
#'
#' @section Hidden graphics parameters:
#' A number of graphics parameters can be passed via `...`. For
#' example, `bgcolor` to control the background color, and
#' `seg_width`, `seg_lwd`, and `seg_col` to control the lines at the
#' confidence intervals. Further, `hlines`, `hlines_col`,
#' `hlines_lwd`, and `hlines_lty` to control the horizontal grid
#' lines. (Use `hlines=NA` to avoid plotting horizontal grid lines.)
#' Similarly `vlines`, `vlines_col`, `vlines_lwd`, and `vlines_lty`
#' for vertical grid lines. These are not included as formal
#' parameters in order to avoid cluttering the function definition.
#'
#' @seealso [plot_coef()], [plot_ci()]
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,"16"]}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # inferred genotype at a 28.6 cM on chr 16
#' geno <- maxmarg(probs, map, chr=16, pos=28.6, return_char=TRUE)
#'
#' # plot phenotype vs genotype
#' plot_pxg(geno, log10(iron$pheno[,1]), ylab=expression(log[10](Liver)))
#'
#' # include +/- 2 SE intervals
#' plot_pxg(geno, log10(iron$pheno[,1]), ylab=expression(log[10](Liver)),
#'          SEmult=2)
#'
#' # plot just the means
#' plot_pxg(geno, log10(iron$pheno[,1]), ylab=expression(log[10](Liver)),
#'          omit_points=TRUE)
#'
#' # plot just the means +/- 2 SEs
#' plot_pxg(geno, log10(iron$pheno[,1]), ylab=expression(log[10](Liver)),
#'          omit_points=TRUE, SEmult=2)
plot_pxg <-
    function(geno, pheno, sort=TRUE, SEmult=NULL, pooledSD=TRUE,
             swap_axes=FALSE, jitter=0.2,
             force_labels=TRUE, alternate_labels=FALSE,
             omit_points=FALSE, ...)
{
    if(is.null(geno)) stop("geno is NULL")
    if(is.null(pheno)) stop("pheno is NULL")

    if(length(jitter) > 1 && length(jitter) != length(geno)) {
        stop("length(jitter) > 1 but length(jitter) != length(geno)")
        if(any(!is.na(jitter) & (jitter < -0.5 | jitter > 0.5)))
            stop("jitter values should be in [-0.5, +0.5]")
    }
    if(length(jitter) == 1) {
        if(jitter < 0 || jitter > 0.5)
            stop("jitter should be in [0,0.5]")
        jitter <- runif(length(geno), -jitter, jitter)
    }
    names(jitter) <- names(geno)

    # force pheno to be numeric
    if(is.matrix(pheno) || is.data.frame(pheno)) {
        if(ncol(pheno)>1) {
            pheno <- pheno[,1,drop=FALSE]
            warning("pheno is a ", class(pheno)[1], "; using the first column")
        }
        pheno <- setNames(as.numeric(unlist(pheno)), rownames(pheno))
    }

    # align geno and pheno
    ind_g <- names(geno)
    ind_p <- names(pheno)
    if(!is.null(ind_g) && !is.null(ind_p)) {
        ind <- ind_g[ind_g %in% ind_p]
        if(length(ind) == 0)
            stop("No individuals in common between geno and pheno")
        geno <- geno[ind]
        jitter <- jitter[ind]
        pheno <- pheno[ind]
    }
    else {
        if(length(geno) != length(pheno))
            stop("length(geno) != length(pheno)")
    }

    # genotype groups
    if(!is.factor(geno)) geno <- factor(geno)
    ugeno <- levels(geno)
    me <- tapply(pheno, geno, mean, na.rm=TRUE)
    if(sort) {
        o <- order(me, decreasing=TRUE, na.last=TRUE)
        ugeno <- names(me[o])
        geno <- factor(geno, levels=ugeno)
        me <- tapply(pheno, geno, mean, na.rm=TRUE)
    }

    # calculate intervals
    se <- NULL
    if(!is.null(SEmult)) {
        if(SEmult < 0) stop("SEmult should be >= 0")
        if(pooledSD) {
            sigma <- summary(lm(pheno ~ geno))$sigma
            se <- sigma / sqrt(tapply(pheno, geno, function(a) sum(!is.na(a))))
        }
        else {
            se <- tapply(pheno, geno, function(a) sd(a, na.rm=TRUE)/sqrt(sum(!is.na(a))))
        }
    }

    plot_pxg_internal <-
        function(geno, pheno, swap_axes=FALSE, bgcolor="gray90",
                 seg_width=NULL, seg_lwd=2, seg_col="black",
                 hlines=NULL, hlines_col=NULL, hlines_lty=NULL, hlines_lwd=NULL,
                 vlines=NULL, vlines_col=NULL, vlines_lty=NULL, vlines_lwd=NULL,
                 xlim=NULL, ylim=NULL,
                 xaxs=NULL, yaxs=NULL, xlab=NULL, ylab=NULL,
                 mgp=c(2.6, 0.3, 0), mgp.x=mgp, mgp.y=mgp, las=NULL,
                 pch=21, bg="lightblue", main="", sub="", ...)
        {
            if(swap_axes) {
                if(is.null(ylim)) ylim <- c(0.5, length(ugeno)+0.5)

                # get x-axis limits
                if(is.null(xlim)) {
                    if(omit_points) {
                        xlim <- range(me, na.rm=TRUE)
                    } else {
                        xlim <- range(pheno, na.rm=TRUE)
                    }
                    if(!is.null(SEmult)) {
                        xlim <- range(c(xlim, me-se*SEmult, me+se*SEmult), na.rm=TRUE)
                    }
                }

                if(is.null(xlab)) xlab <- "Phenotype"
                if(is.null(ylab)) ylab <- "Genotype"

                if(is.null(yaxs)) yaxs <- "i"
                if(is.null(xaxs)) xaxs <- "r"

                if(is.null(las)) las <- 1

                if(is.null(hlines)) hlines <- seq_len(length(ugeno))
                if(is.null(vlines)) vlines <- pretty(pheno)
                if(is.null(hlines_col)) hlines_col <- "gray80"
                if(is.null(hlines_lwd)) hlines_lwd <- 3
                if(is.null(hlines_lty)) hlines_lty <- 1
                if(is.null(vlines_col)) vlines_col <- "white"
                if(is.null(vlines_lwd)) vlines_lwd <- 1
                if(is.null(vlines_lty)) vlines_lty <- 1

                x <- pheno
                y <- match(geno, ugeno) + jitter

            } else {
                if(is.null(xlim)) xlim <- c(0.5, length(ugeno)+0.5)

                # get y-axis limits
                if(is.null(ylim)) {
                    if(omit_points) {
                        ylim <- range(me, na.rm=TRUE)
                    } else {
                        ylim <- range(pheno, na.rm=TRUE)
                    }
                    if(!is.null(SEmult)) {
                        ylim <- range(c(ylim, me-se*SEmult, me+se*SEmult), na.rm=TRUE)
                    }
                }

                if(is.null(ylab)) ylab <- "Phenotype"
                if(is.null(xlab)) xlab <- "Genotype"

                if(is.null(xaxs)) xaxs <- "i"
                if(is.null(yaxs)) yaxs <- "r"

                if(is.null(las)) {
                    las <- ifelse(length(ugeno) > 4, 2, 1)
                }

                if(is.null(hlines)) hlines <- pretty(pheno)
                if(is.null(vlines)) vlines <- seq_len(length(ugeno))
                if(is.null(vlines_col)) vlines_col <- "gray80"
                if(is.null(vlines_lwd)) vlines_lwd <- 3
                if(is.null(vlines_lty)) vlines_lty <- 1
                if(is.null(hlines_col)) hlines_col <- "white"
                if(is.null(hlines_lwd)) hlines_lwd <- 1
                if(is.null(hlines_lty)) hlines_lty <- 1

                y <- pheno
                x <- match(geno, ugeno) + jitter

            }

            plot(0, 0, type="n", xlim=xlim, ylim=ylim, xlab="", ylab="", xaxt="n",
                 yaxt="n", xaxs=xaxs, yaxs=yaxs, main=main, sub=sub)
            title(xlab=xlab, mgp=mgp.x)
            title(ylab=ylab, mgp=mgp.y)

            u <- par("usr")
            rect(u[1], u[3], u[2], u[4], col=bgcolor, border=NA)

            # grid lines (if swap_axes, horizontal over vertical
            if(!swap_axes) {
                if(!(length(vlines)==1 && is.na(vlines)))
                    abline(v=vlines, lwd=vlines_lwd, lty=vlines_lty, col=vlines_col)
            }
            if(!(length(hlines)==1 && is.na(hlines)))
                abline(h=hlines, lwd=hlines_lwd, lty=hlines_lty, col=hlines_col)
            if(swap_axes) {
                if(!(length(vlines)==1 && is.na(vlines)))
                    abline(v=vlines, lwd=vlines_lwd, lty=vlines_lty, col=vlines_col)
            }

            # x-axis labels
            xtick <- seq_along(ugeno)
            side <- swap_axes + 1
            this_mgp <- list(mgp.x, mgp.y)[[side]]
            if(force_labels) {
                    if(alternate_labels) {
                        for(i in xtick)
                            axis(side=side, at=i, labels=ugeno[i], mgp=this_mgp, tick=FALSE,
                                 line=2 - (i %% 2)-1, las=las)
                    } else {
                        for(i in xtick)
                            axis(side=side, at=i, labels=ugeno[i], mgp=this_mgp, tick=FALSE, las=las)
                    }
            } else {
                if(alternate_labels) {
                    odd <- seq(1, length(ugeno), by=2)
                    axis(side=side, at=xtick[odd], labels=ugeno[odd], mgp=this_mgp, tick=FALSE, las=las)
                    if(length(ugeno) > 1) {
                        axis(side=side, at=xtick[-odd], labels=ugeno[-odd], mgp=this_mgp, tick=FALSE,
                             line=1, las=las)
                    }
                } else {
                    axis(side=side, at=xtick, labels=ugeno, mgp=this_mgp, tick=FALSE, las=las)
                }
            }

            # add points
            if(!omit_points)
                points(x, y, pch=pch, bg=bg, ...)

            # y-axis labels
            if(swap_axes) {
                axis(side=1, at=vlines, mgp=mgp.x, tick=FALSE, las=las)
            } else {
                axis(side=2, at=hlines, mgp=mgp.y, tick=FALSE, las=las)
            }

            # determine default CI segment width; use width of plot but make it no bigger than 0.2 and no smaller than 0.05
            if(is.null(seg_width)) {
                seg_width <- ifelse(swap_axes, diff(par("usr"))[3:4], diff(par("usr"))[1:2])/20
                if(seg_width > 0.2) seg_width <- 0.2
                if(seg_width < 0.05) seg_width <- 0.05
            }

            # add CIs
            if(seg_width > 0) {
                if(swap_axes) {
                    segments(me, xtick-seg_width/2, me, xtick+seg_width/2,
                             lwd=seg_lwd, col=seg_col)
                }
                else {
                    segments(xtick-seg_width/2, me, xtick+seg_width/2, me,
                             lwd=seg_lwd, col=seg_col)
                }

                if(!is.null(SEmult)) {
                    if(swap_axes) {
                        segments(me[ugeno]+se[ugeno]*SEmult, xtick, me[ugeno]-se[ugeno]*SEmult, xtick,
                                 lwd=seg_lwd, col=seg_col)
                    } else {
                        segments(xtick, me[ugeno]-se[ugeno]*SEmult, xtick, me[ugeno]+se[ugeno]*SEmult,
                                 lwd=seg_lwd, col=seg_col)
                    }
                }

            }

            box()
        }

    plot_pxg_internal(geno, pheno, swap_axes=swap_axes, ...)


    # return mean and SE
    invisible(cbind(mean=me, SE=se))

}

#' Join Viterbi results for different individuals
#'
#' Join multiple imputed genotype objects, as produced by [viterbi()],
#' for the same set of markers but for different individuals.
#'
#' @param ... Imputed genotype objects as produced by
#' [viterbi()]. Must have the same set of markers.
#'
#' @return An object of class `"viterbi"`, like the input; see [viterbi()].
#'
#' @seealso [cbind.viterbi()], [subset.viterbi()], [viterbi()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' gA <- viterbi(grav2[1:5,], map, error_prob=0.002)
#' gB <- viterbi(grav2[6:12,], map, error_prob=0.002)
#' g <- rbind(gA, gB)
#'
#' @export
rbind.viterbi <-
    function(...)
{
    args <- list(...)

    # to rbind: the data itself
    # to pass through (must match): crosstype, is_x_chr, alleles

    result <- args[[1]]
    if(length(args) == 1) return(result)

    # check that things match
    other_stuff <- c("crosstype", "is_x_chr", "alleles")
    for(i in 2:length(args)) {
        for(obj in other_stuff) {
            if(!is_same(attr(args[[1]], obj), attr(args[[i]], obj)))
                stop("Input objects 1 and ", i, " differ in their ", obj)
        }
    }

    for(i in 2:length(args)) {
        if(length(result) != length(args[[i]]) ||
           !all(names(result) == names(args[[i]])))
            stop("Input arguments have different chromosomes")

        for(s in seq(along=result)) {
            if(!is_same(ncol(result[[s]]), ncol(args[[i]][[s]])))
                stop("input objects have varying numbers of geno columns on chr ", names(result)[s])
            result[[s]] <- rbind(result[[s]], args[[i]][[s]])
        }
    }

    result
}

#' Plot a genome scan
#'
#' Plot LOD curves for a genome scan
#'
#' @param x An object of class `"scan1"`, as output by [scan1()].
#'
#' @param map A list of vectors of marker positions, as produced by
#' [insert_pseudomarkers()].
#'
#' @param lodcolumn LOD score column to plot (a numeric index, or a
#' character string for a column name). Only one value allowed.
#'
#' @param chr Selected chromosomes to plot; a vector of character
#' strings.
#'
#' @param add If TRUE, add to current plot (must have same map and
#' chromosomes).
#'
#' @param gap Gap between chromosomes. The default is 1% of the total genome length.
#'
#' @param ... Additional graphics parameters.
#'
#' @seealso [plot_coef()], [plot_coefCC()], [plot_snpasso()], [plot_scan1_heatmap()]
#'
#' @section Hidden graphics parameters:
#' A number of graphics parameters can be passed via `...`. For
#' example, `bgcolor` to control the background color and
#' `altbgcolor` to control the background color on alternate chromosomes.
#' `col` controls the color of lines/curves; `altcol` can be used if
#' you want alternative chromosomes in different colors.
#' These are not included as formal parameters in order to avoid
#' cluttering the function definition.
#'
#' @export
#' @importFrom graphics rect lines par axis title abline box
#'
#' @return None.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' # plot the results for selected chromosomes
#' ylim <- c(0, maxlod(out)*1.02) # need to strip class to get overall max LOD
#' chr <- c(2,7,8,9,15,16)
#' plot(out, map, chr=chr, ylim=ylim)
#' plot(out, map, lodcolumn=2, chr=chr, col="violetred", add=TRUE)
#' legend("topleft", lwd=2, col=c("darkslateblue", "violetred"), colnames(out),
#'        bg="gray90")
#'
#' # plot just one chromosome
#' plot(out, map, chr=8, ylim=ylim)
#' plot(out, map, chr=8, lodcolumn=2, col="violetred", add=TRUE)
#'
#' # lodcolumn can also be a column name
#' plot(out, map, lodcolumn="liver", ylim=ylim)
#' plot(out, map, lodcolumn="spleen", col="violetred", add=TRUE)
plot_scan1 <-
    function(x, map, lodcolumn=1, chr=NULL, add=FALSE, gap=NULL, ...)
{
    if(is.null(map)) stop("map is NULL")

    if(!is.list(map)) map <- list(" "=map) # if a vector, treat it as a list with no names

    # if input x is a vector, try turning it into a matrix, since we'll be pull out row names
    if(!is.matrix(x) && !is.data.frame(x) && is.numeric(x)) x <- as.matrix(x)

    # subset chromosomes
    if(!is.null(chr)) {
        chri <- match(chr, names(map))
        if(any(is.na(chri)))
            stop("Chromosomes ", paste(chr[is.na(chri)], collapse=", "), " not found")
        x <- subset_scan1(x, map, chr)
        map <- map[chri]
    }

    if(is.null(gap)) gap <- sum(chr_lengths(map))/100
    if(!is_nonneg_number(gap)) stop("gap should be a single non-negative number")

    # align scan1 output and map
    tmp <- align_scan1_map(x, map)
    x <- tmp$scan1
    map <- tmp$map

    if(nrow(x) != length(unlist(map)))
        stop("nrow(x) [", nrow(x), "] != number of positions in map [",
             length(unlist(map)), "]")

    # pull out lod scores
    if(length(lodcolumn)==0) stop("lodcolumn has length 0")
    if(length(lodcolumn) > 1) { # If length > 1, take first value
        warning("lodcolumn should have length 1; only first element used.")
        lodcolumn <- lodcolumn[1]
    }
    if(is.character(lodcolumn)) { # turn column name into integer
        tmp <- match(lodcolumn, colnames(x))
        if(is.na(tmp)) stop('lodcolumn "', lodcolumn, '" not found')
        lodcolumn <- tmp
    }
    if(lodcolumn < 1 || lodcolumn > ncol(x))
        stop("lodcolumn [", lodcolumn, "] out of range (should be in 1, ..., ", ncol(x), ")")
    lod <- unclass(x)[,lodcolumn]

    # internal function; trick to be able to pull things out of "..."
    #    but still have some defaults for them
    plot_scan1_internal <-
        function(map, lod, add=FALSE, gap=NULL,
                 bgcolor="gray90", altbgcolor="gray85",
                 lwd=2, col="darkslateblue", altcol=NULL,
                 xlab=NULL, ylab="LOD score",
                 xlim=NULL, ylim=NULL, xaxs="i", yaxs="i",
                 main="", mgp.x=c(2.6, 0.5, 0), mgp.y=c(2.6, 0.5, 0),
                 mgp=NULL, las=1,
                 hlines=NULL, hlines_col="white", hlines_lwd=1, hlines_lty=1,
                 vlines=NULL, vlines_col="white", vlines_lwd=1, vlines_lty=1,
                 sub="", ...)
        {
            dots <- list(...)
            onechr <- (length(map)==1) # single chromosome

            xpos <- map_to_xpos(map, gap)
            chrbound <- map_to_boundaries(map, gap)

            if(!add) { # new plot
                if(is.null(ylim))
                    ylim <- c(0, max(lod, na.rm=TRUE)*1.02)

                if(is.null(xlim)) {
                    xlim <- range(xpos, na.rm=TRUE)
                    if(!onechr) xlim <- xlim + c(-gap/2, gap/2)
                }

                if(is.null(xlab)) {
                    if(onechr) {
                        if(names(map) == " ") xlab <- "Position"
                        else xlab <- paste("Chr", names(map), "position")
                    }
                    else xlab <- "Chromosome"
                }

                # margin parameters
                if(!is.null(mgp)) mgp.x <- mgp.y <- mgp

                # make basic plot
                plot(xpos, lod, xlab="", ylab="", xlim=xlim, ylim=ylim,
                     xaxs=xaxs, yaxs=yaxs, xaxt="n", yaxt="n", type="n",
                     main=main, sub=sub)

                # add background rectangles
                u <- par("usr")
                if(!is.null(bgcolor))
                    rect(u[1], u[3], u[2], u[4], col=bgcolor, border=NA)
                if(!is.null(altbgcolor) && !onechr) {
                    for(i in seq(2, ncol(chrbound), by=2))
                        rect(chrbound[1,i], u[3], chrbound[2,i], u[4], col=altbgcolor, border=NA)
                }

                # include axis labels?
                if(is.null(dots$xaxt)) dots$xaxt <- par("xaxt")
                if(is.null(dots$yaxt)) dots$yaxt <- par("yaxt")

                # add x axis unless par(xaxt="n")
                if(dots$xaxt != "n") {
                    if(onechr) {
                        axis(side=1, at=pretty(xlim), mgp=mgp.x, las=las, tick=FALSE)
                    }
                    else {
                        loc <- colMeans(chrbound)
                        odd <- seq(1, length(map), by=2)
                        even <- seq(2, length(map), by=2)
                        axis(side=1, at=loc[odd], names(map)[odd],
                             mgp=mgp.x, las=las, tick=FALSE)
                        axis(side=1, at=loc[even], names(map)[even],
                             mgp=mgp.x, las=las, tick=FALSE)
                    }
                }

                # add y axis unless par(yaxt="n")
                if(dots$yaxt != "n") {
                    axis(side=2, at=pretty(ylim), mgp=mgp.y, las=las, tick=FALSE)
                }
                dots$xaxt <- dots$yaxt <- NULL # delete those

                # x and y axis labels
                title(xlab=xlab, mgp=mgp.x)
                title(ylab=ylab, mgp=mgp.y)

                # grid lines
                if(onechr && !(length(vlines)==1 && is.na(vlines))) { # if vlines==NA (or mult chr), skip lines
                    if(is.null(vlines)) vlines <- pretty(xlim)
                    abline(v=vlines, col=vlines_col, lwd=vlines_lwd, lty=vlines_lty)
                }
                if(!(length(hlines)==1 && is.na(hlines))) { # if hlines==NA, skip lines
                    if(is.null(hlines)) hlines <- pretty(ylim)
                    abline(h=hlines, col=hlines_col, lwd=hlines_lwd, lty=hlines_lty)
                }
            }

            # plot each chromosome
            indexes <- map_to_index(map)
            for(i in seq(along=indexes)) {
                # if altcol provided, have chromosomes alternate colors
                if(is.null(altcol) || i %% 2) this_col <- col
                else this_col <- altcol

                lines(xpos[indexes[[i]]], lod[indexes[[i]]],
                           lwd=lwd, col=this_col, ...)
            }

            # add box just in case
            box()
        }

    # make the plot
    plot_scan1_internal(map=map, lod=lod, add=add, gap=gap, ...)
}


#' @export
#' @rdname plot_scan1
plot.scan1 <-
    function(x, map, lodcolumn=1, chr=NULL, add=FALSE, gap=NULL, ...)
{
    if(is.null(map)) stop("map is NULL")

    # if map looks like snpinfo, assume this is a snp asso result and use plot_snpasso()
    if(is.data.frame(map) && "index" %in% names(map)) {
        plot_snpasso(x, snpinfo=map, lodcolumn=lodcolumn, add=add, gap=gap, chr=chr, ...)
    }
    else { # mostly, use plot_scan1()
        plot_scan1(x, map=map, lodcolumn=lodcolumn, chr=chr, add=add, gap=gap, ...)
    }
}


# convert map to x-axis positions for plot_scan1
map_to_xpos <-
    function(map, gap=NULL)
{
    if(is.null(gap)) sum(chr_lengths(map))/100

    if(length(map)==1) return(map[[1]])

    chr_range <- vapply(map, range, c(0,1), na.rm=TRUE)

    result <- map[[1]]-chr_range[1,1] + gap/2
    for(i in 2:length(map)) {
        result <- c(result,
                    map[[i]] - chr_range[1,i] + gap + max(result, na.rm=TRUE))
    }
    result
}

# boundaries of chromosomes in plot_scan1
# first row: left edges
# second row: right edges
map_to_boundaries <-
    function(map, gap=NULL)
{
    if(is.null(gap)) gap <- sum(chr_lengths(map))/100

    if(length(map)==1)
        return(cbind(range(map[[1]], na.rm=TRUE)))

    # range of each chromosome
    chr_range <- lapply(map, range, na.rm=TRUE)

    # corresponding xpos, as matrix with two rows
    startend <- matrix(map_to_xpos(chr_range, gap), nrow=2)

    startend[1,] <- startend[1,] - gap/2
    startend[2,] <- startend[2,] + gap/2

    startend
}

# convert map to list of indexes to LOD vector
map_to_index <-
    function(map)
{
    if(length(map)==1) {
        map[[1]] <- seq(along=map[[1]])
        return(map)
    }

    lengths <- vapply(map, length, 0)
    split(seq_len(sum(lengths)), rep(seq(along=map), lengths))
}

# find_marker
#' Find markers by chromosome position
#'
#' Find markers closest to specified set of positions, or within a specified interval.
#'
#' @param map A map object: a list (corresponding to chromosomes) of
#'     vectors of marker positions. Can also be a snpinfo object (data
#'     frame with columns `chr` and `pos`; marker names taken from
#'     column `snp` or if that doesn't exist from the row names)
#' @param chr A vector of chromosomes
#' @param pos A vector of positions
#' @param interval A pair of positions (provide either `pos` or `interval` but not both)
#'
#' @return A vector of marker names, either closest to the positions
#' specified by `pos`, or within the interval defined by
#' `interval`.
#'
#' @details
#' If `pos` is provided, `interval` should not be, and vice versa.
#'
#' If `pos` is provided, then `chr` and `pos` should
#' either be the same length, or one of them should have length 1 (to
#' be expanded to the length of the other).
#'
#' If `interval` is provided, then `chr` should have length 1.
#'
#' @seealso [find_markerpos()], [find_index_snp()], [pull_genoprobpos()], [pull_genoprobint()]
#' @export
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#'
#' # find markers by their genetic map positions
#' find_marker(iron$gmap, c(8, 11), c(37.7, 56.9))
#'
#' # find markers by their physical map positions (two markers on chr 7)
#' find_marker(iron$pmap, 7, c(44.2, 108.9))
#'
#' # find markers in an interval
#' find_marker(iron$pmap, 16, interval=c(35, 80))
find_marker <-
    function(map, chr, pos=NULL, interval=NULL)
{
    if(is.null(map)) stop("map is NULL")
    if(is.data.frame(map)) { # snp info table; convert to list
        if(!all(c("chr", "pos") %in% colnames(map)))
            stop('If map is a data frame, it should include columns "chr", and "pos".')
        map_chr <- factor(map$chr, unique(map$chr))
        map_pos <- split(map$pos, map_chr)
        mar <- map$snp
        if(is.null(mar)) mar <- rownames(map)
        mar <- split(mar, map_chr)
        for(i in seq_along(map_pos))
            names(map_pos[[i]]) <- mar[[i]]
        map <- map_pos
    }
    chr <- as.character(chr) # treat as a character string

    if(!is.null(interval)) {
        if(!is.null(pos))
            stop("Only one of pos and interval should be given")
        if(length(chr) > 1)
            stop("If interval is provided, chr should have length 1")
        if(length(interval) != 2)
            stop("interval should have length 2")

        # make sure it's sorted
        interval <- sort(interval)

        if(!(chr %in% names(map)))
            stop("Can't find chromosome ", chr)

        map <- map[[chr]]
        return(names(map)[map >= interval[1] & map <= interval[2]])
    }
    else {
        if(length(chr) != length(pos)) {
            if(length(chr) > 1 && length(pos) > 1)
                stop("chr and pos should be the same length, or at least one of length 1")
            if(length(chr)==1 && length(pos) > 1)
                chr <- rep(chr, length(pos))
            if(length(pos)==1 && length(chr) > 1)
                pos <- rep(pos, length(chr))
        }
        if(length(chr) > 1) {
            found <- chr %in% names(map)
            if(!all(found)) {
                if(!any(found))
                    stop("None of the chromosomes found in the map")
                warning("Some chr not found: ", paste(chr[!found], collapse=", "))
                chr <- chr[found]
                pos <- pos[found]
            }

            result <- rep("", length(chr))
            for(i in seq(along=chr))
                result[i] <- find_marker(map, chr[i], pos[i])
            return(result)
        }

        # one position
        if(!(chr %in% names(map)))
            stop("chr ", chr, " not found")
        map <- map[[chr]]
        d <- abs(map - pos)
        closest <- which(d==min(d))
        if(length(closest) > 1) closest <- sample(closest, 1)
        return(names(map)[closest])
    }
}

# subset cross2 objects

# subset.cross2
#' Subsetting data for a QTL experiment
#'
#' Pull out a specified set of individuals and/or chromosomes from a
#' `cross2` object.
#'
#' @param x An object of class `"cross2"`, as output by [read_cross2()]. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param ind A vector of individuals: numeric indices, logical
#' values, or character string IDs.
#' @param chr A vector of chromosomes: numeric indices, logical
#' values, or character string IDs
#' @param ... Ignored.
#'
#' @details
#' When subsetting by individual, if `ind` is numeric, they're
#' assumed to be numeric indices; if character strings, they're
#' assumed to be individual IDs. `ind` can be numeric or logical
#' only if the genotype, phenotype, and covariate data all have the
#' same individuals in the same order.
#'
#' When subsetting by chromosome, `chr` is _always_
#' converted to character strings and treated as chromosome IDs. So if
#' there are three chromosomes with IDs `"18"`, `"19"`, and
#' `"X"`, `mycross[,18]` will give the first of the
#' chromosomes (labeled `"18"`) and `mycross[,3]` will give
#' an error.
#'
#' When using character string IDs for `ind` or `chr`, you
#' can use "negative" subscripts to indicate exclusions, for example
#' `mycross[,c("-18", "-X")]` or `mycross["-Mouse2501",]`.
#' But you can't mix "positive" and "negative" subscripts, and if any
#' of the individuals has an ID that begins with `"-"`, you can't
#' use negative subscripts like this.
#'
#' @return The input `cross2` object, with the selected
#' individuals and/or chromsomes.
#'
#' @section Warning:
#' The order of the two arguments is reversed relative to the related
#' function in [R/qtl](https://rqtl.org).
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' # keep individuals 1-20 and chromosomes 3 and 4
#' grav2sub <- grav2[1:20, c(3,4)]
#' # keep just chromosome 1
#' grav2_c1 <- grav2[,1]
subset.cross2 <-
    function(x, ind=NULL, chr=NULL, ...)
{
    if(is.null(ind) && is.null(chr))
        stop("You must specify either ind or chr.")

    slice_by_chr <- c("geno", "founder_geno", "gmap", "pmap", "is_x_chr")

    slice_by_ind <- c("geno", "is_female", "cross_info", "pheno", "covar")

    if(!is.null(chr)) {
        # first clean up chromosome argument
        chr <- subset_chr(chr, names(x$geno))

        for(obj in slice_by_chr) {
            if(obj %in% names(x)) {
                x[[obj]] <- x[[obj]][chr]
            }
        }
    }

    if(!is.null(ind)) {
        # check that geno, covar, pheno have individuals in same order
        allow_logical <- TRUE
        ind_g <- rownames(x$geno[[1]])
        if(!is.null(x$covar) && (nrow(x$covar) != length(ind_g) ||
           !all(rownames(x$covar) == ind_g)))
            allow_logical <- FALSE
        if(!is.null(x$pheno) && (nrow(x$pheno) != length(ind_g) ||
           !all(rownames(x$pheno) == ind_g)))
            allow_logical <- FALSE

        # first clean up ind argument
        ind <- subset_ind(ind, ind_g, allow_logical)

        # Finally, the actual subsetting
        for(obj in slice_by_ind) {
            if(obj %in% names(x)) {
                # is it a matrix like cross_info?
                if(is.matrix(x[[obj]]) || is.data.frame(x[[obj]])) {
                    x[[obj]] <- x[[obj]][ind[ind %in% rownames(x[[obj]])],,drop=FALSE]
                }
                # is it a list like $geno?
                else if(!is.matrix(x[[obj]]) && is.list(x[[obj]])) {
                    x[[obj]] <- lapply(x[[obj]], function(a, b) {
                        if(is.matrix(a))
                            a <- a[b[b %in% rownames(a)],,drop=FALSE]
                        else
                            a <- a[b[b %in% names(a)]]
                        a }, ind)
                }
                else # it's a vector
                    x[[obj]] <- x[[obj]][ind[ind %in% names(x[[obj]])]]
            }
        }
    }

    x
}

#' @export
#' @rdname subset.cross2
`[.cross2` <-
    function(x, ind=NULL, chr=NULL)
    subset(x, ind, chr)

#' Find name of indexed snp
#'
#' For a particular SNP, find the name of the corresponding indexed SNP.
#'
#' @param snpinfo Data frame with SNP information with the following columns:
#' * `chr` - Character string or factor with chromosome
#' * `index` - Numeric index of equivalent, indexed SNP, as produced by [index_snps()].
#' * `snp` - Character string with SNP identifier (if
#'     missing, the rownames are used).
#' @param snp Name of snp to look for (can be a vector).
#'
#' @return A vector of SNP IDs (the corresponding indexed SNPs), with NA if a SNP is not found.
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DO_Recla/recla.zip")
#' recla <- read_cross2(file)
#'
#' # founder genotypes for a set of SNPs
#' snpgeno <- rbind(m1=c(3,1,1,3,1,1,1,1),
#'                  m2=c(3,1,1,3,1,1,1,1),
#'                  m3=c(1,1,1,1,3,3,3,3),
#'                  m4=c(1,3,1,3,1,3,1,3))
#' sdp <- calc_sdp(snpgeno)
#' snpinfo <- data.frame(chr=c("19", "19", "X", "X"),
#'                       pos=c(40.36, 40.53, 110.91, 111.21),
#'                       sdp=sdp,
#'                       snp=c("m1", "m2", "m3", "m4"), stringsAsFactors=FALSE)
#'
#' # update snp info by adding the SNP index column
#' snpinfo <- index_snps(recla$pmap, snpinfo)
#'
#' # find indexed snp for a particular snp
#' find_index_snp(snpinfo, "m3")
#' }
#'
#' @seealso [find_marker()]
#' @export
find_index_snp <-
    function(snpinfo, snp)
{
    if(!("index" %in% colnames(snpinfo)))
        stop('snpinfo does not contain an "index" column; run index_snps().')
    if(!("chr" %in% colnames(snpinfo)))
        stop('snpinfo does not contain an "chr" column.')

    if(length(unique(snpinfo$chr)) > 1) { # more than one chromosome
        # split the snpinfo table by chromosome
        snpinfo <- split(snpinfo, snpinfo$chr)

        # vector of NAs to contain the results
        result <- as.character(rep(NA, length(snp)))

        for(i in seq(along=snpinfo)) { # loop over chromosomes
            # if found everything already, just return...
            if(!any(is.na(result))) return(result)

            # grab snp names for this chromosome
            snps <- snpinfo[[i]]$snp
            if(is.null(snps)) snps <- rownames(snpinfo[[i]])
            index <- snpinfo[[i]]$index

            # indexed snps for those that have not yet been found
            result[is.na(result)] <- snps[index[match(snp[is.na(result)], snps)]]
        }

        # if any remain not found, issue warning
        if(any(is.na(result))) {
            warning("snps not found: ", paste(snp[is.na(result)], collapse=", "))
        }
        return(result)
    }

    # one chromosome:
    snps <- snpinfo$snp
    if(is.null(snps)) snps <- rownames(snpinfo)

    result <- snps[snpinfo$index[match(snp, snps)]]
    if(any(is.na(result))) {
        warning("snps not found: ", paste(snp[is.na(result)], collapse=", "))
    }

    result
}

context("find index snp")

test_that("find_index_snp works", {

    snpinfo <- structure(list(chr = c("19", "19", "X", "X"),
                              pos = c(40.36, 40.53, 110.91, 111.21), sdp = c(170, 170, 240, 170),
                              snp = c("m1", "m2", "m3", "m4"), index = c(1L, 1L, 1L, 2L),
                              interval = c(91L, 91L, 236L, 236L),
                              on_map = c(FALSE, FALSE, FALSE, FALSE)),
                         row.names = c("m1", "m2", "m3", "m4"), class = "data.frame")

    expect_equal(find_index_snp(snpinfo, "m2"), "m1")
    expect_warning(expect_equal(find_index_snp(snpinfo, "blah"), as.character(NA)))
    expect_equal(find_index_snp(snpinfo, "m1"), "m1")
    expect_equal(find_index_snp(snpinfo, c("m1", "m2", "m3", "m4")), c("m1", "m1", "m3", "m4"))

})

#  plot_ci (taken from broman::ciplot)
#'
#' Effect plot with multiple CIs for different groups
#'
#' Plot a set of confidence intervals.
#'
#' @param est Vector of estimates
#' @param se  Vector of standard errors
#' @param lo  Vector of lower values for the intervals
#' @param hi  Vector of upper values for the intervals
#' @param SEmult SE multiplier to create intervals
#' @param labels Labels for the groups (vector of character strings)
#' @param swap_axes If TRUE, have group as y-axis; default (FALSE) has
#' group on x-axis.
#' @param add If TRUE, add the CIs to a previous plot with the same axes
#' @param offset Amount to offset the positions of the confidence intervals
#' (particularly for the case `add=TRUE`)
#' @param ... Optional graphics arguments
#'
#' @details
#' Provide either `se` or both `lo` and `hi`. In the case that `se` is
#' used, the intervals will be `est` +/- `SEmult * se`.
#'
#' If `labels` is not provided, group names are taken from the `names(est)`.
#' If that is also missing, we use capital letters.
#'
#' You can control the CI line widths with `ci_lwd` and the color of
#' the CI segments with `ci_col`. You can control the width of the
#' segments at the top and bottom with `ci_endseg`.
#'
#' @importFrom graphics axis points segments
#' @export
#'
#' @return
#' None.
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' iron <- iron[,"9"] # only chr 9
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#' pr <- pull_genoprobpos(probs, map, chr="9", pos=56.6)
#' out_fit1 <- fit1(pr, iron$pheno[,"spleen"], se=TRUE, zerosum=FALSE)
#' plot_ci(out_fit1$coef[1:3], out_fit1$SE[1:3],
#'         xlab="Genotype chr 9 @ 56.6 cM", ylab="Spleen phenotype")
#'
#' @seealso [plot_coef()], [plot_pxg()], [fit1()], [pull_genoprobpos()]
#'
#' @keywords
#' graphics
plot_ci <-
    function(est, se=NULL, lo=NULL, hi=NULL, SEmult=2,
             labels=NULL, swap_axes=FALSE, add=FALSE, offset=0, ...)
{
    if(is.null(se) && is.null(lo) && is.null(hi)) {
        se <- rep(0, length(est))
    }
    if(!is.null(se) && (!is.null(lo) || !is.null(hi))) {
        warning("Provide either se or both lo and hi; se is being used")
    }
    if(!is.null(se)) {
        stopifnot(length(se) == length(est))

        lo <- est - se*SEmult
        hi <- est + se*SEmult
    } else {
        stopifnot(length(lo) == length(est), length(hi) == length(est))
    }

    if(is.null(labels)) {
        labels <- names(est)
    }
    if(is.null(labels)) {
        labels <- rep(LETTERS, length(est))[seq_along(est)]
    }
    stopifnot(length(labels) == length(est))

    n_group <- length(est)
    group <- seq_len(n_group)

    # this is to deal with varying inputs
    hide_ciplot <-
        function(est, lo, hi, swap_axes=FALSE,
                 vlines=NULL, vlines.col="white", vlines.lwd=1,
                 hlines=NULL, hlines.col="white", hlines.lwd=1,
                 xat=NULL, xlim=NULL, xaxs="r", xlab=NULL,
                 yat=NULL, ylim=NULL, yaxs="r", ylab=NULL,
                 las=1, pch=21, bg="slateblue", ci_col="black",
                 ci_lwd=2, ci_endseg=0.05, labels=NULL, main="",
                 mgp.x=NULL, mgp.y=NULL, mgp=NULL, add=FALSE,
                 offset=0, ...)

        {
            if(!add) {
                if(!swap_axes) {
                    xlim <- c(0.5, n_group+0.5)
                    vlines <- 1:n_group
                    vlines.col <- "gray70"
                    vlines.lwd <- 4
                    xat <- NA
                    if(is.null(xlab)) xlab <- ""

                    if(is.null(ylim)) ylim <- range(c(lo, hi, est))

                    # deal with vlines/lines ** FIX ME **

                    qtl2_grayplot(group, est,
                                  vlines=vlines, vlines.col=vlines.col, vlines.lwd=vlines.lwd,
                                  hlines=hlines, hlines.col=hlines.col, hlines.lwd=hlines.lwd,
                                  xat=xat, xlim=xlim, xaxs=xaxs, xlab=xlab,
                                  yat=yat, ylim=ylim, yaxs=yaxs, ylab=ylab, las=las, type="n",
                                  main=main, mgp=mgp, mgp.x=mgp.x, mgp.y=mgp.y, ...)
                    axis(side=1, at=vlines, labels, las=las, tick=FALSE, mgp=c(0,0.2,0))

                }
                else {
                    ylim <- c(0.5, n_group+0.5)
                    hlines <- 1:n_group
                    hlines.col <- "gray70"
                    hlines.lwd <- 4
                    yat <- NA
                    if(is.null(ylab)) ylab <- ""

                    if(is.null(xlim)) xlim <- range(c(lo, hi, est))

                    qtl2_grayplot(est, group,
                                  vlines=vlines, vlines.col=vlines.col, vlines.lwd=vlines.lwd,
                                  hlines=hlines, hlines.col=hlines.col, hlines.lwd=hlines.lwd,
                                  xat=xat, xlim=xlim, xaxs=xaxs, xlab=xlab,
                                  yat=yat, ylim=ylim, yaxs=yaxs, ylab=ylab,
                                  v_over_h=TRUE, las=las, type="n", main=main,
                                  mgp=mgp, mgp.x=mgp.x, mgp.y=mgp.y, ...)
                    axis(side=2, at=hlines, labels, las=las, tick=FALSE, mgp=c(0,0.3,0))
                }
            } # end if(!add)

            if(!swap_axes) {
                segments(group+offset, lo, group+offset, hi, col=ci_col, lwd=ci_lwd)
                segments(group+offset-ci_endseg, lo, group+offset+ci_endseg, lo, col=ci_col, lwd=ci_lwd)
                segments(group+offset-ci_endseg, hi, group+offset+ci_endseg, hi, col=ci_col, lwd=ci_lwd)

                points(group+offset, est, pch=pch, bg=bg, ...)
            } else {
                segments(lo, group+offset, hi, group+offset, col=ci_col, lwd=ci_lwd)
                segments(lo, group+offset-ci_endseg, lo, group+offset+ci_endseg, col=ci_col, lwd=ci_lwd)
                segments(hi, group+offset-ci_endseg, hi, group+offset+ci_endseg, col=ci_col, lwd=ci_lwd)

                points(est, group+offset, pch=pch, bg=bg, ...)
            }
        }

    hide_ciplot(est=est, lo=lo, hi=hi, swap_axes=swap_axes, labels=labels,
                add=add, offset=offset, ...)

    invisible()
}

#' Create table of top snp associations
#'
#' Create a table of the top snp associations
#'
#' @param scan1_output Output of [scan1()].
#' Should contain a component `"snpinfo"`, as when
#' [scan1()] is run with SNP probabilities
#' produced by [genoprob_to_snpprob()].
#'
#' @param snpinfo Data frame with SNP information with the following
#'     columns (the last three are generally derived with
#'     [index_snps()]):
#' * `chr` - Character string or factor with chromosome
#' * `pos` - Position (in same units as in the `"map"`
#'     attribute in `genoprobs`.
#' * `sdp` - Strain distribution pattern: an integer, between
#'     1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
#'     binary encoding indicates the founder genotypes
#' * `snp` - Character string with SNP identifier (if
#'     missing, the rownames are used).
#' * `index` - Indices that indicate equivalent
#'     groups of SNPs, calculated by [index_snps()].
#' * `intervals` - Indexes that indicate which marker
#'     intervals the SNPs reside.
#' * `on_map` - Indicate whether SNP coincides with a marker
#'     in the `genoprobs`
#'
#' @param lodcolumn Selected LOD score column to (a numeric index, or a
#' character string for a column name). Only one value allowed.
#'
#' @param chr Selected chromosome; only one value allowed.
#'
#' @param show_all_snps If TRUE, expand to show all SNPs.
#'
#' @param drop Show all SNPs with LOD score within this amount of the
#' maximum SNP association.
#'
#' @return Data frame like the input `snpinfo` with just the selected
#' subset of rows, and with an added column with the LOD score.
#'
#' @examples
#' \dontrun{
#' # load example DO data from web
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#'
#' # subset to chr 2
#' DOex <- DOex[,"2"]
#'
#' # calculate genotype probabilities and convert to allele probabilities
#' pr <- calc_genoprob(DOex, error_prob=0.002)
#' apr <- genoprob_to_alleleprob(pr)
#'
#' # query function for grabbing info about variants in region
#' dbfile <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
#' query_variants <- create_variant_query_func(dbfile)
#'
#' # SNP association scan, keep information on all SNPs
#' out_snps <- scan1snps(apr, DOex$pmap, DOex$pheno, query_func=query_variants,
#'                       chr=2, start=97, end=98, keep_all_snps=TRUE)
#'
#' # table with top SNPs
#' top_snps(out_snps$lod, out_snps$snpinfo)
#'
#' # top SNPs among the distinct subset at which calculations were performed
#' top_snps(out_snps$lod, out_snps$snpinfo, show_all_snps=FALSE)
#'
#' # top SNPs within 0.5 LOD of max
#' top_snps(out_snps$lod, out_snps$snpinfo, drop=0.5)
#' }
#' @export
#' @seealso [index_snps()], [genoprob_to_snpprob()], [scan1snps()], [plot_snpasso()]
top_snps <-
    function(scan1_output, snpinfo, lodcolumn=1, chr=NULL, drop=1.5, show_all_snps=TRUE)
{
    if(is.null(scan1_output)) stop("scan1_output is NULL")
    if(is.null(snpinfo)) stop("snpinfo is NULL")

    # pull out lod scores
    if(length(lodcolumn)==0) stop("lodcolumn has length 0")
    if(length(lodcolumn) > 1) { # If length > 1, take first value
        warning("lodcolumn should have length 1; only first element used.")
        lodcolumn <- lodcolumn[1]
    }
    if(is.character(lodcolumn)) { # turn column name into integer
        tmp <- match(lodcolumn, colnames(scan1_output))
        if(is.na(tmp)) stop('lodcolumn "', lodcolumn, '" not found')
        lodcolumn <- tmp
    }
    if(lodcolumn < 1 || lodcolumn > ncol(scan1_output))
        stop("lodcolumn [", lodcolumn, "] out of range (should be in 1, ..., ", ncol(scan1_output), ")")
    scan1_output <- scan1_output[,lodcolumn,drop=FALSE]

    # reduce to one chromosome
    if(is.null(chr)) {
        uchr <- unique(snpinfo$chr)
        if(length(uchr) > 1) {
            warning("Considering only chr ", uchr[1])
        }
        chr <- uchr[1]
    }
    if(length(chr) > 1) {
        chr <- chr[1]
        warning("Considering only chr ", chr)
    }
    snpinfo <- snpinfo[snpinfo$chr == chr,,drop=FALSE]
    scan1_output <- scan1_output[rownames(scan1_output) %in% snpinfo$snp,,drop=FALSE]

    # check index business
    uindex <- unique(snpinfo$index)
    if(length(uindex) != nrow(scan1_output))
        stop("Something is wrong with snpinfo$index.\n",
             "      length(unique(snpinfo$index)) [",
             length(unique(snpinfo$index)), "] != nrow(scan1_output) [",
             length(scan1_output), "].")

    if(any(snpinfo$index[uindex] != uindex))
        stop("Something is wrong with snpinfo$index.\n",
             "      snpinfo$index[u] should == u for values in snpinfo$index")

    map <- snpinfo_to_map(snpinfo)

    keep <- which(!is.na(scan1_output[,1]) & scan1_output[,1] >= max(scan1_output[ ,1], na.rm=TRUE) - drop)

    # reverse the snp index
    revindex <- rev_snp_index(snpinfo)

   if(show_all_snps) { # expand to all related SNPs
        snpinfo <- snpinfo[revindex %in% keep,,drop=FALSE]
        revindex <- revindex[revindex %in% keep]
        snpinfo$lod <- scan1_output[revindex]
    } else { # just keep the SNPs that were used
        snpinfo$lod <- scan1_output[revindex]
        snpinfo <- snpinfo[snpinfo$snp %in% rownames(scan1_output)[keep],]
    }

    snpinfo
}

# snpinfo to map
snpinfo_to_map <-
    function(snpinfo)
{
    uindex <- sort(unique(snpinfo$index))
    if(any(snpinfo$index < 1 | snpinfo$index > nrow(snpinfo)))
        stop("snpinfo$index values outside of range [1, ",
             nrow(snpinfo), "]")

    uchr <- unique(snpinfo$chr)
    chr <- factor(snpinfo$chr, levels=uchr)

    map <- split(snpinfo$pos, chr)
    snp <- split(snpinfo$snp, chr)
    index <- split(snpinfo$index, chr)
    for(i in seq_along(map)) {
        u <- unique(index[[i]])
        map[[i]] <- map[[i]][u]
        names(map[[i]]) <- snp[[i]][u]
    }

    names(map) <- uchr

    map
}

# reverse index
rev_snp_index <-
    function(snpinfo)
{
    index_spl <- split(seq_len(nrow(snpinfo)), snpinfo$index)
    revindex <- rep(seq_along(index_spl), vapply(index_spl, length, 1))
    revindex[unlist(index_spl)] <- revindex

    revindex
}

# calc_kinship
#' Calculate kinship matrix
#'
#' Calculate genetic similarity among individuals (kinship matrix)
#' from conditional genotype probabilities.
#'
#' @param probs Genotype probabilities, as calculated from
#' [calc_genoprob()].
#' @param type Indicates whether to calculate the overall kinship
#' (`"overall"`, using all chromosomes), the kinship matrix
#' leaving out one chromosome at a time (`"loco"`), or the
#' kinship matrix for each chromosome (`"chr"`).
#' @param omit_x If `TRUE`, only use the autosomes; ignored when
#' `type="chr"`.
#' @param use_allele_probs If `TRUE`, assess similarity with
#' allele probabilities (that is, first run
#' [genoprob_to_alleleprob()]); otherwise use the genotype
#' probabilities.
#' @param quiet IF `FALSE`, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return If `type="overall"` (the default), a matrix of
#' proportion of matching alleles. Otherwise a list with one matrix
#' per chromosome.
#'
#' @details If `use_allele_probs=TRUE` (the default), we first
#' convert the genotype probabilities to allele
#' probabilities (using [genoprob_to_alleleprob()]).
#'
#' We then calculate
#' \eqn{\sum_{kl}(p_{ikl} p_{jkl})}{sum_kl (p_ikl p_jkl)}
#' where \eqn{k} = position, \eqn{l} = allele, and \eqn{i,j}
#' are two individuals.
#'
#' For crosses with just two possible genotypes (e.g., backcross), we
#' don't convert to allele probabilities but just use the original
#' genotype probabilities.
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' probs <- calc_genoprob(grav2, map, error_prob=0.002)
#' K <- calc_kinship(probs)
#'
#' # using only markers/pseudomarkers on the grid
#' grid <- calc_grid(grav2$gmap, step=1)
#' probs_sub <- probs_to_grid(probs, grid)
#' K_grid <- calc_kinship(probs_sub)


calc_kinship <-
    function(probs, type=c("overall", "loco", "chr"),
             omit_x=FALSE, use_allele_probs=TRUE,
             quiet=TRUE, cores=1)
{
    if(is.null(probs)) stop("probs is NULL")
    if(is.cross2(probs))
        stop('Input probs is a "cross2" object but should be genotype probabilities, as from calc_genoprob')

    type <- match.arg(type)

    # make sure probs has chromosome names
    if(is.null(names(probs))) {
        names(probs) <- as.character(seq_len(probs))
        warning("probs is missing chromosome names.")
    }
    allchr <- names(probs)

    if(omit_x && type != "chr") chrs <- which(!attr(probs, "is_x_chr"))
    else chrs <- seq(along=allchr)

    # convert from genotype probabilities to allele probabilities
    ap <- attr(probs, "alleleprobs")
    if(use_allele_probs && (is.null(ap) || !ap)) {
        if(!quiet) message(" - converting to allele probs")
        probs <- genoprob_to_alleleprob(probs, quiet=quiet, cores=cores)
    }

    if(type=="overall") {
        K <- calc_kinship_overall(probs, chrs=chrs, quiet=quiet, cores=cores)
    }
    else if(type=="chr") {
        K <- calc_kinship_bychr(probs, chrs=chrs, scale=TRUE, quiet=quiet, cores=cores)
    }
    else {
        # otherwise LOCO (leave one chromosome out)
        result <- calc_kinship_bychr(probs, chrs=chrs, scale=FALSE, quiet=quiet, cores=cores)
        K <- kinship_bychr2loco(result, allchr)
    }

    K
}

# calculate an overall kinship matrix
calc_kinship_overall <-
    function(probs, chrs, quiet=TRUE, cores=1)
{
    ind_names <- rownames(probs[[1]])
    n_ind <- length(ind_names)

    result <- matrix(0, nrow=n_ind, ncol=n_ind)
    dimnames(result) <- list(ind_names, ind_names)

    # set up cluster; set quiet=TRUE if multi-core
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # function that does the work
    by_chr_func <- function(chr) {
        if(!quiet) message(" - Chr ", names(probs)[chr])
        pr <- aperm(probs[[chr]], c(3,2,1)) # convert to pos x gen x ind
        .calc_kinship(pr)
    }

    # run and combine results
    if(n_cores(cores) == 1) {
        for(chr in chrs)
            result <- result + by_chr_func(chr)
    }
    else {
        by_chr_res <- cluster_lapply(cores, chrs, by_chr_func)
        for(i in seq(along=by_chr_res))
            result <- result + by_chr_res[[i]]
    }

    tot_pos <- sum(dim(probs)[3,chrs])
    result <- result/tot_pos
    attr(result, "n_pos") <- tot_pos

    result
}

# calculate kinship for each chromosome
calc_kinship_bychr <-
    function(probs, chrs, scale=TRUE, quiet=TRUE, cores=1)
{
    ind_names <- rownames(probs[[1]])
    n_ind <- length(ind_names)

    # set up cluster and set quiet=TRUE if multi-core
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # function that does the work
    by_chr_func <- function(chr) {
        if(!quiet) message(" - Chr ", names(probs)[chr])

        n_pos <- dim(probs[[chr]])[3]

        # aperm converts to pos x gen x ind
        pr <- aperm(probs[[chr]], c(3,2,1))
        result <- .calc_kinship(pr)
        if(scale) result <- result/n_pos

        attr(result, "n_pos") <- n_pos
        dimnames(result) <- list(ind_names, ind_names)
        result
    }

    # run and combine results
    result <- cluster_lapply(cores, chrs, by_chr_func)

    names(result) <- names(probs)[chrs]
    result
}

# use kinship for each chromosome
# to calculate kinship leaving one chromosome out at a time
kinship_bychr2loco <-
    function(kinship, allchr)
{
    # sum over chromosomes
    overall <- kinship[[1]]
    tot_pos <- attr(kinship[[1]], "n_pos")
    if(length(kinship) > 1) {
        for(i in 2:length(kinship)) {
            overall <- overall + kinship[[i]]
            tot_pos <- tot_pos + attr(kinship[[i]], "n_pos")
        }
    }
    attr(overall, "n_pos") <- tot_pos

    for(chr in allchr) {
        if(chr %in% names(kinship)) {
            n_pos <- attr(kinship[[chr]], "n_pos")
            kinship[[chr]] <- (overall - kinship[[chr]])/(tot_pos - n_pos)
            attr(kinship[[chr]], "n_pos") <- tot_pos - n_pos
        } else {
            kinship <- c(kinship, list(overall/tot_pos))
            names(kinship)[length(kinship)] <- chr
        }
    }

    # make sure it's in the right order
    kinship[allchr]
}

# normalize kinship as in equation 5 in Kang et al. (2010) Nat Genet '
# 42:348-354. https://www.ncbi.nlm.nih.gov/pubmed/20208533 (doi: 10.1038/ng.548)
#
# (suggested by Petr Simecek)
normalize_kinship <-
    function(kinship)
{
    if(is.list(kinship)) {
        return(lapply(kinship, normalize_kinship))
    }

    n <- nrow(kinship)
    norm_const <- (sum(diag(kinship)) - sum(kinship)/n) / (n-1)
    kinship/norm_const
}

# compare_geno
#' Compare individuals' genotype data
#'
#' Count the number of matching genotypes between all pairs of
#' individuals, to look for unusually closely related individuals.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param omit_x If TRUE, only use autosomal genotypes
#' @param proportion If TRUE (the default), the upper triangle of the
#'     result contains the proportions of matching genotypes. If
#'     FALSE, the upper triangle contains counts of matching
#'     genotypes.
#' @param quiet IF `FALSE`, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A square matrix; diagonal is number of observed genotypes
#' for each individual. The values in the lower triangle are the
#' numbers of markers where both of a pair were genotyped. The
#' values in the upper triangle are either proportions or counts
#' of matching genotypes for each pair (depending on whether
#' `proportion=TRUE` or `=FALSE`). The object is given
#' class `"compare_geno"`.
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' cg <- compare_geno(grav2)
#' summary(cg)

compare_geno <-
    function(cross, omit_x=FALSE, proportion=TRUE, quiet=TRUE, cores=1)
{
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')

    # which chr is X?
    is_x_chr <- cross$is_x_chr
    if(is.null(is_x_chr))
        is_x_chr <- rep(FALSE, length(cross$geno))

    # chr names and numeric indices
    chrnames <- names(cross$geno)
    chrnum <- seq_along(chrnames)
    if(omit_x) { # don't use the X chromosome
        chrnames <- chrnames[!is_x_chr]
        chrnum <- chrnum[!is_x_chr]
    }

    ind_names <- rownames(cross$geno[[1]])
    n_ind <- length(ind_names)

    result <- matrix(0, nrow=n_ind, ncol=n_ind)
    dimnames(result) <- list(ind_names, ind_names)

    # set up cluster; set quiet=TRUE if multi-core
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # function that does the work
    by_chr_func <- function(chr) {
        if(!quiet) message(" - Chr ", chrnames[chr])
        .compare_geno(t(cross$geno[[chr]]))
    }

    # run and combine results
    if(n_cores(cores) == 1) {
        for(chr in chrnum)
            result <- result + by_chr_func(chr)
    }
    else {
        result_list <- cluster_lapply(cores, chrnum, by_chr_func)
        for(i in seq_along(result_list))
            result <- result + result_list[[i]]
    }

    # upper triangle from count -> proportion
    if(proportion) {
        result[upper.tri(result)] <- result[upper.tri(result)]/t(result)[upper.tri(result)]
        result[is.nan(result)] <- NA
    }

    attr(result, "proportion") <- proportion
    class(result) <- c("compare_geno", "matrix")
    result
}


#' Basic summary of compare_geno object
#'
#' From results of [compare_geno()], show pairs of individuals with similar genotypes.
#'
#' @param object A square matrix with genotype comparisons for pairs
#'     of individuals, as output by [compare_geno()].
#' @param threshold Minimum proportion matches for a pair of individuals to be shown.
#' @param ... Ignored
#'
#' @return Data frame with names of individuals in pair, proportion
#'     matches, number of mismatches, number of matches, and total
#'     markers genotyped. Last two columns are the numeric indexes of
#'     the individuals in the pair.
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' cg <- compare_geno(grav2)
#' summary(cg)

summary_compare_geno <-
    function(object, threshold=0.9, ...)
{
    # does upper triangle contain proportions or not?
    proportion <- attr(object, "proportion")
    if(is.null(proportion)) { # missing attribute; guess
        if(max(object[upper.tri(object)], na.rm=TRUE) <= 1)
            proportion <- TRUE
        else proportion <- FALSE
    }

    # proportion matching as symmetric matrix
    if(proportion) {
        p <- object

        # number matching in upper triangle
        object[upper.tri(object)] <- round(object[upper.tri(object)]*
                                           t(object)[upper.tri(object)])
        object[is.na(object)] <- 0
    } else {
        p <- object
        p[upper.tri(p)] <- object[upper.tri(p)]/t(object)[upper.tri(p)]
    }
    p[lower.tri(p)] <- t(p)[lower.tri(p)]
    diag(p) <- NA

    if(sum(!is.na(p) & p >= threshold) == 0) { # no results
        result <- data.frame(ind1=character(0),
                             ind2=character(0),
                             prop_match=numeric(0),
                             n_mismatch=numeric(0),
                             n_typed=numeric(0),
                             n_match=numeric(0),
                             index1=numeric(0),
                             index2=numeric(0),
                             stringsAsFactors=FALSE)

        class(result) <- c("summary.compare_geno", "data.frame")
        attr(result, "threshold") <- threshold
        return(result)
    }

    wh <- which(!is.na(p) & p >= threshold, arr.ind=TRUE)
    wh <- wh[wh[,1] < wh[,2], , drop=FALSE]

    ind <- rownames(object)

    r <- 1:nrow(wh)
    p_match <- vapply(r, function(i) p[wh[i,1],wh[i,2]], 1)
    n_match <- vapply(r, function(i) object[wh[i,1],wh[i,2]], 1)
    n_typed <- vapply(r, function(i) object[wh[i,2],wh[i,1]], 1)

    result <- data.frame(ind1=ind[wh[,1]],
                         ind2=ind[wh[,2]],
                         prop_match=p_match,
                         n_mismatch=n_typed-n_match,
                         n_typed=n_typed,
                         n_match=n_match,
                         index1=wh[,1],
                         index2=wh[,2],
                         stringsAsFactors=FALSE)

    result <- result[order(result$prop_match, decreasing=TRUE),,drop=FALSE]

    attr(result, "threshold") <- threshold
    class(result) <- c("summary.compare_geno", "data.frame")
    result
}

#' @rdname summary_compare_geno
#' @export
summary.compare_geno <- summary_compare_geno

#' @rdname summary_compare_geno
#'
#' @param x Results of [summary.compare_geno()]
#' @param digits Number of digits to print
#' @export
print.summary.compare_geno <-
    function(x, digits=2, ...)
{
    if(nrow(x) == 0)
        cat("No pairs with proportion matches above ", attr(x, "threshold"), ".\n", sep="")
    else {
        print.data.frame(x, digits=digits, row.names=FALSE)
    }
    invisible(x)
}


#' Find pair with most similar genotypes
#'
#' From results of [compare_geno()], show the pair with most similar genotypes.
#'
#' @param object A square matrix with genotype comparisons for pairs
#'     of individuals, as output by [compare_geno()].
#' @param ... Ignored
#'
#' @return Data frame with individual pair, proportion matches, number
#'     of mismatches, number of matches, and total markers genotyped.
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' cg <- compare_geno(grav2)
#' max(cg)

max_compare_geno <-
    function(object, ...)
{
    # does upper triangle contain proportions or not?
    proportion <- attr(object, "proportion")
    if(is.null(proportion)) { # missing attribute; guess
        if(max(object[upper.tri(object)], na.rm=TRUE) <= 1)
            proportion <- TRUE
        else proportion <- FALSE
    }

    # proportion matching as symmetric matrix
    if(proportion) {
        p <- object

        # number matching in upper triangle
        object[upper.tri(object)] <- round(object[upper.tri(object)]*
                                           t(object)[upper.tri(object)])
        object[is.na(object)] <- 0
    } else {
        p <- object
        p[upper.tri(p)] <- object[upper.tri(p)]/t(object)[upper.tri(p)]
    }
    p[lower.tri(p)] <- t(p)[lower.tri(p)]
    diag(p) <- NA

    # maximum proportion of matches
    max_p <- max(unclass(p), na.rm=TRUE)

    wh <- which(!is.na(p) & p == max_p, arr.ind=TRUE)
    wh <- wh[wh[,1] < wh[,2], , drop=FALSE]

    ind <- rownames(object)

    r <- 1:nrow(wh)
    p_match <- vapply(r, function(i) p[wh[i,1],wh[i,2]], 1)
    n_match <- vapply(r, function(i) object[wh[i,1],wh[i,2]], 1)
    n_typed <- vapply(r, function(i) object[wh[i,2],wh[i,1]], 1)

    result <- data.frame(ind1=ind[wh[,1]],
                         ind2=ind[wh[,2]],
                         prop_match=p_match,
                         n_mismatch=n_typed-n_match,
                         n_typed=n_typed,
                         n_match=n_match,
                         index1=wh[,1],
                         index2=wh[,2],
                         stringsAsFactors=FALSE)

    result <- result[order(result$prop_match, decreasing=TRUE),,drop=FALSE]

    class(result) <- c("summary.compare_geno", "data.frame")
    result
}

#' @rdname max_compare_geno
#' @export
max.compare_geno <- max_compare_geno

#' Plot of compare_geno object.
#'
#' From results of [compare_geno()], plot histogram of
#'
#' @param x A square matrix with genotype comparisons for pairs
#'     of individuals, as output by [compare_geno()].
#' @param rug If true, use [rug()] to plot tick marks at observed values below the histogram.
#' @param ... Additional graphics parameters passed to [hist()]
#'
#' @return None.
#'
#' @export
#' @keywords graphics
#' @importFrom graphics hist rug
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' cg <- compare_geno(grav2)
#' plot(cg)
plot_compare_geno <-
    function(x, rug=TRUE, ...)
{
    .plot_compare_geno <-
        function(x, breaks=seq(0, 1, length=101), xlab="Proportion matching genotypes",
                 main="", las=1, rug=TRUE, ...)
        {
            x <- x[upper.tri(x)]
            hist(x, breaks=breaks, las=las, main=main, xlab=xlab)
            if(rug) rug(x)
        }

    .plot_compare_geno(x, rug=rug, ...)
}

#' @rdname plot_compare_geno
#' @export
plot.compare_geno <- plot_compare_geno


# compare_founder_geno
#' Compare founders genotype data
#'
#' Count the number of matching genotypes between all pairs of
#' founder lines.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param omit_x If TRUE, only use autosomal genotypes
#' @param proportion If TRUE (the default), the upper triangle of the
#'     result contains the proportions of matching genotypes. If
#'     FALSE, the upper triangle contains counts of matching
#'     genotypes.
#' @param quiet IF `FALSE`, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A square matrix; diagonal is number of observed genotypes
#' for each founder line. The values in the lower triangle are the
#' numbers of markers where both of a pair were genotyped. The
#' values in the upper triangle are either proportions or counts
#' of matching genotypes for each pair (depending on whether
#' `proportion=TRUE` or `=FALSE`). The object is given
#' class `"compare_geno"`.
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' cg <- compare_founder_geno(DOex)
#' summary(cg)
#' }

compare_founder_geno <-
    function(cross, omit_x=FALSE, proportion=TRUE, quiet=TRUE, cores=1)
{
    if(!("founder_geno" %in% names(cross))) {
        stop("cross does not contain founder genotypes")
    }

    cross$geno <- cross$founder_geno
    compare_geno(cross, omit_x=omit_x, proportion=proportion, quiet=quiet, cores=cores)
}

# find_peaks
#' Find peaks in a set of LOD curves
#'
#' Find peaks in a set of LOD curves (output from [scan1()]
#'
#' @param scan1_output An object of class `"scan1"` as returned by
#' [scan1()].
#' @param map A list of vectors of marker positions, as produced by
#' [insert_pseudomarkers()]. Can also be an indexed SNP info table,
#' as from [index_snps()] or [scan1snps()].
#' @param threshold Minimum LOD score for a peak (can be a vector with
#' separate thresholds for each lod score column in
#' `scan1_output`)
#' @param peakdrop Amount that the LOD score must drop between peaks,
#' if multiple peaks are to be defined on a chromosome. (Can be a vector with
#' separate values for each lod score column in
#' `scan1_output`.)
#' @param drop If provided, LOD support intervals are included in the
#' results, and this indicates the amount to drop in the support
#' interval. (Can be a vector with
#' separate values for each lod score column in
#' `scan1_output`.) Must be \eqn{\le} `peakdrop`
#' @param prob If provided, Bayes credible intervals are included in the
#' results, and this indicates the nominal coverage.
#' (Can be a vector with
#' separate values for each lod score column in
#' `scan1_output`.) Provide just one of `drop` and `prob`.
#' @param thresholdX Separate threshold for the X chromosome; if
#' unspecified, the same threshold is used for both autosomes and the
#' X chromosome. (Like `threshold`, this can be a vector with
#' separate thresholds for each lod score column.)
#' @param peakdropX Like `peakdrop`, but for the X chromosome; if
#' unspecified, the same value is used for both autosomes and the X
#' chromosome.  (Can be a vector with separate values for each lod
#' score column in `scan1_output`.)
#' @param dropX Amount to drop for LOD support intervals on the X
#' chromosome.  Ignored if `drop` is not provided. (Can be a
#' vector with separate values for each lod score column in
#' `scan1_output`.)
#' @param probX Nominal coverage for Bayes intervals on the X
#' chromosome.  Ignored if `prob` is not provided. (Can be a
#' vector with separate values for each lod score column in
#' `scan1_output`.)
#' @param expand2markers If TRUE (and if `drop` or `prob` is
#' provided, so that QTL intervals are calculated), QTL intervals are
#' expanded so that their endpoints are at genetic markers.
#' @param sort_by Indicates whether to sort the rows by lod column,
#' genomic position, or LOD score.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A data frame with each row being a single peak on a single
#' chromosome for a single LOD score column, and with columns
#' * `lodindex` - lod column index
#' * `lodcolumn` - lod column name
#' * `chr` - chromosome ID
#' * `pos` - peak position
#' * `lod` - lod score at peak
#'
#' If `drop` or `prob` is provided, the results will include
#' two additional columns: `ci_lo` and `ci_hi`, with the
#' endpoints of the LOD support intervals or Bayes credible wintervals.
#'
#' @details For each lod score column on each chromosome, we return a
#' set of peaks defined as local maxima that exceed the specified
#' `threshold`, with the requirement that the LOD score must have
#' dropped by at least `peakdrop` below the lowest of any two
#' adjacent peaks.
#'
#' At a given peak, if there are ties, with multiple positions jointly
#' achieving the maximum LOD score, we take the average of these
#' positions as the location of the peak.
#'
#' @export
#'
#' @seealso [scan1()], [lod_int()], [bayes_int()]
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c(1,2,7,8,9,13,15,16,19)]}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' # find just the highest peak on each chromosome
#' find_peaks(out, map, threshold=3)
#'
#' # possibly multiple peaks per chromosome
#' find_peaks(out, map, threshold=3, peakdrop=1)
#'
#' # possibly multiple peaks, also getting 1-LOD support intervals
#' find_peaks(out, map, threshold=3, peakdrop=1, drop=1)
#'
#' # possibly multiple peaks, also getting 90% Bayes intervals
#' find_peaks(out, map, threshold=3, peakdrop=1, prob=0.9)
find_peaks <-
    function(scan1_output, map, threshold=3, peakdrop=Inf, drop=NULL, prob=NULL,
             thresholdX=NULL, peakdropX=NULL, dropX=NULL, probX=NULL,
             expand2markers=TRUE, sort_by=c("column", "pos", "lod"), cores=1)
{
    sort_by <- match.arg(sort_by)

    # check if map input is a snpinfo table; if so convert to
    if(is.data.frame(map) && "index" %in% names(map)) { # looks like snpinfo table
        map <- snpinfo_to_map(map)
    }

    # align scan1_output and map
    tmp <- align_scan1_map(scan1_output, map)
    scan1_output <- tmp$scan1
    map <- tmp$map

    if(nrow(scan1_output) != length(unlist(map)))
        stop("nrow(scan1_output) [", nrow(scan1_output), "] != number of positions in map [",
             length(unlist(map)), "]")

    if(!is.null(drop) && !is.null(prob))
        stop('No more than one of "drop" and "prob" should be provided')

    if(!is.null(drop)) # also include lod support intervals
        return(find_peaks_and_lodint(scan1_output, map, threshold, peakdrop, drop,
                                     thresholdX, peakdropX, dropX,
                                     expand2markers, sort_by, cores))
    if(!is.null(prob)) # also include Bayes credible intervals
        return(find_peaks_and_bayesint(scan1_output, map, threshold, peakdrop, prob,
                                       thresholdX, peakdropX, probX,
                                       expand2markers, sort_by, cores))

    if(!is.null(dropX))
        warning("dropX ignored if drop is not provided")
    if(!is.null(probX))
        warning("probX ignored if prob is not provided")

    lodnames <- colnames(scan1_output)
    if(is.null(lodnames)) {
        warning('scan1_output has no column names; using "lod[#]"')
        lodnames <- colnames(scan1_output) <- paste0("lod", 1:ncol(scan1_output))
    }
    n_lod <- length(lodnames)

    # make the thresholds have length n_lod
    if(length(threshold)==1) threshold <- rep(threshold, n_lod)
    if(length(threshold) != n_lod)
        stop("threshold should have length 1 or ", n_lod)
    if(is.null(thresholdX)) thresholdX <- threshold
    if(length(thresholdX)==1) thresholdX <- rep(thresholdX, n_lod)
    if(length(thresholdX) != n_lod)
        stop("thresholdX should have length 1 or ", n_lod)

    # make the peakdrops have length n_lod
    if(length(peakdrop)==1) peakdrop <- rep(peakdrop, n_lod)
    if(length(peakdrop) != n_lod)
        stop("peakdrop should have length 1 or ", n_lod)
    if(is.null(peakdropX)) peakdropX <- peakdrop
    if(length(peakdropX)==1) peakdropX <- rep(peakdropX, n_lod)
    if(length(peakdropX) != n_lod)
        stop("peakdropX should have length 1 or ", n_lod)

    # split lod scores by column and by chromosome
    lod <- as.list(as.data.frame(unclass(scan1_output)))
    chr <- rep(seq_along(map), vapply(map, length, 1))
    lod <- lapply(lod, split, chr)

    # batch info for parallel-processing
    batch <- cbind(rep(seq_along(lod), vapply(lod, length, 1)),
                   unlist(lapply(lod, function(a) seq_along(a))))
    dimnames(batch) <- NULL
    batch_index <- seq_len(nrow(batch))

    # set up parallel analysis
    cores <- setup_cluster(cores)

    # X chr info
    is_x_chr <- attr(map, "is_x_chr")
    if(is.null(is_x_chr))
        is_x_chr <- rep(FALSE, length(map))

    # function to be applied to each batch
    by_batch_func <- function(i) {
        lodcol <- batch[i,1]
        chr <- batch[i,2]

        this_lod <- lod[[lodcol]][[chr]]

        peaks <- .find_peaks(this_lod,
                             ifelse(is_x_chr[chr], thresholdX[lodcol], threshold[lodcol]),
                             ifelse(is_x_chr[chr], peakdropX[lodcol], peakdrop[lodcol]))

        n_peaks <- length(peaks)
        if(n_peaks==0) return(NULL)

        # calculate peak positions
        # (this deals with ties in the LOD scores:
        #  take average of multiple positions sharing the maximum LOD score)
        # and remember that .find_peaks returns indexes starting at 0
        peak_pos <- vapply(peaks, function(a) mean(map[[chr]][a+1]), 0.0)
        peak_lod <- vapply(peaks, function(a) this_lod[a[1]+1], 0.0)

        data.frame(lodindex=rep(lodcol,n_peaks),
                   lodcolumn=rep(lodnames[lodcol],n_peaks),
                   chr=rep(names(map)[chr],n_peaks),
                   pos=peak_pos,
                   lod=peak_lod,
                   row.names=seq_along(peaks),
                   stringsAsFactors=FALSE)
    }

    if(n_cores(cores)==1) {
        peaks <- lapply(batch_index, by_batch_func)
    } else {
        peaks <- cluster_lapply(cores, batch_index, by_batch_func)
    }

    # empty data frame to start
    result <- data.frame(lodindex=numeric(0),
                         lodcolumn=character(0),
                         chr=character(0),
                         pos=numeric(0),
                         lod=numeric(0),
                         row.names=character(0),
                         stringsAsFactors=FALSE)

    for(p in peaks) result <- rbind(result, p)

    rownames(result) <- NULL
    result$chr <- factor(result$chr, names(map))

    if(nrow(result) > 1) {
        if(sort_by == "pos") {

            result <- result[order(result$chr, result$pos, result$lodindex), ]

        } else if(sort_by == "lod") {

            result <- result[order(result$lod, decreasing=TRUE),]

        }
    }

    result
}

#' Maximum LOD score from genome scan with a single-QTL model
#'
#' Maximum LOD score from genome scan with a single-QTL model by
#' Haley-Knott regression or a linear mixed model, with possible
#' allowance for covariates.
#'
#' @param genoprobs Genotype probabilities as calculated by
#' [calc_genoprob()].
#' @param pheno A numeric matrix of phenotypes, individuals x phenotypes.
#' @param kinship Optional kinship matrix, or a list of kinship matrices (one
#' per chromosome), in order to use the LOCO (leave one chromosome
#' out) method.
#' @param addcovar An optional numeric matrix of additive covariates.
#' @param Xcovar An optional numeric matrix with additional additive covariates used for
#' null hypothesis when scanning the X chromosome.
#' @param intcovar An numeric optional matrix of interactive covariates.
#' @param weights An optional numeric vector of positive weights for the
#' individuals. As with the other inputs, it must have `names`
#' for individual identifiers.
#' @param reml If `kinship` provided: if `reml=TRUE`, use
#' REML; otherwise maximum likelihood.
#' @param model Indicates whether to use a normal model (least
#'     squares) or binary model (logistic regression) for the phenotype.
#'     If `model="binary"`, the phenotypes must have values in \eqn{[0, 1]}.
#' @param hsq Considered only if `kinship` is provided, in which case
#'     this is taken as the assumed value for the residual
#'     heritability. It should be a vector with length corresponding
#'     to the number of columns in `pheno`, or (if `kinship`
#'     corresponds to a list of LOCO kinship matrices) a matrix with dimension
#'     `length(kinship) x ncol(pheno)`.
#' @param by_chr If TRUE, save the individual chromosome maxima.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#' @param ... Additional control parameters; see Details.
#'
#' @return Either a vector of genome-wide maximum LOD scores, or if
#'     `by_chr` is TRUE, a matrix with the chromosome-specific maxima,
#'     with the rows being the chromosomes and the columns being the
#'     phenotypes.
#'
#' @details Equivalent to running `scan1()` and then saving the column
#'     maxima, with some savings in memory usage.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c("19", "X")] # subset to chr 19 and X}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # perform genome scan
#' out <- scan1max(probs, pheno, addcovar=covar, Xcovar=Xcovar)
#'
#' @seealso [scan1()], [scan1perm()]
#'
#' @export
scan1max <-
    function(genoprobs, pheno, kinship=NULL, addcovar=NULL, Xcovar=NULL,
             intcovar=NULL, weights=NULL, reml=TRUE,
             model=c("normal", "binary"), hsq=NULL,
             by_chr=FALSE, cores=1, ...)
{
    if(is.null(genoprobs)) stop("genoprobs is NULL")
    if(is.null(pheno)) stop("pheno is NULL")

    # grab dot args
    dotargs <- list(...)
    if("n_perm" %in% names(dotargs))
        stop("You included n_perm as an argument; you probably want to run scan1perm not scan1.")

    model <- match.arg(model)

    if(!is.null(kinship)) { # fit linear mixed model
        if(model=="binary") warning("Can't fit binary model with kinship matrix; using normal model")
        return(scan1max_pg(genoprobs, pheno, kinship, addcovar, Xcovar, intcovar,
                           weights, reml, hsq, by_chr, cores, ...))
    }

    # deal with the dot args
    tol <- grab_dots(dotargs, "tol", 1e-12)
    if(!is_pos_number(tol)) stop("tol should be a single positive number")
    intcovar_method <- grab_dots(dotargs, "intcovar_method", "lowmem",
                                 c("highmem", "lowmem"))
    quiet <- grab_dots(dotargs, "quiet", TRUE)
    max_batch <- grab_dots(dotargs, "max_batch", NULL)
    if(!is.null(max_batch) && !is_pos_number(max_batch)) stop("max_batch should be a single positive integer")
    if(model=="binary") {
        bintol <- grab_dots(dotargs, "bintol", sqrt(tol)) # for model="binary"
        if(!is_pos_number(bintol)) stop("bintol should be a single positive number")
        eta_max <- grab_dots(dotargs, "eta_max", log(1-tol)-log(tol)) # for model="binary"
        if(!is_pos_number(eta_max)) stop("eta_max should be a single positive number")
        maxit <- grab_dots(dotargs, "maxit", 100) # for model="binary"
        if(!is_nonneg_number(maxit)) stop("maxit should be a single non-negative integer")
        check_extra_dots(dotargs, c("tol", "intcovar_method", "quiet", "max_batch", "maxit", "bintol", "eta_max"))
    }
    else {
        check_extra_dots(dotargs, c("tol", "intcovar_method", "quiet", "max_batch"))
    }

    # check that the objects have rownames
    check4names(pheno, addcovar, Xcovar, intcovar)

    # force things to be matrices
    if(!is.matrix(pheno)) {
        pheno <- as.matrix(pheno)
        if(!is.numeric(pheno)) stop("pheno is not numeric")
    }
    if(is.null(colnames(pheno))) # force column names
        colnames(pheno) <- paste0("pheno", seq_len(ncol(pheno)))
    if(!is.null(addcovar)) {
        if(!is.matrix(addcovar)) addcovar <- as.matrix(addcovar)
        if(!is.numeric(addcovar)) stop("addcovar is not numeric")
    }
    if(!is.null(Xcovar)) {
        if(!is.matrix(Xcovar)) Xcovar <- as.matrix(Xcovar)
        if(!is.numeric(Xcovar)) stop("Xcovar is not numeric")
    }
    if(!is.null(intcovar)) {
        if(!is.matrix(intcovar)) intcovar <- as.matrix(intcovar)
        if(!is.numeric(intcovar)) stop("intcovar is not numeric")
    }

    # for binary model
    if(model=="binary") {
        if(!is.null(kinship))
            stop("Can't yet account for kinship with model = \"binary\"")
        pheno <- check_binary_pheno(pheno)
    }
    else {
        # square-root of weights (only if model="normal")
        weights <- sqrt_weights(weights) # also check >0 (and if all 1's, turn to NULL)
    }

    # find individuals in common across all arguments
    # and drop individuals with missing covariates or missing *all* phenotypes
    ind2keep <- get_common_ids(genoprobs, addcovar, Xcovar, intcovar,
                               weights, complete.cases=TRUE)
    ind2keep <- get_common_ids(ind2keep, pheno[rowSums(is.finite(pheno)) > 0,,drop=FALSE])
    if(length(ind2keep)<=2) {
        if(length(ind2keep)==0)
            stop("No individuals in common.")
        else
            stop("Only ", length(ind2keep), " individuals in common: ",
                 paste(ind2keep, collapse=":"))
    }

    # make sure addcovar is full rank when we add an intercept
    addcovar <- drop_depcols(addcovar, TRUE, tol)

    # make sure columns in intcovar are also in addcovar
    addcovar <- force_intcovar(addcovar, intcovar, tol)

    # drop things from Xcovar that are already in addcovar
    Xcovar <- drop_xcovar(addcovar, Xcovar, tol)

    # batch phenotypes by missing values
    phe_batches <- batch_cols(pheno[ind2keep,,drop=FALSE], max_batch)

    # drop cols in genotype probs that are all 0 (just looking at the X chromosome)
    genoprob_Xcol2drop <- genoprobs_col2drop(genoprobs)
    is_x_chr <- attr(genoprobs, "is_x_chr")
    if(is.null(is_x_chr)) is_x_chr <- rep(FALSE, length(genoprobs))

    # set up parallel analysis
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    n_chr <- length(genoprobs)

    # batches for analysis, to allow parallel analysis
    run_batches <- data.frame(chr=rep(seq_len(length(genoprobs)), length(phe_batches)),
                              phe_batch=rep(seq_along(phe_batches), each=length(genoprobs)))
    run_indexes <- seq_len(length(genoprobs)*length(phe_batches))

    # the function that does the work
    by_group_func <- function(i) {
        # deal with batch information, including individuals to drop due to missing phenotypes
        chr <- run_batches$chr[i]
        chrnam <- names(genoprobs)[chr]
        phebatch <- phe_batches[[run_batches$phe_batch[i]]]
        phecol <- phebatch$cols
        omit <- phebatch$omit
        these2keep <- ind2keep # individuals 2 keep for this batch
        if(length(omit) > 0) these2keep <- ind2keep[-omit]
        if(length(these2keep)<=2) return(NULL) # not enough individuals

        # subset the genotype probabilities: drop cols with all 0s, plus the first column
        Xcol2drop <- genoprob_Xcol2drop[[chrnam]]
        if(length(Xcol2drop) > 0) {
            pr <- genoprobs[[chr]][these2keep,-Xcol2drop,,drop=FALSE]
            pr <- pr[,-1,,drop=FALSE]
        }
        else
            pr <- genoprobs[[chr]][these2keep,-1,,drop=FALSE]

        # subset the rest
        ac <- addcovar; if(!is.null(ac)) { ac <- ac[these2keep,,drop=FALSE]; ac <- drop_depcols(ac, TRUE, tol) }
        Xc <- Xcovar;   if(!is.null(Xc)) Xc <- Xc[these2keep,,drop=FALSE]
        ic <- intcovar; if(!is.null(ic)) { ic <- ic[these2keep,,drop=FALSE]; ic <- drop_depcols(ic, TRUE, tol) }
        ph <- pheno[these2keep,phecol,drop=FALSE]
        wts <- weights[these2keep]

        # if X chr, paste X covariates onto additive covariates
        # (only for the null)
        if(is_x_chr[chr]) ac0 <- drop_depcols(cbind(ac, Xc), add_intercept=FALSE, tol)
        else ac0 <- ac

        if(model=="normal") {
            # FIX_ME: calculating null RSS multiple times :(
            nullrss <- nullrss_clean(ph, ac0, wts, add_intercept=TRUE, tol)

            # scan1 function taking clean data (with no missing values)
            rss <- scan1_clean(pr, ph, ac, ic, wts, add_intercept=TRUE, tol, intcovar_method)

            # calculate LOD score
            lod <- nrow(ph)/2 * (log10(nullrss) - log10(rss))
        }
        else { # binary traits
            # FIX_ME: calculating null LOD multiple times :(
            nulllod <- null_binary_clean(ph, ac0, wts, add_intercept=TRUE, maxit, bintol, tol, eta_max)

            # scan1 function taking clean data (with no missing values)
            lod <- scan1_binary_clean(pr, ph, ac, ic, wts, add_intercept=TRUE,
                                      maxit, bintol, tol, intcovar_method, eta_max)

            # calculate LOD score
            lod <- lod - nulllod
        }

        list(lod=apply(lod, 1, max, na.rm=TRUE), n=nrow(ph)) # return LOD & number of individuals used
    }

    # object to contain the LOD scores; also attr to contain sample size
    result <- matrix(nrow=n_chr, ncol=ncol(pheno))
    n <- rep(NA, ncol(pheno)); names(n) <- colnames(pheno)
    if(n_chr==0) { # edge case of no genoprobs
        return(NULL)
    }

    if(n_cores(cores)==1) { # no parallel processing
        for(i in run_indexes) {
            chr <- run_batches$chr[i]
            chrnam <- names(genoprobs)[chr]
            phebatch <- phe_batches[[run_batches$phe_batch[i]]]
            phecol <- phebatch$cols

            this_result <- by_group_func(i)
            if(!is.null(this_result)) {
                result[chr, phecol] <- this_result$lod
                if(chr==1) n[phecol] <- this_result$n
            }
        }
    }
    else {
        # calculations in parallel
        list_result <- cluster_lapply(cores, run_indexes, by_group_func)

        # check for problems (if clusters run out of memory, they'll return NULL)
        result_is_null <- vapply(list_result, is.null, TRUE)
        if(any(result_is_null))
            stop("cluster problem: returned ", sum(result_is_null), " NULLs.")

        # reorganize results
        for(i in run_indexes) {
            chr <- run_batches$chr[i]
            chrnam <- names(genoprobs)[chr]
            phebatch <- phe_batches[[run_batches$phe_batch[i]]]
            phecol <- phebatch$cols

            if(!is.null(list_result[[i]])) {
                result[chr, phecol] <- list_result[[i]]$lod
                if(chr==1) n[phecol] <- list_result[[i]]$n
            }
        }
    }

    if(by_chr) {
        dimnames(result) <- list(names(genoprobs), colnames(pheno))
    } else {
        result <- apply(result, 2, max, na.rm=TRUE)
        names(result) <- colnames(pheno)
    }

    # add some attributes with details on analysis
    attr(result, "sample_size") <- n

    result
}

#' Summarize scan1perm results
#'
#' Summarize permutation test results from [scan1perm()], as significance thresholds.
#'
#' @param object An object of class `"scanoneperm"`, as output by [scan1perm()]
#' @param alpha Vector of significance levels
#'
#' @return
#' An object of class `summary.scan1perm`. If
#' [scan1perm()] was run with `perm_Xsp=FALSE`, this is
#' a single matrix of significance thresholds, with rows being
#' signicance levels and columns being the columns in the input. If
#' [scan1perm()] was run with `perm_Xsp=TRUE`, this is
#' a list of two matrices, with the significance thresholds for the
#' autosomes and X chromosome, respectively.
#'
#' The result has an attribute `"n_perm"` that has the numbers of
#' permutation replicates (either a matrix or a list of two matrices).
#'
#' @details
#' In the case of X-chromosome-specific permutations (when
#' [scan1perm()] was run with `perm_Xsp=TRUE`, we
#' follow the approach of Broman et al. (2006) to get separate
#' thresholds for the autosomes and X chromosome, using
#'
#' Let \eqn{L_A} and \eqn{L_X} be total the genetic lengths of the
#' autosomes and X chromosome, respectively, and let \eqn{L_T = L_A +
#' L_X} Then in place of \eqn{\alpha}{alpha}, we use \deqn{\alpha_A =
#' 1 - (1-\alpha)^{L_A/L_T}}{alpha_A = 1 - (1 - alpha)^(L_A/L_T)} as
#' the significance level for the autosomes and \deqn{\alpha_X = 1 -
#' (1-\alpha)^{L_X/L_T}}{alpha_x = 1 - (1 - alpha)^(LX/LT)} as the
#' significance level for the X chromosome.
#'
#' @references
#' Broman KW, Sen Ś, Owens SE, Manichaikul A, Southard-Smith EM,
#' Churchill GA (2006) The X chromosome in quantitative trait locus
#' mapping. Genetics 174:2151-2158
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c(10,18,"X")]}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # permutations with genome scan (just 3 replicates, for illustration)
#' operm <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar,
#'                    n_perm=3)
#'
#' summary(operm, alpha=c(0.20, 0.05))
#'
#' @importFrom stats quantile
#' @export
summary_scan1perm <-
    function(object, alpha=0.05)
{

    if(is.matrix(object)) { # not X-chr-specific
        result <- apply(object, 2, quantile, 1-alpha, na.rm=TRUE)
        if(length(alpha)==1) {
            result <- matrix(result, nrow=1)
            colnames(result) <- colnames(object)
        }

        rownames(result) <- alpha

        n_perm <- matrix(colSums(!is.na(object)), nrow=1)
        colnames(n_perm) <- colnames(object)

        class(result) <- c("summary.scan1perm", "matrix")
    } else { # X-chr specific
        if(!is.list(object) || length(object) != 2 || any(sort(names(object)) != c("A", "X")))
            stop('input should be a matrix or a list of two matrices named "A" and "X"')

        chr_lengths <- attr(object, "chr_lengths")
        if(is.null(chr_lengths) || length(chr_lengths) != 2
           || any(names(chr_lengths) != c("A", "X"))) {
            warning("Ill-formed chr_lengths attribute.\n",
                    "Estimating LA/LX from the numbers of permutations, as ",
                    round(nrow(object$A)/nrow(object$X), 2))
            LA <- nrow(object$A)
            LX <- nrow(object$X)
        } else {
            LA <- chr_lengths["A"]
            LX <- chr_lengths["X"]
        }
        Lt <- LA+LX

        alphaA <- 1 - (1-alpha)^(LA/Lt)
        alphaX <- 1 - (1-alpha)^(LX/Lt)

        result <- list(A=apply(object$A, 2, quantile, 1-alphaA, na.rm=TRUE),
                       X=apply(object$X, 2, quantile, 1-alphaX, na.rm=TRUE))
        for(i in seq_along(result)) {
            if(!is.matrix(result[[i]])) {
                result[[i]] <- matrix(result[[i]], nrow=1)
                colnames(result[[i]]) <- colnames(object[[i]])
            }
            rownames(result[[i]]) <- alpha
        }
        n_perm <- rbind(A=colSums(!is.na(object$A)),
                        X=colSums(!is.na(object$X)))
        colnames(n_perm) <- colnames(object$A)

        class(result) <- c("summary.scan1perm", "list")
    }

    attr(result, "n_perm") <- n_perm
    result
}

#' @rdname summary_scan1perm
#' @param ... Ignored
#' @export
summary.scan1perm <-
    function(object, alpha=0.05, ...)
{
    summary_scan1perm(object, alpha=alpha)
}

#' Print summary of scan1perm permutations
#'
#' Print summary of scan1perm permutations
#'
#' @param x Object of class `"summary.scan1perm"`, as produced by [summary_scan1perm()].
#' @param digits Number of digits in printing significance thresholds; passed to [base::print()].
#' @param ... Ignored.
#'
#' @return Invisibly returns the input, `x`.
#'
#' @details This is to go with [summary_scan1perm()], so
#' that the summary output is printed in a nice format. Generally
#' not called directly, but it can be in order to control the
#' number of digits that appear.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' \dontshow{iron <- iron[,c(10,18,"X")]}
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#' Xcovar <- get_x_covar(iron)
#'
#' # permutations with genome scan (just 3 replicates, for illustration)
#' operm <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar,
#'                    n_perm=3)
#'
#' print( summary(operm, alpha=c(0.20, 0.05)), digits=8 )
#'
#' @export
print.summary.scan1perm <-
    function(x, digits=3, ...)
{
    n_perm <- attr(x, "n_perm")

    if(is.list(x)) { # X-chr-specific
        if(length(unique(n_perm[1,])) == 1 &&
           length(unique(n_perm[2,])) == 1) constant_perms <- TRUE
        else constant_perms <- FALSE

        cat("Autosome LOD thresholds ")
        if(constant_perms) cat("(", n_perm[1,1], " permutations)", sep="")
        cat("\n")
        print(x$A, digits=digits)

        cat("\nX chromosome LOD thresholds ")
        if(constant_perms) cat("(", n_perm[2,1], " permutations)", sep="")
        cat("\n")
        print(x$X, digits=digits)

        if(!constant_perms) {
            cat("\n")
            cat("Number of permutations:\n")
            n_spaces <- max(vapply(x, function(a) nchar(rownames(a)), 1))
            rownames(n_perm) <- paste0(paste(rep(" ", n_spaces-1), collapse=""), rownames(n_perm))
            print(n_perm)
        }
    }
    else { # not X-chr-specific
        if(length(unique(as.numeric(n_perm))) == 1) constant_perms <- TRUE
        else constant_perms <- FALSE

        cat("LOD thresholds ")
        if(constant_perms) cat("(", n_perm[1], " permutations)", sep="")
        cat("\n")

        print(x[seq_len(nrow(x)),,drop=FALSE], digits=digits)

        if(!constant_perms) {
            cat("\n")
            cat("Number of permutations:\n")
            n_spaces <- max(nchar(rownames(x)))
            rownames(n_perm) <- paste(rep(" ", n_spaces), collapse="")
            print(n_perm)
        }
    }

    invisible(x)
}

#' Heat map of scan1 results with multiple traits
#'
#' Heat map of multiple scan1 results
#'
#' @param x An object of class `"scan1"`, as output by [scan1()].
#'
#' @param map A list of vectors of marker positions, as produced by
#' [insert_pseudomarkers()].
#'
#' @param y Vector with length equal to `ncol(x)`, defining the
#'     y-axis scale of the heat map. (If the multiple LOD scores are
#'     for a quantitative variable like time, these are the
#'     corresponding times.)
#'
#' @param chr Selected chromosomes to plot; a vector of character
#' strings.
#'
#' @param gap Gap between chromosomes. The default is 1% of the total genome length.
#'
#' @param zlim Z-axis limits (a pair of numbers)
#'
#' @param color_scheme Color scheme for the heatmap (ignored if `col` is provided).
#'
#' @param col Optional vector of colors for the heatmap.
#'
#' @param n_colors Number of z-axis colors (ignored if `col` is provided).
#'
#' @param swap_axes If TRUE, put chromosome position on y-axis and lod columns on x-axis
#' (default is to have chromosome position on x-axis and lod columns on y-axis)
#'
#' @param chrlines Color of lines at chromosome breaks (NULL for default choices; NA to skip them).
#'
#' @param rescale_by Rescale the LOD scores by column, chromosome, or both, by subtracting the
#' maximum value. The default is not to rescale.
#'
#' @param ... Additional graphics paramaters.
#'
#' @return None.
#'
#' @seealso [plot_scan1()], [plot_lodpeaks()], [graphics::image()]; [plot_colorscale()]
#'
#' @importFrom graphics image par axis title
#' @importFrom grDevices rainbow heat.colors terrain.colors topo.colors gray
#'
#' @export
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map <- insert_pseudomarkers(grav2$gmap, step=1)
#' probs <- calc_genoprob(grav2, map, error_prob=0.002)
#' out <- scan1(probs, grav2$pheno)
#' plot_scan1_heatmap(out, map)
#'
#' # make the y-axis quantitative
#' y <- grav2$phenocovar$time
#' plot_scan1_heatmap(out, map, y=y, ylab=colnames(grav2$phenocovar))
#'
#' # plot with colors white -> darkslateblue, and with color scale
#' col <- colorRampPalette(c("white", "darkslateblue"))(256)
#' zlim <- c(0, maxlod(out))
#' layout(cbind(1,2), width=c(3,1))
#' plot_scan1_heatmap(out, map, y=y, ylab=colnames(grav2$phenocovar),
#'                    col=col, zlim=zlim)
#' plot_colorscale(col=col, zlim=zlim)

plot_scan1_heatmap <-
    function(x, map, chr=NULL, y=NULL, gap=NULL, zlim=NULL,
             color_scheme=c("viridis", "gray", "revgray", "heat", "terrain", "topo", "rainbow"),
             col=NULL, n_colors=256, swap_axes=FALSE, chrlines=NULL,
             rescale_by=c("none", "col", "chr", "both"), ...)
{
    if(is.null(map)) stop("map is NULL")

    if(!is.list(map)) map <- list(" "=map) # if a vector, treat it as a list with no names

    if(!is.matrix(x) && !is.data.frame(x)) stop("x must be a matrix or data frame")

    # if input x is a vector, try turning it into a matrix, since we'll be pull out row names
    if(!is.matrix(x) && !is.data.frame(x) && is.numeric(x)) x <- as.matrix(x)

    # subset chromosomes
    if(!is.null(chr)) {
        chri <- match(chr, names(map))
        if(any(is.na(chri)))
            stop("Chromosomes ", paste(chr[is.na(chri)], collapse=", "), " not found")
        x <- subset_scan1(x, map, chr)
        map <- map[chri]
    }

    # align scan1 output and map
    tmp <- align_scan1_map(x, map)
    x <- tmp$scan1
    map <- tmp$map

    if(is.null(gap)) gap <- sum(chr_lengths(map))/100
    if(!is_nonneg_number(gap)) stop("gap should be a single non-negative number")

    if(is.null(col)) {
        color_scheme <- match.arg(color_scheme)

        i <- seq(0, 1, length=n_colors)

        if(color_scheme=="viridis") {
            col <- viridis_qtl2(n_colors)
            if(is.null(chrlines)) chrlines <- "white"
        } else if(color_scheme=="gray") {
            col <- grDevices::gray(i)
            if(is.null(chrlines)) chrlines <- "violetred"
        } else if(color_scheme=="revgray") {
            col <- rev(grDevices::gray(i))
            if(is.null(chrlines)) chrlines <- "violetred"
        } else if(color_scheme=="heat") {
            col <- grDevices::heat.colors(n_colors)
            if(is.null(chrlines)) chrlines <- "white"
        } else if(color_scheme=="terrain") {
            col <- grDevices::terrain.colors(n_colors)
            if(is.null(chrlines)) chrlines <- "black"
        } else if(color_scheme=="topo") {
            col <- grDevices::topo.colors(n_colors)
            if(is.null(chrlines)) chrlines <- "white"
        } else if(color_scheme=="rainbow") {
            col <- rev(grDevices::rainbow(n_colors, start=0, end=2/3))
            if(is.null(chrlines)) chrlines <- "white"
        } else stop('color_scheme "', color_scheme, '" not allowed')
    } else {
        if(is.null(chrlines)) chrlines <- "white"
    }

    # rescale the lod scores if desired
    x <- rescale_lod(x, map, rescale_by)

    plot_scan1_heatmap_internal <-
        function(x, map, y=NULL, gap, col, chrlines, swap_axes=FALSE, xlab=NULL, ylab=NULL,
                 xlim=NULL, ylim=NULL, zlim=NULL,
                 mgp=NULL, mgp.x=NULL, mgp.y=NULL,
                 las=1, xaxt="s", yaxt="s", chrlines_lwd=2, ...)
    {
        # get x-axis range
        if(is.null(xlim)) {
            start <- xpos_scan1(map, chr=names(map), gap=gap,
                                names(map)[1], min(map[[1]]))
            end <- xpos_scan1(map, chr=names(map), gap=gap,
                              names(map)[length(map)], max(map[[length(map)]]))

            xlim <- c(start, end)
        }
        if(is.null(zlim)) {
            zlim <- range(as.numeric(x), na.rm=TRUE)
            if(zlim[1] > 0) zlim[1] <- 0
        }

        if(is.null(mgp.x)) {
            if(!is.null(mgp)) mgp.x <- mgp
            else mgp.x <- c(1.7, 0.3, 0)
        }
        if(is.null(mgp.y)) {
            if(!is.null(mgp)) mgp.y <- mgp
            else mgp.y <- c(2.0, 0.4, 0)
        }

        xpos <- unlist(lapply(seq_along(map), function(chr) xpos_scan1(map, chr=names(map), gap=gap, names(map)[chr], map[[chr]])))

        if(any(diff(xpos)==0)) { # deal with identical positions
            xpos <- xpos + seq(0, 1e-6, length=length(xpos))
        }

        if(is.null(y)) {
            y <- 1:ncol(x)
            ylim <- c(0.5, ncol(x)+0.5)
            yaxis_quant <- FALSE
        } else {
            stopifnot(length(y) == ncol(x))
            ylim <- range(y)
            yaxis_quant <- TRUE
        }

        if(swap_axes) {
            graphics::image(y, xpos, t(x), xaxs="i", yaxs="i", xlim=ylim, ylim=xlim, zlim=zlim, xlab="", ylab="", xaxt="n", yaxt="n", col=col)
        } else {
            graphics::image(xpos, y, x, xaxs="i", yaxs="i", xlim=xlim, ylim=ylim, zlim=zlim, xlab="", ylab="", xaxt="n", yaxt="n", col=col)
        }

        # chromosome axis labels
        chr <- names(map)
        if(length(chr) > 1) {
            chr_midpt <- sapply(seq_along(map), function(i) xpos_scan1(map, names(map), gap=gap,
                                                                       names(map)[i], mean(range(map[[i]], na.rm=TRUE))))

            for(i in seq_along(chr_midpt)) {
                if(!swap_axes && xaxt != "n") graphics::axis(side=1, at=chr_midpt[i], chr[i], mgp=mgp.x, tick=FALSE, las=las)
                if(swap_axes && yaxt != "n") graphics::axis(side=2, at=chr_midpt[i], chr[i], mgp=mgp.y, tick=FALSE, las=las)
            }
        } else { # single chromosome; put position on axis
            if(!swap_axes && xaxt != "n") graphics::axis(side=1, at=pretty(xpos, n=7), mgp=mgp.x, tick=FALSE, las=las)
            if(swap_axes && yaxt != "n") graphics::axis(side=2, at=pretty(xpos, n=7), mgp=mgp.y, tick=FALSE, las=las)
        }

        # lod column axis labels
        if(!swap_axes && yaxt != "n") {
            if(yaxis_quant) graphics::axis(side=2, at=pretty(y), mgp=mgp.y, tick=FALSE, las=las)
            else graphics::axis(side=2, at=1:ncol(x), colnames(x), mgp=mgp.y, tick=FALSE, las=las)
        }
        if(swap_axes && xaxt != "n") {
            if(yaxis_quant) graphics::axis(side=1, at=pretty(y), mgp=mgp.x, tick=FALSE, las=las)
            else graphics::axis(side=1, at=1:ncol(x), colnames(x), mgp=mgp.x, tick=FALSE, las=las)
        }

        # axis titles
        if(!is.null(xlab) && xlab != "") graphics::title(xlab=xlab, mgp=mgp.x)
        if(!is.null(ylab) && ylab != "") graphics::title(ylab=ylab, mgp=mgp.y)

        if(!is.na(chrlines) && !is.null(chrlines) && length(map)>1) {
            minpos <- sapply(map, min)
            chrlines_pos <- sapply(seq_along(map)[-1], function(chri) xpos_scan1(map, names(map), gap=gap, names(map)[chri], minpos[chri])-gap/2)
            if(swap_axes) abline(h=chrlines_pos, col=chrlines, lwd=chrlines_lwd)
            else abline(v=chrlines_pos, col=chrlines, lwd=chrlines_lwd)
        }

    }

    plot_scan1_heatmap_internal(x, map, y=y, gap=gap, col=col, chrlines=chrlines, swap_axes=swap_axes, zlim=zlim, ...)

}



# rescale a matrix of lod scores
#
# none: leave as is
# col: subtract off the max in each column
# chr: on each chr, subtract off the overall max
# both: on each chr, subtract off the max in each column
rescale_lod <-
    function(x, map, rescale_by=c("none", "col", "chr", "both"))
{
    rescale_by <- match.arg(rescale_by)

    if(rescale_by=="col") {
        # subtract off the maximum in each column
        x <- t(t(x) - apply(x, 2, max, na.rm=TRUE))
    } else if(rescale_by=="chr") {
        # on each chromosome, subtract off the overall maximum
        for(chr in names(map)) {
            mar <- names(map[[chr]])
            x[mar,] <- x[mar,] - max(x[mar,], na.rm=TRUE)
        }
    } else if(rescale_by=="both") {
        # on each chromosome, subtract off the max in each column
        for(chr in names(map)) {
            mar <- names(map[[chr]])
            z <- x[mar,,drop=FALSE]
            z <- t(t(z) - apply(z, 2, max, na.rm=TRUE))
            x[mar,] <- z
        }
    }

    x
}

# map_to_grid
#' Subset a map to positions on a grid
#'
#' Subset a map object to the locations on some grid.
#'
#' @param map A list of vectors of marker positions.
#' @param grid A list of logical vectors (aligned with
#' `map`), with TRUE indicating the position is on the grid.
#'
#' @return Same list as input, but subset to just include
#' pseudomarkers along a grid.
#'
#' @details This is generally for the case of a map created with
#' [insert_pseudomarkers()] with `step`>0 and
#' `stepwidth="fixed"`, so that the pseudomarkers form a grid
#' along each chromosome.
#'
#' @export
#' @keywords utilities
#' @seealso [calc_grid()], [probs_to_grid()]
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map_w_pmar <- insert_pseudomarkers(grav2$gmap, step=1)
#' sapply(map_w_pmar, length)
#' grid <- calc_grid(grav2$gmap, step=1)
#' map_sub <- map_to_grid(map_w_pmar, grid)
#' sapply(map_sub, length)

map_to_grid <-
    function(map, grid)
{
    if(is.null(map)) stop("map is NULL")
    if(is.null(grid)) stop("grid is NULL")

    if(length(map) != length(grid))
        stop("length(grid) [", length(grid), "] != length(map) [", length(map), "]")

    for(i in seq(along=map)) {
        if(is.null(grid[[i]]) || all(grid[[i]])) next

        if(length(map[[i]]) != length(grid[[i]]))
            stop("length(grid) [", length(grid[[i]]), "] != length(map) [",
                 length(map[[i]]), "] for chr ", names(map)[i])

        # subset map
        map[[i]] <- map[[i]][grid[[i]]]
    }

    map
}

#' Join phased genotype results for different chromosomes
#'
#' Join multiple phased genotype objects, as produced by [guess_phase()], for the
#' same set of individuals but different chromosomes.
#'
#' @param ... Imputed genotype objects as produced by
#' [guess_phase()]. Must have the same set of individuals.
#'
#' @return An object of class `"phasedgeno"`, like the input; see [guess_phase()].
#'
#' @seealso [rbind.phasedgeno()], [subset.phasedgeno()], [guess_phase()]
#'
#' @examples
#' \dontrun{
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' pr <- calc_genoprob(DOex, error_prob=0.002)
#' m <- maxmarg(pr)
#' phA <- guess_phase(DOex[,2], m[,2])
#' phB <- guess_phase(DOex[,3], m[,3])
#' ph <- cbind(phA, phB)
#' }
#'
#' @export
cbind.phasedgeno <-
    cbind.calc_genoprob

#' Find gaps in a genetic map
#'
#' Find gaps between markers in a genetic map
#'
#' @param map Genetic map as a list of vectors (each vector is a
#' chromosome and contains the marker positions).
#' @param min_gap Minimum gap length to return.
#'
#' @return
#' Data frame with 6 columns: chromosome, marker to left of gap,
#' numeric index of marker to left, marker to right of gap, numeric
#' index of marker to right, and the length of the gap.
#'
#' @seealso [reduce_map_gaps()]
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' find_map_gaps(iron$gmap, 40)
#'
#' @export
find_map_gaps <-
    function(map, min_gap=50)
{
    if(is.null(map)) stop("map is NULL")
    if(!is_nonneg_number(min_gap)) stop("min_gap should be a single non-negative number")

    gaps <- NULL
    for(i in seq_along(map)) {
        # interval widths
        d <- diff(map[[i]])

        index <- seq_along(map[[i]])
        lefti <- index[-length(index)]
        righti <- index[-1]
        marnam <- names(map[[i]])

        df <- data.frame(chr=names(map)[i],
                         left_marker=marnam[lefti],
                         left_index=lefti,
                         right_marker=marnam[righti],
                         right_index=righti,
                         gap=d,
                         stringsAsFactors=FALSE)
        df <- df[df$gap >= min_gap, , drop=FALSE]
        gaps <- rbind(gaps, df)
    }

    rownames(gaps) <- seq_len(nrow(gaps))

    gaps
}

#' Define strata based on rows of a matrix
#'
#' Use the rows of a matrix to define a set of strata for a stratified permutation test
#'
#' @param mat A covariate matrix, as individuals x covariates
#'
#' @return A vector of character strings: for each row of `mat`,
#' we use [base::paste()] with `collapse="|"`.
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#'
#' Xcovar <- get_x_covar(iron)
#' perm_strata <- mat2strata(Xcovar)
#'
#' @seealso [get_x_covar()], [scan1perm()]
#' @export
mat2strata <-
    function(mat)
{
    if(is.null(mat)) return(NULL)
    if(!is.matrix(mat) && is.data.frame(mat)) mat <- as.matrix(mat)
    if(!is.matrix(mat)) stop("mat should be a matrix")

    result <- apply(mat, 1, paste, collapse="|")
    names(result) <- rownames(mat)

    result
}

#' Predict SNP genotypes
#'
#' Predict SNP genotypes in a multiparent population from inferred genotypes plus founder strains' SNP alleles.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param geno Imputed genotypes, as a list of matrices, as from [maxmarg()].
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return
#' A list of matrices with inferred SNP genotypes, coded 1/2/3.
#'
#' @keywords utilities
#' @seealso [maxmarg()], [viterbi()], [calc_errorlod()]
#' @export
#' @importFrom stats setNames
#'
#' @examples
#' \dontrun{
#' # load example data and calculate genotype probabilities
#' file <- paste0("https://raw.githubusercontent.com/rqtl/",
#'                "qtl2data/main/DOex/DOex.zip")
#' DOex <- read_cross2(file)
#' probs <- calc_genoprob(DOex, error_prob=0.002)
#'
#' # inferred genotypes
#' m <- maxmarg(probs, minprob=0.5)
#'
#' # inferred SNP genotypes
#' inferg <- predict_snpgeno(DOex, m)
#' }
#'
predict_snpgeno <-
function(cross, geno, cores=1)
{
   if(!("founder_geno" %in% names(cross))) {
      stop("cross doesn't contain founder genotypes")
   }

   # ensure same chromosomes
   if(n_chr(cross) != length(geno) ||
      any(chr_names(cross) != names(geno))) {
       chr <- get_common_ids(setNames(names(geno), NULL), setNames(chr_names(cross), NULL))
       cross <- cross[,chr]
       geno <- geno[,chr]
   }

   # ensure same individuals
   if(n_ind_geno(cross) != nrow(geno[[1]]) ||
      any(ind_ids_geno(cross) != rownames(geno[[1]]))) {
       ind <- get_common_ids(setNames(rownames(geno[[1]]), NULL), setNames(ind_ids_geno(cross), NULL))
       cross <- cross[ind,]
       geno <- geno[ind,]
   }

   phase_known <- is_phase_known(cross)

   # guess phase
   if(!phase_known) {
       ph <- guess_phase(cross, geno, cores=cores)
   } else {
       ph <- cross$geno
   }

   # subset cross chromosomes
   cross <- cross[,names(ph)]

   # X chr and sexes
   is_x_chr <- cross$is_x_chr
   is_male <- !cross$is_female

   # set up cluster; use quiet=TRUE
   cores <- setup_cluster(cores, TRUE)

   by_chr_func <- function(chr) {
       # ensure the same markers
       fg <- cross$founder_geno[[chr]]
       if(phase_known) {
           ph1 <- ph2 <- geno[[chr]]
       } else {
           ph1 <- ph[[chr]][,,1]
           ph2 <- ph[[chr]][,,2]
       }

       # ugh deal with DOF1/HSF1 specially
       if(cross$crosstype %in% c("dof1", "hsf1")) {
           if(!is_x_chr[chr]) ph2[] <- 9 # one chromosome is from 9th founder
           else {
               if(any(is_male)) ph2[is_male,] <- ph1[is_male,,drop=FALSE]
               if(any(!is_male)) ph2[!is_male,] <- 9
           }
       }

       if(ncol(fg) != ncol(ph1) ||
          any(colnames(fg) != colnames(ph1))) {
           mar <- get_common_ids(setNames(colnames(fg), NULL), setNames(colnames(ph1), NULL))
           fg <- fg[,mar,drop=FALSE]
           ph1 <- ph1[,mar,drop=FALSE]
           ph2 <- ph2[,mar,drop=FALSE]
       }

       result <- .predict_snpgeno(ph1, ph2, fg)
       dimnames(result) <- dimnames(cross$geno[[chr]])

       if(is_x_chr[chr] && any(is_male)) { # some males on X chr
           result[is_male,] <- .predict_snpgeno(ph1[is_male,,drop=FALSE], ph1[is_male,,drop=FALSE], fg)
       }

       result
   }

   result <- cluster_lapply(cores, seq_along(geno), by_chr_func)
   names(result) <- names(geno)
   result

}

# stuff for genome scan with binary trait

# scan1 function taking nicely aligned data with no missing values
#
# Here genoprobs is a plain 3d array
scan1_binary_clean <-
    function(genoprobs, pheno, addcovar, intcovar,
             weights, add_intercept=TRUE, maxit, tol, qr_tol,
             intcovar_method, eta_max=30)
{
    n <- nrow(pheno)
    if(add_intercept)
        addcovar <- cbind(rep(1,n), addcovar) # add intercept

    if(is.null(intcovar)) { # no interactive covariates

        if(is.null(weights)) { # no weights
            return( scan_binary_onechr(genoprobs, pheno, addcovar, maxit, tol, qr_tol, eta_max) )
        } else { # weights included
            return( scan_binary_onechr_weighted(genoprobs, pheno, addcovar, weights, maxit, tol, qr_tol, eta_max) )
        }

    } else { # interactive covariates
        # high- and low-memory versions of functions
        if(intcovar_method=="highmem")
            scanf <- c(scan_binary_onechr_intcovar_highmem,
                       scan_binary_onechr_intcovar_weighted_highmem)
        else
            scanf <- c(scan_binary_onechr_intcovar_lowmem,
                       scan_binary_onechr_intcovar_weighted_lowmem)

        if(is.null(weights)) { # no weights
            return( scanf[[1]](genoprobs, pheno, addcovar, intcovar, maxit, tol, qr_tol, eta_max) )
        } else { # weights included
            return( scanf[[2]](genoprobs, pheno, addcovar, intcovar, weights, maxit, tol, qr_tol, eta_max) )
        }

    }
}

# calculate null log likelihood, with nicely aligned data with no missing values
null_binary_clean <-
    function(pheno, addcovar, weights, add_intercept=TRUE, maxit, tol, qr_tol, eta_max=30)
{
    n <- nrow(pheno)
    if(add_intercept)
        addcovar <- cbind(rep(1,n), addcovar) # add intercept

    if(is.null(weights) || length(weights)==0) { # no weights
        result <- apply(pheno, 2, function(phe) calc_ll_binreg(addcovar, phe, maxit, tol, qr_tol, eta_max))
    } else { # weights included
        result <- apply(pheno, 2, function(phe) calc_ll_binreg_weighted(addcovar, phe, weights, maxit, tol,
                                                                        qr_tol, eta_max))
    }

    as.numeric(result)
}

# check phenotype columns are binary (or at least between 0 and 1)
# if not, omit the non-binary columns
check_binary_pheno <-
    function(pheno, tol=1e-8)
{
    binary_col <- apply(pheno, 2, function(a) all(is.na(a) | (a >= 0 & a <= 1)))

    if(!all(binary_col)) {
        if(!any(binary_col))
            stop("Phenotypes are not binary (values in [0, 1])")
        if(!all(binary_col)) {
            warning("Not all phenotypes are binary. Dropping ",
                    sum(!binary_col), " columns")
            pheno <- pheno[,binary_col,drop=FALSE]
        }
    }

    round(pheno)
}

#' Calculate entropy of genotype probability distribution
#'
#' For each individual at each genomic position, calculate the entropy
#' of the genotype probability distribution, as a quantitative summary
#' of the amount of missing information.
#'
#' @param probs Genotype probabilities, as calculated from
#' [calc_genoprob()].
#' @param quiet IF `FALSE`, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A list of matrices (each matrix is a chromosome and is arranged as individuals x markers).
#'
#' @details We calculate -sum(p log_2 p), where we take 0 log 0 = 0.
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' \dontshow{grav2 <- grav2[,c(4,5)] # subset to chr 3}
#' probs <- calc_genoprob(grav2, error_prob=0.002)
#' e <- calc_entropy(probs)
#' e <- do.call("cbind", e) # combine chromosomes into one big matrix
#'
#' # summarize by individual
#' mean_ind <- rowMeans(e)
#'
#' # summarize by marker
#' mean_marker <- colMeans(e)
calc_entropy <-
    function(probs, quiet=TRUE, cores=1)
{
    if(is.null(probs)) stop("probs is NULL")

    if(is.cross2(probs)) {
        stop('Input probs is a "cross2" object but should be genotype probabilities, as from calc_genoprob')
    }

    # set up cluster; set quiet=TRUE if multi-core
    cores <- setup_cluster(cores, quiet)
    if(!quiet && n_cores(cores)>1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # make the rest quiet
    }

    # function that does the work
    entropy <-
        function(p, tol=1e-256) {
            p <- p[p>tol]
            -sum(p * log2(p))
        }

    result <- cluster_lapply(cores, seq_along(probs),
                             function(i) apply(probs[[i]], c(1,3), entropy))
    names(result) <- names(probs)
    result
}

#' Create a function to query genes
#'
#' Create a function that will connect to a SQLite database of gene
#' information and return a data frame with gene information for a
#' selected region.
#'
#' @param dbfile Name of database file
#' @param db Optional database connection (provide one of `file` and `db`).
#' @param table_name Name of table in the database
#' @param chr_field Name of chromosome field
#' @param start_field Name of field with start position (in basepairs)
#' @param stop_field Name of field with stop position (in basepairs)
#' @param name_field Name of field with gene name
#' @param strand_field Name of field with strand (+/-)
#' @param filter Additional SQL filter (as a character string).
#'
#' @return Function with three arguments, `chr`, `start`,
#'     and `end`, which returns a data frame with the genes
#'     overlapping that region, with `start` and `end` being
#'     in Mbp. The output should contain at least the columns
#'     `Name`, `chr`, `start`, and `stop`, the
#'     latter two being positions in Mbp.
#'
#' @details Note that this function assumes that the database has
#'     `start` and `stop` fields that are in basepairs, but
#'     the selection uses positions in Mbp, and the output data frame
#'     should have `start` and `stop` columns in Mbp.
#'
#' Also note that a SQLite database of MGI mouse genes
#' is available at figshare:
#' \doi{10.6084/m9.figshare.5286019.v7}
#'
#' @export
#' @importFrom RSQLite SQLite dbConnect dbDisconnect dbGetQuery
#'
#' @examples
#' # create query function by connecting to file
#' dbfile <- system.file("extdata", "mouse_genes_small.sqlite", package="qtl2")
#' query_genes <- create_gene_query_func(dbfile, filter="(source=='MGI')")
#' # query_genes will connect and disconnect each time
#' genes <- query_genes("2", 97.0, 98.0)
#'
#' # connect and disconnect separately
#' library(RSQLite)
#' db <- dbConnect(SQLite(), dbfile)
#' query_genes <- create_gene_query_func(db=db, filter="(source=='MGI')")
#' genes <- query_genes("2", 97.0, 98.0)
#' dbDisconnect(db)

create_gene_query_func <-
    function(dbfile=NULL, db=NULL, table_name="genes",
             chr_field="chr", start_field="start", stop_field="stop",
             name_field="Name", strand_field="strand",
             filter=NULL)
{
    if(!is.null(db)) {
        if(!is.null(dbfile))
            warning("Provide just one of dbfile or db; using db")

        query_func <- function(chr, start, end) {

            # convert input positions from Mbp to basepairs
            start <- round(start * 1e6)
            end <- round(end * 1e6)

            # bits of the query
            chrselect <- paste0(chr_field, " == '", chr, "'")
            pos_select1 <- paste0("(", start_field, " >= ", start, " AND ", start_field, " <= ", end, ")")
            pos_select2 <- paste0("(", stop_field, " >= ", start, " AND ", stop_field, " <= ", end, ")")
            pos_select3 <- paste0("(", start_field, " <= ", end, " AND ", stop_field, " >= ", start, ")")

            # the full query
            query <- paste0("SELECT * FROM ", table_name, " WHERE ",
                            chrselect, " AND ( ",
                            pos_select1, " OR ", pos_select2, " OR ", pos_select3, " )")
            if(!is.null(filter) && filter != "")
                query <- paste0(query, " AND (", filter, ")")

            # do the query
            result <- RSQLite::dbGetQuery(db, query)

            # include start/stop columns in Mbp
            result$start <- result[,start_field]/1e6
            result$stop <- result[,stop_field]/1e6

            # name and strand
            result <- swap_colname(result, name_field, "Name")
            result <- swap_colname(result, strand_field, "strand")

            result
        }

    }
    else {
        if(is.null(dbfile) || dbfile=="")
            stop("Provide either dbfile or db")

        query_func <- function(chr, start, end)
        {
            if(!file.exists(dbfile))
                stop("File ", dbfile, " doesn't exist")

            db <- RSQLite::dbConnect(RSQLite::SQLite(), dbfile)
            on.exit(RSQLite::dbDisconnect(db)) # disconnect on exit

            # convert input positions from Mbp to basepairs
            start <- round(start * 1e6)
            end <- round(end * 1e6)

            # bits of the query
            chrselect <- paste0(chr_field, " == '", chr, "'")
            pos_select1 <- paste0("(", start_field, " >= ", start, " AND ", start_field, " <= ", end, ")")
            pos_select2 <- paste0("(", stop_field, " >= ", start, " AND ", stop_field, " <= ", end, ")")
            pos_select3 <- paste0("(", start_field, " <= ", end, " AND ", stop_field, " >= ", start, ")")

            # the full query
            query <- paste0("SELECT * FROM ", table_name, " WHERE ",
                            chrselect, " AND ( ",
                            pos_select1, " OR ", pos_select2, " OR ", pos_select3, " )")
            if(!is.null(filter) && filter != "")
                query <- paste0(query, " AND (", filter, ")")

            # do the query
            result <- RSQLite::dbGetQuery(db, query)

            # include start/stop columns in Mbp
            result$start <- result[,start_field]/1e6
            result$stop <- result[,stop_field]/1e6

            # name and strand
            result <- swap_colname(result, name_field, "Name")
            result <- swap_colname(result, strand_field, "strand")

            result
        }
    }

    query_func
}

context("create_gene_query_func")

test_that("create_gene_query_func works", {

    # use file name
    dbfile <- system.file("extdata", "mouse_genes_small.sqlite", package="qtl2")
    qf <- create_gene_query_func(dbfile)

    expected <- structure(list(chr = c("2", "2", "2"), source = c("MGI", "MGI", "MGI"),
                               type = c("gene", "gene", "pseudogene"),
                               start = c(96.318169, 97.626967, 97.947953),
                               stop = c(97.631672, 97.627052, 97.949756),
                               score = c(NA_real_, NA_real_, NA_real_),
                               strand = c("+", "-", "-"),
                               phase = c(NA_real_, NA_real_, NA_real_),
                               ID = c("MGI_C57BL6J_2442636", "MGI_C57BL6J_5690712", "MGI_C57BL6J_3651056"),
                               Name = c("Lrrc4c", "Gm44320", "Gm13803"),
                               Parent = c(NA_character_, NA_character_, NA_character_),
                               Dbxref = c("NCBI_Gene:241568,ENSEMBL:ENSMUSG00000050587",
                                          "ENSEMBL:ENSMUSG00000105133", "NCBI_Gene:621146,ENSEMBL:ENSMUSG00000082820"),
                               gene_id = c("MGI:2442636", "MGI:5690712", "MGI:3651056"),
                               mgi_type = c("protein coding gene", "miRNA gene", "pseudogene"),
                               description = c("leucine rich repeat containing 4C", "predicted gene%2c 44320",
                                               "predicted gene 13803")),
                          row.names = c(NA, -3L), class = "data.frame")

    expect_equal(qf(2, 97.5, 98.0), expected)

    # use db connection
    library(RSQLite)
    db <- dbConnect(SQLite(), dbfile)
    qf2 <- create_gene_query_func(db=db)
    expect_equal(qf2(2, 97.5, 98.0), expected)
    dbDisconnect(db)

    # include filter
    qf3 <- create_gene_query_func(dbfile, filter="Name = 'Lrrc4c'")
    expected_sub <- expected[1,,drop=FALSE]
    rownames(expected_sub) <- 1L
    expect_equal(qf3(2, 97.5, 98), expected_sub)

})

# est_map
#' Estimate genetic maps
#'
#' Uses a hidden Markov model to re-estimate the genetic map for an
#' experimental cross, with possible allowance for genotyping errors.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param error_prob Assumed genotyping error probability
#' @param map_function Character string indicating the map function to
#' use to convert genetic distances to recombination fractions.
#' @param lowmem If `FALSE`, precalculate initial and emission
#' probabilities, and at each iteration calculate the transition
#' matrices for a chromosome; potentially a lot faster but using
#' more memory. Needs to be tailored somewhat to cross type. For
#' example, multi-way RIL may need to reorder the transition
#' matrix according to cross order, and AIL and DO need separate
#' transition matrices for each generation.
#' @param maxit Maximum number of iterations in EM algorithm.
#' @param tol Tolerance for determining convergence
#' @param quiet If `FALSE`, print progress messages.
#' @param save_rf If `TRUE`, save the estimated recombination
#' fractions as an attribute (`"rf"`) of the result.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return A list of numeric vectors, with the estimated marker
#' locations (in cM). The location of the initial marker on each
#' chromosome is kept the same as in the input `cross`.
#'
#' @details
#' The map is estimated assuming no crossover interference,
#' but a map function (by default, Haldane's) is used to derive the genetic distances.
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' \dontshow{grav2 <- grav2[,"3"]}
#' gmap <- est_map(grav2, error_prob=0.002)

est_map <-
function(cross, error_prob=1e-4,
         map_function=c("haldane", "kosambi", "c-f", "morgan"),
         lowmem=FALSE, maxit=10000, tol=1e-6, quiet=TRUE, save_rf=FALSE,
         cores=1)
{
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')

    map_function <- match.arg(map_function)
    if(!is_nonneg_number(error_prob) || error_prob > 1) stop("error_prob must be a single number in [0,1]")
    if(!is_nonneg_number(maxit)) stop("maxit must be a single non-negative number")
    if(!is_pos_number(tol)) stop("tol must be a single positive number")

    # deal with missing information
    ind <- rownames(cross$geno[[1]])
    chrnames <- names(cross$geno)
    is_x_chr <- handle_null_isxchr(cross$is_x_chr, chrnames)
    cross$is_female <- handle_null_isfemale(cross$is_female, ind)
    cross$cross_info <- handle_null_isfemale(cross$cross_info, ind)

    if(is.null(cross$gmap)) stop("cross needs a genetic map, as a starting point")
    map <- vector("list", length(cross$gmap))

    # set up cluster; make quiet=FALSE if cores>1
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores) > 1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # no more messages
    }

    founder_geno <- cross$founder_geno
    if(is.null(founder_geno))
        founder_geno <- create_empty_founder_geno(cross$geno)

    by_chr_func <- function(chr) {
        # the following avoids a warning in R CMD check
        . <- "avoid R CMD check warning"

        if(!quiet) message("Chr ", names(cross$geno)[chr])

        gmap <- cross$gmap[[chr]]

        # omit individuals with < 2 genotypes
        geno <- cross$geno[[chr]]
        ntyped <- rowSums(geno>0)
        keep <- ntyped >= 2
        geno <- t(geno[keep,,drop=FALSE])
        is_female <- cross$is_female[keep]
        cross_info <- t(cross$cross_info[keep,,drop=FALSE])

        rf_start <- map2rf(gmap) # positions to inter-marker rec frac
        if(lowmem)
            rf <- .est_map(cross$crosstype, geno, founder_geno[[chr]],
                           is_x_chr[chr], is_female, cross_info,
                           rf_start, error_prob, maxit, tol, !quiet)
        else {
            # groups of individuals with common sex and cross_info
            sex_crossinfo <- paste(is_female, apply(cross_info, 2, paste, collapse=":"), sep=":")
            unique_cross_group <- unique(sex_crossinfo)
            cross_group <- match(sex_crossinfo, unique_cross_group)-1 # indexes start at 0
            unique_cross_group <- match(seq_along(unique_cross_group)-1, cross_group)-1 # again start at 0

            rf <- .est_map2(cross$crosstype, geno, founder_geno[[chr]],
                            is_x_chr[chr], is_female, cross_info,
                            cross_group, unique_cross_group,
                            rf_start, error_prob, maxit, tol, !quiet)
        }

        loglik <- attr(rf, "loglik")
        map <- cumsum(c(gmap[1], imf(rf, map_function))) # rec frac to positions

        names(map) <- names(gmap)
        attr(map, "loglik") <- loglik
        if(save_rf) {
            attr(rf, "loglik") <- NULL
            attr(map, "rf") <- rf
        }

        map
    }

    chrs <- seq(along=map)
    map <- cluster_lapply(cores, chrs, by_chr_func) # if cores==1, uses lapply

    names(map) <- names(cross$gmap)
    attr(map, "is_x_chr") <- is_x_chr

    if(save_rf) { # put est'd rec fracs as single attribute, as list
        rfs <- lapply(map, function(a) attr(a, "rf"))
        names(rfs) <- names(map)
        attr(map, "rf") <- rfs

        # strip off the individual rf attributes
        for(i in seq_along(map))
            attr(map[[i]], "rf") <- NULL
    }

    map
}

# sim_geno
#' Simulate genotypes given observed marker data
#'
#' Uses a hidden Markov model to simulate from the joint distribution
#' Pr(g | O) where g is the underlying sequence of true genotypes and
#' O is the observed multipoint marker data, with possible allowance
#' for genotyping errors.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param map Genetic map of markers. May include pseudomarker
#' locations (that is, locations that are not within the marker
#' genotype data). If NULL, the genetic map in `cross` is used.
#' @param n_draws Number of simulations to perform.
#' @param error_prob Assumed genotyping error probability
#' @param map_function Character string indicating the map function to
#' use to convert genetic distances to recombination fractions.
#' @param lowmem If `FALSE`, split individuals into groups with
#' common sex and crossinfo and then precalculate the transition
#' matrices for a chromosome; potentially a lot faster but using more
#' memory.
#' @param quiet If `FALSE`, print progress messages.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return An object of class `"sim_geno"`: a list of three-dimensional arrays of imputed genotypes,
#' individuals x positions x draws. Also contains three attributes:
#' * `crosstype` - The cross type of the input `cross`.
#' * `is_x_chr` - Logical vector indicating whether chromosomes
#'     are to be treated as the X chromosome or not, from input `cross`.
#' * `alleles` - Vector of allele codes, from input
#'     `cross`.
#'
#' @details
#'  After performing the backward equations, we draw from
#'  \eqn{Pr(g_1 = v | O)}{Pr(g[1] = v | O)} and then \eqn{Pr(g_{k+1} = v |
#'    O, g_k = u)}{Pr(g[k+1] = v | O, g[k] = u)}.
#'
#' @seealso [cbind.sim_geno()], [rbind.sim_geno()]
#'
#' @export
#' @keywords utilities
#'
#' @examples
#' grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
#' map_w_pmar <- insert_pseudomarkers(grav2$gmap, step=1)
#' draws <- sim_geno(grav2, map_w_pmar, n_draws=4, error_prob=0.002)

sim_geno <-
function(cross, map=NULL, n_draws=1, error_prob=1e-4,
         map_function=c("haldane", "kosambi", "c-f", "morgan"),
         lowmem=FALSE, quiet=TRUE, cores=1)
{
    # check inputs
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')
    if(error_prob < 0)
        stop("error_prob must be > 0")
    map_function <- match.arg(map_function)

    if(!is_pos_number(n_draws)) stop("n_draws should be a single positive integer")
    if(!is_nonneg_number(error_prob)) stop("error_prob should be a single non-negative number")

    if(!lowmem)
        return(sim_geno2(cross=cross, map=map, n_draws=n_draws,
                         error_prob=error_prob, map_function=map_function, quiet=quiet,
                         cores=cores))

    # set up cluster; make quiet=FALSE if cores>1
    cores <- setup_cluster(cores)
    if(!quiet && n_cores(cores) > 1) {
        message(" - Using ", n_cores(cores), " cores")
        quiet <- TRUE # no more messages
    }

    # pseudomarker map
    if(is.null(map)) {
        if(is.null(cross$gmap)) stop("If cross does not contain a genetic map, map must be provided.")
        map <- insert_pseudomarkers(cross$gmap)
    }
    # possibly subset the map
    if(length(map) != length(cross$geno) || !all(names(map) == names(cross$geno))) {
        chr <- names(cross$geno)
        if(!all(chr %in% names(map)))
            stop("map doesn't contain all of the necessary chromosomes")
        map <- map[chr]
    }
    # calculate marker index object
    index <- create_marker_index(lapply(cross$geno, colnames), map)

    rf <- map2rf(map, map_function)

    # deal with missing information
    ind <- rownames(cross$geno[[1]])
    chrnames <- names(cross$geno)
    is_x_chr <- handle_null_isxchr(cross$is_x_chr, chrnames)
    cross$is_female <- handle_null_isfemale(cross$is_female, ind)
    cross$cross_info <- handle_null_isfemale(cross$cross_info, ind)

    founder_geno <- cross$founder_geno
    if(is.null(founder_geno))
        founder_geno <- create_empty_founder_geno(cross$geno)

    by_group_func <- function(i) {
        dr <- .sim_geno(cross$crosstype, t(cross$geno[[chr]][group[[i]],,drop=FALSE]),
                        founder_geno[[chr]], cross$is_x_chr[chr], cross$is_female[group[[i]]],
                        t(cross$cross_info[group[[i]],,drop=FALSE]), rf[[chr]], index[[chr]],
                        error_prob, n_draws)
        aperm(dr, c(3,1,2))
    }

    group <- parallel::splitIndices(nrow(cross$geno[[1]]), n_cores(cores))
    groupindex <- seq(along=group)

    draws <- vector("list", length(cross$geno))
    names(draws) <- names(cross$geno)
    for(chr in seq(along=cross$geno)) {
        if(!quiet) message("Chr ", names(cross$geno)[chr])

        if(n_cores(cores)==1) { # no parallel processing
            # calculations in one group
            draws[[chr]] <- by_group_func(1)
        }
        else {
            # calculations in parallel
            temp <- cluster_lapply(cores, groupindex, by_group_func)

            # paste them back together
            d <- vapply(temp, dim, rep(0,3))
            nr <- sum(d[1,])
            draws[[chr]] <- array(dim=c(nr, d[2,1], d[3,1]))
            for(i in groupindex)
                draws[[chr]][group[[i]],,] <- temp[[i]]
        }

        dimnames(draws[[chr]]) <- list(rownames(cross$geno[[chr]]),
                                       names(map[[chr]]),
                                       NULL)

    }

    names(draws) <- names(cross$geno)
    attr(draws, "crosstype") <- cross$crosstype
    attr(draws, "is_x_chr") <- cross$is_x_chr
    attr(draws, "alleles") <- cross$alleles

    class(draws) <- c("sim_geno", "list")
    draws
}

#' Chi-square test on all pairs of columns
#'
#' Perform a chi-square test for independence for all pairs of columns of a matrix.
#'
#' @param x A matrix of positive integers. `NA`s and values <= 0 are treated as missing.
#'
#' @return A matrix of size p x p, where p is the number of columns in
#'     the input matrix `x`, containing the chi-square test
#'     statistics for independence, applied to pairs of columns of
#'     `x`. The diagonal of the result will be all `NA`s.
#'
#' @keywords htest
#'
#' @export
#'
#' @examples
#' z <- matrix(sample(1:2, 500, replace=TRUE), ncol=5)
#' chisq_colpairs(z)

chisq_colpairs <-
    function(x)
{
    if(!is.matrix(x) && is.data.frame(x)) x <- as.matrix(x)
    if(!is.matrix(x)) stop("x should be a matrix")

    if(ncol(x) < 2)
        stop("ncol(x) should be >= 2")

    result <- .chisq_colpairs(x)
    dimnames(result) <- list(colnames(x), colnames(x))
    diag(result) <- NA

    result
}

context("Chi-square test for column pairs")

test_that("chisq_colpairs works", {

    set.seed(44552625)
    p <- 5
    z <- matrix(sample(1:2, p*100, replace=TRUE), ncol=p)
    colnames(z) <- LETTERS[1:p]
    z[,2] <- 0

    result <- chisq_colpairs(z)

    expected <- matrix(ncol=p,nrow=p)
    cols <- seq(p)[-2]
    for(i in 1:(p-2)) {
        for(j in (i+1):(p-1)) {
            ii <- cols[i]
            jj <- cols[j]
            suppressWarnings(expected[jj,ii] <- expected[ii,jj] <-
                                 chisq.test(table(z[,ii], z[,jj]), correct=FALSE)$stat)
        }
    }
    dimnames(expected) <- list(LETTERS[1:p], LETTERS[1:p])

    expect_equal(result, expected)

    # throws error if you give a single-column matrix
    expect_error( chisq_colpairs(z[,1,drop=FALSE]) )

    # throws error if you give a vector
    expect_error( chisq_colpairs(z[,1]) )

})

# various internal utility functions

# create empty set of matrices for founder genotype data
create_empty_founder_geno <-
function(geno)
{
    result <- vector("list", length(geno))
    names(result) <- names(geno)
    for(i in seq(along=geno)) {
        result[[i]] <- matrix(0L, nrow=0, ncol=ncol(geno[[i]]))
        colnames(result[[i]]) <- colnames(geno[[i]])
    }
    result
}

# convert map to rf
map2rf <-
    function(map, map_function=c("haldane", "kosambi", "c-f", "morgan"), tol=1e-12)
{
    map_function <- match.arg(map_function)

    # if list, use lapply and call this same function for each chromosome
    if(is.list(map))
        return(lapply(map, map2rf, map_function, tol))

    rf <- mf(diff(map), map_function)
    rf[rf < tol] <- tol # don't let values be too small

    rf
}

#' Plot QTL effects along chromosome
#'
#' Plot estimated QTL effects along a chromosomes.
#'
#' @param x Estimated QTL effects ("coefficients") as obtained from
#' [scan1coef()].
#'
#' @param map A list of vectors of marker positions, as produced by
#' [insert_pseudomarkers()].
#'
#' @param columns Vector of columns to plot
#'
#' @param col Vector of colors, same length as `columns`. If
#' NULL, some default choices are made.
#'
#' @param scan1_output If provided, we make a two-panel plot with
#' coefficients on top and LOD scores below. Should have just one LOD
#' score column; if multiple, only the first is used.
#'
#' @param add If TRUE, add to current plot (must have same map and
#' chromosomes).
#'
#' @param gap Gap between chromosomes. The default is 1% of the total genome length.
#'
#' @param top_panel_prop If `scan1_output` provided, this gives the
#' proportion of the plot that is devoted to the top panel.
#'
#' @param legend Location of legend, such as `"bottomleft"` or `"topright"` (NULL for no legend)
#'
#' @param ... Additional graphics parameters.
#'
#' @return None.
#'
#' @export
#' @importFrom graphics layout par
#'
#' @details
#' `plot_coefCC()` is the same as `plot_coef()`, but forcing
#' `columns=1:8` and using the Collaborative Cross colors,
#' [CCcolors].
#'
#' @seealso [CCcolors], [plot_scan1()], [plot_snpasso()], [plot_ci()]
#'
#' @section Hidden graphics parameters:
#' A number of graphics parameters can be passed via `...`. For
#' example, `bgcolor` to control the background color, and things
#' like `ylab` and `ylim`. These are not included as formal
#' parameters in order to avoid cluttering the function definition.
#'
#' In the case that `scan1_output` is provided, `col`,
#' `ylab`, and `ylim` all control the panel with estimated
#' QTL effects, while `col_lod`, `ylab_lod`, and
#' `ylim_lod` control the LOD curve panel.
#'
#' If `legend` is indicated so that a legend is shown, `legend_lab`
#' controls the labels in the legend, and `legend_ncol` indicates the
#' number of columns in the legend.
#'
#' @examples
#' # read data
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#'
#' # insert pseudomarkers into map
#' map <- insert_pseudomarkers(iron$gmap, step=1)
#'
#' # calculate genotype probabilities
#' probs <- calc_genoprob(iron, map, error_prob=0.002)
#'
#' # grab phenotypes and covariates; ensure that covariates have names attribute
#' pheno <- iron$pheno[,1]
#' covar <- match(iron$covar$sex, c("f", "m")) # make numeric
#' names(covar) <- rownames(iron$covar)
#'
#' # calculate coefficients for chromosome 7
#' coef <- scan1coef(probs[,7], pheno, addcovar=covar)
#'
#' # plot QTL effects (note the need to subset the map object, for chromosome 7)
#' plot(coef, map[7], columns=1:3, col=c("slateblue", "violetred", "green3"))
plot_coef <-
    function(x, map, columns=NULL, col=NULL, scan1_output=NULL,
             add=FALSE, gap=NULL, top_panel_prop=0.65,
             legend=NULL, ...)
{
    if(is.null(map)) stop("map is NULL")

    # align scan1 output and map
    tmp <- align_scan1_map(x, map)
    x <- tmp$scan1
    map <- tmp$map

    if(!is_pos_number(top_panel_prop)) stop("top_panel_prop should be a single non-negative number")

    if(nrow(x) != length(unlist(map)))
        stop("nrow(x) [", nrow(x), "] != number of positions in map [",
             length(unlist(map)), "]")

    if(!is.null(scan1_output)) { # call internal function for both coef and LOD
        return(plot_coef_and_lod(x, map, columns=columns, col=col, scan1_output=scan1_output,
                                 gap=gap, xaxt=NULL, top_panel_prop=top_panel_prop,
                                 legend=legend, ...))
    }

    if(is.null(columns))
        columns <- seq_len(ncol(x))

    if(is.null(col)) {
        n_col <- length(columns)
        if(n_col == 1) col <- c("slateblue")
        else if(n_col == 2) col <- c("slateblue", "violetred")
        else if (n_col==3) col=c("slateblue", "violetred", "green3")
        else if(n_col <= 8) col <- c("#1B9E77", "#D95F02", "#7570B3", "#E7298A",
                "#66A61E", "#E6AB02", "#A6761D", "#666666")[seq_len(n_col)]
        else
            stop("With >8 coefficients, you need to provide colors via col")
    } else {
        if(length(col) < length(columns))
            stop("Need at least ", length(columns), " colors")
    }

    plot_coef_internal <-
        function(x, map, columns, ylim=NULL, col, add, gap, bgcolor="gray90", altbgcolor="gray85",
                 ylab="QTL effects", legend_lab=NULL, legend_ncol=NULL, ...)
        {
            if(is.null(ylim)) {
                ylim <- range(unclass(x)[,columns], na.rm=TRUE)
                d <- diff(ylim) * 0.02 # add 2% on either side
                ylim <- ylim + c(-d, d)
            }

            plot_scan1(x, map, lodcolumn=columns[1], ylim=ylim, col=col[1], add=add,
                       gap=gap, bgcolor=bgcolor, altbgcolor=altbgcolor,
                       ylab=ylab, ...)
            if(length(columns) > 1) {
                for(i in seq(along=columns)[-1])
                    plot_scan1(x, map, lodcolumn=columns[i], col=col[i], gap=gap,
                               add=TRUE, ...)
            }

            if(!is.null(legend)) {
                if(is.null(legend_lab)) {
                    if(!is.null(names(col))) legend_lab <- names(col)
                    else legend_lab <- colnames(x)[columns]
                }
                if(length(legend_lab) > length(columns))
                    legend_lab <- legend_lab[seq_along(columns)]
                this_col <- rep(col, length(legend_lab))[seq_along(legend_lab)]
                if(is.null(legend_ncol)) legend_ncol <- ifelse(length(legend_lab) > 4, 2, 1)
                legend(legend, lwd=2, col=this_col, legend_lab, bg=bgcolor,
                       ncol=legend_ncol)
            }

        }
    plot_coef_internal(unclass(x), map, columns=columns, col=col, add=add, gap=gap, ...)
}

#' @export
#' @rdname plot_coef
plot_coefCC <-
    function(x, map, columns=1:8, col=qtl2::CCcolors, scan1_output=NULL,
             add=FALSE, gap=NULL, top_panel_prop=0.65, legend=NULL, ...)
{
    plot_coef(x, map, columns=columns, col=col,
              scan1_output=scan1_output, add=add, gap=gap,
              top_panel_prop=top_panel_prop, legend=legend, ...)
}

#' @export
#' @rdname plot_coef
plot.scan1coef <-
    function(x, map, columns=1, col=NULL, scan1_output=NULL, add=FALSE, gap=NULL,
             top_panel_prop=0.65, legend=NULL, ...)
{
    plot_coef(x, map, columns=columns, col=col, scan1_output=scan1_output,
              add=add, gap=gap, top_panel_prop=top_panel_prop, legend=legend,
              ...)
}

#' Reduce the lengths of gaps in a map
#'
#' Reduce the lengths of gaps in a map
#'
#' @param map Genetic map as a list of vectors (each vector is a
#' chromosome and contains the marker positions).
#' @param min_gap Minimum gap length to return.
#'
#' @return
#' Input map with any gaps greater than `min_gap` reduced to `min_gap`.
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' rev_map <- reduce_map_gaps(iron$gmap, 30)
#'
#' @seealso [find_map_gaps()]
#'
#' @export
reduce_map_gaps <-
    function(map, min_gap=50)
{
    if(is.null(map)) stop("map is NULL")
    if(!is_nonneg_number(min_gap)) stop("min_gap should be a single non-negative number")

    for(i in seq_along(map)) {

        # inter-marker distances
        d <- diff(map[[i]])
        d[d > min_gap] <- min_gap

        # back to original positions
        m <- cumsum(c(min(map[[i]]), d))

        # substitute back into object
        names(m) <- names(map[[i]])
        map[[i]] <- m
    }

    map
}

#' Guess phase of imputed genotypes
#'
#' Turn imputed genotypes into phased genotypes along chromosomes by
#' attempting to pick the phase that leads to the fewest recombination
#' events.
#'
#' @param cross Object of class `"cross2"`. For details, see the
#' [R/qtl2 developer guide](https://kbroman.org/qtl2/assets/vignettes/developer_guide.html).
#' @param geno Imputed genotypes, as a list of matrices, as from [maxmarg()].
#' @param deterministic If TRUE, preferentially put smaller allele first when there's uncertainty.
#' If FALSE, the order of alleles is random in such cases.
#' @param cores Number of CPU cores to use, for parallel calculations.
#' (If `0`, use [parallel::detectCores()].)
#' Alternatively, this can be links to a set of cluster sockets, as
#' produced by [parallel::makeCluster()].
#'
#' @return
#' If input cross is phase-known (e.g., recombinant inbred lines),
#' the output will be the input `geno`. Otherwise, the output
#' will be a list of three-dimensional arrays of imputed
#' genotypes, individual x position x haplotype (1/2), and will have class `"phasedgeno"`.
#'
#' @details We randomly assign the pair of alleles at the first locus
#'     to two haplotypes, and then work left to right, assigning
#'     alleles to haplotypes one locus at a time seeking the fewest
#'     recombination events. The results are subject to arbitrary and
#'     random choices. For example, to the right of a homozygous
#'     region, either orientation is equally reasonable.
#'
#' @export
#' @keywords utilities
#' @seealso [maxmarg()]
#'
#' @examples
#' iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
#' gmap <- insert_pseudomarkers(iron$gmap, step=1)
#' probs <- calc_genoprob(iron, gmap, error_prob=0.002)
#' imp_geno <- maxmarg(probs)
#' ph_geno <- guess_phase(iron, imp_geno)
guess_phase <-
    function(cross, geno, deterministic=FALSE, cores=1)
{
    # if cross is phase-known, don't change geno
    if(!is.cross2(cross))
        stop('Input cross must have class "cross2"')
    if(is_phase_known(cross)) return(geno)

    geno <- unclass(geno) # treat geno as a plain list
    if(!is.list(geno)) stop("geno should be a list of genotype matrices")

    # match chromosomes
    if(n_chr(cross) != length(geno) ||
       any(names(cross$geno) != names(geno))) {
        chr_cross <- chr_names(cross)
        chr_geno <- names(geno)
        if(!any(chr_cross %in% chr_geno))
            stop("No chromosomes in common between cross and geno")
        common_chr <- chr_cross[chr_cross %in% chr_geno]
        cross <- subset(cross, chr=common_chr)
        geno <- geno[common_chr]
    }

    # match individuals, if X chr
    if(any(cross$is_x_chr)) {
        ind_cross <- names(cross$is_female)
        xchr <- chr_names(cross)[cross$is_x_chr]
        ind_geno <- rownames(geno[[xchr[1]]])

        common_ind <- ind_cross[ind_cross %in% ind_geno]
        if(length(common_ind) == 0)
            stop("No individuals in common between cross and geno")
        cross <- subset(cross, ind=common_ind)
        for(i in names(geno)[cross$is_x_chr])
            geno[[i]] <- geno[[i]][common_ind,,drop=FALSE]
        if(length(common_ind) < length(ind_cross) ||
           length(common_ind) < length(ind_geno))
            warning("On X chr, considering only the ", length(common_ind),
                    ifelse(length(common_ind)==1, " individual", " individuals"),
                    " in common between cross and geno")
    }

    # set up cluster; use quiet=TRUE
    cores <- setup_cluster(cores, TRUE)

    is_x_chr <- cross$is_x_chr
    crosstype <- cross$crosstype
    is_female <- cross$is_female

    by_chr_func <- function(i) {
        if(crosstype=="f2") {
            if(is_x_chr[i]) result <- .guess_phase_f2X(t(geno[[i]]), deterministic)
            else result <- .guess_phase_f2A(t(geno[[i]]), deterministic)
        }
        else {
            if(is_x_chr[i]) result <- .guess_phase_X(t(geno[[i]]), crosstype, is_female, deterministic)
            else result <- .guess_phase_A(t(geno[[i]]), crosstype, deterministic)
        }
        dn <- dimnames(geno[[i]])
        result <- aperm(result, c(3,2,1))
        dimnames(result) <- list(dn[[1]], dn[[2]], c("mom", "dad"))
        return(result)
    }

    result <- cluster_lapply(cores, seq_along(geno), by_chr_func)
    names(result) <- names(geno)
    class(result) <- c("phasedgeno", "list")
    result
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_genoprob.R
\name{calc_genoprob}
\alias{calc_genoprob}
\title{Calculate conditional genotype probabilities}
\usage{
calc_genoprob(
  cross,
  map = NULL,
  error_prob = 0.0001,
  map_function = c("haldane", "kosambi", "c-f", "morgan"),
  lowmem = FALSE,
  quiet = TRUE,
  cores = 1
)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{map}{Genetic map of markers. May include pseudomarker
locations (that is, locations that are not within the marker
genotype data). If NULL, the genetic map in \code{cross} is used.}

\item{error_prob}{Assumed genotyping error probability}

\item{map_function}{Character string indicating the map function to
use to convert genetic distances to recombination fractions.}

\item{lowmem}{If \code{FALSE}, split individuals into groups with
common sex and crossinfo and then precalculate the transition
matrices for a chromosome; potentially a lot faster but using more
memory.}

\item{quiet}{If \code{FALSE}, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
An object of class \code{"calc_genoprob"}: a list of three-dimensional arrays of probabilities,
individuals x genotypes x positions. (Note that the arrangement is
different from R/qtl.) Also contains four attributes:
\itemize{
\item \code{crosstype} - The cross type of the input \code{cross}.
\item \code{is_x_chr} - Logical vector indicating whether chromosomes
are to be treated as the X chromosome or not, from input \code{cross}.
\item \code{alleles} - Vector of allele codes, from input
\code{cross}.
\item \code{alleleprobs} - Logical value (\code{FALSE}) that
indicates whether the probabilities are compressed to allele
probabilities, as from \code{\link[=genoprob_to_alleleprob]{genoprob_to_alleleprob()}}.
}
}
\description{
Uses a hidden Markov model to calculate the probabilities of the
true underlying genotypes given the observed multipoint marker
data, with possible allowance for genotyping errors.
}
\details{
Let \eqn{O_k}{O[k]} denote the observed marker genotype at position
\eqn{k}, and \eqn{g_k}{g[k]} denote the corresponding true underlying
genotype.

We use the forward-backward equations to calculate
\eqn{\alpha_{kv} = \log Pr(O_1, \ldots, O_k, g_k = v)}{%
   a[k][v] = log Pr(O[1], \ldots, O[k], g[k] = v)}
and
\eqn{\beta_{kv} = \log Pr(O_{k+1}, \ldots, O_n | g_k = v)}{%
   b[k][v] = log Pr(O[k+1], \ldots, O[n] | g[k] = v)}

We then obtain
\eqn{Pr(g_k | O_1, \ldots, O_n) = \exp(\alpha_{kv} + \beta_{kv}) / s}{%
   Pr(g[k] | O[1], \ldots, O[n] = exp(a[k][v] + b[k][v]) / s}
where
\eqn{s = \sum_v \exp(\alpha_{kv} + \beta_{kv})}{%
   s = sum_v exp(a[k][v] + b[k][v])}
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
gmap_w_pmar <- insert_pseudomarkers(grav2$gmap, step=1)
probs <- calc_genoprob(grav2, gmap_w_pmar, error_prob=0.002)
}
\seealso{
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scale_kinship.R
\name{scale_kinship}
\alias{scale_kinship}
\title{Scale kinship matrix}
\usage{
scale_kinship(kinship)
}
\arguments{
\item{kinship}{A kinship matrix, or a list of such in the case of
the "leave one chromosome out" method, as calculated by
\code{\link[=calc_kinship]{calc_kinship()}}.}
}
\value{
A matrix or list of matrices, as with the input, but with
the matrices scaled to be like correlation matrices.
}
\description{
Scale kinship matrix to be like a correlation matrix.
}
\details{
We take
\ifelse{html}{\out{<em>c<sub>ij</sub> = k<sub>ij</sub> / &radic;(k<sub>ii</sub> k<sub>jj</sub>)</em>}}{\eqn{c_{ij} = k_{ij} / \sqrt{k_{ii} k_{jj}}}{c_ij = k_ij / sqrt(k_ii k_jj)}}
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
probs <- calc_genoprob(grav2, map, error_prob=0.002)
K <- calc_kinship(probs)
Ka <- scale_kinship(K)
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_marker.R
\name{find_marker}
\alias{find_marker}
\title{Find markers by chromosome position}
\usage{
find_marker(map, chr, pos = NULL, interval = NULL)
}
\arguments{
\item{map}{A map object: a list (corresponding to chromosomes) of
vectors of marker positions. Can also be a snpinfo object (data
frame with columns \code{chr} and \code{pos}; marker names taken from
column \code{snp} or if that doesn't exist from the row names)}

\item{chr}{A vector of chromosomes}

\item{pos}{A vector of positions}

\item{interval}{A pair of positions (provide either \code{pos} or \code{interval} but not both)}
}
\value{
A vector of marker names, either closest to the positions
specified by \code{pos}, or within the interval defined by
\code{interval}.
}
\description{
Find markers closest to specified set of positions, or within a specified interval.
}
\details{
If \code{pos} is provided, \code{interval} should not be, and vice versa.

If \code{pos} is provided, then \code{chr} and \code{pos} should
either be the same length, or one of them should have length 1 (to
be expanded to the length of the other).

If \code{interval} is provided, then \code{chr} should have length 1.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

# find markers by their genetic map positions
find_marker(iron$gmap, c(8, 11), c(37.7, 56.9))

# find markers by their physical map positions (two markers on chr 7)
find_marker(iron$pmap, 7, c(44.2, 108.9))

# find markers in an interval
find_marker(iron$pmap, 16, interval=c(35, 80))
}
\seealso{
\code{\link[=find_markerpos]{find_markerpos()}}, \code{\link[=find_index_snp]{find_index_snp()}}, \code{\link[=pull_genoprobpos]{pull_genoprobpos()}}, \code{\link[=pull_genoprobint]{pull_genoprobint()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cbind_scan1.R
\name{cbind.scan1}
\alias{cbind.scan1}
\title{Join genome scan results for different phenotypes.}
\usage{
\method{cbind}{scan1}(...)
}
\arguments{
\item{...}{Genome scan objects of class \code{"scan1"}, as produced by \code{\link[=scan1]{scan1()}}.
Must have the same map.}
}
\value{
An object of class `"scan1", like the inputs, but with the lod score columns
from the inputs combined as multiple columns in a single object.
}
\description{
Join multiple \code{\link[=scan1]{scan1()}} results for different phenotypes;
must have the same map.
}
\details{
If components \code{addcovar()}, \code{Xcovar},
\code{intcovar}, \code{weights} do not match between objects, we
omit this information.

If \code{hsq} present but has differing numbers of rows, we omit this information.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
probs <- calc_genoprob(grav2, map, error_prob=0.002)
phe1 <- grav2$pheno[,1,drop=FALSE]
phe2 <- grav2$pheno[,2,drop=FALSE]

out1 <- scan1(probs, phe1) # phenotype 1
out2 <- scan1(probs, phe2) # phenotype 2
out <- cbind(out1, out2)

}
\seealso{
\code{\link[=rbind.scan1]{rbind.scan1()}}, \code{\link[=scan1]{scan1()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insert_pseudomarkers.R
\name{insert_pseudomarkers}
\alias{insert_pseudomarkers}
\title{Insert pseudomarkers into a marker map}
\usage{
insert_pseudomarkers(
  map,
  step = 0,
  off_end = 0,
  stepwidth = c("fixed", "max"),
  pseudomarker_map = NULL,
  tol = 0.01,
  cores = 1
)
}
\arguments{
\item{map}{A list of numeric vectors; each vector gives marker
positions for a single chromosome.}

\item{step}{Distance between pseudomarkers and markers; if
\code{step=0} no pseudomarkers are inserted.}

\item{off_end}{Distance beyond terminal markers in which to insert
pseudomarkers.}

\item{stepwidth}{Indicates whether to use a fixed grid
(\code{stepwidth="fixed"}) or to use the maximal distance between
pseudomarkers to ensure that no two adjacent markers/pseudomarkers
are more than \code{step} apart.}

\item{pseudomarker_map}{A map of pseudomarker locations; if provided the
\code{step}, \code{off_end}, and \code{stepwidth} arguments are
ignored.}

\item{tol}{Tolerance for determining whether a pseudomarker would duplicate a marker position.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A list like the input \code{map} with pseudomarkers
inserted. Will also have an attribute \code{"is_x_chr"}, taken
from the input \code{map}.
}
\description{
Insert pseudomarkers into a map of genetic markers
}
\details{
If \code{stepwidth="fixed"}, a grid of pseudomarkers is
added to the marker map.

If \code{stepwidth="max"}, a minimal set of pseudomarkers are
added, so that the maximum distance between adjacent markers or
pseudomarkers is at least \code{step}. If two adjacent markers are
separated by less than \code{step}, no pseudomarkers will be added
to the interval. If they are more then \code{step} apart, a set of
equally-spaced pseudomarkers will be added.

If \code{pseudomarker_map} is provided, then the \code{step},
\code{off_end}, and \code{stepwidth} arguments are ignored, and the
input \code{pseudomarker_map} is taken to be the set of
pseudomarker positions.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
gmap_w_pmar <- insert_pseudomarkers(iron$gmap, step=1)
}
\seealso{
\code{\link[=calc_genoprob]{calc_genoprob()}}, \code{\link[=calc_grid]{calc_grid()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map_to_grid.R
\name{map_to_grid}
\alias{map_to_grid}
\title{Subset a map to positions on a grid}
\usage{
map_to_grid(map, grid)
}
\arguments{
\item{map}{A list of vectors of marker positions.}

\item{grid}{A list of logical vectors (aligned with
\code{map}), with TRUE indicating the position is on the grid.}
}
\value{
Same list as input, but subset to just include
pseudomarkers along a grid.
}
\description{
Subset a map object to the locations on some grid.
}
\details{
This is generally for the case of a map created with
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}} with \code{step}>0 and
\code{stepwidth="fixed"}, so that the pseudomarkers form a grid
along each chromosome.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map_w_pmar <- insert_pseudomarkers(grav2$gmap, step=1)
sapply(map_w_pmar, length)
grid <- calc_grid(grav2$gmap, step=1)
map_sub <- map_to_grid(map_w_pmar, grid)
sapply(map_sub, length)
}
\seealso{
\code{\link[=calc_grid]{calc_grid()}}, \code{\link[=probs_to_grid]{probs_to_grid()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_raw_summaries.R
\name{calc_raw_het}
\alias{calc_raw_het}
\title{Calculate estimated heterozygosity from raw SNP genotypes}
\usage{
calc_raw_het(cross, by = c("individual", "marker"))
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{by}{Indicates whether to summarize by founder strain (\code{"individual"}) or by marker.}
}
\value{
A vector of heterozygosities, one for each individual or marker.
}
\description{
Calculate estimated heterozygosity for each individual from raw SNP genotypes
}
\examples{
\dontrun{
# load example data and calculate genotype probabilities
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
DOex_het <- calc_raw_het(DOex)
}
}
\seealso{
\code{\link[=recode_snps]{recode_snps()}}, \code{\link[=calc_raw_maf]{calc_raw_maf()}}, \code{\link[=calc_raw_founder_maf]{calc_raw_founder_maf()}}, \code{\link[=calc_raw_geno_freq]{calc_raw_geno_freq()}}, \code{\link[=calc_het]{calc_het()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan1coef.R
\name{scan1coef}
\alias{scan1coef}
\title{Calculate QTL effects in scan along one chromosome}
\usage{
scan1coef(
  genoprobs,
  pheno,
  kinship = NULL,
  addcovar = NULL,
  nullcovar = NULL,
  intcovar = NULL,
  weights = NULL,
  contrasts = NULL,
  model = c("normal", "binary"),
  zerosum = TRUE,
  se = FALSE,
  hsq = NULL,
  reml = TRUE,
  ...
)
}
\arguments{
\item{genoprobs}{Genotype probabilities as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{pheno}{A numeric vector of phenotype values (just one phenotype, not a matrix of them)}

\item{kinship}{Optional kinship matrix, or a list of kinship matrices (one
per chromosome), in order to use the LOCO (leave one chromosome
out) method.}

\item{addcovar}{An optional numeric matrix of additive covariates.}

\item{nullcovar}{An optional numeric matrix of additional additive
covariates that are used under the null hypothesis (of no QTL) but
not under the alternative (with a QTL). This is needed for the X
chromosome, where we might need sex as a additive covariate under
the null hypothesis, but we wouldn't want to include it under the
alternative as it would be collinear with the QTL effects. Only
used if \code{kinship} is provided but \code{hsq} is not, to get
estimate of residual heritability.}

\item{intcovar}{An optional numeric matrix of interactive covariates.}

\item{weights}{An optional numeric vector of positive weights for the
individuals. As with the other inputs, it must have \code{names}
for individual identifiers.}

\item{contrasts}{An optional numeric matrix of genotype contrasts, size
genotypes x genotypes. For an intercross, you might use
\code{cbind(mu=c(1,1,1), a=c(-1, 0, 1), d=c(0, 1, 0))} to get
mean, additive effect, and dominance effect. The default is the
identity matrix.}

\item{model}{Indicates whether to use a normal model (least
squares) or binary model (logistic regression) for the phenotype.
If \code{model="binary"}, the phenotypes must have values in \eqn{[0, 1]}.}

\item{zerosum}{If TRUE, force the genotype or allele coefficients
sum to 0 by subtracting their mean and add another column with
the mean. Ignored if \code{contrasts} is provided.
\strong{Note}: the standard errors aren't quite right when \code{zerosum=TRUE};
use \code{\link[=fit1]{fit1()}} to get proper standard errors.}

\item{se}{If TRUE, also calculate the standard errors.}

\item{hsq}{(Optional) residual heritability; used only if
\code{kinship} provided.}

\item{reml}{If \code{kinship} provided: if \code{reml=TRUE}, use
REML; otherwise maximum likelihood.}

\item{...}{Additional control parameters; see Details;}
}
\value{
An object of class \code{"scan1coef"}: a matrix of estimated regression coefficients, of dimension
positions x number of effects. The number of effects is
\code{n_genotypes + n_addcovar + (n_genotypes-1)*n_intcovar}.
May also contain the following attributes:
\itemize{
\item \code{SE} - Present if \code{se=TRUE}: a matrix of estimated
standard errors, of same dimension as \code{coef}.
\item \code{sample_size} - Vector of sample sizes used for each
phenotype
}
}
\description{
Calculate QTL effects in scan along one chromosome with a
single-QTL model using Haley-Knott regression or a linear mixed
model (the latter to account for a residual polygenic effect), with
possible allowance for covariates.
}
\details{
For each of the inputs, the row names are used as
individual identifiers, to align individuals.

If \code{kinship} is absent, Haley-Knott regression is performed.
If \code{kinship} is provided, a linear mixed model is used, with a
polygenic effect estimated under the null hypothesis of no (major)
QTL, and then taken as fixed as known in the genome scan.

If \code{contrasts} is provided, the genotype probability matrix,
\eqn{P}, is post-multiplied by the contrasts matrix, \eqn{A}, prior
to fitting the model. So we use \eqn{P \cdot A}{P A} as the \eqn{X}
matrix in the model. One might view the rows of
\ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}
as the set of contrasts, as the estimated effects are the estimated
genotype effects pre-multiplied by
\ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}.

The \code{...} argument can contain several additional control
parameters; suspended for simplicity (or confusion, depending on
your point of view). \code{tol} is used as a tolerance value for linear
regression by QR decomposition (in determining whether columns are
linearly dependent on others and should be omitted); default
\code{1e-12}. \code{maxit} is the maximum number of iterations for
converence of the iterative algorithm used when \code{model=binary}.
\code{bintol} is used as a tolerance for converence for the iterative
algorithm used when \code{model=binary}. \code{eta_max} is the maximum value
for the "linear predictor" in the case \code{model="binary"} (a bit of a
technicality to avoid fitted values exactly at 0 or 1).
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c(7,19)] # reduce to chr 7 and 19}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)
\dontshow{probs[["7"]] <- probs[["7"]][,,1:5] # reduce to very small number}

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno[,1]
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)

# calculate coefficients for chromosome 7
coef <- scan1coef(probs[,"7"], pheno, addcovar=covar)

# leave-one-chromosome-out kinship matrix for chr 7
kinship7 <- calc_kinship(probs, "loco")[["7"]]

# calculate coefficients for chromosome 7, adjusting for residual polygenic effect
coef_pg <- scan1coef(probs[,"7"], pheno, kinship7, addcovar=covar)

}
\references{
Haley CS, Knott SA (1992) A simple
regression method for mapping quantitative trait loci in line
crosses using flanking markers.  Heredity 69:315--324.

Kang HM, Zaitlen NA, Wade CM, Kirby A, Heckerman D, Daly MJ, Eskin
E (2008) Efficient control of population structure in model
organism association mapping. Genetics 178:1709--1723.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genoprob_to_alleleprob.R
\name{genoprob_to_alleleprob}
\alias{genoprob_to_alleleprob}
\title{Convert genotype probabilities to allele probabilities}
\usage{
genoprob_to_alleleprob(probs, quiet = TRUE, cores = 1)
}
\arguments{
\item{probs}{Genotype probabilities, as calculated from
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{quiet}{IF \code{FALSE}, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
An object of class \code{"calc_genoprob"}, like the input \code{probs},
but with probabilities collapsed to alleles rather than genotypes. See \code{\link[=calc_genoprob]{calc_genoprob()}}.
}
\description{
Reduce genotype probabilities (as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}) to allele probabilities.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
gmap_w_pmar <- insert_pseudomarkers(iron, step=1)
probs <- calc_genoprob(iron, gmap_w_pmar, error_prob=0.002)
allele_probs <- genoprob_to_alleleprob(probs)
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan1gen.R
\name{scan1gen}
\alias{scan1gen}
\title{General genome scan with a single-QTL model}
\usage{
scan1gen(
  genoprobs,
  pheno,
  kinship = NULL,
  addcovar = NULL,
  Xcovar = NULL,
  intcovar = NULL,
  weights = NULL,
  func,
  vectorize_func = TRUE,
  cores = 1,
  ...
)
}
\arguments{
\item{genoprobs}{Genotype probabilities as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{pheno}{A numeric matrix of phenotypes, individuals x phenotypes.}

\item{kinship}{Optional kinship matrix, or a list of kinship matrices (one
per chromosome), in order to use the LOCO (leave one chromosome
out) method.}

\item{addcovar}{An optional numeric matrix of additive covariates.}

\item{Xcovar}{An optional numeric matrix with additional additive covariates used for
null hypothesis when scanning the X chromosome.}

\item{intcovar}{An numeric optional matrix of interactive covariates.}

\item{weights}{An optional numeric vector of positive weights for the
individuals. As with the other inputs, it must have \code{names}
for individual identifiers.}

\item{func}{Function to calculate log10 likelihood. It is called as
\code{func(pr, pheno, kinship, addcovar, intcovar, weights, ...)}
where \code{pr} is for the genotype probabilities at a fixed
position. It will also be used to calculate the null log10
likelihood, with \code{pr} being NULL in that case.}

\item{vectorize_func}{If TRUE (the default), assume that \code{func}
takes only a single phenotype. If FALSE, assume \code{func} is able
to take a matrix of phenotypes and return a vector of log10
likelihood values.}

\item{cores}{Number of CPU cores to use, for parallel calculations.}

\item{...}{Additional arguments passed to \code{func()}.}
}
\value{
An object of class \code{"scan1"}: a matrix of LOD scores, positions x phenotypes.
}
\description{
General genome scan with a single-QTL model, with possible allowance for covariates
and a polygeneic effect, but taking a general function that calculates a LOD score
at a single position.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c("19", "X")] # subset to chr 19 and X}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# covariates for X chr under null
Xcovar <- get_x_covar(iron)

# create binary trait
bin_pheno <- setNames(as.numeric(iron$pheno[,1] > median(iron$pheno[,1])),
                      rownames(iron$pheno))

ll_glm <-
   function(pr, pheno, addcovar=NULL, ...)
{
   formula <- ifelse(is.null(pr), "pheno ~ 1", "pheno ~ pr")
   if(!is.null(addcovar)) formula <- paste(formula, "+ addcovar")

    glm_out <- glm(as.formula(formula), family=binomial(link=probit))
    -glm_out$deviance/(2*log(10)) # log10 likelihood
}

# perform genome scan using glm() with probit link function
out <- scan1gen(probs, bin_pheno, Xcovar=Xcovar, func=ll_glm)

}
\seealso{
\code{\link[=scan1]{scan1()}}, \code{\link[=scan1perm]{scan1perm()}}, \code{\link[=scan1max]{scan1max()}}, \href{https://kbroman.org/qtl2/assets/vignettes/scan1gen.html}{scan1gen examples}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch_vec.R
\name{batch_vec}
\alias{batch_vec}
\title{Split vector into batches}
\usage{
batch_vec(vec, batch_size = NULL, n_cores = 1)
}
\arguments{
\item{vec}{A vector to be split into batches}

\item{batch_size}{Maximum size for each batch}

\item{n_cores}{Number of compute cores, to be used as a minimum number of batches.}
}
\value{
A list of vectors, each no longer than \code{batch_size}, and with at least \code{n_cores} componenets.
}
\description{
Split a vector into batches, each no longer than \code{batch_size} and
creating at least \code{n_cores} batches, for use in parallel
calculations.
}
\examples{
vec_split <- batch_vec(1:304, 50, 8)
vec_split2 <- batch_vec(1:304, 50)
}
\seealso{
\code{\link[=batch_cols]{batch_cols()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_geno_freq.R
\name{calc_geno_freq}
\alias{calc_geno_freq}
\title{Calculate genotype frequencies}
\usage{
calc_geno_freq(probs, by = c("individual", "marker"), omit_x = TRUE)
}
\arguments{
\item{probs}{List of arrays of genotype probabilities, as
calculated by \code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{by}{Whether to summarize by individual or marker}

\item{omit_x}{If TRUE, results are just for the autosomes. If
FALSE, results are a list of length two, containing the results
for the autosomes and those for the X chromosome.}
}
\value{
If \code{omit_x=TRUE}, the result is a matrix of genotype
frequencies; columns are genotypes and rows are either individuals
or markers.

If necessary (that is, if \code{omit_x=FALSE}, the data include the
X chromosome, and the set of genotypes on the X chromosome are
different than on the autosomes), the result is a list with two
components (for the autosomes and for the X chromosome), each being
a matrix of genotype frequencies.
}
\description{
Calculate genotype frequencies, by individual or by marker
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
p <- calc_genoprob(iron, err=0.002)

# genotype frequencies by marker
tab_g <- calc_geno_freq(p, "marker")

# allele frequencies by marker
ap <- genoprob_to_alleleprob(p)
tab_a <- calc_geno_freq(ap, "marker")

}
\seealso{
\code{\link[=calc_raw_geno_freq]{calc_raw_geno_freq()}}, \code{\link[=calc_het]{calc_het()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_peaks.R
\name{find_peaks}
\alias{find_peaks}
\title{Find peaks in a set of LOD curves}
\usage{
find_peaks(
  scan1_output,
  map,
  threshold = 3,
  peakdrop = Inf,
  drop = NULL,
  prob = NULL,
  thresholdX = NULL,
  peakdropX = NULL,
  dropX = NULL,
  probX = NULL,
  expand2markers = TRUE,
  sort_by = c("column", "pos", "lod"),
  cores = 1
)
}
\arguments{
\item{scan1_output}{An object of class \code{"scan1"} as returned by
\code{\link[=scan1]{scan1()}}.}

\item{map}{A list of vectors of marker positions, as produced by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}. Can also be an indexed SNP info table,
as from \code{\link[=index_snps]{index_snps()}} or \code{\link[=scan1snps]{scan1snps()}}.}

\item{threshold}{Minimum LOD score for a peak (can be a vector with
separate thresholds for each lod score column in
\code{scan1_output})}

\item{peakdrop}{Amount that the LOD score must drop between peaks,
if multiple peaks are to be defined on a chromosome. (Can be a vector with
separate values for each lod score column in
\code{scan1_output}.)}

\item{drop}{If provided, LOD support intervals are included in the
results, and this indicates the amount to drop in the support
interval. (Can be a vector with
separate values for each lod score column in
\code{scan1_output}.) Must be \eqn{\le} \code{peakdrop}}

\item{prob}{If provided, Bayes credible intervals are included in the
results, and this indicates the nominal coverage.
(Can be a vector with
separate values for each lod score column in
\code{scan1_output}.) Provide just one of \code{drop} and \code{prob}.}

\item{thresholdX}{Separate threshold for the X chromosome; if
unspecified, the same threshold is used for both autosomes and the
X chromosome. (Like \code{threshold}, this can be a vector with
separate thresholds for each lod score column.)}

\item{peakdropX}{Like \code{peakdrop}, but for the X chromosome; if
unspecified, the same value is used for both autosomes and the X
chromosome.  (Can be a vector with separate values for each lod
score column in \code{scan1_output}.)}

\item{dropX}{Amount to drop for LOD support intervals on the X
chromosome.  Ignored if \code{drop} is not provided. (Can be a
vector with separate values for each lod score column in
\code{scan1_output}.)}

\item{probX}{Nominal coverage for Bayes intervals on the X
chromosome.  Ignored if \code{prob} is not provided. (Can be a
vector with separate values for each lod score column in
\code{scan1_output}.)}

\item{expand2markers}{If TRUE (and if \code{drop} or \code{prob} is
provided, so that QTL intervals are calculated), QTL intervals are
expanded so that their endpoints are at genetic markers.}

\item{sort_by}{Indicates whether to sort the rows by lod column,
genomic position, or LOD score.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A data frame with each row being a single peak on a single
chromosome for a single LOD score column, and with columns
\itemize{
\item \code{lodindex} - lod column index
\item \code{lodcolumn} - lod column name
\item \code{chr} - chromosome ID
\item \code{pos} - peak position
\item \code{lod} - lod score at peak
}

If \code{drop} or \code{prob} is provided, the results will include
two additional columns: \code{ci_lo} and \code{ci_hi}, with the
endpoints of the LOD support intervals or Bayes credible wintervals.
}
\description{
Find peaks in a set of LOD curves (output from \code{\link[=scan1]{scan1()}}
}
\details{
For each lod score column on each chromosome, we return a
set of peaks defined as local maxima that exceed the specified
\code{threshold}, with the requirement that the LOD score must have
dropped by at least \code{peakdrop} below the lowest of any two
adjacent peaks.

At a given peak, if there are ties, with multiple positions jointly
achieving the maximum LOD score, we take the average of these
positions as the location of the peak.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c(1,2,7,8,9,13,15,16,19)]}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# find just the highest peak on each chromosome
find_peaks(out, map, threshold=3)

# possibly multiple peaks per chromosome
find_peaks(out, map, threshold=3, peakdrop=1)

# possibly multiple peaks, also getting 1-LOD support intervals
find_peaks(out, map, threshold=3, peakdrop=1, drop=1)

# possibly multiple peaks, also getting 90\% Bayes intervals
find_peaks(out, map, threshold=3, peakdrop=1, prob=0.9)
}
\seealso{
\code{\link[=scan1]{scan1()}}, \code{\link[=lod_int]{lod_int()}}, \code{\link[=bayes_int]{bayes_int()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_genoprobcomp.R
\name{plot_genoprobcomp}
\alias{plot_genoprobcomp}
\title{Plot comparison of two sets of genotype probabilities}
\usage{
plot_genoprobcomp(
  probs1,
  probs2,
  map,
  ind = 1,
  chr = NULL,
  geno = NULL,
  threshold = 0,
  n_colors = 256,
  swap_axes = FALSE,
  ...
)
}
\arguments{
\item{probs1}{Genotype probabilities (as produced by \code{\link[=calc_genoprob]{calc_genoprob()}})
or allele dosages (as produced by \code{\link[=genoprob_to_alleleprob]{genoprob_to_alleleprob()}}).}

\item{probs2}{A second set of genotype probabilities, just like \code{probs1}.}

\item{map}{Marker map (a list of vectors of marker positions).}

\item{ind}{Individual to plot, either a numeric index or an ID.}

\item{chr}{Selected chromosome to plot; a single character string.}

\item{geno}{Optional vector of genotypes or alleles to be shown
(vector of integers or character strings)}

\item{threshold}{Threshold for genotype probabilities; only genotypes that achieve
this value somewhere on the chromosome (in one or the other set of probabilities) will be shown.}

\item{n_colors}{Number of colors in each color scale.}

\item{swap_axes}{If TRUE, swap the axes, so that the genotypes are
on the x-axis and the chromosome position is on the y-axis.}

\item{...}{Additional graphics parameters passed to \code{\link[graphics:image]{graphics::image()}}.}
}
\value{
None.
}
\description{
Plot a comparison of two sets of genotype probabilities for one individual on one chromosome, as a heat map.
}
\details{
We plot the first set of probabilities in the range white to blue
and the second set in the range white to red and attempt to combine
them, for colors that are white, some amount of blue or red, or
where both are large something like blackish purple.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
iron <- iron[228,"1"]   # subset to one individual on chr 1
map <- insert_pseudomarkers(iron$gmap, step=5)

# introduce genotype error and look at difference in genotype probabilities
pr_ne <- calc_genoprob(iron, map, error_prob=0.002)
iron$geno[[1]][1,2] <- 3
pr_e <- calc_genoprob(iron, map, error_prob=0.002)

# image of probabilities + comparison
\dontshow{old_mfrow <- par("mfrow")}
par(mfrow=c(3,1))
plot_genoprob(pr_ne, map, main="No error")
plot_genoprob(pr_e, map, main="With an error")
plot_genoprobcomp(pr_ne, pr_e, map, main="Comparison")
\dontshow{par(mfrow=old_mfrow)}

}
\seealso{
\code{\link[=plot_genoprob]{plot_genoprob()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/viterbi.R
\name{viterbi}
\alias{viterbi}
\title{Calculate most probable sequence of genotypes}
\usage{
viterbi(
  cross,
  map = NULL,
  error_prob = 0.0001,
  map_function = c("haldane", "kosambi", "c-f", "morgan"),
  lowmem = FALSE,
  quiet = TRUE,
  cores = 1
)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{map}{Genetic map of markers. May include pseudomarker
locations (that is, locations that are not within the marker
genotype data). If NULL, the genetic map in \code{cross} is used.}

\item{error_prob}{Assumed genotyping error probability}

\item{map_function}{Character string indicating the map function to
use to convert genetic distances to recombination fractions.}

\item{lowmem}{If \code{FALSE}, split individuals into groups with
common sex and crossinfo and then precalculate the transition
matrices for a chromosome; potentially a lot faster but using more
memory.}

\item{quiet}{If \code{FALSE}, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
An object of class \code{"viterbi"}: a list of two-dimensional
arrays of imputed genotypes, individuals x positions.
Also contains three attributes:
\itemize{
\item \code{crosstype} - The cross type of the input \code{cross}.
\item \code{is_x_chr} - Logical vector indicating whether chromosomes
are to be treated as the X chromosome or not, from input \code{cross}.
\item \code{alleles} - Vector of allele codes, from input
\code{cross}.
}
}
\description{
Uses a hidden Markov model to calculate arg max Pr(g | O) where g
is the underlying sequence of true genotypes and O is the observed
multipoint marker data, with possible allowance for genotyping
errors.
}
\details{
We use a hidden Markov model to find, for each individual
on each chromosome, the most probable sequence of underlying
genotypes given the observed marker data.

Note that we break ties at random, and our method for doing this
may introduce some bias.

Consider the results with caution; the most probable sequence can
have very low probability, and can have features that are quite
unusual (for example, the number of recombination events can be too
small). In most cases, the results of a single imputation with
\code{\link[=sim_geno]{sim_geno()}} will be more realistic.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map_w_pmar <- insert_pseudomarkers(grav2$gmap, step=1)
g <- viterbi(grav2, map_w_pmar, error_prob=0.002)
}
\seealso{
\code{\link[=sim_geno]{sim_geno()}}, \code{\link[=maxmarg]{maxmarg()}}, \code{\link[=cbind.viterbi]{cbind.viterbi()}}, \code{\link[=rbind.viterbi]{rbind.viterbi()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_coef.R
\name{plot_coef}
\alias{plot_coef}
\alias{plot_coefCC}
\alias{plot.scan1coef}
\title{Plot QTL effects along chromosome}
\usage{
plot_coef(
  x,
  map,
  columns = NULL,
  col = NULL,
  scan1_output = NULL,
  add = FALSE,
  gap = NULL,
  top_panel_prop = 0.65,
  legend = NULL,
  ...
)

plot_coefCC(
  x,
  map,
  columns = 1:8,
  col = qtl2::CCcolors,
  scan1_output = NULL,
  add = FALSE,
  gap = NULL,
  top_panel_prop = 0.65,
  legend = NULL,
  ...
)

\method{plot}{scan1coef}(
  x,
  map,
  columns = 1,
  col = NULL,
  scan1_output = NULL,
  add = FALSE,
  gap = NULL,
  top_panel_prop = 0.65,
  legend = NULL,
  ...
)
}
\arguments{
\item{x}{Estimated QTL effects ("coefficients") as obtained from
\code{\link[=scan1coef]{scan1coef()}}.}

\item{map}{A list of vectors of marker positions, as produced by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}.}

\item{columns}{Vector of columns to plot}

\item{col}{Vector of colors, same length as \code{columns}. If
NULL, some default choices are made.}

\item{scan1_output}{If provided, we make a two-panel plot with
coefficients on top and LOD scores below. Should have just one LOD
score column; if multiple, only the first is used.}

\item{add}{If TRUE, add to current plot (must have same map and
chromosomes).}

\item{gap}{Gap between chromosomes. The default is 1\% of the total genome length.}

\item{top_panel_prop}{If \code{scan1_output} provided, this gives the
proportion of the plot that is devoted to the top panel.}

\item{legend}{Location of legend, such as \code{"bottomleft"} or \code{"topright"} (NULL for no legend)}

\item{...}{Additional graphics parameters.}
}
\value{
None.
}
\description{
Plot estimated QTL effects along a chromosomes.
}
\details{
\code{plot_coefCC()} is the same as \code{plot_coef()}, but forcing
\code{columns=1:8} and using the Collaborative Cross colors,
\link{CCcolors}.
}
\section{Hidden graphics parameters}{

A number of graphics parameters can be passed via \code{...}. For
example, \code{bgcolor} to control the background color, and things
like \code{ylab} and \code{ylim}. These are not included as formal
parameters in order to avoid cluttering the function definition.

In the case that \code{scan1_output} is provided, \code{col},
\code{ylab}, and \code{ylim} all control the panel with estimated
QTL effects, while \code{col_lod}, \code{ylab_lod}, and
\code{ylim_lod} control the LOD curve panel.

If \code{legend} is indicated so that a legend is shown, \code{legend_lab}
controls the labels in the legend, and \code{legend_ncol} indicates the
number of columns in the legend.
}

\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno[,1]
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)

# calculate coefficients for chromosome 7
coef <- scan1coef(probs[,7], pheno, addcovar=covar)

# plot QTL effects (note the need to subset the map object, for chromosome 7)
plot(coef, map[7], columns=1:3, col=c("slateblue", "violetred", "green3"))
}
\seealso{
\link{CCcolors}, \code{\link[=plot_scan1]{plot_scan1()}}, \code{\link[=plot_snpasso]{plot_snpasso()}}, \code{\link[=plot_ci]{plot_ci()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_kinship.R
\name{calc_kinship}
\alias{calc_kinship}
\title{Calculate kinship matrix}
\usage{
calc_kinship(
  probs,
  type = c("overall", "loco", "chr"),
  omit_x = FALSE,
  use_allele_probs = TRUE,
  quiet = TRUE,
  cores = 1
)
}
\arguments{
\item{probs}{Genotype probabilities, as calculated from
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{type}{Indicates whether to calculate the overall kinship
(\code{"overall"}, using all chromosomes), the kinship matrix
leaving out one chromosome at a time (\code{"loco"}), or the
kinship matrix for each chromosome (\code{"chr"}).}

\item{omit_x}{If \code{TRUE}, only use the autosomes; ignored when
\code{type="chr"}.}

\item{use_allele_probs}{If \code{TRUE}, assess similarity with
allele probabilities (that is, first run
\code{\link[=genoprob_to_alleleprob]{genoprob_to_alleleprob()}}); otherwise use the genotype
probabilities.}

\item{quiet}{IF \code{FALSE}, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
If \code{type="overall"} (the default), a matrix of
proportion of matching alleles. Otherwise a list with one matrix
per chromosome.
}
\description{
Calculate genetic similarity among individuals (kinship matrix)
from conditional genotype probabilities.
}
\details{
If \code{use_allele_probs=TRUE} (the default), we first
convert the genotype probabilities to allele
probabilities (using \code{\link[=genoprob_to_alleleprob]{genoprob_to_alleleprob()}}).

We then calculate
\eqn{\sum_{kl}(p_{ikl} p_{jkl})}{sum_kl (p_ikl p_jkl)}
where \eqn{k} = position, \eqn{l} = allele, and \eqn{i,j}
are two individuals.

For crosses with just two possible genotypes (e.g., backcross), we
don't convert to allele probabilities but just use the original
genotype probabilities.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
probs <- calc_genoprob(grav2, map, error_prob=0.002)
K <- calc_kinship(probs)

# using only markers/pseudomarkers on the grid
grid <- calc_grid(grav2$gmap, step=1)
probs_sub <- probs_to_grid(probs, grid)
K_grid <- calc_kinship(probs_sub)
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_ibd_segments.R
\name{find_ibd_segments}
\alias{find_ibd_segments}
\title{Find IBD segments for a set of strains}
\usage{
find_ibd_segments(geno, map, min_lod = 15, error_prob = 0.001, cores = 1)
}
\arguments{
\item{geno}{List of matrices of founder genotypes. The matrices
correspond to the genotypes on chromosomes and are arrayed as
founders x markers.}

\item{map}{List of vectors of marker positions}

\item{min_lod}{Threshold for minimum LOD score for a segment}

\item{error_prob}{Genotyping error/mutation probability}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A data frame whose rows are IBD segments and whose columns
are:
\itemize{
\item Strain 1
\item Strain 2
\item Chromosome
\item Left marker
\item Right marker
\item Left position
\item Right position
\item Left marker index
\item Right marker index
\item Interval length
\item Number of markers
\item Number of mismatches
\item LOD score
}
}
\description{
Find IBD segments (regions with a lot of shared SNP genotypes) for
a set of strains
}
\details{
For each strain pair on each chromosome, we consider all
marker intervals and calculate a LOD score comparing the two
hypotheses: that the strains are IBD in the interval, vs. that
they are not. We assume that the two strains are homozygous at
all markers, and use the model from Broman and Weber (1999),
which assumes linkage equilibrium between markers and uses a
simple model for genotype frequencies in the presence of genotyping
errors or mutations.

Note that inference of IBD segments is heavily dependent on how
SNPs were chosen to be genotyped. (For example, were the SNPs ascertained
based on their polymorphism between a particular strain pair?)
}
\examples{
\dontrun{
# load DO data from Recla et al. (2014) Mamm Genome 25:211-222.
recla <- read_cross2("https://raw.githubusercontent.com/rqtl/qtl2data/main/DO_Recla/recla.zip")

# grab founder genotypes and physical map
fg <- recla$founder_geno
pmap <- recla$pmap

# find shared segments
(segs <- find_ibd_segments(fg, pmap, min_lod=10, error_prob=0.0001))
}

}
\references{
Broman KW, Weber JL (1999) Long homozygous chromosomal segments in
reference families from the Centre d’Étude du Polymorphisme Humain.
Am J Hum Genet 65:1493--1500.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset_sim_geno.R
\name{subset.sim_geno}
\alias{subset.sim_geno}
\alias{[.sim_geno}
\title{Subsetting imputed genotypes}
\usage{
\method{subset}{sim_geno}(x, ind = NULL, chr = NULL, ...)

\method{[}{sim_geno}(x, ind = NULL, chr = NULL)
}
\arguments{
\item{x}{Imputed genotypes as output from \code{\link[=sim_geno]{sim_geno()}}.}

\item{ind}{A vector of individuals: numeric indices, logical
values, or character string IDs}

\item{chr}{A vector of chromosomes: logical values, or character
string IDs. Numbers are interpreted as character string IDs.}

\item{...}{Ignored.}
}
\value{
An object of class \code{"sim_geno"}, like the input
with the selected individuals and/or chromsomes; see \code{\link[=sim_geno]{sim_geno()}}.
}
\description{
Pull out a specified set of individuals and/or chromosomes from
the results of \code{\link[=sim_geno]{sim_geno()}}.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
\dontshow{grav2 <- grav2[1:8,c(1,2)]}
dr <- sim_geno(grav2, n_draws=4)
# keep just individuals 1:5, chromosome 2
drsub <- dr[1:5,2]
# keep just chromosome 2
drsub2 <- dr[,2]
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_entropy.R
\name{calc_entropy}
\alias{calc_entropy}
\title{Calculate entropy of genotype probability distribution}
\usage{
calc_entropy(probs, quiet = TRUE, cores = 1)
}
\arguments{
\item{probs}{Genotype probabilities, as calculated from
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{quiet}{IF \code{FALSE}, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A list of matrices (each matrix is a chromosome and is arranged as individuals x markers).
}
\description{
For each individual at each genomic position, calculate the entropy
of the genotype probability distribution, as a quantitative summary
of the amount of missing information.
}
\details{
We calculate -sum(p log_2 p), where we take 0 log 0 = 0.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
\dontshow{grav2 <- grav2[,c(4,5)] # subset to chr 3}
probs <- calc_genoprob(grav2, error_prob=0.002)
e <- calc_entropy(probs)
e <- do.call("cbind", e) # combine chromosomes into one big matrix

# summarize by individual
mean_ind <- rowMeans(e)

# summarize by marker
mean_marker <- colMeans(e)
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_scan1.R
\name{plot_scan1}
\alias{plot_scan1}
\alias{plot.scan1}
\title{Plot a genome scan}
\usage{
plot_scan1(x, map, lodcolumn = 1, chr = NULL, add = FALSE, gap = NULL, ...)

\method{plot}{scan1}(x, map, lodcolumn = 1, chr = NULL, add = FALSE, gap = NULL, ...)
}
\arguments{
\item{x}{An object of class \code{"scan1"}, as output by \code{\link[=scan1]{scan1()}}.}

\item{map}{A list of vectors of marker positions, as produced by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}.}

\item{lodcolumn}{LOD score column to plot (a numeric index, or a
character string for a column name). Only one value allowed.}

\item{chr}{Selected chromosomes to plot; a vector of character
strings.}

\item{add}{If TRUE, add to current plot (must have same map and
chromosomes).}

\item{gap}{Gap between chromosomes. The default is 1\% of the total genome length.}

\item{...}{Additional graphics parameters.}
}
\value{
None.
}
\description{
Plot LOD curves for a genome scan
}
\section{Hidden graphics parameters}{

A number of graphics parameters can be passed via \code{...}. For
example, \code{bgcolor} to control the background color and
\code{altbgcolor} to control the background color on alternate chromosomes.
\code{col} controls the color of lines/curves; \code{altcol} can be used if
you want alternative chromosomes in different colors.
These are not included as formal parameters in order to avoid
cluttering the function definition.
}

\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# plot the results for selected chromosomes
ylim <- c(0, maxlod(out)*1.02) # need to strip class to get overall max LOD
chr <- c(2,7,8,9,15,16)
plot(out, map, chr=chr, ylim=ylim)
plot(out, map, lodcolumn=2, chr=chr, col="violetred", add=TRUE)
legend("topleft", lwd=2, col=c("darkslateblue", "violetred"), colnames(out),
       bg="gray90")

# plot just one chromosome
plot(out, map, chr=8, ylim=ylim)
plot(out, map, chr=8, lodcolumn=2, col="violetred", add=TRUE)

# lodcolumn can also be a column name
plot(out, map, lodcolumn="liver", ylim=ylim)
plot(out, map, lodcolumn="spleen", col="violetred", add=TRUE)
}
\seealso{
\code{\link[=plot_coef]{plot_coef()}}, \code{\link[=plot_coefCC]{plot_coefCC()}}, \code{\link[=plot_snpasso]{plot_snpasso()}}, \code{\link[=plot_scan1_heatmap]{plot_scan1_heatmap()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan1snps.R
\name{scan1snps}
\alias{scan1snps}
\title{Single-QTL genome scan at imputed SNPs}
\usage{
scan1snps(
  genoprobs,
  map,
  pheno,
  kinship = NULL,
  addcovar = NULL,
  Xcovar = NULL,
  intcovar = NULL,
  weights = NULL,
  reml = TRUE,
  model = c("normal", "binary"),
  query_func = NULL,
  chr = NULL,
  start = NULL,
  end = NULL,
  snpinfo = NULL,
  batch_length = 20,
  keep_all_snps = FALSE,
  cores = 1,
  ...
)
}
\arguments{
\item{genoprobs}{Genotype probabilities as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{map}{Physical map for the positions in the \code{genoprobs}
object: A list of numeric vectors; each vector gives marker
positions for a single chromosome.}

\item{pheno}{A numeric matrix of phenotypes, individuals x phenotypes.}

\item{kinship}{Optional kinship matrix, or a list of kinship matrices (one
per chromosome), in order to use the LOCO (leave one chromosome
out) method.}

\item{addcovar}{An optional numeric matrix of additive covariates.}

\item{Xcovar}{An optional numeric matrix with additional additive covariates used for
null hypothesis when scanning the X chromosome.}

\item{intcovar}{An optional numeric matrix of interactive covariates.}

\item{weights}{An optional numeric vector of positive weights for the
individuals. As with the other inputs, it must have \code{names}
for individual identifiers.}

\item{reml}{If \code{kinship} provided: if \code{reml=TRUE}, use
REML; otherwise maximum likelihood.}

\item{model}{Indicates whether to use a normal model (least
squares) or binary model (logistic regression) for the phenotype.
If \code{model="binary"}, the phenotypes must have values in \eqn{[0, 1]}.}

\item{query_func}{Function for querying SNP information; see
\code{\link[=create_variant_query_func]{create_variant_query_func()}}). Takes arguments
\code{chr}, \code{start}, \code{end}, (with \code{start} and \code{end} in the units in
\code{map}, generally Mbp), and returns a data frame containing
the columns \code{snp}, \code{chr}, \code{pos}, and \code{sdp}. (See \code{snpinfo} below.)}

\item{chr}{Chromosome or chromosomes to scan}

\item{start}{Position defining the start of an interval to scan.
Should be a single number, and if provided, \code{chr} should also
have length 1.}

\item{end}{Position defining the end of an interval to scan.
Should be a single number, and if provided, \code{chr} should also
have length 1.}

\item{snpinfo}{Optional data frame of SNPs to scan; if provided,
\code{query_func}, \code{chr}, \code{start}, and \code{end} are ignored. Should
contain the following columns:
\itemize{
\item \code{chr} - Character string or factor with chromosome
\item \code{pos} - Position (in same units as in the \code{"map"}).
\item \code{sdp} - Strain distribution pattern: an integer, between
1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
binary encoding indicates the founder genotypes
\item \code{snp} - Character string with SNP identifier (if
missing, the rownames are used).
}}

\item{batch_length}{Interval length (in units of \code{map}, generally
Mbp) to scan at one time.}

\item{keep_all_snps}{SNPs are grouped into equivalence classes based
on position and founder genotypes; if \code{keep_all_snps=FALSE},
the return value will contain information only on the indexed
SNPs (one per equivalence class).}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}

\item{...}{Additional control parameters passed to \code{\link[=scan1]{scan1()}}}
}
\value{
A list with two components: \code{lod} (matrix of LOD scores)
and \code{snpinfo} (a data frame of SNPs that were scanned,
including columns \code{index} which indicates groups of equivalent
SNPs)
}
\description{
Perform a single-QTL scan across the genome or a defined region at
SNPs genotyped in the founders, by Haley-Knott regression or a
liear mixed model, with possible allowance for covariates.
}
\details{
The analysis proceeds as follows:
\itemize{
\item Call \code{query_func()} to grab all SNPs over a region.
\item Use \code{\link[=index_snps]{index_snps()}} to group SNPs into equivalence classes.
\item Use \code{\link[=genoprob_to_snpprob]{genoprob_to_snpprob()}} to convert \code{genoprobs} to SNP probabilities.
\item Use \code{\link[=scan1]{scan1()}} to do a single-QTL scan at the SNPs.
}
}
\examples{
\dontrun{
# load example data and calculate genotype probabilities
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
probs <- calc_genoprob(DOex, error_prob=0.002)

snpdb_file <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
queryf <- create_variant_query_func(snpdb_file)

out <- scan1snps(probs, DOex$pmap, DOex$pheno, query_func=queryf, chr=2, start=97, end=98)
}

}
\seealso{
\code{\link[=scan1]{scan1()}}, \code{\link[=genoprob_to_snpprob]{genoprob_to_snpprob()}}, \code{\link[=index_snps]{index_snps()}}, \code{\link[=create_variant_query_func]{create_variant_query_func()}}, \code{\link[=plot_snpasso]{plot_snpasso()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/est_herit.R
\name{est_herit}
\alias{est_herit}
\title{Estimate heritability with a linear mixed model}
\usage{
est_herit(
  pheno,
  kinship,
  addcovar = NULL,
  weights = NULL,
  reml = TRUE,
  cores = 1,
  ...
)
}
\arguments{
\item{pheno}{A numeric matrix of phenotypes, individuals x phenotypes.}

\item{kinship}{A kinship matrix.}

\item{addcovar}{An optional numeric matrix of additive covariates.}

\item{weights}{An optional numeric vector of positive weights for the
individuals. As with the other inputs, it must have \code{names}
for individual identifiers.}

\item{reml}{If true, use REML; otherwise, use maximimum likelihood.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}

\item{...}{Additional control parameters (see details).}
}
\value{
A vector of estimated heritabilities, corresponding to the
columns in \code{pheno}. The result has attributes \code{"sample_size"},
\code{"log10lik"} and \code{"resid_sd"}.
}
\description{
Estimate the heritability of a set of traits via a linear mixed
model, with possible allowance for covariates.
}
\details{
We fit the model \eqn{y = X \beta + \epsilon}{y = Xb + e} where
\eqn{\epsilon}{e} is multivariate normal with mean 0 and covariance
matrix \eqn{\sigma^2 [h^2 (2 K) + (1-h^2) I]}{sigmasq*[hsq*2*K+(1-hsq)*I]} where
\eqn{K} is the kinship matrix and \eqn{I} is the identity matrix.

If \code{weights} are provided, the covariance matrix becomes
\eqn{\sigma^2 [h^2 (2 K) + (1-h^2) D]}{sigmasq*[hsq*2*K+(1-hsq)*D]} where
\eqn{D} is a diagonal matrix with the reciprocal of the weights.

For each of the inputs, the row names are used as
individual identifiers, to align individuals.

If \code{reml=TRUE}, restricted maximum likelihood (reml) is used
to estimate the heritability, separately for each phenotype.

Additional control parameters include \code{tol} for the tolerance
for convergence, \code{quiet} for controlling whether messages will
be display, \code{max_batch} for the maximum number of phenotypes
in a batch, and \code{check_boundary} for whether the 0 and 1
boundary values for the estimated heritability will be checked
explicitly.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c("19","X")] # subset to two chromosomes}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# kinship matrix
kinship <- calc_kinship(probs)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)

# perform genome scan
hsq <- est_herit(pheno, kinship, covar)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_geno.R
\name{compare_geno}
\alias{compare_geno}
\title{Compare individuals' genotype data}
\usage{
compare_geno(cross, omit_x = FALSE, proportion = TRUE, quiet = TRUE, cores = 1)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{omit_x}{If TRUE, only use autosomal genotypes}

\item{proportion}{If TRUE (the default), the upper triangle of the
result contains the proportions of matching genotypes. If
FALSE, the upper triangle contains counts of matching
genotypes.}

\item{quiet}{IF \code{FALSE}, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A square matrix; diagonal is number of observed genotypes
for each individual. The values in the lower triangle are the
numbers of markers where both of a pair were genotyped. The
values in the upper triangle are either proportions or counts
of matching genotypes for each pair (depending on whether
\code{proportion=TRUE} or \verb{=FALSE}). The object is given
class \code{"compare_geno"}.
}
\description{
Count the number of matching genotypes between all pairs of
individuals, to look for unusually closely related individuals.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
cg <- compare_geno(grav2)
summary(cg)
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/align_scan1_map.R, R/qtl2-internal.R
\name{align_scan1_map}
\alias{align_scan1_map}
\alias{qtl2-internal}
\title{Internal functions}
\usage{
align_scan1_map(scan1_output, map)
}
\description{
Functions that are exported and so available to users but not really intended to be used by a typical user.
}
\section{Functions}{
\itemize{
\item \code{align_scan1_map()}: Align \code{scan1()} output with a map

}}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_pheno.R
\name{read_pheno}
\alias{read_pheno}
\title{Read phenotype data}
\usage{
read_pheno(
  file,
  phenocovarfile = NULL,
  sep = ",",
  na.strings = c("-", "NA"),
  comment.char = "#",
  transpose = FALSE,
  quiet = TRUE
)
}
\arguments{
\item{file}{Character string with path to the phenotype data file
(or a zip file containing both the phenotype and phenotype
covariate files).}

\item{phenocovarfile}{Character string with path to the phenotype
covariate file. This can be a separate CSV or zip file; if a zip
file, it must contain exactly one CSV file. Alternatively, if the
\code{file} argument indicates a zip file that contains two files
(phenotypes and phenotype covariates), then this
\code{phenocovarfile} argument must indicate the base name for the
phenotype covariate file.}

\item{sep}{the field separator character}

\item{na.strings}{a character vector of strings which are to be
interpreted as \code{NA} values.}

\item{comment.char}{A character vector of length one containing a
single character to denote comments within the CSV files.}

\item{transpose}{If TRUE, the phenotype data will be transposed. The
phenotype covariate information is \strong{never} transposed.}

\item{quiet}{If \code{FALSE}, print progress messages.}
}
\value{
Either a matrix of phenotype data, or a list containing
\code{pheno} (phenotype matrix) and \code{phenocovar} (phenotype
covariate matrix).
}
\description{
Read phenotype data from a CSV file (and, optionally, phenotype
covariate data from a separate CSV file). The CSV files may be
contained in zip files, separately or together.
}
\examples{
\dontrun{
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/Gough/gough_pheno.csv")
phe <- read_pheno(file)

phecovfile <- paste0("https://raw.githubusercontent.com/rqtl/",
                     "qtl2data/main/Gough/gough_phenocovar.csv")
phe_list <- read_pheno(file, phecovfile)
}
}
\seealso{
\code{\link[=read_cross2]{read_cross2()}},
sample data files at \url{https://kbroman.org/qtl2/pages/sampledata.html}
and \url{https://github.com/rqtl/qtl2data}
}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbind_viterbi.R
\name{rbind.viterbi}
\alias{rbind.viterbi}
\title{Join Viterbi results for different individuals}
\usage{
\method{rbind}{viterbi}(...)
}
\arguments{
\item{...}{Imputed genotype objects as produced by
\code{\link[=viterbi]{viterbi()}}. Must have the same set of markers.}
}
\value{
An object of class \code{"viterbi"}, like the input; see \code{\link[=viterbi]{viterbi()}}.
}
\description{
Join multiple imputed genotype objects, as produced by \code{\link[=viterbi]{viterbi()}},
for the same set of markers but for different individuals.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
gA <- viterbi(grav2[1:5,], map, error_prob=0.002)
gB <- viterbi(grav2[6:12,], map, error_prob=0.002)
g <- rbind(gA, gB)

}
\seealso{
\code{\link[=cbind.viterbi]{cbind.viterbi()}}, \code{\link[=subset.viterbi]{subset.viterbi()}}, \code{\link[=viterbi]{viterbi()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayes_int.R
\name{bayes_int}
\alias{bayes_int}
\title{Calculate Bayes credible intervals}
\usage{
bayes_int(
  scan1_output,
  map,
  chr = NULL,
  lodcolumn = 1,
  threshold = 0,
  peakdrop = Inf,
  prob = 0.95,
  expand2markers = TRUE
)
}
\arguments{
\item{scan1_output}{An object of class \code{"scan1"} as returned by
\code{\link[=scan1]{scan1()}}.}

\item{map}{A list of vectors of marker positions, as produced by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}.}

\item{chr}{Chromosome ID to consider (must be a single value).}

\item{lodcolumn}{LOD score column to consider (must be a single value).}

\item{threshold}{Minimum LOD score for a peak.}

\item{peakdrop}{Amount that the LOD score must drop between peaks,
if multiple peaks are to be defined on a chromosome.}

\item{prob}{Nominal coverage for the interval.}

\item{expand2markers}{If TRUE, QTL intervals are expanded so
that their endpoints are at genetic markers.}
}
\value{
A matrix with three columns:
\itemize{
\item \code{ci_lo} - lower bound of interval
\item \code{pos} - peak position
\item \code{ci_hi} - upper bound of interval
}

Each row corresponds to a different peak.
}
\description{
Calculate Bayes credible intervals for a single LOD curve on a single
chromosome, with the ability to identify intervals for multiple LOD
peaks.
}
\details{
We identify a set of peaks defined as local maxima that
exceed the specified \code{threshold}, with the requirement that
the LOD score must have dropped by at least \code{peakdrop} below
the lowest of any two adjacent peaks.

At a given peak, if there are ties, with multiple positions jointly
achieving the maximum LOD score, we take the average of these
positions as the location of the peak.

The default is to use \code{threshold=0}, \code{peakdrop=Inf}, and
\code{prob=0.95}. We then return results a single peak, no matter the
maximum LOD score, and give a 95\% Bayes credible interval.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,7]}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# 95\% Bayes credible interval for QTL on chr 7, first phenotype
bayes_int(out, map, chr=7, lodcolum=1)
}
\seealso{
\code{\link[=lod_int]{lod_int()}}, \code{\link[=find_peaks]{find_peaks()}}, \code{\link[=scan1]{scan1()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cbind_sim_geno.R
\name{cbind.sim_geno}
\alias{cbind.sim_geno}
\title{Join genotype imputations for different chromosomes}
\usage{
\method{cbind}{sim_geno}(...)
}
\arguments{
\item{...}{Genotype imputation objects as produced by
\code{\link[=sim_geno]{sim_geno()}}. Must have the same set of individuals.}
}
\value{
An object of class \code{"sim_geno"}, like the input; see \code{\link[=sim_geno]{sim_geno()}}.
}
\description{
Join multiple genotype imputation objects, as produced by
\code{\link[=sim_geno]{sim_geno()}}, for the same individuals but different
chromosomes.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
drawsA <- sim_geno(grav2[1:5,1:2], map, error_prob=0.002, n_draws=4)
drawsB <- sim_geno(grav2[1:5,3:4], map, error_prob=0.002, n_draws=4)
draws <- cbind(drawsA, drawsB)

}
\seealso{
\code{\link[=rbind.sim_geno]{rbind.sim_geno()}}, \code{\link[=sim_geno]{sim_geno()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smooth_gmap.R
\name{smooth_gmap}
\alias{smooth_gmap}
\title{Smooth genetic map}
\usage{
smooth_gmap(gmap, pmap, alpha = 0.02)
}
\arguments{
\item{gmap}{Genetic map, as a list of numeric vectors; each vector gives marker
positions for a single chromosome.}

\item{pmap}{Physical map, as a list of numeric vectors; each vector gives marker
positions for a single chromosome, with the same chromosomes and markers as \code{gmap}.}

\item{alpha}{Proportion of mixture to take from constant recombination.}
}
\value{
A genetic map like the input \code{gmap}, but smoothed by mixing
it with a proportion \code{alpha} of constant recombination on each
chromosome.
}
\description{
Smooth a genetic map by mixing it with a bit of constant
recombination (using a separate recombination rate for each
chromosome), to eliminate intervals that have exactly 0
recombination.
}
\details{
An interval of genetic length \eqn{d_g}{dg} and physical
length \eqn{d_p}{dp} is changed to have length
\eqn{(1-\alpha)d_g + \alpha d_p r}{(1-alpha)dg + (alpha*dp*r)}
where \eqn{r = L_g / L_p}{r = Lg/Lp} is the chromosome-specific
recombination rate.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
gmap_adj <- smooth_gmap(iron$gmap, iron$pmap)

}
\seealso{
\code{\link[=unsmooth_gmap]{unsmooth_gmap()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_sdp.R
\name{plot_sdp}
\alias{plot_sdp}
\title{plot strain distribution patterns for SNPs}
\usage{
plot_sdp(pos, sdp, strain_labels = names(qtl2::CCcolors), ...)
}
\arguments{
\item{pos}{vector of SNP positions}

\item{sdp}{vector of strain distribution patterns (as integers)}

\item{strain_labels}{names of the strains}

\item{...}{additional graphic arguments}
}
\value{
None.
}
\description{
plot the strain distribution patterns of SNPs
using tracks of tick-marks for each founder strain
}
\details{
Additional arguments, such as \code{xlab}, \code{ylab}, \code{xlim}, and \code{main},
are passed via \code{...}; also \code{bgcolor} to control the color of the
background, and \code{col} and \code{lwd} to control the color and thickness
of the tick marks.
}
\examples{
n_tick <- 50
plot_sdp(runif(n_tick, 0, 100), sample(0:255, n_tick, replace=TRUE))
}
\seealso{
\code{\link[=calc_sdp]{calc_sdp()}}, \code{\link[=invert_sdp]{invert_sdp()}}
}
\keyword{hgraphics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_raw_summaries.R
\name{calc_raw_geno_freq}
\alias{calc_raw_geno_freq}
\title{Calculate genotype frequencies from raw SNP genotypes}
\usage{
calc_raw_geno_freq(cross, by = c("individual", "marker"), cores = 1)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{by}{Indicates whether to summarize by individual or by marker.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A matrix of genotypes frequencies with 3 columns (AA, AB,
and BB) and with rows being either individuals or markers.
}
\description{
Calculate genotype frequencies from raw SNP genotypes, by individual or by marker
}
\examples{
\dontrun{
# load example data and calculate genotype probabilities
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
gfreq <- calc_raw_geno_freq(DOex)
}
}
\seealso{
\code{\link[=calc_raw_maf]{calc_raw_maf()}}, \code{\link[=calc_raw_het]{calc_raw_het()}}, \code{\link[=recode_snps]{recode_snps()}}, \code{\link[=calc_geno_freq]{calc_geno_freq()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_cross2.R
\name{read_cross2}
\alias{read_cross2}
\title{Read QTL data from files}
\usage{
read_cross2(file, quiet = TRUE)
}
\arguments{
\item{file}{Character string with path to the
\href{https://yaml.org}{YAML} or \href{https://json.org}{JSON} file containing all of the control
information. This could instead be a zip file containing all of the
data files, in which case the contents are unzipped to a temporary
directory and then read.}

\item{quiet}{If \code{FALSE}, print progress messages.}
}
\value{
Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.
}
\description{
Read QTL data from a set of files
}
\details{
A control file in \href{https://yaml.org}{YAML} or
\href{https://json.org}{JSON} format contains information
about basic parameters as well as the names of the series of
data files to be read. See the
\href{https://kbroman.org/qtl2/pages/sampledata.html}{sample data files} and the
\href{https://kbroman.org/qtl2/assets/vignettes/input_files.html}{vignette describing the input file format}.
}
\examples{
\dontrun{
yaml_file <- "https://kbroman.org/qtl2/assets/sampledata/grav2/grav2.yaml"
grav2 <- read_cross2(yaml_file)
}
zip_file <- system.file("extdata", "grav2.zip", package="qtl2")
grav2 <- read_cross2(zip_file)
}
\seealso{
\code{\link[=read_pheno]{read_pheno()}}, \code{\link[=write_control_file]{write_control_file()}},
sample data files at \url{https://kbroman.org/qtl2/pages/sampledata.html}
and \url{https://github.com/rqtl/qtl2data}
}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_markerpos.R
\name{find_markerpos}
\alias{find_markerpos}
\title{Find positions of markers}
\usage{
find_markerpos(cross, markers, na.rm = TRUE)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.
Can also be a map (as a list of vectors of marker positions).}

\item{markers}{A vector of marker names.}

\item{na.rm}{If TRUE, don't include not-found markers in the
results (but issue a warning if some markers weren't found). If
FALSE, include those markers with \code{NA} for chr and position.}
}
\value{
A data frame with chromosome and genetic and physical
positions (in columns \code{"gmap"} and \code{"pmap"}), with markers as
row names. If the input \code{cross} is not a cross2 object but
rather a map, the output contains \code{chr} and \code{pos}.
}
\description{
Find positions of markers within a cross object
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

# find markers
find_markerpos(iron, c("D8Mit294", "D11Mit101"))
}
\seealso{
\code{\link[=find_marker]{find_marker()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_geno.R
\name{summary_compare_geno}
\alias{summary_compare_geno}
\alias{summary.compare_geno}
\alias{print.summary.compare_geno}
\title{Basic summary of compare_geno object}
\usage{
summary_compare_geno(object, threshold = 0.9, ...)

\method{summary}{compare_geno}(object, threshold = 0.9, ...)

\method{print}{summary.compare_geno}(x, digits = 2, ...)
}
\arguments{
\item{object}{A square matrix with genotype comparisons for pairs
of individuals, as output by \code{\link[=compare_geno]{compare_geno()}}.}

\item{threshold}{Minimum proportion matches for a pair of individuals to be shown.}

\item{...}{Ignored}

\item{x}{Results of \code{\link[=summary.compare_geno]{summary.compare_geno()}}}

\item{digits}{Number of digits to print}
}
\value{
Data frame with names of individuals in pair, proportion
matches, number of mismatches, number of matches, and total
markers genotyped. Last two columns are the numeric indexes of
the individuals in the pair.
}
\description{
From results of \code{\link[=compare_geno]{compare_geno()}}, show pairs of individuals with similar genotypes.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
cg <- compare_geno(grav2)
summary(cg)
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fread_csv_numer.R
\name{fread_csv_numer}
\alias{fread_csv_numer}
\title{Read a csv file that has numeric columns}
\usage{
fread_csv_numer(
  filename,
  sep = ",",
  na.strings = c("NA", "-"),
  comment.char = "#",
  transpose = FALSE,
  rownames_included = TRUE
)
}
\arguments{
\item{filename}{Name of input file}

\item{sep}{Field separator}

\item{na.strings}{Missing value codes}

\item{comment.char}{Comment character; rest of line after this character is ignored}

\item{transpose}{If TRUE, transpose the result}

\item{rownames_included}{If TRUE, the first column is taken to be row names.}
}
\value{
Data frame
}
\description{
Read a csv file via \code{\link[data.table:fread]{data.table::fread()}} using a
particular set of options, including the ability to transpose the
result. This version assumes that the contents other than the first
column and the header row are strictly numeric.
}
\details{
Initial two lines can contain comments with number of rows
and columns. Number of columns includes an ID column; number of
rows does not include the header row.

The first column is taken to be a set of row names
}
\examples{
\dontrun{mydata <- fread_csv_numer("myfile.csv", transpose=TRUE)}
}
\seealso{
\code{\link[=fread_csv]{fread_csv()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_cistrans.R
\name{plot_cistrans}
\alias{plot_cistrans}
\title{cis-trans plot for eQTL results}
\usage{
plot_cistrans(
  peaks,
  map,
  pheno_pos,
  gap = 0,
  cis_window = 1,
  pch = 21,
  col = "slateblue",
  bg = "slateblue",
  cex = 0.8,
  pch_cis = 21,
  col_cis = "violetred",
  bg_cis = "violetred",
  cex_cis = 0.8,
  pty = "s",
  bgcolor = "gray90",
  altbgcolor = "gray80",
  ...
)
}
\arguments{
\item{peaks}{Data frame of QTL results, as output by \code{\link[=find_peaks]{find_peaks()}}
Should contain columns \code{lodcolumn}, \code{chr}, and \code{pos}.}

\item{map}{Marker map, as a list of chromosomes, each being a
vector of positions; used to get chromosome names and start/end}

\item{pheno_pos}{Data frame with chromosome, position, and name of
the phenotypes in \code{peaks}. Should contain columns \code{pheno},
\code{chr}, and \code{pos}.}

\item{gap}{Gap between chromosomes in the plot}

\item{cis_window}{Window size that defines cis-QTL (on same
chromosome, and position of gene within this distance of QTL
position)}

\item{pch}{Point type for "trans" points}

\item{col}{Point color for "trans" points}

\item{bg}{Background color for "trans" points}

\item{cex}{Character expansion size for "trans" points}

\item{pch_cis}{Point type for "cis" points}

\item{col_cis}{Point color for "cis" points}

\item{bg_cis}{Background color for "cis" points}

\item{cex_cis}{Character expansion size for "cis" points}

\item{pty}{Plot type; the default "s" forces a square plot; use "m"
to use the maximal plotting region.}

\item{bgcolor}{Background color for chromosome rectangles}

\item{altbgcolor}{Alternate background color for chromosome rectangles}

\item{...}{Additional graphics arguments passed to \code{\link[graphics:points]{graphics::points()}}}
}
\value{
Invisibly returns a data frame with \code{peaks} merged with
\code{pheno_pos}, with an additional column \code{cis} that indicates
which QTL appear to be cis (vs trans).
}
\description{
Scatterplot of gene location vs QTL location to display eQTL/pQTL results
}
\examples{
\dontrun{
# download example pQTL results (from Keele et al. 2026, https://doi.org/10.1016/j.xgen.2025.101069)
# contains qtl, map, and pheno_pos
url <- "https://kbroman.org/qtl2/assets/sampledata/pqtl_data.RData"
tempfile <- file.path(tempdir(), basename(url))
download.file(url, tempfile)
load(tempfile)
unlink(tempfile)

plot_cistrans(qtl, map, pheno_pos, cis_window=5)
}
}
\seealso{
\code{\link[=plot_scan1]{plot_scan1()}}, \code{\link[=find_peaks]{find_peaks()}}, \code{\link[=plot_lodpeaks]{plot_lodpeaks()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_het.R
\name{calc_het}
\alias{calc_het}
\title{Calculate heterozygosities}
\usage{
calc_het(probs, by = c("individual", "marker"), omit_x = TRUE, cores = 1)
}
\arguments{
\item{probs}{List of arrays of genotype probabilities, as
calculated by \code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{by}{Whether to summarize by individual or marker}

\item{omit_x}{If TRUE, omit the X chromosome.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
The result is a vector of estimated heterozygosities
}
\description{
Calculate heterozygosites, by individual or by marker
}
\details{
\code{calc_het()} looks at the genotype names (the 2nd
dimension of the dimnames of the input \code{probs}), which must be
two-letter names, and assumes that when the two letters are
different it's a heterozygous genotype while if they're the
same it's a homozygous genotype
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
p <- calc_genoprob(iron, err=0.002)

# heterozygosities by individual
het_ind <- calc_het(p)

# heterozygosities by marker
het_mar <- calc_het(p, "marker")

}
\seealso{
\code{\link[=calc_raw_het]{calc_raw_het()}}, \code{\link[=calc_geno_freq]{calc_geno_freq()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_x_covar.R
\name{get_x_covar}
\alias{get_x_covar}
\title{Get X chromosome covariates}
\usage{
get_x_covar(cross)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}
}
\value{
A matrix of size individuals x no. covariates.
}
\description{
Get the matrix of covariates to be used for the null hypothesis when
performing QTL analysis with the X chromosome.
}
\details{
For most crosses, the result is either \code{NULL}
(indicating no additional covariates are needed) or a
matrix with a single column containing sex indicators (1 for males
and 0 for females).

For an intercross, we also consider cross direction. There are
four cases: (1) All male or all female but just one direction: no
covariate; (2) All female but both directions: covariate indicating
cross direction; (3) Both sexes, one direction: covariate
indicating sex; (4) Both sexes, both directions: a covariate
indicating sex and a covariate that is 1 for females from the
reverse direction and 0 otherwise.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
xcovar <- get_x_covar(iron)
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qtl2-package.R
\docType{package}
\name{qtl2-package}
\alias{qtl2}
\alias{qtl2-package}
\title{qtl2: Quantitative Trait Locus Mapping in Experimental Crosses}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Provides a set of tools to perform quantitative trait locus (QTL) analysis in experimental crosses. It is a reimplementation of the 'R/qtl' package to better handle high-dimensional data and complex cross designs. Broman et al. (2019) \doi{10.1534/genetics.118.301595}.
}
\section{Vignettes}{

\itemize{
\item \href{https://kbroman.org/qtl2/assets/vignettes/user_guide.html}{user guide}
\item \href{https://kbroman.org/qtl2/pages/rqtl2_functions.html}{categorized list of functions in R/qtl2}
\item \href{https://kbroman.org/qtl2/assets/vignettes/input_files.html}{input file formats}
\item \href{https://kbroman.org/qtl2/assets/vignettes/qtl2fst.html}{using qtl2fst for on-disk genotype probabilities}
\item \href{https://kbroman.org/qtl2/pages/prep_do_data.html}{preparing DO mouse data for R/qtl2}
\item \href{https://kbroman.org/qtl2/assets/vignettes/do_diagnostics.html}{genotype diagnostics for diversity outbred mice}
\item \href{https://kbroman.org/qtl2/assets/vignettes/do_mixups.html}{identifying sample mix-ups in diversity outbred mice}
\item \href{https://kbroman.org/qtl2/assets/vignettes/rqtl_diff.html}{differences between R/qtl and R/qtl2}
\item \href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{developer guide}
\item \href{https://smcclatchy.github.io/mapping/}{Tutorial on R/qtl2} by
\href{https://github.com/smcclatchy}{Susan McClatchy} and \href{https://github.com/dmgatti}{Dan Gatti}
\item \href{https://kbroman.org/qtl2/assets/vignettes/scan1gen.html}{scan1gen examples}
}
}

\section{Related packages}{

\itemize{
\item \href{https://github.com/rqtl/qtl2convert}{qtl2convert}, for converting
data among the R/qtl2, DOQTL, and R/qtl formats
\item \href{https://github.com/rqtl/qtl2fst}{qtl2fst}, for storing genotype
probabilities on disk
\item \href{https://github.com/byandell-sysgen/qtl2ggplot}{qtl2ggplot},
for \href{https://ggplot2.tidyverse.org/}{ggplot2}-based data visualizations
}
}

\section{Google groups}{

\itemize{
\item For discussion/questions about R/qtl2, join the \href{https://groups.google.com/forum/#!forum/rqtl2-disc}{rqtl2-disc google group}
\item Or join \href{https://groups.google.com/forum/#!forum/rqtl-announce}{rqtl-announce} for announcements about R/qtl and R/qtl2
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://kbroman.org/qtl2/}
  \item \url{https://github.com/rqtl/qtl2}
  \item Report bugs at \url{https://github.com/rqtl/qtl2/issues}
}

}
\author{
\strong{Maintainer}: Karl W Broman \email{broman@wisc.edu} (\href{https://orcid.org/0000-0002-4914-6671}{ORCID})

Authors:
\itemize{
  \item Karl W Broman \email{broman@wisc.edu} (\href{https://orcid.org/0000-0002-4914-6671}{ORCID})
}

Other contributors:
\itemize{
  \item R Core Team [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_cross2.R
\name{check_cross2}
\alias{check_cross2}
\title{Check a cross2 object}
\usage{
check_cross2(cross2)
}
\arguments{
\item{cross2}{An object of class \code{"cross2"}, as output by \code{\link[=read_cross2]{read_cross2()}}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}
}
\value{
If everything is correct, returns \code{TRUE}; otherwise \code{FALSE},
with attributes that give the problems.
}
\description{
Check the integrity of the data within a cross2 object.
}
\details{
Checks whether a cross2 object meets the
specifications. Problems are issued as warnings.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
check_cross2(grav2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_map_gaps.R
\name{find_map_gaps}
\alias{find_map_gaps}
\title{Find gaps in a genetic map}
\usage{
find_map_gaps(map, min_gap = 50)
}
\arguments{
\item{map}{Genetic map as a list of vectors (each vector is a
chromosome and contains the marker positions).}

\item{min_gap}{Minimum gap length to return.}
}
\value{
Data frame with 6 columns: chromosome, marker to left of gap,
numeric index of marker to left, marker to right of gap, numeric
index of marker to right, and the length of the gap.
}
\description{
Find gaps between markers in a genetic map
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
find_map_gaps(iron$gmap, 40)

}
\seealso{
\code{\link[=reduce_map_gaps]{reduce_map_gaps()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan1max.R
\name{scan1max}
\alias{scan1max}
\title{Maximum LOD score from genome scan with a single-QTL model}
\usage{
scan1max(
  genoprobs,
  pheno,
  kinship = NULL,
  addcovar = NULL,
  Xcovar = NULL,
  intcovar = NULL,
  weights = NULL,
  reml = TRUE,
  model = c("normal", "binary"),
  hsq = NULL,
  by_chr = FALSE,
  cores = 1,
  ...
)
}
\arguments{
\item{genoprobs}{Genotype probabilities as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{pheno}{A numeric matrix of phenotypes, individuals x phenotypes.}

\item{kinship}{Optional kinship matrix, or a list of kinship matrices (one
per chromosome), in order to use the LOCO (leave one chromosome
out) method.}

\item{addcovar}{An optional numeric matrix of additive covariates.}

\item{Xcovar}{An optional numeric matrix with additional additive covariates used for
null hypothesis when scanning the X chromosome.}

\item{intcovar}{An numeric optional matrix of interactive covariates.}

\item{weights}{An optional numeric vector of positive weights for the
individuals. As with the other inputs, it must have \code{names}
for individual identifiers.}

\item{reml}{If \code{kinship} provided: if \code{reml=TRUE}, use
REML; otherwise maximum likelihood.}

\item{model}{Indicates whether to use a normal model (least
squares) or binary model (logistic regression) for the phenotype.
If \code{model="binary"}, the phenotypes must have values in \eqn{[0, 1]}.}

\item{hsq}{Considered only if \code{kinship} is provided, in which case
this is taken as the assumed value for the residual
heritability. It should be a vector with length corresponding
to the number of columns in \code{pheno}, or (if \code{kinship}
corresponds to a list of LOCO kinship matrices) a matrix with dimension
\verb{length(kinship) x ncol(pheno)}.}

\item{by_chr}{If TRUE, save the individual chromosome maxima.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}

\item{...}{Additional control parameters; see Details.}
}
\value{
Either a vector of genome-wide maximum LOD scores, or if
\code{by_chr} is TRUE, a matrix with the chromosome-specific maxima,
with the rows being the chromosomes and the columns being the
phenotypes.
}
\description{
Maximum LOD score from genome scan with a single-QTL model by
Haley-Knott regression or a linear mixed model, with possible
allowance for covariates.
}
\details{
Equivalent to running \code{scan1()} and then saving the column
maxima, with some savings in memory usage.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c("19", "X")] # subset to chr 19 and X}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1max(probs, pheno, addcovar=covar, Xcovar=Xcovar)

}
\seealso{
\code{\link[=scan1]{scan1()}}, \code{\link[=scan1perm]{scan1perm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset_phasedgeno.R
\name{subset.phasedgeno}
\alias{subset.phasedgeno}
\alias{[.phasedgeno}
\title{Subsetting phased genotype objects}
\usage{
\method{subset}{phasedgeno}(x, ind = NULL, chr = NULL, ...)

\method{[}{phasedgeno}(x, ind = NULL, chr = NULL)
}
\arguments{
\item{x}{Imputed genotypes as output from \code{\link[=guess_phase]{guess_phase()}}, a list of 3d arrays.}

\item{ind}{A vector of individuals: numeric indices, logical
values, or character string IDs}

\item{chr}{A vector of chromosomes: logical values, or character
string IDs. Numbers are interpreted as character string IDs.}

\item{...}{Ignored.}
}
\value{
An object of class \code{"phasedgeno"}, like the input, with the
selected individuals and/or chromosomes; see \code{\link[=guess_phase]{guess_phase()}}.
}
\description{
Pull out a specified set of individuals and/or chromosomes from
the results of \code{\link[=guess_phase]{guess_phase()}}
}
\examples{
\dontrun{
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
pr <- calc_genoprob(DOex, error_prob=0.002)
m <- maxmarg(pr)
ph <- guess_phase(DOex, m)
ph <- ph[1:10,] # first ten individuals
ph <- ph[,2]  # chromosome 2
}
}
\seealso{
\code{\link[=guess_phase]{guess_phase()}}, \code{\link[=cbind.phasedgeno]{cbind.phasedgeno()}}, \code{\link[=rbind.phasedgeno]{rbind.phasedgeno()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_dup_markers.R
\name{find_dup_markers}
\alias{find_dup_markers}
\title{Find markers with identical genotype data}
\usage{
find_dup_markers(cross, chr, exact_only = TRUE, adjacent_only = FALSE)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{chr}{Optional vector specifying which chromosomes to consider.
This may be a logical, numeric, or character string vector.}

\item{exact_only}{If TRUE, look only for markers that have matching
genotypes and the same pattern of missing data; if FALSE, also look for
cases where the observed genotypes at one marker match those at
another, and where the first marker has missing genotype whenever the
genotype for the second marker is missing.}

\item{adjacent_only}{If TRUE, look only for sets of markers that are
adjacent to each other.}
}
\value{
A list of marker names; each component is a set of markers whose
genotypes match one other marker, and the name of the component is the
name of the marker that they match.
}
\description{
Identify sets of markers with identical genotype data.
}
\details{
If \code{exact.only=TRUE}, we look only for groups of markers whose
pattern of missing data and observed genotypes match exactly.  One
marker (chosen at random) is selected as the name of the group (in the
output of the function).

If \code{exact.only=FALSE}, we look also for markers whose observed genotypes
are contained in the observed genotypes of another marker.  We use a
pair of nested loops, working from the markers with the most observed
genotypes to the markers with the fewest observed genotypes.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
dup <- find_dup_markers(grav2)
grav2_nodup <- drop_markers(grav2, unlist(dup))
}
\seealso{
\code{\link[=drop_markers]{drop_markers()}}, \code{\link[=drop_nullmarkers]{drop_nullmarkers()}}, \code{\link[=reduce_markers]{reduce_markers()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_index_snp.R
\name{find_index_snp}
\alias{find_index_snp}
\title{Find name of indexed snp}
\usage{
find_index_snp(snpinfo, snp)
}
\arguments{
\item{snpinfo}{Data frame with SNP information with the following columns:
\itemize{
\item \code{chr} - Character string or factor with chromosome
\item \code{index} - Numeric index of equivalent, indexed SNP, as produced by \code{\link[=index_snps]{index_snps()}}.
\item \code{snp} - Character string with SNP identifier (if
missing, the rownames are used).
}}

\item{snp}{Name of snp to look for (can be a vector).}
}
\value{
A vector of SNP IDs (the corresponding indexed SNPs), with NA if a SNP is not found.
}
\description{
For a particular SNP, find the name of the corresponding indexed SNP.
}
\examples{
\dontrun{
# load example data and calculate genotype probabilities
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DO_Recla/recla.zip")
recla <- read_cross2(file)

# founder genotypes for a set of SNPs
snpgeno <- rbind(m1=c(3,1,1,3,1,1,1,1),
                 m2=c(3,1,1,3,1,1,1,1),
                 m3=c(1,1,1,1,3,3,3,3),
                 m4=c(1,3,1,3,1,3,1,3))
sdp <- calc_sdp(snpgeno)
snpinfo <- data.frame(chr=c("19", "19", "X", "X"),
                      pos=c(40.36, 40.53, 110.91, 111.21),
                      sdp=sdp,
                      snp=c("m1", "m2", "m3", "m4"), stringsAsFactors=FALSE)

# update snp info by adding the SNP index column
snpinfo <- index_snps(recla$pmap, snpinfo)

# find indexed snp for a particular snp
find_index_snp(snpinfo, "m3")
}

}
\seealso{
\code{\link[=find_marker]{find_marker()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/decomp_kinship.R
\name{decomp_kinship}
\alias{decomp_kinship}
\title{Calculate eigen decomposition of kinship matrix}
\usage{
decomp_kinship(kinship, cores = 1)
}
\arguments{
\item{kinship}{A square matrix, or a list of square matrices.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
The eigen values and the \strong{transposed} eigen vectors,
as a list containing a vector \code{values} and a matrix
\code{vectors}.
}
\description{
Calculate the eigen decomposition of a kinship matrix, or of a list of such matrices.
}
\details{
The result contains an attribute \code{"eigen_decomp"}.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[1:30,18:19] # subset to 30 individuals and two chromosomes}
map <- insert_pseudomarkers(iron$gmap, step=1)
probs <- calc_genoprob(iron, map, error_prob=0.002)
K <- calc_kinship(probs)

Ke <- decomp_kinship(K)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rcbind_scan1perm.R
\name{rbind.scan1perm}
\alias{rbind.scan1perm}
\alias{c.scan1perm}
\title{Combine data from scan1perm objects}
\usage{
\method{rbind}{scan1perm}(...)

\method{c}{scan1perm}(...)
}
\arguments{
\item{...}{A set of permutation results from
\code{\link[=scan1perm]{scan1perm()}} (objects of class \code{"scan1perm"}).
They must have the same set of columns. If any include
autosome/X chromosome-specific permutations, they must all be
such.}
}
\value{
The combined row-binded input, as an object of class \code{"scan1perm"}; see \code{\link[=scan1perm]{scan1perm()}}.
}
\description{
Row-bind multiple scan1perm objects with the same set of columns
}
\details{
The aim of this function is to concatenate the results
from multiple runs of a permutation test with
\code{\link[=scan1perm]{scan1perm()}}, to assist in the case that such
permutations are done on multiple processors in parallel.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c("19","X")] # subset to chr 19 and X}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# permutations with genome scan (just 3 replicates, for illustration)
operm1 <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar, n_perm=3)
operm2 <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar, n_perm=3)

operm <- rbind(operm1, operm2)

}
\seealso{
\code{\link[=cbind.scan1perm]{cbind.scan1perm()}}, \code{\link[=scan1perm]{scan1perm()}}, \code{\link[=scan1]{scan1()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean_scan1.R
\name{clean_scan1}
\alias{clean_scan1}
\alias{clean.scan1}
\title{Clean scan1 output}
\usage{
clean_scan1(object, ...)

\method{clean}{scan1}(object, ...)
}
\arguments{
\item{object}{Output of \code{\link[=scan1]{scan1()}}.}

\item{...}{Ignored at present}
}
\value{
The input object with negative values replaced with NAs and then rows with all NAs removed.
}
\description{
Clean scan1 output by replacing negative values with NA and remove
rows where all values are NA.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,"2"]}
pr <- calc_genoprob(iron)
out <- scan1(pr, iron$pheno)

out <- clean(out)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reduce_map_gaps.R
\name{reduce_map_gaps}
\alias{reduce_map_gaps}
\title{Reduce the lengths of gaps in a map}
\usage{
reduce_map_gaps(map, min_gap = 50)
}
\arguments{
\item{map}{Genetic map as a list of vectors (each vector is a
chromosome and contains the marker positions).}

\item{min_gap}{Minimum gap length to return.}
}
\value{
Input map with any gaps greater than \code{min_gap} reduced to \code{min_gap}.
}
\description{
Reduce the lengths of gaps in a map
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
rev_map <- reduce_map_gaps(iron$gmap, 30)

}
\seealso{
\code{\link[=find_map_gaps]{find_map_gaps()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/max_scan1.R
\name{maxlod}
\alias{maxlod}
\title{Overall maximum LOD score}
\usage{
maxlod(scan1_output, map = NULL, chr = NULL, lodcolumn = NULL)
}
\arguments{
\item{scan1_output}{An object of class \code{"scan1"} as returned by
\code{\link[=scan1]{scan1()}}.}

\item{map}{A list of vectors of marker positions, as produced by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}.}

\item{chr}{Optional vector of chromosomes to consider.}

\item{lodcolumn}{An integer or character string indicating the LOD
score column, either as a numeric index or column name.
If \code{NULL}, return maximum for all columns.}
}
\value{
A single number: the maximum LOD score across all columns and positions for
the selected chromosomes.
}
\description{
Find overall maximum LOD score in genome scan results, across all positions and columns.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# overall maximum
maxlod(out)

# maximum on chromosome 2
maxlod(out, map, "2")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/CCcolors-data.R
\name{CCcolors}
\alias{CCcolors}
\alias{CCorigcolors}
\alias{CCaltcolors}
\title{Collaborative Cross colors}
\source{
\url{https://web.archive.org/web/20250215070655/https://csbio.unc.edu/CCstatus/index.py?run=AvailableLines.information}
}
\description{
A vector of 8 colors for use with the mouse
Collaborative Cross and Diversity Outbreds.
}
\details{
\code{CCorigcolors} are the original eight colors for the Collaborative Cross founder strains.
\code{CCaltcolors} are a slightly modified version, but still not color-blind friendly.
\code{CCcolors} are derived from the the Okabe-Ito color blind friendly palette in
Wong (2011) Nature Methods \doi{10.1038/nmeth.1618}.
}
\examples{
data(CCcolors)
data(CCaltcolors)
data(CCorigcolors)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean_genoprob.R
\name{clean_genoprob}
\alias{clean_genoprob}
\alias{clean.calc_genoprob}
\title{Clean genotype probabilities}
\usage{
clean_genoprob(
  object,
  value_threshold = 0.000001,
  column_threshold = 0.01,
  ind = NULL,
  cores = 1,
  ...
)

\method{clean}{calc_genoprob}(
  object,
  value_threshold = 0.000001,
  column_threshold = 0.01,
  ind = NULL,
  cores = 1,
  ...
)
}
\arguments{
\item{object}{Genotype probabilities as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{value_threshold}{Probabilities below this value will be set
to 0.}

\item{column_threshold}{For genotype columns where the maximum
value is below this threshold, all values will be set to 0.
This must be less than \eqn{1/k} where \eqn{k} is the number of
genotypes.}

\item{ind}{Optional vector of individuals (logical, numeric, or
character). If provided, only the genotype probabilities for
these individuals will be cleaned, though the full set will be
returned.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}

\item{...}{Ignored at this point.}
}
\value{
A cleaned version of the input genotype probabilities
object, \code{object}.
}
\description{
Clean up genotype probabilities by setting small values to 0 and
for a genotype column where the maximum value is rather small, set
all values in that column to 0.
}
\details{
In cases where a particular genotype is largely absent,
\code{scan1coef()} and \code{fit1()} can give unstable estimates of the
genotype effects. Cleaning up the genotype probabilities by setting
small values to 0 helps to ensure that such effects get set to
\code{NA}.

At each position and for each genotype column, we find the maximum
probability across individuals. If that maximum is <
\code{column_threshold}, all values in that genotype column at that
position are set to 0.

In addition, any genotype probabilities that are < \code{value_threshold}
(generally < \code{column_threshold}) are set to 0.

The probabilities are then re-scaled so that the probabilities for
each individual at each position sum to 1.

If \code{ind} is provided, the function is applied only to the
designated subset of individuals. This may be useful when only a
subset of individuals have been phenotyped, as you may want to zero
out genotype columns where that subset of individuals has only
negligible probability values.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c("19", "X")] # subset to chr 19 and X}

# calculate genotype probabilities
probs <- calc_genoprob(iron, error_prob=0.002)

# clean the genotype probabilities
# (doesn't really do anything in this case, because there are no small but non-zero values)
probs_clean <- clean(probs)

# clean only the females' genotype probabilities
probs_cleanf <- clean(probs, ind=names(iron$is_female)[iron$is_female])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_colorscale.R
\name{plot_colorscale}
\alias{plot_colorscale}
\title{Heat map color scale}
\usage{
plot_colorscale(
  x = NULL,
  zlim = NULL,
  color_scheme = c("viridis", "gray", "revgray", "heat", "terrain", "topo", "rainbow"),
  col = NULL,
  n_colors = 256,
  gridlines = NULL,
  swap_axes = FALSE,
  ...
)
}
\arguments{
\item{x}{Values that define the color limits (matrix, data frame, or vector); ignored if \code{zlim} is provided.}

\item{zlim}{Z-axis limits (a pair of numbers)}

\item{color_scheme}{Color scheme for the heatmap (ignored if \code{col} is provided).}

\item{col}{Optional vector of colors for the heatmap.}

\item{n_colors}{Number of z-axis colors (ignored if \code{col} is provided).}

\item{gridlines}{Color of gridlines (NULL for default choices; NA to skip them).}

\item{swap_axes}{If TRUE, make the scale horizontal}

\item{...}{Additional graphics paramaters.}
}
\value{
None.
}
\description{
Plot heat map color scale
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
probs <- calc_genoprob(grav2, map, error_prob=0.002)
out <- scan1(probs, grav2$pheno)

layout(cbind(1,2), widths=c(4, 1))
par(mar=c(4.1, 4.1, 1.1, 1.1))
plot_scan1_heatmap(out, map, xlab="Chromosome", ylab="Phenotype", mgp.y=c(2.6, 0.3, 0))
par(mar=c(15.1, 3.1, 15.1, 1.1))
plot_colorscale(out)
}
\seealso{
\code{\link[=plot_scan1_heatmap]{plot_scan1_heatmap()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary_cross2.R
\name{summary.cross2}
\alias{summary.cross2}
\title{Summary of cross2 object}
\usage{
\method{summary}{cross2}(object, ...)
}
\arguments{
\item{object}{An object of class \code{"cross2"}, as output by \code{\link[=read_cross2]{read_cross2()}}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{...}{Ignored.}
}
\value{
None.
}
\description{
Summarize a cross2 object
}
\seealso{
\link{basic_summaries}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_genes.R
\name{plot_genes}
\alias{plot_genes}
\title{Plot gene locations for a genomic interval}
\usage{
plot_genes(
  genes,
  minrow = 4,
  padding = 0.2,
  colors = c("black", "red3", "green4", "blue3", "orange"),
  scale_pos = 1,
  start_field = "start",
  stop_field = "stop",
  strand_field = "strand",
  name_field = "Name",
  ...
)
}
\arguments{
\item{genes}{Data frame containing \code{start} and \code{stop} in
Mbp, \code{strand} (as \code{"-"}, \code{"+"}, or \code{NA}), and
\code{Name}.}

\item{minrow}{Minimum number of rows of genes in the plot}

\item{padding}{Proportion to pad with white space around the genes}

\item{colors}{Vectors of colors, used sequentially and then re-used.}

\item{scale_pos}{Factor by which to scale position (for example, to convert basepairs to Mbp)}

\item{start_field}{Character string with name of column containing the genes' start positions.}

\item{stop_field}{Character string with name of column containing the genes' stop positions.}

\item{strand_field}{Character string with name of column containing the genes' strands.
(The values of the corresponding field can be character strings \code{"+"} or \code{"-"}, or numeric +1 or -1.)}

\item{name_field}{Character string with name of column containing the genes' names.}

\item{...}{Optional arguments passed to \code{plot()}.}
}
\value{
None.
}
\description{
Plot gene locations for a genomic interval, as rectangles with gene
symbol (and arrow indicating strand/direction) below.
}
\section{Hidden graphics parameters}{

Graphics parameters can be passed via \code{...}. For
example, \code{xlim} to control the x-axis limits.
These are not included as formal
}

\examples{
genes <- data.frame(chr = c("6", "6", "6", "6", "6", "6", "6", "6"),
                    start = c(139988753, 140680185, 141708118, 142234227, 142587862,
                              143232344, 144398099, 144993835),
                    stop  = c(140041457, 140826797, 141773810, 142322981, 142702315,
                              143260627, 144399821, 145076184),
                    strand = c("-", "+", "-", "-", "-", NA, "+", "-"),
                    Name = c("Plcz1", "Gm30215", "Gm5724", "Slco1a5", "Abcc9",
                             "4930407I02Rik", "Gm31777", "Bcat1"),
                    stringsAsFactors=FALSE)

# use scale_pos=1e-6 because data in bp but we want the plot in Mbp
plot_genes(genes, xlim=c(140, 146), scale_pos=1e-6)
}
\keyword{hgraphics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbind_scan1.R
\name{rbind.scan1}
\alias{rbind.scan1}
\title{Join genome scan results for different chromosomes.}
\usage{
\method{rbind}{scan1}(...)
}
\arguments{
\item{...}{Genome scan objects of class \code{"scan1"}, as produced by \code{\link[=scan1]{scan1()}}.
Must have the same lod score columns.}
}
\value{
An object of class `"scan1", like the inputs, but with
the results for different sets of chromosomes combined.
}
\description{
Join multiple \code{\link[=scan1]{scan1()}} results for different
chromosomes; must have the same set of lod score column.
}
\details{
If components \code{addcovar}, \code{Xcovar},
\code{intcovar}, \code{weights}, \code{sample_size} do not match
between objects, we omit this information.

If \code{hsq} present, we simply \code{rbind()} the contents.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
probs <- calc_genoprob(grav2, map, error_prob=0.002)
phe <- grav2$pheno[,1,drop=FALSE]

out1 <- scan1(probs[,1], phe) # chr 1
out2 <- scan1(probs[,5], phe) # chr 5
out <- rbind(out1, out2)

}
\seealso{
\code{\link[=cbind.scan1]{cbind.scan1()}}, \code{\link[=scan1]{scan1()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_genoprob.R
\name{compare_genoprob}
\alias{compare_genoprob}
\title{Compare two sets of genotype probabilities}
\usage{
compare_genoprob(
  probs1,
  probs2,
  cross,
  ind = 1,
  chr = NULL,
  minprob = 0.95,
  minmarkers = 10,
  minwidth = 0,
  annotate = FALSE
)
}
\arguments{
\item{probs1}{Genotype probabilities (as produced by \code{\link[=calc_genoprob]{calc_genoprob()}})
or allele dosages (as produced by \code{\link[=genoprob_to_alleleprob]{genoprob_to_alleleprob()}}).}

\item{probs2}{A second set of genotype probabilities, just like \code{probs1}.}

\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{ind}{Individual to plot, either a numeric index or an ID.}

\item{chr}{Selected chromosome; a single character string.}

\item{minprob}{Minimum probability for inferring genotypes (passed to \code{\link[=maxmarg]{maxmarg()}}).}

\item{minmarkers}{Minimum number of markers in results.}

\item{minwidth}{Minimum width in results.}

\item{annotate}{If TRUE, add some annotations to the \code{geno1} and
\code{geno2} columns to indicate, where they differ, which one
matches what appears to be the best genotype. (\code{*} = matches
the best genotype; \code{-} = lower match).}
}
\value{
A data frame with each row corresponding to an interval over which
\code{probs1} and \code{probs2} each have a fixed inferred genotype. Columns
include the two inferred genotypes, the start and end points and
width of the interval, and when founder genotypes are in \code{cross},
the proportions of SNPs where the individual matches each possible
genotypes.
}
\description{
Compare two sets of genotype probabilities for one individual on a single chromosome.
}
\details{
The function does the following:
\itemize{
\item Reduce the probabilities to a set of common locations that also appear in \code{cross}.
\item Use \code{\link[=maxmarg]{maxmarg()}} to infer the genotype at every position using each set of probabilities.
\item Identify intervals where the two inferred genotypes are constant.
\item Within each segment, compare the observed SNP genotypes to the founders' genotypes.
}
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
iron <- iron[1,"2"]   # subset to first individual on chr 2
map <- insert_pseudomarkers(iron$gmap, step=1)

# in presence of a genotyping error, how much does error_prob matter?
iron$geno[[1]][1,3] <- 3
pr_e <- calc_genoprob(iron, map, error_prob=0.002)
pr_ne <- calc_genoprob(iron, map, error_prob=1e-15)

compare_genoprob(pr_e, pr_ne, iron, minmarkers=1, minprob=0.5)

}
\seealso{
\code{\link[=plot_genoprobcomp]{plot_genoprobcomp()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit1.R
\name{fit1}
\alias{fit1}
\title{Fit single-QTL model at a single position}
\usage{
fit1(
  genoprobs,
  pheno,
  kinship = NULL,
  addcovar = NULL,
  nullcovar = NULL,
  intcovar = NULL,
  weights = NULL,
  contrasts = NULL,
  model = c("normal", "binary"),
  zerosum = TRUE,
  se = TRUE,
  hsq = NULL,
  reml = TRUE,
  blup = FALSE,
  var = FALSE,
  ...
)
}
\arguments{
\item{genoprobs}{A matrix of genotype probabilities, individuals x genotypes.
If NULL, we create a single intercept column, matching the individual IDs in \code{pheno}.}

\item{pheno}{A numeric vector of phenotype values (just one phenotype, not a matrix of them)}

\item{kinship}{Optional kinship matrix.}

\item{addcovar}{An optional numeric matrix of additive covariates.}

\item{nullcovar}{An optional numeric matrix of additional additive
covariates that are used under the null hypothesis (of no QTL)
but not under the alternative (with a QTL). This is needed for
the X chromosome, where we might need sex as a additive
covariate under the null hypothesis, but we wouldn't want to
include it under the alternative as it would be collinear with
the QTL effects.}

\item{intcovar}{An optional numeric matrix of interactive covariates.}

\item{weights}{An optional numeric vector of positive weights for the
individuals. As with the other inputs, it must have \code{names}
for individual identifiers.}

\item{contrasts}{An optional numeric matrix of genotype contrasts, size
genotypes x genotypes. For an intercross, you might use
\code{cbind(mu=c(1,1,1), a=c(-1, 0, 1), d=c(0, 1, 0))} to get
mean, additive effect, and dominance effect. The default is the
identity matrix.}

\item{model}{Indicates whether to use a normal model (least
squares) or binary model (logistic regression) for the phenotype.
If \code{model="binary"}, the phenotypes must have values in
\eqn{[0, 1]}.}

\item{zerosum}{If TRUE, force the genotype or allele coefficients
sum to 0 by subtracting their mean and add another column with
the mean. Ignored if \code{contrasts} is provided.}

\item{se}{If TRUE, calculate the standard errors.}

\item{hsq}{(Optional) residual heritability; used only if
\code{kinship} provided.}

\item{reml}{If \code{kinship} provided: if \code{reml=TRUE}, use
REML; otherwise maximum likelihood.}

\item{blup}{If TRUE, fit a model with QTL effects being random, as in \code{\link[=scan1blup]{scan1blup()}}.}

\item{var}{If TRUE, calculate the variance-covariance matrix of the coefficients
(in which case \code{se} is ignored, and standard errors will also be included).
But note that the variance-covariance matrix is only returned
if the covariate matrix is full rank.}

\item{...}{Additional control parameters; see Details;}
}
\value{
A list containing
\itemize{
\item \code{coef} - Vector of estimated coefficients.
\item \code{SE} - Vector of estimated standard errors (included if \code{se=TRUE}).
\item \code{lod} - The overall lod score.
\item \code{ind_lod} - Vector of individual contributions to the LOD score (not provided if \code{kinship} is used).
\item \code{fitted}  - Fitted values.
\item \code{resid}  - Residuals.
If \code{blup==TRUE}, only \code{coef} and \code{SE} are included at present.
}
}
\description{
Fit a single-QTL model at a single putative QTL position and get detailed results
about estimated coefficients and individuals contributions to the LOD score.
}
\details{
For each of the inputs, the row names are used as
individual identifiers, to align individuals.

If \code{kinship} is absent, Haley-Knott regression is performed.
If \code{kinship} is provided, a linear mixed model is used, with a
polygenic effect estimated under the null hypothesis of no (major)
QTL, and then taken as fixed as known in the genome scan.

If \code{contrasts} is provided, the genotype probability matrix,
\eqn{P}, is post-multiplied by the contrasts matrix, \eqn{A}, prior
to fitting the model. So we use \eqn{P \cdot A}{P A} as the \eqn{X}
matrix in the model. One might view the rows of
\ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}
as the set of contrasts, as the estimated effects are the estimated
genotype effects pre-multiplied by
\ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}.

The \code{...} argument can contain several additional control
parameters; suspended for simplicity (or confusion, depending on
your point of view). \code{tol} is used as a tolerance value for linear
regression by QR decomposition (in determining whether columns are
linearly dependent on others and should be omitted); default
\code{1e-12}. \code{maxit} is the maximum number of iterations for
converence of the iterative algorithm used when \code{model=binary}.
\code{bintol} is used as a tolerance for converence for the iterative
algorithm used when \code{model=binary}. \code{eta_max} is the maximum value
for the "linear predictor" in the case \code{model="binary"} (a bit of a
technicality to avoid fitted values exactly at 0 or 1).
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,"7"] # subset to chr 7}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=5)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno[,1]
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)

# scan chromosome 7 to find peak
out <- scan1(probs[,"7"], pheno, addcovar=covar)

# find peak position
max_pos <- max(out, map)

# genoprobs at max position
pr_max <- pull_genoprobpos(probs, map, max_pos$chr, max_pos$pos)

# fit QTL model just at that position
out_fit1 <- fit1(pr_max, pheno, addcovar=covar)

# plot estimated effects
plot_ci(out_fit1$coef[1:3], out_fit1$SE[1:3],
        xlab="Genotype", ylab="QTL effects")

}
\references{
Haley CS, Knott SA (1992) A simple
regression method for mapping quantitative trait loci in line
crosses using flanking markers.  Heredity 69:315--324.

Kang HM, Zaitlen NA, Wade CM, Kirby A, Heckerman D, Daly MJ, Eskin
E (2008) Efficient control of population structure in model
organism association mapping. Genetics 178:1709--1723.
}
\seealso{
\code{\link[=pull_genoprobpos]{pull_genoprobpos()}}, \code{\link[=find_marker]{find_marker()}}, \code{\link[=plot_ci]{plot_ci()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_errorlod.R
\name{calc_errorlod}
\alias{calc_errorlod}
\title{Calculate genotyping error LOD scores}
\usage{
calc_errorlod(cross, probs, quiet = TRUE, cores = 1)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{probs}{Genotype probabilities as calculated from \code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{quiet}{If \code{FALSE}, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A list of matrices of genotyping error LOD scores. Each
matrix corresponds to a chromosome and is arranged as
individuals x markers.
}
\description{
Use the genotype probabilities calculated with
\code{\link[=calc_genoprob]{calc_genoprob()}} to calculate genotyping error LOD
scores, to help identify potential genotyping errors (and problem
markers and/or individuals).
}
\details{
Let \eqn{O_k}{O[k]} denote the observed marker genotype at position
\eqn{k}, and \eqn{g_k}{g[k]} denote the corresponding true underlying
genotype.

Following Lincoln and Lander (1992), we calculate
LOD = \eqn{log_{10} [ Pr(O_k | g_k = O_k) / Pr(O_k | g_k \ne O_K) ]}{
 log10[ Pr(O[k] | g[k] = O[k]) / Pr(O[k] | g[k] != O[k]) ]}
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
probs <- calc_genoprob(iron, error_prob=0.002, map_function="c-f")
errorlod <- calc_errorlod(iron, probs)

# combine into one matrix
errorlod <- do.call("cbind", errorlod)
}
\references{
Lincoln SE, Lander ES (1992) Systematic detection of errors in genetic linkage data. Genomics 14:604--610.
}
\seealso{
\code{\link[=calc_genoprob]{calc_genoprob()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_threshold.R
\name{add_threshold}
\alias{add_threshold}
\title{Add thresholds to genome scan plot}
\usage{
add_threshold(map, thresholdA, thresholdX = NULL, chr = NULL, gap = NULL, ...)
}
\arguments{
\item{map}{Marker map used in the genome scan plot}

\item{thresholdA}{Autosomal threshold. Numeric or a list. (If a
list, the \code{"A"} component is taken to be \code{thresholdA} and the
\code{"X"} component is taken to be \code{thresholdX}.)}

\item{thresholdX}{X chromosome threshold (if missing, assumed to be the same as \code{thresholdA})}

\item{chr}{Chromosomes that were included in the plot}

\item{gap}{Gap between chromosomes in the plot. Default is 1\% of the total genome length.}

\item{...}{Additional arguments passed to \code{\link[graphics:segments]{graphics::segments()}}}
}
\value{
None.
}
\description{
Draw line segments at significance thresholds for a genome scan plot
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c(3,15,"X")]}
map <- insert_pseudomarkers(iron$gmap, step=5)
probs <- calc_genoprob(iron, map, error_prob=0.002)
Xcovar <- get_x_covar(iron)
out <- scan1(probs, iron$pheno[,1], Xcovar=Xcovar)
# run just 3 permutations, as a fast illustration
operm <- scan1perm(probs, iron$pheno[,1], addcovar=Xcovar,
                   n_perm=3, perm_Xsp=TRUE, chr_lengths=chr_lengths(map))

plot(out, map)
add_threshold(map, summary(operm), col="violetred", lty=2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_xo.R
\name{count_xo}
\alias{count_xo}
\title{Count numbers of crossovers}
\usage{
count_xo(geno, quiet = TRUE, cores = 1)
}
\arguments{
\item{geno}{List of matrices of genotypes (output of \code{\link[=maxmarg]{maxmarg()}} or \code{\link[=viterbi]{viterbi()}})
or a list of 3d-arrays of genotypes (output of \code{\link[=sim_geno]{sim_geno()}}).}

\item{quiet}{If FALSE, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A matrix of crossover counts, individuals x chromosomes, or
(if the input was the output of \code{\link[=sim_geno]{sim_geno()}}) a
3d-array of crossover counts, individuals x chromosomes x
imputations.
}
\description{
Estimate the numbers of crossovers in each individual on each chromosome.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[1:20,1:2] # subset to first 20 individuals and to chr 1 and 2}
map <- insert_pseudomarkers(iron$gmap, step=1)
pr <- calc_genoprob(iron, map, error_prob=0.002, map_function="c-f")
g <- maxmarg(pr)
n_xo <- count_xo(g)

# imputations
imp <- sim_geno(iron, map, error_prob=0.002, map_function="c-f", n_draws=32)
n_xo_imp <- count_xo(imp)
# sums across chromosomes
tot_xo_imp <- apply(n_xo_imp, c(1,3), sum)
# mean and SD across imputations
summary_xo <- cbind(mean=rowMeans(tot_xo_imp),
                    sd=apply(tot_xo_imp, 1, sd))

}
\seealso{
\code{\link[=locate_xo]{locate_xo()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interp_genoprob.R
\name{interp_genoprob}
\alias{interp_genoprob}
\title{Interpolate genotype probabilities}
\usage{
interp_genoprob(probs, map, cores = 1)
}
\arguments{
\item{probs}{Genotype probabilities, as calculated from
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{map}{List of vectors of map positions.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
An object of class \code{"calc_genoprob"}, like the input,
but with additional positions present in \code{map}. See \code{\link[=calc_genoprob]{calc_genoprob()}}.
}
\description{
Linear interpolation of genotype probabilities, mostly to get two sets onto the same map for comparison purposes.
}
\details{
We reduce \code{probs} to the positions present in \code{map} and then
interpolate the genotype probabilities at additional positions
in \code{map} by linear interpolation using the two adjacent
positions. Off the ends, we just copy over the first or last
value unchanged.

In general, it's better to use \code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}} and
\code{\link[=calc_genoprob]{calc_genoprob()}} to get genotype probabilities at additional
positions along a chromosome. This function is a \strong{very} crude
alternative that was implemented in order to compare genotype
probabilities derived by different methods, where we first need to
get them onto a common set of positions.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[1:20,c("1", "X")]}
probs <- calc_genoprob(iron, iron$gmap, error_prob=0.002)

# you generally wouldn't want to do this, but this is an illustration
map <- insert_pseudomarkers(iron$gmap, step=1)
probs_map <- interp_genoprob(probs, map)

}
\seealso{
\code{\link[=calc_genoprob]{calc_genoprob()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_sdp.R
\name{sdp2char}
\alias{sdp2char}
\title{Convert strain distribution patterns to character strings}
\usage{
sdp2char(sdp, n_strains = NULL, strains = NULL)
}
\arguments{
\item{sdp}{Vector of strain distribution patterns (integers between
1 and \eqn{2^n-2} where \eqn{n} is the number of strains.}

\item{n_strains}{Number of founder strains (if missing but
\code{strains} is provided, we use the length of \code{strains})}

\item{strains}{Vector of single-letter codes for the strains}
}
\value{
Vector of character strings with the two groups of alleles
separated by a vertical bar (|).
}
\description{
Convert a vector of numeric codes for strain distribution patterns to character strings.
}
\examples{
sdp <- c(m1=1, m2=12, m3=240)
sdp2char(sdp, 8)
sdp2char(sdp, strains=c("A", "B", "1", "D", "Z", "C", "P", "W"))
}
\seealso{
\code{\link[=invert_sdp]{invert_sdp()}}, \code{\link[=calc_sdp]{calc_sdp()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/qtl2version.R
\name{qtl2version}
\alias{qtl2version}
\title{Installed version of R/qtl2}
\usage{
qtl2version()
}
\value{
A character string with the installed version of the R/qtl2 package.
}
\description{
Get installed version of R/qtl2
}
\examples{
qtl2version()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_hotspots.R
\name{calc_hotspots}
\alias{calc_hotspots}
\title{Calculate QTL hotspots}
\usage{
calc_hotspots(peaks, map, window = 1, cores = 1, quiet = TRUE)
}
\arguments{
\item{peaks}{Data frame of QTL results, as output by \code{\link[=find_peaks]{find_peaks()}}
Should contain columns \code{chr} and \code{pos}.}

\item{map}{Marker map, as a list of chromosomes, each being a vector of
positions. Hotspot counts will be calculated at these
positions.}

\item{window}{Window size for counting QTL.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}

\item{quiet}{If TRUE, don't print any messages.}
}
\value{
An object of class \code{"scan1"}: a matrix with a single
column, of counts, with rownames being the marker names in
\code{map}. The column name is \code{"num_qtl"}.
}
\description{
For a set of QTL locations, calculate a running count in a sliding
window across the genome.
}
\examples{
\dontrun{
# download example pQTL results (from Keele et al. 2026, https://doi.org/10.1016/j.xgen.2025.101069)
# contains qtl, map_endpoints, and pheno_pos
url <- "https://kbroman.org/qtl2/assets/sampledata/pqtl_data.RData"
tempfile <- file.path(tempdir(), basename(url))
download.file(url, tempfile)
load(tempfile)
unlink(tempfile)

hotspots <- calc_hotspots(qtl, map, window=2)
plot(hotspots, map, ylab="No. QTL")
find_peaks(hotspots, map, threshold=20)
}
}
\seealso{
\code{\link[=find_peaks]{find_peaks()}}, \code{\link[=plot_lodpeaks]{plot_lodpeaks()}}, \code{\link[=plot_cistrans]{plot_cistrans()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sim_geno.R
\name{sim_geno}
\alias{sim_geno}
\title{Simulate genotypes given observed marker data}
\usage{
sim_geno(
  cross,
  map = NULL,
  n_draws = 1,
  error_prob = 0.0001,
  map_function = c("haldane", "kosambi", "c-f", "morgan"),
  lowmem = FALSE,
  quiet = TRUE,
  cores = 1
)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{map}{Genetic map of markers. May include pseudomarker
locations (that is, locations that are not within the marker
genotype data). If NULL, the genetic map in \code{cross} is used.}

\item{n_draws}{Number of simulations to perform.}

\item{error_prob}{Assumed genotyping error probability}

\item{map_function}{Character string indicating the map function to
use to convert genetic distances to recombination fractions.}

\item{lowmem}{If \code{FALSE}, split individuals into groups with
common sex and crossinfo and then precalculate the transition
matrices for a chromosome; potentially a lot faster but using more
memory.}

\item{quiet}{If \code{FALSE}, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
An object of class \code{"sim_geno"}: a list of three-dimensional arrays of imputed genotypes,
individuals x positions x draws. Also contains three attributes:
\itemize{
\item \code{crosstype} - The cross type of the input \code{cross}.
\item \code{is_x_chr} - Logical vector indicating whether chromosomes
are to be treated as the X chromosome or not, from input \code{cross}.
\item \code{alleles} - Vector of allele codes, from input
\code{cross}.
}
}
\description{
Uses a hidden Markov model to simulate from the joint distribution
Pr(g | O) where g is the underlying sequence of true genotypes and
O is the observed multipoint marker data, with possible allowance
for genotyping errors.
}
\details{
After performing the backward equations, we draw from
\eqn{Pr(g_1 = v | O)}{Pr(g[1] = v | O)} and then \eqn{Pr(g_{k+1} = v |
   O, g_k = u)}{Pr(g[k+1] = v | O, g[k] = u)}.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map_w_pmar <- insert_pseudomarkers(grav2$gmap, step=1)
draws <- sim_geno(grav2, map_w_pmar, n_draws=4, error_prob=0.002)
}
\seealso{
\code{\link[=cbind.sim_geno]{cbind.sim_geno()}}, \code{\link[=rbind.sim_geno]{rbind.sim_geno()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_geno.R
\name{plot_geno}
\alias{plot_geno}
\title{Plot multiple individuals' genome-wide genotypes}
\usage{
plot_geno(
  geno,
  map,
  ind = NULL,
  chr = NULL,
  gap = 0,
  col = NULL,
  na_col = "white",
  chrlines = "white",
  swap_axes = FALSE,
  ...
)
}
\arguments{
\item{geno}{Imputed phase-known genotypes, as a list of matrices
(as produced by \code{\link[=maxmarg]{maxmarg()}}) or a list of
three-dimensional arrays (as produced by \code{\link[=guess_phase]{guess_phase()}}).}

\item{map}{Marker map (a list of vectors of marker positions).}

\item{ind}{Individuals to plot, either a numeric indexes or IDs.}

\item{chr}{Selected chromosomes to plot; a vector of character strings.}

\item{gap}{Gap between chromosomes}

\item{col}{Vector of colors for the different genotypes.}

\item{na_col}{Color for missing segments.}

\item{chrlines}{Color for lines separating chromosomes}

\item{swap_axes}{If TRUE, swap the axes, so that the chromosomes run horizontally.}

\item{...}{Additional graphics parameters}
}
\value{
None.
}
\description{
Plot multiple individuals' genome-wide genotypes
}
\examples{
# load data and calculate genotype probabilities
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
iron <- iron[1:50, ] # subset to first 50 individuals
map <- insert_pseudomarkers(iron$gmap, step=1)
pr <- calc_genoprob(iron, map, error_prob=0.002)

# infer genotypes, as those with maximal marginal probability
m <- maxmarg(pr, minprob=0.5)

# re-code the X chr, (5,6) -> (1,3)
m[["X"]] <- (m[["X"]] - 5)*2 + 1

# plot phased genotypes
plot_geno(m, map, col=c("#FFDC00", "#00C800", "#0064C9"))

# this is more interesting for Diversity Outbred mouse data
\dontrun{
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
# subset to first 25 individuals
DOex <- DOex[1:25, ]
pr <- calc_genoprob(DOex, error_prob=0.002)

# infer genotypes, as those with maximal marginal probability
m <- maxmarg(pr, minprob=0.5)
# guess phase
ph <- guess_phase(DOex, m)

# plot phased genotypes
plot_geno(ph, DOex$gmap)
}

}
\seealso{
\code{\link[=plot_onegeno]{plot_onegeno()}}, \code{\link[=plot_genoprob]{plot_genoprob()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_ci.R
\name{plot_ci}
\alias{plot_ci}
\title{Effect plot with multiple CIs for different groups}
\usage{
plot_ci(
  est,
  se = NULL,
  lo = NULL,
  hi = NULL,
  SEmult = 2,
  labels = NULL,
  swap_axes = FALSE,
  add = FALSE,
  offset = 0,
  ...
)
}
\arguments{
\item{est}{Vector of estimates}

\item{se}{Vector of standard errors}

\item{lo}{Vector of lower values for the intervals}

\item{hi}{Vector of upper values for the intervals}

\item{SEmult}{SE multiplier to create intervals}

\item{labels}{Labels for the groups (vector of character strings)}

\item{swap_axes}{If TRUE, have group as y-axis; default (FALSE) has
group on x-axis.}

\item{add}{If TRUE, add the CIs to a previous plot with the same axes}

\item{offset}{Amount to offset the positions of the confidence intervals
(particularly for the case \code{add=TRUE})}

\item{...}{Optional graphics arguments}
}
\value{
None.
}
\description{
Plot a set of confidence intervals.
}
\details{
Provide either \code{se} or both \code{lo} and \code{hi}. In the case that \code{se} is
used, the intervals will be \code{est} +/- \code{SEmult * se}.

If \code{labels} is not provided, group names are taken from the \code{names(est)}.
If that is also missing, we use capital letters.

You can control the CI line widths with \code{ci_lwd} and the color of
the CI segments with \code{ci_col}. You can control the width of the
segments at the top and bottom with \code{ci_endseg}.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
iron <- iron[,"9"] # only chr 9
map <- insert_pseudomarkers(iron$gmap, step=1)
probs <- calc_genoprob(iron, map, error_prob=0.002)
pr <- pull_genoprobpos(probs, map, chr="9", pos=56.6)
out_fit1 <- fit1(pr, iron$pheno[,"spleen"], se=TRUE, zerosum=FALSE)
plot_ci(out_fit1$coef[1:3], out_fit1$SE[1:3],
        xlab="Genotype chr 9 @ 56.6 cM", ylab="Spleen phenotype")

}
\seealso{
\code{\link[=plot_coef]{plot_coef()}}, \code{\link[=plot_pxg]{plot_pxg()}}, \code{\link[=fit1]{fit1()}}, \code{\link[=pull_genoprobpos]{pull_genoprobpos()}}
}
\keyword{graphics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/probs_to_grid.R
\name{probs_to_grid}
\alias{probs_to_grid}
\title{Subset genotype probability array to pseudomarkers on a grid}
\usage{
probs_to_grid(probs, grid)
}
\arguments{
\item{probs}{Genotype probabilities as output from
\code{\link[=calc_genoprob]{calc_genoprob()}} with \code{stepwidth="fixed"}.}

\item{grid}{List of logical vectors that indicate which positions
are on the grid and should be retained.}
}
\value{
An object of class \code{"calc_genoprob"}, like the input, subset to just include
pseudomarkers along a grid. See \code{\link[=calc_genoprob]{calc_genoprob()}}.
}
\description{
Subset genotype probability array (from \code{\link[=calc_genoprob]{calc_genoprob()}}
to a grid of pseudomarkers along each chromosome.
}
\details{
This only works if \code{\link[=calc_genoprob]{calc_genoprob()}} was run
with \code{stepwidth="fixed"}, so that the genotype
probabilities were calculated at a grid of
markers/pseudomarkers. When this is the case, we omit all but
the probabilities on this grid. Use \code{\link[=calc_grid]{calc_grid()}} to
find the grid positions.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map_w_pmar <- insert_pseudomarkers(grav2$gmap, step=1)
probs <- calc_genoprob(grav2, map_w_pmar, error_prob=0.002)
sapply(probs, dim)
grid <- calc_grid(grav2$gmap, step=1)
probs_sub <- probs_to_grid(probs, grid)
sapply(probs_sub, dim)
}
\seealso{
\code{\link[=calc_grid]{calc_grid()}}, \code{\link[=map_to_grid]{map_to_grid()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/top_snps.R
\name{top_snps}
\alias{top_snps}
\title{Create table of top snp associations}
\usage{
top_snps(
  scan1_output,
  snpinfo,
  lodcolumn = 1,
  chr = NULL,
  drop = 1.5,
  show_all_snps = TRUE
)
}
\arguments{
\item{scan1_output}{Output of \code{\link[=scan1]{scan1()}}.
Should contain a component \code{"snpinfo"}, as when
\code{\link[=scan1]{scan1()}} is run with SNP probabilities
produced by \code{\link[=genoprob_to_snpprob]{genoprob_to_snpprob()}}.}

\item{snpinfo}{Data frame with SNP information with the following
columns (the last three are generally derived with
\code{\link[=index_snps]{index_snps()}}):
\itemize{
\item \code{chr} - Character string or factor with chromosome
\item \code{pos} - Position (in same units as in the \code{"map"}
attribute in \code{genoprobs}.
\item \code{sdp} - Strain distribution pattern: an integer, between
1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
binary encoding indicates the founder genotypes
\item \code{snp} - Character string with SNP identifier (if
missing, the rownames are used).
\item \code{index} - Indices that indicate equivalent
groups of SNPs, calculated by \code{\link[=index_snps]{index_snps()}}.
\item \code{intervals} - Indexes that indicate which marker
intervals the SNPs reside.
\item \code{on_map} - Indicate whether SNP coincides with a marker
in the \code{genoprobs}
}}

\item{lodcolumn}{Selected LOD score column to (a numeric index, or a
character string for a column name). Only one value allowed.}

\item{chr}{Selected chromosome; only one value allowed.}

\item{drop}{Show all SNPs with LOD score within this amount of the
maximum SNP association.}

\item{show_all_snps}{If TRUE, expand to show all SNPs.}
}
\value{
Data frame like the input \code{snpinfo} with just the selected
subset of rows, and with an added column with the LOD score.
}
\description{
Create a table of the top snp associations
}
\examples{
\dontrun{
# load example DO data from web
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)

# subset to chr 2
DOex <- DOex[,"2"]

# calculate genotype probabilities and convert to allele probabilities
pr <- calc_genoprob(DOex, error_prob=0.002)
apr <- genoprob_to_alleleprob(pr)

# query function for grabbing info about variants in region
dbfile <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
query_variants <- create_variant_query_func(dbfile)

# SNP association scan, keep information on all SNPs
out_snps <- scan1snps(apr, DOex$pmap, DOex$pheno, query_func=query_variants,
                      chr=2, start=97, end=98, keep_all_snps=TRUE)

# table with top SNPs
top_snps(out_snps$lod, out_snps$snpinfo)

# top SNPs among the distinct subset at which calculations were performed
top_snps(out_snps$lod, out_snps$snpinfo, show_all_snps=FALSE)

# top SNPs within 0.5 LOD of max
top_snps(out_snps$lod, out_snps$snpinfo, drop=0.5)
}
}
\seealso{
\code{\link[=index_snps]{index_snps()}}, \code{\link[=genoprob_to_snpprob]{genoprob_to_snpprob()}}, \code{\link[=scan1snps]{scan1snps()}}, \code{\link[=plot_snpasso]{plot_snpasso()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_geno.R
\name{plot_compare_geno}
\alias{plot_compare_geno}
\alias{plot.compare_geno}
\title{Plot of compare_geno object.}
\usage{
plot_compare_geno(x, rug = TRUE, ...)

\method{plot}{compare_geno}(x, rug = TRUE, ...)
}
\arguments{
\item{x}{A square matrix with genotype comparisons for pairs
of individuals, as output by \code{\link[=compare_geno]{compare_geno()}}.}

\item{rug}{If true, use \code{\link[=rug]{rug()}} to plot tick marks at observed values below the histogram.}

\item{...}{Additional graphics parameters passed to \code{\link[=hist]{hist()}}}
}
\value{
None.
}
\description{
From results of \code{\link[=compare_geno]{compare_geno()}}, plot histogram of
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
cg <- compare_geno(grav2)
plot(cg)
}
\keyword{graphics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_sdp.R
\name{invert_sdp}
\alias{invert_sdp}
\title{Calculate SNP genotype matrix from strain distribution patterns}
\usage{
invert_sdp(sdp, n_strains)
}
\arguments{
\item{sdp}{Vector of strain distribution patterns (integers between
1 and \eqn{2^n-2} where \eqn{n} is the number of strains.}

\item{n_strains}{Number of strains}
}
\value{
Matrix of SNP genotypes, markers x strains, coded as 1
(AA) and 3 (BB). Markers with values other than 1 or 3 are omitted,
and monomorphic markers, are omitted.
}
\description{
Calculate the matrix of SNP genotypes from a vector of strain distribution patterns (SDPs).
}
\examples{
sdp <- c(m1=1, m2=12, m3=240)
invert_sdp(sdp, 8)
}
\seealso{
\code{\link[=sdp2char]{sdp2char()}}, \code{\link[=calc_sdp]{calc_sdp()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lod_int.R
\name{lod_int}
\alias{lod_int}
\title{Calculate LOD support intervals}
\usage{
lod_int(
  scan1_output,
  map,
  chr = NULL,
  lodcolumn = 1,
  threshold = 0,
  peakdrop = Inf,
  drop = 1.5,
  expand2markers = TRUE
)
}
\arguments{
\item{scan1_output}{An object of class \code{"scan1"} as returned by
\code{\link[=scan1]{scan1()}}.}

\item{map}{A list of vectors of marker positions, as produced by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}.}

\item{chr}{Chromosome ID to consider (must be a single value).}

\item{lodcolumn}{LOD score column to consider (must be a single value).}

\item{threshold}{Minimum LOD score for a peak.}

\item{peakdrop}{Amount that the LOD score must drop between peaks,
if multiple peaks are to be defined on a chromosome.}

\item{drop}{Amount to drop in the support interval.  Must be
\eqn{\le} \code{peakdrop}}

\item{expand2markers}{If TRUE, QTL intervals are expanded so
that their endpoints are at genetic markers.}
}
\value{
A matrix with three columns:
\itemize{
\item \code{ci_lo} - lower bound of interval
\item \code{pos} - peak position
\item \code{ci_hi} - upper bound of interval
}

Each row corresponds to a different peak.
}
\description{
Calculate LOD support intervals for a single LOD curve on a single
chromosome, with the ability to identify intervals for multiple LOD
peaks.
}
\details{
We identify a set of peaks defined as local maxima that
exceed the specified \code{threshold}, with the requirement that
the LOD score must have dropped by at least \code{peakdrop} below
the lowest of any two adjacent peaks.

At a given peak, if there are ties, with multiple positions jointly
achieving the maximum LOD score, we take the average of these
positions as the location of the peak.

The default is to use \code{threshold=0}, \code{peakdrop=Inf}, and
\code{drop=1.5}. We then return results a single peak, no matter the
maximum LOD score, and give a 1.5-LOD support interval.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,7]}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# 1.5-LOD support interval for QTL on chr 7, first phenotype
lod_int(out, map, chr=7, lodcolum=1)
}
\seealso{
\code{\link[=bayes_int]{bayes_int()}}, \code{\link[=find_peaks]{find_peaks()}}, \code{\link[=scan1]{scan1()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interp_map.R
\name{interp_map}
\alias{interp_map}
\title{Interpolate between maps}
\usage{
interp_map(map, oldmap, newmap)
}
\arguments{
\item{map}{The map to be interpolated; a list of vectors.}

\item{oldmap}{Map with positions in the original scale, as in \code{map}.}

\item{newmap}{Map with positions in the new scale.}
}
\value{
Object of same form as input \code{map} but in the units as in \code{newmap}.
}
\description{
Use interpolate to convert from one map to another
}
\examples{
# load example data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

# positions to interpolate from cM to Mbp
tointerp <- list("7" = c(pos7.1= 5, pos7.2=15, pos7.3=25),
                 "9" = c(pos9.1=20, pos9.2=40))

interp_map(tointerp, iron$gmap, iron$pmap)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/locate_xo.R
\name{locate_xo}
\alias{locate_xo}
\title{Locate crossovers}
\usage{
locate_xo(geno, map, quiet = TRUE, cores = 1)
}
\arguments{
\item{geno}{List of matrices of genotypes (output of \code{\link[=maxmarg]{maxmarg()}} or \code{\link[=viterbi]{viterbi()}}).}

\item{map}{List of vectors with the map positions of the markers.}

\item{quiet}{If FALSE, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A list of lists of estimated crossover locations, with
crossovers placed at the midpoint of the intervals that contain
them.
}
\description{
Estimate the locations of crossovers in each individual on each chromosome.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
map <- insert_pseudomarkers(iron$gmap, step=1)
pr <- calc_genoprob(iron, map, error_prob=0.002, map_function="c-f")
g <- maxmarg(pr)
pos <- locate_xo(g, iron$gmap)

}
\seealso{
\code{\link[=count_xo]{count_xo()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/genoprob_to_snpprob.R
\name{genoprob_to_snpprob}
\alias{genoprob_to_snpprob}
\title{Convert genotype probabilities to SNP probabilities}
\usage{
genoprob_to_snpprob(genoprobs, snpinfo)
}
\arguments{
\item{genoprobs}{Genotype probabilities as
calculated by \code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{snpinfo}{Data frame with SNP information with the following
columns (the last three are generally derived with
\code{\link[=index_snps]{index_snps()}}):
\itemize{
\item \code{chr} - Character string or factor with chromosome
\item \code{pos} - Position (in same units as in the \code{"map"}
attribute in \code{genoprobs}.
\item \code{sdp} - Strain distribution pattern: an integer, between
1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
binary encoding indicates the founder genotypes
\item \code{snp} - Character string with SNP identifier (if
missing, the rownames are used).
\item \code{index} - Indices that indicate equivalent
groups of SNPs, calculated by \code{\link[=index_snps]{index_snps()}}.
\item \code{intervals} - Indexes that indicate which marker
intervals the SNPs reside.
\item \code{on_map} - Indicate whether SNP coincides with a marker
in the \code{genoprobs}
}

Alternatively, \code{snpinfo} can be a object of class \code{"cross2"},
as output by \code{\link[=read_cross2]{read_cross2()}}, containing the data for a multi-parent
population with founder genotypes, in which case the SNP
information for all markers with complete founder genotype data is
calculated and then used. But, in this case, the genotype
probabilities must be at the markers in the cross.}
}
\value{
An object of class \code{"calc_genoprob"}, like the input \code{genoprobs},
but with imputed genotype probabilities at the selected SNPs indicated in
\code{snpinfo$index}. See \code{\link[=calc_genoprob]{calc_genoprob()}}.

If the input \code{genoprobs} is for allele probabilities, the
\code{probs} output has just two probability columns (for the two SNP
alleles). If the input has a full set of \eqn{n(n+1)/2}
probabilities for \eqn{n} strains, the \code{probs} output has 3 probabilities
(for the three SNP genotypes). If the input has full genotype
probabilities for the X chromosome (\eqn{n(n+1)/2} genotypes for
the females followed by \eqn{n} hemizygous genotypes for the
males), the output has 5 probabilities: the 3 female SNP genotypes
followed by the two male hemizygous SNP genotypes.
}
\description{
For multi-parent populations, convert use founder genotypes at a
set of SNPs to convert founder-based genotype probabilities to SNP
genotype probabilities.
}
\details{
We first split the SNPs by chromosome and use
\code{snpinfo$index} to subset to non-equivalent SNPs.
\code{snpinfo$interval} indicates the intervals in the genotype
probabilities that contain each. For SNPs contained within an
interval, we use the average of the probabilities for the two
endpoints. We then collapse the probabilities according to the
strain distribution pattern.
}
\examples{
\dontrun{
# load example data and calculate genotype probabilities
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DO_Recla/recla.zip")
recla <- read_cross2(file)
recla <- recla[c(1:2,53:54), c("19","X")] # subset to 4 mice and 2 chromosomes
probs <- calc_genoprob(recla, error_prob=0.002)

# founder genotypes for a set of SNPs
snpgeno <- rbind(m1=c(3,1,1,3,1,1,1,1),
                 m2=c(1,3,1,3,1,3,1,3),
                 m3=c(1,1,1,1,3,3,3,3),
                 m4=c(1,3,1,3,1,3,1,3))
sdp <- calc_sdp(snpgeno)
snpinfo <- data.frame(chr=c("19", "19", "X", "X"),
                      pos=c(40.36, 40.53, 110.91, 111.21),
                      sdp=sdp,
                      snp=c("m1", "m2", "m3", "m4"), stringsAsFactors=FALSE)

# identify groups of equivalent SNPs
snpinfo <- index_snps(recla$pmap, snpinfo)

# collapse to SNP genotype probabilities
snpprobs <- genoprob_to_snpprob(probs, snpinfo)

# could also first convert to allele probs
aprobs <- genoprob_to_alleleprob(probs)
snpaprobs <- genoprob_to_snpprob(aprobs, snpinfo)
}

}
\seealso{
\code{\link[=index_snps]{index_snps()}}, \code{\link[=calc_genoprob]{calc_genoprob()}}, \code{\link[=scan1snps]{scan1snps()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drop_markers.R
\name{drop_nullmarkers}
\alias{drop_nullmarkers}
\title{Drop markers with no genotype data}
\usage{
drop_nullmarkers(cross, quiet = FALSE)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{quiet}{If FALSE, print information about how many markers were dropped.}
}
\value{
The input \code{cross} with the uninformative markers removed.
}
\description{
Drop markers with no genotype data (or no informative genotypes)
}
\details{
We omit any markers that have completely missing data, or
if founder genotypes are present (e.g., for Diversity Outbreds),
the founder genotypes are missing or are all the same.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
# make a couple of markers missing
grav2$geno[[2]][,c(3,25)] <- 0
grav2_rev <- drop_nullmarkers(grav2)
}
\seealso{
\code{\link[=drop_markers]{drop_markers()}}, \code{\link[=pull_markers]{pull_markers()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_snpgeno.R
\name{predict_snpgeno}
\alias{predict_snpgeno}
\title{Predict SNP genotypes}
\usage{
predict_snpgeno(cross, geno, cores = 1)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{geno}{Imputed genotypes, as a list of matrices, as from \code{\link[=maxmarg]{maxmarg()}}.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A list of matrices with inferred SNP genotypes, coded 1/2/3.
}
\description{
Predict SNP genotypes in a multiparent population from inferred genotypes plus founder strains' SNP alleles.
}
\examples{
\dontrun{
# load example data and calculate genotype probabilities
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
probs <- calc_genoprob(DOex, error_prob=0.002)

# inferred genotypes
m <- maxmarg(probs, minprob=0.5)

# inferred SNP genotypes
inferg <- predict_snpgeno(DOex, m)
}

}
\seealso{
\code{\link[=maxmarg]{maxmarg()}}, \code{\link[=viterbi]{viterbi()}}, \code{\link[=calc_errorlod]{calc_errorlod()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cbind_expand.R
\name{cbind_expand}
\alias{cbind_expand}
\title{Combine matrices by columns, expanding and aligning rows}
\usage{
cbind_expand(...)
}
\arguments{
\item{...}{A set of matrices or data frames}
}
\value{
The matrices combined by columns, using row names to align the rows, and expanding with missing values if there are rows in some matrices but not others.
}
\description{
This is like \code{\link[base:cbind]{base::cbind()}} but using row names to align the rows and expanding
with missing values if there are rows in some matrices but not others.
}
\examples{
df1 <- data.frame(x=c(1,2,3,NA,4), y=c(5,8,9,10,11), row.names=c("A", "B", "C", "D", "E"))
df2 <- data.frame(w=c(7,8,0,9,10), z=c(6,NA,NA,9,10), row.names=c("A", "B", "F", "C", "D"))
cbind_expand(df1, df2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset_viterbi.R
\name{subset.viterbi}
\alias{subset.viterbi}
\alias{[.viterbi}
\title{Subsetting Viterbi results}
\usage{
\method{subset}{viterbi}(x, ind = NULL, chr = NULL, ...)

\method{[}{viterbi}(x, ind = NULL, chr = NULL)
}
\arguments{
\item{x}{Imputed genotypes as output from \code{\link[=viterbi]{viterbi()}}.}

\item{ind}{A vector of individuals: numeric indices, logical
values, or character string IDs}

\item{chr}{A vector of chromosomes: logical values, or character
string IDs. Numbers are interpreted as character string IDs.}

\item{...}{Ignored.}
}
\value{
An object of class \code{"viterbi"}, like the input, with the
selected individuals and/or chromosomes; see \code{\link[=viterbi]{viterbi()}}.
}
\description{
Pull out a specified set of individuals and/or chromosomes from
the results of \code{\link[=viterbi]{viterbi()}}
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
\dontshow{grav2 <- grav2[1:8,c(1,2)]}
g <- viterbi(grav2)
# keep just individuals 1:5, chromosome 2
gsub <- g[1:5,2]
# keep just chromosome 2
gsub2 <- g[,2]
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_variant_query_func.R
\name{create_variant_query_func}
\alias{create_variant_query_func}
\title{Create a function to query variants}
\usage{
create_variant_query_func(
  dbfile = NULL,
  db = NULL,
  table_name = "variants",
  chr_field = "chr",
  pos_field = "pos",
  id_field = "snp_id",
  sdp_field = "sdp",
  filter = NULL
)
}
\arguments{
\item{dbfile}{Name of database file}

\item{db}{Optional database connection (provide one of \code{file} and \code{db}).}

\item{table_name}{Name of table in the database}

\item{chr_field}{Name of chromosome field}

\item{pos_field}{Name of position field}

\item{id_field}{Name of SNP/variant ID field}

\item{sdp_field}{Name of strain distribution pattern (SDP) field}

\item{filter}{Additional SQL filter (as a character string)}
}
\value{
Function with three arguments, \code{chr}, \code{start},
and \code{end}, which returns a data frame with the variants in
that region, with \code{start} and \code{end} being in Mbp. The
output should contain at least the columns \code{chr} and
\code{pos}, the latter being position in Mbp.
}
\description{
Create a function that will connect to a SQLite database of founder
variant information and return a data frame with variants for a
selected region.
}
\details{
Note that this function assumes that the database has a
\code{pos} field that is in basepairs, but the selection uses
\code{start} and \code{end} positions in Mbp, and the output
data frame should have \code{pos} in Mbp.

Also note that a SQLite database of variants in the founder strains
of the mouse Collaborative Cross is available at figshare:
\doi{10.6084/m9.figshare.5280229.v3}
}
\examples{
# create query function by connecting to file
dbfile <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
query_variants <- create_variant_query_func(dbfile)
# query_variants will connect and disconnect each time
variants <- query_variants("2", 97.0, 98.0)

# create query function to just grab SNPs
query_snps <- create_variant_query_func(dbfile, filter="type=='snp'")
# query_variants will connect and disconnect each time
snps <- query_snps("2", 97.0, 98.0)

# connect and disconnect separately
library(RSQLite)
db <- dbConnect(SQLite(), dbfile)
query_variants <- create_variant_query_func(db=db)
variants <- query_variants("2", 97.0, 98.0)
dbDisconnect(db)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unsmooth_gmap.R
\name{unsmooth_gmap}
\alias{unsmooth_gmap}
\title{Unsmooth genetic map}
\usage{
unsmooth_gmap(gmap, pmap, alpha = 0.02)
}
\arguments{
\item{gmap}{Genetic map, as a list of numeric vectors; each vector gives marker
positions for a single chromosome.}

\item{pmap}{Physical map, as a list of numeric vectors; each vector gives marker
positions for a single chromosome, with the same chromosomes and markers as \code{gmap}.}

\item{alpha}{Proportion of mixture to take from constant recombination.}
}
\value{
A genetic map like the input \code{gmap}, but with the reverse
operation of \code{\link[=smooth_gmap]{smooth_gmap()}} applied, provided that exactly the
same physical map and \code{alpha} are used.
}
\description{
Performs the reverse operation of \code{\link[=smooth_gmap]{smooth_gmap()}}, in case one wants to go back
to the original genetic map.
}
\details{
An interval of genetic length \eqn{d_g}{dg} and physical
length \eqn{d_p}{dp} is changed to have length
\eqn{(d_g - \alpha d_p r)/(1-\alpha)}{(dg + alpha*dp*r)/(1-alpha)}
where \eqn{r = L_g / L_p}{r = Lg/Lp} is the chromosome-specific
recombination rate.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
gmap_adj <- smooth_gmap(iron$gmap, iron$pmap)
gmap_back <- unsmooth_gmap(gmap_adj, iron$pmap)

}
\seealso{
\code{\link[=smooth_gmap]{smooth_gmap()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_control_file.R
\name{write_control_file}
\alias{write_control_file}
\title{Write a control file for QTL data}
\usage{
write_control_file(
  output_file,
  crosstype = NULL,
  geno_file = NULL,
  founder_geno_file = NULL,
  gmap_file = NULL,
  pmap_file = NULL,
  pheno_file = NULL,
  covar_file = NULL,
  phenocovar_file = NULL,
  sex_file = NULL,
  sex_covar = NULL,
  sex_codes = NULL,
  crossinfo_file = NULL,
  crossinfo_covar = NULL,
  crossinfo_codes = NULL,
  geno_codes = NULL,
  alleles = NULL,
  xchr = NULL,
  sep = ",",
  na.strings = c("-", "NA"),
  comment.char = "#",
  geno_transposed = FALSE,
  founder_geno_transposed = FALSE,
  pheno_transposed = FALSE,
  covar_transposed = FALSE,
  phenocovar_transposed = FALSE,
  description = NULL,
  comments = NULL,
  overwrite = FALSE
)
}
\arguments{
\item{output_file}{File name (with path) of the
\href{https://yaml.org}{YAML} or \href{https://json.org}{JSON} file to be created, as a character
string. If extension is \code{.json}, JSON format is used; otherwise, YAML is used.}

\item{crosstype}{Character string with the cross type.}

\item{geno_file}{File name for genotype data.}

\item{founder_geno_file}{File name for the founder genotype data.}

\item{gmap_file}{File name for genetic map.}

\item{pmap_file}{File name for the physical map.}

\item{pheno_file}{File name for the phenotype data.}

\item{covar_file}{File name for the covariate data.}

\item{phenocovar_file}{File name for the phenotype covariate data
(i.e., metadata about the phenotypes).}

\item{sex_file}{File name for the individuals' sex. (Specify just
one of \code{sex_file} or \code{sex_covar}.)}

\item{sex_covar}{Column name in the covariate data that corresponds
to sex. (Specify just one of \code{sex_file} or \code{sex_covar}.)}

\item{sex_codes}{Named vector of character strings specifying the
encoding of sex. The names attribute should be the codes used in
the data files; the values within the vector should be
\code{"female"} and \code{"male"}.}

\item{crossinfo_file}{File name for the \code{cross_info} data. (Specify just
one of \code{crossinfo_file} or \code{crossinfo_covar}.)}

\item{crossinfo_covar}{Column name in the covariate data that
corresponds to the \code{cross_info} data. (Specify just one of
\code{crossinfo_file} or \code{crossinfo_covar}.)}

\item{crossinfo_codes}{In the case that there is a single cross
info column (whether in a file or as a covariate), you can
provide a named vector of character strings specifying the
encoding of \code{cross_info}. The names attribute should be the
codes used; the values within the vector should be the codes to
which they will be converted (for example, \code{0} and \code{1} for an
intercross).}

\item{geno_codes}{Named vector specifying the encoding of
genotypes. The names attribute has the codes used within the
genotype and founder genotype data files; the values within the
vector should be the integers to which the genotypes will be
converted.}

\item{alleles}{Vector of single-character codes for the founder
alleles.}

\item{xchr}{Character string with the ID for the X chromosome.}

\item{sep}{Character string that separates columns in the data files.}

\item{na.strings}{Vector of character strings with codes to be
treated as missing values.}

\item{comment.char}{Character string that is used as initial
character in a set of leading comment lines in the data files.}

\item{geno_transposed}{If TRUE, genotype file is transposed (with markers as rows).}

\item{founder_geno_transposed}{If TRUE, founder genotype file is transposed (with markers as rows).}

\item{pheno_transposed}{If TRUE, phenotype file is transposed (with phenotypes as rows).}

\item{covar_transposed}{If TRUE, covariate file is transposed (with covariates as rows).}

\item{phenocovar_transposed}{If TRUE, phenotype covariate file is transposed (with phenotype covariates as rows).}

\item{description}{Optional character string describing the data.}

\item{comments}{Vector of character strings to be inserted as
comments at the top of the file (in the case of YAML), with each
string as a line. For JSON, the comments are instead included
within the control object.}

\item{overwrite}{If TRUE, overwrite file if it exists. If FALSE
(the default) and the file exists, stop with an error.}
}
\value{
(Invisibly) The data structure that was written.
}
\description{
Write the control file (in \href{https://yaml.org}{YAML} or \href{https://json.org}{JSON}) needed
by \code{\link[=read_cross2]{read_cross2()}} for a set of QTL data.
}
\details{
This function takes a set of parameters and creates the
control file (in \href{https://yaml.org}{YAML} or \href{https://json.org}{JSON} format) needed
for the new input data file format for
\href{https://kbroman.org/qtl2/}{R/qtl2}.  See the
\href{https://kbroman.org/qtl2/pages/sampledata.html}{sample data files} and the
\href{https://kbroman.org/qtl2/assets/vignettes/input_files.html}{vignette describing the input file format}.
}
\examples{
# Control file for the sample dataset, grav2
grav2_control_file <- file.path(tempdir(), "grav2.yaml")
write_control_file(grav2_control_file,
                   crosstype="riself",
                   geno_file="grav2_geno.csv",
                   gmap_file="grav2_gmap.csv",
                   pheno_file="grav2_pheno.csv",
                   phenocovar_file="grav2_phenocovar.csv",
                   geno_codes=c(L=1L, C=2L),
                   alleles=c("L", "C"),
                   na.strings=c("-", "NA"))

# Control file for the sample dataset, iron
iron_control_file <- file.path(tempdir(), "iron.yaml")
write_control_file(iron_control_file,
                   crosstype="f2",
                   geno_file="iron_geno.csv",
                   gmap_file="iron_gmap.csv",
                   pheno_file="iron_pheno.csv",
                   covar_file="iron_covar.csv",
                   phenocovar_file="iron_phenocovar.csv",
                   geno_codes=c(SS=1L, SB=2L, BB=3L),
                   sex_covar="sex",
                   sex_codes=c(f="female", m="male"),
                   crossinfo_covar="cross_direction",
                   crossinfo_codes=c("(SxB)x(SxB)"=0L, "(BxS)x(BxS)"=1L),
                   xchr="X",
                   alleles=c("S", "B"),
                   na.strings=c("-", "NA"))

# Remove these files, to clean up temporary directory
unlink(c(grav2_control_file, iron_control_file))
}
\seealso{
\code{\link[=read_cross2]{read_cross2()}}, sample data files at
\url{https://kbroman.org/qtl2/pages/sampledata.html}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan1blup.R
\name{scan1blup}
\alias{scan1blup}
\title{Calculate BLUPs of QTL effects in scan along one chromosome}
\usage{
scan1blup(
  genoprobs,
  pheno,
  kinship = NULL,
  addcovar = NULL,
  nullcovar = NULL,
  contrasts = NULL,
  se = FALSE,
  reml = TRUE,
  tol = 0.000000000001,
  cores = 1,
  quiet = TRUE
)
}
\arguments{
\item{genoprobs}{Genotype probabilities as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{pheno}{A numeric vector of phenotype values (just one phenotype, not a matrix of them)}

\item{kinship}{Optional kinship matrix, or a list of kinship matrices (one
per chromosome), in order to use the LOCO (leave one chromosome
out) method.}

\item{addcovar}{An optional numeric matrix of additive covariates.}

\item{nullcovar}{An optional numeric matrix of additional additive
covariates that are used under the null hypothesis (of no QTL) but
not under the alternative (with a QTL). This is needed for the X
chromosome, where we might need sex as a additive covariate under
the null hypothesis, but we wouldn't want to include it under the
alternative as it would be collinear with the QTL effects. Only
used if \code{kinship} is provided but \code{hsq} is not, to get
estimate of residual heritability.}

\item{contrasts}{An optional numeric matrix of genotype contrasts, size
genotypes x genotypes. For an intercross, you might use
\code{cbind(mu=c(1,0,0), a=c(-1, 0, 1), d=c(0, 1, 0))} to get
mean, additive effect, and dominance effect. The default is the
identity matrix.}

\item{se}{If TRUE, also calculate the standard errors.}

\item{reml}{If \code{reml=TRUE}, use
REML to estimate variance components; otherwise maximum likelihood.}

\item{tol}{Tolerance value for convergence of linear mixed model fit.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}

\item{quiet}{If FALSE, print message about number of cores used when multi-core.}
}
\value{
An object of class \code{"scan1coef"}: a matrix of estimated regression coefficients, of dimension
positions x number of effects. The number of effects is
\code{n_genotypes + n_addcovar}.
May also contain the following attributes:
\itemize{
\item \code{SE} - Present if \code{se=TRUE}: a matrix of estimated
standard errors, of same dimension as \code{coef}.
\item \code{sample_size} - Vector of sample sizes used for each
phenotype
}
}
\description{
Calculate BLUPs of QTL effects in scan along one chromosome, with a
single-QTL model treating the QTL effects as random, with possible
allowance for covariates and for a residual polygenic effect.
}
\details{
For each of the inputs, the row names are used as
individual identifiers, to align individuals.

If \code{kinship} is provided, the linear mixed model accounts for
a residual polygenic effect, with a the polygenic variance
estimated under the null hypothesis of no (major) QTL, and then
taken as fixed as known in the scan to estimate QTL effects.

If \code{contrasts} is provided, the genotype probability matrix,
\eqn{P}, is post-multiplied by the contrasts matrix, \eqn{A}, prior
to fitting the model. So we use \eqn{P \cdot A}{P A} as the \eqn{X}
matrix in the model. One might view the rows of
\ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}
as the set of contrasts, as the estimated effects are the estimated
genotype effects pre-multiplied by
\ifelse{html}{\out{<em>A</em><sup>-1</sup>}}{\eqn{A^{-1}}}.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c(7,19)] # reduce to chr 7 and 19}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)
\dontshow{probs[["7"]] <- probs[["7"]][,,1:5] # reduce to very small number}

# convert to allele probabilities
aprobs <- genoprob_to_alleleprob(probs)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno[,1]
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)

# calculate BLUPs of coefficients for chromosome 7
blup <- scan1blup(aprobs[,"7"], pheno, addcovar=covar)

# leave-one-chromosome-out kinship matrix for chr 7
kinship7 <- calc_kinship(probs, "loco")[["7"]]

# calculate BLUPs of coefficients for chromosome 7, adjusting for residual polygenic effect
blup_pg <- scan1blup(aprobs[,"7"], pheno, kinship7, addcovar=covar)

}
\references{
Haley CS, Knott SA (1992) A simple
regression method for mapping quantitative trait loci in line
crosses using flanking markers.  Heredity 69:315--324.

Kang HM, Zaitlen NA, Wade CM, Kirby A, Heckerman D, Daly MJ, Eskin
E (2008) Efficient control of population structure in model
organism association mapping. Genetics 178:1709--1723.

Robinson GK (1991) That BLUP is a good thing: The estimation of
random effects. Statist Sci 6:15--32.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mat2strata.R
\name{mat2strata}
\alias{mat2strata}
\title{Define strata based on rows of a matrix}
\usage{
mat2strata(mat)
}
\arguments{
\item{mat}{A covariate matrix, as individuals x covariates}
}
\value{
A vector of character strings: for each row of \code{mat},
we use \code{\link[base:paste]{base::paste()}} with \code{collapse="|"}.
}
\description{
Use the rows of a matrix to define a set of strata for a stratified permutation test
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

Xcovar <- get_x_covar(iron)
perm_strata <- mat2strata(Xcovar)

}
\seealso{
\code{\link[=get_x_covar]{get_x_covar()}}, \code{\link[=scan1perm]{scan1perm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pull_genoprobint.R
\name{pull_genoprobint}
\alias{pull_genoprobint}
\title{Pull genotype probabilities for an interval}
\usage{
pull_genoprobint(genoprobs, map, chr, interval)
}
\arguments{
\item{genoprobs}{Genotype probabilities as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{map}{The marker map for the genotype probabilities}

\item{chr}{Chromosome ID (single character sting)}

\item{interval}{Interval (pair of numbers)}
}
\value{
A list containing a single 3d array of genotype probabilities, like the input \code{genoprobs}
but for the designated interval.
}
\description{
Pull out the genotype probabilities for a given genomic interval
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,"8"]}
gmap <- insert_pseudomarkers(iron$gmap, step=1)
pr <- calc_genoprob(iron, gmap, error_prob=0.002)

pr_sub <- pull_genoprobint(pr, gmap, "8", c(25, 35))

}
\seealso{
\code{\link[=find_marker]{find_marker()}}, \code{\link[=pull_genoprobpos]{pull_genoprobpos()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/est_map.R
\name{est_map}
\alias{est_map}
\title{Estimate genetic maps}
\usage{
est_map(
  cross,
  error_prob = 0.0001,
  map_function = c("haldane", "kosambi", "c-f", "morgan"),
  lowmem = FALSE,
  maxit = 10000,
  tol = 0.000001,
  quiet = TRUE,
  save_rf = FALSE,
  cores = 1
)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{error_prob}{Assumed genotyping error probability}

\item{map_function}{Character string indicating the map function to
use to convert genetic distances to recombination fractions.}

\item{lowmem}{If \code{FALSE}, precalculate initial and emission
probabilities, and at each iteration calculate the transition
matrices for a chromosome; potentially a lot faster but using
more memory. Needs to be tailored somewhat to cross type. For
example, multi-way RIL may need to reorder the transition
matrix according to cross order, and AIL and DO need separate
transition matrices for each generation.}

\item{maxit}{Maximum number of iterations in EM algorithm.}

\item{tol}{Tolerance for determining convergence}

\item{quiet}{If \code{FALSE}, print progress messages.}

\item{save_rf}{If \code{TRUE}, save the estimated recombination
fractions as an attribute (\code{"rf"}) of the result.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A list of numeric vectors, with the estimated marker
locations (in cM). The location of the initial marker on each
chromosome is kept the same as in the input \code{cross}.
}
\description{
Uses a hidden Markov model to re-estimate the genetic map for an
experimental cross, with possible allowance for genotyping errors.
}
\details{
The map is estimated assuming no crossover interference,
but a map function (by default, Haldane's) is used to derive the genetic distances.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
\dontshow{grav2 <- grav2[,"3"]}
gmap <- est_map(grav2, error_prob=0.002)
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cbind_calc_genoprob.R
\name{cbind.calc_genoprob}
\alias{cbind.calc_genoprob}
\title{Join genotype probabilities for different chromosomes}
\usage{
\method{cbind}{calc_genoprob}(...)
}
\arguments{
\item{...}{Genotype probability objects as produced by
\code{\link[=calc_genoprob]{calc_genoprob()}}. Must have the same set of individuals.}
}
\value{
An object of class \code{"calc_genoprob"}, like the input; see \code{\link[=calc_genoprob]{calc_genoprob()}}.
}
\description{
Join multiple genotype probability objects, as produced by
\code{\link[=calc_genoprob]{calc_genoprob()}}, for the same set of individuals but different
chromosomes.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
probsA <- calc_genoprob(grav2[1:5,1:2], map, error_prob=0.002)
probsB <- calc_genoprob(grav2[1:5,3:4], map, error_prob=0.002)
probs <- cbind(probsA, probsB)

}
\seealso{
\code{\link[=rbind.calc_genoprob]{rbind.calc_genoprob()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cbind_phasedgeno.R
\name{cbind.phasedgeno}
\alias{cbind.phasedgeno}
\title{Join phased genotype results for different chromosomes}
\usage{
\method{cbind}{phasedgeno}(...)
}
\arguments{
\item{...}{Imputed genotype objects as produced by
\code{\link[=guess_phase]{guess_phase()}}. Must have the same set of individuals.}
}
\value{
An object of class \code{"phasedgeno"}, like the input; see \code{\link[=guess_phase]{guess_phase()}}.
}
\description{
Join multiple phased genotype objects, as produced by \code{\link[=guess_phase]{guess_phase()}}, for the
same set of individuals but different chromosomes.
}
\examples{
\dontrun{
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
pr <- calc_genoprob(DOex, error_prob=0.002)
m <- maxmarg(pr)
phA <- guess_phase(DOex[,2], m[,2])
phB <- guess_phase(DOex[,3], m[,3])
ph <- cbind(phA, phB)
}

}
\seealso{
\code{\link[=rbind.phasedgeno]{rbind.phasedgeno()}}, \code{\link[=subset.phasedgeno]{subset.phasedgeno()}}, \code{\link[=guess_phase]{guess_phase()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbind_calc_genoprob.R
\name{rbind.calc_genoprob}
\alias{rbind.calc_genoprob}
\title{Join genotype probabilities for different individuals}
\usage{
\method{rbind}{calc_genoprob}(...)
}
\arguments{
\item{...}{Genotype probability objects as produced by
\code{\link[=calc_genoprob]{calc_genoprob()}}. Must have the same set of markers and
genotypes.}
}
\value{
An object of class \code{"calc_genoprob"}, like the input; see \code{\link[=calc_genoprob]{calc_genoprob()}}.
}
\description{
Join multiple genotype probability objects, as produced by
\code{\link[=calc_genoprob]{calc_genoprob()}}, for the same set of markers and genotypes but for
different individuals.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
probsA <- calc_genoprob(grav2[1:5,], map, error_prob=0.002)
probsB <- calc_genoprob(grav2[6:12,], map, error_prob=0.002)
probs <- rbind(probsA, probsB)

}
\seealso{
\code{\link[=cbind.calc_genoprob]{cbind.calc_genoprob()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary_cross2.R
\name{basic_summaries}
\alias{basic_summaries}
\alias{n_ind}
\alias{n_ind_geno}
\alias{n_ind_pheno}
\alias{n_ind_covar}
\alias{n_ind_gnp}
\alias{ind_ids}
\alias{ind_ids_geno}
\alias{ind_ids_pheno}
\alias{ind_ids_covar}
\alias{ind_ids_gnp}
\alias{n_chr}
\alias{n_founders}
\alias{founders}
\alias{chr_names}
\alias{tot_mar}
\alias{n_mar}
\alias{marker_names}
\alias{n_pheno}
\alias{pheno_names}
\alias{n_covar}
\alias{covar_names}
\alias{n_phenocovar}
\alias{phenocovar_names}
\title{Basic summaries of a cross2 object}
\usage{
n_ind(cross2)

n_ind_geno(cross2)

n_ind_pheno(cross2)

n_ind_covar(cross2)

n_ind_gnp(cross2)

ind_ids(cross2)

ind_ids_geno(cross2)

ind_ids_pheno(cross2)

ind_ids_covar(cross2)

ind_ids_gnp(cross2)

n_chr(cross2)

n_founders(cross2)

founders(cross2)

chr_names(cross2)

tot_mar(cross2)

n_mar(cross2)

marker_names(cross2)

n_pheno(cross2)

pheno_names(cross2)

n_covar(cross2)

covar_names(cross2)

n_phenocovar(cross2)

phenocovar_names(cross2)
}
\arguments{
\item{cross2}{An object of class \code{"cross2"}, as output by \code{\link[=read_cross2]{read_cross2()}}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}
}
\value{
Variously a number, vector of numbers, or vector of character strings.
}
\description{
Basic summaries of a cross2 object.
}
\section{Functions}{
\itemize{
\item \code{n_ind()}: Number of individuals (either genotyped or phenotyped)

\item \code{n_ind_geno()}: Number of genotyped individuals

\item \code{n_ind_pheno()}: Number of phenotyped individuals

\item \code{n_ind_covar()}: Number of individuals with covariate data

\item \code{n_ind_gnp()}: Number of individuals with both genotype and phenotype data

\item \code{ind_ids()}: IDs of individuals (either genotyped or phenotyped)

\item \code{ind_ids_geno()}: IDs of genotyped individuals

\item \code{ind_ids_pheno()}: IDs of phenotyped individuals

\item \code{ind_ids_covar()}: IDs of individuals with covariate data

\item \code{ind_ids_gnp()}: IDs of individuals with both genotype and phenotype data

\item \code{n_chr()}: Number of chromosomes

\item \code{n_founders()}: Number of founder strains

\item \code{founders()}: Names of founder strains

\item \code{chr_names()}: Chromosome names

\item \code{tot_mar()}: Total number of markers

\item \code{n_mar()}: Number of markers on each chromosome

\item \code{marker_names()}: Marker names

\item \code{n_pheno()}: Number of phenotypes

\item \code{pheno_names()}: Phenotype names

\item \code{n_covar()}: Number of covariates

\item \code{covar_names()}: Covariate names

\item \code{n_phenocovar()}: Number of phenotype covariates

\item \code{phenocovar_names()}: Phenotype covariate names

}}
\seealso{
\code{\link[=summary.cross2]{summary.cross2()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index_snps.R
\name{index_snps}
\alias{index_snps}
\title{Create index of equivalent SNPs}
\usage{
index_snps(map, snpinfo, tol = 0.00000001)
}
\arguments{
\item{map}{Physical map of markers and pseudomarkers; generally
created from \code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}} and
used for a set of genotype probabilities (calculated with
\code{\link[=calc_genoprob]{calc_genoprob()}}) that are to be used to
interpolate SNP genotype probabilities (with
\code{\link[=genoprob_to_snpprob]{genoprob_to_snpprob()}}).}

\item{snpinfo}{Data frame with SNP information with the following columns:
\itemize{
\item \code{chr} - Character string or factor with chromosome
\item \code{pos} - Position (in same units as in the \code{"map"}).
\item \code{sdp} - Strain distribution pattern: an integer, between
1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
binary encoding indicates the founder genotypes
\item \code{snp} - Character string with SNP identifier (if
missing, the rownames are used).
}}

\item{tol}{Tolerance for determining whether a SNP is exactly at a
position at which genotype probabilities were already calculated.}
}
\value{
A data frame containing the input \code{snpinfo} with three
added columns: \code{"index"} (which indicates the groups of
equivalent SNPs), \code{"interval"} (which indicates the map
interval containing the SNP, with values starting at 0), and
\code{on_map} (which indicates that the SNP is within
\code{tol} of a position on the map). The rows get reordered,
so that they are ordered by chromosome and position, and the
values in the \code{"index"} column are \emph{by chromosome}.
}
\description{
For a set of SNPs and a map of marker/pseudomarkers, partition the
SNPs into groups that are contained within common intervals and
have the same strain distribution pattern, and then create an index
to a set of distinct SNPs, one per partition.
}
\details{
We split the SNPs by chromosome and identify the intervals
in the \code{map} that contain each. For SNPs within \code{tol}
of a position at which the genotype probabilities were
calculated, we take the SNP to be at that position. For each
marker position or interval, we then partition the SNPs into
groups that have distinct strain distribution patterns, and
choose a single index SNP for each partition.
}
\examples{
\dontrun{
# load example data and calculate genotype probabilities
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DO_Recla/recla.zip")
recla <- read_cross2(file)

# founder genotypes for a set of SNPs
snpgeno <- rbind(m1=c(3,1,1,3,1,1,1,1),
                 m2=c(1,3,1,3,1,3,1,3),
                 m3=c(1,1,1,1,3,3,3,3),
                 m4=c(1,3,1,3,1,3,1,3))
sdp <- calc_sdp(snpgeno)
snpinfo <- data.frame(chr=c("19", "19", "X", "X"),
                      pos=c(40.36, 40.53, 110.91, 111.21),
                      sdp=sdp,
                      snp=c("m1", "m2", "m3", "m4"), stringsAsFactors=FALSE)

# update snp info by adding the SNP index column
snpinfo <- index_snps(recla$pmap, snpinfo)
}

}
\seealso{
\code{\link[=genoprob_to_snpprob]{genoprob_to_snpprob()}}, \code{\link[=scan1snps]{scan1snps()}}, \code{\link[=find_index_snp]{find_index_snp()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_common_ids.R
\name{get_common_ids}
\alias{get_common_ids}
\title{Get common set of IDs from objects}
\usage{
get_common_ids(..., complete.cases = FALSE)
}
\arguments{
\item{...}{A set of objects: vectors, lists, matrices, data frames,
and/or arrays. If one is a character vector with no names
attribute, it's taken to be a set of IDs, itself.}

\item{complete.cases}{If TRUE, look at matrices and non-character
vectors and keep only individuals with no missing values.}
}
\value{
A vector of character strings for the individuals that are
in common.
}
\description{
For a set objects with IDs as row names (or, for a vector, just
names), find the IDs that are present in all of the objects.
}
\details{
This is used (mostly internally) to align phenotypes,
genotype probabilities, and covariates in preparation for a genome
scan. The \code{complete.cases} argument is used to omit
individuals with any missing covariate values.
}
\examples{
x <- matrix(0, nrow=10, ncol=5); rownames(x) <- LETTERS[1:10]
y <- matrix(0, nrow=5, ncol=5);  rownames(y) <- LETTERS[(1:5)+7]
z <- LETTERS[5:15]
get_common_ids(x, y, z)

x[8,1] <- NA
get_common_ids(x, y, z)
get_common_ids(x, y, z, complete.cases=TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbind_sim_geno.R
\name{rbind.sim_geno}
\alias{rbind.sim_geno}
\title{Join genotype imputations for different individuals}
\usage{
\method{rbind}{sim_geno}(...)
}
\arguments{
\item{...}{Genotype imputation objects as produced by
\code{\link[=sim_geno]{sim_geno()}}. Must have the same set of markers and
genotypes.}
}
\value{
An object of class \code{"sim_geno"}, like the input; see \code{\link[=sim_geno]{sim_geno()}}.
}
\description{
Join multiple genotype imputation objects, as produced by
\code{\link[=sim_geno]{sim_geno()}}, for the same set of markers but for different
individuals.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
drawsA <- sim_geno(grav2[1:5,], map, error_prob=0.002, n_draws=4)
drawsB <- sim_geno(grav2[6:12,], map, error_prob=0.002, n_draws=4)
draws <- rbind(drawsA, drawsB)

}
\seealso{
\code{\link[=cbind.sim_geno]{cbind.sim_geno()}}, \code{\link[=sim_geno]{sim_geno()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary_scan1perm.R
\name{print.summary.scan1perm}
\alias{print.summary.scan1perm}
\title{Print summary of scan1perm permutations}
\usage{
\method{print}{summary.scan1perm}(x, digits = 3, ...)
}
\arguments{
\item{x}{Object of class \code{"summary.scan1perm"}, as produced by \code{\link[=summary_scan1perm]{summary_scan1perm()}}.}

\item{digits}{Number of digits in printing significance thresholds; passed to \code{\link[base:print]{base::print()}}.}

\item{...}{Ignored.}
}
\value{
Invisibly returns the input, \code{x}.
}
\description{
Print summary of scan1perm permutations
}
\details{
This is to go with \code{\link[=summary_scan1perm]{summary_scan1perm()}}, so
that the summary output is printed in a nice format. Generally
not called directly, but it can be in order to control the
number of digits that appear.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c(10,18,"X")]}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# permutations with genome scan (just 3 replicates, for illustration)
operm <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar,
                   n_perm=3)

print( summary(operm, alpha=c(0.20, 0.05)), digits=8 )

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan1.R
\name{scan1}
\alias{scan1}
\title{Genome scan with a single-QTL model}
\usage{
scan1(
  genoprobs,
  pheno,
  kinship = NULL,
  addcovar = NULL,
  Xcovar = NULL,
  intcovar = NULL,
  weights = NULL,
  reml = TRUE,
  model = c("normal", "binary"),
  hsq = NULL,
  cores = 1,
  ...
)
}
\arguments{
\item{genoprobs}{Genotype probabilities as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{pheno}{A numeric matrix of phenotypes, individuals x phenotypes.}

\item{kinship}{Optional kinship matrix, or a list of kinship matrices (one
per chromosome), in order to use the LOCO (leave one chromosome
out) method.}

\item{addcovar}{An optional numeric matrix of additive covariates.}

\item{Xcovar}{An optional numeric matrix with additional additive covariates used for
null hypothesis when scanning the X chromosome.}

\item{intcovar}{An numeric optional matrix of interactive covariates.}

\item{weights}{An optional numeric vector of positive weights for the
individuals. As with the other inputs, it must have \code{names}
for individual identifiers.}

\item{reml}{If \code{kinship} provided: if \code{reml=TRUE}, use
REML; otherwise maximum likelihood.}

\item{model}{Indicates whether to use a normal model (least
squares) or binary model (logistic regression) for the phenotype.
If \code{model="binary"}, the phenotypes must have values in \eqn{[0, 1]}.}

\item{hsq}{Considered only if \code{kinship} is provided, in which case
this is taken as the assumed value for the residual
heritability. It should be a vector with length corresponding
to the number of columns in \code{pheno}, or (if \code{kinship}
corresponds to a list of LOCO kinship matrices) a matrix with dimension
\verb{length(kinship) x ncol(pheno)}.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}

\item{...}{Additional control parameters; see Details.}
}
\value{
An object of class \code{"scan1"}: a matrix of LOD scores, positions x phenotypes.
Also contains one or more of the following attributes:
\itemize{
\item \code{sample_size} - Vector of sample sizes used for each
phenotype
\item \code{hsq} - Included if \code{kinship} provided: A matrix of
estimated heritabilities under the null hypothesis of no
QTL. Columns are the phenotypes. If the \code{"loco"} method was
used with \code{\link[=calc_kinship]{calc_kinship()}} to calculate a list
of kinship matrices, one per chromosome, the rows of \code{hsq}
will be the heritabilities for the different chromosomes (well,
leaving out each one). If \code{Xcovar} was not NULL, there will at
least be an autosome and X chromosome row.
}
}
\description{
Genome scan with a single-QTL model by Haley-Knott regression or a
linear mixed model, with possible allowance for covariates.
}
\details{
We first fit the model \eqn{y = X \beta + \epsilon}{y = Xb + e}
where \eqn{X} is a matrix of covariates (or just an intercept) and
\eqn{\epsilon}{e} is multivariate normal with mean 0 and covariance
matrix \eqn{\sigma^2 [h^2 (2 K) + (1-h^2) I]}{sigmasq*[hsq*2*K+(1-hsq)*I]} where
\eqn{K} is the kinship matrix and \eqn{I} is the identity matrix.

We then take \eqn{h^2}{hsq} as fixed and then scan the genome, at
each genomic position fitting the model \eqn{y = P \alpha + X \beta
+ \epsilon}{y = Xb + e} where \eqn{P} is a matrix of genotype
probabilities for the current position and again \eqn{X} is a
matrix of covariates \eqn{\epsilon}{e} is multivariate normal with
mean 0 and covariance matrix \eqn{\sigma^2 [h^2 (2 K) +
(1-h^2) I]}{sigmasq*[hsq*2*K+(1-hsq)*I]}, taking \eqn{h^2}{hsq} to be known.

Note that if \code{weights} are provided, the covariance matrix becomes
\eqn{\sigma^2 [h^2 (2 K) + (1-h^2) D]}{sigmasq*[hsq*2*K+(1-hsq)*D]} where
\eqn{D} is a diagonal matrix with the reciprocal of the weights.

For each of the inputs, the row names are used as
individual identifiers, to align individuals. The \code{genoprobs}
object should have a component \code{"is_x_chr"} that indicates
which of the chromosomes is the X chromosome, if any.

The \code{...} argument can contain several additional control
parameters; suspended for simplicity (or confusion, depending on
your point of view). \code{tol} is used as a tolerance value for linear
regression by QR decomposition (in determining whether columns are
linearly dependent on others and should be omitted); default
\code{1e-12}. \code{intcovar_method} indicates whether to use a high-memory
(but potentially faster) method or a low-memory (and possibly
slower) method, with values \code{"highmem"} or \code{"lowmem"}; default
\code{"lowmem"}. \code{max_batch} indicates the maximum number of phenotypes
to run together; default is unlimited. \code{maxit} is the maximum
number of iterations for converence of the iterative algorithm
used when \code{model=binary}. \code{bintol} is used as a tolerance for
converence for the iterative algorithm used when \code{model=binary}.
\code{eta_max} is the maximum value for the "linear predictor" in the
case \code{model="binary"} (a bit of a technicality to avoid fitted
values exactly at 0 or 1).

If \code{kinship} is absent, Haley-Knott regression is performed.
If \code{kinship} is provided, a linear mixed model is used, with a
polygenic effect estimated under the null hypothesis of no (major)
QTL, and then taken as fixed as known in the genome scan.

If \code{kinship} is a single matrix, then the \code{hsq}
in the results is a vector of heritabilities (one value for each phenotype). If
\code{kinship} is a list (one matrix per chromosome), then
\code{hsq} is a matrix, chromosomes x phenotypes.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c("19", "X")] # subset to chr 19 and X}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# leave-one-chromosome-out kinship matrices
kinship <- calc_kinship(probs, "loco")

# genome scan with a linear mixed model
out_lmm <- scan1(probs, pheno, kinship, covar, Xcovar)

}
\references{
Haley CS, Knott SA (1992) A simple
regression method for mapping quantitative trait loci in line
crosses using flanking markers.  Heredity 69:315--324.

Kang HM, Zaitlen NA, Wade CM, Kirby A, Heckerman D, Daly MJ, Eskin
E (2008) Efficient control of population structure in model
organism association mapping. Genetics 178:1709--1723.
}
\seealso{
\code{\link[=scan1perm]{scan1perm()}}, \code{\link[=scan1coef]{scan1coef()}}, \code{\link[=cbind.scan1]{cbind.scan1()}}, \code{\link[=rbind.scan1]{rbind.scan1()}}, \code{\link[=scan1max]{scan1max()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/batch_cols.R
\name{batch_cols}
\alias{batch_cols}
\title{Batch columns by pattern of missing values}
\usage{
batch_cols(mat, max_batch = NULL)
}
\arguments{
\item{mat}{A numeric matrix}

\item{max_batch}{Maximum batch size}
}
\value{
A list containing the batches, each with two components:
\code{cols} containing numeric indices of the columns in the
corresponding batch, and \code{omit} containing a vector of row indices
that have missing values in this batch.
}
\description{
Identify batches of columns of a matrix that have the same pattern
of missing values.
}
\examples{
x <- rbind(c( 1,  2,  3, 13, 16),
           c( 4,  5,  6, 14, 17),
           c( 7, NA,  8, NA, 18),
           c(NA, NA, NA, NA, 19),
           c(10, 11, 12, 15, 20))
batch_cols(x)
}
\seealso{
\code{\link[=batch_vec]{batch_vec()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_geno.R
\name{max_compare_geno}
\alias{max_compare_geno}
\alias{max.compare_geno}
\title{Find pair with most similar genotypes}
\usage{
max_compare_geno(object, ...)

\method{max}{compare_geno}(object, ...)
}
\arguments{
\item{object}{A square matrix with genotype comparisons for pairs
of individuals, as output by \code{\link[=compare_geno]{compare_geno()}}.}

\item{...}{Ignored}
}
\value{
Data frame with individual pair, proportion matches, number
of mismatches, number of matches, and total markers genotyped.
}
\description{
From results of \code{\link[=compare_geno]{compare_geno()}}, show the pair with most similar genotypes.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
cg <- compare_geno(grav2)
max(cg)
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/max_scan1.R
\name{max_scan1}
\alias{max_scan1}
\alias{max.scan1}
\title{Find position with maximum LOD score}
\usage{
max_scan1(
  scan1_output,
  map = NULL,
  lodcolumn = 1,
  chr = NULL,
  na.rm = TRUE,
  ...
)

\method{max}{scan1}(scan1_output, map = NULL, lodcolumn = 1, chr = NULL, na.rm = TRUE, ...)
}
\arguments{
\item{scan1_output}{An object of class \code{"scan1"} as returned by
\code{\link[=scan1]{scan1()}}.}

\item{map}{A list of vectors of marker positions, as produced by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}. Can also be an indexed SNP info table,
as from \code{\link[=index_snps]{index_snps()}} or \code{\link[=scan1snps]{scan1snps()}}.}

\item{lodcolumn}{An integer or character string indicating the LOD
score column, either as a numeric index or column name.
If \code{NULL}, return maximum for all columns.}

\item{chr}{Optional vector of chromosomes to consider.}

\item{na.rm}{Ignored (take to be TRUE)}

\item{...}{Ignored}
}
\value{
If \code{map} is NULL, the genome-wide maximum LOD score for the selected column is returned.
If also \code{lodcolumn} is NULL, you get a vector with the maximum LOD for each column.

If \code{map} is provided, the return value is a data.frame with three columns: chr, pos, and lod score.
But if \code{lodcolumn} is NULL, you get the maximum for each lod score column, in the format provided by
\code{\link[=find_peaks]{find_peaks()}}, so a data.frame with five columns: lodindex, lodcolumn, chr, pos, and lod.
}
\description{
Return data frame with the positions having maximum LOD score for a
particular LOD score column
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# maximum of first column
max(out, map)

# maximum of spleen column
max(out, map, lodcolumn="spleen")

# maximum of first column on chr 2
max(out, map, chr="2")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_grid.R
\name{calc_grid}
\alias{calc_grid}
\title{Calculate indicators of which marker/pseudomarker positions are along a fixed grid}
\usage{
calc_grid(map, step = 0, off_end = 0, tol = 0.01)
}
\arguments{
\item{map}{A list of numeric vectors; each vector gives marker
positions for a single chromosome.}

\item{step}{Distance between pseudomarkers and markers; if
\code{step=0} no pseudomarkers are inserted.}

\item{off_end}{Distance beyond terminal markers in which to insert
pseudomarkers.}

\item{tol}{Tolerance for determining whether a pseudomarker would
duplicate a marker position.}
}
\value{
A list of logical (TRUE/FALSE) vectors that indicate, for a
marker/pseudomarker map created by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}} with \code{step}>0 and
\code{stepwidth="fixed"}, which positions correspond to he
locations along the fixed grid.
}
\description{
Construct vectors of logical indicators that indicate which
positions correspond to locations along a grid
}
\details{
The function \code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}, with
\code{stepwidth="fixed"}, will insert a grid of pseudomarkers,
to a marker map. The present function gives a series of
TRUE/FALSE vectors that indicate which positions fall on the
grid. This is for use with \code{\link[=probs_to_grid]{probs_to_grid()}}, for
reducing genotype probabilities, calculated with
\code{\link[=calc_genoprob]{calc_genoprob()}}, to just the positions on the grid.
The main value of this is to speed up genome scan computations
in the case of very dense markers, by focusing on just a grid
of positions rather than on all marker locations.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
gmap_w_pmar <- insert_pseudomarkers(iron$gmap, step=1)
grid <- calc_grid(iron$gmap, step=1)
}
\seealso{
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}, \code{\link[=probs_to_grid]{probs_to_grid()}},
\code{\link[=map_to_grid]{map_to_grid()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_genoprob.R
\name{plot_genoprob}
\alias{plot_genoprob}
\alias{plot.calc_genoprob}
\title{Plot genotype probabilities for one individual on one chromosome.}
\usage{
plot_genoprob(
  probs,
  map,
  ind = 1,
  chr = NULL,
  geno = NULL,
  color_scheme = c("gray", "viridis"),
  col = NULL,
  threshold = 0,
  swap_axes = FALSE,
  ...
)

\method{plot}{calc_genoprob}(x, ...)
}
\arguments{
\item{probs}{Genotype probabilities (as produced by \code{\link[=calc_genoprob]{calc_genoprob()}})
or allele dosages (as produced by \code{\link[=genoprob_to_alleleprob]{genoprob_to_alleleprob()}}).}

\item{map}{Marker map (a list of vectors of marker positions).}

\item{ind}{Individual to plot, either a numeric index or an ID.}

\item{chr}{Selected chromosome to plot; a single character string.}

\item{geno}{Optional vector of genotypes or alleles to be shown
(vector of integers or character strings)}

\item{color_scheme}{Color scheme for the heatmap (ignored if \code{col} is provided).}

\item{col}{Optional vector of colors for the heatmap.}

\item{threshold}{Threshold for genotype probabilities; only genotypes that achieve
this value somewhere on the chromosome will be shown.}

\item{swap_axes}{If TRUE, swap the axes, so that the genotypes are
on the x-axis and the chromosome position is on the y-axis.}

\item{...}{Additional graphics parameters passed to \code{\link[graphics:image]{graphics::image()}}.}

\item{x}{Genotype probabilities (as produced by
\code{\link[=calc_genoprob]{calc_genoprob()}}) or allele dosages (as produced by
\code{\link[=genoprob_to_alleleprob]{genoprob_to_alleleprob()}}). (For the S3 type plot
function, this has to be called \code{x}.)}
}
\value{
None.
}
\description{
Plot the genotype probabilities for one individual on one chromosome, as a heat map.
}
\section{Hidden graphics parameters}{

A number of graphics parameters can be passed via \code{...}. For
example, \code{hlines}, \code{hlines_col}, \code{hlines_lwd}, and \code{hlines_lty} to
control the horizontal grid lines. (Use \code{hlines=NA} to avoid
plotting horizontal grid lines.) Similarly \code{vlines}, \code{vlines_col},
\code{vlines_lwd}, and \code{vlines_lty} for vertical grid lines. You can
also use many standard graphics parameters like \code{xlab} and \code{xlim}.
These are not included as formal parameters in order to avoid
cluttering the function definition.
}

\examples{
# load data and calculate genotype probabilities
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
iron <- iron[,"2"] # subset to chr 2
map <- insert_pseudomarkers(iron$gmap, step=1)
pr <- calc_genoprob(iron, map, error_prob=0.002)

# plot the probabilities for the individual labeled "262"
#  (white = 0, black = 1)
plot_genoprob(pr, map, ind="262")

# change the x-axis label
plot_genoprob(pr, map, ind="262", xlab="Position (cM)")

# swap the axes so that the chromosome runs vertically
plot_genoprob(pr, map, ind="262", swap_axes=TRUE, ylab="Position (cM)")

# This is more interesting for a Diversity Outbred mouse example
\dontrun{
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
# subset to chr 2 and X and individuals labeled "232" and "256"
DOex <- DOex[c("232", "256"), c("2", "X")]
pr <- calc_genoprob(DOex, error_prob=0.002)
# plot individual "256" on chr 2 (default is to pick first chr in the probs)
plot_genoprob(pr, DOex$pmap, ind="256")

# omit states that never have probability >= 0.5
plot_genoprob(pr, DOex$pmap, ind="256", threshold=0.05)

# X chr male 232: just show the AY-HY genotype probabilities
plot_genoprob(pr, DOex$pmap, ind="232", chr="X", geno=paste0(LETTERS[1:8], "Y"))
# could also indicate genotypes by number
plot_genoprob(pr, DOex$pmap, ind="232", chr="X", geno=37:44)
# and can use negative indexes
plot_genoprob(pr, DOex$pmap, ind="232", chr="X", geno=-(1:36))

# X chr female 256: just show the first 36 genotype probabilities
plot_genoprob(pr, DOex$pmap, ind="256", chr="X", geno=1:36)

# again, can give threshold to omit genotypes whose probabilities never reach that threshold
plot_genoprob(pr, DOex$pmap, ind="256", chr="X", geno=1:36, threshold=0.5)

# can also look at the allele dosages
apr <- genoprob_to_alleleprob(pr)
plot_genoprob(apr, DOex$pmap, ind="232")
}

}
\seealso{
\code{\link[=plot_genoprobcomp]{plot_genoprobcomp()}}, \code{\link[=plot_onegeno]{plot_onegeno()}}, \code{\link[=plot_geno]{plot_geno()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_scan1_heatmap.R
\name{plot_scan1_heatmap}
\alias{plot_scan1_heatmap}
\title{Heat map of scan1 results with multiple traits}
\usage{
plot_scan1_heatmap(
  x,
  map,
  chr = NULL,
  y = NULL,
  gap = NULL,
  zlim = NULL,
  color_scheme = c("viridis", "gray", "revgray", "heat", "terrain", "topo", "rainbow"),
  col = NULL,
  n_colors = 256,
  swap_axes = FALSE,
  chrlines = NULL,
  rescale_by = c("none", "col", "chr", "both"),
  ...
)
}
\arguments{
\item{x}{An object of class \code{"scan1"}, as output by \code{\link[=scan1]{scan1()}}.}

\item{map}{A list of vectors of marker positions, as produced by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}.}

\item{chr}{Selected chromosomes to plot; a vector of character
strings.}

\item{y}{Vector with length equal to \code{ncol(x)}, defining the
y-axis scale of the heat map. (If the multiple LOD scores are
for a quantitative variable like time, these are the
corresponding times.)}

\item{gap}{Gap between chromosomes. The default is 1\% of the total genome length.}

\item{zlim}{Z-axis limits (a pair of numbers)}

\item{color_scheme}{Color scheme for the heatmap (ignored if \code{col} is provided).}

\item{col}{Optional vector of colors for the heatmap.}

\item{n_colors}{Number of z-axis colors (ignored if \code{col} is provided).}

\item{swap_axes}{If TRUE, put chromosome position on y-axis and lod columns on x-axis
(default is to have chromosome position on x-axis and lod columns on y-axis)}

\item{chrlines}{Color of lines at chromosome breaks (NULL for default choices; NA to skip them).}

\item{rescale_by}{Rescale the LOD scores by column, chromosome, or both, by subtracting the
maximum value. The default is not to rescale.}

\item{...}{Additional graphics paramaters.}
}
\value{
None.
}
\description{
Heat map of multiple scan1 results
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
map <- insert_pseudomarkers(grav2$gmap, step=1)
probs <- calc_genoprob(grav2, map, error_prob=0.002)
out <- scan1(probs, grav2$pheno)
plot_scan1_heatmap(out, map)

# make the y-axis quantitative
y <- grav2$phenocovar$time
plot_scan1_heatmap(out, map, y=y, ylab=colnames(grav2$phenocovar))

# plot with colors white -> darkslateblue, and with color scale
col <- colorRampPalette(c("white", "darkslateblue"))(256)
zlim <- c(0, maxlod(out))
layout(cbind(1,2), width=c(3,1))
plot_scan1_heatmap(out, map, y=y, ylab=colnames(grav2$phenocovar),
                   col=col, zlim=zlim)
plot_colorscale(col=col, zlim=zlim)
}
\seealso{
\code{\link[=plot_scan1]{plot_scan1()}}, \code{\link[=plot_lodpeaks]{plot_lodpeaks()}}, \code{\link[graphics:image]{graphics::image()}}; \code{\link[=plot_colorscale]{plot_colorscale()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drop_markers.R
\name{pull_markers}
\alias{pull_markers}
\title{Drop all but a specified set of markers}
\usage{
pull_markers(cross, markers)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{markers}{A vector of marker names.}
}
\value{
The input \code{cross} with only the specified markers.
}
\description{
Drop all markers from a cross2 object expect those in a specified vector.
}
\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
markers2drop <- c("BH.342C/347L-Col", "GH.94L", "EG.357C/359L-Col", "CD.245L", "ANL2")
grav2_rev <- pull_markers(grav2, markers2drop)
}
\seealso{
\code{\link[=drop_markers]{drop_markers()}}, \code{\link[=drop_nullmarkers]{drop_nullmarkers()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/compare_maps.R
\name{compare_maps}
\alias{compare_maps}
\title{Compare two marker maps}
\usage{
compare_maps(map1, map2)
}
\arguments{
\item{map1}{A list of numeric vectors; each vector gives marker
positions for a single chromosome.}

\item{map2}{A second map, in the same format as \code{map1}.}
}
\value{
A data frame containing
\itemize{
\item \code{marker} - marker name
\item \code{chr_map1} - chromosome ID on \code{map1}
\item \code{pos_map1} - position on \code{map1}
\item \code{chr_map2} - chromosome ID on \code{map2}
\item \code{pos_map2} - position on \code{map2}
}
}
\description{
Compare two marker maps, identifying markers that are only in one
of the two maps, or that are in different orders on the two maps.
}
\examples{
# load some data
iron <- read_cross2( system.file("extdata", "iron.zip", package="qtl2") )
gmap <- iron$gmap
pmap <- iron$pmap

# omit a marker from each map
gmap[[7]] <- gmap[[7]][-3]
pmap[[8]] <- pmap[[8]][-7]
# swap order of a couple of markers on the physical map
names(pmap[[9]])[3:4] <- names(pmap[[9]])[4:3]
# move a marker to a different chromosome
pmap[[10]] <- c(pmap[[10]], pmap[[1]][2])[c(1,3,2)]
pmap[[1]] <- pmap[[1]][-2]

# compare these messed-up maps
compare_maps(gmap, pmap)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_peaks.R
\name{plot_peaks}
\alias{plot_peaks}
\title{Plot QTL peak locations}
\usage{
plot_peaks(
  peaks,
  map,
  chr = NULL,
  tick_height = 0.3,
  gap = NULL,
  lod_labels = FALSE,
  ...
)
}
\arguments{
\item{peaks}{Data frame such as that produced by
\code{\link[=find_peaks]{find_peaks()}}) containing columns
\code{chr}, \code{pos}, \code{lodindex}, and \code{lodcolumn}.
May also contain columns \code{ci_lo} and \code{ci_hi}, in
which case intervals will be plotted.}

\item{map}{Marker map, used to get chromosome lengths (and start
and end positions).}

\item{chr}{Selected chromosomes to plot; a vector of character
strings.}

\item{tick_height}{Height of tick marks at the peaks (a number between 0 and 1).}

\item{gap}{Gap between chromosomes. The default is 1\% of the total genome length.}

\item{lod_labels}{If TRUE, plot LOD scores near the intervals. Uses
three hidden graphics parameters, \code{label_gap} (distance between
CI and LOD text label), \code{label_left} (vector that indicates
whether the labels should go on the left side; TRUE=on left,
FALSE=on right, NA=put into larger gap on that chromosome), and
\code{label_cex} that controls the size of these labels}

\item{...}{Additional graphics parameters}
}
\value{
None.
}
\description{
Plot QTL peak locations (possibly with intervals) for multiple traits.
}
\section{Hidden graphics parameters}{

A number of graphics parameters can be passed via \code{...}. For
example, \code{bgcolor} to control the background color and
\code{altbgcolor} to control the background color on alternate chromosomes.
These are not included as formal parameters in order to avoid
cluttering the function definition.
}

\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c(1,2,7,8,9,13,15,16,19)]}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# find peaks above lod=3.5 (and calculate 1.5-LOD support intervals)
peaks <- find_peaks(out, map, threshold=3.5, drop=1.5)

plot_peaks(peaks, map)

# show LOD scores
plot_peaks(peaks, map, lod_labels=TRUE)

# show LOD scores, controlling whether they go on the left or right
plot_peaks(peaks, map, lod_labels=TRUE,
           label_left=c(TRUE, FALSE, TRUE, FALSE, TRUE, TRUE))
}
\seealso{
\code{\link[=find_peaks]{find_peaks()}}, \code{\link[=plot_lodpeaks]{plot_lodpeaks()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_snpasso.R
\name{plot_snpasso}
\alias{plot_snpasso}
\title{Plot SNP associations}
\usage{
plot_snpasso(
  scan1output,
  snpinfo,
  genes = NULL,
  lodcolumn = 1,
  show_all_snps = TRUE,
  chr = NULL,
  add = FALSE,
  drop_hilit = NA,
  col_hilit = "violetred",
  col = "darkslateblue",
  gap = NULL,
  minlod = 0,
  sdp_panel = FALSE,
  strain_labels = names(qtl2::CCcolors),
  ...
)
}
\arguments{
\item{scan1output}{Output of \code{\link[=scan1]{scan1()}} using
SNP probabilities derived by
\code{\link[=genoprob_to_snpprob]{genoprob_to_snpprob()}}.}

\item{snpinfo}{Data frame with SNP information with the following
columns (the last three are generally derived from with
\code{\link[=index_snps]{index_snps()}}):
\itemize{
\item \code{chr} - Character string or factor with chromosome
\item \code{pos} - Position (in same units as in the \code{"map"}
attribute in \code{genoprobs}.
\item \code{sdp} - Strain distribution pattern: an integer, between
1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
binary encoding indicates the founder genotypes
\item \code{snp} - Character string with SNP identifier (if
missing, the rownames are used).
\item \code{index} - Indices that indicate equivalent
groups of SNPs.
\item \code{intervals} - Indexes that indicate which marker
intervals the SNPs reside.
\item \code{on_map} - Indicate whether SNP coincides with a marker
in the \code{genoprobs}
}}

\item{genes}{Optional data frame containing gene information for
the region, with columns \code{start} and \code{stop} in Mbp, \code{strand}
(as \code{"-"}, \code{"+"}, or \code{NA}), and \code{Name}. If included, a
two-panel plot is produced, with SNP associations above and
gene locations below.}

\item{lodcolumn}{LOD score column to plot (a numeric index, or a
character string for a column name). Only one value allowed.}

\item{show_all_snps}{If TRUE, expand to show all SNPs.}

\item{chr}{Vector of character strings with chromosome IDs to plot.}

\item{add}{If TRUE, add to current plot (must have same map and
chromosomes).}

\item{drop_hilit}{SNPs with LOD score within this amount of the maximum SNP association will be highlighted.}

\item{col_hilit}{Color of highlighted points}

\item{col}{Color of other points}

\item{gap}{Gap between chromosomes. The default is 1\% of the total genome length.}

\item{minlod}{Minimum LOD to display. (Mostly for GWAS, in which
case using \code{minlod=1} will greatly increase the plotting speed,
since the vast majority of points would be omittted.}

\item{sdp_panel}{Include a panel with the strain distribution patterns for the highlighted SNPs}

\item{strain_labels}{Labels for the strains, if \code{sdp_panel=TRUE}.}

\item{...}{Additional graphics parameters.}
}
\value{
None.
}
\description{
Plot SNP associations, with possible expansion from distinct snps to all snps.
}
\section{Hidden graphics parameters}{

A number of graphics parameters can be passed via \code{...}. For
example, \code{bgcolor} to control the background color,\code{altbgcolor}
to control the background color on alternate chromosomes,
\code{altcol} to control the point color on alternate chromosomes,
\code{cex} for character expansion for the points (default 0.5),
\code{pch} for the plotting character for the points (default 16),
and \code{ylim} for y-axis limits.
If you are including genes and/or SDP panels, you can use
\code{panel_prop} to control the relative heights of the panels,
from top to bottom.
}

\examples{
\dontrun{
# load example DO data from web
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)

# subset to chr 2
DOex <- DOex[,"2"]

# calculate genotype probabilities and convert to allele probabilities
pr <- calc_genoprob(DOex, error_prob=0.002)
apr <- genoprob_to_alleleprob(pr)

# query function for grabbing info about variants in region
snp_dbfile <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
query_variants <- create_variant_query_func(snp_dbfile)

# SNP association scan
out_snps <- scan1snps(apr, DOex$pmap, DOex$pheno, query_func=query_variants,
                      chr=2, start=97, end=98, keep_all_snps=TRUE)

# plot results
plot_snpasso(out_snps$lod, out_snps$snpinfo)

# can also just type plot()
plot(out_snps$lod, out_snps$snpinfo)

# plot just subset of distinct SNPs
plot(out_snps$lod, out_snps$snpinfo, show_all_snps=FALSE)

# highlight the top snps (with LOD within 1.5 of max)
plot(out_snps$lod, out_snps$snpinfo, drop_hilit=1.5)

# query function for finding genes in region
gene_dbfile <- system.file("extdata", "mouse_genes_small.sqlite", package="qtl2")
query_genes <- create_gene_query_func(gene_dbfile)
genes <- query_genes(2, 97, 98)

# plot SNP association results with gene locations
plot(out_snps$lod, out_snps$snpinfo, drop_hilit=1.5, genes=genes)

# plot SNP asso results with genes plus SDPs of highlighted SNPs
plot(out_snps$lod, out_snps$snpinfo, drop_hilit=2, genes=genes, sdp_panel=TRUE)
}

}
\seealso{
\code{\link[=plot_scan1]{plot_scan1()}}, \code{\link[=plot_coef]{plot_coef()}}, \code{\link[=plot_coefCC]{plot_coefCC()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_gene_query_func.R
\name{create_gene_query_func}
\alias{create_gene_query_func}
\title{Create a function to query genes}
\usage{
create_gene_query_func(
  dbfile = NULL,
  db = NULL,
  table_name = "genes",
  chr_field = "chr",
  start_field = "start",
  stop_field = "stop",
  name_field = "Name",
  strand_field = "strand",
  filter = NULL
)
}
\arguments{
\item{dbfile}{Name of database file}

\item{db}{Optional database connection (provide one of \code{file} and \code{db}).}

\item{table_name}{Name of table in the database}

\item{chr_field}{Name of chromosome field}

\item{start_field}{Name of field with start position (in basepairs)}

\item{stop_field}{Name of field with stop position (in basepairs)}

\item{name_field}{Name of field with gene name}

\item{strand_field}{Name of field with strand (+/-)}

\item{filter}{Additional SQL filter (as a character string).}
}
\value{
Function with three arguments, \code{chr}, \code{start},
and \code{end}, which returns a data frame with the genes
overlapping that region, with \code{start} and \code{end} being
in Mbp. The output should contain at least the columns
\code{Name}, \code{chr}, \code{start}, and \code{stop}, the
latter two being positions in Mbp.
}
\description{
Create a function that will connect to a SQLite database of gene
information and return a data frame with gene information for a
selected region.
}
\details{
Note that this function assumes that the database has
\code{start} and \code{stop} fields that are in basepairs, but
the selection uses positions in Mbp, and the output data frame
should have \code{start} and \code{stop} columns in Mbp.

Also note that a SQLite database of MGI mouse genes
is available at figshare:
\doi{10.6084/m9.figshare.5286019.v7}
}
\examples{
# create query function by connecting to file
dbfile <- system.file("extdata", "mouse_genes_small.sqlite", package="qtl2")
query_genes <- create_gene_query_func(dbfile, filter="(source=='MGI')")
# query_genes will connect and disconnect each time
genes <- query_genes("2", 97.0, 98.0)

# connect and disconnect separately
library(RSQLite)
db <- dbConnect(SQLite(), dbfile)
query_genes <- create_gene_query_func(db=db, filter="(source=='MGI')")
genes <- query_genes("2", 97.0, 98.0)
dbDisconnect(db)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary_scan1perm.R
\name{summary_scan1perm}
\alias{summary_scan1perm}
\alias{summary.scan1perm}
\title{Summarize scan1perm results}
\usage{
summary_scan1perm(object, alpha = 0.05)

\method{summary}{scan1perm}(object, alpha = 0.05, ...)
}
\arguments{
\item{object}{An object of class \code{"scanoneperm"}, as output by \code{\link[=scan1perm]{scan1perm()}}}

\item{alpha}{Vector of significance levels}

\item{...}{Ignored}
}
\value{
An object of class \code{summary.scan1perm}. If
\code{\link[=scan1perm]{scan1perm()}} was run with \code{perm_Xsp=FALSE}, this is
a single matrix of significance thresholds, with rows being
signicance levels and columns being the columns in the input. If
\code{\link[=scan1perm]{scan1perm()}} was run with \code{perm_Xsp=TRUE}, this is
a list of two matrices, with the significance thresholds for the
autosomes and X chromosome, respectively.

The result has an attribute \code{"n_perm"} that has the numbers of
permutation replicates (either a matrix or a list of two matrices).
}
\description{
Summarize permutation test results from \code{\link[=scan1perm]{scan1perm()}}, as significance thresholds.
}
\details{
In the case of X-chromosome-specific permutations (when
\code{\link[=scan1perm]{scan1perm()}} was run with \code{perm_Xsp=TRUE}, we
follow the approach of Broman et al. (2006) to get separate
thresholds for the autosomes and X chromosome, using

Let \eqn{L_A} and \eqn{L_X} be total the genetic lengths of the
autosomes and X chromosome, respectively, and let \eqn{L_T = L_A +
L_X} Then in place of \eqn{\alpha}{alpha}, we use \deqn{\alpha_A =
1 - (1-\alpha)^{L_A/L_T}}{alpha_A = 1 - (1 - alpha)^(L_A/L_T)} as
the significance level for the autosomes and \deqn{\alpha_X = 1 -
(1-\alpha)^{L_X/L_T}}{alpha_x = 1 - (1 - alpha)^(LX/LT)} as the
significance level for the X chromosome.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c(10,18,"X")]}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# permutations with genome scan (just 3 replicates, for illustration)
operm <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar,
                   n_perm=3)

summary(operm, alpha=c(0.20, 0.05))

}
\references{
Broman KW, Sen Ś, Owens SE, Manichaikul A, Southard-Smith EM,
Churchill GA (2006) The X chromosome in quantitative trait locus
mapping. Genetics 174:2151-2158
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_pxg.R
\name{plot_pxg}
\alias{plot_pxg}
\title{Plot phenotype vs genotype}
\usage{
plot_pxg(
  geno,
  pheno,
  sort = TRUE,
  SEmult = NULL,
  pooledSD = TRUE,
  swap_axes = FALSE,
  jitter = 0.2,
  force_labels = TRUE,
  alternate_labels = FALSE,
  omit_points = FALSE,
  ...
)
}
\arguments{
\item{geno}{Vector of genotypes, for example as produced by
\code{\link[=maxmarg]{maxmarg()}} with specific \code{chr} and \code{pos}.}

\item{pheno}{Vector of phenotypes.}

\item{sort}{If TRUE, sort genotypes from largest to smallest.}

\item{SEmult}{If specified, interval estimates of the within-group
averages will be displayed, as \verb{mean +/- SE * SEmult}.}

\item{pooledSD}{If TRUE and \code{SEmult} is specified, calculated
a pooled within-group SD. Otherwise, get separate estimates of
the within-group SD for each group.}

\item{swap_axes}{If TRUE, swap the axes, so that the genotypes are
on the y-axis and the phenotype is on the x-axis.}

\item{jitter}{Amount to jitter the points horizontally, if a vector
of length > 0, it is taken to be the actual jitter amounts
(with values between -0.5 and 0.5).}

\item{force_labels}{If TRUE, force all genotype labels to be shown.}

\item{alternate_labels}{If TRUE, place genotype labels in two rows}

\item{omit_points}{If TRUE, omit the points, just plotting the averages (and, potentially, the +/- SE intervals).}

\item{...}{Additional graphics parameters, passed to \code{plot()}.}
}
\value{
(Invisibly) A matrix with rows being the genotype groups
and columns for the means and (if \code{SEmult} is specified) the SEs.
}
\description{
Plot phenotype vs genotype for a single putative QTL and a single phenotype.
}
\section{Hidden graphics parameters}{

A number of graphics parameters can be passed via \code{...}. For
example, \code{bgcolor} to control the background color, and
\code{seg_width}, \code{seg_lwd}, and \code{seg_col} to control the lines at the
confidence intervals. Further, \code{hlines}, \code{hlines_col},
\code{hlines_lwd}, and \code{hlines_lty} to control the horizontal grid
lines. (Use \code{hlines=NA} to avoid plotting horizontal grid lines.)
Similarly \code{vlines}, \code{vlines_col}, \code{vlines_lwd}, and \code{vlines_lty}
for vertical grid lines. These are not included as formal
parameters in order to avoid cluttering the function definition.
}

\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,"16"]}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# inferred genotype at a 28.6 cM on chr 16
geno <- maxmarg(probs, map, chr=16, pos=28.6, return_char=TRUE)

# plot phenotype vs genotype
plot_pxg(geno, log10(iron$pheno[,1]), ylab=expression(log[10](Liver)))

# include +/- 2 SE intervals
plot_pxg(geno, log10(iron$pheno[,1]), ylab=expression(log[10](Liver)),
         SEmult=2)

# plot just the means
plot_pxg(geno, log10(iron$pheno[,1]), ylab=expression(log[10](Liver)),
         omit_points=TRUE)

# plot just the means +/- 2 SEs
plot_pxg(geno, log10(iron$pheno[,1]), ylab=expression(log[10](Liver)),
         omit_points=TRUE, SEmult=2)
}
\seealso{
\code{\link[=plot_coef]{plot_coef()}}, \code{\link[=plot_ci]{plot_ci()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rcbind_scan1perm.R
\name{cbind.scan1perm}
\alias{cbind.scan1perm}
\title{Combine columns from multiple scan1 permutation results}
\usage{
\method{cbind}{scan1perm}(...)
}
\arguments{
\item{...}{A set of permutation results from
\code{\link[=scan1perm]{scan1perm()}} (objects of class \code{"scan1perm"}. If
different numbers of permutation replicates were used, those
columns with fewer replicates are padded with missing values
\code{NA}. However, if any include autosome/X
chromosome-specific permutations, they must all be such.}
}
\value{
The combined column-binded input, as an object of class \code{"scan1perm"}; see \code{\link[=scan1perm]{scan1perm()}}.
}
\description{
Column-bind multiple scan1perm objects with the same numbers of rows.
}
\details{
The aim of this function is to concatenate the results
from multiple runs of a permutation test with
\code{\link[=scan1perm]{scan1perm()}}, generally with different phenotypes
and/or methods, to be used in parallel with
\code{\link[=rbind.scan1perm]{rbind.scan1perm()}}.
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c("19","X")] # subset to chr 19 and X}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# permutations with genome scan (just 3 replicates, for illustration)
operm1 <- scan1perm(probs, pheno[,1,drop=FALSE], addcovar=covar, Xcovar=Xcovar, n_perm=3)
operm2 <- scan1perm(probs, pheno[,2,drop=FALSE], addcovar=covar, Xcovar=Xcovar, n_perm=3)

operm <- cbind(operm1, operm2)

}
\seealso{
\code{\link[=rbind.scan1perm]{rbind.scan1perm()}}, \code{\link[=scan1perm]{scan1perm()}}, \code{\link[=scan1]{scan1()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/calc_sdp.R
\name{calc_sdp}
\alias{calc_sdp}
\title{Calculate strain distribution pattern from SNP genotypes}
\usage{
calc_sdp(geno)
}
\arguments{
\item{geno}{Matrix of SNP genotypes, markers x strains, coded as 1
(AA) and 3 (BB). Markers with values other than 1 or 3 are omitted,
and monomorphic markers, are omitted.}
}
\value{
A vector of strain distribution patterns: integers between
1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
binary representation indicates the strain genotypes.
}
\description{
Calculate the strain distribution patterns (SDPs) from the strain
genotypes at a set of SNPs.
}
\examples{
x <- rbind(m1=c(3, 1, 1, 1, 1, 1, 1, 1),
           m2=c(1, 3, 3, 1, 1, 1, 1, 1),
           m3=c(1, 1, 1, 1, 3, 3, 3, 3))
calc_sdp(x)
}
\seealso{
\code{\link[=invert_sdp]{invert_sdp()}}, \code{\link[=sdp2char]{sdp2char()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xpos_scan1.R
\name{xpos_scan1}
\alias{xpos_scan1}
\title{Get x-axis position for genomic location}
\usage{
xpos_scan1(map, chr = NULL, gap = NULL, thechr, thepos)
}
\arguments{
\item{map}{A list of vectors of marker positions, as produced by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}.}

\item{chr}{Selected chromosomes that were plotted (if used in the
call to \code{\link[=plot_scan1]{plot_scan1()}}.}

\item{gap}{The gap between chromosomes used in the call to
\code{\link[=plot_scan1]{plot_scan1()}}.}

\item{thechr}{Vector of chromosome IDs}

\item{thepos}{Vector of chromosomal positions}
}
\value{
A vector of x-axis locations.
}
\description{
For a plot of \code{\link[=scan1]{scan1()}} results, get the x-axis
location that corresponds to a particular genomic location
(chromosome ID and position).
}
\details{
\code{thechr} and \code{thepos} should be the same length,
or should have length 1 (in which case they are expanded to the
length of the other vector).
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# plot the results for selected chromosomes
ylim <- c(0, maxlod(out)*1.02) # need to strip class to get overall max LOD
chr <- c(2,7,8,9,15,16)
plot(out, map, chr=chr, ylim=ylim)
plot(out, map, lodcolumn=2, chr=chr, col="violetred", add=TRUE)
legend("topleft", lwd=2, col=c("darkslateblue", "violetred"), colnames(out),
       bg="gray90")

# Use xpos_scan1 to add points at the peaks
# first find the peaks with LOD > 3
peaks <- find_peaks(out, map)

# keep just the peaks for chromosomes that were plotted
peaks <- peaks[peaks$chr \%in\% chr,]

# find x-axis positions
xpos <- xpos_scan1(map, chr=chr, thechr=peaks$chr, thepos=peaks$pos)

# point colors
ptcolor <- c("darkslateblue", "violetred")[match(peaks$lodcolumn, c("liver", "spleen"))]

# plot points
points(xpos, peaks$lod, pch=21, bg=ptcolor)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chisq_colpairs.R
\name{chisq_colpairs}
\alias{chisq_colpairs}
\title{Chi-square test on all pairs of columns}
\usage{
chisq_colpairs(x)
}
\arguments{
\item{x}{A matrix of positive integers. \code{NA}s and values <= 0 are treated as missing.}
}
\value{
A matrix of size p x p, where p is the number of columns in
the input matrix \code{x}, containing the chi-square test
statistics for independence, applied to pairs of columns of
\code{x}. The diagonal of the result will be all \code{NA}s.
}
\description{
Perform a chi-square test for independence for all pairs of columns of a matrix.
}
\examples{
z <- matrix(sample(1:2, 500, replace=TRUE), ncol=5)
chisq_colpairs(z)
}
\keyword{htest}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reduce_markers.R
\name{reduce_markers}
\alias{reduce_markers}
\title{Reduce markers to a subset of more-evenly-spaced ones}
\usage{
reduce_markers(
  map,
  min_distance = 1,
  weights = NULL,
  max_batch = 10000,
  batch_distance_mult = 1,
  cores = 1
)
}
\arguments{
\item{map}{A list with each component being a vector with the
marker positions for a chromosome.}

\item{min_distance}{Minimum distance between markers.}

\item{weights}{A (optional) list of weights on the markers; same
size as \code{map}.}

\item{max_batch}{Maximum number of markers to consider in a batch}

\item{batch_distance_mult}{If working with batches of markers,
reduce \code{min_distance} by this multiple.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
A list like the input \code{map}, but with the selected
subset of markers.
}
\description{
Find the largest subset of markers such that no two adjacent
markers are separated by less than some distance.
}
\details{
Uses a dynamic programming algorithm to find, for each
chromosome, the subset of markers for with max(\code{weights}) is
maximal, subject to the constraint that no two adjacent markers may
be separated by more than \code{min_distance}.

The computation time for the algorithm grows with like the square
of the number of markers, like 1 sec for 10k markers
but 30 sec for 50k markers. If the number of markers on a chromosome
is greater than \code{max_batch}, the markers are split into batches and
the algorithm applied to each batch with min_distance smaller by a
factor \code{min_distance_mult}, and then merged together for one last pass.
}
\examples{
# read data
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))

# grab genetic map
gmap <- grav2$gmap

# subset to markers that are >= 1 cM apart
gmap_sub <- reduce_markers(gmap, 1)

# drop all of the other markers from the cross
markers2keep <- unlist(lapply(gmap_sub, names))
grav2_sub <- pull_markers(grav2, markers2keep)
}
\references{
Broman KW, Weber JL (1999) Method for constructing
confidently ordered linkage maps. Genet Epidemiol 16:337--343
}
\seealso{
\code{\link[=find_dup_markers]{find_dup_markers()}}, \code{\link[=drop_markers]{drop_markers()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert2cross2.R
\name{convert2cross2}
\alias{convert2cross2}
\title{Convert R/qtl cross object to new format}
\usage{
convert2cross2(cross)
}
\arguments{
\item{cross}{An object of class \code{"cross"}; see
\code{\link[qtl:read.cross]{qtl::read.cross()}} for details.}
}
\value{
Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.
}
\description{
Convert a cross object from the R/qtl format to the R/qtl2 format
}
\examples{
library(qtl)
data(hyper)
hyper2 <- convert2cross2(hyper)
}
\seealso{
\code{\link[=read_cross2]{read_cross2()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/maxmarg.R
\name{maxmarg}
\alias{maxmarg}
\title{Find genotypes with maximum marginal probabilities}
\usage{
maxmarg(
  probs,
  map = NULL,
  minprob = 0.95,
  chr = NULL,
  pos = NULL,
  return_char = FALSE,
  quiet = TRUE,
  cores = 1,
  tol = 0.0000000000001
)
}
\arguments{
\item{probs}{Genotype probabilities, as calculated from
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{map}{Map of pseudomarkers in \code{probs}. Used only if \code{chr} and \code{pos} are provided.}

\item{minprob}{Minimum probability for making a call. If maximum
probability is less then this value, give \code{NA}.}

\item{chr}{If provided (along with \code{pos}), consider only the single specified position.}

\item{pos}{If provided (along with \code{chr}), consider only the single specified position.}

\item{return_char}{If TRUE, return genotype names as character strings.}

\item{quiet}{IF \code{FALSE}, print progress messages.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}

\item{tol}{Tolerance value; genotypes with probability that are within
\code{tol} of each other are treated as equivalent.}
}
\value{
If \code{chr} and \code{pos} are provided, a vector of
genotypes is returned. In this case, \code{map} is needed.

Otherwise, the result is a object like that returned by \code{\link[=viterbi]{viterbi()}},
A list of two-dimensional arrays of imputed genotypes,
individuals x positions. Also includes these attributes:
\itemize{
\item \code{crosstype} - The cross type of the input \code{cross}.
\item \code{is_x_chr} - Logical vector indicating whether chromosomes
are to be treated as the X chromosome or not, from input \code{cross}.
\item \code{alleles} - Vector of allele codes, from input
\code{cross}.
}
}
\description{
For each individual at each position, find the genotype with the maximum marginal probability.
}
\details{
If multiple genotypes share the maximum probability, one is chosen at random.
}
\examples{
# load data and calculate genotype probabilities
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
pr <- calc_genoprob(iron, error_prob=0.002)

# full set of imputed genotypes
ginf <- maxmarg(pr)

# imputed genotypes at a fixed position
g <- maxmarg(pr, iron$gmap, chr=8, pos=45.5)

# return genotype names rather than integers
g <- maxmarg(pr, iron$gmap, chr=8, pos=45.5, return_char=TRUE)
}
\seealso{
\code{\link[=sim_geno]{sim_geno()}}, \code{\link[=viterbi]{viterbi()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_lodpeaks.R
\name{plot_lodpeaks}
\alias{plot_lodpeaks}
\title{Plot LOD scores vs QTL peak locations}
\usage{
plot_lodpeaks(peaks, map, chr = NULL, gap = NULL, intervals = FALSE, ...)
}
\arguments{
\item{peaks}{Data frame such as that produced by
\code{\link[=find_peaks]{find_peaks()}}) containing columns
\code{chr}, \code{pos}, \code{lodindex}, and \code{lodcolumn}.
May also contain columns \code{ci_lo} and \code{ci_hi}, in
which case intervals will be plotted.}

\item{map}{Marker map, used to get chromosome lengths (and start
and end positions).}

\item{chr}{Selected chromosomes to plot; a vector of character
strings.}

\item{gap}{Gap between chromosomes. The default is 1\% of the total genome length.}

\item{intervals}{If TRUE and \code{peaks} contains QTL intervals, plot the intervals.}

\item{...}{Additional graphics parameters}
}
\value{
None.
}
\description{
Create a scatterplot of LOD scores vs QTL peak locations (possibly with intervals) for multiple traits.
}
\section{Hidden graphics parameters}{

A number of graphics parameters can be passed via \code{...}. For
example, \code{bgcolor} to control the background color and
\code{altbgcolor} to control the background color on alternate chromosomes.
These are not included as formal parameters in order to avoid
cluttering the function definition.
}

\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c(1,2,7,8,9,13,15,16,19)]}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# find peaks above lod=3.5 (and calculate 1.5-LOD support intervals)
peaks <- find_peaks(out, map, threshold=3.5, drop=1.5)

plot_lodpeaks(peaks, map)
}
\seealso{
\code{\link[=find_peaks]{find_peaks()}}, \code{\link[=plot_peaks]{plot_peaks()}}, \code{\link[=plot_cistrans]{plot_cistrans()}}, \code{\link[=plot_scan1]{plot_scan1()}}, \code{\link[=plot_scan1_heatmap]{plot_scan1_heatmap()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replace_ids.R
\name{replace_ids}
\alias{replace_ids}
\alias{replace_ids.cross2}
\alias{replace_ids.calc_genoprob}
\alias{replace_ids.viterbi}
\alias{replace_ids.sim_geno}
\alias{replace_ids.matrix}
\alias{replace_ids.data.frame}
\title{Replace individual IDs}
\usage{
replace_ids(x, ids)

\method{replace_ids}{cross2}(x, ids)

\method{replace_ids}{calc_genoprob}(x, ids)

\method{replace_ids}{viterbi}(x, ids)

\method{replace_ids}{sim_geno}(x, ids)

\method{replace_ids}{matrix}(x, ids)

\method{replace_ids}{data.frame}(x, ids)
}
\arguments{
\item{x}{Object whose IDs will be replaced}

\item{ids}{Vector of character strings with the new individual IDs, with the names being the original IDs.}
}
\value{
The input \code{x} object, but with individual IDs replaced.
}
\description{
Replace the individual IDs in an object with new ones
}
\section{Methods (by class)}{
\itemize{
\item \code{replace_ids(cross2)}: Replace IDs in a \code{"cross2"} object

\item \code{replace_ids(calc_genoprob)}: Replace IDs in output from \code{\link[=calc_genoprob]{calc_genoprob()}}

\item \code{replace_ids(viterbi)}: Replace IDs in output from \code{\link[=viterbi]{viterbi()}}

\item \code{replace_ids(sim_geno)}: Replace IDs in output from \code{\link[=sim_geno]{sim_geno()}}

\item \code{replace_ids(matrix)}: Replace IDs in a matrix

\item \code{replace_ids(data.frame)}: Replace IDs in a data frame

}}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
ids <- as.numeric(ind_ids(iron))

# replace the numeric IDs with IDs like "mouse003"
new_ids <- setNames( sprintf("mouse\%03d", as.numeric(ids)), ids)

iron <- replace_ids(iron, new_ids)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset_cross2.R
\name{subset.cross2}
\alias{subset.cross2}
\alias{[.cross2}
\title{Subsetting data for a QTL experiment}
\usage{
\method{subset}{cross2}(x, ind = NULL, chr = NULL, ...)

\method{[}{cross2}(x, ind = NULL, chr = NULL)
}
\arguments{
\item{x}{An object of class \code{"cross2"}, as output by \code{\link[=read_cross2]{read_cross2()}}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{ind}{A vector of individuals: numeric indices, logical
values, or character string IDs.}

\item{chr}{A vector of chromosomes: numeric indices, logical
values, or character string IDs}

\item{...}{Ignored.}
}
\value{
The input \code{cross2} object, with the selected
individuals and/or chromsomes.
}
\description{
Pull out a specified set of individuals and/or chromosomes from a
\code{cross2} object.
}
\details{
When subsetting by individual, if \code{ind} is numeric, they're
assumed to be numeric indices; if character strings, they're
assumed to be individual IDs. \code{ind} can be numeric or logical
only if the genotype, phenotype, and covariate data all have the
same individuals in the same order.

When subsetting by chromosome, \code{chr} is \emph{always}
converted to character strings and treated as chromosome IDs. So if
there are three chromosomes with IDs \code{"18"}, \code{"19"}, and
\code{"X"}, \code{mycross[,18]} will give the first of the
chromosomes (labeled \code{"18"}) and \code{mycross[,3]} will give
an error.

When using character string IDs for \code{ind} or \code{chr}, you
can use "negative" subscripts to indicate exclusions, for example
\code{mycross[,c("-18", "-X")]} or \code{mycross["-Mouse2501",]}.
But you can't mix "positive" and "negative" subscripts, and if any
of the individuals has an ID that begins with \code{"-"}, you can't
use negative subscripts like this.
}
\section{Warning}{

The order of the two arguments is reversed relative to the related
function in \href{https://rqtl.org}{R/qtl}.
}

\examples{
grav2 <- read_cross2(system.file("extdata", "grav2.zip", package="qtl2"))
# keep individuals 1-20 and chromosomes 3 and 4
grav2sub <- grav2[1:20, c(3,4)]
# keep just chromosome 1
grav2_c1 <- grav2[,1]
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recode_snps.R
\name{recode_snps}
\alias{recode_snps}
\title{Recode SNPs by major allele}
\usage{
recode_snps(cross)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}
}
\value{
The input cross object with the raw SNP genotypes recoded so that
\code{1} is homozygous for the major alleles in the founders.
}
\description{
For multi-parent populations with founder genotypes, recode the raw
SNP genotypes so that \code{1} means homozygous for the major allele in the
founders.
}
\examples{
\dontrun{
# load example data and calculate genotype probabilities
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
DOex <- recode_snps(DOex)
}
}
\seealso{
\code{\link[=calc_raw_founder_maf]{calc_raw_founder_maf()}}, \code{\link[=calc_raw_maf]{calc_raw_maf()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/guess_phase.R
\name{guess_phase}
\alias{guess_phase}
\title{Guess phase of imputed genotypes}
\usage{
guess_phase(cross, geno, deterministic = FALSE, cores = 1)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{geno}{Imputed genotypes, as a list of matrices, as from \code{\link[=maxmarg]{maxmarg()}}.}

\item{deterministic}{If TRUE, preferentially put smaller allele first when there's uncertainty.
If FALSE, the order of alleles is random in such cases.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}
}
\value{
If input cross is phase-known (e.g., recombinant inbred lines),
the output will be the input \code{geno}. Otherwise, the output
will be a list of three-dimensional arrays of imputed
genotypes, individual x position x haplotype (1/2), and will have class \code{"phasedgeno"}.
}
\description{
Turn imputed genotypes into phased genotypes along chromosomes by
attempting to pick the phase that leads to the fewest recombination
events.
}
\details{
We randomly assign the pair of alleles at the first locus
to two haplotypes, and then work left to right, assigning
alleles to haplotypes one locus at a time seeking the fewest
recombination events. The results are subject to arbitrary and
random choices. For example, to the right of a homozygous
region, either orientation is equally reasonable.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
gmap <- insert_pseudomarkers(iron$gmap, step=1)
probs <- calc_genoprob(iron, gmap, error_prob=0.002)
imp_geno <- maxmarg(probs)
ph_geno <- guess_phase(iron, imp_geno)
}
\seealso{
\code{\link[=maxmarg]{maxmarg()}}
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pull_genoprobpos.R
\name{pull_genoprobpos}
\alias{pull_genoprobpos}
\title{Pull genotype probabilities for a particular position}
\usage{
pull_genoprobpos(genoprobs, map = NULL, chr = NULL, pos = NULL, marker = NULL)
}
\arguments{
\item{genoprobs}{Genotype probabilities as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{map}{A map object: a list (corresponding to chromosomes) of
vectors of marker positions. Can also be a snpinfo object (data
frame with columns \code{chr} and \code{pos}; marker names taken from
column \code{snp} or if that doesn't exist from the row names)}

\item{chr}{A chromosome ID}

\item{pos}{A numeric position}

\item{marker}{A single character string with the name of the
position to pull out.}
}
\value{
A matrix of genotype probabilities for the specified position.
}
\description{
Pull out the genotype probabilities for a particular position (by
name)
}
\details{
Provide either a marker/pseudomarker name (with the argument \code{marker})
or all of \code{map}, \code{chr}, and \code{pos}.
}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,"8"]}
gmap <- insert_pseudomarkers(iron$gmap, step=1)
pr <- calc_genoprob(iron, gmap, error_prob=0.002)

pmar <- find_marker(gmap, 8, 40)
pr_8_40 <- pull_genoprobpos(pr, pmar)

pr_8_40_alt <- pull_genoprobpos(pr, gmap, 8, 40)

}
\seealso{
\code{\link[=find_marker]{find_marker()}}, \code{\link[=fit1]{fit1()}}, \code{\link[=pull_genoprobint]{pull_genoprobint()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subset_scan1.R
\name{subset_scan1}
\alias{subset_scan1}
\alias{subset.scan1}
\title{Subset scan1 output}
\usage{
subset_scan1(x, map = NULL, chr = NULL, lodcolumn = NULL, ...)

\method{subset}{scan1}(x, map = NULL, chr = NULL, lodcolumn = NULL, ...)
}
\arguments{
\item{x}{An object of class \code{"scan1"} as returned by
\code{\link[=scan1]{scan1()}}.}

\item{map}{A list of vectors of marker positions, as produced by
\code{\link[=insert_pseudomarkers]{insert_pseudomarkers()}}.}

\item{chr}{Vector of chromosomes.}

\item{lodcolumn}{Vector of integers or character strings indicating the LOD
score columns, either as a numeric indexes or column names.}

\item{...}{Ignored}
}
\value{
Object of class \code{"scan1"}, like the input, but subset by chromosome and/or column. See \code{\link[=scan1]{scan1()}}.
}
\description{
Subset the output of \code{\link[=scan1]{scan1()}} by chromosome or column
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# perform genome scan
out <- scan1(probs, pheno, addcovar=covar, Xcovar=Xcovar)

# pull out chromosome 8
out_c8 <- subset(out, map, chr="8")

# just the second column on chromosome 2
out_c2_spleen <- subset(out, map, "2", "spleen")

# all positions, but just the "liver" column
out_spleen <- subset(out, map, lodcolumn="spleen")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scan1perm.R
\name{scan1perm}
\alias{scan1perm}
\title{Permutation test for genome scan with a single-QTL model}
\usage{
scan1perm(
  genoprobs,
  pheno,
  kinship = NULL,
  addcovar = NULL,
  Xcovar = NULL,
  intcovar = NULL,
  weights = NULL,
  reml = TRUE,
  model = c("normal", "binary"),
  n_perm = 1,
  perm_Xsp = FALSE,
  perm_strata = NULL,
  chr_lengths = NULL,
  cores = 1,
  scan_func = NULL,
  ...
)
}
\arguments{
\item{genoprobs}{Genotype probabilities as calculated by
\code{\link[=calc_genoprob]{calc_genoprob()}}.}

\item{pheno}{A numeric matrix of phenotypes, individuals x phenotypes.}

\item{kinship}{Optional kinship matrix, or a list of kinship matrices (one
per chromosome), in order to use the LOCO (leave one chromosome
out) method.}

\item{addcovar}{An optional numeric matrix of additive covariates.}

\item{Xcovar}{An optional numeric matrix with additional additive covariates used for
null hypothesis when scanning the X chromosome.}

\item{intcovar}{An optional numeric matrix of interactive covariates.}

\item{weights}{An optional numeric vector of positive weights for the
individuals. As with the other inputs, it must have \code{names}
for individual identifiers.}

\item{reml}{If \code{kinship} provided: if \code{reml=TRUE}, use
REML; otherwise maximum likelihood.}

\item{model}{Indicates whether to use a normal model (least
squares) or binary model (logistic regression) for the phenotype.
If \code{model="binary"}, the phenotypes must have values in \eqn{[0, 1]}.}

\item{n_perm}{Number of permutation replicates.}

\item{perm_Xsp}{If TRUE, do separate permutations for the autosomes
and the X chromosome.}

\item{perm_strata}{Vector of strata, for a stratified permutation
test. Should be named in the same way as the rows of
\code{pheno}. The unique values define the strata.}

\item{chr_lengths}{Lengths of the chromosomes; needed only if
\code{perm_Xsp=TRUE}. See \code{\link[=chr_lengths]{chr_lengths()}}.}

\item{cores}{Number of CPU cores to use, for parallel calculations.
(If \code{0}, use \code{\link[parallel:detectCores]{parallel::detectCores()}}.)
Alternatively, this can be links to a set of cluster sockets, as
produced by \code{\link[parallel:makeCluster]{parallel::makeCluster()}}.}

\item{scan_func}{If provided, this function is used for the genome scans.
It must take arguments \code{genoprobs}, \code{pheno}, \code{kinship}, \code{addcovar},
\code{Xcovar}, \code{intcovar}, \code{weights}, and possibly further arguments.}

\item{...}{Additional control parameters; see Details.}
}
\value{
If \code{perm_Xsp=FALSE}, the result is matrix of
genome-wide maximum LOD scores, permutation replicates x
phenotypes. If \code{perm_Xsp=TRUE}, the result is a list of
two matrices, one for the autosomes and one for the X
chromosome. The object is given class \code{"scan1perm"}.
}
\description{
Permutation test for a enome scan with a single-QTL model by
Haley-Knott regression or a linear mixed model, with possible
allowance for covariates.
}
\details{
If \code{kinship} is not provided, so that analysis proceeds by
Haley-Knott regression, we permute the rows of the phenotype data;
the same permutations are also applied to the rows of the
covariates (\code{addcovar}, \code{Xcovar}, and \code{intcovar})
are permuted.

If \code{kinship} is provided, we instead permute the rows of the
genotype data and fit an LMM with the same residual heritability
(estimated under the null hypothesis of no QTL).

If \code{Xcovar} is provided and \code{perm_strata=NULL}, we do a
stratified permutation test with the strata defined by the rows of
\code{Xcovar}. If a simple permutation test is desired, provide
\code{perm_strata} that is a vector containing a single repeated
value.

The \code{...} argument can contain several additional control
parameters; suspended for simplicity (or confusion, depending on
your point of view). \code{tol} is used as a tolerance value for linear
regression by QR decomposition (in determining whether columns are
linearly dependent on others and should be omitted); default
\code{1e-12}. \code{maxit} is the maximum number of iterations for
converence of the iterative algorithm used when \code{model=binary}.
\code{bintol} is used as a tolerance for converence for the iterative
algorithm used when \code{model=binary}. \code{eta_max} is the maximum value
for the "linear predictor" in the case \code{model="binary"} (a bit of a
technicality to avoid fitted values exactly at 0 or 1).
}
\examples{
# read data
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
\dontshow{iron <- iron[,c(10,18,"X")]}

# insert pseudomarkers into map
map <- insert_pseudomarkers(iron$gmap, step=1)

# calculate genotype probabilities
probs <- calc_genoprob(iron, map, error_prob=0.002)

# grab phenotypes and covariates; ensure that covariates have names attribute
pheno <- iron$pheno
covar <- match(iron$covar$sex, c("f", "m")) # make numeric
names(covar) <- rownames(iron$covar)
Xcovar <- get_x_covar(iron)

# strata for permutations
perm_strata <- mat2strata(Xcovar)

# permutations with genome scan (just 3 replicates, for illustration)
operm <- scan1perm(probs, pheno, addcovar=covar, Xcovar=Xcovar,
                   n_perm=3, perm_strata=perm_strata)
summary(operm)

# leave-one-chromosome-out kinship matrices
kinship <- calc_kinship(probs, "loco")

# permutations of genome scan with a linear mixed model
operm_lmm <- scan1perm(probs, pheno, kinship, covar, Xcovar, n_perm=3,
                       perm_Xsp=TRUE, perm_strata=perm_strata,
                       chr_lengths=chr_lengths(map))
summary(operm_lmm)


# permutations of scan1gen, fitting glm with probit link
# fitting function
ll_glm <-
   function(pr, pheno, addcovar=NULL, ...)
{
   formula <- ifelse(is.null(pr), "pheno ~ 1", "pheno ~ pr")
   if(!is.null(addcovar)) formula <- paste(formula, "+ addcovar")

    glm_out <- glm(as.formula(formula), family=binomial(link=probit))
    -glm_out$deviance/(2*log(10)) # log10 likelihood
}

# create binary trait
bin_pheno <- setNames(as.numeric(iron$pheno[,1] > median(iron$pheno[,1])),
                      rownames(iron$pheno))

# permutations with glm
operm_glm <- scan1perm(probs, bin_pheno, kinship, covar, Xcovar, n_perm=3,
                       perm_Xsp=TRUE, perm_strata=perm_strata,
                       chr_lengths=chr_lengths(map),
                       scan_func=scan1gen, func=ll_glm)
summary(operm_glm)

\dontrun{
# example of scan1perm with scan1snps
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
DOex <- DOex[,c("2", "3")] # subset to chr 2 and 3
probs <- calc_genoprob(DOex, error_prob=0.002)

snpdb_file <- system.file("extdata", "cc_variants_small.sqlite", package="qtl2")
queryf <- create_variant_query_func(snpdb_file)

operm <- scan1perm(genoprobs=probs, map=DOex$pmap, pheno=DOex$pheno,
                   scan_func=scan1snps, query_func=queryf, n_perm=3)
}

}
\references{
Churchill GA, Doerge RW (1994) Empirical threshold
values for quantitative trait mapping. Genetics 138:963--971.

Manichaikul A, Palmer AA, Sen S, Broman KW (2007) Significance
thresholds for quantitative trait locus mapping under selective
genotyping. Genetics 177:1963--1966.

Haley CS, Knott SA (1992) A simple regression method for mapping
quantitative trait loci in line crosses using flanking markers.
Heredity 69:315--324.

Kang HM, Zaitlen NA, Wade CM, Kirby A, Heckerman D, Daly MJ, Eskin
E (2008) Efficient control of population structure in model
organism association mapping. Genetics 178:1709--1723.
}
\seealso{
\code{\link[=scan1]{scan1()}}, \code{\link[=chr_lengths]{chr_lengths()}}, \code{\link[=mat2strata]{mat2strata()}}, \href{https://kbroman.org/qtl2/assets/vignettes/scan1gen.html}{scan1gen examples}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chr_lengths.R
\name{chr_lengths}
\alias{chr_lengths}
\title{Calculate chromosome lengths}
\usage{
chr_lengths(map, collapse_to_AX = FALSE)
}
\arguments{
\item{map}{A list of vectors, each specifying locations of the markers.
Can also be a cross2 object, in which case the genetic map (if present) is used.}

\item{collapse_to_AX}{If TRUE, collapse to the total lengths of the
autosomes and X chromosome.}
}
\value{
A vector of chromosome lengths. If
\code{collapse_to_AX=TRUE}, the result is a vector of length 2
(autosomal and X chromosome lengths).
}
\description{
Calculate chromosome lengths for a map object
}
\details{
We take \code{diff(range(v))} for each vector, \code{v}.
}
\seealso{
\code{\link[=scan1perm]{scan1perm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zip_datafiles.R
\name{zip_datafiles}
\alias{zip_datafiles}
\title{Zip a set of data files}
\usage{
zip_datafiles(control_file, zip_file = NULL, overwrite = FALSE, quiet = TRUE)
}
\arguments{
\item{control_file}{Character string with path to the control file
(\href{https://yaml.org}{YAML} or \href{https://json.org}{JSON})
containing all of the control information.}

\item{zip_file}{Name of zip file to use. If NULL, we use the
stem of \code{control_file} but with a \code{.zip} extension.}

\item{overwrite}{If \code{TRUE}, overwrite file if it exists. If \code{FALSE}
(the default) and the file exists, stop with an error.}

\item{quiet}{If \code{FALSE}, print progress messages.}
}
\value{
Character string with the file name of the zip file that
was created.
}
\description{
Zip a set of data files (in format read by \code{\link[=read_cross2]{read_cross2()}}).
}
\details{
The input \code{control_file} is the control file (in
\href{https://yaml.org}{YAML} or \href{https://json.org}{JSON} format)
to be read by \code{\link[=read_cross2]{read_cross2()}}.  (See the
\href{https://kbroman.org/qtl2/pages/sampledata.html}{sample data files} and the
\href{https://kbroman.org/qtl2/assets/vignettes/input_files.html}{vignette describing the input file format}.)

The \code{\link[utils:zip]{utils::zip()}} function is used to do the zipping.

The files should all be contained within the directory where the
\code{control_file} sits, or in a subdirectory of that directory.
If file paths use \code{..}, these get stripped by zip, and so the
resulting zip file may not work with \code{\link[=read_cross2]{read_cross2()}}.
}
\examples{
\dontrun{
zipfile <- file.path(tempdir(), "grav2.zip")
zip_datafiles("grav2.yaml", zipfile)
}
}
\seealso{
\code{\link[=read_cross2]{read_cross2()}}, sample data files at \url{https://kbroman.org/qtl2/pages/sampledata.html}
}
\keyword{IO}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_snpinfo.R
\name{create_snpinfo}
\alias{create_snpinfo}
\title{Create snp information table for a cross}
\usage{
create_snpinfo(cross)
}
\arguments{
\item{cross}{Object of class \code{"cross2"}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}
}
\value{
A data frame of SNP information with the following columns:
\itemize{
\item \code{chr} - Character string or factor with chromosome
\item \code{pos} - Position (in same units as in the \code{"map"}
attribute in \code{genoprobs}.
\item \code{snp} - Character string with SNP identifier (if
missing, the rownames are used).
\item \code{sdp} - Strain distribution pattern: an integer, between
1 and \eqn{2^n - 2} where \eqn{n} is the number of strains, whose
binary encoding indicates the founder genotypes
SNPs with missing founder genotypes are omitted.
}
}
\description{
Create a table of snp information from a cross, for use with \code{\link[=scan1snps]{scan1snps()}}.
}
\examples{
\dontrun{
# load example data and calculate genotype probabilities
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DO_Recla/recla.zip")
recla <- read_cross2(file)
snpinfo <- create_snpinfo(recla)

# calculate genotype probabilities
pr <- calc_genoprob(recla, error_prob=0.002, map_function="c-f")

# index the snp information
snpinfo <- index_snps(recla$pmap, snpinfo)

# sex covariate
sex <- setNames((recla$covar$Sex=="female")*1, rownames(recla$covar))

# perform a SNP scan
out <- scan1snps(pr, recla$pmap, recla$pheno[,"bw"], addcovar=sex, snpinfo=snpinfo)

# plot the LOD scores
plot(out$lod, snpinfo, altcol="green3")
}

}
\seealso{
\code{\link[=index_snps]{index_snps()}}, \code{\link[=scan1snps]{scan1snps()}}, \code{\link[=genoprob_to_snpprob]{genoprob_to_snpprob()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clean.R
\name{clean}
\alias{clean}
\title{Clean an object}
\usage{
clean(object, ...)
}
\arguments{
\item{object}{Object to be cleaned}

\item{...}{Other arguments}
}
\value{
Input object with messy values cleaned up
}
\description{
Clean an object by removing messy values
}
\seealso{
\code{\link[=clean.scan1]{clean.scan1()}}, \code{\link[=clean.calc_genoprob]{clean.calc_genoprob()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_onegeno.R
\name{plot_onegeno}
\alias{plot_onegeno}
\title{Plot one individual's genome-wide genotypes}
\usage{
plot_onegeno(
  geno,
  map,
  ind = 1,
  chr = NULL,
  col = NULL,
  na_col = "white",
  swap_axes = FALSE,
  border = "black",
  shift = FALSE,
  chrwidth = 0.5,
  ...
)
}
\arguments{
\item{geno}{Imputed phase-known genotypes, as a list of matrices
(as produced by \code{\link[=maxmarg]{maxmarg()}}) or a list of
three-dimensional arrays (as produced by \code{\link[=guess_phase]{guess_phase()}}).}

\item{map}{Marker map (a list of vectors of marker positions).}

\item{ind}{Individual to plot, either a numeric index or an ID.}

\item{chr}{Selected chromosomes to plot; a vector of character strings.}

\item{col}{Vector of colors for the different genotypes.}

\item{na_col}{Color for missing segments.}

\item{swap_axes}{If TRUE, swap the axes, so that the chromosomes run horizontally.}

\item{border}{Color of outer border around chromosome rectangles.}

\item{shift}{If TRUE, shift the chromosomes so they all start at 0.}

\item{chrwidth}{Total width of rectangles for each chromosome, as a
fraction of the distance between them.}

\item{...}{Additional graphics parameters}
}
\value{
None.
}
\description{
Plot one individual's genome-wide genotypes
}
\section{Hidden graphics parameters}{

A number of graphics parameters can be passed via \code{...}. For
example, \code{bgcolor} to control the background color.
These are not included as formal parameters in order to avoid
cluttering the function definition.
}

\examples{
# load data and calculate genotype probabilities
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
iron <- iron["146", ] # subset to individual 146
map <- insert_pseudomarkers(iron$gmap, step=1)
pr <- calc_genoprob(iron, map, error_prob=0.002)

# infer genotypes, as those with maximal marginal probability
m <- maxmarg(pr)

# guess phase
ph <- guess_phase(iron, m)

# plot phased genotypes
plot_onegeno(ph, map, shift=TRUE, col=c("slateblue", "Orchid"))

# this is more interesting for Diversity Outbred mouse data
\dontrun{
file <- paste0("https://raw.githubusercontent.com/rqtl/",
               "qtl2data/main/DOex/DOex.zip")
DOex <- read_cross2(file)
# subset to individuals labeled "232" and "256"
DOex <- DOex[c("232", "256"), ]
pr <- calc_genoprob(DOex, error_prob=0.002)

# infer genotypes, as those with maximal marginal probability
m <- maxmarg(pr, minprob=0.5)
# guess phase
ph <- guess_phase(DOex, m)

# plot phased genotypes
plot_onegeno(ph, DOex$gmap, shift=TRUE)
plot_onegeno(ph, DOex$gmap, ind="256", shift=TRUE)
}

}
\seealso{
\code{\link[=plot_geno]{plot_geno()}}, \code{\link[=plot_genoprob]{plot_genoprob()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/n_missing.R
\name{n_missing}
\alias{n_missing}
\alias{n_typed}
\title{Count missing genotypes}
\usage{
n_missing(
  cross,
  by = c("individual", "marker"),
  summary = c("count", "proportion")
)

n_typed(
  cross,
  by = c("individual", "marker"),
  summary = c("count", "proportion")
)
}
\arguments{
\item{cross}{An object of class \code{"cross2"}, as output by \code{\link[=read_cross2]{read_cross2()}}. For details, see the
\href{https://kbroman.org/qtl2/assets/vignettes/developer_guide.html}{R/qtl2 developer guide}.}

\item{by}{Whether to summarize by individual or marker}

\item{summary}{Whether to take count or proportion}
}
\value{
Vector of counts (or proportions) of missing (or non-missing) genotypes.
}
\description{
Number (or proportion) of missing (or non-missing) genotypes by individual or marker
}
\section{Functions}{
\itemize{
\item \code{n_missing()}: Count missing genotypes

\item \code{n_typed()}: Count genotypes

}}
\examples{
iron <- read_cross2(system.file("extdata", "iron.zip", package="qtl2"))
nmis_ind <- n_missing(iron)
pmis_mar <- n_typed(iron, "mar", "proportion")
plot(nmis_ind, xlab="Individual", ylab="No. missing genotypes")
plot(pmis_mar, xlab="Markers", ylab="Prop. genotyped")
}
